def write_kinetic():
    import os
    if not os.path.exists('_kinetic_mech'):  os.mkdir('_kinetic_mech')
    os.chdir('_kinetic_mech')

    fichier_data = open('C0_Konnov.cti', 'w')
    fichier_data.write("""#-------------------------------------------------------------------------------
# Konnov et al. 2019  https://doi.org/10.1016/j.combustflame.2019.01.032
#-------------------------------------------------------------------------------
        
units(length='cm', time='s', quantity='mol', act_energy='cal/mol')

ideal_gas(name='gas',
          elements="O N Ar H He",
          species=\"""H     H2    O     O2    H2O   
                     OH    H2O2  HO2   OX    OHX   
                     O2X   O3    AR    HE    N2\""",
          reactions='all',
          transport='Mix',
          initial_state=state(temperature=300.0, pressure=OneAtm))

#-------------------------------------------------------------------------------
# Species data
#-------------------------------------------------------------------------------

species(name='H',
        atoms='H:1',
        thermo=(NASA([200.00, 1000.00],
                     [ 2.50000000E+00,  0.00000000E+00,  0.00000000E+00,
                       0.00000000E+00,  0.00000000E+00,  2.54736600E+04,
                      -4.46682850E-01]),
                NASA([1000.00, 6000.00],
                     [ 2.50000000E+00,  0.00000000E+00,  0.00000000E+00,
                       0.00000000E+00,  0.00000000E+00,  2.54736600E+04,
                      -4.46682850E-01])),
        transport=gas_transport(geom='atom',
                                diam=1.53,
                                well_depth=541.672,
                                polar=0.666),
        note='L6/94')

species(name='H2',
        atoms='H:2',
        thermo=(NASA([200.00, 1000.00],
                     [ 2.34433112E+00,  7.98052075E-03, -1.94781510E-05,
                       2.01572094E-08, -7.37611761E-12, -9.17935173E+02,
                       6.83010238E-01]),
                NASA([1000.00, 6000.00],
                     [ 2.93286575E+00,  8.26608026E-04, -1.46402364E-07,
                       1.54100414E-11, -6.88804800E-16, -8.13065581E+02,
                      -1.02432865E+00])),
        transport=gas_transport(geom='linear',
                                diam=2.19,
                                well_depth=304.69,
                                polar=0.775,
                                rot_relax=280.0),
        note='REFELEMENTRUS78')

species(name='O',
        atoms='O:1',
        thermo=(NASA([200.00, 1000.00],
                     [ 3.16826710E+00, -3.27931884E-03,  6.64306396E-06,
                      -6.12806624E-09,  2.11265971E-12,  2.91222592E+04,
                       2.05193346E+00]),
                NASA([1000.00, 6000.00],
                     [ 2.54363697E+00, -2.73162486E-05, -4.19029520E-09,
                       4.95481845E-12, -4.79553694E-16,  2.92260120E+04,
                       4.92229457E+00])),
        transport=gas_transport(geom='atom',
                                diam=2.485,
                                well_depth=235.686,
                                polar=0.744),
        note='ATcT')

species(name='O2',
        atoms='O:2',
        thermo=(NASA([200.00, 1000.00],
                     [ 3.78245636E+00, -2.99673416E-03,  9.84730201E-06,
                      -9.68129509E-09,  3.24372837E-12, -1.06394356E+03,
                       3.65767573E+00]),
                NASA([1000.00, 6000.00],
                     [ 3.66096065E+00,  6.56365811E-04, -1.41149627E-07,
                       2.05797935E-11, -1.29913436E-15, -1.21597718E+03,
                       3.41536279E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.069,
                                well_depth=676.424,
                                polar=1.487,
                                rot_relax=3.8),
        note='ATcT')

species(name='H2O',
        atoms='H:2 O:1',
        thermo=(NASA([200.00, 1000.00],
                     [ 4.19863520E+00, -2.03640170E-03,  6.52034160E-06,
                      -5.48792690E-09,  1.77196800E-12, -3.02937260E+04,
                      -8.49009010E-01]),
                NASA([1000.00, 6000.00],
                     [ 2.67703890E+00,  2.97318160E-03, -7.73768890E-07,
                       9.44335140E-11, -4.26899910E-15, -2.98858940E+04,
                       6.88255000E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=2.943,
                                well_depth=637.056,
                                dipole=1.851,
                                polar=1.407,
                                rot_relax=4.0),
        note='L5/89')

species(name='OH',
        atoms='O:1 H:1',
        thermo=(NASA([200.00, 1000.00],
                     [ 3.99198424E+00, -2.40106655E-03,  4.61664033E-06,
                      -3.87916306E-09,  1.36319502E-12,  3.36889836E+03,
                      -1.03998477E-01]),
                NASA([1000.00, 6000.00],
                     [ 2.83853033E+00,  1.10741289E-03, -2.94000209E-07,
                       4.20698729E-11, -2.42289890E-15,  3.69780808E+03,
                       5.84494652E+00])),
        transport=gas_transport(geom='linear',
                                diam=2.582,
                                well_depth=514.598,
                                dipole=1.635,
                                polar=1.077),
        note='ATcT')

species(name='H2O2',
        atoms='H:2 O:2',
        thermo=(NASA([200.00, 1000.00],
                     [ 4.31515149E+00, -8.47390622E-04,  1.76404323E-05,
                      -2.26762944E-08,  9.08950158E-12, -1.77067437E+04,
                       3.27373319E+00]),
                NASA([1000.00, 6000.00],
                     [ 4.57977305E+00,  4.05326003E-03, -1.29844730E-06,
                       1.98211400E-10, -1.13968792E-14, -1.80071775E+04,
                       6.64970694E-01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.179,
                                well_depth=1361.148,
                                dipole=1.702,
                                polar=2.239,
                                rot_relax=3.8),
        note='L2/93')

species(name='HO2',
        atoms='H:1 O:2',
        thermo=(NASA([200.00, 1000.00],
                     [ 4.30179807E+00, -4.74912097E-03,  2.11582905E-05,
                      -2.42763914E-08,  9.29225225E-12,  2.64018485E+02,
                       3.71666220E+00]),
                NASA([1000.00, 5000.00],
                     [ 4.17228741E+00,  1.88117627E-03, -3.46277286E-07,
                       1.94657549E-11,  1.76256905E-16,  3.10206839E+01,
                       2.95767672E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.129,
                                well_depth=963.003,
                                dipole=2.146,
                                polar=1.959,
                                rot_relax=3.8),
        note='L5/89')

species(name='OX',
        atoms='O:1',
        thermo=(NASA([200.00, 1000.00],
                     [ 2.49993786E+00,  1.71935346E-07, -3.45215267E-10,
                       3.71342028E-13, -1.70964494E-16,  5.19965317E+04,
                       4.61684555E+00]),
                NASA([1000.00, 6000.00],
                     [ 2.49368475E+00,  1.37617903E-05, -1.00401058E-08,
                       2.76012182E-12, -2.01597513E-16,  5.19986304E+04,
                       4.65050950E+00])),
        transport=gas_transport(geom='atom',
                                diam=2.485,
                                well_depth=235.686,
                                polar=0.744),
        note='singletATcT')

species(name='OHX',
        atoms='O:1 H:1',
        thermo=(NASA([200.00, 1000.00],
                     [ 3.46084428E+00,  5.01872172E-04, -2.00254474E-06,
                       3.18901984E-09, -1.35451838E-12,  5.07349466E+04,
                       1.73976415E+00]),
                NASA([1000.00, 6000.00],
                     [ 2.75582920E+00,  1.39848756E-03, -4.19428493E-07,
                       6.33453282E-11, -3.56042218E-15,  5.09751756E+04,
                       5.62581429E+00])),
        transport=gas_transport(geom='linear',
                                diam=2.582,
                                well_depth=514.598,
                                dipole=1.635,
                                polar=1.077),
        note='A2Sigma+ATcT')

species(name='O2X',
        atoms='O:2',
        thermo=(NASA([200.00, 1000.00],
                     [ 3.78535371E+00, -3.21928540E-03,  1.12323443E-05,
                      -1.17254068E-08,  4.17659585E-12,  1.02922572E+04,
                       3.27320239E+00]),
                NASA([1000.00, 6000.00],
                     [ 3.45852381E+00,  1.04045351E-03, -2.79664041E-07,
                       3.11439672E-11, -8.55656058E-16,  1.02229063E+04,
                       4.15264119E+00])),
        transport=gas_transport(geom='linear',
                                diam=2.769,
                                well_depth=1836.0,
                                polar=1.434,
                                rot_relax=3.8),
        note='singletATcT')

species(name='O3',
        atoms='O:3',
        thermo=(NASA([200.00, 1000.00],
                     [ 3.40738221E+00,  2.05379063E-03,  1.38486052E-05,
                      -2.23311542E-08,  9.76073226E-12,  1.58644979E+04,
                       8.28247580E+00]),
                NASA([1000.00, 6000.00],
                     [ 1.23302914E+01, -1.19324783E-02,  7.98741278E-06,
                      -1.77194552E-09,  1.26075824E-13,  1.26755831E+04,
                      -4.08823374E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.302,
                                well_depth=1758.196,
                                dipole=0.55,
                                polar=2.82,
                                rot_relax=1.0),
        note='L5/90')

species(name='AR',
        atoms='Ar:1',
        thermo=(NASA([200.00, 1000.00],
                     [ 2.50000000E+00,  0.00000000E+00,  0.00000000E+00,
                       0.00000000E+00,  0.00000000E+00, -7.45375000E+02,
                       4.37967491E+00]),
                NASA([1000.00, 6000.00],
                     [ 2.50000000E+00,  0.00000000E+00,  0.00000000E+00,
                       0.00000000E+00,  0.00000000E+00, -7.45375000E+02,
                       4.37967491E+00])),
        transport=gas_transport(geom='atom',
                                diam=3.462,
                                well_depth=127.697,
                                polar=1.609,
                                rot_relax=2.0),
        note='REFELEMENTL6/88')

species(name='HE',
        atoms='He:1',
        thermo=(NASA([200.00, 1000.00],
                     [ 2.50000000E+00,  0.00000000E+00,  0.00000000E+00,
                       0.00000000E+00,  0.00000000E+00, -7.45375000E+02,
                       9.28723974E-01]),
                NASA([1000.00, 6000.00],
                     [ 2.50000000E+00,  0.00000000E+00,  0.00000000E+00,
                       0.00000000E+00,  0.00000000E+00, -7.45375000E+02,
                       9.28723974E-01])),
        transport=gas_transport(geom='atom',
                                diam=2.715,
                                well_depth=11.442,
                                polar=0.204,
                                rot_relax=2.0),
        note='REFELEMENTg5/97')

species(name='N2',
        atoms='N:2',
        thermo=(NASA([200.00, 1000.00],
                     [ 3.53100528E+00, -1.23660988E-04, -5.02999433E-07,
                       2.43530612E-09, -1.40881235E-12, -1.04697628E+03,
                       2.96747038E+00]),
                NASA([1000.00, 6000.00],
                     [ 2.95257637E+00,  1.39690040E-03, -4.92631603E-07,
                       7.86010195E-11, -4.60755204E-15, -9.23948688E+02,
                       5.87188762E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.61,
                                well_depth=97.839,
                                polar=1.756,
                                rot_relax=4.0),
        note='REFELEMENT8/02')

#-------------------------------------------------------------------------------
# Reaction data
#-------------------------------------------------------------------------------

# Reaction 1
three_body_reaction('H + H + M <=> H2 + M', [7.000000e+17, -1.0, 0.0],
                    efficiencies='H2:0.0 N2:0.0 H:0.0 H2O:14.3')

# Reaction 2
reaction('H + H + H2 <=> H2 + H2', [1.000000e+17, -0.6, 0.0])

# Reaction 3
reaction('H + H + N2 <=> H2 + N2', [5.400000e+18, -1.3, 0.0])

# Reaction 4
reaction('H + H + H <=> H2 + H', [3.200000e+15, 0.0, 0.0])

# Reaction 5
three_body_reaction('O + O + M <=> O2 + M', [1.000000e+17, -1.0, 0.0],
                    efficiencies='O:28.8 O2:8.0 N2:2.0 H2O:5.0 O3:8.0')

# Reaction 6
three_body_reaction('O + H + M <=> OH + M', [6.750000e+18, -1.0, 0.0],
                    efficiencies='H2O:5.0')

# Reaction 7
three_body_reaction('H2O + M <=> H + OH + M', [6.060000e+27, -3.312, 120770.0],
                    efficiencies='H2O:0.0 H2:3.0 N2:2.0 O2:1.5 HE:1.1')

# Reaction 8
reaction('H2O + H2O <=> H + OH + H2O', [1.000000e+26, -2.44, 120160.0])

# Reaction 9
falloff_reaction('H + O2 (+ M) <=> HO2 (+ M)',
                 kf=[4.660000e+12, 0.44, 0.0],
                 kf0=[1.225000e+19, -1.2, 0.0],
                 efficiencies='AR:0.72 H2O:16.6 O2:1.0 H2:1.5 HE:0.57',
                 falloff=Troe(A=0.5, T3=1.0, T1=10000000000.0))

# Reaction 10
falloff_reaction('H2O2 (+ M) <=> OH + OH (+ M)',
                 kf=[2.000000e+12, 0.9, 48750.0],
                 kf0=[2.490000e+24, -2.3, 48750.0],
                 efficiencies='H2O:7.5 H2O2:7.7 O2:1.2 N2:1.5 HE:0.65 H2:3.7',
                 falloff=Troe(A=0.42, T3=1.0, T1=10000000000.0))

# Reaction 11
reaction('O + H2 <=> OH + H', [5.080000e+04, 2.67, 6292.0])

# Reaction 12
reaction('H + O2 <=> OH + O', [1.040000e+14, 0.0, 15286.0])

# Reaction 13
reaction('H2 + OH <=> H2O + H', [2.140000e+08, 1.52, 3450.0])

# Reaction 14
reaction('OH + OH <=> H2O + O', [2.668000e+06, 1.82, -1647.0])

# Reaction 15
reaction('HO2 + O <=> OH + O2', [2.850000e+10, 1.0, -723.9])

# Reaction 16
reaction('H + HO2 <=> OH + OH', [7.080000e+13, 0.0, 300.0])

# Reaction 17
reaction('H2O + O <=> H + HO2', [2.200000e+08, 2.0, 61600.0])

# Reaction 18
reaction('H2 + O2 <=> H + HO2', [7.400000e+05, 2.43, 53500.0])

# Reaction 19
reaction('HO2 + OH <=> H2O + O2', [7.000000e+12, 0.0, -1093.0],
         options='duplicate')

# Reaction 20
reaction('HO2 + OH <=> H2O + O2', [4.500000e+14, 0.0, 10930.0],
         options='duplicate')

# Reaction 21
reaction('HO2 + HO2 <=> H2O2 + O2', [1.030000e+14, 0.0, 11040.0],
         options='duplicate')

# Reaction 22
reaction('HO2 + HO2 <=> H2O2 + O2', [1.940000e+11, 0.0, -1409.0],
         options='duplicate')

# Reaction 23
reaction('H2O2 + H <=> HO2 + H2', [5.020000e+06, 2.07, 4300.0])

# Reaction 24
reaction('H2O2 + H <=> H2O + OH', [2.030000e+07, 2.02, 2620.0])

# Reaction 25
reaction('H2O2 + O <=> HO2 + OH', [9.550000e+06, 2.0, 3970.0])

# Reaction 26
reaction('H2O2 + OH <=> HO2 + H2O', [1.740000e+12, 0.0, 318.0],
         options='duplicate')

# Reaction 27
reaction('H2O2 + OH <=> HO2 + H2O', [7.590000e+13, 0.0, 7269.0],
         options='duplicate')

# Reaction 28
reaction('O2 + O + AR <=> O3 + AR', [4.290000e+17, -1.5, 0.0],
         options='duplicate')

# Reaction 29
reaction('O2 + O + AR <=> O3 + AR', [5.100000e+21, -3.2, 0.0],
         options='duplicate')

# Reaction 30
three_body_reaction('O2 + O + M <=> O3 + M', [6.530000e+17, -1.5, 0.0],
                    efficiencies='AR:0.0 O2:0.95 O3:2.5 O:4.0',
                    options='duplicate')

# Reaction 31
three_body_reaction('O2 + O + M <=> O3 + M', [1.330000e+22, -3.3, 0.0],
                    efficiencies='AR:0.0 O2:1.07 O3:2.5 O:4.0',
                    options='duplicate')

# Reaction 32
reaction('O3 + O <=> O2 + O2', [4.820000e+12, 0.0, 4094.0])

# Reaction 33
reaction('O3 + O <=> O2X + O2', [1.440000e+11, 0.0, 4094.0])

# Reaction 34
three_body_reaction('O + O + M <=> O2X + M', [7.000000e+15, -1.0, 0.0],
                    efficiencies='O:28.8 O2:8.0 N2:2.0 H2O:5.0 O3:8.0')

# Reaction 35
three_body_reaction('O2X + M <=> O2 + M', [1.800000e+06, 0.0, 400.0],
                    efficiencies='O:0.0 H:0.0 AR:0.005 HE:0.005 N2:0.002 H2O:3.3 H2:2.5')

# Reaction 36
reaction('O2X + O <=> O2 + O', [7.800000e+07, 0.0, 0.0])

# Reaction 37
reaction('O2X + H <=> O2 + H', [4.000000e+13, 0.0, 5030.0])

# Reaction 38
three_body_reaction('O2X + O + M <=> O + O2 + M', [3.600000e+15, 0.0, 0.0],
                    efficiencies='AR:0.63')

# Reaction 39
reaction('O2X + O3 <=> O2 + O2 + O', [3.130000e+13, 0.0, 5644.0])

# Reaction 40
reaction('OX + O2X <=> O + O2', [6.030000e+12, 0.0, 0.0])

# Reaction 41
reaction('OX + O2 <=> O + O2X', [1.590000e+13, 0.0, -139.0])

# Reaction 42
reaction('OX + O2 <=> O + O2', [2.810000e+12, 0.0, -139.0])

# Reaction 43
three_body_reaction('OX + M <=> O + M', [4.800000e+11, 0.0, 0.0],
                    efficiencies='O2:0.0 N2:0.0 O:10.0 H2O:3.0')

# Reaction 44
reaction('OX + N2 <=> O + N2', [1.260000e+13, 0.0, -230.0])

# Reaction 45
reaction('OX + O3 <=> O2 + O + O', [7.230000e+13, 0.0, 0.0])

# Reaction 46
reaction('OX + O3 <=> O2 + O2', [7.230000e+13, 0.0, 0.0])

# Reaction 47
reaction('H2 + O2X <=> H + HO2', [6.160000e+05, 2.335, 31080.0])

# Reaction 48
reaction('H + O2X <=> OH + O', [3.500000e+08, 1.45, 4508.0])

# Reaction 49
three_body_reaction('H + O2X + M <=> HO2 + M', [9.890000e+09, 2.03, 3360.0])

# Reaction 50
reaction('HO2 + OH <=> H2O + O2X', [2.140000e+06, 1.65, 2180.0])

# Reaction 51
reaction('OH + O2X <=> O + HO2', [1.300000e+13, 0.0, 34000.0])

# Reaction 52
reaction('O3 + H <=> OH + O2', [8.430000e+13, 0.0, 934.0])

# Reaction 53
reaction('O3 + OH <=> HO2 + O2', [1.000000e+12, 0.0, 1870.0])

# Reaction 54
reaction('O3 + HO2 <=> OH + O2 + O2', [5.850000e-04, 4.57, -1377.0])

# Reaction 55
reaction('H + HO2 <=> H2O + OX', [2.500000e+12, 0.0, 300.0])

# Reaction 56
reaction('OX + H2 <=> OH + H', [8.100000e+13, 0.0, 0.0])

# Reaction 57
reaction('OX + H2O <=> OH + OH', [1.000000e+14, 0.0, -71.0])

# Reaction 58
three_body_reaction('O + H + M <=> OHX + M', [1.500000e+13, 0.0, 5970.0],
                    efficiencies='AR:0.35 H2O:6.5 O2:0.4 N2:0.4')

# Reaction 59
reaction('OHX + O2 <=> OH + O2', [8.400000e+11, 0.5, -482.0])

# Reaction 60
reaction('OHX + N2 <=> OH + N2', [1.080000e+11, 0.5, -1238.0])

# Reaction 61
reaction('OHX + H2O <=> OH + H2O', [2.960000e+12, 0.5, -861.0])

# Reaction 62
reaction('OHX + H2 <=> OH + H2', [3.540000e+11, 0.5, -444.0])

# Reaction 63
reaction('OHX + OH <=> OH + OH', [1.500000e+12, 0.5, 0.0])

# Reaction 64
reaction('OHX + H <=> OH + H', [1.500000e+12, 0.5, 0.0])

# Reaction 65
reaction('OHX + O <=> OH + O', [1.500000e+12, 0.5, 0.0])

# Reaction 66
reaction('OHX + AR <=> OH + AR', [2.170000e+10, 0.5, 2060.0])

# Reaction 67
reaction('OHX + H2 <=> H2O + H', [2.600000e+12, 0.5, -444.0])

# Reaction 68
reaction('OHX + O2 <=> O3 + H', [2.520000e+11, 0.5, -482.0])

# Reaction 69
reaction('OHX + O2 <=> HO2 + O', [1.008000e+12, 0.5, -482.0])

# Reaction 70
reaction('OHX + H2O <=> H2O2 + H', [2.960000e+12, 0.5, -861.0])

# Reaction 71
reaction('OHX <=> OH', [1.400000e+06, 0.0, 0.0])

# Reaction 72
reaction('H + O2 + H <=> H2 + O2', [8.800000e+22, -1.835, 800.0])

# Reaction 73
reaction('H + O2 + H <=> OH + OH', [4.000000e+22, -1.835, 800.0])

# Reaction 74
reaction('H + O2 + O <=> OH + O2', [7.350000e+22, -1.835, 800.0])

# Reaction 75
reaction('H + O2 + OH <=> H2O + O2', [2.560000e+22, -1.835, 800.0])

    """)

    # ================================================================
    # ================================================================
    # ================================================================
    # ================================================================

    fichier_data = open('C1_GRI30.cti', 'w')
    fichier_data.write("""#---------------------------------------------------------------------------------
# Smith et al. 1999  http://combustion.berkeley.edu/gri-mech/version30/text30.html
#---------------------------------------------------------------------------------

units(length='cm', time='s', quantity='mol', act_energy='cal/mol')

ideal_gas(name='gas',
        elements="O H C N Ar",
        species=\"""H2      H       O       O2      OH
                    H2O     HO2     H2O2    C       CH
                    CH2     CH2(S)  CH3     CH4     CO
                    CO2     HCO     CH2O    CH2OH   CH3O
                    CH3OH   C2H     C2H2    C2H3    C2H4
                    C2H5    C2H6    HCCO    CH2CO   HCCOH
                    N       NH      NH2     NH3     NNH
                    NO      NO2     N2O     HNO     CN
                    HCN     H2CN    HCNN    HCNO    HOCN
                    HNCO    NCO     N2      AR      C3H7
                    C3H8    CH2CHO  CH3CHO\""",
        reactions='all',
        transport='Mix',
        initial_state=state(temperature=300.0, pressure=OneAtm))

#-------------------------------------------------------------------------------
# Species data
#-------------------------------------------------------------------------------

species(name='H2',
        atoms='H:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 2.34433112E+00,  7.98052075E-03, -1.94781510E-05,
                    2.01572094E-08, -7.37611761E-12, -9.17935173E+02,
                    6.83010238E-01]),
                NASA([1000.00, 3500.00],
                    [ 3.33727920E+00, -4.94024731E-05,  4.99456778E-07,
                    -1.79566394E-10,  2.00255376E-14, -9.50158922E+02,
                    -3.20502331E+00])),
        transport=gas_transport(geom='linear',
                                diam=2.92,
                                well_depth=38.0,
                                polar=0.79,
                                rot_relax=280.0),
        note='TPIS78')

species(name='H',
        atoms='H:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 2.50000000E+00,  7.05332819E-13, -1.99591964E-15,
                    2.30081632E-18, -9.27732332E-22,  2.54736599E+04,
                    -4.46682853E-01]),
                NASA([1000.00, 3500.00],
                    [ 2.50000001E+00, -2.30842973E-11,  1.61561948E-14,
                    -4.73515235E-18,  4.98197357E-22,  2.54736599E+04,
                    -4.46682914E-01])),
        transport=gas_transport(geom='atom',
                                diam=2.05,
                                well_depth=145.0),
        note='L7/88')

species(name='O',
        atoms='O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.16826710E+00, -3.27931884E-03,  6.64306396E-06,
                    -6.12806624E-09,  2.11265971E-12,  2.91222592E+04,
                    2.05193346E+00]),
                NASA([1000.00, 3500.00],
                    [ 2.56942078E+00, -8.59741137E-05,  4.19484589E-08,
                    -1.00177799E-11,  1.22833691E-15,  2.92175791E+04,
                    4.78433864E+00])),
        transport=gas_transport(geom='atom',
                                diam=2.75,
                                well_depth=80.0),
        note='L1/90')

species(name='O2',
        atoms='O:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.78245636E+00, -2.99673416E-03,  9.84730201E-06,
                    -9.68129509E-09,  3.24372837E-12, -1.06394356E+03,
                    3.65767573E+00]),
                NASA([1000.00, 3500.00],
                    [ 3.28253784E+00,  1.48308754E-03, -7.57966669E-07,
                    2.09470555E-10, -2.16717794E-14, -1.08845772E+03,
                    5.45323129E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.458,
                                well_depth=107.4,
                                polar=1.6,
                                rot_relax=3.8),
        note='TPIS89')

species(name='OH',
        atoms='H:1 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.99201543E+00, -2.40131752E-03,  4.61793841E-06,
                    -3.88113333E-09,  1.36411470E-12,  3.61508056E+03,
                    -1.03925458E-01]),
                NASA([1000.00, 3500.00],
                    [ 3.09288767E+00,  5.48429716E-04,  1.26505228E-07,
                    -8.79461556E-11,  1.17412376E-14,  3.85865700E+03,
                    4.47669610E+00])),
        transport=gas_transport(geom='linear',
                                diam=2.75,
                                well_depth=80.0),
        note='RUS78')

species(name='H2O',
        atoms='H:2 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.19864056E+00, -2.03643410E-03,  6.52040211E-06,
                    -5.48797062E-09,  1.77197817E-12, -3.02937267E+04,
                    -8.49032208E-01]),
                NASA([1000.00, 3500.00],
                    [ 3.03399249E+00,  2.17691804E-03, -1.64072518E-07,
                    -9.70419870E-11,  1.68200992E-14, -3.00042971E+04,
                    4.96677010E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=2.605,
                                well_depth=572.4,
                                dipole=1.844,
                                rot_relax=4.0),
        note='L8/89')

species(name='HO2',
        atoms='H:1 O:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.30179801E+00, -4.74912051E-03,  2.11582891E-05,
                    -2.42763894E-08,  9.29225124E-12,  2.94808040E+02,
                    3.71666245E+00]),
                NASA([1000.00, 3500.00],
                    [ 4.01721090E+00,  2.23982013E-03, -6.33658150E-07,
                    1.14246370E-10, -1.07908535E-14,  1.11856713E+02,
                    3.78510215E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.458,
                                well_depth=107.4,
                                rot_relax=1.0),
        note='L5/89')

species(name='H2O2',
        atoms='H:2 O:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.27611269E+00, -5.42822417E-04,  1.67335701E-05,
                    -2.15770813E-08,  8.62454363E-12, -1.77025821E+04,
                    3.43505074E+00]),
                NASA([1000.00, 3500.00],
                    [ 4.16500285E+00,  4.90831694E-03, -1.90139225E-06,
                    3.71185986E-10, -2.87908305E-14, -1.78617877E+04,
                    2.91615662E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.458,
                                well_depth=107.4,
                                rot_relax=3.8),
        note='L7/88')

species(name='C',
        atoms='C:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 2.55423955E+00, -3.21537724E-04,  7.33792245E-07,
                    -7.32234889E-10,  2.66521446E-13,  8.54438832E+04,
                    4.53130848E+00]),
                NASA([1000.00, 3500.00],
                    [ 2.49266888E+00,  4.79889284E-05, -7.24335020E-08,
                    3.74291029E-11, -4.87277893E-15,  8.54512953E+04,
                    4.80150373E+00])),
        transport=gas_transport(geom='atom',
                                diam=3.298,
                                well_depth=71.4),
        note='L11/88')

species(name='CH',
        atoms='H:1 C:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.48981665E+00,  3.23835541E-04, -1.68899065E-06,
                    3.16217327E-09, -1.40609067E-12,  7.07972934E+04,
                    2.08401108E+00]),
                NASA([1000.00, 3500.00],
                    [ 2.87846473E+00,  9.70913681E-04,  1.44445655E-07,
                    -1.30687849E-10,  1.76079383E-14,  7.10124364E+04,
                    5.48497999E+00])),
        transport=gas_transport(geom='linear',
                                diam=2.75,
                                well_depth=80.0),
        note='TPIS79')

species(name='CH2',
        atoms='H:2 C:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.76267867E+00,  9.68872143E-04,  2.79489841E-06,
                    -3.85091153E-09,  1.68741719E-12,  4.60040401E+04,
                    1.56253185E+00]),
                NASA([1000.00, 3500.00],
                    [ 2.87410113E+00,  3.65639292E-03, -1.40894597E-06,
                    2.60179549E-10, -1.87727567E-14,  4.62636040E+04,
                    6.17119324E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.8,
                                well_depth=144.0),
        note='LS/93')

species(name='CH2(S)',
        atoms='H:2 C:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.19860411E+00, -2.36661419E-03,  8.23296220E-06,
                    -6.68815981E-09,  1.94314737E-12,  5.04968163E+04,
                    -7.69118967E-01]),
                NASA([1000.00, 3500.00],
                    [ 2.29203842E+00,  4.65588637E-03, -2.01191947E-06,
                    4.17906000E-10, -3.39716365E-14,  5.09259997E+04,
                    8.62650169E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.8,
                                well_depth=144.0),
        note='LS/93')

species(name='CH3',
        atoms='H:3 C:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.67359040E+00,  2.01095175E-03,  5.73021856E-06,
                    -6.87117425E-09,  2.54385734E-12,  1.64449988E+04,
                    1.60456433E+00]),
                NASA([1000.00, 3500.00],
                    [ 2.28571772E+00,  7.23990037E-03, -2.98714348E-06,
                    5.95684644E-10, -4.67154394E-14,  1.67755843E+04,
                    8.48007179E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.8,
                                well_depth=144.0),
        note='L11/89')

species(name='CH4',
        atoms='H:4 C:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 5.14987613E+00, -1.36709788E-02,  4.91800599E-05,
                    -4.84743026E-08,  1.66693956E-11, -1.02466476E+04,
                    -4.64130376E+00]),
                NASA([1000.00, 3500.00],
                    [ 7.48514950E-02,  1.33909467E-02, -5.73285809E-06,
                    1.22292535E-09, -1.01815230E-13, -9.46834459E+03,
                    1.84373180E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.746,
                                well_depth=141.4,
                                polar=2.6,
                                rot_relax=13.0),
        note='L8/88')

species(name='CO',
        atoms='C:1 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.57953347E+00, -6.10353680E-04,  1.01681433E-06,
                    9.07005884E-10, -9.04424499E-13, -1.43440860E+04,
                    3.50840928E+00]),
                NASA([1000.00, 3500.00],
                    [ 2.71518561E+00,  2.06252743E-03, -9.98825771E-07,
                    2.30053008E-10, -2.03647716E-14, -1.41518724E+04,
                    7.81868772E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.65,
                                well_depth=98.1,
                                polar=1.95,
                                rot_relax=1.8),
        note='TPIS79')

species(name='CO2',
        atoms='C:1 O:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 2.35677352E+00,  8.98459677E-03, -7.12356269E-06,
                    2.45919022E-09, -1.43699548E-13, -4.83719697E+04,
                    9.90105222E+00]),
                NASA([1000.00, 3500.00],
                    [ 3.85746029E+00,  4.41437026E-03, -2.21481404E-06,
                    5.23490188E-10, -4.72084164E-14, -4.87591660E+04,
                    2.27163806E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.763,
                                well_depth=244.0,
                                polar=2.65,
                                rot_relax=2.1),
        note='L7/88')

species(name='HCO',
        atoms='H:1 C:1 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.22118584E+00, -3.24392532E-03,  1.37799446E-05,
                    -1.33144093E-08,  4.33768865E-12,  3.83956496E+03,
                    3.39437243E+00]),
                NASA([1000.00, 3500.00],
                    [ 2.77217438E+00,  4.95695526E-03, -2.48445613E-06,
                    5.89161778E-10, -5.33508711E-14,  4.01191815E+03,
                    9.79834492E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.59,
                                well_depth=498.0),
        note='L12/89')

species(name='CH2O',
        atoms='H:2 C:1 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.79372315E+00, -9.90833369E-03,  3.73220008E-05,
                    -3.79285261E-08,  1.31772652E-11, -1.43089567E+04,
                    6.02812900E-01]),
                NASA([1000.00, 3500.00],
                    [ 1.76069008E+00,  9.20000082E-03, -4.42258813E-06,
                    1.00641212E-09, -8.83855640E-14, -1.39958323E+04,
                    1.36563230E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.59,
                                well_depth=498.0,
                                rot_relax=2.0),
        note='L8/88')

species(name='CH2OH',
        atoms='H:3 C:1 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.86388918E+00,  5.59672304E-03,  5.93271791E-06,
                    -1.04532012E-08,  4.36967278E-12, -3.19391367E+03,
                    5.47302243E+00]),
                NASA([1000.00, 3500.00],
                    [ 3.69266569E+00,  8.64576797E-03, -3.75101120E-06,
                    7.87234636E-10, -6.48554201E-14, -3.24250627E+03,
                    5.81043215E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.69,
                                well_depth=417.0,
                                dipole=1.7,
                                rot_relax=2.0),
        note='GUNL93')

species(name='CH3O',
        atoms='H:3 C:1 O:1',
        thermo=(NASA([300.00, 1000.00],
                    [ 2.10620400E+00,  7.21659500E-03,  5.33847200E-06,
                    -7.37763600E-09,  2.07561000E-12,  9.78601100E+02,
                    1.31521770E+01]),
                NASA([1000.00, 3000.00],
                    [ 3.77079900E+00,  7.87149700E-03, -2.65638400E-06,
                    3.94443100E-10, -2.11261600E-14,  1.27832520E+02,
                    2.92957500E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.69,
                                well_depth=417.0,
                                dipole=1.7,
                                rot_relax=2.0),
        note='121686')

species(name='CH3OH',
        atoms='H:4 C:1 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 5.71539582E+00, -1.52309129E-02,  6.52441155E-05,
                    -7.10806889E-08,  2.61352698E-11, -2.56427656E+04,
                    -1.50409823E+00]),
                NASA([1000.00, 3500.00],
                    [ 1.78970791E+00,  1.40938292E-02, -6.36500835E-06,
                    1.38171085E-09, -1.17060220E-13, -2.53748747E+04,
                    1.45023623E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.626,
                                well_depth=481.8,
                                rot_relax=1.0),
        note='L8/88')

species(name='C2H',
        atoms='H:1 C:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 2.88965733E+00,  1.34099611E-02, -2.84769501E-05,
                    2.94791045E-08, -1.09331511E-11,  6.68393932E+04,
                    6.22296438E+00]),
                NASA([1000.00, 3500.00],
                    [ 3.16780652E+00,  4.75221902E-03, -1.83787077E-06,
                    3.04190252E-10, -1.77232770E-14,  6.71210650E+04,
                    6.63589475E+00])),
        transport=gas_transport(geom='linear',
                                diam=4.1,
                                well_depth=209.0,
                                rot_relax=2.5),
        note='L1/91')

species(name='C2H2',
        atoms='H:2 C:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 8.08681094E-01,  2.33615629E-02, -3.55171815E-05,
                    2.80152437E-08, -8.50072974E-12,  2.64289807E+04,
                    1.39397051E+01]),
                NASA([1000.00, 3500.00],
                    [ 4.14756964E+00,  5.96166664E-03, -2.37294852E-06,
                    4.67412171E-10, -3.61235213E-14,  2.59359992E+04,
                    -1.23028121E+00])),
        transport=gas_transport(geom='linear',
                                diam=4.1,
                                well_depth=209.0,
                                rot_relax=2.5),
        note='L1/91')

species(name='C2H3',
        atoms='H:3 C:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.21246645E+00,  1.51479162E-03,  2.59209412E-05,
                    -3.57657847E-08,  1.47150873E-11,  3.48598468E+04,
                    8.51054025E+00]),
                NASA([1000.00, 3500.00],
                    [ 3.01672400E+00,  1.03302292E-02, -4.68082349E-06,
                    1.01763288E-09, -8.62607041E-14,  3.46128739E+04,
                    7.78732378E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.1,
                                well_depth=209.0,
                                rot_relax=1.0),
        note='L2/92')

species(name='C2H4',
        atoms='H:4 C:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.95920148E+00, -7.57052247E-03,  5.70990292E-05,
                    -6.91588753E-08,  2.69884373E-11,  5.08977593E+03,
                    4.09733096E+00]),
                NASA([1000.00, 3500.00],
                    [ 2.03611116E+00,  1.46454151E-02, -6.71077915E-06,
                    1.47222923E-09, -1.25706061E-13,  4.93988614E+03,
                    1.03053693E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.971,
                                well_depth=280.8,
                                rot_relax=1.5),
        note='L1/91')

species(name='C2H5',
        atoms='H:5 C:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.30646568E+00, -4.18658892E-03,  4.97142807E-05,
                    -5.99126606E-08,  2.30509004E-11,  1.28416265E+04,
                    4.70720924E+00]),
                NASA([1000.00, 3500.00],
                    [ 1.95465642E+00,  1.73972722E-02, -7.98206668E-06,
                    1.75217689E-09, -1.49641576E-13,  1.28575200E+04,
                    1.34624343E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.302,
                                well_depth=252.3,
                                rot_relax=1.5),
        note='L12/92')

species(name='C2H6',
        atoms='H:6 C:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.29142492E+00, -5.50154270E-03,  5.99438288E-05,
                    -7.08466285E-08,  2.68685771E-11, -1.15222055E+04,
                    2.66682316E+00]),
                NASA([1000.00, 3500.00],
                    [ 1.07188150E+00,  2.16852677E-02, -1.00256067E-05,
                    2.21412001E-09, -1.90002890E-13, -1.14263932E+04,
                    1.51156107E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.302,
                                well_depth=252.3,
                                rot_relax=1.5),
        note='L8/88')

species(name='HCCO',
        atoms='H:1 C:2 O:1',
        thermo=(NASA([300.00, 1000.00],
                    [ 2.25172140E+00,  1.76550210E-02, -2.37291010E-05,
                    1.72757590E-08, -5.06648110E-12,  2.00594490E+04,
                    1.24904170E+01]),
                NASA([1000.00, 4000.00],
                    [ 5.62820580E+00,  4.08534010E-03, -1.59345470E-06,
                    2.86260520E-10, -1.94078320E-14,  1.93272150E+04,
                    -3.93025950E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=2.5,
                                well_depth=150.0,
                                rot_relax=1.0),
        note='SRIC91')

species(name='CH2CO',
        atoms='H:2 C:2 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 2.13583630E+00,  1.81188721E-02, -1.73947474E-05,
                    9.34397568E-09, -2.01457615E-12, -7.04291804E+03,
                    1.22156480E+01]),
                NASA([1000.00, 3500.00],
                    [ 4.51129732E+00,  9.00359745E-03, -4.16939635E-06,
                    9.23345882E-10, -7.94838201E-14, -7.55105311E+03,
                    6.32247205E-01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='L5/90')

species(name='HCCOH',
        atoms='H:2 C:2 O:1',
        thermo=(NASA([300.00, 1000.00],
                    [ 1.24237330E+00,  3.10722010E-02, -5.08668640E-05,
                    4.31371310E-08, -1.40145940E-11,  8.03161430E+03,
                    1.38743190E+01]),
                NASA([1000.00, 5000.00],
                    [ 5.92382910E+00,  6.79236000E-03, -2.56585640E-06,
                    4.49878410E-10, -2.99401010E-14,  7.26462600E+03,
                    -7.60177420E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='SRI91')

species(name='N',
        atoms='N:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 2.50000000E+00,  0.00000000E+00,  0.00000000E+00,
                    0.00000000E+00,  0.00000000E+00,  5.61046370E+04,
                    4.19390870E+00]),
                NASA([1000.00, 6000.00],
                    [ 2.41594290E+00,  1.74890650E-04, -1.19023690E-07,
                    3.02262450E-11, -2.03609820E-15,  5.61337730E+04,
                    4.64960960E+00])),
        transport=gas_transport(geom='atom',
                                diam=3.298,
                                well_depth=71.4),
        note='L6/88')

species(name='NH',
        atoms='H:1 N:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.49290850E+00,  3.11791980E-04, -1.48904840E-06,
                    2.48164420E-09, -1.03569670E-12,  4.18806290E+04,
                    1.84832780E+00]),
                NASA([1000.00, 6000.00],
                    [ 2.78369280E+00,  1.32984300E-03, -4.24780470E-07,
                    7.83485010E-11, -5.50444700E-15,  4.21208480E+04,
                    5.74077990E+00])),
        transport=gas_transport(geom='linear',
                                diam=2.65,
                                well_depth=80.0,
                                rot_relax=4.0),
        note='And94')

species(name='NH2',
        atoms='H:2 N:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.20400290E+00, -2.10613850E-03,  7.10683480E-06,
                    -5.61151970E-09,  1.64407170E-12,  2.18859100E+04,
                    -1.41842480E-01]),
                NASA([1000.00, 6000.00],
                    [ 2.83474210E+00,  3.20730820E-03, -9.33908040E-07,
                    1.37029530E-10, -7.92061440E-15,  2.21719570E+04,
                    6.52041630E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=2.65,
                                well_depth=80.0,
                                polar=2.26,
                                rot_relax=4.0),
        note='And89')

species(name='NH3',
        atoms='H:3 N:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.28602740E+00, -4.66052300E-03,  2.17185130E-05,
                    -2.28088870E-08,  8.26380460E-12, -6.74172850E+03,
                    -6.25372770E-01]),
                NASA([1000.00, 6000.00],
                    [ 2.63445210E+00,  5.66625600E-03, -1.72786760E-06,
                    2.38671610E-10, -1.25787860E-14, -6.54469580E+03,
                    6.56629280E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=2.92,
                                well_depth=481.0,
                                dipole=1.47,
                                rot_relax=10.0),
        note='J6/77')

species(name='NNH',
        atoms='H:1 N:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.34469270E+00, -4.84970720E-03,  2.00594590E-05,
                    -2.17264640E-08,  7.94695390E-12,  2.87919730E+04,
                    2.97794100E+00]),
                NASA([1000.00, 6000.00],
                    [ 3.76675440E+00,  2.89150820E-03, -1.04166200E-06,
                    1.68425940E-10, -1.00918960E-14,  2.86506970E+04,
                    4.47050670E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.798,
                                well_depth=71.4,
                                rot_relax=1.0),
        note='T07/93')

species(name='NO',
        atoms='O:1 N:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.21847630E+00, -4.63897600E-03,  1.10410220E-05,
                    -9.33613540E-09,  2.80357700E-12,  9.84462300E+03,
                    2.28084640E+00]),
                NASA([1000.00, 6000.00],
                    [ 3.26060560E+00,  1.19110430E-03, -4.29170480E-07,
                    6.94576690E-11, -4.03360990E-15,  9.92097460E+03,
                    6.36930270E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.621,
                                well_depth=97.53,
                                polar=1.76,
                                rot_relax=4.0),
        note='RUS78')

species(name='NO2',
        atoms='O:2 N:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.94403120E+00, -1.58542900E-03,  1.66578120E-05,
                    -2.04754260E-08,  7.83505640E-12,  2.89661790E+03,
                    6.31199170E+00]),
                NASA([1000.00, 6000.00],
                    [ 4.88475420E+00,  2.17239560E-03, -8.28069060E-07,
                    1.57475100E-10, -1.05108950E-14,  2.31649830E+03,
                    -1.17416950E-01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.5,
                                well_depth=200.0,
                                rot_relax=1.0),
        note='L7/88')

species(name='N2O',
        atoms='O:1 N:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 2.25715020E+00,  1.13047280E-02, -1.36713190E-05,
                    9.68198060E-09, -2.93071820E-12,  8.74177440E+03,
                    1.07579920E+01]),
                NASA([1000.00, 6000.00],
                    [ 4.82307290E+00,  2.62702510E-03, -9.58508740E-07,
                    1.60007120E-10, -9.77523030E-15,  8.07340480E+03,
                    -2.20172070E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.828,
                                well_depth=232.4,
                                rot_relax=1.0),
        note='L7/88')

species(name='HNO',
        atoms='H:1 O:1 N:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.53349160E+00, -5.66961710E-03,  1.84732070E-05,
                    -1.71370940E-08,  5.54545730E-12,  1.15482970E+04,
                    1.74984170E+00]),
                NASA([1000.00, 6000.00],
                    [ 2.97925090E+00,  3.49440590E-03, -7.85497780E-07,
                    5.74795940E-11, -1.93359160E-16,  1.17505820E+04,
                    8.60637280E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.492,
                                well_depth=116.7,
                                rot_relax=1.0),
        note='And93')

species(name='CN',
        atoms='C:1 N:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.61293510E+00, -9.55513270E-04,  2.14429770E-06,
                    -3.15163230E-10, -4.64303560E-13,  5.17083400E+04,
                    3.98049950E+00]),
                NASA([1000.00, 6000.00],
                    [ 3.74598050E+00,  4.34507750E-05,  2.97059840E-07,
                    -6.86518060E-11,  4.41341730E-15,  5.15361880E+04,
                    2.78676010E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.856,
                                well_depth=75.0,
                                rot_relax=1.0),
        note='HBH92')

species(name='HCN',
        atoms='H:1 C:1 N:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 2.25898860E+00,  1.00511700E-02, -1.33517630E-05,
                    1.00923490E-08, -3.00890280E-12,  1.47126330E+04,
                    8.91644190E+00]),
                NASA([1000.00, 6000.00],
                    [ 3.80223920E+00,  3.14642280E-03, -1.06321850E-06,
                    1.66197570E-10, -9.79975700E-15,  1.44072920E+04,
                    1.57546010E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.63,
                                well_depth=569.0,
                                rot_relax=1.0),
        note='GRI/98')

species(name='H2CN',
        atoms='H:2 C:1 N:1',
        thermo=(NASA([300.00, 1000.00],
                    [ 2.85166100E+00,  5.69523310E-03,  1.07114000E-06,
                    -1.62261200E-09, -2.35110810E-13,  2.86378200E+04,
                    8.99275110E+00]),
                NASA([1000.00, 4000.00],
                    [ 5.20970300E+00,  2.96929110E-03, -2.85558910E-07,
                    -1.63555000E-10,  3.04325890E-14,  2.76771090E+04,
                    -4.44447800E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.63,
                                well_depth=569.0,
                                rot_relax=1.0),
        note='41687')

species(name='HCNN',
        atoms='H:1 C:1 N:2',
        thermo=(NASA([300.00, 1000.00],
                    [ 2.52431940E+00,  1.59606190E-02, -1.88163540E-05,
                    1.21255400E-08, -3.23573780E-12,  5.42619840E+04,
                    1.16758700E+01]),
                NASA([1000.00, 5000.00],
                    [ 5.89463620E+00,  3.98959590E-03, -1.59823800E-06,
                    2.92493950E-10, -2.00946860E-14,  5.34529410E+04,
                    -5.10305020E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=2.5,
                                well_depth=150.0,
                                rot_relax=1.0),
        note='SRI/94')

species(name='HCNO',
        atoms='H:1 C:1 O:1 N:1',
        thermo=(NASA([300.00, 1382.00],
                    [ 2.64727989E+00,  1.27505342E-02, -1.04794236E-05,
                    4.41432836E-09, -7.57521466E-13,  1.92990252E+04,
                    1.07332972E+01]),
                NASA([1382.00, 5000.00],
                    [ 6.59860456E+00,  3.02778626E-03, -1.07704346E-06,
                    1.71666528E-10, -1.01439391E-14,  1.79661339E+04,
                    -1.03306599E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.828,
                                well_depth=232.4,
                                rot_relax=1.0),
        note='BDEA94')

species(name='HOCN',
        atoms='H:1 C:1 O:1 N:1',
        thermo=(NASA([300.00, 1368.00],
                    [ 3.78604952E+00,  6.88667922E-03, -3.21487864E-06,
                    5.17195767E-10,  1.19360788E-14, -2.82698400E+03,
                    5.63292162E+00]),
                NASA([1368.00, 5000.00],
                    [ 5.89784885E+00,  3.16789393E-03, -1.11801064E-06,
                    1.77243144E-10, -1.04339177E-14, -3.70653331E+03,
                    -6.18167825E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.828,
                                well_depth=232.4,
                                rot_relax=1.0),
        note='BDEA94')

species(name='HNCO',
        atoms='H:1 C:1 O:1 N:1',
        thermo=(NASA([300.00, 1478.00],
                    [ 3.63096317E+00,  7.30282357E-03, -2.28050003E-06,
                    -6.61271298E-10,  3.62235752E-13, -1.55873636E+04,
                    6.19457727E+00]),
                NASA([1478.00, 5000.00],
                    [ 6.22395134E+00,  3.17864004E-03, -1.09378755E-06,
                    1.70735163E-10, -9.95021955E-15, -1.66599344E+04,
                    -8.38224741E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.828,
                                well_depth=232.4,
                                rot_relax=1.0),
        note='BDEA94')

species(name='NCO',
        atoms='C:1 O:1 N:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 2.82693080E+00,  8.80516880E-03, -8.38661340E-06,
                    4.80169640E-09, -1.33135950E-12,  1.46824770E+04,
                    9.55046460E+00]),
                NASA([1000.00, 6000.00],
                    [ 5.15218450E+00,  2.30517610E-03, -8.80331530E-07,
                    1.47890980E-10, -9.09779960E-15,  1.40041230E+04,
                    -2.54426600E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.828,
                                well_depth=232.4,
                                rot_relax=1.0),
        note='EA93')

species(name='N2',
        atoms='N:2',
        thermo=(NASA([300.00, 1000.00],
                    [ 3.29867700E+00,  1.40824040E-03, -3.96322200E-06,
                    5.64151500E-09, -2.44485400E-12, -1.02089990E+03,
                    3.95037200E+00]),
                NASA([1000.00, 5000.00],
                    [ 2.92664000E+00,  1.48797680E-03, -5.68476000E-07,
                    1.00970380E-10, -6.75335100E-15, -9.22797700E+02,
                    5.98052800E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.621,
                                well_depth=97.53,
                                polar=1.76,
                                rot_relax=4.0),
        note='121286')

species(name='AR',
        atoms='Ar:1',
        thermo=(NASA([300.00, 1000.00],
                    [ 2.50000000E+00,  0.00000000E+00,  0.00000000E+00,
                    0.00000000E+00,  0.00000000E+00, -7.45375000E+02,
                    4.36600000E+00]),
                NASA([1000.00, 5000.00],
                    [ 2.50000000E+00,  0.00000000E+00,  0.00000000E+00,
                    0.00000000E+00,  0.00000000E+00, -7.45375000E+02,
                    4.36600000E+00])),
        transport=gas_transport(geom='atom',
                                diam=3.33,
                                well_depth=136.5),
        note='120186')

species(name='C3H7',
        atoms='H:7 C:3',
        thermo=(NASA([300.00, 1000.00],
                    [ 1.05155180E+00,  2.59919800E-02,  2.38005400E-06,
                    -1.96095690E-08,  9.37324700E-12,  1.06318630E+04,
                    2.11225590E+01]),
                NASA([1000.00, 5000.00],
                    [ 7.70269870E+00,  1.60442030E-02, -5.28332200E-06,
                    7.62985900E-10, -3.93922840E-14,  8.29843360E+03,
                    -1.54801800E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.982,
                                well_depth=266.8,
                                rot_relax=1.0),
        note='L9/84')

species(name='C3H8',
        atoms='H:8 C:3',
        thermo=(NASA([300.00, 1000.00],
                    [ 9.33553810E-01,  2.64245790E-02,  6.10597270E-06,
                    -2.19774990E-08,  9.51492530E-12, -1.39585200E+04,
                    1.92016910E+01]),
                NASA([1000.00, 5000.00],
                    [ 7.53413680E+00,  1.88722390E-02, -6.27184910E-06,
                    9.14756490E-10, -4.78380690E-14, -1.64675160E+04,
                    -1.78923490E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.982,
                                well_depth=266.8,
                                rot_relax=1.0),
        note='L4/85')

species(name='CH2CHO',
        atoms='H:3 C:2 O:1',
        thermo=(NASA([300.00, 1000.00],
                    [ 3.40906200E+00,  1.07385740E-02,  1.89149200E-06,
                    -7.15858300E-09,  2.86738500E-12,  1.52147660E+03,
                    9.55829000E+00]),
                NASA([1000.00, 5000.00],
                    [ 5.97567000E+00,  8.13059100E-03, -2.74362400E-06,
                    4.07030400E-10, -2.17601700E-14,  4.90321800E+02,
                    -5.04525100E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='SAND86')

species(name='CH3CHO',
        atoms='H:4 C:2 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.72945950E+00, -3.19328580E-03,  4.75349210E-05,
                    -5.74586110E-08,  2.19311120E-11, -2.15728780E+04,
                    4.10301590E+00]),
                NASA([1000.00, 6000.00],
                    [ 5.40411080E+00,  1.17230590E-02, -4.22631370E-06,
                    6.83724510E-10, -4.09848630E-14, -2.25931220E+04,
                    -3.48079170E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='L8/88')

#-------------------------------------------------------------------------------
# Reaction data
#-------------------------------------------------------------------------------

# Reaction 1
three_body_reaction('2 O + M <=> O2 + M', [1.200000e+17, -1.0, 0.0],
                    efficiencies='CO2:3.6 CO:1.75 H2:2.4 H2O:15.4 AR:0.83 CH4:2.0 C2H6:3.0')

# Reaction 2
three_body_reaction('O + H + M <=> OH + M', [5.000000e+17, -1.0, 0.0],
                    efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0')

# Reaction 3
reaction('O + H2 <=> H + OH', [3.870000e+04, 2.7, 6260.0])

# Reaction 4
reaction('O + HO2 <=> OH + O2', [2.000000e+13, 0.0, 0.0])

# Reaction 5
reaction('O + H2O2 <=> OH + HO2', [9.630000e+06, 2.0, 4000.0])

# Reaction 6
reaction('O + CH <=> H + CO', [5.700000e+13, 0.0, 0.0])

# Reaction 7
reaction('O + CH2 <=> H + HCO', [8.000000e+13, 0.0, 0.0])

# Reaction 8
reaction('O + CH2(S) <=> H2 + CO', [1.500000e+13, 0.0, 0.0])

# Reaction 9
reaction('O + CH2(S) <=> H + HCO', [1.500000e+13, 0.0, 0.0])

# Reaction 10
reaction('O + CH3 <=> H + CH2O', [5.060000e+13, 0.0, 0.0])

# Reaction 11
reaction('O + CH4 <=> OH + CH3', [1.020000e+09, 1.5, 8600.0])

# Reaction 12
falloff_reaction('O + CO (+ M) <=> CO2 (+ M)',
                kf=[1.800000e+10, 0.0, 2385.0],
                kf0=[6.020000e+14, 0.0, 3000.0],
                efficiencies='CO2:3.5 CO:1.5 H2:2.0 H2O:6.0 AR:0.5 CH4:2.0 C2H6:3.0 O2:6.0')

# Reaction 13
reaction('O + HCO <=> OH + CO', [3.000000e+13, 0.0, 0.0])

# Reaction 14
reaction('O + HCO <=> H + CO2', [3.000000e+13, 0.0, 0.0])

# Reaction 15
reaction('O + CH2O <=> OH + HCO', [3.900000e+13, 0.0, 3540.0])

# Reaction 16
reaction('O + CH2OH <=> OH + CH2O', [1.000000e+13, 0.0, 0.0])

# Reaction 17
reaction('O + CH3O <=> OH + CH2O', [1.000000e+13, 0.0, 0.0])

# Reaction 18
reaction('O + CH3OH <=> OH + CH2OH', [3.880000e+05, 2.5, 3100.0])

# Reaction 19
reaction('O + CH3OH <=> OH + CH3O', [1.300000e+05, 2.5, 5000.0])

# Reaction 20
reaction('O + C2H <=> CH + CO', [5.000000e+13, 0.0, 0.0])

# Reaction 21
reaction('O + C2H2 <=> H + HCCO', [1.350000e+07, 2.0, 1900.0])

# Reaction 22
reaction('O + C2H2 <=> OH + C2H', [4.600000e+19, -1.41, 28950.0])

# Reaction 23
reaction('O + C2H2 <=> CO + CH2', [6.940000e+06, 2.0, 1900.0])

# Reaction 24
reaction('O + C2H3 <=> H + CH2CO', [3.000000e+13, 0.0, 0.0])

# Reaction 25
reaction('O + C2H4 <=> CH3 + HCO', [1.250000e+07, 1.83, 220.0])

# Reaction 26
reaction('O + C2H5 <=> CH3 + CH2O', [2.240000e+13, 0.0, 0.0])

# Reaction 27
reaction('O + C2H6 <=> OH + C2H5', [8.980000e+07, 1.92, 5690.0])

# Reaction 28
reaction('O + HCCO <=> H + 2 CO', [1.000000e+14, 0.0, 0.0])

# Reaction 29
reaction('O + CH2CO <=> OH + HCCO', [1.000000e+13, 0.0, 8000.0])

# Reaction 30
reaction('O + CH2CO <=> CH2 + CO2', [1.750000e+12, 0.0, 1350.0])

# Reaction 31
reaction('O2 + CO <=> O + CO2', [2.500000e+12, 0.0, 47800.0])

# Reaction 32
reaction('O2 + CH2O <=> HO2 + HCO', [1.000000e+14, 0.0, 40000.0])

# Reaction 33
three_body_reaction('H + O2 + M <=> HO2 + M', [2.800000e+18, -0.86, 0.0],
                    efficiencies='CO2:1.5 CO:0.75 H2O:0.0 AR:0.0 C2H6:1.5 N2:0.0 O2:0.0')

# Reaction 34
reaction('H + 2 O2 <=> HO2 + O2', [2.080000e+19, -1.24, 0.0])

# Reaction 35
reaction('H + O2 + H2O <=> HO2 + H2O', [1.126000e+19, -0.76, 0.0])

# Reaction 36
reaction('H + O2 + N2 <=> HO2 + N2', [2.600000e+19, -1.24, 0.0])

# Reaction 37
reaction('H + O2 + AR <=> HO2 + AR', [7.000000e+17, -0.8, 0.0])

# Reaction 38
reaction('H + O2 <=> O + OH', [2.650000e+16, -0.6707, 17041.0])

# Reaction 39
three_body_reaction('2 H + M <=> H2 + M', [1.000000e+18, -1.0, 0.0],
                    efficiencies='CO2:0.0 H2:0.0 H2O:0.0 AR:0.63 CH4:2.0 C2H6:3.0')

# Reaction 40
reaction('2 H + H2 <=> 2 H2', [9.000000e+16, -0.6, 0.0])

# Reaction 41
reaction('2 H + H2O <=> H2 + H2O', [6.000000e+19, -1.25, 0.0])

# Reaction 42
reaction('2 H + CO2 <=> H2 + CO2', [5.500000e+20, -2.0, 0.0])

# Reaction 43
three_body_reaction('H + OH + M <=> H2O + M', [2.200000e+22, -2.0, 0.0],
                    efficiencies='H2:0.73 C2H6:3.0 H2O:3.65 AR:0.38 CH4:2.0')

# Reaction 44
reaction('H + HO2 <=> O + H2O', [3.970000e+12, 0.0, 671.0])

# Reaction 45
reaction('H + HO2 <=> O2 + H2', [4.480000e+13, 0.0, 1068.0])

# Reaction 46
reaction('H + HO2 <=> 2 OH', [8.400000e+13, 0.0, 635.0])

# Reaction 47
reaction('H + H2O2 <=> HO2 + H2', [1.210000e+07, 2.0, 5200.0])

# Reaction 48
reaction('H + H2O2 <=> OH + H2O', [1.000000e+13, 0.0, 3600.0])

# Reaction 49
reaction('H + CH <=> C + H2', [1.650000e+14, 0.0, 0.0])

# Reaction 50
falloff_reaction('H + CH2 (+ M) <=> CH3 (+ M)',
                kf=[6.000000e+14, 0.0, 0.0],
                kf0=[1.040000e+26, -2.76, 1600.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.562, T3=91.0, T1=5836.0, T2=8552.0))

# Reaction 51
reaction('H + CH2(S) <=> CH + H2', [3.000000e+13, 0.0, 0.0])

# Reaction 52
falloff_reaction('H + CH3 (+ M) <=> CH4 (+ M)',
                kf=[1.390000e+16, -0.534, 536.0],
                kf0=[2.620000e+33, -4.76, 2440.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:3.0 C2H6:3.0',
                falloff=Troe(A=0.783, T3=74.0, T1=2941.0, T2=6964.0))

# Reaction 53
reaction('H + CH4 <=> CH3 + H2', [6.600000e+08, 1.62, 10840.0])

# Reaction 54
falloff_reaction('H + HCO (+ M) <=> CH2O (+ M)',
                kf=[1.090000e+12, 0.48, -260.0],
                kf0=[2.470000e+24, -2.57, 425.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.7824, T3=271.0, T1=2755.0, T2=6570.0))

# Reaction 55
reaction('H + HCO <=> H2 + CO', [7.340000e+13, 0.0, 0.0])

# Reaction 56
falloff_reaction('H + CH2O (+ M) <=> CH2OH (+ M)',
                kf=[5.400000e+11, 0.454, 3600.0],
                kf0=[1.270000e+32, -4.82, 6530.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.7187, T3=103.0, T1=1291.0, T2=4160.0))

# Reaction 57
falloff_reaction('H + CH2O (+ M) <=> CH3O (+ M)',
                kf=[5.400000e+11, 0.454, 2600.0],
                kf0=[2.200000e+30, -4.8, 5560.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.758, T3=94.0, T1=1555.0, T2=4200.0))

# Reaction 58
reaction('H + CH2O <=> HCO + H2', [5.740000e+07, 1.9, 2742.0])

# Reaction 59
falloff_reaction('H + CH2OH (+ M) <=> CH3OH (+ M)',
                kf=[1.055000e+12, 0.5, 86.0],
                kf0=[4.360000e+31, -4.65, 5080.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.6, T3=100.0, T1=90000.0, T2=10000.0))

# Reaction 60
reaction('H + CH2OH <=> H2 + CH2O', [2.000000e+13, 0.0, 0.0])

# Reaction 61
reaction('H + CH2OH <=> OH + CH3', [1.650000e+11, 0.65, -284.0])

# Reaction 62
reaction('H + CH2OH <=> CH2(S) + H2O', [3.280000e+13, -0.09, 610.0])

# Reaction 63
falloff_reaction('H + CH3O (+ M) <=> CH3OH (+ M)',
                kf=[2.430000e+12, 0.515, 50.0],
                kf0=[4.660000e+41, -7.44, 14080.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.7, T3=100.0, T1=90000.0, T2=10000.0))

# Reaction 64
reaction('H + CH3O <=> H + CH2OH', [4.150000e+07, 1.63, 1924.0])

# Reaction 65
reaction('H + CH3O <=> H2 + CH2O', [2.000000e+13, 0.0, 0.0])

# Reaction 66
reaction('H + CH3O <=> OH + CH3', [1.500000e+12, 0.5, -110.0])

# Reaction 67
reaction('H + CH3O <=> CH2(S) + H2O', [2.620000e+14, -0.23, 1070.0])

# Reaction 68
reaction('H + CH3OH <=> CH2OH + H2', [1.700000e+07, 2.1, 4870.0])

# Reaction 69
reaction('H + CH3OH <=> CH3O + H2', [4.200000e+06, 2.1, 4870.0])

# Reaction 70
falloff_reaction('H + C2H (+ M) <=> C2H2 (+ M)',
                kf=[1.000000e+17, -1.0, 0.0],
                kf0=[3.750000e+33, -4.8, 1900.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.6464, T3=132.0, T1=1315.0, T2=5566.0))

# Reaction 71
falloff_reaction('H + C2H2 (+ M) <=> C2H3 (+ M)',
                kf=[5.600000e+12, 0.0, 2400.0],
                kf0=[3.800000e+40, -7.27, 7220.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.7507, T3=98.5, T1=1302.0, T2=4167.0))

# Reaction 72
falloff_reaction('H + C2H3 (+ M) <=> C2H4 (+ M)',
                kf=[6.080000e+12, 0.27, 280.0],
                kf0=[1.400000e+30, -3.86, 3320.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.782, T3=207.5, T1=2663.0, T2=6095.0))

# Reaction 73
reaction('H + C2H3 <=> H2 + C2H2', [3.000000e+13, 0.0, 0.0])

# Reaction 74
falloff_reaction('H + C2H4 (+ M) <=> C2H5 (+ M)',
                kf=[5.400000e+11, 0.454, 1820.0],
                kf0=[6.000000e+41, -7.62, 6970.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.9753, T3=210.0, T1=984.0, T2=4374.0))

# Reaction 75
reaction('H + C2H4 <=> C2H3 + H2', [1.325000e+06, 2.53, 12240.0])

# Reaction 76
falloff_reaction('H + C2H5 (+ M) <=> C2H6 (+ M)',
                kf=[5.210000e+17, -0.99, 1580.0],
                kf0=[1.990000e+41, -7.08, 6685.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.8422, T3=125.0, T1=2219.0, T2=6882.0))

# Reaction 77
reaction('H + C2H5 <=> H2 + C2H4', [2.000000e+12, 0.0, 0.0])

# Reaction 78
reaction('H + C2H6 <=> C2H5 + H2', [1.150000e+08, 1.9, 7530.0])

# Reaction 79
reaction('H + HCCO <=> CH2(S) + CO', [1.000000e+14, 0.0, 0.0])

# Reaction 80
reaction('H + CH2CO <=> HCCO + H2', [5.000000e+13, 0.0, 8000.0])

# Reaction 81
reaction('H + CH2CO <=> CH3 + CO', [1.130000e+13, 0.0, 3428.0])

# Reaction 82
reaction('H + HCCOH <=> H + CH2CO', [1.000000e+13, 0.0, 0.0])

# Reaction 83
falloff_reaction('H2 + CO (+ M) <=> CH2O (+ M)',
                kf=[4.300000e+07, 1.5, 79600.0],
                kf0=[5.070000e+27, -3.42, 84350.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.932, T3=197.0, T1=1540.0, T2=10300.0))

# Reaction 84
reaction('OH + H2 <=> H + H2O', [2.160000e+08, 1.51, 3430.0])

# Reaction 85
falloff_reaction('2 OH (+ M) <=> H2O2 (+ M)',
                kf=[7.400000e+13, -0.37, 0.0],
                kf0=[2.300000e+18, -0.9, -1700.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.7346, T3=94.0, T1=1756.0, T2=5182.0))

# Reaction 86
reaction('2 OH <=> O + H2O', [3.570000e+04, 2.4, -2110.0])

# Reaction 87
reaction('OH + HO2 <=> O2 + H2O', [1.450000e+13, 0.0, -500.0],
        options='duplicate')

# Reaction 88
reaction('OH + H2O2 <=> HO2 + H2O', [2.000000e+12, 0.0, 427.0],
        options='duplicate')

# Reaction 89
reaction('OH + H2O2 <=> HO2 + H2O', [1.700000e+18, 0.0, 29410.0],
        options='duplicate')

# Reaction 90
reaction('OH + C <=> H + CO', [5.000000e+13, 0.0, 0.0])

# Reaction 91
reaction('OH + CH <=> H + HCO', [3.000000e+13, 0.0, 0.0])

# Reaction 92
reaction('OH + CH2 <=> H + CH2O', [2.000000e+13, 0.0, 0.0])

# Reaction 93
reaction('OH + CH2 <=> CH + H2O', [1.130000e+07, 2.0, 3000.0])

# Reaction 94
reaction('OH + CH2(S) <=> H + CH2O', [3.000000e+13, 0.0, 0.0])

# Reaction 95
falloff_reaction('OH + CH3 (+ M) <=> CH3OH (+ M)',
                kf=[2.790000e+18, -1.43, 1330.0],
                kf0=[4.000000e+36, -5.92, 3140.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.412, T3=195.0, T1=5900.0, T2=6394.0))

# Reaction 96
reaction('OH + CH3 <=> CH2 + H2O', [5.600000e+07, 1.6, 5420.0])

# Reaction 97
reaction('OH + CH3 <=> CH2(S) + H2O', [6.440000e+17, -1.34, 1417.0])

# Reaction 98
reaction('OH + CH4 <=> CH3 + H2O', [1.000000e+08, 1.6, 3120.0])

# Reaction 99
reaction('OH + CO <=> H + CO2', [4.760000e+07, 1.228, 70.0])

# Reaction 100
reaction('OH + HCO <=> H2O + CO', [5.000000e+13, 0.0, 0.0])

# Reaction 101
reaction('OH + CH2O <=> HCO + H2O', [3.430000e+09, 1.18, -447.0])

# Reaction 102
reaction('OH + CH2OH <=> H2O + CH2O', [5.000000e+12, 0.0, 0.0])

# Reaction 103
reaction('OH + CH3O <=> H2O + CH2O', [5.000000e+12, 0.0, 0.0])

# Reaction 104
reaction('OH + CH3OH <=> CH2OH + H2O', [1.440000e+06, 2.0, -840.0])

# Reaction 105
reaction('OH + CH3OH <=> CH3O + H2O', [6.300000e+06, 2.0, 1500.0])

# Reaction 106
reaction('OH + C2H <=> H + HCCO', [2.000000e+13, 0.0, 0.0])

# Reaction 107
reaction('OH + C2H2 <=> H + CH2CO', [2.180000e-04, 4.5, -1000.0])

# Reaction 108
reaction('OH + C2H2 <=> H + HCCOH', [5.040000e+05, 2.3, 13500.0])

# Reaction 109
reaction('OH + C2H2 <=> C2H + H2O', [3.370000e+07, 2.0, 14000.0])

# Reaction 110
reaction('OH + C2H2 <=> CH3 + CO', [4.830000e-04, 4.0, -2000.0])

# Reaction 111
reaction('OH + C2H3 <=> H2O + C2H2', [5.000000e+12, 0.0, 0.0])

# Reaction 112
reaction('OH + C2H4 <=> C2H3 + H2O', [3.600000e+06, 2.0, 2500.0])

# Reaction 113
reaction('OH + C2H6 <=> C2H5 + H2O', [3.540000e+06, 2.12, 870.0])

# Reaction 114
reaction('OH + CH2CO <=> HCCO + H2O', [7.500000e+12, 0.0, 2000.0])

# Reaction 115
reaction('2 HO2 <=> O2 + H2O2', [1.300000e+11, 0.0, -1630.0],
        options='duplicate')

# Reaction 116
reaction('2 HO2 <=> O2 + H2O2', [4.200000e+14, 0.0, 12000.0],
        options='duplicate')

# Reaction 117
reaction('HO2 + CH2 <=> OH + CH2O', [2.000000e+13, 0.0, 0.0])

# Reaction 118
reaction('HO2 + CH3 <=> O2 + CH4', [1.000000e+12, 0.0, 0.0])

# Reaction 119
reaction('HO2 + CH3 <=> OH + CH3O', [3.780000e+13, 0.0, 0.0])

# Reaction 120
reaction('HO2 + CO <=> OH + CO2', [1.500000e+14, 0.0, 23600.0])

# Reaction 121
reaction('HO2 + CH2O <=> HCO + H2O2', [5.600000e+06, 2.0, 12000.0])

# Reaction 122
reaction('C + O2 <=> O + CO', [5.800000e+13, 0.0, 576.0])

# Reaction 123
reaction('C + CH2 <=> H + C2H', [5.000000e+13, 0.0, 0.0])

# Reaction 124
reaction('C + CH3 <=> H + C2H2', [5.000000e+13, 0.0, 0.0])

# Reaction 125
reaction('CH + O2 <=> O + HCO', [6.710000e+13, 0.0, 0.0])

# Reaction 126
reaction('CH + H2 <=> H + CH2', [1.080000e+14, 0.0, 3110.0])

# Reaction 127
reaction('CH + H2O <=> H + CH2O', [5.710000e+12, 0.0, -755.0])

# Reaction 128
reaction('CH + CH2 <=> H + C2H2', [4.000000e+13, 0.0, 0.0])

# Reaction 129
reaction('CH + CH3 <=> H + C2H3', [3.000000e+13, 0.0, 0.0])

# Reaction 130
reaction('CH + CH4 <=> H + C2H4', [6.000000e+13, 0.0, 0.0])

# Reaction 131
falloff_reaction('CH + CO (+ M) <=> HCCO (+ M)',
                kf=[5.000000e+13, 0.0, 0.0],
                kf0=[2.690000e+28, -3.74, 1936.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.5757, T3=237.0, T1=1652.0, T2=5069.0))

# Reaction 132
reaction('CH + CO2 <=> HCO + CO', [1.900000e+14, 0.0, 15792.0])

# Reaction 133
reaction('CH + CH2O <=> H + CH2CO', [9.460000e+13, 0.0, -515.0])

# Reaction 134
reaction('CH + HCCO <=> CO + C2H2', [5.000000e+13, 0.0, 0.0])

# Reaction 135
reaction('CH2 + O2 => OH + H + CO', [5.000000e+12, 0.0, 1500.0])

# Reaction 136
reaction('CH2 + H2 <=> H + CH3', [5.000000e+05, 2.0, 7230.0])

# Reaction 137
reaction('2 CH2 <=> H2 + C2H2', [1.600000e+15, 0.0, 11944.0])

# Reaction 138
reaction('CH2 + CH3 <=> H + C2H4', [4.000000e+13, 0.0, 0.0])

# Reaction 139
reaction('CH2 + CH4 <=> 2 CH3', [2.460000e+06, 2.0, 8270.0])

# Reaction 140
falloff_reaction('CH2 + CO (+ M) <=> CH2CO (+ M)',
                kf=[8.100000e+11, 0.5, 4510.0],
                kf0=[2.690000e+33, -5.11, 7095.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.5907, T3=275.0, T1=1226.0, T2=5185.0))

# Reaction 141
reaction('CH2 + HCCO <=> C2H3 + CO', [3.000000e+13, 0.0, 0.0])

# Reaction 142
reaction('CH2(S) + N2 <=> CH2 + N2', [1.500000e+13, 0.0, 600.0])

# Reaction 143
reaction('CH2(S) + AR <=> CH2 + AR', [9.000000e+12, 0.0, 600.0])

# Reaction 144
reaction('CH2(S) + O2 <=> H + OH + CO', [2.800000e+13, 0.0, 0.0])

# Reaction 145
reaction('CH2(S) + O2 <=> CO + H2O', [1.200000e+13, 0.0, 0.0])

# Reaction 146
reaction('CH2(S) + H2 <=> CH3 + H', [7.000000e+13, 0.0, 0.0])

# Reaction 147
falloff_reaction('CH2(S) + H2O (+ M) <=> CH3OH (+ M)',
                kf=[4.820000e+17, -1.16, 1145.0],
                kf0=[1.880000e+38, -6.36, 5040.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.6027, T3=208.0, T1=3922.0, T2=10180.0))

# Reaction 148
reaction('CH2(S) + H2O <=> CH2 + H2O', [3.000000e+13, 0.0, 0.0])

# Reaction 149
reaction('CH2(S) + CH3 <=> H + C2H4', [1.200000e+13, 0.0, -570.0])

# Reaction 150
reaction('CH2(S) + CH4 <=> 2 CH3', [1.600000e+13, 0.0, -570.0])

# Reaction 151
reaction('CH2(S) + CO <=> CH2 + CO', [9.000000e+12, 0.0, 0.0])

# Reaction 152
reaction('CH2(S) + CO2 <=> CH2 + CO2', [7.000000e+12, 0.0, 0.0])

# Reaction 153
reaction('CH2(S) + CO2 <=> CO + CH2O', [1.400000e+13, 0.0, 0.0])

# Reaction 154
reaction('CH2(S) + C2H6 <=> CH3 + C2H5', [4.000000e+13, 0.0, -550.0])

# Reaction 155
reaction('CH3 + O2 <=> O + CH3O', [3.560000e+13, 0.0, 30480.0])

# Reaction 156
reaction('CH3 + O2 <=> OH + CH2O', [2.310000e+12, 0.0, 20315.0])

# Reaction 157
reaction('CH3 + H2O2 <=> HO2 + CH4', [2.450000e+04, 2.47, 5180.0])

# Reaction 158
falloff_reaction('2 CH3 (+ M) <=> C2H6 (+ M)',
                kf=[6.770000e+16, -1.18, 654.0],
                kf0=[3.400000e+41, -7.03, 2762.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.619, T3=73.2, T1=1180.0, T2=9999.0))

# Reaction 159
reaction('2 CH3 <=> H + C2H5', [6.840000e+12, 0.1, 10600.0])

# Reaction 160
reaction('CH3 + HCO <=> CH4 + CO', [2.648000e+13, 0.0, 0.0])

# Reaction 161
reaction('CH3 + CH2O <=> HCO + CH4', [3.320000e+03, 2.81, 5860.0])

# Reaction 162
reaction('CH3 + CH3OH <=> CH2OH + CH4', [3.000000e+07, 1.5, 9940.0])

# Reaction 163
reaction('CH3 + CH3OH <=> CH3O + CH4', [1.000000e+07, 1.5, 9940.0])

# Reaction 164
reaction('CH3 + C2H4 <=> C2H3 + CH4', [2.270000e+05, 2.0, 9200.0])

# Reaction 165
reaction('CH3 + C2H6 <=> C2H5 + CH4', [6.140000e+06, 1.74, 10450.0])

# Reaction 166
reaction('HCO + H2O <=> H + CO + H2O', [1.500000e+18, -1.0, 17000.0])

# Reaction 167
three_body_reaction('HCO + M <=> H + CO + M', [1.870000e+17, -1.0, 17000.0],
                    efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:0.0 CH4:2.0 C2H6:3.0')

# Reaction 168
reaction('HCO + O2 <=> HO2 + CO', [1.345000e+13, 0.0, 400.0])

# Reaction 169
reaction('CH2OH + O2 <=> HO2 + CH2O', [1.800000e+13, 0.0, 900.0])

# Reaction 170
reaction('CH3O + O2 <=> HO2 + CH2O', [4.280000e-13, 7.6, -3530.0])

# Reaction 171
reaction('C2H + O2 <=> HCO + CO', [1.000000e+13, 0.0, -755.0])

# Reaction 172
reaction('C2H + H2 <=> H + C2H2', [5.680000e+10, 0.9, 1993.0])

# Reaction 173
reaction('C2H3 + O2 <=> HCO + CH2O', [4.580000e+16, -1.39, 1015.0])

# Reaction 174
falloff_reaction('C2H4 (+ M) <=> H2 + C2H2 (+ M)',
                kf=[8.000000e+12, 0.44, 86770.0],
                kf0=[1.580000e+51, -9.3, 97800.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.7345, T3=180.0, T1=1035.0, T2=5417.0))

# Reaction 175
reaction('C2H5 + O2 <=> HO2 + C2H4', [8.400000e+11, 0.0, 3875.0])

# Reaction 176
reaction('HCCO + O2 <=> OH + 2 CO', [3.200000e+12, 0.0, 854.0])

# Reaction 177
reaction('2 HCCO <=> 2 CO + C2H2', [1.000000e+13, 0.0, 0.0])

# Reaction 178
reaction('N + NO <=> N2 + O', [2.700000e+13, 0.0, 355.0])

# Reaction 179
reaction('N + O2 <=> NO + O', [9.000000e+09, 1.0, 6500.0])

# Reaction 180
reaction('N + OH <=> NO + H', [3.360000e+13, 0.0, 385.0])

# Reaction 181
reaction('N2O + O <=> N2 + O2', [1.400000e+12, 0.0, 10810.0])

# Reaction 182
reaction('N2O + O <=> 2 NO', [2.900000e+13, 0.0, 23150.0])

# Reaction 183
reaction('N2O + H <=> N2 + OH', [3.870000e+14, 0.0, 18880.0])

# Reaction 184
reaction('N2O + OH <=> N2 + HO2', [2.000000e+12, 0.0, 21060.0])

# Reaction 185
falloff_reaction('N2O (+ M) <=> N2 + O (+ M)',
                kf=[7.910000e+10, 0.0, 56020.0],
                kf0=[6.370000e+14, 0.0, 56640.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.625 CH4:2.0 C2H6:3.0')

# Reaction 186
reaction('HO2 + NO <=> NO2 + OH', [2.110000e+12, 0.0, -480.0])

# Reaction 187
three_body_reaction('NO + O + M <=> NO2 + M', [1.060000e+20, -1.41, 0.0],
                    efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0')

# Reaction 188
reaction('NO2 + O <=> NO + O2', [3.900000e+12, 0.0, -240.0])

# Reaction 189
reaction('NO2 + H <=> NO + OH', [1.320000e+14, 0.0, 360.0])

# Reaction 190
reaction('NH + O <=> NO + H', [4.000000e+13, 0.0, 0.0])

# Reaction 191
reaction('NH + H <=> N + H2', [3.200000e+13, 0.0, 330.0])

# Reaction 192
reaction('NH + OH <=> HNO + H', [2.000000e+13, 0.0, 0.0])

# Reaction 193
reaction('NH + OH <=> N + H2O', [2.000000e+09, 1.2, 0.0])

# Reaction 194
reaction('NH + O2 <=> HNO + O', [4.610000e+05, 2.0, 6500.0])

# Reaction 195
reaction('NH + O2 <=> NO + OH', [1.280000e+06, 1.5, 100.0])

# Reaction 196
reaction('NH + N <=> N2 + H', [1.500000e+13, 0.0, 0.0])

# Reaction 197
reaction('NH + H2O <=> HNO + H2', [2.000000e+13, 0.0, 13850.0])

# Reaction 198
reaction('NH + NO <=> N2 + OH', [2.160000e+13, -0.23, 0.0])

# Reaction 199
reaction('NH + NO <=> N2O + H', [3.650000e+14, -0.45, 0.0])

# Reaction 200
reaction('NH2 + O <=> OH + NH', [3.000000e+12, 0.0, 0.0])

# Reaction 201
reaction('NH2 + O <=> H + HNO', [3.900000e+13, 0.0, 0.0])

# Reaction 202
reaction('NH2 + H <=> NH + H2', [4.000000e+13, 0.0, 3650.0])

# Reaction 203
reaction('NH2 + OH <=> NH + H2O', [9.000000e+07, 1.5, -460.0])

# Reaction 204
reaction('NNH <=> N2 + H', [3.300000e+08, 0.0, 0.0])

# Reaction 205
three_body_reaction('NNH + M <=> N2 + H + M', [1.300000e+14, -0.11, 4980.0],
                    efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0')

# Reaction 206
reaction('NNH + O2 <=> HO2 + N2', [5.000000e+12, 0.0, 0.0])

# Reaction 207
reaction('NNH + O <=> OH + N2', [2.500000e+13, 0.0, 0.0])

# Reaction 208
reaction('NNH + O <=> NH + NO', [7.000000e+13, 0.0, 0.0])

# Reaction 209
reaction('NNH + H <=> H2 + N2', [5.000000e+13, 0.0, 0.0])

# Reaction 210
reaction('NNH + OH <=> H2O + N2', [2.000000e+13, 0.0, 0.0])

# Reaction 211
reaction('NNH + CH3 <=> CH4 + N2', [2.500000e+13, 0.0, 0.0])

# Reaction 212
three_body_reaction('H + NO + M <=> HNO + M', [4.480000e+19, -1.32, 740.0],
                    efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0')

# Reaction 213
reaction('HNO + O <=> NO + OH', [2.500000e+13, 0.0, 0.0])

# Reaction 214
reaction('HNO + H <=> H2 + NO', [9.000000e+11, 0.72, 660.0])

# Reaction 215
reaction('HNO + OH <=> NO + H2O', [1.300000e+07, 1.9, -950.0])

# Reaction 216
reaction('HNO + O2 <=> HO2 + NO', [1.000000e+13, 0.0, 13000.0])

# Reaction 217
reaction('CN + O <=> CO + N', [7.700000e+13, 0.0, 0.0])

# Reaction 218
reaction('CN + OH <=> NCO + H', [4.000000e+13, 0.0, 0.0])

# Reaction 219
reaction('CN + H2O <=> HCN + OH', [8.000000e+12, 0.0, 7460.0])

# Reaction 220
reaction('CN + O2 <=> NCO + O', [6.140000e+12, 0.0, -440.0])

# Reaction 221
reaction('CN + H2 <=> HCN + H', [2.950000e+05, 2.45, 2240.0])

# Reaction 222
reaction('NCO + O <=> NO + CO', [2.350000e+13, 0.0, 0.0])

# Reaction 223
reaction('NCO + H <=> NH + CO', [5.400000e+13, 0.0, 0.0])

# Reaction 224
reaction('NCO + OH <=> NO + H + CO', [2.500000e+12, 0.0, 0.0])

# Reaction 225
reaction('NCO + N <=> N2 + CO', [2.000000e+13, 0.0, 0.0])

# Reaction 226
reaction('NCO + O2 <=> NO + CO2', [2.000000e+12, 0.0, 20000.0])

# Reaction 227
three_body_reaction('NCO + M <=> N + CO + M', [3.100000e+14, 0.0, 54050.0],
                    efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0')

# Reaction 228
reaction('NCO + NO <=> N2O + CO', [1.900000e+17, -1.52, 740.0])

# Reaction 229
reaction('NCO + NO <=> N2 + CO2', [3.800000e+18, -2.0, 800.0])

# Reaction 230
three_body_reaction('HCN + M <=> H + CN + M', [1.040000e+29, -3.3, 126600.0],
                    efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0')

# Reaction 231
reaction('HCN + O <=> NCO + H', [2.030000e+04, 2.64, 4980.0])

# Reaction 232
reaction('HCN + O <=> NH + CO', [5.070000e+03, 2.64, 4980.0])

# Reaction 233
reaction('HCN + O <=> CN + OH', [3.910000e+09, 1.58, 26600.0])

# Reaction 234
reaction('HCN + OH <=> HOCN + H', [1.100000e+06, 2.03, 13370.0])

# Reaction 235
reaction('HCN + OH <=> HNCO + H', [4.400000e+03, 2.26, 6400.0])

# Reaction 236
reaction('HCN + OH <=> NH2 + CO', [1.600000e+02, 2.56, 9000.0])

# Reaction 237
falloff_reaction('H + HCN (+ M) <=> H2CN (+ M)',
                kf=[3.300000e+13, 0.0, 0.0],
                kf0=[1.400000e+26, -3.4, 1900.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0')

# Reaction 238
reaction('H2CN + N <=> N2 + CH2', [6.000000e+13, 0.0, 400.0])

# Reaction 239
reaction('C + N2 <=> CN + N', [6.300000e+13, 0.0, 46020.0])

# Reaction 240
reaction('CH + N2 <=> HCN + N', [3.120000e+09, 0.88, 20130.0])

# Reaction 241
falloff_reaction('CH + N2 (+ M) <=> HCNN (+ M)',
                kf=[3.100000e+12, 0.15, 0.0],
                kf0=[1.300000e+25, -3.16, 740.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:1.0 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.667, T3=235.0, T1=2117.0, T2=4536.0))

# Reaction 242
reaction('CH2 + N2 <=> HCN + NH', [1.000000e+13, 0.0, 74000.0])

# Reaction 243
reaction('CH2(S) + N2 <=> NH + HCN', [1.000000e+11, 0.0, 65000.0])

# Reaction 244
reaction('C + NO <=> CN + O', [1.900000e+13, 0.0, 0.0])

# Reaction 245
reaction('C + NO <=> CO + N', [2.900000e+13, 0.0, 0.0])

# Reaction 246
reaction('CH + NO <=> HCN + O', [4.100000e+13, 0.0, 0.0])

# Reaction 247
reaction('CH + NO <=> H + NCO', [1.620000e+13, 0.0, 0.0])

# Reaction 248
reaction('CH + NO <=> N + HCO', [2.460000e+13, 0.0, 0.0])

# Reaction 249
reaction('CH2 + NO <=> H + HNCO', [3.100000e+17, -1.38, 1270.0])

# Reaction 250
reaction('CH2 + NO <=> OH + HCN', [2.900000e+14, -0.69, 760.0])

# Reaction 251
reaction('CH2 + NO <=> H + HCNO', [3.800000e+13, -0.36, 580.0])

# Reaction 252
reaction('CH2(S) + NO <=> H + HNCO', [3.100000e+17, -1.38, 1270.0])

# Reaction 253
reaction('CH2(S) + NO <=> OH + HCN', [2.900000e+14, -0.69, 760.0])

# Reaction 254
reaction('CH2(S) + NO <=> H + HCNO', [3.800000e+13, -0.36, 580.0])

# Reaction 255
reaction('CH3 + NO <=> HCN + H2O', [9.600000e+13, 0.0, 28800.0])

# Reaction 256
reaction('CH3 + NO <=> H2CN + OH', [1.000000e+12, 0.0, 21750.0])

# Reaction 257
reaction('HCNN + O <=> CO + H + N2', [2.200000e+13, 0.0, 0.0])

# Reaction 258
reaction('HCNN + O <=> HCN + NO', [2.000000e+12, 0.0, 0.0])

# Reaction 259
reaction('HCNN + O2 <=> O + HCO + N2', [1.200000e+13, 0.0, 0.0])

# Reaction 260
reaction('HCNN + OH <=> H + HCO + N2', [1.200000e+13, 0.0, 0.0])

# Reaction 261
reaction('HCNN + H <=> CH2 + N2', [1.000000e+14, 0.0, 0.0])

# Reaction 262
reaction('HNCO + O <=> NH + CO2', [9.800000e+07, 1.41, 8500.0])

# Reaction 263
reaction('HNCO + O <=> HNO + CO', [1.500000e+08, 1.57, 44000.0])

# Reaction 264
reaction('HNCO + O <=> NCO + OH', [2.200000e+06, 2.11, 11400.0])

# Reaction 265
reaction('HNCO + H <=> NH2 + CO', [2.250000e+07, 1.7, 3800.0])

# Reaction 266
reaction('HNCO + H <=> H2 + NCO', [1.050000e+05, 2.5, 13300.0])

# Reaction 267
reaction('HNCO + OH <=> NCO + H2O', [3.300000e+07, 1.5, 3600.0])

# Reaction 268
reaction('HNCO + OH <=> NH2 + CO2', [3.300000e+06, 1.5, 3600.0])

# Reaction 269
three_body_reaction('HNCO + M <=> NH + CO + M', [1.180000e+16, 0.0, 84720.0],
                    efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0')

# Reaction 270
reaction('HCNO + H <=> H + HNCO', [2.100000e+15, -0.69, 2850.0])

# Reaction 271
reaction('HCNO + H <=> OH + HCN', [2.700000e+11, 0.18, 2120.0])

# Reaction 272
reaction('HCNO + H <=> NH2 + CO', [1.700000e+14, -0.75, 2890.0])

# Reaction 273
reaction('HOCN + H <=> H + HNCO', [2.000000e+07, 2.0, 2000.0])

# Reaction 274
reaction('HCCO + NO <=> HCNO + CO', [9.000000e+12, 0.0, 0.0])

# Reaction 275
reaction('CH3 + N <=> H2CN + H', [6.100000e+14, -0.31, 290.0])

# Reaction 276
reaction('CH3 + N <=> HCN + H2', [3.700000e+12, 0.15, -90.0])

# Reaction 277
reaction('NH3 + H <=> NH2 + H2', [5.400000e+05, 2.4, 9915.0])

# Reaction 278
reaction('NH3 + OH <=> NH2 + H2O', [5.000000e+07, 1.6, 955.0])

# Reaction 279
reaction('NH3 + O <=> NH2 + OH', [9.400000e+06, 1.94, 6460.0])

# Reaction 280
reaction('NH + CO2 <=> HNO + CO', [1.000000e+13, 0.0, 14350.0])

# Reaction 281
reaction('CN + NO2 <=> NCO + NO', [6.160000e+15, -0.752, 345.0])

# Reaction 282
reaction('NCO + NO2 <=> N2O + CO2', [3.250000e+12, 0.0, -705.0])

# Reaction 283
reaction('N + CO2 <=> NO + CO', [3.000000e+12, 0.0, 11300.0])

# Reaction 284
reaction('O + CH3 => H + H2 + CO', [3.370000e+13, 0.0, 0.0])

# Reaction 285
reaction('O + C2H4 <=> H + CH2CHO', [6.700000e+06, 1.83, 220.0])

# Reaction 286
reaction('O + C2H5 <=> H + CH3CHO', [1.096000e+14, 0.0, 0.0])

# Reaction 287
reaction('OH + HO2 <=> O2 + H2O', [5.000000e+15, 0.0, 17330.0],
        options='duplicate')

# Reaction 288
reaction('OH + CH3 => H2 + CH2O', [8.000000e+09, 0.5, -1755.0])

# Reaction 289
falloff_reaction('CH + H2 (+ M) <=> CH3 (+ M)',
                kf=[1.970000e+12, 0.43, -370.0],
                kf0=[4.820000e+25, -2.8, 590.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.578, T3=122.0, T1=2535.0, T2=9365.0))

# Reaction 290
reaction('CH2 + O2 => 2 H + CO2', [5.800000e+12, 0.0, 1500.0])

# Reaction 291
reaction('CH2 + O2 <=> O + CH2O', [2.400000e+12, 0.0, 1500.0])

# Reaction 292
reaction('CH2 + CH2 => 2 H + C2H2', [2.000000e+14, 0.0, 10989.0])

# Reaction 293
reaction('CH2(S) + H2O => H2 + CH2O', [6.820000e+10, 0.25, -935.0])

# Reaction 294
reaction('C2H3 + O2 <=> O + CH2CHO', [3.030000e+11, 0.29, 11.0])

# Reaction 295
reaction('C2H3 + O2 <=> HO2 + C2H2', [1.337000e+06, 1.61, -384.0])

# Reaction 296
reaction('O + CH3CHO <=> OH + CH2CHO', [2.920000e+12, 0.0, 1808.0])

# Reaction 297
reaction('O + CH3CHO => OH + CH3 + CO', [2.920000e+12, 0.0, 1808.0])

# Reaction 298
reaction('O2 + CH3CHO => HO2 + CH3 + CO', [3.010000e+13, 0.0, 39150.0])

# Reaction 299
reaction('H + CH3CHO <=> CH2CHO + H2', [2.050000e+09, 1.16, 2405.0])

# Reaction 300
reaction('H + CH3CHO => CH3 + H2 + CO', [2.050000e+09, 1.16, 2405.0])

# Reaction 301
reaction('OH + CH3CHO => CH3 + H2O + CO', [2.343000e+10, 0.73, -1113.0])

# Reaction 302
reaction('HO2 + CH3CHO => CH3 + H2O2 + CO', [3.010000e+12, 0.0, 11923.0])

# Reaction 303
reaction('CH3 + CH3CHO => CH3 + CH4 + CO', [2.720000e+06, 1.77, 5920.0])

# Reaction 304
falloff_reaction('H + CH2CO (+ M) <=> CH2CHO (+ M)',
                kf=[4.865000e+11, 0.422, -1755.0],
                kf0=[1.012000e+42, -7.63, 3854.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.465, T3=201.0, T1=1773.0, T2=5333.0))

# Reaction 305
reaction('O + CH2CHO => H + CH2 + CO2', [1.500000e+14, 0.0, 0.0])

# Reaction 306
reaction('O2 + CH2CHO => OH + CO + CH2O', [1.810000e+10, 0.0, 0.0])

# Reaction 307
reaction('O2 + CH2CHO => OH + 2 HCO', [2.350000e+10, 0.0, 0.0])

# Reaction 308
reaction('H + CH2CHO <=> CH3 + HCO', [2.200000e+13, 0.0, 0.0])

# Reaction 309
reaction('H + CH2CHO <=> CH2CO + H2', [1.100000e+13, 0.0, 0.0])

# Reaction 310
reaction('OH + CH2CHO <=> H2O + CH2CO', [1.200000e+13, 0.0, 0.0])

# Reaction 311
reaction('OH + CH2CHO <=> HCO + CH2OH', [3.010000e+13, 0.0, 0.0])

# Reaction 312
falloff_reaction('CH3 + C2H5 (+ M) <=> C3H8 (+ M)',
                kf=[9.430000e+12, 0.0, 0.0],
                kf0=[2.710000e+74, -16.82, 13065.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.1527, T3=291.0, T1=2742.0, T2=7748.0))

# Reaction 313
reaction('O + C3H8 <=> OH + C3H7', [1.930000e+05, 2.68, 3716.0])

# Reaction 314
reaction('H + C3H8 <=> C3H7 + H2', [1.320000e+06, 2.54, 6756.0])

# Reaction 315
reaction('OH + C3H8 <=> C3H7 + H2O', [3.160000e+07, 1.8, 934.0])

# Reaction 316
reaction('C3H7 + H2O2 <=> HO2 + C3H8', [3.780000e+02, 2.72, 1500.0])

# Reaction 317
reaction('CH3 + C3H8 <=> C3H7 + CH4', [9.030000e-01, 3.65, 7154.0])

# Reaction 318
falloff_reaction('CH3 + C2H4 (+ M) <=> C3H7 (+ M)',
                kf=[2.550000e+06, 1.6, 5700.0],
                kf0=[3.000000e+63, -14.6, 18170.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.1894, T3=277.0, T1=8748.0, T2=7891.0))

# Reaction 319
reaction('O + C3H7 <=> C2H5 + CH2O', [9.640000e+13, 0.0, 0.0])

# Reaction 320
falloff_reaction('H + C3H7 (+ M) <=> C3H8 (+ M)',
                kf=[3.613000e+13, 0.0, 0.0],
                kf0=[4.420000e+61, -13.545, 11357.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.315, T3=369.0, T1=3285.0, T2=6667.0))

# Reaction 321
reaction('H + C3H7 <=> CH3 + C2H5', [4.060000e+06, 2.19, 890.0])

# Reaction 322
reaction('OH + C3H7 <=> C2H5 + CH2OH', [2.410000e+13, 0.0, 0.0])

# Reaction 323
reaction('HO2 + C3H7 <=> O2 + C3H8', [2.550000e+10, 0.255, -943.0])

# Reaction 324
reaction('HO2 + C3H7 => OH + C2H5 + CH2O', [2.410000e+13, 0.0, 0.0])

# Reaction 325
reaction('CH3 + C3H7 <=> 2 C2H5', [1.927000e+13, -0.32, 0.0])

    """)

    # ================================================================
    # ================================================================
    # ================================================================
    # ================================================================

    fichier_data = open('C3_Gong.cti', 'w')
    fichier_data.write("""#-------------------------------------------------------------------------------
# Gong et al. 2015  http://dx.doi.org/10.1016/j.proci.2014.05.066
#-------------------------------------------------------------------------------
        
units(length='cm', time='s', quantity='mol', act_energy='cal/mol')

ideal_gas(name='gas',
        elements="C H N O Ar He",
        species=\"""H              H2             O              O2             OH
                    OH*            H2O            N2             HO2            H2O2
                    AR             CO             CO2            CH2O           HCO
                    HO2CHO         HCOH           O2CHO          HOCHO          OCHO
                    HOCH2O2H       HOCH2O2        OCH2O2H        HOCH2O         CH3OH
                    CH2OH          CH3O           CH3O2H         CH3O2          CH4
                    CH3            CH2            CH2(S)         C              CH
                    CH*            C2H6           C2H5           C2H4           C2H3
                    C2H2           C2H            CH3CHO         C2H3OH         C2H2OH
                    CH3CO          CH2CHO         O2CH2CHO       HO2CH2CO       CH2CO
                    HCCO           HCCOH          CH3CO3H        CH3CO3         CH3CO2
                    C2H5OH         C2H5O          PC2H4OH        SC2H4OH        O2C2H4OH
                    C2H5O2H        C2H5O2         C2H4O2H        C2H4O1-2       C2H3O1-2
                    CH3COCH3       CH3COCH2       CH3COCH2O2     C3KET21        C2H3CHO
                    C2H3CO         C2H5CHO        C2H5CO         CH3OCH3        CH3OCH2
                    CH3OCH2O2      CH2OCH2O2H     CH3OCH2O2H     CH3OCH2O       O2CH2OCH2O2H
                    HO2CH2OCHO     OCH2OCHO       HOCH2OCO       CH3OCHO        CH3OCO
                    CH2OCHO        HE             C3H8           IC3H7          NC3H7
                    C3H6           C3H5-A         C3H5-S         C3H5-T         C3H4-P
                    C3H4-A         C3H3           C3H5O          C3H6OOH1-2     C3H6OOH1-3
                    C3H6OOH2-1     C3H6OOH1-2O2   C3H6OOH1-3O2   C3H6OOH2-1O2   C3H6OOH2-2
                    NC3H7O2H       IC3H7O2H       NC3H7O2        IC3H7O2        NC3H7O
                    IC3H7O         C3H6O1-2       C3H6O1-3       C3KET12        C3KET13
                    C3H51-2,3OOH   C3H52-1,3OOH   C3H6OH         HOC3H6O2       CH3CHCO
                    AC3H5OOH       C2H3OOH        CC3H4          H2CC           C4H10
                    C4H8-1         C4H8-2         PC4H9          SC4H9          C4H71-1
                    C4H71-2        C4H71-3        C4H71-4        C4H72-2        C4H6
                    PC4H9O2H       SC4H9O2H       PC4H9O2        SC4H9O2        PC4H9O
                    SC4H9O         C4H7O          C4H8O1-2       C4H8O1-3       C4H8O1-4
                    C4H8O2-3       PC4H8OH        SC4H8OH        C4H8OH-1O2     C4H8OH-2O2
                    C4H8OOH1-1     C4H8OOH1-2     C4H8OOH1-3     C4H8OOH1-4     C4H8OOH2-1
                    C4H8OOH2-2     C4H8OOH2-3     C4H8OOH2-4     C4H8OOH1-2O2   C4H8OOH1-3O2
                    C4H8OOH1-4O2   C4H8OOH2-1O2   C4H8OOH2-3O2   C4H8OOH2-4O2   NC4KET12
                    NC4KET13       NC4KET14       NC4KET21       NC4KET23       NC4KET24
                    C2H5COCH3      C2H5COCH2      CH2CH2COCH3    CH3CHCOCH3     C2H3COCH3
                    CH3CHOOCOCH3   CH2CHOOHCOCH3  NC3H7CHO       NC3H7CO        C3H6CHO-1
                    C3H6CHO-2      C3H6CHO-3      C2H5CHCO       SC3H5CHO       SC3H5CO
                    CH2CH2CHO      IC4H10         IC4H9          TC4H9          IC4H8
                    IC4H7          TC4H9O2        IC4H9O2        TC4H8O2H-I     IC4H8O2H-I
                    IC4H8O2H-T     IC4H8O         CC4H8O         IC4H9O         TC4H9O
                    IC4H9O2H       TC4H9O2H       IC4H7O         IC4H8OH        IO2C4H8OH
                    IC3H7CHO       TC3H6CHO       IC3H7CO        IC3H6CHO       TC4H8OOH-IO2
                    IC4H8OOH-IO2   IC4H8OOH-TO2   IC4KETII       IC4KETIT       IC4H7OH
                    IC4H6OH        IC3H5CHO       IC3H5CO        TC3H6OCHO      IC3H6CO
                    IC4H7OOH       TC3H6OHCHO     TC3H6OH        IC3H5OH        TC3H6O2CHO
                    TC3H6O2HCO     IC3H5O2HCHO    CH2CCH2OH      TC4H8CHO       O2C4H8CHO
                    O2HC4H8CO      C3H5OH         TIC4H7Q2-I     IIC4H7Q2-T     IIC4H7Q2-I
                    CH2O2H         C4H4           C4H3-I         C4H3-N         C4H612
                    C4H2           C4H5-I         C4H5-N         CH3CHCHCO      CH2CHCHCHO
                    C4H6O25        C2H3CHOCH2     C4H5-2         C4H6-2         C4H6O23
                    CH3CHCHCHO     C4H4O          H2C4O          NC3H7OH        C3H6OH-3
                    C3H6OH-2       C3H6OH-1       IC3H7OH        IC3H6OH        CH3CHCHOH\""",
        reactions='all',
        transport='Mix',
        initial_state=state(temperature=300.0, pressure=OneAtm))

#-------------------------------------------------------------------------------
# Species data
#-------------------------------------------------------------------------------

species(name='H',
        atoms='H:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 2.50000000E+00,  0.00000000E+00,  0.00000000E+00,
                    0.00000000E+00,  0.00000000E+00,  2.54736600E+04,
                    -4.46682850E-01]),
                NASA([1000.00, 6000.00],
                    [ 2.50000000E+00,  0.00000000E+00,  0.00000000E+00,
                    0.00000000E+00,  0.00000000E+00,  2.54736600E+04,
                    -4.46682850E-01])),
        transport=gas_transport(geom='atom',
                                diam=2.05,
                                well_depth=145.0),
        note='L6/94')

species(name='H2',
        atoms='H:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 2.34433112E+00,  7.98052075E-03, -1.94781510E-05,
                    2.01572094E-08, -7.37611761E-12, -9.17935173E+02,
                    6.83010238E-01]),
                NASA([1000.00, 6000.00],
                    [ 2.93286575E+00,  8.26608026E-04, -1.46402364E-07,
                    1.54100414E-11, -6.88804800E-16, -8.13065581E+02,
                    -1.02432865E+00])),
        transport=gas_transport(geom='linear',
                                diam=2.92,
                                well_depth=38.0,
                                polar=0.79,
                                rot_relax=280.0),
        note='TPIS78')

species(name='O',
        atoms='O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.16826710E+00, -3.27931884E-03,  6.64306396E-06,
                    -6.12806624E-09,  2.11265971E-12,  2.91222592E+04,
                    2.05193346E+00]),
                NASA([1000.00, 6000.00],
                    [ 2.54363697E+00, -2.73162486E-05, -4.19029520E-09,
                    4.95481845E-12, -4.79553694E-16,  2.92260120E+04,
                    4.92229457E+00])),
        transport=gas_transport(geom='atom',
                                diam=2.75,
                                well_depth=80.0),
        note='L1/90')

species(name='O2',
        atoms='O:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.78245636E+00, -2.99673416E-03,  9.84730201E-06,
                    -9.68129509E-09,  3.24372837E-12, -1.06394356E+03,
                    3.65767573E+00]),
                NASA([1000.00, 6000.00],
                    [ 3.66096065E+00,  6.56365811E-04, -1.41149627E-07,
                    2.05797935E-11, -1.29913436E-15, -1.21597718E+03,
                    3.41536279E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.458,
                                well_depth=107.4,
                                polar=1.6,
                                rot_relax=3.8),
        note='RUS89')

species(name='OH',
        atoms='H:1 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.99198424E+00, -2.40106655E-03,  4.61664033E-06,
                    -3.87916306E-09,  1.36319502E-12,  3.36889836E+03,
                    -1.03998477E-01]),
                NASA([1000.00, 6000.00],
                    [ 2.83853033E+00,  1.10741289E-03, -2.94000209E-07,
                    4.20698729E-11, -2.42289890E-15,  3.69780808E+03,
                    5.84494652E+00])),
        transport=gas_transport(geom='linear',
                                diam=2.75,
                                well_depth=80.0),
        note='IU3/03')

species(name='OH*',
        atoms='H:1 O:1',
        thermo=(NASA([300.00, 1000.00],
                    [ 3.63726600E+00,  1.85091000E-04, -1.67616460E-06,
                    2.38720200E-09, -8.43144200E-13,  5.00213000E+04,
                    1.35886050E+00]),
                NASA([1000.00, 5000.00],
                    [ 2.88273000E+00,  1.01397430E-03, -2.27687700E-07,
                    2.17468300E-11, -5.12630500E-16,  5.02650000E+04,
                    5.59571200E+00])),
        transport=gas_transport(geom='linear',
                                diam=2.75,
                                well_depth=80.0),
        note='121286')

species(name='H2O',
        atoms='H:2 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.19863520E+00, -2.03640170E-03,  6.52034160E-06,
                    -5.48792690E-09,  1.77196800E-12, -3.02937260E+04,
                    -8.49009010E-01]),
                NASA([1000.00, 6000.00],
                    [ 2.67703890E+00,  2.97318160E-03, -7.73768890E-07,
                    9.44335140E-11, -4.26899910E-15, -2.98858940E+04,
                    6.88255000E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=2.605,
                                well_depth=572.4,
                                dipole=1.844,
                                rot_relax=4.0),
        note='L5/89')

species(name='N2',
        atoms='N:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.53100528E+00, -1.23660988E-04, -5.02999433E-07,
                    2.43530612E-09, -1.40881235E-12, -1.04697628E+03,
                    2.96747038E+00]),
                NASA([1000.00, 6000.00],
                    [ 2.95257637E+00,  1.39690040E-03, -4.92631603E-07,
                    7.86010195E-11, -4.60755204E-15, -9.23948688E+02,
                    5.87188762E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.621,
                                well_depth=97.53,
                                polar=1.76,
                                rot_relax=4.0),
        note='G8/02')

species(name='HO2',
        atoms='H:1 O:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.30179807E+00, -4.74912097E-03,  2.11582905E-05,
                    -2.42763914E-08,  9.29225225E-12,  2.64018485E+02,
                    3.71666220E+00]),
                NASA([1000.00, 5000.00],
                    [ 4.17228741E+00,  1.88117627E-03, -3.46277286E-07,
                    1.94657549E-11,  1.76256905E-16,  3.10206839E+01,
                    2.95767672E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.458,
                                well_depth=107.4,
                                rot_relax=1.0),
        note='T1/09')

species(name='H2O2',
        atoms='H:2 O:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.31515149E+00, -8.47390622E-04,  1.76404323E-05,
                    -2.26762944E-08,  9.08950158E-12, -1.77067437E+04,
                    3.27373319E+00]),
                NASA([1000.00, 6000.00],
                    [ 4.57977305E+00,  4.05326003E-03, -1.29844730E-06,
                    1.98211400E-10, -1.13968792E-14, -1.80071775E+04,
                    6.64970694E-01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.458,
                                well_depth=107.4,
                                rot_relax=3.8),
        note='T8/03')

species(name='AR',
        atoms='Ar:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 2.50000000E+00,  0.00000000E+00,  0.00000000E+00,
                    0.00000000E+00,  0.00000000E+00, -7.45375000E+02,
                    4.37967491E+00]),
                NASA([1000.00, 6000.00],
                    [ 2.50000000E+00,  0.00000000E+00,  0.00000000E+00,
                    0.00000000E+00,  0.00000000E+00, -7.45375000E+02,
                    4.37967491E+00])),
        transport=gas_transport(geom='atom',
                                diam=3.33,
                                well_depth=136.5),
        note='G5/97')

species(name='CO',
        atoms='C:1 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.57953350E+00, -6.10353690E-04,  1.01681430E-06,
                    9.07005860E-10, -9.04424490E-13, -1.43440860E+04,
                    3.50840930E+00]),
                NASA([1000.00, 6000.00],
                    [ 3.04848590E+00,  1.35172810E-03, -4.85794050E-07,
                    7.88536440E-11, -4.69807460E-15, -1.42661170E+04,
                    6.01709770E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.65,
                                well_depth=98.1,
                                polar=1.95,
                                rot_relax=1.8),
        note='RUS79')

species(name='CO2',
        atoms='C:1 O:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 2.35681300E+00,  8.98412990E-03, -7.12206320E-06,
                    2.45730080E-09, -1.42885480E-13, -4.83719710E+04,
                    9.90090350E+00]),
                NASA([1000.00, 6000.00],
                    [ 4.63651110E+00,  2.74145690E-03, -9.95897590E-07,
                    1.60386660E-10, -9.16198570E-15, -4.90249040E+04,
                    -1.93489550E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.763,
                                well_depth=244.0,
                                polar=2.65,
                                rot_relax=2.1),
        note='L7/88')

species(name='CH2O',
        atoms='H:2 C:1 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.79372312E+00, -9.90833322E-03,  3.73219990E-05,
                    -3.79285237E-08,  1.31772641E-11, -1.43791953E+04,
                    6.02798058E-01]),
                NASA([1000.00, 6000.00],
                    [ 3.16952665E+00,  6.19320560E-03, -2.25056366E-06,
                    3.65975660E-10, -2.20149458E-14, -1.45486831E+04,
                    6.04207898E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.59,
                                well_depth=498.0,
                                rot_relax=2.0),
        note='T5/11')

species(name='HCO',
        atoms='H:1 C:1 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.23754610E+00, -3.32075257E-03,  1.40030264E-05,
                    -1.34239995E-08,  4.37416208E-12,  3.87241185E+03,
                    3.30834869E+00]),
                NASA([1000.00, 6000.00],
                    [ 3.92001542E+00,  2.52279324E-03, -6.71004164E-07,
                    1.05615948E-10, -7.43798261E-15,  3.65342928E+03,
                    3.58077056E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.59,
                                well_depth=498.0),
        note='T5/03')

species(name='HO2CHO',
        atoms='H:2 C:1 O:3',
        thermo=(NASA([300.00, 1378.00],
                    [ 2.42464726E+00,  2.19706380E-02, -1.68705546E-05,
                    6.25612194E-09, -9.11645843E-13, -3.54828006E+04,
                    1.75027796E+01]),
                NASA([1378.00, 5000.00],
                    [ 9.87503878E+00,  4.64663708E-03, -1.67230522E-06,
                    2.68624413E-10, -1.59595232E-14, -3.80502496E+04,
                    -2.24939155E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='6/26/95THERM')

species(name='HCOH',
        atoms='H:2 C:1 O:1',
        thermo=(NASA([300.00, 1398.00],
                    [-2.82157421E+00,  3.57331702E-02, -3.80861580E-05,
                    1.86205951E-08, -3.45957838E-12,  1.12956672E+04,
                    3.48487757E+01]),
                NASA([1398.00, 5000.00],
                    [ 9.18749272E+00,  1.52011152E-03, -6.27603516E-07,
                    1.09727989E-10, -6.89655128E-15,  7.81364593E+03,
                    -2.73434214E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.59,
                                well_depth=498.0,
                                rot_relax=1.0),
        note='MAR94')

species(name='O2CHO',
        atoms='H:1 C:1 O:3',
        thermo=(NASA([300.00, 1368.00],
                    [ 3.96059309E+00,  1.06002279E-02, -5.25713351E-06,
                    1.01716726E-09, -2.87487602E-14, -1.73599383E+04,
                    1.17807483E+01]),
                NASA([1368.00, 5000.00],
                    [ 7.24075139E+00,  4.63312951E-03, -1.63693995E-06,
                    2.59706693E-10, -1.52964699E-14, -1.87027618E+04,
                    -6.49547212E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='6/26/95THERM')

species(name='HOCHO',
        atoms='H:2 C:1 O:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.89836160E+00, -3.55877950E-03,  3.55205380E-05,
                    -4.38499590E-08,  1.71077690E-11, -4.67706090E+04,
                    7.34953970E+00]),
                NASA([1000.00, 6000.00],
                    [ 4.61383160E+00,  6.44963640E-03, -2.29082510E-06,
                    3.67160470E-10, -2.18736750E-14, -4.75148500E+04,
                    8.47883830E-01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='L8/88')

species(name='OCHO',
        atoms='H:1 C:1 O:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.68825921E+00, -4.14871834E-03,  2.55066010E-05,
                    -2.84473900E-08,  1.04422559E-11, -1.69867041E+04,
                    4.28426480E+00]),
                NASA([1000.00, 6000.00],
                    [ 4.14394211E+00,  5.59738818E-03, -1.99794019E-06,
                    3.16179193E-10, -1.85614483E-14, -1.72459887E+04,
                    5.07778617E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.59,
                                well_depth=498.0,
                                rot_relax=2.0),
        note='ATCT/A')

species(name='HOCH2O2H',
        atoms='H:4 C:1 O:3',
        thermo=(NASA([300.00, 1422.00],
                    [ 1.85716693E+00,  3.23153132E-02, -2.69928902E-05,
                    1.11694484E-08, -1.81284103E-12, -4.00314471E+04,
                    1.90917729E+01]),
                NASA([1422.00, 5000.00],
                    [ 1.16303827E+01,  7.15133688E-03, -2.39035030E-06,
                    3.65772791E-10, -2.10199524E-14, -4.31079242E+04,
                    -3.24276725E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.626,
                                well_depth=481.8,
                                dipole=1.7,
                                rot_relax=1.0),
        note='4/9/98THERM')

species(name='HOCH2O2',
        atoms='H:3 C:1 O:3',
        thermo=(NASA([300.00, 1412.00],
                    [ 2.85441621E+00,  2.33663535E-02, -1.88115990E-05,
                    7.96709515E-09, -1.36346618E-12, -2.29866196E+04,
                    1.51730565E+01]),
                NASA([1412.00, 5000.00],
                    [ 9.04545938E+00,  7.15223373E-03, -2.37005676E-06,
                    3.60083481E-10, -2.05750228E-14, -2.49414886E+04,
                    -1.74210530E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.626,
                                well_depth=481.8,
                                dipole=1.7,
                                rot_relax=1.0),
        note='4/9/98THERM')

species(name='OCH2O2H',
        atoms='H:3 C:1 O:3',
        thermo=(NASA([300.00, 1420.00],
                    [ 1.93823075E+00,  3.01465730E-02, -2.61053152E-05,
                    1.09463562E-08, -1.78312692E-12, -1.38166625E+04,
                    1.85042002E+01]),
                NASA([1420.00, 5000.00],
                    [ 1.15398246E+01,  5.34291432E-03, -1.81878917E-06,
                    2.81968625E-10, -1.63584348E-14, -1.68237489E+04,
                    -3.20700633E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.626,
                                well_depth=481.8,
                                dipole=1.7,
                                rot_relax=1.0),
        note='4/9/98THERM')

species(name='HOCH2O',
        atoms='H:3 C:1 O:2',
        thermo=(NASA([300.00, 1452.00],
                    [ 4.11183145E+00,  7.53850697E-03,  3.77337370E-06,
                    -5.38746005E-09,  1.45615887E-12, -2.28023001E+04,
                    7.46807254E+00]),
                NASA([1452.00, 5000.00],
                    [ 6.39521515E+00,  7.43673043E-03, -2.50422354E-06,
                    3.84879712E-10, -2.21778689E-14, -2.41108840E+04,
                    -6.63865583E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.41,
                                well_depth=470.6,
                                rot_relax=1.5),
        note='2/16/99THERM')

species(name='CH3OH',
        atoms='H:4 C:1 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 5.65851051E+00, -1.62983419E-02,  6.91938156E-05,
                    -7.58372926E-08,  2.80427550E-11, -2.56119736E+04,
                    -8.97330508E-01]),
                NASA([1000.00, 6000.00],
                    [ 3.52726795E+00,  1.03178783E-02, -3.62892944E-06,
                    5.77448016E-10, -3.42182632E-14, -2.60028834E+04,
                    5.16758693E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.626,
                                well_depth=481.8,
                                rot_relax=1.0),
        note='T06/02')

species(name='CH2OH',
        atoms='H:3 C:1 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.47834367E+00, -1.35070310E-03,  2.78484980E-05,
                    -3.64869060E-08,  1.47907450E-11, -3.50072890E+03,
                    3.30913500E+00]),
                NASA([1000.00, 6000.00],
                    [ 5.09314370E+00,  5.94761260E-03, -2.06497460E-06,
                    3.23008173E-10, -1.88125902E-14, -4.03409640E+03,
                    -1.84691493E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.69,
                                well_depth=417.0,
                                dipole=1.7,
                                rot_relax=2.0),
        note='IU2/03')

species(name='CH3O',
        atoms='H:3 C:1 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.71180502E+00, -2.80463306E-03,  3.76550971E-05,
                    -4.73072089E-08,  1.86588420E-11,  1.29569760E+03,
                    6.57240864E+00]),
                NASA([1000.00, 6000.00],
                    [ 4.75779238E+00,  7.44142474E-03, -2.69705176E-06,
                    4.38090504E-10, -2.63537098E-14,  3.78111940E+02,
                    -1.96680028E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.69,
                                well_depth=417.0,
                                dipole=1.7,
                                rot_relax=2.0),
        note='IU1/03')

species(name='CH3O2H',
        atoms='H:4 C:1 O:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 2.90540897E+00,  1.74994735E-02,  5.28243630E-06,
                    -2.52827275E-08,  1.34368212E-11, -1.68894632E+04,
                    1.13741987E+01]),
                NASA([1000.00, 6000.00],
                    [ 7.76538058E+00,  8.61499712E-03, -2.98006935E-06,
                    4.68638071E-10, -2.75339255E-14, -1.82979984E+04,
                    -1.43992663E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.626,
                                well_depth=481.8,
                                rot_relax=1.0),
        note='A7/05')

species(name='CH3O2',
        atoms='H:3 C:1 O:2',
        thermo=(NASA([300.00, 1000.00],
                    [ 3.62497097E+00,  3.59397933E-03,  2.26538097E-05,
                    -2.95391947E-08,  1.11977570E-11,  7.93040410E+01,
                    9.96382194E+00]),
                NASA([1000.00, 5000.00],
                    [ 4.80390863E+00,  9.95844638E-03, -3.85301026E-06,
                    6.84740497E-10, -4.58402955E-14, -7.47135460E+02,
                    1.45281400E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.626,
                                well_depth=481.8,
                                rot_relax=1.0))

species(name='CH4',
        atoms='H:4 C:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 5.14911468E+00, -1.36622009E-02,  4.91453921E-05,
                    -4.84246767E-08,  1.66603441E-11, -1.02465983E+04,
                    -4.63848842E+00]),
                NASA([1000.00, 6000.00],
                    [ 1.65326226E+00,  1.00263099E-02, -3.31661238E-06,
                    5.36483138E-10, -3.14696758E-14, -1.00095936E+04,
                    9.90506283E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.746,
                                well_depth=141.4,
                                polar=2.6,
                                rot_relax=13.0),
        note='G8/99')

species(name='CH3',
        atoms='H:3 C:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.65717970E+00,  2.12659790E-03,  5.45838830E-06,
                    -6.61810030E-09,  2.46570740E-12,  1.64227160E+04,
                    1.67353540E+00]),
                NASA([1000.00, 6000.00],
                    [ 2.97812060E+00,  5.79785200E-03, -1.97558000E-06,
                    3.07297900E-10, -1.79174160E-14,  1.65095130E+04,
                    4.72247990E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.8,
                                well_depth=144.0),
        note='IU0702')

species(name='CH2',
        atoms='H:2 C:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.71757846E+00,  1.27391260E-03,  2.17347251E-06,
                    -3.48858500E-09,  1.65208866E-12,  4.58723866E+04,
                    1.75297945E+00]),
                NASA([1000.00, 6000.00],
                    [ 3.14631886E+00,  3.03671259E-03, -9.96474439E-07,
                    1.50483580E-10, -8.57335515E-15,  4.60412605E+04,
                    4.72341711E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.8,
                                well_depth=144.0),
        note='IU3/03')

species(name='CH2(S)',
        atoms='H:2 C:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.19331325E+00, -2.33105184E-03,  8.15676451E-06,
                    -6.62985981E-09,  1.93233199E-12,  5.03662246E+04,
                    -7.46734310E-01]),
                NASA([1000.00, 6000.00],
                    [ 3.13501686E+00,  2.89593926E-03, -8.16668090E-07,
                    1.13572697E-10, -6.36262835E-15,  5.05040504E+04,
                    4.06030621E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.8,
                                well_depth=144.0),
        note='IU6/03')

species(name='C',
        atoms='C:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 2.55423950E+00, -3.21537720E-04,  7.33792230E-07,
                    -7.32234870E-10,  2.66521440E-13,  8.54426810E+04,
                    4.53130850E+00]),
                NASA([1000.00, 6000.00],
                    [ 2.60558300E+00, -1.95934340E-04,  1.06737220E-07,
                    -1.64239400E-11,  8.18705800E-16,  8.54117420E+04,
                    4.19238680E+00])),
        transport=gas_transport(geom='atom',
                                diam=3.298,
                                well_depth=71.4),
        note='L7/88')

species(name='CH',
        atoms='H:1 C:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.48975830E+00,  3.24321600E-04, -1.68997510E-06,
                    3.16284200E-09, -1.40618030E-12,  7.06126460E+04,
                    2.08428410E+00]),
                NASA([1000.00, 6000.00],
                    [ 2.52093690E+00,  1.76536390E-03, -4.61476600E-07,
                    5.92896750E-11, -3.34745010E-15,  7.09467690E+04,
                    7.40518290E+00])),
        transport=gas_transport(geom='linear',
                                diam=2.75,
                                well_depth=80.0),
        note='IU3/03')

species(name='CH*',
        atoms='H:1 C:1',
        thermo=(NASA([300.00, 1000.00],
                    [ 3.20020200E+00,  2.07287500E-03, -5.13443100E-06,
                    5.73389000E-09, -1.95553300E-12,  1.03937140E+05,
                    3.33158700E+00]),
                NASA([1000.00, 5000.00],
                    [ 2.19622300E+00,  2.34038100E-03, -7.05820100E-07,
                    9.00758200E-11, -3.85504000E-15,  1.04195590E+05,
                    9.17837300E+00])),
        transport=gas_transport(geom='linear',
                                diam=2.75,
                                well_depth=80.0),
        note='073003')

species(name='C2H6',
        atoms='H:6 C:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.29142572E+00, -5.50154901E-03,  5.99438458E-05,
                    -7.08466469E-08,  2.68685836E-11, -1.15222056E+04,
                    2.66678994E+00]),
                NASA([1000.00, 6000.00],
                    [ 4.04666411E+00,  1.53538802E-02, -5.47039485E-06,
                    8.77826544E-10, -5.23167531E-14, -1.24473499E+04,
                    -9.68698313E-01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.35,
                                well_depth=247.5,
                                rot_relax=1.5),
        note='G8/88')

species(name='C2H5',
        atoms='H:5 C:2',
        thermo=(NASA([300.00, 1387.00],
                    [ 1.32730217E+00,  1.76656753E-02, -6.14926558E-06,
                    -3.01143466E-10,  4.38617775E-13,  1.34284028E+04,
                    1.71789216E+01]),
                NASA([1387.00, 5000.00],
                    [ 5.88784390E+00,  1.03076793E-02, -3.46844396E-06,
                    5.32499257E-10, -3.06512651E-14,  1.15065499E+04,
                    -8.49651771E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.35,
                                well_depth=247.5,
                                rot_relax=1.5),
        note='8/4/4THERM')

species(name='C2H4',
        atoms='H:4 C:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.95920063E+00, -7.57051373E-03,  5.70989993E-05,
                    -6.91588352E-08,  2.69884190E-11,  5.08977598E+03,
                    4.09730213E+00]),
                NASA([1000.00, 6000.00],
                    [ 3.99182724E+00,  1.04833908E-02, -3.71721342E-06,
                    5.94628366E-10, -3.53630386E-14,  4.26865851E+03,
                    -2.69081762E-01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.496,
                                well_depth=238.4,
                                rot_relax=1.5),
        note='G1/00')

species(name='C2H3',
        atoms='H:3 C:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.36377642E+00,  2.65765722E-04,  2.79620704E-05,
                    -3.72986942E-08,  1.51590176E-11,  3.44749589E+04,
                    7.91510092E+00]),
                NASA([1000.00, 6000.00],
                    [ 4.15026763E+00,  7.54021341E-03, -2.62997847E-06,
                    4.15974048E-10, -2.45407509E-14,  3.38566380E+04,
                    1.72812235E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.721,
                                well_depth=265.3,
                                rot_relax=1.0),
        note='ATCT/A')

species(name='C2H2',
        atoms='H:2 C:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 8.08679682E-01,  2.33615762E-02, -3.55172234E-05,
                    2.80152958E-08, -8.50075165E-12,  2.64289808E+04,
                    1.39396761E+01]),
                NASA([1000.00, 6000.00],
                    [ 4.65878489E+00,  4.88396667E-03, -1.60828888E-06,
                    2.46974544E-10, -1.38605959E-14,  2.57594042E+04,
                    -3.99838194E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.721,
                                well_depth=265.3,
                                rot_relax=2.5),
        note='G1/91')

species(name='C2H',
        atoms='H:1 C:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 2.89867676E+00,  1.32988489E-02, -2.80733327E-05,
                    2.89484755E-08, -1.07502351E-11,  6.70616050E+04,
                    6.18547632E+00]),
                NASA([1000.00, 6000.00],
                    [ 3.66270248E+00,  3.82492252E-03, -1.36632500E-06,
                    2.13455040E-10, -1.23216848E-14,  6.71683790E+04,
                    3.92205792E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.721,
                                well_depth=265.3,
                                rot_relax=2.5),
        note='T5/10')

species(name='CH3CHO',
        atoms='H:4 C:2 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.72945950E+00, -3.19328580E-03,  4.75349210E-05,
                    -5.74586110E-08,  2.19311120E-11, -2.15728780E+04,
                    4.10301590E+00]),
                NASA([1000.00, 6000.00],
                    [ 5.40411080E+00,  1.17230590E-02, -4.22631370E-06,
                    6.83724510E-10, -4.09848630E-14, -2.25931220E+04,
                    -3.48079170E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='L8/88')

species(name='C2H3OH',
        atoms='H:4 C:2 O:1',
        thermo=(NASA([300.00, 1410.00],
                    [-1.27972260E-01,  3.38506073E-02, -3.30644935E-05,
                    1.64858739E-08, -3.19935455E-12, -1.59914544E+04,
                    2.30438601E+01]),
                NASA([1410.00, 5000.00],
                    [ 8.32598158E+00,  8.03387281E-03, -2.63928405E-06,
                    3.98410726E-10, -2.26551155E-14, -1.83221436E+04,
                    -2.02080305E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.41,
                                well_depth=470.6,
                                rot_relax=1.5),
        note='2/3/9THERM')

species(name='C2H2OH',
        atoms='H:3 C:2 O:1',
        thermo=(NASA([300.00, 1000.00],
                    [ 5.62397968E-01,  2.64598620E-02, -2.28453705E-05,
                    9.14478695E-09, -1.06711936E-12,  1.48374643E+04,
                    2.10439544E+01]),
                NASA([1000.00, 5000.00],
                    [ 6.81339897E+00,  8.05827723E-03, -3.11728612E-06,
                    5.54590896E-10, -3.71804255E-14,  1.31583034E+04,
                    -1.09863525E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.162,
                                well_depth=224.7,
                                rot_relax=1.0))

species(name='CH3CO',
        atoms='H:3 C:2 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.03587050E+00,  8.77294870E-04,  3.07100100E-05,
                    -3.92475650E-08,  1.52968690E-11, -2.68207380E+03,
                    7.86176820E+00]),
                NASA([1000.00, 6000.00],
                    [ 5.31371650E+00,  9.17377930E-03, -3.32203860E-06,
                    5.39474560E-10, -3.24523680E-14, -3.64504140E+03,
                    -1.67575580E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='IU2/03')

species(name='CH2CHO',
        atoms='H:3 C:2 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 2.79502600E+00,  1.01099472E-02,  1.61750645E-05,
                    -3.10303145E-08,  1.39436139E-11,  1.62944975E+02,
                    1.23646657E+01]),
                NASA([1000.00, 6000.00],
                    [ 6.53928338E+00,  7.80238629E-03, -2.76413612E-06,
                    4.42098906E-10, -2.62954290E-14, -1.18858659E+03,
                    -8.72091393E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='T03/10')

species(name='O2CH2CHO',
        atoms='H:3 C:2 O:3',
        thermo=(NASA([300.00, 1393.00],
                    [-1.29465843E+00,  4.44936393E-02, -4.26577074E-05,
                    2.07391950E-08, -3.96828771E-12, -1.18275628E+04,
                    3.60778797E+01]),
                NASA([1393.00, 5000.00],
                    [ 1.11807543E+01,  9.14479256E-03, -3.15089833E-06,
                    4.91944238E-10, -2.86639180E-14, -1.55790331E+04,
                    -2.87892740E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.428,
                                well_depth=275.049,
                                rot_relax=1.0),
        note='BOZ_03')

species(name='HO2CH2CO',
        atoms='H:3 C:2 O:3',
        thermo=(NASA([300.00, 1386.00],
                    [ 2.22681686E+00,  3.56781380E-02, -3.26401909E-05,
                    1.47651988E-08, -2.64794380E-12, -1.18735095E+04,
                    1.91581197E+01]),
                NASA([1386.00, 5000.00],
                    [ 1.04146322E+01,  1.12680116E-02, -5.17494839E-06,
                    1.00333285E-09, -6.68165911E-14, -1.40955672E+04,
                    -2.27894400E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.505,
                                well_depth=279.007,
                                dipole=1.3,
                                rot_relax=1.0),
        note='BOZ_03')

species(name='CH2CO',
        atoms='H:2 C:2 O:1',
        thermo=(NASA([300.00, 1000.00],
                    [ 1.81422511E+00,  1.99008590E-02, -2.21416008E-05,
                    1.45028521E-08, -3.98877068E-12, -7.05394926E+03,
                    1.36079359E+01]),
                NASA([1000.00, 5000.00],
                    [ 5.35869367E+00,  6.95641586E-03, -2.64802637E-06,
                    4.65067592E-10, -3.08641820E-14, -7.90294013E+03,
                    -3.98525731E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0))

species(name='HCCO',
        atoms='H:1 C:2 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 1.87607969E+00,  2.21205418E-02, -3.58869325E-05,
                    3.05402541E-08, -1.01281069E-11,  2.01633840E+04,
                    1.36968290E+01]),
                NASA([1000.00, 6000.00],
                    [ 5.91479333E+00,  3.71408730E-03, -1.30137010E-06,
                    2.06473345E-10, -1.21476759E-14,  1.93596301E+04,
                    -5.50567269E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=2.5,
                                well_depth=150.0,
                                rot_relax=1.0),
        note='T4/09')

species(name='HCCOH',
        atoms='H:2 C:2 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 2.05541154E+00,  2.52003372E-02, -3.80821654E-05,
                    3.09890632E-08, -9.89799902E-12,  9.76872113E+03,
                    1.22271534E+01]),
                NASA([1000.00, 6000.00],
                    [ 6.37509678E+00,  5.49429011E-03, -1.88136576E-06,
                    2.93803536E-10, -1.71771901E-14,  8.93277676E+03,
                    -8.24498007E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='T12/09')

species(name='CH3CO3H',
        atoms='H:4 C:2 O:3',
        thermo=(NASA([300.00, 1391.00],
                    [ 2.24135876E+00,  3.37963514E-02, -2.53887482E-05,
                    9.67583587E-09, -1.49266157E-12, -4.24677831E+04,
                    1.70668133E+01]),
                NASA([1391.00, 5000.00],
                    [ 1.25060485E+01,  9.47789695E-03, -3.30402246E-06,
                    5.19630793E-10, -3.04233568E-14, -4.59856703E+04,
                    -3.79195947E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='6/26/95THERM')

species(name='CH3CO3',
        atoms='H:3 C:2 O:3',
        thermo=(NASA([300.00, 1391.00],
                    [ 3.60373432E+00,  2.70080341E-02, -2.08293438E-05,
                    8.50541104E-09, -1.43846110E-12, -2.34205171E+04,
                    1.12014914E+01]),
                NASA([1391.00, 5000.00],
                    [ 1.12522498E+01,  8.33652672E-03, -2.89014530E-06,
                    4.52781734E-10, -2.64354456E-14, -2.60238584E+04,
                    -2.96370457E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='4/3/0THERM')

species(name='CH3CO2',
        atoms='H:3 C:2 O:2',
        thermo=(NASA([300.00, 1395.00],
                    [ 1.37440768E+00,  2.49115604E-02, -1.74308894E-05,
                    6.24799508E-09, -9.09516835E-13, -2.72330150E+04,
                    1.81405454E+01]),
                NASA([1395.00, 5000.00],
                    [ 8.54059736E+00,  8.32951214E-03, -2.84722010E-06,
                    4.41927196E-10, -2.56373394E-14, -2.97290678E+04,
                    -2.03883545E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='2/14/95THERM')

species(name='C2H5OH',
        atoms='H:6 C:2 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.85869570E+00, -3.74017260E-03,  6.95553780E-05,
                    -8.86547960E-08,  3.51688350E-11, -2.99961320E+04,
                    4.80185450E+00]),
                NASA([1000.00, 6000.00],
                    [ 6.56243650E+00,  1.52042220E-02, -5.38967950E-06,
                    8.62250110E-10, -5.12897870E-14, -3.15256210E+04,
                    -9.47302020E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.41,
                                well_depth=470.6,
                                rot_relax=1.5),
        note='L8/88')

species(name='C2H5O',
        atoms='H:5 C:2 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.30742680E+00,  6.41472050E-03,  3.11397140E-05,
                    -4.33140830E-08,  1.72761840E-11, -3.40275240E+03,
                    5.90258370E+00]),
                NASA([1000.00, 6000.00],
                    [ 6.68899820E+00,  1.31256760E-02, -4.70388400E-06,
                    7.58585520E-10, -4.54133060E-14, -4.74578320E+03,
                    -9.69837550E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.41,
                                well_depth=470.6,
                                rot_relax=1.5),
        note='IU2/03')

species(name='PC2H4OH',
        atoms='H:5 C:2 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.47893092E+00,  7.59782301E-03,  2.81794908E-05,
                    -4.26953487E-08,  1.78878934E-11, -4.71446256E+03,
                    6.38921206E+00]),
                NASA([1000.00, 6000.00],
                    [ 7.02824536E+00,  1.20037746E-02, -4.21306455E-06,
                    6.69471213E-10, -3.96371893E-14, -5.92493321E+03,
                    -9.40355948E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.41,
                                well_depth=470.6,
                                rot_relax=1.5),
        note='T12/01')

species(name='SC2H4OH',
        atoms='H:5 C:2 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.22283250E+00,  5.12174798E-03,  3.48386522E-05,
                    -4.91943637E-08,  2.01183723E-11, -8.20503939E+03,
                    8.01675700E+00]),
                NASA([1000.00, 6000.00],
                    [ 6.35842302E+00,  1.24356276E-02, -4.33096839E-06,
                    6.84530381E-10, -4.03713238E-14, -9.37900432E+03,
                    -6.05106112E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.41,
                                well_depth=470.6,
                                rot_relax=1.5),
        note='T10/04')

species(name='O2C2H4OH',
        atoms='H:5 C:2 O:3',
        thermo=(NASA([300.00, 1392.00],
                    [ 4.11839445E+00,  2.72240632E-02, -1.60824430E-05,
                    5.17033408E-09, -7.31610168E-13, -2.30857785E+04,
                    1.28482112E+01]),
                NASA([1392.00, 5000.00],
                    [ 1.07432659E+01,  1.30957787E-02, -4.45370088E-06,
                    6.88548738E-10, -3.98230113E-14, -2.55911274E+04,
                    -2.33254953E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='2/14/95THERM')

species(name='C2H5O2H',
        atoms='H:6 C:2 O:2',
        thermo=(NASA([300.00, 1391.00],
                    [ 1.57329011E+00,  3.52379996E-02, -2.53203993E-05,
                    9.56802476E-09, -1.48167375E-12, -2.15278368E+04,
                    1.90472032E+01]),
                NASA([1391.00, 5000.00],
                    [ 1.12305737E+01,  1.20482120E-02, -3.96730201E-06,
                    6.00754632E-10, -3.42657803E-14, -2.47977531E+04,
                    -3.25607232E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.41,
                                well_depth=470.6,
                                rot_relax=1.5),
        note='1/14/5THERM')

species(name='C2H5O2',
        atoms='H:5 C:2 O:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.50099327E+00,  6.87965342E-03,  4.74143971E-05,
                    -6.92287127E-08,  2.87395324E-11, -5.39547911E+03,
                    7.91490068E+00]),
                NASA([1000.00, 6000.00],
                    [ 8.88872432E+00,  1.35833179E-02, -4.91116949E-06,
                    7.92343362E-10, -4.73525704E-14, -7.44107388E+03,
                    -1.90789836E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.41,
                                well_depth=470.6,
                                rot_relax=1.5),
        note='T10/10')

species(name='C2H4O2H',
        atoms='H:5 C:2 O:2',
        thermo=(NASA([300.00, 1397.00],
                    [ 3.46916874E+00,  2.71188626E-02, -2.08022550E-05,
                    8.44284845E-09, -1.40756215E-12,  3.89688270E+03,
                    1.43400726E+01]),
                NASA([1397.00, 5000.00],
                    [ 1.05228954E+01,  9.48091381E-03, -3.55727763E-06,
                    6.41445994E-10, -4.21232247E-14,  1.55718322E+03,
                    -2.31413632E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.41,
                                well_depth=470.6,
                                rot_relax=1.5),
        note='1/14/5THERM')

species(name='C2H4O1-2',
        atoms='H:4 C:2 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.75905320E+00, -9.44121800E-03,  8.03097210E-05,
                    -1.00807880E-07,  4.00399210E-11, -7.56081430E+03,
                    7.84974750E+00]),
                NASA([1000.00, 6000.00],
                    [ 5.48876410E+00,  1.20461900E-02, -4.33369310E-06,
                    7.00283110E-10, -4.19490880E-14, -9.18042510E+03,
                    -7.07996050E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='L8/88')

species(name='C2H3O1-2',
        atoms='H:3 C:2 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.58349017E+00, -6.02275805E-03,  6.32426867E-05,
                    -8.18540707E-08,  3.30444505E-11,  1.85681353E+04,
                    9.59725926E+00]),
                NASA([1000.00, 6000.00],
                    [ 5.60158035E+00,  9.17613962E-03, -3.28028902E-06,
                    5.27903888E-10, -3.15362241E-14,  1.71446252E+04,
                    -5.47228512E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='A1/05')

species(name='CH3COCH3',
        atoms='H:6 C:3 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 5.55638920E+00, -2.83863547E-03,  7.05722951E-05,
                    -8.78130984E-08,  3.40290951E-11, -2.78325393E+04,
                    2.31960221E+00]),
                NASA([1000.00, 6000.00],
                    [ 7.29796974E+00,  1.75656913E-02, -6.31678065E-06,
                    1.02025553E-09, -6.10903592E-14, -2.95368927E+04,
                    -1.27591704E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.86,
                                well_depth=435.5,
                                rot_relax=1.0),
        note='ATCTA')

species(name='CH3COCH2',
        atoms='H:5 C:3 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.70187196E+00,  5.51653762E-03,  4.27505858E-05,
                    -5.94680816E-08,  2.40685378E-11, -5.92845491E+03,
                    7.12932590E+00]),
                NASA([1000.00, 6000.00],
                    [ 7.54410697E+00,  1.43443222E-02, -5.08381081E-06,
                    8.13200521E-10, -4.83673315E-14, -7.48672286E+03,
                    -1.14792587E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.86,
                                well_depth=435.5,
                                rot_relax=1.0),
        note='A10/04')

species(name='CH3COCH2O2',
        atoms='H:5 C:3 O:3',
        thermo=(NASA([300.00, 1379.00],
                    [ 5.95535468E+00,  2.70255205E-02, -1.37385031E-05,
                    3.53735851E-09, -4.03922557E-13, -2.06679464E+04,
                    5.21436049E+00]),
                NASA([1379.00, 5000.00],
                    [ 1.27690342E+01,  1.42554828E-02, -4.92821461E-06,
                    7.70448921E-10, -4.49110534E-14, -2.34798669E+04,
                    -3.27155799E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.429,
                                well_depth=502.115),
        note='2/14/95')

species(name='C3KET21',
        atoms='H:6 C:3 O:3',
        thermo=(NASA([300.00, 1371.00],
                    [ 4.55686367E+00,  3.57076837E-02, -1.94712054E-05,
                    4.70695431E-09, -3.69753807E-13, -3.86710975E+04,
                    9.97761694E+00]),
                NASA([1371.00, 5000.00],
                    [ 1.56377776E+01,  1.44059342E-02, -5.08808082E-06,
                    8.07076119E-10, -4.75295650E-14, -4.30657975E+04,
                    -5.13105869E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.009,
                                well_depth=464.2,
                                dipole=2.6,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C2H3CHO',
        atoms='H:4 C:3 O:1',
        thermo=(NASA([300.00, 1393.00],
                    [ 2.92355162E-01,  3.54321417E-02, -2.94936324E-05,
                    1.28100124E-08, -2.26144108E-12, -1.16521584E+04,
                    2.28878280E+01]),
                NASA([1393.00, 5000.00],
                    [ 1.04184959E+01,  9.48963321E-03, -3.29310529E-06,
                    5.16279203E-10, -3.01587291E-14, -1.49630281E+04,
                    -3.07235061E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.958,
                                well_depth=428.8,
                                dipole=2.9,
                                rot_relax=1.0),
        note='6/26/95THERM')

species(name='C2H3CO',
        atoms='H:3 C:3 O:1',
        thermo=(NASA([300.00, 1402.00],
                    [ 1.36242013E+00,  3.15273972E-02, -3.00218935E-05,
                    1.48167112E-08, -2.87971530E-12,  4.25770215E+03,
                    1.72626546E+01]),
                NASA([1402.00, 5000.00],
                    [ 9.37467676E+00,  7.91296900E-03, -2.67198280E-06,
                    4.11115430E-10, -2.36978981E-14,  1.92969514E+03,
                    -2.40892696E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.12,
                                well_depth=443.2,
                                rot_relax=1.0),
        note='4/3/0THERM')

species(name='C2H5CHO',
        atoms='H:6 C:3 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.24529681E+00,  6.68296706E-03,  4.93337933E-05,
                    -6.71986124E-08,  2.67262347E-11, -2.41473007E+04,
                    6.90738560E+00]),
                NASA([1000.00, 6000.00],
                    [ 7.44085690E+00,  1.77301764E-02, -6.34081568E-06,
                    1.02040803E-09, -6.09461714E-14, -2.60055814E+04,
                    -1.44195446E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.662,
                                well_depth=435.2,
                                dipole=2.7,
                                rot_relax=1.0),
        note='T05/10')

species(name='C2H5CO',
        atoms='H:5 C:3 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 6.25722402E+00, -9.17612184E-03,  7.61190493E-05,
                    -9.05514997E-08,  3.46198215E-11, -5.91616484E+03,
                    2.23330599E+00]),
                NASA([1000.00, 6000.00],
                    [ 6.52325448E+00,  1.54211952E-02, -5.50898157E-06,
                    8.85889862E-10, -5.28846399E-14, -7.19631634E+03,
                    -5.19862218E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.82,
                                well_depth=424.6,
                                rot_relax=1.0),
        note='A10/04')

species(name='CH3OCH3',
        atoms='H:6 C:2 O:1',
        thermo=(NASA([300.00, 1367.00],
                    [ 2.41860337E+00,  1.87279640E-02, -1.40894592E-06,
                    -4.28367822E-09,  1.36818123E-12, -2.36771735E+04,
                    1.28877967E+01]),
                NASA([1367.00, 5000.00],
                    [ 7.36358679E+00,  1.38910153E-02, -4.74083257E-06,
                    7.34874498E-10, -4.25867578E-14, -2.61148074E+04,
                    -1.61332876E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.037,
                                well_depth=395.0,
                                dipole=1.3,
                                rot_relax=1.0),
        note='1/20/11')

species(name='CH3OCH2',
        atoms='H:5 C:2 O:1',
        thermo=(NASA([300.00, 1389.00],
                    [ 2.19172957E+00,  1.90390966E-02, -6.81450700E-06,
                    8.63577609E-11,  3.23902492E-13, -1.38948253E+03,
                    1.62440756E+01]),
                NASA([1389.00, 5000.00],
                    [ 6.73134670E+00,  1.18022530E-02, -4.00104434E-06,
                    6.17227325E-10, -3.56435023E-14, -3.34933412E+03,
                    -9.40782513E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.037,
                                well_depth=395.0,
                                dipole=1.3,
                                rot_relax=1.0),
        note='1/20/11')

species(name='CH3OCH2O2',
        atoms='H:5 C:2 O:3',
        thermo=(NASA([300.00, 1399.00],
                    [ 1.72101041E+00,  3.78155492E-02, -2.89668870E-05,
                    1.16796117E-08, -1.93234479E-12, -2.05308597E+04,
                    2.19296296E+01]),
                NASA([1399.00, 5000.00],
                    [ 1.22024149E+01,  1.19225104E-02, -4.06624713E-06,
                    6.30079576E-10, -3.65065312E-14, -2.40327175E+04,
                    -3.38665462E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.037,
                                well_depth=395.0,
                                dipole=1.3,
                                rot_relax=1.0),
        note='1/20/11')

species(name='CH2OCH2O2H',
        atoms='H:5 C:2 O:3',
        thermo=(NASA([300.00, 1419.00],
                    [ 9.08332173E-01,  4.47997444E-02, -3.69441578E-05,
                    1.48366604E-08, -2.32279459E-12, -1.50992319E+04,
                    2.66419885E+01]),
                NASA([1419.00, 5000.00],
                    [ 1.51031233E+01,  9.18263324E-03, -3.15895361E-06,
                    4.92789411E-10, -2.87001298E-14, -1.96611854E+04,
                    -4.85172618E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.037,
                                well_depth=395.0,
                                dipole=1.3,
                                rot_relax=1.0),
        note='1/20/11')

species(name='CH3OCH2O2H',
        atoms='H:6 C:2 O:3',
        thermo=(NASA([300.00, 1402.00],
                    [ 7.37369081E-01,  4.67410003E-02, -3.73051936E-05,
                    1.50735309E-08, -2.43781978E-12, -3.73260287E+04,
                    2.57932851E+01]),
                NASA([1402.00, 5000.00],
                    [ 1.46617126E+01,  1.20544674E-02, -4.13604704E-06,
                    6.43721637E-10, -3.74188832E-14, -4.18896186E+04,
                    -4.81293252E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.037,
                                well_depth=395.0,
                                dipole=1.3,
                                rot_relax=1.0),
        note='1/20/11')

species(name='CH3OCH2O',
        atoms='H:5 C:2 O:2',
        thermo=(NASA([300.00, 2012.00],
                    [ 3.25889339E+00,  2.22146359E-02, -7.78556340E-06,
                    -2.41484158E-10,  4.51914496E-13, -1.92377212E+04,
                    1.23680069E+01]),
                NASA([2012.00, 5000.00],
                    [ 8.60261845E+00,  1.35772195E-02, -4.84661602E-06,
                    7.77766193E-10, -4.62633624E-14, -2.13762444E+04,
                    -1.75775023E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.037,
                                well_depth=395.0,
                                dipole=1.3,
                                rot_relax=1.0),
        note='2/9/96THERM')

species(name='O2CH2OCH2O2H',
        atoms='H:5 C:2 O:5',
        thermo=(NASA([300.00, 1411.00],
                    [-1.22579809E-01,  6.64376755E-02, -6.55340609E-05,
                    3.13441786E-08, -5.79042960E-12, -3.41530358E+04,
                    3.49112501E+01]),
                NASA([1411.00, 5000.00],
                    [ 1.96207800E+01,  9.96131702E-03, -3.41480475E-06,
                    5.31364852E-10, -3.08902100E-14, -3.98679641E+04,
                    -6.72721405E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.037,
                                well_depth=395.0,
                                dipole=1.3,
                                rot_relax=1.0),
        note='1/20/11')

species(name='HO2CH2OCHO',
        atoms='H:4 C:2 O:4',
        thermo=(NASA([300.00, 1428.00],
                    [ 5.15432996E-01,  4.76081860E-02, -3.63877925E-05,
                    1.24225089E-08, -1.50813145E-12, -5.82730214E+04,
                    2.81992694E+01]),
                NASA([1428.00, 5000.00],
                    [ 1.78528337E+01,  7.54042388E-03, -2.74895624E-06,
                    4.45343367E-10, -2.66154007E-14, -6.41380252E+04,
                    -6.47357373E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.037,
                                well_depth=395.0,
                                dipole=1.3,
                                rot_relax=1.0),
        note='1/20/11')

species(name='OCH2OCHO',
        atoms='H:3 C:2 O:3',
        thermo=(NASA([300.00, 1426.00],
                    [ 3.92797655E+00,  1.85731220E-02, -4.14656828E-07,
                    -6.94060480E-09,  2.32311215E-12, -4.07070844E+04,
                    1.02048630E+01]),
                NASA([1426.00, 5000.00],
                    [ 1.24378042E+01,  7.82259106E-03, -2.82874305E-06,
                    4.55808935E-10, -2.71391601E-14, -4.44526361E+04,
                    -3.85337353E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.037,
                                well_depth=395.0,
                                dipole=1.3,
                                rot_relax=1.0),
        note='1/20/11')

species(name='HOCH2OCO',
        atoms='H:3 C:2 O:3',
        thermo=(NASA([300.00, 1603.00],
                    [ 6.08180801E+00,  1.28768359E-02,  2.04419418E-06,
                    -6.10154921E-09,  1.79820559E-12, -4.39526183E+04,
                    2.54054449E+00]),
                NASA([1603.00, 5000.00],
                    [ 1.13737391E+01,  8.17663898E-03, -2.92034021E-06,
                    4.66695616E-10, -2.76276823E-14, -4.65575743E+04,
                    -2.86035265E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.037,
                                well_depth=395.0,
                                dipole=1.3,
                                rot_relax=1.0),
        note='8/31/99THERM')

species(name='CH3OCHO',
        atoms='H:4 C:2 O:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 5.96757028E+00, -9.38085425E-03,  7.07648417E-05,
                    -8.29932227E-08,  3.13522917E-11, -4.55713267E+04,
                    7.50341113E-01]),
                NASA([1000.00, 6000.00],
                    [ 6.33360880E+00,  1.34851485E-02, -4.84305805E-06,
                    7.81719241E-10, -4.67917447E-14, -4.68316521E+04,
                    -6.91542601E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.037,
                                well_depth=395.0,
                                dipole=1.3,
                                rot_relax=1.0),
        note='T6/08')

species(name='CH3OCO',
        atoms='H:3 C:2 O:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.75563598E+00,  7.80915313E-03,  1.62272935E-05,
                    -2.41210787E-08,  9.42644561E-12, -2.15157456E+04,
                    4.78096491E+00]),
                NASA([1000.00, 6000.00],
                    [ 7.00171955E+00,  1.01977290E-02, -3.65621800E-06,
                    5.89475086E-10, -3.52561321E-14, -2.26135780E+04,
                    -9.05267669E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.037,
                                well_depth=395.0,
                                dipole=1.3,
                                rot_relax=1.0),
        note='T10/07')

species(name='CH2OCHO',
        atoms='H:3 C:2 O:2',
        thermo=(NASA([300.00, 1442.00],
                    [ 2.31031671E+00,  1.80474065E-02, -2.71519637E-06,
                    -4.60918579E-09,  1.70037078E-12, -2.02910878E+04,
                    1.71549722E+01]),
                NASA([1442.00, 5000.00],
                    [ 1.00960096E+01,  7.19887066E-03, -2.59813465E-06,
                    4.18110812E-10, -2.48723387E-14, -2.36389018E+04,
                    -2.71144175E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.037,
                                well_depth=395.0,
                                dipole=1.3,
                                rot_relax=1.0),
        note='4/15/8THERM')

species(name='HE',
        atoms='He:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 2.50000000E+00,  0.00000000E+00,  0.00000000E+00,
                    0.00000000E+00,  0.00000000E+00, -7.45375000E+02,
                    9.28723974E-01]),
                NASA([1000.00, 6000.00],
                    [ 2.50000000E+00,  0.00000000E+00,  0.00000000E+00,
                    0.00000000E+00,  0.00000000E+00, -7.45375000E+02,
                    9.28723974E-01])),
        transport=gas_transport(geom='atom',
                                diam=2.576,
                                well_depth=10.2),
        note='G5/97')

species(name='C3H8',
        atoms='H:8 C:3',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.21093013E+00,  1.70886504E-03,  7.06530164E-05,
                    -9.20060565E-08,  3.64618453E-11, -1.43810883E+04,
                    5.61004451E+00]),
                NASA([1000.00, 6000.00],
                    [ 6.66919760E+00,  2.06108751E-02, -7.36512349E-06,
                    1.18434262E-09, -7.06914630E-14, -1.62754066E+04,
                    -1.31943379E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.81,
                                well_depth=303.4,
                                rot_relax=1.0),
        note='G2/00')

species(name='IC3H7',
        atoms='H:7 C:3',
        thermo=(NASA([200.00, 1000.00],
                    [ 5.47421257E+00, -8.42536682E-03,  8.04607759E-05,
                    -9.49287824E-08,  3.59830971E-11,  9.04939013E+03,
                    3.40542323E+00]),
                NASA([1000.00, 6000.00],
                    [ 5.30597255E+00,  1.89854588E-02, -6.74315384E-06,
                    1.07993730E-09, -6.42785036E-14,  7.78748910E+03,
                    -2.23233935E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.81,
                                well_depth=303.4,
                                rot_relax=1.0),
        note='A5/05')

species(name='NC3H7',
        atoms='H:7 C:3',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.08211458E+00,  5.23240341E-03,  5.13554466E-05,
                    -6.99343598E-08,  2.81819493E-11,  1.04074558E+04,
                    8.39534919E+00]),
                NASA([1000.00, 6000.00],
                    [ 6.49636579E+00,  1.77337992E-02, -6.24898046E-06,
                    9.95389495E-10, -5.90199770E-14,  8.85973885E+03,
                    -8.56389710E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.81,
                                well_depth=303.4,
                                rot_relax=1.0),
        note='A5/05')

species(name='C3H6',
        atoms='H:6 C:3',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.83464468E+00,  3.29078952E-03,  5.05228001E-05,
                    -6.66251176E-08,  2.63707473E-11,  7.88717123E+02,
                    7.53408013E+00]),
                NASA([1000.00, 6000.00],
                    [ 6.03870234E+00,  1.62963931E-02, -5.82130800E-06,
                    9.35936829E-10, -5.58603143E-14, -7.41715057E+02,
                    -8.43825992E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.14,
                                well_depth=307.8,
                                rot_relax=1.0),
        note='G2/00')

species(name='C3H5-A',
        atoms='H:5 C:3',
        thermo=(NASA([300.00, 1404.00],
                    [-4.35181829E-01,  3.40776627E-02, -2.66098022E-05,
                    1.09853986E-08, -1.85107479E-12,  1.94181924E+04,
                    2.45509858E+01]),
                NASA([1404.00, 5000.00],
                    [ 8.71682542E+00,  1.08257699E-02, -3.63655884E-06,
                    5.57706282E-10, -3.20793321E-14,  1.64375315E+04,
                    -2.39184681E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.22,
                                well_depth=316.0,
                                rot_relax=1.0),
        note='PD5/98')

species(name='C3H5-S',
        atoms='H:5 C:3',
        thermo=(NASA([300.00, 1394.00],
                    [ 5.27650639E-01,  2.95749512E-02, -2.05159708E-05,
                    7.69163477E-09, -1.21882509E-12,  3.09544881E+04,
                    2.15637313E+01]),
                NASA([1394.00, 5000.00],
                    [ 8.30447254E+00,  1.13525433E-02, -3.85583517E-06,
                    5.95730442E-10, -3.44434061E-14,  2.82273490E+04,
                    -2.02359952E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.22,
                                well_depth=316.0,
                                rot_relax=1.0),
        note='PD5/98')

species(name='C3H5-T',
        atoms='H:5 C:3',
        thermo=(NASA([300.00, 1389.00],
                    [ 1.12339597E+00,  2.68562417E-02, -1.65102273E-05,
                    5.28890293E-09, -7.08002939E-13,  2.91053720E+04,
                    1.91241364E+01]),
                NASA([1389.00, 5000.00],
                    [ 8.25977858E+00,  1.14281582E-02, -3.89051874E-06,
                    6.02061188E-10, -3.48497868E-14,  2.64682815E+04,
                    -1.97010509E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.22,
                                well_depth=316.0,
                                rot_relax=1.0),
        note='PD5/98')

species(name='C3H4-P',
        atoms='H:4 C:3',
        thermo=(NASA([200.00, 1000.00],
                    [ 2.68038690E+00,  1.57996510E-02,  2.50705960E-06,
                    -1.36576230E-08,  6.61542850E-12,  2.08023740E+04,
                    9.87693510E+00]),
                NASA([1000.00, 6000.00],
                    [ 6.02524000E+00,  1.13365420E-02, -4.02233910E-06,
                    6.43760630E-10, -3.82996350E-14,  1.96209420E+04,
                    -8.60437850E+00])),
        transport=gas_transport(geom='linear',
                                diam=4.29,
                                well_depth=324.8,
                                rot_relax=1.0),
        note='T2/90')

species(name='C3H4-A',
        atoms='H:4 C:3',
        thermo=(NASA([200.00, 1000.00],
                    [ 2.61304450E+00,  1.21225750E-02,  1.85398800E-05,
                    -3.45251490E-08,  1.53350790E-11,  2.15415670E+04,
                    1.02261390E+01]),
                NASA([1000.00, 6000.00],
                    [ 6.31687220E+00,  1.11337280E-02, -3.96293780E-06,
                    6.35642380E-10, -3.78755400E-14,  2.01174950E+04,
                    -1.09957660E+01])),
        transport=gas_transport(geom='linear',
                                diam=4.29,
                                well_depth=324.8,
                                rot_relax=1.0),
        note='L8/89')

species(name='C3H3',
        atoms='H:3 C:3',
        thermo=(NASA([200.00, 1000.00],
                    [ 1.35110873E+00,  3.27411291E-02, -4.73827407E-05,
                    3.76310220E-08, -1.18541128E-11,  4.07679941E+04,
                    1.52058598E+01]),
                NASA([1000.00, 6000.00],
                    [ 7.14221719E+00,  7.61902211E-03, -2.67460030E-06,
                    4.24914904E-10, -2.51475443E-14,  3.95709594E+04,
                    -1.25848690E+01])),
        transport=gas_transport(geom='linear',
                                diam=4.29,
                                well_depth=324.8,
                                rot_relax=1.0),
        note='T7/11')

species(name='C3H5O',
        atoms='H:5 C:3 O:1',
        thermo=(NASA([300.00, 1380.00],
                    [ 1.19822582E+00,  3.05579837E-02, -1.80630276E-05,
                    4.86150033E-09, -4.19854562E-13,  9.58217784E+03,
                    2.15566221E+01]),
                NASA([1380.00, 5000.00],
                    [ 1.02551752E+01,  1.14983720E-02, -3.84645659E-06,
                    5.88910346E-10, -3.38557923E-14,  6.26560810E+03,
                    -2.77655042E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.82,
                                well_depth=411.0,
                                rot_relax=1.0),
        note='7/20/95THERM')

species(name='C3H6OOH1-2',
        atoms='H:7 C:3 O:2',
        thermo=(NASA([300.00, 1402.00],
                    [ 2.87774562E+00,  3.74166999E-02, -2.36058063E-05,
                    7.79930860E-09, -1.06042562E-12, -7.82368119E+02,
                    1.93941340E+01]),
                NASA([1402.00, 5000.00],
                    [ 1.24605763E+01,  1.58889526E-02, -5.32742106E-06,
                    8.15818791E-10, -4.68723250E-14, -4.20305196E+03,
                    -3.23858947E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.662,
                                well_depth=435.2,
                                dipole=2.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C3H6OOH1-3',
        atoms='H:7 C:3 O:2',
        thermo=(NASA([300.00, 1384.00],
                    [ 1.88331465E+00,  4.40156051E-02, -3.07858462E-05,
                    1.13214862E-08, -1.75323184E-12, -1.68779168E+02,
                    2.14040621E+01]),
                NASA([1384.00, 5000.00],
                    [ 1.46881564E+01,  1.49941200E-02, -5.24056505E-06,
                    8.25462711E-10, -4.83758952E-14, -4.77342863E+03,
                    -4.77984492E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.662,
                                well_depth=435.2,
                                dipole=2.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C3H6OOH2-1',
        atoms='H:7 C:3 O:2',
        thermo=(NASA([300.00, 1407.00],
                    [ 2.09193950E+00,  4.69220394E-02, -3.90280831E-05,
                    1.72381453E-08, -3.07968979E-12, -1.89377918E+03,
                    2.00178282E+01]),
                NASA([1407.00, 5000.00],
                    [ 1.42163221E+01,  1.43382450E-02, -4.78004477E-06,
                    7.29133134E-10, -4.17761973E-14, -5.67381620E+03,
                    -4.35770997E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.662,
                                well_depth=435.2,
                                dipole=2.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C3H6OOH1-2O2',
        atoms='H:7 C:3 O:4',
        thermo=(NASA([300.00, 1386.00],
                    [ 2.65196584E+00,  5.74638149E-02, -4.72190867E-05,
                    2.05591557E-08, -3.68787387E-12, -2.08829371E+04,
                    2.01547955E+01]),
                NASA([1386.00, 5000.00],
                    [ 1.91759159E+01,  1.59857013E-02, -5.61306378E-06,
                    8.86880495E-10, -5.20877040E-14, -2.64412115E+04,
                    -6.77512936E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.662,
                                well_depth=435.2,
                                dipole=2.7,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C3H6OOH1-3O2',
        atoms='H:7 C:3 O:4',
        thermo=(NASA([300.00, 1382.00],
                    [ 3.14864588E+00,  5.33542571E-02, -4.08330611E-05,
                    1.67542220E-08, -2.89288326E-12, -1.85473645E+04,
                    1.86305431E+01]),
                NASA([1382.00, 5000.00],
                    [ 1.85916698E+01,  1.65328553E-02, -5.81343626E-06,
                    9.19396939E-10, -5.40320702E-14, -2.39598698E+04,
                    -6.42544402E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.662,
                                well_depth=435.2,
                                dipole=2.7,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C3H6OOH2-1O2',
        atoms='H:7 C:3 O:4',
        thermo=(NASA([300.00, 1386.00],
                    [ 2.65196584E+00,  5.74638149E-02, -4.72190867E-05,
                    2.05591557E-08, -3.68787387E-12, -2.08829371E+04,
                    2.01547955E+01]),
                NASA([1386.00, 5000.00],
                    [ 1.91759159E+01,  1.59857013E-02, -5.61306378E-06,
                    8.86880495E-10, -5.20877040E-14, -2.64412115E+04,
                    -6.77512936E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.662,
                                well_depth=435.2,
                                dipole=2.7,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C3H6OOH2-2',
        atoms='H:7 C:3 O:2',
        thermo=(NASA([300.00, 1407.00],
                    [ 2.09193950E+00,  4.69220394E-02, -3.90280831E-05,
                    1.72381453E-08, -3.07968979E-12, -1.89377918E+03,
                    2.00178282E+01]),
                NASA([1407.00, 5000.00],
                    [ 1.42163221E+01,  1.43382450E-02, -4.78004477E-06,
                    7.29133134E-10, -4.17761973E-14, -5.67381620E+03,
                    -4.35770997E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.352,
                                well_depth=436.4,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC3H7O2H',
        atoms='H:8 C:3 O:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 7.08977756E+00, -4.86163264E-03,  1.03253531E-04,
                    -1.33200956E-07,  5.30799252E-11, -2.44194556E+04,
                    1.74859215E+00]),
                NASA([1000.00, 6000.00],
                    [ 1.04115631E+01,  2.13763889E-02, -7.55819870E-06,
                    1.20207180E-09, -7.11798090E-14, -2.68935445E+04,
                    -2.35428789E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.997,
                                well_depth=481.5,
                                dipole=1.7,
                                rot_relax=1.0),
        note='T02/10')

species(name='IC3H7O2H',
        atoms='H:8 C:3 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 5.19265570E-01,  5.32111228E-02, -4.05156892E-05,
                    1.63346713E-08, -2.73751233E-12, -2.71048486E+04,
                    2.40815065E+01]),
                NASA([1389.00, 5000.00],
                    [ 1.57046391E+01,  1.65924516E-02, -5.77250934E-06,
                    9.06453070E-10, -5.30084111E-14, -3.23268564E+04,
                    -5.71735156E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.036,
                                well_depth=459.5,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC3H7O2',
        atoms='H:7 C:3 O:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 6.99488957E+00, -7.50355088E-03,  1.01250579E-04,
                    -1.25012534E-07,  4.84675978E-11, -7.53004240E+03,
                    -2.03740242E+00]),
                NASA([1000.00, 6000.00],
                    [ 9.24489759E+00,  2.14800809E-02, -7.69335773E-06,
                    1.23945160E-09, -7.40911693E-14, -9.81187945E+03,
                    -2.20401438E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.997,
                                well_depth=481.5,
                                dipole=1.7,
                                rot_relax=1.0),
        note='T04/10')

species(name='IC3H7O2',
        atoms='H:7 C:3 O:2',
        thermo=(NASA([300.00, 1388.00],
                    [ 1.49941639E+00,  4.43081205E-02, -3.22414456E-05,
                    1.29687136E-08, -2.23370569E-12, -1.02587980E+04,
                    2.02336490E+01]),
                NASA([1388.00, 5000.00],
                    [ 1.32493493E+01,  1.64082190E-02, -5.67432062E-06,
                    8.87336340E-10, -5.17361535E-14, -1.44109855E+04,
                    -4.29066213E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.036,
                                well_depth=459.5,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC3H7O',
        atoms='H:7 C:3 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.21934640E+00,  7.38556641E-03,  6.02825529E-05,
                    -8.38680247E-08,  3.39623435E-11, -6.23491852E+03,
                    8.08139850E+00]),
                NASA([1000.00, 6000.00],
                    [ 8.38041157E+00,  1.95206120E-02, -6.97374143E-06,
                    1.12144919E-09, -6.69467831E-14, -8.48625211E+03,
                    -1.89916219E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.997,
                                well_depth=481.5,
                                dipole=1.7,
                                rot_relax=1.0),
        note='T07/10')

species(name='IC3H7O',
        atoms='H:7 C:3 O:1',
        thermo=(NASA([300.00, 1393.00],
                    [-8.18260771E-01,  4.68807147E-02, -3.74369377E-05,
                    1.55917264E-08, -2.64308171E-12, -8.15153216E+03,
                    2.79483196E+01]),
                NASA([1393.00, 5000.00],
                    [ 1.23135031E+01,  1.38062606E-02, -4.91585531E-06,
                    8.21075632E-10, -5.07493001E-14, -1.24717023E+04,
                    -4.17270532E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.036,
                                well_depth=459.5,
                                dipole=1.7,
                                rot_relax=1.0),
        note='8/9/4THERM')

species(name='C3H6O1-2',
        atoms='H:6 C:3 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.42806676E+00,  6.25176642E-03,  6.13196311E-05,
                    -8.60387185E-08,  3.51371393E-11, -1.28446646E+04,
                    1.04244994E+01]),
                NASA([1000.00, 6000.00],
                    [ 8.01491079E+00,  1.73919953E-02, -6.26027968E-06,
                    1.01188256E-09, -6.06239111E-14, -1.51980838E+04,
                    -1.88279964E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.968,
                                well_depth=403.6,
                                dipole=2.0,
                                rot_relax=1.0),
        note='A01/05')

species(name='C3H6O1-3',
        atoms='H:6 C:3 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 5.15283752E+00, -1.86401716E-02,  1.29980652E-04,
                    -1.58629974E-07,  6.20668783E-11, -1.13243512E+04,
                    4.73561224E+00]),
                NASA([1000.00, 6000.00],
                    [ 6.80716906E+00,  1.88824545E-02, -6.79082475E-06,
                    1.09713919E-09, -6.57154952E-14, -1.36547629E+04,
                    -1.35382154E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.968,
                                well_depth=403.6,
                                dipole=2.0,
                                rot_relax=1.0),
        note='A11/04')

species(name='C3KET12',
        atoms='H:6 C:3 O:3',
        thermo=(NASA([300.00, 1388.00],
                    [ 1.10507238E+00,  5.27396706E-02, -4.31805774E-05,
                    1.81618292E-08, -3.10535568E-12, -3.63627536E+04,
                    2.54111636E+01]),
                NASA([1388.00, 5000.00],
                    [ 1.70756225E+01,  1.31013491E-02, -4.61949408E-06,
                    7.31991329E-10, -4.30788679E-14, -4.17008637E+04,
                    -5.95778952E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.009,
                                well_depth=464.2,
                                dipole=2.6,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C3KET13',
        atoms='H:6 C:3 O:3',
        thermo=(NASA([300.00, 1379.00],
                    [ 3.55241022E+00,  4.18720270E-02, -2.94550370E-05,
                    1.09982900E-08, -1.75045977E-12, -3.48902671E+04,
                    1.42082894E+01]),
                NASA([1379.00, 5000.00],
                    [ 1.58927479E+01,  1.40990923E-02, -4.96118851E-06,
                    7.84992198E-10, -4.61488928E-14, -3.93774829E+04,
                    -5.26049341E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.009,
                                well_depth=464.2,
                                dipole=2.6,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C3H51-2,3OOH',
        atoms='H:7 C:3 O:4',
        thermo=(NASA([300.00, 1386.00],
                    [ 2.55619708E+00,  6.13504487E-02, -5.23205391E-05,
                    2.28208029E-08, -4.02231508E-12, -1.31353414E+04,
                    2.21043799E+01]),
                NASA([1386.00, 5000.00],
                    [ 2.12378169E+01,  1.39519596E-02, -4.94539222E-06,
                    7.86381389E-10, -4.63925564E-14, -1.92864584E+04,
                    -7.69636561E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.716,
                                well_depth=549.655),
        note='8/26/3THRM')

species(name='C3H52-1,3OOH',
        atoms='H:7 C:3 O:4',
        thermo=(NASA([300.00, 1379.00],
                    [ 4.12253742E+00,  5.19553611E-02, -3.83733727E-05,
                    1.45851637E-08, -2.29820536E-12, -1.22759164E+04,
                    1.48367359E+01]),
                NASA([1379.00, 5000.00],
                    [ 2.02817964E+01,  1.48155431E-02, -5.25503386E-06,
                    8.35963453E-10, -4.93308915E-14, -1.80085066E+04,
                    -7.22688262E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.716,
                                well_depth=549.655),
        note='8/26/3THRM')

species(name='C3H6OH',
        atoms='H:7 C:3 O:1',
        thermo=(NASA([300.00, 1674.00],
                    [ 1.20494302E+00,  3.30857885E-02, -1.63893637E-05,
                    3.18103918E-09, -6.84229288E-14, -9.12961720E+03,
                    2.47155202E+01]),
                NASA([1674.00, 5000.00],
                    [ 9.31287816E+00,  1.67579212E-02, -5.75555480E-06,
                    9.00584362E-10, -5.26566836E-14, -1.20169635E+04,
                    -1.95011064E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.82,
                                well_depth=487.9,
                                rot_relax=1.0),
        note='1/7/98THERM')

species(name='HOC3H6O2',
        atoms='H:7 C:3 O:3',
        thermo=(NASA([300.00, 1397.00],
                    [ 3.06954289E+00,  4.43105640E-02, -3.18636260E-05,
                    1.26409287E-08, -2.12751901E-12, -2.75893676E+04,
                    1.83891218E+01]),
                NASA([1397.00, 5000.00],
                    [ 1.42691004E+01,  1.71837825E-02, -5.83419536E-06,
                    9.00994191E-10, -5.20716210E-14, -3.14501750E+04,
                    -4.15295035E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.82,
                                well_depth=487.9,
                                rot_relax=1.0),
        note='1/7/98THERM')

species(name='CH3CHCO',
        atoms='H:4 C:3 O:1',
        thermo=(NASA([300.00, 1400.00],
                    [ 1.48380119E+00,  3.22203013E-02, -2.70250033E-05,
                    1.20499164E-08, -2.18365931E-12, -1.15276540E+04,
                    1.71552068E+01]),
                NASA([1400.00, 5000.00],
                    [ 1.00219123E+01,  9.56966300E-03, -3.26221644E-06,
                    5.05231706E-10, -2.92593257E-14, -1.42482738E+04,
                    -2.77829973E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.12,
                                well_depth=443.2,
                                rot_relax=1.0),
        note='03/03/95THERM')

species(name='AC3H5OOH',
        atoms='H:6 C:3 O:2',
        thermo=(NASA([300.00, 1394.00],
                    [ 2.43934647E+00,  4.02070638E-02, -2.95322679E-05,
                    1.14715600E-08, -1.85170511E-12, -9.43679906E+03,
                    1.70549969E+01]),
                NASA([1394.00, 5000.00],
                    [ 1.36837693E+01,  1.33968049E-02, -4.61533631E-06,
                    7.19988958E-10, -4.19109988E-14, -1.33165248E+04,
                    -4.31904193E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.997,
                                well_depth=481.5,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/20/98THERM')

species(name='C2H3OOH',
        atoms='H:4 C:2 O:2',
        thermo=(NASA([300.00, 1397.00],
                    [ 1.35644398E+00,  3.37002447E-02, -2.75988500E-05,
                    1.14222854E-08, -1.89488886E-12, -5.49996692E+03,
                    1.98354466E+01]),
                NASA([1397.00, 5000.00],
                    [ 1.15749951E+01,  8.09909174E-03, -2.81808668E-06,
                    4.42697954E-10, -2.58998042E-14, -8.84852664E+03,
                    -3.43859117E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='4/18/8THERM')

species(name='CC3H4',
        atoms='H:4 C:3',
        thermo=(NASA([300.00, 1000.00],
                    [-2.46210470E-02,  2.31972150E-02, -1.84743570E-06,
                    -1.59275930E-08,  8.68461550E-12,  3.23341370E+04,
                    2.27297620E+01]),
                NASA([1000.00, 5000.00],
                    [ 6.69999310E+00,  1.03573720E-02, -3.45511670E-06,
                    5.06529490E-10, -2.66822760E-14,  3.01990510E+04,
                    -1.33787700E+01])),
        transport=gas_transport(geom='linear',
                                diam=4.76,
                                well_depth=252.0,
                                rot_relax=1.0),
        note='T12/81')

species(name='H2CC',
        atoms='H:2 C:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.28154830E+00,  6.97647910E-03, -2.38552440E-06,
                    -1.21044320E-09,  9.81895450E-13,  4.86217940E+04,
                    5.92039100E+00]),
                NASA([1000.00, 6000.00],
                    [ 4.27803400E+00,  4.75628040E-03, -1.63010090E-06,
                    2.54628060E-10, -1.48863790E-14,  4.83166880E+04,
                    6.40237010E-01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.1,
                                well_depth=209.0,
                                rot_relax=2.5),
        note='L12/89')

species(name='C4H10',
        atoms='H:10 C:4',
        thermo=(NASA([300.00, 1392.00],
                    [-4.55756824E-01,  4.80323389E-02, -2.65497552E-05,
                    6.92544700E-09, -6.38317504E-13, -1.68960904E+04,
                    2.64870966E+01]),
                NASA([1392.00, 5000.00],
                    [ 1.24940183E+01,  2.17726258E-02, -7.44272215E-06,
                    1.15487023E-09, -6.69712949E-14, -2.18403437E+04,
                    -4.45558921E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.206,
                                well_depth=350.9,
                                rot_relax=1.0),
        note='8/4/4THERM')

species(name='C4H8-1',
        atoms='H:8 C:4',
        thermo=(NASA([200.00, 1000.00],
                    [ 5.13226136E+00,  5.33862838E-03,  6.02928960E-05,
                    -7.60364685E-08,  2.87324693E-11, -2.16718358E+03,
                    3.82936810E+00]),
                NASA([1000.00, 6000.00],
                    [ 7.86795262E+00,  2.24448843E-02, -8.07705438E-06,
                    1.30179988E-09, -7.77958472E-14, -4.23853340E+03,
                    -1.65662549E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.65,
                                well_depth=355.0,
                                rot_relax=1.0),
        note='T05/09')

species(name='C4H8-2',
        atoms='H:8 C:4',
        thermo=(NASA([200.00, 1000.00],
                    [ 5.57278967E+00,  3.76541017E-03,  6.52226708E-05,
                    -8.30909522E-08,  3.20311342E-11, -3.60128327E+03,
                    5.37796708E-01]),
                NASA([1000.00, 6000.00],
                    [ 7.89114667E+00,  2.24970532E-02, -8.12143779E-06,
                    1.31273568E-09, -7.84451632E-14, -5.51643171E+03,
                    -1.76436027E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.65,
                                well_depth=355.0,
                                rot_relax=1.0),
        note='T5/09')

species(name='PC4H9',
        atoms='H:9 C:4',
        thermo=(NASA([300.00, 1391.00],
                    [ 3.20730933E-01,  4.34654454E-02, -2.40584970E-05,
                    6.28245308E-09, -5.80113166E-13,  7.71490893E+03,
                    2.57301085E+01]),
                NASA([1391.00, 5000.00],
                    [ 1.20779744E+01,  1.96264778E-02, -6.71302199E-06,
                    1.04206424E-09, -6.04469282E-14,  3.22550473E+03,
                    -3.87719384E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.24,
                                well_depth=352.0,
                                rot_relax=1.0),
        note='8/4/4THERM')

species(name='SC4H9',
        atoms='H:9 C:4',
        thermo=(NASA([300.00, 1381.00],
                    [ 8.49159986E-01,  3.82085320E-02, -1.49626797E-05,
                    2.04499211E-10,  8.24254437E-13,  6.38832956E+03,
                    2.44466606E+01]),
                NASA([1381.00, 5000.00],
                    [ 1.16934304E+01,  1.96402287E-02, -6.65306517E-06,
                    1.02631895E-09, -5.92826294E-14,  1.96382429E+03,
                    -3.61626672E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.24,
                                well_depth=352.0,
                                rot_relax=1.0),
        note='8/4/4THERM')

species(name='C4H71-1',
        atoms='H:7 C:4',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.19857522E+00,  1.19616999E-02,  4.23864923E-05,
                    -6.30299109E-08,  2.59475110E-11,  2.75256555E+04,
                    8.57181248E+00]),
                NASA([1000.00, 6000.00],
                    [ 8.15646382E+00,  1.90308835E-02, -6.73262214E-06,
                    1.07333098E-09, -6.36886441E-14,  2.55826427E+04,
                    -1.61428872E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.72,
                                well_depth=357.1,
                                rot_relax=1.0),
        note='T05/04')

species(name='C4H71-2',
        atoms='H:7 C:4',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.77145965E+00,  1.46544157E-02,  3.70080802E-05,
                    -5.72714455E-08,  2.36641011E-11,  2.58014506E+04,
                    9.11906641E+00]),
                NASA([1000.00, 6000.00],
                    [ 8.16688868E+00,  1.95680375E-02, -6.95694878E-06,
                    1.11504166E-09, -6.64079384E-14,  2.37537003E+04,
                    -1.77041242E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.72,
                                well_depth=357.1,
                                rot_relax=1.0),
        note='T05/04')

species(name='C4H71-3',
        atoms='H:7 C:4',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.54746808E+00,  4.63771460E-03,  6.61340221E-05,
                    -8.97456502E-08,  3.61716165E-11,  1.43843217E+04,
                    7.30313471E+00]),
                NASA([1000.00, 6000.00],
                    [ 8.08107449E+00,  1.95526544E-02, -6.93149115E-06,
                    1.10889183E-09, -6.59584410E-14,  1.22822959E+04,
                    -1.67137903E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.72,
                                well_depth=357.1,
                                rot_relax=1.0),
        note='T05/04')

species(name='C4H71-4',
        atoms='H:7 C:4',
        thermo=(NASA([200.00, 1000.00],
                    [ 5.07355313E+00,  5.27619329E-03,  6.23441322E-05,
                    -8.54203458E-08,  3.45890031E-11,  2.24615054E+04,
                    5.60318035E+00]),
                NASA([1000.00, 6000.00],
                    [ 8.49073768E+00,  1.91056974E-02, -6.74370664E-06,
                    1.07343267E-09, -6.36251837E-14,  2.04659294E+04,
                    -1.74555814E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.72,
                                well_depth=357.1,
                                rot_relax=1.0),
        note='T05/04')

species(name='C4H72-2',
        atoms='H:7 C:4',
        thermo=(NASA([200.00, 1000.00],
                    [ 7.61389036E+00, -9.06922602E-03,  8.28486476E-05,
                    -9.61203624E-08,  3.59333528E-11,  2.44971584E+04,
                    -5.90519467E+00]),
                NASA([1000.00, 6000.00],
                    [ 7.26612168E+00,  1.99858497E-02, -7.12030976E-06,
                    1.14276142E-09, -6.81206632E-14,  2.31915554E+04,
                    -1.09941637E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.72,
                                well_depth=357.1,
                                rot_relax=1.0),
        note='T05/04')

species(name='C4H6',
        atoms='H:6 C:4',
        thermo=(NASA([300.00, 1395.00],
                    [-7.89604392E-01,  4.13361127E-02, -2.97167286E-05,
                    1.13512399E-08, -1.80261184E-12,  1.18902568E+04,
                    2.68208195E+01]),
                NASA([1395.00, 5000.00],
                    [ 1.04518247E+01,  1.45471475E-02, -4.94228994E-06,
                    7.63833067E-10, -4.41755383E-14,  8.02177132E+03,
                    -3.33939064E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.72,
                                well_depth=357.0,
                                rot_relax=1.0),
        note='H6W/94')

species(name='PC4H9O2H',
        atoms='H:10 C:4 O:2',
        thermo=(NASA([300.00, 1387.00],
                    [ 6.34644286E-01,  6.20007657E-02, -4.34477749E-05,
                    1.61407616E-08, -2.53121507E-12, -2.76345805E+04,
                    2.67026090E+01]),
                NASA([1387.00, 5000.00],
                    [ 1.82687454E+01,  2.16940079E-02, -7.54629828E-06,
                    1.18484806E-09, -6.92818498E-14, -3.39441633E+04,
                    -6.84816300E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='SC4H9O2H',
        atoms='H:10 C:4 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [ 3.40587657E-01,  6.52078069E-02, -4.84840460E-05,
                    1.91011830E-08, -3.13833073E-12, -3.00007960E+04,
                    2.72835451E+01]),
                NASA([1390.00, 5000.00],
                    [ 1.88587916E+01,  2.11200987E-02, -7.33205568E-06,
                    1.14969842E-09, -6.71656991E-14, -3.64236058E+04,
                    -7.20026458E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='PC4H9O2',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1385.00],
                    [ 1.94363650E+00,  5.15513163E-02, -3.28284400E-05,
                    1.13064860E-08, -1.70118606E-12, -1.08358103E+04,
                    2.13503149E+01]),
                NASA([1385.00, 5000.00],
                    [ 1.57845448E+01,  2.15210910E-02, -7.44909017E-06,
                    1.16558071E-09, -6.79885609E-14, -1.60146054E+04,
                    -5.40388525E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='SC4H9O2',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 1.32689044E+00,  5.62785583E-02, -4.01717786E-05,
                    1.57120967E-08, -2.62948443E-12, -1.31557057E+04,
                    2.34069659E+01]),
                NASA([1389.00, 5000.00],
                    [ 1.64031135E+01,  2.09361006E-02, -7.23393011E-06,
                    1.13058996E-09, -6.58938667E-14, -1.85074517E+04,
                    -5.77331636E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='PC4H9O',
        atoms='H:9 C:4 O:1',
        thermo=(NASA([300.00, 1403.00],
                    [-4.99964924E-01,  5.37157310E-02, -3.44426650E-05,
                    1.08145957E-08, -1.29600044E-12, -9.11644218E+03,
                    3.09183423E+01]),
                NASA([1403.00, 5000.00],
                    [ 1.49315588E+01,  1.95927054E-02, -6.66958265E-06,
                    1.03222635E-09, -5.97583630E-14, -1.46178979E+04,
                    -5.25561919E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='8/9/4THERM')

species(name='SC4H9O',
        atoms='H:9 C:4 O:1',
        thermo=(NASA([300.00, 1679.00],
                    [-4.09000900E-01,  5.38573427E-02, -3.39105695E-05,
                    1.01023834E-08, -1.11268025E-12, -1.11109627E+04,
                    2.88554742E+01]),
                NASA([1679.00, 5000.00],
                    [ 1.43323395E+01,  2.04542365E-02, -7.12271896E-06,
                    1.12545447E-09, -6.62697853E-14, -1.60006806E+04,
                    -5.02030895E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='8/9/4THERM')

species(name='C4H7O',
        atoms='H:7 C:4 O:1',
        thermo=(NASA([300.00, 1395.00],
                    [-1.60619192E+00,  5.58562682E-02, -4.35595767E-05,
                    1.70589279E-08, -2.65635180E-12,  4.85090326E+03,
                    3.47112559E+01]),
                NASA([1395.00, 5000.00],
                    [ 1.53137780E+01,  1.43427017E-02, -4.81625517E-06,
                    7.39574839E-10, -4.26140814E-14, -7.29342884E+02,
                    -5.52937859E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='4/3/0THERM')

species(name='C4H8O1-2',
        atoms='H:8 C:4 O:1',
        thermo=(NASA([300.00, 1399.00],
                    [-2.42033073E+00,  5.79308508E-02, -4.30236499E-05,
                    1.66679793E-08, -2.64714350E-12, -1.53950880E+04,
                    3.66425242E+01]),
                NASA([1399.00, 5000.00],
                    [ 1.39197290E+01,  1.85551492E-02, -6.36014179E-06,
                    9.88844645E-10, -5.74274779E-14, -2.09452548E+04,
                    -5.06788231E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='4/3/0THERM')

species(name='C4H8O1-3',
        atoms='H:8 C:4 O:1',
        thermo=(NASA([300.00, 1371.00],
                    [-2.53690104E+00,  5.43995707E-02, -3.43390305E-05,
                    1.01079922E-08, -1.10262736E-12, -1.52980680E+04,
                    3.67400719E+01]),
                NASA([1371.00, 5000.00],
                    [ 1.54227092E+01,  1.70211052E-02, -6.06347951E-06,
                    9.67354762E-10, -5.71992419E-14, -2.20194174E+04,
                    -6.13871862E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='1/22/95THERM')

species(name='C4H8O1-4',
        atoms='H:8 C:4 O:1',
        thermo=(NASA([300.00, 1361.00],
                    [-3.28505561E+00,  5.04800902E-02, -2.51998984E-05,
                    3.65743744E-09,  3.94863161E-13, -2.30067422E+04,
                    4.19349253E+01]),
                NASA([1361.00, 5000.00],
                    [ 1.42360731E+01,  1.81176105E-02, -6.46263809E-06,
                    1.03194372E-09, -6.10557331E-14, -2.99478670E+04,
                    -5.52081041E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='1/22/95THERM')

species(name='C4H8O2-3',
        atoms='H:8 C:4 O:1',
        thermo=(NASA([300.00, 1384.00],
                    [-2.98701197E+00,  6.32678880E-02, -5.20855777E-05,
                    2.24064364E-08, -3.95416618E-12, -1.71957196E+04,
                    3.66920390E+01]),
                NASA([1384.00, 5000.00],
                    [ 1.58263859E+01,  1.64400938E-02, -5.80680311E-06,
                    9.21146000E-10, -5.42511002E-14, -2.35334573E+04,
                    -6.34844562E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='1/22/95THERM')

species(name='PC4H8OH',
        atoms='H:9 C:4 O:1',
        thermo=(NASA([300.00, 1503.00],
                    [ 1.20329680E+00,  4.40218297E-02, -2.12295754E-05,
                    3.03714993E-09,  3.72027554E-13, -1.18622440E+04,
                    2.65515175E+01]),
                NASA([1503.00, 5000.00],
                    [ 1.40357398E+01,  1.94173155E-02, -6.50230141E-06,
                    9.95481694E-10, -5.72034525E-14, -1.67631276E+04,
                    -4.41162528E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='2/12/9THERM')

species(name='SC4H8OH',
        atoms='H:9 C:4 O:1',
        thermo=(NASA([300.00, 1392.00],
                    [-1.04565044E+00,  5.72423420E-02, -4.43109142E-05,
                    1.80944672E-08, -3.02843329E-12, -1.28718032E+04,
                    3.48383489E+01]),
                NASA([1392.00, 5000.00],
                    [ 1.48421985E+01,  1.78435665E-02, -6.16709050E-06,
                    1.00443458E-09, -6.10697983E-14, -1.81714637E+04,
                    -4.96976895E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='2/12/9THERM')

species(name='C4H8OH-1O2',
        atoms='H:9 C:4 O:3',
        thermo=(NASA([300.00, 1396.00],
                    [ 2.88497398E+00,  5.63287929E-02, -3.98403503E-05,
                    1.53891643E-08, -2.51940787E-12, -3.05246721E+04,
                    2.09293355E+01]),
                NASA([1396.00, 5000.00],
                    [ 1.74383247E+01,  2.16778876E-02, -7.37772628E-06,
                    1.14128811E-09, -6.60391451E-14, -3.55892620E+04,
                    -5.71247140E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.356,
                                well_depth=502.2,
                                dipole=1.8,
                                rot_relax=1.0),
        note='6/26/95THERM')

species(name='C4H8OH-2O2',
        atoms='H:9 C:4 O:3',
        thermo=(NASA([300.00, 1399.00],
                    [ 2.10003504E+00,  6.17528218E-02, -4.77238541E-05,
                    1.99145784E-08, -3.44416275E-12, -3.26698755E+04,
                    2.30901271E+01]),
                NASA([1399.00, 5000.00],
                    [ 1.82942871E+01,  2.09395389E-02, -7.12096934E-06,
                    1.10103094E-09, -6.36889768E-14, -3.80505855E+04,
                    -6.29154326E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.356,
                                well_depth=502.2,
                                dipole=1.8,
                                rot_relax=1.0),
        note='6/26/95THERM')

species(name='C4H8OOH1-1',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1379.00],
                    [ 4.36087539E+00,  4.46153162E-02, -2.14784588E-05,
                    3.14619957E-09,  3.03152649E-13, -6.89950326E+03,
                    1.24757248E+01]),
                NASA([1379.00, 5000.00],
                    [ 1.80477436E+01,  1.92137176E-02, -6.66503008E-06,
                    1.04468117E-09, -6.10170737E-14, -1.23069570E+04,
                    -6.33580990E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.716,
                                well_depth=549.655),
        note='3/27/97THERM')

species(name='C4H8OOH1-2',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1393.00],
                    [ 1.51698234E+00,  5.44584331E-02, -3.84524319E-05,
                    1.48145523E-08, -2.42809434E-12, -3.49728423E+03,
                    2.64916926E+01]),
                NASA([1393.00, 5000.00],
                    [ 1.58327906E+01,  2.07220436E-02, -7.09869028E-06,
                    1.10302045E-09, -6.40251984E-14, -8.53034314E+03,
                    -5.04399828E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C4H8OOH1-3',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1377.00],
                    [ 1.94106276E+00,  5.18789351E-02, -3.10411683E-05,
                    8.63568881E-09, -8.42841994E-13, -4.34315962E+03,
                    2.40230471E+01]),
                NASA([1377.00, 5000.00],
                    [ 1.76442170E+01,  1.91706536E-02, -6.57168641E-06,
                    1.02246571E-09, -5.94304735E-14, -1.01859280E+04,
                    -6.17115813E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C4H8OOH1-4',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1386.00],
                    [ 1.45017084E+00,  5.72418991E-02, -4.06431218E-05,
                    1.52821793E-08, -2.42123218E-12, -3.02905522E+03,
                    2.50737890E+01]),
                NASA([1386.00, 5000.00],
                    [ 1.78623700E+01,  1.95212501E-02, -6.80351173E-06,
                    1.06960529E-09, -6.26000540E-14, -8.87923008E+03,
                    -6.34393359E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C4H8OOH2-1',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [ 1.01711681E+00,  6.10560866E-02, -4.65789185E-05,
                    1.88024138E-08, -3.15265575E-12, -5.37391474E+03,
                    2.63059691E+01]),
                NASA([1390.00, 5000.00],
                    [ 1.84555673E+01,  1.89543719E-02, -6.59381457E-06,
                    1.03538370E-09, -6.05467101E-14, -1.13637100E+04,
                    -6.69839138E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C4H8OOH2-2',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1383.00],
                    [ 4.40329034E+00,  4.30478989E-02, -2.14076129E-05,
                    4.28267268E-09, -1.27804885E-13, -9.79432029E+03,
                    1.30514046E+01]),
                NASA([1383.00, 5000.00],
                    [ 1.63837700E+01,  2.04422768E-02, -7.04524557E-06,
                    1.09927552E-09, -6.39967086E-14, -1.45616229E+04,
                    -5.33147431E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.716,
                                well_depth=549.655),
        note='3/27/97THERM')

species(name='C4H8OOH2-3',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1387.00],
                    [ 1.45436266E+00,  5.63121286E-02, -3.84161425E-05,
                    1.34229297E-08, -1.92382914E-12, -6.68916116E+03,
                    2.54247554E+01]),
                NASA([1387.00, 5000.00],
                    [ 1.79338472E+01,  1.92153623E-02, -6.64486961E-06,
                    1.03940007E-09, -6.06239862E-14, -1.25767913E+04,
                    -6.36194089E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C4H8OOH2-4',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [ 1.01711681E+00,  6.10560866E-02, -4.65789185E-05,
                    1.88024138E-08, -3.15265575E-12, -5.37391474E+03,
                    2.63059691E+01]),
                NASA([1390.00, 5000.00],
                    [ 1.84555673E+01,  1.89543719E-02, -6.59381457E-06,
                    1.03538370E-09, -6.05467101E-14, -1.13637100E+04,
                    -6.69839138E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C4H8OOH1-2O2',
        atoms='H:9 C:4 O:4',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.22400728E+00,  7.04994620E-02, -5.66978827E-05,
                    2.42627751E-08, -4.29715459E-12, -2.37391490E+04,
                    2.38371533E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.23244015E+01,  2.05474775E-02, -7.19076348E-06,
                    1.13361536E-09, -6.64744383E-14, -3.05468277E+04,
                    -8.32666070E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C4H8OOH1-3O2',
        atoms='H:9 C:4 O:4',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.22400728E+00,  7.04994620E-02, -5.66978827E-05,
                    2.42627751E-08, -4.29715459E-12, -2.37391490E+04,
                    2.38371533E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.23244015E+01,  2.05474775E-02, -7.19076348E-06,
                    1.13361536E-09, -6.64744383E-14, -3.05468277E+04,
                    -8.32666070E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C4H8OOH1-4O2',
        atoms='H:9 C:4 O:4',
        thermo=(NASA([300.00, 1383.00],
                    [ 3.12351717E+00,  6.46725908E-02, -4.78357914E-05,
                    1.89656817E-08, -3.18025828E-12, -2.14664300E+04,
                    2.04322494E+01]),
                NASA([1383.00, 5000.00],
                    [ 2.17116470E+01,  2.11216172E-02, -7.40101448E-06,
                    1.16772959E-09, -6.85139846E-14, -2.80483124E+04,
                    -7.95885249E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C4H8OOH2-1O2',
        atoms='H:9 C:4 O:4',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.22400728E+00,  7.04994620E-02, -5.66978827E-05,
                    2.42627751E-08, -4.29715459E-12, -2.37391490E+04,
                    2.38371533E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.23244015E+01,  2.05474775E-02, -7.19076348E-06,
                    1.13361536E-09, -6.64744383E-14, -3.05468277E+04,
                    -8.32666070E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C4H8OOH2-3O2',
        atoms='H:9 C:4 O:4',
        thermo=(NASA([300.00, 1389.00],
                    [ 2.19248006E+00,  7.36763430E-02, -6.24428300E-05,
                    2.79371307E-08, -5.10126013E-12, -2.61773538E+04,
                    2.29568752E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.29951044E+01,  1.99238919E-02, -6.96326765E-06,
                    1.09677832E-09, -6.42747281E-14, -3.30095408E+04,
                    -8.71461958E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C4H8OOH2-4O2',
        atoms='H:9 C:4 O:4',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.22400728E+00,  7.04994620E-02, -5.66978827E-05,
                    2.42627751E-08, -4.29715459E-12, -2.37391490E+04,
                    2.38371533E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.23244015E+01,  2.05474775E-02, -7.19076348E-06,
                    1.13361536E-09, -6.64744383E-14, -3.05468277E+04,
                    -8.32666070E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='NC4KET12',
        atoms='H:8 C:4 O:3',
        thermo=(NASA([300.00, 1388.00],
                    [ 8.05387149E-01,  6.52375119E-02, -5.18373651E-05,
                    2.13454896E-08, -3.59860255E-12, -3.92395961E+04,
                    2.84908472E+01]),
                NASA([1388.00, 5000.00],
                    [ 2.02458485E+01,  1.76440360E-02, -6.19054598E-06,
                    9.77688024E-10, -5.74053258E-14, -4.58141511E+04,
                    -7.52145443E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.778,
                                well_depth=476.0,
                                dipole=2.6,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC4KET13',
        atoms='H:8 C:4 O:3',
        thermo=(NASA([300.00, 1386.00],
                    [ 2.74883461E+00,  5.86936745E-02, -4.49605895E-05,
                    1.83200130E-08, -3.11765369E-12, -4.01065878E+04,
                    1.88072090E+01]),
                NASA([1386.00, 5000.00],
                    [ 1.96430808E+01,  1.80940566E-02, -6.33063232E-06,
                    9.97860399E-10, -5.85076458E-14, -4.59588851E+04,
                    -7.16905094E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.778,
                                well_depth=476.0,
                                dipole=2.6,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC4KET14',
        atoms='H:8 C:4 O:3',
        thermo=(NASA([300.00, 1382.00],
                    [ 3.40797147E+00,  5.37386096E-02, -3.72613233E-05,
                    1.36862855E-08, -2.13798779E-12, -3.77919233E+04,
                    1.65580190E+01]),
                NASA([1382.00, 5000.00],
                    [ 1.90283822E+01,  1.86604765E-02, -6.53616227E-06,
                    1.03102482E-09, -6.04833256E-14, -4.34678787E+04,
                    -6.80210373E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.778,
                                well_depth=476.0,
                                dipole=2.6,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC4KET21',
        atoms='H:8 C:4 O:3',
        thermo=(NASA([300.00, 1376.00],
                    [ 4.69785973E+00,  4.83859623E-02, -3.05576939E-05,
                    1.02705744E-08, -1.51713483E-12, -4.18665528E+04,
                    1.06715264E+01]),
                NASA([1376.00, 5000.00],
                    [ 1.86231473E+01,  1.90969401E-02, -6.70745998E-06,
                    1.05994690E-09, -6.22564248E-14, -4.71929933E+04,
                    -6.55477892E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.778,
                                well_depth=476.0,
                                dipole=2.6,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC4KET23',
        atoms='H:8 C:4 O:3',
        thermo=(NASA([300.00, 1387.00],
                    [ 4.50575961E-01,  6.48862527E-02, -5.04446464E-05,
                    2.03085220E-08, -3.35952497E-12, -4.28779359E+04,
                    3.00153676E+01]),
                NASA([1387.00, 5000.00],
                    [ 1.99512969E+01,  1.79558596E-02, -6.31230878E-06,
                    9.98217386E-10, -5.86636973E-14, -4.95585495E+04,
                    -7.43071328E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.778,
                                well_depth=476.0,
                                dipole=2.6,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC4KET24',
        atoms='H:8 C:4 O:3',
        thermo=(NASA([300.00, 1380.00],
                    [ 2.93007283E+00,  5.39296076E-02, -3.66164919E-05,
                    1.30880106E-08, -1.99245456E-12, -4.14118815E+04,
                    1.86567511E+01]),
                NASA([1380.00, 5000.00],
                    [ 1.87707201E+01,  1.89490823E-02, -6.65182145E-06,
                    1.05080860E-09, -6.17069393E-14, -4.72319802E+04,
                    -6.73335107E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.778,
                                well_depth=476.0,
                                dipole=2.6,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C2H5COCH3',
        atoms='H:8 C:4 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 6.61978185E+00,  8.51847835E-03,  5.10322077E-05,
                    -6.58433042E-08,  2.49110484E-11, -3.15251691E+04,
                    -1.09485469E+00]),
                NASA([1000.00, 6000.00],
                    [ 9.29655016E+00,  2.29172746E-02, -8.22048591E-06,
                    1.32404838E-09, -7.91751980E-14, -3.34442311E+04,
                    -2.04993263E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.413,
                                well_depth=454.0,
                                dipole=3.3,
                                rot_relax=1.0),
        note='T09/10')

species(name='C2H5COCH2',
        atoms='H:7 C:4 O:1',
        thermo=(NASA([300.00, 1383.00],
                    [ 1.54013856E+00,  4.39486258E-02, -2.97002421E-05,
                    1.05495313E-08, -1.58598769E-12, -9.50796505E+03,
                    1.99706641E+01]),
                NASA([1383.00, 5000.00],
                    [ 1.42098738E+01,  1.57866459E-02, -5.50529183E-06,
                    8.65870540E-10, -5.06913329E-14, -1.41284951E+04,
                    -4.87132911E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.413,
                                well_depth=454.0,
                                dipole=3.3,
                                rot_relax=1.0),
        note='4/3/0THERM')

species(name='CH2CH2COCH3',
        atoms='H:7 C:4 O:1',
        thermo=(NASA([300.00, 1380.00],
                    [ 2.40255609E+00,  3.67294268E-02, -1.97316510E-05,
                    5.07323216E-09, -4.99655275E-13, -6.15006886E+03,
                    1.93993386E+01]),
                NASA([1380.00, 5000.00],
                    [ 1.24694368E+01,  1.71022143E-02, -5.92156726E-06,
                    9.26816806E-10, -5.40730504E-14, -1.01378242E+04,
                    -3.62186375E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.04,
                                well_depth=440.584),
        note='6/21/95THER')

species(name='CH3CHCOCH3',
        atoms='H:7 C:4 O:1',
        thermo=(NASA([300.00, 1384.00],
                    [ 8.12941079E-01,  4.29256944E-02, -2.69230252E-05,
                    8.59326807E-09, -1.13188129E-12, -1.05247481E+04,
                    2.32952685E+01]),
                NASA([1384.00, 5000.00],
                    [ 1.31388032E+01,  1.66091073E-02, -5.76924215E-06,
                    9.04978165E-10, -5.28826527E-14, -1.51162098E+04,
                    -4.38876580E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.04,
                                well_depth=440.584),
        note='4/3/0THERM')

species(name='C2H3COCH3',
        atoms='H:6 C:4 O:1',
        thermo=(NASA([300.00, 1386.00],
                    [ 2.45578501E-01,  4.26432049E-02, -2.91126822E-05,
                    1.03478392E-08, -1.53551381E-12, -1.70305379E+04,
                    2.64430799E+01]),
                NASA([1386.00, 5000.00],
                    [ 1.25571995E+01,  1.49672645E-02, -5.20015351E-06,
                    8.15864365E-10, -4.76824406E-14, -2.14622958E+04,
                    -4.01434299E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.413,
                                well_depth=454.0,
                                dipole=3.3,
                                rot_relax=1.0),
        note='6/19/95THERM')

species(name='CH3CHOOCOCH3',
        atoms='H:7 C:4 O:3',
        thermo=(NASA([300.00, 1393.00],
                    [ 1.59405312E+00,  5.74729487E-02, -4.65071459E-05,
                    2.01816805E-08, -3.61175600E-12, -2.53445178E+04,
                    2.60857480E+01]),
                NASA([1393.00, 5000.00],
                    [ 1.72170832E+01,  1.76740453E-02, -6.10387321E-06,
                    9.53711693E-10, -5.55757409E-14, -3.05280370E+04,
                    -5.67959657E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.698,
                                well_depth=546.562),
        note='6/27/95')

species(name='CH2CHOOHCOCH3',
        atoms='H:7 C:4 O:3',
        thermo=(NASA([300.00, 1394.00],
                    [ 1.55204769E+00,  6.08201262E-02, -5.04704383E-05,
                    2.16347587E-08, -3.75674081E-12, -1.89563540E+04,
                    2.71011896E+01]),
                NASA([1394.00, 5000.00],
                    [ 1.93158016E+01,  1.56592602E-02, -5.45395604E-06,
                    8.57160086E-10, -5.01582885E-14, -2.47752201E+04,
                    -6.70296581E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.698,
                                well_depth=546.562),
        note='6/27/95')

species(name='NC3H7CHO',
        atoms='H:8 C:4 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 5.30068149E+00,  5.00213349E-03,  8.12219686E-05,
                    -1.07815910E-07,  4.25781054E-11, -2.71198341E+04,
                    4.93592991E+00]),
                NASA([1000.00, 6000.00],
                    [ 1.02351219E+01,  2.32201057E-02, -8.46144199E-06,
                    1.37589764E-09, -8.27046434E-14, -3.00345804E+04,
                    -2.82583105E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.009,
                                well_depth=464.2,
                                dipole=2.6,
                                rot_relax=1.0),
        note='T05/09')

species(name='NC3H7CO',
        atoms='H:7 C:4 O:1',
        thermo=(NASA([300.00, 1380.00],
                    [ 2.67256826E+00,  3.71198825E-02, -2.06862859E-05,
                    5.48873888E-09, -5.35864183E-13, -8.58050888E+03,
                    1.64848950E+01]),
                NASA([1380.00, 5000.00],
                    [ 1.30026331E+01,  1.63104877E-02, -5.57642899E-06,
                    8.65670629E-10, -5.02255667E-14, -1.25523385E+04,
                    -4.02608515E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.009,
                                well_depth=464.2,
                                dipole=2.6,
                                rot_relax=1.0),
        note='9/27/95THERM')

species(name='C3H6CHO-1',
        atoms='H:7 C:4 O:1',
        thermo=(NASA([300.00, 1379.00],
                    [ 2.67672303E+00,  3.73064128E-02, -2.11281405E-05,
                    5.80472681E-09, -6.09688236E-13, -2.49714183E+03,
                    1.75750933E+01]),
                NASA([1379.00, 5000.00],
                    [ 1.30322954E+01,  1.62418373E-02, -5.54388124E-06,
                    8.59723685E-10, -4.98459726E-14, -6.45915975E+03,
                    -3.92399021E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.009,
                                well_depth=464.2,
                                dipole=2.6,
                                rot_relax=1.0),
        note='9/27/95THERM')

species(name='C3H6CHO-2',
        atoms='H:7 C:4 O:1',
        thermo=(NASA([300.00, 1682.00],
                    [ 2.95067531E+00,  3.34223079E-02, -1.45356815E-05,
                    1.67282048E-09,  2.62011555E-13, -3.79034324E+03,
                    1.74324072E+01]),
                NASA([1682.00, 5000.00],
                    [ 1.11942816E+01,  1.81806772E-02, -6.35916662E-06,
                    1.00727333E-09, -5.93943618E-14, -6.86826460E+03,
                    -2.80298956E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.009,
                                well_depth=464.2,
                                dipole=2.6,
                                rot_relax=1.0),
        note='11/15/95THERM')

species(name='C3H6CHO-3',
        atoms='H:7 C:4 O:1',
        thermo=(NASA([300.00, 1387.00],
                    [ 1.35168659E+00,  4.23714083E-02, -2.69123192E-05,
                    8.70132567E-09, -1.15287455E-12, -6.91166610E+03,
                    2.09381509E+01]),
                NASA([1387.00, 5000.00],
                    [ 1.34301738E+01,  1.62250792E-02, -5.60631713E-06,
                    8.76356273E-10, -5.10864241E-14, -1.13561098E+04,
                    -4.47371262E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.009,
                                well_depth=464.2,
                                dipole=2.6,
                                rot_relax=1.0),
        note='11/15/95THERM')

species(name='C2H5CHCO',
        atoms='H:6 C:4 O:1',
        thermo=(NASA([300.00, 1400.00],
                    [ 2.76600655E-01,  4.82966593E-02, -4.02376928E-05,
                    1.76564254E-08, -3.14365357E-12, -1.41717351E+04,
                    2.47118671E+01]),
                NASA([1400.00, 5000.00],
                    [ 1.34100537E+01,  1.38766678E-02, -4.74130143E-06,
                    7.35504188E-10, -4.26457752E-14, -1.83790837E+04,
                    -4.45295352E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.016,
                                well_depth=436.95),
        note='9/27/95THERM')

species(name='SC3H5CHO',
        atoms='H:6 C:4 O:1',
        thermo=(NASA([300.00, 1390.00],
                    [ 4.35795171E-01,  4.48719314E-02, -3.36582931E-05,
                    1.33066870E-08, -2.17839128E-12, -1.60394651E+04,
                    2.37597452E+01]),
                NASA([1390.00, 5000.00],
                    [ 1.31695904E+01,  1.42484434E-02, -4.90843998E-06,
                    7.65789041E-10, -4.45834896E-14, -2.04032613E+04,
                    -4.43673205E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.009,
                                well_depth=464.2,
                                dipole=2.6,
                                rot_relax=1.0),
        note='11/15/95THER')

species(name='SC3H5CO',
        atoms='H:5 C:4 O:1',
        thermo=(NASA([300.00, 1392.00],
                    [ 1.74191343E+00,  3.97229536E-02, -3.20061901E-05,
                    1.38227925E-08, -2.46272017E-12, -6.64428100E+02,
                    1.70762023E+01]),
                NASA([1392.00, 5000.00],
                    [ 1.25514754E+01,  1.22521948E-02, -4.22382101E-06,
                    6.59184896E-10, -3.83818826E-14, -4.25349795E+03,
                    -4.02864145E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.009,
                                well_depth=464.2,
                                dipole=2.6,
                                rot_relax=1.0),
        note='11/15/95THERM')

species(name='CH2CH2CHO',
        atoms='H:5 C:3 O:1',
        thermo=(NASA([300.00, 1363.00],
                    [ 8.36272659E+00, -3.90712606E-03,  3.59751257E-05,
                    -2.69209237E-08,  5.98920261E-12, -7.55243244E+02,
                    -9.53140102E+00]),
                NASA([1363.00, 5000.00],
                    [ 9.96855598E+00,  1.61917946E-02, -6.64161740E-06,
                    1.15444520E-09, -7.21981790E-14, -4.16450354E+03,
                    -2.73971573E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.82,
                                well_depth=424.6,
                                rot_relax=1.0),
        note='8/10/4THERM')

species(name='IC4H10',
        atoms='H:10 C:4',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.45479140E+00,  8.26058864E-03,  8.29886433E-05,
                    -1.14647616E-07,  4.64569994E-11, -1.84593929E+04,
                    4.92740653E+00]),
                NASA([1000.00, 6000.00],
                    [ 9.76991697E+00,  2.54997141E-02, -9.14142587E-06,
                    1.47328201E-09, -8.80799697E-14, -2.14052667E+04,
                    -3.00329670E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.208,
                                well_depth=335.7,
                                dipole=0.1,
                                rot_relax=1.0),
        note='G8/00')

species(name='IC4H9',
        atoms='H:9 C:4',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.34476784E+00,  2.31869650E-02,  3.28261040E-05,
                    -5.96398514E-08,  2.58980820E-11,  6.66201200E+03,
                    9.68860372E+00]),
                NASA([1000.00, 6000.00],
                    [ 9.61250942E+00,  2.28581786E-02, -8.06391309E-06,
                    1.28556553E-09, -7.62730799E-14,  4.15218608E+03,
                    -2.66485099E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.24,
                                well_depth=352.0,
                                rot_relax=1.0),
        note='T6/04')

species(name='TC4H9',
        atoms='H:9 C:4',
        thermo=(NASA([200.00, 1000.00],
                    [ 6.45910754E+00, -1.02015930E-02,  1.06310577E-04,
                    -1.25717030E-07,  4.75543216E-11,  4.43420391E+03,
                    1.30648608E+00]),
                NASA([1000.00, 6000.00],
                    [ 6.72557390E+00,  2.53649194E-02, -9.05306262E-06,
                    1.45474620E-09, -8.67934112E-14,  2.57430692E+03,
                    -8.89920414E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.24,
                                well_depth=352.0,
                                rot_relax=1.0),
        note='T6/04')

species(name='IC4H8',
        atoms='H:8 C:4',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.30612340E+00,  1.33377057E-02,  5.65726066E-05,
                    -8.46898088E-08,  3.52403438E-11, -4.04128388E+03,
                    9.92304633E+00]),
                NASA([1000.00, 6000.00],
                    [ 8.94232121E+00,  2.12900129E-02, -7.61851464E-06,
                    1.22641473E-09, -7.32634442E-14, -6.67292929E+03,
                    -2.46775148E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.089,
                                well_depth=344.5,
                                dipole=0.5,
                                rot_relax=1.0),
        note='T05/09')

species(name='IC4H7',
        atoms='H:7 C:4',
        thermo=(NASA([200.00, 1000.00],
                    [ 2.38739541E+00,  2.06784631E-02,  2.89299685E-05,
                    -5.37553477E-08,  2.35670326E-11,  1.47584145E+04,
                    1.55529104E+01]),
                NASA([1000.00, 6000.00],
                    [ 8.34970451E+00,  1.92508033E-02, -6.81360221E-06,
                    1.08484853E-09, -6.42422082E-14,  1.24406647E+04,
                    -1.87060633E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.65,
                                well_depth=355.0,
                                rot_relax=1.0),
        note='T05/04')

species(name='TC4H9O2',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.45494881E+00,  2.89185105E-02,  3.68274347E-05,
                    -7.18116292E-08,  3.18779709E-11, -1.56103943E+04,
                    9.17831465E+00]),
                NASA([1000.00, 6000.00],
                    [ 1.34566146E+01,  2.40864396E-02, -8.56308159E-06,
                    1.36853372E-09, -8.12706683E-14, -1.89690450E+04,
                    -4.18792381E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.43,
                                well_depth=502.252),
        note='T04/08')

species(name='IC4H9O2',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1387.00],
                    [ 1.21434293E+00,  5.45388311E-02, -3.67001593E-05,
                    1.34131042E-08, -2.11741793E-12, -1.18482450E+04,
                    2.34153048E+01]),
                NASA([1387.00, 5000.00],
                    [ 1.59741221E+01,  2.13534740E-02, -7.39001105E-06,
                    1.15624411E-09, -6.74408046E-14, -1.72329304E+04,
                    -5.65302409E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.43,
                                well_depth=502.252),
        note='7/19/0THERM')

species(name='TC4H8O2H-I',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1405.00],
                    [ 1.57469723E+00,  6.04648659E-02, -4.84991814E-05,
                    2.05236181E-08, -3.52974722E-12, -6.54190901E+03,
                    2.16852700E+01]),
                NASA([1405.00, 5000.00],
                    [ 1.78060253E+01,  1.84952716E-02, -6.21268423E-06,
                    9.52753962E-10, -5.48010330E-14, -1.17555816E+04,
                    -6.40180236E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.356,
                                well_depth=502.2,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='IC4H8O2H-I',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1388.00],
                    [ 9.94784793E-01,  5.89212240E-02, -4.25202225E-05,
                    1.61370574E-08, -2.55904902E-12, -4.08029057E+03,
                    2.58950880E+01]),
                NASA([1388.00, 5000.00],
                    [ 1.80246456E+01,  1.93668264E-02, -6.74676496E-06,
                    1.06040129E-09, -6.20506445E-14, -1.00858977E+04,
                    -6.57629693E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.356,
                                well_depth=502.2,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='IC4H8O2H-T',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1397.00],
                    [ 3.77633001E+00,  4.50889269E-02, -2.62423875E-05,
                    7.74736900E-09, -9.11520393E-13, -5.98295842E+03,
                    1.53503399E+01]),
                NASA([1397.00, 5000.00],
                    [ 1.52395012E+01,  2.06496983E-02, -6.94791927E-06,
                    1.06651097E-09, -6.13782754E-14, -1.02271199E+04,
                    -4.70908585E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.356,
                                well_depth=502.2,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='IC4H8O',
        atoms='H:8 C:4 O:1',
        thermo=(NASA([300.00, 1399.00],
                    [-2.97373741E+00,  6.24618637E-02, -5.04348211E-05,
                    2.13345360E-08, -3.67382824E-12, -1.73274318E+04,
                    3.62336362E+01]),
                NASA([1399.00, 5000.00],
                    [ 1.44624830E+01,  1.80562860E-02, -6.18008791E-06,
                    9.59941341E-10, -5.57132297E-14, -2.30230616E+04,
                    -5.61190165E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.352,
                                well_depth=436.4,
                                rot_relax=1.0),
        note='7/19/95THERM')

species(name='CC4H8O',
        atoms='H:8 C:4 O:1',
        thermo=(NASA([300.00, 1401.00],
                    [-5.27768230E+00,  6.52482242E-02, -5.00209030E-05,
                    1.97452179E-08, -3.15605813E-12, -1.37706522E+04,
                    4.78500688E+01]),
                NASA([1401.00, 5000.00],
                    [ 1.35131704E+01,  1.91666441E-02, -6.53902359E-06,
                    1.01366570E-09, -5.87546671E-14, -2.00311488E+04,
                    -5.22146967E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.063,
                                well_depth=444.197),
        note='6/29/95THERM')

species(name='IC4H9O',
        atoms='H:9 C:4 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.80297372E+00,  1.56874209E-02,  6.81105412E-05,
                    -9.83346774E-08,  3.95261902E-11, -1.00832243E+04,
                    9.78963305E+00]),
                NASA([1000.00, 6000.00],
                    [ 1.16309708E+01,  2.47981574E-02, -9.01550536E-06,
                    1.46714720E-09, -8.83214518E-14, -1.37854612E+04,
                    -3.81956151E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='A08/04')

species(name='TC4H9O',
        atoms='H:9 C:4 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 2.77057100E+00,  2.68033175E-02,  4.12718360E-05,
                    -7.22054739E-08,  3.02642276E-11, -1.27079262E+04,
                    1.21532856E+01]),
                NASA([1000.00, 6000.00],
                    [ 1.27371509E+01,  2.33707342E-02, -8.50516678E-06,
                    1.38519973E-09, -8.34398061E-14, -1.66940150E+04,
                    -4.53156321E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='T08/04')

species(name='IC4H9O2H',
        atoms='H:10 C:4 O:2',
        thermo=(NASA([300.00, 1388.00],
                    [ 2.44181899E-01,  6.34027841E-02, -4.49162379E-05,
                    1.67406877E-08, -2.61216478E-12, -2.86959808E+04,
                    2.72156229E+01]),
                NASA([1388.00, 5000.00],
                    [ 1.84308794E+01,  2.15337606E-02, -7.48623965E-06,
                    1.17499240E-09, -6.86890788E-14, -3.51483277E+04,
                    -7.08030846E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.45,
                                well_depth=505.54),
        note='7/19/0THERM')

species(name='TC4H9O2H',
        atoms='H:10 C:4 O:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.96033042E+00,  4.06267078E-02,  1.85652131E-05,
                    -5.72128820E-08,  2.73014984E-11, -3.20467014E+04,
                    7.55565609E+00]),
                NASA([1000.00, 6000.00],
                    [ 1.52702114E+01,  2.56462637E-02, -9.02796199E-06,
                    1.43759278E-09, -8.52381573E-14, -3.58539067E+04,
                    -5.44966764E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.45,
                                well_depth=505.54),
        note='T02/10')

species(name='IC4H7O',
        atoms='H:7 C:4 O:1',
        thermo=(NASA([300.00, 1386.00],
                    [ 1.74700687E+00,  4.07783436E-02, -2.44750243E-05,
                    7.06502958E-09, -7.51570589E-13,  4.86979233E+03,
                    1.94535999E+01]),
                NASA([1386.00, 5000.00],
                    [ 1.33457615E+01,  1.61218588E-02, -5.44376403E-06,
                    8.38199374E-10, -4.83608280E-14,  6.11443644E+02,
                    -4.36818838E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='4/3/0THERM')

species(name='IC4H8OH',
        atoms='H:9 C:4 O:1',
        thermo=(NASA([300.00, 1376.00],
                    [ 3.29612707E+00,  3.47649647E-02, -1.02505618E-05,
                    -2.04641931E-09,  1.18879408E-12, -1.45627247E+04,
                    1.58606320E+01]),
                NASA([1376.00, 5000.00],
                    [ 1.25605997E+01,  2.10637488E-02, -7.15019648E-06,
                    1.10439262E-09, -6.38428695E-14, -1.86203249E+04,
                    -3.67889430E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='2/14/95THERM')

species(name='IO2C4H8OH',
        atoms='H:9 C:4 O:3',
        thermo=(NASA([300.00, 1395.00],
                    [ 2.88826008E+00,  5.80363124E-02, -4.26690874E-05,
                    1.71759332E-08, -2.91842186E-12, -3.33881503E+04,
                    1.83381149E+01]),
                NASA([1395.00, 5000.00],
                    [ 1.79798208E+01,  2.13452115E-02, -7.29108198E-06,
                    1.13068846E-09, -6.55402236E-14, -3.85769917E+04,
                    -6.23553522E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='8/23/95THERM')

species(name='IC3H7CHO',
        atoms='H:8 C:4 O:1',
        thermo=(NASA([300.00, 1391.00],
                    [-2.73021382E-01,  4.89696307E-02, -3.12770049E-05,
                    1.00052945E-08, -1.27512074E-12, -2.76054737E+04,
                    2.83451139E+01]),
                NASA([1391.00, 5000.00],
                    [ 1.37501656E+01,  1.83126722E-02, -6.28572629E-06,
                    9.78250756E-10, -5.68538653E-14, -3.26936771E+04,
                    -4.77270548E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.352,
                                well_depth=436.4,
                                rot_relax=1.0),
        note='2/22/96THERM')

species(name='TC3H6CHO',
        atoms='H:7 C:4 O:1',
        thermo=(NASA([300.00, 1389.00],
                    [ 1.87052762E+00,  4.14869677E-02, -2.66815701E-05,
                    9.01531610E-09, -1.27870633E-12, -8.97730744E+03,
                    1.66174178E+01]),
                NASA([1389.00, 5000.00],
                    [ 1.31013047E+01,  1.66391865E-02, -5.68457623E-06,
                    8.81808351E-10, -5.11290161E-14, -1.30638647E+04,
                    -4.42705813E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.352,
                                well_depth=436.4,
                                rot_relax=1.0),
        note='2/22/96THERM')

species(name='IC3H7CO',
        atoms='H:7 C:4 O:1',
        thermo=(NASA([300.00, 1390.00],
                    [ 5.03452639E-01,  4.41607510E-02, -2.82139091E-05,
                    8.93548675E-09, -1.11327422E-12, -9.07755468E+03,
                    2.61991461E+01]),
                NASA([1390.00, 5000.00],
                    [ 1.33305736E+01,  1.61873930E-02, -5.56711402E-06,
                    8.67575951E-10, -5.04696549E-14, -1.37307001E+04,
                    -4.33958746E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.352,
                                well_depth=436.4,
                                rot_relax=1.0),
        note='2/22/96THERM')

species(name='IC3H6CHO',
        atoms='H:7 C:4 O:1',
        thermo=(NASA([300.00, 1390.00],
                    [ 5.21481767E-01,  4.43114357E-02, -2.86617314E-05,
                    9.30319894E-09, -1.20761563E-12, -2.99677086E+03,
                    2.68182130E+01]),
                NASA([1390.00, 5000.00],
                    [ 1.33102250E+01,  1.62097959E-02, -5.57575891E-06,
                    8.69003718E-10, -5.05554202E-14, -7.62177931E+03,
                    -4.25050854E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.352,
                                well_depth=436.4,
                                rot_relax=1.0),
        note='2/22/96THERM')

species(name='TC4H8OOH-IO2',
        atoms='H:9 C:4 O:4',
        thermo=(NASA([300.00, 1386.00],
                    [ 2.17782096E+00,  7.17785118E-02, -5.88836046E-05,
                    2.56759171E-08, -4.61678643E-12, -2.55622378E+04,
                    2.07846231E+01]),
                NASA([1386.00, 5000.00],
                    [ 2.26382889E+01,  2.03246143E-02, -7.12390770E-06,
                    1.12425028E-09, -6.59732993E-14, -3.24406877E+04,
                    -8.80387843E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.009,
                                well_depth=600.078),
        note='7/19/0TRM')

species(name='IC4H8OOH-IO2',
        atoms='H:9 C:4 O:4',
        thermo=(NASA([300.00, 1385.00],
                    [ 2.39424426E+00,  6.76572549E-02, -5.17083682E-05,
                    2.10796041E-08, -3.59960373E-12, -2.24787495E+04,
                    2.25029839E+01]),
                NASA([1385.00, 5000.00],
                    [ 2.18969581E+01,  2.09637874E-02, -7.34664900E-06,
                    1.15926594E-09, -6.80225413E-14, -2.92664889E+04,
                    -8.20540807E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.009,
                                well_depth=600.078),
        note='7/19/0TRM')

species(name='IC4H8OOH-TO2',
        atoms='H:9 C:4 O:4',
        thermo=(NASA([300.00, 1386.00],
                    [ 2.17782096E+00,  7.17785118E-02, -5.88836046E-05,
                    2.56759171E-08, -4.61678643E-12, -2.55622378E+04,
                    2.07846231E+01]),
                NASA([1386.00, 5000.00],
                    [ 2.26382889E+01,  2.03246143E-02, -7.12390770E-06,
                    1.12425028E-09, -6.59732993E-14, -3.24406877E+04,
                    -8.80387843E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.009,
                                well_depth=600.078),
        note='7/19/0TRM')

species(name='IC4KETII',
        atoms='H:8 C:4 O:3',
        thermo=(NASA([300.00, 1387.00],
                    [ 1.15501614E+00,  6.10622345E-02, -4.49711323E-05,
                    1.70514654E-08, -2.65948602E-12, -3.82747956E+04,
                    2.69612235E+01]),
                NASA([1387.00, 5000.00],
                    [ 1.95143059E+01,  1.82377395E-02, -6.38908606E-06,
                    1.00801571E-09, -5.91440350E-14, -4.46884836E+04,
                    -7.17167584E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.716,
                                well_depth=549.655),
        note='7/19/0THERM')

species(name='IC4KETIT',
        atoms='H:8 C:4 O:3',
        thermo=(NASA([300.00, 1388.00],
                    [ 1.14243741E+00,  6.33840797E-02, -4.73084738E-05,
                    1.77145373E-08, -2.67265475E-12, -4.09366796E+04,
                    2.34844867E+01]),
                NASA([1388.00, 5000.00],
                    [ 2.09369850E+01,  1.71090955E-02, -6.01892169E-06,
                    9.52353863E-10, -5.59926176E-14, -4.77819819E+04,
                    -8.27717611E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.716,
                                well_depth=549.655),
        note='7/19/0THERM')

species(name='IC4H7OH',
        atoms='H:8 C:4 O:1',
        thermo=(NASA([300.00, 1384.00],
                    [ 1.69099899E+00,  4.27168891E-02, -2.49281695E-05,
                    7.00961522E-09, -7.23262828E-13, -2.14512334E+04,
                    1.99500833E+01]),
                NASA([1384.00, 5000.00],
                    [ 1.35043419E+01,  1.78646968E-02, -5.99304371E-06,
                    9.18717641E-10, -5.28435302E-14, -2.58255688E+04,
                    -4.44645715E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.356,
                                well_depth=502.2,
                                dipole=1.8,
                                rot_relax=1.0),
        note='4/8/97THERM')

species(name='IC4H6OH',
        atoms='H:7 C:4 O:1',
        thermo=(NASA([300.00, 1390.00],
                    [ 8.63371227E-01,  4.68711282E-02, -3.43580339E-05,
                    1.33031052E-08, -2.13914975E-12, -3.14948305E+03,
                    2.29075523E+01]),
                NASA([1390.00, 5000.00],
                    [ 1.40310926E+01,  1.55317541E-02, -5.32754961E-06,
                    8.28785902E-10, -4.81545257E-14, -7.69378228E+03,
                    -4.76555306E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.356,
                                well_depth=502.2,
                                dipole=1.8,
                                rot_relax=1.0),
        note='8/19/95THERM')

species(name='IC3H5CHO',
        atoms='H:6 C:4 O:1',
        thermo=(NASA([300.00, 1396.00],
                    [ 6.27183793E-01,  4.66780254E-02, -3.74430631E-05,
                    1.58330542E-08, -2.73952155E-12, -1.57203117E+04,
                    2.16034294E+01]),
                NASA([1396.00, 5000.00],
                    [ 1.36203958E+01,  1.37917192E-02, -4.73370118E-06,
                    7.36655226E-10, -4.20097974E-14, -2.00025274E+04,
                    -4.73184531E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.352,
                                well_depth=436.4,
                                rot_relax=1.0),
        note='7/19/95THERM')

species(name='IC3H5CO',
        atoms='H:5 C:4 O:1',
        thermo=(NASA([300.00, 1397.00],
                    [ 1.85097069E+00,  4.18855846E-02, -3.62553731E-05,
                    1.65690659E-08, -3.05850846E-12,  1.70381441E+02,
                    1.53014433E+01]),
                NASA([1397.00, 5000.00],
                    [ 1.30667437E+01,  1.16704244E-02, -3.99106523E-06,
                    6.19498148E-10, -3.59348249E-14, -3.36519344E+03,
                    -4.35803090E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.352,
                                well_depth=436.4,
                                rot_relax=1.0),
        note='4/3/0THERM')

species(name='TC3H6OCHO',
        atoms='H:7 C:4 O:2',
        thermo=(NASA([300.00, 1394.00],
                    [ 3.70830259E-01,  5.38475661E-02, -3.82477565E-05,
                    1.32882237E-08, -1.79228730E-12, -2.18391262E+04,
                    2.58142112E+01]),
                NASA([1394.00, 5000.00],
                    [ 1.70371287E+01,  1.54400645E-02, -5.28332886E-06,
                    8.21085347E-10, -4.76898429E-14, -2.75871941E+04,
                    -6.37271230E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.352,
                                well_depth=436.4,
                                rot_relax=1.0),
        note='8/25/95THERM')

species(name='IC3H6CO',
        atoms='H:6 C:4 O:1',
        thermo=(NASA([300.00, 1397.00],
                    [ 2.28039055E+00,  4.17016989E-02, -3.25089661E-05,
                    1.37243419E-08, -2.40573132E-12, -1.63939712E+04,
                    1.38187714E+01]),
                NASA([1397.00, 5000.00],
                    [ 1.32548232E+01,  1.40142787E-02, -4.78910215E-06,
                    7.42924342E-10, -4.30737566E-14, -2.00529779E+04,
                    -4.44810221E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.352,
                                well_depth=436.4,
                                rot_relax=1.0),
        note='03/03/95THERM')

species(name='IC4H7OOH',
        atoms='H:8 C:4 O:2',
        thermo=(NASA([300.00, 1392.00],
                    [ 2.99117402E+00,  5.03349278E-02, -3.56280061E-05,
                    1.33952154E-08, -2.11053409E-12, -1.51095046E+04,
                    1.54537413E+01]),
                NASA([1392.00, 5000.00],
                    [ 1.69234564E+01,  1.78396769E-02, -6.14273279E-06,
                    9.57895028E-10, -5.57438304E-14, -2.00040686E+04,
                    -5.94746070E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.352,
                                well_depth=436.4,
                                rot_relax=1.0),
        note='7/19/95THERM')

species(name='TC3H6OHCHO',
        atoms='H:8 C:4 O:2',
        thermo=(NASA([300.00, 1396.00],
                    [ 2.10033500E-01,  5.63086284E-02, -3.95158669E-05,
                    1.37780428E-08, -1.89068805E-12, -4.80402638E+04,
                    2.67836070E+01]),
                NASA([1396.00, 5000.00],
                    [ 1.70788996E+01,  1.74005554E-02, -5.93008112E-06,
                    9.18860909E-10, -5.32512140E-14, -5.38730427E+04,
                    -6.38638754E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.352,
                                well_depth=436.4,
                                rot_relax=1.0),
        note='8/1/95THERM')

species(name='TC3H6OH',
        atoms='H:7 C:3 O:1',
        thermo=(NASA([300.00, 1392.00],
                    [ 1.09670360E+00,  3.80727565E-02, -2.75022497E-05,
                    1.07477493E-08, -1.74895773E-12, -1.40764487E+04,
                    2.22475799E+01]),
                NASA([1392.00, 5000.00],
                    [ 1.12222277E+01,  1.36444398E-02, -4.51406709E-06,
                    7.10523275E-10, -4.22690392E-14, -1.75350136E+04,
                    -3.18911926E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.036,
                                well_depth=459.5,
                                dipole=1.7,
                                rot_relax=1.0),
        note='8/9/4THERM')

species(name='IC3H5OH',
        atoms='H:6 C:3 O:1',
        thermo=(NASA([300.00, 1374.00],
                    [ 1.58376391E+00,  3.16215366E-02, -1.73664942E-05,
                    4.18927663E-09, -2.79899620E-13, -2.12643496E+04,
                    1.88313766E+01]),
                NASA([1374.00, 5000.00],
                    [ 1.07381025E+01,  1.31698194E-02, -4.41529622E-06,
                    6.77009837E-10, -3.89608901E-14, -2.47298321E+04,
                    -3.13634050E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.036,
                                well_depth=459.5,
                                dipole=1.7,
                                rot_relax=1.0),
        note='8/1/95THERM')

species(name='TC3H6O2CHO',
        atoms='H:7 C:4 O:3',
        thermo=(NASA([300.00, 1386.00],
                    [ 2.17883383E+00,  5.41595832E-02, -3.83435886E-05,
                    1.38308104E-08, -2.04190147E-12, -2.27394154E+04,
                    2.00751264E+01]),
                NASA([1386.00, 5000.00],
                    [ 1.85534443E+01,  1.68774389E-02, -5.90752965E-06,
                    9.31518085E-10, -5.46345187E-14, -2.85447191E+04,
                    -6.82486667E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.352,
                                well_depth=436.4,
                                rot_relax=1.0),
        note='8/2/95THERM')

species(name='TC3H6O2HCO',
        atoms='H:7 C:4 O:3',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.03864428E+00,  5.80421003E-02, -4.32123528E-05,
                    1.58792094E-08, -2.32209543E-12, -2.24284673E+04,
                    2.03680990E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.06472678E+01,  1.48526500E-02, -5.25104875E-06,
                    8.33619219E-10, -4.91256069E-14, -2.88719869E+04,
                    -7.95951389E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.352,
                                well_depth=436.4,
                                rot_relax=1.0),
        note='8/2/95THERM')

species(name='IC3H5O2HCHO',
        atoms='H:7 C:4 O:3',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.05984770E+00,  5.82331716E-02, -4.37672100E-05,
                    1.63249918E-08, -2.43462051E-12, -1.63496250E+04,
                    2.13687921E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.06288832E+01,  1.48625539E-02, -5.25305276E-06,
                    8.33772951E-10, -4.91277401E-14, -2.27589076E+04,
                    -7.82962888E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.352,
                                well_depth=436.4,
                                rot_relax=1.0),
        note='8/2/95THERM')

species(name='CH2CCH2OH',
        atoms='H:5 C:3 O:1',
        thermo=(NASA([300.00, 1372.00],
                    [ 2.88422544E+00,  2.42428071E-02, -1.14152268E-05,
                    1.71775334E-09,  1.42177454E-13,  1.17935615E+04,
                    1.52102335E+01]),
                NASA([1372.00, 5000.00],
                    [ 9.70702027E+00,  1.13972660E-02, -3.77993962E-06,
                    5.75209277E-10, -3.29229125E-14,  9.13212884E+03,
                    -2.25012933E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.997,
                                well_depth=481.5,
                                dipole=1.7,
                                rot_relax=1.0),
        note='9/8/95THERM')

species(name='TC4H8CHO',
        atoms='H:9 C:5 O:1',
        thermo=(NASA([300.00, 1397.00],
                    [-9.58078294E-01,  6.42003258E-02, -4.70776827E-05,
                    1.75737698E-08, -2.64896151E-12, -6.86582501E+03,
                    3.33781112E+01]),
                NASA([1397.00, 5000.00],
                    [ 1.79663933E+01,  1.94207117E-02, -6.67409451E-06,
                    1.03969221E-09, -6.04702651E-14, -1.33368585E+04,
                    -6.79819424E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.778,
                                well_depth=476.0,
                                dipole=2.6,
                                rot_relax=1.0),
        note='9/7/95THERM')

species(name='O2C4H8CHO',
        atoms='H:9 C:5 O:3',
        thermo=(NASA([300.00, 1395.00],
                    [ 1.91847699E+00,  6.67245869E-02, -4.80871046E-05,
                    1.78588690E-08, -2.71163880E-12, -2.49837984E+04,
                    2.38577867E+01]),
                NASA([1395.00, 5000.00],
                    [ 2.12629904E+01,  2.14072282E-02, -7.38342949E-06,
                    1.15281523E-09, -6.71508438E-14, -3.16854524E+04,
                    -7.99828703E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.942,
                                well_depth=588.531),
        note='9/7/95THERM')

species(name='O2HC4H8CO',
        atoms='H:9 C:5 O:3',
        thermo=(NASA([300.00, 1394.00],
                    [ 1.82607262E+00,  6.93466111E-02, -4.93125140E-05,
                    1.69848340E-08, -2.26117657E-12, -2.46578311E+04,
                    2.41167544E+01]),
                NASA([1394.00, 5000.00],
                    [ 2.38219630E+01,  1.91411448E-02, -6.67919154E-06,
                    1.05127303E-09, -6.15876805E-14, -3.23093973E+04,
                    -9.42580755E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.942,
                                well_depth=588.531),
        note='9/7/95THERM')

species(name='C3H5OH',
        atoms='H:6 C:3 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 3.15011905E+00,  1.28538274E-02,  4.28438434E-05,
                    -6.67818707E-08,  2.80408237E-11, -1.66413668E+04,
                    1.35066359E+01]),
                NASA([1000.00, 6000.00],
                    [ 8.72477114E+00,  1.63942712E-02, -5.90852993E-06,
                    9.53262253E-10, -5.70318010E-14, -1.90496618E+04,
                    -1.97198674E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.997,
                                well_depth=481.5,
                                dipole=1.7,
                                rot_relax=1.0),
        note='T06/10')

species(name='TIC4H7Q2-I',
        atoms='H:9 C:4 O:4',
        thermo=(NASA([300.00, 1400.00],
                    [ 4.48426361E+00,  6.61225007E-02, -5.27349018E-05,
                    2.18215585E-08, -3.66788946E-12, -1.98906586E+04,
                    1.26719614E+01]),
                NASA([1400.00, 5000.00],
                    [ 2.33848631E+01,  1.87070035E-02, -6.44021945E-06,
                    1.00428123E-09, -5.84468189E-14, -2.61180902E+04,
                    -8.76610135E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.009,
                                well_depth=600.078),
        note='5/6/96THERM')

species(name='IIC4H7Q2-T',
        atoms='H:9 C:4 O:4',
        thermo=(NASA([300.00, 1377.00],
                    [ 8.16274487E+00,  4.34463050E-02, -1.76972456E-05,
                    4.88790666E-10,  9.03915465E-13, -1.96501749E+04,
                    2.62067299E-01]),
                NASA([1377.00, 5000.00],
                    [ 2.15070321E+01,  2.05359839E-02, -7.12383399E-06,
                    1.11655053E-09, -6.52112103E-14, -2.51117508E+04,
                    -7.43379783E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.009,
                                well_depth=600.078),
        note='7/15/96THERM')

species(name='IIC4H7Q2-I',
        atoms='H:9 C:4 O:4',
        thermo=(NASA([300.00, 1394.00],
                    [ 4.93055661E+00,  6.05819201E-02, -4.23665566E-05,
                    1.49122008E-08, -2.10978665E-12, -1.68415495E+04,
                    1.36228018E+01]),
                NASA([1394.00, 5000.00],
                    [ 2.30500244E+01,  1.92149194E-02, -6.66622576E-06,
                    1.04495725E-09, -6.10370520E-14, -2.32086881E+04,
                    -8.39949885E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.009,
                                well_depth=600.078),
        note='7/15/96THERM')

species(name='CH2O2H',
        atoms='H:3 C:1 O:2',
        thermo=(NASA([300.00, 1357.00],
                    [ 4.47228333E+00,  1.33401095E-02, -5.92919725E-06,
                    4.44481025E-10,  2.12699899E-13,  5.67413711E+03,
                    4.72608208E+00]),
                NASA([1357.00, 5000.00],
                    [ 9.10784249E+00,  5.27260434E-03, -1.88170543E-06,
                    3.00561364E-10, -1.77865959E-14,  3.77440183E+03,
                    -2.11741044E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.496,
                                well_depth=238.4,
                                rot_relax=1.5),
        note='1/14/5THERM')

species(name='C4H4',
        atoms='H:4 C:4',
        thermo=(NASA([300.00, 1000.00],
                    [-1.91524790E+00,  5.27508780E-02, -7.16559440E-05,
                    5.50724230E-08, -1.72862280E-11,  3.29785040E+04,
                    3.14199830E+01]),
                NASA([1000.00, 3000.00],
                    [ 6.65070920E+00,  1.61294340E-02, -7.19388750E-06,
                    1.49817870E-09, -1.18641100E-13,  3.11959920E+04,
                    -9.79521180E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.18,
                                well_depth=357.0,
                                rot_relax=1.0),
        note='H6W/94')

species(name='C4H3-I',
        atoms='H:3 C:4',
        thermo=(NASA([300.00, 1000.00],
                    [ 2.08304120E+00,  4.08342740E-02, -6.21596850E-05,
                    5.16793580E-08, -1.70291840E-11,  5.80051290E+04,
                    1.36174620E+01]),
                NASA([1000.00, 3000.00],
                    [ 9.09781650E+00,  9.22071190E-03, -3.38784410E-06,
                    4.91604980E-10, -1.45297800E-14,  5.66005740E+04,
                    -1.98025970E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.18,
                                well_depth=357.0,
                                rot_relax=1.0),
        note='AB1/93')

species(name='C4H3-N',
        atoms='H:3 C:4',
        thermo=(NASA([300.00, 1000.00],
                    [-3.16841130E-01,  4.69121000E-02, -6.80938100E-05,
                    5.31799210E-08, -1.65230050E-11,  6.24761990E+04,
                    2.46225590E+01]),
                NASA([1000.00, 3000.00],
                    [ 5.43282790E+00,  1.68609810E-02, -9.43131090E-06,
                    2.57038950E-09, -2.74563090E-13,  6.16006800E+04,
                    -1.56739810E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.18,
                                well_depth=357.0,
                                rot_relax=1.0),
        note='H6W/94')

species(name='C4H612',
        atoms='H:6 C:4',
        thermo=(NASA([300.00, 1000.00],
                    [ 1.02346700E+00,  3.49591900E-02, -2.20090500E-05,
                    6.94227200E-09, -7.87918700E-13,  1.81179900E+04,
                    1.97506600E+01]),
                NASA([1000.00, 3000.00],
                    [ 1.78155700E+01, -4.25750200E-03,  1.05118500E-05,
                    -4.47384400E-09,  5.84813800E-13,  1.26734200E+04,
                    -6.98266200E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.18,
                                well_depth=357.0,
                                rot_relax=1.0),
        note='A8/83')

species(name='C4H2',
        atoms='H:2 C:4',
        thermo=(NASA([300.00, 1000.00],
                    [ 1.05439780E+00,  4.16269600E-02, -6.58717840E-05,
                    5.32570750E-08, -1.66831620E-11,  5.41852110E+04,
                    1.48665910E+01]),
                NASA([1000.00, 3000.00],
                    [ 9.15763280E+00,  5.54305180E-03, -1.35916040E-06,
                    1.87800750E-11,  2.31895360E-14,  5.25880390E+04,
                    -2.37114600E+01])),
        transport=gas_transport(geom='linear',
                                diam=4.72,
                                well_depth=357.0,
                                rot_relax=1.0),
        note='D11/99')

species(name='C4H5-I',
        atoms='H:5 C:4',
        thermo=(NASA([300.00, 1000.00],
                    [-1.99329000E-02,  3.80056720E-02, -2.75594500E-05,
                    7.78355510E-09,  4.02093830E-13,  3.74962230E+04,
                    2.43942410E+01]),
                NASA([1000.00, 3000.00],
                    [ 1.02290920E+01,  9.48501380E-03, -9.04064450E-08,
                    -1.25961000E-09,  2.47814680E-13,  3.46428120E+04,
                    -2.85645290E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.18,
                                well_depth=357.0,
                                rot_relax=1.0),
        note='H6W/94')

species(name='C4H5-N',
        atoms='H:5 C:4',
        thermo=(NASA([300.00, 1000.00],
                    [ 1.63053210E-01,  3.98301370E-02, -3.40001280E-05,
                    1.51472330E-08, -2.46658250E-12,  4.14297660E+04,
                    2.35361630E+01]),
                NASA([1000.00, 3000.00],
                    [ 9.85019780E+00,  1.07790080E-02, -1.36721250E-06,
                    -7.72005350E-10,  1.83663140E-13,  3.88403010E+04,
                    -2.60018460E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.18,
                                well_depth=357.0,
                                rot_relax=1.0),
        note='H6W/94')

species(name='CH3CHCHCO',
        atoms='H:5 C:4 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [-1.08199860E+00,  3.64929760E-02, -1.52255950E-05,
                    -5.62607170E-18,  2.16113750E-21,  3.56713230E+03,
                    3.27142550E+01]),
                NASA([1000.00, 6000.00],
                    [ 8.90967920E+00,  1.34364140E-02, -7.62977390E-07,
                    -1.69114810E-09,  2.95540440E-13,  1.48898740E+03,
                    -1.79662460E+01])),
        transport=gas_transport(geom='linear',
                                diam=5.18,
                                well_depth=357.0,
                                rot_relax=1.0),
        note='T03/97')

species(name='CH2CHCHCHO',
        atoms='H:5 C:4 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [-1.08199860E+00,  3.64929760E-02, -1.52255950E-05,
                    -5.62607170E-18,  2.16113750E-21,  3.56713230E+03,
                    3.27142550E+01]),
                NASA([1000.00, 6000.00],
                    [ 8.90967920E+00,  1.34364140E-02, -7.62977390E-07,
                    -1.69114810E-09,  2.95540440E-13,  1.48898740E+03,
                    -1.79662460E+01])),
        transport=gas_transport(geom='linear',
                                diam=5.18,
                                well_depth=357.0,
                                rot_relax=1.0),
        note='T03/97')

species(name='C4H6O25',
        atoms='H:6 C:4 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 2.67053463E+00,  4.92586420E-03,  8.86967406E-05,
                    -1.26219194E-07,  5.23991321E-11, -1.46572472E+04,
                    1.45722395E+01]),
                NASA([1000.00, 5000.00],
                    [ 8.60658242E+00,  2.08310051E-02, -8.42229481E-06,
                    1.56717640E-09, -1.09391202E-13, -1.76177415E+04,
                    -2.32464750E+01])),
        transport=gas_transport(geom='linear',
                                diam=5.18,
                                well_depth=357.0,
                                rot_relax=1.0),
        note='T3/97')

species(name='C2H3CHOCH2',
        atoms='H:6 C:4 O:1',
        thermo=(NASA([300.00, 1000.00],
                    [ 7.97985440E-01,  3.44034320E-02, -1.24598510E-05,
                    -5.18062790E-18,  1.99359540E-21, -6.48927540E+02,
                    2.18896980E+01]),
                NASA([1000.00, 3000.00],
                    [-4.72093360E+00,  3.91413780E-02, -6.52872650E-06,
                    -7.68209500E-09,  2.51473310E-12,  1.75352252E+03,
                    5.17190420E+01])),
        transport=gas_transport(geom='linear',
                                diam=5.18,
                                well_depth=357.0,
                                rot_relax=1.0),
        note='A8/83')

species(name='C4H5-2',
        atoms='H:5 C:4',
        thermo=(NASA([300.00, 1000.00],
                    [ 2.96962800E+00,  2.44422450E-02, -9.12514240E-06,
                    -4.24668710E-18,  1.63047280E-21,  3.55033160E+04,
                    1.20360510E+01]),
                NASA([1000.00, 3000.00],
                    [ 1.45381710E+01, -8.56770560E-03,  2.35595240E-05,
                    -1.36763790E-08,  2.44369270E-12,  3.32590950E+04,
                    -4.53694970E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.18,
                                well_depth=357.0,
                                rot_relax=1.0),
        note='H6W/94')

species(name='C4H6-2',
        atoms='H:6 C:4',
        thermo=(NASA([300.00, 1000.00],
                    [ 2.13733380E+00,  2.64862290E-02, -9.05687110E-06,
                    -5.53863970E-19,  2.12818840E-22,  1.57109020E+04,
                    1.35294260E+01]),
                NASA([1000.00, 3000.00],
                    [ 9.03381330E+00,  8.21245100E-03,  7.17539520E-06,
                    -5.88343340E-09,  1.03439150E-12,  1.43350680E+04,
                    -2.09857620E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.18,
                                well_depth=357.0,
                                rot_relax=1.0),
        note='A8/83')

species(name='C4H6O23',
        atoms='H:6 C:4 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 2.67053463E+00,  4.92586420E-03,  8.86967406E-05,
                    -1.26219194E-07,  5.23991321E-11, -1.02787872E+04,
                    1.45722395E+01]),
                NASA([1000.00, 5000.00],
                    [ 8.60658242E+00,  2.08310051E-02, -8.42229481E-06,
                    1.56717640E-09, -1.09391202E-13, -1.32392815E+04,
                    -2.32464750E+01])),
        transport=gas_transport(geom='linear',
                                diam=5.18,
                                well_depth=357.0,
                                rot_relax=1.0),
        note='T3/97')

species(name='CH3CHCHCHO',
        atoms='H:6 C:4 O:1',
        thermo=(NASA([298.15, 1000.00],
                    [-1.55577660E+00,  4.09640630E-02, -1.69868810E-05,
                    -6.00928140E-18,  2.31368530E-21, -1.41394920E+04,
                    3.74707580E+01]),
                NASA([1000.00, 3000.00],
                    [ 1.98794540E+01, -2.09130550E-02,  4.45360508E-05,
                    -2.60374870E-08,  4.86836120E-12, -1.95278768E+04,
                    -6.87200320E+01])),
        transport=gas_transport(geom='linear',
                                diam=5.18,
                                well_depth=357.0,
                                rot_relax=1.0),
        note='T5/92')

species(name='C4H4O',
        atoms='H:4 C:4 O:1',
        thermo=(NASA([200.00, 1000.00],
                    [ 8.47469463E-01,  1.31773796E-02,  5.99735901E-05,
                    -9.71562904E-08,  4.22733796E-11, -5.36785445E+03,
                    2.14945172E+01]),
                NASA([1000.00, 6000.00],
                    [ 9.38935003E+00,  1.40291241E-02, -5.07755110E-06,
                    8.24137332E-10, -4.95319963E-14, -8.68241814E+03,
                    -2.79162920E+01])),
        transport=gas_transport(geom='linear',
                                diam=5.18,
                                well_depth=357.0,
                                rot_relax=1.0),
        note='T03/97')

species(name='H2C4O',
        atoms='H:2 C:4 O:1',
        thermo=(NASA([300.00, 1000.00],
                    [ 4.81097100E+00,  1.31399900E-02,  9.86507300E-07,
                    -6.12072000E-09,  1.64000300E-12,  2.54580300E+04,
                    2.11342400E+00]),
                NASA([1000.00, 4000.00],
                    [ 1.02688800E+01,  4.89616400E-03, -4.88508100E-07,
                    -2.70856600E-10,  5.10701300E-14,  2.34690300E+04,
                    -2.81598500E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.18,
                                well_depth=357.0,
                                rot_relax=1.0),
        note='120189')

species(name='NC3H7OH',
        atoms='H:8 C:3 O:1',
        thermo=(NASA([300.00, 1400.00],
                    [-2.40806690E-01,  4.34611469E-02, -2.75532776E-05,
                    8.90741295E-09, -1.15919009E-12, -3.23582496E+04,
                    2.82034963E+01]),
                NASA([1400.00, 5000.00],
                    [ 1.15330230E+01,  1.73694541E-02, -5.89694949E-06,
                    9.10813224E-10, -5.26492908E-14, -3.65903879E+04,
                    -3.55321352E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.997,
                                well_depth=481.5,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/18/8THERM')

species(name='C3H6OH-3',
        atoms='H:7 C:3 O:1',
        thermo=(NASA([300.00, 1401.00],
                    [ 4.79055844E-01,  3.90840706E-02, -2.52934637E-05,
                    8.39036258E-09, -1.12650227E-12, -7.73699631E+03,
                    2.70272522E+01]),
                NASA([1401.00, 5000.00],
                    [ 1.11187211E+01,  1.52263759E-02, -5.16931790E-06,
                    7.98435556E-10, -4.61541396E-14, -1.15296599E+04,
                    -3.04632777E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.81,
                                well_depth=403.4,
                                rot_relax=1.0),
        note='7/18/8THERM')

species(name='C3H6OH-2',
        atoms='H:7 C:3 O:1',
        thermo=(NASA([300.00, 1530.00],
                    [ 1.38218609E+00,  3.22641651E-02, -1.39820294E-05,
                    9.59442093E-10,  5.73234299E-13, -8.29246569E+03,
                    2.39928393E+01]),
                NASA([1530.00, 5000.00],
                    [ 1.08612250E+01,  1.49944440E-02, -4.99956622E-06,
                    7.63352693E-10, -4.37872385E-14, -1.20025758E+04,
                    -2.85286286E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.81,
                                well_depth=403.4,
                                rot_relax=1.0),
        note='7/18/8THERM')

species(name='C3H6OH-1',
        atoms='H:7 C:3 O:1',
        thermo=(NASA([300.00, 1401.00],
                    [ 1.31781034E+00,  3.85018240E-02, -2.61358793E-05,
                    9.39954865E-09, -1.40537401E-12, -1.09450432E+04,
                    2.11274386E+01]),
                NASA([1401.00, 5000.00],
                    [ 1.14914460E+01,  1.48556953E-02, -5.02888291E-06,
                    7.75157618E-10, -4.47423059E-14, -1.45014891E+04,
                    -3.35712772E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.81,
                                well_depth=403.4,
                                rot_relax=1.0),
        note='7/18/8THERM')

species(name='IC3H7OH',
        atoms='H:8 C:3 O:1',
        thermo=(NASA([300.00, 1396.00],
                    [-1.76838921E-01,  4.52559563E-02, -3.19190718E-05,
                    1.23062878E-08, -2.01412538E-12, -3.45587071E+04,
                    2.63322249E+01]),
                NASA([1396.00, 5000.00],
                    [ 1.15109036E+01,  1.75006903E-02, -5.96531139E-06,
                    9.23747459E-10, -5.34895475E-14, -3.86375916E+04,
                    -3.63855211E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.036,
                                well_depth=459.5,
                                dipole=1.7,
                                rot_relax=1.0),
        note='2/14/3THERM')

species(name='IC3H6OH',
        atoms='H:7 C:3 O:1',
        thermo=(NASA([300.00, 1396.00],
                    [ 5.84672920E-01,  4.07370189E-02, -2.94865043E-05,
                    1.16950656E-08, -1.96228356E-12, -9.94491545E+03,
                    2.49591243E+01]),
                NASA([1396.00, 5000.00],
                    [ 1.10944203E+01,  1.53549108E-02, -5.23574640E-06,
                    8.10964124E-10, -4.69665855E-14, -1.35725752E+04,
                    -3.12908162E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.81,
                                well_depth=403.4,
                                rot_relax=1.0),
        note='2/14/3THERM')

species(name='CH3CHCHOH',
        atoms='H:6 C:3 O:1',
        thermo=(NASA([300.00, 1404.00],
                    [-3.53977226E-02,  4.34969453E-02, -3.74479918E-05,
                    1.70906074E-08, -3.13775054E-12, -2.02502608E+04,
                    2.41528201E+01]),
                NASA([1404.00, 5000.00],
                    [ 1.11222064E+01,  1.27745410E-02, -4.25315532E-06,
                    6.48216484E-10, -3.71190850E-14, -2.36690795E+04,
                    -3.41335182E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='2/3/9THERM')

#-------------------------------------------------------------------------------
# Reaction data
#-------------------------------------------------------------------------------

# Reaction 1
reaction('H + O2 <=> O + OH', [1.040000e+14, 0.0, 15286.0])

# Reaction 2
reaction('O + H2 <=> H + OH', [5.080000e+04, 2.67, 6292.0])

# Reaction 3
reaction('OH + H2 <=> H + H2O', [4.380000e+13, 0.0, 6990.0])

# Reaction 4
reaction('O + H2O <=> OH + OH', [2.970000e+06, 2.02, 13400.0])

# Reaction 5
three_body_reaction('H2 + M <=> H + H + M', [4.577000e+19, -1.4, 104400.0],
                    efficiencies='CO2:3.8 CO:1.9 H2:2.5 H2O:12.0 CH4:2.0 C2H6:3.0 HE:0.83')

# Reaction 6
three_body_reaction('O + O + M <=> O2 + M', [6.165000e+15, -0.5, 0.0],
                    efficiencies='CO2:3.8 CO:1.9 H2:2.5 H2O:12.0 AR:0.83 CH4:2.0 C2H6:3.0 HE:0.83')

# Reaction 7
three_body_reaction('O + H + M <=> OH + M', [4.714000e+18, -1.0, 0.0],
                    efficiencies='CO2:2.0 CO:1.5 H2:2.5 H2O:12.0 AR:0.75 CH4:2.0 C2H6:3.0 HE:0.75')

# Reaction 8
three_body_reaction('H + OH + M <=> H2O + M', [3.500000e+22, -2.0, 0.0],
                    efficiencies='H2:0.73 C2H6:3.0 H2O:3.65 AR:0.38 CH4:2.0')

# Reaction 9
falloff_reaction('H + O2 (+ M) <=> HO2 (+ M)',
                kf=[4.650000e+12, 0.44, 0.0],
                kf0=[1.737000e+19, -1.23, 0.0],
                efficiencies='CO2:3.8 CO:1.9 H2:1.3 H2O:10.0 AR:0.0 CH4:2.0 C2H6:3.0 HE:0.0',
                falloff=Troe(A=0.67, T3=1e-30, T1=1e+30, T2=1e+30))

# Reaction 10
falloff_reaction('H + O2 (+ AR) <=> HO2 (+ AR)',
                kf=[4.650000e+12, 0.44, 0.0],
                kf0=[6.810000e+18, -1.2, 0.0],
                falloff=Troe(A=0.7, T3=1e-30, T1=1e+30, T2=1e+30))

# Reaction 11
falloff_reaction('H + O2 (+ HE) <=> HO2 (+ HE)',
                kf=[4.650000e+12, 0.44, 0.0],
                kf0=[9.192000e+18, -1.2, 0.0],
                falloff=Troe(A=0.59, T3=1e-30, T1=1e+30, T2=1e+30))

# Reaction 12
reaction('HO2 + H <=> OH + OH', [7.079000e+13, 0.0, 295.0])

# Reaction 13
reaction('H2 + O2 <=> H + HO2', [5.176000e+05, 2.433, 53502.0])

# Reaction 14
reaction('HO2 + O <=> OH + O2', [3.250000e+13, 0.0, 0.0])

# Reaction 15
reaction('HO2 + OH <=> H2O + O2', [2.456000e+13, 0.0, -497.0])

# Reaction 16
reaction('HO2 + HO2 <=> H2O2 + O2', [1.300000e+11, 0.0, -1630.0],
        options='duplicate')

# Reaction 17
reaction('HO2 + HO2 <=> H2O2 + O2', [3.658000e+14, 0.0, 12000.0],
        options='duplicate')

# Reaction 18
falloff_reaction('H2O2 (+ H2O) <=> OH + OH (+ H2O)',
                kf=[2.000000e+12, 0.9, 48749.0],
                kf0=[1.865000e+25, -2.3, 48749.0],
                falloff=Troe(A=0.51, T3=1e-30, T1=1e+30))

# Reaction 19
falloff_reaction('H2O2 (+ M) <=> OH + OH (+ M)',
                kf=[2.000000e+12, 0.9, 48749.0],
                kf0=[2.490000e+24, -2.3, 48749.0],
                efficiencies='CO2:1.6 CO:2.8 H2O2:7.7 H2:3.7 H2O:0.0 N2:1.5 O2:1.2 HE:0.65',
                falloff=Troe(A=0.43, T3=1e-30, T1=1e+30))

# Reaction 20
reaction('H2O2 + H <=> H2O + OH', [2.410000e+13, 0.0, 3970.0])

# Reaction 21
reaction('H2O2 + H <=> H2 + HO2', [2.150000e+10, 1.0, 6000.0])

# Reaction 22
reaction('H2O2 + O <=> OH + HO2', [9.550000e+06, 2.0, 3970.0])

# Reaction 23
reaction('H2O2 + OH <=> H2O + HO2', [1.740000e+12, 0.0, 318.0],
        options='duplicate')

# Reaction 24
reaction('H2O2 + OH <=> H2O + HO2', [7.590000e+13, 0.0, 7269.0],
        options='duplicate')

# Reaction 25
falloff_reaction('CO + O (+ M) <=> CO2 (+ M)',
                kf=[1.362000e+10, 0.0, 2384.0],
                kf0=[1.173000e+24, -2.79, 4191.0],
                efficiencies='CO2:3.6 CO:1.75 H2:2.0 H2O:12.0 AR:0.7 HE:0.7')

# Reaction 26
reaction('CO + O2 <=> CO2 + O', [1.119000e+12, 0.0, 47700.0])

# Reaction 27
reaction('CO + OH <=> CO2 + H', [7.015000e+04, 2.053, -355.7],
        options='duplicate')

# Reaction 28
reaction('CO + OH <=> CO2 + H', [5.757000e+12, -0.664, 331.8],
        options='duplicate')

# Reaction 29
reaction('CO + HO2 <=> CO2 + OH', [1.570000e+05, 2.18, 17940.0])

# Reaction 30
three_body_reaction('HCO + M <=> H + CO + M', [5.700000e+11, 0.66, 14870.0],
                    efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:12.0 CH4:2.0 C2H6:3.0')

# Reaction 31
reaction('HCO + O2 <=> CO + HO2', [7.580000e+12, 0.0, 410.0])

# Reaction 32
reaction('HCO + H <=> CO + H2', [7.340000e+13, 0.0, 0.0])

# Reaction 33
reaction('HCO + O <=> CO + OH', [3.020000e+13, 0.0, 0.0])

# Reaction 34
reaction('HCO + O <=> CO2 + H', [3.000000e+13, 0.0, 0.0])

# Reaction 35
reaction('HCO + OH <=> CO + H2O', [1.020000e+14, 0.0, 0.0])

# Reaction 36
reaction('HCO + HO2 => CO2 + H + OH', [3.000000e+13, 0.0, 0.0])

# Reaction 37
reaction('HCO + HCO => H2 + CO + CO', [3.000000e+12, 0.0, 0.0])

# Reaction 38
reaction('HCO + CH3 <=> CH4 + CO', [2.650000e+13, 0.0, 0.0])

# Reaction 39
reaction('CH2O + O2 <=> HCO + HO2', [8.070000e+15, 0.0, 53420.0])

# Reaction 40
reaction('HCO + O2 <=> O2CHO', [1.200000e+11, 0.0, -1100.0])

# Reaction 41
reaction('CH2O + O2CHO <=> HCO + HO2CHO', [1.990000e+12, 0.0, 11660.0])

# Reaction 42
reaction('OCHO + OH <=> HO2CHO', [2.000000e+13, 0.0, 0.0])

# Reaction 43
reaction('H + CO2 <=> OCHO', [7.500000e+13, 0.0, 29000.0])

# Reaction 44
reaction('HCO + HCO <=> CH2O + CO', [1.800000e+13, 0.0, 0.0])

# Reaction 45
three_body_reaction('H + O + M <=> OH* + M', [1.500000e+13, 0.0, 5975.0],
                    efficiencies='H2:1.0 H2O:6.5 AR:0.35 N2:0.4 O2:0.4')

# Reaction 46
reaction('OH* + H2O <=> OH + H2O', [5.930000e+12, 0.5, -860.0])

# Reaction 47
reaction('OH* + H2 <=> OH + H2', [2.950000e+12, 0.5, -444.0])

# Reaction 48
reaction('OH* + N2 <=> OH + N2', [1.080000e+11, 0.5, -1242.0])

# Reaction 49
reaction('OH* + OH <=> OH + OH', [6.010000e+12, 0.5, -764.0])

# Reaction 50
reaction('OH* + H <=> OH + H', [1.310000e+12, 0.5, -167.0])

# Reaction 51
reaction('OH* + AR <=> OH + AR', [1.690000e+12, 0.0, 4135.0])

# Reaction 52
reaction('OH* <=> OH', [1.450000e+06, 0.0, 0.0])

# Reaction 53
reaction('OH* + O2 <=> OH + O2', [2.100000e+12, 0.5, -478.0])

# Reaction 54
reaction('OH* + CO2 <=> OH + CO2', [2.750000e+12, 0.5, -968.0])

# Reaction 55
reaction('OH* + CO <=> OH + CO', [3.230000e+12, 0.5, -787.0])

# Reaction 56
reaction('OH* + CH4 <=> OH + CH4', [3.360000e+12, 0.5, -635.0])

# Reaction 57
reaction('CH + O2 <=> CO + OH*', [4.040000e+13, 0.0, 0.0])

# Reaction 58
reaction('C2H + O <=> CO + CH*', [6.200000e+12, 0.0, 0.0])

# Reaction 59
three_body_reaction('C + H + M <=> CH* + M', [6.000000e+14, 0.0, 6940.0])

# Reaction 60
reaction('C2H + O2 <=> CO2 + CH*', [2.170000e+10, 0.0, 0.0])

# Reaction 61
reaction('CH* + AR <=> CH + AR', [4.000000e+11, 0.5, 0.0])

# Reaction 62
reaction('CH* + H2O <=> CH + H2O', [5.300000e+13, 0.0, 0.0])

# Reaction 63
reaction('CH* + CO <=> CH + CO', [2.440000e+12, 0.5, 0.0])

# Reaction 64
reaction('CH* + CO2 <=> CH + CO2', [2.410000e-01, 4.3, -1694.0])

# Reaction 65
reaction('CH* + O2 <=> CH + O2', [2.480000e+06, 2.14, -1720.0])

# Reaction 66
reaction('CH* + H2 <=> CH + H2', [1.470000e+14, 0.0, 1361.0])

# Reaction 67
reaction('CH* + CH4 <=> CH + CH4', [1.730000e+13, 0.0, 167.0])

# Reaction 68
reaction('CH* <=> CH', [1.860000e+06, 0.0, 0.0])

# Reaction 69
reaction('CH* + N2 <=> CH + N2', [3.030000e+02, 3.4, -381.0])

# Reaction 70
falloff_reaction('HCO + H (+ M) <=> CH2O (+ M)',
                kf=[1.090000e+12, 0.48, -260.0],
                kf0=[1.350000e+24, -2.57, 1425.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0 HE:0.7',
                falloff=Troe(A=0.7824, T3=271.0, T1=2755.0, T2=6570.0))

# Reaction 71
falloff_reaction('CO + H2 (+ M) <=> CH2O (+ M)',
                kf=[4.300000e+07, 1.5, 79600.0],
                kf0=[5.070000e+27, -3.42, 84348.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0 HE:0.7',
                falloff=Troe(A=0.932, T3=197.0, T1=1540.0, T2=10300.0))

# Reaction 72
reaction('CH2O + OH <=> HCO + H2O', [7.820000e+07, 1.63, -1055.0])

# Reaction 73
reaction('CH2O + H <=> HCO + H2', [5.740000e+07, 1.9, 2740.0])

# Reaction 74
reaction('CH2O + O <=> HCO + OH', [6.260000e+09, 1.15, 2260.0])

# Reaction 75
reaction('CH2O + CH3 <=> HCO + CH4', [3.830000e+01, 3.36, 4312.0])

# Reaction 76
reaction('CH2O + HO2 <=> HCO + H2O2', [1.880000e+04, 2.7, 11520.0])

# Reaction 77
reaction('CH2O + OH <=> HOCH2O', [4.500000e+15, -1.1, 0.0])

# Reaction 78
reaction('HOCH2O <=> HOCHO + H', [1.000000e+14, 0.0, 14900.0])

# Reaction 79
reaction('HOCHO <=> CO + H2O', [2.450000e+12, 0.0, 60470.0])

# Reaction 80
reaction('HOCHO <=> CO2 + H2', [2.950000e+09, 0.0, 48520.0])

# Reaction 81
reaction('OCHO + HO2 <=> HOCHO + O2', [3.500000e+10, 0.0, -3275.0])

# Reaction 82
reaction('HOCHO + OH => H2O + CO2 + H', [2.620000e+06, 2.06, 916.0])

# Reaction 83
reaction('HOCHO + OH => H2O + CO + OH', [1.850000e+07, 1.51, -962.0])

# Reaction 84
reaction('HOCHO + H => H2 + CO2 + H', [4.240000e+06, 2.1, 4868.0])

# Reaction 85
reaction('HOCHO + H => H2 + CO + OH', [6.030000e+13, -0.35, 2988.0])

# Reaction 86
reaction('HOCHO + CH3 => CH4 + CO + OH', [3.900000e-07, 5.8, 2200.0])

# Reaction 87
reaction('OCHO + H2O2 <=> HOCHO + HO2', [2.400000e+12, 0.0, 10000.0])

# Reaction 88
reaction('HOCHO + HO2 => H2O2 + CO + OH', [1.000000e+12, 0.0, 11920.0])

# Reaction 89
reaction('HOCHO + O => CO + OH + OH', [1.770000e+18, -1.9, 2975.0])

# Reaction 90
reaction('CH2O + OCHO <=> HOCHO + HCO', [5.600000e+12, 0.0, 13600.0])

# Reaction 91
falloff_reaction('CH3O (+ M) <=> CH2O + H (+ M)',
                kf=[6.800000e+13, 0.0, 26170.0],
                kf0=[1.867000e+25, -3.0, 24307.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.9, T3=2500.0, T1=1300.0, T2=1e+99))

# Reaction 92
reaction('CH3O + O2 <=> CH2O + HO2', [4.380000e-19, 9.5, -5501.0])

# Reaction 93
reaction('CH2O + CH3O <=> CH3OH + HCO', [6.620000e+11, 0.0, 2294.0])

# Reaction 94
reaction('CH3 + CH3OH <=> CH4 + CH3O', [1.440000e+01, 3.1, 6935.0])

# Reaction 95
reaction('CH3O + CH3 <=> CH2O + CH4', [1.200000e+13, 0.0, 0.0])

# Reaction 96
reaction('CH3O + H <=> CH2O + H2', [2.000000e+13, 0.0, 0.0])

# Reaction 97
reaction('CH3O + HO2 <=> CH2O + H2O2', [3.010000e+11, 0.0, 0.0])

# Reaction 98
falloff_reaction('CH2O + H (+ M) <=> CH2OH (+ M)',
                kf=[5.400000e+11, 0.454, 3600.0],
                kf0=[1.270000e+32, -4.82, 6530.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.7187, T3=103.0, T1=1291.0, T2=4160.0))

# Reaction 99
reaction('CH2OH + O2 <=> CH2O + HO2', [1.510000e+15, -1.0, 0.0],
        options='duplicate')

# Reaction 100
reaction('CH2OH + O2 <=> CH2O + HO2', [2.410000e+14, 0.0, 5017.0],
        options='duplicate')

# Reaction 101
reaction('CH2OH + H <=> CH2O + H2', [6.000000e+12, 0.0, 0.0])

# Reaction 102
reaction('CH2OH + HO2 <=> CH2O + H2O2', [1.200000e+13, 0.0, 0.0])

# Reaction 103
reaction('CH2OH + HCO <=> CH2O + CH2O', [1.800000e+14, 0.0, 0.0])

# Reaction 104
reaction('CH2OH + CH3O <=> CH2O + CH3OH', [2.400000e+13, 0.0, 0.0])

# Reaction 105
reaction('CH3OH + HCO <=> CH2OH + CH2O', [9.630000e+03, 2.9, 13110.0])

# Reaction 106
reaction('OH + CH2OH <=> H2O + CH2O', [2.400000e+13, 0.0, 0.0])

# Reaction 107
reaction('O + CH2OH <=> OH + CH2O', [4.200000e+13, 0.0, 0.0])

# Reaction 108
reaction('CH2OH + CH2OH <=> CH2O + CH3OH', [3.000000e+12, 0.0, 0.0])

# Reaction 109
reaction('CH2OH + HO2 <=> HOCH2O + OH', [1.000000e+13, 0.0, 0.0])

# Reaction 110
reaction('CH2O + HO2 <=> OCH2O2H', [1.500000e+11, 0.0, 11900.0])

# Reaction 111
reaction('OCH2O2H <=> HOCH2O2', [3.000000e+11, 0.0, 8600.0])

# Reaction 112
reaction('HOCH2O2 + HO2 <=> HOCH2O2H + O2', [3.500000e+10, 0.0, -3275.0])

# Reaction 113
reaction('HOCH2O + OH <=> HOCH2O2H', [1.000000e+13, 0.0, 0.0])

# Reaction 114
falloff_reaction('CH3OH (+ M) <=> CH3 + OH (+ M)',
                kf=[2.084000e+18, -0.615, 92540.6],
                kf0=[1.500000e+43, -6.995, 97992.2],
                falloff=Troe(A=-0.4748, T3=35580.0, T1=1116.0, T2=9023.0))

# Reaction 115
falloff_reaction('CH3OH (+ M) <=> CH2(S) + H2O (+ M)',
                kf=[3.121000e+18, -1.017, 91712.0],
                kf0=[1.430000e+47, -8.227, 99417.1],
                falloff=Troe(A=2.545, T3=3290.0, T1=47320.0, T2=47110.0))

# Reaction 116
falloff_reaction('CH3OH (+ M) <=> CH2OH + H (+ M)',
                kf=[7.896000e-03, 5.038, 84467.4],
                kf0=[3.390000e+42, -7.244, 105230.3],
                falloff=Troe(A=-73.91, T3=37050.0, T1=41500.0, T2=5220.0))

# Reaction 117
reaction('CH3OH + H <=> CH2OH + H2', [3.070000e+05, 2.55, 5440.0])

# Reaction 118
reaction('CH3OH + H <=> CH3O + H2', [1.990000e+05, 2.56, 10300.0])

# Reaction 119
reaction('CH3OH + O <=> CH2OH + OH', [3.880000e+05, 2.5, 3080.0])

# Reaction 120
reaction('CH3OH + OH <=> CH2OH + H2O', [3.080000e+04, 2.65, -806.7])

# Reaction 121
reaction('CH3OH + OH <=> CH3O + H2O', [1.500000e+02, 3.03, -763.0])

# Reaction 122
reaction('CH3OH + O2 <=> CH2OH + HO2', [2.050000e+13, 0.0, 44900.0])

# Reaction 123
reaction('CH3OH + HO2 <=> CH2OH + H2O2', [1.080000e+04, 2.55, 10530.0])

# Reaction 124
reaction('CH3OH + CH3 <=> CH2OH + CH4', [3.190000e+01, 3.17, 7172.0])

# Reaction 125
reaction('CH3O + CH3OH <=> CH2OH + CH3OH', [3.000000e+11, 0.0, 4074.0])

# Reaction 126
reaction('CH3O + CH3O <=> CH3OH + CH2O', [6.030000e+13, 0.0, 0.0])

# Reaction 127
falloff_reaction('CH3 + H (+ M) <=> CH4 (+ M)',
                kf=[1.270000e+16, -0.63, 383.0],
                kf0=[2.477000e+33, -4.76, 2440.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0 HE:0.7',
                falloff=Troe(A=0.783, T3=74.0, T1=2941.0, T2=6964.0))

# Reaction 128
reaction('CH4 + H <=> CH3 + H2', [6.140000e+05, 2.5, 9587.0])

# Reaction 129
reaction('CH4 + OH <=> CH3 + H2O', [5.830000e+04, 2.6, 2190.0])

# Reaction 130
reaction('CH4 + O <=> CH3 + OH', [1.020000e+09, 1.5, 8600.0])

# Reaction 131
reaction('CH4 + HO2 <=> CH3 + H2O2', [1.695000e+01, 3.74, 21010.0])

# Reaction 132
reaction('CH4 + CH2 <=> CH3 + CH3', [2.460000e+06, 2.0, 8270.0])

# Reaction 133
pdep_arrhenius('CH3 + OH <=> CH2(S) + H2O',
            [(0.01, 'atm'), 4.936000e+14, -0.669, -445.8],
            [(0.1, 'atm'), 1.207000e+15, -0.778, -175.6],
            [(1.0, 'atm'), 5.282000e+17, -1.518, 1772.0],
            [(10.0, 'atm'), 4.788000e+23, -3.155, 7003.0],
            [(100.0, 'atm'), 8.433000e+19, -1.962, 8244.0])

# Reaction 134
pdep_arrhenius('CH3 + OH <=> CH2O + H2',
            [(0.01, 'atm'), 3.502000e+05, 1.441, -3244.0],
            [(0.1, 'atm'), 8.854000e+05, 1.327, -2975.0],
            [(1.0, 'atm'), 1.650000e+07, 0.973, -2010.0],
            [(10.0, 'atm'), 5.374000e+09, 0.287, 280.0],
            [(100.0, 'atm'), 9.494000e+18, -2.199, 9769.0])

# Reaction 135
pdep_arrhenius('CH3 + OH <=> CH2OH + H',
            [(0.01, 'atm'), 1.621000e+10, 0.965, 3214.0],
            [(0.1, 'atm'), 1.807000e+10, 0.95, 3247.0],
            [(1.0, 'atm'), 4.686000e+10, 0.833, 3566.0],
            [(10.0, 'atm'), 1.525000e+13, 0.134, 5641.0],
            [(100.0, 'atm'), 3.590000e+14, -0.186, 8601.0])

# Reaction 136
pdep_arrhenius('CH3 + OH <=> H + CH3O',
            [(0.01, 'atm'), 1.186000e+09, 1.016, 11940.0],
            [(0.1, 'atm'), 1.188000e+09, 1.016, 11940.0],
            [(1.0, 'atm'), 1.230000e+09, 1.011, 11950.0],
            [(10.0, 'atm'), 1.798000e+09, 0.965, 12060.0],
            [(100.0, 'atm'), 5.242000e+10, 0.551, 13070.0])

# Reaction 137
pdep_arrhenius('CH3 + OH <=> HCOH + H2',
            [(0.01, 'atm'), 8.674000e+08, 0.787, -3046.0],
            [(0.1, 'atm'), 3.115000e+09, 0.63, -2669.0],
            [(1.0, 'atm'), 1.557000e+11, 0.156, -1368.0],
            [(10.0, 'atm'), 1.704000e+21, -2.641, 6412.0],
            [(100.0, 'atm'), 7.250000e+20, -2.402, 9639.0])

# Reaction 138
reaction('HCOH + OH <=> HCO + H2O', [2.000000e+13, 0.0, 0.0])

# Reaction 139
reaction('HCOH + H <=> CH2O + H', [2.000000e+14, 0.0, 0.0])

# Reaction 140
reaction('HCOH + O => CO2 + H + H', [5.000000e+13, 0.0, 0.0])

# Reaction 141
reaction('HCOH + O => CO + OH + H', [3.000000e+13, 0.0, 0.0])

# Reaction 142
reaction('HCOH + O2 => CO2 + H + OH', [5.000000e+12, 0.0, 0.0])

# Reaction 143
reaction('HCOH + O2 <=> CO2 + H2O', [3.000000e+13, 0.0, 0.0])

# Reaction 144
reaction('CH3 + HO2 <=> CH3O + OH', [1.000000e+12, 0.269, -687.5])

# Reaction 145
reaction('CH3 + HO2 <=> CH4 + O2', [1.160000e+05, 2.23, -3022.0])

# Reaction 146
reaction('CH3 + O <=> CH2O + H', [5.540000e+13, 0.05, -136.0])

# Reaction 147
reaction('CH3 + O2 <=> CH3O + O', [7.546000e+12, 0.0, 28320.0])

# Reaction 148
reaction('CH3 + O2 <=> CH2O + OH', [2.641000e+00, 3.283, 8105.0])

# Reaction 149
falloff_reaction('CH3 + O2 (+ M) <=> CH3O2 (+ M)',
                kf=[7.812000e+09, 0.9, 0.0],
                kf0=[6.850000e+24, -3.0, 0.0],
                falloff=Troe(A=0.6, T3=1000.0, T1=70.0, T2=1700.0))

# Reaction 150
reaction('CH3O2 + CH2O <=> CH3O2H + HCO', [1.990000e+12, 0.0, 11660.0])

# Reaction 151
reaction('CH4 + CH3O2 <=> CH3 + CH3O2H', [9.600000e-01, 3.77, 17810.0])

# Reaction 152
reaction('CH3OH + CH3O2 <=> CH2OH + CH3O2H', [1.810000e+12, 0.0, 13710.0])

# Reaction 153
reaction('CH3O2 + CH3 <=> CH3O + CH3O', [5.080000e+12, 0.0, -1411.0])

# Reaction 154
reaction('CH3O2 + HO2 <=> CH3O2H + O2', [2.470000e+11, 0.0, -1570.0])

# Reaction 155
reaction('CH3O2 + CH3O2 => CH2O + CH3OH + O2', [3.110000e+14, -1.61, -1051.0])

# Reaction 156
reaction('CH3O2 + CH3O2 => O2 + CH3O + CH3O', [1.400000e+16, -1.61, 1860.0])

# Reaction 157
reaction('CH3O2 + H <=> CH3O + OH', [9.600000e+13, 0.0, 0.0])

# Reaction 158
reaction('CH3O2 + O <=> CH3O + O2', [3.600000e+13, 0.0, 0.0])

# Reaction 159
reaction('CH3O2 + OH <=> CH3OH + O2', [6.000000e+13, 0.0, 0.0])

# Reaction 160
reaction('CH3O2H <=> CH3O + OH', [6.310000e+14, 0.0, 42300.0])

# Reaction 161
reaction('CH2(S) + N2 <=> CH2 + N2', [1.500000e+13, 0.0, 600.0])

# Reaction 162
reaction('CH2(S) + AR <=> CH2 + AR', [9.000000e+12, 0.0, 600.0])

# Reaction 163
reaction('CH2(S) + H <=> CH + H2', [3.000000e+13, 0.0, 0.0])

# Reaction 164
reaction('CH2(S) + O <=> CO + H2', [1.500000e+13, 0.0, 0.0])

# Reaction 165
reaction('CH2(S) + O <=> HCO + H', [1.500000e+13, 0.0, 0.0])

# Reaction 166
reaction('CH2(S) + OH <=> CH2O + H', [3.000000e+13, 0.0, 0.0])

# Reaction 167
reaction('CH2(S) + H2 <=> CH3 + H', [7.000000e+13, 0.0, 0.0])

# Reaction 168
reaction('CH2(S) + O2 => H + OH + CO', [2.800000e+13, 0.0, 0.0])

# Reaction 169
reaction('CH2(S) + O2 <=> CO + H2O', [1.200000e+13, 0.0, 0.0])

# Reaction 170
reaction('CH2(S) + H2O <=> CH2 + H2O', [3.000000e+13, 0.0, 0.0])

# Reaction 171
reaction('CH2(S) + CO <=> CH2 + CO', [9.000000e+12, 0.0, 0.0])

# Reaction 172
reaction('CH2(S) + CO2 <=> CH2 + CO2', [7.000000e+12, 0.0, 0.0])

# Reaction 173
reaction('CH2(S) + CO2 <=> CH2O + CO', [1.400000e+13, 0.0, 0.0])

# Reaction 174
falloff_reaction('CH2 + H (+ M) <=> CH3 (+ M)',
                kf=[2.500000e+16, -0.8, 0.0],
                kf0=[3.200000e+27, -3.14, 1230.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0 HE:0.7',
                falloff=Troe(A=0.68, T3=78.0, T1=1995.0, T2=5590.0))

# Reaction 175
reaction('CH2 + O2 <=> HCO + OH', [1.060000e+13, 0.0, 1500.0])

# Reaction 176
reaction('CH2 + O2 => CO2 + H + H', [2.640000e+12, 0.0, 1500.0])

# Reaction 177
reaction('CH2 + O => CO + H + H', [5.000000e+13, 0.0, 0.0])

# Reaction 178
reaction('CH2 + H <=> CH + H2', [1.000000e+18, -1.56, 0.0],
        options='duplicate')

# Reaction 179
reaction('CH2 + OH <=> CH + H2O', [1.130000e+07, 2.0, 3000.0])

# Reaction 180
reaction('CH + O2 <=> HCO + O', [3.300000e+13, 0.0, 0.0])

# Reaction 181
reaction('C + OH <=> CO + H', [5.000000e+13, 0.0, 0.0])

# Reaction 182
reaction('C + O2 <=> CO + O', [5.000000e+13, 0.0, 0.0])

# Reaction 183
reaction('CH + H <=> C + H2', [1.100000e+14, 0.0, 0.0])

# Reaction 184
reaction('CH + O <=> CO + H', [5.700000e+13, 0.0, 0.0])

# Reaction 185
reaction('CH + OH <=> HCO + H', [3.000000e+13, 0.0, 0.0])

# Reaction 186
reaction('CH2 + H <=> CH + H2', [2.700000e+11, 0.67, 25700.0],
        options='duplicate')

# Reaction 187
reaction('CH + H2O <=> H + CH2O', [1.713000e+13, 0.0, -755.0])

# Reaction 188
reaction('CH + CO2 <=> HCO + CO', [1.700000e+12, 0.0, 685.0])

# Reaction 189
falloff_reaction('CH3 + CH3 (+ M) <=> C2H6 (+ M)',
                kf=[2.277000e+15, -0.69, 174.9],
                kf0=[8.054000e+31, -3.75, 981.6],
                efficiencies='H2O:5.0 CO2:3.0 CO:2.0',
                falloff=Troe(A=0.0, T3=570.0, T1=1e+30, T2=1e+30))

# Reaction 190
falloff_reaction('C2H5 + H (+ M) <=> C2H6 (+ M)',
                kf=[5.210000e+17, -0.99, 1580.0],
                kf0=[1.990000e+41, -7.08, 6685.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0 HE:0.7',
                falloff=Troe(A=0.842, T3=125.0, T1=2219.0, T2=6882.0))

# Reaction 191
reaction('C2H6 + H <=> C2H5 + H2', [1.150000e+08, 1.9, 7530.0])

# Reaction 192
reaction('C2H6 + O <=> C2H5 + OH', [3.550000e+06, 2.4, 5830.0])

# Reaction 193
reaction('C2H6 + OH <=> C2H5 + H2O', [1.480000e+07, 1.9, 950.0])

# Reaction 194
reaction('C2H6 + O2 <=> C2H5 + HO2', [6.030000e+13, 0.0, 51870.0])

# Reaction 195
reaction('C2H6 + CH3 <=> C2H5 + CH4', [5.480000e-01, 4.0, 8280.0])

# Reaction 196
reaction('C2H6 + HO2 <=> C2H5 + H2O2', [3.460000e+01, 3.61, 16920.0])

# Reaction 197
reaction('C2H6 + CH3O2 <=> C2H5 + CH3O2H', [1.940000e+01, 3.64, 17100.0])

# Reaction 198
reaction('C2H6 + CH3O <=> C2H5 + CH3OH', [2.410000e+11, 0.0, 7090.0])

# Reaction 199
reaction('C2H6 + CH <=> C2H5 + CH2', [1.100000e+14, 0.0, -260.0])

# Reaction 200
reaction('CH2(S) + C2H6 <=> CH3 + C2H5', [1.200000e+14, 0.0, 0.0])

# Reaction 201
falloff_reaction('C2H4 + H (+ M) <=> C2H5 (+ M)',
                kf=[9.569000e+08, 1.463, 1355.0],
                kf0=[1.419000e+39, -6.642, 5769.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=-0.569, T3=299.0, T1=-9147.0, T2=152.4))

# Reaction 202
reaction('H2 + CH3O2 <=> H + CH3O2H', [1.500000e+14, 0.0, 26030.0])

# Reaction 203
reaction('H2 + C2H5O2 <=> H + C2H5O2H', [1.500000e+14, 0.0, 26030.0])

# Reaction 204
reaction('C2H4 + C2H4 <=> C2H5 + C2H3', [4.820000e+14, 0.0, 71530.0])

# Reaction 205
reaction('CH3 + C2H5 <=> CH4 + C2H4', [1.180000e+04, 2.45, -2921.0])

# Reaction 206
pdep_arrhenius('CH3 + CH3 <=> H + C2H5',
            [(0.01, 'atm'), 4.740000e+12, 0.105, 10664.3],
            [(0.1, 'atm'), 2.570000e+13, -0.096, 11406.1],
            [(1.0, 'atm'), 3.100000e+14, -0.362, 13372.5],
            [(10.0, 'atm'), 2.150000e+10, 0.885, 13532.5],
            [(100.0, 'atm'), 1.032000e+02, 3.23, 11236.1])

# Reaction 207
reaction('C2H5 + H <=> C2H4 + H2', [2.000000e+12, 0.0, 0.0])

# Reaction 208
reaction('C2H5 + O <=> CH3CHO + H', [1.100000e+14, 0.0, 0.0])

# Reaction 209
reaction('C2H5 + HO2 <=> C2H5O + OH', [1.100000e+13, 0.0, 0.0])

# Reaction 210
reaction('CH3O2 + C2H5 <=> CH3O + C2H5O', [8.000000e+12, 0.0, -1000.0])

# Reaction 211
reaction('C2H5O + O2 <=> CH3CHO + HO2', [4.280000e+10, 0.0, 1097.0])

# Reaction 212
reaction('CH3 + CH2O <=> C2H5O', [3.000000e+11, 0.0, 6336.0])

# Reaction 213
reaction('CH3CHO + H <=> C2H5O', [4.610000e+07, 1.71, 7090.0])

# Reaction 214
reaction('C2H5O2 + CH2O <=> C2H5O2H + HCO', [1.990000e+12, 0.0, 11660.0])

# Reaction 215
reaction('CH4 + C2H5O2 <=> CH3 + C2H5O2H', [1.810000e+11, 0.0, 18480.0])

# Reaction 216
reaction('CH3OH + C2H5O2 <=> CH2OH + C2H5O2H', [1.810000e+12, 0.0, 13710.0])

# Reaction 217
reaction('C2H5O2 + HO2 <=> C2H5O2H + O2', [1.750000e+10, 0.0, -3275.0])

# Reaction 218
reaction('C2H6 + C2H5O2 <=> C2H5 + C2H5O2H', [8.600000e+00, 3.76, 17200.0])

# Reaction 219
reaction('C2H5O2H <=> C2H5O + OH', [6.310000e+14, 0.0, 42300.0])

# Reaction 220
pdep_arrhenius('C2H5 + O2 <=> C2H5O2',
            [(0.04, 'atm'), 3.398000e+53, -13.9, 9279.0],
            [(1.0, 'atm'), 9.362000e+59, -15.28, 14240.0],
            [(10.0, 'atm'), 1.262000e+60, -14.91, 16240.0])

# Reaction 221
pdep_arrhenius('C2H5 + O2 <=> C2H4O2H',
            [(0.04, 'atm'), 2.103000e+34, -9.01, 5444.0],
            [(1.0, 'atm'), 4.884000e+33, -8.31, 7710.0],
            [(10.0, 'atm'), 1.705000e+45, -11.49, 14590.0])

# Reaction 222
pdep_arrhenius('C2H5 + O2 <=> C2H4 + HO2',
            [(0.04, 'atm'), 2.094000e+09, 0.49, -391.4],
            [(1.0, 'atm'), 1.843000e+07, 1.13, -720.6],
            [(10.0, 'atm'), 7.561000e+14, -1.01, 4749.0],
            options='duplicate')

# Reaction 223
reaction('C2H5 + O2 <=> C2H4 + HO2', [6.609000e+00, 3.51, 14160.0],
        options='duplicate')

# Reaction 224
pdep_arrhenius('C2H5 + O2 <=> C2H4O1-2 + OH',
            [(0.04, 'atm'), 1.303000e+03, 1.93, -502.7],
            [(1.0, 'atm'), 2.438000e+02, 2.18, -62.5],
            [(10.0, 'atm'), 4.621000e+09, 0.15, 5409.0])

# Reaction 225
pdep_arrhenius('C2H5 + O2 <=> CH3CHO + OH',
            [(0.04, 'atm'), 4.908000e-06, 4.76, 254.3],
            [(1.0, 'atm'), 6.803000e-02, 3.57, 2643.0],
            [(10.0, 'atm'), 8.265000e+02, 2.41, 5285.0])

# Reaction 226
pdep_arrhenius('C2H4O2H <=> C2H5O2',
            [(0.04, 'atm'), 2.653000e-16, 6.96, 2396.0],
            [(1.0, 'atm'), 1.064000e+41, -10.1, 26030.0],
            [(10.0, 'atm'), 1.203000e+36, -8.13, 27020.0])

# Reaction 227
pdep_arrhenius('C2H5O2 <=> CH3CHO + OH',
            [(0.04, 'atm'), 1.237000e+35, -9.42, 36360.0],
            [(1.0, 'atm'), 1.687000e+36, -9.22, 38700.0],
            [(10.0, 'atm'), 2.520000e+41, -10.2, 43710.0])

# Reaction 228
pdep_arrhenius('C2H5O2 <=> C2H4 + HO2',
            [(0.04, 'atm'), 1.782000e+32, -7.1, 32840.0],
            [(1.0, 'atm'), 2.701000e+37, -8.47, 35840.0],
            [(10.0, 'atm'), 1.980000e+38, -8.46, 37900.0])

# Reaction 229
pdep_arrhenius('C2H5O2 <=> C2H4O1-2 + OH',
            [(0.04, 'atm'), 5.778000e+45, -11.9, 4112.0],
            [(1.0, 'atm'), 1.916000e+43, -10.75, 42400.0],
            [(10.0, 'atm'), 3.965000e+43, -10.46, 45580.0])

# Reaction 230
pdep_arrhenius('C2H4O2H <=> C2H4O1-2 + OH',
            [(0.04, 'atm'), 8.959000e+38, -9.4, 20660.0],
            [(1.0, 'atm'), 1.224000e+37, -8.32, 21460.0],
            [(10.0, 'atm'), 8.848000e+30, -6.08, 20660.0])

# Reaction 231
pdep_arrhenius('C2H4O2H <=> C2H4 + HO2',
            [(0.04, 'atm'), 3.918000e+40, -10.2, 22250.0],
            [(1.0, 'atm'), 6.825000e+40, -9.61, 23840.0],
            [(10.0, 'atm'), 3.980000e+34, -7.25, 23250.0])

# Reaction 232
pdep_arrhenius('C2H4O2H <=> CH3CHO + OH',
            [(0.04, 'atm'), 5.819000e+26, -7.97, 20860.0],
            [(1.0, 'atm'), 5.520000e+34, -9.88, 26230.0],
            [(10.0, 'atm'), 1.188000e+34, -9.02, 29210.0])

# Reaction 233
reaction('C2H4O1-2 <=> CH3 + HCO', [3.630000e+13, 0.0, 57200.0])

# Reaction 234
reaction('C2H4O1-2 <=> CH3CHO', [7.407000e+12, 0.0, 53800.0])

# Reaction 235
reaction('C2H4O1-2 + OH <=> C2H3O1-2 + H2O', [1.780000e+13, 0.0, 3610.0])

# Reaction 236
reaction('C2H4O1-2 + H <=> C2H3O1-2 + H2', [8.000000e+13, 0.0, 9680.0])

# Reaction 237
reaction('C2H4O1-2 + HO2 <=> C2H3O1-2 + H2O2', [1.130000e+13, 0.0, 30430.0])

# Reaction 238
reaction('C2H4O1-2 + CH3O2 <=> C2H3O1-2 + CH3O2H', [1.130000e+13, 0.0, 30430.0])

# Reaction 239
reaction('C2H4O1-2 + C2H5O2 <=> C2H3O1-2 + C2H5O2H', [1.130000e+13, 0.0, 30430.0])

# Reaction 240
reaction('C2H4O1-2 + CH3 <=> C2H3O1-2 + CH4', [1.070000e+12, 0.0, 11830.0])

# Reaction 241
reaction('C2H4O1-2 + CH3O <=> C2H3O1-2 + CH3OH', [1.200000e+11, 0.0, 6750.0])

# Reaction 242
reaction('C2H3O1-2 <=> CH3CO', [8.500000e+14, 0.0, 14000.0])

# Reaction 243
reaction('C2H3O1-2 <=> CH2CHO', [1.000000e+14, 0.0, 14000.0])

# Reaction 244
falloff_reaction('CH3CHO (+ M) <=> CH3 + HCO (+ M)',
                kf=[2.450000e+22, -1.74, 86355.0],
                kf0=[1.030000e+59, -11.3, 95912.5],
                falloff=Troe(A=0.00249, T3=718.1, T1=6.089, T2=3780.0))

# Reaction 245
falloff_reaction('CH3CHO (+ M) <=> CH4 + CO (+ M)',
                kf=[2.720000e+21, -1.74, 86355.0],
                kf0=[1.144000e+58, -11.3, 95912.5],
                falloff=Troe(A=0.00249, T3=718.1, T1=6.089, T2=3780.0))

# Reaction 246
reaction('CH3CHO + H <=> CH3CO + H2', [1.310000e+05, 2.58, 1220.0])

# Reaction 247
reaction('CH3CHO + H <=> CH2CHO + H2', [2.720000e+03, 3.1, 5210.0])

# Reaction 248
reaction('CH3CHO + O <=> CH3CO + OH', [5.940000e+12, 0.0, 1868.0])

# Reaction 249
reaction('CH3CHO + OH <=> CH3CO + H2O', [3.370000e+12, 0.0, -619.0])

# Reaction 250
reaction('CH3CHO + O2 <=> CH3CO + HO2', [3.010000e+13, 0.0, 39150.0])

# Reaction 251
reaction('CH3CHO + CH3 <=> CH3CO + CH4', [7.080000e-04, 4.58, 1966.0])

# Reaction 252
reaction('CH3CHO + HO2 <=> CH3CO + H2O2', [3.010000e+12, 0.0, 11920.0])

# Reaction 253
reaction('CH3O2 + CH3CHO <=> CH3O2H + CH3CO', [3.010000e+12, 0.0, 11920.0])

# Reaction 254
reaction('CH3CHO + CH3CO3 <=> CH3CO + CH3CO3H', [3.010000e+12, 0.0, 11920.0])

# Reaction 255
reaction('CH3CHO + OH <=> CH3 + HOCHO', [3.000000e+15, -1.076, 0.0])

# Reaction 256
reaction('CH3CHO + OH <=> CH2CHO + H2O', [1.720000e+05, 2.4, 815.0])

# Reaction 257
falloff_reaction('CH3CO (+ M) <=> CH3 + CO (+ M)',
                kf=[1.070000e+12, 0.63, 16900.0],
                kf0=[5.650000e+18, -0.97, 14600.0],
                falloff=Troe(A=0.629, T3=8730000000.0, T1=5.52, T2=76000000.0))

# Reaction 258
reaction('CH3CO + H <=> CH2CO + H2', [2.000000e+13, 0.0, 0.0])

# Reaction 259
reaction('CH3CO + O <=> CH2CO + OH', [2.000000e+13, 0.0, 0.0])

# Reaction 260
reaction('CH3CO + CH3 <=> CH2CO + CH4', [5.000000e+13, 0.0, 0.0])

# Reaction 261
reaction('CH3CO + O2 <=> CH3CO3', [1.200000e+11, 0.0, -1100.0])

# Reaction 262
reaction('CH3CO3 + HO2 <=> CH3CO3H + O2', [1.750000e+10, 0.0, -3275.0])

# Reaction 263
reaction('H2O2 + CH3CO3 <=> HO2 + CH3CO3H', [2.410000e+12, 0.0, 9936.0])

# Reaction 264
reaction('CH4 + CH3CO3 <=> CH3 + CH3CO3H', [1.810000e+11, 0.0, 18480.0])

# Reaction 265
reaction('CH2O + CH3CO3 <=> HCO + CH3CO3H', [1.990000e+12, 0.0, 11660.0])

# Reaction 266
reaction('C2H6 + CH3CO3 <=> C2H5 + CH3CO3H', [1.700000e+13, 0.0, 20460.0])

# Reaction 267
reaction('CH3CO3H <=> CH3CO2 + OH', [5.010000e+14, 0.0, 40150.0])

# Reaction 268
three_body_reaction('CH3CO2 + M <=> CH3 + CO2 + M', [4.400000e+15, 0.0, 10500.0])

# Reaction 269
falloff_reaction('CH2CHO (+ M) <=> CH2CO + H (+ M)',
                kf=[1.430000e+15, -0.15, 45600.0],
                kf0=[6.000000e+29, -3.8, 43423.9],
                falloff=Troe(A=0.985, T3=393.0, T1=9800000000.0, T2=5000000000.0))

# Reaction 270
falloff_reaction('CH2CHO (+ M) <=> CH3 + CO (+ M)',
                kf=[2.930000e+12, 0.29, 40300.0],
                kf0=[9.520000e+33, -5.07, 41300.0],
                falloff=Troe(A=7.13e-17, T3=1150.0, T1=4990000000.0, T2=1790000000.0))

# Reaction 271
pdep_arrhenius('CH2CHO + O2 <=> O2CH2CHO',
            [(0.01, 'atm'), 1.580000e+77, -21.9, 19350.0],
            [(0.1, 'atm'), 3.880000e+69, -18.84, 19240.0],
            [(1.0, 'atm'), 7.800000e+59, -15.4, 17650.0],
            [(10.0, 'atm'), 3.050000e+50, -12.2, 15630.0])

# Reaction 272
pdep_arrhenius('CH2CHO + O2 <=> CH2CO + HO2',
            [(0.01, 'atm'), 1.880000e+05, 2.37, 23730.0],
            [(0.1, 'atm'), 1.880000e+05, 2.37, 27370.0],
            [(1.0, 'atm'), 2.510000e+05, 2.33, 23800.0],
            [(10.0, 'atm'), 7.050000e+07, 1.63, 25290.0])

# Reaction 273
pdep_arrhenius('CH2CHO + O2 => CH2O + CO + OH',
            [(0.01, 'atm'), 2.680000e+17, -1.84, 6530.0],
            [(0.1, 'atm'), 1.520000e+20, -2.58, 8980.0],
            [(1.0, 'atm'), 1.650000e+19, -2.22, 10340.0],
            [(10.0, 'atm'), 8.953000e+13, -0.6, 10120.0])

# Reaction 274
pdep_arrhenius('CH2CHO + O2 <=> HO2CH2CO',
            [(0.01, 'atm'), 3.640000e+65, -21.87, 19020.0],
            [(0.1, 'atm'), 3.640000e+58, -19.0, 19090.0],
            [(1.0, 'atm'), 6.650000e+48, -15.55, 17460.0],
            [(10.0, 'atm'), 4.800000e+38, -12.14, 14960.0])

# Reaction 275
pdep_arrhenius('O2CH2CHO <=> HO2CH2CO',
            [(0.01, 'atm'), 8.270000e+30, -6.65, 24500.0],
            [(0.1, 'atm'), 1.730000e+26, -4.99, 23760.0],
            [(1.0, 'atm'), 9.030000e+19, -2.92, 22170.0],
            [(10.0, 'atm'), 1.430000e+16, -1.67, 21210.0])

# Reaction 276
pdep_arrhenius('O2CH2CHO <=> CH2CO + HO2',
            [(0.01, 'atm'), 2.050000e+40, -13.31, 52150.0],
            [(0.1, 'atm'), 5.720000e+45, -14.0, 52200.0],
            [(1.0, 'atm'), 4.160000e+55, -15.76, 55080.0],
            [(10.0, 'atm'), 1.120000e+61, -16.04, 60010.0])

# Reaction 277
pdep_arrhenius('HO2CH2CO => CO + CH2O + OH',
            [(0.01, 'atm'), 2.360000e+17, -2.95, 8100.0],
            [(0.1, 'atm'), 2.380000e+18, -2.95, 8100.0],
            [(1.0, 'atm'), 2.510000e+19, -2.95, 8110.0],
            [(10.0, 'atm'), 4.160000e+20, -3.02, 8240.0])

# Reaction 278
pdep_arrhenius('HO2CH2CO <=> CH2CO + HO2',
            [(0.01, 'atm'), 1.120000e+07, -3.76, 21680.0],
            [(0.1, 'atm'), 1.100000e+08, -3.76, 21680.0],
            [(1.0, 'atm'), 9.200000e+08, -3.73, 21630.0],
            [(10.0, 'atm'), 2.090000e+09, -3.55, 21220.0])

# Reaction 279
falloff_reaction('CH2 + CO (+ M) <=> CH2CO (+ M)',
                kf=[8.100000e+11, 0.0, 0.0],
                kf0=[2.690000e+33, -5.11, 7095.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0 HE:0.7',
                falloff=Troe(A=0.5907, T3=275.0, T1=1226.0, T2=5185.0))

# Reaction 280
falloff_reaction('CH3CO (+ M) <=> CH2CO + H (+ M)',
                kf=[9.413000e+07, 1.917, 44987.2],
                kf0=[1.516000e+51, -10.27, 55390.0],
                falloff=Troe(A=0.6009, T3=8103000000.0, T1=667.7, T2=5000000000.0))

# Reaction 281
reaction('CH2CO + H <=> HCCO + H2', [1.401000e+15, -0.171, 8783.2])

# Reaction 282
reaction('CH2CO + H <=> CH3 + CO', [7.704000e+13, -0.171, 4183.2])

# Reaction 283
reaction('CH2CO + O <=> CH2 + CO2', [1.750000e+12, 0.0, 1350.0])

# Reaction 284
reaction('CH2CO + O <=> HCCO + OH', [1.000000e+13, 0.0, 8000.0])

# Reaction 285
reaction('CH2CO + OH <=> HCCO + H2O', [1.000000e+13, 0.0, 2000.0])

# Reaction 286
reaction('CH2CO + OH <=> CH2OH + CO', [2.000000e+12, 0.0, -1010.0])

# Reaction 287
reaction('CH2CO + CH3 <=> C2H5 + CO', [4.769000e+04, 2.312, 9468.0])

# Reaction 288
reaction('CH2(S) + CH2CO <=> C2H4 + CO', [1.600000e+14, 0.0, 0.0])

# Reaction 289
reaction('HCCO + OH => H2 + CO + CO', [1.000000e+14, 0.0, 0.0])

# Reaction 290
reaction('HCCO + O => H + CO + CO', [8.000000e+13, 0.0, 0.0])

# Reaction 291
reaction('HCCO + H <=> CH2(S) + CO', [1.000000e+14, 0.0, 0.0])

# Reaction 292
reaction('HCCO + O2 => OH + CO + CO', [1.910000e+11, -0.02, 1020.0])

# Reaction 293
reaction('HCCO + O2 => CO2 + CO + H', [4.780000e+12, -0.142, 1150.0])

# Reaction 294
three_body_reaction('CH + CO + M <=> HCCO + M', [7.570000e+22, -1.9, 0.0])

# Reaction 295
reaction('CH + CH2O <=> H + CH2CO', [9.460000e+13, 0.0, -515.0])

# Reaction 296
reaction('CH + HCCO <=> CO + C2H2', [5.000000e+13, 0.0, 0.0])

# Reaction 297
falloff_reaction('C2H3 + H (+ M) <=> C2H4 (+ M)',
                kf=[6.080000e+12, 0.27, 280.0],
                kf0=[1.400000e+30, -3.86, 3320.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0 HE:0.7',
                falloff=Troe(A=0.782, T3=207.5, T1=2663.0, T2=6095.0))

# Reaction 298
falloff_reaction('C2H4 (+ M) <=> H2 + H2CC (+ M)',
                kf=[8.000000e+12, 0.44, 88770.0],
                kf0=[7.000000e+50, -9.31, 99860.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.7345, T3=180.0, T1=1035.0, T2=5417.0))

# Reaction 299
reaction('C2H4 + H <=> C2H3 + H2', [5.070000e+07, 1.93, 12950.0])

# Reaction 300
reaction('C2H4 + O <=> CH3 + HCO', [7.453000e+06, 1.88, 183.0])

# Reaction 301
reaction('C2H4 + O <=> CH2CHO + H', [6.098000e+06, 1.88, 183.0])

# Reaction 302
reaction('C2H4 + OH <=> C2H3 + H2O', [2.230000e+04, 2.745, 2215.5])

# Reaction 303
pdep_arrhenius('C2H4 + OH <=> CH3 + CH2O',
            [(0.01, 'atm'), 5.350000e+00, 2.92, -1732.7],
            [(0.025, 'atm'), 3.190000e+01, 2.71, -1172.3],
            [(0.1, 'atm'), 5.550000e+02, 2.36, -180.8],
            [(1.0, 'atm'), 1.780000e+05, 1.68, 2060.5],
            [(10.0, 'atm'), 2.370000e+09, 0.56, 6006.7],
            [(100.0, 'atm'), 2.760000e+13, -0.5, 11455.1])

# Reaction 304
pdep_arrhenius('C2H4 + OH <=> CH3CHO + H',
            [(0.01, 'atm'), 2.370000e-07, 5.3, -2050.6],
            [(0.025, 'atm'), 8.730000e-05, 4.57, -618.0],
            [(0.1, 'atm'), 4.030000e-01, 3.54, 1881.7],
            [(1.0, 'atm'), 2.380000e-02, 3.91, 1722.7],
            [(10.0, 'atm'), 8.250000e+08, 1.01, 10507.3],
            [(100.0, 'atm'), 6.800000e+09, 0.81, 13867.3])

# Reaction 305
pdep_arrhenius('C2H4 + OH <=> C2H3OH + H',
            [(0.01, 'atm'), 1.040000e+04, 2.6, 4121.0],
            [(0.025, 'atm'), 1.070000e+04, 2.6, 4129.0],
            [(0.1, 'atm'), 1.520000e+04, 2.56, 4238.3],
            [(1.0, 'atm'), 3.190000e+05, 2.19, 5255.6],
            [(10.0, 'atm'), 1.940000e+08, 1.43, 7828.8],
            [(100.0, 'atm'), 8.550000e+10, 0.75, 11490.8])

# Reaction 306
pdep_arrhenius('C2H4 + OH <=> PC2H4OH',
            [(0.01, 'atm'), 1.740000e+43, -10.461, 7698.7],
            [(0.025, 'atm'), 3.250000e+37, -8.629, 5214.7],
            [(0.1, 'atm'), 1.840000e+35, -7.75, 4908.9],
            [(1.0, 'atm'), 2.560000e+36, -7.752, 6946.1],
            [(10.0, 'atm'), 3.700000e+33, -6.573, 7605.9],
            [(100.0, 'atm'), 1.120000e+26, -4.101, 5757.0])

# Reaction 307
reaction('C2H3OH + O2 <=> CH2CHO + HO2', [5.310000e+11, 0.21, 39830.0])

# Reaction 308
reaction('C2H3OH + O <=> CH2CHO + OH', [1.875000e+06, 1.9, -860.0])

# Reaction 309
reaction('C2H3OH + OH <=> CH2CHO + H2O', [3.330000e+09, 1.1, 540.5])

# Reaction 310
reaction('C2H3OH + CH3 <=> CH2CHO + CH4', [2.030000e-08, 5.9, 1052.0])

# Reaction 311
reaction('C2H3OH + CH3O2 <=> CH2CHO + CH3O2H', [3.400000e+03, 2.5, 8922.0])

# Reaction 312
reaction('C2H3OH + H <=> CH2CHO + H2', [1.480000e+03, 3.077, 7220.0])

# Reaction 313
reaction('C2H3OH + H <=> C2H2OH + H2', [2.470000e+07, 2.03, 15200.0])

# Reaction 314
reaction('C2H3OH + H <=> PC2H4OH', [3.010000e+08, 1.577, 3670.0])

# Reaction 315
reaction('C2H3OH + HO2 <=> CH3CHO + HO2', [1.490000e+05, 1.67, 6810.0])

# Reaction 316
pdep_arrhenius('C2H3OH <=> CH3CHO',
            [(0.1, 'atm'), 7.420000e+46, -10.56, 67420.0],
            [(1.0, 'atm'), 4.420000e+42, -9.09, 67069.2],
            [(100.0, 'atm'), 2.900000e+27, -4.35, 61612.9])

# Reaction 317
reaction('C2H4 + CH3 <=> C2H3 + CH4', [6.620000e+00, 3.7, 9500.0])

# Reaction 318
reaction('C2H4 + O2 <=> C2H3 + HO2', [4.220000e+13, 0.0, 57623.1])

# Reaction 319
reaction('C2H4 + CH3O <=> C2H3 + CH3OH', [1.200000e+11, 0.0, 6750.0])

# Reaction 320
reaction('C2H4 + CH3O2 <=> C2H3 + CH3O2H', [8.590000e+00, 3.754, 27132.0])

# Reaction 321
reaction('C2H4 + C2H5O2 <=> C2H3 + C2H5O2H', [8.590000e+00, 3.754, 27132.0])

# Reaction 322
reaction('C2H4 + CH3CO3 <=> C2H3 + CH3CO3H', [1.130000e+13, 0.0, 30430.0])

# Reaction 323
reaction('C2H4 + CH3O2 <=> C2H4O1-2 + CH3O', [2.820000e+12, 0.0, 17110.0])

# Reaction 324
reaction('C2H4 + C2H5O2 <=> C2H4O1-2 + C2H5O', [2.820000e+12, 0.0, 17110.0])

# Reaction 325
reaction('C2H4 + HO2 <=> C2H4O1-2 + OH', [5.575000e+11, 0.0, 17190.0])

# Reaction 326
reaction('CH + CH4 <=> C2H4 + H', [6.000000e+13, 0.0, 0.0])

# Reaction 327
reaction('CH2(S) + CH3 <=> C2H4 + H', [2.000000e+13, 0.0, 0.0])

# Reaction 328
falloff_reaction('C2H2 + H (+ M) <=> C2H3 (+ M)',
                kf=[1.710000e+10, 1.266, 2709.0],
                kf0=[6.346000e+31, -4.664, 3780.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0 HE:0.7',
                falloff=Troe(A=0.788, T3=-10200.0, T1=1e-30))

# Reaction 329
reaction('C2H3 + O2 <=> CH2O + HCO', [1.700000e+29, -5.312, 6503.1])

# Reaction 330
reaction('C2H3 + O2 <=> CH2CHO + O', [7.000000e+14, -0.611, 5262.4])

# Reaction 331
reaction('C2H3 + O2 => H + CO + CH2O', [5.190000e+15, -1.26, 3312.6])

# Reaction 332
reaction('CH3 + C2H3 <=> CH4 + C2H2', [3.920000e+11, 0.0, 0.0])

# Reaction 333
reaction('C2H3 + H <=> C2H2 + H2', [9.000000e+13, 0.0, 0.0])

# Reaction 334
reaction('C2H3 + H <=> H2CC + H2', [6.000000e+13, 0.0, 0.0])

# Reaction 335
reaction('C2H3 + OH <=> C2H2 + H2O', [3.011000e+13, 0.0, 0.0])

# Reaction 336
reaction('C2H3 + C2H3 <=> C2H2 + C2H4', [9.600000e+11, 0.0, 0.0])

# Reaction 337
falloff_reaction('C2H + H (+ M) <=> C2H2 (+ M)',
                kf=[1.000000e+17, 0.0, 0.0],
                kf0=[3.750000e+33, -4.8, 1900.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0 HE:0.7',
                falloff=Troe(A=0.646, T3=132.0, T1=1315.0, T2=5566.0))

# Reaction 338
reaction('C2H + O <=> CH + CO', [5.000000e+13, 0.0, 0.0])

# Reaction 339
reaction('C2H + OH <=> H + HCCO', [2.000000e+13, 0.0, 0.0])

# Reaction 340
reaction('C2H + O2 <=> HCO + CO', [5.000000e+13, 0.0, 1500.0])

# Reaction 341
reaction('C2H + H2 <=> H + C2H2', [4.900000e+05, 2.5, 560.0])

# Reaction 342
falloff_reaction('C2H2 (+ M) <=> H2CC (+ M)',
                kf=[8.000000e+14, -0.52, 50750.0],
                kf0=[2.450000e+15, -0.64, 49700.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 CH4:2.0 C2H6:3.0 C2H4:2.5 C2H2:2.5')

# Reaction 343
reaction('C2H2 + O <=> CH2 + CO', [7.395000e+08, 1.28, 2472.0])

# Reaction 344
reaction('C2H2 + O <=> HCCO + H', [2.958000e+09, 1.28, 2472.0])

# Reaction 345
reaction('C2H2 + OH <=> C2H + H2O', [2.632000e+06, 2.14, 17060.0])

# Reaction 346
pdep_arrhenius('C2H2 + OH <=> HCCOH + H',
            [(0.01, 'atm'), 2.800000e+05, 2.28, 12420.0],
            [(0.025, 'atm'), 7.467000e+05, 2.16, 12550.0],
            [(0.1, 'atm'), 1.776000e+06, 2.04, 12670.0],
            [(1.0, 'atm'), 2.415000e+06, 2.0, 12710.0],
            [(10.0, 'atm'), 3.210000e+06, 1.97, 12810.0],
            [(100.0, 'atm'), 7.347000e+06, 1.89, 13600.0])

# Reaction 347
pdep_arrhenius('C2H2 + OH <=> CH2CO + H',
            [(0.01, 'atm'), 1.578000e+03, 2.56, -844.5],
            [(0.025, 'atm'), 1.518000e+04, 2.28, -292.1],
            [(0.1, 'atm'), 3.017000e+05, 1.92, 598.1],
            [(1.0, 'atm'), 7.528000e+06, 1.55, 2106.0],
            [(10.0, 'atm'), 5.101000e+06, 1.65, 3400.0],
            [(100.0, 'atm'), 1.457000e+04, 2.45, 4477.0])

# Reaction 348
pdep_arrhenius('C2H2 + OH <=> CH3 + CO',
            [(0.01, 'atm'), 4.757000e+05, 1.68, -329.8],
            [(0.025, 'atm'), 4.372000e+06, 1.4, 226.5],
            [(0.1, 'atm'), 7.648000e+07, 1.05, 1115.0],
            [(1.0, 'atm'), 1.277000e+09, 0.73, 2579.0],
            [(10.0, 'atm'), 4.312000e+08, 0.92, 3736.0],
            [(100.0, 'atm'), 8.250000e+05, 1.77, 4697.0])

# Reaction 349
pdep_arrhenius('C2H2 + OH <=> C2H2OH',
            [(0.01, 'atm'), 3.913000e+32, -7.126, 5824.0],
            [(0.025, 'atm'), 1.067000e+32, -6.847, 5508.0],
            [(0.1, 'atm'), 1.646000e+32, -6.717, 5822.0],
            [(1.0, 'atm'), 1.387000e+31, -6.087, 6348.0],
            [(10.0, 'atm'), 2.892000e+29, -5.288, 7055.0],
            [(100.0, 'atm'), 1.367000e+25, -3.754, 6543.0])

# Reaction 350
reaction('C2H2 + HCO <=> C2H3 + CO', [1.000000e+07, 2.0, 6000.0])

# Reaction 351
reaction('C2H2 + CH2 <=> C3H3 + H', [1.200000e+13, 0.0, 6620.0])

# Reaction 352
reaction('C2H2 + CH2(S) <=> C3H3 + H', [2.000000e+13, 0.0, 0.0])

# Reaction 353
reaction('C2H2 + HCCO <=> C3H3 + CO', [1.000000e+11, 0.0, 3000.0])

# Reaction 354
reaction('H2CC + H <=> C2H2 + H', [1.000000e+14, 0.0, 0.0])

# Reaction 355
reaction('H2CC + OH <=> CH2CO + H', [2.000000e+13, 0.0, 0.0])

# Reaction 356
reaction('H2CC + O2 <=> HCO + HCO', [1.000000e+13, 0.0, 0.0])

# Reaction 357
reaction('H + HCCOH <=> H + CH2CO', [1.000000e+13, 0.0, 0.0])

# Reaction 358
pdep_arrhenius('C2H5OH <=> C2H4 + H2O',
            [(0.001, 'atm'), 3.410000e+59, -14.2, 83672.6],
            [(0.01, 'atm'), 2.620000e+57, -13.3, 85262.2],
            [(0.1, 'atm'), 1.650000e+52, -11.5, 84745.6],
            [(1.0, 'atm'), 5.230000e+43, -8.9, 81506.7],
            [(10.0, 'atm'), 4.590000e+32, -5.6, 76062.4],
            [(100.0, 'atm'), 3.840000e+20, -2.06, 69465.5])

# Reaction 359
pdep_arrhenius('C2H5OH <=> CH3 + CH2OH',
            [(0.001, 'atm'), 1.200000e+54, -12.9, 100005.7],
            [(0.01, 'atm'), 5.180000e+59, -14.0, 99906.4],
            [(0.1, 'atm'), 1.620000e+66, -15.3, 105390.5],
            [(1.0, 'atm'), 5.550000e+64, -14.5, 106183.0],
            [(10.0, 'atm'), 1.550000e+58, -12.3, 105768.0],
            [(100.0, 'atm'), 1.780000e+47, -8.96, 101058.8])

# Reaction 360
pdep_arrhenius('C2H5OH <=> C2H5 + OH',
            [(0.001, 'atm'), 8.100000e+46, -11.3, 111053.4],
            [(0.01, 'atm'), 1.860000e+56, -13.5, 107238.4],
            [(0.1, 'atm'), 4.650000e+63, -15.0, 109622.8],
            [(1.0, 'atm'), 4.460000e+65, -14.9, 112345.0],
            [(10.0, 'atm'), 2.790000e+61, -13.4, 113080.2],
            [(100.0, 'atm'), 6.170000e+51, -10.3, 109940.7])

# Reaction 361
reaction('C2H5OH + O2 <=> PC2H4OH + HO2', [2.000000e+13, 0.0, 52800.0])

# Reaction 362
reaction('C2H5OH + O2 <=> SC2H4OH + HO2', [1.500000e+13, 0.0, 50150.0])

# Reaction 363
reaction('C2H5OH + H <=> SC2H4OH + H2', [8.790000e+04, 2.68, 2910.0])

# Reaction 364
reaction('C2H5OH + H <=> PC2H4OH + H2', [5.310000e+04, 2.81, 7490.0])

# Reaction 365
reaction('C2H5OH + H <=> C2H5O + H2', [9.450000e+02, 3.14, 8701.1])

# Reaction 366
reaction('C2H5OH + OH <=> SC2H4OH + H2O', [7.170000e+04, 2.54, -1534.0])

# Reaction 367
reaction('C2H5OH + OH <=> PC2H4OH + H2O', [5.700000e+00, 3.38, -2394.3])

# Reaction 368
reaction('C2H5OH + OH <=> C2H5O + H2O', [5.810000e-03, 4.28, -3560.0])

# Reaction 369
reaction('C2H5OH + HO2 <=> SC2H4OH + H2O2', [3.500000e-05, 5.26, 7475.1])

# Reaction 370
reaction('C2H5OH + HO2 <=> PC2H4OH + H2O2', [3.986000e-02, 4.3, 15333.0])

# Reaction 371
reaction('C2H5OH + HO2 <=> C2H5O + H2O2', [6.470000e-07, 5.3, 10533.1])

# Reaction 372
reaction('C2H5OH + CH3O2 <=> PC2H4OH + CH3O2H', [1.230000e+04, 2.55, 15750.0])

# Reaction 373
reaction('C2H5OH + CH3O2 <=> SC2H4OH + CH3O2H', [8.200000e+03, 2.55, 10750.0])

# Reaction 374
reaction('C2H5OH + CH3O2 <=> C2H5O + CH3O2H', [2.500000e+12, 0.0, 24000.0])

# Reaction 375
reaction('C2H5OH + O <=> PC2H4OH + OH', [9.690000e+02, 3.23, 4658.0])

# Reaction 376
reaction('C2H5OH + O <=> SC2H4OH + OH', [1.450000e+05, 2.47, 876.0])

# Reaction 377
reaction('C2H5OH + O <=> C2H5O + OH', [1.460000e-03, 4.73, 1727.0])

# Reaction 378
reaction('C2H5OH + CH3 <=> PC2H4OH + CH4', [3.300000e+02, 3.3, 12290.0])

# Reaction 379
reaction('C2H5OH + CH3 <=> SC2H4OH + CH4', [1.993000e+01, 3.37, 7634.0])

# Reaction 380
reaction('C2H5OH + CH3 <=> C2H5O + CH4', [2.035000e+00, 3.57, 7721.0])

# Reaction 381
reaction('C2H5OH + C2H5 <=> PC2H4OH + C2H6', [5.000000e+10, 0.0, 13400.0])

# Reaction 382
reaction('C2H5OH + C2H5 <=> SC2H4OH + C2H6', [5.000000e+10, 0.0, 10400.0])

# Reaction 383
pdep_arrhenius('SC2H4OH <=> CH3CHO + H',
            [(0.001, 'atm'), 5.690000e+52, -13.38, 45049.0],
            [(0.01, 'atm'), 3.290000e+56, -14.12, 48129.0],
            [(0.1, 'atm'), 8.580000e+57, -14.16, 50743.0],
            [(1.0, 'atm'), 5.360000e+55, -13.15, 51886.0],
            [(10.0, 'atm'), 1.660000e+48, -10.64, 50297.0],
            [(20.0, 'atm'), 8.260000e+44, -9.59, 49218.0],
            [(50.0, 'atm'), 1.010000e+40, -8.06, 47439.0],
            [(100.0, 'atm'), 1.100000e+36, -6.84, 45899.0])

# Reaction 384
pdep_arrhenius('SC2H4OH <=> C2H3OH + H',
            [(0.001, 'atm'), 5.400000e+46, -11.63, 44323.0],
            [(0.01, 'atm'), 1.210000e+51, -12.55, 47240.0],
            [(0.1, 'atm'), 2.870000e+54, -13.15, 50702.0],
            [(1.0, 'atm'), 3.790000e+53, -12.51, 52560.0],
            [(10.0, 'atm'), 6.330000e+46, -10.2, 51441.0],
            [(20.0, 'atm'), 3.870000e+43, -9.17, 50440.0],
            [(50.0, 'atm'), 5.080000e+38, -7.65, 48713.0],
            [(100.0, 'atm'), 5.120000e+34, -6.41, 47182.0])

# Reaction 385
pdep_arrhenius('SC2H4OH <=> C2H5O',
            [(0.001, 'atm'), 5.480000e+45, -11.63, 44328.0],
            [(0.01, 'atm'), 2.540000e+49, -12.37, 46445.0],
            [(0.1, 'atm'), 1.650000e+54, -13.4, 50330.0],
            [(1.0, 'atm'), 1.810000e+55, -13.31, 53132.0],
            [(10.0, 'atm'), 4.580000e+49, -11.32, 52714.0],
            [(20.0, 'atm'), 4.110000e+46, -10.33, 51834.0],
            [(50.0, 'atm'), 6.680000e+41, -8.83, 50202.0],
            [(100.0, 'atm'), 6.540000e+37, -7.58, 48697.0])

# Reaction 386
pdep_arrhenius('SC2H4OH <=> PC2H4OH',
            [(0.001, 'atm'), 2.650000e+36, -8.86, 51019.0],
            [(0.01, 'atm'), 3.560000e+37, -8.89, 51114.0],
            [(0.1, 'atm'), 4.140000e+39, -9.19, 51912.0],
            [(1.0, 'atm'), 5.820000e+44, -10.34, 55296.0],
            [(10.0, 'atm'), 4.260000e+48, -11.06, 59458.0],
            [(20.0, 'atm'), 8.840000e+47, -10.74, 59901.0],
            [(50.0, 'atm'), 2.230000e+45, -9.84, 59604.0],
            [(100.0, 'atm'), 1.700000e+42, -8.83, 58737.0])

# Reaction 387
reaction('O2C2H4OH <=> PC2H4OH + O2', [3.900000e+16, -1.0, 30000.0])

# Reaction 388
reaction('O2C2H4OH => OH + CH2O + CH2O', [3.125000e+09, 0.0, 18900.0])

# Reaction 389
pdep_arrhenius('SC2H4OH + O2 <=> CH3CHO + HO2',
            [(0.01, 'atm'), 5.260000e+17, -1.637, 838.0],
            [(0.1, 'atm'), 5.260000e+17, -1.637, 838.0],
            [(1.0, 'atm'), 5.280000e+17, -1.638, 839.0],
            [(10.0, 'atm'), 1.540000e+18, -1.771, 1120.0],
            [(100.0, 'atm'), 3.780000e+20, -2.429, 3090.0])

# Reaction 390
pdep_arrhenius('SC2H4OH + O2 <=> C2H3OH + HO2',
            [(0.01, 'atm'), 5.120000e+02, 2.496, -414.0],
            [(0.1, 'atm'), 5.330000e+02, 2.49, -402.0],
            [(1.0, 'atm'), 7.620000e+02, 2.446, -296.0],
            [(10.0, 'atm'), 8.920000e+03, 2.146, 470.0],
            [(100.0, 'atm'), 4.380000e+05, 1.699, 2330.0])

# Reaction 391
pdep_arrhenius('CH3COCH3 <=> CH3CO + CH3',
            [(0.01, 'atm'), 2.050000e+58, -12.796, 100030.1],
            [(0.1, 'atm'), 3.300000e+51, -10.574, 98221.2],
            [(1.0, 'atm'), 1.310000e+42, -7.657, 94660.6],
            [(10.0, 'atm'), 2.160000e+33, -4.989, 90916.5],
            [(100.0, 'atm'), 9.400000e+28, -3.669, 89022.8])

# Reaction 392
reaction('CH3COCH3 + OH <=> CH3COCH2 + H2O', [1.250000e+05, 2.483, 445.0])

# Reaction 393
reaction('CH3COCH3 + H <=> CH3COCH2 + H2', [9.800000e+05, 2.43, 5160.0])

# Reaction 394
reaction('CH3COCH3 + O <=> CH3COCH2 + OH', [5.130000e+11, 0.211, 4890.0])

# Reaction 395
reaction('CH3COCH3 + CH3 <=> CH3COCH2 + CH4', [3.960000e+11, 0.0, 9784.0])

# Reaction 396
reaction('CH3COCH3 + CH3O <=> CH3COCH2 + CH3OH', [4.340000e+11, 0.0, 6460.0])

# Reaction 397
reaction('CH3COCH3 + O2 <=> CH3COCH2 + HO2', [6.030000e+13, 0.0, 48500.0])

# Reaction 398
reaction('CH3COCH3 + HO2 <=> CH3COCH2 + H2O2', [1.700000e+13, 0.0, 20460.0])

# Reaction 399
reaction('CH3COCH3 + CH3O2 <=> CH3COCH2 + CH3O2H', [1.700000e+13, 0.0, 20460.0])

# Reaction 400
reaction('CH2CO + CH3 <=> CH3COCH2', [1.760000e+04, 2.48, 6130.0])

# Reaction 401
reaction('CH3COCH2 + O2 <=> CH3COCH2O2', [1.200000e+11, 0.0, -1100.0])

# Reaction 402
reaction('CH3COCH3 + CH3COCH2O2 <=> CH3COCH2 + C3KET21', [1.000000e+11, 0.0, 5000.0])

# Reaction 403
reaction('CH2O + CH3COCH2O2 <=> HCO + C3KET21', [1.288000e+11, 0.0, 9000.0])

# Reaction 404
reaction('HO2 + CH3COCH2O2 <=> C3KET21 + O2', [1.000000e+12, 0.0, 0.0])

# Reaction 405
reaction('C2H3 + HCO <=> C2H3CHO', [1.810000e+13, 0.0, 0.0])

# Reaction 406
reaction('C2H3CHO + H <=> C2H3CO + H2', [1.340000e+13, 0.0, 3300.0])

# Reaction 407
reaction('C2H3CHO + O <=> C2H3CO + OH', [5.940000e+12, 0.0, 1868.0])

# Reaction 408
reaction('C2H3CHO + OH <=> C2H3CO + H2O', [9.240000e+06, 1.5, -962.0])

# Reaction 409
reaction('C2H3CHO + O2 <=> C2H3CO + HO2', [1.005000e+13, 0.0, 40700.0])

# Reaction 410
reaction('C2H3CHO + HO2 <=> C2H3CO + H2O2', [3.010000e+12, 0.0, 11920.0])

# Reaction 411
reaction('C2H3CHO + CH3 <=> C2H3CO + CH4', [2.608000e+06, 1.78, 5911.0])

# Reaction 412
reaction('C2H3CHO + C2H3 <=> C2H3CO + C2H4', [1.740000e+12, 0.0, 8440.0])

# Reaction 413
reaction('C2H3CHO + CH3O <=> C2H3CO + CH3OH', [1.000000e+12, 0.0, 3300.0])

# Reaction 414
reaction('C2H3CHO + CH3O2 <=> C2H3CO + CH3O2H', [3.010000e+12, 0.0, 11920.0])

# Reaction 415
reaction('C2H3 + CO <=> C2H3CO', [1.510000e+11, 0.0, 4810.0])

# Reaction 416
reaction('C2H5 + HCO <=> C2H5CHO', [1.810000e+13, 0.0, 0.0])

# Reaction 417
reaction('C2H5CHO + H <=> C2H5CO + H2', [4.000000e+13, 0.0, 4200.0])

# Reaction 418
reaction('C2H5CHO + O <=> C2H5CO + OH', [5.000000e+12, 0.0, 1790.0])

# Reaction 419
reaction('C2H5CHO + OH <=> C2H5CO + H2O', [2.690000e+10, 0.76, -340.0])

# Reaction 420
reaction('C2H5CHO + CH3 <=> C2H5CO + CH4', [2.608000e+06, 1.78, 5911.0])

# Reaction 421
reaction('C2H5CHO + HO2 <=> C2H5CO + H2O2', [2.800000e+12, 0.0, 13600.0])

# Reaction 422
reaction('C2H5CHO + CH3O <=> C2H5CO + CH3OH', [1.000000e+12, 0.0, 3300.0])

# Reaction 423
reaction('C2H5CHO + CH3O2 <=> C2H5CO + CH3O2H', [3.010000e+12, 0.0, 11920.0])

# Reaction 424
reaction('C2H5CHO + C2H5 <=> C2H5CO + C2H6', [1.000000e+12, 0.0, 8000.0])

# Reaction 425
reaction('C2H5CHO + C2H5O <=> C2H5CO + C2H5OH', [6.026000e+11, 0.0, 3300.0])

# Reaction 426
reaction('C2H5CHO + C2H5O2 <=> C2H5CO + C2H5O2H', [3.010000e+12, 0.0, 11920.0])

# Reaction 427
reaction('C2H5CHO + O2 <=> C2H5CO + HO2', [1.005000e+13, 0.0, 40700.0])

# Reaction 428
reaction('C2H5CHO + CH3CO3 <=> C2H5CO + CH3CO3H', [3.010000e+12, 0.0, 11920.0])

# Reaction 429
reaction('C2H5CHO + C2H3 <=> C2H5CO + C2H4', [1.700000e+12, 0.0, 8440.0])

# Reaction 430
reaction('C2H5 + CO <=> C2H5CO', [1.510000e+11, 0.0, 4810.0])

# Reaction 431
falloff_reaction('CH3OCH3 (+ M) <=> CH3 + CH3O (+ M)',
                kf=[4.380000e+21, -1.57, 83890.0],
                kf0=[7.520000e+15, 0.0, 42790.0],
                falloff=Troe(A=0.454, T3=1e-30, T1=2510.0))

# Reaction 432
reaction('CH3OCH3 + OH <=> CH3OCH2 + H2O', [6.324000e+06, 2.0, -651.7])

# Reaction 433
reaction('CH3OCH3 + H <=> CH3OCH2 + H2', [7.721000e+06, 2.09, 3384.0])

# Reaction 434
reaction('CH3OCH3 + O <=> CH3OCH2 + OH', [7.750000e+08, 1.36, 2250.0])

# Reaction 435
reaction('CH3OCH3 + HO2 <=> CH3OCH2 + H2O2', [8.670000e+02, 3.01, 12090.0])

# Reaction 436
reaction('CH3OCH3 + CH3O2 <=> CH3OCH2 + CH3O2H', [3.120000e+02, 3.12, 13190.0])

# Reaction 437
reaction('CH3OCH3 + CH3 <=> CH3OCH2 + CH4', [1.445000e-06, 5.73, 5700.0])

# Reaction 438
reaction('CH3OCH3 + O2 <=> CH3OCH2 + HO2', [4.100000e+13, 0.0, 44910.0])

# Reaction 439
reaction('CH3OCH3 + CH3O <=> CH3OCH2 + CH3OH', [6.020000e+11, 0.0, 4074.0])

# Reaction 440
reaction('CH3OCH3 + CH3OCH2O2 <=> CH3OCH2 + CH3OCH2O2H', [5.000000e+12, 0.0, 17690.0])

# Reaction 441
reaction('CH3OCH3 + O2CHO <=> CH3OCH2 + HO2CHO', [4.425000e+04, 2.6, 13910.0])

# Reaction 442
reaction('CH3OCH3 + OCHO <=> CH3OCH2 + HOCHO', [1.000000e+13, 0.0, 17690.0])

# Reaction 443
reaction('CH3OCH2 <=> CH2O + CH3', [1.600000e+13, 0.0, 25500.0])

# Reaction 444
reaction('CH3OCH2 + CH3O <=> CH3OCH3 + CH2O', [2.410000e+13, 0.0, 0.0])

# Reaction 445
reaction('CH3OCH2 + CH2O <=> CH3OCH3 + HCO', [5.490000e+03, 2.8, 5862.0])

# Reaction 446
reaction('CH3OCH2 + CH3CHO <=> CH3OCH3 + CH3CO', [1.260000e+12, 0.0, 8499.0])

# Reaction 447
reaction('CH3OCH2 + O2 <=> CH3OCH2O2', [2.000000e+12, 0.0, 0.0])

# Reaction 448
reaction('CH3OCH2O2 + CH2O <=> CH3OCH2O2H + HCO', [1.000000e+12, 0.0, 11660.0])

# Reaction 449
reaction('CH3OCH2O2 + CH3CHO <=> CH3OCH2O2H + CH3CO', [2.800000e+12, 0.0, 13600.0])

# Reaction 450
reaction('CH3OCH2O2 + CH3OCH2O2 => O2 + CH3OCH2O + CH3OCH2O', [1.547000e+23, -4.5, 0.0])

# Reaction 451
reaction('CH3OCH2O + OH <=> CH3OCH2O2H', [2.000000e+13, 0.0, 0.0])

# Reaction 452
reaction('CH3O + CH2O <=> CH3OCH2O', [1.000000e+11, 0.0, 7960.0])

# Reaction 453
reaction('CH3OCH2O + O2 <=> CH3OCHO + HO2', [5.000000e+10, 0.0, 500.0])

# Reaction 454
reaction('CH3OCHO + H <=> CH3OCH2O', [1.000000e+13, 0.0, 7838.0])

# Reaction 455
reaction('CH3OCH2O2 <=> CH2OCH2O2H', [6.000000e+10, 0.0, 21580.0])

# Reaction 456
reaction('CH2OCH2O2H => OH + CH2O + CH2O', [1.500000e+13, 0.0, 20760.0])

# Reaction 457
reaction('CH2OCH2O2H + O2 <=> O2CH2OCH2O2H', [7.000000e+11, 0.0, 0.0])

# Reaction 458
reaction('O2CH2OCH2O2H <=> HO2CH2OCHO + OH', [4.000000e+10, 0.0, 18580.0])

# Reaction 459
reaction('HO2CH2OCHO <=> OCH2OCHO + OH', [2.000000e+16, 0.0, 40500.0])

# Reaction 460
reaction('CH2O + OCHO <=> OCH2OCHO', [1.250000e+11, 0.0, 11900.0])

# Reaction 461
reaction('OCH2OCHO <=> HOCH2OCO', [1.000000e+11, 0.0, 14000.0])

# Reaction 462
reaction('HOCH2O + CO <=> HOCH2OCO', [1.500000e+11, 0.0, 4800.0])

# Reaction 463
reaction('CH2OH + CO2 <=> HOCH2OCO', [1.500000e+11, 0.0, 35720.0])

# Reaction 464
reaction('CH2OCHO + H <=> CH3OCHO', [1.000000e+14, 0.0, 0.0])

# Reaction 465
reaction('CH3OCO + H <=> CH3OCHO', [1.000000e+14, 0.0, 0.0])

# Reaction 466
falloff_reaction('CH3OCHO (+ M) <=> CH3OH + CO (+ M)',
                kf=[1.000000e+14, 0.0, 62500.0],
                kf0=[6.143000e+60, -12.07, 75400.0],
                falloff=Troe(A=0.78, T3=8280000000.0, T1=438.9, T2=670000000.0))

# Reaction 467
reaction('CH3O + HCO <=> CH3OCHO', [3.000000e+13, 0.0, 0.0])

# Reaction 468
reaction('CH3 + OCHO <=> CH3OCHO', [1.000000e+13, 0.0, 0.0])

# Reaction 469
reaction('CH3OCHO + O2 <=> CH3OCO + HO2', [1.000000e+13, 0.0, 49700.0])

# Reaction 470
reaction('CH3OCHO + O2 <=> CH2OCHO + HO2', [2.050000e+13, 0.0, 52000.0])

# Reaction 471
reaction('CH3OCHO + OH <=> CH3OCO + H2O', [1.580000e+07, 1.8, 934.0])

# Reaction 472
reaction('CH3OCHO + OH <=> CH2OCHO + H2O', [5.270000e+09, 0.97, 1586.0])

# Reaction 473
reaction('CH3OCHO + HO2 <=> CH3OCO + H2O2', [4.820000e+03, 2.6, 13910.0])

# Reaction 474
reaction('CH3OCHO + HO2 <=> CH2OCHO + H2O2', [2.380000e+04, 2.55, 16490.0])

# Reaction 475
reaction('CH3OCHO + O <=> CH3OCO + OH', [2.755000e+05, 2.45, 2830.0])

# Reaction 476
reaction('CH3OCHO + O <=> CH2OCHO + OH', [9.800000e+05, 2.43, 4750.0])

# Reaction 477
reaction('CH3OCHO + H <=> CH3OCO + H2', [6.500000e+05, 2.4, 4471.0])

# Reaction 478
reaction('CH3OCHO + H <=> CH2OCHO + H2', [6.650000e+05, 2.54, 6756.0])

# Reaction 479
reaction('CH3OCHO + CH3 <=> CH3OCO + CH4', [7.550000e-01, 3.46, 5481.0])

# Reaction 480
reaction('CH3OCHO + CH3 <=> CH2OCHO + CH4', [4.520000e-01, 3.65, 7154.0])

# Reaction 481
reaction('CH3OCHO + CH3O <=> CH3OCO + CH3OH', [5.480000e+11, 0.0, 5000.0])

# Reaction 482
reaction('CH3OCHO + CH3O <=> CH2OCHO + CH3OH', [2.170000e+11, 0.0, 6458.0])

# Reaction 483
reaction('CH3OCHO + CH3O2 <=> CH3OCO + CH3O2H', [4.820000e+03, 2.6, 13910.0])

# Reaction 484
reaction('CH3OCHO + CH3O2 <=> CH2OCHO + CH3O2H', [2.380000e+04, 2.55, 16490.0])

# Reaction 485
reaction('CH3OCHO + HCO <=> CH3OCO + CH2O', [5.400000e+06, 1.9, 17010.0])

# Reaction 486
reaction('CH3OCHO + HCO <=> CH2OCHO + CH2O', [1.025000e+05, 2.5, 18430.0])

# Reaction 487
reaction('CH3OCO <=> CH2OCHO', [1.629000e+12, -0.18, 40670.0])

# Reaction 488
reaction('CH3 + CO2 <=> CH3OCO', [4.760000e+07, 1.54, 34700.0])

# Reaction 489
reaction('CH3O + CO <=> CH3OCO', [1.550000e+06, 2.02, 5730.0])

# Reaction 490
reaction('CH2O + HCO <=> CH2OCHO', [1.500000e+11, 0.0, 11900.0])

# Reaction 491
falloff_reaction('C3H8 (+ M) <=> CH3 + C2H5 (+ M)',
                kf=[1.290000e+37, -5.84, 97380.0],
                kf0=[5.640000e+74, -15.74, 98714.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0 HE:0.7',
                falloff=Troe(A=0.31, T3=50.0, T1=3000.0, T2=9000.0))

# Reaction 492
reaction('NC3H7 + H <=> C3H8', [1.000000e+14, 0.0, 0.0])

# Reaction 493
reaction('IC3H7 + H <=> C3H8', [1.000000e+14, 0.0, 0.0])

# Reaction 494
reaction('C3H8 + O2 <=> IC3H7 + HO2', [2.000000e+13, 0.0, 49640.0])

# Reaction 495
reaction('C3H8 + O2 <=> NC3H7 + HO2', [6.000000e+13, 0.0, 52290.0])

# Reaction 496
reaction('H + C3H8 <=> H2 + IC3H7', [1.300000e+06, 2.4, 4471.0])

# Reaction 497
reaction('H + C3H8 <=> H2 + NC3H7', [3.490000e+05, 2.69, 6450.0])

# Reaction 498
reaction('C3H8 + O <=> IC3H7 + OH', [5.490000e+05, 2.5, 3140.0])

# Reaction 499
reaction('C3H8 + O <=> NC3H7 + OH', [3.710000e+06, 2.4, 5505.0])

# Reaction 500
reaction('C3H8 + OH <=> NC3H7 + H2O', [1.054000e+10, 0.97, 1586.0])

# Reaction 501
reaction('C3H8 + OH <=> IC3H7 + H2O', [4.670000e+07, 1.61, -35.0])

# Reaction 502
reaction('C3H8 + HO2 <=> IC3H7 + H2O2', [6.320000e+01, 3.37, 13720.0])

# Reaction 503
reaction('C3H8 + HO2 <=> NC3H7 + H2O2', [4.080000e+01, 3.59, 17160.0])

# Reaction 504
reaction('CH3 + C3H8 <=> CH4 + IC3H7', [6.400000e+04, 2.17, 7520.0])

# Reaction 505
reaction('CH3 + C3H8 <=> CH4 + NC3H7', [9.040000e-01, 3.65, 7154.0])

# Reaction 506
reaction('IC3H7 + C3H8 <=> NC3H7 + C3H8', [3.000000e+10, 0.0, 12900.0])

# Reaction 507
reaction('C2H3 + C3H8 <=> C2H4 + IC3H7', [1.000000e+11, 0.0, 10400.0])

# Reaction 508
reaction('C2H3 + C3H8 <=> C2H4 + NC3H7', [1.000000e+11, 0.0, 10400.0])

# Reaction 509
reaction('C2H5 + C3H8 <=> C2H6 + IC3H7', [1.000000e+11, 0.0, 10400.0])

# Reaction 510
reaction('C2H5 + C3H8 <=> C2H6 + NC3H7', [1.000000e+11, 0.0, 10400.0])

# Reaction 511
reaction('C3H8 + C3H5-A <=> NC3H7 + C3H6', [7.940000e+11, 0.0, 20500.0])

# Reaction 512
reaction('C3H8 + C3H5-A <=> IC3H7 + C3H6', [7.940000e+11, 0.0, 16200.0])

# Reaction 513
reaction('C3H8 + CH3O <=> NC3H7 + CH3OH', [3.000000e+11, 0.0, 7000.0])

# Reaction 514
reaction('C3H8 + CH3O <=> IC3H7 + CH3OH', [3.000000e+11, 0.0, 7000.0])

# Reaction 515
reaction('CH3O2 + C3H8 <=> CH3O2H + NC3H7', [1.386000e+00, 3.97, 18280.0])

# Reaction 516
reaction('CH3O2 + C3H8 <=> CH3O2H + IC3H7', [1.019000e+01, 3.58, 14810.0])

# Reaction 517
reaction('C2H5O2 + C3H8 <=> C2H5O2H + NC3H7', [1.386000e+00, 3.97, 18280.0])

# Reaction 518
reaction('C2H5O2 + C3H8 <=> C2H5O2H + IC3H7', [1.019000e+01, 3.58, 14810.0])

# Reaction 519
reaction('NC3H7O2 + C3H8 <=> NC3H7O2H + NC3H7', [1.700000e+13, 0.0, 20460.0])

# Reaction 520
reaction('NC3H7O2 + C3H8 <=> NC3H7O2H + IC3H7', [2.000000e+12, 0.0, 17000.0])

# Reaction 521
reaction('IC3H7O2 + C3H8 <=> IC3H7O2H + NC3H7', [1.700000e+13, 0.0, 20460.0])

# Reaction 522
reaction('IC3H7O2 + C3H8 <=> IC3H7O2H + IC3H7', [2.000000e+12, 0.0, 17000.0])

# Reaction 523
reaction('C3H8 + CH3CO3 <=> IC3H7 + CH3CO3H', [2.000000e+12, 0.0, 17000.0])

# Reaction 524
reaction('C3H8 + CH3CO3 <=> NC3H7 + CH3CO3H', [1.700000e+13, 0.0, 20460.0])

# Reaction 525
reaction('C3H8 + O2CHO <=> NC3H7 + HO2CHO', [5.520000e+04, 2.55, 16480.0])

# Reaction 526
reaction('C3H8 + O2CHO <=> IC3H7 + HO2CHO', [1.475000e+04, 2.6, 13910.0])

# Reaction 527
reaction('H + C3H6 <=> IC3H7', [4.240000e+11, 0.51, 1230.0])

# Reaction 528
reaction('IC3H7 + H <=> C2H5 + CH3', [2.000000e+13, 0.0, 0.0])

# Reaction 529
reaction('IC3H7 + O2 <=> C3H6 + HO2', [4.500000e-19, 0.0, 5020.0])

# Reaction 530
reaction('IC3H7 + OH <=> C3H6 + H2O', [2.410000e+13, 0.0, 0.0])

# Reaction 531
reaction('IC3H7 + O <=> CH3COCH3 + H', [4.818000e+13, 0.0, 0.0])

# Reaction 532
reaction('IC3H7 + O <=> CH3CHO + CH3', [4.818000e+13, 0.0, 0.0])

# Reaction 533
reaction('CH3 + C2H4 <=> NC3H7', [1.760000e+04, 2.48, 6130.0])

# Reaction 534
reaction('H + C3H6 <=> NC3H7', [2.500000e+11, 0.51, 2620.0])

# Reaction 535
reaction('NC3H7 + O2 <=> C3H6 + HO2', [3.000000e-19, 0.0, 3000.0])

# Reaction 536
reaction('C2H5CHO + NC3H7 <=> C2H5CO + C3H8', [1.700000e+12, 0.0, 8440.0])

# Reaction 537
reaction('C2H5CHO + IC3H7 <=> C2H5CO + C3H8', [1.700000e+12, 0.0, 8440.0])

# Reaction 538
reaction('C2H5CHO + C3H5-A <=> C2H5CO + C3H6', [1.700000e+12, 0.0, 8440.0])

# Reaction 539
falloff_reaction('C2H3 + CH3 (+ M) <=> C3H6 (+ M)',
                kf=[2.500000e+13, 0.0, 0.0],
                kf0=[4.270000e+58, -11.94, 9769.8],
                falloff=Troe(A=0.175, T3=1341.0, T1=60000.0, T2=10140.0))

# Reaction 540
falloff_reaction('C3H5-A + H (+ M) <=> C3H6 (+ M)',
                kf=[2.000000e+14, 0.0, 0.0],
                kf0=[1.330000e+60, -12.0, 5967.8],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.02, T3=1097.0, T1=1097.0, T2=6860.0))

# Reaction 541
reaction('C3H6 <=> C3H5-S + H', [7.710000e+69, -16.09, 140000.0])

# Reaction 542
reaction('C3H6 <=> C3H5-T + H', [5.620000e+71, -16.58, 139300.0])

# Reaction 543
reaction('C3H6 + O <=> C2H5 + HCO', [1.580000e+07, 1.76, -1216.0])

# Reaction 544
reaction('C3H6 + O => CH2CO + CH3 + H', [2.500000e+07, 1.76, 76.0])

# Reaction 545
reaction('C3H6 + O => CH3CHCO + H + H', [2.500000e+07, 1.76, 76.0])

# Reaction 546
reaction('C3H6 + O <=> C3H5-A + OH', [5.240000e+11, 0.7, 5884.0])

# Reaction 547
reaction('C3H6 + O <=> C3H5-S + OH', [1.200000e+11, 0.7, 8959.0])

# Reaction 548
reaction('C3H6 + O <=> C3H5-T + OH', [6.030000e+10, 0.7, 7632.0])

# Reaction 549
reaction('C3H6 + OH <=> C3H5-A + H2O', [1.970000e+06, 2.2, 540.0])

# Reaction 550
reaction('C3H6 + OH <=> C3H5-S + H2O', [2.110000e+06, 2.0, 2778.0])

# Reaction 551
reaction('C3H6 + OH <=> C3H5-T + H2O', [1.110000e+06, 2.0, 1451.0])

# Reaction 552
reaction('C3H6 + HO2 <=> C3H5-A + H2O2', [2.700000e+04, 2.5, 12340.0])

# Reaction 553
reaction('C3H6 + HO2 <=> C3H5-S + H2O2', [1.800000e+04, 2.5, 27620.0])

# Reaction 554
reaction('C3H6 + HO2 <=> C3H5-T + H2O2', [9.000000e+03, 2.5, 23590.0])

# Reaction 555
reaction('C3H6 + H <=> C3H5-A + H2', [1.730000e+05, 2.5, 2492.0])

# Reaction 556
reaction('C3H6 + H <=> C3H5-S + H2', [8.040000e+05, 2.5, 12280.0])

# Reaction 557
reaction('C3H6 + H <=> C3H5-T + H2', [4.050000e+05, 2.5, 9794.0])

# Reaction 558
pdep_arrhenius('C3H6 + H <=> C2H4 + CH3',
            [(0.1, 'atm'), 8.800000e+16, -1.05, 6461.0],
            [(1.0, 'atm'), 8.000000e+21, -2.39, 11180.0],
            [(10.0, 'atm'), 3.300000e+24, -3.04, 15610.0])

# Reaction 559
reaction('C3H6 + O2 <=> C3H5-A + HO2', [4.000000e+12, 0.0, 39900.0])

# Reaction 560
reaction('C3H6 + O2 <=> C3H5-S + HO2', [2.000000e+12, 0.0, 62900.0])

# Reaction 561
reaction('C3H6 + O2 <=> C3H5-T + HO2', [1.400000e+12, 0.0, 60700.0])

# Reaction 562
reaction('C3H6 + CH3 <=> C3H5-A + CH4', [2.210000e+00, 3.5, 5675.0])

# Reaction 563
reaction('C3H6 + CH3 <=> C3H5-S + CH4', [1.348000e+00, 3.5, 12850.0])

# Reaction 564
reaction('C3H6 + CH3 <=> C3H5-T + CH4', [8.400000e-01, 3.5, 11660.0])

# Reaction 565
reaction('C3H6 + C2H5 <=> C3H5-A + C2H6', [1.000000e+11, 0.0, 9800.0])

# Reaction 566
reaction('C3H6 + CH3CO3 <=> C3H5-A + CH3CO3H', [3.240000e+11, 0.0, 14900.0])

# Reaction 567
reaction('C3H6 + CH3O2 <=> C3H5-A + CH3O2H', [3.240000e+11, 0.0, 14900.0])

# Reaction 568
reaction('C3H6 + HO2 <=> C3H6O1-2 + OH', [1.290000e+12, 0.0, 14900.0])

# Reaction 569
reaction('C3H6 + C2H5O2 <=> C3H5-A + C2H5O2H', [3.240000e+11, 0.0, 14900.0])

# Reaction 570
reaction('C3H6 + NC3H7O2 <=> C3H5-A + NC3H7O2H', [3.240000e+11, 0.0, 14900.0])

# Reaction 571
reaction('C3H6 + IC3H7O2 <=> C3H5-A + IC3H7O2H', [3.240000e+11, 0.0, 14900.0])

# Reaction 572
reaction('C3H6 + OH <=> C3H6OH', [9.930000e+11, 0.0, -960.0])

# Reaction 573
reaction('C3H6OH + O2 <=> HOC3H6O2', [1.200000e+11, 0.0, -1100.0])

# Reaction 574
reaction('HOC3H6O2 => CH3CHO + CH2O + OH', [1.250000e+10, 0.0, 18900.0])

# Reaction 575
pdep_arrhenius('C2H2 + CH3 <=> C3H5-A',
            [(0.1, 'atm'), 8.200000e+53, -13.32, 33200.0],
            [(1.0, 'atm'), 2.680000e+53, -12.82, 35730.0],
            [(2.0, 'atm'), 3.640000e+52, -12.46, 36127.0],
            [(5.0, 'atm'), 1.040000e+51, -11.89, 36476.0],
            [(10.0, 'atm'), 4.400000e+49, -11.4, 36700.0],
            [(100.0, 'atm'), 3.800000e+44, -9.63, 37600.0])

# Reaction 576
pdep_arrhenius('C3H4-A + H <=> C3H5-A',
            [(0.1, 'atm'), 9.600000e+61, -14.67, 26000.0],
            [(1.0, 'atm'), 1.520000e+59, -13.54, 26949.0],
            [(2.0, 'atm'), 3.780000e+57, -12.98, 26785.0],
            [(5.0, 'atm'), 7.340000e+54, -12.09, 26187.0],
            [(10.0, 'atm'), 2.400000e+52, -11.3, 25400.0],
            [(100.0, 'atm'), 6.900000e+41, -8.06, 21300.0])

# Reaction 577
reaction('C3H5-A + HO2 <=> C3H5O + OH', [7.000000e+12, 0.0, -1000.0])

# Reaction 578
reaction('C3H5-A + CH3O2 <=> C3H5O + CH3O', [7.000000e+12, 0.0, -1000.0])

# Reaction 579
reaction('C3H5-A + H <=> C3H4-A + H2', [1.232000e+03, 3.035, 2582.0])

# Reaction 580
reaction('C3H5-A + O <=> C2H3CHO + H', [6.000000e+13, 0.0, 0.0])

# Reaction 581
pdep_arrhenius('C3H5-A + OH => C2H3CHO + H + H',
            [(0.1, 'atm'), 5.300000e+37, -6.71, 29306.0],
            [(1.0, 'atm'), 4.200000e+32, -5.16, 30126.0],
            [(10.0, 'atm'), 1.600000e+20, -1.56, 26330.0])

# Reaction 582
reaction('C3H5-A + OH <=> C3H4-A + H2O', [6.000000e+12, 0.0, 0.0])

# Reaction 583
reaction('C3H5-A + CH3 <=> C3H4-A + CH4', [3.000000e+12, -0.32, -131.0])

# Reaction 584
reaction('C3H5-A + C2H5 <=> C2H6 + C3H4-A', [4.000000e+11, 0.0, 0.0])

# Reaction 585
reaction('C3H5-A + C2H5 <=> C2H4 + C3H6', [4.000000e+11, 0.0, 0.0])

# Reaction 586
reaction('C3H5-A + C2H3 <=> C2H4 + C3H4-A', [1.000000e+12, 0.0, 0.0])

# Reaction 587
reaction('C3H5-A + C3H5-A <=> C3H4-A + C3H6', [8.430000e+10, 0.0, -262.0])

# Reaction 588
pdep_arrhenius('C3H5-A + O2 <=> C3H4-A + HO2',
            [(1.0, 'atm'), 4.990000e+15, -1.4, 22428.0],
            [(10.0, 'atm'), 2.180000e+21, -2.85, 30755.0])

# Reaction 589
pdep_arrhenius('C3H5-A + O2 <=> CH3CO + CH2O',
            [(1.0, 'atm'), 1.190000e+15, -1.01, 20128.0],
            [(10.0, 'atm'), 7.140000e+15, -1.21, 21046.0])

# Reaction 590
pdep_arrhenius('C3H5-A + O2 <=> C2H3CHO + OH',
            [(1.0, 'atm'), 1.820000e+13, -0.41, 22859.0],
            [(10.0, 'atm'), 2.470000e+13, -0.45, 23017.0])

# Reaction 591
reaction('C3H5-A + HCO <=> C3H6 + CO', [6.000000e+13, 0.0, 0.0])

# Reaction 592
reaction('C2H3 + CH3 <=> C3H5-A + H', [1.500000e+24, -2.83, 18618.0])

# Reaction 593
pdep_arrhenius('C3H5-A <=> C3H5-T',
            [(0.1, 'atm'), 3.900000e+59, -15.42, 75400.0],
            [(1.0, 'atm'), 7.060000e+56, -14.08, 75868.0],
            [(2.0, 'atm'), 4.800000e+55, -13.59, 75949.0],
            [(5.0, 'atm'), 4.860000e+53, -12.81, 75883.0],
            [(10.0, 'atm'), 6.400000e+51, -12.12, 75700.0],
            [(100.0, 'atm'), 2.800000e+43, -9.27, 74000.0])

# Reaction 594
pdep_arrhenius('C3H5-A <=> C3H5-S',
            [(0.1, 'atm'), 1.300000e+55, -14.53, 73800.0],
            [(1.0, 'atm'), 5.000000e+51, -13.02, 73300.0],
            [(10.0, 'atm'), 9.700000e+48, -11.73, 73700.0],
            [(100.0, 'atm'), 4.860000e+44, -9.84, 73400.0])

# Reaction 595
pdep_arrhenius('C2H2 + CH3 <=> C3H5-S',
            [(0.1, 'atm'), 1.400000e+32, -7.14, 10000.0],
            [(1.0, 'atm'), 3.200000e+35, -7.76, 13300.0],
            [(10.0, 'atm'), 2.400000e+38, -8.21, 17100.0],
            [(100.0, 'atm'), 1.400000e+39, -8.06, 20200.0])

# Reaction 596
reaction('C3H5-S + O2 <=> CH3CHO + HCO', [1.000000e+11, 0.0, 0.0])

# Reaction 597
reaction('C3H5-S + H <=> C3H4-A + H2', [3.333000e+12, 0.0, 0.0])

# Reaction 598
reaction('C3H5-S + CH3 <=> C3H4-A + CH4', [1.000000e+11, 0.0, 0.0])

# Reaction 599
pdep_arrhenius('C2H2 + CH3 <=> C3H5-T',
            [(0.1, 'atm'), 6.800000e+20, -4.16, 18000.0],
            [(1.0, 'atm'), 4.990000e+22, -4.39, 18850.0],
            [(2.0, 'atm'), 6.000000e+23, -4.6, 19571.0],
            [(5.0, 'atm'), 7.310000e+25, -5.06, 21150.0],
            [(10.0, 'atm'), 9.300000e+27, -5.55, 22900.0],
            [(100.0, 'atm'), 3.800000e+36, -7.58, 31300.0])

# Reaction 600
pdep_arrhenius('C3H4-A + H <=> C3H5-T',
            [(0.1, 'atm'), 9.200000e+38, -8.65, 7000.0],
            [(1.0, 'atm'), 9.460000e+42, -9.43, 11190.0],
            [(2.0, 'atm'), 8.470000e+43, -9.59, 12462.0],
            [(5.0, 'atm'), 6.980000e+44, -9.7, 14032.0],
            [(10.0, 'atm'), 1.500000e+45, -9.69, 15100.0],
            [(100.0, 'atm'), 1.800000e+43, -8.78, 16800.0])

# Reaction 601
pdep_arrhenius('C3H4-P + H <=> C3H5-T',
            [(0.1, 'atm'), 4.600000e+44, -10.21, 10200.0],
            [(1.0, 'atm'), 1.660000e+47, -10.58, 13690.0],
            [(2.0, 'atm'), 5.040000e+47, -10.61, 14707.0],
            [(5.0, 'atm'), 9.620000e+47, -10.55, 15910.0],
            [(10.0, 'atm'), 7.000000e+47, -10.4, 16600.0],
            [(100.0, 'atm'), 3.200000e+44, -9.11, 17400.0])

# Reaction 602
reaction('C3H5-S + H <=> C3H4-P + H2', [3.340000e+12, 0.0, 0.0])

# Reaction 603
reaction('C3H5-S + O <=> C2H4 + HCO', [6.000000e+13, 0.0, 0.0])

# Reaction 604
reaction('C3H5-S + OH => C2H4 + HCO + H', [5.000000e+12, 0.0, 0.0])

# Reaction 605
reaction('C3H5-S + HO2 => C2H4 + HCO + OH', [2.000000e+13, 0.0, 0.0])

# Reaction 606
reaction('C3H5-S + HCO <=> C3H6 + CO', [9.000000e+13, 0.0, 0.0])

# Reaction 607
reaction('C3H5-S + CH3 <=> C3H4-P + CH4', [1.000000e+11, 0.0, 0.0])

# Reaction 608
pdep_arrhenius('C3H5-T <=> C3H5-S',
            [(0.1, 'atm'), 1.600000e+44, -12.16, 52200.0],
            [(1.0, 'atm'), 1.500000e+48, -12.71, 53900.0],
            [(10.0, 'atm'), 5.100000e+52, -13.37, 57200.0],
            [(100.0, 'atm'), 5.800000e+51, -12.43, 59200.0])

# Reaction 609
pdep_arrhenius('C3H4-A + H <=> C3H5-S',
            [(0.1, 'atm'), 1.100000e+30, -6.52, 15200.0],
            [(1.0, 'atm'), 5.400000e+29, -6.09, 16300.0],
            [(10.0, 'atm'), 2.600000e+31, -6.23, 18700.0],
            [(100.0, 'atm'), 3.200000e+31, -5.88, 21500.0])

# Reaction 610
reaction('C3H4-A + C3H4-A <=> C3H5-A + C3H3', [5.000000e+14, 0.0, 64746.7])

# Reaction 611
reaction('C3H5-T + O2 <=> C3H4-A + HO2', [1.890000e+30, -5.59, 15540.0])

# Reaction 612
reaction('C3H5-T + O2 <=> CH3COCH2 + O', [3.810000e+17, -1.36, 5580.0])

# Reaction 613
reaction('C3H5-T + O2 <=> CH3CO + CH2O', [1.000000e+11, 0.0, 0.0])

# Reaction 614
reaction('C3H5-T + H <=> C3H4-P + H2', [3.340000e+12, 0.0, 0.0])

# Reaction 615
reaction('C3H5-T + CH3 <=> C3H4-P + CH4', [1.000000e+11, 0.0, 0.0])

# Reaction 616
reaction('C3H5-T + O <=> CH3 + CH2CO', [6.000000e+13, 0.0, 0.0])

# Reaction 617
reaction('C3H5-T + OH => CH3 + CH2CO + H', [5.000000e+12, 0.0, 0.0])

# Reaction 618
reaction('C3H5-T + HO2 => CH3 + CH2CO + OH', [2.000000e+13, 0.0, 0.0])

# Reaction 619
reaction('C3H5-T + HCO <=> C3H6 + CO', [9.000000e+13, 0.0, 0.0])

# Reaction 620
pdep_arrhenius('C3H4-P <=> C3H4-A',
            [(0.1, 'atm'), 6.400000e+61, -14.59, 88200.0],
            [(0.4, 'atm'), 5.810000e+62, -14.63, 91211.0],
            [(1.0, 'atm'), 5.150000e+60, -13.93, 91117.0],
            [(2.0, 'atm'), 7.640000e+59, -13.59, 91817.0],
            [(5.0, 'atm'), 3.120000e+58, -13.07, 92680.0],
            [(10.0, 'atm'), 1.900000e+57, -12.62, 93300.0],
            [(100.0, 'atm'), 1.400000e+52, -10.86, 95400.0])

# Reaction 621
reaction('C3H3 + HO2 <=> C3H4-A + O2', [3.000000e+11, 0.0, 0.0])

# Reaction 622
reaction('C3H4-A + HO2 => CH2CO + CH2 + OH', [4.000000e+12, 0.0, 19000.0])

# Reaction 623
reaction('C3H4-A + OH <=> CH2CO + CH3', [3.120000e+12, 0.0, -397.0])

# Reaction 624
reaction('C3H4-A + OH <=> C3H3 + H2O', [5.300000e+06, 2.0, 2000.0])

# Reaction 625
reaction('C3H4-A + O <=> C2H4 + CO', [2.000000e+07, 1.8, 1000.0])

# Reaction 626
reaction('C3H4-A + O <=> C2H2 + CH2O', [3.000000e-03, 4.61, -4243.0])

# Reaction 627
reaction('C3H4-A + H <=> C3H3 + H2', [1.300000e+06, 2.0, 5500.0])

# Reaction 628
reaction('C3H4-A + CH3 <=> C3H3 + CH4', [1.300000e+12, 0.0, 7700.0])

# Reaction 629
reaction('C3H4-A + C3H5-A <=> C3H3 + C3H6', [2.000000e+11, 0.0, 7700.0])

# Reaction 630
reaction('C3H4-A + C2H <=> C2H2 + C3H3', [1.000000e+13, 0.0, 0.0])

# Reaction 631
pdep_arrhenius('C3H4-P <=> CC3H4',
            [(0.1, 'atm'), 3.400000e+46, -10.97, 68900.0],
            [(0.4, 'atm'), 2.840000e+45, -10.45, 69284.0],
            [(1.0, 'atm'), 1.200000e+44, -9.92, 69250.0],
            [(2.0, 'atm'), 5.470000e+42, -9.43, 69089.0],
            [(5.0, 'atm'), 3.920000e+40, -8.69, 68706.0],
            [(10.0, 'atm'), 5.300000e+38, -8.06, 68300.0],
            [(100.0, 'atm'), 2.800000e+31, -5.69, 66400.0])

# Reaction 632
pdep_arrhenius('C3H4-P + H <=> C3H4-A + H',
            [(0.1, 'atm'), 2.300000e+15, -0.26, 7600.0],
            [(1.0, 'atm'), 6.270000e+17, -0.91, 10079.0],
            [(2.0, 'atm'), 1.500000e+18, -1.0, 10756.0],
            [(5.0, 'atm'), 1.930000e+18, -1.01, 11523.0],
            [(10.0, 'atm'), 3.100000e+22, -2.18, 14800.0],
            [(100.0, 'atm'), 6.400000e+27, -3.58, 21200.0])

# Reaction 633
pdep_arrhenius('C3H4-P + H <=> C3H5-S',
            [(0.1, 'atm'), 1.000000e+25, -5.0, 1800.0],
            [(1.0, 'atm'), 5.500000e+28, -5.74, 4300.0],
            [(10.0, 'atm'), 1.000000e+34, -6.88, 8900.0],
            [(100.0, 'atm'), 9.700000e+37, -7.63, 13800.0])

# Reaction 634
pdep_arrhenius('C3H4-P + H <=> C3H5-A',
            [(0.1, 'atm'), 1.100000e+60, -14.56, 28100.0],
            [(1.0, 'atm'), 4.910000e+60, -14.37, 31644.0],
            [(2.0, 'atm'), 3.040000e+60, -14.19, 32642.0],
            [(5.0, 'atm'), 9.020000e+59, -13.89, 33953.0],
            [(10.0, 'atm'), 2.200000e+59, -13.61, 34900.0],
            [(100.0, 'atm'), 1.600000e+55, -12.07, 37500.0])

# Reaction 635
reaction('C3H4-P + H <=> C3H3 + H2', [1.300000e+06, 2.0, 5500.0])

# Reaction 636
reaction('C3H4-P + C3H3 <=> C3H4-A + C3H3', [6.140000e+06, 1.74, 10450.0])

# Reaction 637
reaction('C3H4-P + O <=> HCCO + CH3', [7.300000e+12, 0.0, 2250.0])

# Reaction 638
reaction('C3H4-P + O <=> C2H4 + CO', [1.000000e+13, 0.0, 2250.0])

# Reaction 639
reaction('C3H4-P + OH <=> C3H3 + H2O', [1.000000e+06, 2.0, 100.0])

# Reaction 640
reaction('C3H4-P + C2H <=> C2H2 + C3H3', [1.000000e+13, 0.0, 0.0])

# Reaction 641
reaction('C3H4-P + CH3 <=> C3H3 + CH4', [1.800000e+12, 0.0, 7700.0])

# Reaction 642
pdep_arrhenius('C2H2 + CH3 <=> C3H4-P + H',
            [(0.1, 'atm'), 4.500000e+06, 1.86, 11600.0],
            [(1.0, 'atm'), 2.560000e+09, 1.1, 13644.0],
            [(2.0, 'atm'), 2.070000e+10, 0.85, 14415.0],
            [(5.0, 'atm'), 2.510000e+11, 0.56, 15453.0],
            [(10.0, 'atm'), 1.100000e+12, 0.39, 16200.0],
            [(100.0, 'atm'), 2.100000e+12, 0.37, 18100.0])

# Reaction 643
pdep_arrhenius('CC3H4 <=> C3H4-A',
            [(0.1, 'atm'), 2.300000e+39, -8.81, 47800.0],
            [(0.4, 'atm'), 7.590000e+40, -9.07, 48831.0],
            [(1.0, 'atm'), 4.890000e+41, -9.17, 49594.0],
            [(2.0, 'atm'), 8.810000e+41, -9.15, 50073.0],
            [(5.0, 'atm'), 4.330000e+41, -8.93, 50475.0],
            [(10.0, 'atm'), 7.200000e+40, -8.6, 50600.0],
            [(100.0, 'atm'), 1.600000e+35, -6.64, 49500.0])

# Reaction 644
reaction('C3H3 + H <=> C3H4-P', [1.500000e+13, 0.0, 0.0])

# Reaction 645
reaction('C3H3 + H <=> C3H4-A', [2.500000e+12, 0.0, 0.0])

# Reaction 646
reaction('C2H + CH3 <=> C3H4-P', [8.000000e+13, 0.0, 0.0])

# Reaction 647
reaction('C3H3 + HO2 <=> C3H4-P + O2', [2.500000e+12, 0.0, 0.0])

# Reaction 648
reaction('C3H4-P + HO2 => C2H4 + CO + OH', [3.000000e+12, 0.0, 19000.0])

# Reaction 649
reaction('C3H4-P + OH <=> CH2CO + CH3', [5.000000e-04, 4.5, -1000.0])

# Reaction 650
reaction('C3H4-P + O <=> C2H3 + HCO', [3.200000e+12, 0.0, 2010.0])

# Reaction 651
reaction('C3H4-P + O <=> C3H3 + OH', [7.650000e+08, 1.5, 8600.0])

# Reaction 652
reaction('C3H4-P + C2H3 <=> C3H3 + C2H4', [1.000000e+12, 0.0, 7700.0])

# Reaction 653
reaction('C3H4-P + C3H5-A <=> C3H3 + C3H6', [1.000000e+12, 0.0, 7700.0])

# Reaction 654
reaction('C3H3 + O <=> CH2O + C2H', [2.000000e+13, 0.0, 0.0])

# Reaction 655
reaction('C3H3 + O2 <=> CH2CO + HCO', [3.000000e+10, 0.0, 2868.0])

# Reaction 656
reaction('C3H3 + HO2 => OH + CO + C2H3', [8.000000e+11, 0.0, 0.0])

# Reaction 657
reaction('C3H3 + HCO <=> C3H4-A + CO', [2.500000e+13, 0.0, 0.0])

# Reaction 658
reaction('C3H3 + HCO <=> C3H4-P + CO', [2.500000e+13, 0.0, 0.0])

# Reaction 659
reaction('C2H5 + C2H <=> C3H3 + CH3', [1.810000e+13, 0.0, 0.0])

# Reaction 660
reaction('C3H4-A + HO2 => C2H4 + CO + OH', [1.000000e+12, 0.0, 14000.0])

# Reaction 661
reaction('C3H4-A + HO2 <=> C3H3 + H2O2', [3.000000e+13, 0.0, 14000.0])

# Reaction 662
pdep_arrhenius('C2H2 + CH3 <=> C3H4-A + H',
            [(0.1, 'atm'), 2.400000e+09, 0.91, 20700.0],
            [(1.0, 'atm'), 5.140000e+09, 0.86, 22153.0],
            [(2.0, 'atm'), 1.330000e+10, 0.75, 22811.0],
            [(5.0, 'atm'), 9.200000e+10, 0.54, 23950.0],
            [(10.0, 'atm'), 5.100000e+11, 0.35, 25000.0],
            [(100.0, 'atm'), 7.300000e+12, 0.11, 28500.0])

# Reaction 663
reaction('CH3CHCO + OH <=> C2H5 + CO2', [1.730000e+12, 0.0, -1010.0])

# Reaction 664
reaction('CH3CHCO + OH <=> SC2H4OH + CO', [2.000000e+12, 0.0, -1010.0])

# Reaction 665
reaction('CH3CHCO + H <=> C2H5 + CO', [4.400000e+12, 0.0, 1459.0])

# Reaction 666
reaction('CH3CHCO + O <=> CH3CHO + CO', [3.200000e+12, 0.0, -437.0])

# Reaction 667
reaction('NC3H7 + HO2 <=> NC3H7O + OH', [7.000000e+12, 0.0, -1000.0])

# Reaction 668
reaction('IC3H7 + HO2 <=> IC3H7O + OH', [7.000000e+12, 0.0, -1000.0])

# Reaction 669
reaction('CH3O2 + NC3H7 <=> CH3O + NC3H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 670
reaction('CH3O2 + IC3H7 <=> CH3O + IC3H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 671
reaction('NC3H7 + O2 <=> NC3H7O2', [4.520000e+12, 0.0, 0.0])

# Reaction 672
reaction('IC3H7 + O2 <=> IC3H7O2', [7.540000e+12, 0.0, 0.0])

# Reaction 673
reaction('NC3H7O2 + CH2O <=> NC3H7O2H + HCO', [5.600000e+12, 0.0, 13600.0])

# Reaction 674
reaction('NC3H7O2 + CH3CHO <=> NC3H7O2H + CH3CO', [2.800000e+12, 0.0, 13600.0])

# Reaction 675
reaction('IC3H7O2 + CH2O <=> IC3H7O2H + HCO', [5.600000e+12, 0.0, 13600.0])

# Reaction 676
reaction('IC3H7O2 + CH3CHO <=> IC3H7O2H + CH3CO', [2.800000e+12, 0.0, 13600.0])

# Reaction 677
reaction('NC3H7O2 + HO2 <=> NC3H7O2H + O2', [1.750000e+10, 0.0, -3275.0])

# Reaction 678
reaction('IC3H7O2 + HO2 <=> IC3H7O2H + O2', [1.750000e+10, 0.0, -3275.0])

# Reaction 679
reaction('C2H4 + NC3H7O2 <=> C2H3 + NC3H7O2H', [1.130000e+13, 0.0, 30430.0])

# Reaction 680
reaction('C2H4 + IC3H7O2 <=> C2H3 + IC3H7O2H', [1.130000e+13, 0.0, 30430.0])

# Reaction 681
reaction('CH3OH + NC3H7O2 <=> CH2OH + NC3H7O2H', [6.300000e+12, 0.0, 19360.0])

# Reaction 682
reaction('CH3OH + IC3H7O2 <=> CH2OH + IC3H7O2H', [6.300000e+12, 0.0, 19360.0])

# Reaction 683
reaction('C2H3CHO + NC3H7O2 <=> C2H3CO + NC3H7O2H', [2.800000e+12, 0.0, 13600.0])

# Reaction 684
reaction('C2H3CHO + IC3H7O2 <=> C2H3CO + IC3H7O2H', [2.800000e+12, 0.0, 13600.0])

# Reaction 685
reaction('CH4 + NC3H7O2 <=> CH3 + NC3H7O2H', [1.120000e+13, 0.0, 24640.0])

# Reaction 686
reaction('CH4 + IC3H7O2 <=> CH3 + IC3H7O2H', [1.120000e+13, 0.0, 24640.0])

# Reaction 687
reaction('NC3H7O2 + CH3O2 => NC3H7O + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 688
reaction('IC3H7O2 + CH3O2 => IC3H7O + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 689
reaction('H2 + NC3H7O2 <=> H + NC3H7O2H', [3.010000e+13, 0.0, 26030.0])

# Reaction 690
reaction('H2 + IC3H7O2 <=> H + IC3H7O2H', [3.010000e+13, 0.0, 26030.0])

# Reaction 691
reaction('IC3H7O2 + C2H6 <=> IC3H7O2H + C2H5', [1.700000e+13, 0.0, 20460.0])

# Reaction 692
reaction('NC3H7O2 + C2H6 <=> NC3H7O2H + C2H5', [1.700000e+13, 0.0, 20460.0])

# Reaction 693
reaction('IC3H7O2 + C2H5CHO <=> IC3H7O2H + C2H5CO', [2.000000e+11, 0.0, 9500.0])

# Reaction 694
reaction('NC3H7O2 + C2H5CHO <=> NC3H7O2H + C2H5CO', [2.000000e+11, 0.0, 9500.0])

# Reaction 695
reaction('IC3H7O2 + CH3CO3 => IC3H7O + CH3CO2 + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 696
reaction('NC3H7O2 + CH3CO3 => NC3H7O + CH3CO2 + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 697
reaction('IC3H7O2 + C2H5O2 => IC3H7O + C2H5O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 698
reaction('NC3H7O2 + C2H5O2 => NC3H7O + C2H5O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 699
reaction('IC3H7O2 + IC3H7O2 => O2 + IC3H7O + IC3H7O', [1.400000e+16, -1.61, 1860.0])

# Reaction 700
reaction('NC3H7O2 + NC3H7O2 => O2 + NC3H7O + NC3H7O', [1.400000e+16, -1.61, 1860.0])

# Reaction 701
reaction('IC3H7O2 + NC3H7O2 => IC3H7O + NC3H7O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 702
reaction('IC3H7O2 + CH3 <=> IC3H7O + CH3O', [7.000000e+12, 0.0, -1000.0])

# Reaction 703
reaction('IC3H7O2 + C2H5 <=> IC3H7O + C2H5O', [7.000000e+12, 0.0, -1000.0])

# Reaction 704
reaction('IC3H7O2 + IC3H7 <=> IC3H7O + IC3H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 705
reaction('IC3H7O2 + NC3H7 <=> IC3H7O + NC3H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 706
reaction('IC3H7O2 + C3H5-A <=> IC3H7O + C3H5O', [7.000000e+12, 0.0, -1000.0])

# Reaction 707
reaction('NC3H7O2 + CH3 <=> NC3H7O + CH3O', [7.000000e+12, 0.0, -1000.0])

# Reaction 708
reaction('NC3H7O2 + C2H5 <=> NC3H7O + C2H5O', [7.000000e+12, 0.0, -1000.0])

# Reaction 709
reaction('NC3H7O2 + IC3H7 <=> NC3H7O + IC3H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 710
reaction('NC3H7O2 + NC3H7 <=> NC3H7O + NC3H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 711
reaction('NC3H7O2 + C3H5-A <=> NC3H7O + C3H5O', [7.000000e+12, 0.0, -1000.0])

# Reaction 712
reaction('NC3H7O2H <=> NC3H7O + OH', [1.500000e+16, 0.0, 42500.0])

# Reaction 713
reaction('IC3H7O + OH <=> IC3H7O2H', [1.000000e+15, -0.8, 0.0])

# Reaction 714
reaction('C2H5 + CH2O <=> NC3H7O', [1.000000e+11, 0.0, 3496.0])

# Reaction 715
reaction('C2H5CHO + H <=> NC3H7O', [4.000000e+12, 0.0, 6260.0])

# Reaction 716
reaction('CH3 + CH3CHO <=> IC3H7O', [1.000000e+11, 0.0, 9256.0])

# Reaction 717
reaction('CH3COCH3 + H <=> IC3H7O', [2.000000e+12, 0.0, 7270.0])

# Reaction 718
reaction('IC3H7O + O2 <=> CH3COCH3 + HO2', [9.090000e+09, 0.0, 390.0])

# Reaction 719
reaction('NC3H7O2 <=> C3H6OOH1-2', [6.000000e+11, 0.0, 26850.0])

# Reaction 720
reaction('NC3H7O2 <=> C3H6OOH1-3', [1.125000e+11, 0.0, 24400.0])

# Reaction 721
reaction('IC3H7O2 <=> C3H6OOH2-1', [1.800000e+12, 0.0, 29400.0])

# Reaction 722
reaction('IC3H7O2 <=> C3H6OOH2-2', [1.230000e+35, -6.96, 48880.0])

# Reaction 723
reaction('C3H6OOH1-2 <=> C3H6O1-2 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 724
reaction('C3H6OOH1-3 <=> C3H6O1-3 + OH', [7.500000e+10, 0.0, 15250.0])

# Reaction 725
reaction('C3H6OOH2-1 <=> C3H6O1-2 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 726
reaction('C3H6 + HO2 <=> C3H6OOH1-2', [1.000000e+11, 0.0, 11000.0])

# Reaction 727
reaction('C3H6 + HO2 <=> C3H6OOH2-1', [1.000000e+11, 0.0, 11750.0])

# Reaction 728
reaction('C3H6OOH1-3 => OH + CH2O + C2H4', [3.035000e+15, -0.79, 27400.0])

# Reaction 729
reaction('C3H6OOH2-1 <=> C2H3OOH + CH3', [6.540000e+27, -5.14, 38320.0])

# Reaction 730
reaction('C3H6OOH1-2 => C2H4 + CH2O + OH', [1.310000e+33, -7.01, 48120.0])

# Reaction 731
reaction('C3H6OOH2-2 <=> CH3COCH3 + OH', [9.000000e+14, 0.0, 1500.0])

# Reaction 732
reaction('C3H6OOH1-2 + O2 <=> C3H6OOH1-2O2', [5.000000e+12, 0.0, 0.0])

# Reaction 733
reaction('C3H6OOH1-3 + O2 <=> C3H6OOH1-3O2', [4.520000e+12, 0.0, 0.0])

# Reaction 734
reaction('C3H6OOH2-1 + O2 <=> C3H6OOH2-1O2', [4.520000e+12, 0.0, 0.0])

# Reaction 735
reaction('C3H6OOH1-2O2 <=> C3KET12 + OH', [6.000000e+11, 0.0, 26400.0])

# Reaction 736
reaction('C3H6OOH1-3O2 <=> C3KET13 + OH', [7.500000e+10, 0.0, 21400.0])

# Reaction 737
reaction('C3H6OOH2-1O2 <=> C3KET21 + OH', [3.000000e+11, 0.0, 23850.0])

# Reaction 738
reaction('C3H6OOH2-1O2 <=> C3H51-2,3OOH', [1.125000e+11, 0.0, 24400.0])

# Reaction 739
reaction('C3H6OOH1-2O2 <=> C3H51-2,3OOH', [9.000000e+11, 0.0, 29400.0])

# Reaction 740
reaction('C3H51-2,3OOH <=> AC3H5OOH + HO2', [2.560000e+13, -0.49, 17770.0])

# Reaction 741
reaction('C3H6OOH1-3O2 <=> C3H52-1,3OOH', [6.000000e+11, 0.0, 26850.0])

# Reaction 742
reaction('C3H52-1,3OOH <=> AC3H5OOH + HO2', [1.150000e+14, -0.63, 17250.0])

# Reaction 743
reaction('C3KET12 => CH3CHO + HCO + OH', [9.450000e+15, 0.0, 43000.0])

# Reaction 744
reaction('C3KET13 => CH2O + CH2CHO + OH', [1.000000e+16, 0.0, 43000.0])

# Reaction 745
reaction('C3KET21 => CH2O + CH3CO + OH', [1.000000e+16, 0.0, 43000.0])

# Reaction 746
reaction('C3H5O + OH <=> AC3H5OOH', [2.000000e+13, 0.0, 0.0])

# Reaction 747
reaction('C3H5O <=> C2H3CHO + H', [1.000000e+14, 0.0, 29100.0])

# Reaction 748
reaction('C2H3 + CH2O <=> C3H5O', [1.500000e+11, 0.0, 10600.0])

# Reaction 749
reaction('C3H5O + O2 <=> C2H3CHO + HO2', [1.000000e+12, 0.0, 6000.0])

# Reaction 750
reaction('C2H3OOH <=> CH2CHO + OH', [8.400000e+14, 0.0, 43000.0])

# Reaction 751
reaction('C3H6O1-2 <=> C2H4 + CH2O', [6.000000e+14, 0.0, 60000.0])

# Reaction 752
reaction('C3H6O1-2 + OH => CH2O + C2H3 + H2O', [5.000000e+12, 0.0, 0.0])

# Reaction 753
reaction('C3H6O1-2 + H => CH2O + C2H3 + H2', [2.630000e+07, 2.0, 5000.0])

# Reaction 754
reaction('C3H6O1-2 + O => CH2O + C2H3 + OH', [8.430000e+13, 0.0, 5200.0])

# Reaction 755
reaction('C3H6O1-2 + HO2 => CH2O + C2H3 + H2O2', [1.000000e+13, 0.0, 15000.0])

# Reaction 756
reaction('C3H6O1-2 + CH3O2 => CH2O + C2H3 + CH3O2H', [1.000000e+13, 0.0, 19000.0])

# Reaction 757
reaction('C3H6O1-2 + CH3 => CH2O + C2H3 + CH4', [2.000000e+11, 0.0, 10000.0])

# Reaction 758
reaction('C3H6O1-3 <=> C2H4 + CH2O', [6.000000e+14, 0.0, 60000.0])

# Reaction 759
reaction('C3H6O1-3 + OH => CH2O + C2H3 + H2O', [5.000000e+12, 0.0, 0.0])

# Reaction 760
reaction('C3H6O1-3 + O => CH2O + C2H3 + OH', [8.430000e+13, 0.0, 5200.0])

# Reaction 761
reaction('C3H6O1-3 + H => CH2O + C2H3 + H2', [2.630000e+07, 2.0, 5000.0])

# Reaction 762
reaction('C3H6O1-3 + CH3O2 => CH2O + C2H3 + CH3O2H', [1.000000e+13, 0.0, 19000.0])

# Reaction 763
reaction('C3H6O1-3 + HO2 => CH2O + C2H3 + H2O2', [1.000000e+13, 0.0, 15000.0])

# Reaction 764
reaction('C3H6O1-3 + CH3 => CH2O + C2H3 + CH4', [2.000000e+11, 0.0, 10000.0])

# Reaction 765
reaction('IC3H7O2 <=> C3H6 + HO2', [1.196000e+43, -9.43, 41530.0])

# Reaction 766
reaction('NC3H7O2 <=> C3H6 + HO2', [4.308000e+36, -7.5, 39510.0])

# Reaction 767
falloff_reaction('C4H10 (+ M) <=> C2H5 + C2H5 (+ M)',
                kf=[1.355000e+37, -6.036, 92929.0],
                kf0=[4.720000e+18, 0.0, 49578.0],
                falloff=Troe(A=0.07998, T3=1e-20, T1=32430.0, T2=4858.0))

# Reaction 768
falloff_reaction('C4H10 (+ M) <=> NC3H7 + CH3 (+ M)',
                kf=[6.600000e+52, -10.626, 100330.0],
                kf0=[5.340000e+17, 0.0, 42959.0],
                falloff=Troe(A=0.09502, T3=1e-20, T1=5348.0, T2=4326.0))

# Reaction 769
reaction('PC4H9 + H <=> C4H10', [3.610000e+13, 0.0, 0.0])

# Reaction 770
reaction('SC4H9 + H <=> C4H10', [3.610000e+13, 0.0, 0.0])

# Reaction 771
reaction('C4H10 + O2 <=> PC4H9 + HO2', [6.000000e+13, 0.0, 52340.0])

# Reaction 772
reaction('C4H10 + O2 <=> SC4H9 + HO2', [4.000000e+13, 0.0, 49800.0])

# Reaction 773
reaction('C4H10 + C3H5-A <=> PC4H9 + C3H6', [7.940000e+11, 0.0, 20500.0])

# Reaction 774
reaction('C4H10 + C3H5-A <=> SC4H9 + C3H6', [3.160000e+11, 0.0, 16400.0])

# Reaction 775
reaction('C4H10 + C2H5 <=> PC4H9 + C2H6', [1.580000e+11, 0.0, 12300.0])

# Reaction 776
reaction('C4H10 + C2H5 <=> SC4H9 + C2H6', [1.000000e+11, 0.0, 10400.0])

# Reaction 777
reaction('C4H10 + C2H3 <=> PC4H9 + C2H4', [1.000000e+12, 0.0, 18000.0])

# Reaction 778
reaction('C4H10 + C2H3 <=> SC4H9 + C2H4', [8.000000e+11, 0.0, 16800.0])

# Reaction 779
reaction('C4H10 + CH3 <=> PC4H9 + CH4', [9.040000e-01, 3.65, 7154.0])

# Reaction 780
reaction('C4H10 + CH3 <=> SC4H9 + CH4', [3.020000e+00, 3.46, 5481.0])

# Reaction 781
reaction('C4H10 + H <=> PC4H9 + H2', [3.490000e+05, 2.69, 6450.0])

# Reaction 782
reaction('C4H10 + H <=> SC4H9 + H2', [2.600000e+06, 2.4, 4471.0])

# Reaction 783
reaction('C4H10 + OH <=> PC4H9 + H2O', [1.054000e+10, 0.97, 1586.0])

# Reaction 784
reaction('C4H10 + OH <=> SC4H9 + H2O', [9.340000e+07, 1.61, -35.0])

# Reaction 785
reaction('C4H10 + O <=> PC4H9 + OH', [1.130000e+14, 0.0, 7850.0])

# Reaction 786
reaction('C4H10 + O <=> SC4H9 + OH', [5.620000e+13, 0.0, 5200.0])

# Reaction 787
reaction('C4H10 + HO2 <=> PC4H9 + H2O2', [4.080000e+01, 3.59, 17160.0])

# Reaction 788
reaction('C4H10 + HO2 <=> SC4H9 + H2O2', [1.264000e+02, 3.37, 13720.0])

# Reaction 789
reaction('C4H10 + CH3O <=> PC4H9 + CH3OH', [3.000000e+11, 0.0, 7000.0])

# Reaction 790
reaction('C4H10 + CH3O <=> SC4H9 + CH3OH', [6.000000e+11, 0.0, 7000.0])

# Reaction 791
reaction('C4H10 + C2H5O <=> PC4H9 + C2H5OH', [3.000000e+11, 0.0, 7000.0])

# Reaction 792
reaction('C4H10 + C2H5O <=> SC4H9 + C2H5OH', [6.000000e+11, 0.0, 7000.0])

# Reaction 793
reaction('C4H10 + PC4H9 <=> SC4H9 + C4H10', [1.000000e+11, 0.0, 10400.0])

# Reaction 794
reaction('C4H10 + CH3CO3 <=> PC4H9 + CH3CO3H', [1.700000e+13, 0.0, 20460.0])

# Reaction 795
reaction('C4H10 + CH3CO3 <=> SC4H9 + CH3CO3H', [1.120000e+13, 0.0, 17700.0])

# Reaction 796
reaction('C4H10 + O2CHO <=> PC4H9 + HO2CHO', [1.680000e+13, 0.0, 20440.0])

# Reaction 797
reaction('C4H10 + O2CHO <=> SC4H9 + HO2CHO', [1.120000e+13, 0.0, 17690.0])

# Reaction 798
reaction('CH3O2 + C4H10 <=> CH3O2H + PC4H9', [1.386000e+00, 3.97, 18280.0])

# Reaction 799
reaction('CH3O2 + C4H10 <=> CH3O2H + SC4H9', [2.037000e+01, 3.58, 14810.0])

# Reaction 800
reaction('C2H5O2 + C4H10 <=> C2H5O2H + PC4H9', [4.080000e+01, 3.59, 17160.0])

# Reaction 801
reaction('C2H5O2 + C4H10 <=> C2H5O2H + SC4H9', [1.264000e+02, 3.37, 13720.0])

# Reaction 802
reaction('NC3H7O2 + C4H10 <=> NC3H7O2H + PC4H9', [1.700000e+13, 0.0, 20460.0])

# Reaction 803
reaction('NC3H7O2 + C4H10 <=> NC3H7O2H + SC4H9', [1.120000e+13, 0.0, 17700.0])

# Reaction 804
reaction('IC3H7O2 + C4H10 <=> IC3H7O2H + PC4H9', [1.700000e+13, 0.0, 20460.0])

# Reaction 805
reaction('IC3H7O2 + C4H10 <=> IC3H7O2H + SC4H9', [1.120000e+13, 0.0, 17700.0])

# Reaction 806
reaction('PC4H9O2 + C3H8 <=> PC4H9O2H + NC3H7', [1.700000e+13, 0.0, 20460.0])

# Reaction 807
reaction('PC4H9O2 + C3H8 <=> PC4H9O2H + IC3H7', [2.000000e+12, 0.0, 17000.0])

# Reaction 808
reaction('PC4H9O2 + C4H10 <=> PC4H9O2H + PC4H9', [1.700000e+13, 0.0, 20460.0])

# Reaction 809
reaction('PC4H9O2 + C4H10 <=> PC4H9O2H + SC4H9', [1.120000e+13, 0.0, 17700.0])

# Reaction 810
reaction('SC4H9O2 + C3H8 <=> SC4H9O2H + NC3H7', [1.700000e+13, 0.0, 20460.0])

# Reaction 811
reaction('SC4H9O2 + C3H8 <=> SC4H9O2H + IC3H7', [2.000000e+12, 0.0, 17000.0])

# Reaction 812
reaction('SC4H9O2 + C4H10 <=> SC4H9O2H + PC4H9', [1.700000e+13, 0.0, 20460.0])

# Reaction 813
reaction('SC4H9O2 + C4H10 <=> SC4H9O2H + SC4H9', [1.120000e+13, 0.0, 17700.0])

# Reaction 814
reaction('PC4H9 <=> SC4H9', [3.560000e+10, 0.88, 37300.0],
        options='duplicate')

# Reaction 815
reaction('PC4H9 <=> SC4H9', [3.800000e+10, 0.67, 36600.0],
        options='duplicate')

# Reaction 816
reaction('C2H5 + C2H4 <=> PC4H9', [1.320000e+04, 2.48, 6130.0])

# Reaction 817
reaction('C3H6 + CH3 <=> SC4H9', [1.760000e+04, 2.48, 6130.0])

# Reaction 818
reaction('C4H8-1 + H <=> PC4H9', [2.500000e+11, 0.51, 2620.0])

# Reaction 819
reaction('C4H8-2 + H <=> SC4H9', [2.500000e+11, 0.51, 2620.0])

# Reaction 820
reaction('C4H8-1 + H <=> SC4H9', [4.240000e+11, 0.51, 1230.0])

# Reaction 821
reaction('PC4H9 + O2 <=> C4H8-1 + HO2', [8.370000e-01, 3.59, 11960.0])

# Reaction 822
reaction('SC4H9 + O2 <=> C4H8-1 + HO2', [5.350000e-01, 3.71, 9322.0])

# Reaction 823
reaction('SC4H9 + O2 <=> C4H8-2 + HO2', [1.070000e+00, 3.71, 9322.0])

# Reaction 824
reaction('C2H3 + C2H5 <=> C4H8-1', [9.000000e+12, 0.0, 0.0])

# Reaction 825
falloff_reaction('C3H5-A + CH3 (+ M) <=> C4H8-1 (+ M)',
                kf=[1.000000e+14, -0.32, -262.3],
                kf0=[3.910000e+60, -12.81, 6250.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.104, T3=1606.0, T1=60000.0, T2=6118.0))

# Reaction 826
reaction('H + C4H71-3 <=> C4H8-1', [5.000000e+13, 0.0, 0.0])

# Reaction 827
reaction('C4H8-1 + O2 <=> C4H71-3 + HO2', [2.000000e+13, 0.0, 37190.0])

# Reaction 828
reaction('C4H8-1 + O <=> C4H71-3 + OH', [1.750000e+11, 0.7, 5884.0])

# Reaction 829
reaction('C4H8-2 + O <=> C4H71-3 + OH', [2.190000e+11, 0.81, 7550.0])

# Reaction 830
reaction('C4H8-1 + H <=> C4H71-3 + H2', [1.730000e+05, 2.5, 2492.0])

# Reaction 831
reaction('C4H8-1 + H <=> C4H71-4 + H2', [6.651000e+05, 2.54, 6756.0])

# Reaction 832
reaction('C4H8-1 + OH <=> C4H71-1 + H2O', [9.000000e+05, 2.0, 2500.0])

# Reaction 833
reaction('C4H8-1 + OH <=> C4H71-2 + H2O', [2.220000e+06, 2.0, 1451.0])

# Reaction 834
reaction('C4H8-1 + OH <=> C4H71-3 + H2O', [3.120000e+06, 2.0, -298.0])

# Reaction 835
reaction('C4H8-1 + OH <=> C4H71-4 + H2O', [5.270000e+09, 0.97, 1586.0])

# Reaction 836
reaction('C4H8-1 + CH3 <=> C4H71-3 + CH4', [2.210000e+00, 3.5, 5675.0])

# Reaction 837
reaction('C4H8-1 + CH3 <=> C4H71-4 + CH4', [4.520000e-01, 3.65, 7154.0])

# Reaction 838
reaction('C4H8-1 + HO2 <=> C4H71-3 + H2O2', [2.700000e+04, 0.7, 5884.0])

# Reaction 839
reaction('C4H8-1 + HO2 <=> C4H71-4 + H2O2', [2.380000e+03, 2.55, 16490.0])

# Reaction 840
reaction('C4H8-1 + CH3O2 <=> C4H71-3 + CH3O2H', [2.700000e+04, 0.7, 5884.0])

# Reaction 841
reaction('C4H8-1 + CH3O2 <=> C4H71-4 + CH3O2H', [2.380000e+03, 2.55, 16490.0])

# Reaction 842
reaction('C4H8-1 + CH3O <=> C4H71-3 + CH3OH', [4.000000e+01, 2.9, 8609.0])

# Reaction 843
reaction('C4H8-1 + CH3O <=> C4H71-4 + CH3OH', [2.170000e+11, 0.0, 6458.0])

# Reaction 844
reaction('C4H8-1 + CH3CO3 <=> C4H71-3 + CH3CO3H', [1.000000e+11, 0.0, 8000.0])

# Reaction 845
reaction('C4H8-1 + C3H5-A <=> C4H71-3 + C3H6', [7.900000e+10, 0.0, 12400.0])

# Reaction 846
reaction('C4H71-3 + C4H71-3 <=> C4H8-1 + C4H6', [1.600000e+12, 0.0, 0.0])

# Reaction 847
reaction('C4H8-1 + C2H5O2 <=> C4H71-3 + C2H5O2H', [1.400000e+12, 0.0, 14900.0])

# Reaction 848
reaction('C4H8-1 + NC3H7O2 <=> C4H71-3 + NC3H7O2H', [1.400000e+12, 0.0, 14900.0])

# Reaction 849
reaction('C4H8-1 + IC3H7O2 <=> C4H71-3 + IC3H7O2H', [1.400000e+12, 0.0, 14900.0])

# Reaction 850
reaction('C4H8-1 + PC4H9O2 <=> C4H71-3 + PC4H9O2H', [1.400000e+12, 0.0, 14900.0])

# Reaction 851
reaction('C4H8-1 + SC4H9O2 <=> C4H71-3 + SC4H9O2H', [1.400000e+12, 0.0, 14900.0])

# Reaction 852
reaction('C4H8-1 + CH3O2 <=> C4H8O1-2 + CH3O', [1.000000e+12, 0.0, 14340.0])

# Reaction 853
reaction('H + C4H71-3 <=> C4H8-2', [5.000000e+13, 0.0, 0.0])

# Reaction 854
reaction('C4H8-2 + O2 <=> C4H71-3 + HO2', [4.000000e+13, 0.0, 39390.0])

# Reaction 855
reaction('C4H8-2 + H <=> C4H71-3 + H2', [4.440000e+04, 2.81, 4414.0])

# Reaction 856
reaction('C4H8-2 + OH <=> C4H71-3 + H2O', [5.100000e+08, 1.4, 1250.0])

# Reaction 857
reaction('C4H8-2 + CH3 <=> C4H71-3 + CH4', [7.140000e+00, 3.57, 7642.0])

# Reaction 858
reaction('C4H8-2 + HO2 <=> C4H71-3 + H2O2', [5.940000e+04, 2.57, 16140.0])

# Reaction 859
reaction('C4H8-2 + CH3O2 <=> C4H71-3 + CH3O2H', [5.940000e+04, 2.57, 16140.0])

# Reaction 860
reaction('C4H8-2 + CH3O <=> C4H71-3 + CH3OH', [1.800000e+01, 2.95, 11990.0])

# Reaction 861
reaction('C4H8-2 + C2H5O2 <=> C4H71-3 + C2H5O2H', [3.200000e+12, 0.0, 14900.0])

# Reaction 862
reaction('C4H8-2 + NC3H7O2 <=> C4H71-3 + NC3H7O2H', [3.200000e+12, 0.0, 14900.0])

# Reaction 863
reaction('C4H8-2 + IC3H7O2 <=> C4H71-3 + IC3H7O2H', [3.200000e+12, 0.0, 14900.0])

# Reaction 864
reaction('C4H8-2 + PC4H9O2 <=> C4H71-3 + PC4H9O2H', [3.200000e+12, 0.0, 14900.0])

# Reaction 865
reaction('C4H8-2 + SC4H9O2 <=> C4H71-3 + SC4H9O2H', [3.200000e+12, 0.0, 14900.0])

# Reaction 866
reaction('C4H8-1 + HO2 <=> C4H8O1-2 + OH', [1.000000e+12, 0.0, 14340.0])

# Reaction 867
reaction('C4H8-2 + HO2 <=> C4H8O2-3 + OH', [5.620000e+11, 0.0, 12310.0])

# Reaction 868
reaction('C4H8-2 + CH3O2 <=> C4H8O2-3 + CH3O', [5.620000e+11, 0.0, 12310.0])

# Reaction 869
reaction('C4H8-1 + OH <=> PC4H8OH', [4.750000e+12, 0.0, -782.0])

# Reaction 870
reaction('C4H8-2 + OH <=> SC4H8OH', [4.750000e+12, 0.0, -782.0])

# Reaction 871
reaction('PC4H8OH + O2 <=> C4H8OH-1O2', [2.000000e+12, 0.0, 0.0])

# Reaction 872
reaction('SC4H8OH + O2 <=> C4H8OH-2O2', [2.000000e+12, 0.0, 0.0])

# Reaction 873
reaction('C4H8OH-1O2 => C2H5CHO + CH2O + OH', [1.000000e+16, 0.0, 25000.0])

# Reaction 874
reaction('C4H8OH-2O2 => OH + CH3CHO + CH3CHO', [1.000000e+16, 0.0, 25000.0])

# Reaction 875
reaction('C2H2 + C2H5 <=> C4H71-1', [2.000000e+11, 0.0, 7800.0])

# Reaction 876
reaction('C3H4-A + CH3 <=> C4H71-2', [2.000000e+11, 0.0, 7800.0])

# Reaction 877
reaction('C2H4 + C2H3 <=> C4H71-4', [2.000000e+11, 0.0, 7800.0])

# Reaction 878
reaction('C3H4-P + CH3 <=> C4H72-2', [1.000000e+11, 0.0, 7800.0])

# Reaction 879
reaction('C4H6 + H <=> C4H71-3', [4.000000e+13, 0.0, 1300.0])

# Reaction 880
reaction('C4H71-3 + C2H5 <=> C4H8-1 + C2H4', [2.590000e+12, 0.0, -131.0])

# Reaction 881
reaction('C4H71-3 + CH3O <=> C4H8-1 + CH2O', [2.410000e+13, 0.0, 0.0])

# Reaction 882
reaction('C4H71-3 + O <=> C2H3CHO + CH3', [6.030000e+13, 0.0, 0.0])

# Reaction 883
reaction('C4H71-3 + HO2 <=> C4H7O + OH', [9.640000e+12, 0.0, 0.0])

# Reaction 884
reaction('C4H71-3 + CH3O2 <=> C4H7O + CH3O', [9.640000e+12, 0.0, 0.0])

# Reaction 885
reaction('C3H5-A + C4H71-3 <=> C3H6 + C4H6', [6.310000e+12, 0.0, 0.0])

# Reaction 886
reaction('C4H71-3 + O2 <=> C4H6 + HO2', [1.000000e+09, 0.0, 0.0])

# Reaction 887
reaction('H + C4H71-3 <=> C4H6 + H2', [3.160000e+13, 0.0, 0.0])

# Reaction 888
reaction('C2H5 + C4H71-3 <=> C4H6 + C2H6', [3.980000e+12, 0.0, 0.0])

# Reaction 889
reaction('C2H3 + C4H71-3 <=> C2H4 + C4H6', [3.980000e+12, 0.0, 0.0])

# Reaction 890
reaction('C4H71-3 + C2H5O2 <=> C4H7O + C2H5O', [3.800000e+12, 0.0, -1200.0])

# Reaction 891
reaction('IC3H7O2 + C4H71-3 <=> IC3H7O + C4H7O', [3.800000e+12, 0.0, -1200.0])

# Reaction 892
reaction('NC3H7O2 + C4H71-3 <=> NC3H7O + C4H7O', [3.800000e+12, 0.0, -1200.0])

# Reaction 893
reaction('C4H7O <=> CH3CHO + C2H3', [7.940000e+14, 0.0, 19000.0])

# Reaction 894
reaction('C4H7O <=> C2H3CHO + CH3', [7.940000e+14, 0.0, 19000.0])

# Reaction 895
reaction('C4H6 <=> C4H5-I + H', [5.700000e+36, -6.27, 112353.0])

# Reaction 896
reaction('C4H6 <=> C4H5-N + H', [5.300000e+44, -8.62, 123608.0])

# Reaction 897
reaction('C4H6 <=> C4H4 + H2', [2.500000e+15, 0.0, 94700.0])

# Reaction 898
reaction('C4H6 + H <=> C4H5-N + H2', [1.330000e+06, 2.53, 12240.0])

# Reaction 899
reaction('C4H6 + H <=> C4H5-I + H2', [6.650000e+05, 2.53, 9240.0])

# Reaction 900
pdep_arrhenius('C4H6 + H <=> C2H4 + C2H3',
            [(1.0, 'atm'), 1.460000e+30, -4.34, 21647.0],
            [(10.0, 'atm'), 5.450000e+30, -4.51, 21877.0])

# Reaction 901
reaction('C4H6 + H <=> C3H4-P + CH3', [2.000000e+12, 0.0, 7000.0])

# Reaction 902
reaction('C4H6 + H <=> C3H4-A + CH3', [2.000000e+12, 0.0, 7000.0])

# Reaction 903
reaction('C4H6 + O <=> C4H5-N + OH', [7.500000e+06, 1.9, 3740.0])

# Reaction 904
reaction('C4H6 + O <=> C4H5-I + OH', [7.500000e+06, 1.9, 3740.0])

# Reaction 905
reaction('C4H6 + O <=> C2H2 + C2H4O1-2', [1.000000e+08, 1.45, -860.0])

# Reaction 906
reaction('C4H6 + O <=> CH3CHCHCO + H', [5.000000e+07, 1.45, -860.0])

# Reaction 907
reaction('C4H6 + O <=> CH2CHCHCHO + H', [4.500000e+08, 1.45, -860.0])

# Reaction 908
reaction('C4H6 + OH <=> C4H5-N + H2O', [6.200000e+06, 2.0, 3430.0])

# Reaction 909
reaction('C4H6 + OH <=> C4H5-I + H2O', [3.100000e+06, 2.0, 430.0])

# Reaction 910
reaction('C4H6 + HO2 <=> C4H6O25 + OH', [1.200000e+12, 0.0, 14000.0])

# Reaction 911
reaction('C4H6 + HO2 <=> C2H3CHOCH2 + OH', [4.800000e+12, 0.0, 14000.0])

# Reaction 912
reaction('C4H6 + CH3 <=> C4H5-N + CH4', [2.000000e+14, 0.0, 22800.0])

# Reaction 913
reaction('C4H6 + CH3 <=> C4H5-I + CH4', [1.000000e+14, 0.0, 19800.0])

# Reaction 914
reaction('C4H6 + C2H3 <=> C4H5-N + C2H4', [5.000000e+13, 0.0, 22800.0])

# Reaction 915
reaction('C4H6 + C2H3 <=> C4H5-I + C2H4', [2.500000e+13, 0.0, 19800.0])

# Reaction 916
reaction('C4H6 + C3H3 <=> C4H5-N + C3H4-A', [1.000000e+13, 0.0, 22500.0])

# Reaction 917
reaction('C4H6 + C3H3 <=> C4H5-I + C3H4-A', [5.000000e+12, 0.0, 19500.0])

# Reaction 918
reaction('C4H6 + C3H5-A <=> C4H5-N + C3H6', [1.000000e+13, 0.0, 22500.0])

# Reaction 919
reaction('C4H6 + C3H5-A <=> C4H5-I + C3H6', [5.000000e+12, 0.0, 19500.0])

# Reaction 920
pdep_arrhenius('C4H71-4 <=> C4H6 + H',
            [(1.0, 'atm'), 2.480000e+53, -12.3, 52000.0],
            [(10.0, 'atm'), 1.850000e+48, -10.5, 51770.0])

# Reaction 921
pdep_arrhenius('C2H3 + C2H2 <=> C4H4 + H',
            [(0.0132, 'atm'), 7.200000e+13, -0.48, 6100.0],
            [(0.0263, 'atm'), 5.000000e+14, -0.71, 6700.0],
            [(0.12, 'atm'), 4.600000e+16, -1.25, 8400.0],
            [(1.0, 'atm'), 2.000000e+18, -1.68, 10600.0],
            [(10.0, 'atm'), 4.900000e+16, -1.13, 11800.0])

# Reaction 922
pdep_arrhenius('C2H3 + C2H2 <=> C4H5-N',
            [(0.0132, 'atm'), 1.100000e+31, -7.14, 5600.0],
            [(0.0263, 'atm'), 1.100000e+32, -7.33, 6200.0],
            [(0.12, 'atm'), 2.400000e+31, -6.95, 5600.0],
            [(1.0, 'atm'), 9.300000e+38, -8.76, 12000.0],
            [(10.0, 'atm'), 8.100000e+37, -8.09, 13400.0])

# Reaction 923
pdep_arrhenius('C2H3 + C2H2 <=> C4H5-I',
            [(0.0132, 'atm'), 5.000000e+34, -8.42, 7900.0],
            [(0.0263, 'atm'), 2.100000e+36, -8.78, 9100.0],
            [(0.12, 'atm'), 1.000000e+37, -8.77, 9800.0],
            [(1.0, 'atm'), 1.600000e+46, -10.98, 18600.0],
            [(10.0, 'atm'), 5.100000e+53, -12.64, 28800.0])

# Reaction 924
pdep_arrhenius('C2H3 + C2H3 <=> C4H6',
            [(0.0263, 'atm'), 7.000000e+57, -13.82, 17629.0],
            [(0.12, 'atm'), 1.500000e+52, -11.97, 16056.0],
            [(1.0, 'atm'), 1.500000e+42, -8.84, 12483.0])

# Reaction 925
pdep_arrhenius('C2H3 + C2H3 <=> C4H5-I + H',
            [(0.0263, 'atm'), 1.500000e+30, -4.95, 12958.0],
            [(0.12, 'atm'), 7.200000e+28, -4.49, 14273.0],
            [(1.0, 'atm'), 1.200000e+22, -2.44, 13654.0])

# Reaction 926
pdep_arrhenius('C2H3 + C2H3 <=> C4H5-N + H',
            [(0.0263, 'atm'), 1.100000e+24, -3.28, 12395.0],
            [(0.12, 'atm'), 4.600000e+24, -3.38, 14650.0],
            [(1.0, 'atm'), 2.400000e+20, -2.04, 15361.0])

# Reaction 927
pdep_arrhenius('C4H5-N <=> C4H5-I',
            [(0.0132, 'atm'), 2.400000e+60, -16.08, 47500.0],
            [(0.0263, 'atm'), 1.300000e+62, -16.38, 49600.0],
            [(0.12, 'atm'), 4.900000e+66, -17.26, 55400.0],
            [(1.0, 'atm'), 1.500000e+67, -16.89, 59100.0],
            [(10.0, 'atm'), 2.000000e+60, -14.46, 58600.0])

# Reaction 928
reaction('C4H5-N + H <=> C4H5-I + H', [3.100000e+26, -3.35, 17423.0])

# Reaction 929
reaction('C4H5-N + H <=> C4H4 + H2', [1.500000e+13, 0.0, 0.0])

# Reaction 930
reaction('C4H5-N + OH <=> C4H4 + H2O', [2.000000e+12, 0.0, 0.0])

# Reaction 931
reaction('C4H5-N + HCO <=> C4H6 + CO', [5.000000e+12, 0.0, 0.0])

# Reaction 932
reaction('C4H5-N + HO2 => C2H3 + CH2CO + OH', [6.600000e+12, 0.0, 0.0])

# Reaction 933
reaction('C4H5-N + H2O2 <=> C4H6 + HO2', [1.210000e+10, 0.0, -596.0])

# Reaction 934
reaction('C4H5-N + HO2 <=> C4H6 + O2', [6.000000e+11, 0.0, 0.0])

# Reaction 935
reaction('C4H5-N + O2 <=> CH2CHCHCHO + O', [3.000000e+11, 0.29, 11.0])

# Reaction 936
reaction('C4H5-N + O2 <=> HCO + C2H3CHO', [9.200000e+16, -1.39, 1010.0])

# Reaction 937
reaction('C4H5-I + H <=> C4H4 + H2', [3.000000e+13, 0.0, 0.0])

# Reaction 938
reaction('C4H5-I + H <=> C3H3 + CH3', [2.000000e+13, 0.0, 2000.0])

# Reaction 939
reaction('C4H5-I + OH <=> C4H4 + H2O', [4.000000e+12, 0.0, 0.0])

# Reaction 940
reaction('C4H5-I + HCO <=> C4H6 + CO', [5.000000e+12, 0.0, 0.0])

# Reaction 941
reaction('C4H5-I + HO2 <=> C4H6 + O2', [6.000000e+11, 0.0, 0.0])

# Reaction 942
reaction('C4H5-I + HO2 => C2H3 + CH2CO + OH', [6.600000e+12, 0.0, 0.0])

# Reaction 943
reaction('C4H5-I + H2O2 <=> C4H6 + HO2', [1.210000e+10, 0.0, -596.0])

# Reaction 944
reaction('C4H5-I + O2 <=> CH2CO + CH2CHO', [2.160000e+10, 0.0, 2500.0])

# Reaction 945
reaction('C4H5-2 <=> C4H5-I', [1.500000e+67, -16.89, 59100.0])

# Reaction 946
reaction('C4H5-2 + H <=> C4H5-I + H', [3.100000e+26, -3.35, 17423.0])

# Reaction 947
reaction('C4H5-2 + HO2 => OH + C2H2 + CH3CO', [8.000000e+11, 0.0, 0.0])

# Reaction 948
reaction('C4H5-2 + O2 <=> CH3CO + CH2CO', [2.160000e+10, 0.0, 2500.0])

# Reaction 949
reaction('C4H612 <=> C4H5-I + H', [4.200000e+15, 0.0, 92600.0])

# Reaction 950
reaction('C4H612 + H <=> C4H6 + H', [2.000000e+13, 0.0, 4000.0])

# Reaction 951
reaction('C4H612 + H <=> C4H5-I + H2', [1.700000e+05, 2.5, 2490.0])

# Reaction 952
reaction('C4H612 + H <=> C3H4-A + CH3', [2.000000e+13, 0.0, 2000.0])

# Reaction 953
reaction('C4H612 + H <=> C3H4-P + CH3', [2.000000e+13, 0.0, 2000.0])

# Reaction 954
reaction('C4H612 + CH3 <=> C4H5-I + CH4', [7.000000e+13, 0.0, 18500.0])

# Reaction 955
reaction('C4H612 + O <=> CH2CO + C2H4', [1.200000e+08, 1.65, 327.0])

# Reaction 956
reaction('C4H612 + O <=> C4H5-I + OH', [1.800000e+11, 0.7, 5880.0])

# Reaction 957
reaction('C4H612 + OH <=> C4H5-I + H2O', [3.100000e+06, 2.0, -298.0])

# Reaction 958
reaction('C4H612 <=> C4H6', [3.000000e+13, 0.0, 65000.0])

# Reaction 959
reaction('C4H6-2 <=> C4H6', [3.000000e+13, 0.0, 65000.0])

# Reaction 960
reaction('C4H6-2 <=> C4H612', [3.000000e+13, 0.0, 67000.0])

# Reaction 961
reaction('C4H6-2 + H <=> C4H612 + H', [2.000000e+13, 0.0, 4000.0])

# Reaction 962
reaction('C4H6-2 + H <=> C4H5-2 + H2', [3.400000e+05, 2.5, 2490.0])

# Reaction 963
reaction('C4H6-2 + H <=> CH3 + C3H4-P', [2.600000e+05, 2.5, 1000.0])

# Reaction 964
reaction('C4H6-2 <=> H + C4H5-2', [5.000000e+15, 0.0, 87300.0])

# Reaction 965
reaction('C4H6-2 + CH3 <=> C4H5-2 + CH4', [1.400000e+14, 0.0, 18500.0])

# Reaction 966
reaction('C2H3CHOCH2 <=> C4H6O23', [2.000000e+14, 0.0, 50600.0])

# Reaction 967
reaction('C4H6O23 <=> CH3CHCHCHO', [1.950000e+13, 0.0, 49400.0])

# Reaction 968
reaction('C4H6O23 <=> C2H4 + CH2CO', [5.750000e+15, 0.0, 69300.0])

# Reaction 969
reaction('C4H6O23 <=> C2H2 + C2H4O1-2', [1.000000e+16, 0.0, 75800.0])

# Reaction 970
reaction('C4H6O25 <=> C4H4O + H2', [5.300000e+12, 0.0, 48500.0])

# Reaction 971
reaction('C4H4O <=> CO + C3H4-P', [1.780000e+15, 0.0, 77500.0])

# Reaction 972
reaction('C4H4O <=> C2H2 + CH2CO', [5.010000e+14, 0.0, 77500.0])

# Reaction 973
reaction('CH3CHCHCHO <=> C3H6 + CO', [3.900000e+14, 0.0, 69000.0])

# Reaction 974
reaction('CH3CHCHCHO + H <=> CH2CHCHCHO + H2', [1.700000e+05, 2.5, 2490.0])

# Reaction 975
reaction('CH3CHCHCHO + H <=> CH3CHCHCO + H2', [1.000000e+05, 2.5, 2490.0])

# Reaction 976
reaction('CH3CHCHCHO + H <=> CH3 + C2H3CHO', [4.000000e+21, -2.39, 11180.0])

# Reaction 977
reaction('CH3CHCHCHO + H <=> C3H6 + HCO', [4.000000e+21, -2.39, 11180.0])

# Reaction 978
reaction('CH3CHCHCHO + CH3 <=> CH2CHCHCHO + CH4', [2.100000e+00, 3.5, 5675.0])

# Reaction 979
reaction('CH3CHCHCHO + CH3 <=> CH3CHCHCO + CH4', [1.100000e+00, 3.5, 5675.0])

# Reaction 980
reaction('CH3CHCHCHO + C2H3 <=> CH2CHCHCHO + C2H4', [2.210000e+00, 3.5, 4682.0])

# Reaction 981
reaction('CH3CHCHCHO + C2H3 <=> CH3CHCHCO + C2H4', [1.110000e+00, 3.5, 4682.0])

# Reaction 982
reaction('CH3CHCHCO <=> C3H5-S + CO', [1.000000e+14, 0.0, 30000.0])

# Reaction 983
reaction('CH3CHCHCO + H <=> CH3CHCHCHO', [1.000000e+14, 0.0, 0.0])

# Reaction 984
reaction('CH2CHCHCHO <=> C3H5-A + CO', [1.000000e+14, 0.0, 25000.0])

# Reaction 985
reaction('CH2CHCHCHO + H <=> CH3CHCHCHO', [1.000000e+14, 0.0, 0.0])

# Reaction 986
pdep_arrhenius('C4H4 + H <=> C4H5-N',
            [(0.0132, 'atm'), 1.200000e+51, -12.57, 12300.0],
            [(0.0263, 'atm'), 4.200000e+50, -12.34, 12500.0],
            [(0.12, 'atm'), 1.100000e+50, -11.94, 13400.0],
            [(1.0, 'atm'), 1.300000e+51, -11.92, 16500.0],
            [(10.0, 'atm'), 6.200000e+45, -10.08, 15800.0])

# Reaction 987
pdep_arrhenius('C4H4 + H <=> C4H5-I',
            [(0.0132, 'atm'), 6.100000e+53, -13.19, 14200.0],
            [(0.0263, 'atm'), 9.600000e+52, -12.85, 14300.0],
            [(0.12, 'atm'), 2.100000e+52, -12.44, 15500.0],
            [(1.0, 'atm'), 4.900000e+51, -11.92, 17700.0],
            [(10.0, 'atm'), 1.500000e+48, -10.58, 18800.0])

# Reaction 988
reaction('C4H4 + H <=> C4H3-N + H2', [6.650000e+05, 2.53, 12240.0])

# Reaction 989
reaction('C4H4 + H <=> C4H3-I + H2', [3.330000e+05, 2.53, 9240.0])

# Reaction 990
reaction('C4H4 + OH <=> C4H3-N + H2O', [3.100000e+07, 2.0, 3430.0])

# Reaction 991
reaction('C4H4 + OH <=> C4H3-I + H2O', [1.550000e+07, 2.0, 430.0])

# Reaction 992
reaction('C4H4 + O <=> C3H3 + HCO', [6.000000e+08, 1.45, -860.0])

# Reaction 993
reaction('C3H3 + HCCO <=> C4H4 + CO', [2.500000e+13, 0.0, 0.0])

# Reaction 994
reaction('C3H3 + CH <=> C4H3-I + H', [5.000000e+13, 0.0, 0.0])

# Reaction 995
reaction('C3H3 + CH2 <=> C4H4 + H', [5.000000e+13, 0.0, 0.0])

# Reaction 996
falloff_reaction('C3H3 + CH3 (+ M) <=> C4H612 (+ M)',
                kf=[1.500000e+12, 0.0, 0.0],
                kf0=[2.600000e+57, -11.94, 9770.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.175, T3=1341.0, T1=60000.0, T2=9770.0))

# Reaction 997
falloff_reaction('C2H2 + C2H (+ M) <=> C4H3-N (+ M)',
                kf=[8.300000e+10, 0.899, -363.0],
                kf0=[1.240000e+31, -4.718, 1871.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 CH4:2.0 C2H6:3.0 C2H4:2.5 C2H2:2.5',
                falloff=Troe(A=1.0, T3=100.0, T1=5613.0, T2=13390.0))

# Reaction 998
reaction('C4H3-N <=> C4H3-I', [4.100000e+43, -9.49, 53000.0])

# Reaction 999
reaction('C4H3-N + H <=> C4H3-I + H', [2.500000e+20, -1.67, 10800.0])

# Reaction 1000
reaction('C4H3-N + H <=> C2H2 + H2CC', [6.300000e+25, -3.34, 10014.0])

# Reaction 1001
reaction('C4H3-N + H <=> C4H4', [2.000000e+47, -10.26, 13070.0])

# Reaction 1002
reaction('C4H3-N + H <=> C4H2 + H2', [3.000000e+13, 0.0, 0.0])

# Reaction 1003
reaction('C4H3-N + OH <=> C4H2 + H2O', [2.000000e+12, 0.0, 0.0])

# Reaction 1004
falloff_reaction('C2H2 + C2H (+ M) <=> C4H3-I (+ M)',
                kf=[8.300000e+10, 0.899, -363.0],
                kf0=[1.240000e+31, -4.718, 1871.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 CH4:2.0 C2H6:3.0 C2H4:2.5 C2H2:2.5',
                falloff=Troe(A=1.0, T3=100.0, T1=5613.0, T2=13390.0))

# Reaction 1005
reaction('C4H3-I + H <=> C2H2 + H2CC', [2.800000e+23, -2.55, 10780.0])

# Reaction 1006
reaction('C4H3-I + H <=> C4H4', [3.400000e+43, -9.01, 12120.0])

# Reaction 1007
reaction('C4H3-I + H <=> C4H2 + H2', [6.000000e+13, 0.0, 0.0])

# Reaction 1008
reaction('C4H3-I + OH <=> C4H2 + H2O', [4.000000e+12, 0.0, 0.0])

# Reaction 1009
reaction('C4H3-I + O2 <=> HCCO + CH2CO', [7.860000e+16, -1.8, 0.0])

# Reaction 1010
reaction('C4H3-I + CH2 <=> C3H4-A + C2H', [2.000000e+13, 0.0, 0.0])

# Reaction 1011
reaction('C2H2 + C2H <=> C4H2 + H', [9.600000e+13, 0.0, 0.0])

# Reaction 1012
reaction('C4H2 + H <=> C4H3-N', [1.100000e+42, -8.72, 15300.0])

# Reaction 1013
reaction('C4H2 + H <=> C4H3-I', [1.100000e+30, -4.92, 10800.0])

# Reaction 1014
reaction('C4H2 + OH <=> H2C4O + H', [6.600000e+12, 0.0, -410.0])

# Reaction 1015
reaction('H2C4O + H <=> C2H2 + HCCO', [5.000000e+13, 0.0, 3000.0])

# Reaction 1016
reaction('H2C4O + OH <=> CH2CO + HCCO', [1.000000e+07, 2.0, 2000.0])

# Reaction 1017
falloff_reaction('H2CC + C2H2 (+ M) <=> C4H4 (+ M)',
                kf=[3.500000e+05, 2.055, -2400.0],
                kf0=[1.400000e+60, -12.599, 7417.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 CH4:2.0 C2H6:3.0 C2H4:3.0 C2H2:3.0',
                falloff=Troe(A=0.98, T3=56.0, T1=580.0, T2=4164.0))

# Reaction 1018
reaction('H2CC + C2H4 <=> C4H6', [1.000000e+12, 0.0, 0.0])

# Reaction 1019
reaction('C4H8O1-2 + OH => CH2O + C3H5-A + H2O', [5.000000e+12, 0.0, 0.0])

# Reaction 1020
reaction('C4H8O1-2 + H => CH2O + C3H5-A + H2', [5.000000e+12, 0.0, 0.0])

# Reaction 1021
reaction('C4H8O1-2 + O => CH2O + C3H5-A + OH', [5.000000e+12, 0.0, 0.0])

# Reaction 1022
reaction('C4H8O1-2 + HO2 => CH2O + C3H5-A + H2O2', [1.000000e+13, 0.0, 15000.0])

# Reaction 1023
reaction('C4H8O1-2 + CH3O2 => CH2O + C3H5-A + CH3O2H', [1.000000e+13, 0.0, 19000.0])

# Reaction 1024
reaction('C4H8O1-2 + CH3 => CH2O + C3H5-A + CH4', [2.000000e+11, 0.0, 10000.0])

# Reaction 1025
reaction('C4H8O1-3 + OH => CH2O + C3H5-A + H2O', [5.000000e+12, 0.0, 0.0])

# Reaction 1026
reaction('C4H8O1-3 + H => CH2O + C3H5-A + H2', [5.000000e+12, 0.0, 0.0])

# Reaction 1027
reaction('C4H8O1-3 + O => CH2O + C3H5-A + OH', [5.000000e+12, 0.0, 0.0])

# Reaction 1028
reaction('C4H8O1-3 + HO2 => CH2O + C3H5-A + H2O2', [1.000000e+13, 0.0, 15000.0])

# Reaction 1029
reaction('C4H8O1-3 + CH3O2 => CH2O + C3H5-A + CH3O2H', [1.000000e+13, 0.0, 19000.0])

# Reaction 1030
reaction('C4H8O1-3 + CH3 => CH2O + C3H5-A + CH4', [2.000000e+11, 0.0, 10000.0])

# Reaction 1031
reaction('C4H8O1-4 + OH => CH2O + C3H5-A + H2O', [5.000000e+12, 0.0, 0.0])

# Reaction 1032
reaction('C4H8O1-4 + H => CH2O + C3H5-A + H2', [5.000000e+12, 0.0, 0.0])

# Reaction 1033
reaction('C4H8O1-4 + O => CH2O + C3H5-A + OH', [5.000000e+12, 0.0, 0.0])

# Reaction 1034
reaction('C4H8O1-4 + HO2 => CH2O + C3H5-A + H2O2', [1.000000e+13, 0.0, 15000.0])

# Reaction 1035
reaction('C4H8O1-4 + CH3O2 => CH2O + C3H5-A + CH3O2H', [1.000000e+13, 0.0, 19000.0])

# Reaction 1036
reaction('C4H8O1-4 + CH3 => CH2O + C3H5-A + CH4', [2.000000e+11, 0.0, 10000.0])

# Reaction 1037
reaction('C4H8O2-3 + OH => CH2O + C3H5-A + H2O', [5.000000e+12, 0.0, 0.0])

# Reaction 1038
reaction('C4H8O2-3 + H => CH2O + C3H5-A + H2', [5.000000e+12, 0.0, 0.0])

# Reaction 1039
reaction('C4H8O2-3 + O => CH2O + C3H5-A + OH', [5.000000e+12, 0.0, 0.0])

# Reaction 1040
reaction('C4H8O2-3 + HO2 => CH2O + C3H5-A + H2O2', [1.000000e+13, 0.0, 15000.0])

# Reaction 1041
reaction('C4H8O2-3 + CH3O2 => CH2O + C3H5-A + CH3O2H', [1.000000e+13, 0.0, 19000.0])

# Reaction 1042
reaction('C4H8O2-3 + CH3 => CH2O + C3H5-A + CH4', [2.000000e+11, 0.0, 10000.0])

# Reaction 1043
reaction('PC4H9 + O2 <=> PC4H9O2', [4.520000e+12, 0.0, 0.0])

# Reaction 1044
reaction('SC4H9 + O2 <=> SC4H9O2', [7.540000e+12, 0.0, 0.0])

# Reaction 1045
reaction('SC4H9O2 + CH2O <=> SC4H9O2H + HCO', [5.600000e+12, 0.0, 13600.0])

# Reaction 1046
reaction('SC4H9O2 + CH3CHO <=> SC4H9O2H + CH3CO', [2.800000e+12, 0.0, 13600.0])

# Reaction 1047
reaction('SC4H9O2 + HO2 <=> SC4H9O2H + O2', [1.750000e+10, 0.0, -3275.0])

# Reaction 1048
reaction('IC3H7O2 + PC4H9 <=> IC3H7O + PC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1049
reaction('IC3H7O2 + SC4H9 <=> IC3H7O + SC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1050
reaction('NC3H7O2 + PC4H9 <=> NC3H7O + PC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1051
reaction('NC3H7O2 + SC4H9 <=> NC3H7O + SC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1052
reaction('SC4H9O2 + SC4H9O2 => O2 + SC4H9O + SC4H9O', [1.400000e+16, -1.61, 1860.0])

# Reaction 1053
reaction('SC4H9O2 + NC3H7O2 => SC4H9O + NC3H7O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1054
reaction('SC4H9O2 + IC3H7O2 => SC4H9O + IC3H7O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1055
reaction('SC4H9O2 + C2H5O2 => SC4H9O + C2H5O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1056
reaction('SC4H9O2 + CH3O2 => SC4H9O + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1057
reaction('SC4H9O2 + CH3CO3 => SC4H9O + CH3CO2 + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1058
reaction('H2 + PC4H9O2 <=> H + PC4H9O2H', [3.010000e+13, 0.0, 26030.0])

# Reaction 1059
reaction('H2 + SC4H9O2 <=> H + SC4H9O2H', [3.010000e+13, 0.0, 26030.0])

# Reaction 1060
reaction('C2H6 + PC4H9O2 <=> C2H5 + PC4H9O2H', [1.700000e+13, 0.0, 20460.0])

# Reaction 1061
reaction('C2H6 + SC4H9O2 <=> C2H5 + SC4H9O2H', [1.700000e+13, 0.0, 20460.0])

# Reaction 1062
reaction('PC4H9O2 + C2H5CHO <=> PC4H9O2H + C2H5CO', [2.000000e+11, 0.0, 9500.0])

# Reaction 1063
reaction('SC4H9O2 + C2H5CHO <=> SC4H9O2H + C2H5CO', [2.000000e+11, 0.0, 9500.0])

# Reaction 1064
reaction('SC4H9O2 + CH3 <=> SC4H9O + CH3O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1065
reaction('SC4H9O2 + C2H5 <=> SC4H9O + C2H5O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1066
reaction('SC4H9O2 + IC3H7 <=> SC4H9O + IC3H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1067
reaction('SC4H9O2 + NC3H7 <=> SC4H9O + NC3H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1068
reaction('SC4H9O2 + PC4H9 <=> SC4H9O + PC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1069
reaction('SC4H9O2 + SC4H9 <=> SC4H9O + SC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1070
reaction('SC4H9O2 + C3H5-A <=> SC4H9O + C3H5O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1071
reaction('PC4H9O2 + CH2O <=> PC4H9O2H + HCO', [5.600000e+12, 0.0, 13600.0])

# Reaction 1072
reaction('PC4H9O2 + CH3CHO <=> PC4H9O2H + CH3CO', [2.800000e+12, 0.0, 13600.0])

# Reaction 1073
reaction('PC4H9O2 + HO2 <=> PC4H9O2H + O2', [1.750000e+10, 0.0, -3275.0])

# Reaction 1074
reaction('C3H6 + PC4H9O2 <=> C3H5-A + PC4H9O2H', [3.240000e+11, 0.0, 14900.0])

# Reaction 1075
reaction('C3H6 + SC4H9O2 <=> C3H5-A + SC4H9O2H', [3.240000e+11, 0.0, 14900.0])

# Reaction 1076
reaction('C2H4 + PC4H9O2 <=> C2H3 + PC4H9O2H', [1.130000e+13, 0.0, 30430.0])

# Reaction 1077
reaction('C2H4 + SC4H9O2 <=> C2H3 + SC4H9O2H', [1.130000e+13, 0.0, 30430.0])

# Reaction 1078
reaction('CH3OH + PC4H9O2 <=> CH2OH + PC4H9O2H', [6.300000e+12, 0.0, 19360.0])

# Reaction 1079
reaction('CH3OH + SC4H9O2 <=> CH2OH + SC4H9O2H', [6.300000e+12, 0.0, 19360.0])

# Reaction 1080
reaction('C2H3CHO + PC4H9O2 <=> C2H3CO + PC4H9O2H', [2.800000e+12, 0.0, 13600.0])

# Reaction 1081
reaction('C2H3CHO + SC4H9O2 <=> C2H3CO + SC4H9O2H', [2.800000e+12, 0.0, 13600.0])

# Reaction 1082
reaction('CH4 + PC4H9O2 <=> CH3 + PC4H9O2H', [1.120000e+13, 0.0, 24640.0])

# Reaction 1083
reaction('CH4 + SC4H9O2 <=> CH3 + SC4H9O2H', [1.120000e+13, 0.0, 24640.0])

# Reaction 1084
reaction('C4H71-3 + PC4H9O2 <=> C4H7O + PC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1085
reaction('C4H71-3 + SC4H9O2 <=> C4H7O + SC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1086
reaction('H2O2 + PC4H9O2 <=> HO2 + PC4H9O2H', [2.400000e+12, 0.0, 10000.0])

# Reaction 1087
reaction('H2O2 + SC4H9O2 <=> HO2 + SC4H9O2H', [2.400000e+12, 0.0, 10000.0])

# Reaction 1088
reaction('PC4H9O2 + PC4H9O2 => O2 + PC4H9O + PC4H9O', [1.400000e+16, -1.61, 1860.0])

# Reaction 1089
reaction('PC4H9O2 + SC4H9O2 => PC4H9O + SC4H9O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1090
reaction('PC4H9O2 + NC3H7O2 => PC4H9O + NC3H7O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1091
reaction('PC4H9O2 + IC3H7O2 => PC4H9O + IC3H7O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1092
reaction('PC4H9O2 + C2H5O2 => PC4H9O + C2H5O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1093
reaction('PC4H9O2 + CH3O2 => PC4H9O + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1094
reaction('PC4H9O2 + CH3CO3 => PC4H9O + CH3CO2 + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1095
reaction('PC4H9O2 + CH3 <=> PC4H9O + CH3O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1096
reaction('PC4H9O2 + C2H5 <=> PC4H9O + C2H5O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1097
reaction('PC4H9O2 + IC3H7 <=> PC4H9O + IC3H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1098
reaction('PC4H9O2 + NC3H7 <=> PC4H9O + NC3H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1099
reaction('PC4H9O2 + PC4H9 <=> PC4H9O + PC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1100
reaction('PC4H9O2 + SC4H9 <=> PC4H9O + SC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1101
reaction('PC4H9O2 + C3H5-A <=> PC4H9O + C3H5O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1102
reaction('PC4H9 + HO2 <=> PC4H9O + OH', [7.000000e+12, 0.0, -1000.0])

# Reaction 1103
reaction('SC4H9 + HO2 <=> SC4H9O + OH', [7.000000e+12, 0.0, -1000.0])

# Reaction 1104
reaction('CH3O2 + PC4H9 <=> CH3O + PC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1105
reaction('CH3O2 + SC4H9 <=> CH3O + SC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1106
reaction('PC4H9O2H <=> PC4H9O + OH', [1.500000e+16, 0.0, 42500.0])

# Reaction 1107
reaction('SC4H9O + OH <=> SC4H9O2H', [1.000000e+15, -0.8, 0.0])

# Reaction 1108
reaction('NC3H7 + CH2O <=> PC4H9O', [5.000000e+10, 0.0, 3457.0])

# Reaction 1109
reaction('CH3 + C2H5CHO <=> SC4H9O', [5.000000e+10, 0.0, 9043.0])

# Reaction 1110
reaction('C2H5 + CH3CHO <=> SC4H9O', [3.330000e+10, 0.0, 6397.0])

# Reaction 1111
reaction('PC4H9O2 <=> C4H8OOH1-2', [2.000000e+11, 0.0, 31850.0])

# Reaction 1112
reaction('PC4H9O2 <=> C4H8OOH1-3', [2.500000e+10, 0.0, 20850.0])

# Reaction 1113
reaction('PC4H9O2 <=> C4H8OOH1-4', [4.688000e+09, 0.0, 22350.0])

# Reaction 1114
reaction('SC4H9O2 <=> C4H8OOH2-1', [3.000000e+11, 0.0, 34500.0])

# Reaction 1115
reaction('SC4H9O2 <=> C4H8OOH2-3', [2.000000e+11, 0.0, 31850.0])

# Reaction 1116
reaction('SC4H9O2 <=> C4H8OOH2-4', [3.750000e+10, 0.0, 24400.0])

# Reaction 1117
reaction('PC4H9O2 <=> C4H8-1 + HO2', [4.308000e+36, -7.5, 39510.0])

# Reaction 1118
reaction('SC4H9O2 <=> C4H8-1 + HO2', [5.980000e+42, -9.43, 41530.0])

# Reaction 1119
reaction('SC4H9O2 <=> C4H8-2 + HO2', [4.308000e+36, -7.5, 39510.0])

# Reaction 1120
reaction('C4H8-1 + HO2 <=> C4H8OOH1-2', [1.000000e+11, 0.0, 11000.0])

# Reaction 1121
reaction('C4H8-1 + HO2 <=> C4H8OOH2-1', [1.000000e+11, 0.0, 11750.0])

# Reaction 1122
reaction('C4H8-2 + HO2 <=> C4H8OOH2-3', [1.000000e+11, 0.0, 11750.0])

# Reaction 1123
reaction('C4H8OOH1-2 <=> C4H8O1-2 + OH', [1.380000e+12, 0.0, 15900.0])

# Reaction 1124
reaction('C4H8OOH1-3 <=> C4H8O1-3 + OH', [2.040000e+11, 0.0, 19500.0])

# Reaction 1125
reaction('C4H8OOH1-4 <=> C4H8O1-4 + OH', [5.130000e+10, 0.0, 14800.0])

# Reaction 1126
reaction('C4H8OOH2-1 <=> C4H8O1-2 + OH', [3.980000e+12, 0.0, 17000.0])

# Reaction 1127
reaction('C4H8OOH2-3 <=> C4H8O2-3 + OH', [1.380000e+12, 0.0, 15900.0])

# Reaction 1128
reaction('C4H8OOH2-4 <=> C4H8O1-3 + OH', [4.470000e+11, 0.0, 21900.0])

# Reaction 1129
reaction('C4H8OOH1-1 <=> NC3H7CHO + OH', [9.000000e+14, 0.0, 1500.0])

# Reaction 1130
reaction('C4H8OOH2-2 <=> C2H5COCH3 + OH', [9.000000e+14, 0.0, 1500.0])

# Reaction 1131
reaction('C4H8OOH1-3 => OH + CH2O + C3H6', [6.635000e+13, -0.16, 29900.0])

# Reaction 1132
reaction('C4H8OOH2-4 => OH + CH3CHO + C2H4', [1.945000e+18, -1.63, 26790.0])

# Reaction 1133
reaction('C4H8OOH1-2 + O2 <=> C4H8OOH1-2O2', [7.540000e+12, 0.0, 0.0])

# Reaction 1134
reaction('C4H8OOH1-3 + O2 <=> C4H8OOH1-3O2', [7.540000e+12, 0.0, 0.0])

# Reaction 1135
reaction('C4H8OOH1-4 + O2 <=> C4H8OOH1-4O2', [4.520000e+12, 0.0, 0.0])

# Reaction 1136
reaction('C4H8OOH2-1 + O2 <=> C4H8OOH2-1O2', [4.520000e+12, 0.0, 0.0])

# Reaction 1137
reaction('C4H8OOH2-3 + O2 <=> C4H8OOH2-3O2', [7.540000e+12, 0.0, 0.0])

# Reaction 1138
reaction('C4H8OOH2-4 + O2 <=> C4H8OOH2-4O2', [4.520000e+12, 0.0, 0.0])

# Reaction 1139
reaction('C4H8OOH1-2O2 <=> NC4KET12 + OH', [2.000000e+11, 0.0, 31500.0])

# Reaction 1140
reaction('C4H8OOH1-3O2 <=> NC4KET13 + OH', [2.500000e+10, 0.0, 21400.0])

# Reaction 1141
reaction('C4H8OOH1-4O2 <=> NC4KET14 + OH', [3.125000e+09, 0.0, 19350.0])

# Reaction 1142
reaction('C4H8OOH2-1O2 <=> NC4KET21 + OH', [1.000000e+11, 0.0, 28850.0])

# Reaction 1143
reaction('C4H8OOH2-3O2 <=> NC4KET23 + OH', [1.000000e+11, 0.0, 28850.0])

# Reaction 1144
reaction('C4H8OOH2-4O2 <=> NC4KET24 + OH', [1.250000e+10, 0.0, 17850.0])

# Reaction 1145
reaction('NC4KET12 => C2H5CHO + HCO + OH', [1.050000e+16, 0.0, 41600.0])

# Reaction 1146
reaction('NC4KET13 => CH3CHO + CH2CHO + OH', [1.050000e+16, 0.0, 41600.0])

# Reaction 1147
reaction('NC4KET14 => CH2CH2CHO + CH2O + OH', [1.500000e+16, 0.0, 42000.0])

# Reaction 1148
reaction('NC4KET21 => CH2O + C2H5CO + OH', [1.500000e+16, 0.0, 42000.0])

# Reaction 1149
reaction('NC4KET23 => CH3CHO + CH3CO + OH', [1.050000e+16, 0.0, 41600.0])

# Reaction 1150
reaction('NC4KET24 => CH2O + CH3COCH2 + OH', [1.500000e+16, 0.0, 42000.0])

# Reaction 1151
reaction('C2H5COCH3 + OH <=> CH2CH2COCH3 + H2O', [7.550000e+09, 0.97, 1586.0])

# Reaction 1152
reaction('C2H5COCH3 + OH <=> CH3CHCOCH3 + H2O', [8.450000e+11, 0.0, -228.0])

# Reaction 1153
reaction('C2H5COCH3 + OH <=> C2H5COCH2 + H2O', [5.100000e+11, 0.0, 1192.0])

# Reaction 1154
reaction('C2H5COCH3 + HO2 <=> CH2CH2COCH3 + H2O2', [2.380000e+04, 2.55, 16490.0])

# Reaction 1155
reaction('C2H5COCH3 + HO2 <=> CH3CHCOCH3 + H2O2', [2.000000e+11, 0.0, 8698.0])

# Reaction 1156
reaction('C2H5COCH3 + HO2 <=> C2H5COCH2 + H2O2', [2.380000e+04, 2.55, 14690.0])

# Reaction 1157
reaction('C2H5COCH3 + O <=> CH2CH2COCH3 + OH', [2.250000e+13, 0.0, 7700.0])

# Reaction 1158
reaction('C2H5COCH3 + O <=> CH3CHCOCH3 + OH', [3.070000e+13, 0.0, 3400.0])

# Reaction 1159
reaction('C2H5COCH3 + O <=> C2H5COCH2 + OH', [5.000000e+12, 0.0, 5962.0])

# Reaction 1160
reaction('C2H5COCH3 + H <=> CH2CH2COCH3 + H2', [9.160000e+06, 2.0, 7700.0])

# Reaction 1161
reaction('C2H5COCH3 + H <=> CH3CHCOCH3 + H2', [4.460000e+06, 2.0, 3200.0])

# Reaction 1162
reaction('C2H5COCH3 + H <=> C2H5COCH2 + H2', [9.300000e+12, 0.0, 6357.0])

# Reaction 1163
reaction('C2H5COCH3 + O2 <=> CH2CH2COCH3 + HO2', [2.050000e+13, 0.0, 51310.0])

# Reaction 1164
reaction('C2H5COCH3 + O2 <=> CH3CHCOCH3 + HO2', [1.550000e+13, 0.0, 41970.0])

# Reaction 1165
reaction('C2H5COCH3 + O2 <=> C2H5COCH2 + HO2', [2.050000e+13, 0.0, 49150.0])

# Reaction 1166
reaction('C2H5COCH3 + CH3 <=> CH2CH2COCH3 + CH4', [3.190000e+01, 3.17, 7172.0])

# Reaction 1167
reaction('C2H5COCH3 + CH3 <=> CH3CHCOCH3 + CH4', [1.740000e+00, 3.46, 3680.0])

# Reaction 1168
reaction('C2H5COCH3 + CH3 <=> C2H5COCH2 + CH4', [1.620000e+11, 0.0, 9630.0])

# Reaction 1169
reaction('C2H5COCH3 + CH3O <=> CH2CH2COCH3 + CH3OH', [2.170000e+11, 0.0, 6460.0])

# Reaction 1170
reaction('C2H5COCH3 + CH3O <=> CH3CHCOCH3 + CH3OH', [1.450000e+11, 0.0, 2771.0])

# Reaction 1171
reaction('C2H5COCH3 + CH3O <=> C2H5COCH2 + CH3OH', [2.170000e+11, 0.0, 4660.0])

# Reaction 1172
reaction('C2H5COCH3 + CH3O2 <=> CH2CH2COCH3 + CH3O2H', [3.010000e+12, 0.0, 19380.0])

# Reaction 1173
reaction('C2H5COCH3 + CH3O2 <=> CH3CHCOCH3 + CH3O2H', [2.000000e+12, 0.0, 15250.0])

# Reaction 1174
reaction('C2H5COCH3 + CH3O2 <=> C2H5COCH2 + CH3O2H', [3.010000e+12, 0.0, 17580.0])

# Reaction 1175
reaction('C2H5COCH3 + C2H3 <=> CH2CH2COCH3 + C2H4', [5.000000e+11, 0.0, 10400.0])

# Reaction 1176
reaction('C2H5COCH3 + C2H3 <=> CH3CHCOCH3 + C2H4', [3.000000e+11, 0.0, 3400.0])

# Reaction 1177
reaction('C2H5COCH3 + C2H3 <=> C2H5COCH2 + C2H4', [6.150000e+10, 0.0, 4278.0])

# Reaction 1178
reaction('C2H5COCH3 + C2H5 <=> CH2CH2COCH3 + C2H6', [5.000000e+10, 0.0, 13400.0])

# Reaction 1179
reaction('C2H5COCH3 + C2H5 <=> CH3CHCOCH3 + C2H6', [3.000000e+10, 0.0, 8600.0])

# Reaction 1180
reaction('C2H5COCH3 + C2H5 <=> C2H5COCH2 + C2H6', [5.000000e+10, 0.0, 11600.0])

# Reaction 1181
reaction('CH3CHCOCH3 + O2 <=> CH3CHOOCOCH3', [1.000000e+11, 0.0, 0.0])

# Reaction 1182
reaction('CH3CHOOCOCH3 <=> CH2CHOOHCOCH3', [8.900000e+12, 0.0, 29700.0])

# Reaction 1183
reaction('C2H3COCH3 + HO2 <=> CH2CHOOHCOCH3', [7.000000e+10, 0.0, 7800.0])

# Reaction 1184
reaction('CH2CH2CHO <=> C2H4 + HCO', [3.127000e+13, -0.52, 24590.0])

# Reaction 1185
reaction('CH2CH2COCH3 <=> C2H4 + CH3CO', [1.000000e+14, 0.0, 18000.0])

# Reaction 1186
reaction('C2H5COCH2 <=> CH2CO + C2H5', [1.000000e+14, 0.0, 35000.0])

# Reaction 1187
reaction('C2H3COCH3 + H <=> CH3CHCOCH3', [5.000000e+12, 0.0, 1200.0])

# Reaction 1188
reaction('CH3CHCO + CH3 <=> CH3CHCOCH3', [1.230000e+11, 0.0, 7800.0])

# Reaction 1189
reaction('NC3H7CHO + O2 <=> NC3H7CO + HO2', [1.200000e+05, 2.5, 37560.0])

# Reaction 1190
reaction('NC3H7CHO + OH <=> NC3H7CO + H2O', [2.000000e+06, 1.8, -1300.0])

# Reaction 1191
reaction('NC3H7CHO + H <=> NC3H7CO + H2', [4.140000e+09, 1.12, 2320.0])

# Reaction 1192
reaction('NC3H7CHO + O <=> NC3H7CO + OH', [5.940000e+12, 0.0, 1868.0])

# Reaction 1193
reaction('NC3H7CHO + HO2 <=> NC3H7CO + H2O2', [4.090000e+04, 2.5, 10200.0])

# Reaction 1194
reaction('NC3H7CHO + CH3 <=> NC3H7CO + CH4', [2.890000e-03, 4.62, 3210.0])

# Reaction 1195
reaction('NC3H7CHO + CH3O <=> NC3H7CO + CH3OH', [1.000000e+12, 0.0, 3300.0])

# Reaction 1196
reaction('NC3H7CHO + CH3O2 <=> NC3H7CO + CH3O2H', [4.090000e+04, 2.5, 10200.0])

# Reaction 1197
reaction('NC3H7CHO + OH <=> C3H6CHO-1 + H2O', [5.280000e+09, 0.97, 1586.0])

# Reaction 1198
reaction('NC3H7CHO + OH <=> C3H6CHO-2 + H2O', [4.680000e+07, 1.61, -35.0])

# Reaction 1199
reaction('NC3H7CHO + OH <=> C3H6CHO-3 + H2O', [5.520000e+02, 3.12, -1176.0])

# Reaction 1200
reaction('NC3H7CHO + HO2 <=> C3H6CHO-1 + H2O2', [2.379000e+04, 2.55, 16490.0])

# Reaction 1201
reaction('NC3H7CHO + HO2 <=> C3H6CHO-2 + H2O2', [9.640000e+03, 2.6, 13910.0])

# Reaction 1202
reaction('NC3H7CHO + HO2 <=> C3H6CHO-3 + H2O2', [3.440000e+12, 0.05, 17880.0])

# Reaction 1203
reaction('NC3H7CHO + CH3O2 <=> C3H6CHO-1 + CH3O2H', [2.379000e+04, 2.55, 16490.0])

# Reaction 1204
reaction('NC3H7CHO + CH3O2 <=> C3H6CHO-2 + CH3O2H', [9.640000e+03, 2.6, 13910.0])

# Reaction 1205
reaction('NC3H7CHO + CH3O2 <=> C3H6CHO-3 + CH3O2H', [3.440000e+12, 0.05, 17880.0])

# Reaction 1206
reaction('NC3H7CO <=> NC3H7 + CO', [1.000000e+11, 0.0, 9600.0])

# Reaction 1207
reaction('C3H6CHO-1 <=> C2H4 + CH2CHO', [7.400000e+11, 0.0, 21970.0])

# Reaction 1208
reaction('C2H5CHCO + H <=> C3H6CHO-3', [5.000000e+12, 0.0, 1200.0])

# Reaction 1209
reaction('C2H3CHO + CH3 <=> C3H6CHO-3', [1.230000e+11, 0.0, 7800.0])

# Reaction 1210
reaction('SC3H5CHO + H <=> C3H6CHO-2', [5.000000e+12, 0.0, 2900.0])

# Reaction 1211
reaction('C3H6 + HCO <=> C3H6CHO-2', [1.000000e+11, 0.0, 6000.0])

# Reaction 1212
reaction('C2H5CHCO + OH <=> NC3H7 + CO2', [3.730000e+12, 0.0, -1010.0])

# Reaction 1213
reaction('C2H5CHCO + H <=> NC3H7 + CO', [4.400000e+12, 0.0, 1459.0])

# Reaction 1214
reaction('C2H5CHCO + O <=> C3H6 + CO2', [3.200000e+12, 0.0, -437.0])

# Reaction 1215
reaction('SC3H5CHO + OH <=> SC3H5CO + H2O', [2.690000e+10, 0.76, -340.0])

# Reaction 1216
reaction('C3H5-S + CO <=> SC3H5CO', [5.000000e+12, 0.0, 8000.0])

# Reaction 1217
reaction('SC3H5CHO + HO2 <=> SC3H5CO + H2O2', [1.000000e+12, 0.0, 11920.0])

# Reaction 1218
reaction('SC3H5CHO + CH3 <=> SC3H5CO + CH4', [3.980000e+12, 0.0, 8700.0])

# Reaction 1219
reaction('SC3H5CHO + O <=> SC3H5CO + OH', [7.180000e+12, 0.0, 1389.0])

# Reaction 1220
reaction('SC3H5CHO + O2 <=> SC3H5CO + HO2', [4.000000e+13, 0.0, 37600.0])

# Reaction 1221
reaction('SC3H5CHO + H <=> SC3H5CO + H2', [2.600000e+12, 0.0, 2600.0])

# Reaction 1222
reaction('C2H3COCH3 + OH <=> CH3CHO + CH3CO', [1.000000e+11, 0.0, 0.0])

# Reaction 1223
reaction('C2H3COCH3 + OH => CH2CO + C2H3 + H2O', [5.100000e+11, 0.0, 1192.0])

# Reaction 1224
reaction('C2H3COCH3 + HO2 => CH2CHO + CH3CO + OH', [6.030000e+09, 0.0, 7949.0])

# Reaction 1225
reaction('C2H3COCH3 + HO2 => CH2CO + C2H3 + H2O2', [8.500000e+12, 0.0, 20460.0])

# Reaction 1226
reaction('C2H3COCH3 + CH3O2 => CH2CHO + CH3CO + CH3O', [3.970000e+11, 0.0, 17050.0])

# Reaction 1227
reaction('C2H3COCH3 + CH3O2 => CH2CO + C2H3 + CH3O2H', [3.010000e+12, 0.0, 17580.0])

# Reaction 1228
falloff_reaction('IC4H10 (+ M) <=> CH3 + IC3H7 (+ M)',
                kf=[2.520000e+31, -4.102, 91495.0],
                kf0=[2.410000e+19, 0.0, 52576.0],
                falloff=Troe(A=0.3662, T3=815.3, T1=60.79, T2=1e+20))

# Reaction 1229
reaction('IC4H10 <=> TC4H9 + H', [2.510000e+98, -23.81, 145300.0])

# Reaction 1230
reaction('IC4H10 <=> IC4H9 + H', [9.850000e+95, -23.11, 147600.0])

# Reaction 1231
reaction('IC4H10 + H <=> TC4H9 + H2', [6.020000e+05, 2.4, 2583.0])

# Reaction 1232
reaction('IC4H10 + H <=> IC4H9 + H2', [1.810000e+06, 2.54, 6756.0])

# Reaction 1233
reaction('IC4H10 + CH3 <=> TC4H9 + CH4', [9.040000e-01, 3.46, 4598.0])

# Reaction 1234
reaction('IC4H10 + CH3 <=> IC4H9 + CH4', [1.360000e+00, 3.65, 7154.0])

# Reaction 1235
reaction('IC4H10 + OH <=> TC4H9 + H2O', [2.925000e+04, 2.531, -1659.0])

# Reaction 1236
reaction('IC4H10 + OH <=> IC4H9 + H2O', [6.654000e+04, 2.665, -168.9])

# Reaction 1237
reaction('IC4H10 + C2H5 <=> IC4H9 + C2H6', [1.510000e+12, 0.0, 10400.0])

# Reaction 1238
reaction('IC4H10 + C2H5 <=> TC4H9 + C2H6', [1.000000e+11, 0.0, 7900.0])

# Reaction 1239
reaction('IC4H10 + HO2 <=> IC4H9 + H2O2', [6.120000e+01, 3.59, 17160.0])

# Reaction 1240
reaction('IC4H10 + HO2 <=> TC4H9 + H2O2', [4.332000e+02, 3.01, 12090.0])

# Reaction 1241
reaction('IC4H10 + O <=> TC4H9 + OH', [1.968000e+05, 2.402, 1150.0])

# Reaction 1242
reaction('IC4H10 + O <=> IC4H9 + OH', [4.046000e+07, 2.034, 5136.0])

# Reaction 1243
reaction('IC4H10 + CH3O <=> IC4H9 + CH3OH', [4.800000e+11, 0.0, 7000.0])

# Reaction 1244
reaction('IC4H10 + CH3O <=> TC4H9 + CH3OH', [1.900000e+10, 0.0, 2800.0])

# Reaction 1245
reaction('IC4H10 + O2 <=> IC4H9 + HO2', [9.000000e+13, 0.0, 52290.0])

# Reaction 1246
reaction('IC4H10 + O2 <=> TC4H9 + HO2', [1.000000e+13, 0.0, 48200.0])

# Reaction 1247
reaction('IC4H10 + CH3O2 <=> IC4H9 + CH3O2H', [2.079000e+00, 3.97, 18280.0])

# Reaction 1248
reaction('IC4H10 + C2H5O2 <=> IC4H9 + C2H5O2H', [2.550000e+13, 0.0, 20460.0])

# Reaction 1249
reaction('IC4H10 + CH3CO3 <=> IC4H9 + CH3CO3H', [2.550000e+13, 0.0, 20460.0])

# Reaction 1250
reaction('IC4H10 + NC3H7O2 <=> IC4H9 + NC3H7O2H', [2.550000e+13, 0.0, 20460.0])

# Reaction 1251
reaction('IC4H10 + IC3H7O2 <=> IC4H9 + IC3H7O2H', [2.550000e+13, 0.0, 20460.0])

# Reaction 1252
reaction('IC4H10 + IC4H9O2 <=> IC4H9 + IC4H9O2H', [2.550000e+13, 0.0, 20460.0])

# Reaction 1253
reaction('IC4H10 + TC4H9O2 <=> IC4H9 + TC4H9O2H', [2.550000e+13, 0.0, 20460.0])

# Reaction 1254
reaction('IC4H10 + O2CHO <=> IC4H9 + HO2CHO', [2.520000e+13, 0.0, 20440.0])

# Reaction 1255
reaction('IC4H10 + O2CHO <=> TC4H9 + HO2CHO', [2.800000e+12, 0.0, 16010.0])

# Reaction 1256
reaction('IC4H10 + SC4H9O2 <=> IC4H9 + SC4H9O2H', [2.250000e+13, 0.0, 20460.0])

# Reaction 1257
reaction('IC4H10 + SC4H9O2 <=> TC4H9 + SC4H9O2H', [2.800000e+12, 0.0, 16000.0])

# Reaction 1258
reaction('IC4H10 + PC4H9O2 <=> IC4H9 + PC4H9O2H', [2.250000e+13, 0.0, 20460.0])

# Reaction 1259
reaction('IC4H10 + PC4H9O2 <=> TC4H9 + PC4H9O2H', [2.800000e+12, 0.0, 16000.0])

# Reaction 1260
reaction('IC4H10 + CH3O2 <=> TC4H9 + CH3O2H', [1.366000e+02, 3.12, 13190.0])

# Reaction 1261
reaction('IC4H10 + C2H5O2 <=> TC4H9 + C2H5O2H', [2.800000e+12, 0.0, 16000.0])

# Reaction 1262
reaction('IC4H10 + CH3CO3 <=> TC4H9 + CH3CO3H', [2.800000e+12, 0.0, 16000.0])

# Reaction 1263
reaction('IC4H10 + NC3H7O2 <=> TC4H9 + NC3H7O2H', [2.800000e+12, 0.0, 16000.0])

# Reaction 1264
reaction('IC4H10 + IC3H7O2 <=> TC4H9 + IC3H7O2H', [2.800000e+12, 0.0, 16000.0])

# Reaction 1265
reaction('IC4H10 + IC4H9O2 <=> TC4H9 + IC4H9O2H', [2.800000e+12, 0.0, 16000.0])

# Reaction 1266
reaction('IC4H10 + TC4H9O2 <=> TC4H9 + TC4H9O2H', [2.800000e+12, 0.0, 16000.0])

# Reaction 1267
reaction('IC4H10 + IC4H9 <=> TC4H9 + IC4H10', [2.500000e+10, 0.0, 7900.0])

# Reaction 1268
reaction('IC4H9 + HO2 <=> IC4H9O + OH', [7.000000e+12, 0.0, -1000.0])

# Reaction 1269
reaction('TC4H9 + HO2 <=> TC4H9O + OH', [7.000000e+12, 0.0, -1000.0])

# Reaction 1270
reaction('CH3O2 + IC4H9 <=> CH3O + IC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1271
reaction('CH3O2 + TC4H9 <=> CH3O + TC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1272
reaction('IC4H9 <=> TC4H9', [3.560000e+10, 0.88, 34600.0])

# Reaction 1273
reaction('IC4H8 + H <=> IC4H9', [6.250000e+11, 0.51, 2620.0])

# Reaction 1274
reaction('C3H6 + CH3 <=> IC4H9', [1.890000e+03, 2.67, 6850.0])

# Reaction 1275
reaction('H + IC4H8 <=> TC4H9', [1.060000e+12, 0.51, 1230.0])

# Reaction 1276
reaction('TC4H9 + O2 <=> IC4H8 + HO2', [8.370000e-01, 3.59, 11960.0])

# Reaction 1277
reaction('IC4H9 + O2 <=> IC4H8 + HO2', [1.070000e+00, 3.71, 9322.0])

# Reaction 1278
reaction('NC3H7O2 + IC4H9 <=> NC3H7O + IC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1279
reaction('NC3H7O2 + TC4H9 <=> NC3H7O + TC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1280
reaction('NC3H7O2 + IC4H7 <=> NC3H7O + IC4H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1281
reaction('SC4H9O2 + IC4H9 <=> SC4H9O + IC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1282
reaction('SC4H9O2 + TC4H9 <=> SC4H9O + TC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1283
reaction('PC4H9O2 + IC4H9 <=> PC4H9O + IC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1284
reaction('PC4H9O2 + TC4H9 <=> PC4H9O + TC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1285
reaction('PC4H9O2 + IC4H7 <=> PC4H9O + IC4H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1286
reaction('SC4H9O2 + IC4H7 <=> SC4H9O + IC4H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1287
reaction('IC4H9 + O2 <=> IC4H9O2', [2.260000e+12, 0.0, 0.0])

# Reaction 1288
reaction('TC4H9 + O2 <=> TC4H9O2', [1.410000e+13, 0.0, 0.0])

# Reaction 1289
reaction('IC4H9O2 + C4H10 <=> IC4H9O2H + SC4H9', [1.120000e+13, 0.0, 17700.0])

# Reaction 1290
reaction('TC4H9O2 + C4H10 <=> TC4H9O2H + SC4H9', [1.120000e+13, 0.0, 17700.0])

# Reaction 1291
reaction('IC4H9O2 + C4H10 <=> IC4H9O2H + PC4H9', [1.700000e+13, 0.0, 20460.0])

# Reaction 1292
reaction('TC4H9O2 + C4H10 <=> TC4H9O2H + PC4H9', [1.700000e+13, 0.0, 20460.0])

# Reaction 1293
reaction('IC3H7O2 + IC4H9 <=> IC3H7O + IC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1294
reaction('IC3H7O2 + TC4H9 <=> IC3H7O + TC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1295
reaction('IC3H7O2 + IC4H7 <=> IC3H7O + IC4H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1296
reaction('IC4H9O2 + C3H6 <=> IC4H9O2H + C3H5-A', [3.240000e+11, 0.0, 14900.0])

# Reaction 1297
reaction('TC4H9O2 + C3H6 <=> TC4H9O2H + C3H5-A', [3.240000e+11, 0.0, 14900.0])

# Reaction 1298
reaction('IC4H9O2 + IC4H8 <=> IC4H9O2H + IC4H7', [1.400000e+12, 0.0, 14900.0])

# Reaction 1299
reaction('TC4H9O2 + IC4H8 <=> TC4H9O2H + IC4H7', [1.400000e+12, 0.0, 14900.0])

# Reaction 1300
reaction('PC4H9O2 + IC4H8 <=> PC4H9O2H + IC4H7', [1.400000e+12, 0.0, 14900.0])

# Reaction 1301
reaction('SC4H9O2 + IC4H8 <=> SC4H9O2H + IC4H7', [1.400000e+12, 0.0, 14900.0])

# Reaction 1302
reaction('IC3H7O2 + IC4H8 <=> IC3H7O2H + IC4H7', [1.400000e+12, 0.0, 14900.0])

# Reaction 1303
reaction('NC3H7O2 + IC4H8 <=> NC3H7O2H + IC4H7', [1.400000e+12, 0.0, 14900.0])

# Reaction 1304
reaction('IC4H9O2 + C4H8-1 <=> IC4H9O2H + C4H71-3', [1.400000e+12, 0.0, 14900.0])

# Reaction 1305
reaction('TC4H9O2 + C4H8-1 <=> TC4H9O2H + C4H71-3', [1.400000e+12, 0.0, 14900.0])

# Reaction 1306
reaction('IC4H9O2 + C4H8-2 <=> IC4H9O2H + C4H71-3', [1.400000e+12, 0.0, 14900.0])

# Reaction 1307
reaction('TC4H9O2 + C4H8-2 <=> TC4H9O2H + C4H71-3', [1.400000e+12, 0.0, 14900.0])

# Reaction 1308
reaction('CC4H8O + OH => CH2O + C3H5-A + H2O', [5.000000e+12, 0.0, 0.0])

# Reaction 1309
reaction('CC4H8O + H => CH2O + C3H5-A + H2', [3.510000e+07, 2.0, 5000.0])

# Reaction 1310
reaction('CC4H8O + O => CH2O + C3H5-A + OH', [1.124000e+14, 0.0, 5200.0])

# Reaction 1311
reaction('CC4H8O + HO2 => CH2O + C3H5-A + H2O2', [1.000000e+13, 0.0, 15000.0])

# Reaction 1312
reaction('CC4H8O + CH3O2 => CH2O + C3H5-A + CH3O2H', [1.000000e+13, 0.0, 19000.0])

# Reaction 1313
reaction('CC4H8O + CH3 => CH2O + C3H5-A + CH4', [2.000000e+11, 0.0, 10000.0])

# Reaction 1314
reaction('C2H4 + TC4H9O2 <=> C2H3 + TC4H9O2H', [8.590000e+00, 3.754, 27132.0])

# Reaction 1315
reaction('TC4H9O2 + CH4 <=> TC4H9O2H + CH3', [1.130000e+13, 0.0, 20460.0])

# Reaction 1316
reaction('H2 + TC4H9O2 <=> H + TC4H9O2H', [3.010000e+13, 0.0, 26030.0])

# Reaction 1317
reaction('TC4H9O2 + C2H6 <=> TC4H9O2H + C2H5', [1.700000e+13, 0.0, 20460.0])

# Reaction 1318
reaction('TC4H9O2 + C3H8 <=> TC4H9O2H + IC3H7', [2.000000e+12, 0.0, 17000.0])

# Reaction 1319
reaction('TC4H9O2 + C3H8 <=> TC4H9O2H + NC3H7', [1.700000e+13, 0.0, 20460.0])

# Reaction 1320
reaction('TC4H9O2 + CH3OH <=> TC4H9O2H + CH2OH', [6.300000e+12, 0.0, 19360.0])

# Reaction 1321
reaction('TC4H9O2 + C2H5OH <=> TC4H9O2H + PC2H4OH', [6.300000e+12, 0.0, 19360.0])

# Reaction 1322
reaction('TC4H9O2 + C2H5OH <=> TC4H9O2H + SC2H4OH', [4.200000e+12, 0.0, 15000.0])

# Reaction 1323
reaction('IC4H9O2 + CH3CHO <=> IC4H9O2H + CH3CO', [2.800000e+12, 0.0, 13600.0])

# Reaction 1324
reaction('TC4H9O2 + CH3CHO <=> TC4H9O2H + CH3CO', [2.800000e+12, 0.0, 13600.0])

# Reaction 1325
reaction('IC4H9O2 + C2H3CHO <=> IC4H9O2H + C2H3CO', [2.800000e+12, 0.0, 13600.0])

# Reaction 1326
reaction('TC4H9O2 + C2H3CHO <=> TC4H9O2H + C2H3CO', [2.800000e+12, 0.0, 13600.0])

# Reaction 1327
reaction('IC4H9O2 + C2H5CHO <=> IC4H9O2H + C2H5CO', [2.800000e+12, 0.0, 13600.0])

# Reaction 1328
reaction('TC4H9O2 + C2H5CHO <=> TC4H9O2H + C2H5CO', [2.800000e+12, 0.0, 13600.0])

# Reaction 1329
reaction('IC4H9O2 + HO2 <=> IC4H9O2H + O2', [1.750000e+10, 0.0, -3275.0])

# Reaction 1330
reaction('TC4H9O2 + HO2 <=> TC4H9O2H + O2', [1.750000e+10, 0.0, -3275.0])

# Reaction 1331
reaction('IC4H9O2 + H2O2 <=> IC4H9O2H + HO2', [2.400000e+12, 0.0, 10000.0])

# Reaction 1332
reaction('TC4H9O2 + H2O2 <=> TC4H9O2H + HO2', [2.400000e+12, 0.0, 10000.0])

# Reaction 1333
reaction('IC4H9O2 + CH2O <=> IC4H9O2H + HCO', [1.300000e+11, 0.0, 9000.0])

# Reaction 1334
reaction('TC4H9O2 + CH2O <=> TC4H9O2H + HCO', [1.300000e+11, 0.0, 9000.0])

# Reaction 1335
reaction('IC4H9O2 + CH3O2 => IC4H9O + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1336
reaction('TC4H9O2 + CH3O2 => TC4H9O + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1337
reaction('IC4H9O2 + C2H5O2 => IC4H9O + C2H5O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1338
reaction('TC4H9O2 + C2H5O2 => TC4H9O + C2H5O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1339
reaction('IC4H9O2 + CH3CO3 => IC4H9O + CH3CO2 + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1340
reaction('TC4H9O2 + CH3CO3 => TC4H9O + CH3CO2 + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1341
reaction('IC4H9O2 + IC4H9O2 => O2 + IC4H9O + IC4H9O', [1.400000e+16, -1.61, 1860.0])

# Reaction 1342
reaction('IC4H9O2 + TC4H9O2 => IC4H9O + TC4H9O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1343
reaction('TC4H9O2 + TC4H9O2 => O2 + TC4H9O + TC4H9O', [1.400000e+16, -1.61, 1860.0])

# Reaction 1344
reaction('IC4H9O2 + PC4H9O2 => IC4H9O + PC4H9O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1345
reaction('TC4H9O2 + PC4H9O2 => TC4H9O + PC4H9O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1346
reaction('IC4H9O2 + SC4H9O2 => IC4H9O + SC4H9O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1347
reaction('TC4H9O2 + SC4H9O2 => TC4H9O + SC4H9O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1348
reaction('IC4H9O2 + NC3H7O2 => IC4H9O + NC3H7O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1349
reaction('TC4H9O2 + NC3H7O2 => TC4H9O + NC3H7O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1350
reaction('IC4H9O2 + IC3H7O2 => IC4H9O + IC3H7O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1351
reaction('TC4H9O2 + IC3H7O2 => TC4H9O + IC3H7O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1352
reaction('IC4H9O2 + HO2 => IC4H9O + OH + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1353
reaction('TC4H9O2 + HO2 => TC4H9O + OH + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1354
reaction('IC4H9O2 + CH3 <=> IC4H9O + CH3O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1355
reaction('IC4H9O2 + C2H5 <=> IC4H9O + C2H5O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1356
reaction('IC4H9O2 + IC3H7 <=> IC4H9O + IC3H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1357
reaction('IC4H9O2 + NC3H7 <=> IC4H9O + NC3H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1358
reaction('IC4H9O2 + PC4H9 <=> IC4H9O + PC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1359
reaction('IC4H9O2 + SC4H9 <=> IC4H9O + SC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1360
reaction('IC4H9O2 + IC4H9 <=> IC4H9O + IC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1361
reaction('IC4H9O2 + TC4H9 <=> IC4H9O + TC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1362
reaction('IC4H9O2 + C3H5-A <=> IC4H9O + C3H5O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1363
reaction('IC4H9O2 + C4H71-3 <=> IC4H9O + C4H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1364
reaction('IC4H9O2 + IC4H7 <=> IC4H9O + IC4H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1365
reaction('TC4H9O2 + CH3 <=> TC4H9O + CH3O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1366
reaction('TC4H9O2 + C2H5 <=> TC4H9O + C2H5O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1367
reaction('TC4H9O2 + IC3H7 <=> TC4H9O + IC3H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1368
reaction('TC4H9O2 + NC3H7 <=> TC4H9O + NC3H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1369
reaction('TC4H9O2 + PC4H9 <=> TC4H9O + PC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1370
reaction('TC4H9O2 + SC4H9 <=> TC4H9O + SC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1371
reaction('TC4H9O2 + IC4H9 <=> TC4H9O + IC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1372
reaction('TC4H9O2 + TC4H9 <=> TC4H9O + TC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1373
reaction('TC4H9O2 + C3H5-A <=> TC4H9O + C3H5O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1374
reaction('TC4H9O2 + C4H71-3 <=> TC4H9O + C4H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1375
reaction('TC4H9O2 + IC4H7 <=> TC4H9O + IC4H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1376
reaction('IC4H9O2 + C2H4 <=> IC4H9O2H + C2H3', [2.000000e+11, 0.0, 6000.0])

# Reaction 1377
reaction('IC4H9O2 + CH4 <=> IC4H9O2H + CH3', [1.130000e+13, 0.0, 20460.0])

# Reaction 1378
reaction('H2 + IC4H9O2 <=> H + IC4H9O2H', [3.010000e+13, 0.0, 26030.0])

# Reaction 1379
reaction('IC4H9O2 + C2H6 <=> IC4H9O2H + C2H5', [1.700000e+13, 0.0, 20460.0])

# Reaction 1380
reaction('IC4H9O2 + C3H8 <=> IC4H9O2H + IC3H7', [2.000000e+12, 0.0, 17000.0])

# Reaction 1381
reaction('IC4H9O2 + C3H8 <=> IC4H9O2H + NC3H7', [1.700000e+13, 0.0, 20460.0])

# Reaction 1382
reaction('IC4H9O2 + CH3OH <=> IC4H9O2H + CH2OH', [6.300000e+12, 0.0, 19360.0])

# Reaction 1383
reaction('IC4H9O2 + C2H5OH <=> IC4H9O2H + PC2H4OH', [6.300000e+12, 0.0, 19360.0])

# Reaction 1384
reaction('IC4H9O2 + C2H5OH <=> IC4H9O2H + SC2H4OH', [4.200000e+12, 0.0, 15000.0])

# Reaction 1385
reaction('IC4H9O2H <=> IC4H9O + OH', [1.500000e+16, 0.0, 42500.0])

# Reaction 1386
reaction('TC4H9O2H <=> TC4H9O + OH', [5.950000e+15, 0.0, 42540.0])

# Reaction 1387
reaction('IC4H9O + HO2 <=> IC3H7CHO + H2O2', [1.000000e+12, 0.0, 0.0])

# Reaction 1388
reaction('IC4H9O + OH <=> IC3H7CHO + H2O', [1.810000e+13, 0.0, 0.0])

# Reaction 1389
reaction('IC4H9O + CH3 <=> IC3H7CHO + CH4', [2.400000e+13, 0.0, 0.0])

# Reaction 1390
reaction('IC4H9O + O <=> IC3H7CHO + OH', [6.000000e+12, 0.0, 0.0])

# Reaction 1391
reaction('IC4H9O + H <=> IC3H7CHO + H2', [1.990000e+13, 0.0, 0.0])

# Reaction 1392
reaction('IC3H7CHO + H <=> IC4H9O', [1.000000e+12, 0.0, 5860.0])

# Reaction 1393
reaction('CH2O + IC3H7 <=> IC4H9O', [5.000000e+10, 0.0, 2330.0])

# Reaction 1394
reaction('CH3COCH3 + CH3 <=> TC4H9O', [1.500000e+11, 0.0, 11900.0])

# Reaction 1395
reaction('IC4H9O + O2 <=> IC3H7CHO + HO2', [1.930000e+11, 0.0, 1660.0])

# Reaction 1396
reaction('TC4H9O + O2 <=> IC4H8O + HO2', [8.100000e+11, 0.0, 4700.0])

# Reaction 1397
reaction('IC4H8O <=> IC3H7CHO', [4.180000e+13, 0.0, 52720.0])

# Reaction 1398
reaction('IC4H8O + OH <=> IC3H6CHO + H2O', [1.250000e+12, 0.0, 0.0])

# Reaction 1399
reaction('IC4H8O + H <=> IC3H6CHO + H2', [1.250000e+12, 0.0, 0.0])

# Reaction 1400
reaction('IC4H8O + HO2 <=> IC3H6CHO + H2O2', [2.500000e+12, 0.0, 15000.0])

# Reaction 1401
reaction('IC4H8O + CH3O2 <=> IC3H6CHO + CH3O2H', [2.500000e+12, 0.0, 19000.0])

# Reaction 1402
reaction('IC4H8O + CH3 <=> IC3H6CHO + CH4', [5.000000e+10, 0.0, 10000.0])

# Reaction 1403
reaction('IC4H8O + O <=> IC3H6CHO + OH', [1.250000e+12, 0.0, 0.0])

# Reaction 1404
reaction('TC3H6CHO + H <=> IC3H7CHO', [2.000000e+14, 0.0, 0.0])

# Reaction 1405
reaction('IC3H7 + HCO <=> IC3H7CHO', [1.810000e+13, 0.0, 0.0])

# Reaction 1406
reaction('IC3H7CHO + HO2 <=> IC3H7CO + H2O2', [3.000000e+12, 0.0, 11920.0])

# Reaction 1407
reaction('IC3H7CHO + HO2 <=> TC3H6CHO + H2O2', [8.000000e+10, 0.0, 11920.0])

# Reaction 1408
reaction('IC3H7CHO + CH3 <=> IC3H7CO + CH4', [3.980000e+12, 0.0, 8700.0])

# Reaction 1409
reaction('IC3H7CHO + O <=> IC3H7CO + OH', [7.180000e+12, 0.0, 1389.0])

# Reaction 1410
reaction('IC3H7CHO + O2 <=> IC3H7CO + HO2', [4.000000e+13, 0.0, 37600.0])

# Reaction 1411
reaction('IC3H7CHO + OH <=> IC3H7CO + H2O', [2.690000e+10, 0.76, -340.0])

# Reaction 1412
reaction('IC3H7CHO + OH <=> TC3H6CHO + H2O', [1.684000e+12, 0.0, -781.0])

# Reaction 1413
reaction('IC3H7CHO + H <=> IC3H7CO + H2', [2.600000e+12, 0.0, 2600.0])

# Reaction 1414
reaction('IC3H7CHO + OH <=> IC3H6CHO + H2O', [3.120000e+06, 2.0, -298.0])

# Reaction 1415
reaction('IC3H7CHO + HO2 <=> IC3H6CHO + H2O2', [2.740000e+04, 2.55, 15500.0])

# Reaction 1416
reaction('IC3H7CHO + CH3O2 <=> IC3H6CHO + CH3O2H', [4.760000e+04, 2.55, 16490.0])

# Reaction 1417
reaction('IC3H7 + CO <=> IC3H7CO', [1.500000e+11, 0.0, 4810.0])

# Reaction 1418
reaction('C3H6 + HCO <=> IC3H6CHO', [1.000000e+11, 0.0, 7800.0])

# Reaction 1419
reaction('C2H3CHO + CH3 <=> IC3H6CHO', [1.000000e+11, 0.0, 7800.0])

# Reaction 1420
reaction('IC4H8 + OH <=> IC4H8OH', [9.930000e+11, 0.0, -960.0])

# Reaction 1421
reaction('IC4H8OH + O2 <=> IO2C4H8OH', [1.200000e+11, 0.0, -1100.0])

# Reaction 1422
reaction('IO2C4H8OH => CH3COCH3 + CH2O + OH', [1.250000e+10, 0.0, 18900.0])

# Reaction 1423
reaction('IC4H9O2 <=> IC4H8O2H-I', [7.500000e+10, 0.0, 24400.0])

# Reaction 1424
reaction('TC4H9O2 <=> TC4H8O2H-I', [9.000000e+11, 0.0, 34500.0])

# Reaction 1425
reaction('IC4H9O2 <=> IC4H8O2H-T', [1.000000e+11, 0.0, 29200.0])

# Reaction 1426
reaction('IC4H9O2 <=> IC4H8 + HO2', [2.265000e+35, -7.22, 39490.0])

# Reaction 1427
reaction('TC4H9O2 <=> IC4H8 + HO2', [7.612000e+42, -9.41, 41490.0])

# Reaction 1428
reaction('IC4H8O2H-I + O2 <=> IC4H8OOH-IO2', [2.260000e+12, 0.0, 0.0])

# Reaction 1429
reaction('TC4H8O2H-I + O2 <=> TC4H8OOH-IO2', [2.260000e+12, 0.0, 0.0])

# Reaction 1430
reaction('IC4H8O2H-T + O2 <=> IC4H8OOH-TO2', [1.410000e+13, 0.0, 0.0])

# Reaction 1431
reaction('IC4H8OOH-IO2 <=> IC4KETII + OH', [5.000000e+10, 0.0, 21400.0])

# Reaction 1432
reaction('IC4H8OOH-TO2 <=> IC4KETIT + OH', [4.000000e+11, 0.0, 31500.0])

# Reaction 1433
reaction('TC4H8OOH-IO2 <=> TIC4H7Q2-I', [7.500000e+10, 0.0, 24400.0])

# Reaction 1434
reaction('IC4H7OOH + HO2 <=> TIC4H7Q2-I', [1.000000e+11, 0.0, 10600.0])

# Reaction 1435
reaction('IC4H8OOH-IO2 <=> IIC4H7Q2-I', [3.750000e+10, 0.0, 24400.0])

# Reaction 1436
reaction('IC4H8OOH-IO2 <=> IIC4H7Q2-T', [1.000000e+11, 0.0, 29200.0])

# Reaction 1437
reaction('IC4H8OOH-TO2 <=> TIC4H7Q2-I', [6.000000e+11, 0.0, 34500.0])

# Reaction 1438
reaction('AC3H5OOH + CH2O2H <=> IIC4H7Q2-I', [8.500000e+10, 0.0, 10600.0])

# Reaction 1439
reaction('IC4H7OOH + HO2 <=> IIC4H7Q2-T', [1.000000e+11, 0.0, 7800.0])

# Reaction 1440
reaction('CH2O2H <=> CH2O + OH', [9.000000e+14, 0.0, 1500.0])

# Reaction 1441
reaction('IC4KETII => CH2O + C2H5CO + OH', [1.500000e+16, 0.0, 42000.0])

# Reaction 1442
reaction('IC4KETIT => CH3COCH3 + HCO + OH', [9.500000e+15, 0.0, 42540.0])

# Reaction 1443
reaction('IC4H8 + HO2 <=> TC4H8O2H-I', [3.970000e+11, 0.0, 12620.0])

# Reaction 1444
reaction('IC4H8 + HO2 <=> IC4H8O2H-T', [3.970000e+11, 0.0, 12620.0])

# Reaction 1445
reaction('IC4H8O2H-I <=> CC4H8O + OH', [4.470000e+11, 0.0, 21900.0])

# Reaction 1446
reaction('IC4H8O2H-T <=> IC4H8O + OH', [3.090000e+12, 0.0, 13400.0])

# Reaction 1447
reaction('TC4H8O2H-I <=> IC4H8O + OH', [3.980000e+12, 0.0, 17000.0])

# Reaction 1448
reaction('IC4H8O2H-I => OH + CH2O + C3H6', [8.451000e+15, -0.68, 29170.0])

# Reaction 1449
reaction('IC4H8 <=> C3H5-T + CH3', [1.920000e+66, -14.22, 128100.0])

# Reaction 1450
reaction('IC4H8 <=> IC4H7 + H', [3.070000e+55, -11.49, 114300.0])

# Reaction 1451
reaction('IC4H8 + H <=> C3H6 + CH3', [5.680000e+33, -5.72, 20000.0])

# Reaction 1452
reaction('IC4H8 + H <=> IC4H7 + H2', [3.400000e+05, 2.5, 2492.0])

# Reaction 1453
reaction('IC4H8 + O => CH2CO + CH3 + CH3', [3.330000e+07, 1.76, 76.0])

# Reaction 1454
reaction('IC4H8 + O => IC3H6CO + H + H', [1.660000e+07, 1.76, 76.0])

# Reaction 1455
reaction('IC4H8 + O <=> IC4H7 + OH', [1.206000e+11, 0.7, 7633.0])

# Reaction 1456
reaction('IC4H8 + CH3 <=> IC4H7 + CH4', [4.420000e+00, 3.5, 5675.0])

# Reaction 1457
reaction('IC4H8 + HO2 <=> IC4H7 + H2O2', [1.928000e+04, 2.6, 13910.0])

# Reaction 1458
reaction('IC4H8 + O2CHO <=> IC4H7 + HO2CHO', [1.928000e+04, 2.6, 13910.0])

# Reaction 1459
reaction('IC4H8 + O2 <=> IC4H7 + HO2', [6.000000e+12, 0.0, 39900.0])

# Reaction 1460
reaction('IC4H8 + C3H5-A <=> IC4H7 + C3H6', [7.940000e+11, 0.0, 20500.0])

# Reaction 1461
reaction('IC4H8 + C3H5-S <=> IC4H7 + C3H6', [7.940000e+11, 0.0, 20500.0])

# Reaction 1462
reaction('IC4H8 + C3H5-T <=> IC4H7 + C3H6', [7.940000e+11, 0.0, 20500.0])

# Reaction 1463
reaction('IC4H8 + OH <=> IC4H7 + H2O', [5.200000e+06, 2.0, -298.0])

# Reaction 1464
reaction('IC4H8 + O <=> IC3H7 + HCO', [1.580000e+07, 1.76, -1216.0])

# Reaction 1465
reaction('IC4H8 + CH3O2 <=> IC4H7 + CH3O2H', [1.928000e+04, 2.6, 13910.0])

# Reaction 1466
reaction('IC4H8 + HO2 <=> IC4H8O + OH', [1.290000e+12, 0.0, 13340.0])

# Reaction 1467
reaction('IC4H7 + O2 <=> IC3H5CHO + OH', [2.470000e+13, -0.45, 23020.0])

# Reaction 1468
reaction('IC4H7 + O2 <=> CH3COCH2 + CH2O', [7.140000e+15, -1.21, 21050.0])

# Reaction 1469
reaction('IC4H7 + O2 => C3H4-A + CH2O + OH', [7.290000e+29, -5.71, 21450.0])

# Reaction 1470
reaction('IC4H7 + O <=> IC3H5CHO + H', [6.030000e+13, 0.0, 0.0])

# Reaction 1471
reaction('IC4H7 <=> C3H4-A + CH3', [1.230000e+47, -9.74, 74260.0])

# Reaction 1472
reaction('CH3O2 + IC4H7 <=> CH3O + IC4H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1473
reaction('IC4H7 + HO2 <=> IC4H7O + OH', [7.000000e+12, 0.0, -1000.0])

# Reaction 1474
reaction('C3H5-T + CH2O <=> IC4H7O', [1.000000e+11, 0.0, 12600.0])

# Reaction 1475
reaction('IC4H7O <=> IC4H6OH', [1.391000e+11, 0.0, 15600.0])

# Reaction 1476
reaction('IC4H7O <=> IC3H5CHO + H', [5.000000e+13, 0.0, 29100.0])

# Reaction 1477
reaction('IC4H6OH + H2 <=> IC4H7OH + H', [2.160000e+04, 2.38, 18990.0])

# Reaction 1478
reaction('IC4H7OH + O2 <=> IC4H6OH + HO2', [6.000000e+13, 0.0, 39900.0])

# Reaction 1479
reaction('IC4H6OH + CH2O <=> IC4H7OH + HCO', [6.300000e+08, 1.9, 18190.0])

# Reaction 1480
reaction('IC4H6OH + IC4H8 <=> IC4H7OH + IC4H7', [4.700000e+02, 3.3, 19840.0])

# Reaction 1481
reaction('IC4H6OH + H <=> IC4H7OH', [1.000000e+14, 0.0, 0.0])

# Reaction 1482
reaction('IC4H6OH + H2O2 <=> IC4H7OH + HO2', [7.830000e+05, 2.05, 13580.0])

# Reaction 1483
reaction('C3H4-A + CH2OH <=> IC4H6OH', [1.000000e+11, 0.0, 9200.0])

# Reaction 1484
reaction('IC4H7O + O2 <=> IC3H5CHO + HO2', [3.000000e+10, 0.0, 1649.0])

# Reaction 1485
reaction('IC4H7O + HO2 <=> IC3H5CHO + H2O2', [3.000000e+11, 0.0, 0.0])

# Reaction 1486
reaction('IC4H7O + CH3 <=> IC3H5CHO + CH4', [2.400000e+13, 0.0, 0.0])

# Reaction 1487
reaction('IC4H7O + O <=> IC3H5CHO + OH', [6.000000e+12, 0.0, 0.0])

# Reaction 1488
reaction('IC4H7O + OH <=> IC3H5CHO + H2O', [1.810000e+13, 0.0, 0.0])

# Reaction 1489
reaction('IC4H7O + H <=> IC3H5CHO + H2', [1.990000e+13, 0.0, 0.0])

# Reaction 1490
reaction('IC3H5CHO + OH <=> IC3H5CO + H2O', [2.690000e+10, 0.76, -340.0])

# Reaction 1491
reaction('IC3H5CHO + HO2 <=> IC3H5CO + H2O2', [1.000000e+12, 0.0, 11920.0])

# Reaction 1492
reaction('IC3H5CHO + CH3 <=> IC3H5CO + CH4', [3.980000e+12, 0.0, 8700.0])

# Reaction 1493
reaction('IC3H5CHO + O <=> IC3H5CO + OH', [7.180000e+12, 0.0, 1389.0])

# Reaction 1494
reaction('IC3H5CHO + O2 <=> IC3H5CO + HO2', [2.000000e+13, 0.0, 40700.0])

# Reaction 1495
reaction('IC3H5CHO + H <=> IC3H5CO + H2', [2.600000e+12, 0.0, 2600.0])

# Reaction 1496
reaction('IC3H5CO <=> C3H5-T + CO', [1.278000e+20, -1.89, 34460.0])

# Reaction 1497
reaction('TC3H6OCHO + OH <=> TC3H6CHO + HO2', [2.018000e+17, -1.2, 21010.0])

# Reaction 1498
reaction('TC3H6OCHO <=> CH3COCH3 + HCO', [3.980000e+13, 0.0, 9700.0])

# Reaction 1499
reaction('IC3H5CHO + H <=> TC3H6CHO', [1.300000e+13, 0.0, 1200.0])

# Reaction 1500
reaction('IC3H6CO + H <=> TC3H6CHO', [1.300000e+13, 0.0, 4800.0])

# Reaction 1501
reaction('TC3H6CHO + H2 <=> IC3H7CHO + H', [2.160000e+05, 2.38, 18990.0])

# Reaction 1502
reaction('IC4H7O + OH <=> IC4H7OOH', [1.000000e+11, 0.0, 0.0])

# Reaction 1503
reaction('IC4H7O + H <=> IC4H7OH', [4.000000e+13, 0.0, 0.0])

# Reaction 1504
reaction('IC4H7OH + H <=> IC4H8OH', [1.000000e+13, 0.0, 1200.0])

# Reaction 1505
reaction('IC4H7O + H2 <=> IC4H7OH + H', [9.050000e+06, 2.0, 17830.0])

# Reaction 1506
reaction('IC4H7 + OH <=> IC4H7OH', [3.000000e+13, 0.0, 0.0])

# Reaction 1507
reaction('IC4H7OH + HCO <=> IC4H7O + CH2O', [3.020000e+11, 0.0, 18160.0])

# Reaction 1508
reaction('TC3H6CHO + CH2O <=> IC3H7CHO + HCO', [2.520000e+08, 1.9, 18190.0])

# Reaction 1509
reaction('TC3H6CHO + IC4H8 <=> IC3H7CHO + IC4H7', [4.700000e+02, 3.3, 19840.0])

# Reaction 1510
reaction('IC3H6CO + OH <=> IC3H7 + CO2', [1.730000e+12, 0.0, -1010.0])

# Reaction 1511
reaction('TC3H6CHO + OH <=> TC3H6OHCHO', [5.000000e+13, 0.0, 0.0])

# Reaction 1512
reaction('TC3H6OH + HCO <=> TC3H6OHCHO', [1.810000e+13, 0.0, 0.0])

# Reaction 1513
reaction('CH3COCH3 + H <=> TC3H6OH', [8.000000e+12, 0.0, 9500.0])

# Reaction 1514
reaction('IC3H5OH + H <=> TC3H6OH', [6.250000e+11, 0.51, 4020.0])

# Reaction 1515
reaction('C3H5-T + OH <=> IC3H5OH', [5.000000e+13, 0.0, 0.0])

# Reaction 1516
reaction('TC3H6CHO + O2 <=> TC3H6O2CHO', [1.990000e+17, -2.1, 0.0])

# Reaction 1517
reaction('TC3H6O2CHO <=> IC3H5O2HCHO', [6.000000e+11, 0.0, 29880.0])

# Reaction 1518
reaction('TC3H6O2CHO <=> TC3H6O2HCO', [1.000000e+11, 0.0, 25750.0])

# Reaction 1519
reaction('IC3H5CHO + HO2 <=> IC3H5O2HCHO', [2.230000e+11, 0.0, 10600.0])

# Reaction 1520
reaction('TC3H6O2HCO => CH3COCH3 + CO + OH', [4.244000e+18, -1.43, 4800.0])

# Reaction 1521
reaction('TC3H6OH + O2 <=> CH3COCH3 + HO2', [2.230000e+13, 0.0, 0.0])

# Reaction 1522
reaction('IC3H6CO + OH <=> TC3H6OH + CO', [2.000000e+12, 0.0, -1010.0])

# Reaction 1523
reaction('TC3H6CHO + HO2 <=> IC3H7CHO + O2', [3.675000e+12, 0.0, 1310.0])

# Reaction 1524
reaction('TC3H6CHO + CH3 <=> IC3H5CHO + CH4', [3.010000e+12, -0.32, -131.0])

# Reaction 1525
reaction('TC4H8CHO <=> IC3H5CHO + CH3', [1.000000e+13, 0.0, 26290.0])

# Reaction 1526
reaction('TC4H8CHO <=> IC4H8 + HCO', [8.520000e+12, 0.0, 20090.0])

# Reaction 1527
reaction('TC4H8CHO + O2 <=> O2C4H8CHO', [2.000000e+12, 0.0, 0.0])

# Reaction 1528
reaction('O2C4H8CHO <=> O2HC4H8CO', [2.160000e+11, 0.0, 15360.0])

# Reaction 1529
reaction('IC4H8O2H-T + CO <=> O2HC4H8CO', [1.500000e+11, 0.0, 4809.0])

# Reaction 1530
reaction('IC4H7O + IC4H8 <=> IC4H7OH + IC4H7', [2.700000e+11, 0.0, 4000.0])

# Reaction 1531
reaction('IC4H6OH + HO2 => CH2CCH2OH + CH2O + OH', [1.446000e+13, 0.0, 0.0])

# Reaction 1532
reaction('IC4H8 + CH2CCH2OH <=> IC4H7 + C3H5OH', [7.940000e+11, 0.0, 20500.0])

# Reaction 1533
reaction('CH2CCH2OH + H2O2 <=> C3H5OH + HO2', [3.010000e+09, 0.0, 2583.0])

# Reaction 1534
reaction('C3H5OH + OH <=> CH2CCH2OH + H2O', [5.060000e+12, 0.0, 5960.0])

# Reaction 1535
reaction('C3H5OH + H <=> CH2CCH2OH + H2', [3.900000e+05, 2.5, 5821.0])

# Reaction 1536
reaction('C3H5OH + O2 <=> CH2CCH2OH + HO2', [4.000000e+13, 0.0, 60690.0])

# Reaction 1537
reaction('C3H5OH + CH3 <=> CH2CCH2OH + CH4', [2.400000e+11, 0.0, 8030.0])

# Reaction 1538
reaction('CH2CCH2OH + CH3 <=> IC4H7OH', [3.000000e+13, 0.0, 0.0])

# Reaction 1539
reaction('CH2CCH2OH + H <=> C3H5OH', [1.000000e+14, 0.0, 0.0])

# Reaction 1540
reaction('CH2CCH2OH + O2 => CH2OH + CO + CH2O', [4.335000e+12, 0.0, 0.0])

# Reaction 1541
reaction('CH2CCH2OH <=> C2H2 + CH2OH', [2.163000e+40, -8.31, 45110.0])

# Reaction 1542
reaction('C3H4-A + OH <=> CH2CCH2OH', [8.500000e+12, 0.0, 2000.0])

# Reaction 1543
reaction('IC3H5OH + HOCHO <=> CH3COCH3 + HOCHO', [2.810000e-02, 3.29, -4510.0])

# Reaction 1544
falloff_reaction('NC3H7OH (+ M) <=> C3H6 + H2O (+ M)',
                kf=[3.520000e+13, 0.0, 67260.0],
                kf0=[1.988000e+81, -18.59, 78190.0],
                falloff=Troe(A=0.13, T3=1.0, T1=8938000000.0, T2=9330000000.0))

# Reaction 1545
falloff_reaction('NC3H7OH (+ M) <=> CH3 + PC2H4OH (+ M)',
                kf=[8.806000e+23, -2.119, 89820.0],
                kf0=[2.204000e+64, -13.39, 95330.0],
                falloff=Troe(A=0.307, T3=1109.8, T1=9330000000.0, T2=9330000000.0))

# Reaction 1546
falloff_reaction('NC3H7OH (+ M) <=> C2H5 + CH2OH (+ M)',
                kf=[1.031000e+24, -2.203, 88260.0],
                kf0=[1.447000e+65, -13.66, 93980.0],
                falloff=Troe(A=0.293, T3=1101.7, T1=9627000000.0, T2=9330000000.0))

# Reaction 1547
falloff_reaction('NC3H7OH (+ M) <=> NC3H7 + OH (+ M)',
                kf=[5.257000e+20, -1.339, 94630.0],
                kf0=[8.920000e+57, -11.78, 99410.0],
                falloff=Troe(A=0.355, T3=941.95, T1=9920000000.0, T2=9920000000.0))

# Reaction 1548
reaction('NC3H7OH => C3H6OH-3 + H', [1.837000e+17, -0.356, 101200.0])

# Reaction 1549
reaction('C3H6OH-3 + H => NC3H7OH', [1.000000e+14, 0.0, 0.0])

# Reaction 1550
reaction('NC3H7OH => C3H6OH-2 + H', [3.520000e+18, -0.728, 100400.0])

# Reaction 1551
reaction('C3H6OH-2 + H => NC3H7OH', [1.000000e+14, 0.0, 0.0])

# Reaction 1552
reaction('NC3H7OH => C3H6OH-1 + H', [1.778000e+16, -0.173, 95310.0])

# Reaction 1553
reaction('C3H6OH-1 + H => NC3H7OH', [1.000000e+14, 0.0, 0.0])

# Reaction 1554
reaction('NC3H7OH => NC3H7O + H', [7.029000e+14, 0.074, 104900.0])

# Reaction 1555
reaction('NC3H7O + H => NC3H7OH', [1.000000e+14, 0.0, 0.0])

# Reaction 1556
reaction('NC3H7OH + O2 => C3H6OH-3 + HO2', [3.000000e+13, 0.0, 52290.0])

# Reaction 1557
reaction('C3H6OH-3 + HO2 => NC3H7OH + O2', [3.421000e+10, 0.284, -59.0])

# Reaction 1558
reaction('NC3H7OH + O2 => C3H6OH-2 + HO2', [2.440000e+12, 0.22, 49760.0])

# Reaction 1559
reaction('C3H6OH-2 + HO2 => NC3H7OH + O2', [1.452000e+08, 0.876, -1764.0])

# Reaction 1560
reaction('NC3H7OH + O2 => C3H6OH-1 + HO2', [2.000000e+13, 0.0, 49460.0])

# Reaction 1561
reaction('C3H6OH-1 + HO2 => NC3H7OH + O2', [2.356000e+11, 0.101, 3021.0])

# Reaction 1562
reaction('NC3H7OH + O2 => NC3H7O + HO2', [1.000000e+13, 0.0, 48200.0])

# Reaction 1563
reaction('NC3H7O + HO2 => NC3H7OH + O2', [2.980000e+12, -0.146, -7791.0])

# Reaction 1564
reaction('NC3H7OH + H => C3H6OH-3 + H2', [6.660000e+05, 2.54, 6756.0])

# Reaction 1565
reaction('C3H6OH-3 + H2 => NC3H7OH + H', [1.448000e+02, 3.172, 9096.0])

# Reaction 1566
reaction('NC3H7OH + H => C3H6OH-2 + H2', [1.113000e+05, 2.4, 2010.0])

# Reaction 1567
reaction('C3H6OH-2 + H2 => NC3H7OH + H', [1.263000e+00, 3.404, 5170.0])

# Reaction 1568
reaction('NC3H7OH + H => C3H6OH-1 + H2', [1.790000e+05, 2.53, 3420.0])

# Reaction 1569
reaction('C3H6OH-1 + H2 => NC3H7OH + H', [4.021000e+02, 2.979, 11670.0])

# Reaction 1570
reaction('NC3H7OH + H => NC3H7O + H2', [5.360000e+04, 2.53, 4405.0])

# Reaction 1571
reaction('NC3H7O + H2 => NC3H7OH + H', [3.044000e+03, 2.732, 3103.0])

# Reaction 1572
reaction('NC3H7OH + OH => C3H6OH-3 + H2O', [5.280000e+09, 0.97, 1586.0])

# Reaction 1573
reaction('C3H6OH-3 + H2O => NC3H7OH + OH', [1.217000e+07, 1.496, 18820.0])

# Reaction 1574
reaction('NC3H7OH + OH => C3H6OH-2 + H2O', [5.588000e+07, 1.61, 1176.0])

# Reaction 1575
reaction('C3H6OH-2 + H2O => NC3H7OH + OH', [6.724000e+03, 2.508, 19230.0])

# Reaction 1576
reaction('NC3H7OH + OH => C3H6OH-1 + H2O', [5.560000e+10, 0.5, -380.0])

# Reaction 1577
reaction('C3H6OH-1 + H2O => NC3H7OH + OH', [1.324000e+09, 0.843, 22760.0])

# Reaction 1578
reaction('NC3H7OH + OH => NC3H7O + H2O', [1.500000e+10, 0.8, 2534.0])

# Reaction 1579
reaction('NC3H7O + H2O => NC3H7OH + OH', [9.038000e+09, 0.896, 16120.0])

# Reaction 1580
reaction('NC3H7OH + O => C3H6OH-3 + OH', [9.810000e+05, 2.43, 4750.0])

# Reaction 1581
reaction('C3H6OH-3 + OH => NC3H7OH + O', [1.107000e+02, 3.043, 5678.0])

# Reaction 1582
reaction('NC3H7OH + O => C3H6OH-2 + OH', [1.440000e+05, 2.61, 3029.0])

# Reaction 1583
reaction('C3H6OH-2 + OH => NC3H7OH + O', [8.484000e-01, 3.595, 4777.0])

# Reaction 1584
reaction('NC3H7OH + O => C3H6OH-1 + OH', [1.450000e+05, 2.47, 876.0])

# Reaction 1585
reaction('C3H6OH-1 + OH => NC3H7OH + O', [1.691000e+02, 2.9, 7714.0])

# Reaction 1586
reaction('NC3H7OH + O => NC3H7O + OH', [1.580000e+07, 2.0, 4448.0])

# Reaction 1587
reaction('NC3H7O + OH => NC3H7OH + O', [4.661000e+05, 2.183, 1734.0])

# Reaction 1588
reaction('NC3H7OH + HO2 => C3H6OH-3 + H2O2', [2.380000e+04, 2.55, 16490.0])

# Reaction 1589
reaction('C3H6OH-3 + H2O2 => NC3H7OH + HO2', [2.995000e+03, 2.487, 2833.0])

# Reaction 1590
reaction('NC3H7OH + HO2 => C3H6OH-2 + H2O2', [2.300000e-03, 3.37, 466.0])

# Reaction 1591
reaction('C3H6OH-2 + H2O2 => NC3H7OH + HO2', [1.510000e-05, 3.679, -12380.0])

# Reaction 1592
reaction('NC3H7OH + HO2 => C3H6OH-1 + H2O2', [6.000000e+12, 0.0, 16000.0])

# Reaction 1593
reaction('C3H6OH-1 + H2O2 => NC3H7OH + HO2', [7.798000e+12, -0.246, 8249.0])

# Reaction 1594
reaction('NC3H7OH + HO2 => NC3H7O + H2O2', [2.500000e+12, 0.0, 24000.0])

# Reaction 1595
reaction('NC3H7O + H2O2 => NC3H7OH + HO2', [8.220000e+13, -0.493, 6697.0])

# Reaction 1596
reaction('NC3H7OH + CH3O2 => C3H6OH-3 + CH3O2H', [2.380000e+04, 2.55, 16490.0])

# Reaction 1597
reaction('C3H6OH-3 + CH3O2H => NC3H7OH + CH3O2', [5.826000e+04, 2.042, 1238.0])

# Reaction 1598
reaction('NC3H7OH + CH3O2 => C3H6OH-2 + CH3O2H', [1.560000e+03, 2.81, 14280.0])

# Reaction 1599
reaction('C3H6OH-2 + CH3O2H => NC3H7OH + CH3O2', [1.993000e+02, 2.674, -152.0])

# Reaction 1600
reaction('NC3H7OH + CH3O2 => C3H6OH-1 + CH3O2H', [6.000000e+12, 0.0, 16000.0])

# Reaction 1601
reaction('C3H6OH-1 + CH3O2H => NC3H7OH + CH3O2', [1.517000e+14, -0.691, 6654.0])

# Reaction 1602
reaction('NC3H7OH + CH3O2 => NC3H7O + CH3O2H', [2.500000e+12, 0.0, 24000.0])

# Reaction 1603
reaction('NC3H7O + CH3O2H => NC3H7OH + CH3O2', [1.599000e+15, -0.938, 5102.0])

# Reaction 1604
reaction('NC3H7OH + CH3 => C3H6OH-3 + CH4', [4.530000e-01, 3.65, 7154.0])

# Reaction 1605
reaction('C3H6OH-3 + CH4 => NC3H7OH + CH3', [8.987000e-02, 3.836, 11030.0])

# Reaction 1606
reaction('NC3H7OH + CH3 => C3H6OH-2 + CH4', [8.020000e+00, 3.23, 6461.0])

# Reaction 1607
reaction('C3H6OH-2 + CH4 => NC3H7OH + CH3', [8.304000e-02, 3.788, 11160.0])

# Reaction 1608
reaction('NC3H7OH + CH3 => C3H6OH-1 + CH4', [1.993000e+01, 3.37, 7634.0])

# Reaction 1609
reaction('C3H6OH-1 + CH4 => NC3H7OH + CH3', [4.084000e+01, 3.373, 17420.0])

# Reaction 1610
reaction('NC3H7OH + CH3 => NC3H7O + CH4', [2.040000e+00, 3.57, 7721.0])

# Reaction 1611
reaction('NC3H7O + CH4 => NC3H7OH + CH3', [1.058000e+02, 3.326, 7959.0])

# Reaction 1612
reaction('NC3H7OH + C2H5 => C3H6OH-3 + C2H6', [4.620000e-01, 3.65, 9142.0])

# Reaction 1613
reaction('C3H6OH-3 + C2H6 => NC3H7OH + C2H5', [1.088000e+00, 3.6, 9252.0])

# Reaction 1614
reaction('NC3H7OH + C2H5 => C3H6OH-2 + C2H6', [7.660000e+00, 3.11, 12210.0])

# Reaction 1615
reaction('C3H6OH-2 + C2H6 => NC3H7OH + C2H5', [9.411000e-01, 3.432, 13140.0])

# Reaction 1616
reaction('NC3H7OH + C2H5 => C3H6OH-1 + C2H6', [2.000000e+11, 0.0, 11000.0])

# Reaction 1617
reaction('C3H6OH-1 + C2H6 => NC3H7OH + C2H5', [4.863000e+12, -0.233, 17020.0])

# Reaction 1618
reaction('NC3H7OH + C2H5 => NC3H7O + C2H6', [1.000000e+11, 0.0, 9200.0])

# Reaction 1619
reaction('NC3H7O + C2H6 => NC3H7OH + C2H5', [6.152000e+13, -0.48, 5668.0])

# Reaction 1620
reaction('NC3H7OH + HCO => C3H6OH-3 + CH2O', [1.020000e+05, 2.5, 18440.0])

# Reaction 1621
reaction('C3H6OH-3 + CH2O => NC3H7OH + HCO', [3.756000e+03, 2.845, 5592.0])

# Reaction 1622
reaction('NC3H7OH + HCO => C3H6OH-2 + CH2O', [5.160000e+05, 2.25, 16760.0])

# Reaction 1623
reaction('C3H6OH-2 + CH2O => NC3H7OH + HCO', [9.917000e+02, 2.967, 4730.0])

# Reaction 1624
reaction('NC3H7OH + HCO => C3H6OH-1 + CH2O', [1.000000e+07, 1.9, 17000.0])

# Reaction 1625
reaction('C3H6OH-1 + CH2O => NC3H7OH + HCO', [3.804000e+06, 2.062, 10060.0])

# Reaction 1626
reaction('NC3H7OH + HCO => NC3H7O + CH2O', [3.400000e+04, 2.5, 13500.0])

# Reaction 1627
reaction('NC3H7O + CH2O => NC3H7OH + HCO', [3.272000e+05, 2.415, -2992.0])

# Reaction 1628
reaction('NC3H7OH + CH2OH => C3H6OH-3 + CH3OH', [1.010000e+02, 2.95, 13970.0])

# Reaction 1629
reaction('C3H6OH-3 + CH3OH => NC3H7OH + CH2OH', [1.908000e+00, 3.473, 10020.0])

# Reaction 1630
reaction('NC3H7OH + CH2OH => C3H6OH-2 + CH3OH', [1.530000e+01, 3.11, 12210.0])

# Reaction 1631
reaction('C3H6OH-2 + CH3OH => NC3H7OH + CH2OH', [1.508000e-02, 4.005, 9082.0])

# Reaction 1632
reaction('NC3H7OH + CH2OH => C3H6OH-1 + CH3OH', [6.000000e+01, 2.95, 12000.0])

# Reaction 1633
reaction('C3H6OH-1 + CH3OH => NC3H7OH + CH2OH', [1.171000e+01, 3.29, 13960.0])

# Reaction 1634
reaction('NC3H7OH + CH2OH => NC3H7O + CH3OH', [1.200000e+02, 2.76, 10800.0])

# Reaction 1635
reaction('NC3H7O + CH3OH => NC3H7OH + CH2OH', [5.923000e+02, 2.854, 3210.0])

# Reaction 1636
reaction('NC3H7OH + CH3O => C3H6OH-3 + CH3OH', [2.170000e+11, 0.0, 6458.0])

# Reaction 1637
reaction('C3H6OH-3 + CH3OH => NC3H7OH + CH3O', [1.013000e+09, 0.419, 9025.0])

# Reaction 1638
reaction('NC3H7OH + CH3O => C3H6OH-2 + CH3OH', [3.020000e+00, 0.18, 4703.0])

# Reaction 1639
reaction('C3H6OH-2 + CH3OH => NC3H7OH + CH3O', [7.358000e-04, 0.971, 8090.0])

# Reaction 1640
reaction('NC3H7OH + CH3O => C3H6OH-1 + CH3OH', [1.500000e+11, 0.0, 4500.0])

# Reaction 1641
reaction('C3H6OH-1 + CH3OH => NC3H7OH + CH3O', [7.234000e+09, 0.236, 12980.0])

# Reaction 1642
reaction('NC3H7OH + CH3O => NC3H7O + CH3OH', [2.300000e+10, 0.0, 2900.0])

# Reaction 1643
reaction('NC3H7O + CH3OH => NC3H7OH + CH3O', [2.806000e+10, -0.01, 1825.0])

# Reaction 1644
reaction('C3H6OH-3 => C2H4 + CH2OH', [1.541000e+11, 0.833, 28660.0])

# Reaction 1645
reaction('C2H4 + CH2OH => C3H6OH-3', [1.320000e+04, 2.48, 6130.0])

# Reaction 1646
reaction('C3H6OH-3 <=> C3H5OH + H', [3.000000e+13, 0.0, 36000.0])

# Reaction 1647
reaction('O2 + C3H6OH-3 <=> HO2 + C3H5OH', [1.500000e+12, 0.0, 5000.0])

# Reaction 1648
reaction('C3H6OH-2 => C3H6 + OH', [5.093000e+15, -0.95, 26260.0])

# Reaction 1649
reaction('C3H6 + OH => C3H6OH-2', [4.950000e+11, 0.0, -960.0])

# Reaction 1650
reaction('H + C3H5OH <=> C3H6OH-2', [4.240000e+11, 0.51, 1230.0])

# Reaction 1651
reaction('H + CH3CHCHOH <=> C3H6OH-2', [2.500000e+11, 0.51, 2620.0])

# Reaction 1652
reaction('O2 + C3H6OH-2 <=> HO2 + C2H5CHO', [1.500000e+12, 0.0, 5000.0])

# Reaction 1653
reaction('O2 + C3H6OH-2 <=> HO2 + C3H5OH', [1.500000e+12, 0.0, 5000.0])

# Reaction 1654
reaction('C3H6OH-1 => C2H3OH + CH3', [5.011000e+10, 1.038, 30450.0])

# Reaction 1655
reaction('C2H3OH + CH3 => C3H6OH-1', [1.760000e+04, 2.48, 6130.0])

# Reaction 1656
reaction('C3H6OH-1 => C2H5CHO + H', [7.035000e+09, 0.993, 32590.0])

# Reaction 1657
reaction('C2H5CHO + H => C3H6OH-1', [8.000000e+12, 0.0, 9500.0])

# Reaction 1658
reaction('H + CH3CHCHOH <=> C3H6OH-1', [2.500000e+11, 0.51, 2620.0])

# Reaction 1659
pdep_arrhenius('O2 + C3H6OH-1 <=> HO2 + C2H5CHO',
            [(0.001, 'atm'), 5.260000e+17, -1.637, 838.0],
            [(0.01, 'atm'), 5.260000e+17, -1.637, 838.0],
            [(0.1, 'atm'), 5.260000e+17, -1.637, 838.0],
            [(1.0, 'atm'), 5.280000e+17, -1.638, 839.0],
            [(10.0, 'atm'), 1.540000e+18, -1.771, 1120.0],
            [(100.0, 'atm'), 3.780000e+20, -2.429, 3090.0])

# Reaction 1660
reaction('C2H3OH => C2H3 + OH', [6.899000e+21, -1.564, 110700.0])

# Reaction 1661
reaction('C2H3 + OH => C2H3OH', [2.410000e+13, 0.0, 0.0])

# Reaction 1662
reaction('C2H3OH => CH2CHO + H', [3.643000e+15, -0.397, 95390.0])

# Reaction 1663
reaction('CH2CHO + H => C2H3OH', [2.400000e+13, 0.0, 12510.0])

# Reaction 1664
falloff_reaction('IC3H7OH (+ M) <=> C3H6 + H2O (+ M)',
                kf=[2.110000e+14, 0.0, 67260.0],
                kf0=[3.611000e+82, -18.85, 79340.0],
                falloff=Troe(A=0.136, T3=1.0, T1=9937000000.0, T2=9330000000.0))

# Reaction 1665
falloff_reaction('IC3H7OH (+ M) <=> CH3 + SC2H4OH (+ M)',
                kf=[1.345000e+24, -2.274, 86960.0],
                kf0=[2.852000e+69, -14.89, 94650.0],
                falloff=Troe(A=0.336, T3=774.5, T1=9684000000.0, T2=8214000000.0))

# Reaction 1666
falloff_reaction('IC3H7OH (+ M) <=> IC3H7 + OH (+ M)',
                kf=[1.753000e+22, -1.747, 96320.0],
                kf0=[3.924000e+61, -12.8, 102950.0],
                falloff=Troe(A=0.57, T3=8830000000.0, T1=754.8, T2=9330000000.0))

# Reaction 1667
reaction('IC3H7OH + O2 => IC3H6OH + HO2', [4.200000e+13, 0.0, 52000.0])

# Reaction 1668
reaction('IC3H6OH + HO2 => IC3H7OH + O2', [3.536000e+10, 0.32, -1212.0])

# Reaction 1669
reaction('IC3H7OH + O2 => TC3H6OH + HO2', [7.000000e+12, 0.0, 43700.0])

# Reaction 1670
reaction('TC3H6OH + HO2 => IC3H7OH + O2', [6.722000e+08, 0.67, -1402.0])

# Reaction 1671
reaction('IC3H7OH + OH => IC3H6OH + H2O', [1.054000e+10, 0.97, 1586.0])

# Reaction 1672
reaction('IC3H6OH + H2O => IC3H7OH + OH', [2.115000e+09, 0.97, 21010.0])

# Reaction 1673
reaction('IC3H7OH + OH => TC3H6OH + H2O', [2.990000e+13, -0.49, -235.0])

# Reaction 1674
reaction('TC3H6OH + H2O => IC3H7OH + OH', [2.365000e+08, 0.904, 24350.0])

# Reaction 1675
reaction('IC3H7OH + H => IC3H6OH + H2', [1.330000e+06, 2.54, 6756.0])

# Reaction 1676
reaction('IC3H6OH + H2 => IC3H7OH + H', [6.165000e+04, 2.54, 11030.0])

# Reaction 1677
reaction('IC3H7OH + H => TC3H6OH + H2', [2.300000e+05, 2.72, 6260.0])

# Reaction 1678
reaction('TC3H6OH + H2 => IC3H7OH + H', [1.216000e+03, 3.06, 18640.0])

# Reaction 1679
reaction('IC3H7OH + O => IC3H6OH + OH', [1.960000e+06, 2.43, 4750.0])

# Reaction 1680
reaction('IC3H6OH + OH => IC3H7OH + O', [3.990000e+04, 2.43, 6925.0])

# Reaction 1681
reaction('IC3H7OH + O => TC3H6OH + OH', [5.130000e+11, 0.21, 4890.0])

# Reaction 1682
reaction('TC3H6OH + OH => IC3H7OH + O', [1.191000e+09, 0.55, 15180.0])

# Reaction 1683
reaction('IC3H7OH + CH3 => IC3H6OH + CH4', [9.040000e-01, 3.65, 7154.0])

# Reaction 1684
reaction('IC3H6OH + CH4 => IC3H7OH + CH3', [1.095000e+00, 3.65, 11900.0])

# Reaction 1685
reaction('IC3H7OH + CH3 => TC3H6OH + CH4', [2.390000e-04, 4.97, 5710.0])

# Reaction 1686
reaction('TC3H6OH + CH4 => IC3H7OH + CH3', [3.300000e-05, 5.31, 18570.0])

# Reaction 1687
reaction('IC3H7OH + CH3O => IC3H6OH + CH3OH', [3.000000e+11, 0.0, 7000.0])

# Reaction 1688
reaction('IC3H6OH + CH3OH => IC3H7OH + CH3O', [6.704000e+09, 0.02, 9540.0])

# Reaction 1689
reaction('IC3H7OH + CH3O => TC3H6OH + CH3OH', [4.340000e+11, 0.0, 6460.0])

# Reaction 1690
reaction('TC3H6OH + CH3OH => IC3H7OH + CH3O', [1.106000e+09, 0.37, 17110.0])

# Reaction 1691
reaction('IC3H7OH + HO2 => IC3H6OH + H2O2', [1.680000e+13, 0.0, 20440.0])

# Reaction 1692
reaction('IC3H6OH + H2O2 => IC3H7OH + HO2', [2.001000e+13, -0.33, 8397.0])

# Reaction 1693
reaction('IC3H7OH + HO2 => TC3H6OH + H2O2', [2.800000e+12, 0.0, 14860.0])

# Reaction 1694
reaction('TC3H6OH + H2O2 => IC3H7OH + HO2', [3.804000e+11, 0.02, 10930.0])

# Reaction 1695
reaction('IC3H7OH + CH3O2 => IC3H6OH + CH3O2H', [1.680000e+13, 0.0, 20440.0])

# Reaction 1696
reaction('IC3H6OH + CH3O2H => IC3H7OH + CH3O2', [4.167000e+13, -0.51, 5377.0])

# Reaction 1697
reaction('IC3H7OH + CH3O2 => TC3H6OH + CH3O2H', [2.800000e+12, 0.0, 14860.0])

# Reaction 1698
reaction('TC3H6OH + CH3O2H => IC3H7OH + CH3O2', [7.921000e+11, -0.16, 7912.0])

# Reaction 1699
reaction('H + TC3H6OH <=> IC3H7OH', [4.000000e+13, 0.0, 0.0])

# Reaction 1700
reaction('H + IC3H6OH <=> IC3H7OH', [4.000000e+13, 0.0, 0.0])

# Reaction 1701
reaction('H + IC3H7O <=> IC3H7OH', [1.000000e+14, 0.0, 0.0])

# Reaction 1702
reaction('OH + IC3H7OH <=> H2O + IC3H7O', [5.880000e+02, 2.82, -585.0])

# Reaction 1703
reaction('O + IC3H7OH => OH + IC3H7O', [1.460000e-03, 4.73, 1727.0])

# Reaction 1704
reaction('HO2 + IC3H7OH => H2O2 + IC3H7O', [6.470000e-07, 5.3, 10530.0])

# Reaction 1705
reaction('HCO + IC3H7OH => CH2O + IC3H7O', [3.400000e+04, 2.5, 13500.0])

# Reaction 1706
reaction('CH3 + IC3H7OH => CH4 + IC3H7O', [1.020000e+00, 3.57, 8221.0])

# Reaction 1707
reaction('O2 + IC3H7OH => HO2 + IC3H7O', [3.000000e+13, 0.0, 52340.0])

# Reaction 1708
reaction('H + IC3H7OH => H2 + IC3H7O', [9.450000e+02, 3.14, 8701.0])

# Reaction 1709
reaction('C2H5 + IC3H7OH => C2H6 + IC3H7O', [1.670000e+10, 0.0, 13400.0])

# Reaction 1710
reaction('IC3H7OH + HCO <=> IC3H6OH + CH2O', [1.020000e+05, 2.5, 18440.0])

# Reaction 1711
reaction('IC3H7OH + HCO <=> TC3H6OH + CH2O', [5.000000e+06, 1.9, 17000.0])

# Reaction 1712
reaction('IC3H7OH + C2H5 <=> IC3H6OH + C2H6', [5.010000e+10, 0.0, 13400.0])

# Reaction 1713
reaction('IC3H7OH + C2H5 <=> TC3H6OH + C2H6', [1.010000e+11, 0.0, 7900.0])

# Reaction 1714
reaction('IC3H6OH => C3H6 + OH', [6.408000e+15, -0.6, 28840.0])

# Reaction 1715
reaction('C3H6 + OH => IC3H6OH', [2.920000e+12, 0.0, -1000.0])

# Reaction 1716
reaction('IC3H6OH => C2H3OH + CH3', [2.689000e+09, 1.319, 29480.0])

# Reaction 1717
reaction('C2H3OH + CH3 => IC3H6OH', [1.890000e+03, 2.67, 6850.0])

# Reaction 1718
reaction('IC3H5OH <=> CH3COCH3', [8.590000e+11, 0.318, 55900.0])

# Reaction 1719
reaction('IC3H5OH + H <=> IC3H6OH', [6.250000e+11, 0.51, 2620.0])

# Reaction 1720
reaction('IC3H5OH + HO2 <=> CH3COCH3 + HO2', [1.490000e+05, 1.67, 6810.0])

# Reaction 1721
reaction('O2 + NC3H7O <=> HO2 + C2H5CHO', [1.500000e+12, 0.0, 5000.0])

# Reaction 1722
reaction('C3H6OH-3 <=> NC3H7O', [1.000000e+11, 0.0, 20600.0])

# Reaction 1723
reaction('CH3CHCHOH + HO2 <=> C2H5CHO + HO2', [1.490000e+05, 1.67, 6810.0])

# Reaction 1724
reaction('CH3CHCHOH + HOCHO <=> C2H5CHO + HOCHO', [2.810000e-02, 3.286, -4509.0])

# Reaction 1725
reaction('CH3CHCHOH <=> C2H5CHO', [8.590000e+11, 0.318, 55900.0])


    """)


    # ================================================================
    # ================================================================
    # ================================================================
    # ================================================================

    #fichier_data = open('C6_Zhang.cti', 'w')
    #fichier_data.write("""##-------------------------------------------------------------------------------
## Zhang et al. 2015  http://dx.doi.org/10.1016/j.combustflame.2015.08.001
##-------------------------------------------------------------------------------
        
#units(length='cm', time='s', quantity='mol', act_energy='cal/mol')

#ideal_gas(name='gas',
        #elements="C H N O Ar He",
        #species=\"""H                 H2                O                 O2                OH
                    #OH*               H2O               N2                HO2               H2O2
                    #AR                CO                CO2               CH2O              HCO
                    #HO2CHO            HCOH              O2CHO             HOCHO             OCHO
                    #HOCH2O2H          HOCH2O2           OCH2O2H           HOCH2O            CH3OH
                    #CH2OH             CH3O              CH3O2H            CH3O2             CH4
                    #CH3               CH2               CH2(S)            C                 CH
                    #CH*               C2H6              C2H5              C2H4              C2H3
                    #C2H2              C2H               CH3CHO            C2H3OH            C2H2OH
                    #CH3CO             CH2CHO            O2CH2CHO          HO2CH2CO          CH2CO
                    #HCCO              HCCOH             CH3CO3H           CH3CO3            CH3CO2
                    #C2H5OH            C2H5O             PC2H4OH           SC2H4OH           O2C2H4OH
                    #C2H5O2H           C2H5O2            C2H4O2H           C2H4O1-2          C2H3O1-2
                    #CH3COCH3          CH3COCH2          CH3COCH2O2        C3KET21           C2H3CHO
                    #C2H3CO            C2H5CHO           C2H5CO            CH3OCH3           CH3OCH2
                    #CH3OCH2O2         CH2OCH2O2H        CH3OCH2O2H        CH3OCH2O          O2CH2OCH2O2H
                    #HO2CH2OCHO        OCH2OCHO          HOCH2OCO          CH3OCHO           CH3OCO
                    #CH2OCHO           HE                C3H8              IC3H7             NC3H7
                    #C3H6              C3H5-A            C3H5-S            C3H5-T            C3H4-P
                    #C3H4-A            C3H3              C3H5O             C3H6OOH1-2        C3H6OOH1-3
                    #C3H6OOH2-1        C3H6OOH1-2O2      C3H6OOH1-3O2      C3H6OOH2-1O2      NC3H7O2H
                    #IC3H7O2H          NC3H7O2           IC3H7O2           NC3H7O            IC3H7O
                    #C3H6O1-2          C3H6O1-3          C3KET12           C3KET13           C3H51-2,3OOH
                    #C3H52-1,3OOH      C3H5O1-2OOH-3     C3H5O1-3OOH-2     C3H6OH1-2         CH3CHCO
                    #AC3H5OOH          C2H3OOH           CC3H4             H2CC              FULVENE
                    #CC3H6             C4H10             C4H8-1            C4H8-2            PC4H9
                    #SC4H9             C4H71-1           C4H71-2           C4H71-3           C4H71-4
                    #C4H72-2           C4H6              PC4H9O2H          SC4H9O2H          PC4H9O2
                    #SC4H9O2           PC4H9O            SC4H9O            C4H71-O           C4H8O1-2
                    #C4H8O1-3          C4H8O1-4          C4H8O2-3          SC4H8OH-3         C4H8OH-1O2
                    #C4H8OH-2O2        C4H8OOH1-2        C4H8OOH1-3        C4H8OOH1-4        C4H8OOH2-1
                    #C4H8OOH2-3        C4H8OOH2-4        C4H8OOH1-2O2      C4H8OOH1-3O2      C4H8OOH1-4O2
                    #C4H8OOH2-1O2      C4H8OOH2-3O2      C4H8OOH2-4O2      C4H72-1OOH        C4H71-4OOH
                    #C4H71-3OOH        C4H71-3,4OOH      C4H72-3,4OOH      C4H71-2,4OOH      C4H72-1,3OOH
                    #C4H72-1,4OOH      C4H71-2,3OOH      C4H7O1-3OOH-4     C4H7O1-3OOH-2     C4H7O1-2OOH-4
                    #C4H7O1-4OOH-2     C4H7O1-2OOH-3     C4H7O2-3OOH-1     HO2CH2CHO         NC4KET12
                    #NC4KET13          NC4KET14          NC4KET21          NC4KET23          NC4KET24
                    #C2H5COCH3         C2H5COCH2         CH2CH2COCH3       CH3CHCOCH3        C2H3COCH3
                    #CH3CHOOCOCH3      CH2CHOOHCOCH3     NC3H7CHO          NC3H7CO           C3H6CHO-3
                    #C3H6CHO-2         C3H6CHO-1         C2H5CHCO          SC3H5CHO          SC3H5CO
                    #CH2CH2CHO         IC4H10            IC4H9             IC4H8             IC4H7
                    #TC4H9             TC4H9O2           IC4H7OOCH3        IC4H7OOIC4H7      IC3H5OCH2
                    #IC4H7-I1          IC4H8OH-IT        IC4H8OH-TI        IQJC4H8OH         IQC4H8OT
                    #TQJC4H8OH         TQC4H8OI          TQC4H7OHI         IC3H4CHO-A        QC4H7OHP
                    #IQC4H7OHT         SC4H7OH-IP        C2CY(COC)OH       CCY(CCO)COH       CCY(CCOC)OH
                    #IC3H6OHCHO        CH3COCOOH         IC3H5COHQ         IC3H5Q            TQC4H7OHIO2
                    #TQC4H7OHIQ-I      TQC4H7OHIQ-P      CH2CQCOHQ         COHQCYC(COC)      QCYC(CCOC)OH
                    #TQC4H7OHTO2       HOCOCQ(CH3)2      IQC4H7OHTO2       TC3H6OH           IQC4H7OHTQ-P
                    #IQC4H8OTQ-I       CH2COHCH2OOH      IC4H6Q2-II        CO(CH2OOH)2       CHOC(CH3)OHCH2Q
                    #CH3COCHO          SC4H7OH-I         CCYCCC            H15DE25DM         H15DE25DM-S
                    #H15DE25DM-A       H15DE25DM-AO      H15DE2M-T         H15DE25DM-SO      IC4H7CHO
                    #IC4H7O2           IC4H6OOH-I        C*CYCCOC          CCYCCOOC-T1       CCYCCOOC-I2
                    #CHOIC3H6O         CCY(C2O)CO        CCYCCO-T1         C2CYCOOC-I1       IC3H5OOCH2
                    #B2E3M1OJ          B13DE2M           B13DE2MJ          B2E2M1OJ          B12DE3M
                    #IC4H9O2           TC4H8O2H-I        IC4H8O2H-I        IC4H8O2H-T        IC4H8O
                    #CC4H8O            IC4H9O            TC4H9O            IC4H9O2H          TC4H9O2H
                    #IC4H7O            IC4H8OH           IC3H7CHO          TC3H6CHO          IC3H7CO
                    #IC3H6CHO          TC4H8OOH-IO2      IC4H8OOH-IO2      IC4H8OOH-TO2      IC4KETII
                    #IC4KETIT          IC4H7OH           IC4H6OH           IC3H5CHO          IC3H5CO
                    #TC3H6OCHO         IC3H6CO           IC4H7OOH          TC3H6OHCHO        C3H6OH2-1
                    #IC3H5OH           TC3H6O2CHO        TC3H6O2HCO        IC3H5O2HCHO       CH2CCH2OH
                    #TC4H8CHO          O2C4H8CHO         O2HC4H8CO         C3H5OH            TIC4H7Q2-I
                    #IIC4H7Q2-T        IIC4H7Q2-I        CH2O2H            C4H4              C4H3-I
                    #C4H3-N            C4H612            C4H2              C4H5-I            C4H5-N
                    #CH2CHCHCHO        C4H6O25           C2H3CHOCH2        C4H5-2            C4H6-2
                    #C4H6O23           C4H4O             H2C4O             C2H3OCH2          C6H6
                    #C*CCJC*C          C6H5              C6H2              C6H3              L-C6H4
                    #C-C6H4            C6H5O             C6H5OH            P-C6H4O2          O-C6H4O2
                    #P-C6H3O2          C5H6              C5H5              C5H4O             C5H5O
                    #C5H5OH            C5H4OH            C5H3O             C6H5OO            C6H5OOH
                    #OC6H4OH           C6H4OH            O-OC6H5OJ         P-OC6H5OJ         C#CC*CCJ
                    #C5H6-L            CJ*CC*CC*O        C*CC*CCJ*O        CJ*CC*O           C5H7
                    #OC5H7O            C*CCJC*COH        C*CC*CCJ          C*CC*CCOH         OC4H6O
                    #OC4H5O            HOCO              HOC*CC*O          HOC*CCJ*O         O2CCHOOJ
                    #C6H101-5          C6H9-A            C3H2              C3H2(S)           C3H2C
                    #H2CCC(S)          C3H3O             C3H3O2H           C2HCHO            CH3COCH2O
                    #PC3H4OH-2         SC3H4OH           SC3H5OH           CH3CHCHO          AC4H7OOH
                    #TQJC3H6OH         TQC3H6OI          QC3H5OHP          TQC3H5OHI         CCY(COC)OH
                    #IQJC3H6OH         IQC3H6OT          IQC3H5OHPJ        CHOCH2OOH         CY(CCOC)OH
                    #CHOCOHCH3         CY(COC)COH        TQC3H5OHIO2       TQC3H5OHIQ-I      TQC3H5OHIQ-P
                    #C2H3COHOOH        COHOOHCY(COC)     OHCY(COCC)OOH     IQC3H5OHPJO2      IQC3H5OTQ-I
                    #IQC3H5OHQ-SJ      CHOCOHCH2OOH      CHOCHO            TQC3H5OHTO2       OHCOCOOHCH3
                    #C2H3OO            CHCHO             C4H71-3OH         C4H71-1OH         C4H71-2OH
                    #PC4H8OH-2         SC4H8OH-1         C-2-C4H8          C4H6-1            C3H5OCH2-1
                    #NC5H12            C5H11-1           C5H11-2           C5H11-3           C5H10-1
                    #C5H10-2           C5H81-3           C5H91-3           C5H91-4           C5H91-5
                    #C5H92-4           C5H92-5           C5H9O1-3          C5H9O2-4          C5H11O2H-1
                    #C5H11O2H-2        C5H11O2H-3        C5H11O2-1         C5H11O2-2         C5H11O2-3
                    #C5H11O-1          C5H11O-2          C5H11O-3          C5H10OOH1-2       C5H10OOH1-3
                    #C5H10OOH1-4       C5H10OOH1-5       C5H10OOH2-1       C5H10OOH2-3       C5H10OOH2-4
                    #C5H10OOH2-5       C5H10OOH3-1       C5H10OOH3-2       C5H10O1-2         C5H10O1-3
                    #C5H10O1-4         C5H10O1-5         C5H10O2-3         C5H10O2-4         C5H10OOH1-2O2
                    #C5H10OOH1-3O2     C5H10OOH1-4O2     C5H10OOH1-5O2     C5H10OOH2-1O2     C5H10OOH2-3O2
                    #C5H10OOH2-4O2     C5H10OOH2-5O2     C5H10OOH3-1O2     C5H10OOH3-2O2     C5H91-2,3OOH
                    #C5H91-2,4OOH      C5H91-2,5OOH      C5H91-3,4OOH      C5H91-3,5OOH      C5H91-4,5OOH
                    #C5H92-1,3OOH      C5H92-1,4OOH      C5H92-1,5OOH      C5H92-3,4OOH      C5H92-3,5OOH
                    #C5H92-4,5OOH      C5H93-1,2OOH      C5H93-1,4OOH      C5H93-1,5OOH      C5H93-2,4OOH
                    #C5H91-3OOH        C5H91-4OOH        C5H91-5OOH        C5H92-1OOH        C5H92-4OOH
                    #C5H92-5OOH        C5H9O1-2OOH-3     C5H9O1-2OOH-4     C5H9O1-2OOH-5     C5H9O1-3OOH-2
                    #C5H9O1-3OOH-4     C5H9O1-3OOH-5     C5H9O1-4OOH-2     C5H9O1-4OOH-3     C5H9O1-4OOH-5
                    #C5H9O1-5OOH-2     C5H9O1-5OOH-3     C5H9O2-3OOH-1     C5H9O2-3OOH-4     C5H9O2-3OOH-5
                    #C5H9O2-4OOH-1     C5H9O2-4OOH-3     OCH2CHO           C5H9O1-2O-5       CH2OCH2CHO
                    #C2H4OCHO          CH2CH2OCH2CH2CHO  C4H7O1-4          NC5KET12          NC5KET13
                    #NC5KET14          NC5KET15          NC5KET21          NC5KET23          NC5KET24
                    #NC5KET25          NC5KET31          NC5KET32          CHOC4H8O          CHOCH2CH2C2H4O
                    #CHOCH2CH2CH2CH2O  C3H7COCH2O        CH3COC3H6O        CH3COCH2CH2CH2O   C2H5COC2H4O
                    #CH3COCH2C2H4O     CHOCH2C3H6O       C2H5COCH2CH2O     C5H10OH-1         C5H10OH-2
                    #O2C5H10OH-1       O2C5H10OH-2       IC5H12            AC5H11            BC5H11
                    #CC5H11            DC5H11            AC5H10            BC5H10            CC5H10
                    #AC5H9-A2          AC5H9-C           AC5H9-D           CC5H9-A           CC5H9-B
                    #AC5H9O-A2         AC5H9O-C          CC5H9O-B          AC5H11O2H         BC5H11O2H
                    #CC5H11O2H         DC5H11O2H         AC5H11O2          BC5H11O2          CC5H11O2
                    #DC5H11O2          AC5H11O           BC5H11O           CC5H11O           DC5H11O
                    #AC5H10OOH-A       AC5H10OOH-B       AC5H10OOH-C       AC5H10OOH-D       BC5H10OOH-A
                    #BC5H10OOH-C       BC5H10OOH-D       CC5H10OOH-A       CC5H10OOH-B       CC5H10OOH-D
                    #DC5H10OOH-A       DC5H10OOH-B       DC5H10OOH-C       A-AC5H10O         A-BC5H10O
                    #A-CC5H10O         A-DC5H10O         B-CC5H10O         B-DC5H10O         C-DC5H10O
                    #AC5H10OOH-AO2     AC5H10OOH-BO2     AC5H10OOH-CO2     AC5H10OOH-DO2     BC5H10OOH-AO2
                    #BC5H10OOH-CO2     BC5H10OOH-DO2     CC5H10OOH-AO2     CC5H10OOH-BO2     CC5H10OOH-DO2
                    #DC5H10OOH-AO2     DC5H10OOH-BO2     DC5H10OOH-CO2     C5H9A-A,BOOH      C5H9A-A,COOH
                    #C5H9A-A,DOOH      C5H9A-B,COOH      C5H9A-B,DOOH      C5H9A-C,DOOH      C5H9B-A,AOOH
                    #C5H9B-A,COOH      C5H9B-A,DOOH      C5H9B-C,DOOH      C5H9C-A,AOOH      C5H9C-A,BOOH
                    #C5H9C-A,DOOH      C5H9C-B,DOOH      C5H9D-A,AOOH      C5H9D-A,BOOH      C5H9D-A,COOH
                    #C5H9D-B,COOH      C5H9A-AOOH        C5H9A-COOH        C5H9A-DOOH        C5H9B-AOOH
                    #C5H9B-DOOH        C5H9C-AOOH        C5H9C-BOOH        C5H9OA-AOOH-B     C5H9OA-AOOH-C
                    #C5H9OA-AOOH-D     C5H9OA-BOOH-A     C5H9OA-BOOH-C     C5H9OA-BOOH-D     C5H9OA-COOH-A
                    #C5H9OA-COOH-B     C5H9OA-COOH-D     C5H9OA-DOOH-A     C5H9OA-DOOH-B     C5H9OA-DOOH-C
                    #C5H9OB-COOH-A     C5H9OB-COOH-D     C5H9OB-DOOH-A     C5H9OB-DOOH-C     C5H9OC-DOOH-A
                    #C5H9OC-DOOH-B     C4H7O2-1          AC3H5OCH2         CH3COCH2OCH2CH2   IC5KETAA
                    #IC5KETAB          IC5KETAC          IC5KETAD          IC5KETCA          IC5KETCB
                    #IC5KETCD          IC5KETDA          IC5KETDB          IC5KETDC          C2H5CH(CH2O)CHO
                    #C3H5O(CH3)CHO     C2H4OCH(CH3)CHO   CH2OC3H6CHO       CH3COCH(CH3)CH2O  CH3CO(CH3)C2H3O
                    #CH2OCOCH(CH3)CH3  CHOC3H6CH2O       CHOCH2(CH3)C2H3O  CHOCHOCH(CH3)CH3  AC5H10OH
                    #BC5H10OH          CC5H10OH          AO2C5H10OH        BO2C5H10OH        CO2C5H10OH
                    #IC3H5COCH3        IC3H5COCH2        AC3H4COCH3        NEOC5H12          NEOC5H11
                    #NEOC5H11O2H       NEOC5H11O2        NEOC5H11O         NEOC5H10OOH       NEO-C5H10O
                    #NEOC5H10OOH-O2    NEOC5H9Q2         NEOC5KET          NEOC5KETOX        NEOC5KEJOL
                    #NEOC5H9O-OOH      SC3H5OOH          C5H10OOH1-2O      C5H10OOH1-3O      C5H10OOH1-4O
                    #C5H10OOH1-5O      C5H10OOH2-1O      C5H10OOH2-3O      C5H10OOH2-4O      C5H10OOH2-5O
                    #C5H10OOH3-1O      C5H10OOH3-2O      NC6H14            C6H13-1           C6H13-2
                    #C6H13-3           C6H12-1           C6H12-2           C6H12-3           C6H111-3
                    #C6H111-4          C6H111-5          C6H111-6          C6H112-4          C6H112-5
                    #C6H112-6          C6H113-1          C6H12OH-1J2       C6H12OH-2J1       C6H12OH-2J3
                    #C6H12OH-3J2       C6H12OH-3J4       C6H12OH-1O2-2     C6H12OH-2O2-1     C6H12OH-2O2-3
                    #C6H12OH-3O2-2     C6H12OH-3O2-4     C6H11-1D3O        C6H11-1D4O        C6H11-1D5O
                    #C6H11-1D6O        C6H11-2D1O        C6H11-2D4O        C6H11-2D5O        C6H11-2D6O
                    #C6H11-3D1O        C6H11-3D2O        C6H13OOH-1        C6H13OOH-2        C6H13OOH-3
                    #C6H13O2-1         C6H13O2-2         C6H13O2-3         C6H13O-1          C6H13O-2
                    #C6H13O-3          C6H12OOH1-2       C6H12OOH1-3       C6H12OOH1-4       C6H12OOH1-5
                    #C6H12OOH2-1       C6H12OOH2-3       C6H12OOH2-4       C6H12OOH2-5       C6H12OOH2-6
                    #C6H12OOH3-1       C6H12OOH3-2       C6H12OOH3-4       C6H12OOH3-5       C6H12OOH3-6
                    #C6H12O1-2         C6H12O1-3         C6H12O1-4         C6H12O1-5         C6H12O2-3
                    #C6H12O2-4         C6H12O2-5         C6H12O3-4         C6H12OOH1-2O2     C6H12OOH1-3O2
                    #C6H12OOH1-4O2     C6H12OOH1-5O2     C6H12OOH2-1O2     C6H12OOH2-3O2     C6H12OOH2-4O2
                    #C6H12OOH2-5O2     C6H12OOH2-6O2     C6H12OOH3-1O2     C6H12OOH3-2O2     C6H12OOH3-4O2
                    #C6H12OOH3-5O2     C6H12OOH3-6O2     C6H11-1D3OOH      C6H11-1D4OOH      C6H11-1D5OOH
                    #C6H11-1D6OOH      C6H11-2D1OOH      C6H11-2D4OOH      C6H11-2D5OOH      C6H11-2D6OOH
                    #C6H11-3D1OOH      C6H11-3D2OOH      C6H11Q12-3        C6H11Q12-4        C6H11Q12-5
                    #C6H11Q12-6        C6H11Q13-2        C6H11Q13-4        C6H11Q13-5        C6H11Q13-6
                    #C6H11Q14-2        C6H11Q14-3        C6H11Q14-5        C6H11Q14-6        C6H11Q15-2
                    #C6H11Q15-3        C6H11Q15-4        C6H11Q15-6        C6H11Q23-1        C6H11Q23-4
                    #C6H11Q23-5        C6H11Q23-6        C6H11Q24-1        C6H11Q24-3        C6H11Q24-5
                    #C6H11Q24-6        C6H11Q25-1        C6H11Q25-3        C6H11Q34-1        C6H11Q34-2
                    #C6KET12           C6KET13           C6KET14           C6KET15           C6KET21
                    #C6KET23           C6KET24           C6KET25           C6KET26           C6KET31
                    #C6KET32           C6KET34           C6KET35           C6KET36           C5Y2
                    #C5Y2-1J           C5Y2-5J           C5Y3-1J           C5H91-1           C5AL-4D
                    #C5AL-2D           C6H11O12-3OOH     C6H11O12-4OOH     C6H11O12-5OOH     C6H11O12-6OOH
                    #C6H11O13-2OOH     C6H11O13-4OOH     C6H11O13-5OOH     C6H11O13-6OOH     C6H11O14-2OOH
                    #C6H11O14-3OOH     C6H11O14-5OOH     C6H11O14-6OOH     C6H11O15-2OOH     C6H11O15-3OOH
                    #C6H11O15-4OOH     C6H11O15-6OOH     C6H11O23-1OOH     C6H11O23-4OOH     C6H11O23-5OOH
                    #C6H11O23-6OOH     C6H11O24-1OOH     C6H11O24-3OOH     C6H11O24-5OOH     C6H11O24-6OOH
                    #C6H11O25-1OOH     C6H11O25-3OOH     C6H11O34-1OOH     C6H11O34-2OOH     PC4H9CHO
                    #PC4H8CHO-1        PC4H8CHO-2        PC4H8CHO-3        PC4H8CHO-4        C4H7O12-4
                    #C4H7O13-4         C4H7O23-1         C5H9O12-5         C5H9O13-5         C5H9O14-5
                    #C5H9O23-1         C5H9O23-5         C5H9O24-1         C3H6CHO-3O        AC3H5CHO
                    #PC4H9CO           C6KET12O          C6KET13O          C6KET14O          C6KET15O
                    #C6KET21O          C6KET23O          C6KET24O          C6KET25O          C6KET26O
                    #C6KET31O          C6KET32O          C6KET34O          C6KET35O          C6KET36O
                    #C6H10D13          C6H10D24          C6H12OOH1-2O      C6H12OOH1-3O      C6H12OOH1-4O
                    #C6H12OOH1-5O      C6H12OOH2-1O      C6H12OOH2-3O      C6H12OOH2-4O      C6H12OOH2-5O
                    #C6H12OOH2-6O      C6H12OOH3-1O      C6H12OOH3-2O      C6H12OOH3-4O      C6H12OOH3-5O
                    #C6H12OOH3-6O      CH3COOH           C2H5COOH          I3C6              I3C6-1
                    #I3C6-2            I3C6-3            I3C6-4            I3C6D1            I3C6D2
                    #I3C6D3            I3C6D1-3          I3C6D1-4          I3C6D1-5          I3C6D1-6
                    #I3C6D2-5          I3C6D2-6          I3C6D3-1          I3C6D3-2          I3C6OH1-2J
                    #I3C6OH2-1J        I3C6OH2-3J        I3C6OH3-2J        I3C6OH3-4J        I3C6OH4-3J
                    #I3C6OH1-2O2       I3C6OH2-1O2       I3C6OH2-3O2       I3C6OH3-2O2       I3C6OH3-4O2
                    #I3C6OH4-3O2       I3C6D1-3O         I3C6D1-4O         I3C6D1-5O         I3C6D1-6O
                    #I3C6D2-5O         I3C6D2-6O         I3C6D3-1O         I3C6D3-2O         I3C6-1OOH
                    #I3C6-1O2          I3C6-2OOH         I3C6-2O2          I3C6-3OOH         I3C6-3O2
                    #I3C6-4OOH         I3C6-4O2          I3C6-1O           I3C6-2O           I3C6-3O
                    #I3C6-4O           I3C6OOH1-2        I3C6OOH1-3        I3C6OOH1-4        I3C6OOH1-5
                    #I3C6OOH1-6        I3C6OOH2-1        I3C6OOH2-3        I3C6OOH2-4        I3C6OOH2-5
                    #I3C6OOH2-6        I3C6OOH3-1        I3C6OOH3-2        I3C6OOH3-4        I3C6OOH4-1
                    #I3C6OOH4-2        I3C6OOH4-3        I3C6O1-2          I3C6O1-3          I3C6O1-4
                    #I3C6O1-5          I3C6O1-6          I3C6O2-3          I3C6O2-4          I3C6O2-5
                    #I3C6O3-4          I3C6OOH1-2O2      I3C6OOH1-3O2      I3C6OOH1-4O2      I3C6OOH1-5O2
                    #I3C6OOH1-6O2      I3C6OOH2-1O2      I3C6OOH2-3O2      I3C6OOH2-4O2      I3C6OOH2-5O2
                    #I3C6OOH2-6O2      I3C6OOH3-1O2      I3C6OOH3-2O2      I3C6OOH3-4O2      I3C6OOH4-1O2
                    #I3C6OOH4-2O2      I3C6OOH4-3O2      I3C6KET1-2        I3C6KET1-3        I3C6KET1-4
                    #I3C6KET1-5        I3C6KET1-6        I3C6KET2-1        I3C6KET2-3        I3C6KET2-4
                    #I3C6KET2-5        I3C6KET2-6        I3C6KET4-1        I3C6KET4-2        I3C6KET4-3
                    #I3C6D1-3OOH       I3C6D1-4OOH       I3C6D1-5OOH       I3C6D1-6OOH       I3C6D2-1OOH
                    #I3C6D2-4OOH       I3C6D2-5OOH       I3C6D2-6OOH       I3C6D3-1OOH       I3C6D3-2OOH
                    #I3C6Q12-3         I3C6Q12-4         I3C6Q12-5         I3C6Q12-6         I3C6Q13-2
                    #I3C6Q13-4         I3C6Q13-5         I3C6Q13-6         I3C6Q14-2         I3C6Q14-3
                    #I3C6Q14-5         I3C6Q14-6         I3C6Q15-2         I3C6Q15-3         I3C6Q15-4
                    #I3C6Q15-6         I3C6Q16-2         I3C6Q16-3         I3C6Q16-4         I3C6Q23-1
                    #I3C6Q23-4         I3C6Q23-5         I3C6Q23-6         I3C6Q24-1         I3C6Q24-3
                    #I3C6Q24-5         I3C6Q24-6         I3C6Q25-1         I3C6Q25-3         I3C6Q25-4
                    #I3C6Q34-1         I3C6Q34-2         I3C6O12-3OOH      I3C6O12-4OOH      I3C6O12-5OOH
                    #I3C6O12-6OOH      I3C6O13-2OOH      I3C6O13-4OOH      I3C6O13-5OOH      I3C6O13-6OOH
                    #I3C6O14-2OOH      I3C6O14-3OOH      I3C6O14-5OOH      I3C6O14-6OOH      I3C6O15-2OOH
                    #I3C6O15-3OOH      I3C6O15-4OOH      I3C6O15-6OOH      I3C6O16-2OOH      I3C6O16-3OOH
                    #I3C6O16-4OOH      I3C6O23-1OOH      I3C6O23-4OOH      I3C6O23-5OOH      I3C6O23-6OOH
                    #I3C6O24-1OOH      I3C6O24-3OOH      I3C6O24-5OOH      I3C6O24-6OOH      I3C6O25-1OOH
                    #I3C6O25-3OOH      I3C6O25-4OOH      I3C6O34-1OOH      I3C6O34-2OOH      IC4H8CHO-I
                    #IC5Y3-1J          C5H9O2-1          IC5O34-1J         C5Y2-3J           C5H9O15-3
                    #AC5H9O-D          IC5D1Y3           SC4H9CHO          SC4H9CO           SC4H8CHO-1
                    #SC4H8CHO-2        SC4H8CHO-3        SC4H8CHO-4        C5Y3              C5Y3-2J
                    #I3C6KET1-2O       I3C6KET1-3O       I3C6KET1-4O       I3C6KET1-5O       I3C6KET1-6O
                    #I3C6KET2-1O       I3C6KET2-3O       I3C6KET2-4O       I3C6KET2-5O       I3C6KET2-6O
                    #I3C6KET4-1O       I3C6KET4-2O       I3C6KET4-3O\""",
        #reactions='all',
        #initial_state=state(temperature=300.0, pressure=OneAtm))

##-------------------------------------------------------------------------------
## Species data
##-------------------------------------------------------------------------------

#species(name='H',
        #atoms='H:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 2.50000000E+00,  0.00000000E+00,  0.00000000E+00,
                    #0.00000000E+00,  0.00000000E+00,  2.54736600E+04,
                    #-4.46682850E-01]),
                #NASA([1000.00, 6000.00],
                    #[ 2.50000000E+00,  0.00000000E+00,  0.00000000E+00,
                    #0.00000000E+00,  0.00000000E+00,  2.54736600E+04,
                    #-4.46682850E-01])),
        #note='L6/94')

#species(name='H2',
        #atoms='H:2',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 2.34433112E+00,  7.98052075E-03, -1.94781510E-05,
                    #2.01572094E-08, -7.37611761E-12, -9.17935173E+02,
                    #6.83010238E-01]),
                #NASA([1000.00, 6000.00],
                    #[ 2.93286575E+00,  8.26608026E-04, -1.46402364E-07,
                    #1.54100414E-11, -6.88804800E-16, -8.13065581E+02,
                    #-1.02432865E+00])),
        #note='TPIS78')

#species(name='O',
        #atoms='O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 3.16826710E+00, -3.27931884E-03,  6.64306396E-06,
                    #-6.12806624E-09,  2.11265971E-12,  2.91222592E+04,
                    #2.05193346E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 2.54363697E+00, -2.73162486E-05, -4.19029520E-09,
                    #4.95481845E-12, -4.79553694E-16,  2.92260120E+04,
                    #4.92229457E+00])),
        #note='L1/90')

#species(name='O2',
        #atoms='O:2',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 3.78245636E+00, -2.99673416E-03,  9.84730201E-06,
                    #-9.68129509E-09,  3.24372837E-12, -1.06394356E+03,
                    #3.65767573E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 3.66096065E+00,  6.56365811E-04, -1.41149627E-07,
                    #2.05797935E-11, -1.29913436E-15, -1.21597718E+03,
                    #3.41536279E+00])),
        #note='RUS89')

#species(name='OH',
        #atoms='O:1 H:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 3.99198424E+00, -2.40106655E-03,  4.61664033E-06,
                    #-3.87916306E-09,  1.36319502E-12,  3.36889836E+03,
                    #-1.03998477E-01]),
                #NASA([1000.00, 6000.00],
                    #[ 2.83853033E+00,  1.10741289E-03, -2.94000209E-07,
                    #4.20698729E-11, -2.42289890E-15,  3.69780808E+03,
                    #5.84494652E+00])),
        #note='IU3/03')

#species(name='OH*',
        #atoms='O:1 H:1',
        #thermo=(NASA([300.00, 1000.00],
                    #[ 3.63726600E+00,  1.85091000E-04, -1.67616460E-06,
                    #2.38720200E-09, -8.43144200E-13,  5.00213000E+04,
                    #1.35886050E+00]),
                #NASA([1000.00, 5000.00],
                    #[ 2.88273000E+00,  1.01397430E-03, -2.27687700E-07,
                    #2.17468300E-11, -5.12630500E-16,  5.02650000E+04,
                    #5.59571200E+00])),
        #note='121286')

#species(name='H2O',
        #atoms='H:2 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 4.19863520E+00, -2.03640170E-03,  6.52034160E-06,
                    #-5.48792690E-09,  1.77196800E-12, -3.02937260E+04,
                    #-8.49009010E-01]),
                #NASA([1000.00, 6000.00],
                    #[ 2.67703890E+00,  2.97318160E-03, -7.73768890E-07,
                    #9.44335140E-11, -4.26899910E-15, -2.98858940E+04,
                    #6.88255000E+00])),
        #note='L5/89')

#species(name='N2',
        #atoms='N:2',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 3.53100528E+00, -1.23660988E-04, -5.02999433E-07,
                    #2.43530612E-09, -1.40881235E-12, -1.04697628E+03,
                    #2.96747038E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 2.95257637E+00,  1.39690040E-03, -4.92631603E-07,
                    #7.86010195E-11, -4.60755204E-15, -9.23948688E+02,
                    #5.87188762E+00])),
        #note='G8/02')

#species(name='HO2',
        #atoms='H:1 O:2',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 4.30179807E+00, -4.74912097E-03,  2.11582905E-05,
                    #-2.42763914E-08,  9.29225225E-12,  2.64018485E+02,
                    #3.71666220E+00]),
                #NASA([1000.00, 5000.00],
                    #[ 4.17228741E+00,  1.88117627E-03, -3.46277286E-07,
                    #1.94657549E-11,  1.76256905E-16,  3.10206839E+01,
                    #2.95767672E+00])),
        #note='T1/09')

#species(name='H2O2',
        #atoms='H:2 O:2',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 4.31515149E+00, -8.47390622E-04,  1.76404323E-05,
                    #-2.26762944E-08,  9.08950158E-12, -1.77067437E+04,
                    #3.27373319E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 4.57977305E+00,  4.05326003E-03, -1.29844730E-06,
                    #1.98211400E-10, -1.13968792E-14, -1.80071775E+04,
                    #6.64970694E-01])),
        #note='T8/03')

#species(name='AR',
        #atoms='Ar:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 2.50000000E+00,  0.00000000E+00,  0.00000000E+00,
                    #0.00000000E+00,  0.00000000E+00, -7.45375000E+02,
                    #4.37967491E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 2.50000000E+00,  0.00000000E+00,  0.00000000E+00,
                    #0.00000000E+00,  0.00000000E+00, -7.45375000E+02,
                    #4.37967491E+00])),
        #note='G5/97')

#species(name='CO',
        #atoms='C:1 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 3.57953350E+00, -6.10353690E-04,  1.01681430E-06,
                    #9.07005860E-10, -9.04424490E-13, -1.43440860E+04,
                    #3.50840930E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 3.04848590E+00,  1.35172810E-03, -4.85794050E-07,
                    #7.88536440E-11, -4.69807460E-15, -1.42661170E+04,
                    #6.01709770E+00])),
        #note='RUS79')

#species(name='CO2',
        #atoms='C:1 O:2',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 2.35681300E+00,  8.98412990E-03, -7.12206320E-06,
                    #2.45730080E-09, -1.42885480E-13, -4.83719710E+04,
                    #9.90090350E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 4.63651110E+00,  2.74145690E-03, -9.95897590E-07,
                    #1.60386660E-10, -9.16198570E-15, -4.90249040E+04,
                    #-1.93489550E+00])),
        #note='L7/88')

#species(name='CH2O',
        #atoms='H:2 C:1 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 4.79372312E+00, -9.90833322E-03,  3.73219990E-05,
                    #-3.79285237E-08,  1.31772641E-11, -1.43791953E+04,
                    #6.02798058E-01]),
                #NASA([1000.00, 6000.00],
                    #[ 3.16952665E+00,  6.19320560E-03, -2.25056366E-06,
                    #3.65975660E-10, -2.20149458E-14, -1.45486831E+04,
                    #6.04207898E+00])),
        #note='T5/11')

#species(name='HCO',
        #atoms='C:1 H:1 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 4.23754610E+00, -3.32075257E-03,  1.40030264E-05,
                    #-1.34239995E-08,  4.37416208E-12,  3.87241185E+03,
                    #3.30834869E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 3.92001542E+00,  2.52279324E-03, -6.71004164E-07,
                    #1.05615948E-10, -7.43798261E-15,  3.65342928E+03,
                    #3.58077056E+00])),
        #note='T5/03')

#species(name='HO2CHO',
        #atoms='C:1 H:2 O:3',
        #thermo=(NASA([300.00, 1378.00],
                    #[ 2.42464726E+00,  2.19706380E-02, -1.68705546E-05,
                    #6.25612194E-09, -9.11645843E-13, -3.54828006E+04,
                    #1.75027796E+01]),
                #NASA([1378.00, 5000.00],
                    #[ 9.87503878E+00,  4.64663708E-03, -1.67230522E-06,
                    #2.68624413E-10, -1.59595232E-14, -3.80502496E+04,
                    #-2.24939155E+01])),
        #note='6/26/95THERM')

#species(name='HCOH',
        #atoms='C:1 H:2 O:1',
        #thermo=(NASA([300.00, 1398.00],
                    #[-2.82157421E+00,  3.57331702E-02, -3.80861580E-05,
                    #1.86205951E-08, -3.45957838E-12,  1.12956672E+04,
                    #3.48487757E+01]),
                #NASA([1398.00, 5000.00],
                    #[ 9.18749272E+00,  1.52011152E-03, -6.27603516E-07,
                    #1.09727989E-10, -6.89655128E-15,  7.81364593E+03,
                    #-2.73434214E+01])),
        #note='MAR94')

#species(name='O2CHO',
        #atoms='C:1 H:1 O:3',
        #thermo=(NASA([300.00, 1368.00],
                    #[ 3.96059309E+00,  1.06002279E-02, -5.25713351E-06,
                    #1.01716726E-09, -2.87487602E-14, -1.73599383E+04,
                    #1.17807483E+01]),
                #NASA([1368.00, 5000.00],
                    #[ 7.24075139E+00,  4.63312951E-03, -1.63693995E-06,
                    #2.59706693E-10, -1.52964699E-14, -1.87027618E+04,
                    #-6.49547212E+00])),
        #note='6/26/95THERM')

#species(name='HOCHO',
        #atoms='H:2 C:1 O:2',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 3.89836160E+00, -3.55877950E-03,  3.55205380E-05,
                    #-4.38499590E-08,  1.71077690E-11, -4.67706090E+04,
                    #7.34953970E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 4.61383160E+00,  6.44963640E-03, -2.29082510E-06,
                    #3.67160470E-10, -2.18736750E-14, -4.75148500E+04,
                    #8.47883830E-01])),
        #note='L8/88')

#species(name='OCHO',
        #atoms='C:1 O:2 H:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 4.68825921E+00, -4.14871834E-03,  2.55066010E-05,
                    #-2.84473900E-08,  1.04422559E-11, -1.69867041E+04,
                    #4.28426480E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 4.14394211E+00,  5.59738818E-03, -1.99794019E-06,
                    #3.16179193E-10, -1.85614483E-14, -1.72459887E+04,
                    #5.07778617E+00])),
        #note='ATCT/A')

#species(name='HOCH2O2H',
        #atoms='C:1 H:4 O:3',
        #thermo=(NASA([300.00, 1398.00],
                    #[ 5.35189713E-01,  3.73266553E-02, -3.15299511E-05,
                    #1.30352583E-08, -2.11473264E-12, -3.86609415E+04,
                    #2.71776082E+01]),
                #NASA([1398.00, 5000.00],
                    #[ 1.24531886E+01,  7.18221110E-03, -2.47029548E-06,
                    #3.85611737E-10, -2.24774193E-14, -4.24862928E+04,
                    #-3.58745197E+01])),
        #note='9/1/12')

#species(name='HOCH2O2',
        #atoms='C:1 H:3 O:3',
        #thermo=(NASA([300.00, 1377.00],
                    #[ 2.82068616E+00,  2.47857094E-02, -1.66186399E-05,
                    #4.79633095E-09, -4.28087766E-13, -2.22077036E+04,
                    #1.70599803E+01]),
                #NASA([1377.00, 5000.00],
                    #[ 1.16406115E+01,  5.72826040E-03, -2.05362036E-06,
                    #3.29070695E-10, -1.95188360E-14, -2.53505769E+04,
                    #-3.07332064E+01])),
        #note='9/1/12')

#species(name='OCH2O2H',
        #atoms='C:1 H:3 O:3',
        #thermo=(NASA([300.00, 1418.00],
                    #[ 4.46349361E-01,  3.63049606E-02, -3.26130978E-05,
                    #1.37050551E-08, -2.20872791E-12, -1.41972598E+04,
                    #2.72960376E+01]),
                #NASA([1418.00, 5000.00],
                    #[ 1.29622491E+01,  4.21948855E-03, -1.54275194E-06,
                    #2.50413077E-10, -1.49855537E-14, -1.81326406E+04,
                    #-3.87016356E+01])),
        #note='7/21/14THERM')

#species(name='HOCH2O',
        #atoms='C:1 H:3 O:2',
        #thermo=(NASA([300.00, 1452.00],
                    #[ 4.11183145E+00,  7.53850697E-03,  3.77337370E-06,
                    #-5.38746005E-09,  1.45615887E-12, -2.28023001E+04,
                    #7.46807254E+00]),
                #NASA([1452.00, 5000.00],
                    #[ 6.39521515E+00,  7.43673043E-03, -2.50422354E-06,
                    #3.84879712E-10, -2.21778689E-14, -2.41108840E+04,
                    #-6.63865583E+00])),
        #note='2/16/99THERM')

#species(name='CH3OH',
        #atoms='C:1 H:4 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 5.65851051E+00, -1.62983419E-02,  6.91938156E-05,
                    #-7.58372926E-08,  2.80427550E-11, -2.56119736E+04,
                    #-8.97330508E-01]),
                #NASA([1000.00, 6000.00],
                    #[ 3.52726795E+00,  1.03178783E-02, -3.62892944E-06,
                    #5.77448016E-10, -3.42182632E-14, -2.60028834E+04,
                    #5.16758693E+00])),
        #note='T06/02')

#species(name='CH2OH',
        #atoms='C:1 H:3 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 4.47834367E+00, -1.35070310E-03,  2.78484980E-05,
                    #-3.64869060E-08,  1.47907450E-11, -3.50072890E+03,
                    #3.30913500E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 5.09314370E+00,  5.94761260E-03, -2.06497460E-06,
                    #3.23008173E-10, -1.88125902E-14, -4.03409640E+03,
                    #-1.84691493E+00])),
        #note='IU2/03')

#species(name='CH3O',
        #atoms='C:1 H:3 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 3.71180502E+00, -2.80463306E-03,  3.76550971E-05,
                    #-4.73072089E-08,  1.86588420E-11,  1.29569760E+03,
                    #6.57240864E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 4.75779238E+00,  7.44142474E-03, -2.69705176E-06,
                    #4.38090504E-10, -2.63537098E-14,  3.78111940E+02,
                    #-1.96680028E+00])),
        #note='IU1/03')

#species(name='CH3O2H',
        #atoms='C:1 H:4 O:2',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 2.90540897E+00,  1.74994735E-02,  5.28243630E-06,
                    #-2.52827275E-08,  1.34368212E-11, -1.68894632E+04,
                    #1.13741987E+01]),
                #NASA([1000.00, 6000.00],
                    #[ 7.76538058E+00,  8.61499712E-03, -2.98006935E-06,
                    #4.68638071E-10, -2.75339255E-14, -1.82979984E+04,
                    #-1.43992663E+01])),
        #note='A7/05')

#species(name='CH3O2',
        #atoms='H:3 C:1 O:2',
        #thermo=(NASA([300.00, 1374.00],
                    #[ 1.97339205E+00,  1.53542340E-02, -6.37314891E-06,
                    #3.19930565E-10,  2.82193915E-13,  2.54278835E+02,
                    #1.69194215E+01]),
                #NASA([1374.00, 5000.00],
                    #[ 6.47970487E+00,  7.44401080E-03, -2.52348555E-06,
                    #3.89577296E-10, -2.25182399E-14, -1.56285441E+03,
                    #-8.19477074E+00])))

#species(name='CH4',
        #atoms='C:1 H:4',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 5.14911468E+00, -1.36622009E-02,  4.91453921E-05,
                    #-4.84246767E-08,  1.66603441E-11, -1.02465983E+04,
                    #-4.63848842E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 1.65326226E+00,  1.00263099E-02, -3.31661238E-06,
                    #5.36483138E-10, -3.14696758E-14, -1.00095936E+04,
                    #9.90506283E+00])),
        #note='G8/99')

#species(name='CH3',
        #atoms='C:1 H:3',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 3.65717970E+00,  2.12659790E-03,  5.45838830E-06,
                    #-6.61810030E-09,  2.46570740E-12,  1.64227160E+04,
                    #1.67353540E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 2.97812060E+00,  5.79785200E-03, -1.97558000E-06,
                    #3.07297900E-10, -1.79174160E-14,  1.65095130E+04,
                    #4.72247990E+00])),
        #note='IU0702')

#species(name='CH2',
        #atoms='C:1 H:2',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 3.71757846E+00,  1.27391260E-03,  2.17347251E-06,
                    #-3.48858500E-09,  1.65208866E-12,  4.58723866E+04,
                    #1.75297945E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 3.14631886E+00,  3.03671259E-03, -9.96474439E-07,
                    #1.50483580E-10, -8.57335515E-15,  4.60412605E+04,
                    #4.72341711E+00])),
        #note='IU3/03')

#species(name='CH2(S)',
        #atoms='C:1 H:2',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 4.19331325E+00, -2.33105184E-03,  8.15676451E-06,
                    #-6.62985981E-09,  1.93233199E-12,  5.03662246E+04,
                    #-7.46734310E-01]),
                #NASA([1000.00, 6000.00],
                    #[ 3.13501686E+00,  2.89593926E-03, -8.16668090E-07,
                    #1.13572697E-10, -6.36262835E-15,  5.05040504E+04,
                    #4.06030621E+00])),
        #note='IU6/03')

#species(name='C',
        #atoms='C:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 2.55423950E+00, -3.21537720E-04,  7.33792230E-07,
                    #-7.32234870E-10,  2.66521440E-13,  8.54426810E+04,
                    #4.53130850E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 2.60558300E+00, -1.95934340E-04,  1.06737220E-07,
                    #-1.64239400E-11,  8.18705800E-16,  8.54117420E+04,
                    #4.19238680E+00])),
        #note='L7/88')

#species(name='CH',
        #atoms='C:1 H:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 3.48975830E+00,  3.24321600E-04, -1.68997510E-06,
                    #3.16284200E-09, -1.40618030E-12,  7.06126460E+04,
                    #2.08428410E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 2.52093690E+00,  1.76536390E-03, -4.61476600E-07,
                    #5.92896750E-11, -3.34745010E-15,  7.09467690E+04,
                    #7.40518290E+00])),
        #note='IU3/03')

#species(name='CH*',
        #atoms='C:1 H:1',
        #thermo=(NASA([300.00, 1000.00],
                    #[ 3.20020200E+00,  2.07287500E-03, -5.13443100E-06,
                    #5.73389000E-09, -1.95553300E-12,  1.03937140E+05,
                    #3.33158700E+00]),
                #NASA([1000.00, 5000.00],
                    #[ 2.19622300E+00,  2.34038100E-03, -7.05820100E-07,
                    #9.00758200E-11, -3.85504000E-15,  1.04195590E+05,
                    #9.17837300E+00])),
        #note='073003')

#species(name='C2H6',
        #atoms='C:2 H:6',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 4.29142572E+00, -5.50154901E-03,  5.99438458E-05,
                    #-7.08466469E-08,  2.68685836E-11, -1.15222056E+04,
                    #2.66678994E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 4.04666411E+00,  1.53538802E-02, -5.47039485E-06,
                    #8.77826544E-10, -5.23167531E-14, -1.24473499E+04,
                    #-9.68698313E-01])),
        #note='G8/88')

#species(name='C2H5',
        #atoms='C:2 H:5',
        #thermo=(NASA([300.00, 1387.00],
                    #[ 1.32730217E+00,  1.76656753E-02, -6.14926558E-06,
                    #-3.01143466E-10,  4.38617775E-13,  1.34284028E+04,
                    #1.71789216E+01]),
                #NASA([1387.00, 5000.00],
                    #[ 5.88784390E+00,  1.03076793E-02, -3.46844396E-06,
                    #5.32499257E-10, -3.06512651E-14,  1.15065499E+04,
                    #-8.49651771E+00])),
        #note='8/4/4THERM')

#species(name='C2H4',
        #atoms='C:2 H:4',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 3.95920063E+00, -7.57051373E-03,  5.70989993E-05,
                    #-6.91588352E-08,  2.69884190E-11,  5.08977598E+03,
                    #4.09730213E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 3.99182724E+00,  1.04833908E-02, -3.71721342E-06,
                    #5.94628366E-10, -3.53630386E-14,  4.26865851E+03,
                    #-2.69081762E-01])),
        #note='G1/00')

#species(name='C2H3',
        #atoms='C:2 H:3',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 3.36377642E+00,  2.65765722E-04,  2.79620704E-05,
                    #-3.72986942E-08,  1.51590176E-11,  3.44749589E+04,
                    #7.91510092E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 4.15026763E+00,  7.54021341E-03, -2.62997847E-06,
                    #4.15974048E-10, -2.45407509E-14,  3.38566380E+04,
                    #1.72812235E+00])),
        #note='ATCT/A')

#species(name='C2H2',
        #atoms='C:2 H:2',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 8.08679682E-01,  2.33615762E-02, -3.55172234E-05,
                    #2.80152958E-08, -8.50075165E-12,  2.64289808E+04,
                    #1.39396761E+01]),
                #NASA([1000.00, 6000.00],
                    #[ 4.65878489E+00,  4.88396667E-03, -1.60828888E-06,
                    #2.46974544E-10, -1.38605959E-14,  2.57594042E+04,
                    #-3.99838194E+00])),
        #note='G1/91')

#species(name='C2H',
        #atoms='C:2 H:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 2.89867676E+00,  1.32988489E-02, -2.80733327E-05,
                    #2.89484755E-08, -1.07502351E-11,  6.70616050E+04,
                    #6.18547632E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 3.66270248E+00,  3.82492252E-03, -1.36632500E-06,
                    #2.13455040E-10, -1.23216848E-14,  6.71683790E+04,
                    #3.92205792E+00])),
        #note='T5/10')

#species(name='CH3CHO',
        #atoms='C:2 H:4 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 4.72945950E+00, -3.19328580E-03,  4.75349210E-05,
                    #-5.74586110E-08,  2.19311120E-11, -2.15728780E+04,
                    #4.10301590E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 5.40411080E+00,  1.17230590E-02, -4.22631370E-06,
                    #6.83724510E-10, -4.09848630E-14, -2.25931220E+04,
                    #-3.48079170E+00])),
        #note='L8/88')

#species(name='C2H3OH',
        #atoms='C:2 H:4 O:1',
        #thermo=(NASA([300.00, 1410.00],
                    #[-1.27972260E-01,  3.38506073E-02, -3.30644935E-05,
                    #1.64858739E-08, -3.19935455E-12, -1.59914544E+04,
                    #2.30438601E+01]),
                #NASA([1410.00, 5000.00],
                    #[ 8.32598158E+00,  8.03387281E-03, -2.63928405E-06,
                    #3.98410726E-10, -2.26551155E-14, -1.83221436E+04,
                    #-2.02080305E+01])),
        #note='2/3/9THERM')

#species(name='C2H2OH',
        #atoms='H:3 C:2 O:1',
        #thermo=(NASA([300.00, 1401.00],
                    #[ 6.41642616E-01,  2.61903633E-02, -2.30385370E-05,
                    #1.02804704E-08, -1.81971416E-12,  1.48276951E+04,
                    #2.06750999E+01]),
                #NASA([1401.00, 5000.00],
                    #[ 8.20268447E+00,  5.92989165E-03, -1.99194448E-06,
                    #3.05794341E-10, -1.76114732E-14,  1.24881328E+04,
                    #-1.89670436E+01])))

#species(name='CH3CO',
        #atoms='C:2 H:3 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 4.03587050E+00,  8.77294870E-04,  3.07100100E-05,
                    #-3.92475650E-08,  1.52968690E-11, -2.68207380E+03,
                    #7.86176820E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 5.31371650E+00,  9.17377930E-03, -3.32203860E-06,
                    #5.39474560E-10, -3.24523680E-14, -3.64504140E+03,
                    #-1.67575580E+00])),
        #note='IU2/03')

#species(name='CH2CHO',
        #atoms='C:2 H:3 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 2.79502600E+00,  1.01099472E-02,  1.61750645E-05,
                    #-3.10303145E-08,  1.39436139E-11,  1.62944975E+02,
                    #1.23646657E+01]),
                #NASA([1000.00, 6000.00],
                    #[ 6.53928338E+00,  7.80238629E-03, -2.76413612E-06,
                    #4.42098906E-10, -2.62954290E-14, -1.18858659E+03,
                    #-8.72091393E+00])),
        #note='T03/10')

#species(name='O2CH2CHO',
        #atoms='C:2 H:3 O:3',
        #thermo=(NASA([300.00, 1393.00],
                    #[-1.29465843E+00,  4.44936393E-02, -4.26577074E-05,
                    #2.07391950E-08, -3.96828771E-12, -1.18275628E+04,
                    #3.60778797E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 1.11807543E+01,  9.14479256E-03, -3.15089833E-06,
                    #4.91944238E-10, -2.86639180E-14, -1.55790331E+04,
                    #-2.87892740E+01])),
        #note='BOZ_03')

#species(name='HO2CH2CO',
        #atoms='C:2 H:3 O:3',
        #thermo=(NASA([300.00, 1386.00],
                    #[ 2.22681686E+00,  3.56781380E-02, -3.26401909E-05,
                    #1.47651988E-08, -2.64794380E-12, -1.18735095E+04,
                    #1.91581197E+01]),
                #NASA([1386.00, 5000.00],
                    #[ 1.04146322E+01,  1.12680116E-02, -5.17494839E-06,
                    #1.00333285E-09, -6.68165911E-14, -1.40955672E+04,
                    #-2.27894400E+01])),
        #note='BOZ_03')

#species(name='CH2CO',
        #atoms='H:2 C:2 O:1',
        #thermo=(NASA([300.00, 1000.00],
                    #[ 1.81422511E+00,  1.99008590E-02, -2.21416008E-05,
                    #1.45028521E-08, -3.98877068E-12, -7.05394926E+03,
                    #1.36079359E+01]),
                #NASA([1000.00, 5000.00],
                    #[ 5.35869367E+00,  6.95641586E-03, -2.64802637E-06,
                    #4.65067592E-10, -3.08641820E-14, -7.90294013E+03,
                    #-3.98525731E+00])))

#species(name='HCCO',
        #atoms='H:1 C:2 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 1.87607969E+00,  2.21205418E-02, -3.58869325E-05,
                    #3.05402541E-08, -1.01281069E-11,  2.01633840E+04,
                    #1.36968290E+01]),
                #NASA([1000.00, 6000.00],
                    #[ 5.91479333E+00,  3.71408730E-03, -1.30137010E-06,
                    #2.06473345E-10, -1.21476759E-14,  1.93596301E+04,
                    #-5.50567269E+00])),
        #note='T4/09')

#species(name='HCCOH',
        #atoms='C:2 H:2 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 2.05541154E+00,  2.52003372E-02, -3.80821654E-05,
                    #3.09890632E-08, -9.89799902E-12,  9.76872113E+03,
                    #1.22271534E+01]),
                #NASA([1000.00, 6000.00],
                    #[ 6.37509678E+00,  5.49429011E-03, -1.88136576E-06,
                    #2.93803536E-10, -1.71771901E-14,  8.93277676E+03,
                    #-8.24498007E+00])),
        #note='T12/09')

#species(name='CH3CO3H',
        #atoms='C:2 H:4 O:3',
        #thermo=(NASA([300.00, 1391.00],
                    #[ 2.24135876E+00,  3.37963514E-02, -2.53887482E-05,
                    #9.67583587E-09, -1.49266157E-12, -4.24677831E+04,
                    #1.70668133E+01]),
                #NASA([1391.00, 5000.00],
                    #[ 1.25060485E+01,  9.47789695E-03, -3.30402246E-06,
                    #5.19630793E-10, -3.04233568E-14, -4.59856703E+04,
                    #-3.79195947E+01])),
        #note='6/26/95THERM')

#species(name='CH3CO3',
        #atoms='C:2 H:3 O:3',
        #thermo=(NASA([300.00, 1391.00],
                    #[ 3.60373432E+00,  2.70080341E-02, -2.08293438E-05,
                    #8.50541104E-09, -1.43846110E-12, -2.34205171E+04,
                    #1.12014914E+01]),
                #NASA([1391.00, 5000.00],
                    #[ 1.12522498E+01,  8.33652672E-03, -2.89014530E-06,
                    #4.52781734E-10, -2.64354456E-14, -2.60238584E+04,
                    #-2.96370457E+01])),
        #note='4/3/0THERM')

#species(name='CH3CO2',
        #atoms='C:2 H:3 O:2',
        #thermo=(NASA([300.00, 1395.00],
                    #[ 1.37440768E+00,  2.49115604E-02, -1.74308894E-05,
                    #6.24799508E-09, -9.09516835E-13, -2.72330150E+04,
                    #1.81405454E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 8.54059736E+00,  8.32951214E-03, -2.84722010E-06,
                    #4.41927196E-10, -2.56373394E-14, -2.97290678E+04,
                    #-2.03883545E+01])),
        #note='2/14/95THERM')

#species(name='C2H5OH',
        #atoms='C:2 H:6 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 4.85869570E+00, -3.74017260E-03,  6.95553780E-05,
                    #-8.86547960E-08,  3.51688350E-11, -2.99961320E+04,
                    #4.80185450E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 6.56243650E+00,  1.52042220E-02, -5.38967950E-06,
                    #8.62250110E-10, -5.12897870E-14, -3.15256210E+04,
                    #-9.47302020E+00])),
        #note='L8/88')

#species(name='C2H5O',
        #atoms='C:2 H:5 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 4.30742680E+00,  6.41472050E-03,  3.11397140E-05,
                    #-4.33140830E-08,  1.72761840E-11, -3.40275240E+03,
                    #5.90258370E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 6.68899820E+00,  1.31256760E-02, -4.70388400E-06,
                    #7.58585520E-10, -4.54133060E-14, -4.74578320E+03,
                    #-9.69837550E+00])),
        #note='IU2/03')

#species(name='PC2H4OH',
        #atoms='C:2 H:5 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 4.47893092E+00,  7.59782301E-03,  2.81794908E-05,
                    #-4.26953487E-08,  1.78878934E-11, -4.71446256E+03,
                    #6.38921206E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 7.02824536E+00,  1.20037746E-02, -4.21306455E-06,
                    #6.69471213E-10, -3.96371893E-14, -5.92493321E+03,
                    #-9.40355948E+00])),
        #note='T12/01')

#species(name='SC2H4OH',
        #atoms='C:2 H:5 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 4.22283250E+00,  5.12174798E-03,  3.48386522E-05,
                    #-4.91943637E-08,  2.01183723E-11, -8.20503939E+03,
                    #8.01675700E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 6.35842302E+00,  1.24356276E-02, -4.33096839E-06,
                    #6.84530381E-10, -4.03713238E-14, -9.37900432E+03,
                    #-6.05106112E+00])),
        #note='T10/04')

#species(name='O2C2H4OH',
        #atoms='C:2 H:5 O:3',
        #thermo=(NASA([300.00, 1506.00],
                    #[ 7.04009800E+00,  1.59564166E-02,  2.21097416E-06,
                    #-7.05197355E-09,  2.08266026E-12, -2.24524432E+04,
                    #-1.75361758E+00]),
                #NASA([1506.00, 5000.00],
                    #[ 1.27503881E+01,  1.11514325E-02, -3.83473891E-06,
                    #5.98155829E-10, -3.48372108E-14, -2.52770876E+04,
                    #-3.54317608E+01])),
        #note='9/1/12THERM')

#species(name='C2H5O2H',
        #atoms='C:2 H:6 O:2',
        #thermo=(NASA([300.00, 1390.00],
                    #[ 1.83755328E+00,  3.38053586E-02, -2.37548140E-05,
                    #9.31974865E-09, -1.58003428E-12, -2.15814086E+04,
                    #1.80977584E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 1.04823538E+01,  1.34779879E-02, -4.62179078E-06,
                    #7.18618519E-10, -4.17307436E-14, -2.46578171E+04,
                    #-2.84294243E+01])),
        #note='9/1/12')

#species(name='C2H5O2',
        #atoms='C:2 H:5 O:2',
        #thermo=(NASA([300.00, 1389.00],
                    #[ 3.90351912E+00,  2.22599212E-02, -1.01610079E-05,
                    #1.71709751E-09,  1.88166738E-14, -5.09654081E+03,
                    #8.98722750E+00]),
                #NASA([1389.00, 5000.00],
                    #[ 9.50282570E+00,  1.20429839E-02, -4.09491581E-06,
                    #6.33049241E-10, -3.66133788E-14, -7.37069391E+03,
                    #-2.21717130E+01])),
        #note='9/1/12')

#species(name='C2H4O2H',
        #atoms='C:2 H:5 O:2',
        #thermo=(NASA([300.00, 1389.00],
                    #[ 2.75788364E+00,  2.88271987E-02, -2.08302264E-05,
                    #8.47401397E-09, -1.48617610E-12,  3.00153893E+03,
                    #1.59921711E+01]),
                #NASA([1389.00, 5000.00],
                    #[ 1.00590614E+01,  1.13378955E-02, -3.89403387E-06,
                    #6.06090687E-10, -3.52212353E-14,  4.24048653E+02,
                    #-2.32086536E+01])),
        #note='9/1/12')

#species(name='C2H4O1-2',
        #atoms='C:2 H:4 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 3.75905320E+00, -9.44121800E-03,  8.03097210E-05,
                    #-1.00807880E-07,  4.00399210E-11, -7.56081430E+03,
                    #7.84974750E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 5.48876410E+00,  1.20461900E-02, -4.33369310E-06,
                    #7.00283110E-10, -4.19490880E-14, -9.18042510E+03,
                    #-7.07996050E+00])),
        #note='L8/88')

#species(name='C2H3O1-2',
        #atoms='C:2 H:3 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 3.58349017E+00, -6.02275805E-03,  6.32426867E-05,
                    #-8.18540707E-08,  3.30444505E-11,  1.85681353E+04,
                    #9.59725926E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 5.60158035E+00,  9.17613962E-03, -3.28028902E-06,
                    #5.27903888E-10, -3.15362241E-14,  1.71446252E+04,
                    #-5.47228512E+00])),
        #note='A1/05')

#species(name='CH3COCH3',
        #atoms='C:3 H:6 O:1',
        #thermo=(NASA([300.00, 1381.00],
                    #[ 1.14767529E+00,  3.00731373E-02, -1.37643759E-05,
                    #1.89028808E-09,  2.02896771E-13, -2.76607074E+04,
                    #2.06153026E+01]),
                #NASA([1381.00, 5000.00],
                    #[ 9.68478208E+00,  1.45488533E-02, -4.98681534E-06,
                    #7.75343101E-10, -4.50297606E-14, -3.10899721E+04,
                    #-2.68395746E+01])),
        #note='2/14/13THERM')

#species(name='CH3COCH2',
        #atoms='C:3 H:5 O:1',
        #thermo=(NASA([300.00, 1387.00],
                    #[ 1.13381826E+00,  3.25095045E-02, -2.10424651E-05,
                    #6.64421151E-09, -8.12618901E-13, -6.04868361E+03,
                    #2.17158655E+01]),
                #NASA([1387.00, 5000.00],
                    #[ 1.09524298E+01,  1.11458668E-02, -3.86262877E-06,
                    #6.05088857E-10, -3.53293362E-14, -9.60833727E+03,
                    #-3.15622776E+01])),
        #note='2/14/13THERM')

#species(name='CH3COCH2O2',
        #atoms='C:3 H:5 O:3',
        #thermo=(NASA([300.00, 1397.00],
                    #[ 1.19378141E+00,  4.98027161E-02, -4.17999508E-05,
                    #1.74527607E-08, -2.88198761E-12, -1.93244224E+04,
                    #2.67877493E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 1.65756401E+01,  1.06465489E-02, -3.61368681E-06,
                    #5.59053564E-10, -3.23832271E-14, -2.42541401E+04,
                    #-5.45304899E+01])),
        #note='2/14/13THERM')

#species(name='C3KET21',
        #atoms='C:3 H:6 O:3',
        #thermo=(NASA([300.00, 1394.00],
                    #[-8.74352903E-01,  6.12501498E-02, -5.51474542E-05,
                    #2.48491014E-08, -4.42613472E-12, -3.58060819E+04,
                    #3.59306224E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 1.75768076E+01,  1.20311704E-02, -4.11633942E-06,
                    #6.40149366E-10, -3.72127562E-14, -4.15502347E+04,
                    #-6.09097100E+01])),
        #note='2/14/13THERM')

#species(name='C2H3CHO',
        #atoms='C:3 H:4 O:1',
        #thermo=(NASA([300.00, 1398.00],
                    #[ 7.33844455E-01,  3.17482671E-02, -2.29599468E-05,
                    #8.42104232E-09, -1.23613478E-12, -9.38473548E+03,
                    #2.10308851E+01]),
                #NASA([1398.00, 5000.00],
                    #[ 9.99155394E+00,  9.82348001E-03, -3.31203088E-06,
                    #5.09524422E-10, -2.93821890E-14, -1.25303509E+04,
                    #-2.85168883E+01])),
        #note='KPS12')

#species(name='C2H3CO',
        #atoms='C:3 H:3 O:1',
        #thermo=(NASA([300.00, 1395.00],
                    #[ 1.65335195E+00,  2.57402596E-02, -1.89009911E-05,
                    #7.29174972E-09, -1.16083226E-12,  1.02020654E+04,
                    #1.78705872E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 8.86032735E+00,  8.48985205E-03, -2.90350080E-06,
                    #4.50763986E-10, -2.61524281E-14,  7.73489171E+03,
                    #-2.06978792E+01])),
        #note='KPS12')

#species(name='C2H5CHO',
        #atoms='C:3 H:6 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 4.24529681E+00,  6.68296706E-03,  4.93337933E-05,
                    #-6.71986124E-08,  2.67262347E-11, -2.41473007E+04,
                    #6.90738560E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 7.44085690E+00,  1.77301764E-02, -6.34081568E-06,
                    #1.02040803E-09, -6.09461714E-14, -2.60055814E+04,
                    #-1.44195446E+01])),
        #note='T05/10')

#species(name='C2H5CO',
        #atoms='C:3 H:5 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 6.25722402E+00, -9.17612184E-03,  7.61190493E-05,
                    #-9.05514997E-08,  3.46198215E-11, -5.91616484E+03,
                    #2.23330599E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 6.52325448E+00,  1.54211952E-02, -5.50898157E-06,
                    #8.85889862E-10, -5.28846399E-14, -7.19631634E+03,
                    #-5.19862218E+00])),
        #note='A10/04')

#species(name='CH3OCH3',
        #atoms='C:2 H:6 O:1',
        #thermo=(NASA([300.00, 1999.00],
                    #[ 2.05597390E+00,  2.07019456E-02, -5.00382376E-06,
                    #-1.62279885E-09,  6.84330155E-13, -2.35494445E+04,
                    #1.45029944E+01]),
                #NASA([1999.00, 5000.00],
                    #[ 6.03232751E+00,  1.56155270E-02, -5.50761030E-06,
                    #8.75666140E-10, -5.17180562E-14, -2.52690354E+04,
                    #-8.25885183E+00])),
        #note='2/11/14THERM')

#species(name='CH3OCH2',
        #atoms='C:2 H:5 O:1',
        #thermo=(NASA([300.00, 1395.00],
                    #[ 1.58874948E+00,  2.24414123E-02, -1.19434933E-05,
                    #3.37160213E-09, -4.15077249E-13, -1.37208255E+03,
                    #1.87548958E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 6.62621974E+00,  1.22219496E-02, -4.12416696E-06,
                    #6.34127512E-10, -3.65317390E-14, -3.33965890E+03,
                    #-8.95305753E+00])),
        #note='2/11/14THERM')

#species(name='CH3OCH2O2',
        #atoms='C:2 H:5 O:3',
        #thermo=(NASA([300.00, 1441.00],
                    #[ 3.39930541E+00,  3.09460407E-02, -1.92548181E-05,
                    #5.76033887E-09, -6.16081571E-13, -2.04433218E+04,
                    #1.39429608E+01]),
                #NASA([1441.00, 5000.00],
                    #[ 1.19179361E+01,  1.19412867E-02, -3.93526185E-06,
                    #5.95756132E-10, -3.39597705E-14, -2.34231833E+04,
                    #-3.20096863E+01])),
        #note='2/12/14THERM')

#species(name='CH2OCH2O2H',
        #atoms='C:2 H:5 O:3',
        #thermo=(NASA([300.00, 1418.00],
                    #[ 1.62245477E-01,  4.76101093E-02, -4.52046954E-05,
                    #2.18379311E-08, -4.11295947E-12, -1.46498100E+04,
                    #2.98253164E+01]),
                #NASA([1418.00, 5000.00],
                    #[ 1.23892901E+01,  1.11758961E-02, -3.59249095E-06,
                    #5.34196366E-10, -3.00536541E-14, -1.80551598E+04,
                    #-3.29576862E+01])),
        #note='2/12/14THERM')

#species(name='CH3OCH2O2H',
        #atoms='C:2 H:6 O:3',
        #thermo=(NASA([300.00, 1404.00],
                    #[ 1.05786981E+00,  4.36787095E-02, -3.46383899E-05,
                    #1.44808830E-08, -2.46100643E-12, -3.68851076E+04,
                    #2.43391936E+01]),
                #NASA([1404.00, 5000.00],
                    #[ 1.28159161E+01,  1.34818095E-02, -4.50397729E-06,
                    #6.88229286E-10, -3.94883680E-14, -4.06745921E+04,
                    #-3.78047802E+01])),
        #note='2/12/14THERM')

#species(name='CH3OCH2O',
        #atoms='C:2 H:5 O:2',
        #thermo=(NASA([300.00, 1523.00],
                    #[ 5.63414373E+00,  8.92830283E-03,  1.37225633E-05,
                    #-1.40497059E-08,  3.54625624E-12, -2.22825214E+04,
                    #1.93588846E+00]),
                #NASA([1523.00, 5000.00],
                    #[ 9.81288609E+00,  1.21313106E-02, -4.30285768E-06,
                    #6.84443177E-10, -4.03862658E-14, -2.50760742E+04,
                    #-2.51866352E+01])),
        #note='5/15/14THERM')

#species(name='O2CH2OCH2O2H',
        #atoms='C:2 H:5 O:5',
        #thermo=(NASA([300.00, 1433.00],
                    #[ 2.39977678E+00,  5.39881943E-02, -4.87969524E-05,
                    #2.19792134E-08, -3.86106979E-12, -3.37824638E+04,
                    #2.30683371E+01]),
                #NASA([1433.00, 5000.00],
                    #[ 1.77378326E+01,  1.13589899E-02, -3.67382539E-06,
                    #5.49255712E-10, -3.10405899E-14, -3.82903058E+04,
                    #-5.66609932E+01])),
        #note='2/12/14ERM')

#species(name='HO2CH2OCHO',
        #atoms='C:2 H:4 O:4',
        #thermo=(NASA([300.00, 1386.00],
                    #[ 1.21909586E+00,  4.28858235E-02, -3.17634222E-05,
                    #1.11542676E-08, -1.49753153E-12, -5.79287926E+04,
                    #2.49759193E+01]),
                #NASA([1386.00, 5000.00],
                    #[ 1.57136128E+01,  9.64430166E-03, -3.44136025E-06,
                    #5.49722196E-10, -3.25360322E-14, -6.29409094E+04,
                    #-5.29505242E+01])),
        #note='2/12/14THERM')

#species(name='OCH2OCHO',
        #atoms='C:2 H:3 O:3',
        #thermo=(NASA([300.00, 1523.00],
                    #[ 1.89539692E+00,  2.74118545E-02, -1.36476090E-05,
                    #1.26325603E-09,  5.17970476E-13, -4.27879440E+04,
                    #2.02333278E+01]),
                #NASA([1523.00, 5000.00],
                    #[ 1.24013200E+01,  7.83738243E-03, -2.82992688E-06,
                    #4.55558739E-10, -2.71061389E-14, -4.68453470E+04,
                    #-3.78084549E+01])),
        #note='5/29/14THERM')

#species(name='HOCH2OCO',
        #atoms='C:2 H:3 O:3',
        #thermo=(NASA([300.00, 1443.00],
                    #[ 5.95255071E+00,  8.42196282E-03,  1.36741678E-05,
                    #-1.46786275E-08,  3.84143533E-12, -4.44470269E+04,
                    #2.85657217E+00]),
                #NASA([1443.00, 5000.00],
                    #[ 1.11498410E+01,  9.34736520E-03, -3.35541548E-06,
                    #5.38037115E-10, -3.19260183E-14, -4.75012119E+04,
                    #-2.95983867E+01])),
        #note='5/29/14THERM')

#species(name='CH3OCHO',
        #atoms='C:2 H:4 O:2',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 5.96757028E+00, -9.38085425E-03,  7.07648417E-05,
                    #-8.29932227E-08,  3.13522917E-11, -4.55713267E+04,
                    #7.50341113E-01]),
                #NASA([1000.00, 6000.00],
                    #[ 6.33360880E+00,  1.34851485E-02, -4.84305805E-06,
                    #7.81719241E-10, -4.67917447E-14, -4.68316521E+04,
                    #-6.91542601E+00])),
        #note='T6/08')

#species(name='CH3OCO',
        #atoms='C:2 H:3 O:2',
        #thermo=(NASA([300.00, 1601.00],
                    #[ 4.16215406E+00,  1.38037511E-02, -3.08486109E-07,
                    #-4.56430814E-09,  1.46909632E-12, -2.10130301E+04,
                    #8.64301044E+00]),
                #NASA([1601.00, 5000.00],
                    #[ 9.73659803E+00,  7.42432713E-03, -2.65641779E-06,
                    #4.25031143E-10, -2.51824924E-14, -2.36015721E+04,
                    #-2.36353471E+01])),
        #note='5/8/3THERM')

#species(name='CH2OCHO',
        #atoms='C:2 H:3 O:2',
        #thermo=(NASA([300.00, 1442.00],
                    #[ 2.31031671E+00,  1.80474065E-02, -2.71519637E-06,
                    #-4.60918579E-09,  1.70037078E-12, -2.02910878E+04,
                    #1.71549722E+01]),
                #NASA([1442.00, 5000.00],
                    #[ 1.00960096E+01,  7.19887066E-03, -2.59813465E-06,
                    #4.18110812E-10, -2.48723387E-14, -2.36389018E+04,
                    #-2.71144175E+01])),
        #note='4/15/8THERM')

#species(name='HE',
        #atoms='He:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 2.50000000E+00,  0.00000000E+00,  0.00000000E+00,
                    #0.00000000E+00,  0.00000000E+00, -7.45375000E+02,
                    #9.28723974E-01]),
                #NASA([1000.00, 6000.00],
                    #[ 2.50000000E+00,  0.00000000E+00,  0.00000000E+00,
                    #0.00000000E+00,  0.00000000E+00, -7.45375000E+02,
                    #9.28723974E-01])),
        #note='G5/97')

#species(name='C3H8',
        #atoms='C:3 H:8',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 4.21093013E+00,  1.70886504E-03,  7.06530164E-05,
                    #-9.20060565E-08,  3.64618453E-11, -1.43810883E+04,
                    #5.61004451E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 6.66919760E+00,  2.06108751E-02, -7.36512349E-06,
                    #1.18434262E-09, -7.06914630E-14, -1.62754066E+04,
                    #-1.31943379E+01])),
        #note='G2/00')

#species(name='IC3H7',
        #atoms='C:3 H:7',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 5.47421257E+00, -8.42536682E-03,  8.04607759E-05,
                    #-9.49287824E-08,  3.59830971E-11,  9.04939013E+03,
                    #3.40542323E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 5.30597255E+00,  1.89854588E-02, -6.74315384E-06,
                    #1.07993730E-09, -6.42785036E-14,  7.78748910E+03,
                    #-2.23233935E+00])),
        #note='A5/05')

#species(name='NC3H7',
        #atoms='C:3 H:7',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 4.08211458E+00,  5.23240341E-03,  5.13554466E-05,
                    #-6.99343598E-08,  2.81819493E-11,  1.04074558E+04,
                    #8.39534919E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 6.49636579E+00,  1.77337992E-02, -6.24898046E-06,
                    #9.95389495E-10, -5.90199770E-14,  8.85973885E+03,
                    #-8.56389710E+00])),
        #note='A5/05')

#species(name='C3H6',
        #atoms='C:3 H:6',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 3.83464468E+00,  3.29078952E-03,  5.05228001E-05,
                    #-6.66251176E-08,  2.63707473E-11,  7.88717123E+02,
                    #7.53408013E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 6.03870234E+00,  1.62963931E-02, -5.82130800E-06,
                    #9.35936829E-10, -5.58603143E-14, -7.41715057E+02,
                    #-8.43825992E+00])),
        #note='G2/00')

#species(name='C3H5-A',
        #atoms='C:3 H:5',
        #thermo=(NASA([300.00, 1000.00],
                    #[ 1.36318350E+00,  1.98138210E-02,  1.24970600E-05,
                    #-3.33555550E-08,  1.58465710E-11,  1.92456290E+04,
                    #1.71732140E+01]),
                #NASA([1000.00, 3000.00],
                    #[ 6.50078770E+00,  1.43247310E-02, -5.67816320E-06,
                    #1.10808010E-09, -9.03638870E-14,  1.74824490E+04,
                    #-1.12430500E+01])),
        #note='PD5/98')

#species(name='C3H5-S',
        #atoms='C:3 H:5',
        #thermo=(NASA([300.00, 1000.00],
                    #[ 9.13729310E-01,  2.64323430E-02, -1.17589500E-05,
                    #-2.30356780E-09,  2.77154880E-12,  3.09168670E+04,
                    #1.99892690E+01]),
                #NASA([1000.00, 3000.00],
                    #[ 5.37252810E+00,  1.57805090E-02, -5.99228500E-06,
                    #9.30896640E-10, -3.65509660E-14,  2.96147600E+04,
                    #-3.41864780E+00])),
        #note='PD5/98')

#species(name='C3H5-T',
        #atoms='C:3 H:5',
        #thermo=(NASA([300.00, 1000.00],
                    #[ 1.73292090E+00,  2.23946200E-02, -5.14906110E-06,
                    #-6.75964660E-09,  3.82532110E-12,  2.90404980E+04,
                    #1.65688780E+01]),
                #NASA([1000.00, 3000.00],
                    #[ 5.42555280E+00,  1.55110720E-02, -5.66783500E-06,
                    #7.92243880E-10, -1.68780340E-14,  2.78430270E+04,
                    #-3.35271840E+00])),
        #note='PD5/98')

#species(name='C3H4-P',
        #atoms='H:4 C:3',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 2.68038690E+00,  1.57996510E-02,  2.50705960E-06,
                    #-1.36576230E-08,  6.61542850E-12,  2.08023740E+04,
                    #9.87693510E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 6.02524000E+00,  1.13365420E-02, -4.02233910E-06,
                    #6.43760630E-10, -3.82996350E-14,  1.96209420E+04,
                    #-8.60437850E+00])),
        #note='T2/90')

#species(name='C3H4-A',
        #atoms='C:3 H:4',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 2.61304450E+00,  1.21225750E-02,  1.85398800E-05,
                    #-3.45251490E-08,  1.53350790E-11,  2.15415670E+04,
                    #1.02261390E+01]),
                #NASA([1000.00, 6000.00],
                    #[ 6.31687220E+00,  1.11337280E-02, -3.96293780E-06,
                    #6.35642380E-10, -3.78755400E-14,  2.01174950E+04,
                    #-1.09957660E+01])),
        #note='L8/89')

#species(name='C3H3',
        #atoms='C:3 H:3',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 1.35110873E+00,  3.27411291E-02, -4.73827407E-05,
                    #3.76310220E-08, -1.18541128E-11,  4.07679941E+04,
                    #1.52058598E+01]),
                #NASA([1000.00, 6000.00],
                    #[ 7.14221719E+00,  7.61902211E-03, -2.67460030E-06,
                    #4.24914904E-10, -2.51475443E-14,  3.95709594E+04,
                    #-1.25848690E+01])),
        #note='T7/11')

#species(name='C3H5O',
        #atoms='C:3 H:5 O:1',
        #thermo=(NASA([300.00, 1402.00],
                    #[ 8.24068673E-01,  3.46749909E-02, -2.51786795E-05,
                    #9.56781953E-09, -1.48085302E-12,  1.04203725E+04,
                    #2.28283070E+01]),
                #NASA([1402.00, 5000.00],
                    #[ 1.02638186E+01,  1.17609932E-02, -3.89837957E-06,
                    #5.92650815E-10, -3.38867417E-14,  7.25938472E+03,
                    #-2.75108651E+01])),
        #note='KPS12')

#species(name='C3H6OOH1-2',
        #atoms='C:3 H:7 O:2',
        #thermo=(NASA([300.00, 1387.00],
                    #[ 2.83631132E+00,  3.88229894E-02, -2.47944364E-05,
                    #7.85644898E-09, -9.58634300E-13, -1.26002528E+03,
                    #1.72549973E+01]),
                #NASA([1387.00, 5000.00],
                    #[ 1.38088686E+01,  1.43845650E-02, -4.74440961E-06,
                    #7.19308280E-10, -4.10654123E-14, -5.14352831E+03,
                    #-4.20210765E+01])),
        #note='9/1/12')

#species(name='C3H6OOH1-3',
        #atoms='C:3 H:7 O:2',
        #thermo=(NASA([300.00, 1401.00],
                    #[ 1.74271107E+00,  4.53733504E-02, -3.57580373E-05,
                    #1.48540053E-08, -2.49981756E-12,  2.32580844E+02,
                    #2.20973041E+01]),
                #NASA([1401.00, 5000.00],
                    #[ 1.39130757E+01,  1.40218463E-02, -4.55921149E-06,
                    #6.84182417E-10, -3.87696213E-14, -3.65650518E+03,
                    #-4.21532559E+01])),
        #note='9/1/12')

#species(name='C3H6OOH2-1',
        #atoms='C:3 H:7 O:2',
        #thermo=(NASA([300.00, 1393.00],
                    #[ 2.38465746E+00,  4.42928555E-02, -3.50977087E-05,
                    #1.53695144E-08, -2.81167824E-12, -1.80979612E+03,
                    #1.69923285E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 1.36645362E+01,  1.54329764E-02, -5.29285952E-06,
                    #8.23001262E-10, -4.77931121E-14, -5.58295862E+03,
                    #-4.28758364E+01])),
        #note='9/1/12')

#species(name='C3H6OOH1-2O2',
        #atoms='C:3 H:7 O:4',
        #thermo=(NASA([300.00, 1404.00],
                    #[ 3.99085043E+00,  5.31865338E-02, -4.28597948E-05,
                    #1.77187019E-08, -2.92768695E-12, -2.02143526E+04,
                    #1.34150719E+01]),
                #NASA([1404.00, 5000.00],
                    #[ 1.91044980E+01,  1.44076100E-02, -4.72127814E-06,
                    #7.12631642E-10, -4.05578490E-14, -2.50270510E+04,
                    #-6.63747978E+01])),
        #note='9/1/12')

#species(name='C3H6OOH1-3O2',
        #atoms='C:3 H:7 O:4',
        #thermo=(NASA([300.00, 1416.00],
                    #[ 5.56933350E+00,  4.68523421E-02, -3.58917784E-05,
                    #1.43314525E-08, -2.29776083E-12, -1.86065694E+04,
                    #7.18655005E+00]),
                #NASA([1416.00, 5000.00],
                    #[ 1.81661664E+01,  1.47644887E-02, -4.74842743E-06,
                    #7.06972467E-10, -3.98305587E-14, -2.26256376E+04,
                    #-5.93719393E+01])),
        #note='9/1/12')

#species(name='C3H6OOH2-1O2',
        #atoms='C:3 H:7 O:4',
        #thermo=(NASA([300.00, 1404.00],
                    #[ 3.99085043E+00,  5.31865338E-02, -4.28597948E-05,
                    #1.77187019E-08, -2.92768695E-12, -2.02143526E+04,
                    #1.34150719E+01]),
                #NASA([1404.00, 5000.00],
                    #[ 1.91044980E+01,  1.44076100E-02, -4.72127814E-06,
                    #7.12631642E-10, -4.05578490E-14, -2.50270510E+04,
                    #-6.63747978E+01])),
        #note='9/1/12')

#species(name='NC3H7O2H',
        #atoms='C:3 H:8 O:2',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 9.13626931E-01,  4.89566251E-02, -3.51850148E-05,
                    #1.30546796E-08, -1.95621807E-12, -2.43818102E+04,
                    #2.40611651E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 1.46883407E+01,  1.60212483E-02, -5.27531221E-06,
                    #7.98877253E-10, -4.55711618E-14, -2.90318646E+04,
                    #-4.95529916E+01])),
        #note='9/1/12')

#species(name='IC3H7O2H',
        #atoms='C:3 H:8 O:2',
        #thermo=(NASA([300.00, 1393.00],
                    #[ 1.47731961E+00,  4.92440994E-02, -3.79913951E-05,
                    #1.61970607E-08, -2.90125059E-12, -2.63956602E+04,
                    #1.90105627E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 1.40855241E+01,  1.75805904E-02, -6.02432523E-06,
                    #9.36214739E-10, -5.43465682E-14, -3.06618857E+04,
                    #-4.81005453E+01])),
        #note='9/1/12')

#species(name='NC3H7O2',
        #atoms='C:3 H:7 O:2',
        #thermo=(NASA([300.00, 1500.00],
                    #[ 3.13948561E+00,  3.66456006E-02, -2.03703424E-05,
                    #4.70670546E-09, -1.99573208E-13, -7.91997431E+03,
                    #1.42198984E+01]),
                #NASA([1500.00, 5000.00],
                    #[ 1.36927178E+01,  1.46801304E-02, -4.79793976E-06,
                    #7.22768575E-10, -4.10732852E-14, -1.17561732E+04,
                    #-4.32476343E+01])),
        #note='9/1/12')

#species(name='IC3H7O2',
        #atoms='C:3 H:7 O:2',
        #thermo=(NASA([300.00, 1395.00],
                    #[ 3.42416052E+00,  3.84321103E-02, -2.50950244E-05,
                    #8.80154103E-09, -1.31154059E-12, -9.90211798E+03,
                    #1.03892668E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 1.33991937E+01,  1.58889581E-02, -5.40791162E-06,
                    #8.36616758E-10, -4.84119240E-14, -1.34911944E+04,
                    #-4.35329618E+01])),
        #note='9/1/12')

#species(name='NC3H7O',
        #atoms='C:3 H:7 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 4.21934640E+00,  7.38556641E-03,  6.02825529E-05,
                    #-8.38680247E-08,  3.39623435E-11, -6.23491852E+03,
                    #8.08139850E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 8.38041157E+00,  1.95206120E-02, -6.97374143E-06,
                    #1.12144919E-09, -6.69467831E-14, -8.48625211E+03,
                    #-1.89916219E+01])),
        #note='T07/10')

#species(name='IC3H7O',
        #atoms='C:3 H:7 O:1',
        #thermo=(NASA([300.00, 1386.00],
                    #[ 1.69886473E+00,  3.63643211E-02, -2.24587302E-05,
                    #7.15111944E-09, -9.33278403E-13, -7.77098005E+03,
                    #1.69680075E+01]),
                #NASA([1386.00, 5000.00],
                    #[ 1.14627518E+01,  1.51011907E-02, -5.07470532E-06,
                    #7.78736627E-10, -4.48206286E-14, -1.13388060E+04,
                    #-3.60564653E+01])),
        #note='9/1/12')

#species(name='C3H6O1-2',
        #atoms='C:3 H:6 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 3.42806676E+00,  6.25176642E-03,  6.13196311E-05,
                    #-8.60387185E-08,  3.51371393E-11, -1.28446646E+04,
                    #1.04244994E+01]),
                #NASA([1000.00, 6000.00],
                    #[ 8.01491079E+00,  1.73919953E-02, -6.26027968E-06,
                    #1.01188256E-09, -6.06239111E-14, -1.51980838E+04,
                    #-1.88279964E+01])),
        #note='A01/05')

#species(name='C3H6O1-3',
        #atoms='C:3 H:6 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 5.15283752E+00, -1.86401716E-02,  1.29980652E-04,
                    #-1.58629974E-07,  6.20668783E-11, -1.13243512E+04,
                    #4.73561224E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 6.80716906E+00,  1.88824545E-02, -6.79082475E-06,
                    #1.09713919E-09, -6.57154952E-14, -1.36547629E+04,
                    #-1.35382154E+01])),
        #note='A11/04')

#species(name='C3KET12',
        #atoms='C:3 H:6 O:3',
        #thermo=(NASA([300.00, 1385.00],
                    #[ 1.03882879E+00,  5.34180080E-02, -4.47684141E-05,
                    #1.94651680E-08, -3.45055244E-12, -3.70308881E+04,
                    #2.56511209E+01]),
                #NASA([1385.00, 5000.00],
                    #[ 1.70187760E+01,  1.32097361E-02, -4.67054741E-06,
                    #7.41411770E-10, -4.36869787E-14, -4.23572589E+04,
                    #-5.92615939E+01])),
        #note='10/17/12')

#species(name='C3KET13',
        #atoms='C:3 H:6 O:3',
        #thermo=(NASA([300.00, 1508.00],
                    #[ 4.74956819E+00,  3.14080991E-02, -6.83838427E-06,
                    #-5.67123901E-09,  2.27686972E-12, -3.51924570E+04,
                    #9.83753744E+00]),
                #NASA([1508.00, 5000.00],
                    #[ 1.73612692E+01,  1.32330813E-02, -4.75332110E-06,
                    #7.62529227E-10, -4.52613717E-14, -4.06248060E+04,
                    #-6.17768199E+01])),
        #note='10/17/12')

#species(name='C3H51-2,3OOH',
        #atoms='C:3 H:7 O:4',
        #thermo=(NASA([300.00, 1386.00],
                    #[ 2.55619708E+00,  6.13504487E-02, -5.23205391E-05,
                    #2.28208029E-08, -4.02231508E-12, -1.31353414E+04,
                    #2.21043799E+01]),
                #NASA([1386.00, 5000.00],
                    #[ 2.12378169E+01,  1.39519596E-02, -4.94539222E-06,
                    #7.86381389E-10, -4.63925564E-14, -1.92864584E+04,
                    #-7.69636561E+01])),
        #note='8/26/3THRM')

#species(name='C3H52-1,3OOH',
        #atoms='C:3 H:7 O:4',
        #thermo=(NASA([300.00, 1379.00],
                    #[ 4.12253742E+00,  5.19553611E-02, -3.83733727E-05,
                    #1.45851637E-08, -2.29820536E-12, -1.22759164E+04,
                    #1.48367359E+01]),
                #NASA([1379.00, 5000.00],
                    #[ 2.02817964E+01,  1.48155431E-02, -5.25503386E-06,
                    #8.35963453E-10, -4.93308915E-14, -1.80085066E+04,
                    #-7.22688262E+01])),
        #note='8/26/3THRM')

#species(name='C3H5O1-2OOH-3',
        #atoms='C:3 H:6 O:3',
        #thermo=(NASA([300.00, 1432.00],
                    #[-3.25001215E+00,  6.65787151E-02, -6.18859778E-05,
                    #2.84638649E-08, -5.08511634E-12, -2.22371240E+04,
                    #4.30381280E+01]),
                #NASA([1432.00, 5000.00],
                    #[ 1.57042382E+01,  1.30255692E-02, -4.23544254E-06,
                    #6.35555595E-10, -3.60110207E-14, -2.77269333E+04,
                    #-5.51895464E+01])),
        #note='10/13THER')

#species(name='C3H5O1-3OOH-2',
        #atoms='C:3 H:6 O:3',
        #thermo=(NASA([300.00, 1434.00],
                    #[-4.43959178E+00,  7.16532928E-02, -7.15032351E-05,
                    #3.51737842E-08, -6.63682938E-12, -2.27250412E+04,
                    #4.56394038E+01]),
                #NASA([1434.00, 5000.00],
                    #[ 1.44493479E+01,  1.36372560E-02, -4.25836513E-06,
                    #6.20006211E-10, -3.43451580E-14, -2.77372360E+04,
                    #-5.06099103E+01])),
        #note='10/13THER')

#species(name='C3H6OH1-2',
        #atoms='C:3 H:7 O:1',
        #thermo=(NASA([300.00, 1395.00],
                    #[ 5.05207596E-01,  3.63869988E-02, -2.15530901E-05,
                    #6.45584786E-09, -7.71267046E-13, -9.26980840E+03,
                    #2.79804349E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 1.00338281E+01,  1.60227373E-02, -5.41658448E-06,
                    #8.34191172E-10, -4.81215988E-14, -1.27912397E+04,
                    #-2.39034395E+01])),
        #note='9/1/12')

#species(name='CH3CHCO',
        #atoms='C:3 H:4 O:1',
        #thermo=(NASA([300.00, 1400.00],
                    #[ 1.48380119E+00,  3.22203013E-02, -2.70250033E-05,
                    #1.20499164E-08, -2.18365931E-12, -1.15276540E+04,
                    #1.71552068E+01]),
                #NASA([1400.00, 5000.00],
                    #[ 1.00219123E+01,  9.56966300E-03, -3.26221644E-06,
                    #5.05231706E-10, -2.92593257E-14, -1.42482738E+04,
                    #-2.77829973E+01])),
        #note='03/03/95THERM')

#species(name='AC3H5OOH',
        #atoms='C:3 H:6 O:2',
        #thermo=(NASA([298.00, 1000.00],
                    #[ 3.18124993E+00,  4.35233041E-02, -5.16277353E-05,
                    #4.32011427E-08, -1.57714983E-11, -7.63521503E+03,
                    #1.21725683E+01]),
                #NASA([1000.00, 6000.00],
                    #[ 1.20838649E+01,  1.47946591E-02, -5.13212591E-06,
                    #8.07504999E-10, -4.74394983E-14, -1.02184463E+04,
                    #-3.36434791E+01])),
        #note='GOLDSMITH')

#species(name='C2H3OOH',
        #atoms='C:2 H:4 O:2',
        #thermo=(NASA([300.00, 1397.00],
                    #[ 1.35644398E+00,  3.37002447E-02, -2.75988500E-05,
                    #1.14222854E-08, -1.89488886E-12, -5.49996692E+03,
                    #1.98354466E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 1.15749951E+01,  8.09909174E-03, -2.81808668E-06,
                    #4.42697954E-10, -2.58998042E-14, -8.84852664E+03,
                    #-3.43859117E+01])),
        #note='4/18/8THERM')

#species(name='CC3H4',
        #atoms='C:3 H:4',
        #thermo=(NASA([300.00, 1000.00],
                    #[-2.46210470E-02,  2.31972150E-02, -1.84743570E-06,
                    #-1.59275930E-08,  8.68461550E-12,  3.23341370E+04,
                    #2.27297620E+01]),
                #NASA([1000.00, 5000.00],
                    #[ 6.69999310E+00,  1.03573720E-02, -3.45511670E-06,
                    #5.06529490E-10, -2.66822760E-14,  3.01990510E+04,
                    #-1.33787700E+01])),
        #note='T12/81')

#species(name='H2CC',
        #atoms='H:2 C:2',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 3.28154830E+00,  6.97647910E-03, -2.38552440E-06,
                    #-1.21044320E-09,  9.81895450E-13,  4.86217940E+04,
                    #5.92039100E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 4.27803400E+00,  4.75628040E-03, -1.63010090E-06,
                    #2.54628060E-10, -1.48863790E-14,  4.83166880E+04,
                    #6.40237010E-01])),
        #note='L12/89')

#species(name='FULVENE',
        #atoms='C:6 H:6',
        #thermo=(NASA([200.00, 1000.00],
                    #[-7.18131910E-01,  3.79343120E-02,  1.13988370E-05,
                    #-4.13335030E-08,  1.80559270E-11,  2.42238250E+04,
                    #2.78557140E+01]),
                #NASA([1000.00, 5000.00],
                    #[ 1.11035250E+01,  2.06006850E-02, -7.53022240E-06,
                    #1.23886950E-09, -7.54159760E-14,  2.03618430E+04,
                    #-3.66651970E+01])),
        #note='0')

#species(name='CC3H6',
        #atoms='C:3 H:6',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 2.83278674E+00, -5.21028618E-03,  9.29583210E-05,
                    #-1.22753194E-07,  4.99191366E-11,  5.19520048E+03,
                    #1.08306333E+01]),
                #NASA([1000.00, 6000.00],
                    #[ 6.21663437E+00,  1.65393591E-02, -5.90075838E-06,
                    #9.48095199E-10, -5.65661522E-14,  2.95937491E+03,
                    #-1.36041009E+01])))

#species(name='C4H10',
        #atoms='C:4 H:10',
        #thermo=(NASA([300.00, 1389.00],
                    #[-4.02516492E-01,  4.79826446E-02, -2.65875714E-05,
                    #7.07078402E-09, -6.98493372E-13, -1.68794837E+04,
                    #2.63198718E+01]),
                #NASA([1389.00, 5000.00],
                    #[ 1.25231536E+01,  2.18666638E-02, -7.50125881E-06,
                    #1.16671087E-09, -6.77696640E-14, -2.18475689E+04,
                    #-4.46643459E+01])),
        #note='9/1/12')

#species(name='C4H8-1',
        #atoms='C:4 H:8',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 5.13226136E+00,  5.33862838E-03,  6.02928960E-05,
                    #-7.60364685E-08,  2.87324693E-11, -2.16718358E+03,
                    #3.82936810E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 7.86795262E+00,  2.24448843E-02, -8.07705438E-06,
                    #1.30179988E-09, -7.77958472E-14, -4.23853340E+03,
                    #-1.65662549E+01])),
        #note='T05/09')

#species(name='C4H8-2',
        #atoms='C:4 H:8',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 5.57278967E+00,  3.76541017E-03,  6.52226708E-05,
                    #-8.30909522E-08,  3.20311342E-11, -3.60128327E+03,
                    #5.37796708E-01]),
                #NASA([1000.00, 6000.00],
                    #[ 7.89114667E+00,  2.24970532E-02, -8.12143779E-06,
                    #1.31273568E-09, -7.84451632E-14, -5.51643171E+03,
                    #-1.76436027E+01])),
        #note='T5/09')

#species(name='PC4H9',
        #atoms='C:4 H:9',
        #thermo=(NASA([300.00, 1393.00],
                    #[ 3.50939027E-01,  4.47616979E-02, -2.76331489E-05,
                    #9.10558432E-09, -1.28318459E-12,  7.74508072E+03,
                    #2.53924028E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 1.18227273E+01,  1.97584602E-02, -6.73845199E-06,
                    #1.04383408E-09, -6.04576130E-14,  3.50333683E+03,
                    #-3.69846129E+01])),
        #note='9/1/12')

#species(name='SC4H9',
        #atoms='C:4 H:9',
        #thermo=(NASA([300.00, 1388.00],
                    #[ 6.23603962E-01,  3.99289701E-02, -1.89596886E-05,
                    #3.45938210E-09, -2.39235619E-14,  6.17688777E+03,
                    #2.52737918E+01]),
                #NASA([1388.00, 5000.00],
                    #[ 1.11370287E+01,  2.04038703E-02, -6.97571297E-06,
                    #1.08244679E-09, -6.27714790E-14,  1.95591491E+03,
                    #-3.30856128E+01])),
        #note='9/1/12')

#species(name='C4H71-1',
        #atoms='C:4 H:7',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 4.19857522E+00,  1.19616999E-02,  4.23864923E-05,
                    #-6.30299109E-08,  2.59475110E-11,  2.75256555E+04,
                    #8.57181248E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 8.15646382E+00,  1.90308835E-02, -6.73262214E-06,
                    #1.07333098E-09, -6.36886441E-14,  2.55826427E+04,
                    #-1.61428872E+01])),
        #note='T05/04')

#species(name='C4H71-2',
        #atoms='C:4 H:7',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 3.77145965E+00,  1.46544157E-02,  3.70080802E-05,
                    #-5.72714455E-08,  2.36641011E-11,  2.58014506E+04,
                    #9.11906641E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 8.16688868E+00,  1.95680375E-02, -6.95694878E-06,
                    #1.11504166E-09, -6.64079384E-14,  2.37537003E+04,
                    #-1.77041242E+01])),
        #note='T05/04')

#species(name='C4H71-3',
        #atoms='C:4 H:7',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 4.54746808E+00,  4.63771460E-03,  6.61340221E-05,
                    #-8.97456502E-08,  3.61716165E-11,  1.43843217E+04,
                    #7.30313471E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 8.08107449E+00,  1.95526544E-02, -6.93149115E-06,
                    #1.10889183E-09, -6.59584410E-14,  1.22822959E+04,
                    #-1.67137903E+01])),
        #note='T05/04')

#species(name='C4H71-4',
        #atoms='C:4 H:7',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 5.07355313E+00,  5.27619329E-03,  6.23441322E-05,
                    #-8.54203458E-08,  3.45890031E-11,  2.24615054E+04,
                    #5.60318035E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 8.49073768E+00,  1.91056974E-02, -6.74370664E-06,
                    #1.07343267E-09, -6.36251837E-14,  2.04659294E+04,
                    #-1.74555814E+01])),
        #note='T05/04')

#species(name='C4H72-2',
        #atoms='C:4 H:7',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 7.61389036E+00, -9.06922602E-03,  8.28486476E-05,
                    #-9.61203624E-08,  3.59333528E-11,  2.44971584E+04,
                    #-5.90519467E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 7.26612168E+00,  1.99858497E-02, -7.12030976E-06,
                    #1.14276142E-09, -6.81206632E-14,  2.31915554E+04,
                    #-1.09941637E+01])),
        #note='T05/04')

#species(name='C4H6',
        #atoms='C:4 H:6',
        #thermo=(NASA([300.00, 1000.00],
                    #[ 1.12844650E-01,  3.43690220E-02, -1.11073920E-05,
                    #-9.21066600E-09,  6.20651790E-12,  1.18022700E+04,
                    #2.30899960E+01]),
                #NASA([1000.00, 3000.00],
                    #[ 8.86731340E+00,  1.49186700E-02, -3.15487160E-06,
                    #-4.18413300E-10,  1.57612580E-13,  9.13385160E+03,
                    #-2.33281710E+01])),
        #note='H6W/94')

#species(name='PC4H9O2H',
        #atoms='C:4 H:10 O:2',
        #thermo=(NASA([300.00, 1410.00],
                    #[ 1.00788926E-01,  6.35127410E-02, -4.55348612E-05,
                    #1.63427400E-08, -2.29549654E-12, -2.71975592E+04,
                    #2.95249361E+01]),
                #NASA([1410.00, 5000.00],
                    #[ 1.85422067E+01,  1.95532704E-02, -6.40903366E-06,
                    #9.67561826E-10, -5.50744230E-14, -3.33659478E+04,
                    #-6.89574879E+01])),
        #note='9/1/12')

#species(name='SC4H9O2H',
        #atoms='C:4 H:10 O:2',
        #thermo=(NASA([300.00, 1395.00],
                    #[ 4.05944553E-01,  6.48970390E-02, -4.98857846E-05,
                    #2.00008685E-08, -3.25194364E-12, -2.91704448E+04,
                    #2.63739532E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 1.85276559E+01,  1.96998004E-02, -6.48703965E-06,
                    #9.82426308E-10, -5.60434220E-14, -3.51152892E+04,
                    #-6.98322937E+01])),
        #note='9/1/12')

#species(name='PC4H9O2',
        #atoms='C:4 H:9 O:2',
        #thermo=(NASA([300.00, 1663.00],
                    #[ 2.16957521E+00,  5.22744687E-02, -3.32954958E-05,
                    #1.03030945E-08, -1.20770740E-12, -1.07180659E+04,
                    #2.03541558E+01]),
                #NASA([1663.00, 5000.00],
                    #[ 1.58558021E+01,  2.04563545E-02, -6.93230110E-06,
                    #1.07504998E-09, -6.24729238E-14, -1.51651087E+04,
                    #-5.27319441E+01])),
        #note='9/1/12')

#species(name='SC4H9O2',
        #atoms='C:4 H:9 O:2',
        #thermo=(NASA([300.00, 1408.00],
                    #[ 2.68574393E+00,  5.26273779E-02, -3.56094836E-05,
                    #1.22899616E-08, -1.68623273E-12, -1.27229421E+04,
                    #1.62316995E+01]),
                #NASA([1408.00, 5000.00],
                    #[ 1.71094567E+01,  1.90236957E-02, -6.29508198E-06,
                    #9.55825976E-10, -5.46024443E-14, -1.76672012E+04,
                    #-6.11399832E+01])),
        #note='9/1/12')

#species(name='PC4H9O',
        #atoms='C:4 H:9 O:1',
        #thermo=(NASA([300.00, 1403.00],
                    #[-4.99964924E-01,  5.37157310E-02, -3.44426650E-05,
                    #1.08145957E-08, -1.29600044E-12, -9.11644218E+03,
                    #3.09183423E+01]),
                #NASA([1403.00, 5000.00],
                    #[ 1.49315588E+01,  1.95927054E-02, -6.66958265E-06,
                    #1.03222635E-09, -5.97583630E-14, -1.46178979E+04,
                    #-5.25561919E+01])),
        #note='8/9/4THERM')

#species(name='SC4H9O',
        #atoms='C:4 H:9 O:1',
        #thermo=(NASA([300.00, 1679.00],
                    #[-4.09000900E-01,  5.38573427E-02, -3.39105695E-05,
                    #1.01023834E-08, -1.11268025E-12, -1.11109627E+04,
                    #2.88554742E+01]),
                #NASA([1679.00, 5000.00],
                    #[ 1.43323395E+01,  2.04542365E-02, -7.12271896E-06,
                    #1.12545447E-09, -6.62697853E-14, -1.60006806E+04,
                    #-5.02030895E+01])),
        #note='8/9/4THERM')

#species(name='C4H71-O',
        #atoms='C:4 H:7 O:1',
        #thermo=(NASA([300.00, 1395.00],
                    #[-1.60619192E+00,  5.58562682E-02, -4.35595767E-05,
                    #1.70589279E-08, -2.65635180E-12,  4.85090326E+03,
                    #3.47112559E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 1.53137780E+01,  1.43427017E-02, -4.81625517E-06,
                    #7.39574839E-10, -4.26140814E-14, -7.29342884E+02,
                    #-5.52937859E+01])),
        #note='4/3/0THERM')

#species(name='C4H8O1-2',
        #atoms='C:4 H:8 O:1',
        #thermo=(NASA([300.00, 1399.00],
                    #[-2.42033073E+00,  5.79308508E-02, -4.30236499E-05,
                    #1.66679793E-08, -2.64714350E-12, -1.53950880E+04,
                    #3.66425242E+01]),
                #NASA([1399.00, 5000.00],
                    #[ 1.39197290E+01,  1.85551492E-02, -6.36014179E-06,
                    #9.88844645E-10, -5.74274779E-14, -2.09452548E+04,
                    #-5.06788231E+01])),
        #note='4/3/0THERM')

#species(name='C4H8O1-3',
        #atoms='C:4 H:8 O:1',
        #thermo=(NASA([300.00, 1371.00],
                    #[-2.53690104E+00,  5.43995707E-02, -3.43390305E-05,
                    #1.01079922E-08, -1.10262736E-12, -1.52980680E+04,
                    #3.67400719E+01]),
                #NASA([1371.00, 5000.00],
                    #[ 1.54227092E+01,  1.70211052E-02, -6.06347951E-06,
                    #9.67354762E-10, -5.71992419E-14, -2.20194174E+04,
                    #-6.13871862E+01])),
        #note='1/22/95THERM')

#species(name='C4H8O1-4',
        #atoms='C:4 H:8 O:1',
        #thermo=(NASA([300.00, 1361.00],
                    #[-3.28505561E+00,  5.04800902E-02, -2.51998984E-05,
                    #3.65743744E-09,  3.94863161E-13, -2.30067422E+04,
                    #4.19349253E+01]),
                #NASA([1361.00, 5000.00],
                    #[ 1.42360731E+01,  1.81176105E-02, -6.46263809E-06,
                    #1.03194372E-09, -6.10557331E-14, -2.99478670E+04,
                    #-5.52081041E+01])),
        #note='1/22/95THERM')

#species(name='C4H8O2-3',
        #atoms='C:4 H:8 O:1',
        #thermo=(NASA([300.00, 1384.00],
                    #[-2.98701197E+00,  6.32678880E-02, -5.20855777E-05,
                    #2.24064364E-08, -3.95416618E-12, -1.71957196E+04,
                    #3.66920390E+01]),
                #NASA([1384.00, 5000.00],
                    #[ 1.58263859E+01,  1.64400938E-02, -5.80680311E-06,
                    #9.21146000E-10, -5.42511002E-14, -2.35334573E+04,
                    #-6.34844562E+01])),
        #note='1/22/95THERM')

#species(name='SC4H8OH-3',
        #atoms='C:4 H:9 O:1',
        #thermo=(NASA([300.00, 1403.00],
                    #[ 1.93185505E+00,  4.79506224E-02, -3.23719194E-05,
                    #1.16232751E-08, -1.72498918E-12, -1.39431370E+04,
                    #2.03995003E+01]),
                #NASA([1403.00, 5000.00],
                    #[ 1.42453780E+01,  1.89491000E-02, -6.27403528E-06,
                    #9.52692609E-10, -5.44157896E-14, -1.81803642E+04,
                    #-4.56206494E+01])))

#species(name='C4H8OH-1O2',
        #atoms='C:4 H:9 O:3',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 2.88497398E+00,  5.63287929E-02, -3.98403503E-05,
                    #1.53891643E-08, -2.51940787E-12, -3.05246721E+04,
                    #2.09293355E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 1.74383247E+01,  2.16778876E-02, -7.37772628E-06,
                    #1.14128811E-09, -6.60391451E-14, -3.55892620E+04,
                    #-5.71247140E+01])),
        #note='6/26/95THERM')

#species(name='C4H8OH-2O2',
        #atoms='C:4 H:9 O:3',
        #thermo=(NASA([300.00, 1399.00],
                    #[ 2.10003504E+00,  6.17528218E-02, -4.77238541E-05,
                    #1.99145784E-08, -3.44416275E-12, -3.26698755E+04,
                    #2.30901271E+01]),
                #NASA([1399.00, 5000.00],
                    #[ 1.82942871E+01,  2.09395389E-02, -7.12096934E-06,
                    #1.10103094E-09, -6.36889768E-14, -3.80505855E+04,
                    #-6.29154326E+01])),
        #note='6/26/95THERM')

#species(name='C4H8OOH1-2',
        #atoms='C:4 H:9 O:2',
        #thermo=(NASA([300.00, 1543.00],
                    #[ 4.11088786E+00,  4.40707552E-02, -2.05662367E-05,
                    #2.16287651E-09,  6.28147736E-13, -4.21015542E+03,
                    #1.31340908E+01]),
                #NASA([1543.00, 5000.00],
                    #[ 1.79390997E+01,  1.82089140E-02, -6.09336386E-06,
                    #9.33343734E-10, -5.36802933E-14, -9.54818055E+03,
                    #-6.32376942E+01])),
        #note='8/23/12')

#species(name='C4H8OOH1-3',
        #atoms='C:4 H:9 O:2',
        #thermo=(NASA([300.00, 1407.00],
                    #[ 3.38642457E+00,  4.94794333E-02, -3.11731015E-05,
                    #9.34788901E-09, -9.92348731E-13, -4.34048889E+03,
                    #1.65183354E+01]),
                #NASA([1407.00, 5000.00],
                    #[ 1.75957868E+01,  1.78823172E-02, -5.84601568E-06,
                    #8.80852127E-10, -5.00662655E-14, -9.31608811E+03,
                    #-6.01625528E+01])),
        #note='9/1/12')

#species(name='C4H8OOH1-4',
        #atoms='C:4 H:9 O:2',
        #thermo=(NASA([300.00, 1400.00],
                    #[ 1.16451160E+00,  5.72955861E-02, -3.91319051E-05,
                    #1.28721497E-08, -1.58263019E-12, -2.58786910E+03,
                    #2.67435065E+01]),
                #NASA([1400.00, 5000.00],
                    #[ 1.85835797E+01,  1.74940527E-02, -5.81082562E-06,
                    #8.85581555E-10, -5.07537090E-14, -8.58260234E+03,
                    #-6.68841973E+01])),
        #note='9/1/12')

#species(name='C4H8OOH2-1',
        #atoms='C:4 H:9 O:2',
        #thermo=(NASA([300.00, 1399.00],
                    #[ 1.17740818E+00,  6.15982544E-02, -5.08527521E-05,
                    #2.20415677E-08, -3.84993017E-12, -4.54848974E+03,
                    #2.46740872E+01]),
                #NASA([1399.00, 5000.00],
                    #[ 1.77802286E+01,  1.77014238E-02, -5.77663712E-06,
                    #8.69169184E-10, -4.93475318E-14, -9.75728180E+03,
                    #-6.26039433E+01])),
        #note='9/1/12')

#species(name='C4H8OOH2-3',
        #atoms='C:4 H:9 O:2',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 1.51848988E+00,  5.66020683E-02, -4.19819291E-05,
                    #1.62832583E-08, -2.56698689E-12, -6.13134926E+03,
                    #2.42065717E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 1.71222128E+01,  1.83275129E-02, -6.00816831E-06,
                    #9.06985051E-10, -5.16189415E-14, -1.13097779E+04,
                    #-5.88518523E+01])),
        #note='9/1/12')

#species(name='C4H8OOH2-4',
        #atoms='C:4 H:9 O:2',
        #thermo=(NASA([300.00, 1399.00],
                    #[ 1.17740818E+00,  6.15982544E-02, -5.08527521E-05,
                    #2.20415677E-08, -3.84993017E-12, -4.54848974E+03,
                    #2.46740872E+01]),
                #NASA([1399.00, 5000.00],
                    #[ 1.77802286E+01,  1.77014238E-02, -5.77663712E-06,
                    #8.69169184E-10, -4.93475318E-14, -9.75728180E+03,
                    #-6.26039433E+01])),
        #note='9/1/12')

#species(name='C4H8OOH1-2O2',
        #atoms='C:4 H:9 O:4',
        #thermo=(NASA([300.00, 1407.00],
                    #[ 4.97809196E+00,  6.30221979E-02, -5.08369093E-05,
                    #2.14022338E-08, -3.61837516E-12, -2.33785644E+04,
                    #9.92185868E+00]),
                #NASA([1407.00, 5000.00],
                    #[ 2.20047701E+01,  1.84524759E-02, -5.96541861E-06,
                    #8.91556992E-10, -5.03719127E-14, -2.87195147E+04,
                    #-7.96660066E+01])),
        #note='9/1/12')

#species(name='C4H8OOH1-3O2',
        #atoms='C:4 H:9 O:4',
        #thermo=(NASA([300.00, 1403.00],
                    #[ 4.62285172E+00,  6.29062788E-02, -4.88786183E-05,
                    #1.94853380E-08, -3.10968451E-12, -2.32812507E+04,
                    #1.19093259E+01]),
                #NASA([1403.00, 5000.00],
                    #[ 2.25294170E+01,  1.79966270E-02, -5.89860391E-06,
                    #8.90459870E-10, -5.06832571E-14, -2.90832590E+04,
                    #-8.29865593E+01])),
        #note='9/1/12')

#species(name='C4H8OOH1-4O2',
        #atoms='C:4 H:9 O:4',
        #thermo=(NASA([300.00, 1401.00],
                    #[ 5.66005233E+00,  5.67747056E-02, -4.09134514E-05,
                    #1.53803855E-08, -2.33839484E-12, -2.15285481E+04,
                    #8.66614371E+00]),
                #NASA([1401.00, 5000.00],
                    #[ 2.10565572E+01,  1.93146062E-02, -6.27532497E-06,
                    #9.41231913E-10, -5.33168957E-14, -2.66456739E+04,
                    #-7.33584419E+01])),
        #note='9/1/12')

#species(name='C4H8OOH2-1O2',
        #atoms='C:4 H:9 O:4',
        #thermo=(NASA([300.00, 1407.00],
                    #[ 4.97809196E+00,  6.30221979E-02, -5.08369093E-05,
                    #2.14022338E-08, -3.61837516E-12, -2.33785644E+04,
                    #9.92185868E+00]),
                #NASA([1407.00, 5000.00],
                    #[ 2.20047701E+01,  1.84524759E-02, -5.96541861E-06,
                    #8.91556992E-10, -5.03719127E-14, -2.87195147E+04,
                    #-7.96660066E+01])),
        #note='9/1/12')

#species(name='C4H8OOH2-3O2',
        #atoms='C:4 H:9 O:4',
        #thermo=(NASA([300.00, 1406.00],
                    #[ 3.80491691E+00,  6.83934891E-02, -5.79295908E-05,
                    #2.53748849E-08, -4.43236163E-12, -2.50643608E+04,
                    #1.41238878E+01]),
                #NASA([1406.00, 5000.00],
                    #[ 2.25193043E+01,  1.81552061E-02, -5.89326339E-06,
                    #8.83422284E-10, -5.00238043E-14, -3.08191398E+04,
                    #-8.39166749E+01])),
        #note='9/1/12')

#species(name='C4H8OOH2-4O2',
        #atoms='C:4 H:9 O:4',
        #thermo=(NASA([300.00, 1407.00],
                    #[ 4.97809196E+00,  6.30221979E-02, -5.08369093E-05,
                    #2.14022338E-08, -3.61837516E-12, -2.33785644E+04,
                    #9.92185868E+00]),
                #NASA([1407.00, 5000.00],
                    #[ 2.20047701E+01,  1.84524759E-02, -5.96541861E-06,
                    #8.91556992E-10, -5.03719127E-14, -2.87195147E+04,
                    #-7.96660066E+01])),
        #note='9/1/12')

#species(name='C4H72-1OOH',
        #atoms='C:4 H:8 O:2',
        #thermo=(NASA([300.00, 1381.00],
                    #[ 1.29755275E+00,  5.59252255E-02, -4.08890003E-05,
                    #1.54880526E-08, -2.42412478E-12, -1.16046928E+04,
                    #2.43621382E+01]),
                #NASA([1381.00, 5000.00],
                    #[ 1.80122740E+01,  1.70340943E-02, -5.89884086E-06,
                    #9.23962123E-10, -5.39539803E-14, -1.74585465E+04,
                    #-6.55209757E+01])),
        #note='9/8/14')

#species(name='C4H71-4OOH',
        #atoms='C:4 H:8 O:2',
        #thermo=(NASA([300.00, 1392.00],
                    #[ 1.31653247E+00,  5.16546159E-02, -3.47310922E-05,
                    #1.20405755E-08, -1.71650639E-12, -1.17955885E+04,
                    #2.46384545E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 1.59871304E+01,  1.86028119E-02, -6.40003189E-06,
                    #9.97531711E-10, -5.80334203E-14, -1.70353450E+04,
                    #-5.46227137E+01])),
        #note='9/8/14')

#species(name='C4H71-3OOH',
        #atoms='C:4 H:8 O:2',
        #thermo=(NASA([300.00, 1392.00],
                    #[-1.50977396E+00,  6.85369305E-02, -5.75193633E-05,
                    #2.43179107E-08, -4.09788488E-12, -1.18018961E+04,
                    #3.50420113E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 1.92985494E+01,  1.54534427E-02, -5.25460431E-06,
                    #8.13772446E-10, -4.71689947E-14, -1.85003480E+04,
                    #-7.49926639E+01])),
        #note='9/8/14')

#species(name='C4H71-3,4OOH',
        #atoms='C:4 H:9 O:4',
        #thermo=(NASA([300.00, 1407.00],
                    #[ 1.86723456E+00,  7.44877061E-02, -6.25572873E-05,
                    #2.63558512E-08, -4.38696236E-12, -1.48595004E+04,
                    #2.66340450E+01]),
                #NASA([1407.00, 5000.00],
                    #[ 2.39953392E+01,  1.71031027E-02, -5.66760181E-06,
                    #8.62413511E-10, -4.93737188E-14, -2.18188653E+04,
                    #-8.99282558E+01])),
        #note='1/14')

#species(name='C4H72-3,4OOH',
        #atoms='C:4 H:9 O:4',
        #thermo=(NASA([300.00, 1400.00],
                    #[ 2.70891700E+00,  6.63784916E-02, -4.79787328E-05,
                    #1.67756211E-08, -2.23855120E-12, -1.64965328E+04,
                    #2.33313413E+01]),
                #NASA([1400.00, 5000.00],
                    #[ 2.34794236E+01,  1.78306825E-02, -5.98251496E-06,
                    #9.18228543E-10, -5.28933083E-14, -2.35324885E+04,
                    #-8.79191723E+01])),
        #note='1/14')

#species(name='C4H71-2,4OOH',
        #atoms='C:4 H:9 O:4',
        #thermo=(NASA([300.00, 1407.00],
                    #[ 1.86723456E+00,  7.44877061E-02, -6.25572873E-05,
                    #2.63558512E-08, -4.38696236E-12, -1.48595004E+04,
                    #2.66340450E+01]),
                #NASA([1407.00, 5000.00],
                    #[ 2.39953392E+01,  1.71031027E-02, -5.66760181E-06,
                    #8.62413511E-10, -4.93737188E-14, -2.18188653E+04,
                    #-8.99282558E+01])),
        #note='1/14THERM')

#species(name='C4H72-1,3OOH',
        #atoms='C:4 H:9 O:4',
        #thermo=(NASA([300.00, 1400.00],
                    #[ 2.70891700E+00,  6.63784916E-02, -4.79787328E-05,
                    #1.67756211E-08, -2.23855120E-12, -1.64965328E+04,
                    #2.33313413E+01]),
                #NASA([1400.00, 5000.00],
                    #[ 2.34794236E+01,  1.78306825E-02, -5.98251496E-06,
                    #9.18228543E-10, -5.28933083E-14, -2.35324885E+04,
                    #-8.79191723E+01])),
        #note='1/14THERM')

#species(name='C4H72-1,4OOH',
        #atoms='C:4 H:9 O:4',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 1.44670410E+00,  7.28736321E-02, -6.26626253E-05,
                    #2.73079747E-08, -4.77166491E-12, -1.44467413E+04,
                    #3.09980074E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 2.09572073E+01,  2.06411107E-02, -8.38936394E-06,
                    #1.53960504E-09, -1.00462875E-13, -2.04921532E+04,
                    #-7.13195428E+01])),
        #note='1/14THERM')

#species(name='C4H71-2,3OOH',
        #atoms='C:4 H:9 O:4',
        #thermo=(NASA([300.00, 1403.00],
                    #[ 2.39280462E+00,  7.70376238E-02, -7.30090599E-05,
                    #3.52799789E-08, -6.67480668E-12, -1.69057880E+04,
                    #2.14225214E+01]),
                #NASA([1403.00, 5000.00],
                    #[ 2.27731474E+01,  1.75034209E-02, -5.66457024E-06,
                    #8.47219472E-10, -4.78921429E-14, -2.27515175E+04,
                    #-8.37370284E+01])),
        #note='1/14THERM')

#species(name='C4H7O1-3OOH-4',
        #atoms='C:4 H:8 O:3',
        #thermo=(NASA([300.00, 1414.00],
                    #[-5.87403962E+00,  8.33039901E-02, -7.00220183E-05,
                    #2.93119233E-08, -4.84389336E-12, -2.60133188E+04,
                    #6.77644454E+01]),
                #NASA([1414.00, 5000.00],
                    #[ 1.92854938E+01,  1.85951439E-02, -6.36386688E-06,
                    #9.88877447E-10, -5.74229199E-14, -3.39951341E+04,
                    #-6.49803548E+01])),
        #note='1/14THERM')

#species(name='C4H7O1-3OOH-2',
        #atoms='C:4 H:8 O:3',
        #thermo=(NASA([300.00, 1426.00],
                    #[-5.80699080E+00,  8.70554703E-02, -8.07046622E-05,
                    #3.73023383E-08, -6.71892329E-12, -2.74592775E+04,
                    #5.31942410E+01]),
                #NASA([1426.00, 5000.00],
                    #[ 1.85731016E+01,  1.79855347E-02, -5.89151227E-06,
                    #8.88341423E-10, -5.04988876E-14, -3.45330146E+04,
                    #-7.31374827E+01])),
        #note='1/14THERM')

#species(name='C4H7O1-2OOH-4',
        #atoms='C:4 H:8 O:3',
        #thermo=(NASA([300.00, 1321.00],
                    #[-3.03909300E+00,  7.70427849E-02, -6.78116281E-05,
                    #2.87134868E-08, -4.75633566E-12, -2.52096262E+04,
                    #4.37178696E+01]),
                #NASA([1321.00, 5000.00],
                    #[ 1.21804875E+01,  2.36602376E-02, -7.63821374E-06,
                    #8.99674655E-10, -2.40399075E-14, -2.76074166E+04,
                    #-2.98669581E+01])),
        #note='1/14THERM')

#species(name='C4H7O1-4OOH-2',
        #atoms='C:4 H:8 O:3',
        #thermo=(NASA([300.00, 1318.00],
                    #[-7.95551052E+00,  9.00599921E-02, -8.03662894E-05,
                    #3.43520365E-08, -5.72381630E-12, -3.49921422E+04,
                    #6.49968122E+01]),
                #NASA([1318.00, 5000.00],
                    #[ 1.02421776E+01,  2.68371730E-02, -8.94220565E-06,
                    #1.11033806E-09, -3.52804404E-14, -3.80202869E+04,
                    #-2.33637448E+01])),
        #note='1/14THERM')

#species(name='C4H7O1-2OOH-3',
        #atoms='C:4 H:8 O:3',
        #thermo=(NASA([300.00, 1425.00],
                    #[-2.59179532E+00,  7.75547001E-02, -6.80402455E-05,
                    #2.98114704E-08, -5.12684762E-12, -2.72177540E+04,
                    #3.91709078E+01]),
                #NASA([1425.00, 5000.00],
                    #[ 1.99393696E+01,  1.69060349E-02, -5.62817044E-06,
                    #8.58382596E-10, -4.91993361E-14, -3.40672730E+04,
                    #-7.86999130E+01])),
        #note='1/14THERM')

#species(name='C4H7O2-3OOH-1',
        #atoms='C:4 H:8 O:3',
        #thermo=(NASA([300.00, 1434.00],
                    #[-7.98811313E-01,  6.85676116E-02, -5.52683004E-05,
                    #2.23194926E-08, -3.54844373E-12, -2.76378733E+04,
                    #3.15075609E+01]),
                #NASA([1434.00, 5000.00],
                    #[ 1.92543019E+01,  1.73732227E-02, -5.76537817E-06,
                    #8.77411418E-10, -5.02138537E-14, -3.39904310E+04,
                    #-7.43499141E+01])),
        #note='1/14THERM')

#species(name='HO2CH2CHO',
        #atoms='C:2 H:4 O:3',
        #thermo=(NASA([300.00, 1391.00],
                    #[-1.32768631E+00,  5.21618601E-02, -4.97327645E-05,
                    #2.31272366E-08, -4.20787867E-12, -2.90608844E+04,
                    #3.61860491E+01]),
                #NASA([1391.00, 5000.00],
                    #[ 1.51554685E+01,  7.57240000E-03, -2.72693024E-06,
                    #4.38217189E-10, -2.60434287E-14, -3.41419680E+04,
                    #-5.01255068E+01])),
        #note='9/8/14')

#species(name='NC4KET12',
        #atoms='C:4 H:8 O:3',
        #thermo=(NASA([300.00, 1391.00],
                    #[ 6.79868708E-01,  6.64490814E-02, -5.41920789E-05,
                    #2.29108842E-08, -3.94482528E-12, -3.99225985E+04,
                    #2.89413177E+01]),
                #NASA([1391.00, 5000.00],
                    #[ 2.01935363E+01,  1.75757627E-02, -6.14119142E-06,
                    #9.67228346E-10, -5.66826486E-14, -4.64040370E+04,
                    #-7.47504131E+01])),
        #note='10/17/12')

#species(name='NC4KET13',
        #atoms='C:4 H:8 O:3',
        #thermo=(NASA([300.00, 1359.00],
                    #[ 5.01329203E+00,  4.46935759E-02, -1.93602546E-05,
                    #9.36914700E-10,  8.79544966E-13, -4.01116170E+04,
                    #8.45810839E+00]),
                #NASA([1359.00, 5000.00],
                    #[ 2.05329077E+01,  1.77923667E-02, -6.33303705E-06,
                    #1.00980249E-09, -5.96860625E-14, -4.64515541E+04,
                    #-7.82394029E+01])),
        #note='10/17/12')

#species(name='NC4KET14',
        #atoms='C:4 H:8 O:3',
        #thermo=(NASA([300.00, 1683.00],
                    #[ 3.81084680E+00,  4.79846992E-02, -2.40272245E-05,
                    #4.06503766E-09,  1.02115318E-13, -3.80119322E+04,
                    #1.56667637E+01]),
                #NASA([1683.00, 5000.00],
                    #[ 1.75047160E+01,  2.13987514E-02, -7.88768858E-06,
                    #1.29161450E-09, -7.78674250E-14, -4.30038194E+04,
                    #-5.94011957E+01])),
        #note='10/17/12')

#species(name='NC4KET21',
        #atoms='C:4 H:8 O:3',
        #thermo=(NASA([300.00, 1530.00],
                    #[ 6.73128471E+00,  3.47415959E-02, -2.00380045E-06,
                    #-1.06311860E-08,  3.51986112E-12, -4.19538953E+04,
                    #2.64836589E+00]),
                #NASA([1530.00, 5000.00],
                    #[ 2.09112593E+01,  1.76512301E-02, -6.32733352E-06,
                    #1.01363708E-09, -6.01083707E-14, -4.84340373E+04,
                    #-7.91228611E+01])),
        #note='10/17/12')

#species(name='NC4KET23',
        #atoms='C:4 H:8 O:3',
        #thermo=(NASA([300.00, 1386.00],
                    #[ 7.72137962E-01,  6.37996750E-02, -4.99093248E-05,
                    #2.04982417E-08, -3.48371157E-12, -4.30966399E+04,
                    #2.90104109E+01]),
                #NASA([1386.00, 5000.00],
                    #[ 1.95748872E+01,  1.82621559E-02, -6.41442968E-06,
                    #1.01374326E-09, -5.95494716E-14, -4.95474113E+04,
                    #-7.15402539E+01])),
        #note='10/17/12')

#species(name='NC4KET24',
        #atoms='C:4 H:8 O:3',
        #thermo=(NASA([300.00, 1683.00],
                    #[ 4.30704813E+00,  4.36065214E-02, -1.73729572E-05,
                    #3.60442667E-10,  8.08121495E-13, -4.12480820E+04,
                    #1.38604048E+01]),
                #NASA([1683.00, 5000.00],
                    #[ 1.66466077E+01,  2.24308129E-02, -8.31906841E-06,
                    #1.36749860E-09, -8.26526159E-14, -4.60267778E+04,
                    #-5.47993275E+01])),
        #note='10/17/12')

#species(name='C2H5COCH3',
        #atoms='C:4 H:8 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 6.61978185E+00,  8.51847835E-03,  5.10322077E-05,
                    #-6.58433042E-08,  2.49110484E-11, -3.15251691E+04,
                    #-1.09485469E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 9.29655016E+00,  2.29172746E-02, -8.22048591E-06,
                    #1.32404838E-09, -7.91751980E-14, -3.34442311E+04,
                    #-2.04993263E+01])),
        #note='T09/10')

#species(name='C2H5COCH2',
        #atoms='C:4 H:7 O:1',
        #thermo=(NASA([300.00, 1383.00],
                    #[ 1.54013856E+00,  4.39486258E-02, -2.97002421E-05,
                    #1.05495313E-08, -1.58598769E-12, -9.50796505E+03,
                    #1.99706641E+01]),
                #NASA([1383.00, 5000.00],
                    #[ 1.42098738E+01,  1.57866459E-02, -5.50529183E-06,
                    #8.65870540E-10, -5.06913329E-14, -1.41284951E+04,
                    #-4.87132911E+01])),
        #note='4/3/0THERM')

#species(name='CH2CH2COCH3',
        #atoms='C:4 H:7 O:1',
        #thermo=(NASA([300.00, 1380.00],
                    #[ 2.40255609E+00,  3.67294268E-02, -1.97316510E-05,
                    #5.07323216E-09, -4.99655275E-13, -6.15006886E+03,
                    #1.93993386E+01]),
                #NASA([1380.00, 5000.00],
                    #[ 1.24694368E+01,  1.71022143E-02, -5.92156726E-06,
                    #9.26816806E-10, -5.40730504E-14, -1.01378242E+04,
                    #-3.62186375E+01])),
        #note='6/21/95THER')

#species(name='CH3CHCOCH3',
        #atoms='C:4 H:7 O:1',
        #thermo=(NASA([300.00, 1384.00],
                    #[ 8.12941079E-01,  4.29256944E-02, -2.69230252E-05,
                    #8.59326807E-09, -1.13188129E-12, -1.05247481E+04,
                    #2.32952685E+01]),
                #NASA([1384.00, 5000.00],
                    #[ 1.31388032E+01,  1.66091073E-02, -5.76924215E-06,
                    #9.04978165E-10, -5.28826527E-14, -1.51162098E+04,
                    #-4.38876580E+01])),
        #note='4/3/0THERM')

#species(name='C2H3COCH3',
        #atoms='C:4 H:6 O:1',
        #thermo=(NASA([300.00, 1386.00],
                    #[ 2.45578501E-01,  4.26432049E-02, -2.91126822E-05,
                    #1.03478392E-08, -1.53551381E-12, -1.70305379E+04,
                    #2.64430799E+01]),
                #NASA([1386.00, 5000.00],
                    #[ 1.25571995E+01,  1.49672645E-02, -5.20015351E-06,
                    #8.15864365E-10, -4.76824406E-14, -2.14622958E+04,
                    #-4.01434299E+01])),
        #note='6/19/95THERM')

#species(name='CH3CHOOCOCH3',
        #atoms='C:4 H:7 O:3',
        #thermo=(NASA([300.00, 1393.00],
                    #[ 1.59405312E+00,  5.74729487E-02, -4.65071459E-05,
                    #2.01816805E-08, -3.61175600E-12, -2.53445178E+04,
                    #2.60857480E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 1.72170832E+01,  1.76740453E-02, -6.10387321E-06,
                    #9.53711693E-10, -5.55757409E-14, -3.05280370E+04,
                    #-5.67959657E+01])),
        #note='6/27/95')

#species(name='CH2CHOOHCOCH3',
        #atoms='C:4 H:7 O:3',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 1.55204769E+00,  6.08201262E-02, -5.04704383E-05,
                    #2.16347587E-08, -3.75674081E-12, -1.89563540E+04,
                    #2.71011896E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 1.93158016E+01,  1.56592602E-02, -5.45395604E-06,
                    #8.57160086E-10, -5.01582885E-14, -2.47752201E+04,
                    #-6.70296581E+01])),
        #note='6/27/95')

#species(name='NC3H7CHO',
        #atoms='C:4 H:8 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 5.30068149E+00,  5.00213349E-03,  8.12219686E-05,
                    #-1.07815910E-07,  4.25781054E-11, -2.71198341E+04,
                    #4.93592991E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 1.02351219E+01,  2.32201057E-02, -8.46144199E-06,
                    #1.37589764E-09, -8.27046434E-14, -3.00345804E+04,
                    #-2.82583105E+01])),
        #note='T05/09')

#species(name='NC3H7CO',
        #atoms='C:4 H:7 O:1',
        #thermo=(NASA([300.00, 1380.00],
                    #[ 2.67256826E+00,  3.71198825E-02, -2.06862859E-05,
                    #5.48873888E-09, -5.35864183E-13, -8.58050888E+03,
                    #1.64848950E+01]),
                #NASA([1380.00, 5000.00],
                    #[ 1.30026331E+01,  1.63104877E-02, -5.57642899E-06,
                    #8.65670629E-10, -5.02255667E-14, -1.25523385E+04,
                    #-4.02608515E+01])),
        #note='9/27/95THERM')

#species(name='C3H6CHO-3',
        #atoms='C:4 H:7 O:1',
        #thermo=(NASA([300.00, 1379.00],
                    #[ 2.67672303E+00,  3.73064128E-02, -2.11281405E-05,
                    #5.80472681E-09, -6.09688236E-13, -2.49714183E+03,
                    #1.75750933E+01]),
                #NASA([1379.00, 5000.00],
                    #[ 1.30322954E+01,  1.62418373E-02, -5.54388124E-06,
                    #8.59723685E-10, -4.98459726E-14, -6.45915975E+03,
                    #-3.92399021E+01])),
        #note='9/27/95THERM')

#species(name='C3H6CHO-2',
        #atoms='C:4 H:7 O:1',
        #thermo=(NASA([300.00, 1682.00],
                    #[ 2.95067531E+00,  3.34223079E-02, -1.45356815E-05,
                    #1.67282048E-09,  2.62011555E-13, -3.79034324E+03,
                    #1.74324072E+01]),
                #NASA([1682.00, 5000.00],
                    #[ 1.11942816E+01,  1.81806772E-02, -6.35916662E-06,
                    #1.00727333E-09, -5.93943618E-14, -6.86826460E+03,
                    #-2.80298956E+01])),
        #note='11/15/95THERM')

#species(name='C3H6CHO-1',
        #atoms='C:4 H:7 O:1',
        #thermo=(NASA([300.00, 1387.00],
                    #[ 1.35168659E+00,  4.23714083E-02, -2.69123192E-05,
                    #8.70132567E-09, -1.15287455E-12, -6.91166610E+03,
                    #2.09381509E+01]),
                #NASA([1387.00, 5000.00],
                    #[ 1.34301738E+01,  1.62250792E-02, -5.60631713E-06,
                    #8.76356273E-10, -5.10864241E-14, -1.13561098E+04,
                    #-4.47371262E+01])),
        #note='11/15/95THERM')

#species(name='C2H5CHCO',
        #atoms='C:4 H:6 O:1',
        #thermo=(NASA([300.00, 1400.00],
                    #[ 2.76600655E-01,  4.82966593E-02, -4.02376928E-05,
                    #1.76564254E-08, -3.14365357E-12, -1.41717351E+04,
                    #2.47118671E+01]),
                #NASA([1400.00, 5000.00],
                    #[ 1.34100537E+01,  1.38766678E-02, -4.74130143E-06,
                    #7.35504188E-10, -4.26457752E-14, -1.83790837E+04,
                    #-4.45295352E+01])),
        #note='9/27/95THERM')

#species(name='SC3H5CHO',
        #atoms='C:4 H:6 O:1',
        #thermo=(NASA([300.00, 1390.00],
                    #[ 4.35795171E-01,  4.48719314E-02, -3.36582931E-05,
                    #1.33066870E-08, -2.17839128E-12, -1.60394651E+04,
                    #2.37597452E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 1.31695904E+01,  1.42484434E-02, -4.90843998E-06,
                    #7.65789041E-10, -4.45834896E-14, -2.04032613E+04,
                    #-4.43673205E+01])),
        #note='11/15/95THER')

#species(name='SC3H5CO',
        #atoms='C:4 H:5 O:1',
        #thermo=(NASA([300.00, 1392.00],
                    #[ 1.74191343E+00,  3.97229536E-02, -3.20061901E-05,
                    #1.38227925E-08, -2.46272017E-12, -6.64428100E+02,
                    #1.70762023E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 1.25514754E+01,  1.22521948E-02, -4.22382101E-06,
                    #6.59184896E-10, -3.83818826E-14, -4.25349795E+03,
                    #-4.02864145E+01])),
        #note='11/15/95THERM')

#species(name='CH2CH2CHO',
        #atoms='C:3 H:5 O:1',
        #thermo=(NASA([300.00, 1363.00],
                    #[ 8.36272659E+00, -3.90712606E-03,  3.59751257E-05,
                    #-2.69209237E-08,  5.98920261E-12, -7.55243244E+02,
                    #-9.53140102E+00]),
                #NASA([1363.00, 5000.00],
                    #[ 9.96855598E+00,  1.61917946E-02, -6.64161740E-06,
                    #1.15444520E-09, -7.21981790E-14, -4.16450354E+03,
                    #-2.73971573E+01])),
        #note='8/10/4THERM')

#species(name='IC4H10',
        #atoms='C:4 H:10',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 4.45479140E+00,  8.26058864E-03,  8.29886433E-05,
                    #-1.14647616E-07,  4.64569994E-11, -1.84593929E+04,
                    #4.92740653E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 9.76991697E+00,  2.54997141E-02, -9.14142587E-06,
                    #1.47328201E-09, -8.80799697E-14, -2.14052667E+04,
                    #-3.00329670E+01])),
        #note='G8/00')

#species(name='IC4H9',
        #atoms='C:4 H:9',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 3.34476784E+00,  2.31869650E-02,  3.28261040E-05,
                    #-5.96398514E-08,  2.58980820E-11,  6.66201200E+03,
                    #9.68860372E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 9.61250942E+00,  2.28581786E-02, -8.06391309E-06,
                    #1.28556553E-09, -7.62730799E-14,  4.15218608E+03,
                    #-2.66485099E+01])),
        #note='T6/04')

#species(name='IC4H8',
        #atoms='C:4 H:8',
        #thermo=(NASA([300.00, 1392.00],
                    #[ 5.72478139E-02,  4.17768938E-02, -2.49095729E-05,
                    #7.54294402E-09, -9.23202212E-13, -3.72166259E+03,
                    #2.35698905E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 1.11444028E+01,  1.81609265E-02, -6.17791116E-06,
                    #9.55481871E-10, -5.52826092E-14, -7.84024684E+03,
                    #-3.68508829E+01])))

#species(name='IC4H7',
        #atoms='C:4 H:7',
        #thermo=(NASA([300.00, 1384.00],
                    #[-2.29578762E-01,  4.17842986E-02, -2.66885700E-05,
                    #8.42205744E-09, -1.03175361E-12,  1.43946680E+04,
                    #2.54797645E+01]),
                #NASA([1384.00, 5000.00],
                    #[ 1.18999143E+01,  1.51569859E-02, -5.09995449E-06,
                    #7.83722199E-10, -4.51660275E-14,  1.00363555E+04,
                    #-4.02286635E+01])))

#species(name='TC4H9',
        #atoms='C:4 H:9',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 6.45910754E+00, -1.02015930E-02,  1.06310577E-04,
                    #-1.25717030E-07,  4.75543216E-11,  4.43420391E+03,
                    #1.30648608E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 6.72557390E+00,  2.53649194E-02, -9.05306262E-06,
                    #1.45474620E-09, -8.67934112E-14,  2.57430692E+03,
                    #-8.89920414E+00])),
        #note='T6/04')

#species(name='TC4H9O2',
        #atoms='C:4 H:9 O:2',
        #thermo=(NASA([300.00, 1380.00],
                    #[ 2.63892371E+00,  5.44717499E-02, -3.75504698E-05,
                    #1.40479250E-08, -2.27968600E-12, -1.47598933E+04,
                    #1.40325533E+01]),
                #NASA([1380.00, 5000.00],
                    #[ 1.80863238E+01,  1.99282971E-02, -6.98287309E-06,
                    #1.10171726E-09, -6.46381057E-14, -2.04420664E+04,
                    #-6.97533212E+01])),
        #note='9/1/12')

#species(name='IC4H7OOCH3',
        #atoms='C:5 H:10 O:2',
        #thermo=(NASA([300.00, 1386.00],
                    #[ 1.26784161E+00,  6.82442475E-02, -5.30807931E-05,
                    #2.27496198E-08, -4.11475065E-12, -1.16767937E+04,
                    #2.44900544E+01]),
                #NASA([1386.00, 5000.00],
                    #[ 1.95896715E+01,  2.31057369E-02, -8.02911330E-06,
                    #1.25969946E-09, -7.36169360E-14, -1.80069088E+04,
                    #-7.34192482E+01])))

#species(name='IC4H7OOIC4H7',
        #atoms='C:8 H:14 O:2',
        #thermo=(NASA([300.00, 1390.00],
                    #[-2.07881710E-01,  1.04075647E-01, -8.33973077E-05,
                    #3.60897624E-08, -6.47906520E-12, -7.79062693E+03,
                    #3.50446555E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 2.80447245E+01,  3.27505220E-02, -1.13220177E-05,
                    #1.77027767E-09, -1.03211988E-13, -1.72683337E+04,
                    #-1.15134796E+02])))

#species(name='IC3H5OCH2',
        #atoms='H:7 C:4 O:1',
        #thermo=(NASA([298.15, 1000.00],
                    #[-2.14798958E+00,  7.00225553E-02, -8.21595440E-05,
                    #5.22589946E-08, -1.34176532E-11,  3.26474773E+03,
                    #3.55145589E+01]),
                #NASA([1000.00, 2000.00],
                    #[ 6.64731727E+00,  3.08190709E-02, -1.73209320E-05,
                    #5.01099629E-09, -6.00089387E-13,  1.70679921E+03,
                    #-5.91214819E+00])),
        #note='6/2/14CZHOU')

#species(name='IC4H7-I1',
        #atoms='C:4 H:7',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 9.12464579E-01,  3.88654394E-02, -2.57575714E-05,
                    #9.07760026E-09, -1.33946902E-12,  2.55635553E+04,
                    #2.08634918E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 1.11158752E+01,  1.55127192E-02, -5.23769366E-06,
                    #8.05998394E-10, -4.64703390E-14,  2.19488297E+04,
                    #-3.41440480E+01])),
        #note='5/13/15')

#species(name='IC4H8OH-IT',
        #atoms='C:4 H:9 O:1',
        #thermo=(NASA([300.00, 1391.00],
                    #[ 3.05275715E+00,  3.93926461E-02, -1.90686417E-05,
                    #3.86408022E-09, -1.48005244E-13, -1.42263749E+04,
                    #1.60840537E+01]),
                #NASA([1391.00, 5000.00],
                    #[ 1.29136746E+01,  2.06583409E-02, -6.98445966E-06,
                    #1.07562552E-09, -6.20443876E-14, -1.81394866E+04,
                    #-3.84972088E+01])))

#species(name='IC4H8OH-TI',
        #atoms='C:4 H:9 O:1',
        #thermo=(NASA([300.00, 1402.00],
                    #[ 2.33169342E+00,  5.13017040E-02, -4.02698872E-05,
                    #1.75150405E-08, -3.16001727E-12, -1.48318978E+04,
                    #1.55368130E+01]),
                #NASA([1402.00, 5000.00],
                    #[ 1.46323607E+01,  1.88895981E-02, -6.30561450E-06,
                    #9.62474230E-10, -5.51640163E-14, -1.87976018E+04,
                    #-4.93218793E+01])))

#species(name='IQJC4H8OH',
        #atoms='C:4 H:9 O:3',
        #thermo=(NASA([300.00, 1410.00],
                    #[ 1.81448831E+00,  7.47452750E-02, -7.10895172E-05,
                    #3.44973679E-08, -6.54646593E-12, -3.44023586E+04,
                    #1.77380434E+01]),
                #NASA([1410.00, 5000.00],
                    #[ 2.11752212E+01,  1.75144254E-02, -5.73227292E-06,
                    #8.63386596E-10, -4.90282414E-14, -3.98881576E+04,
                    #-8.19187015E+01])),
        #note='L2/00')

#species(name='IQC4H8OT',
        #atoms='C:4 H:9 O:3',
        #thermo=(NASA([300.00, 1405.00],
                    #[ 3.72211529E+00,  6.42864861E-02, -5.52809053E-05,
                    #2.50036630E-08, -4.53471908E-12, -2.43110525E+04,
                    #1.21981167E+01]),
                #NASA([1405.00, 5000.00],
                    #[ 2.04823628E+01,  1.82966721E-02, -6.04413378E-06,
                    #9.16380548E-10, -5.22866551E-14, -2.94287153E+04,
                    #-7.53563247E+01])))

#species(name='TQJC4H8OH',
        #atoms='C:4 H:9 O:3',
        #thermo=(NASA([300.00, 1415.00],
                    #[-6.43419503E-01,  8.49131517E-02, -8.17210578E-05,
                    #3.90979927E-08, -7.27092842E-12, -3.42375932E+04,
                    #2.84394025E+01]),
                #NASA([1415.00, 5000.00],
                    #[ 2.29681617E+01,  1.65162786E-02, -5.50247318E-06,
                    #8.39335285E-10, -4.81030625E-14, -4.10051460E+04,
                    #-9.34897892E+01])))

#species(name='TQC4H8OI',
        #atoms='C:4 H:9 O:3',
        #thermo=(NASA([300.00, 1411.00],
                    #[ 7.45747835E-02,  7.46499596E-02, -6.42255048E-05,
                    #2.80908988E-08, -4.87692045E-12, -2.47182737E+04,
                    #2.94511549E+01]),
                #NASA([1411.00, 5000.00],
                    #[ 2.13200701E+01,  1.80489663E-02, -6.06124072E-06,
                    #9.29740751E-10, -5.34977374E-14, -3.12966663E+04,
                    #-8.20046659E+01])))

#species(name='TQC4H7OHI',
        #atoms='C:4 H:9 O:3',
        #thermo=(NASA([300.00, 1404.00],
                    #[ 2.55843807E+00,  6.37086077E-02, -4.97169945E-05,
                    #1.99225109E-08, -3.21373436E-12, -2.77526909E+04,
                    #1.95001368E+01]),
                #NASA([1404.00, 5000.00],
                    #[ 2.08281225E+01,  1.81675094E-02, -6.12943202E-06,
                    #9.43194119E-10, -5.43937008E-14, -3.37386684E+04,
                    #-7.74823720E+01])),
        #note='L2/00')

#species(name='IC3H4CHO-A',
        #atoms='C:4 H:5 O:1',
        #thermo=(NASA([300.00, 1392.00],
                    #[ 7.64345054E-01,  4.45242412E-02, -3.61033720E-05,
                    #1.48295287E-08, -2.43809290E-12,  2.44732544E+03,
                    #2.08541848E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 1.41736959E+01,  1.09161978E-02, -3.69020878E-06,
                    #5.69228087E-10, -3.29023246E-14, -1.92867979E+03,
                    #-5.02663740E+01])))

#species(name='QC4H7OHP',
        #atoms='C:4 H:9 O:3',
        #thermo=(NASA([300.00, 1416.00],
                    #[-1.27864186E+00,  8.94492926E-02, -8.78565423E-05,
                    #4.22110919E-08, -7.83450876E-12, -2.58975226E+04,
                    #3.53963909E+01]),
                #NASA([1416.00, 5000.00],
                    #[ 2.43481084E+01,  1.50316366E-02, -5.01788017E-06,
                    #7.66774357E-10, -4.40093220E-14, -3.31922320E+04,
                    #-9.68211106E+01])))

#species(name='IQC4H7OHT',
        #atoms='C:4 H:9 O:3',
        #thermo=(NASA([300.00, 1413.00],
                    #[ 3.58900054E+00,  7.25591129E-02, -7.14080484E-05,
                    #3.55250907E-08, -6.84991795E-12, -2.57241250E+04,
                    #1.23766657E+01]),
                #NASA([1413.00, 5000.00],
                    #[ 2.19945886E+01,  1.62011186E-02, -5.23758492E-06,
                    #7.81898296E-10, -4.41125515E-14, -3.07383725E+04,
                    #-8.16613568E+01])))

#species(name='SC4H7OH-IP',
        #atoms='C:4 H:7 O:1',
        #thermo=(NASA([300.00, 1402.00],
                    #[-1.46664187E+00,  6.03351671E-02, -5.43112644E-05,
                    #2.49299933E-08, -4.52282491E-12, -6.95012413E+03,
                    #3.20768458E+01]),
                #NASA([1402.00, 5000.00],
                    #[ 1.53490714E+01,  1.38856699E-02, -4.56427754E-06,
                    #6.90418690E-10, -3.93540403E-14, -1.20164758E+04,
                    #-5.55975530E+01])))

#species(name='C2CY(COC)OH',
        #atoms='C:4 H:8 O:2',
        #thermo=(NASA([300.00, 1393.00],
                    #[ 5.92324183E-01,  5.52429007E-02, -4.02419018E-05,
                    #1.57152217E-08, -2.57388393E-12, -3.58241476E+04,
                    #2.23378086E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 1.56829970E+01,  1.92910506E-02, -6.63718495E-06,
                    #1.03441014E-09, -6.01715267E-14, -4.10598236E+04,
                    #-5.85686221E+01])))

#species(name='CCY(CCO)COH',
        #atoms='C:4 H:8 O:2',
        #thermo=(NASA([300.00, 1412.00],
                    #[-7.10048774E+00,  9.53371808E-02, -9.78701612E-05,
                    #4.90005646E-08, -9.41685766E-12, -3.98987202E+04,
                    #5.60924667E+01]),
                #NASA([1412.00, 5000.00],
                    #[ 1.91884885E+01,  1.56255714E-02, -5.22569800E-06,
                    #7.99171074E-10, -4.58832445E-14, -4.71120302E+04,
                    #-7.84579023E+01])))

#species(name='CCY(CCOC)OH',
        #atoms='C:4 H:8 O:2',
        #thermo=(NASA([300.00, 1404.00],
                    #[-2.84914896E+00,  6.23736856E-02, -4.70326536E-05,
                    #1.84908286E-08, -2.94976027E-12, -3.74257004E+04,
                    #3.83163588E+01]),
                #NASA([1404.00, 5000.00],
                    #[ 1.43404718E+01,  1.98311504E-02, -6.60657660E-06,
                    #1.00779570E-09, -5.77614335E-14, -4.30959414E+04,
                    #-5.30535015E+01])),
        #note='L2/00')

#species(name='IC3H6OHCHO',
        #atoms='C:4 H:8 O:2',
        #thermo=(NASA([300.00, 1393.00],
                    #[ 1.84080874E+00,  5.29601347E-02, -3.94261774E-05,
                    #1.59063430E-08, -2.69565279E-12, -5.01437169E+04,
                    #1.75482756E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 1.60254376E+01,  1.85402212E-02, -6.36973877E-06,
                    #9.91732739E-10, -5.76472640E-14, -5.50198923E+04,
                    #-5.83074874E+01])))

#species(name='CH3COCOOH',
        #atoms='C:3 H:6 O:3',
        #thermo=(NASA([300.00, 1405.00],
                    #[ 4.22993148E+00,  4.45552095E-02, -3.76424204E-05,
                    #1.66306528E-08, -2.94629936E-12, -3.72634997E+04,
                    #9.58272910E+00]),
                #NASA([1405.00, 5000.00],
                    #[ 1.61208655E+01,  1.25067303E-02, -4.10793603E-06,
                    #6.20553031E-10, -3.53216368E-14, -4.09312648E+04,
                    #-5.27056280E+01])))

#species(name='IC3H5COHQ',
        #atoms='C:4 H:8 O:3',
        #thermo=(NASA([300.00, 1504.00],
                    #[ 2.64360992E+00,  5.72485066E-02, -3.95907807E-05,
                    #1.27775635E-08, -1.47241476E-12, -3.80099993E+04,
                    #1.77322273E+01]),
                #NASA([1504.00, 5000.00],
                    #[ 2.07387831E+01,  1.58360934E-02, -5.27614462E-06,
                    #8.05932218E-10, -4.62682425E-14, -4.41821241E+04,
                    #-7.94239893E+01])))

#species(name='IC3H5Q',
        #atoms='C:3 H:6 O:2',
        #thermo=(NASA([300.00, 1397.00],
                    #[ 1.32903007E+00,  4.49170722E-02, -3.51235127E-05,
                    #1.41982181E-08, -2.33335008E-12, -1.21898396E+04,
                    #2.02696565E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 1.43424294E+01,  1.28053632E-02, -4.40584813E-06,
                    #6.86848148E-10, -3.99675209E-14, -1.65261025E+04,
                    #-4.89934539E+01])))

#species(name='TQC4H7OHIO2',
        #atoms='C:4 H:9 O:5',
        #thermo=(NASA([300.00, 1402.00],
                    #[ 3.17336206E+00,  7.94005900E-02, -6.51165712E-05,
                    #2.62035931E-08, -4.13406290E-12, -4.84943162E+04,
                    #1.77867184E+01]),
                #NASA([1402.00, 5000.00],
                    #[ 2.82564819E+01,  1.66969871E-02, -5.67314614E-06,
                    #8.78350442E-10, -5.09090253E-14, -5.66017464E+04,
                    #-1.15147927E+02])))

#species(name='TQC4H7OHIQ-I',
        #atoms='C:4 H:9 O:5',
        #thermo=(NASA([300.00, 1384.00],
                    #[ 6.09881562E+00,  6.93745451E-02, -5.12049498E-05,
                    #1.81276221E-08, -2.46329781E-12, -3.91393707E+04,
                    #2.93862639E+00]),
                #NASA([1384.00, 5000.00],
                    #[ 2.88466964E+01,  1.66289773E-02, -5.74301906E-06,
                    #8.98791324E-10, -5.24794891E-14, -4.69249234E+04,
                    #-1.19117836E+02])))

#species(name='TQC4H7OHIQ-P',
        #atoms='C:4 H:9 O:5',
        #thermo=(NASA([300.00, 1400.00],
                    #[ 3.36947511E+00,  7.96855226E-02, -6.74365765E-05,
                    #2.82233283E-08, -4.65270820E-12, -4.05567534E+04,
                    #1.92726029E+01]),
                #NASA([1400.00, 5000.00],
                    #[ 2.81439191E+01,  1.63524649E-02, -5.54998081E-06,
                    #8.58603978E-10, -4.97359401E-14, -4.84502814E+04,
                    #-1.11572783E+02])))

#species(name='CH2CQCOHQ',
        #atoms='C:3 H:6 O:5',
        #thermo=(NASA([300.00, 1418.00],
                    #[-1.61171759E+01,  1.78866440E-01, -2.16751308E-04,
                    #1.15450289E-07, -2.27163078E-11, -5.21165145E+04,
                    #1.14441286E+02]),
                #NASA([1418.00, 5000.00],
                    #[ 3.86574091E+01,  4.83815026E-04, -2.10413843E-07,
                    #3.81490832E-11, -2.46187754E-15, -6.56392184E+04,
                    #-1.61083579E+02])),
        #note='7/1/14')

#species(name='COHQCYC(COC)',
        #atoms='C:4 H:8 O:4',
        #thermo=(NASA([300.00, 1319.00],
                    #[ 2.40687943E+00,  5.95180442E-02, -3.16913659E-05,
                    #4.23694824E-09,  8.42033474E-13, -5.19694864E+04,
                    #1.88941416E+01]),
                #NASA([1319.00, 5000.00],
                    #[ 2.44599226E+01,  1.64187782E-02, -5.74024372E-06,
                    #9.05710407E-10, -5.31829152E-14, -6.01811228E+04,
                    #-1.02049722E+02])))

#species(name='QCYC(CCOC)OH',
        #atoms='C:4 H:8 O:4',
        #thermo=(NASA([300.00, 1411.00],
                    #[-2.09099972E+00,  8.02656739E-02, -6.69756245E-05,
                    #2.79282034E-08, -4.60981823E-12, -5.12529252E+04,
                    #4.11897260E+01]),
                #NASA([1411.00, 5000.00],
                    #[ 2.20860197E+01,  1.83743733E-02, -6.29478769E-06,
                    #9.78787994E-10, -5.68621317E-14, -5.89652037E+04,
                    #-8.64964494E+01])))

#species(name='TQC4H7OHTO2',
        #atoms='C:4 H:9 O:5',
        #thermo=(NASA([300.00, 1402.00],
                    #[ 3.17336206E+00,  7.94005900E-02, -6.51165712E-05,
                    #2.62035931E-08, -4.13406290E-12, -4.84943162E+04,
                    #1.77867184E+01]),
                #NASA([1402.00, 5000.00],
                    #[ 2.82564819E+01,  1.66969871E-02, -5.67314614E-06,
                    #8.78350442E-10, -5.09090253E-14, -5.66017464E+04,
                    #-1.15147927E+02])))

#species(name='HOCOCQ(CH3)2',
        #atoms='C:4 H:8 O:4',
        #thermo=(NASA([300.00, 1380.00],
                    #[ 1.56326363E+00,  6.77165643E-02, -5.14600149E-05,
                    #1.99184629E-08, -3.16007445E-12, -7.30943302E+04,
                    #2.37255421E+01]),
                #NASA([1380.00, 5000.00],
                    #[ 2.28935401E+01,  1.78627606E-02, -6.34627520E-06,
                    #1.01068546E-09, -5.96885712E-14, -8.05400680E+04,
                    #-9.08953779E+01])))

#species(name='IQC4H7OHTO2',
        #atoms='C:4 H:9 O:5',
        #thermo=(NASA([300.00, 1389.00],
                    #[ 3.83287364E+00,  6.90926177E-02, -5.15031144E-05,
                    #1.99235553E-08, -3.17456897E-12, -4.34442789E+04,
                    #1.94648622E+01]),
                #NASA([1389.00, 5000.00],
                    #[ 2.41720261E+01,  2.09397176E-02, -7.29061406E-06,
                    #1.14554056E-09, -6.70217517E-14, -5.04772118E+04,
                    #-8.95996685E+01])))

#species(name='TC3H6OH',
        #atoms='C:3 H:7 O:1',
        #thermo=(NASA([300.00, 1392.00],
                    #[ 1.09670360E+00,  3.80727565E-02, -2.75022497E-05,
                    #1.07477493E-08, -1.74895773E-12, -1.40764487E+04,
                    #2.22475799E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 1.12222277E+01,  1.36444398E-02, -4.51406709E-06,
                    #7.10523275E-10, -4.22690392E-14, -1.75350136E+04,
                    #-3.18911926E+01])),
        #note='8/9/4THERM')

#species(name='IQC4H7OHTQ-P',
        #atoms='C:4 H:9 O:5',
        #thermo=(NASA([300.00, 1391.00],
                    #[ 3.45242802E+00,  7.14541224E-02, -5.54616790E-05,
                    #2.22754676E-08, -3.66067450E-12, -3.54133076E+04,
                    #2.23024492E+01]),
                #NASA([1391.00, 5000.00],
                    #[ 2.45769593E+01,  2.01889393E-02, -7.03584537E-06,
                    #1.10623327E-09, -6.47523225E-14, -4.25783112E+04,
                    #-9.05052742E+01])))

#species(name='IQC4H8OTQ-I',
        #atoms='C:4 H:9 O:5',
        #thermo=(NASA([300.00, 1386.00],
                    #[ 5.90219800E+00,  6.26952142E-02, -4.17351838E-05,
                    #1.35422162E-08, -1.71169663E-12, -3.39620020E+04,
                    #7.89290087E+00]),
                #NASA([1386.00, 5000.00],
                    #[ 2.53249383E+01,  2.00634820E-02, -7.01249558E-06,
                    #1.10475030E-09, -6.47561724E-14, -4.09610593E+04,
                    #-9.73592779E+01])))

#species(name='CH2COHCH2OOH',
        #atoms='C:3 H:6 O:3',
        #thermo=(NASA([300.00, 1398.00],
                    #[-3.89823383E-01,  7.01531131E-02, -7.42036788E-05,
                    #3.84181056E-08, -7.63555985E-12, -3.07879938E+04,
                    #2.86873505E+01]),
                #NASA([1398.00, 5000.00],
                    #[ 1.87971268E+01,  1.12783442E-02, -3.90789058E-06,
                    #6.12064651E-10, -3.57305453E-14, -3.61154867E+04,
                    #-6.94914300E+01])))

#species(name='IC4H6Q2-II',
        #atoms='C:4 H:8 O:4',
        #thermo=(NASA([300.00, 1386.00],
                    #[ 2.16408482E-01,  8.25572149E-02, -7.64540445E-05,
                    #3.58211787E-08, -6.67718658E-12, -2.05694845E+04,
                    #3.36943223E+01]),
                #NASA([1386.00, 5000.00],
                    #[ 2.50360805E+01,  1.60230197E-02, -5.70704966E-06,
                    #9.10412038E-10, -5.38294659E-14, -2.84196401E+04,
                    #-9.67186452E+01])),
        #note='9/8/14')

#species(name='CO(CH2OOH)2',
        #atoms='C:3 H:6 O:5',
        #thermo=(NASA([300.00, 1393.00],
                    #[-2.47626577E+00,  8.93736793E-02, -9.25891121E-05,
                    #4.63168490E-08, -8.93300309E-12, -4.38924057E+04,
                    #4.84479477E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 2.43376341E+01,  1.14074110E-02, -4.08931881E-06,
                    #6.55183244E-10, -3.88570518E-14, -5.16862647E+04,
                    #-9.01518175E+01])))

#species(name='CHOC(CH3)OHCH2Q',
        #atoms='C:4 H:8 O:4',
        #thermo=(NASA([300.00, 1383.00],
                    #[ 2.40933530E+00,  6.13326038E-02, -4.03874986E-05,
                    #1.22738404E-08, -1.32997480E-12, -5.83049516E+04,
                    #2.29870742E+01]),
                #NASA([1383.00, 5000.00],
                    #[ 2.24480753E+01,  1.78753902E-02, -6.26904597E-06,
                    #9.90084046E-10, -5.81417382E-14, -6.55093075E+04,
                    #-8.56747327E+01])))

#species(name='CH3COCHO',
        #atoms='C:3 H:4 O:2',
        #thermo=(NASA([300.00, 1381.00],
                    #[ 2.08731049E+00,  3.09032484E-02, -1.98794164E-05,
                    #6.26174519E-09, -7.69945504E-13, -3.43989451E+04,
                    #1.82839639E+01]),
                #NASA([1381.00, 5000.00],
                    #[ 1.14371190E+01,  1.06773624E-02, -3.68967757E-06,
                    #5.77006752E-10, -3.36532201E-14, -3.78079398E+04,
                    #-3.25054087E+01])))

#species(name='SC4H7OH-I',
        #atoms='C:4 H:8 O:1',
        #thermo=(NASA([300.00, 1395.00],
                    #[ 2.70103499E+00,  4.17950180E-02, -2.67860575E-05,
                    #9.38191037E-09, -1.41171285E-12, -2.73561190E+04,
                    #1.35316306E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 1.30299481E+01,  1.83782479E-02, -6.18529878E-06,
                    #9.49578099E-10, -5.46526348E-14, -3.10723026E+04,
                    #-4.22891828E+01])),
        #note='L2/00')

#species(name='CCYCCC',
        #atoms='H:8 C:4',
        #thermo=(NASA([298.15, 1000.00],
                    #[-2.42295323E+00,  4.78931182E-02, -2.71594188E-05,
                    #4.04374801E-09,  1.34174013E-12,  1.74370161E+03,
                    #3.49485618E+01]),
                #NASA([1000.00, 2000.00],
                    #[ 3.30809027E+00,  3.21789790E-02, -1.44032621E-05,
                    #2.74967834E-09, -1.37251143E-13,  2.36991172E+02,
                    #5.49708116E+00])))

#species(name='H15DE25DM',
        #atoms='C:8 H:14',
        #thermo=(NASA([300.00, 1395.00],
                    #[-1.71853441E+00,  8.82613783E-02, -6.03140500E-05,
                    #2.15862289E-08, -3.19690882E-12, -1.95255748E+03,
                    #3.86048681E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 2.25355644E+01,  3.23955734E-02, -1.10270814E-05,
                    #1.70640907E-09, -9.87758061E-14, -1.04808866E+04,
                    #-9.19784905E+01])))

#species(name='H15DE25DM-S',
        #atoms='C:8 H:13',
        #thermo=(NASA([300.00, 1395.00],
                    #[-2.04235551E+00,  8.66261688E-02, -5.99110976E-05,
                    #2.15551630E-08, -3.18976697E-12,  1.50224279E+04,
                    #3.96917763E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 2.21422958E+01,  3.06966055E-02, -1.04617380E-05,
                    #1.62037747E-09, -9.38578511E-14,  6.56045671E+03,
                    #-9.04059585E+01])))

#species(name='H15DE25DM-A',
        #atoms='C:8 H:13',
        #thermo=(NASA([300.00, 1391.00],
                    #[-2.21863377E+00,  8.91652629E-02, -6.33360636E-05,
                    #2.32007379E-08, -3.46329567E-12,  1.61971179E+04,
                    #4.15116620E+01]),
                #NASA([1391.00, 5000.00],
                    #[ 2.33282646E+01,  2.93694154E-02, -9.94258906E-06,
                    #1.53368063E-09, -8.86030202E-14,  7.37319036E+03,
                    #-9.55902415E+01])))

#species(name='H15DE25DM-AO',
        #atoms='C:8 H:13 O:1',
        #thermo=(NASA([300.00, 1378.00],
                    #[ 3.03691957E+00,  7.50640438E-02, -4.17779534E-05,
                    #1.07915725E-08, -9.97734397E-13,  6.11649780E+03,
                    #2.04746031E+01]),
                #NASA([1378.00, 5000.00],
                    #[ 2.50985951E+01,  3.15085628E-02, -1.10076915E-05,
                    #1.73332426E-09, -1.01557468E-13, -2.49465338E+03,
                    #-1.01097942E+02])))

#species(name='H15DE2M-T',
        #atoms='C:7 H:11',
        #thermo=(NASA([300.00, 1389.00],
                    #[ 3.88856066E-01,  6.70215700E-02, -4.29671268E-05,
                    #1.42489875E-08, -1.96105571E-12,  3.03627273E+04,
                    #2.95444186E+01]),
                #NASA([1389.00, 5000.00],
                    #[ 1.90144729E+01,  2.61932198E-02, -9.01183878E-06,
                    #1.40455988E-09, -8.17075251E-14,  2.35551101E+04,
                    #-7.15568628E+01])))

#species(name='H15DE25DM-SO',
        #atoms='C:8 H:13 O:1',
        #thermo=(NASA([300.00, 1388.00],
                    #[ 9.30464015E-01,  8.71669873E-02, -5.99582593E-05,
                    #2.15026041E-08, -3.21687573E-12,  4.19310073E+03,
                    #2.90924508E+01]),
                #NASA([1388.00, 5000.00],
                    #[ 2.60152179E+01,  3.04295766E-02, -1.05676513E-05,
                    #1.65746547E-09, -9.68470091E-14, -4.79568520E+03,
                    #-1.06442472E+02])))

#species(name='IC4H7CHO',
        #atoms='C:5 H:8 O:1',
        #thermo=(NASA([300.00, 1391.00],
                    #[-1.20982776E+00,  5.92603375E-02, -4.26960089E-05,
                    #1.60356164E-08, -2.49347521E-12, -1.56205327E+04,
                    #3.53137305E+01]),
                #NASA([1391.00, 5000.00],
                    #[ 1.59171638E+01,  1.93357284E-02, -6.70857943E-06,
                    #1.05155191E-09, -6.14175906E-14, -2.16140735E+04,
                    #-5.67616631E+01])))

#species(name='IC4H7O2',
        #atoms='C:4 H:7 O:2',
        #thermo=(NASA([300.00, 1404.00],
                    #[ 1.43532045E+00,  4.89026570E-02, -3.63600970E-05,
                    #1.41906420E-08, -2.24557878E-12,  3.09284623E+03,
                    #2.41320196E+01]),
                #NASA([1404.00, 5000.00],
                    #[ 1.45791608E+01,  1.62136068E-02, -5.26957103E-06,
                    #7.90454323E-10, -4.47755051E-14, -1.20848042E+03,
                    #-4.56459433E+01])),
        #note='L2/00')

#species(name='IC4H6OOH-I',
        #atoms='C:4 H:7 O:2',
        #thermo=(NASA([300.00, 1398.00],
                    #[ 6.06669321E+00,  3.94950065E-02, -2.52721718E-05,
                    #7.84485641E-09, -8.98876886E-13,  2.87448422E+03,
                    #-3.78377026E-01]),
                #NASA([1398.00, 5000.00],
                    #[ 1.73601429E+01,  1.42046196E-02, -4.65348263E-06,
                    #7.02209785E-10, -3.99552775E-14, -1.07509366E+03,
                    #-6.12822377E+01])),
        #note='L2/00')

#species(name='C*CYCCOC',
        #atoms='C:4 H:6 O:1',
        #thermo=(NASA([300.00, 1397.00],
                    #[-2.45566808E+00,  4.88377067E-02, -3.46186324E-05,
                    #1.26590643E-08, -1.88737988E-12, -6.43460880E+02,
                    #3.58434061E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 1.14363523E+01,  1.63250968E-02, -5.57146218E-06,
                    #8.63795613E-10, -5.00701899E-14, -5.44653450E+03,
                    #-3.87179379E+01])),
        #note='L2/00')

#species(name='CCYCCOOC-T1',
        #atoms='C:4 H:7 O:2',
        #thermo=(NASA([300.00, 1394.00],
                    #[-5.29767923E+00,  6.65082201E-02, -4.67054235E-05,
                    #1.51029775E-08, -1.74322091E-12,  3.97935712E+03,
                    #5.16412476E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 1.68269657E+01,  1.64471921E-02, -5.63767184E-06,
                    #8.78001611E-10, -5.10959632E-14, -3.65710841E+03,
                    #-6.74096786E+01])),
        #note='THERM')

#species(name='CCYCCOOC-I2',
        #atoms='C:4 H:7 O:2',
        #thermo=(NASA([300.00, 1398.00],
                    #[-4.55772603E+00,  7.10221387E-02, -6.05228430E-05,
                    #2.61773944E-08, -4.51390528E-12,  1.32922215E+04,
                    #4.60828737E+01]),
                #NASA([1398.00, 5000.00],
                    #[ 1.63388791E+01,  1.65879474E-02, -5.61829108E-06,
                    #8.67331033E-10, -5.01476782E-14,  6.66946988E+03,
                    #-6.40322040E+01])),
        #note='L2/00')

#species(name='CHOIC3H6O',
        #atoms='C:4 H:7 O:2',
        #thermo=(NASA([300.00, 1386.00],
                    #[ 2.55559437E-01,  5.22086026E-02, -3.72283766E-05,
                    #1.36714420E-08, -2.05638885E-12, -1.97284649E+04,
                    #2.99210251E+01]),
                #NASA([1386.00, 5000.00],
                    #[ 1.55511970E+01,  1.67360034E-02, -5.73573457E-06,
                    #8.92095191E-10, -5.18351848E-14, -2.50674335E+04,
                    #-5.23215658E+01])),
        #note='L2/00')

#species(name='CCY(C2O)CO',
        #atoms='C:4 H:7 O:2',
        #thermo=(NASA([300.00, 1401.00],
                    #[-2.97381263E+00,  6.75615160E-02, -5.79901709E-05,
                    #2.55213006E-08, -4.49991674E-12, -9.93607156E+03,
                    #4.25127997E+01]),
                #NASA([1401.00, 5000.00],
                    #[ 1.64352871E+01,  1.65843210E-02, -5.71790511E-06,
                    #8.92586562E-10, -5.19865043E-14, -1.60938498E+04,
                    #-5.96946106E+01])),
        #note='L2/00')

#species(name='CCYCCO-T1',
        #atoms='C:3 H:5 O:1',
        #thermo=(NASA([300.00, 1389.00],
                    #[-1.39311392E+00,  3.81789194E-02, -2.62316288E-05,
                    #8.74877238E-09, -1.11296763E-12,  1.12534953E+04,
                    #3.41877003E+01]),
                #NASA([1389.00, 5000.00],
                    #[ 1.03394781E+01,  1.15180335E-02, -3.87496644E-06,
                    #5.95744364E-10, -3.43539552E-14,  7.17658970E+03,
                    #-2.89687593E+01])),
        #note='L2/00')

#species(name='C2CYCOOC-I1',
        #atoms='C:4 H:7 O:2',
        #thermo=(NASA([300.00, 1388.00],
                    #[-2.04718031E+00,  7.26608379E-02, -6.82960279E-05,
                    #3.27505762E-08, -6.23802161E-12, -3.71475599E+02,
                    #1.92056057E+01]),
                #NASA([1388.00, 5000.00],
                    #[ 1.89745085E+01,  1.50113808E-02, -5.30728309E-06,
                    #8.42454740E-10, -4.96392976E-14, -6.93111358E+03,
                    #-9.08581019E+01])),
        #note='7/14')

#species(name='IC3H5OOCH2',
        #atoms='C:4 H:7 O:2',
        #thermo=(NASA([300.00, 1410.00],
                    #[-1.26548168E+00, -3.61337863E-03,  7.31803998E-05,
                    #-5.49845697E-08,  1.20447492E-11,  1.33108969E+04,
                    #5.33428723E+01]),
                #NASA([1410.00, 5000.00],
                    #[ 5.05335676E+00,  3.62904388E-02, -1.54012992E-05,
                    #2.74341619E-09, -1.74718602E-13,  4.08326144E+03,
                    #-2.55389509E+00])),
        #note='L2/00')

#species(name='B2E3M1OJ',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 2003.00],
                    #[ 5.64762692E+00,  3.43609862E-02, -5.60649548E-06,
                    #-5.19111402E-09,  1.73155461E-12, -6.65737793E+00,
                    #3.13825267E+00]),
                #NASA([2003.00, 5000.00],
                    #[ 1.35666458E+01,  2.56547934E-02, -9.37226256E-06,
                    #1.52518398E-09, -9.15378763E-14, -3.58767067E+03,
                    #-4.27272791E+01])))

#species(name='B13DE2M',
        #atoms='C:5 H:8',
        #thermo=(NASA([300.00, 1397.00],
                    #[-4.72170009E-01,  5.59413257E-02, -4.50458826E-05,
                    #1.96181377E-08, -3.52009647E-12,  7.14960489E+03,
                    #2.56276807E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 1.40757545E+01,  1.82375566E-02, -6.20747687E-06,
                    #9.60350919E-10, -5.55741525E-14,  2.40033881E+03,
                    #-5.12988284E+01])))

#species(name='B13DE2MJ',
        #atoms='C:5 H:7',
        #thermo=(NASA([300.00, 1393.00],
                    #[-8.98037227E-01,  5.65174321E-02, -4.76107829E-05,
                    #2.09672106E-08, -3.73076319E-12,  2.52881763E+04,
                    #2.74966602E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 1.48413344E+01,  1.52425753E-02, -5.13559858E-06,
                    #7.89787260E-10, -4.55353492E-14,  2.02661570E+04,
                    #-5.54474955E+01])))

#species(name='B2E2M1OJ',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 2003.00],
                    #[ 5.64762692E+00,  3.43609862E-02, -5.60649548E-06,
                    #-5.19111402E-09,  1.73155461E-12, -6.65737793E+00,
                    #3.13825267E+00]),
                #NASA([2003.00, 5000.00],
                    #[ 1.35666458E+01,  2.56547934E-02, -9.37226256E-06,
                    #1.52518398E-09, -9.15378763E-14, -3.58767067E+03,
                    #-4.27272791E+01])))

#species(name='B12DE3M',
        #atoms='C:5 H:8',
        #thermo=(NASA([300.00, 1388.00],
                    #[ 1.27860173E+00,  4.54917474E-02, -2.82334169E-05,
                    #8.98449372E-09, -1.17220042E-12,  1.31637341E+04,
                    #1.87039373E+01]),
                #NASA([1388.00, 5000.00],
                    #[ 1.37093177E+01,  1.85726726E-02, -6.33115904E-06,
                    #9.80719059E-10, -5.68099894E-14,  8.59518752E+03,
                    #-4.88749621E+01])),
        #note='11/12/12THERM')

#species(name='IC4H9O2',
        #atoms='C:4 H:9 O:2',
        #thermo=(NASA([300.00, 1432.00],
                    #[ 1.77219624E+00,  5.34032789E-02, -3.31041810E-05,
                    #9.24465657E-09, -8.01706642E-13, -1.17768774E+04,
                    #2.09581481E+01]),
                #NASA([1432.00, 5000.00],
                    #[ 1.78793870E+01,  1.82474607E-02, -6.01252193E-06,
                    #9.11106794E-10, -5.20018932E-14, -1.74569774E+04,
                    #-6.61552973E+01])),
        #note='9/1/12')

#species(name='TC4H8O2H-I',
        #atoms='C:4 H:9 O:2',
        #thermo=(NASA([300.00, 1379.00],
                    #[ 3.54378349E+00,  5.25201369E-02, -3.69898493E-05,
                    #1.44634925E-08, -2.47536050E-12, -6.98183185E+03,
                    #1.27623539E+01]),
                #NASA([1379.00, 5000.00],
                    #[ 1.81415374E+01,  1.94699499E-02, -6.82750014E-06,
                    #1.07773311E-09, -6.32519099E-14, -1.23570939E+04,
                    #-6.63491602E+01])),
        #note='9/1/12')

#species(name='IC4H8O2H-I',
        #atoms='C:4 H:9 O:2',
        #thermo=(NASA([300.00, 1414.00],
                    #[ 1.86432620E-01,  6.26430177E-02, -4.83690886E-05,
                    #1.88657148E-08, -2.91189385E-12, -3.59086611E+03,
                    #2.97635367E+01]),
                #NASA([1414.00, 5000.00],
                    #[ 1.83915486E+01,  1.73042831E-02, -5.66841018E-06,
                    #8.55414265E-10, -4.86781778E-14, -9.48569748E+03,
                    #-6.67673286E+01])),
        #note='9/1/12')

#species(name='IC4H8O2H-T',
        #atoms='C:4 H:9 O:2',
        #thermo=(NASA([300.00, 1413.00],
                    #[ 3.84374544E+00,  4.36800978E-02, -2.07599526E-05,
                    #2.51709167E-09,  5.41306513E-13, -6.50766215E+03,
                    #1.34244877E+01]),
                #NASA([1413.00, 5000.00],
                    #[ 1.69753885E+01,  1.85198010E-02, -6.09075415E-06,
                    #9.21673609E-10, -5.25502501E-14, -1.14812757E+04,
                    #-5.88259039E+01])),
        #note='9/1/12')

#species(name='IC4H8O',
        #atoms='C:4 H:8 O:1',
        #thermo=(NASA([300.00, 1394.00],
                    #[-5.02573822E+00,  7.51340960E-02, -6.88668822E-05,
                    #3.12223247E-08, -5.60128818E-12, -3.07481413E+04,
                    #2.96284295E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 1.40433578E+01,  2.05733637E-02, -9.09519220E-06,
                    #1.73417298E-09, -1.14908544E-13, -3.62275308E+04,
                    #-6.90009668E+01])),
        #note='9/1/12')

#species(name='CC4H8O',
        #atoms='C:4 H:8 O:1',
        #thermo=(NASA([300.00, 1431.00],
                    #[-6.56746688E+00,  7.87298554E-02, -7.33065478E-05,
                    #3.40602701E-08, -6.15674656E-12, -2.71582518E+04,
                    #3.80875851E+01]),
                #NASA([1431.00, 5000.00],
                    #[ 1.51841776E+01,  1.64656666E-02, -5.33483091E-06,
                    #7.98149768E-10, -4.51160381E-14, -3.33923434E+04,
                    #-7.43746988E+01])),
        #note='9/1/12')

#species(name='IC4H9O',
        #atoms='C:4 H:9 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 3.80297372E+00,  1.56874209E-02,  6.81105412E-05,
                    #-9.83346774E-08,  3.95261902E-11, -1.00832243E+04,
                    #9.78963305E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 1.16309708E+01,  2.47981574E-02, -9.01550536E-06,
                    #1.46714720E-09, -8.83214518E-14, -1.37854612E+04,
                    #-3.81956151E+01])),
        #note='A08/04')

#species(name='TC4H9O',
        #atoms='C:4 H:9 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 2.77057100E+00,  2.68033175E-02,  4.12718360E-05,
                    #-7.22054739E-08,  3.02642276E-11, -1.27079262E+04,
                    #1.21532856E+01]),
                #NASA([1000.00, 6000.00],
                    #[ 1.27371509E+01,  2.33707342E-02, -8.50516678E-06,
                    #1.38519973E-09, -8.34398061E-14, -1.66940150E+04,
                    #-4.53156321E+01])),
        #note='T08/04')

#species(name='IC4H9O2H',
        #atoms='C:4 H:10 O:2',
        #thermo=(NASA([300.00, 1402.00],
                    #[-5.47969191E-01,  6.57428787E-02, -4.70123149E-05,
                    #1.65991096E-08, -2.27331437E-12, -2.82180462E+04,
                    #3.12961385E+01]),
                #NASA([1402.00, 5000.00],
                    #[ 1.91651624E+01,  1.93678648E-02, -6.41984435E-06,
                    #9.76947752E-10, -5.59299558E-14, -3.48790978E+04,
                    #-7.42063787E+01])),
        #note='9/1/12')

#species(name='TC4H9O2H',
        #atoms='C:4 H:10 O:2',
        #thermo=(NASA([300.00, 1382.00],
                    #[ 4.45573540E-01,  6.66153523E-02, -5.20932123E-05,
                    #2.22301799E-08, -4.00189859E-12, -3.12260714E+04,
                    #2.37278262E+01]),
                #NASA([1382.00, 5000.00],
                    #[ 1.90926853E+01,  2.12697804E-02, -7.46252626E-06,
                    #1.17841127E-09, -6.91795087E-14, -3.77278405E+04,
                    #-7.61321196E+01])),
        #note='9/1/12')

#species(name='IC4H7O',
        #atoms='C:4 H:7 O:1',
        #thermo=(NASA([300.00, 1386.00],
                    #[ 1.74700687E+00,  4.07783436E-02, -2.44750243E-05,
                    #7.06502958E-09, -7.51570589E-13,  4.86979233E+03,
                    #1.94535999E+01]),
                #NASA([1386.00, 5000.00],
                    #[ 1.33457615E+01,  1.61218588E-02, -5.44376403E-06,
                    #8.38199374E-10, -4.83608280E-14,  6.11443644E+02,
                    #-4.36818838E+01])),
        #note='4/3/0THERM')

#species(name='IC4H8OH',
        #atoms='C:4 H:9 O:1',
        #thermo=(NASA([300.00, 1376.00],
                    #[ 3.29612707E+00,  3.47649647E-02, -1.02505618E-05,
                    #-2.04641931E-09,  1.18879408E-12, -1.45627247E+04,
                    #1.58606320E+01]),
                #NASA([1376.00, 5000.00],
                    #[ 1.25605997E+01,  2.10637488E-02, -7.15019648E-06,
                    #1.10439262E-09, -6.38428695E-14, -1.86203249E+04,
                    #-3.67889430E+01])),
        #note='2/14/95THERM')

#species(name='IC3H7CHO',
        #atoms='C:4 H:8 O:1',
        #thermo=(NASA([300.00, 1391.00],
                    #[-2.73021382E-01,  4.89696307E-02, -3.12770049E-05,
                    #1.00052945E-08, -1.27512074E-12, -2.76054737E+04,
                    #2.83451139E+01]),
                #NASA([1391.00, 5000.00],
                    #[ 1.37501656E+01,  1.83126722E-02, -6.28572629E-06,
                    #9.78250756E-10, -5.68538653E-14, -3.26936771E+04,
                    #-4.77270548E+01])),
        #note='2/22/96THERM')

#species(name='TC3H6CHO',
        #atoms='C:4 H:7 O:1',
        #thermo=(NASA([300.00, 1389.00],
                    #[ 1.87052762E+00,  4.14869677E-02, -2.66815701E-05,
                    #9.01531610E-09, -1.27870633E-12, -8.97730744E+03,
                    #1.66174178E+01]),
                #NASA([1389.00, 5000.00],
                    #[ 1.31013047E+01,  1.66391865E-02, -5.68457623E-06,
                    #8.81808351E-10, -5.11290161E-14, -1.30638647E+04,
                    #-4.42705813E+01])),
        #note='2/22/96THERM')

#species(name='IC3H7CO',
        #atoms='C:4 H:7 O:1',
        #thermo=(NASA([300.00, 1390.00],
                    #[ 5.03452639E-01,  4.41607510E-02, -2.82139091E-05,
                    #8.93548675E-09, -1.11327422E-12, -9.07755468E+03,
                    #2.61991461E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 1.33305736E+01,  1.61873930E-02, -5.56711402E-06,
                    #8.67575951E-10, -5.04696549E-14, -1.37307001E+04,
                    #-4.33958746E+01])),
        #note='2/22/96THERM')

#species(name='IC3H6CHO',
        #atoms='C:4 H:7 O:1',
        #thermo=(NASA([300.00, 1390.00],
                    #[ 5.21481767E-01,  4.43114357E-02, -2.86617314E-05,
                    #9.30319894E-09, -1.20761563E-12, -2.99677086E+03,
                    #2.68182130E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 1.33102250E+01,  1.62097959E-02, -5.57575891E-06,
                    #8.69003718E-10, -5.05554202E-14, -7.62177931E+03,
                    #-4.25050854E+01])),
        #note='2/22/96THERM')

#species(name='TC4H8OOH-IO2',
        #atoms='C:4 H:9 O:4',
        #thermo=(NASA([300.00, 1385.00],
                    #[ 3.36413530E+00,  6.93742776E-02, -5.70416393E-05,
                    #2.46040165E-08, -4.32848680E-12, -2.51137558E+04,
                    #1.65767339E+01]),
                #NASA([1385.00, 5000.00],
                    #[ 2.32464612E+01,  1.88384513E-02, -6.40938087E-06,
                    #9.92649459E-10, -5.75275879E-14, -3.16533132E+04,
                    #-8.88301710E+01])),
        #note='9/1/12')

#species(name='IC4H8OOH-IO2',
        #atoms='C:4 H:9 O:4',
        #thermo=(NASA([300.00, 1367.00],
                    #[ 4.23354857E+00,  5.63088857E-02, -3.15672522E-05,
                    #7.79536931E-09, -6.21665008E-13, -2.22782534E+04,
                    #1.52623111E+01]),
                #NASA([1367.00, 5000.00],
                    #[ 2.24664901E+01,  2.09351287E-02, -7.44324128E-06,
                    #1.18589255E-09, -7.00546897E-14, -2.94495457E+04,
                    #-8.54241451E+01])),
        #note='9/1/12')

#species(name='IC4H8OOH-TO2',
        #atoms='C:4 H:9 O:4',
        #thermo=(NASA([300.00, 1385.00],
                    #[ 3.36413530E+00,  6.93742776E-02, -5.70416393E-05,
                    #2.46040165E-08, -4.32848680E-12, -2.51137558E+04,
                    #1.65767339E+01]),
                #NASA([1385.00, 5000.00],
                    #[ 2.32464612E+01,  1.88384513E-02, -6.40938087E-06,
                    #9.92649459E-10, -5.75275879E-14, -3.16533132E+04,
                    #-8.88301710E+01])),
        #note='9/1/12')

#species(name='IC4KETII',
        #atoms='C:4 H:8 O:3',
        #thermo=(NASA([300.00, 1387.00],
                    #[ 1.15501614E+00,  6.10622345E-02, -4.49711323E-05,
                    #1.70514654E-08, -2.65948602E-12, -3.82747956E+04,
                    #2.69612235E+01]),
                #NASA([1387.00, 5000.00],
                    #[ 1.95143059E+01,  1.82377395E-02, -6.38908606E-06,
                    #1.00801571E-09, -5.91440350E-14, -4.46884836E+04,
                    #-7.17167584E+01])),
        #note='7/19/0THERM')

#species(name='IC4KETIT',
        #atoms='C:4 H:8 O:3',
        #thermo=(NASA([300.00, 1388.00],
                    #[ 1.14243741E+00,  6.33840797E-02, -4.73084738E-05,
                    #1.77145373E-08, -2.67265475E-12, -4.09366796E+04,
                    #2.34844867E+01]),
                #NASA([1388.00, 5000.00],
                    #[ 2.09369850E+01,  1.71090955E-02, -6.01892169E-06,
                    #9.52353863E-10, -5.59926176E-14, -4.77819819E+04,
                    #-8.27717611E+01])),
        #note='7/19/0THERM')

#species(name='IC4H7OH',
        #atoms='C:4 H:8 O:1',
        #thermo=(NASA([300.00, 1398.00],
                    #[ 2.04124240E+00,  4.14387207E-02, -2.55228632E-05,
                    #8.28133017E-09, -1.10654457E-12, -2.22709637E+04,
                    #1.88699473E+01]),
                #NASA([1398.00, 5000.00],
                    #[ 1.23304221E+01,  1.83885172E-02, -6.06721733E-06,
                    #9.19054723E-10, -5.24036171E-14, -2.59452023E+04,
                    #-3.67418286E+01])))

#species(name='IC4H6OH',
        #atoms='C:4 H:7 O:1',
        #thermo=(NASA([300.00, 1390.00],
                    #[ 8.63371227E-01,  4.68711282E-02, -3.43580339E-05,
                    #1.33031052E-08, -2.13914975E-12, -3.14948305E+03,
                    #2.29075523E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 1.40310926E+01,  1.55317541E-02, -5.32754961E-06,
                    #8.28785902E-10, -4.81545257E-14, -7.69378228E+03,
                    #-4.76555306E+01])),
        #note='8/19/95THERM')

#species(name='IC3H5CHO',
        #atoms='C:4 H:6 O:1',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 1.09372823E+00,  4.43315368E-02, -3.41918451E-05,
                    #1.39369607E-08, -2.33791460E-12, -1.56745978E+04,
                    #1.94458467E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 1.33892118E+01,  1.39115420E-02, -4.75820958E-06,
                    #7.38736618E-10, -4.28606559E-14, -1.97917448E+04,
                    #-4.60146004E+01])))

#species(name='IC3H5CO',
        #atoms='C:4 H:5 O:1',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 1.87306990E+00,  3.95188508E-02, -3.11404053E-05,
                    #1.28844447E-08, -2.18165308E-12,  2.85270691E+03,
                    #1.68774016E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 1.29634401E+01,  1.17954996E-02, -4.04361488E-06,
                    #6.28771516E-10, -3.65209867E-14, -8.26519462E+02,
                    #-4.20562575E+01])))

#species(name='TC3H6OCHO',
        #atoms='C:4 H:7 O:2',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 3.70830259E-01,  5.38475661E-02, -3.82477565E-05,
                    #1.32882237E-08, -1.79228730E-12, -2.18391262E+04,
                    #2.58142112E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 1.70371287E+01,  1.54400645E-02, -5.28332886E-06,
                    #8.21085347E-10, -4.76898429E-14, -2.75871941E+04,
                    #-6.37271230E+01])),
        #note='8/25/95THERM')

#species(name='IC3H6CO',
        #atoms='C:4 H:6 O:1',
        #thermo=(NASA([300.00, 1397.00],
                    #[ 2.28039055E+00,  4.17016989E-02, -3.25089661E-05,
                    #1.37243419E-08, -2.40573132E-12, -1.63939712E+04,
                    #1.38187714E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 1.32548232E+01,  1.40142787E-02, -4.78910215E-06,
                    #7.42924342E-10, -4.30737566E-14, -2.00529779E+04,
                    #-4.44810221E+01])),
        #note='03/03/95THERM')

#species(name='IC4H7OOH',
        #atoms='C:4 H:8 O:2',
        #thermo=(NASA([300.00, 1386.00],
                    #[ 1.31851762E-01,  6.19561224E-02, -4.99343877E-05,
                    #2.09628211E-08, -3.59717924E-12, -1.21399925E+04,
                    #2.93905962E+01]),
                #NASA([1386.00, 5000.00],
                    #[ 1.82897194E+01,  1.67815784E-02, -5.80668193E-06,
                    #9.08949180E-10, -5.30513302E-14, -1.82046522E+04,
                    #-6.72111342E+01])),
        #note='4/15/15')

#species(name='TC3H6OHCHO',
        #atoms='C:4 H:8 O:2',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 2.10033500E-01,  5.63086284E-02, -3.95158669E-05,
                    #1.37780428E-08, -1.89068805E-12, -4.80402638E+04,
                    #2.67836070E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 1.70788996E+01,  1.74005554E-02, -5.93008112E-06,
                    #9.18860909E-10, -5.32512140E-14, -5.38730427E+04,
                    #-6.38638754E+01])),
        #note='8/1/95THERM')

#species(name='C3H6OH2-1',
        #atoms='C:3 H:7 O:1',
        #thermo=(NASA([300.00, 1392.00],
                    #[ 1.09670360E+00,  3.80727565E-02, -2.75022497E-05,
                    #1.07477493E-08, -1.74895773E-12, -1.40764487E+04,
                    #2.22475799E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 1.12222277E+01,  1.36444398E-02, -4.51406709E-06,
                    #7.10523275E-10, -4.22690392E-14, -1.75350136E+04,
                    #-3.18911926E+01])),
        #note='8/9/4THERM')

#species(name='IC3H5OH',
        #atoms='C:3 H:6 O:1',
        #thermo=(NASA([300.00, 1374.00],
                    #[ 1.58376391E+00,  3.16215366E-02, -1.73664942E-05,
                    #4.18927663E-09, -2.79899620E-13, -2.12643496E+04,
                    #1.88313766E+01]),
                #NASA([1374.00, 5000.00],
                    #[ 1.07381025E+01,  1.31698194E-02, -4.41529622E-06,
                    #6.77009837E-10, -3.89608901E-14, -2.47298321E+04,
                    #-3.13634050E+01])),
        #note='8/1/95THERM')

#species(name='TC3H6O2CHO',
        #atoms='C:4 H:7 O:3',
        #thermo=(NASA([300.00, 1386.00],
                    #[ 2.17883383E+00,  5.41595832E-02, -3.83435886E-05,
                    #1.38308104E-08, -2.04190147E-12, -2.27394154E+04,
                    #2.00751264E+01]),
                #NASA([1386.00, 5000.00],
                    #[ 1.85534443E+01,  1.68774389E-02, -5.90752965E-06,
                    #9.31518085E-10, -5.46345187E-14, -2.85447191E+04,
                    #-6.82486667E+01])),
        #note='8/2/95THERM')

#species(name='TC3H6O2HCO',
        #atoms='C:4 H:7 O:3',
        #thermo=(NASA([300.00, 1387.00],
                    #[ 2.03864428E+00,  5.80421003E-02, -4.32123528E-05,
                    #1.58792094E-08, -2.32209543E-12, -2.24284673E+04,
                    #2.03680990E+01]),
                #NASA([1387.00, 5000.00],
                    #[ 2.06472678E+01,  1.48526500E-02, -5.25104875E-06,
                    #8.33619219E-10, -4.91256069E-14, -2.88719869E+04,
                    #-7.95951389E+01])),
        #note='8/2/95THERM')

#species(name='IC3H5O2HCHO',
        #atoms='C:4 H:7 O:3',
        #thermo=(NASA([300.00, 1387.00],
                    #[ 2.05984770E+00,  5.82331716E-02, -4.37672100E-05,
                    #1.63249918E-08, -2.43462051E-12, -1.63496250E+04,
                    #2.13687921E+01]),
                #NASA([1387.00, 5000.00],
                    #[ 2.06288832E+01,  1.48625539E-02, -5.25305276E-06,
                    #8.33772951E-10, -4.91277401E-14, -2.27589076E+04,
                    #-7.82962888E+01])),
        #note='8/2/95THERM')

#species(name='CH2CCH2OH',
        #atoms='C:3 H:5 O:1',
        #thermo=(NASA([300.00, 1372.00],
                    #[ 2.88422544E+00,  2.42428071E-02, -1.14152268E-05,
                    #1.71775334E-09,  1.42177454E-13,  1.17935615E+04,
                    #1.52102335E+01]),
                #NASA([1372.00, 5000.00],
                    #[ 9.70702027E+00,  1.13972660E-02, -3.77993962E-06,
                    #5.75209277E-10, -3.29229125E-14,  9.13212884E+03,
                    #-2.25012933E+01])),
        #note='9/8/95THERM')

#species(name='TC4H8CHO',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1397.00],
                    #[-9.58078294E-01,  6.42003258E-02, -4.70776827E-05,
                    #1.75737698E-08, -2.64896151E-12, -6.86582501E+03,
                    #3.33781112E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 1.79663933E+01,  1.94207117E-02, -6.67409451E-06,
                    #1.03969221E-09, -6.04702651E-14, -1.33368585E+04,
                    #-6.79819424E+01])),
        #note='9/7/95THERM')

#species(name='O2C4H8CHO',
        #atoms='C:5 H:9 O:3',
        #thermo=(NASA([300.00, 1395.00],
                    #[ 1.91847699E+00,  6.67245869E-02, -4.80871046E-05,
                    #1.78588690E-08, -2.71163880E-12, -2.49837984E+04,
                    #2.38577867E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 2.12629904E+01,  2.14072282E-02, -7.38342949E-06,
                    #1.15281523E-09, -6.71508438E-14, -3.16854524E+04,
                    #-7.99828703E+01])),
        #note='9/7/95THERM')

#species(name='O2HC4H8CO',
        #atoms='C:5 H:9 O:3',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 1.82607262E+00,  6.93466111E-02, -4.93125140E-05,
                    #1.69848340E-08, -2.26117657E-12, -2.46578311E+04,
                    #2.41167544E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 2.38219630E+01,  1.91411448E-02, -6.67919154E-06,
                    #1.05127303E-09, -6.15876805E-14, -3.23093973E+04,
                    #-9.42580755E+01])),
        #note='9/7/95THERM')

#species(name='C3H5OH',
        #atoms='C:3 H:6 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 3.15011905E+00,  1.28538274E-02,  4.28438434E-05,
                    #-6.67818707E-08,  2.80408237E-11, -1.66413668E+04,
                    #1.35066359E+01]),
                #NASA([1000.00, 6000.00],
                    #[ 8.72477114E+00,  1.63942712E-02, -5.90852993E-06,
                    #9.53262253E-10, -5.70318010E-14, -1.90496618E+04,
                    #-1.97198674E+01])),
        #note='T06/10')

#species(name='TIC4H7Q2-I',
        #atoms='C:4 H:9 O:4',
        #thermo=(NASA([300.00, 1400.00],
                    #[ 4.48426361E+00,  6.61225007E-02, -5.27349018E-05,
                    #2.18215585E-08, -3.66788946E-12, -1.98906586E+04,
                    #1.26719614E+01]),
                #NASA([1400.00, 5000.00],
                    #[ 2.33848631E+01,  1.87070035E-02, -6.44021945E-06,
                    #1.00428123E-09, -5.84468189E-14, -2.61180902E+04,
                    #-8.76610135E+01])),
        #note='5/6/96THERM')

#species(name='IIC4H7Q2-T',
        #atoms='C:4 H:9 O:4',
        #thermo=(NASA([300.00, 1377.00],
                    #[ 8.16274487E+00,  4.34463050E-02, -1.76972456E-05,
                    #4.88790666E-10,  9.03915465E-13, -1.96501749E+04,
                    #2.62067299E-01]),
                #NASA([1377.00, 5000.00],
                    #[ 2.15070321E+01,  2.05359839E-02, -7.12383399E-06,
                    #1.11655053E-09, -6.52112103E-14, -2.51117508E+04,
                    #-7.43379783E+01])),
        #note='7/15/96THERM')

#species(name='IIC4H7Q2-I',
        #atoms='C:4 H:9 O:4',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 4.93055661E+00,  6.05819201E-02, -4.23665566E-05,
                    #1.49122008E-08, -2.10978665E-12, -1.68415495E+04,
                    #1.36228018E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 2.30500244E+01,  1.92149194E-02, -6.66622576E-06,
                    #1.04495725E-09, -6.10370520E-14, -2.32086881E+04,
                    #-8.39949885E+01])),
        #note='7/15/96THERM')

#species(name='CH2O2H',
        #atoms='C:1 H:3 O:2',
        #thermo=(NASA([300.00, 1410.00],
                    #[ 2.88976454E+00,  2.09465776E-02, -1.75190772E-05,
                    #7.27819787E-09, -1.18912344E-12,  6.12390620E+03,
                    #1.23802076E+01]),
                #NASA([1410.00, 5000.00],
                    #[ 9.24697852E+00,  4.60845541E-03, -1.53501472E-06,
                    #2.34434830E-10, -1.34573106E-14,  4.11529953E+03,
                    #-2.11503248E+01])),
        #note='9/1/12')

#species(name='C4H4',
        #atoms='C:4 H:4',
        #thermo=(NASA([300.00, 1000.00],
                    #[-1.91524790E+00,  5.27508780E-02, -7.16559440E-05,
                    #5.50724230E-08, -1.72862280E-11,  3.29785040E+04,
                    #3.14199830E+01]),
                #NASA([1000.00, 3000.00],
                    #[ 6.65070920E+00,  1.61294340E-02, -7.19388750E-06,
                    #1.49817870E-09, -1.18641100E-13,  3.11959920E+04,
                    #-9.79521180E+00])),
        #note='H6W/94')

#species(name='C4H3-I',
        #atoms='C:4 H:3',
        #thermo=(NASA([300.00, 1000.00],
                    #[ 2.08304120E+00,  4.08342740E-02, -6.21596850E-05,
                    #5.16793580E-08, -1.70291840E-11,  5.80051290E+04,
                    #1.36174620E+01]),
                #NASA([1000.00, 3000.00],
                    #[ 9.09781650E+00,  9.22071190E-03, -3.38784410E-06,
                    #4.91604980E-10, -1.45297800E-14,  5.66005740E+04,
                    #-1.98025970E+01])),
        #note='AB1/93')

#species(name='C4H3-N',
        #atoms='C:4 H:3',
        #thermo=(NASA([300.00, 1000.00],
                    #[-3.16841130E-01,  4.69121000E-02, -6.80938100E-05,
                    #5.31799210E-08, -1.65230050E-11,  6.24761990E+04,
                    #2.46225590E+01]),
                #NASA([1000.00, 3000.00],
                    #[ 5.43282790E+00,  1.68609810E-02, -9.43131090E-06,
                    #2.57038950E-09, -2.74563090E-13,  6.16006800E+04,
                    #-1.56739810E+00])),
        #note='H6W/94')

#species(name='C4H612',
        #atoms='C:4 H:6',
        #thermo=(NASA([300.00, 1374.00],
                    #[ 9.45515689E-01,  3.46162239E-02, -1.98590697E-05,
                    #5.02139421E-09, -3.67977164E-13,  1.81439079E+04,
                    #2.02191143E+01]),
                #NASA([1374.00, 5000.00],
                    #[ 1.14059885E+01,  1.31489843E-02, -4.43542071E-06,
                    #6.83028825E-10, -3.94289265E-14,  1.42427294E+04,
                    #-3.69674067E+01])),
        #note='A8/83')

#species(name='C4H2',
        #atoms='C:4 H:2',
        #thermo=(NASA([300.00, 1000.00],
                    #[ 1.05439780E+00,  4.16269600E-02, -6.58717840E-05,
                    #5.32570750E-08, -1.66831620E-11,  5.41852110E+04,
                    #1.48665910E+01]),
                #NASA([1000.00, 3000.00],
                    #[ 9.15763280E+00,  5.54305180E-03, -1.35916040E-06,
                    #1.87800750E-11,  2.31895360E-14,  5.25880390E+04,
                    #-2.37114600E+01])),
        #note='D11/99')

#species(name='C4H5-I',
        #atoms='C:4 H:5',
        #thermo=(NASA([300.00, 1000.00],
                    #[-1.99329000E-02,  3.80056720E-02, -2.75594500E-05,
                    #7.78355510E-09,  4.02093830E-13,  3.74962230E+04,
                    #2.43942410E+01]),
                #NASA([1000.00, 3000.00],
                    #[ 1.02290920E+01,  9.48501380E-03, -9.04064450E-08,
                    #-1.25961000E-09,  2.47814680E-13,  3.46428120E+04,
                    #-2.85645290E+01])),
        #note='H6W/94')

#species(name='C4H5-N',
        #atoms='C:4 H:5',
        #thermo=(NASA([300.00, 1000.00],
                    #[ 1.63053210E-01,  3.98301370E-02, -3.40001280E-05,
                    #1.51472330E-08, -2.46658250E-12,  4.14297660E+04,
                    #2.35361630E+01]),
                #NASA([1000.00, 3000.00],
                    #[ 9.85019780E+00,  1.07790080E-02, -1.36721250E-06,
                    #-7.72005350E-10,  1.83663140E-13,  3.88403010E+04,
                    #-2.60018460E+01])),
        #note='H6W/94')

#species(name='CH2CHCHCHO',
        #atoms='C:4 H:5 O:1',
        #thermo=(NASA([300.00, 1385.00],
                    #[ 4.51553480E-01,  4.37530418E-02, -3.37273602E-05,
                    #1.31111496E-08, -2.04732021E-12,  2.14943141E+03,
                    #2.40835867E+01]),
                #NASA([1385.00, 5000.00],
                    #[ 1.39299886E+01,  1.13228814E-02, -3.87393567E-06,
                    #6.02378620E-10, -3.50120301E-14, -2.39589898E+03,
                    #-4.79026849E+01])))

#species(name='C4H6O25',
        #atoms='C:4 H:6 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 2.67053463E+00,  4.92586420E-03,  8.86967406E-05,
                    #-1.26219194E-07,  5.23991321E-11, -1.46572472E+04,
                    #1.45722395E+01]),
                #NASA([1000.00, 5000.00],
                    #[ 8.60658242E+00,  2.08310051E-02, -8.42229481E-06,
                    #1.56717640E-09, -1.09391202E-13, -1.76177415E+04,
                    #-2.32464750E+01])),
        #note='T3/97')

#species(name='C2H3CHOCH2',
        #atoms='C:4 H:6 O:1',
        #thermo=(NASA([300.00, 1431.00],
                    #[-3.59388437E+00,  5.79063450E-02, -4.97163294E-05,
                    #2.15818682E-08, -3.69199072E-12,  8.58852628E+02,
                    #4.28475443E+01]),
                #NASA([1431.00, 5000.00],
                    #[ 1.26762790E+01,  1.40819509E-02, -4.63473868E-06,
                    #7.01090838E-10, -3.99438277E-14, -4.08065264E+03,
                    #-4.22515995E+01])))

#species(name='C4H5-2',
        #atoms='C:4 H:5',
        #thermo=(NASA([300.00, 1385.00],
                    #[ 2.31011292E+00,  2.83747046E-02, -1.63836755E-05,
                    #4.46251967E-09, -4.30510879E-13,  3.55842945E+04,
                    #1.49105965E+01]),
                #NASA([1385.00, 5000.00],
                    #[ 1.03230828E+01,  1.17625574E-02, -4.00004665E-06,
                    #6.18727929E-10, -3.58083530E-14,  3.25861413E+04,
                    #-2.88794317E+01])),
        #note='H6W/94')

#species(name='C4H6-2',
        #atoms='C:4 H:6',
        #thermo=(NASA([300.00, 1377.00],
                    #[ 1.97152408E+00,  2.76790997E-02, -1.13396645E-05,
                    #1.02970745E-09,  2.75290944E-13,  1.57283766E+04,
                    #1.42147356E+01]),
                #NASA([1377.00, 5000.00],
                    #[ 9.60305554E+00,  1.48972169E-02, -5.16751230E-06,
                    #8.09757170E-10, -4.72817668E-14,  1.24831314E+04,
                    #-2.87129792E+01])),
        #note='A8/83')

#species(name='C4H6O23',
        #atoms='C:4 H:6 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 2.67053463E+00,  4.92586420E-03,  8.86967406E-05,
                    #-1.26219194E-07,  5.23991321E-11, -1.02787872E+04,
                    #1.45722395E+01]),
                #NASA([1000.00, 5000.00],
                    #[ 8.60658242E+00,  2.08310051E-02, -8.42229481E-06,
                    #1.56717640E-09, -1.09391202E-13, -1.32392815E+04,
                    #-2.32464750E+01])),
        #note='T3/97')

#species(name='C4H4O',
        #atoms='C:4 H:4 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 8.47469463E-01,  1.31773796E-02,  5.99735901E-05,
                    #-9.71562904E-08,  4.22733796E-11, -5.36785445E+03,
                    #2.14945172E+01]),
                #NASA([1000.00, 6000.00],
                    #[ 9.38935003E+00,  1.40291241E-02, -5.07755110E-06,
                    #8.24137332E-10, -4.95319963E-14, -8.68241814E+03,
                    #-2.79162920E+01])),
        #note='T03/97')

#species(name='H2C4O',
        #atoms='H:2 C:4 O:1',
        #thermo=(NASA([300.00, 1000.00],
                    #[ 4.81097100E+00,  1.31399900E-02,  9.86507300E-07,
                    #-6.12072000E-09,  1.64000300E-12,  2.54580300E+04,
                    #2.11342400E+00]),
                #NASA([1000.00, 4000.00],
                    #[ 1.02688800E+01,  4.89616400E-03, -4.88508100E-07,
                    #-2.70856600E-10,  5.10701300E-14,  2.34690300E+04,
                    #-2.81598500E+01])),
        #note='120189')

#species(name='C2H3OCH2',
        #atoms='H:5 C:3 O:1',
        #thermo=(NASA([298.00, 1000.00],
                    #[-1.44860882E+00,  5.42514911E-02, -7.45556749E-05,
                    #6.03597706E-08, -2.01938036E-11,  1.01608402E+04,
                    #3.16318657E+01]),
                #NASA([1000.00, 6000.00],
                    #[ 1.00397568E+01,  1.18952742E-02, -4.13512716E-06,
                    #6.51590471E-10, -3.83196617E-14,  7.27301530E+03,
                    #-2.57171542E+01])))

#species(name='C6H6',
        #atoms='C:6 H:6',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 5.04818632E-01,  1.85020642E-02,  7.38345881E-05,
                    #-1.18135741E-07,  5.07210429E-11,  8.55247913E+03,
                    #2.16412893E+01]),
                #NASA([1000.00, 6000.00],
                    #[ 1.10809576E+01,  2.07176746E-02, -7.52145991E-06,
                    #1.22320984E-09, -7.36091279E-14,  4.30641035E+03,
                    #-4.00413310E+01])),
        #note='G6/01')

#species(name='C*CCJC*C',
        #atoms='C:5 H:7',
        #thermo=(NASA([300.00, 1388.00],
                    #[-2.94595603E+00,  5.68783623E-02, -4.31336497E-05,
                    #1.68169537E-08, -2.67926433E-12,  2.35156925E+04,
                    #3.98188778E+01]),
                #NASA([1388.00, 5000.00],
                    #[ 1.40879309E+01,  1.62398907E-02, -5.64768950E-06,
                    #8.86857524E-10, -5.18698993E-14,  1.76798698E+04,
                    #-5.13735038E+01])),
        #note='3/1/95Z&B')

#species(name='C6H5',
        #atoms='C:6 H:5',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 2.10306633E-01,  2.04745507E-02,  5.89743006E-05,
                    #-1.01534255E-07,  4.47105660E-11,  3.95468722E+04,
                    #2.52910455E+01]),
                #NASA([1000.00, 6000.00],
                    #[ 1.08444762E+01,  1.73212473E-02, -6.29233249E-06,
                    #1.02369961E-09, -6.16216828E-14,  3.55598475E+04,
                    #-3.53735134E+01])),
        #note='T04/02')

#species(name='C6H2',
        #atoms='C:6 H:2',
        #thermo=(NASA([300.00, 1000.00],
                    #[-1.59326240E+00,  8.05301450E-02, -1.48006490E-04,
                    #1.33000310E-07, -4.53323130E-11,  8.32732270E+04,
                    #2.79808730E+01]),
                #NASA([1000.00, 3000.00],
                    #[ 1.32262810E+01,  7.39043020E-03, -2.27153810E-06,
                    #2.58752170E-10, -5.53567410E-15,  8.05652580E+04,
                    #-4.12011760E+01])),
        #note='P1/93')

#species(name='C6H3',
        #atoms='C:6 H:3',
        #thermo=(NASA([300.00, 1000.00],
                    #[ 1.17906190E+00,  5.55473600E-02, -7.30761680E-05,
                    #5.20767360E-08, -1.50469640E-11,  8.56473120E+04,
                    #1.91791990E+01]),
                #NASA([1000.00, 3000.00],
                    #[ 5.81883430E+00,  2.79334080E-02, -1.78254270E-05,
                    #5.37025360E-09, -6.17076270E-13,  8.51882500E+04,
                    #-9.21478270E-01])),
        #note='H6W/94')

#species(name='L-C6H4',
        #atoms='C:6 H:4',
        #thermo=(NASA([300.00, 1000.00],
                    #[ 2.95902250E-01,  5.80533180E-02, -6.77667560E-05,
                    #4.33767620E-08, -1.14188640E-11,  6.00013710E+04,
                    #2.23189700E+01]),
                #NASA([1000.00, 3000.00],
                    #[ 1.27151820E+01,  1.38396620E-02, -4.37654400E-06,
                    #3.15416360E-10,  4.66190260E-14,  5.70311480E+04,
                    #-3.94646000E+01])),
        #note='H6W/94')

#species(name='C-C6H4',
        #atoms='C:6 H:4',
        #thermo=(NASA([300.00, 1000.00],
                    #[-3.09912680E+00,  5.40305640E-02, -4.08390040E-05,
                    #1.07388370E-08,  9.80784900E-13,  5.22057110E+04,
                    #3.74152070E+01]),
                #NASA([1000.00, 3000.00],
                    #[ 1.38492090E+01,  7.88079200E-03,  1.82438360E-06,
                    #-2.11691660E-09,  3.74599770E-13,  4.74463400E+04,
                    #-5.04049530E+01])),
        #note='H6W/94')

#species(name='C6H5O',
        #atoms='C:6 H:5 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[-4.66204455E-01,  4.13443975E-02,  1.32412991E-05,
                    #-5.72872769E-08,  2.89763707E-11,  4.77858391E+03,
                    #2.76990274E+01]),
                #NASA([1000.00, 6000.00],
                    #[ 1.37221720E+01,  1.74688771E-02, -6.35504520E-06,
                    #1.03492308E-09, -6.23410504E-14,  2.87274751E+02,
                    #-4.88181680E+01])),
        #note='T05/02')

#species(name='C6H5OH',
        #atoms='C:6 H:6 O:1',
        #thermo=(NASA([300.00, 1000.00],
                    #[-1.69565390E+00,  5.22712990E-02, -7.20240500E-06,
                    #-3.58596030E-08,  2.04490730E-11, -1.32841210E+04,
                    #3.25421600E+01]),
                #NASA([1000.00, 5000.00],
                    #[ 1.49120730E+01,  1.83781350E-02, -6.19831280E-06,
                    #9.19832210E-10, -4.92095650E-14, -1.83751990E+04,
                    #-5.59241030E+01])),
        #note='L4/84')

#species(name='P-C6H4O2',
        #atoms='C:6 H:4 O:2',
        #thermo=(NASA([270.00, 1370.00],
                    #[-2.43170113E+00,  6.87937608E-02, -6.41382837E-05,
                    #3.08126855E-08, -5.99832072E-12, -1.65696994E+04,
                    #3.48309430E+01]),
                #NASA([1370.00, 3000.00],
                    #[ 1.23423732E+01,  2.40612690E-02, -1.16565184E-05,
                    #2.71393504E-09, -2.47643065E-13, -2.06185312E+04,
                    #-4.08244024E+01])),
        #note='AK0405')

#species(name='O-C6H4O2',
        #atoms='C:6 H:4 O:2',
        #thermo=(NASA([270.00, 1370.00],
                    #[-2.36179712E+00,  6.86058343E-02, -6.39129516E-05,
                    #3.06903009E-08, -5.97357785E-12, -1.26704431E+04,
                    #3.53724482E+01]),
                #NASA([1370.00, 3000.00],
                    #[ 1.23614349E+01,  2.40491397E-02, -1.16529057E-05,
                    #2.71332785E-09, -2.47593219E-13, -1.67079717E+04,
                    #-4.00310857E+01])),
        #note='AK0405')

#species(name='P-C6H3O2',
        #atoms='C:6 H:3 O:2',
        #thermo=(NASA([270.00, 1290.00],
                    #[-1.57852347E+00,  6.55376473E-02, -6.50308721E-05,
                    #3.32026554E-08, -6.86665555E-12,  1.51750093E+04,
                    #3.31518638E+01]),
                #NASA([1290.00, 3000.00],
                    #[ 1.22963699E+01,  2.15055142E-02, -1.07516136E-05,
                    #2.57528163E-09, -2.41023652E-13,  1.15428998E+04,
                    #-3.72584002E+01])),
        #note='AK0505')

#species(name='C5H6',
        #atoms='C:5 H:6',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 8.61089570E-01,  1.48040310E-02,  7.21088950E-05,
                    #-1.13380550E-07,  4.86899720E-11,  1.48017550E+04,
                    #2.13534530E+01]),
                #NASA([1000.00, 6000.00],
                    #[ 9.97578480E+00,  1.89055430E-02, -6.84114610E-06,
                    #1.10993400E-09, -6.66802360E-14,  1.10816930E+04,
                    #-3.22094540E+01])),
        #note='T1/90')

#species(name='C5H5',
        #atoms='C:5 H:5',
        #thermo=(NASA([298.15, 969.35],
                    #[-3.97555452E+00,  7.41370991E-02, -1.11803345E-04,
                    #9.04628776E-08, -2.80999747E-11,  3.01769405E+04,
                    #3.67153636E+01]),
                #NASA([969.35, 3500.00],
                    #[ 1.33675715E+00,  3.24793912E-02, -1.67587774E-05,
                    #4.03514137E-09, -3.70739036E-13,  3.00730524E+04,
                    #1.60315806E+01])),
        #note='TAK0505')

#species(name='C5H4O',
        #atoms='C:5 H:4 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 2.64576497E-01,  3.34873827E-02,  1.67738470E-06,
                    #-2.96207455E-08,  1.54431476E-11,  5.11159287E+03,
                    #2.35409513E+01]),
                #NASA([1000.00, 6000.00],
                    #[ 1.00806824E+01,  1.61143465E-02, -5.83314509E-06,
                    #9.46759320E-10, -5.68972206E-14,  1.94364771E+03,
                    #-2.94521623E+01])),
        #note='T8/99')

#species(name='C5H5O',
        #atoms='C:5 H:5 O:1',
        #thermo=(NASA([300.00, 1392.00],
                    #[-2.83112840E+00,  5.67277287E-02, -4.44757303E-05,
                    #1.74924447E-08, -2.76004847E-12,  2.04992154E+04,
                    #3.69634411E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 1.48322894E+01,  1.40483376E-02, -4.92302051E-06,
                    #7.77041219E-10, -4.56103939E-14,  1.45523665E+04,
                    #-5.73228191E+01])),
        #note='5/16/90THERM')

#species(name='C5H5OH',
        #atoms='C:5 H:6 O:1',
        #thermo=(NASA([300.00, 1398.00],
                    #[-4.26822012E+00,  6.62446749E-02, -5.68494038E-05,
                    #2.46858526E-08, -4.26820696E-12, -5.75581338E+03,
                    #4.47962850E+01]),
                #NASA([1398.00, 5000.00],
                    #[ 1.53433477E+01,  1.50754059E-02, -5.13553582E-06,
                    #7.95807816E-10, -4.61311517E-14, -1.19645453E+04,
                    #-5.85204430E+01])),
        #note='5/2/91THE.M')

#species(name='C5H4OH',
        #atoms='C:5 H:5 O:1',
        #thermo=(NASA([200.00, 1000.00],
                    #[-1.28398054E+00,  4.90298511E-02, -1.35844414E-05,
                    #-2.92983743E-08,  1.90820619E-11,  6.37364803E+03,
                    #3.08073591E+01]),
                #NASA([1000.00, 6000.00],
                    #[ 1.33741248E+01,  1.51996469E-02, -5.45685046E-06,
                    #8.80944866E-10, -5.27493258E-14,  2.20358027E+03,
                    #-4.59569069E+01])),
        #note='T8/99')

#species(name='C5H3O',
        #atoms='C:5 H:3 O:1',
        #thermo=(NASA([300.00, 1500.00],
                    #[-3.03242604E+00,  5.43937201E-02, -4.95018348E-05,
                    #2.25523751E-08, -4.10727920E-12,  3.35644081E+04,
                    #3.78374823E+01]),
                #NASA([1500.00, 3500.00],
                    #[ 1.19961781E+01,  1.34287065E-02, -5.90045309E-06,
                    #1.22553862E-09, -9.86114716E-14,  2.89592010E+04,
                    #-4.07548249E+01])),
        #note='TAK0905')

#species(name='C6H5OO',
        #atoms='C:6 H:5 O:2',
        #thermo=(NASA([300.00, 1403.00],
                    #[-2.99164672E+00,  7.03857150E-02, -6.34400574E-05,
                    #2.91548920E-08, -5.30706938E-12,  1.41320240E+04,
                    #4.20142955E+01]),
                #NASA([1403.00, 5000.00],
                    #[ 1.67078262E+01,  1.62326229E-02, -5.47969630E-06,
                    #8.43510060E-10, -4.86562431E-14,  8.14242915E+03,
                    #-6.08346973E+01])),
        #note='3/26/9THERM')

#species(name='C6H5OOH',
        #atoms='C:6 H:6 O:2',
        #thermo=(NASA([300.00, 1404.00],
                    #[-4.03105975E+00,  7.96101888E-02, -7.21655013E-05,
                    #3.27610696E-08, -5.85584239E-12, -3.10973017E+03,
                    #4.54324978E+01]),
                #NASA([1404.00, 5000.00],
                    #[ 1.92317474E+01,  1.63154699E-02, -5.53448904E-06,
                    #8.55059974E-10, -4.94583790E-14, -1.01971012E+04,
                    #-7.61674471E+01])),
        #note='3/26/9THERM')

#species(name='OC6H4OH',
        #atoms='C:6 H:5 O:2',
        #thermo=(NASA([300.00, 1403.00],
                    #[-8.02205657E+00,  1.09403210E-01, -1.23489276E-04,
                    #6.56286805E-08, -1.31527870E-11, -1.55949156E+04,
                    #5.72175202E+01]),
                #NASA([1403.00, 5000.00],
                    #[ 2.22718210E+01,  1.21038561E-02, -4.18429526E-06,
                    #6.54475399E-10, -3.81746504E-14, -2.34827539E+04,
                    #-9.61035467E+01])),
        #note='4/9/9THERM')

#species(name='C6H4OH',
        #atoms='C:6 H:5 O:1',
        #thermo=(NASA([300.00, 1402.00],
                    #[-5.99875435E+00,  8.59063379E-02, -9.12525636E-05,
                    #4.72275890E-08, -9.35576749E-12,  1.78621926E+04,
                    #4.99931427E+01]),
                #NASA([1402.00, 5000.00],
                    #[ 1.73187560E+01,  1.36366984E-02, -4.68316332E-06,
                    #7.29071204E-10, -4.23805358E-14,  1.14990276E+04,
                    #-6.89986593E+01])),
        #note='4/9/9THERM')

#species(name='O-OC6H5OJ',
        #atoms='C:6 O:2 H:5',
        #thermo=(NASA([300.00, 1400.00],
                    #[-2.65459198E+00,  7.17179095E-02, -6.31552372E-05,
                    #2.81132946E-08, -4.97463333E-12,  6.45283150E+03,
                    #3.81123139E+01]),
                #NASA([1400.00, 5000.00],
                    #[ 1.84625733E+01,  1.57607263E-02, -5.44671499E-06,
                    #8.51765760E-10, -4.96759541E-14, -1.72770226E+02,
                    #-7.28742484E+01])),
        #note='WKM')

#species(name='P-OC6H5OJ',
        #atoms='C:6 O:2 H:5',
        #thermo=(NASA([300.00, 1400.00],
                    #[-3.29683290E+00,  7.27365977E-02, -6.36158220E-05,
                    #2.80683553E-08, -4.92279426E-12,  6.73402222E+03,
                    #4.09349895E+01]),
                #NASA([1400.00, 5000.00],
                    #[ 1.82799770E+01,  1.59280974E-02, -5.50765220E-06,
                    #8.61649836E-10, -5.02677539E-14, -6.25907994E+01,
                    #-7.25809444E+01])),
        #note='WKM')

#species(name='C#CC*CCJ',
        #atoms='C:5 H:5',
        #thermo=(NASA([300.00, 1396.00],
                    #[-6.16143558E-01,  5.06466579E-02, -4.48561743E-05,
                    #2.02459419E-08, -3.64542145E-12,  4.71532377E+04,
                    #2.71623299E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 1.41230912E+01,  1.14233190E-02, -3.95851276E-06,
                    #6.20128961E-10, -3.62097887E-14,  4.25158384E+04,
                    #-5.02942871E+01])),
        #note='GLAR')

#species(name='C5H6-L',
        #atoms='C:5 H:6',
        #thermo=(NASA([300.00, 1372.00],
                    #[ 3.58448213E+00,  3.24459626E-02, -1.70150991E-05,
                    #4.22715914E-09, -4.18452556E-13,  2.76514681E+04,
                    #9.60644208E+00]),
                #NASA([1372.00, 5000.00],
                    #[ 1.29600892E+01,  1.48953758E-02, -5.23622902E-06,
                    #8.27916389E-10, -4.86464523E-14,  2.38180800E+04,
                    #-4.25312093E+01])),
        #note='2/5/9THERM')

#species(name='CJ*CC*CC*O',
        #atoms='C:5 H:5 O:1',
        #thermo=(NASA([300.00, 1396.00],
                    #[-5.06628841E-01,  6.04671965E-02, -5.97396749E-05,
                    #2.96804228E-08, -5.76240010E-12,  2.42765544E+04,
                    #2.82994148E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 1.62360823E+01,  1.18297101E-02, -4.11454219E-06,
                    #6.46026823E-10, -3.77767639E-14,  1.93499885E+04,
                    #-5.83498817E+01])),
        #note='2/5/9THERM')

#species(name='C*CC*CCJ*O',
        #atoms='C:5 H:5 O:1',
        #thermo=(NASA([300.00, 1399.00],
                    #[-2.18492198E-01,  5.92100223E-02, -5.89241174E-05,
                    #2.97411920E-08, -5.85244770E-12,  1.20600764E+04,
                    #2.55968530E+01]),
                #NASA([1399.00, 5000.00],
                    #[ 1.53178248E+01,  1.27352911E-02, -4.35882964E-06,
                    #6.76912763E-10, -3.92771371E-14,  7.60582726E+03,
                    #-5.43599625E+01])),
        #note='2/5/9THERM')

#species(name='CJ*CC*O',
        #atoms='C:3 H:3 O:1',
        #thermo=(NASA([300.00, 1402.00],
                    #[ 1.46654466E+00,  3.23390476E-02, -3.05588208E-05,
                    #1.44081861E-08, -2.65600505E-12,  1.78850058E+04,
                    #1.80850321E+01]),
                #NASA([1402.00, 5000.00],
                    #[ 1.07482537E+01,  6.19822688E-03, -2.06130981E-06,
                    #3.14418872E-10, -1.80309517E-14,  1.51410162E+04,
                    #-3.01266033E+01])),
        #note='4/8/94THERM')

#species(name='C5H7',
        #atoms='C:5 H:7',
        #thermo=(NASA([300.00, 1377.00],
                    #[-6.75118368E+00,  6.06461693E-02, -4.01260152E-05,
                    #1.22051562E-08, -1.33459844E-12,  2.01365277E+04,
                    #5.62694938E+01]),
                #NASA([1377.00, 5000.00],
                    #[ 1.36630213E+01,  1.68061358E-02, -5.98746539E-06,
                    #9.55341072E-10, -5.64951981E-14,  1.27238941E+04,
                    #-5.46331286E+01])),
        #note='1/22/9WKM')

#species(name='OC5H7O',
        #atoms='C:5 H:7 O:2',
        #thermo=(NASA([300.00, 1375.00],
                    #[ 4.88394767E+00,  4.03401300E-02, -1.97774150E-05,
                    #3.68903501E-09, -3.40202384E-14, -2.35295942E+04,
                    #9.97070337E+00]),
                #NASA([1375.00, 5000.00],
                    #[ 1.65416953E+01,  1.86677673E-02, -6.44836048E-06,
                    #1.00787611E-09, -5.87521858E-14, -2.82017168E+04,
                    #-5.47258181E+01])),
        #note='1/22/9WKM')

#species(name='C*CCJC*COH',
        #atoms='C:5 H:7 O:1',
        #thermo=(NASA([300.00, 1397.00],
                    #[-2.91175436E+00,  6.69362484E-02, -5.71603047E-05,
                    #2.48753749E-08, -4.33243894E-12,  1.96441523E+03,
                    #4.17454344E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 1.67465815E+01,  1.58357240E-02, -5.44954706E-06,
                    #8.49881387E-10, -4.94743246E-14, -4.30972870E+03,
                    #-6.19378748E+01])),
        #note='10/6/95Z&B')

#species(name='C*CC*CCJ',
        #atoms='C:5 H:7',
        #thermo=(NASA([300.00, 1386.00],
                    #[-1.60087476E+00,  5.38764703E-02, -3.96302225E-05,
                    #1.49599474E-08, -2.31995284E-12,  2.31199746E+04,
                    #3.35492960E+01]),
                #NASA([1386.00, 5000.00],
                    #[ 1.47302883E+01,  1.59030900E-02, -5.57729508E-06,
                    #8.80604825E-10, -5.16963733E-14,  1.74050791E+04,
                    #-5.42670706E+01])),
        #note='Z&B')

#species(name='C*CC*CCOH',
        #atoms='C:5 H:8 O:1',
        #thermo=(NASA([300.00, 1396.00],
                    #[-5.31488384E-01,  6.06983915E-02, -4.81499862E-05,
                    #2.00308244E-08, -3.38987282E-12, -1.03301302E+04,
                    #3.07961436E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 1.63079670E+01,  1.79957763E-02, -6.03115896E-06,
                    #9.23992259E-10, -5.31254053E-14, -1.58204603E+04,
                    #-5.84137244E+01])),
        #note='1/23/9WKM')

#species(name='OC4H6O',
        #atoms='C:4 H:6 O:2',
        #thermo=(NASA([300.00, 1382.00],
                    #[ 4.21628848E+00,  3.57422725E-02, -2.04226185E-05,
                    #5.63821367E-09, -5.88888993E-13, -3.72055911E+04,
                    #1.02814620E+01]),
                #NASA([1382.00, 5000.00],
                    #[ 1.41894774E+01,  1.53345510E-02, -5.24594862E-06,
                    #8.14655154E-10, -4.72759368E-14, -4.10001835E+04,
                    #-4.43771751E+01])),
        #note='1/23/9WKM')

#species(name='OC4H5O',
        #atoms='C:4 H:5 O:2',
        #thermo=(NASA([300.00, 1388.00],
                    #[ 4.60550978E+00,  3.30498712E-02, -2.13102363E-05,
                    #7.37021089E-09, -1.08289438E-12, -1.85460831E+04,
                    #1.01599453E+01]),
                #NASA([1388.00, 5000.00],
                    #[ 1.32138775E+01,  1.37339051E-02, -4.62639517E-06,
                    #7.10941370E-10, -4.09538499E-14, -2.16535271E+04,
                    #-3.64185255E+01])),
        #note='1/23/9WKM')

#species(name='HOCO',
        #atoms='H:1 C:1 O:2',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 2.92207919E+00,  7.62453859E-03,  3.29884437E-06,
                    #-1.07135205E-08,  5.11587057E-12, -2.30281524E+04,
                    #1.12925886E+01]),
                #NASA([1000.00, 6000.00],
                    #[ 5.39206152E+00,  4.11221455E-03, -1.48194900E-06,
                    #2.39875460E-10, -1.43903104E-14, -2.38606717E+04,
                    #-2.23529091E+00])),
        #note='T05/06')

#species(name='HOC*CC*O',
        #atoms='C:3 H:4 O:2',
        #thermo=(NASA([300.00, 1413.00],
                    #[-2.01837189E+00,  6.26539783E-02, -6.73359280E-05,
                    #3.39430425E-08, -6.48917648E-12, -3.31367523E+04,
                    #3.18162860E+01]),
                #NASA([1413.00, 5000.00],
                    #[ 1.66505478E+01,  6.11745137E-03, -2.09080785E-06,
                    #3.24985683E-10, -1.88875073E-14, -3.82179939E+04,
                    #-6.36794754E+01])),
        #note='1/26/9WKM')

#species(name='HOC*CCJ*O',
        #atoms='C:3 H:3 O:2',
        #thermo=(NASA([300.00, 1414.00],
                    #[ 6.07270082E-01,  4.96011303E-02, -5.32300885E-05,
                    #2.68392951E-08, -5.13094510E-12, -1.58814562E+04,
                    #1.94817133E+01]),
                #NASA([1414.00, 5000.00],
                    #[ 1.52720985E+01,  5.02586331E-03, -1.68408578E-06,
                    #2.58390706E-10, -1.48849424E-14, -1.98506828E+04,
                    #-5.54641734E+01])),
        #note='1/26/9WKM')

#species(name='O2CCHOOJ',
        #atoms='C:2 H:1 O:4',
        #thermo=(NASA([300.00, 1682.00],
                    #[ 8.91497688E+00,  8.60571847E-03,  5.24416766E-07,
                    #-2.79301331E-09,  7.62963051E-13, -3.40867754E+04,
                    #-8.72978273E+00]),
                #NASA([1682.00, 5000.00],
                    #[ 1.09910849E+01,  7.46985861E-03, -2.75568271E-06,
                    #4.51353051E-10, -2.72108652E-14, -3.51335323E+04,
                    #-2.11652231E+01])),
        #note='Z&B')

#species(name='C6H101-5',
        #atoms='C:6 H:10',
        #thermo=(NASA([300.00, 1413.00],
                    #[-1.01375402E+00,  6.38242808E-02, -4.40653860E-05,
                    #1.58295163E-08, -2.30830701E-12,  7.94033696E+03,
                    #3.25056094E+01]),
                #NASA([1413.00, 5000.00],
                    #[ 1.60456030E+01,  2.34774145E-02, -7.85797929E-06,
                    #1.20200542E-09, -6.90100029E-14,  2.11899382E+03,
                    #-5.88452460E+01])),
        #note='4/12/13THERM')

#species(name='C6H9-A',
        #atoms='C:6 H:9',
        #thermo=(NASA([300.00, 1400.00],
                    #[-2.66715213E+00,  7.26196475E-02, -6.05323920E-05,
                    #2.66000571E-08, -4.74613408E-12,  2.64415017E+04,
                    #4.02220332E+01]),
                #NASA([1400.00, 5000.00],
                    #[ 1.70842767E+01,  2.08842788E-02, -7.14529004E-06,
                    #1.10943563E-09, -6.43676989E-14,  2.01040204E+04,
                    #-6.39326012E+01])),
        #note='12/5/12THERM')

#species(name='C3H2',
        #atoms='C:3 H:2',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 2.43417332E+00,  1.73013063E-02, -1.18294047E-05,
                    #1.02756396E-09,  1.62626314E-12,  7.69074892E+04,
                    #1.21012230E+01]),
                #NASA([1000.00, 6000.00],
                    #[ 6.67324762E+00,  5.57728845E-03, -1.99180164E-06,
                    #3.20289156E-10, -1.91216272E-14,  7.57571184E+04,
                    #-9.72894405E+00])),
        #note='T12/00')

#species(name='C3H2(S)',
        #atoms='C:3 H:2',
        #thermo=(NASA([200.00, 900.00],
                    #[ 5.29764820E+00,  1.69874660E-02, -2.42665170E-05,
                    #1.86536810E-08, -5.57630010E-12,  6.72404660E+04,
                    #-3.75400410E+00]),
                #NASA([900.00, 5000.00],
                    #[ 7.76425700E+00,  4.71127740E-03, -1.61706370E-06,
                    #2.54724060E-10, -1.50385720E-14,  6.68496720E+04,
                    #-1.50985490E+01])),
        #note='0')

#species(name='C3H2C',
        #atoms='C:3 H:2',
        #thermo=(NASA([200.00, 1500.00],
                    #[ 1.12958880E+00,  1.72874010E-02, -1.13668230E-05,
                    #3.45692960E-09, -3.66159510E-13,  5.84190800E+04,
                    #1.73314480E+01]),
                #NASA([1500.00, 5000.00],
                    #[ 6.56326800E+00,  5.23632560E-03, -1.75448300E-06,
                    #2.68661060E-10, -1.54285090E-14,  5.65146180E+04,
                    #-1.20006070E+01])),
        #note='0')

#species(name='H2CCC(S)',
        #atoms='C:3 H:2',
        #thermo=(NASA([200.00, 1500.00],
                    #[ 3.72297260E+00,  9.25898540E-03, -2.30061910E-06,
                    #-1.02008080E-09,  4.53743570E-13,  6.48772890E+04,
                    #5.68659360E+00]),
                #NASA([1500.00, 5000.00],
                    #[ 6.48887620E+00,  5.31127890E-03, -1.78094900E-06,
                    #2.72526420E-10, -1.56195900E-14,  6.36618640E+04,
                    #-1.00642830E+01])),
        #note='0')

#species(name='C3H3O',
        #atoms='H:3 C:3 O:1',
        #thermo=(NASA([298.15, 1000.00],
                    #[ 8.75023836E-01,  3.51184068E-02, -3.89901356E-05,
                    #2.40255750E-08, -6.10883631E-12,  3.20427921E+04,
                    #2.04717253E+01]),
                #NASA([1000.00, 2000.00],
                    #[ 4.19355696E+00,  1.95625103E-02, -1.22336450E-05,
                    #3.90615061E-09, -5.08539231E-13,  3.14931737E+04,
                    #5.03216224E+00])),
        #note='2/17/14CZHOU')

#species(name='C3H3O2H',
        #atoms='C:3 H:4 O:2',
        #thermo=(NASA([300.00, 1385.00],
                    #[ 1.09787313E+00,  4.22717882E-02, -3.83969355E-05,
                    #1.77405069E-08, -3.27674312E-12,  1.03592314E+04,
                    #2.30651783E+01]),
                #NASA([1385.00, 5000.00],
                    #[ 1.38152174E+01,  8.62174763E-03, -3.06710006E-06,
                    #4.88874247E-10, -2.88888385E-14,  6.29182941E+03,
                    #-4.39151257E+01])),
        #note='1/31/13')

#species(name='C2HCHO',
        #atoms='C:3 H:2 O:1',
        #thermo=(NASA([300.00, 2012.00],
                    #[ 4.20776611E+00,  1.34382727E-02, -5.15442099E-06,
                    #-2.24570818E-11,  2.74111284E-13,  1.02117375E+04,
                    #5.43871873E+00]),
                #NASA([2012.00, 5000.00],
                    #[ 7.99952054E+00,  7.07825497E-03, -2.63086819E-06,
                    #4.33073185E-10, -2.62003284E-14,  8.71863156E+03,
                    #-1.57226237E+01])),
        #note='1/31/13')

#species(name='CH3COCH2O',
        #atoms='C:3 H:5 O:2',
        #thermo=(NASA([300.00, 2002.00],
                    #[ 5.85960137E+00,  1.78954926E-02,  7.41506398E-07,
                    #-5.40032753E-09,  1.47393197E-12, -1.90714739E+04,
                    #2.70987883E+00]),
                #NASA([2002.00, 5000.00],
                    #[ 9.84061707E+00,  1.59181106E-02, -5.85164644E-06,
                    #9.56160073E-10, -5.75477263E-14, -2.11214823E+04,
                    #-2.12330791E+01])),
        #note='2/8/13THERM')

#species(name='PC3H4OH-2',
        #atoms='C:3 H:5 O:1',
        #thermo=(NASA([300.00, 1403.00],
                    #[ 1.42757363E+00,  3.64825569E-02, -3.18007132E-05,
                    #1.46914605E-08, -2.72331227E-12,  7.80342663E+03,
                    #1.85890339E+01]),
                #NASA([1403.00, 5000.00],
                    #[ 1.07164095E+01,  1.06066461E-02, -3.51374060E-06,
                    #5.33713932E-10, -3.04901511E-14,  4.98486803E+03,
                    #-2.98329329E+01])),
        #note='4/2/13THERM')

#species(name='SC3H4OH',
        #atoms='C:3 H:5 O:1',
        #thermo=(NASA([300.00, 1407.00],
                    #[ 1.72870561E+00,  4.41015870E-02, -4.72013860E-05,
                    #2.52073596E-08, -5.13375710E-12,  2.22720503E+03,
                    #1.43928257E+01]),
                #NASA([1407.00, 5000.00],
                    #[ 1.20968484E+01,  9.43976596E-03, -3.10773897E-06,
                    #4.69609188E-10, -2.67165710E-14, -3.85854894E+02,
                    #-3.76795997E+01])),
        #note='3/28/13')

#species(name='SC3H5OH',
        #atoms='C:3 H:6 O:1',
        #thermo=(NASA([300.00, 1404.00],
                    #[-3.53977226E-02,  4.34969453E-02, -3.74479918E-05,
                    #1.70906074E-08, -3.13775054E-12, -2.02502608E+04,
                    #2.41528201E+01]),
                #NASA([1404.00, 5000.00],
                    #[ 1.11222064E+01,  1.27745410E-02, -4.25315532E-06,
                    #6.48216484E-10, -3.71190850E-14, -2.36690795E+04,
                    #-3.41335182E+01])),
        #note='2/3/9')

#species(name='CH3CHCHO',
        #atoms='C:3 H:5 O:1',
        #thermo=(NASA([300.00, 1253.00],
                    #[-2.72811212E+00,  3.80354483E-02, -1.96782132E-05,
                    #5.09554570E-09, -7.28789092E-13, -3.51122739E+03,
                    #4.04496677E+01]),
                #NASA([1253.00, 5000.00],
                    #[ 8.27772090E+00,  1.95687188E-02, -8.47575089E-06,
                    #1.53249728E-09, -9.86441542E-14, -8.34456062E+03,
                    #-2.17652483E+01])),
        #note='8/10/4THERM')

#species(name='AC4H7OOH',
        #atoms='C:4 H:8 O:2',
        #thermo=(NASA([300.00, 1395.00],
                    #[ 1.33470633E+00,  5.27831440E-02, -3.58861360E-05,
                    #1.32495013E-08, -2.06619289E-12, -8.87891782E+03,
                    #2.43857336E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 1.47661443E+01,  2.12235231E-02, -7.09403390E-06,
                    #1.08423759E-09, -6.22145708E-14, -1.35617411E+04,
                    #-4.77449138E+01])),
        #note='6/17/13THERM')

#species(name='TQJC3H6OH',
        #atoms='C:3 H:7 O:3',
        #thermo=(NASA([300.00, 1411.00],
                    #[ 2.29728737E+00,  4.96725824E-02, -3.84714645E-05,
                    #1.56038090E-08, -2.56748169E-12, -2.80429510E+04,
                    #2.03767936E+01]),
                #NASA([1411.00, 5000.00],
                    #[ 1.57530365E+01,  1.56289438E-02, -5.23690385E-06,
                    #8.01741456E-10, -4.60590728E-14, -3.24130576E+04,
                    #-5.08911217E+01])))

#species(name='TQC3H6OI',
        #atoms='C:3 H:7 O:3',
        #thermo=(NASA([300.00, 1417.00],
                    #[ 3.77984842E+00,  4.51246380E-02, -3.03819021E-05,
                    #9.92338649E-09, -1.22338206E-12, -1.84544341E+04,
                    #1.23546787E+01]),
                #NASA([1417.00, 5000.00],
                    #[ 1.71497525E+01,  1.47440875E-02, -4.99973196E-06,
                    #7.71920259E-10, -4.46175680E-14, -2.30944769E+04,
                    #-5.96032656E+01])))

#species(name='QC3H5OHP',
        #atoms='C:3 H:7 O:3',
        #thermo=(NASA([300.00, 1415.00],
                    #[ 1.76951547E+00,  5.23733147E-02, -4.19658180E-05,
                    #1.72471815E-08, -2.83710014E-12, -1.99859353E+04,
                    #2.46273265E+01]),
                #NASA([1415.00, 5000.00],
                    #[ 1.65390643E+01,  1.45962463E-02, -4.89569717E-06,
                    #7.50190330E-10, -4.31310565E-14, -2.47102324E+04,
                    #-5.34015731E+01])))

#species(name='TQC3H5OHI',
        #atoms='C:3 H:7 O:3',
        #thermo=(NASA([300.00, 1526.00],
                    #[ 5.58028765E+00,  3.00649782E-02, -6.38519910E-06,
                    #-4.74322634E-09,  1.89822226E-12, -2.19537872E+04,
                    #8.64790031E+00]),
                #NASA([1526.00, 5000.00],
                    #[ 1.55469726E+01,  1.61841763E-02, -5.60787877E-06,
                    #8.78452695E-10, -5.12904040E-14, -2.63164910E+04,
                    #-4.81509679E+01])))

#species(name='CCY(COC)OH',
        #atoms='C:3 H:6 O:2',
        #thermo=(NASA([300.00, 1427.00],
                    #[-7.68565795E-01,  4.79603507E-02, -3.53347028E-05,
                    #1.27934873E-08, -1.79555386E-12, -3.59177637E+04,
                    #2.98848715E+01]),
                #NASA([1427.00, 5000.00],
                    #[ 1.37176873E+01,  1.33634818E-02, -4.51490840E-06,
                    #6.95521032E-10, -4.01451866E-14, -4.07527954E+04,
                    #-4.74449508E+01])))

#species(name='IQJC3H6OH',
        #atoms='C:3 H:7 O:3',
        #thermo=(NASA([300.00, 1408.00],
                    #[ 1.74093905E+00,  5.10931764E-02, -4.02179535E-05,
                    #1.66120004E-08, -2.78425352E-12, -2.79317094E+04,
                    #2.33734165E+01]),
                #NASA([1408.00, 5000.00],
                    #[ 1.55719527E+01,  1.57683363E-02, -5.28242575E-06,
                    #8.08628793E-10, -4.64527621E-14, -3.23985113E+04,
                    #-4.97782092E+01])))

#species(name='IQC3H6OT',
        #atoms='C:3 H:7 O:3',
        #thermo=(NASA([300.00, 1406.00],
                    #[ 2.97554636E+00,  4.80962295E-02, -3.51787999E-05,
                    #1.32008460E-08, -2.00208137E-12, -1.83154710E+04,
                    #1.64182550E+01]),
                #NASA([1406.00, 5000.00],
                    #[ 1.67044548E+01,  1.52497437E-02, -5.19996341E-06,
                    #8.05676289E-10, -4.66785083E-14, -2.29660389E+04,
                    #-5.69717854E+01])))

#species(name='IQC3H5OHPJ',
        #atoms='C:3 H:7 O:3',
        #thermo=(NASA([300.00, 1409.00],
                    #[ 1.35929059E+00,  5.34300669E-02, -4.41305751E-05,
                    #1.89646022E-08, -3.27838412E-12, -1.98996889E+04,
                    #2.69126449E+01]),
                #NASA([1409.00, 5000.00],
                    #[ 1.59552559E+01,  1.50895186E-02, -5.06382338E-06,
                    #7.76108726E-10, -4.46235236E-14, -2.45035338E+04,
                    #-4.99018045E+01])))

#species(name='CHOCH2OOH',
        #atoms='C:2 H:4 O:3',
        #thermo=(NASA([300.00, 1391.00],
                    #[-1.32768631E+00,  5.21618601E-02, -4.97327645E-05,
                    #2.31272366E-08, -4.20787867E-12, -2.90608844E+04,
                    #3.54965675E+01]),
                #NASA([1391.00, 5000.00],
                    #[ 1.51554685E+01,  7.57240000E-03, -2.72693024E-06,
                    #4.38217189E-10, -2.60434287E-14, -3.41419680E+04,
                    #-5.08149884E+01])))

#species(name='CY(CCOC)OH',
        #atoms='C:3 H:6 O:2',
        #thermo=(NASA([300.00, 1433.00],
                    #[-5.94615865E+00,  6.41555008E-02, -5.85235665E-05,
                    #2.68222743E-08, -4.79465381E-12, -3.07734013E+04,
                    #5.39539327E+01]),
                #NASA([1433.00, 5000.00],
                    #[ 1.16477060E+01,  1.42006208E-02, -4.52110513E-06,
                    #6.68183015E-10, -3.74416389E-14, -3.58421461E+04,
                    #-3.71322765E+01])))

#species(name='CHOCOHCH3',
        #atoms='C:3 H:6 O:2',
        #thermo=(NASA([300.00, 1379.00],
                    #[ 9.59562607E-01,  4.04355781E-02, -2.52333819E-05,
                    #7.47748757E-09, -8.33450351E-13, -4.38727633E+04,
                    #2.52653125E+01]),
                #NASA([1379.00, 5000.00],
                    #[ 1.36133542E+01,  1.38716512E-02, -4.87445259E-06,
                    #7.70639716E-10, -4.52818606E-14, -4.85836793E+04,
                    #-4.37863752E+01])))

#species(name='CY(COC)COH',
        #atoms='C:3 H:6 O:2',
        #thermo=(NASA([300.00, 1427.00],
                    #[-3.28993462E+00,  5.79159124E-02, -5.21642132E-05,
                    #2.38463400E-08, -4.27986682E-12, -3.04628787E+04,
                    #4.23855523E+01]),
                #NASA([1427.00, 5000.00],
                    #[ 1.23650178E+01,  1.37866121E-02, -4.50198310E-06,
                    #6.77071535E-10, -3.84092609E-14, -3.50388267E+04,
                    #-3.88318770E+01])))

#species(name='TQC3H5OHIO2',
        #atoms='C:3 H:7 O:5',
        #thermo=(NASA([300.00, 1387.00],
                    #[ 2.11774097E+00,  7.25997025E-02, -6.81667819E-05,
                    #3.13796075E-08, -5.68565078E-12, -4.35518506E+04,
                    #2.23586550E+01]),
                #NASA([1387.00, 5000.00],
                    #[ 2.02916342E+01,  1.92145529E-02, -8.58823433E-06,
                    #1.62509037E-09, -1.06225627E-13, -4.85976512E+04,
                    #-7.10879629E+01])))

#species(name='TQC3H5OHIQ-I',
        #atoms='C:3 H:7 O:5',
        #thermo=(NASA([300.00, 1391.00],
                    #[ 3.46235675E+00,  6.98705440E-02, -6.48577576E-05,
                    #2.96626742E-08, -5.34536953E-12, -3.39681675E+04,
                    #1.47596723E+01]),
                #NASA([1391.00, 5000.00],
                    #[ 2.28708932E+01,  1.56658128E-02, -6.48662482E-06,
                    #1.17068411E-09, -7.44373426E-14, -3.97155885E+04,
                    #-8.61588096E+01])))

#species(name='TQC3H5OHIQ-P',
        #atoms='C:3 H:7 O:5',
        #thermo=(NASA([300.00, 1478.00],
                    #[ 2.45629666E+00,  7.08507900E-02, -6.43845421E-05,
                    #2.79222006E-08, -4.68412156E-12, -3.56101660E+04,
                    #2.27189955E+01]),
                #NASA([1478.00, 5000.00],
                    #[ 2.59345778E+01,  6.66525395E-03, -1.66651304E-07,
                    #-4.20552574E-10,  5.49726117E-14, -4.21899085E+04,
                    #-9.90618026E+01])))

#species(name='C2H3COHOOH',
        #atoms='C:3 H:6 O:3',
        #thermo=(NASA([300.00, 1393.00],
                    #[ 3.73734021E-01,  5.84060291E-02, -5.29627415E-05,
                    #2.38734015E-08, -4.26063052E-12, -3.27577137E+04,
                    #2.67905679E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 1.69349460E+01,  1.32757225E-02, -5.30958950E-06,
                    #9.43268930E-10, -5.97339692E-14, -3.77990580E+04,
                    #-5.97614008E+01])))

#species(name='COHOOHCY(COC)',
        #atoms='C:3 H:6 O:4',
        #thermo=(NASA([300.00, 1394.00],
                    #[-3.16180219E+00,  7.92240221E-02, -8.28674521E-05,
                    #4.32049848E-08, -8.69506663E-12, -4.59891154E+04,
                    #4.31128002E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 1.78808618E+01,  1.47154928E-02, -5.11766843E-06,
                    #8.03382779E-10, -4.69696797E-14, -5.19254526E+04,
                    #-6.47302080E+01])))

#species(name='OHCY(COCC)OOH',
        #atoms='C:3 H:6 O:4',
        #thermo=(NASA([300.00, 1520.00],
                    #[-2.13807799E-01,  5.79347134E-02, -4.16882857E-05,
                    #1.38590975E-08, -1.65613593E-12, -4.67090699E+04,
                    #3.18890641E+01]),
                #NASA([1520.00, 5000.00],
                    #[ 1.88928431E+01,  1.38516388E-02, -4.73692460E-06,
                    #7.36282623E-10, -4.27819063E-14, -5.31875949E+04,
                    #-7.05664934E+01])))

#species(name='IQC3H5OHPJO2',
        #atoms='C:3 H:7 O:5',
        #thermo=(NASA([300.00, 1407.00],
                    #[ 2.83411409E+00,  6.29377015E-02, -5.35228290E-05,
                    #2.33491916E-08, -4.06934890E-12, -3.80451295E+04,
                    #2.27789799E+01]),
                #NASA([1407.00, 5000.00],
                    #[ 2.05841169E+01,  1.60145253E-02, -5.42436519E-06,
                    #8.36749101E-10, -4.83324213E-14, -4.36087438E+04,
                    #-7.05205391E+01])))

#species(name='IQC3H5OTQ-I',
        #atoms='C:3 H:7 O:5',
        #thermo=(NASA([300.00, 1415.00],
                    #[ 4.80614714E+00,  5.68137508E-02, -4.39675105E-05,
                    #1.70399951E-08, -2.61590876E-12, -2.85442590E+04,
                    #1.23781282E+01]),
                #NASA([1415.00, 5000.00],
                    #[ 2.17590493E+01,  1.51627633E-02, -5.16409734E-06,
                    #7.99713840E-10, -4.63243290E-14, -3.41193762E+04,
                    #-7.77619217E+01])))

#species(name='IQC3H5OHQ-SJ',
        #atoms='C:3 H:7 O:5',
        #thermo=(NASA([300.00, 1430.00],
                    #[ 3.75150220E+00,  5.60926263E-02, -4.26964877E-05,
                    #1.62818685E-08, -2.44662581E-12, -3.16549841E+04,
                    #2.15564729E+01]),
                #NASA([1430.00, 5000.00],
                    #[ 2.01633063E+01,  1.57355922E-02, -5.28426264E-06,
                    #8.10536228E-10, -4.66373245E-14, -3.70242993E+04,
                    #-6.56524053E+01])))

#species(name='CHOCOHCH2OOH',
        #atoms='C:3 H:6 O:4',
        #thermo=(NASA([300.00, 1384.00],
                    #[ 1.90697565E+00,  5.29537602E-02, -3.95708722E-05,
                    #1.48146113E-08, -2.23976588E-12, -5.39646718E+04,
                    #2.60308355E+01]),
                #NASA([1384.00, 5000.00],
                    #[ 1.86816278E+01,  1.39658539E-02, -4.94259642E-06,
                    #7.85147166E-10, -4.62881931E-14, -5.97999375E+04,
                    #-6.41155772E+01])))

#species(name='CHOCHO',
        #atoms='C:2 H:2 O:2',
        #thermo=(NASA([300.00, 1386.00],
                    #[ 1.88105120E+00,  2.36386368E-02, -1.83443295E-05,
                    #6.84842963E-09, -9.92733674E-13, -2.69280190E+04,
                    #1.59154793E+01]),
                #NASA([1386.00, 5000.00],
                    #[ 9.75438561E+00,  4.97645947E-03, -1.74410483E-06,
                    #2.75586994E-10, -1.61969892E-14, -2.95832896E+04,
                    #-2.61878329E+01])))

#species(name='TQC3H5OHTO2',
        #atoms='C:3 H:7 O:5',
        #thermo=(NASA([300.00, 1387.00],
                    #[ 2.11774097E+00,  7.25997025E-02, -6.81667819E-05,
                    #3.13796075E-08, -5.68565078E-12, -4.35518506E+04,
                    #2.23586550E+01]),
                #NASA([1387.00, 5000.00],
                    #[ 2.02916342E+01,  1.92145529E-02, -8.58823433E-06,
                    #1.62509037E-09, -1.06225627E-13, -4.85976512E+04,
                    #-7.10879629E+01])))

#species(name='OHCOCOOHCH3',
        #atoms='C:3 H:6 O:4',
        #thermo=(NASA([300.00, 1393.00],
                    #[-1.96293682E-01,  6.45815031E-02, -5.49196882E-05,
                    #2.34355354E-08, -4.01181684E-12, -6.66388685E+04,
                    #3.25552353E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 2.01819586E+01,  1.34325042E-02, -4.81163867E-06,
                    #7.56987137E-10, -4.37448189E-14, -7.33623538E+04,
                    #-7.56343874E+01])))

#species(name='C2H3OO',
        #atoms='H:3 C:2 O:2',
        #thermo=(NASA([298.15, 1000.00],
                    #[ 1.09784776E+00,  2.95333237E-02, -2.27744360E-05,
                    #7.20559155E-09, -3.07929092E-13,  1.13996101E+04,
                    #2.13563583E+01]),
                #NASA([1000.00, 2000.00],
                    #[ 6.04483828E+00,  1.45511127E-02, -7.50974622E-06,
                    #1.83488280E-09, -1.66689681E-13,  1.01699244E+04,
                    #-3.71144913E+00])))

#species(name='CHCHO',
        #atoms='H:2 C:2 O:1',
        #thermo=(NASA([298.15, 1000.00],
                    #[ 2.33256751E+00,  1.62952986E-02, -9.72052177E-06,
                    #5.15124155E-10,  1.03836514E-12,  2.96585452E+04,
                    #1.39904923E+01]),
                #NASA([1000.00, 2000.00],
                    #[ 4.92632910E+00,  9.71712147E-03, -5.54855980E-06,
                    #1.53068537E-09, -1.64742462E-13,  2.89499494E+04,
                    #5.27874677E-01])))

#species(name='C4H71-3OH',
        #atoms='C:4 H:8 O:1',
        #thermo=(NASA([300.00, 1385.00],
                    #[ 7.16707858E-02,  5.04772909E-02, -3.50356983E-05,
                    #1.30539199E-08, -2.07783174E-12, -2.16983717E+04,
                    #2.82072724E+01]),
                #NASA([1385.00, 5000.00],
                    #[ 1.42401634E+01,  1.83038695E-02, -6.37213444E-06,
                    #1.00099740E-09, -5.85511411E-14, -2.68183960E+04,
                    #-4.83949774E+01])))

#species(name='C4H71-1OH',
        #atoms='C:4 H:8 O:1',
        #thermo=(NASA([300.00, 1402.00],
                    #[-4.65981165E-01,  5.56573217E-02, -4.50578305E-05,
                    #1.93726056E-08, -3.38967639E-12, -2.31007894E+04,
                    #2.79744544E+01]),
                #NASA([1402.00, 5000.00],
                    #[ 1.42586569E+01,  1.71932504E-02, -5.74956414E-06,
                    #8.79089774E-10, -5.04586493E-14, -2.78051048E+04,
                    #-4.96581579E+01])))

#species(name='C4H71-2OH',
        #atoms='C:4 H:8 O:1',
        #thermo=(NASA([300.00, 1404.00],
                    #[-1.28560619E+00,  6.35758020E-02, -5.85328191E-05,
                    #2.80452715E-08, -5.32187122E-12, -2.51538600E+04,
                    #2.93802791E+01]),
                #NASA([1404.00, 5000.00],
                    #[ 1.50658194E+01,  1.65276030E-02, -5.52197706E-06,
                    #8.43591643E-10, -4.83870716E-14, -2.99737946E+04,
                    #-5.53405024E+01])))

#species(name='PC4H8OH-2',
        #atoms='C:4 H:9 O:1',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 1.89476538E+00,  4.87803114E-02, -3.26702980E-05,
                    #1.17119949E-08, -1.77100238E-12, -1.07455371E+04,
                    #2.06226225E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 1.47198620E+01,  1.93691825E-02, -6.59356634E-06,
                    #1.02018538E-09, -5.90410746E-14, -1.53023402E+04,
                    #-4.85296396E+01])))

#species(name='SC4H8OH-1',
        #atoms='C:4 H:9 O:1',
        #thermo=(NASA([300.00, 1405.00],
                    #[ 1.78282749E+00,  5.18872497E-02, -3.89600849E-05,
                    #1.57818686E-08, -2.64133852E-12, -1.28305019E+04,
                    #2.00308244E+01]),
                #NASA([1405.00, 5000.00],
                    #[ 1.50517242E+01,  1.84748629E-02, -6.15327969E-06,
                    #9.38017339E-10, -5.37214281E-14, -1.71974443E+04,
                    #-5.03733848E+01])))

#species(name='C-2-C4H8',
        #atoms='C:4 H:8',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 5.44417817E+00, -5.20451694E-03,  9.62906577E-05,
                    #-1.20068814E-07,  4.68194825E-11, -2.91019842E+03,
                    #3.46050733E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 7.37356737E+00,  2.29134106E-02, -8.26008983E-06,
                    #1.33374014E-09, -7.96410694E-14, -5.01911609E+03,
                    #-1.44525321E+01])),
        #note='T5/09')

#species(name='C4H6-1',
        #atoms='C:4 H:6',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 2.42819263E+00,  2.49821955E-02,  6.27370548E-06,
                    #-2.61747866E-08,  1.26585079E-11,  1.80248564E+04,
                    #1.36683982E+01]),
                #NASA([1000.00, 6000.00],
                    #[ 7.81179394E+00,  1.79733772E-02, -6.61044149E-06,
                    #1.05501491E-09, -6.19297169E-14,  1.61770171E+04,
                    #-1.59658015E+01])),
        #note='L10/93')

#species(name='C3H5OCH2-1',
        #atoms='C:4 H:7 O:1',
        #thermo=(NASA([300.00, 1393.00],
                    #[ 9.16272411E-01,  4.50800020E-02, -3.08192305E-05,
                    #1.08179680E-08, -1.54953325E-12,  7.01857266E+03,
                    #2.44883193E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 1.37864631E+01,  1.57304959E-02, -5.39376646E-06,
                    #8.38852214E-10, -4.87289477E-14,  2.47659491E+03,
                    #-4.48877302E+01])))

#species(name='NC5H12',
        #atoms='C:5 H:12',
        #thermo=(NASA([300.00, 1393.00],
                    #[-2.99551806E-01,  5.94963054E-02, -3.41764359E-05,
                    #9.47896058E-09, -9.73675086E-13, -1.98959978E+04,
                    #2.75742132E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 1.58289132E+01,  2.59344752E-02, -8.83016276E-06,
                    #1.36654986E-09, -7.91029283E-14, -2.59397429E+04,
                    #-6.05558457E+01])),
        #note='9/8/14')

#species(name='C5H11-1',
        #atoms='C:5 H:11',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 9.83190194E-02,  5.58653605E-02, -3.28855625E-05,
                    #9.58366873E-09, -1.08641370E-12,  4.82065818E+03,
                    #2.86921367E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 1.51918481E+01,  2.40339049E-02, -8.19717624E-06,
                    #1.27002751E-09, -7.35727956E-14, -8.02148517E+02,
                    #-5.36479311E+01])),
        #note='9/8/14')

#species(name='C5H11-2',
        #atoms='C:5 H:11',
        #thermo=(NASA([300.00, 1395.00],
                    #[ 8.17689455E-01,  4.92653514E-02, -2.13785619E-05,
                    #1.85532454E-09,  7.06259101E-13,  3.26221492E+03,
                    #2.65876082E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 1.47177050E+01,  2.41667557E-02, -8.18647646E-06,
                    #1.26271867E-09, -7.29267957E-14, -2.26027771E+03,
                    #-5.06071115E+01])),
        #note='9/8/14')

#species(name='C5H11-3',
        #atoms='C:5 H:11',
        #thermo=(NASA([300.00, 1395.00],
                    #[ 8.17689455E-01,  4.92653514E-02, -2.13785619E-05,
                    #1.85532454E-09,  7.06259101E-13,  3.26221492E+03,
                    #2.58981266E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 1.47177050E+01,  2.41667557E-02, -8.18647646E-06,
                    #1.26271867E-09, -7.29267957E-14, -2.26027771E+03,
                    #-5.12965931E+01])),
        #note='9/8/14')

#species(name='C5H10-1',
        #atoms='C:5 H:10',
        #thermo=(NASA([300.00, 1390.00],
                    #[-1.65023816E-01,  5.30727359E-02, -3.10861587E-05,
                    #8.92413402E-09, -9.81619602E-13, -4.59376228E+03,
                    #2.80570113E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 1.43624894E+01,  2.26076154E-02, -7.70500843E-06,
                    #1.19329968E-09, -6.91126022E-14, -1.00192871E+04,
                    #-5.12512094E+01])),
        #note='9/8/14')

#species(name='C5H10-2',
        #atoms='C:5 H:10',
        #thermo=(NASA([300.00, 1385.00],
                    #[ 5.90528835E-01,  4.85275113E-02, -2.43231598E-05,
                    #4.86096027E-09, -1.13099201E-13, -5.99777644E+03,
                    #2.41816377E+01]),
                #NASA([1385.00, 5000.00],
                    #[ 1.39425521E+01,  2.28734997E-02, -7.77800113E-06,
                    #1.20284861E-09, -6.95972140E-14, -1.12165700E+04,
                    #-4.95379542E+01])),
        #note='9/8/14')

#species(name='C5H81-3',
        #atoms='C:5 H:8',
        #thermo=(NASA([300.00, 1385.00],
                    #[ 1.54882436E+00,  4.15042709E-02, -2.14359890E-05,
                    #4.71145517E-09, -2.42142508E-13,  9.05636062E+03,
                    #1.95665910E+01]),
                #NASA([1385.00, 5000.00],
                    #[ 1.29945372E+01,  1.92678312E-02, -6.58966712E-06,
                    #1.02295969E-09, -5.93441369E-14,  4.59040047E+03,
                    #-4.35689825E+01])),
        #note='9/8/14')

#species(name='C5H91-3',
        #atoms='C:5 H:9',
        #thermo=(NASA([300.00, 1390.00],
                    #[-4.68159636E-01,  5.13472600E-02, -3.05648794E-05,
                    #8.82809744E-09, -9.61458255E-13,  1.23781446E+04,
                    #2.83587835E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 1.39593933E+01,  2.09180210E-02, -7.14307074E-06,
                    #1.10781333E-09, -6.42268160E-14,  7.02651017E+03,
                    #-5.03103610E+01])),
        #note='9/8/14')

#species(name='C5H91-4',
        #atoms='C:5 H:9',
        #thermo=(NASA([300.00, 1387.00],
                    #[ 1.58797615E+00,  4.01577239E-02, -1.50062687E-05,
                    #-3.94608061E-10,  1.02064182E-12,  1.84683983E+04,
                    #2.34273438E+01]),
                #NASA([1387.00, 5000.00],
                    #[ 1.29229725E+01,  2.10931621E-02, -7.14227112E-06,
                    #1.10137556E-09, -6.35983966E-14,  1.38192408E+04,
                    #-4.00293126E+01])),
        #note='9/8/14')

#species(name='C5H91-5',
        #atoms='C:5 H:9',
        #thermo=(NASA([300.00, 1391.00],
                    #[ 2.07676634E-01,  4.96049077E-02, -3.00621602E-05,
                    #9.19962077E-09, -1.13246061E-12,  2.01251866E+04,
                    #2.85856261E+01]),
                #NASA([1391.00, 5000.00],
                    #[ 1.37334869E+01,  2.07003006E-02, -7.06962737E-06,
                    #1.09638953E-09, -6.35591113E-14,  1.51176253E+04,
                    #-4.50794850E+01])),
        #note='9/8/14')

#species(name='C5H92-4',
        #atoms='C:5 H:9',
        #thermo=(NASA([300.00, 1384.00],
                    #[ 2.55028348E-01,  4.69860501E-02, -2.40447582E-05,
                    #4.89006694E-09, -1.15557679E-13,  1.09775103E+04,
                    #2.46226810E+01]),
                #NASA([1384.00, 5000.00],
                    #[ 1.35637956E+01,  2.11621316E-02, -7.20826071E-06,
                    #1.11611006E-09, -6.46370085E-14,  5.82258629E+03,
                    #-4.87323606E+01])),
        #note='9/8/14')

#species(name='C5H92-5',
        #atoms='C:5 H:9',
        #thermo=(NASA([300.00, 1386.00],
                    #[ 9.96363945E-01,  4.49148898E-02, -2.30951498E-05,
                    #5.01698735E-09, -2.38646470E-13,  1.87161506E+04,
                    #2.45616052E+01]),
                #NASA([1386.00, 5000.00],
                    #[ 1.32988753E+01,  2.09845923E-02, -7.14999241E-06,
                    #1.10717347E-09, -6.41183619E-14,  1.39270369E+04,
                    #-4.32758927E+01])),
        #note='9/8/14')

#species(name='C5H9O1-3',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1379.00],
                    #[ 2.49666166E+00,  5.20326105E-02, -3.09828268E-05,
                    #9.02567482E-09, -1.04311127E-12,  1.54758549E+03,
                    #1.70868328E+01]),
                #NASA([1379.00, 5000.00],
                    #[ 1.78860333E+01,  2.04837367E-02, -7.16613461E-06,
                    #1.12948597E-09, -6.62221468E-14, -4.33139568E+03,
                    #-6.72851315E+01])),
        #note='9/8/14')

#species(name='C5H9O2-4',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1376.00],
                    #[ 3.38899911E+00,  4.71704753E-02, -2.40389823E-05,
                    #4.99166621E-09, -2.03242522E-13,  1.15232772E+02,
                    #1.25183625E+01]),
                #NASA([1376.00, 5000.00],
                    #[ 1.73707371E+01,  2.09044827E-02, -7.30723872E-06,
                    #1.15106814E-09, -6.74601377E-14, -5.48020434E+03,
                    #-6.50109045E+01])),
        #note='9/8/14')

#species(name='C5H11O2H-1',
        #atoms='C:5 H:12 O:2',
        #thermo=(NASA([300.00, 1393.00],
                    #[ 7.56966700E-01,  7.13994143E-02, -4.75006334E-05,
                    #1.62337006E-08, -2.26834390E-12, -3.00009521E+04,
                    #2.85558826E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 2.08826791E+01,  2.61587380E-02, -8.96860003E-06,
                    #1.39463023E-09, -8.10033260E-14, -3.71864314E+04,
                    #-8.01917306E+01])),
        #note='9/8/14')

#species(name='C5H11O2H-2',
        #atoms='C:5 H:12 O:2',
        #thermo=(NASA([300.00, 1400.00],
                    #[ 1.06841342E+00,  7.36472213E-02, -5.31234084E-05,
                    #2.02380955E-08, -3.18671855E-12, -3.21746715E+04,
                    #2.51901593E+01]),
                #NASA([1400.00, 5000.00],
                    #[ 2.11347974E+01,  2.56494323E-02, -8.72650240E-06,
                    #1.34983451E-09, -7.81086942E-14, -3.90416300E+04,
                    #-8.21964012E+01])),
        #note='9/8/14')

#species(name='C5H11O2H-3',
        #atoms='C:5 H:12 O:2',
        #thermo=(NASA([300.00, 1400.00],
                    #[ 1.06841342E+00,  7.36472213E-02, -5.31234084E-05,
                    #2.02380955E-08, -3.18671855E-12, -3.21746715E+04,
                    #2.38111961E+01]),
                #NASA([1400.00, 5000.00],
                    #[ 2.11347974E+01,  2.56494323E-02, -8.72650240E-06,
                    #1.34983451E-09, -7.81086942E-14, -3.90416300E+04,
                    #-8.35753644E+01])),
        #note='9/8/14')

#species(name='C5H11O2-1',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1392.00],
                    #[ 1.47849612E+00,  6.56222617E-02, -4.24002295E-05,
                    #1.39455731E-08, -1.86035665E-12, -1.33065245E+04,
                    #2.57184992E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 1.99508904E+01,  2.48182054E-02, -8.50415197E-06,
                    #1.32191009E-09, -7.67596518E-14, -1.99782879E+04,
                    #-7.43587406E+01])),
        #note='9/8/14')

#species(name='C5H11O2-2',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1401.00],
                    #[ 1.95945265E+00,  6.73012650E-02, -4.74261817E-05,
                    #1.76453313E-08, -2.71599328E-12, -1.55102022E+04,
                    #2.15326067E+01]),
                #NASA([1401.00, 5000.00],
                    #[ 2.01379084E+01,  2.43048677E-02, -8.24744582E-06,
                    #1.27344965E-09, -7.35953601E-14, -2.17777897E+04,
                    #-7.59194124E+01])),
        #note='9/8/14')

#species(name='C5H11O2-3',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1401.00],
                    #[ 1.95945265E+00,  6.73012650E-02, -4.74261817E-05,
                    #1.76453313E-08, -2.71599328E-12, -1.55102022E+04,
                    #2.01536434E+01]),
                #NASA([1401.00, 5000.00],
                    #[ 2.01379084E+01,  2.43048677E-02, -8.24744582E-06,
                    #1.27344965E-09, -7.35953601E-14, -2.17777897E+04,
                    #-7.72983757E+01])),
        #note='9/8/14')

#species(name='C5H11O-1',
        #atoms='C:5 H:11 O:1',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 1.61443159E+00,  5.84449099E-02, -3.26027513E-05,
                    #7.76525028E-09, -4.48739932E-13, -1.20210227E+04,
                    #2.23620687E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 1.87118024E+01,  2.36054696E-02, -8.05610528E-06,
                    #1.24900841E-09, -7.23989034E-14, -1.84133036E+04,
                    #-7.11769229E+01])),
        #note='9/8/14')

#species(name='C5H11O-2',
        #atoms='C:5 H:11 O:1',
        #thermo=(NASA([300.00, 1407.00],
                    #[ 1.80797841E+00,  5.95687062E-02, -3.62936785E-05,
                    #1.05785660E-08, -1.11086649E-12, -1.41294991E+04,
                    #2.01085450E+01]),
                #NASA([1407.00, 5000.00],
                    #[ 1.85178187E+01,  2.33560329E-02, -7.87896503E-06,
                    #1.21190761E-09, -6.98572825E-14, -2.01488741E+04,
                    #-7.05224357E+01])),
        #note='9/8/14')

#species(name='C5H11O-3',
        #atoms='C:5 H:11 O:1',
        #thermo=(NASA([300.00, 1407.00],
                    #[ 1.80797841E+00,  5.95687062E-02, -3.62936785E-05,
                    #1.05785660E-08, -1.11086649E-12, -1.41294991E+04,
                    #1.87295817E+01]),
                #NASA([1407.00, 5000.00],
                    #[ 1.85178187E+01,  2.33560329E-02, -7.87896503E-06,
                    #1.21190761E-09, -6.98572825E-14, -2.01488741E+04,
                    #-7.19013990E+01])),
        #note='9/8/14')

#species(name='C5H10OOH1-2',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 2.65531081E+00,  6.13777489E-02, -3.72526890E-05,
                    #1.10683497E-08, -1.26575773E-12, -6.87569412E+03,
                    #2.18610512E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 2.00795991E+01,  2.43589475E-02, -8.36202795E-06,
                    #1.30140406E-09, -7.56329790E-14, -1.33166011E+04,
                    #-7.30604965E+01])),
        #note='9/8/14')

#species(name='C5H10OOH1-3',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 2.21641946E+00,  5.96426514E-02, -3.30151174E-05,
                    #7.82801873E-09, -4.53626906E-13, -6.89038067E+03,
                    #2.53165655E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 1.94584083E+01,  2.46031733E-02, -8.38534822E-06,
                    #1.29883762E-09, -7.52367289E-14, -1.33548974E+04,
                    #-6.90616123E+01])),
        #note='9/8/14')

#species(name='C5H10OOH1-4',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 2.21641946E+00,  5.96426514E-02, -3.30151174E-05,
                    #7.82801873E-09, -4.53626906E-13, -6.89038067E+03,
                    #2.53165655E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 1.94584083E+01,  2.46031733E-02, -8.38534822E-06,
                    #1.29883762E-09, -7.52367289E-14, -1.33548974E+04,
                    #-6.90616123E+01])),
        #note='9/8/14')

#species(name='C5H10OOH1-5',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1393.00],
                    #[ 9.60306204E-01,  6.85338215E-02, -4.72580697E-05,
                    #1.69913164E-08, -2.53055987E-12, -5.25227210E+03,
                    #2.99005842E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 2.02627427E+01,  2.43041392E-02, -8.36491479E-06,
                    #1.30408546E-09, -7.58781596E-14, -1.20792388E+04,
                    #-7.41332935E+01])),
        #note='9/8/14')

#species(name='C5H10OOH2-1',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1401.00],
                    #[ 1.97482622E+00,  6.85026751E-02, -4.88793184E-05,
                    #1.80505063E-08, -2.71515676E-12, -7.58936343E+03,
                    #2.31709703E+01]),
                #NASA([1401.00, 5000.00],
                    #[ 2.11683533E+01,  2.30995637E-02, -7.85421525E-06,
                    #1.21454488E-09, -7.02697376E-14, -1.41652579E+04,
                    #-7.96524097E+01])),
        #note='9/8/14')

#species(name='C5H10OOH2-3',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1410.00],
                    #[ 3.03844851E+00,  6.35523371E-02, -4.29455468E-05,
                    #1.50665567E-08, -2.16139351E-12, -9.06691319E+03,
                    #1.81118164E+01]),
                #NASA([1410.00, 5000.00],
                    #[ 2.03599510E+01,  2.36444020E-02, -8.00996599E-06,
                    #1.23543694E-09, -7.13457143E-14, -1.51264004E+04,
                    #-7.50912668E+01])),
        #note='9/8/14')

#species(name='C5H10OOH2-4',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1410.00],
                    #[ 2.69411113E+00,  6.13272504E-02, -3.77802938E-05,
                    #1.11809673E-08, -1.20043769E-12, -9.09500363E+03,
                    #2.11388648E+01]),
                #NASA([1410.00, 5000.00],
                    #[ 1.98843630E+01,  2.37470675E-02, -7.98273943E-06,
                    #1.22497373E-09, -7.04946509E-14, -1.52413050E+04,
                    #-7.19593692E+01])),
        #note='9/8/14')

#species(name='C5H10OOH2-5',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1401.00],
                    #[ 1.46239008E+00,  7.00773603E-02, -5.18935110E-05,
                    #2.03023923E-08, -3.27212476E-12, -7.45896384E+03,
                    #2.56255897E+01]),
                #NASA([1401.00, 5000.00],
                    #[ 2.05595155E+01,  2.35923560E-02, -8.01736916E-06,
                    #1.23918525E-09, -7.16674070E-14, -1.39059417E+04,
                    #-7.62751847E+01])),
        #note='9/8/14')

#species(name='C5H10OOH3-1',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1401.00],
                    #[ 1.46239008E+00,  7.00773603E-02, -5.18935110E-05,
                    #2.03023923E-08, -3.27212476E-12, -7.45896384E+03,
                    #2.49361080E+01]),
                #NASA([1401.00, 5000.00],
                    #[ 2.05595155E+01,  2.35923560E-02, -8.01736916E-06,
                    #1.23918525E-09, -7.16674070E-14, -1.39059417E+04,
                    #-7.69646663E+01])),
        #note='9/8/14')

#species(name='C5H10OOH3-2',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1410.00],
                    #[ 3.03844851E+00,  6.35523371E-02, -4.29455468E-05,
                    #1.50665567E-08, -2.16139351E-12, -9.06691319E+03,
                    #1.74223348E+01]),
                #NASA([1410.00, 5000.00],
                    #[ 2.03599510E+01,  2.36444020E-02, -8.00996599E-06,
                    #1.23543694E-09, -7.13457143E-14, -1.51264004E+04,
                    #-7.57807484E+01])),
        #note='9/8/14')

#species(name='C5H10O1-2',
        #atoms='C:5 H:10 O:1',
        #thermo=(NASA([300.00, 1424.00],
                    #[-5.22384549E+00,  7.95280569E-02, -6.27918049E-05,
                    #2.51187341E-08, -3.99106436E-12, -1.76152192E+04,
                    #5.15328120E+01]),
                #NASA([1424.00, 5000.00],
                    #[ 1.76453140E+01,  2.17647043E-02, -7.28831391E-06,
                    #1.11571673E-09, -6.41056178E-14, -2.49627812E+04,
                    #-6.94780751E+01])),
        #note='9/8/14')

#species(name='C5H10O1-3',
        #atoms='C:5 H:10 O:1',
        #thermo=(NASA([300.00, 1445.00],
                    #[-7.38770073E+00,  8.62948016E-02, -7.34565945E-05,
                    #3.18221366E-08, -5.43057938E-12, -1.80648633E+04,
                    #6.09884425E+01]),
                #NASA([1445.00, 5000.00],
                    #[ 1.61536944E+01,  2.22699220E-02, -7.20967982E-06,
                    #1.07781327E-09, -6.08828236E-14, -2.51216994E+04,
                    #-6.18744318E+01])),
        #note='9/8/14')

#species(name='C5H10O1-4',
        #atoms='C:5 H:10 O:1',
        #thermo=(NASA([300.00, 1470.00],
                    #[-7.99511075E+00,  8.48027993E-02, -6.84820969E-05,
                    #2.79646707E-08, -4.48601572E-12, -2.79347448E+04,
                    #6.36835324E+01]),
                #NASA([1470.00, 5000.00],
                    #[ 1.56838302E+01,  2.28597182E-02, -7.36869093E-06,
                    #1.09855828E-09, -6.19425653E-14, -3.52365120E+04,
                    #-6.06986511E+01])),
        #note='9/8/14')

#species(name='C5H10O1-5',
        #atoms='C:5 H:10 O:1',
        #thermo=(NASA([300.00, 1447.00],
                    #[-8.45066821E+00,  8.36041573E-02, -6.24375031E-05,
                    #2.32498119E-08, -3.38258239E-12, -2.79060038E+04,
                    #6.19634163E+01]),
                #NASA([1447.00, 5000.00],
                    #[ 1.58722584E+01,  2.41484940E-02, -8.02830905E-06,
                    #1.22306406E-09, -7.00384536E-14, -3.58667196E+04,
                    #-6.73548909E+01])),
        #note='9/8/14')

#species(name='C5H10O2-3',
        #atoms='C:5 H:10 O:1',
        #thermo=(NASA([300.00, 1415.00],
                    #[-6.83688134E+00,  9.00167224E-02, -8.09486382E-05,
                    #3.71523167E-08, -6.73312332E-12, -1.94907358E+04,
                    #5.73459463E+01]),
                #NASA([1415.00, 5000.00],
                    #[ 1.76695497E+01,  2.16741351E-02, -7.23760886E-06,
                    #1.10546424E-09, -6.34044008E-14, -2.67999740E+04,
                    #-7.01762004E+01])),
        #note='9/8/14')

#species(name='C5H10O2-4',
        #atoms='C:5 H:10 O:1',
        #thermo=(NASA([300.00, 1396.00],
                    #[-6.91381150E+00,  8.65695113E-02, -7.51445092E-05,
                    #3.29307332E-08, -5.77631101E-12, -2.02336617E+04,
                    #5.66267224E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 1.58873476E+01,  2.48573824E-02, -1.03962573E-05,
                    #1.95192708E-09, -1.29464891E-13, -2.72224431E+04,
                    #-6.26943208E+01])),
        #note='9/8/14')

#species(name='C5H10OOH1-2O2',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 2.92123878E+00,  7.67601721E-02, -5.10111840E-05,
                    #1.59643052E-08, -1.82979631E-12, -2.55486792E+04,
                    #2.27306226E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 2.71105703E+01,  2.34082367E-02, -8.11719657E-06,
                    #1.27230537E-09, -7.43221831E-14, -3.41406290E+04,
                    #-1.08093476E+02])),
        #note='9/14')

#species(name='C5H10OOH1-3O2',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 2.92123878E+00,  7.67601721E-02, -5.10111840E-05,
                    #1.59643052E-08, -1.82979631E-12, -2.55486792E+04,
                    #2.27306226E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 2.71105703E+01,  2.34082367E-02, -8.11719657E-06,
                    #1.27230537E-09, -7.43221831E-14, -3.41406290E+04,
                    #-1.08093476E+02])),
        #note='9/14')

#species(name='C5H10OOH1-4O2',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 2.92123878E+00,  7.67601721E-02, -5.10111840E-05,
                    #1.59643052E-08, -1.82979631E-12, -2.55486792E+04,
                    #2.27306226E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 2.71105703E+01,  2.34082367E-02, -8.11719657E-06,
                    #1.27230537E-09, -7.43221831E-14, -3.41406290E+04,
                    #-1.08093476E+02])),
        #note='9/14')

#species(name='C5H10OOH1-5O2',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1390.00],
                    #[ 2.61115242E+00,  7.58124541E-02, -5.03939636E-05,
                    #1.63241251E-08, -2.04839461E-12, -2.33970141E+04,
                    #2.58821117E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 2.58212210E+01,  2.46140917E-02, -8.55565938E-06,
                    #1.34293065E-09, -7.85178268E-14, -3.17158002E+04,
                    #-9.97698267E+01])),
        #note='9/14')

#species(name='C5H10OOH2-1O2',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 2.92123878E+00,  7.67601721E-02, -5.10111840E-05,
                    #1.59643052E-08, -1.82979631E-12, -2.55486792E+04,
                    #2.27306226E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 2.71105703E+01,  2.34082367E-02, -8.11719657E-06,
                    #1.27230537E-09, -7.43221831E-14, -3.41406290E+04,
                    #-1.08093476E+02])),
        #note='9/14')

#species(name='C5H10OOH2-3O2',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1405.00],
                    #[ 2.94958658E+00,  7.94696642E-02, -6.12300085E-05,
                    #2.45360891E-08, -3.99243941E-12, -2.76263457E+04,
                    #2.10066180E+01]),
                #NASA([1405.00, 5000.00],
                    #[ 2.51554088E+01,  2.42476887E-02, -8.22109032E-06,
                    #1.26886123E-09, -7.33157779E-14, -3.49522107E+04,
                    #-9.69775586E+01])),
        #note='9/14')

#species(name='C5H10OOH2-4O2',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1405.00],
                    #[ 2.94958658E+00,  7.94696642E-02, -6.12300085E-05,
                    #2.45360891E-08, -3.99243941E-12, -2.76263457E+04,
                    #2.10066180E+01]),
                #NASA([1405.00, 5000.00],
                    #[ 2.51554088E+01,  2.42476887E-02, -8.22109032E-06,
                    #1.26886123E-09, -7.33157779E-14, -3.49522107E+04,
                    #-9.69775586E+01])),
        #note='9/14')

#species(name='C5H10OOH2-5O2',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 2.92123878E+00,  7.67601721E-02, -5.10111840E-05,
                    #1.59643052E-08, -1.82979631E-12, -2.55486792E+04,
                    #2.27306226E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 2.71105703E+01,  2.34082367E-02, -8.11719657E-06,
                    #1.27230537E-09, -7.43221831E-14, -3.41406290E+04,
                    #-1.08093476E+02])),
        #note='9/14')

#species(name='C5H10OOH3-1O2',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 2.92123878E+00,  7.67601721E-02, -5.10111840E-05,
                    #1.59643052E-08, -1.82979631E-12, -2.55486792E+04,
                    #2.27306226E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 2.71105703E+01,  2.34082367E-02, -8.11719657E-06,
                    #1.27230537E-09, -7.43221831E-14, -3.41406290E+04,
                    #-1.08093476E+02])),
        #note='9/14')

#species(name='C5H10OOH3-2O2',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1405.00],
                    #[ 2.94958658E+00,  7.94696642E-02, -6.12300085E-05,
                    #2.45360891E-08, -3.99243941E-12, -2.76263457E+04,
                    #2.10066180E+01]),
                #NASA([1405.00, 5000.00],
                    #[ 2.51554088E+01,  2.42476887E-02, -8.22109032E-06,
                    #1.26886123E-09, -7.33157779E-14, -3.49522107E+04,
                    #-9.69775586E+01])),
        #note='9/14')

#species(name='C5H91-2,3OOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1405.00],
                    #[ 2.91226160E+00,  7.95651114E-02, -6.13230261E-05,
                    #2.44460527E-08, -3.95034020E-12, -1.96516956E+04,
                    #2.32188202E+01]),
                #NASA([1405.00, 5000.00],
                    #[ 2.55056147E+01,  2.36990320E-02, -8.07032909E-06,
                    #1.24934920E-09, -7.23427367E-14, -2.71353489E+04,
                    #-9.69348818E+01])),
        #note='9/14')

#species(name='C5H91-2,4OOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1405.00],
                    #[ 2.91226160E+00,  7.95651114E-02, -6.13230261E-05,
                    #2.44460527E-08, -3.95034020E-12, -1.96516956E+04,
                    #2.32188202E+01]),
                #NASA([1405.00, 5000.00],
                    #[ 2.55056147E+01,  2.36990320E-02, -8.07032909E-06,
                    #1.24934920E-09, -7.23427367E-14, -2.71353489E+04,
                    #-9.69348818E+01])),
        #note='9/14')

#species(name='C5H91-2,5OOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1397.00],
                    #[ 2.56086129E+00,  7.80408775E-02, -5.74863330E-05,
                    #2.18761227E-08, -3.39924240E-12, -1.74851061E+04,
                    #2.66588551E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 2.50741397E+01,  2.44380043E-02, -8.40704928E-06,
                    #1.31038422E-09, -7.62381905E-14, -2.51871158E+04,
                    #-9.38667514E+01])),
        #note='9/14')

#species(name='C5H91-3,4OOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1405.00],
                    #[ 2.48775154E+00,  8.15271831E-02, -6.44013124E-05,
                    #2.63443736E-08, -4.35804025E-12, -1.95635349E+04,
                    #2.50626503E+01]),
                #NASA([1405.00, 5000.00],
                    #[ 2.54680852E+01,  2.36307420E-02, -8.02474436E-06,
                    #1.23994908E-09, -7.17032304E-14, -2.70635665E+04,
                    #-9.67605580E+01])),
        #note='9/14')

#species(name='C5H91-3,5OOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1397.00],
                    #[ 2.50132978E+00,  7.93141150E-02, -5.51608762E-05,
                    #1.83963651E-08, -2.32880956E-12, -1.75129588E+04,
                    #2.64329747E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 2.75716451E+01,  2.26282394E-02, -7.85651555E-06,
                    #1.23251566E-09, -7.20426920E-14, -2.62702922E+04,
                    #-1.08648340E+02])),
        #note='9/14')

#species(name='C5H91-4,5OOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1397.00],
                    #[ 2.50132978E+00,  7.93141150E-02, -5.51608762E-05,
                    #1.83963651E-08, -2.32880956E-12, -1.75129588E+04,
                    #2.64329747E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 2.75716451E+01,  2.26282394E-02, -7.85651555E-06,
                    #1.23251566E-09, -7.20426920E-14, -2.62702922E+04,
                    #-1.08648340E+02])),
        #note='9/14')

#species(name='C5H92-1,3OOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1413.00],
                    #[ 3.42922915E+00,  7.33174178E-02, -5.05533009E-05,
                    #1.77709491E-08, -2.53421378E-12, -1.89212844E+04,
                    #2.26144479E+01]),
                #NASA([1413.00, 5000.00],
                    #[ 2.46778451E+01,  2.47896676E-02, -8.53168232E-06,
                    #1.33018418E-09, -7.74050315E-14, -2.64008851E+04,
                    #-9.18794025E+01])),
        #note='9/14')

#species(name='C5H92-1,4OOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1413.00],
                    #[ 3.42922915E+00,  7.33174178E-02, -5.05533009E-05,
                    #1.77709491E-08, -2.53421378E-12, -1.89212844E+04,
                    #2.26144479E+01]),
                #NASA([1413.00, 5000.00],
                    #[ 2.46778451E+01,  2.47896676E-02, -8.53168232E-06,
                    #1.33018418E-09, -7.74050315E-14, -2.64008851E+04,
                    #-9.18794025E+01])),
        #note='9/14')

#species(name='C5H92-1,5OOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1399.00],
                    #[ 3.43107332E+00,  7.25133538E-02, -4.53395240E-05,
                    #1.29499313E-08, -1.26534785E-12, -1.69000809E+04,
                    #2.37843821E+01]),
                #NASA([1399.00, 5000.00],
                    #[ 2.63836136E+01,  2.38373273E-02, -8.31679219E-06,
                    #1.30877666E-09, -7.66591074E-14, -2.52930801E+04,
                    #-1.01118754E+02])),
        #note='9/14')

#species(name='C5H92-3,4OOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1418.00],
                    #[ 3.76788328E+00,  7.47668469E-02, -5.41780015E-05,
                    #2.01841133E-08, -3.04867679E-12, -2.10823293E+04,
                    #1.92561961E+01]),
                #NASA([1418.00, 5000.00],
                    #[ 2.51070211E+01,  2.40686276E-02, -8.20467040E-06,
                    #1.27103065E-09, -7.36339262E-14, -2.83591157E+04,
                    #-9.49637963E+01])),
        #note='9/14')

#species(name='C5H92-3,5OOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1413.00],
                    #[ 3.42922915E+00,  7.33174178E-02, -5.05533009E-05,
                    #1.77709491E-08, -2.53421378E-12, -1.89212844E+04,
                    #2.26144479E+01]),
                #NASA([1413.00, 5000.00],
                    #[ 2.46778451E+01,  2.47896676E-02, -8.53168232E-06,
                    #1.33018418E-09, -7.74050315E-14, -2.64008851E+04,
                    #-9.18794025E+01])),
        #note='9/14')

#species(name='C5H92-4,5OOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1413.00],
                    #[ 4.06621416E+00,  7.20273834E-02, -4.95956461E-05,
                    #1.70908613E-08, -2.31584179E-12, -1.92523266E+04,
                    #1.99316930E+01]),
                #NASA([1413.00, 5000.00],
                    #[ 2.47514622E+01,  2.41459993E-02, -8.18238994E-06,
                    #1.26250806E-09, -7.29352454E-14, -2.63782993E+04,
                    #-9.11444444E+01])),
        #note='9/14')

#species(name='C5H93-1,2OOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1413.00],
                    #[ 3.42922915E+00,  7.33174178E-02, -5.05533009E-05,
                    #1.77709491E-08, -2.53421378E-12, -1.89212844E+04,
                    #2.26144479E+01]),
                #NASA([1413.00, 5000.00],
                    #[ 2.46778451E+01,  2.47896676E-02, -8.53168232E-06,
                    #1.33018418E-09, -7.74050315E-14, -2.64008851E+04,
                    #-9.18794025E+01])),
        #note='9/14')

#species(name='C5H93-1,4OOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1413.00],
                    #[ 3.42922915E+00,  7.33174178E-02, -5.05533009E-05,
                    #1.77709491E-08, -2.53421378E-12, -1.89212844E+04,
                    #2.26144479E+01]),
                #NASA([1413.00, 5000.00],
                    #[ 2.46778451E+01,  2.47896676E-02, -8.53168232E-06,
                    #1.33018418E-09, -7.74050315E-14, -2.64008851E+04,
                    #-9.18794025E+01])),
        #note='9/14')

#species(name='C5H93-1,5OOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1399.00],
                    #[ 3.67511835E+00,  6.98557277E-02, -4.39193701E-05,
                    #1.31290152E-08, -1.43358803E-12, -1.70587842E+04,
                    #2.30262907E+01]),
                #NASA([1399.00, 5000.00],
                    #[ 2.44184921E+01,  2.49501505E-02, -8.57440942E-06,
                    #1.33558596E-09, -7.76697869E-14, -2.45529807E+04,
                    #-8.95288488E+01])),
        #note='9/14')

#species(name='C5H93-2,4OOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1418.00],
                    #[ 3.76788328E+00,  7.47668469E-02, -5.41780015E-05,
                    #2.01841133E-08, -3.04867679E-12, -2.10823293E+04,
                    #1.85667144E+01]),
                #NASA([1418.00, 5000.00],
                    #[ 2.51070211E+01,  2.40686276E-02, -8.20467040E-06,
                    #1.27103065E-09, -7.36339262E-14, -2.83591157E+04,
                    #-9.56532779E+01])),
        #note='9/14')

#species(name='C5H91-3OOH',
        #atoms='C:5 H:10 O:2',
        #thermo=(NASA([300.00, 1394.00],
                    #[-1.89589507E+00,  8.18483531E-02, -6.75161021E-05,
                    #2.83331852E-08, -4.76826956E-12, -1.47140793E+04,
                    #3.85399143E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 2.24660467E+01,  2.00408413E-02, -6.84550040E-06,
                    #1.06266050E-09, -6.16741604E-14, -2.26148518E+04,
                    #-9.04469895E+01])),
        #note='9/8/14')

#species(name='C5H91-4OOH',
        #atoms='C:5 H:10 O:2',
        #thermo=(NASA([300.00, 1402.00],
                    #[ 1.38243484E+00,  6.67661361E-02, -4.96344237E-05,
                    #1.94278468E-08, -3.12175781E-12, -1.69086691E+04,
                    #2.40815942E+01]),
                #NASA([1402.00, 5000.00],
                    #[ 1.96307052E+01,  2.22000091E-02, -7.52489271E-06,
                    #1.16109117E-09, -6.70722374E-14, -2.30400033E+04,
                    #-7.32135194E+01])),
        #note='9/8/14')

#species(name='C5H91-5OOH',
        #atoms='C:5 H:10 O:2',
        #thermo=(NASA([300.00, 1392.00],
                    #[ 9.93373748E-01,  6.46025217E-02, -4.40191862E-05,
                    #1.55375290E-08, -2.26236370E-12, -1.47156820E+04,
                    #2.78588419E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 1.93103284E+01,  2.29913370E-02, -7.91269843E-06,
                    #1.23358507E-09, -7.17773402E-14, -2.12246688E+04,
                    #-7.09825558E+01])),
        #note='9/8/14')

#species(name='C5H92-1OOH',
        #atoms='C:5 H:10 O:2',
        #thermo=(NASA([300.00, 1382.00],
                    #[ 5.77880009E-01,  6.91528527E-02, -5.00924279E-05,
                    #1.88100079E-08, -2.91765412E-12, -1.44043272E+04,
                    #2.97056559E+01]),
                #NASA([1382.00, 5000.00],
                    #[ 2.10623695E+01,  2.15772208E-02, -7.44876502E-06,
                    #1.16423903E-09, -6.78819533E-14, -2.15816942E+04,
                    #-8.04705541E+01])),
        #note='9/8/14')

#species(name='C5H92-4OOH',
        #atoms='C:5 H:10 O:2',
        #thermo=(NASA([300.00, 1389.00],
                    #[-1.10664142E+00,  7.70652343E-02, -6.01781689E-05,
                    #2.37846489E-08, -3.77298555E-12, -1.61217479E+04,
                    #3.45219377E+01]),
                #NASA([1389.00, 5000.00],
                    #[ 2.22714879E+01,  1.99891269E-02, -6.78760139E-06,
                    #1.05004677E-09, -6.08129638E-14, -2.39098291E+04,
                    #-9.00213678E+01])),
        #note='9/8/14')

#species(name='C5H92-5OOH',
        #atoms='C:5 H:10 O:2',
        #thermo=(NASA([300.00, 1389.00],
                    #[ 1.76944896E+00,  6.00153444E-02, -3.72031347E-05,
                    #1.14551487E-08, -1.39226390E-12, -1.61243818E+04,
                    #2.38818885E+01]),
                #NASA([1389.00, 5000.00],
                    #[ 1.88657282E+01,  2.33222202E-02, -8.01683595E-06,
                    #1.24882522E-09, -7.26244846E-14, -2.24129903E+04,
                    #-6.91316665E+01])),
        #note='9/8/14')

#species(name='C5H9O1-2OOH-3',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1422.00],
                    #[-1.63458528E+00,  8.37428576E-02, -6.46448858E-05,
                    #2.45435115E-08, -3.63637746E-12, -3.02443795E+04,
                    #3.81009801E+01]),
                #NASA([1422.00, 5000.00],
                    #[ 2.40742385E+01,  2.10049879E-02, -7.12965508E-06,
                    #1.10196053E-09, -6.37577378E-14, -3.86905727E+04,
                    #-9.86614868E+01])),
        #note='9/14')

#species(name='C5H9O1-2OOH-4',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1422.00],
                    #[-1.63458528E+00,  8.37428576E-02, -6.46448858E-05,
                    #2.45435115E-08, -3.63637746E-12, -3.02443795E+04,
                    #3.81009801E+01]),
                #NASA([1422.00, 5000.00],
                    #[ 2.40742385E+01,  2.10049879E-02, -7.12965508E-06,
                    #1.10196053E-09, -6.37577378E-14, -3.86905727E+04,
                    #-9.86614868E+01])),
        #note='9/14')

#species(name='C5H9O1-2OOH-5',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1425.00],
                    #[-2.91502569E+00,  8.63720024E-02, -6.80858869E-05,
                    #2.66150873E-08, -4.08265327E-12, -2.79298051E+04,
                    #4.58804153E+01]),
                #NASA([1425.00, 5000.00],
                    #[ 2.31521586E+01,  2.15716913E-02, -7.27787584E-06,
                    #1.12021946E-09, -6.46248595E-14, -3.63754336E+04,
                    #-9.23672941E+01])),
        #note='9/14')

#species(name='C5H9O1-3OOH-2',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1421.00],
                    #[-4.67708778E+00,  9.25191737E-02, -7.56245586E-05,
                    #3.08655754E-08, -4.96974700E-12, -3.05178886E+04,
                    #5.19807893E+01]),
                #NASA([1421.00, 5000.00],
                    #[ 2.29238871E+01,  2.22244781E-02, -7.50929996E-06,
                    #1.15689292E-09, -6.67794310E-14, -3.93149495E+04,
                    #-9.38426660E+01])),
        #note='9/14')

#species(name='C5H9O1-3OOH-4',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1421.00],
                    #[-4.67708778E+00,  9.25191737E-02, -7.56245586E-05,
                    #3.08655754E-08, -4.96974700E-12, -3.05178886E+04,
                    #5.19807893E+01]),
                #NASA([1421.00, 5000.00],
                    #[ 2.29238871E+01,  2.22244781E-02, -7.50929996E-06,
                    #1.15689292E-09, -6.67794310E-14, -3.93149495E+04,
                    #-9.38426660E+01])),
        #note='9/14')

#species(name='C5H9O1-3OOH-5',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1421.00],
                    #[-5.55071358E+00,  9.29081113E-02, -7.49705054E-05,
                    #3.02210571E-08, -4.80852681E-12, -2.82561346E+04,
                    #5.79510604E+01]),
                #NASA([1421.00, 5000.00],
                    #[ 2.21702811E+01,  2.28760461E-02, -7.73630997E-06,
                    #1.19254044E-09, -6.88630410E-14, -3.71502930E+04,
                    #-8.87113724E+01])),
        #note='9/14')

#species(name='C5H9O1-4OOH-2',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1417.00],
                    #[-9.18084516E+00,  1.08786371E-01, -9.86756121E-05,
                    #4.49207522E-08, -8.03087742E-12, -3.98013189E+04,
                    #7.26760969E+01]),
                #NASA([1417.00, 5000.00],
                    #[ 2.17260410E+01,  2.34146274E-02, -7.86187868E-06,
                    #1.20571561E-09, -6.93653621E-14, -4.90509096E+04,
                    #-8.83590566E+01])),
        #note='9/14')

#species(name='C5H9O1-4OOH-3',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1417.00],
                    #[-9.18084516E+00,  1.08786371E-01, -9.86756121E-05,
                    #4.49207522E-08, -8.03087742E-12, -3.98013189E+04,
                    #7.26760969E+01]),
                #NASA([1417.00, 5000.00],
                    #[ 2.17260410E+01,  2.34146274E-02, -7.86187868E-06,
                    #1.20571561E-09, -6.93653621E-14, -4.90509096E+04,
                    #-8.83590566E+01])),
        #note='9/14')

#species(name='C5H9O1-4OOH-5',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1411.00],
                    #[-9.29661331E+00,  1.04958911E-01, -9.00892767E-05,
                    #3.89843596E-08, -6.68819849E-12, -3.76382956E+04,
                    #7.52773731E+01]),
                #NASA([1411.00, 5000.00],
                    #[ 2.13868464E+01,  2.42467386E-02, -8.26745278E-06,
                    #1.28123479E-09, -7.42527951E-14, -4.72411464E+04,
                    #-8.60552096E+01])),
        #note='9/14')

#species(name='C5H9O1-5OOH-2',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1407.00],
                    #[-9.63082131E+00,  1.06539535E-01, -8.95953916E-05,
                    #3.77712411E-08, -6.32248900E-12, -3.97499469E+04,
                    #7.18166046E+01]),
                #NASA([1407.00, 5000.00],
                    #[ 2.24618796E+01,  2.42252273E-02, -8.36381388E-06,
                    #1.30712783E-09, -7.62010208E-14, -5.00102121E+04,
                    #-9.76835754E+01])),
        #note='9/14')

#species(name='C5H9O1-5OOH-3',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1407.00],
                    #[-9.63082131E+00,  1.06539535E-01, -8.95953916E-05,
                    #3.77712411E-08, -6.32248900E-12, -3.97499469E+04,
                    #7.11220903E+01]),
                #NASA([1407.00, 5000.00],
                    #[ 2.24618796E+01,  2.42252273E-02, -8.36381388E-06,
                    #1.30712783E-09, -7.62010208E-14, -5.00102121E+04,
                    #-9.83780898E+01])),
        #note='9/14')

#species(name='C5H9O2-3OOH-1',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1436.00],
                    #[-2.95211763E+00,  8.90845071E-02, -7.32339295E-05,
                    #2.98425405E-08, -4.76173344E-12, -3.00280666E+04,
                    #4.45334764E+01]),
                #NASA([1436.00, 5000.00],
                    #[ 2.37309690E+01,  2.06751517E-02, -6.88444402E-06,
                    #1.05044759E-09, -6.02362275E-14, -3.84362991E+04,
                    #-9.61945548E+01])),
        #note='9/14')

#species(name='C5H9O2-3OOH-4',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1430.00],
                    #[-4.57026018E+00,  1.05196229E-01, -1.05063534E-04,
                    #5.15547206E-08, -9.70822253E-12, -3.20442434E+04,
                    #4.90445735E+01]),
                #NASA([1430.00, 5000.00],
                    #[ 2.36691815E+01,  1.92560232E-02, -6.07856844E-06,
                    #8.92371727E-10, -4.97448691E-14, -3.96296162E+04,
                    #-9.51544015E+01])),
        #note='9/14')

#species(name='C5H9O2-3OOH-5',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1436.00],
                    #[-2.95211763E+00,  8.90845071E-02, -7.32339295E-05,
                    #2.98425405E-08, -4.76173344E-12, -3.00280666E+04,
                    #4.45334764E+01]),
                #NASA([1436.00, 5000.00],
                    #[ 2.37309690E+01,  2.06751517E-02, -6.88444402E-06,
                    #1.05044759E-09, -6.02362275E-14, -3.84362991E+04,
                    #-9.61945548E+01])),
        #note='9/14')

#species(name='C5H9O2-4OOH-1',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1429.00],
                    #[-5.54951868E+00,  9.55182117E-02, -8.01619637E-05,
                    #3.35588175E-08, -5.52327105E-12, -3.03610773E+04,
                    #5.64175360E+01]),
                #NASA([1429.00, 5000.00],
                    #[ 2.26534501E+01,  2.20609304E-02, -7.36975144E-06,
                    #1.12674313E-09, -6.46941132E-14, -3.91616134E+04,
                    #-9.19653554E+01])),
        #note='9/14')

#species(name='C5H9O2-4OOH-3',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1427.00],
                    #[-4.69459258E+00,  9.49457030E-02, -7.98763001E-05,
                    #3.34048886E-08, -5.48568419E-12, -3.26159152E+04,
                    #4.98789194E+01]),
                #NASA([1427.00, 5000.00],
                    #[ 2.36121225E+01,  2.13802117E-02, -7.16433060E-06,
                    #1.09773009E-09, -6.31280098E-14, -4.14634634E+04,
                    #-9.91050535E+01])),
        #note='9/14')

#species(name='OCH2CHO',
        #atoms='C:2 H:3 O:2',
        #thermo=(NASA([300.00, 1468.00],
                    #[ 6.12784775E+00,  6.67998271E-03,  1.00125301E-05,
                    #-1.06887671E-08,  2.76652784E-12, -1.28216539E+04,
                    #-8.09763223E-01]),
                #NASA([1468.00, 5000.00],
                    #[ 9.83472746E+00,  7.77895515E-03, -2.78174926E-06,
                    #4.44911597E-10, -2.63530280E-14, -1.50740413E+04,
                    #-2.41644767E+01])),
        #note='9/8/14')

#species(name='C5H9O1-2O-5',
        #atoms='C:5 H:9 O:2',
        #thermo=(NASA([300.00, 1421.00],
                    #[-6.46243908E-01,  6.53341554E-02, -4.13295267E-05,
                    #1.15534981E-08, -9.90013428E-13, -1.01037208E+04,
                    #3.35841975E+01]),
                #NASA([1421.00, 5000.00],
                    #[ 2.01220485E+01,  2.03728898E-02, -6.96642194E-06,
                    #1.08201705E-09, -6.28152071E-14, -1.74886279E+04,
                    #-7.89086914E+01])),
        #note='9/8/14')

#species(name='CH2OCH2CHO',
        #atoms='C:3 H:5 O:2',
        #thermo=(NASA([300.00, 1386.00],
                    #[ 4.09796526E+00,  2.50971166E-02, -1.04380487E-05,
                    #6.95156168E-10,  3.93228907E-13, -1.51203149E+04,
                    #1.16942054E+01]),
                #NASA([1386.00, 5000.00],
                    #[ 1.11192754E+01,  1.27912900E-02, -4.35464414E-06,
                    #6.73922275E-10, -3.90118722E-14, -1.79694096E+04,
                    #-2.74688068E+01])),
        #note='9/8/14')

#species(name='C2H4OCHO',
        #atoms='C:3 H:5 O:2',
        #thermo=(NASA([300.00, 1684.00],
                    #[ 3.11985293E+00,  3.19514120E-02, -1.67508873E-05,
                    #3.25914973E-09, -4.96818754E-14, -1.76239471E+04,
                    #1.42845094E+01]),
                #NASA([1684.00, 5000.00],
                    #[ 1.22682078E+01,  1.37336751E-02, -5.06732537E-06,
                    #8.30265992E-10, -5.00728023E-14, -2.09125606E+04,
                    #-3.56985038E+01])),
        #note='9/8/14')

#species(name='CH2CH2OCH2CH2CHO',
        #atoms='C:5 H:9 O:2',
        #thermo=(NASA([300.00, 1447.00],
                    #[-6.18508829E-01,  6.43128028E-02, -4.60370456E-05,
                    #1.65228932E-08, -2.31382679E-12, -2.02103439E+04,
                    #3.75876155E+01]),
                #NASA([1447.00, 5000.00],
                    #[ 1.75698412E+01,  2.05459822E-02, -6.77593971E-06,
                    #1.02644892E-09, -5.85405221E-14, -2.62366696E+04,
                    #-5.93672925E+01])),
        #note='4')

#species(name='C4H7O1-4',
        #atoms='C:4 H:7 O:1',
        #thermo=(NASA([300.00, 1388.00],
                    #[ 2.44895430E+00,  3.69423519E-02, -1.77510852E-05,
                    #2.64537530E-09,  2.43666218E-13,  6.16385921E+03,
                    #1.73174889E+01]),
                #NASA([1388.00, 5000.00],
                    #[ 1.33251126E+01,  1.65057558E-02, -5.65235890E-06,
                    #8.78319847E-10, -5.09914580E-14,  1.91281414E+03,
                    #-4.28181479E+01])),
        #note='9/8/14')

#species(name='NC5KET12',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1386.00],
                    #[-1.29174799E+00,  8.64138926E-02, -7.02456071E-05,
                    #2.86657248E-08, -4.68321816E-12, -4.01718501E+04,
                    #4.00248297E+01]),
                #NASA([1386.00, 5000.00],
                    #[ 2.56958620E+01,  1.97581564E-02, -6.83008913E-06,
                    #1.06937987E-09, -6.24522080E-14, -4.91027960E+04,
                    #-1.03505108E+02])),
        #note='9/8/14')

#species(name='NC5KET13',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1405.00],
                    #[ 3.00527399E+00,  6.57859047E-02, -4.36348536E-05,
                    #1.39803001E-08, -1.68002311E-12, -4.24094861E+04,
                    #1.90078452E+01]),
                #NASA([1405.00, 5000.00],
                    #[ 2.25901874E+01,  2.17812117E-02, -7.38063794E-06,
                    #1.13910638E-09, -6.58295642E-14, -4.92679139E+04,
                    #-8.65971472E+01])),
        #note='9/8/14')

#species(name='NC5KET14',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1405.00],
                    #[ 3.00527399E+00,  6.57859047E-02, -4.36348536E-05,
                    #1.39803001E-08, -1.68002311E-12, -4.24094861E+04,
                    #1.90078452E+01]),
                #NASA([1405.00, 5000.00],
                    #[ 2.25901874E+01,  2.17812117E-02, -7.38063794E-06,
                    #1.13910638E-09, -6.58295642E-14, -4.92679139E+04,
                    #-8.65971472E+01])),
        #note='9/8/14')

#species(name='NC5KET15',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1389.00],
                    #[ 2.63079123E+00,  6.36042087E-02, -3.80050385E-05,
                    #1.00772174E-08, -8.15255520E-13, -4.02201480E+04,
                    #2.27068767E+01]),
                #NASA([1389.00, 5000.00],
                    #[ 2.22256536E+01,  2.26347088E-02, -7.79347724E-06,
                    #1.21574025E-09, -7.07808477E-14, -4.74266046E+04,
                    #-8.41025053E+01])),
        #note='9/8/14')

#species(name='NC5KET21',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1388.00],
                    #[-1.97528990E-01,  7.92952922E-02, -6.01376298E-05,
                    #2.26192730E-08, -3.36572580E-12, -4.18751690E+04,
                    #3.59471835E+01]),
                #NASA([1388.00, 5000.00],
                    #[ 2.45822715E+01,  2.02064434E-02, -6.87782149E-06,
                    #1.06593008E-09, -6.18171916E-14, -5.02288346E+04,
                    #-9.64925259E+01])),
        #note='9/8/14')

#species(name='NC5KET23',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1390.00],
                    #[-4.59062579E-01,  8.23364070E-02, -6.69200221E-05,
                    #2.76548457E-08, -4.58692734E-12, -4.40257491E+04,
                    #3.61810105E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 2.42608382E+01,  2.03281864E-02, -6.88576173E-06,
                    #1.06357099E-09, -6.15324796E-14, -5.21063936E+04,
                    #-9.49416999E+01])),
        #note='9/8/14')

#species(name='NC5KET24',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1417.00],
                    #[ 3.70377687E+00,  6.22140345E-02, -4.08903819E-05,
                    #1.31907442E-08, -1.60742182E-12, -4.62411438E+04,
                    #1.58017148E+01]),
                #NASA([1417.00, 5000.00],
                    #[ 2.13557506E+01,  2.20059280E-02, -7.28286282E-06,
                    #1.10610540E-09, -6.32057101E-14, -5.23446138E+04,
                    #-7.91468873E+01])),
        #note='9/8/14')

#species(name='NC5KET25',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1395.00],
                    #[ 3.30376112E+00,  6.01763378E-02, -3.55337358E-05,
                    #9.49960340E-09, -7.96339342E-13, -4.40485734E+04,
                    #1.96186974E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 2.09051806E+01,  2.29956432E-02, -7.75343394E-06,
                    #1.19262969E-09, -6.87612656E-14, -5.04678003E+04,
                    #-7.61623703E+01])),
        #note='9/8/14')

#species(name='NC5KET31',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1431.00],
                    #[ 4.06393861E+00,  5.46274576E-02, -2.50666165E-05,
                    #2.02312915E-09,  1.00150274E-12, -4.41625000E+04,
                    #1.59654525E+01]),
                #NASA([1431.00, 5000.00],
                    #[ 2.14962676E+01,  2.20658980E-02, -7.35490387E-06,
                    #1.12358588E-09, -6.45028163E-14, -5.08471346E+04,
                    #-8.02429953E+01])),
        #note='9/8/14')

#species(name='NC5KET32',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1388.00],
                    #[ 3.80342038E-01,  7.67072849E-02, -5.67635185E-05,
                    #2.07068449E-08, -2.96497951E-12, -4.41579478E+04,
                    #3.21097700E+01]),
                #NASA([1388.00, 5000.00],
                    #[ 2.43326367E+01,  2.02926765E-02, -6.88030752E-06,
                    #1.06352004E-09, -6.15648338E-14, -5.22889844E+04,
                    #-9.61325613E+01])),
        #note='9/8/14')

#species(name='CHOC4H8O',
        #atoms='C:5 H:9 O:2',
        #thermo=(NASA([300.00, 1377.00],
                    #[ 2.82328315E+00,  5.56013887E-02, -3.02956498E-05,
                    #6.22735484E-09, -8.25379616E-14, -2.35055878E+04,
                    #1.92120018E+01]),
                #NASA([1377.00, 5000.00],
                    #[ 2.09367959E+01,  2.00874425E-02, -7.04829670E-06,
                    #1.11330293E-09, -6.53771464E-14, -3.04314417E+04,
                    #-8.04081881E+01])),
        #note='9/8/14')

#species(name='CHOCH2CH2C2H4O',
        #atoms='C:5 H:9 O:2',
        #thermo=(NASA([300.00, 1323.00],
                    #[ 3.92095399E+00,  4.94038774E-02, -1.99599619E-05,
                    #-1.05666833E-09,  1.70331197E-12, -2.43649054E+04,
                    #1.33502699E+01]),
                #NASA([1323.00, 5000.00],
                    #[ 2.11225120E+01,  1.90830240E-02, -6.52403075E-06,
                    #1.01368310E-09, -5.88793478E-14, -3.11151636E+04,
                    #-8.21812017E+01])),
        #note='914')

#species(name='CHOCH2CH2CH2CH2O',
        #atoms='C:5 H:9 O:2',
        #thermo=(NASA([300.00, 1489.00],
                    #[ 3.41522476E+00,  4.72023685E-02, -1.66453220E-05,
                    #-2.39548810E-09,  1.83814508E-12, -2.21147912E+04,
                    #1.77035103E+01]),
                #NASA([1489.00, 5000.00],
                    #[ 1.92273388E+01,  2.11348590E-02, -7.32088660E-06,
                    #1.14667396E-09, -6.69507752E-14, -2.85866827E+04,
                    #-7.08964819E+01])),
        #note='4')

#species(name='C3H7COCH2O',
        #atoms='C:5 H:9 O:2',
        #thermo=(NASA([300.00, 1669.00],
                    #[ 7.57725086E+00,  3.09672386E-02,  2.94905694E-06,
                    #-1.10969895E-08,  3.01674992E-12, -2.56340563E+04,
                    #-2.16720088E+00]),
                #NASA([1669.00, 5000.00],
                    #[ 1.55374345E+01,  2.65967153E-02, -9.82520086E-06,
                    #1.61087197E-09, -9.71864472E-14, -2.96220202E+04,
                    #-4.97983106E+01])),
        #note='9/8/14')

#species(name='CH3COC3H6O',
        #atoms='C:5 H:9 O:2',
        #thermo=(NASA([300.00, 1383.00],
                    #[ 3.56460361E+00,  5.18213432E-02, -2.70330790E-05,
                    #5.19426451E-09,  1.69499746E-14, -2.73448675E+04,
                    #1.58030522E+01]),
                #NASA([1383.00, 5000.00],
                    #[ 1.95245872E+01,  2.08858647E-02, -7.23257140E-06,
                    #1.13238277E-09, -6.60902126E-14, -3.34926354E+04,
                    #-7.21145428E+01])),
        #note='9/8/14')

#species(name='CH3COCH2CH2CH2O',
        #atoms='C:5 H:9 O:2',
        #thermo=(NASA([300.00, 1433.00],
                    #[ 4.40785464E+00,  4.34772079E-02, -1.40284695E-05,
                    #-3.02310104E-09,  1.86768676E-12, -2.60265017E+04,
                    #1.28766258E+01]),
                #NASA([1433.00, 5000.00],
                    #[ 1.81303967E+01,  2.13611121E-02, -7.24506108E-06,
                    #1.11897303E-09, -6.47006005E-14, -3.16682816E+04,
                    #-6.41505194E+01])),
        #note='14')

#species(name='C2H5COC2H4O',
        #atoms='C:5 H:9 O:2',
        #thermo=(NASA([300.00, 1539.00],
                    #[ 4.55782533E+00,  4.55546283E-02, -1.73485031E-05,
                    #-9.24152417E-10,  1.38889438E-12, -2.74919397E+04,
                    #1.10596626E+01]),
                #NASA([1539.00, 5000.00],
                    #[ 1.90206162E+01,  2.10022758E-02, -7.20628551E-06,
                    #1.12155484E-09, -6.51936687E-14, -3.33553472E+04,
                    #-6.97524186E+01])),
        #note='9/8/14')

#species(name='CH3COCH2C2H4O',
        #atoms='C:5 H:9 O:2',
        #thermo=(NASA([300.00, 1382.00],
                    #[ 4.40946289E+00,  4.75586902E-02, -2.19238261E-05,
                    #2.43250490E-09,  5.88504938E-13, -2.76535939E+04,
                    #1.17394140E+01]),
                #NASA([1382.00, 5000.00],
                    #[ 1.87900628E+01,  2.10010240E-02, -7.16188369E-06,
                    #1.11001257E-09, -6.43339326E-14, -3.32898658E+04,
                    #-6.78862933E+01])),
        #note='9/14')

#species(name='CHOCH2C3H6O',
        #atoms='C:5 H:9 O:2',
        #thermo=(NASA([300.00, 1323.00],
                    #[ 3.92095399E+00,  4.94038774E-02, -1.99599619E-05,
                    #-1.05666833E-09,  1.70331197E-12, -2.43649054E+04,
                    #1.33502699E+01]),
                #NASA([1323.00, 5000.00],
                    #[ 2.11225120E+01,  1.90830240E-02, -6.52403075E-06,
                    #1.01368310E-09, -5.88793478E-14, -3.11151636E+04,
                    #-8.21812017E+01])),
        #note='9/8/14')

#species(name='C2H5COCH2CH2O',
        #atoms='C:5 H:9 O:2',
        #thermo=(NASA([300.00, 1426.00],
                    #[ 5.37524822E+00,  3.49659305E-02,  3.67920037E-06,
                    #-1.56750631E-08,  4.84743343E-12, -2.61248405E+04,
                    #8.64755527E+00]),
                #NASA([1426.00, 5000.00],
                    #[ 1.92416140E+01,  2.07624762E-02, -7.12773261E-06,
                    #1.11092657E-09, -6.46775655E-14, -3.25710561E+04,
                    #-7.19638232E+01])),
        #note='9/14')

#species(name='C5H10OH-1',
        #atoms='C:5 H:11 O:1',
        #thermo=(NASA([300.00, 1402.00],
                    #[ 5.41167710E-01,  5.66862244E-02, -2.94678472E-05,
                    #5.84904634E-09, -3.31866473E-14, -1.50943752E+04,
                    #3.18827466E+01]),
                #NASA([1402.00, 5000.00],
                    #[ 1.68143234E+01,  2.44876194E-02, -8.29372543E-06,
                    #1.27921101E-09, -7.38804613E-14, -2.12567818E+04,
                    #-5.74573295E+01])),
        #note='9/8/14')

#species(name='C5H10OH-2',
        #atoms='C:5 H:11 O:1',
        #thermo=(NASA([300.00, 1447.00],
                    #[ 9.23102001E-01,  5.84732430E-02, -3.47946675E-05,
                    #9.61256846E-09, -8.64039220E-13, -1.72758426E+04,
                    #2.84227761E+01]),
                #NASA([1447.00, 5000.00],
                    #[ 1.70150141E+01,  2.35005444E-02, -7.77847758E-06,
                    #1.18102788E-09, -6.74584591E-14, -2.30013982E+04,
                    #-5.87178603E+01])),
        #note='9/8/14')

#species(name='O2C5H10OH-1',
        #atoms='C:5 H:11 O:3',
        #thermo=(NASA([300.00, 1406.00],
                    #[ 1.54218231E+00,  7.57844091E-02, -5.63028041E-05,
                    #2.16374600E-08, -3.37143496E-12, -3.38651232E+04,
                    #2.73478003E+01]),
                #NASA([1406.00, 5000.00],
                    #[ 2.28162462E+01,  2.40750309E-02, -8.15387271E-06,
                    #1.25759554E-09, -7.26292545E-14, -4.09893865E+04,
                    #-8.60865796E+01])),
        #note='9/8/14')

#species(name='O2C5H10OH-2',
        #atoms='C:5 H:11 O:3',
        #thermo=(NASA([300.00, 1413.00],
                    #[ 1.48292990E+00,  7.92100619E-02, -6.45238925E-05,
                    #2.75597603E-08, -4.74651399E-12, -3.59663215E+04,
                    #2.60236189E+01]),
                #NASA([1413.00, 5000.00],
                    #[ 2.25663309E+01,  2.36960784E-02, -7.89320309E-06,
                    #1.20343029E-09, -6.89316529E-14, -4.25964462E+04,
                    #-8.48729630E+01])),
        #note='9/8/14')

#species(name='IC5H12',
        #atoms='C:5 H:12',
        #thermo=(NASA([300.00, 1397.00],
                    #[-1.38253385E+00,  6.53645895E-02, -4.34522951E-05,
                    #1.52986287E-08, -2.25248819E-12, -2.08449123E+04,
                    #3.17292839E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 1.59790958E+01,  2.57511190E-02, -8.75373554E-06,
                    #1.35318160E-09, -7.82642580E-14, -2.70131035E+04,
                    #-6.19207317E+01])))

#species(name='AC5H11',
        #atoms='C:5 H:11',
        #thermo=(NASA([300.00, 1397.00],
                    #[-1.05353982E+00,  6.19729558E-02, -4.24710792E-05,
                    #1.55773628E-08, -2.40104166E-12,  4.23634137E+03,
                    #3.31805856E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 1.53440125E+01,  2.38465896E-02, -8.11897135E-06,
                    #1.25634599E-09, -7.27146606E-14, -1.52767896E+03,
                    #-5.50323375E+01])),
        #note='9/8/14')

#species(name='BC5H11',
        #atoms='C:5 H:11',
        #thermo=(NASA([300.00, 1385.00],
                    #[ 7.44237180E-01,  4.70591752E-02, -1.83107291E-05,
                    #5.03910927E-10,  8.85776586E-13,  1.30430752E+03,
                    #2.59689042E+01]),
                #NASA([1385.00, 5000.00],
                    #[ 1.36183345E+01,  2.53158683E-02, -8.62827133E-06,
                    #1.33623402E-09, -7.73862050E-14, -4.03058789E+03,
                    #-4.61784698E+01])),
        #note='9/8/14')

#species(name='CC5H11',
        #atoms='C:5 H:11',
        #thermo=(NASA([300.00, 1402.00],
                    #[-2.37043597E-01,  5.49326482E-02, -3.05921538E-05,
                    #7.74913666E-09, -6.05124619E-13,  2.66541988E+03,
                    #2.99488121E+01]),
                #NASA([1402.00, 5000.00],
                    #[ 1.47105935E+01,  2.40981788E-02, -8.14500318E-06,
                    #1.25428598E-09, -7.23532844E-14, -2.91057300E+03,
                    #-5.17362733E+01])),
        #note='9/8/14')

#species(name='DC5H11',
        #atoms='C:5 H:11',
        #thermo=(NASA([300.00, 1397.00],
                    #[-1.05353982E+00,  6.19729558E-02, -4.24710792E-05,
                    #1.55773628E-08, -2.40104166E-12,  4.23634137E+03,
                    #3.24911039E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 1.53440125E+01,  2.38465896E-02, -8.11897135E-06,
                    #1.25634599E-09, -7.27146606E-14, -1.52767896E+03,
                    #-5.57218191E+01])),
        #note='9/8/14')

#species(name='AC5H10',
        #atoms='C:5 H:10',
        #thermo=(NASA([300.00, 1392.00],
                    #[-5.39429136E-01,  5.44489715E-02, -3.32707895E-05,
                    #1.03047694E-08, -1.28363329E-12, -6.53967251E+03,
                    #2.90349986E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 1.41931279E+01,  2.26551019E-02, -7.70008627E-06,
                    #1.19031326E-09, -6.88489604E-14, -1.19491010E+04,
                    #-5.10688681E+01])))

#species(name='BC5H10',
        #atoms='C:5 H:10',
        #thermo=(NASA([300.00, 1389.00],
                    #[ 6.04882839E-01,  4.96635256E-02, -2.66571142E-05,
                    #6.47312078E-09, -4.84017883E-13, -8.06312207E+03,
                    #2.23818263E+01]),
                #NASA([1389.00, 5000.00],
                    #[ 1.40426423E+01,  2.27915348E-02, -7.74902598E-06,
                    #1.19814522E-09, -6.93127003E-14, -1.32160483E+04,
                    #-5.14469569E+01])),
        #note='9/8/14')

#species(name='CC5H10',
        #atoms='C:5 H:10',
        #thermo=(NASA([300.00, 1394.00],
                    #[-9.64397004E-01,  5.68482691E-02, -3.66324275E-05,
                    #1.23008921E-08, -1.71392531E-12, -5.93502026E+03,
                    #3.02998433E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 1.44242929E+01,  2.26454984E-02, -7.73646334E-06,
                    #1.19999861E-09, -6.95711276E-14, -1.14981027E+04,
                    #-5.30391197E+01])),
        #note='9/8/14')

#species(name='AC5H9-A2',
        #atoms='C:5 H:9',
        #thermo=(NASA([300.00, 1385.00],
                    #[-9.54047710E-01,  5.50153471E-02, -3.58307910E-05,
                    #1.16444141E-08, -1.49086833E-12,  1.15959514E+04,
                    #3.08476806E+01]),
                #NASA([1385.00, 5000.00],
                    #[ 1.50019889E+01,  1.95965428E-02, -6.60205927E-06,
                    #1.01535593E-09, -5.85454957E-14,  5.90081014E+03,
                    #-5.54528992E+01])))

#species(name='AC5H9-C',
        #atoms='C:5 H:9',
        #thermo=(NASA([300.00, 1392.00],
                    #[-8.41372657E-01,  5.27157283E-02, -3.27334707E-05,
                    #1.01974035E-08, -1.26084730E-12,  1.04321036E+04,
                    #2.93316754E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 1.37918110E+01,  2.09644966E-02, -7.13793293E-06,
                    #1.10480881E-09, -6.39628018E-14,  5.09555600E+03,
                    #-5.01391436E+01])))

#species(name='AC5H9-D',
        #atoms='C:5 H:9',
        #thermo=(NASA([300.00, 1393.00],
                    #[-1.67398152E-01,  5.09876974E-02, -3.22615638E-05,
                    #1.05911843E-08, -1.43705463E-12,  1.81792947E+04,
                    #2.95710715E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 1.35607521E+01,  2.07514878E-02, -7.06608626E-06,
                    #1.09362345E-09, -6.33083050E-14,  1.31896844E+04,
                    #-4.48718988E+01])))

#species(name='CC5H9-A',
        #atoms='C:5 H:9',
        #thermo=(NASA([300.00, 1393.00],
                    #[-6.78909715E-01,  5.37496002E-02, -3.61116825E-05,
                    #1.28563411E-08, -1.92155723E-12,  1.87974429E+04,
                    #3.12403382E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 1.38635892E+01,  2.06362872E-02, -7.05717373E-06,
                    #1.09539676E-09, -6.35381849E-14,  1.36104251E+04,
                    #-4.72507189E+01])))

#species(name='CC5H9-B',
        #atoms='C:5 H:9',
        #thermo=(NASA([300.00, 1392.00],
                    #[-6.91501860E-01,  5.09804846E-02, -2.96448293E-05,
                    #8.28273022E-09, -8.57575066E-13,  1.07478467E+04,
                    #2.76711552E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 1.35249511E+01,  2.14363755E-02, -7.35304116E-06,
                    #1.14368637E-09, -6.64363518E-14,  5.41381916E+03,
                    #-5.00302394E+01])),
        #note='9/8/14')

#species(name='AC5H9O-A2',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1364.00],
                    #[ 4.66263417E+00,  3.89775126E-02, -1.12815501E-05,
                    #-2.66550677E-09,  1.40115432E-12,  1.47086581E+03,
                    #8.21346271E+00]),
                #NASA([1364.00, 5000.00],
                    #[ 1.67516787E+01,  2.16463662E-02, -7.61269705E-06,
                    #1.20404408E-09, -7.07632803E-14, -3.95848360E+03,
                    #-6.08201916E+01])),
        #note='9/8/14')

#species(name='AC5H9O-C',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1382.00],
                    #[ 2.04951318E+00,  5.36584877E-02, -3.35086710E-05,
                    #1.06174235E-08, -1.39224338E-12, -3.85112518E+02,
                    #1.91077262E+01]),
                #NASA([1382.00, 5000.00],
                    #[ 1.76789601E+01,  2.05915073E-02, -7.18764438E-06,
                    #1.13116483E-09, -6.62505034E-14, -6.25146818E+03,
                    #-6.62131721E+01])))

#species(name='CC5H9O-B',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1377.00],
                    #[ 2.75614808E+00,  5.20514036E-02, -3.10966938E-05,
                    #8.25895065E-09, -6.59917121E-13, -1.35162741E+03,
                    #1.40990119E+01]),
                #NASA([1377.00, 5000.00],
                    #[ 1.86974377E+01,  1.84542537E-02, -6.18682537E-06,
                    #9.48850967E-10, -5.46203980E-14, -7.15897769E+03,
                    #-7.26540429E+01])),
        #note='9/8/14')

#species(name='AC5H11O2H',
        #atoms='C:5 H:12 O:2',
        #thermo=(NASA([300.00, 1397.00],
                    #[-3.83097656E-01,  7.75293305E-02, -5.71652441E-05,
                    #2.22670177E-08, -3.58721009E-12, -3.05891551E+04,
                    #3.29686174E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 2.10463833E+01,  2.59225280E-02, -8.86503936E-06,
                    #1.37611919E-09, -7.98291025E-14, -3.79034414E+04,
                    #-8.16163782E+01])),
        #note='9/8/14')

#species(name='BC5H11O2H',
        #atoms='C:5 H:12 O:2',
        #thermo=(NASA([300.00, 1389.00],
                    #[-1.96398881E-01,  8.00539195E-02, -6.06245513E-05,
                    #2.38763967E-08, -3.84750148E-12, -3.34622874E+04,
                    #2.95005821E+01]),
                #NASA([1389.00, 5000.00],
                    #[ 2.29240164E+01,  2.40774597E-02, -8.18457843E-06,
                    #1.26619943E-09, -7.33059813E-14, -4.12839493E+04,
                    #-9.39570707E+01])),
        #note='9/8/14')

#species(name='CC5H11O2H',
        #atoms='C:5 H:12 O:2',
        #thermo=(NASA([300.00, 1406.00],
                    #[ 4.28477267E-02,  7.95135147E-02, -6.24988771E-05,
                    #2.59693876E-08, -4.40242741E-12, -3.27873015E+04,
                    #2.83342901E+01]),
                #NASA([1406.00, 5000.00],
                    #[ 2.13766018E+01,  2.51045757E-02, -8.46450799E-06,
                    #1.30130796E-09, -7.49768283E-14, -3.97208337E+04,
                    #-8.45864663E+01])),
        #note='9/8/14')

#species(name='DC5H11O2H',
        #atoms='C:5 H:12 O:2',
        #thermo=(NASA([300.00, 1397.00],
                    #[-3.83097656E-01,  7.75293305E-02, -5.71652441E-05,
                    #2.22670177E-08, -3.58721009E-12, -3.05891551E+04,
                    #3.22791357E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 2.10463833E+01,  2.59225280E-02, -8.86503936E-06,
                    #1.37611919E-09, -7.98291025E-14, -3.79034414E+04,
                    #-8.23058599E+01])),
        #note='9/8/14')

#species(name='AC5H11O2',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 4.33973910E-01,  7.13982313E-02, -5.16583407E-05,
                    #1.97686122E-08, -3.13751422E-12, -1.39106770E+04,
                    #2.96761102E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 2.00710052E+01,  2.46047087E-02, -8.40507957E-06,
                    #1.30373672E-09, -7.55899526E-14, -2.06670757E+04,
                    #-7.55077454E+01])),
        #note='9/8/14')

#species(name='BC5H11O2',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1386.00],
                    #[ 5.99419477E-01,  7.39607301E-02, -5.50443925E-05,
                    #2.12526793E-08, -3.35787966E-12, -1.67795760E+04,
                    #2.63156366E+01]),
                #NASA([1386.00, 5000.00],
                    #[ 2.21004974E+01,  2.25392426E-02, -7.63272744E-06,
                    #1.17814793E-09, -6.81106788E-14, -2.41130117E+04,
                    #-8.87144463E+01])),
        #note='9/8/14')

#species(name='CC5H11O2',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1405.00],
                    #[ 7.68208766E-01,  7.37734356E-02, -5.75364545E-05,
                    #2.37895752E-08, -4.02044039E-12, -1.60946468E+04,
                    #2.54692494E+01]),
                #NASA([1405.00, 5000.00],
                    #[ 2.04381948E+01,  2.37428212E-02, -7.98714323E-06,
                    #1.22601475E-09, -7.05615887E-14, -2.25023727E+04,
                    #-7.86954028E+01])),
        #note='9/8/14')

#species(name='DC5H11O2',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 4.33973910E-01,  7.13982313E-02, -5.16583407E-05,
                    #1.97686122E-08, -3.13751422E-12, -1.39106770E+04,
                    #2.89866286E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 2.00710052E+01,  2.46047087E-02, -8.40507957E-06,
                    #1.30373672E-09, -7.55899526E-14, -2.06670757E+04,
                    #-7.61972270E+01])),
        #note='9/8/14')

#species(name='AC5H11O',
        #atoms='C:5 H:11 O:1',
        #thermo=(NASA([300.00, 1400.00],
                    #[ 6.12135299E-01,  6.37939332E-02, -4.13544665E-05,
                    #1.33882853E-08, -1.70346830E-12, -1.26228778E+04,
                    #2.61856024E+01]),
                #NASA([1400.00, 5000.00],
                    #[ 1.86713806E+01,  2.35577702E-02, -8.02035115E-06,
                    #1.24132611E-09, -7.18636960E-14, -1.90530781E+04,
                    #-7.14330583E+01])),
        #note='9/8/14')

#species(name='BC5H11O',
        #atoms='C:5 H:11 O:1',
        #thermo=(NASA([300.00, 1378.00],
                    #[ 3.67366095E+00,  5.17139422E-02, -2.49457898E-05,
                    #4.54084943E-09, -3.31184982E-14, -1.63731271E+04,
                    #9.40461313E+00]),
                #NASA([1378.00, 5000.00],
                    #[ 1.85549393E+01,  2.44786662E-02, -8.51926880E-06,
                    #1.33803108E-09, -7.82547188E-14, -2.24167627E+04,
                    #-7.33946136E+01])),
        #note='9/8/14')

#species(name='CC5H11O',
        #atoms='C:5 H:11 O:1',
        #thermo=(NASA([300.00, 1409.00],
                    #[ 6.11304005E-01,  6.60484477E-02, -4.65814257E-05,
                    #1.70410121E-08, -2.53050937E-12, -1.47116675E+04,
                    #2.40791759E+01]),
                #NASA([1409.00, 5000.00],
                    #[ 1.86565428E+01,  2.31561973E-02, -7.79088683E-06,
                    #1.19608628E-09, -6.88497016E-14, -2.08514389E+04,
                    #-7.24815819E+01])),
        #note='9/8/14')

#species(name='DC5H11O',
        #atoms='C:5 H:11 O:1',
        #thermo=(NASA([300.00, 1400.00],
                    #[ 6.12135299E-01,  6.37939332E-02, -4.13544665E-05,
                    #1.33882853E-08, -1.70346830E-12, -1.26228778E+04,
                    #2.54961208E+01]),
                #NASA([1400.00, 5000.00],
                    #[ 1.86713806E+01,  2.35577702E-02, -8.02035115E-06,
                    #1.24132611E-09, -7.18636960E-14, -1.90530781E+04,
                    #-7.21225400E+01])),
        #note='9/8/14')

#species(name='AC5H10OOH-A',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1398.00],
                    #[-1.70898042E-03,  7.39903239E-02, -5.59517601E-05,
                    #2.23367236E-08, -3.67395948E-12, -5.87093071E+03,
                    #3.34672541E+01]),
                #NASA([1398.00, 5000.00],
                    #[ 2.04730841E+01,  2.38768580E-02, -8.16265188E-06,
                    #1.26680916E-09, -7.34771763E-14, -1.27728563E+04,
                    #-7.57172349E+01])),
        #note='9/8/14')

#species(name='AC5H10OOH-B',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1390.00],
                    #[ 3.72381154E+00,  5.63566061E-02, -3.12255592E-05,
                    #8.06035571E-09, -7.14647749E-13, -9.18272496E+03,
                    #1.71586685E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 1.92392806E+01,  2.49670792E-02, -8.54731996E-06,
                    #1.32768701E-09, -7.70536860E-14, -1.51054781E+04,
                    #-6.79724766E+01])),
        #note='9/8/14')

#species(name='AC5H10OOH-C',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1402.00],
                    #[ 1.28401222E+00,  6.48936005E-02, -4.13324696E-05,
                    #1.29875776E-08, -1.57178209E-12, -7.51170024E+03,
                    #2.87556593E+01]),
                #NASA([1402.00, 5000.00],
                    #[ 1.96523154E+01,  2.43128018E-02, -8.25709081E-06,
                    #1.27586796E-09, -7.37791680E-14, -1.40744871E+04,
                    #-7.06356333E+01])),
        #note='9/8/14')

#species(name='AC5H10OOH-D',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1398.00],
                    #[-1.70898042E-03,  7.39903239E-02, -5.59517601E-05,
                    #2.23367236E-08, -3.67395948E-12, -5.87093071E+03,
                    #3.34672541E+01]),
                #NASA([1398.00, 5000.00],
                    #[ 2.04730841E+01,  2.38768580E-02, -8.16265188E-06,
                    #1.26680916E-09, -7.34771763E-14, -1.27728563E+04,
                    #-7.57172349E+01])),
        #note='9/8/14')

#species(name='BC5H10OOH-A',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1389.00],
                    #[ 7.27230216E-01,  7.50438273E-02, -5.76165408E-05,
                    #2.29759736E-08, -3.74255781E-12, -8.87953820E+03,
                    #2.80769882E+01]),
                #NASA([1389.00, 5000.00],
                    #[ 2.25053262E+01,  2.19567103E-02, -7.46854531E-06,
                    #1.15601991E-09, -6.69539257E-14, -1.62100144E+04,
                    #-8.80836630E+01])),
        #note='9/8/14')

#species(name='BC5H10OOH-C',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1390.00],
                    #[ 1.70518427E+00,  7.00032147E-02, -5.03007757E-05,
                    #1.86407873E-08, -2.82557457E-12, -1.03372446E+04,
                    #2.27931660E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 2.21637663E+01,  2.22104861E-02, -7.54936430E-06,
                    #1.16803395E-09, -6.76321121E-14, -1.74314866E+04,
                    #-8.70670513E+01])),
        #note='9/8/14')

#species(name='BC5H10OOH-D',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1389.00],
                    #[ 1.74814357E-01,  7.65426234E-02, -5.94208599E-05,
                    #2.39377588E-08, -3.93095818E-12, -8.74215935E+03,
                    #3.00494349E+01]),
                #NASA([1389.00, 5000.00],
                    #[ 2.23966989E+01,  2.19719191E-02, -7.45848073E-06,
                    #1.15298534E-09, -6.67219770E-14, -1.61758674E+04,
                    #-8.83258681E+01])),
        #note='9/8/14')

#species(name='CC5H10OOH-A',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1407.00],
                    #[ 3.79656831E-01,  7.61630881E-02, -6.15527602E-05,
                    #2.61999847E-08, -4.52431537E-12, -8.06210669E+03,
                    #2.97307114E+01]),
                #NASA([1407.00, 5000.00],
                    #[ 2.08189235E+01,  2.30417107E-02, -7.75559619E-06,
                    #1.19093894E-09, -6.85622462E-14, -1.45987511E+04,
                    #-7.80916538E+01])),
        #note='9/8/14')

#species(name='CC5H10OOH-B',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1403.00],
                    #[ 4.01390573E+00,  5.89581793E-02, -3.74602189E-05,
                    #1.23076066E-08, -1.64804436E-12, -1.13608688E+04,
                    #1.31497594E+01]),
                #NASA([1403.00, 5000.00],
                    #[ 1.96099931E+01,  2.41017207E-02, -8.12789220E-06,
                    #1.24968767E-09, -7.20069527E-14, -1.69405994E+04,
                    #-7.11778005E+01])),
        #note='9/8/14')

#species(name='CC5H10OOH-D',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1407.00],
                    #[ 9.62809685E-01,  7.45146309E-02, -5.94943176E-05,
                    #2.50597392E-08, -4.29318374E-12, -8.20391048E+03,
                    #2.62389931E+01]),
                #NASA([1407.00, 5000.00],
                    #[ 2.09661599E+01,  2.29667003E-02, -7.74034026E-06,
                    #1.18962848E-09, -6.85281914E-14, -1.46486669E+04,
                    #-7.94457953E+01])),
        #note='9/8/14')

#species(name='DC5H10OOH-A',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1398.00],
                    #[-1.70898042E-03,  7.39903239E-02, -5.59517601E-05,
                    #2.23367236E-08, -3.67395948E-12, -5.87093071E+03,
                    #3.34672541E+01]),
                #NASA([1398.00, 5000.00],
                    #[ 2.04730841E+01,  2.38768580E-02, -8.16265188E-06,
                    #1.26680916E-09, -7.34771763E-14, -1.27728563E+04,
                    #-7.57172349E+01])),
        #note='9/8/14')

#species(name='DC5H10OOH-B',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1390.00],
                    #[ 1.88576509E+00,  5.86556146E-02, -3.11812158E-05,
                    #6.92384226E-09, -3.20217597E-13, -8.81575060E+03,
                    #2.65402746E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 1.86667193E+01,  2.54911514E-02, -8.73776931E-06,
                    #1.35852736E-09, -7.88977778E-14, -1.52508543E+04,
                    #-6.57304763E+01])),
        #note='9/8/14')

#species(name='DC5H10OOH-C',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1402.00],
                    #[ 1.55689223E+00,  6.72873158E-02, -4.65676311E-05,
                    #1.68717112E-08, -2.53126154E-12, -7.46917998E+03,
                    #2.53991051E+01]),
                #NASA([1402.00, 5000.00],
                    #[ 2.02367841E+01,  2.41185200E-02, -8.25487443E-06,
                    #1.28211115E-09, -7.44043362E-14, -1.40294794E+04,
                    #-7.51327525E+01])),
        #note='9/8/14')

#species(name='A-AC5H10O',
        #atoms='C:5 H:10 O:1',
        #thermo=(NASA([300.00, 1452.00],
                    #[-7.01151849E+00,  8.32843802E-02, -6.78574322E-05,
                    #2.80656722E-08, -4.57849326E-12, -1.67872253E+04,
                    #5.91579158E+01]),
                #NASA([1452.00, 5000.00],
                    #[ 1.61470282E+01,  2.24544126E-02, -7.31234234E-06,
                    #1.09786539E-09, -6.22123297E-14, -2.39328324E+04,
                    #-6.24512542E+01])),
        #note='9/8/14')

#species(name='A-BC5H10O',
        #atoms='C:5 H:10 O:1',
        #thermo=(NASA([300.00, 1424.00],
                    #[-2.81578161E+00,  7.04149950E-02, -5.23593700E-05,
                    #1.99929340E-08, -3.05569594E-12, -1.99582779E+04,
                    #3.88311343E+01]),
                #NASA([1424.00, 5000.00],
                    #[ 1.66888391E+01,  2.23672932E-02, -7.44828561E-06,
                    #1.13557959E-09, -6.50516730E-14, -2.63630388E+04,
                    #-6.48349416E+01])),
        #note='9/8/14')

#species(name='A-CC5H10O',
        #atoms='C:5 H:10 O:1',
        #thermo=(NASA([300.00, 1429.00],
                    #[-8.91216215E+00,  9.53495145E-02, -8.84353223E-05,
                    #4.15290469E-08, -7.61492452E-12, -1.89806474E+04,
                    #6.69247389E+01]),
                #NASA([1429.00, 5000.00],
                    #[ 1.62697447E+01,  2.21770870E-02, -7.17473979E-06,
                    #1.07173572E-09, -6.04939326E-14, -2.61236414E+04,
                    #-6.29431949E+01])),
        #note='9/8/14')

#species(name='A-DC5H10O',
        #atoms='C:5 H:10 O:1',
        #thermo=(NASA([300.00, 1453.00],
                    #[-8.95290911E+00,  8.74962592E-02, -7.11264892E-05,
                    #2.92430785E-08, -4.73462606E-12, -2.68017891E+04,
                    #6.87664513E+01]),
                #NASA([1453.00, 5000.00],
                    #[ 1.56366471E+01,  2.32330072E-02, -7.56609834E-06,
                    #1.13614434E-09, -6.43938663E-14, -3.44157719E+04,
                    #-6.04636117E+01])),
        #note='9/8/14')

#species(name='B-CC5H10O',
        #atoms='C:5 H:10 O:1',
        #thermo=(NASA([300.00, 1413.00],
                    #[-4.02733813E+00,  7.75766076E-02, -6.31068363E-05,
                    #2.66344724E-08, -4.52085229E-12, -2.22133168E+04,
                    #4.23801047E+01]),
                #NASA([1413.00, 5000.00],
                    #[ 1.72822083E+01,  2.21811831E-02, -7.45086135E-06,
                    #1.14266526E-09, -6.57264262E-14, -2.89756909E+04,
                    #-6.99425869E+01])),
        #note='9/8/14')

#species(name='B-DC5H10O',
        #atoms='C:5 H:10 O:1',
        #thermo=(NASA([300.00, 1426.00],
                    #[-4.57207704E+00,  7.33362569E-02, -5.36913946E-05,
                    #2.00360735E-08, -2.97590047E-12, -2.07808158E+04,
                    #4.60627306E+01]),
                #NASA([1426.00, 5000.00],
                    #[ 1.59398262E+01,  2.34035968E-02, -7.80111462E-06,
                    #1.19023227E-09, -6.82188270E-14, -2.75659851E+04,
                    #-6.31518634E+01])),
        #note='9/8/14')

#species(name='C-DC5H10O',
        #atoms='C:5 H:10 O:1',
        #thermo=(NASA([300.00, 1415.00],
                    #[-7.85249202E+00,  9.24572706E-02, -8.26859508E-05,
                    #3.75476189E-08, -6.72954433E-12, -1.79834003E+04,
                    #6.21111673E+01]),
                #NASA([1415.00, 5000.00],
                    #[ 1.78096288E+01,  2.16569592E-02, -7.25509968E-06,
                    #1.11070510E-09, -6.38134365E-14, -2.56961971E+04,
                    #-7.16670670E+01])),
        #note='9/8/14')

#species(name='AC5H10OOH-AO2',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1397.00],
                    #[ 1.56380071E+00,  8.29798833E-02, -6.44672355E-05,
                    #2.61513298E-08, -4.33760267E-12, -2.40270626E+04,
                    #2.96266556E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 2.51246846E+01,  2.47883083E-02, -8.52048861E-06,
                    #1.32728569E-09, -7.71886075E-14, -3.18988778E+04,
                    #-9.57993329E+01])),
        #note='9/14')

#species(name='AC5H10OOH-BO2',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1389.00],
                    #[ 1.60288955E+00,  8.61877924E-02, -6.89692568E-05,
                    #2.84213101E-08, -4.74629543E-12, -2.68785795E+04,
                    #2.75211615E+01]),
                #NASA([1389.00, 5000.00],
                    #[ 2.70053318E+01,  2.29516730E-02, -7.84494464E-06,
                    #1.21831089E-09, -7.07267404E-14, -3.52825198E+04,
                    #-1.07479938E+02])),
        #note='9/14')

#species(name='AC5H10OOH-CO2',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1405.00],
                    #[ 1.88094547E+00,  8.54729606E-02, -7.05748924E-05,
                    #3.03432704E-08, -5.26290667E-12, -2.62094659E+04,
                    #2.61752480E+01]),
                #NASA([1405.00, 5000.00],
                    #[ 2.54672904E+01,  2.39654991E-02, -8.11944280E-06,
                    #1.25250847E-09, -7.23427482E-14, -3.37234155E+04,
                    #-9.81621694E+01])),
        #note='9/14')

#species(name='AC5H10OOH-DO2',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1397.00],
                    #[ 1.56380071E+00,  8.29798833E-02, -6.44672355E-05,
                    #2.61513298E-08, -4.33760267E-12, -2.40270626E+04,
                    #2.96266556E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 2.51246846E+01,  2.47883083E-02, -8.52048861E-06,
                    #1.32728569E-09, -7.71886075E-14, -3.18988778E+04,
                    #-9.57993329E+01])),
        #note='9/14')

#species(name='BC5H10OOH-AO2',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1389.00],
                    #[ 1.60288955E+00,  8.61877924E-02, -6.89692568E-05,
                    #2.84213101E-08, -4.74629543E-12, -2.68785795E+04,
                    #2.75211615E+01]),
                #NASA([1389.00, 5000.00],
                    #[ 2.70053318E+01,  2.29516730E-02, -7.84494464E-06,
                    #1.21831089E-09, -7.07267404E-14, -3.52825198E+04,
                    #-1.07479938E+02])),
        #note='9/14')

#species(name='BC5H10OOH-CO2',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 1.98239726E+00,  8.83401126E-02, -7.45427451E-05,
                    #3.22832372E-08, -5.60168617E-12, -2.90685309E+04,
                    #2.31106182E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 2.74085371E+01,  2.20461555E-02, -7.41197549E-06,
                    #1.13845768E-09, -6.55893154E-14, -3.71427772E+04,
                    #-1.10888518E+02])),
        #note='9/14')

#species(name='BC5H10OOH-DO2',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1389.00],
                    #[ 1.60288955E+00,  8.61877924E-02, -6.89692568E-05,
                    #2.84213101E-08, -4.74629543E-12, -2.68785795E+04,
                    #2.68316798E+01]),
                #NASA([1389.00, 5000.00],
                    #[ 2.70053318E+01,  2.29516730E-02, -7.84494464E-06,
                    #1.21831089E-09, -7.07267404E-14, -3.52825198E+04,
                    #-1.08169419E+02])),
        #note='9/14')

#species(name='CC5H10OOH-AO2',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1405.00],
                    #[ 1.88094547E+00,  8.54729606E-02, -7.05748924E-05,
                    #3.03432704E-08, -5.26290667E-12, -2.62094659E+04,
                    #2.61752480E+01]),
                #NASA([1405.00, 5000.00],
                    #[ 2.54672904E+01,  2.39654991E-02, -8.11944280E-06,
                    #1.25250847E-09, -7.23427482E-14, -3.37234155E+04,
                    #-9.81621694E+01])),
        #note='9/14')

#species(name='CC5H10OOH-BO2',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 1.98239726E+00,  8.83401126E-02, -7.45427451E-05,
                    #3.22832372E-08, -5.60168617E-12, -2.90685309E+04,
                    #2.31106182E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 2.74085371E+01,  2.20461555E-02, -7.41197549E-06,
                    #1.13845768E-09, -6.55893154E-14, -3.71427772E+04,
                    #-1.10888518E+02])),
        #note='9/14')

#species(name='CC5H10OOH-DO2',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1405.00],
                    #[ 1.88094547E+00,  8.54729606E-02, -7.05748924E-05,
                    #3.03432704E-08, -5.26290667E-12, -2.62094659E+04,
                    #2.54857664E+01]),
                #NASA([1405.00, 5000.00],
                    #[ 2.54672904E+01,  2.39654991E-02, -8.11944280E-06,
                    #1.25250847E-09, -7.23427482E-14, -3.37234155E+04,
                    #-9.88516510E+01])),
        #note='9/14')

#species(name='DC5H10OOH-AO2',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1397.00],
                    #[ 1.56380071E+00,  8.29798833E-02, -6.44672355E-05,
                    #2.61513298E-08, -4.33760267E-12, -2.40270626E+04,
                    #2.96266556E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 2.51246846E+01,  2.47883083E-02, -8.52048861E-06,
                    #1.32728569E-09, -7.71886075E-14, -3.18988778E+04,
                    #-9.57993329E+01])),
        #note='9/14')

#species(name='DC5H10OOH-BO2',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1389.00],
                    #[ 1.60288955E+00,  8.61877924E-02, -6.89692568E-05,
                    #2.84213101E-08, -4.74629543E-12, -2.68785795E+04,
                    #2.68316798E+01]),
                #NASA([1389.00, 5000.00],
                    #[ 2.70053318E+01,  2.29516730E-02, -7.84494464E-06,
                    #1.21831089E-09, -7.07267404E-14, -3.52825198E+04,
                    #-1.08169419E+02])),
        #note='9/14')

#species(name='DC5H10OOH-CO2',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1405.00],
                    #[ 1.88094547E+00,  8.54729606E-02, -7.05748924E-05,
                    #3.03432704E-08, -5.26290667E-12, -2.62094659E+04,
                    #2.54857664E+01]),
                #NASA([1405.00, 5000.00],
                    #[ 2.54672904E+01,  2.39654991E-02, -8.11944280E-06,
                    #1.25250847E-09, -7.23427482E-14, -3.37234155E+04,
                    #-9.88516510E+01])),
        #note='9/14')

#species(name='C5H9A-A,BOOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1391.00],
                    #[ 1.74209472E+00,  8.72007406E-02, -7.13938821E-05,
                    #3.00232775E-08, -5.09932418E-12, -1.89798996E+04,
                    #2.85434780E+01]),
                #NASA([1391.00, 5000.00],
                    #[ 2.74649238E+01,  2.22841428E-02, -7.64480931E-06,
                    #1.19001972E-09, -6.91930019E-14, -2.74019678E+04,
                    #-1.07847544E+02])),
        #note='9/14')

#species(name='C5H9A-A,COOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1406.00],
                    #[ 1.49498946E+00,  8.78482767E-02, -7.45627852E-05,
                    #3.27306977E-08, -5.76071753E-12, -1.81772711E+04,
                    #2.97356404E+01]),
                #NASA([1406.00, 5000.00],
                    #[ 2.58537301E+01,  2.32542931E-02, -7.88336202E-06,
                    #1.21662388E-09, -7.02924813E-14, -2.58216418E+04,
                    #-9.82790978E+01])),
        #note='9/14')

#species(name='C5H9A-A,DOOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1399.00],
                    #[ 1.23277981E+00,  8.51858495E-02, -6.84068676E-05,
                    #2.85833888E-08, -4.85431762E-12, -1.60042253E+04,
                    #3.29170068E+01]),
                #NASA([1399.00, 5000.00],
                    #[ 2.54389507E+01,  2.41064865E-02, -8.28741475E-06,
                    #1.29112221E-09, -7.50916362E-14, -2.39536208E+04,
                    #-9.54706001E+01])),
        #note='9/14')

#species(name='C5H9A-B,COOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1398.00],
                    #[ 2.12165412E+00,  8.93553243E-02, -7.69753685E-05,
                    #3.38931026E-08, -5.95683390E-12, -2.11699109E+04,
                    #2.48217368E+01]),
                #NASA([1398.00, 5000.00],
                    #[ 2.78585754E+01,  2.13921889E-02, -7.21720450E-06,
                    #1.11103758E-09, -6.41065604E-14, -2.92578030E+04,
                    #-1.10511702E+02])),
        #note='9/14')

#species(name='C5H9A-B,DOOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1391.00],
                    #[ 1.74209472E+00,  8.72007406E-02, -7.13938821E-05,
                    #3.00232775E-08, -5.09932418E-12, -1.89798996E+04,
                    #2.85434780E+01]),
                #NASA([1391.00, 5000.00],
                    #[ 2.74649238E+01,  2.22841428E-02, -7.64480931E-06,
                    #1.19001972E-09, -6.91930019E-14, -2.74019678E+04,
                    #-1.07847544E+02])),
        #note='9/14')

#species(name='C5H9A-C,DOOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1406.00],
                    #[ 1.49498946E+00,  8.78482767E-02, -7.45627852E-05,
                    #3.27306977E-08, -5.76071753E-12, -1.81772711E+04,
                    #2.97356404E+01]),
                #NASA([1406.00, 5000.00],
                    #[ 2.58537301E+01,  2.32542931E-02, -7.88336202E-06,
                    #1.21662388E-09, -7.02924813E-14, -2.58216418E+04,
                    #-9.82790978E+01])),
        #note='9/14')

#species(name='C5H9B-A,AOOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1392.00],
                    #[ 4.46595997E+00,  6.76842861E-02, -4.27291923E-05,
                    #1.34530161E-08, -1.68677293E-12, -1.91819189E+04,
                    #1.86545708E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 2.40001772E+01,  2.54913091E-02, -8.80097909E-06,
                    #1.37500879E-09, -8.01262867E-14, -2.63422161E+04,
                    #-8.75309232E+01])),
        #note='9/14')

#species(name='C5H9B-A,COOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1675.00],
                    #[ 5.11997216E+00,  7.06862242E-02, -5.05395936E-05,
                    #1.88850975E-08, -2.89544987E-12, -2.14746499E+04,
                    #1.38868784E+01]),
                #NASA([1675.00, 5000.00],
                    #[ 2.46412701E+01,  2.43204620E-02, -8.25843742E-06,
                    #1.27587110E-09, -7.37687166E-14, -2.81624190E+04,
                    #-9.06566816E+01])),
        #note='9/14')

#species(name='C5H9B-A,DOOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1392.00],
                    #[ 4.46595997E+00,  6.76842861E-02, -4.27291923E-05,
                    #1.34530161E-08, -1.68677293E-12, -1.91819189E+04,
                    #1.93440525E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 2.40001772E+01,  2.54913091E-02, -8.80097909E-06,
                    #1.37500879E-09, -8.01262867E-14, -2.63422161E+04,
                    #-8.68414415E+01])),
        #note='9/14')

#species(name='C5H9B-C,DOOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1675.00],
                    #[ 5.11997216E+00,  7.06862242E-02, -5.05395936E-05,
                    #1.88850975E-08, -2.89544987E-12, -2.14746499E+04,
                    #1.25079152E+01]),
                #NASA([1675.00, 5000.00],
                    #[ 2.46412701E+01,  2.43204620E-02, -8.25843742E-06,
                    #1.27587110E-09, -7.37687166E-14, -2.81624190E+04,
                    #-9.20356449E+01])),
        #note='9/14')

#species(name='C5H9C-A,AOOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1681.00],
                    #[ 1.63830609E+00,  7.61674011E-02, -4.75989477E-05,
                    #1.35364970E-08, -1.35802072E-12, -1.74301183E+04,
                    #3.22851396E+01]),
                #NASA([1681.00, 5000.00],
                    #[ 2.35258181E+01,  2.77182294E-02, -1.00794505E-05,
                    #1.63650323E-09, -9.81022627E-14, -2.48044613E+04,
                    #-8.55170933E+01])),
        #note='9/14')

#species(name='C5H9C-A,BOOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1393.00],
                    #[ 2.71890486E+00,  8.21702316E-02, -6.41039066E-05,
                    #2.57115711E-08, -4.18873475E-12, -2.04375295E+04,
                    #2.39535941E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 2.71089434E+01,  2.25601968E-02, -7.73502682E-06,
                    #1.20364175E-09, -6.99693974E-14, -2.86174586E+04,
                    #-1.06059958E+02])),
        #note='9/14')

#species(name='C5H9C-A,DOOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1681.00],
                    #[ 1.63830609E+00,  7.61674011E-02, -4.75989477E-05,
                    #1.35364970E-08, -1.35802072E-12, -1.74301183E+04,
                    #3.29746213E+01]),
                #NASA([1681.00, 5000.00],
                    #[ 2.35258181E+01,  2.77182294E-02, -1.00794505E-05,
                    #1.63650323E-09, -9.81022627E-14, -2.48044613E+04,
                    #-8.48276117E+01])),
        #note='9/14')

#species(name='C5H9C-B,DOOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1393.00],
                    #[ 2.71890486E+00,  8.21702316E-02, -6.41039066E-05,
                    #2.57115711E-08, -4.18873475E-12, -2.04375295E+04,
                    #2.32641124E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 2.71089434E+01,  2.25601968E-02, -7.73502682E-06,
                    #1.20364175E-09, -6.99693974E-14, -2.86174586E+04,
                    #-1.06749440E+02])),
        #note='9/14')

#species(name='C5H9D-A,AOOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1399.00],
                    #[ 1.23277981E+00,  8.51858495E-02, -6.84068676E-05,
                    #2.85833888E-08, -4.85431762E-12, -1.60042253E+04,
                    #3.22275252E+01]),
                #NASA([1399.00, 5000.00],
                    #[ 2.54389507E+01,  2.41064865E-02, -8.28741475E-06,
                    #1.29112221E-09, -7.50916362E-14, -2.39536208E+04,
                    #-9.61600818E+01])),
        #note='9/14')

#species(name='C5H9D-A,BOOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1391.00],
                    #[ 1.18667054E+00,  8.87250785E-02, -7.32632472E-05,
                    #3.10459232E-08, -5.30475828E-12, -1.88422966E+04,
                    #3.12173183E+01]),
                #NASA([1391.00, 5000.00],
                    #[ 2.73248174E+01,  2.23513350E-02, -7.65730755E-06,
                    #1.19091789E-09, -6.92045182E-14, -2.73558944E+04,
                    #-1.07224812E+02])),
        #note='9/14')

#species(name='C5H9D-A,COOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1406.00],
                    #[ 2.18806983E+00,  8.56911733E-02, -7.20711649E-05,
                    #3.14354507E-08, -5.50793784E-12, -1.83327730E+04,
                    #2.64408072E+01]),
                #NASA([1406.00, 5000.00],
                    #[ 2.58515091E+01,  2.32105631E-02, -7.85811852E-06,
                    #1.21161952E-09, -6.99578340E-14, -2.57865221E+04,
                    #-9.80163771E+01])),
        #note='9/14')

#species(name='C5H9D-B,COOH',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 1398.00],
                    #[ 2.12165412E+00,  8.93553243E-02, -7.69753685E-05,
                    #3.38931026E-08, -5.95683390E-12, -2.11699109E+04,
                    #2.41322551E+01]),
                #NASA([1398.00, 5000.00],
                    #[ 2.78585754E+01,  2.13921889E-02, -7.21720450E-06,
                    #1.11103758E-09, -6.41065604E-14, -2.92578030E+04,
                    #-1.11201184E+02])),
        #note='9/14')

#species(name='C5H9A-AOOH',
        #atoms='C:5 H:10 O:2',
        #thermo=(NASA([300.00, 1388.00],
                    #[-5.66437031E-01,  7.51592771E-02, -5.92588303E-05,
                    #2.44464873E-08, -4.14011509E-12, -1.49441419E+04,
                    #3.46236211E+01]),
                #NASA([1388.00, 5000.00],
                    #[ 2.12338541E+01,  2.14695668E-02, -7.41711017E-06,
                    #1.15964346E-09, -6.76214281E-14, -2.22824726E+04,
                    #-8.15535062E+01])),
        #note='9/8/14')

#species(name='C5H9A-COOH',
        #atoms='C:5 H:10 O:2',
        #thermo=(NASA([300.00, 1396.00],
                    #[-2.23218718E+00,  8.31068782E-02, -6.96695008E-05,
                    #2.98090622E-08, -5.11291423E-12, -1.66665651E+04,
                    #3.93336979E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 2.21762503E+01,  2.02997794E-02, -6.93527858E-06,
                    #1.07653985E-09, -6.24712973E-14, -2.45061123E+04,
                    #-8.96059444E+01])),
        #note='9/8/14')

#species(name='C5H9A-DOOH',
        #atoms='C:5 H:10 O:2',
        #thermo=(NASA([300.00, 1393.00],
                    #[ 5.84051402E-01,  6.61435369E-02, -4.64794214E-05,
                    #1.71320929E-08, -2.62129070E-12, -1.66564666E+04,
                    #2.90021482E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 1.91192483E+01,  2.31083713E-02, -7.94263265E-06,
                    #1.23713692E-09, -7.19377465E-14, -2.31563003E+04,
                    #-7.06976781E+01])),
        #note='9/8/14')

#species(name='C5H9B-AOOH',
        #atoms='C:5 H:10 O:2',
        #thermo=(NASA([300.00, 1385.00],
                    #[ 5.49656384E-01,  7.04990310E-02, -5.27825801E-05,
                    #2.06470792E-08, -3.33664189E-12, -1.64635942E+04,
                    #2.87891957E+01]),
                #NASA([1385.00, 5000.00],
                    #[ 2.11330268E+01,  2.15086873E-02, -7.42113144E-06,
                    #1.15934717E-09, -6.75687044E-14, -2.35608139E+04,
                    #-8.15040563E+01])),
        #note='9/8/14')

#species(name='C5H9B-DOOH',
        #atoms='C:5 H:10 O:2',
        #thermo=(NASA([300.00, 1385.00],
                    #[ 5.49656384E-01,  7.04990310E-02, -5.27825801E-05,
                    #2.06470792E-08, -3.33664189E-12, -1.64635942E+04,
                    #2.87891957E+01]),
                #NASA([1385.00, 5000.00],
                    #[ 2.11330268E+01,  2.15086873E-02, -7.42113144E-06,
                    #1.15934717E-09, -6.75687044E-14, -2.35608139E+04,
                    #-8.15040563E+01])),
        #note='9/8/14')

#species(name='C5H9C-AOOH',
        #atoms='C:5 H:10 O:2',
        #thermo=(NASA([300.00, 1396.00],
                    #[-1.33187657E-02,  6.92568397E-02, -5.06647207E-05,
                    #1.94153559E-08, -3.06921462E-12, -1.60254301E+04,
                    #3.17599798E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 1.95116301E+01,  2.27723060E-02, -7.82629606E-06,
                    #1.21895173E-09, -7.08783753E-14, -2.27377278E+04,
                    #-7.28224624E+01])),
        #note='9/8/14')

#species(name='C5H9C-BOOH',
        #atoms='C:5 H:10 O:2',
        #thermo=(NASA([300.00, 1393.00],
                    #[-1.04862343E+00,  8.26699722E-02, -7.17941940E-05,
                    #3.18245309E-08, -5.62912876E-12, -1.90383401E+04,
                    #3.41116802E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 2.30088665E+01,  1.91355851E-02, -6.44358123E-06,
                    #9.91225120E-10, -5.71817398E-14, -2.66103173E+04,
                    #-9.24108219E+01])),
        #note='9/8/14')

#species(name='C5H9OA-AOOH-B',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1421.00],
                    #[-6.75374276E+00,  1.02663107E-01, -9.33055784E-05,
                    #4.25654630E-08, -7.61110350E-12, -2.96264363E+04,
                    #6.11388448E+01]),
                #NASA([1421.00, 5000.00],
                    #[ 2.21045706E+01,  2.23104424E-02, -7.40451010E-06,
                    #1.12656049E-09, -6.44483178E-14, -3.81730717E+04,
                    #-8.89517851E+01])),
        #note='9/14')

#species(name='C5H9OA-AOOH-C',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1426.00],
                    #[-6.22575239E+00,  9.78862329E-02, -8.25002804E-05,
                    #3.45778461E-08, -5.69404109E-12, -2.89425112E+04,
                    #5.91111169E+01]),
                #NASA([1426.00, 5000.00],
                    #[ 2.29635481E+01,  2.19928354E-02, -7.38851185E-06,
                    #1.13400285E-09, -6.52900790E-14, -3.80694629E+04,
                    #-9.45175782E+01])),
        #note='9/14')

#species(name='C5H9OA-AOOH-D',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1411.00],
                    #[-8.29760859E+00,  1.03901358E-01, -9.08046967E-05,
                    #3.99841755E-08, -6.96787716E-12, -2.65057037E+04,
                    #6.98840126E+01]),
                #NASA([1411.00, 5000.00],
                    #[ 2.19233422E+01,  2.33749910E-02, -7.96814249E-06,
                    #1.23462014E-09, -7.15415752E-14, -3.58700293E+04,
                    #-8.86666847E+01])),
        #note='9/14')

#species(name='C5H9OA-BOOH-A',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1417.00],
                    #[-1.90416408E+00,  8.28866784E-02, -6.64998506E-05,
                    #2.72221761E-08, -4.44711330E-12, -3.00389816E+04,
                    #3.98116691E+01]),
                #NASA([1417.00, 5000.00],
                    #[ 2.17022735E+01,  2.26119914E-02, -7.58981668E-06,
                    #1.16366327E-09, -6.69308130E-14, -3.75868512E+04,
                    #-8.49179287E+01])),
        #note='9/14')

#species(name='C5H9OA-BOOH-C',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1417.00],
                    #[-4.48273125E-01,  7.90873425E-02, -5.97079860E-05,
                    #2.24183926E-08, -3.31048774E-12, -3.23783865E+04,
                    #3.12502863E+01]),
                #NASA([1417.00, 5000.00],
                    #[ 2.32822727E+01,  2.16339746E-02, -7.33435063E-06,
                    #1.13242966E-09, -6.54655290E-14, -4.02487769E+04,
                    #-9.51976254E+01])),
        #note='9/14')

#species(name='C5H9OA-BOOH-D',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1417.00],
                    #[-1.90416408E+00,  8.28866784E-02, -6.64998506E-05,
                    #2.72221761E-08, -4.44711330E-12, -3.00389816E+04,
                    #3.98116691E+01]),
                #NASA([1417.00, 5000.00],
                    #[ 2.17022735E+01,  2.26119914E-02, -7.58981668E-06,
                    #1.16366327E-09, -6.69308130E-14, -3.75868512E+04,
                    #-8.49179287E+01])),
        #note='9/14')

#species(name='C5H9OA-COOH-A',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1416.00],
                    #[-8.12800697E+00,  1.05722616E-01, -9.50365933E-05,
                    #4.28440557E-08, -7.59517099E-12, -2.89856598E+04,
                    #6.89715637E+01]),
                #NASA([1416.00, 5000.00],
                    #[ 2.22283125E+01,  2.27380710E-02, -7.66670701E-06,
                    #1.17919710E-09, -6.79799943E-14, -3.81557392E+04,
                    #-8.95003647E+01])),
        #note='9/14')

#species(name='C5H9OA-COOH-B',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1427.00],
                    #[-6.50835976E+00,  1.04136207E-01, -9.72241217E-05,
                    #4.52696757E-08, -8.20508870E-12, -3.21159886E+04,
                    #5.91862621E+01]),
                #NASA([1427.00, 5000.00],
                    #[ 2.24138827E+01,  2.15060611E-02, -7.01359864E-06,
                    #1.05420532E-09, -5.97909281E-14, -4.04392956E+04,
                    #-9.04376317E+01])),
        #note='9/14')

#species(name='C5H9OA-COOH-D',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1416.00],
                    #[-8.12800697E+00,  1.05722616E-01, -9.50365933E-05,
                    #4.28440557E-08, -7.59517099E-12, -2.89856598E+04,
                    #6.89715637E+01]),
                #NASA([1416.00, 5000.00],
                    #[ 2.22283125E+01,  2.27380710E-02, -7.66670701E-06,
                    #1.17919710E-09, -6.79799943E-14, -3.81557392E+04,
                    #-8.95003647E+01])),
        #note='9/14')

#species(name='C5H9OA-DOOH-A',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1412.00],
                    #[-1.02378200E+01,  1.07917027E-01, -9.36515087E-05,
                    #4.08759994E-08, -7.05929060E-12, -3.65150948E+04,
                    #8.02224716E+01]),
                #NASA([1412.00, 5000.00],
                    #[ 2.13403622E+01,  2.42805757E-02, -8.27822066E-06,
                    #1.28284876E-09, -7.43449277E-14, -4.63370599E+04,
                    #-8.56056983E+01])),
        #note='9/14')

#species(name='C5H9OA-DOOH-B',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1424.00],
                    #[-8.36309520E+00,  1.04931903E-01, -9.32716329E-05,
                    #4.15748714E-08, -7.27559342E-12, -3.96787628E+04,
                    #6.93062651E+01]),
                #NASA([1424.00, 5000.00],
                    #[ 2.15643048E+01,  2.31721670E-02, -7.69883177E-06,
                    #1.17234482E-09, -6.71124276E-14, -4.86745380E+04,
                    #-8.68606431E+01])),
        #note='9/14')

#species(name='C5H9OA-DOOH-C',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1418.00],
                    #[-1.02222511E+01,  1.11938112E-01, -1.02203494E-04,
                    #4.66370825E-08, -8.33864572E-12, -3.86568367E+04,
                    #7.81342579E+01]),
                #NASA([1418.00, 5000.00],
                    #[ 2.17583837E+01,  2.33472354E-02, -7.83108254E-06,
                    #1.20022182E-09, -6.90207652E-14, -4.81877268E+04,
                    #-8.83811218E+01])),
        #note='9/14')

#species(name='C5H9OB-COOH-A',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1427.00],
                    #[-2.03626914E+00,  8.44316222E-02, -6.63448448E-05,
                    #2.58620472E-08, -3.95367034E-12, -3.20953603E+04,
                    #3.91516619E+01]),
                #NASA([1427.00, 5000.00],
                    #[ 2.33660026E+01,  2.12967547E-02, -7.16311642E-06,
                    #1.10028483E-09, -6.33839052E-14, -4.03214180E+04,
                    #-9.55650042E+01])),
        #note='9/14')

#species(name='C5H9OB-COOH-D',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1427.00],
                    #[-2.03626914E+00,  8.44316222E-02, -6.63448448E-05,
                    #2.58620472E-08, -3.95367034E-12, -3.20953603E+04,
                    #3.84621803E+01]),
                #NASA([1427.00, 5000.00],
                    #[ 2.33660026E+01,  2.12967547E-02, -7.16311642E-06,
                    #1.10028483E-09, -6.33839052E-14, -4.03214180E+04,
                    #-9.62544859E+01])),
        #note='9/14')

#species(name='C5H9OB-DOOH-A',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1420.00],
                    #[-3.52560638E+00,  8.53243101E-02, -6.71300274E-05,
                    #2.68114187E-08, -4.26214321E-12, -3.08855835E+04,
                    #4.70812100E+01]),
                #NASA([1420.00, 5000.00],
                    #[ 2.09989877E+01,  2.35749161E-02, -7.91024357E-06,
                    #1.21256198E-09, -6.97360491E-14, -3.87964686E+04,
                    #-8.27787933E+01])),
        #note='9/14')

#species(name='C5H9OB-DOOH-C',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1416.00],
                    #[-3.01430534E+00,  8.55622233E-02, -6.67458085E-05,
                    #2.61563238E-08, -4.06382943E-12, -3.30739466E+04,
                    #4.22585328E+01]),
                #NASA([1416.00, 5000.00],
                    #[ 2.23010712E+01,  2.28353062E-02, -7.73563047E-06,
                    #1.19365671E-09, -6.89722094E-14, -4.13407064E+04,
                    #-9.21482846E+01])),
        #note='9/14')

#species(name='C5H9OC-DOOH-A',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1424.00],
                    #[-4.57472336E+00,  9.50200664E-02, -8.14529347E-05,
                    #3.48249376E-08, -5.85417355E-12, -2.84463800E+04,
                    #5.26542848E+01]),
                #NASA([1424.00, 5000.00],
                    #[ 2.34532517E+01,  2.11267245E-02, -7.08308588E-06,
                    #1.08555326E-09, -6.24351088E-14, -3.71203864E+04,
                    #-9.45207023E+01])),
        #note='9/14')

#species(name='C5H9OC-DOOH-B',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1424.00],
                    #[-4.82866538E+00,  1.01816578E-01, -9.70871530E-05,
                    #4.61039008E-08, -8.49951271E-12, -3.12884271E+04,
                    #5.08785569E+01]),
                #NASA([1424.00, 5000.00],
                    #[ 2.30916286E+01,  2.05720891E-02, -6.70383557E-06,
                    #1.00697145E-09, -5.70805180E-14, -3.91980266E+04,
                    #-9.30753568E+01])),
        #note='9/14')

#species(name='C4H7O2-1',
        #atoms='C:4 H:7 O:1',
        #thermo=(NASA([300.00, 1684.00],
                    #[ 5.16434924E+00,  2.53427725E-02, -1.39669041E-06,
                    #-6.00713944E-09,  1.76696879E-12,  5.03210693E+03,
                    #4.37441148E+00]),
                #NASA([1684.00, 5000.00],
                    #[ 1.12685690E+01,  2.01724820E-02, -7.41925667E-06,
                    #1.21275519E-09, -7.30115348E-14,  2.13876442E+03,
                    #-3.15208064E+01])),
        #note='9/8/14')

#species(name='AC3H5OCH2',
        #atoms='C:4 H:7 O:1',
        #thermo=(NASA([300.00, 1397.00],
                    #[ 1.20089076E+00,  4.24760987E-02, -2.94114641E-05,
                    #1.11824860E-08, -1.81309530E-12,  7.26792055E+03,
                    #2.43628654E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 1.19880368E+01,  1.70263516E-02, -5.78617265E-06,
                    #8.94150465E-10, -5.16994613E-14,  3.48367473E+03,
                    #-3.35860630E+01])),
        #note='9/8/14')

#species(name='CH3COCH2OCH2CH2',
        #atoms='C:5 H:9 O:2',
        #thermo=(NASA([300.00, 1398.00],
                    #[ 2.46265398E+00,  5.31605978E-02, -3.44158813E-05,
                    #1.17106545E-08, -1.63587278E-12, -2.35149168E+04,
                    #2.32544951E+01]),
                #NASA([1398.00, 5000.00],
                    #[ 1.61915801E+01,  2.17259962E-02, -7.18255818E-06,
                    #1.08967551E-09, -6.22058703E-14, -2.83312411E+04,
                    #-5.06778787E+01])),
        #note='14')

#species(name='IC5KETAA',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 2.55453432E+00,  6.48725206E-02, -4.21449634E-05,
                    #1.35396181E-08, -1.69771989E-12, -4.10439664E+04,
                    #2.25509833E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 2.14062122E+01,  2.31814714E-02, -7.94585296E-06,
                    #1.23549697E-09, -7.17602293E-14, -4.77919593E+04,
                    #-7.94699242E+01])),
        #note='9/8/14')

#species(name='IC5KETAB',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1387.00],
                    #[-2.76644890E-01,  8.19422293E-02, -6.36165752E-05,
                    #2.46645924E-08, -3.82375149E-12, -4.22291828E+04,
                    #3.35972645E+01]),
                #NASA([1387.00, 5000.00],
                    #[ 2.54563544E+01,  2.01871439E-02, -7.02056943E-06,
                    #1.10298380E-09, -6.45487927E-14, -5.09169381E+04,
                    #-1.03884759E+02])),
        #note='9/8/14')

#species(name='IC5KETAC',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1530.00],
                    #[ 3.38364660E+00,  6.08702747E-02, -3.15372777E-05,
                    #4.94036887E-09,  4.82022927E-13, -4.32141831E+04,
                    #1.74830999E+01]),
                #NASA([1530.00, 5000.00],
                    #[ 2.36202007E+01,  2.16574723E-02, -7.50493626E-06,
                    #1.17597427E-09, -6.86855912E-14, -5.08902285E+04,
                    #-9.37964825E+01])),
        #note='9/8/14')

#species(name='IC5KETAD',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 2.55453432E+00,  6.48725206E-02, -4.21449634E-05,
                    #1.35396181E-08, -1.69771989E-12, -4.10439664E+04,
                    #2.25509833E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 2.14062122E+01,  2.31814714E-02, -7.94585296E-06,
                    #1.23549697E-09, -7.17602293E-14, -4.77919593E+04,
                    #-7.94699242E+01])),
        #note='9/8/14')

#species(name='IC5KETCA',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1403.00],
                    #[ 3.30504848E+00,  6.12212470E-02, -3.95733068E-05,
                    #1.30417749E-08, -1.72022312E-12, -4.52388966E+04,
                    #1.90774795E+01]),
                #NASA([1403.00, 5000.00],
                    #[ 1.99169154E+01,  2.37975805E-02, -8.01511093E-06,
                    #1.23140060E-09, -7.09187874E-14, -5.11169360E+04,
                    #-7.05780003E+01])),
        #note='9/8/14')

#species(name='IC5KETCB',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1388.00],
                    #[ 4.99341353E-01,  7.79929244E-02, -6.02225926E-05,
                    #2.34360987E-08, -3.65044242E-12, -4.64243747E+04,
                    #2.93525587E+01]),
                #NASA([1388.00, 5000.00],
                    #[ 2.43081284E+01,  2.03116947E-02, -6.88557448E-06,
                    #1.06413626E-09, -6.15901378E-14, -5.43890101E+04,
                    #-9.76241400E+01])),
        #note='9/8/14')

#species(name='IC5KETCD',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1394.00],
                    #[-2.83543507E-01,  8.06299025E-02, -6.46191514E-05,
                    #2.62468255E-08, -4.26140982E-12, -4.30486170E+04,
                    #3.51463538E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 2.37898908E+01,  2.04031709E-02, -6.84168010E-06,
                    #1.04954951E-09, -6.04313994E-14, -5.08956742E+04,
                    #-9.25390969E+01])),
        #note='9/8/14')

#species(name='IC5KETDA',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1395.00],
                    #[ 1.38997831E+00,  7.01103489E-02, -4.80214291E-05,
                    #1.61542500E-08, -2.10987314E-12, -4.07920247E+04,
                    #2.76016990E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 2.25488083E+01,  2.21036481E-02, -7.55547668E-06,
                    #1.17303142E-09, -6.80737156E-14, -4.81889779E+04,
                    #-8.63830583E+01])),
        #note='9/8/14')

#species(name='IC5KETDB',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1381.00],
                    #[ 1.80844843E+00,  7.16334880E-02, -5.01029117E-05,
                    #1.70680523E-08, -2.25260208E-12, -4.37006021E+04,
                    #2.30513467E+01]),
                #NASA([1381.00, 5000.00],
                    #[ 2.43539807E+01,  2.04860817E-02, -6.99182132E-06,
                    #1.08544233E-09, -6.30195097E-14, -5.15777648E+04,
                    #-9.84002879E+01])),
        #note='9/8/14')

#species(name='IC5KETDC',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1391.00],
                    #[-2.39968171E+00,  9.23788752E-02, -7.95459934E-05,
                    #3.43729808E-08, -5.91010616E-12, -4.07649190E+04,
                    #4.36007006E+01]),
                #NASA([1391.00, 5000.00],
                    #[ 2.59294325E+01,  1.93869680E-02, -6.66456173E-06,
                    #1.03969171E-09, -6.05693148E-14, -4.98346105E+04,
                    #-1.05986051E+02])),
        #note='9/8/14')

#species(name='C2H5CH(CH2O)CHO',
        #atoms='C:5 H:9 O:2',
        #thermo=(NASA([300.00, 1670.00],
                    #[ 6.20340512E+00,  3.78488938E-02, -8.21632367E-06,
                    #-4.45998794E-09,  1.69104440E-12, -2.34210066E+04,
                    #3.88282175E+00]),
                #NASA([1670.00, 5000.00],
                    #[ 1.54491093E+01,  2.58830088E-02, -9.38194545E-06,
                    #1.51919963E-09, -9.08764412E-14, -2.73632213E+04,
                    #-4.89349130E+01])),
        #note='14')

#species(name='C3H5O(CH3)CHO',
        #atoms='C:5 H:9 O:2',
        #thermo=(NASA([300.00, 1680.00],
                    #[ 2.44492593E+00,  5.39988220E-02, -2.76353434E-05,
                    #5.14986988E-09, -2.28548561E-14, -2.43042020E+04,
                    #1.91347808E+01]),
                #NASA([1680.00, 5000.00],
                    #[ 1.74003757E+01,  2.43634015E-02, -8.89217693E-06,
                    #1.44672950E-09, -8.68336325E-14, -2.96862586E+04,
                    #-6.26150618E+01])),
        #note='9/14')

#species(name='C2H4OCH(CH3)CHO',
        #atoms='C:5 H:9 O:2',
        #thermo=(NASA([300.00, 1405.00],
                    #[ 3.17102493E+00,  5.52252992E-02, -3.48045866E-05,
                    #1.07470812E-08, -1.26264879E-12, -2.54634549E+04,
                    #1.60763592E+01]),
                #NASA([1405.00, 5000.00],
                    #[ 1.87759820E+01,  2.09363794E-02, -7.11851372E-06,
                    #1.10072541E-09, -6.36812526E-14, -3.10560920E+04,
                    #-6.84256552E+01])),
        #note='14')

#species(name='CH2OC3H6CHO',
        #atoms='C:5 H:9 O:2',
        #thermo=(NASA([300.00, 1670.00],
                    #[ 6.20340512E+00,  3.78488938E-02, -8.21632367E-06,
                    #-4.45998794E-09,  1.69104440E-12, -2.34210066E+04,
                    #3.88282175E+00]),
                #NASA([1670.00, 5000.00],
                    #[ 1.54491093E+01,  2.58830088E-02, -9.38194545E-06,
                    #1.51919963E-09, -9.08764412E-14, -2.73632213E+04,
                    #-4.89349130E+01])),
        #note='9/14')

#species(name='CH3COCH(CH3)CH2O',
        #atoms='C:5 H:9 O:2',
        #thermo=(NASA([300.00, 1514.00],
                    #[ 4.75737279E+00,  4.52905574E-02, -2.09520730E-05,
                    #2.65371883E-09,  4.60703138E-13, -2.73333149E+04,
                    #1.02325004E+01]),
                #NASA([1514.00, 5000.00],
                    #[ 1.72485932E+01,  2.16326796E-02, -7.22890702E-06,
                    #1.10467817E-09, -6.33813429E-14, -3.21371394E+04,
                    #-5.86643290E+01])),
        #note='4')

#species(name='CH3CO(CH3)C2H3O',
        #atoms='C:5 H:9 O:2',
        #thermo=(NASA([300.00, 1510.00],
                    #[ 4.60922258E+00,  4.63105317E-02, -1.85546802E-05,
                    #-2.86158793E-10,  1.27532426E-12, -2.87895902E+04,
                    #7.87701574E+00]),
                #NASA([1510.00, 5000.00],
                    #[ 1.93742496E+01,  2.06425485E-02, -7.06983683E-06,
                    #1.09902869E-09, -6.38342204E-14, -3.47003768E+04,
                    #-7.43848166E+01])),
        #note='14')

#species(name='CH2OCOCH(CH3)CH3',
        #atoms='C:5 H:9 O:2',
        #thermo=(NASA([300.00, 1489.00],
                    #[ 7.66113508E+00,  3.31212888E-02, -2.76253288E-06,
                    #-8.32131128E-09,  2.78275744E-12, -2.68823428E+04,
                    #-4.11946991E+00]),
                #NASA([1489.00, 5000.00],
                    #[ 1.78986194E+01,  2.13374499E-02, -7.19313241E-06,
                    #1.10668165E-09, -6.38266549E-14, -3.16063716E+04,
                    #-6.33320448E+01])),
        #note='4')

#species(name='CHOC3H6CH2O',
        #atoms='C:5 H:9 O:2',
        #thermo=(NASA([300.00, 1485.00],
                    #[ 2.83718968E+00,  5.41714197E-02, -2.93161004E-05,
                    #5.70409714E-09,  8.40746252E-14, -2.28851788E+04,
                    #1.87850931E+01]),
                #NASA([1485.00, 5000.00],
                    #[ 1.98811047E+01,  1.99858558E-02, -6.79549451E-06,
                    #1.05139684E-09, -6.08714283E-14, -2.92192606E+04,
                    #-7.45004128E+01])),
        #note='9/8/14')

#species(name='CHOCH2(CH3)C2H3O',
        #atoms='C:5 H:9 O:2',
        #thermo=(NASA([300.00, 1519.00],
                    #[ 5.21028340E+00,  4.44018385E-02, -1.51327339E-05,
                    #-2.59536762E-09,  1.79445594E-12, -2.65143041E+04,
                    #5.32661738E+00]),
                #NASA([1519.00, 5000.00],
                    #[ 2.00669200E+01,  2.02402238E-02, -6.97500713E-06,
                    #1.08902290E-09, -6.34524956E-14, -3.26362270E+04,
                    #-7.80517736E+01])),
        #note='4')

#species(name='CHOCHOCH(CH3)CH3',
        #atoms='C:5 H:9 O:2',
        #thermo=(NASA([300.00, 1385.00],
                    #[ 1.81302803E+00,  6.07908730E-02, -3.85150152E-05,
                    #1.13873917E-08, -1.21465641E-12, -2.41018252E+04,
                    #2.24191537E+01]),
                #NASA([1385.00, 5000.00],
                    #[ 2.09665488E+01,  1.99702092E-02, -6.98669046E-06,
                    #1.10141723E-09, -6.45912153E-14, -3.11088773E+04,
                    #-8.17784579E+01])),
        #note='4')

#species(name='AC5H10OH',
        #atoms='C:5 H:11 O:1',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 5.07415221E-01,  5.51668766E-02, -2.78117387E-05,
                    #5.35272238E-09, -2.01780596E-14, -1.70837852E+04,
                    #3.22675556E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 1.58342161E+01,  2.53792543E-02, -8.61212271E-06,
                    #1.32986083E-09, -7.68625656E-14, -2.29861585E+04,
                    #-5.21422138E+01])),
        #note='9/8/14')

#species(name='BC5H10OH',
        #atoms='C:5 H:11 O:1',
        #thermo=(NASA([300.00, 1384.00],
                    #[ 2.16094248E+00,  5.26897228E-02, -2.56800949E-05,
                    #4.72180512E-09, -1.32098054E-14, -1.94067682E+04,
                    #2.07680728E+01]),
                #NASA([1384.00, 5000.00],
                    #[ 1.69551854E+01,  2.50397308E-02, -8.62897621E-06,
                    #1.34632322E-09, -7.83770200E-14, -2.53059646E+04,
                    #-6.12549178E+01])),
        #note='9/8/14')

#species(name='CC5H10OH',
        #atoms='C:5 H:11 O:1',
        #thermo=(NASA([300.00, 1408.00],
                    #[-1.70282457E+00,  7.44975249E-02, -6.01709532E-05,
                    #2.56502987E-08, -4.42828776E-12, -1.61906329E+04,
                    #3.78082093E+01]),
                #NASA([1408.00, 5000.00],
                    #[ 1.80009484E+01,  2.28638177E-02, -7.59934779E-06,
                    #1.15698844E-09, -6.62082931E-14, -2.24290830E+04,
                    #-6.59495284E+01])),
        #note='9/8/14')

#species(name='AO2C5H10OH',
        #atoms='C:5 H:11 O:3',
        #thermo=(NASA([300.00, 1391.00],
                    #[ 2.30425989E-01,  8.25763029E-02, -6.53026006E-05,
                    #2.66261960E-08, -4.39439687E-12, -3.51451206E+04,
                    #3.25506893E+01]),
                #NASA([1391.00, 5000.00],
                    #[ 2.43089997E+01,  2.26687195E-02, -7.65285953E-06,
                    #1.17885952E-09, -6.80575830E-14, -4.30900126E+04,
                    #-9.53853124E+01])),
        #note='9/8/14')

#species(name='BO2C5H10OH',
        #atoms='C:5 H:11 O:3',
        #thermo=(NASA([300.00, 1395.00],
                    #[ 2.71004906E+00,  7.28069687E-02, -5.37414655E-05,
                    #2.08815586E-08, -3.34176454E-12, -3.80779771E+04,
                    #1.85552934E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 2.28865971E+01,  2.40778073E-02, -8.16957904E-06,
                    #1.26165301E-09, -7.29326115E-14, -4.49268135E+04,
                    #-8.92415513E+01])),
        #note='9/8/14')

#species(name='CO2C5H10OH',
        #atoms='C:5 H:11 O:3',
        #thermo=(NASA([300.00, 1406.00],
                    #[-8.20429195E-02,  8.32830873E-02, -6.85842582E-05,
                    #2.95735681E-08, -5.14907737E-12, -3.43552675E+04,
                    #3.37030942E+01]),
                #NASA([1406.00, 5000.00],
                    #[ 2.24710201E+01,  2.40941302E-02, -8.10010400E-06,
                    #1.24293522E-09, -7.15225197E-14, -4.15001520E+04,
                    #-8.50497640E+01])),
        #note='9/8/14')

#species(name='IC3H5COCH3',
        #atoms='C:5 H:8 O:1',
        #thermo=(NASA([300.00, 1392.00],
                    #[ 1.54292407E+00,  5.05432300E-02, -3.42434122E-05,
                    #1.25252368E-08, -1.95302231E-12, -2.12702785E+04,
                    #1.99855884E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 1.49935813E+01,  1.97788403E-02, -6.78313207E-06,
                    #1.05481230E-09, -6.12614698E-14, -2.60877929E+04,
                    #-5.26202086E+01])),
        #note='9/8/14')

#species(name='IC3H5COCH2',
        #atoms='C:5 H:7 O:1',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 1.42097745E+00,  5.34336638E-02, -4.21903633E-05,
                    #1.76937615E-08, -3.06075290E-12,  6.91082980E+02,
                    #1.99522060E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 1.62814836E+01,  1.63492954E-02, -5.64820836E-06,
                    #8.82756224E-10, -5.14522209E-14, -4.28650692E+03,
                    #-5.91075528E+01])),
        #note='9/8/14')

#species(name='AC3H4COCH3',
        #atoms='C:5 H:7 O:1',
        #thermo=(NASA([300.00, 1384.00],
                    #[ 9.48806278E-01,  5.17869173E-02, -3.74867620E-05,
                    #1.40294881E-08, -2.14698694E-12, -3.10553556E+03,
                    #2.26479362E+01]),
                #NASA([1384.00, 5000.00],
                    #[ 1.60405803E+01,  1.62938876E-02, -5.49275040E-06,
                    #8.45442040E-10, -4.87866397E-14, -8.30972832E+03,
                    #-5.83003416E+01])),
        #note='9/8/14')

#species(name='NEOC5H12',
        #atoms='C:5 H:12',
        #thermo=(NASA([300.00, 1459.00],
                    #[ 8.60482114E-01,  4.63444251E-02,  2.26557531E-06,
                    #-1.91739545E-08,  6.07291520E-12, -2.24020628E+04,
                    #1.78235950E+01]),
                #NASA([1459.00, 5000.00],
                    #[ 2.13410465E+01,  2.32500105E-02, -8.36387815E-06,
                    #1.34306692E-09, -7.97752777E-14, -3.18056043E+04,
                    #-1.00649903E+02])),
        #note='9/8/14')

#species(name='NEOC5H11',
        #atoms='C:5 H:11',
        #thermo=(NASA([300.00, 1375.00],
                    #[ 9.68072098E+00,  7.07718151E-03,  4.53301778E-05,
                    #-3.61952123E-08,  8.06654396E-12,  1.05116941E+03,
                    #-1.94358274E+01]),
                #NASA([1375.00, 5000.00],
                    #[ 1.17983312E+01,  3.59579648E-02, -1.44108662E-05,
                    #2.46307665E-09, -1.52118262E-13, -3.86101387E+03,
                    #-4.41297337E+01])),
        #note='9/8/14')

#species(name='NEOC5H11O2H',
        #atoms='C:5 H:12 O:2',
        #thermo=(NASA([300.00, 2019.00],
                    #[ 3.55225748E-01,  6.91131421E-02, -3.61242956E-05,
                    #6.97537323E-09, -9.54855674E-14, -3.23388889E+04,
                    #2.61109340E+01]),
                #NASA([2019.00, 5000.00],
                    #[ 2.01975284E+01,  2.97380281E-02, -1.09930813E-05,
                    #1.80338361E-09, -1.08851734E-13, -3.95165952E+04,
                    #-8.23753662E+01])),
        #note='9/8/14')

#species(name='NEOC5H11O2',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1682.00],
                    #[ 1.08144664E+00,  6.33534647E-02, -3.11938784E-05,
                    #4.88883786E-09,  2.48584189E-13, -1.56455031E+04,
                    #2.32475108E+01]),
                #NASA([1682.00, 5000.00],
                    #[ 1.93713105E+01,  2.82612571E-02, -1.04670110E-05,
                    #1.71917414E-09, -1.03854375E-13, -2.23544131E+04,
                    #-7.71662358E+01])),
        #note='9/8/14')

#species(name='NEOC5H11O',
        #atoms='C:5 H:11 O:1',
        #thermo=(NASA([300.00, 1374.00],
                    #[ 1.11399567E+01,  9.95551445E-03,  4.37547263E-05,
                    #-3.60663886E-08,  8.13722634E-12, -1.57739963E+04,
                    #-2.54644054E+01]),
                #NASA([1374.00, 5000.00],
                    #[ 1.47920426E+01,  3.60239861E-02, -1.44548775E-05,
                    #2.47285304E-09, -1.52828569E-13, -2.12880602E+04,
                    #-5.86605592E+01])),
        #note='9/8/14')

#species(name='NEOC5H10OOH',
        #atoms='C:5 H:11 O:2',
        #thermo=(NASA([300.00, 1686.00],
                    #[ 7.26388728E-01,  6.54763558E-02, -3.46433563E-05,
                    #6.82220142E-09, -1.22217205E-13, -7.61522056E+03,
                    #2.77840799E+01]),
                #NASA([1686.00, 5000.00],
                    #[ 1.98216577E+01,  2.74351153E-02, -1.01763800E-05,
                    #1.67311287E-09, -1.01142157E-13, -1.44846637E+04,
                    #-7.65499851E+01])),
        #note='9/8/14')

#species(name='NEO-C5H10O',
        #atoms='C:5 H:10 O:1',
        #thermo=(NASA([300.00, 1465.00],
                    #[-5.80125483E+00,  7.19986445E-02, -4.12060995E-05,
                    #8.52947622E-09,  2.15532151E-14, -1.85918283E+04,
                    #5.13449073E+01]),
                #NASA([1465.00, 5000.00],
                    #[ 1.88696363E+01,  2.18446364E-02, -7.48656898E-06,
                    #1.16574180E-09, -6.78305009E-14, -2.76538327E+04,
                    #-8.33848438E+01])),
        #note='9/8/14')

#species(name='NEOC5H10OOH-O2',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 2038.00],
                    #[ 2.12490773E+00,  7.51548100E-02, -4.39349794E-05,
                    #1.09212325E-08, -8.06703541E-13, -2.57458785E+04,
                    #2.47165260E+01]),
                #NASA([2038.00, 5000.00],
                    #[ 2.43460937E+01,  2.84840536E-02, -1.05961421E-05,
                    #1.74545495E-09, -1.05654571E-13, -3.35348323E+04,
                    #-9.58566184E+01])),
        #note='914')

#species(name='NEOC5H9Q2',
        #atoms='C:5 H:11 O:4',
        #thermo=(NASA([300.00, 2050.00],
                    #[ 1.90419491E+00,  7.68023849E-02, -4.67666506E-05,
                    #1.25175307E-08, -1.11211508E-12, -1.77393477E+04,
                    #2.75076242E+01]),
                #NASA([2050.00, 5000.00],
                    #[ 2.48112747E+01,  2.76461534E-02, -1.03017544E-05,
                    #1.69884890E-09, -1.02912689E-13, -2.56663214E+04,
                    #-9.64118626E+01])),
        #note='9/8/14')

#species(name='NEOC5KET',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 1.99299705E+00,  6.63799852E-02, -4.41434233E-05,
                    #1.47656228E-08, -1.97164356E-12, -4.31727732E+04,
                    #2.30090729E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 2.11457411E+01,  2.34263362E-02, -8.03514708E-06,
                    #1.24990645E-09, -7.26173329E-14, -4.99812693E+04,
                    #-8.04502172E+01])),
        #note='9/8/14')

#species(name='NEOC5KETOX',
        #atoms='C:5 H:9 O:2',
        #thermo=(NASA([300.00, 1677.00],
                    #[ 3.76958673E+00,  4.84037082E-02, -2.22215659E-05,
                    #2.98251470E-09,  2.98064950E-13, -2.52993970E+04,
                    #1.27918996E+01]),
                #NASA([1677.00, 5000.00],
                    #[ 1.62834087E+01,  2.47667618E-02, -8.89982283E-06,
                    #1.43325482E-09, -8.54232652E-14, -2.99100005E+04,
                    #-5.60198008E+01])),
        #note='9/8/14')

#species(name='NEOC5KEJOL',
        #atoms='C:5 H:9 O:2',
        #thermo=(NASA([300.00, 1402.00],
                    #[ 1.30298806E+00,  5.84132694E-02, -3.81961088E-05,
                    #1.24622441E-08, -1.59437417E-12, -3.28976593E+04,
                    #2.59154966E+01]),
                #NASA([1402.00, 5000.00],
                    #[ 1.78858167E+01,  2.12423369E-02, -7.21947714E-06,
                    #1.11609095E-09, -6.45625691E-14, -3.87680922E+04,
                    #-6.36233643E+01])),
        #note='9/8/14')

#species(name='NEOC5H9O-OOH',
        #atoms='C:5 H:10 O:3',
        #thermo=(NASA([300.00, 1429.00],
                    #[-4.33879286E+00,  8.23095013E-02, -5.24634537E-05,
                    #1.40859094E-08, -1.00626405E-12, -2.87630624E+04,
                    #5.04065995E+01]),
                #NASA([1429.00, 5000.00],
                    #[ 2.34301312E+01,  2.26473800E-02, -7.84730788E-06,
                    #1.22994662E-09, -7.18633905E-14, -3.86378493E+04,
                    #-1.00093171E+02])),
        #note='9/14')

#species(name='SC3H5OOH',
        #atoms='C:3 H:6 O:2',
        #thermo=(NASA([300.00, 1397.00],
                    #[ 1.32903007E+00,  4.49170722E-02, -3.51235127E-05,
                    #1.41982181E-08, -2.33335008E-12, -1.21898396E+04,
                    #2.02696565E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 1.43424294E+01,  1.28053632E-02, -4.40584813E-06,
                    #6.86848148E-10, -3.99675209E-14, -1.65261025E+04,
                    #-4.89934539E+01])),
        #note='9/8/14')

#species(name='C5H10OOH1-2O',
        #atoms='C:5 H:11 O:3',
        #thermo=(NASA([300.00, 1399.00],
                    #[ 2.27626580E+00,  7.41677471E-02, -5.37102267E-05,
                    #2.00876552E-08, -3.06562840E-12, -2.41478738E+04,
                    #2.31073065E+01]),
                #NASA([1399.00, 5000.00],
                    #[ 2.34744091E+01,  2.40289087E-02, -8.24957196E-06,
                    #1.28401429E-09, -7.46276890E-14, -3.14288055E+04,
                    #-9.04874276E+01])))

#species(name='C5H10OOH1-3O',
        #atoms='C:5 H:11 O:3',
        #thermo=(NASA([300.00, 1399.00],
                    #[ 2.27626580E+00,  7.41677471E-02, -5.37102267E-05,
                    #2.00876552E-08, -3.06562840E-12, -2.41478738E+04,
                    #2.31073065E+01]),
                #NASA([1399.00, 5000.00],
                    #[ 2.34744091E+01,  2.40289087E-02, -8.24957196E-06,
                    #1.28401429E-09, -7.46276890E-14, -3.14288055E+04,
                    #-9.04874276E+01])))

#species(name='C5H10OOH1-4O',
        #atoms='C:5 H:11 O:3',
        #thermo=(NASA([300.00, 1399.00],
                    #[ 2.27626580E+00,  7.41677471E-02, -5.37102267E-05,
                    #2.00876552E-08, -3.06562840E-12, -2.41478738E+04,
                    #2.31073065E+01]),
                #NASA([1399.00, 5000.00],
                    #[ 2.34744091E+01,  2.40289087E-02, -8.24957196E-06,
                    #1.28401429E-09, -7.46276890E-14, -3.14288055E+04,
                    #-9.04874276E+01])))

#species(name='C5H10OOH1-5O',
        #atoms='C:5 H:11 O:3',
        #thermo=(NASA([300.00, 1398.00],
                    #[ 3.33591252E+00,  6.71158698E-02, -4.18533324E-05,
                    #1.23681897E-08, -1.32406720E-12, -2.22125123E+04,
                    #1.96613241E+01]),
                #NASA([1398.00, 5000.00],
                    #[ 2.32107059E+01,  2.42653307E-02, -8.33521218E-06,
                    #1.29791364E-09, -7.54618456E-14, -2.94127692E+04,
                    #-8.82467525E+01])))

#species(name='C5H10OOH2-1O',
        #atoms='C:5 H:11 O:3',
        #thermo=(NASA([300.00, 1413.00],
                    #[ 3.59931598E+00,  6.97303823E-02, -4.81110105E-05,
                    #1.66633403E-08, -2.27605640E-12, -2.43831559E+04,
                    #1.64907485E+01]),
                #NASA([1413.00, 5000.00],
                    #[ 2.35346179E+01,  2.34722862E-02, -7.94776963E-06,
                    #1.22562706E-09, -7.07764677E-14, -3.12413772E+04,
                    #-9.05213665E+01])))

#species(name='C5H10OOH2-3O',
        #atoms='C:5 H:11 O:3',
        #thermo=(NASA([300.00, 1422.00],
                    #[ 3.52101067E+00,  7.27990609E-02, -5.45969132E-05,
                    #2.10575994E-08, -3.25724347E-12, -2.64753038E+04,
                    #1.53171528E+01]),
                #NASA([1422.00, 5000.00],
                    #[ 2.37039661E+01,  2.28944277E-02, -7.65396504E-06,
                    #1.17004168E-09, -6.71498252E-14, -3.30935488E+04,
                    #-9.19037151E+01])))

#species(name='C5H10OOH2-4O',
        #atoms='C:5 H:11 O:3',
        #thermo=(NASA([300.00, 1422.00],
                    #[ 3.52101067E+00,  7.27990609E-02, -5.45969132E-05,
                    #2.10575994E-08, -3.25724347E-12, -2.64753038E+04,
                    #1.53171528E+01]),
                #NASA([1422.00, 5000.00],
                    #[ 2.37039661E+01,  2.28944277E-02, -7.65396504E-06,
                    #1.17004168E-09, -6.71498252E-14, -3.30935488E+04,
                    #-9.19037151E+01])))

#species(name='C5H10OOH2-5O',
        #atoms='C:5 H:11 O:3',
        #thermo=(NASA([300.00, 1413.00],
                    #[ 3.59931598E+00,  6.97303823E-02, -4.81110105E-05,
                    #1.66633403E-08, -2.27605640E-12, -2.43831559E+04,
                    #1.64907485E+01]),
                #NASA([1413.00, 5000.00],
                    #[ 2.35346179E+01,  2.34722862E-02, -7.94776963E-06,
                    #1.22562706E-09, -7.07764677E-14, -3.12413772E+04,
                    #-9.05213665E+01])))

#species(name='C5H10OOH3-1O',
        #atoms='C:5 H:11 O:3',
        #thermo=(NASA([300.00, 1413.00],
                    #[ 3.59931598E+00,  6.97303823E-02, -4.81110105E-05,
                    #1.66633403E-08, -2.27605640E-12, -2.43831559E+04,
                    #1.64907485E+01]),
                #NASA([1413.00, 5000.00],
                    #[ 2.35346179E+01,  2.34722862E-02, -7.94776963E-06,
                    #1.22562706E-09, -7.07764677E-14, -3.12413772E+04,
                    #-9.05213665E+01])))

#species(name='C5H10OOH3-2O',
        #atoms='C:5 H:11 O:3',
        #thermo=(NASA([300.00, 1422.00],
                    #[ 3.52101067E+00,  7.27990609E-02, -5.45969132E-05,
                    #2.10575994E-08, -3.25724347E-12, -2.64753038E+04,
                    #1.53171528E+01]),
                #NASA([1422.00, 5000.00],
                    #[ 2.37039661E+01,  2.28944277E-02, -7.65396504E-06,
                    #1.17004168E-09, -6.71498252E-14, -3.30935488E+04,
                    #-9.19037151E+01])))

#species(name='NC6H14',
        #atoms='C:6 H:14',
        #thermo=(NASA([300.00, 1394.00],
                    #[-6.06787842E-01,  7.23956364E-02, -4.33845424E-05,
                    #1.28945357E-08, -1.49361322E-12, -2.28192378E+04,
                    #3.07154747E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 1.91649837E+01,  3.02733796E-02, -1.03172746E-05,
                    #1.59774518E-09, -9.25291178E-14, -3.01230801E+04,
                    #-7.69633109E+01])))

#species(name='C6H13-1',
        #atoms='C:6 H:13',
        #thermo=(NASA([300.00, 1395.00],
                    #[-2.12213030E-01,  6.87756030E-02, -4.21067325E-05,
                    #1.30061090E-08, -1.60770046E-12,  1.89801840E+03,
                    #3.18494469E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 1.85287088E+01,  2.83718554E-02, -9.68391797E-06,
                    #1.50116241E-09, -8.69954114E-14, -4.98606882E+03,
                    #-7.00604665E+01])))

#species(name='C6H13-2',
        #atoms='C:6 H:13',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 6.20068248E-01,  6.16195063E-02, -2.96980348E-05,
                    #4.69998266E-09,  3.13755723E-13,  3.23741699E+02,
                    #2.92321681E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 1.80640936E+01,  2.85111833E-02, -9.67861302E-06,
                    #1.49502045E-09, -8.64303710E-14, -6.45433301E+03,
                    #-6.70878929E+01])))

#species(name='C6H13-3',
        #atoms='C:6 H:13',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 6.20068248E-01,  6.16195063E-02, -2.96980348E-05,
                    #4.69998266E-09,  3.13755723E-13,  3.23741699E+02,
                    #2.92321681E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 1.80640936E+01,  2.85111833E-02, -9.67861302E-06,
                    #1.49502045E-09, -8.64303710E-14, -6.45433301E+03,
                    #-6.70878929E+01])))

#species(name='C6H12-1',
        #atoms='C:6 H:12',
        #thermo=(NASA([300.00, 1391.00],
                    #[-4.88101166E-01,  6.60223990E-02, -4.03430121E-05,
                    #1.23616901E-08, -1.50560345E-12, -7.51412714E+03,
                    #3.12755041E+01]),
                #NASA([1391.00, 5000.00],
                    #[ 1.77068656E+01,  2.69421757E-02, -9.19125439E-06,
                    #1.42442840E-09, -8.25377749E-14, -1.42085292E+04,
                    #-6.77121636E+01])))

#species(name='C6H12-2',
        #atoms='C:6 H:12',
        #thermo=(NASA([300.00, 1388.00],
                    #[ 2.64795074E-01,  6.14938008E-02, -3.36108501E-05,
                    #8.31993597E-09, -6.42021164E-13, -8.91785584E+03,
                    #2.74114997E+01]),
                #NASA([1388.00, 5000.00],
                    #[ 1.72811801E+01,  2.72154214E-02, -9.26717645E-06,
                    #1.43446313E-09, -8.30514585E-14, -1.54028154E+04,
                    #-6.59650838E+01])))

#species(name='C6H12-3',
        #atoms='C:6 H:12',
        #thermo=(NASA([300.00, 1386.00],
                    #[-1.72607040E-01,  6.19357715E-02, -3.37261047E-05,
                    #8.23624938E-09, -6.03434172E-13, -8.79082505E+03,
                    #2.90378440E+01]),
                #NASA([1386.00, 5000.00],
                    #[ 1.70313735E+01,  2.73270171E-02, -9.28504971E-06,
                    #1.43523149E-09, -8.30182955E-14, -1.53433694E+04,
                    #-6.53708588E+01])))

#species(name='C6H111-3',
        #atoms='C:6 H:11',
        #thermo=(NASA([300.00, 1391.00],
                    #[-7.91618996E-01,  6.42973572E-02, -3.98197537E-05,
                    #1.22637824E-08, -1.48498610E-12,  9.45786039E+03,
                    #3.15792511E+01]),
                #NASA([1391.00, 5000.00],
                    #[ 1.73040900E+01,  2.52533901E-02, -8.62981839E-06,
                    #1.33904230E-09, -7.76586750E-14,  2.83687143E+03,
                    #-6.67738819E+01])))

#species(name='C6H111-4',
        #atoms='C:6 H:11',
        #thermo=(NASA([300.00, 1393.00],
                    #[ 1.15589239E+00,  5.34977768E-02, -2.48120970E-05,
                    #3.38660024E-09,  4.19889741E-13,  1.55673894E+04,
                    #2.71720322E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 1.61989292E+01,  2.55342673E-02, -8.67483385E-06,
                    #1.34064100E-09, -7.75312952E-14,  9.65114345E+03,
                    #-5.61196642E+01])))

#species(name='C6H111-5',
        #atoms='C:6 H:11',
        #thermo=(NASA([300.00, 1393.00],
                    #[ 1.15589239E+00,  5.34977768E-02, -2.48120970E-05,
                    #3.38660024E-09,  4.19889741E-13,  1.55673894E+04,
                    #2.71720322E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 1.61989292E+01,  2.55342673E-02, -8.67483385E-06,
                    #1.34064100E-09, -7.75312952E-14,  9.65114345E+03,
                    #-5.61196642E+01])))

#species(name='C6H111-6',
        #atoms='C:6 H:11',
        #thermo=(NASA([300.00, 1387.00],
                    #[ 8.71879475E-02,  6.14727795E-02, -3.75054886E-05,
                    #1.14767733E-08, -1.41643272E-12,  1.71786596E+04,
                    #3.09015588E+01]),
                #NASA([1387.00, 5000.00],
                    #[ 1.74644753E+01,  2.47005987E-02, -8.53334148E-06,
                    #1.33367297E-09, -7.77343118E-14,  1.06887470E+04,
                    #-6.39032857E+01])))

#species(name='C6H112-4',
        #atoms='C:6 H:11',
        #thermo=(NASA([300.00, 1388.00],
                    #[-8.09214028E-02,  6.00167919E-02, -3.34681107E-05,
                    #8.44580243E-09, -6.67121954E-13,  8.05863014E+03,
                    #2.78967527E+01]),
                #NASA([1388.00, 5000.00],
                    #[ 1.68906239E+01,  2.55067554E-02, -8.69671971E-06,
                    #1.34743880E-09, -7.80675027E-14,  1.64435470E+03,
                    #-6.50840113E+01])))

#species(name='C6H112-5',
        #atoms='C:6 H:11',
        #thermo=(NASA([300.00, 1385.00],
                    #[ 2.08038945E+00,  4.83569471E-02, -1.72566989E-05,
                    #-1.13784525E-09,  1.38654068E-12,  1.41333417E+04,
                    #2.24803229E+01]),
                #NASA([1385.00, 5000.00],
                    #[ 1.57816863E+01,  2.57956633E-02, -8.74566769E-06,
                    #1.34977950E-09, -7.79887258E-14,  8.46300849E+03,
                    #-5.43988313E+01])))

#species(name='C6H112-6',
        #atoms='C:6 H:11',
        #thermo=(NASA([300.00, 1385.00],
                    #[ 6.38205633E-01,  5.78758020E-02, -3.22194721E-05,
                    #8.38982851E-09, -7.74542374E-13,  1.58047078E+04,
                    #2.79713260E+01]),
                #NASA([1385.00, 5000.00],
                    #[ 1.69591869E+01,  2.51413831E-02, -8.68761395E-06,
                    #1.35795747E-09, -7.91556719E-14,  9.51660897E+03,
                    #-6.17306897E+01])))

#species(name='C6H113-1',
        #atoms='C:6 H:11',
        #thermo=(NASA([300.00, 1387.00],
                    #[ 2.39410938E-01,  5.82991009E-02, -3.24715921E-05,
                    #8.38108128E-09, -7.27332655E-13,  1.59221147E+04,
                    #3.00781535E+01]),
                #NASA([1387.00, 5000.00],
                    #[ 1.63801476E+01,  2.54479755E-02, -8.66105290E-06,
                    #1.34023418E-09, -7.75806403E-14,  9.80356390E+03,
                    #-5.83757912E+01])))

#species(name='C6H12OH-1J2',
        #atoms='C:6 H:13 O:1',
        #thermo=(NASA([300.00, 1395.00],
                    #[ 1.52879690E+00,  7.02411418E-02, -4.39606425E-05,
                    #1.40750120E-08, -1.82860462E-12, -1.77206486E+04,
                    #2.72130211E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 2.05219095E+01,  2.85941986E-02, -9.71574333E-06,
                    #1.50154769E-09, -8.68352394E-14, -2.46176531E+04,
                    #-7.58039610E+01])))

#species(name='C6H12OH-2J1',
        #atoms='C:6 H:13 O:1',
        #thermo=(NASA([300.00, 1403.00],
                    #[ 1.12882665E+00,  7.79334290E-02, -5.78628264E-05,
                    #2.29540709E-08, -3.76055142E-12, -1.86725358E+04,
                    #2.64835374E+01]),
                #NASA([1403.00, 5000.00],
                    #[ 2.16778753E+01,  2.71851643E-02, -9.13650238E-06,
                    #1.40152530E-09, -8.06255019E-14, -2.55354913E+04,
                    #-8.29034809E+01])))

#species(name='C6H12OH-2J3',
        #atoms='C:6 H:13 O:1',
        #thermo=(NASA([300.00, 1402.00],
                    #[ 1.24040698E+00,  7.41986609E-02, -5.16292083E-05,
                    #1.90231170E-08, -2.89350308E-12, -1.97802722E+04,
                    #2.70192329E+01]),
                #NASA([1402.00, 5000.00],
                    #[ 2.08484871E+01,  2.76635202E-02, -9.25462670E-06,
                    #1.41532103E-09, -8.12482889E-14, -2.65002333E+04,
                    #-7.79996791E+01])))

#species(name='C6H12OH-3J2',
        #atoms='C:6 H:13 O:1',
        #thermo=(NASA([300.00, 1402.00],
                    #[ 1.24040698E+00,  7.41986609E-02, -5.16292083E-05,
                    #1.90231170E-08, -2.89350308E-12, -1.97802722E+04,
                    #2.70192329E+01]),
                #NASA([1402.00, 5000.00],
                    #[ 2.08484871E+01,  2.76635202E-02, -9.25462670E-06,
                    #1.41532103E-09, -8.12482889E-14, -2.65002333E+04,
                    #-7.79996791E+01])))

#species(name='C6H12OH-3J4',
        #atoms='C:6 H:13 O:1',
        #thermo=(NASA([300.00, 1402.00],
                    #[ 1.24040698E+00,  7.41986609E-02, -5.16292083E-05,
                    #1.90231170E-08, -2.89350308E-12, -1.97802722E+04,
                    #2.70192329E+01]),
                #NASA([1402.00, 5000.00],
                    #[ 2.08484871E+01,  2.76635202E-02, -9.25462670E-06,
                    #1.41532103E-09, -8.12482889E-14, -2.65002333E+04,
                    #-7.79996791E+01])))

#species(name='C6H12OH-1O2-2',
        #atoms='C:6 H:13 O:3',
        #thermo=(NASA([300.00, 1404.00],
                    #[ 1.21185034E+00,  8.86935209E-02, -6.52629776E-05,
                    #2.48084263E-08, -3.82705801E-12, -3.67836493E+04,
                    #3.06087511E+01]),
                #NASA([1404.00, 5000.00],
                    #[ 2.62416220E+01,  2.83873841E-02, -9.64269140E-06,
                    #1.49018659E-09, -8.61823559E-14, -4.52283711E+04,
                    #-1.03056223E+02])))

#species(name='C6H12OH-2O2-1',
        #atoms='C:6 H:13 O:3',
        #thermo=(NASA([300.00, 1402.00],
                    #[ 8.48833477E-01,  8.96057240E-02, -6.76377691E-05,
                    #2.68760659E-08, -4.38117112E-12, -3.67038364E+04,
                    #3.26746588E+01]),
                #NASA([1402.00, 5000.00],
                    #[ 2.54686767E+01,  2.90602674E-02, -9.87649038E-06,
                    #1.52670395E-09, -8.83047932E-14, -4.49414723E+04,
                    #-9.84561539E+01])))

#species(name='C6H12OH-2O2-3',
        #atoms='C:6 H:13 O:3',
        #thermo=(NASA([300.00, 1409.00],
                    #[ 1.25084224E+00,  9.16687695E-02, -7.30413883E-05,
                    #3.06072325E-08, -5.20205053E-12, -3.88977871E+04,
                    #2.88405166E+01]),
                #NASA([1409.00, 5000.00],
                    #[ 2.58464390E+01,  2.81918402E-02, -9.45730022E-06,
                    #1.44892549E-09, -8.32798272E-14, -4.67848660E+04,
                    #-1.01026245E+02])))

#species(name='C6H12OH-3O2-2',
        #atoms='C:6 H:13 O:3',
        #thermo=(NASA([300.00, 1409.00],
                    #[ 1.25084224E+00,  9.16687695E-02, -7.30413883E-05,
                    #3.06072325E-08, -5.20205053E-12, -3.88977871E+04,
                    #2.88405166E+01]),
                #NASA([1409.00, 5000.00],
                    #[ 2.58464390E+01,  2.81918402E-02, -9.45730022E-06,
                    #1.44892549E-09, -8.32798272E-14, -4.67848660E+04,
                    #-1.01026245E+02])))

#species(name='C6H12OH-3O2-4',
        #atoms='C:6 H:13 O:3',
        #thermo=(NASA([300.00, 1409.00],
                    #[ 1.25084224E+00,  9.16687695E-02, -7.30413883E-05,
                    #3.06072325E-08, -5.20205053E-12, -3.88977871E+04,
                    #2.88405166E+01]),
                #NASA([1409.00, 5000.00],
                    #[ 2.58464390E+01,  2.81918402E-02, -9.45730022E-06,
                    #1.44892549E-09, -8.32798272E-14, -4.67848660E+04,
                    #-1.01026245E+02])))

#species(name='C6H11-1D3O',
        #atoms='C:6 H:11 O:1',
        #thermo=(NASA([300.00, 1383.00],
                    #[ 2.10835922E+00,  6.53603616E-02, -4.08332113E-05,
                    #1.28408111E-08, -1.65145256E-12, -1.36559988E+03,
                    #2.12769181E+01]),
                #NASA([1383.00, 5000.00],
                    #[ 2.12331424E+01,  2.48369812E-02, -8.66311197E-06,
                    #1.36270774E-09, -7.97852110E-14, -8.51940477E+03,
                    #-8.30721992E+01])))

#species(name='C6H11-1D4O',
        #atoms='C:6 H:11 O:1',
        #thermo=(NASA([300.00, 1407.00],
                    #[ 1.74896641E+00,  6.59595125E-02, -4.28715185E-05,
                    #1.38805462E-08, -1.74690339E-12, -1.77738274E+03,
                    #2.24374197E+01]),
                #NASA([1407.00, 5000.00],
                    #[ 2.02215422E+01,  2.43935819E-02, -8.22353442E-06,
                    #1.26435910E-09, -7.28591420E-14, -8.27983812E+03,
                    #-7.72129963E+01])))

#species(name='C6H11-1D5O',
        #atoms='C:6 H:11 O:1',
        #thermo=(NASA([300.00, 1407.00],
                    #[ 1.74896641E+00,  6.59595125E-02, -4.28715185E-05,
                    #1.38805462E-08, -1.74690339E-12, -1.77738274E+03,
                    #2.24374197E+01]),
                #NASA([1407.00, 5000.00],
                    #[ 2.02215422E+01,  2.43935819E-02, -8.22353442E-06,
                    #1.26435910E-09, -7.28591420E-14, -8.27983812E+03,
                    #-7.72129963E+01])))

#species(name='C6H11-1D6O',
        #atoms='C:6 H:11 O:1',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 1.65674953E+00,  6.36239531E-02, -3.67332362E-05,
                    #9.43984310E-09, -7.26849847E-13,  3.36166059E+02,
                    #2.43747538E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 2.04129591E+01,  2.48384017E-02, -8.51026795E-06,
                    #1.32292012E-09, -7.68252415E-14, -6.63100699E+03,
                    #-7.80582419E+01])))

#species(name='C6H11-2D1O',
        #atoms='C:6 H:11 O:1',
        #thermo=(NASA([300.00, 1999.00],
                    #[ 6.40521825E+00,  4.23558507E-02, -7.63213993E-06,
                    #-5.93172496E-09,  2.04790110E-12, -1.07194029E+03,
                    #2.22105053E+00]),
                #NASA([1999.00, 5000.00],
                    #[ 1.63147059E+01,  3.08756722E-02, -1.12659286E-05,
                    #1.83193378E-09, -1.09891277E-13, -5.48521972E+03,
                    #-5.49439851E+01])))

#species(name='C6H11-2D4O',
        #atoms='C:6 H:11 O:1',
        #thermo=(NASA([300.00, 1380.00],
                    #[ 2.99930917E+00,  6.04876255E-02, -3.38261990E-05,
                    #8.74283850E-09, -7.93336016E-13, -2.79750804E+03,
                    #1.67172352E+01]),
                #NASA([1380.00, 5000.00],
                    #[ 2.07337615E+01,  2.52406898E-02, -8.79759458E-06,
                    #1.38318161E-09, -8.09557361E-14, -9.67476730E+03,
                    #-8.08901300E+01])))

#species(name='C6H11-2D5O',
        #atoms='C:6 H:11 O:1',
        #thermo=(NASA([300.00, 1403.00],
                    #[ 2.54000411E+00,  6.12908043E-02, -3.59176813E-05,
                    #9.70008242E-09, -8.53771601E-13, -3.18795419E+03,
                    #1.83891887E+01]),
                #NASA([1403.00, 5000.00],
                    #[ 1.98339795E+01,  2.46335527E-02, -8.28848420E-06,
                    #1.27278690E-09, -7.32851452E-14, -9.49287419E+03,
                    #-7.56901783E+01])))

#species(name='C6H11-2D6O',
        #atoms='C:6 H:11 O:1',
        #thermo=(NASA([300.00, 1390.00],
                    #[ 2.70082133E+00,  5.78441974E-02, -2.89980180E-05,
                    #5.19918235E-09,  1.16017596E-13, -1.10722562E+03,
                    #1.91870295E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 1.95292766E+01,  2.55039957E-02, -8.72047811E-06,
                    #1.35366557E-09, -7.85302933E-14, -7.61837294E+03,
                    #-7.36136387E+01])))

#species(name='C6H11-3D1O',
        #atoms='C:6 H:11 O:1',
        #thermo=(NASA([300.00, 1387.00],
                    #[ 2.34238003E+00,  5.79452100E-02, -2.86377646E-05,
                    #4.84570137E-09,  2.08901971E-13, -9.92276019E+02,
                    #2.11357667E+01]),
                #NASA([1387.00, 5000.00],
                    #[ 1.92747614E+01,  2.56242111E-02, -8.74228366E-06,
                    #1.35514342E-09, -7.85422000E-14, -7.55927229E+03,
                    #-7.23056917E+01])))

#species(name='C6H11-3D2O',
        #atoms='C:6 H:11 O:1',
        #thermo=(NASA([300.00, 1380.00],
                    #[ 2.54762399E+00,  6.07926451E-02, -3.36415232E-05,
                    #8.47505102E-09, -7.17131188E-13, -2.66243550E+03,
                    #1.91434934E+01]),
                #NASA([1380.00, 5000.00],
                    #[ 2.03827442E+01,  2.55217123E-02, -8.89072063E-06,
                    #1.39731901E-09, -8.17627196E-14, -9.59093507E+03,
                    #-7.90711932E+01])))

#species(name='C6H13OOH-1',
        #atoms='C:6 H:14 O:2',
        #thermo=(NASA([300.00, 1393.00],
                    #[ 3.76304221E-01,  8.42019657E-02, -5.55911124E-05,
                    #1.86319290E-08, -2.52819359E-12, -3.29059850E+04,
                    #3.21031523E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 2.45163599E+01,  3.04358125E-02, -1.04768961E-05,
                    #1.63358168E-09, -9.50618695E-14, -4.15677013E+04,
                    #-9.85011918E+01])))

#species(name='C6H13OOH-2',
        #atoms='C:6 H:14 O:2',
        #thermo=(NASA([300.00, 1401.00],
                    #[ 7.33867061E-01,  8.64149041E-02, -6.11174825E-05,
                    #2.24268392E-08, -3.36429856E-12, -3.50923807E+04,
                    #2.84824773E+01]),
                #NASA([1401.00, 5000.00],
                    #[ 2.48832119E+01,  2.96512073E-02, -1.01015279E-05,
                    #1.56407586E-09, -9.05729002E-14, -4.34204125E+04,
                    #-1.01048943E+02])))

#species(name='C6H13OOH-3',
        #atoms='C:6 H:14 O:2',
        #thermo=(NASA([300.00, 1401.00],
                    #[ 7.33867061E-01,  8.64149041E-02, -6.11174825E-05,
                    #2.24268392E-08, -3.36429856E-12, -3.50923807E+04,
                    #2.84824773E+01]),
                #NASA([1401.00, 5000.00],
                    #[ 2.48832119E+01,  2.96512073E-02, -1.01015279E-05,
                    #1.56407586E-09, -9.05729002E-14, -4.34204125E+04,
                    #-1.01048943E+02])))

#species(name='C6H13O2-1',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1392.00],
                    #[ 1.05761492E+00,  7.89345805E-02, -5.23116185E-05,
                    #1.79548180E-08, -2.54740591E-12, -1.62091180E+04,
                    #2.94104766E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 2.31785440E+01,  2.94255210E-02, -1.01218190E-05,
                    #1.57735564E-09, -9.17519123E-14, -2.41616714E+04,
                    #-9.02503944E+01])))

#species(name='C6H13O2-2',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1399.00],
                    #[ 1.59315499E+00,  8.03858598E-02, -5.68690473E-05,
                    #2.12270568E-08, -3.27992517E-12, -1.84223008E+04,
                    #2.49641944E+01]),
                #NASA([1399.00, 5000.00],
                    #[ 2.34641557E+01,  2.87030846E-02, -9.76607754E-06,
                    #1.51070014E-09, -8.74195382E-14, -2.59745219E+04,
                    #-9.23115240E+01])))

#species(name='C6H13O2-3',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1399.00],
                    #[ 1.59315499E+00,  8.03858598E-02, -5.68690473E-05,
                    #2.12270568E-08, -3.27992517E-12, -1.84223008E+04,
                    #2.49641944E+01]),
                #NASA([1399.00, 5000.00],
                    #[ 2.34641557E+01,  2.87030846E-02, -9.76607754E-06,
                    #1.51070014E-09, -8.74195382E-14, -2.59745219E+04,
                    #-9.23115240E+01])))

#species(name='C6H13O-1',
        #atoms='C:6 H:13 O:1',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 1.37476919E+00,  7.09707741E-02, -4.11061030E-05,
                    #1.06839948E-08, -8.50881624E-13, -1.49531106E+04,
                    #2.52025044E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 2.20928922E+01,  2.79119481E-02, -9.53349733E-06,
                    #1.47885354E-09, -8.57536878E-14, -2.26216369E+04,
                    #-8.78582845E+01])))

#species(name='C6H13O-2',
        #atoms='C:6 H:13 O:1',
        #thermo=(NASA([300.00, 1405.00],
                    #[ 1.53124348E+00,  7.23874176E-02, -4.54632890E-05,
                    #1.39938106E-08, -1.63272186E-12, -1.70585882E+04,
                    #2.30983559E+01]),
                #NASA([1405.00, 5000.00],
                    #[ 2.18292052E+01,  2.77008387E-02, -9.36457645E-06,
                    #1.44248495E-09, -8.32314882E-14, -2.43135941E+04,
                    #-8.67680797E+01])))

#species(name='C6H13O-3',
        #atoms='C:6 H:13 O:1',
        #thermo=(NASA([300.00, 1405.00],
                    #[ 1.53124348E+00,  7.23874176E-02, -4.54632890E-05,
                    #1.39938106E-08, -1.63272186E-12, -1.70585882E+04,
                    #2.30983559E+01]),
                #NASA([1405.00, 5000.00],
                    #[ 2.18292052E+01,  2.77008387E-02, -9.36457645E-06,
                    #1.44248495E-09, -8.32314882E-14, -2.43135941E+04,
                    #-8.67680797E+01])))

#species(name='C6H12OOH1-2',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1392.00],
                    #[ 2.35234877E+00,  7.42464378E-02, -4.64307066E-05,
                    #1.44953455E-08, -1.79518968E-12, -9.79913038E+03,
                    #2.49856776E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 2.33958970E+01,  2.87500915E-02, -9.87479228E-06,
                    #1.53738184E-09, -8.93686071E-14, -1.75002122E+04,
                    #-8.93745541E+01])))

#species(name='C6H12OOH1-3',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 1.92496561E+00,  7.24529009E-02, -4.20779710E-05,
                    #1.11647255E-08, -9.59696372E-13, -9.81550072E+03,
                    #2.83886795E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 2.27895146E+01,  2.89701974E-02, -9.88728294E-06,
                    #1.53286753E-09, -8.88486536E-14, -1.75428625E+04,
                    #-8.54563755E+01])))

#species(name='C6H12OOH1-4',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 1.92496561E+00,  7.24529009E-02, -4.20779710E-05,
                    #1.11647255E-08, -9.59696372E-13, -9.81550072E+03,
                    #2.83886795E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 2.27895146E+01,  2.89701974E-02, -9.88728294E-06,
                    #1.53286753E-09, -8.88486536E-14, -1.75428625E+04,
                    #-8.54563755E+01])))

#species(name='C6H12OOH1-5',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 1.92496561E+00,  7.24529009E-02, -4.20779710E-05,
                    #1.11647255E-08, -9.59696372E-13, -9.81550072E+03,
                    #2.83886795E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 2.27895146E+01,  2.89701974E-02, -9.88728294E-06,
                    #1.53286753E-09, -8.88486536E-14, -1.75428625E+04,
                    #-8.54563755E+01])))

#species(name='C6H12OOH2-1',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1401.00],
                    #[ 1.63351733E+00,  8.14796740E-02, -5.79995011E-05,
                    #2.13418964E-08, -3.19994195E-12, -1.05064447E+04,
                    #2.64803283E+01]),
                #NASA([1401.00, 5000.00],
                    #[ 2.45762848E+01,  2.74155943E-02, -9.34192750E-06,
                    #1.44672243E-09, -8.37892149E-14, -1.83935404E+04,
                    #-9.65123105E+01])))

#species(name='C6H12OOH2-3',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1400.00],
                    #[ 2.72663117E+00,  7.64651999E-02, -5.21833045E-05,
                    #1.85187539E-08, -2.69369864E-12, -1.19891933E+04,
                    #2.12761646E+01]),
                #NASA([1400.00, 5000.00],
                    #[ 2.36864065E+01,  2.80115990E-02, -9.51112192E-06,
                    #1.46925773E-09, -8.49419123E-14, -1.93109406E+04,
                    #-9.14557581E+01])))

#species(name='C6H12OOH2-4',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1408.00],
                    #[ 2.26007590E+00,  7.48358213E-02, -4.80375241E-05,
                    #1.52775741E-08, -1.87091068E-12, -1.19995231E+04,
                    #2.48618430E+01]),
                #NASA([1408.00, 5000.00],
                    #[ 2.31606108E+01,  2.80956105E-02, -9.46369942E-06,
                    #1.45419355E-09, -8.37638675E-14, -1.93834354E+04,
                    #-8.79850394E+01])))

#species(name='C6H12OOH2-5',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1408.00],
                    #[ 2.26007590E+00,  7.48358213E-02, -4.80375241E-05,
                    #1.52775741E-08, -1.87091068E-12, -1.19995231E+04,
                    #2.48618430E+01]),
                #NASA([1408.00, 5000.00],
                    #[ 2.31606108E+01,  2.80956105E-02, -9.46369942E-06,
                    #1.45419355E-09, -8.37638675E-14, -1.93834354E+04,
                    #-8.79850394E+01])))

#species(name='C6H12OOH2-6',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1402.00],
                    #[ 1.09455724E+00,  8.29967617E-02, -6.01254476E-05,
                    #2.26497954E-08, -3.48706819E-12, -1.03716865E+04,
                    #2.90715700E+01]),
                #NASA([1402.00, 5000.00],
                    #[ 2.43039070E+01,  2.76074809E-02, -9.39913372E-06,
                    #1.45469578E-09, -8.42141472E-14, -1.82854496E+04,
                    #-9.51106853E+01])))

#species(name='C6H12OOH3-1',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1402.00],
                    #[ 1.09455724E+00,  8.29967617E-02, -6.01254476E-05,
                    #2.26497954E-08, -3.48706819E-12, -1.03716865E+04,
                    #2.90715700E+01]),
                #NASA([1402.00, 5000.00],
                    #[ 2.43039070E+01,  2.76074809E-02, -9.39913372E-06,
                    #1.45469578E-09, -8.42141472E-14, -1.82854496E+04,
                    #-9.51106853E+01])))

#species(name='C6H12OOH3-2',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1400.00],
                    #[ 2.72663117E+00,  7.64651999E-02, -5.21833045E-05,
                    #1.85187539E-08, -2.69369864E-12, -1.19891933E+04,
                    #2.12761646E+01]),
                #NASA([1400.00, 5000.00],
                    #[ 2.36864065E+01,  2.80115990E-02, -9.51112192E-06,
                    #1.46925773E-09, -8.49419123E-14, -1.93109406E+04,
                    #-9.14557581E+01])))

#species(name='C6H12OOH3-4',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1400.00],
                    #[ 2.72663117E+00,  7.64651999E-02, -5.21833045E-05,
                    #1.85187539E-08, -2.69369864E-12, -1.19891933E+04,
                    #2.12761646E+01]),
                #NASA([1400.00, 5000.00],
                    #[ 2.36864065E+01,  2.80115990E-02, -9.51112192E-06,
                    #1.46925773E-09, -8.49419123E-14, -1.93109406E+04,
                    #-9.14557581E+01])))

#species(name='C6H12OOH3-5',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1408.00],
                    #[ 2.26007590E+00,  7.48358213E-02, -4.80375241E-05,
                    #1.52775741E-08, -1.87091068E-12, -1.19995231E+04,
                    #2.48618430E+01]),
                #NASA([1408.00, 5000.00],
                    #[ 2.31606108E+01,  2.80956105E-02, -9.46369942E-06,
                    #1.45419355E-09, -8.37638675E-14, -1.93834354E+04,
                    #-8.79850394E+01])))

#species(name='C6H12OOH3-6',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1402.00],
                    #[ 1.09455724E+00,  8.29967617E-02, -6.01254476E-05,
                    #2.26497954E-08, -3.48706819E-12, -1.03716865E+04,
                    #2.90715700E+01]),
                #NASA([1402.00, 5000.00],
                    #[ 2.43039070E+01,  2.76074809E-02, -9.39913372E-06,
                    #1.45469578E-09, -8.42141472E-14, -1.82854496E+04,
                    #-9.51106853E+01])))

#species(name='C6H12O1-2',
        #atoms='C:6 H:12 O:1',
        #thermo=(NASA([300.00, 1418.00],
                    #[-5.27479473E+00,  9.15832630E-02, -7.09194990E-05,
                    #2.79461610E-08, -4.39482049E-12, -2.05857903E+04,
                    #5.34224762E+01]),
                #NASA([1418.00, 5000.00],
                    #[ 2.09805167E+01,  2.61456832E-02, -8.79805046E-06,
                    #1.35121022E-09, -7.78107305E-14, -2.91366825E+04,
                    #-8.58639749E+01])))

#species(name='C6H12O1-3',
        #atoms='C:6 H:12 O:1',
        #thermo=(NASA([300.00, 1412.00],
                    #[-8.93042069E+00,  1.02982605E-01, -8.53638224E-05,
                    #3.63896486E-08, -6.21235185E-12, -2.07664357E+04,
                    #7.01497676E+01]),
                #NASA([1412.00, 5000.00],
                    #[ 1.99712630E+01,  2.75739130E-02, -9.32003684E-06,
                    #1.43550676E-09, -8.28258579E-14, -2.99062932E+04,
                    #-8.20867887E+01])))

#species(name='C6H12O1-4',
        #atoms='C:6 H:12 O:1',
        #thermo=(NASA([300.00, 1451.00],
                    #[-8.12435380E+00,  9.70640114E-02, -7.69405734E-05,
                    #3.10474493E-08, -4.95396033E-12, -3.08886808E+04,
                    #6.59704356E+01]),
                #NASA([1451.00, 5000.00],
                    #[ 1.87967925E+01,  2.75535527E-02, -9.00773471E-06,
                    #1.35601108E-09, -7.69845934E-14, -3.93246166E+04,
                    #-7.58383918E+01])))

#species(name='C6H12O1-5',
        #atoms='C:6 H:12 O:1',
        #thermo=(NASA([300.00, 1411.00],
                    #[-1.13294962E+01,  1.08590522E-01, -8.83887484E-05,
                    #3.67290337E-08, -6.10865223E-12, -3.24988594E+04,
                    #7.65557826E+01]),
                #NASA([1411.00, 5000.00],
                    #[ 1.99284959E+01,  2.86967410E-02, -9.76084126E-06,
                    #1.50996782E-09, -8.73930262E-14, -4.25348728E+04,
                    #-8.86574195E+01])))

#species(name='C6H12O2-3',
        #atoms='C:6 H:12 O:1',
        #thermo=(NASA([300.00, 1425.00],
                    #[-5.25994961E+00,  9.41384839E-02, -7.60798059E-05,
                    #3.12814710E-08, -5.11045005E-12, -2.26926336E+04,
                    #5.18211467E+01]),
                #NASA([1425.00, 5000.00],
                    #[ 2.14267419E+01,  2.53570884E-02, -8.43894210E-06,
                    #1.28635266E-09, -7.36861092E-14, -3.11276934E+04,
                    #-8.88992853E+01])))

#species(name='C6H12O2-4',
        #atoms='C:6 H:12 O:1',
        #thermo=(NASA([300.00, 1417.00],
                    #[-8.71729383E+00,  1.04644598E-01, -8.94323784E-05,
                    #3.91535754E-08, -6.81619305E-12, -2.29012647E+04,
                    #6.76413269E+01]),
                #NASA([1417.00, 5000.00],
                    #[ 2.02636052E+01,  2.68979324E-02, -8.99457950E-06,
                    #1.37527090E-09, -7.89420963E-14, -3.18212913E+04,
                    #-8.42025816E+01])))

#species(name='C6H12O2-5',
        #atoms='C:6 H:12 O:1',
        #thermo=(NASA([300.00, 1418.00],
                    #[-1.07496328E+01,  1.08968762E-01, -9.26948775E-05,
                    #4.02824223E-08, -6.95618656E-12, -3.25413826E+04,
                    #7.63557140E+01]),
                #NASA([1418.00, 5000.00],
                    #[ 1.96691261E+01,  2.77995312E-02, -9.30042882E-06,
                    #1.42255648E-09, -8.16795227E-14, -4.19335698E+04,
                    #-8.31528662E+01])))

#species(name='C6H12O3-4',
        #atoms='C:6 H:12 O:1',
        #thermo=(NASA([300.00, 1425.00],
                    #[-5.25994961E+00,  9.41384839E-02, -7.60798059E-05,
                    #3.12814710E-08, -5.11045005E-12, -2.26926336E+04,
                    #5.11316651E+01]),
                #NASA([1425.00, 5000.00],
                    #[ 2.14267419E+01,  2.53570884E-02, -8.43894210E-06,
                    #1.28635266E-09, -7.36861092E-14, -3.11276934E+04,
                    #-8.95887669E+01])))

#species(name='C6H12OOH1-2O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 2.62680845E+00,  8.94856642E-02, -5.98279275E-05,
                    #1.91011928E-08, -2.28528990E-12, -2.84709479E+04,
                    #2.58356924E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 3.04683025E+01,  2.77522708E-02, -9.61137583E-06,
                    #1.50516303E-09, -8.78680262E-14, -3.83461529E+04,
                    #-1.24655802E+02])))

#species(name='C6H12OOH1-3O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 2.62680845E+00,  8.94856642E-02, -5.98279275E-05,
                    #1.91011928E-08, -2.28528990E-12, -2.84709479E+04,
                    #2.58356924E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 3.04683025E+01,  2.77522708E-02, -9.61137583E-06,
                    #1.50516303E-09, -8.78680262E-14, -3.83461529E+04,
                    #-1.24655802E+02])))

#species(name='C6H12OOH1-4O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 2.62680845E+00,  8.94856642E-02, -5.98279275E-05,
                    #1.91011928E-08, -2.28528990E-12, -2.84709479E+04,
                    #2.58356924E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 3.04683025E+01,  2.77522708E-02, -9.61137583E-06,
                    #1.50516303E-09, -8.78680262E-14, -3.83461529E+04,
                    #-1.24655802E+02])))

#species(name='C6H12OOH1-5O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 2.62680845E+00,  8.94856642E-02, -5.98279275E-05,
                    #1.91011928E-08, -2.28528990E-12, -2.84709479E+04,
                    #2.58356924E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 3.04683025E+01,  2.77522708E-02, -9.61137583E-06,
                    #1.50516303E-09, -8.78680262E-14, -3.83461529E+04,
                    #-1.24655802E+02])))

#species(name='C6H12OOH2-1O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 2.62680845E+00,  8.94856642E-02, -5.98279275E-05,
                    #1.91011928E-08, -2.28528990E-12, -2.84709479E+04,
                    #2.58356924E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 3.04683025E+01,  2.77522708E-02, -9.61137583E-06,
                    #1.50516303E-09, -8.78680262E-14, -3.83461529E+04,
                    #-1.24655802E+02])))

#species(name='C6H12OOH2-3O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1403.00],
                    #[ 2.53306668E+00,  9.21554288E-02, -6.98231363E-05,
                    #2.75462625E-08, -4.42827896E-12, -3.05121822E+04,
                    #2.48102913E+01]),
                #NASA([1403.00, 5000.00],
                    #[ 2.83590321E+01,  2.87735842E-02, -9.78879765E-06,
                    #1.51428335E-09, -8.76360651E-14, -3.91336282E+04,
                    #-1.12736119E+02])))

#species(name='C6H12OOH2-4O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1403.00],
                    #[ 2.53306668E+00,  9.21554288E-02, -6.98231363E-05,
                    #2.75462625E-08, -4.42827896E-12, -3.05121822E+04,
                    #2.48102913E+01]),
                #NASA([1403.00, 5000.00],
                    #[ 2.83590321E+01,  2.87735842E-02, -9.78879765E-06,
                    #1.51428335E-09, -8.76360651E-14, -3.91336282E+04,
                    #-1.12736119E+02])))

#species(name='C6H12OOH2-5O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1403.00],
                    #[ 2.53306668E+00,  9.21554288E-02, -6.98231363E-05,
                    #2.75462625E-08, -4.42827896E-12, -3.05121822E+04,
                    #2.48102913E+01]),
                #NASA([1403.00, 5000.00],
                    #[ 2.83590321E+01,  2.87735842E-02, -9.78879765E-06,
                    #1.51428335E-09, -8.76360651E-14, -3.91336282E+04,
                    #-1.12736119E+02])))

#species(name='C6H12OOH2-6O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 2.62680845E+00,  8.94856642E-02, -5.98279275E-05,
                    #1.91011928E-08, -2.28528990E-12, -2.84709479E+04,
                    #2.58356924E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 3.04683025E+01,  2.77522708E-02, -9.61137583E-06,
                    #1.50516303E-09, -8.78680262E-14, -3.83461529E+04,
                    #-1.24655802E+02])))

#species(name='C6H12OOH3-1O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 2.62680845E+00,  8.94856642E-02, -5.98279275E-05,
                    #1.91011928E-08, -2.28528990E-12, -2.84709479E+04,
                    #2.58356924E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 3.04683025E+01,  2.77522708E-02, -9.61137583E-06,
                    #1.50516303E-09, -8.78680262E-14, -3.83461529E+04,
                    #-1.24655802E+02])))

#species(name='C6H12OOH3-2O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1403.00],
                    #[ 2.53306668E+00,  9.21554288E-02, -6.98231363E-05,
                    #2.75462625E-08, -4.42827896E-12, -3.05121822E+04,
                    #2.48102913E+01]),
                #NASA([1403.00, 5000.00],
                    #[ 2.83590321E+01,  2.87735842E-02, -9.78879765E-06,
                    #1.51428335E-09, -8.76360651E-14, -3.91336282E+04,
                    #-1.12736119E+02])))

#species(name='C6H12OOH3-4O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1403.00],
                    #[ 2.53306668E+00,  9.21554288E-02, -6.98231363E-05,
                    #2.75462625E-08, -4.42827896E-12, -3.05121822E+04,
                    #2.48102913E+01]),
                #NASA([1403.00, 5000.00],
                    #[ 2.83590321E+01,  2.87735842E-02, -9.78879765E-06,
                    #1.51428335E-09, -8.76360651E-14, -3.91336282E+04,
                    #-1.12736119E+02])))

#species(name='C6H12OOH3-5O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1403.00],
                    #[ 2.53306668E+00,  9.21554288E-02, -6.98231363E-05,
                    #2.75462625E-08, -4.42827896E-12, -3.05121822E+04,
                    #2.48102913E+01]),
                #NASA([1403.00, 5000.00],
                    #[ 2.83590321E+01,  2.87735842E-02, -9.78879765E-06,
                    #1.51428335E-09, -8.76360651E-14, -3.91336282E+04,
                    #-1.12736119E+02])))

#species(name='C6H12OOH3-6O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 2.62680845E+00,  8.94856642E-02, -5.98279275E-05,
                    #1.91011928E-08, -2.28528990E-12, -2.84709479E+04,
                    #2.58356924E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 3.04683025E+01,  2.77522708E-02, -9.61137583E-06,
                    #1.50516303E-09, -8.78680262E-14, -3.83461529E+04,
                    #-1.24655802E+02])))

#species(name='C6H11-1D3OOH',
        #atoms='C:6 H:12 O:2',
        #thermo=(NASA([300.00, 1393.00],
                    #[-2.31977696E+00,  9.52241599E-02, -7.73053497E-05,
                    #3.20512066E-08, -5.34750057E-12, -1.76191545E+04,
                    #4.22269905E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 2.59035094E+01,  2.42696112E-02, -8.29035467E-06,
                    #1.28689728E-09, -7.46830616E-14, -2.68476823E+04,
                    #-1.07452182E+02])))

#species(name='C6H11-1D4OOH',
        #atoms='C:6 H:12 O:2',
        #thermo=(NASA([300.00, 1401.00],
                    #[ 1.13416759E+00,  7.94193160E-02, -5.85260848E-05,
                    #2.26979408E-08, -3.62009744E-12, -1.98410684E+04,
                    #2.69474429E+01]),
                #NASA([1401.00, 5000.00],
                    #[ 2.29270511E+01,  2.66225664E-02, -9.05133554E-06,
                    #1.39947254E-09, -8.09578355E-14, -2.72150149E+04,
                    #-8.94129595E+01])))

#species(name='C6H11-1D5OOH',
        #atoms='C:6 H:12 O:2',
        #thermo=(NASA([300.00, 1401.00],
                    #[ 1.13416759E+00,  7.94193160E-02, -5.85260848E-05,
                    #2.26979408E-08, -3.62009744E-12, -1.98410684E+04,
                    #2.69474429E+01]),
                #NASA([1401.00, 5000.00],
                    #[ 2.29270511E+01,  2.66225664E-02, -9.05133554E-06,
                    #1.39947254E-09, -8.09578355E-14, -2.72150149E+04,
                    #-8.94129595E+01])))

#species(name='C6H11-1D6OOH',
        #atoms='C:6 H:12 O:2',
        #thermo=(NASA([300.00, 1393.00],
                    #[ 6.48701799E-01,  7.76335542E-02, -5.34143826E-05,
                    #1.90936716E-08, -2.82021629E-12, -1.76322212E+04,
                    #3.11812799E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 2.26383519E+01,  2.73757543E-02, -9.42397309E-06,
                    #1.46942591E-09, -8.55091451E-14, -2.54169255E+04,
                    #-8.73725472E+01])))

#species(name='C6H11-2D1OOH',
        #atoms='C:6 H:12 O:2',
        #thermo=(NASA([300.00, 1385.00],
                    #[ 2.11692928E-01,  8.23288253E-02, -5.97664416E-05,
                    #2.25592257E-08, -3.51959321E-12, -1.73187911E+04,
                    #3.31182384E+01]),
                #NASA([1385.00, 5000.00],
                    #[ 2.43428219E+01,  2.60142938E-02, -8.97949936E-06,
                    #1.40314527E-09, -8.17901261E-14, -2.57469168E+04,
                    #-9.65749831E+01])))

#species(name='C6H11-2D4OOH',
        #atoms='C:6 H:12 O:2',
        #thermo=(NASA([300.00, 1392.00],
                    #[-1.46245581E+00,  9.02938420E-02, -7.01570906E-05,
                    #2.78745585E-08, -4.47562215E-12, -1.90395758E+04,
                    #3.78704689E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 2.53272397E+01,  2.47729296E-02, -8.46596932E-06,
                    #1.31443483E-09, -7.62890280E-14, -2.79852809E+04,
                    #-1.04859678E+02])))

#species(name='C6H11-2D5OOH',
        #atoms='C:6 H:12 O:2',
        #thermo=(NASA([300.00, 1399.00],
                    #[ 1.81623661E+00,  7.52045876E-02, -5.22511183E-05,
                    #1.89442291E-08, -2.82183511E-12, -2.12341296E+04,
                    #2.34113083E+01]),
                #NASA([1399.00, 5000.00],
                    #[ 2.25048513E+01,  2.69114345E-02, -9.13654120E-06,
                    #1.41134534E-09, -8.15937722E-14, -2.84153206E+04,
                    #-8.76985785E+01])))

#species(name='C6H11-2D6OOH',
        #atoms='C:6 H:12 O:2',
        #thermo=(NASA([300.00, 1390.00],
                    #[ 1.51940041E+00,  7.26401571E-02, -4.60784955E-05,
                    #1.47295637E-08, -1.89409854E-12, -1.90551401E+04,
                    #2.67655882E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 2.21518946E+01,  2.77388788E-02, -9.53838077E-06,
                    #1.48617108E-09, -8.64396575E-14, -2.65848852E+04,
                    #-8.52712241E+01])))

#species(name='C6H11-3D1OOH',
        #atoms='C:6 H:12 O:2',
        #thermo=(NASA([300.00, 1390.00],
                    #[ 1.17006930E+00,  7.27400593E-02, -4.57945723E-05,
                    #1.44941072E-08, -1.84089055E-12, -1.89423975E+04,
                    #2.86651787E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 2.18169743E+01,  2.80152259E-02, -9.63205183E-06,
                    #1.50063123E-09, -8.72754508E-14, -2.65017535E+04,
                    #-8.35297001E+01])))

#species(name='C6H11-3D2OOH',
        #atoms='C:6 H:12 O:2',
        #thermo=(NASA([300.00, 1391.00],
                    #[-1.79725520E+00,  9.02776158E-02, -6.96160763E-05,
                    #2.74275414E-08, -4.36716320E-12, -1.89278837E+04,
                    #3.97122202E+01]),
                #NASA([1391.00, 5000.00],
                    #[ 2.50642842E+01,  2.49351789E-02, -8.51039657E-06,
                    #1.32031801E-09, -7.65936140E-14, -2.79315737E+04,
                    #-1.03524314E+02])))

#species(name='C6H11Q12-3',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 2.99684347E+00,  8.60938560E-02, -5.92978062E-05,
                    #2.08719630E-08, -2.98876240E-12, -2.18052539E+04,
                    #2.64874463E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 2.78943291E+01,  2.92847808E-02, -1.00847792E-05,
                    #1.57294057E-09, -9.15554523E-14, -3.05839622E+04,
                    #-1.07701979E+02])))

#species(name='C6H11Q12-4',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1409.00],
                    #[ 3.80954684E+00,  8.47207481E-02, -5.85351924E-05,
                    #2.03764845E-08, -2.81456479E-12, -2.21836150E+04,
                    #2.28354127E+01]),
                #NASA([1409.00, 5000.00],
                    #[ 2.80581380E+01,  2.85464861E-02, -9.69832158E-06,
                    #1.49895066E-09, -8.66960196E-14, -3.05545861E+04,
                    #-1.07396273E+02])))

#species(name='C6H11Q12-5',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1409.00],
                    #[ 3.80954684E+00,  8.47207481E-02, -5.85351924E-05,
                    #2.03764845E-08, -2.81456479E-12, -2.21836150E+04,
                    #2.28354127E+01]),
                #NASA([1409.00, 5000.00],
                    #[ 2.80581380E+01,  2.85464861E-02, -9.69832158E-06,
                    #1.49895066E-09, -8.66960196E-14, -3.05545861E+04,
                    #-1.07396273E+02])))

#species(name='C6H11Q12-6',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 2.20320858E+00,  9.20185664E-02, -6.38825199E-05,
                    #2.14499236E-08, -2.76271950E-12, -2.04338777E+04,
                    #2.95618135E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 3.09468859E+01,  2.69623754E-02, -9.34843940E-06,
                    #1.46514666E-09, -8.55805589E-14, -3.04867475E+04,
                    #-1.25320374E+02])))

#species(name='C6H11Q13-2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 2.99684347E+00,  8.60938560E-02, -5.92978062E-05,
                    #2.08719630E-08, -2.98876240E-12, -2.18052539E+04,
                    #2.64874463E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 2.78943291E+01,  2.92847808E-02, -1.00847792E-05,
                    #1.57294057E-09, -9.15554523E-14, -3.05839622E+04,
                    #-1.07701979E+02])))

#species(name='C6H11Q13-4',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 2.99684347E+00,  8.60938560E-02, -5.92978062E-05,
                    #2.08719630E-08, -2.98876240E-12, -2.18052539E+04,
                    #2.64874463E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 2.78943291E+01,  2.92847808E-02, -1.00847792E-05,
                    #1.57294057E-09, -9.15554523E-14, -3.05839622E+04,
                    #-1.07701979E+02])))

#species(name='C6H11Q13-5',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1409.00],
                    #[ 3.80954684E+00,  8.47207481E-02, -5.85351924E-05,
                    #2.03764845E-08, -2.81456479E-12, -2.21836150E+04,
                    #2.28354127E+01]),
                #NASA([1409.00, 5000.00],
                    #[ 2.80581380E+01,  2.85464861E-02, -9.69832158E-06,
                    #1.49895066E-09, -8.66960196E-14, -3.05545861E+04,
                    #-1.07396273E+02])))

#species(name='C6H11Q13-6',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 2.20320858E+00,  9.20185664E-02, -6.38825199E-05,
                    #2.14499236E-08, -2.76271950E-12, -2.04338777E+04,
                    #2.95618135E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 3.09468859E+01,  2.69623754E-02, -9.34843940E-06,
                    #1.46514666E-09, -8.55805589E-14, -3.04867475E+04,
                    #-1.25320374E+02])))

#species(name='C6H11Q14-2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 2.99684347E+00,  8.60938560E-02, -5.92978062E-05,
                    #2.08719630E-08, -2.98876240E-12, -2.18052539E+04,
                    #2.64874463E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 2.78943291E+01,  2.92847808E-02, -1.00847792E-05,
                    #1.57294057E-09, -9.15554523E-14, -3.05839622E+04,
                    #-1.07701979E+02])))

#species(name='C6H11Q14-3',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 2.99684347E+00,  8.60938560E-02, -5.92978062E-05,
                    #2.08719630E-08, -2.98876240E-12, -2.18052539E+04,
                    #2.64874463E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 2.78943291E+01,  2.92847808E-02, -1.00847792E-05,
                    #1.57294057E-09, -9.15554523E-14, -3.05839622E+04,
                    #-1.07701979E+02])))

#species(name='C6H11Q14-5',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 2.99684347E+00,  8.60938560E-02, -5.92978062E-05,
                    #2.08719630E-08, -2.98876240E-12, -2.18052539E+04,
                    #2.64874463E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 2.78943291E+01,  2.92847808E-02, -1.00847792E-05,
                    #1.57294057E-09, -9.15554523E-14, -3.05839622E+04,
                    #-1.07701979E+02])))

#species(name='C6H11Q14-6',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 2.20320858E+00,  9.20185664E-02, -6.38825199E-05,
                    #2.14499236E-08, -2.76271950E-12, -2.04338777E+04,
                    #2.95618135E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 3.09468859E+01,  2.69623754E-02, -9.34843940E-06,
                    #1.46514666E-09, -8.55805589E-14, -3.04867475E+04,
                    #-1.25320374E+02])))

#species(name='C6H11Q15-2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 2.99684347E+00,  8.60938560E-02, -5.92978062E-05,
                    #2.08719630E-08, -2.98876240E-12, -2.18052539E+04,
                    #2.64874463E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 2.78943291E+01,  2.92847808E-02, -1.00847792E-05,
                    #1.57294057E-09, -9.15554523E-14, -3.05839622E+04,
                    #-1.07701979E+02])))

#species(name='C6H11Q15-3',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1409.00],
                    #[ 3.80954684E+00,  8.47207481E-02, -5.85351924E-05,
                    #2.03764845E-08, -2.81456479E-12, -2.21836150E+04,
                    #2.28354127E+01]),
                #NASA([1409.00, 5000.00],
                    #[ 2.80581380E+01,  2.85464861E-02, -9.69832158E-06,
                    #1.49895066E-09, -8.66960196E-14, -3.05545861E+04,
                    #-1.07396273E+02])))

#species(name='C6H11Q15-4',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 2.99684347E+00,  8.60938560E-02, -5.92978062E-05,
                    #2.08719630E-08, -2.98876240E-12, -2.18052539E+04,
                    #2.64874463E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 2.78943291E+01,  2.92847808E-02, -1.00847792E-05,
                    #1.57294057E-09, -9.15554523E-14, -3.05839622E+04,
                    #-1.07701979E+02])))

#species(name='C6H11Q15-6',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 2.12435546E+00,  9.08091150E-02, -6.62026612E-05,
                    #2.49559670E-08, -3.84884889E-12, -2.03676921E+04,
                    #3.05566985E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 2.82883945E+01,  2.89369583E-02, -9.96190727E-06,
                    #1.55346138E-09, -9.04091590E-14, -2.93711746E+04,
                    #-1.09680591E+02])))

#species(name='C6H11Q23-1',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1402.00],
                    #[ 2.50274303E+00,  9.23699697E-02, -7.01978035E-05,
                    #2.76525328E-08, -4.43054874E-12, -2.25427916E+04,
                    #2.69579667E+01]),
                #NASA([1402.00, 5000.00],
                    #[ 2.87206101E+01,  2.81926003E-02, -9.62199118E-06,
                    #1.49176614E-09, -8.64680578E-14, -3.13085620E+04,
                    #-1.12729880E+02])))

#species(name='C6H11Q23-4',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1400.00],
                    #[ 3.37103808E+00,  8.76299651E-02, -6.32127378E-05,
                    #2.35062741E-08, -3.55550322E-12, -2.39785960E+04,
                    #2.29119049E+01]),
                #NASA([1400.00, 5000.00],
                    #[ 2.83307086E+01,  2.85436428E-02, -9.74746596E-06,
                    #1.51180692E-09, -8.76533346E-14, -3.25271405E+04,
                    #-1.10789659E+02])))

#species(name='C6H11Q23-5',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1421.00],
                    #[ 4.01277430E+00,  8.75230514E-02, -6.49830548E-05,
                    #2.47375697E-08, -3.77261053E-12, -2.43431110E+04,
                    #1.99545556E+01]),
                #NASA([1421.00, 5000.00],
                    #[ 2.84438571E+01,  2.76525438E-02, -9.26702496E-06,
                    #1.41900244E-09, -8.15353368E-14, -3.24103661E+04,
                    #-1.10029025E+02])))

#species(name='C6H11Q23-6',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1403.00],
                    #[ 2.07300290E+00,  9.43156884E-02, -7.32191610E-05,
                    #2.95085081E-08, -4.82851589E-12, -2.24527685E+04,
                    #2.88342959E+01]),
                #NASA([1403.00, 5000.00],
                    #[ 2.86807869E+01,  2.81353066E-02, -9.58211151E-06,
                    #1.48344887E-09, -8.58991774E-14, -3.12397030E+04,
                    #-1.12551544E+02])))

#species(name='C6H11Q24-1',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1402.00],
                    #[ 2.50274303E+00,  9.23699697E-02, -7.01978035E-05,
                    #2.76525328E-08, -4.43054874E-12, -2.25427916E+04,
                    #2.69579667E+01]),
                #NASA([1402.00, 5000.00],
                    #[ 2.87206101E+01,  2.81926003E-02, -9.62199118E-06,
                    #1.49176614E-09, -8.64680578E-14, -3.13085620E+04,
                    #-1.12729880E+02])))

#species(name='C6H11Q24-3',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1400.00],
                    #[ 3.37103808E+00,  8.76299651E-02, -6.32127378E-05,
                    #2.35062741E-08, -3.55550322E-12, -2.39785960E+04,
                    #2.29119049E+01]),
                #NASA([1400.00, 5000.00],
                    #[ 2.83307086E+01,  2.85436428E-02, -9.74746596E-06,
                    #1.51180692E-09, -8.76533346E-14, -3.25271405E+04,
                    #-1.10789659E+02])))

#species(name='C6H11Q24-5',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1400.00],
                    #[ 3.37103808E+00,  8.76299651E-02, -6.32127378E-05,
                    #2.35062741E-08, -3.55550322E-12, -2.39785960E+04,
                    #2.29119049E+01]),
                #NASA([1400.00, 5000.00],
                    #[ 2.83307086E+01,  2.85436428E-02, -9.74746596E-06,
                    #1.51180692E-09, -8.76533346E-14, -3.25271405E+04,
                    #-1.10789659E+02])))

#species(name='C6H11Q24-6',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1403.00],
                    #[ 2.07300290E+00,  9.43156884E-02, -7.32191610E-05,
                    #2.95085081E-08, -4.82851589E-12, -2.24527685E+04,
                    #2.88342959E+01]),
                #NASA([1403.00, 5000.00],
                    #[ 2.86807869E+01,  2.81353066E-02, -9.58211151E-06,
                    #1.48344887E-09, -8.58991774E-14, -3.12397030E+04,
                    #-1.12551544E+02])))

#species(name='C6H11Q25-1',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1402.00],
                    #[ 2.50274303E+00,  9.23699697E-02, -7.01978035E-05,
                    #2.76525328E-08, -4.43054874E-12, -2.25427916E+04,
                    #2.69579667E+01]),
                #NASA([1402.00, 5000.00],
                    #[ 2.87206101E+01,  2.81926003E-02, -9.62199118E-06,
                    #1.49176614E-09, -8.64680578E-14, -3.13085620E+04,
                    #-1.12729880E+02])))

#species(name='C6H11Q25-3',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1400.00],
                    #[ 3.37103808E+00,  8.76299651E-02, -6.32127378E-05,
                    #2.35062741E-08, -3.55550322E-12, -2.39785960E+04,
                    #2.29119049E+01]),
                #NASA([1400.00, 5000.00],
                    #[ 2.83307086E+01,  2.85436428E-02, -9.74746596E-06,
                    #1.51180692E-09, -8.76533346E-14, -3.25271405E+04,
                    #-1.10789659E+02])))

#species(name='C6H11Q34-1',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1403.00],
                    #[ 2.07300290E+00,  9.43156884E-02, -7.32191610E-05,
                    #2.95085081E-08, -4.82851589E-12, -2.24527685E+04,
                    #2.88342959E+01]),
                #NASA([1403.00, 5000.00],
                    #[ 2.86807869E+01,  2.81353066E-02, -9.58211151E-06,
                    #1.48344887E-09, -8.58991774E-14, -3.12397030E+04,
                    #-1.12551544E+02])))

#species(name='C6H11Q34-2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1400.00],
                    #[ 3.37103808E+00,  8.76299651E-02, -6.32127378E-05,
                    #2.35062741E-08, -3.55550322E-12, -2.39785960E+04,
                    #2.29119049E+01]),
                #NASA([1400.00, 5000.00],
                    #[ 2.83307086E+01,  2.85436428E-02, -9.74746596E-06,
                    #1.51180692E-09, -8.76533346E-14, -3.25271405E+04,
                    #-1.10789659E+02])))

#species(name='C6KET12',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1387.00],
                    #[-1.70742664E+00,  9.97939251E-02, -8.01901862E-05,
                    #3.25346440E-08, -5.30122784E-12, -4.30784665E+04,
                    #4.36695548E+01]),
                #NASA([1387.00, 5000.00],
                    #[ 2.90256185E+01,  2.40820489E-02, -8.30639705E-06,
                    #1.29822444E-09, -7.57124082E-14, -5.32782165E+04,
                    #-1.19864821E+02])))

#species(name='C6KET13',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1403.00],
                    #[ 2.70686236E+00,  7.86587312E-02, -5.28474860E-05,
                    #1.74402371E-08, -2.21798402E-12, -4.53342182E+04,
                    #2.21065759E+01]),
                #NASA([1403.00, 5000.00],
                    #[ 2.58825405E+01,  2.62029247E-02, -8.90539787E-06,
                    #1.37704650E-09, -7.96812443E-14, -5.34388150E+04,
                    #-1.02769588E+02])))

#species(name='C6KET14',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1403.00],
                    #[ 2.70686236E+00,  7.86587312E-02, -5.28474860E-05,
                    #1.74402371E-08, -2.21798402E-12, -4.53342182E+04,
                    #2.21065759E+01]),
                #NASA([1403.00, 5000.00],
                    #[ 2.58825405E+01,  2.62029247E-02, -8.90539787E-06,
                    #1.37704650E-09, -7.96812443E-14, -5.34388150E+04,
                    #-1.02769588E+02])))

#species(name='C6KET15',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1403.00],
                    #[ 2.70686236E+00,  7.86587312E-02, -5.28474860E-05,
                    #1.74402371E-08, -2.21798402E-12, -4.53342182E+04,
                    #2.21065759E+01]),
                #NASA([1403.00, 5000.00],
                    #[ 2.58825405E+01,  2.62029247E-02, -8.90539787E-06,
                    #1.37704650E-09, -7.96812443E-14, -5.34388150E+04,
                    #-1.02769588E+02])))

#species(name='C6KET21',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1389.00],
                    #[-5.14482555E-01,  9.21784846E-02, -6.94513152E-05,
                    #2.60396658E-08, -3.86537916E-12, -4.47938808E+04,
                    #3.91532333E+01]),
                #NASA([1389.00, 5000.00],
                    #[ 2.79607714E+01,  2.42379297E-02, -8.20343155E-06,
                    #1.26648645E-09, -7.32501537E-14, -5.43857540E+04,
                    #-1.13016914E+02])))

#species(name='C6KET23',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1389.00],
                    #[-8.86325259E-01,  9.56932379E-02, -7.66622031E-05,
                    #3.13003623E-08, -5.14052764E-12, -4.69292576E+04,
                    #3.98915047E+01]),
                #NASA([1389.00, 5000.00],
                    #[ 2.77518536E+01,  2.44228814E-02, -8.26691484E-06,
                    #1.27620900E-09, -7.38039402E-14, -5.63515030E+04,
                    #-1.12222879E+02])))

#species(name='C6KET24',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1410.00],
                    #[ 3.38656723E+00,  7.51931934E-02, -5.03122375E-05,
                    #1.68156568E-08, -2.18752991E-12, -4.91634551E+04,
                    #1.89837711E+01]),
                #NASA([1410.00, 5000.00],
                    #[ 2.45827447E+01,  2.65291227E-02, -8.85055847E-06,
                    #1.35140650E-09, -7.75076694E-14, -5.64887218E+04,
                    #-9.49507504E+01])))

#species(name='C6KET25',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1410.00],
                    #[ 3.38656723E+00,  7.51931934E-02, -5.03122375E-05,
                    #1.68156568E-08, -2.18752991E-12, -4.91634551E+04,
                    #1.89837711E+01]),
                #NASA([1410.00, 5000.00],
                    #[ 2.45827447E+01,  2.65291227E-02, -8.85055847E-06,
                    #1.35140650E-09, -7.75076694E-14, -5.64887218E+04,
                    #-9.49507504E+01])))

#species(name='C6KET26',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 2.91522244E+00,  7.34293670E-02, -4.53418223E-05,
                    #1.33776385E-08, -1.43669218E-12, -4.69588500E+04,
                    #2.31405529E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 2.41192811E+01,  2.75685546E-02, -9.34714604E-06,
                    #1.44293337E-09, -8.33930684E-14, -5.46176520E+04,
                    #-9.19195513E+01])))

#species(name='C6KET31',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1513.00],
                    #[ 3.68645338E+00,  6.79151683E-02, -3.49002426E-05,
                    #5.86626515E-09,  3.80680692E-13, -4.70773283E+04,
                    #1.94153944E+01]),
                #NASA([1513.00, 5000.00],
                    #[ 2.47427707E+01,  2.65724465E-02, -8.91511787E-06,
                    #1.36741583E-09, -7.87033305E-14, -5.49929590E+04,
                    #-9.61506620E+01])))

#species(name='C6KET32',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1388.00],
                    #[-3.40401334E-02,  9.00007504E-02, -6.64122175E-05,
                    #2.42781825E-08, -3.49817449E-12, -4.70631692E+04,
                    #3.57622287E+01]),
                #NASA([1388.00, 5000.00],
                    #[ 2.78409221E+01,  2.43352401E-02, -8.23613901E-06,
                    #1.27151170E-09, -7.35397855E-14, -5.65357668E+04,
                    #-1.13496792E+02])))

#species(name='C6KET34',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1388.00],
                    #[-3.40401334E-02,  9.00007504E-02, -6.64122175E-05,
                    #2.42781825E-08, -3.49817449E-12, -4.70631692E+04,
                    #3.57622287E+01]),
                #NASA([1388.00, 5000.00],
                    #[ 2.78409221E+01,  2.43352401E-02, -8.23613901E-06,
                    #1.27151170E-09, -7.35397855E-14, -5.65357668E+04,
                    #-1.13496792E+02])))

#species(name='C6KET35',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1422.00],
                    #[ 4.11000703E+00,  6.98568662E-02, -4.01291482E-05,
                    #9.49599829E-09, -4.21088500E-13, -4.92736613E+04,
                    #1.54877708E+01]),
                #NASA([1422.00, 5000.00],
                    #[ 2.52008823E+01,  2.55867980E-02, -8.44943107E-06,
                    #1.28209867E-09, -7.32384055E-14, -5.68781816E+04,
                    #-9.91884828E+01])))

#species(name='C6KET36',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1513.00],
                    #[ 3.68645338E+00,  6.79151683E-02, -3.49002426E-05,
                    #5.86626515E-09,  3.80680692E-13, -4.70773283E+04,
                    #1.94153944E+01]),
                #NASA([1513.00, 5000.00],
                    #[ 2.47427707E+01,  2.65724465E-02, -8.91511787E-06,
                    #1.36741583E-09, -7.87033305E-14, -5.49929590E+04,
                    #-9.61506620E+01])))

#species(name='C5Y2',
        #atoms='C:5 H:10 O:1',
        #thermo=(NASA([300.00, 1387.00],
                    #[ 2.22465879E+00,  4.82791863E-02, -2.21019858E-05,
                    #2.64029906E-09,  5.23720489E-13, -3.39379060E+04,
                    #1.94495002E+01]),
                #NASA([1387.00, 5000.00],
                    #[ 1.59632618E+01,  2.26394805E-02, -7.56430418E-06,
                    #1.15632032E-09, -6.63757297E-14, -3.92822369E+04,
                    #-5.65033421E+01])))

#species(name='C5Y2-1J',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1393.00],
                    #[ 2.14964962E+00,  5.09715972E-02, -2.97744827E-05,
                    #7.64275255E-09, -5.47753503E-13, -1.19838512E+04,
                    #1.91969246E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 1.72611127E+01,  1.91828746E-02, -6.41680319E-06,
                    #9.82027213E-10, -5.64270287E-14, -1.74829806E+04,
                    #-6.30397964E+01])))

#species(name='C5Y2-5J',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 2010.00],
                    #[ 2.03928898E+00,  4.76715898E-02, -2.60373643E-05,
                    #6.44110554E-09, -5.11322699E-13, -9.14243040E+03,
                    #2.25010591E+01]),
                #NASA([2010.00, 5000.00],
                    #[ 1.36121876E+01,  2.29083157E-02, -7.89568575E-06,
                    #1.23743112E-09, -7.24018569E-14, -1.31339181E+04,
                    #-4.00994585E+01])))

#species(name='C5Y3-1J',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1418.00],
                    #[ 3.36988365E+00,  3.93333507E-02, -1.08692381E-05,
                    #-4.28049004E-09,  2.08635748E-12, -9.33794632E+03,
                    #1.62475992E+01]),
                #NASA([1418.00, 5000.00],
                    #[ 1.58030158E+01,  2.00450814E-02, -6.64102020E-06,
                    #1.01015297E-09, -5.78091691E-14, -1.44839773E+04,
                    #-5.37408145E+01])))

#species(name='C5H91-1',
        #atoms='C:5 H:9',
        #thermo=(NASA([300.00, 1392.00],
                    #[ 5.53601698E-01,  5.07680916E-02, -3.27903964E-05,
                    #1.09672330E-08, -1.50701628E-12,  2.47118210E+04,
                    #2.52922332E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 1.43746526E+01,  1.99244772E-02, -6.75299634E-06,
                    #1.04202761E-09, -6.01991076E-14,  1.97494378E+04,
                    #-4.94774744E+01])))

#species(name='C5AL-4D',
        #atoms='C:5 H:8 O:1',
        #thermo=(NASA([300.00, 2020.00],
                    #[ 1.15814338E+00,  4.81746752E-02, -2.67363172E-05,
                    #6.50396537E-09, -4.78416928E-13, -1.47390215E+04,
                    #2.46792248E+01]),
                #NASA([2020.00, 5000.00],
                    #[ 1.37156197E+01,  2.16425024E-02, -7.67112142E-06,
                    #1.22483309E-09, -7.25906717E-14, -1.91110565E+04,
                    #-4.33806662E+01])))

#species(name='C5AL-2D',
        #atoms='C:5 H:8 O:1',
        #thermo=(NASA([300.00, 1390.00],
                    #[ 3.69688874E-04,  5.71382743E-02, -4.15287925E-05,
                    #1.58289787E-08, -2.49867087E-12, -1.87655817E+04,
                    #2.75869712E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 1.62588601E+01,  1.87787531E-02, -6.45805165E-06,
                    #1.00641261E-09, -5.85466414E-14, -2.44023233E+04,
                    #-5.96450353E+01])))

#species(name='C6H11O12-3OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1432.00],
                    #[-2.27599349E+00,  9.86141518E-02, -7.74723381E-05,
                    #3.01376954E-08, -4.58669108E-12, -3.31277592E+04,
                    #4.27007503E+01]),
                #NASA([1432.00, 5000.00],
                    #[ 2.74037429E+01,  2.47674765E-02, -8.29487594E-06,
                    #1.27052501E-09, -7.30489658E-14, -4.27151891E+04,
                    #-1.14645759E+02])))

#species(name='C6H11O12-4OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1432.00],
                    #[-2.27599349E+00,  9.86141518E-02, -7.74723381E-05,
                    #3.01376954E-08, -4.58669108E-12, -3.31277592E+04,
                    #4.27007503E+01]),
                #NASA([1432.00, 5000.00],
                    #[ 2.74037429E+01,  2.47674765E-02, -8.29487594E-06,
                    #1.27052501E-09, -7.30489658E-14, -4.27151891E+04,
                    #-1.14645759E+02])))

#species(name='C6H11O12-5OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1432.00],
                    #[-2.27599349E+00,  9.86141518E-02, -7.74723381E-05,
                    #3.01376954E-08, -4.58669108E-12, -3.31277592E+04,
                    #4.27007503E+01]),
                #NASA([1432.00, 5000.00],
                    #[ 2.74037429E+01,  2.47674765E-02, -8.29487594E-06,
                    #1.27052501E-09, -7.30489658E-14, -4.27151891E+04,
                    #-1.14645759E+02])))

#species(name='C6H11O12-6OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1418.00],
                    #[-2.95097578E+00,  9.82464841E-02, -7.59224537E-05,
                    #2.92820284E-08, -4.45657904E-12, -3.08990846E+04,
                    #4.77269227E+01]),
                #NASA([1418.00, 5000.00],
                    #[ 2.64398182E+01,  2.60261489E-02, -8.81968488E-06,
                    #1.36137474E-09, -7.86848040E-14, -4.05410765E+04,
                    #-1.08505405E+02])))

#species(name='C6H11O13-2OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1425.00],
                    #[-5.04452953E+00,  1.05918270E-01, -8.58242816E-05,
                    #3.47733637E-08, -5.55440543E-12, -3.34375726E+04,
                    #5.53559272E+01]),
                #NASA([1425.00, 5000.00],
                    #[ 2.63085971E+01,  2.61546948E-02, -8.77769854E-06,
                    #1.34612150E-09, -7.74534976E-14, -4.34234785E+04,
                    #-1.10295283E+02])))

#species(name='C6H11O13-4OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1425.00],
                    #[-5.04452953E+00,  1.05918270E-01, -8.58242816E-05,
                    #3.47733637E-08, -5.55440543E-12, -3.34375726E+04,
                    #5.53559272E+01]),
                #NASA([1425.00, 5000.00],
                    #[ 2.63085971E+01,  2.61546948E-02, -8.77769854E-06,
                    #1.34612150E-09, -7.74534976E-14, -4.34234785E+04,
                    #-1.10295283E+02])))

#species(name='C6H11O13-5OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1425.00],
                    #[-5.04452953E+00,  1.05918270E-01, -8.58242816E-05,
                    #3.47733637E-08, -5.55440543E-12, -3.34375726E+04,
                    #5.53559272E+01]),
                #NASA([1425.00, 5000.00],
                    #[ 2.63085971E+01,  2.61546948E-02, -8.77769854E-06,
                    #1.34612150E-09, -7.74534976E-14, -4.34234785E+04,
                    #-1.10295283E+02])))

#species(name='C6H11O13-6OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1416.00],
                    #[-5.60563964E+00,  1.04991651E-01, -8.32471205E-05,
                    #3.31747712E-08, -5.24343976E-12, -3.12256985E+04,
                    #5.98605002E+01]),
                #NASA([1416.00, 5000.00],
                    #[ 2.54629106E+01,  2.72623626E-02, -9.24125642E-06,
                    #1.42660936E-09, -8.24582230E-14, -4.12981961E+04,
                    #-1.04829559E+02])))

#species(name='C6H11O14-2OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1425.00],
                    #[-7.31154133E+00,  1.11367218E-01, -9.09451363E-05,
                    #3.71421722E-08, -5.98132077E-12, -4.30436125E+04,
                    #6.58337228E+01]),
                #NASA([1425.00, 5000.00],
                    #[ 2.56382200E+01,  2.71509666E-02, -9.12183288E-06,
                    #1.39988818E-09, -8.05862525E-14, -5.35027729E+04,
                    #-1.08120825E+02])))

#species(name='C6H11O14-3OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1425.00],
                    #[-7.31154133E+00,  1.11367218E-01, -9.09451363E-05,
                    #3.71421722E-08, -5.98132077E-12, -4.30436125E+04,
                    #6.58337228E+01]),
                #NASA([1425.00, 5000.00],
                    #[ 2.56382200E+01,  2.71509666E-02, -9.12183288E-06,
                    #1.39988818E-09, -8.05862525E-14, -5.35027729E+04,
                    #-1.08120825E+02])))

#species(name='C6H11O14-5OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1425.00],
                    #[-7.31154133E+00,  1.11367218E-01, -9.09451363E-05,
                    #3.71421722E-08, -5.98132077E-12, -4.30436125E+04,
                    #6.58337228E+01]),
                #NASA([1425.00, 5000.00],
                    #[ 2.56382200E+01,  2.71509666E-02, -9.12183288E-06,
                    #1.39988818E-09, -8.05862525E-14, -5.35027729E+04,
                    #-1.08120825E+02])))

#species(name='C6H11O14-6OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1412.00],
                    #[-9.39812599E+00,  1.17516430E-01, -9.95761400E-05,
                    #4.27858226E-08, -7.31024149E-12, -4.06059311E+04,
                    #7.73575995E+01]),
                #NASA([1412.00, 5000.00],
                    #[ 2.45656595E+01,  2.85283025E-02, -9.69282356E-06,
                    #1.49844211E-09, -8.66890301E-14, -5.12777917E+04,
                    #-1.01359408E+02])))

#species(name='C6H11O15-2OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1413.00],
                    #[-9.56356958E+00,  1.21160691E-01, -1.04208598E-04,
                    #4.51613897E-08, -7.75046859E-12, -4.48492256E+04,
                    #7.15665603E+01]),
                #NASA([1413.00, 5000.00],
                    #[ 2.57514330E+01,  2.79489406E-02, -9.49498364E-06,
                    #1.46787433E-09, -8.49251091E-14, -5.58563407E+04,
                    #-1.13983278E+02])))

#species(name='C6H11O15-3OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1413.00],
                    #[-9.56356958E+00,  1.21160691E-01, -1.04208598E-04,
                    #4.51613897E-08, -7.75046859E-12, -4.48492256E+04,
                    #7.15665603E+01]),
                #NASA([1413.00, 5000.00],
                    #[ 2.57514330E+01,  2.79489406E-02, -9.49498364E-06,
                    #1.46787433E-09, -8.49251091E-14, -5.58563407E+04,
                    #-1.13983278E+02])))

#species(name='C6H11O15-4OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1413.00],
                    #[-9.56356958E+00,  1.21160691E-01, -1.04208598E-04,
                    #4.51613897E-08, -7.75046859E-12, -4.48492256E+04,
                    #7.15665603E+01]),
                #NASA([1413.00, 5000.00],
                    #[ 2.57514330E+01,  2.79489406E-02, -9.49498364E-06,
                    #1.46787433E-09, -8.49251091E-14, -5.58563407E+04,
                    #-1.13983278E+02])))

#species(name='C6H11O15-6OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1410.00],
                    #[-9.86762792E+00,  1.19225986E-01, -1.00035576E-04,
                    #4.24450450E-08, -7.16631181E-12, -4.26812046E+04,
                    #7.48455212E+01]),
                #NASA([1410.00, 5000.00],
                    #[ 2.50812660E+01,  2.87541510E-02, -9.82374389E-06,
                    #1.52441144E-09, -8.84255093E-14, -5.37741881E+04,
                    #-1.09448133E+02])))

#species(name='C6H11O23-1OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1425.00],
                    #[-3.00051751E+00,  1.01052251E-01, -8.12852734E-05,
                    #3.26822710E-08, -5.17958864E-12, -3.29962026E+04,
                    #4.64312180E+01]),
                #NASA([1425.00, 5000.00],
                    #[ 2.69564039E+01,  2.52133315E-02, -8.46015924E-06,
                    #1.29728465E-09, -7.46387394E-14, -4.25735727E+04,
                    #-1.11973745E+02])))

#species(name='C6H11O23-4OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1443.00],
                    #[-2.32023655E+00,  1.01282059E-01, -8.24064664E-05,
                    #3.31693978E-08, -5.21408840E-12, -3.52234339E+04,
                    #4.13949391E+01]),
                #NASA([1443.00, 5000.00],
                    #[ 2.80583928E+01,  2.37928220E-02, -7.87310190E-06,
                    #1.19630476E-09, -6.84039421E-14, -4.48154981E+04,
                    #-1.18929909E+02])))

#species(name='C6H11O23-5OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1443.00],
                    #[-2.32023655E+00,  1.01282059E-01, -8.24064664E-05,
                    #3.31693978E-08, -5.21408840E-12, -3.52234339E+04,
                    #4.13949391E+01]),
                #NASA([1443.00, 5000.00],
                    #[ 2.80583928E+01,  2.37928220E-02, -7.87310190E-06,
                    #1.19630476E-09, -6.84039421E-14, -4.48154981E+04,
                    #-1.18929909E+02])))

#species(name='C6H11O23-6OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1425.00],
                    #[-3.00051751E+00,  1.01052251E-01, -8.12852734E-05,
                    #3.26822710E-08, -5.17958864E-12, -3.29962026E+04,
                    #4.64312180E+01]),
                #NASA([1425.00, 5000.00],
                    #[ 2.69564039E+01,  2.52133315E-02, -8.46015924E-06,
                    #1.29728465E-09, -7.46387394E-14, -4.25735727E+04,
                    #-1.11973745E+02])))

#species(name='C6H11O24-1OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1423.00],
                    #[-5.62254622E+00,  1.07701280E-01, -8.86273652E-05,
                    #3.66476695E-08, -5.99023124E-12, -3.33282520E+04,
                    #5.84076967E+01]),
                #NASA([1423.00, 5000.00],
                    #[ 2.59111942E+01,  2.64899573E-02, -8.89134224E-06,
                    #1.36353586E-09, -7.84507854E-14, -4.32925163E+04,
                    #-1.07879889E+02])))

#species(name='C6H11O24-3OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1433.00],
                    #[-5.05790795E+00,  1.08518606E-01, -9.08699191E-05,
                    #3.79673868E-08, -6.23050248E-12, -3.55387589E+04,
                    #5.38980265E+01]),
                #NASA([1433.00, 5000.00],
                    #[ 2.68522312E+01,  2.52818726E-02, -8.39099827E-06,
                    #1.27724274E-09, -7.31102872E-14, -4.54678687E+04,
                    #-1.13918662E+02])))

#species(name='C6H11O24-5OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1433.00],
                    #[-5.05790795E+00,  1.08518606E-01, -9.08699191E-05,
                    #3.79673868E-08, -6.23050248E-12, -3.55387589E+04,
                    #5.38980265E+01]),
                #NASA([1433.00, 5000.00],
                    #[ 2.68522312E+01,  2.52818726E-02, -8.39099827E-06,
                    #1.27724274E-09, -7.31102872E-14, -4.54678687E+04,
                    #-1.13918662E+02])))

#species(name='C6H11O24-6OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1423.00],
                    #[-5.62254622E+00,  1.07701280E-01, -8.86273652E-05,
                    #3.66476695E-08, -5.99023124E-12, -3.33282520E+04,
                    #5.84076967E+01]),
                #NASA([1423.00, 5000.00],
                    #[ 2.59111942E+01,  2.64899573E-02, -8.89134224E-06,
                    #1.36353586E-09, -7.84507854E-14, -4.32925163E+04,
                    #-1.07879889E+02])))

#species(name='C6H11O25-1OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1416.00],
                    #[-9.24035424E+00,  1.19425795E-01, -1.03974569E-04,
                    #4.57446356E-08, -7.95660813E-12, -4.27327015E+04,
                    #7.51088527E+01]),
                #NASA([1416.00, 5000.00],
                    #[ 2.48755287E+01,  2.78565097E-02, -9.37289923E-06,
                    #1.43948573E-09, -8.28962898E-14, -5.32048100E+04,
                    #-1.03584543E+02])))

#species(name='C6H11O25-3OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1421.00],
                    #[-8.92888836E+00,  1.22020603E-01, -1.09968258E-04,
                    #4.98119645E-08, -8.85746002E-12, -4.49187054E+04,
                    #7.16554572E+01]),
                #NASA([1421.00, 5000.00],
                    #[ 2.54725769E+01,  2.68960204E-02, -8.94476389E-06,
                    #1.36283531E-09, -7.80436306E-14, -5.51746673E+04,
                    #-1.07502255E+02])))

#species(name='C6H11O34-1OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1425.00],
                    #[-3.00051751E+00,  1.01052251E-01, -8.12852734E-05,
                    #3.26822710E-08, -5.17958864E-12, -3.29962026E+04,
                    #4.64312180E+01]),
                #NASA([1425.00, 5000.00],
                    #[ 2.69564039E+01,  2.52133315E-02, -8.46015924E-06,
                    #1.29728465E-09, -7.46387394E-14, -4.25735727E+04,
                    #-1.11973745E+02])))

#species(name='C6H11O34-2OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1443.00],
                    #[-2.32023655E+00,  1.01282059E-01, -8.24064664E-05,
                    #3.31693978E-08, -5.21408840E-12, -3.52234339E+04,
                    #4.13949391E+01]),
                #NASA([1443.00, 5000.00],
                    #[ 2.80583928E+01,  2.37928220E-02, -7.87310190E-06,
                    #1.19630476E-09, -6.84039421E-14, -4.48154981E+04,
                    #-1.18929909E+02])))

#species(name='PC4H9CHO',
        #atoms='C:5 H:10 O:1',
        #thermo=(NASA([300.00, 1383.00],
                    #[ 1.55168051E+00,  5.17580624E-02, -2.47521866E-05,
                    #3.39585733E-09,  4.54214964E-13, -3.01104924E+04,
                    #2.25340321E+01]),
                #NASA([1383.00, 5000.00],
                    #[ 1.71928366E+01,  2.24205079E-02, -7.66524563E-06,
                    #1.18999819E-09, -6.90481012E-14, -3.62048450E+04,
                    #-6.39265450E+01])))

#species(name='PC4H8CHO-1',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1391.00],
                    #[ 8.84469923E-01,  5.26232725E-02, -2.83373818E-05,
                    #5.66303879E-09, -3.63459101E-15, -9.89241254E+03,
                    #2.50250388E+01]),
                #NASA([1391.00, 5000.00],
                    #[ 1.73203841E+01,  2.00050584E-02, -6.88462599E-06,
                    #1.07357298E-09, -6.24869113E-14, -1.60879440E+04,
                    #-6.51443168E+01])))

#species(name='PC4H8CHO-2',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1426.00],
                    #[ 3.19129086E+00,  3.91487154E-02, -8.88573585E-06,
                    #-5.86340400E-09,  2.44961462E-12, -7.02633915E+03,
                    #1.84682536E+01]),
                #NASA([1426.00, 5000.00],
                    #[ 1.58456286E+01,  2.08430527E-02, -7.08562851E-06,
                    #1.09605491E-09, -6.34445999E-14, -1.24300179E+04,
                    #-5.32908403E+01])))

#species(name='PC4H8CHO-3',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1426.00],
                    #[ 3.19129086E+00,  3.91487154E-02, -8.88573585E-06,
                    #-5.86340400E-09,  2.44961462E-12, -7.02633915E+03,
                    #1.84682536E+01]),
                #NASA([1426.00, 5000.00],
                    #[ 1.58456286E+01,  2.08430527E-02, -7.08562851E-06,
                    #1.09605491E-09, -6.34445999E-14, -1.24300179E+04,
                    #-5.32908403E+01])))

#species(name='PC4H8CHO-4',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1383.00],
                    #[ 1.89962807E+00,  4.83598549E-02, -2.37381500E-05,
                    #3.58762168E-09,  3.40201608E-13, -5.38702187E+03,
                    #2.31840311E+01]),
                #NASA([1383.00, 5000.00],
                    #[ 1.66333143E+01,  2.03619734E-02, -6.95844731E-06,
                    #1.08002411E-09, -6.26591600E-14, -1.10824930E+04,
                    #-5.81167187E+01])))

#species(name='C4H7O12-4',
        #atoms='C:4 H:7 O:1',
        #thermo=(NASA([300.00, 1521.00],
                    #[-3.37974855E+00,  6.03280881E-02, -5.19900485E-05,
                    #2.23629697E-08, -3.72920327E-12,  9.80004298E+03,
                    #4.31033481E+01]),
                #NASA([1521.00, 5000.00],
                    #[ 1.34570917E+01,  1.40456552E-02, -4.31981858E-06,
                    #6.22563489E-10, -3.42463069E-14,  4.88696367E+03,
                    #-4.44520295E+01])))

#species(name='C4H7O13-4',
        #atoms='C:4 H:7 O:1',
        #thermo=(NASA([300.00, 1436.00],
                    #[-4.66267157E+00,  6.15207910E-02, -5.06948989E-05,
                    #2.11863240E-08, -3.49163560E-12,  9.24461326E+03,
                    #4.87364061E+01]),
                #NASA([1436.00, 5000.00],
                    #[ 1.26383103E+01,  1.59453761E-02, -5.13309123E-06,
                    #7.64898728E-10, -4.31253622E-14,  3.91648984E+03,
                    #-4.20740863E+01])))

#species(name='C4H7O23-1',
        #atoms='C:4 H:7 O:1',
        #thermo=(NASA([300.00, 1324.00],
                    #[-3.45625302E+00,  6.17954277E-02, -5.36422388E-05,
                    #2.24943234E-08, -3.70306875E-12,  7.74377568E+03,
                    #4.22227921E+01]),
                #NASA([1324.00, 5000.00],
                    #[ 8.53917898E+00,  1.94224428E-02, -6.06403336E-06,
                    #6.79865833E-10, -1.50192364E-14,  5.95005552E+03,
                    #-1.55618089E+01])))

#species(name='C5H9O12-5',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1428.00],
                    #[-4.74570948E+00,  7.55874963E-02, -6.08534529E-05,
                    #2.46895048E-08, -3.96232760E-12,  7.08653513E+03,
                    #5.15668737E+01]),
                #NASA([1428.00, 5000.00],
                    #[ 1.71692866E+01,  1.96439921E-02, -6.56270756E-06,
                    #1.00313537E-09, -5.75797826E-14,  1.20372734E+02,
                    #-6.41594420E+01])))

#species(name='C5H9O13-5',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1458.00],
                    #[-6.25393631E+00,  7.92184870E-02, -6.70994450E-05,
                    #2.87608390E-08, -4.83752138E-12,  6.54668516E+03,
                    #5.80476602E+01]),
                #NASA([1458.00, 5000.00],
                    #[ 1.55990028E+01,  2.00371413E-02, -6.41154602E-06,
                    #9.50807571E-10, -5.34042125E-14,  3.37602077E+00,
                    #-5.60383294E+01])))

#species(name='C5H9O14-5',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1447.00],
                    #[-7.49219030E+00,  8.07085780E-02, -6.60336535E-05,
                    #2.72423613E-08, -4.41750289E-12, -3.23749643E+03,
                    #6.36020351E+01]),
                #NASA([1447.00, 5000.00],
                    #[ 1.54610559E+01,  2.06809909E-02, -6.66141956E-06,
                    #9.93099401E-10, -5.60118090E-14, -1.03327057E+04,
                    #-5.70023035E+01])))

#species(name='C5H9O23-1',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1424.00],
                    #[-4.67636252E+00,  7.74809264E-02, -6.42244373E-05,
                    #2.68440080E-08, -4.43868315E-12,  4.97926865E+03,
                    #4.97877901E+01]),
                #NASA([1424.00, 5000.00],
                    #[ 1.77401551E+01,  1.92298085E-02, -6.43473125E-06,
                    #9.84654362E-10, -5.65624830E-14, -2.05448135E+03,
                    #-6.82408862E+01])))

#species(name='C5H9O23-5',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1424.00],
                    #[-4.67636252E+00,  7.74809264E-02, -6.42244373E-05,
                    #2.68440080E-08, -4.43868315E-12,  4.97926865E+03,
                    #4.97877901E+01]),
                #NASA([1424.00, 5000.00],
                    #[ 1.77401551E+01,  1.92298085E-02, -6.43473125E-06,
                    #9.84654362E-10, -5.65624830E-14, -2.05448135E+03,
                    #-6.82408862E+01])))

#species(name='C5H9O24-1',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1398.00],
                    #[-6.45048964E+00,  8.25684509E-02, -7.30635550E-05,
                    #3.23767886E-08, -5.72480023E-12,  4.47417427E+03,
                    #5.74524602E+01]),
                #NASA([1398.00, 5000.00],
                    #[ 1.40157807E+01,  2.51554423E-02, -1.11149003E-05,
                    #2.14885206E-09, -1.44559627E-13, -1.54769089E+03,
                    #-4.88521063E+01])))

#species(name='C3H6CHO-3O',
        #atoms='C:4 H:7 O:2',
        #thermo=(NASA([300.00, 1417.00],
                    #[ 3.94683511E+00,  3.65984570E-02, -1.24268392E-05,
                    #-2.56058969E-09,  1.66288190E-12, -2.13506683E+04,
                    #1.18427218E+01]),
                #NASA([1417.00, 5000.00],
                    #[ 1.64998473E+01,  1.58390093E-02, -5.41015815E-06,
                    #8.39905247E-10, -4.87505919E-14, -2.64232542E+04,
                    #-5.83737300E+01])))

#species(name='AC3H5CHO',
        #atoms='C:4 H:6 O:1',
        #thermo=(NASA([300.00, 1387.00],
                    #[-1.35627837E-01,  4.32102800E-02, -2.86018998E-05,
                    #9.48372797E-09, -1.26553781E-12, -1.08101418E+04,
                    #2.93765145E+01]),
                #NASA([1387.00, 5000.00],
                    #[ 1.27609075E+01,  1.48416812E-02, -5.16797249E-06,
                    #8.12052956E-10, -4.75111888E-14, -1.54802190E+04,
                    #-4.05362549E+01])))

#species(name='PC4H9CO',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1380.00],
                    #[ 2.32674439E+00,  4.69567990E-02, -2.17002843E-05,
                    #2.33166983E-09,  6.15109331E-13, -1.15824032E+04,
                    #1.99866279E+01]),
                #NASA([1380.00, 5000.00],
                    #[ 1.67761337E+01,  2.02898448E-02, -6.94418179E-06,
                    #1.07888297E-09, -6.26360391E-14, -1.72426942E+04,
                    #-6.00185187E+01])))

#species(name='C6KET12O',
        #atoms='C:6 H:11 O:2',
        #thermo=(NASA([300.00, 1380.00],
                    #[ 2.48709886E+00,  6.84851704E-02, -3.91082619E-05,
                    #9.26081053E-09, -5.01384589E-13, -2.64222925E+04,
                    #2.25084325E+01]),
                #NASA([1380.00, 5000.00],
                    #[ 2.44011335E+01,  2.43695802E-02, -8.52760748E-06,
                    #1.34452656E-09, -7.88560069E-14, -3.46915506E+04,
                    #-9.76141351E+01])))

#species(name='C6KET13O',
        #atoms='C:6 H:11 O:2',
        #thermo=(NASA([300.00, 1420.00],
                    #[ 3.39465791E+00,  6.34450550E-02, -3.11900396E-05,
                    #3.70312592E-09,  8.80779018E-13, -2.72582338E+04,
                    #1.74777979E+01]),
                #NASA([1420.00, 5000.00],
                    #[ 2.43086453E+01,  2.35169486E-02, -8.03437802E-06,
                    #1.24733142E-09, -7.23954247E-14, -3.52041500E+04,
                    #-9.76594163E+01])))

#species(name='C6KET14O',
        #atoms='C:6 H:11 O:2',
        #thermo=(NASA([300.00, 1420.00],
                    #[ 3.39465791E+00,  6.34450550E-02, -3.11900396E-05,
                    #3.70312592E-09,  8.80779018E-13, -2.72582338E+04,
                    #1.74777979E+01]),
                #NASA([1420.00, 5000.00],
                    #[ 2.43086453E+01,  2.35169486E-02, -8.03437802E-06,
                    #1.24733142E-09, -7.23954247E-14, -3.52041500E+04,
                    #-9.76594163E+01])))

#species(name='C6KET15O',
        #atoms='C:6 H:11 O:2',
        #thermo=(NASA([300.00, 1420.00],
                    #[ 3.39465791E+00,  6.34450550E-02, -3.11900396E-05,
                    #3.70312592E-09,  8.80779018E-13, -2.72582338E+04,
                    #1.74777979E+01]),
                #NASA([1420.00, 5000.00],
                    #[ 2.43086453E+01,  2.35169486E-02, -8.03437802E-06,
                    #1.24733142E-09, -7.23954247E-14, -3.52041500E+04,
                    #-9.76594163E+01])))

#species(name='C6KET21O',
        #atoms='C:6 H:11 O:2',
        #thermo=(NASA([300.00, 1487.00],
                    #[ 7.39116829E+00,  4.44580281E-02, -6.38228035E-06,
                    #-9.61459324E-09,  3.45678863E-12, -2.86152512E+04,
                    #1.28093150E-01]),
                #NASA([1487.00, 5000.00],
                    #[ 2.25293622E+01,  2.47975925E-02, -8.43579282E-06,
                    #1.30658010E-09, -7.57309929E-14, -3.53514289E+04,
                    #-8.65954633E+01])))

#species(name='C6KET23O',
        #atoms='C:6 H:11 O:2',
        #thermo=(NASA([300.00, 1386.00],
                    #[ 3.25181786E+00,  6.46413698E-02, -3.58852109E-05,
                    #8.30469730E-09, -4.25746250E-13, -3.02655065E+04,
                    #1.89862996E+01]),
                #NASA([1386.00, 5000.00],
                    #[ 2.29408073E+01,  2.51899756E-02, -8.71525786E-06,
                    #1.36369341E-09, -7.95562705E-14, -3.77254587E+04,
                    #-8.90236293E+01])))

#species(name='C6KET24O',
        #atoms='C:6 H:11 O:2',
        #thermo=(NASA([300.00, 1530.00],
                    #[ 4.41342175E+00,  5.98412718E-02, -3.19612065E-05,
                    #6.55909263E-09, -5.07843738E-14, -3.11596284E+04,
                    #1.25825004E+01]),
                #NASA([1530.00, 5000.00],
                    #[ 2.17184075E+01,  2.47007497E-02, -8.21787847E-06,
                    #1.25253279E-09, -7.17484997E-14, -3.75473266E+04,
                    #-8.19767338E+01])))

#species(name='C6KET25O',
        #atoms='C:6 H:11 O:2',
        #thermo=(NASA([300.00, 1530.00],
                    #[ 4.41342175E+00,  5.98412718E-02, -3.19612065E-05,
                    #6.55909263E-09, -5.07843738E-14, -3.11596284E+04,
                    #1.25825004E+01]),
                #NASA([1530.00, 5000.00],
                    #[ 2.17184075E+01,  2.47007497E-02, -8.21787847E-06,
                    #1.25253279E-09, -7.17484997E-14, -3.75473266E+04,
                    #-8.19767338E+01])))

#species(name='C6KET26O',
        #atoms='C:6 H:11 O:2',
        #thermo=(NASA([300.00, 1527.00],
                    #[ 4.01263738E+00,  5.66756784E-02, -2.38571625E-05,
                    #8.77122160E-10,  1.22542395E-12, -2.89322231E+04,
                    #1.64535234E+01]),
                #NASA([1527.00, 5000.00],
                    #[ 2.13007803E+01,  2.58716731E-02, -8.79653453E-06,
                    #1.36062148E-09, -7.87479444E-14, -3.57843109E+04,
                    #-7.96089891E+01])))

#species(name='C6KET31O',
        #atoms='C:6 H:11 O:2',
        #thermo=(NASA([300.00, 1305.00],
                    #[ 5.02858055E+00,  5.12156892E-02, -1.40537055E-05,
                    #-6.12620240E-09,  2.92422949E-12, -2.91213820E+04,
                    #1.13420391E+01]),
                #NASA([1305.00, 5000.00],
                    #[ 2.20864552E+01,  2.46779992E-02, -8.28346973E-06,
                    #1.27113585E-09, -7.31921565E-14, -3.61540517E+04,
                    #-8.46188998E+01])))

#species(name='C6KET32O',
        #atoms='C:6 H:11 O:2',
        #thermo=(NASA([300.00, 1381.00],
                    #[ 4.28777527E+00,  5.83079361E-02, -2.63826419E-05,
                    #2.44071634E-09,  8.67814900E-13, -3.04212126E+04,
                    #1.40253081E+01]),
                #NASA([1381.00, 5000.00],
                    #[ 2.23075626E+01,  2.54572934E-02, -8.74888878E-06,
                    #1.36294097E-09, -7.92726719E-14, -3.75305099E+04,
                    #-8.59089550E+01])))

#species(name='C6KET34O',
        #atoms='C:6 H:11 O:2',
        #thermo=(NASA([300.00, 1381.00],
                    #[ 4.28777527E+00,  5.83079361E-02, -2.63826419E-05,
                    #2.44071634E-09,  8.67814900E-13, -3.04212126E+04,
                    #1.40253081E+01]),
                #NASA([1381.00, 5000.00],
                    #[ 2.23075626E+01,  2.54572934E-02, -8.74888878E-06,
                    #1.36294097E-09, -7.92726719E-14, -3.75305099E+04,
                    #-8.59089550E+01])))

#species(name='C6KET35O',
        #atoms='C:6 H:11 O:2',
        #thermo=(NASA([300.00, 1416.00],
                    #[ 4.84610323E+00,  5.24943003E-02, -1.55391890E-05,
                    #-5.44987122E-09,  2.81073621E-12, -3.11401488E+04,
                    #1.11045620E+01]),
                #NASA([1416.00, 5000.00],
                    #[ 2.26318948E+01,  2.41684263E-02, -8.10341688E-06,
                    #1.24303233E-09, -7.15692685E-14, -3.84096424E+04,
                    #-8.87241802E+01])))

#species(name='C6KET36O',
        #atoms='C:6 H:11 O:2',
        #thermo=(NASA([300.00, 1305.00],
                    #[ 5.02858055E+00,  5.12156892E-02, -1.40537055E-05,
                    #-6.12620240E-09,  2.92422949E-12, -2.91213820E+04,
                    #1.13420391E+01]),
                #NASA([1305.00, 5000.00],
                    #[ 2.20864552E+01,  2.46779992E-02, -8.28346973E-06,
                    #1.27113585E-09, -7.31921565E-14, -3.61540517E+04,
                    #-8.46188998E+01])))

#species(name='C6H10D13',
        #atoms='C:6 H:10',
        #thermo=(NASA([300.00, 1395.00],
                    #[-1.48701712E+00,  6.83526916E-02, -5.17682192E-05,
                    #2.10789514E-08, -3.57266133E-12,  4.56149351E+03,
                    #3.34156371E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 1.69088929E+01,  2.30881104E-02, -7.89077821E-06,
                    #1.22430123E-09, -7.09963130E-14, -1.66488681E+03,
                    #-6.46806187E+01])))

#species(name='C6H10D24',
        #atoms='C:6 H:10',
        #thermo=(NASA([300.00, 1393.00],
                    #[-3.63783422E-01,  6.36719104E-02, -4.53090160E-05,
                    #1.73368393E-08, -2.79133016E-12,  3.04087286E+03,
                    #2.75469470E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 1.67662657E+01,  2.32109616E-02, -7.93355119E-06,
                    #1.23101995E-09, -7.13892514E-14, -2.93324539E+03,
                    #-6.44104411E+01])))

#species(name='C6H12OOH1-2O',
        #atoms='C:6 H:13 O:3',
        #thermo=(NASA([300.00, 1535.00],
                    #[ 2.80591255E+00,  7.79229941E-02, -4.02499146E-05,
                    #6.01126951E-09,  7.21748984E-13, -2.70819447E+04,
                    #2.33451434E+01]),
                #NASA([1535.00, 5000.00],
                    #[ 2.92647989E+01,  2.69146554E-02, -9.38475487E-06,
                    #1.47659709E-09, -8.64902012E-14, -3.71424073E+04,
                    #-1.22240832E+02])))

#species(name='C6H12OOH1-3O',
        #atoms='C:6 H:13 O:3',
        #thermo=(NASA([300.00, 1535.00],
                    #[ 2.80591255E+00,  7.79229941E-02, -4.02499146E-05,
                    #6.01126951E-09,  7.21748984E-13, -2.70819447E+04,
                    #2.33451434E+01]),
                #NASA([1535.00, 5000.00],
                    #[ 2.92647989E+01,  2.69146554E-02, -9.38475487E-06,
                    #1.47659709E-09, -8.64902012E-14, -3.71424073E+04,
                    #-1.22240832E+02])))

#species(name='C6H12OOH1-4O',
        #atoms='C:6 H:13 O:3',
        #thermo=(NASA([300.00, 1535.00],
                    #[ 2.80591255E+00,  7.79229941E-02, -4.02499146E-05,
                    #6.01126951E-09,  7.21748984E-13, -2.70819447E+04,
                    #2.33451434E+01]),
                #NASA([1535.00, 5000.00],
                    #[ 2.92647989E+01,  2.69146554E-02, -9.38475487E-06,
                    #1.47659709E-09, -8.64902012E-14, -3.71424073E+04,
                    #-1.22240832E+02])))

#species(name='C6H12OOH1-5O',
        #atoms='C:6 H:13 O:3',
        #thermo=(NASA([300.00, 1535.00],
                    #[ 2.80591255E+00,  7.79229941E-02, -4.02499146E-05,
                    #6.01126951E-09,  7.21748984E-13, -2.70819447E+04,
                    #2.33451434E+01]),
                #NASA([1535.00, 5000.00],
                    #[ 2.92647989E+01,  2.69146554E-02, -9.38475487E-06,
                    #1.47659709E-09, -8.64902012E-14, -3.71424073E+04,
                    #-1.22240832E+02])))

#species(name='C6H12OOH2-1O',
        #atoms='C:6 H:13 O:3',
        #thermo=(NASA([300.00, 1409.00],
                    #[ 3.34629056E+00,  8.24137185E-02, -5.70430278E-05,
                    #1.99491279E-08, -2.77573029E-12, -2.73151509E+04,
                    #1.93763739E+01]),
                #NASA([1409.00, 5000.00],
                    #[ 2.68394200E+01,  2.78771803E-02, -9.46581223E-06,
                    #1.46245818E-09, -8.45621758E-14, -3.54171892E+04,
                    #-1.06763190E+02])))

#species(name='C6H12OOH2-3O',
        #atoms='C:6 H:13 O:3',
        #thermo=(NASA([300.00, 1408.00],
                    #[ 2.14026689E+00,  9.05994482E-02, -7.11844189E-05,
                    #2.92349633E-08, -4.87305224E-12, -2.92400104E+04,
                    #2.34016654E+01]),
                #NASA([1408.00, 5000.00],
                    #[ 2.70100981E+01,  2.75803123E-02, -9.32839137E-06,
                    #1.43724826E-09, -8.29386001E-14, -3.73322411E+04,
                    #-1.08327965E+02])))

#species(name='C6H12OOH2-4O',
        #atoms='C:6 H:13 O:3',
        #thermo=(NASA([300.00, 1408.00],
                    #[ 2.14026689E+00,  9.05994482E-02, -7.11844189E-05,
                    #2.92349633E-08, -4.87305224E-12, -2.92400104E+04,
                    #2.34016654E+01]),
                #NASA([1408.00, 5000.00],
                    #[ 2.70100981E+01,  2.75803123E-02, -9.32839137E-06,
                    #1.43724826E-09, -8.29386001E-14, -3.73322411E+04,
                    #-1.08327965E+02])))

#species(name='C6H12OOH2-5O',
        #atoms='C:6 H:13 O:3',
        #thermo=(NASA([300.00, 1408.00],
                    #[ 2.14026689E+00,  9.05994482E-02, -7.11844189E-05,
                    #2.92349633E-08, -4.87305224E-12, -2.92400104E+04,
                    #2.34016654E+01]),
                #NASA([1408.00, 5000.00],
                    #[ 2.70100981E+01,  2.75803123E-02, -9.32839137E-06,
                    #1.43724826E-09, -8.29386001E-14, -3.73322411E+04,
                    #-1.08327965E+02])))

#species(name='C6H12OOH2-6O',
        #atoms='C:6 H:13 O:3',
        #thermo=(NASA([300.00, 1409.00],
                    #[ 3.34629056E+00,  8.24137185E-02, -5.70430278E-05,
                    #1.99491279E-08, -2.77573029E-12, -2.73151509E+04,
                    #1.93763739E+01]),
                #NASA([1409.00, 5000.00],
                    #[ 2.68394200E+01,  2.78771803E-02, -9.46581223E-06,
                    #1.46245818E-09, -8.45621758E-14, -3.54171892E+04,
                    #-1.06763190E+02])))

#species(name='C6H12OOH3-1O',
        #atoms='C:6 H:13 O:3',
        #thermo=(NASA([300.00, 1409.00],
                    #[ 3.34629056E+00,  8.24137185E-02, -5.70430278E-05,
                    #1.99491279E-08, -2.77573029E-12, -2.73151509E+04,
                    #1.93763739E+01]),
                #NASA([1409.00, 5000.00],
                    #[ 2.68394200E+01,  2.78771803E-02, -9.46581223E-06,
                    #1.46245818E-09, -8.45621758E-14, -3.54171892E+04,
                    #-1.06763190E+02])))

#species(name='C6H12OOH3-2O',
        #atoms='C:6 H:13 O:3',
        #thermo=(NASA([300.00, 1408.00],
                    #[ 2.14026689E+00,  9.05994482E-02, -7.11844189E-05,
                    #2.92349633E-08, -4.87305224E-12, -2.92400104E+04,
                    #2.34016654E+01]),
                #NASA([1408.00, 5000.00],
                    #[ 2.70100981E+01,  2.75803123E-02, -9.32839137E-06,
                    #1.43724826E-09, -8.29386001E-14, -3.73322411E+04,
                    #-1.08327965E+02])))

#species(name='C6H12OOH3-4O',
        #atoms='C:6 H:13 O:3',
        #thermo=(NASA([300.00, 1408.00],
                    #[ 2.14026689E+00,  9.05994482E-02, -7.11844189E-05,
                    #2.92349633E-08, -4.87305224E-12, -2.92400104E+04,
                    #2.34016654E+01]),
                #NASA([1408.00, 5000.00],
                    #[ 2.70100981E+01,  2.75803123E-02, -9.32839137E-06,
                    #1.43724826E-09, -8.29386001E-14, -3.73322411E+04,
                    #-1.08327965E+02])))

#species(name='C6H12OOH3-5O',
        #atoms='C:6 H:13 O:3',
        #thermo=(NASA([300.00, 1408.00],
                    #[ 2.14026689E+00,  9.05994482E-02, -7.11844189E-05,
                    #2.92349633E-08, -4.87305224E-12, -2.92400104E+04,
                    #2.34016654E+01]),
                #NASA([1408.00, 5000.00],
                    #[ 2.70100981E+01,  2.75803123E-02, -9.32839137E-06,
                    #1.43724826E-09, -8.29386001E-14, -3.73322411E+04,
                    #-1.08327965E+02])))

#species(name='C6H12OOH3-6O',
        #atoms='C:6 H:13 O:3',
        #thermo=(NASA([300.00, 1409.00],
                    #[ 3.34629056E+00,  8.24137185E-02, -5.70430278E-05,
                    #1.99491279E-08, -2.77573029E-12, -2.73151509E+04,
                    #1.93763739E+01]),
                #NASA([1409.00, 5000.00],
                    #[ 2.68394200E+01,  2.78771803E-02, -9.46581223E-06,
                    #1.46245818E-09, -8.45621758E-14, -3.54171892E+04,
                    #-1.06763190E+02])))

#species(name='CH3COOH',
        #atoms='C:2 H:4 O:2',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 2.78950201E+00,  9.99941719E-03,  3.42572245E-05,
                    #-5.09031329E-08,  2.06222185E-11, -5.34752488E+04,
                    #1.41053123E+01]),
                #NASA([1000.00, 6000.00],
                    #[ 7.67084601E+00,  1.35152602E-02, -5.25874333E-06,
                    #8.93184479E-10, -5.53180543E-14, -5.57560970E+04,
                    #-1.54677315E+01])),
        #note='g6/00')

#species(name='C2H5COOH',
        #atoms='C:3 H:6 O:2',
        #thermo=(NASA([200.00, 1000.00],
                    #[ 5.51515351E+00, -1.02654246E-05,  7.54897239E-05,
                    #-9.81744541E-08,  3.88829515E-11, -5.63618075E+04,
                    #6.11744431E+00]),
                #NASA([1000.00, 6000.00],
                    #[ 8.61036813E+00,  1.87894582E-02, -6.70711294E-06,
                    #1.07428300E-09, -6.38870386E-14, -5.84807549E+04,
                    #-1.63110488E+01])),
        #note='T11/07')

#species(name='I3C6',
        #atoms='C:6 H:14',
        #thermo=(NASA([300.00, 1397.00],
                    #[-1.68896577E+00,  7.82590693E-02, -5.26518260E-05,
                    #1.87083154E-08, -2.77104943E-12, -2.29630072E+04,
                    #3.48670784E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 1.93157326E+01,  3.00892098E-02, -1.02405066E-05,
                    #1.58431849E-09, -9.16868661E-14, -3.03915102E+04,
                    #-7.83315381E+01])))

#species(name='I3C6-1',
        #atoms='C:6 H:13',
        #thermo=(NASA([300.00, 1397.00],
                    #[-1.35705050E+00,  7.48590679E-02, -5.16606774E-05,
                    #1.89817723E-08, -2.91856788E-12,  1.76537974E+03,
                    #3.63038339E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 1.86813264E+01,  2.81840981E-02, -9.60554817E-06,
                    #1.48745388E-09, -8.61356516E-14, -5.25846140E+03,
                    #-7.14465957E+01])))

#species(name='I3C6-2',
        #atoms='C:6 H:13',
        #thermo=(NASA([300.00, 1401.00],
                    #[-4.40822702E-01,  6.73884228E-02, -3.90682317E-05,
                    #1.06823783E-08, -1.01515549E-12,  1.78426906E+02,
                    #3.32938534E+01]),
                #NASA([1401.00, 5000.00],
                    #[ 1.80865593E+01,  2.84123255E-02, -9.62566091E-06,
                    #1.48469670E-09, -8.57429880E-14, -6.65973812E+03,
                    #-6.76896956E+01])))

#species(name='I3C6-3',
        #atoms='C:6 H:13',
        #thermo=(NASA([300.00, 1389.00],
                    #[ 2.28121426E-01,  6.06666579E-02, -2.83718958E-05,
                    #4.40800491E-09,  2.59783976E-13, -7.76274694E+02,
                    #3.01229520E+01]),
                #NASA([1389.00, 5000.00],
                    #[ 1.69399268E+01,  2.97496678E-02, -1.01657554E-05,
                    #1.57707940E-09, -9.14450630E-14, -7.43011243E+03,
                    #-6.25727480E+01])))

#species(name='I3C6-4',
        #atoms='C:6 H:13',
        #thermo=(NASA([300.00, 1397.00],
                    #[-1.35705050E+00,  7.48590679E-02, -5.16606774E-05,
                    #1.89817723E-08, -2.91856788E-12,  1.76537974E+03,
                    #3.56143523E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 1.86813264E+01,  2.81840981E-02, -9.60554817E-06,
                    #1.48745388E-09, -8.61356516E-14, -5.25846140E+03,
                    #-7.21360773E+01])))

#species(name='I3C6D1',
        #atoms='C:6 H:12',
        #thermo=(NASA([300.00, 1395.00],
                    #[-1.28061255E+00,  6.97935486E-02, -4.59082978E-05,
                    #1.57564454E-08, -2.24229481E-12, -8.85710162E+03,
                    #3.41707239E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 1.77667979E+01,  2.69756406E-02, -9.21989087E-06,
                    #1.43055518E-09, -8.29577133E-14, -1.56836997E+04,
                    #-6.87930377E+01])))

#species(name='I3C6D2',
        #atoms='C:6 H:12',
        #thermo=(NASA([300.00, 1390.00],
                    #[-1.04785477E-01,  6.28453125E-02, -3.57571402E-05,
                    #9.67773505E-09, -9.39346628E-13, -1.08644015E+04,
                    #2.83728899E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 1.71128150E+01,  2.72628755E-02, -9.26247860E-06,
                    #1.43154043E-09, -8.27928142E-14, -1.73337710E+04,
                    #-6.57848171E+01])))

#species(name='I3C6D3',
        #atoms='C:6 H:12',
        #thermo=(NASA([300.00, 1392.00],
                    #[-1.24661445E+00,  6.76124917E-02, -4.23393767E-05,
                    #1.34898469E-08, -1.73480140E-12, -9.34112148E+03,
                    #3.36321323E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 1.72611895E+01,  2.71307676E-02, -9.21555704E-06,
                    #1.42407568E-09, -8.23524881E-14, -1.60667148E+04,
                    #-6.67804053E+01])))

#species(name='I3C6D1-3',
        #atoms='C:6 H:11',
        #thermo=(NASA([300.00, 1393.00],
                    #[-1.15675656E+00,  6.44703234E-02, -3.96482718E-05,
                    #1.21633493E-08, -1.47656661E-12,  7.85162825E+03,
                    #3.22576024E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 1.68635880E+01,  2.57812938E-02, -8.84405949E-06,
                    #1.37568378E-09, -7.99170732E-14,  1.22096473E+03,
                    #-6.57835946E+01])))

#species(name='I3C6D1-4',
        #atoms='C:6 H:11',
        #thermo=(NASA([300.00, 1395.00],
                    #[-1.59271036E+00,  6.93277449E-02, -4.92183481E-05,
                    #1.86651110E-08, -2.96734679E-12,  1.59659476E+04,
                    #3.78815278E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 1.72207487E+01,  2.50334500E-02, -8.58057575E-06,
                    #1.33389118E-09, -7.74538777E-14,  9.40188509E+03,
                    #-6.31436249E+01])))

#species(name='I3C6D1-5',
        #atoms='C:6 H:11',
        #thermo=(NASA([300.00, 1398.00],
                    #[-4.59832762E-01,  6.08412541E-02, -3.50511287E-05,
                    #9.34427858E-09, -8.29144025E-13,  1.43475465E+04,
                    #3.38786051E+01]),
                #NASA([1398.00, 5000.00],
                    #[ 1.66110857E+01,  2.52471277E-02, -8.58934301E-06,
                    #1.32868114E-09, -7.68901099E-14,  8.01581001E+03,
                    #-5.92786504E+01])))

#species(name='I3C6D1-6',
        #atoms='C:6 H:11',
        #thermo=(NASA([300.00, 1395.00],
                    #[-1.59271036E+00,  6.93277449E-02, -4.92183481E-05,
                    #1.86651110E-08, -2.96734679E-12,  1.59659476E+04,
                    #3.78815278E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 1.72207487E+01,  2.50334500E-02, -8.58057575E-06,
                    #1.33389118E-09, -7.74538777E-14,  9.40188509E+03,
                    #-6.31436249E+01])))

#species(name='I3C6D2-5',
        #atoms='C:6 H:11',
        #thermo=(NASA([300.00, 1389.00],
                    #[-4.20476241E-01,  6.11706644E-02, -3.53044604E-05,
                    #9.62017932E-09, -9.26998242E-13,  6.10952324E+03,
                    #2.80392510E+01]),
                #NASA([1389.00, 5000.00],
                    #[ 1.67138603E+01,  2.55674696E-02, -8.69811395E-06,
                    #1.34563774E-09, -7.78814655E-14, -2.89864535E+02,
                    #-6.55625925E+01])))

#species(name='I3C6D2-6',
        #atoms='C:6 H:11',
        #thermo=(NASA([300.00, 1390.00],
                    #[ 3.03049977E-01,  5.92249143E-02, -3.45203481E-05,
                    #9.82907960E-09, -1.06393838E-12,  1.38492060E+04,
                    #2.87434280E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 1.64692980E+01,  2.53723246E-02, -8.63360755E-06,
                    #1.33570282E-09, -7.73034643E-14,  7.81000051E+03,
                    #-5.95229040E+01])))

#species(name='I3C6D3-1',
        #atoms='C:6 H:11',
        #thermo=(NASA([300.00, 1393.00],
                    #[-8.36604831E-01,  6.39827718E-02, -4.10905950E-05,
                    #1.36352102E-08, -1.85837064E-12,  1.53721575E+04,
                    #3.46820617E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 1.66173114E+01,  2.52409297E-02, -8.58703166E-06,
                    #1.32830316E-09, -7.68674005E-14,  9.07702266E+03,
                    #-5.98272402E+01])))

#species(name='I3C6D3-2',
        #atoms='C:6 H:11',
        #thermo=(NASA([300.00, 1392.00],
                    #[-1.56488055E+00,  6.59496241E-02, -4.19010961E-05,
                    #1.34392828E-08, -1.72364804E-12,  7.63316147E+03,
                    #3.46942308E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 1.68637069E+01,  2.54339391E-02, -8.65066181E-06,
                    #1.33808590E-09, -7.74358916E-14,  9.76674560E+02,
                    #-6.51825889E+01])))

#species(name='I3C6OH1-2J',
        #atoms='C:6 H:13 O:1',
        #thermo=(NASA([300.00, 1398.00],
                    #[ 3.93738874E-01,  7.63734213E-02, -5.36451960E-05,
                    #2.01477829E-08, -3.16252906E-12, -1.78573998E+04,
                    #3.15970947E+01]),
                #NASA([1398.00, 5000.00],
                    #[ 2.06784155E+01,  2.84012023E-02, -9.63501781E-06,
                    #1.48740498E-09, -8.59472200E-14, -2.48857094E+04,
                    #-7.72053110E+01])))

#species(name='I3C6OH2-1J',
        #atoms='C:6 H:13 O:1',
        #thermo=(NASA([300.00, 1404.00],
                    #[ 5.96645387E-02,  8.36748032E-02, -6.68275598E-05,
                    #2.85411678E-08, -4.98386357E-12, -1.88170113E+04,
                    #3.05858223E+01]),
                #NASA([1404.00, 5000.00],
                    #[ 2.18633529E+01,  2.69916660E-02, -9.06097182E-06,
                    #1.38875585E-09, -7.98403124E-14, -2.58277049E+04,
                    #-8.44943426E+01])))

#species(name='I3C6OH2-3J',
        #atoms='C:6 H:13 O:1',
        #thermo=(NASA([300.00, 1403.00],
                    #[ 1.48076386E-01,  8.00236805E-02, -6.07049394E-05,
                    #2.46769166E-08, -4.13123552E-12, -1.99207021E+04,
                    #3.05434747E+01]),
                #NASA([1403.00, 5000.00],
                    #[ 2.10169280E+01,  2.74990635E-02, -9.19185966E-06,
                    #1.40479337E-09, -8.06026153E-14, -2.67867784E+04,
                    #-8.01875837E+01])))

#species(name='I3C6OH3-2J',
        #atoms='C:6 H:13 O:1',
        #thermo=(NASA([300.00, 1384.00],
                    #[ 3.15050305E+00,  6.51866880E-02, -3.78234118E-05,
                    #1.13306453E-08, -1.44651791E-12, -2.15513834E+04,
                    #1.70013978E+01]),
                #NASA([1384.00, 5000.00],
                    #[ 2.06965148E+01,  2.92417750E-02, -1.01134682E-05,
                    #1.58164865E-09, -9.22239821E-14, -2.83290692E+04,
                    #-7.93224338E+01])))

#species(name='I3C6OH3-4J',
        #atoms='C:6 H:13 O:1',
        #thermo=(NASA([300.00, 1384.00],
                    #[ 3.03395385E+00,  6.89126979E-02, -4.39186105E-05,
                    #1.50915652E-08, -2.26334414E-12, -2.04426869E+04,
                    #1.58023932E+01]),
                #NASA([1384.00, 5000.00],
                    #[ 2.17171805E+01,  2.84798352E-02, -9.87444944E-06,
                    #1.54686588E-09, -9.03023622E-14, -2.74432765E+04,
                    #-8.59986115E+01])))

#species(name='I3C6OH4-3J',
        #atoms='C:6 H:13 O:1',
        #thermo=(NASA([300.00, 1398.00],
                    #[ 3.93738874E-01,  7.63734213E-02, -5.36451960E-05,
                    #2.01477829E-08, -3.16252906E-12, -1.78573998E+04,
                    #3.09076131E+01]),
                #NASA([1398.00, 5000.00],
                    #[ 2.06784155E+01,  2.84012023E-02, -9.63501781E-06,
                    #1.48740498E-09, -8.59472200E-14, -2.48857094E+04,
                    #-7.78947926E+01])))

#species(name='I3C6OH1-2O2',
        #atoms='C:6 H:13 O:3',
        #thermo=(NASA([300.00, 1408.00],
                    #[ 1.98460027E-01,  9.48343555E-02, -7.63785557E-05,
                    #3.22459441E-08, -5.51470652E-12, -3.69469429E+04,
                    #3.36598251E+01]),
                #NASA([1408.00, 5000.00],
                    #[ 2.59350350E+01,  2.83082430E-02, -9.53943557E-06,
                    #1.46608879E-09, -8.44540460E-14, -4.51981609E+04,
                    #-1.02209481E+02])))

#species(name='I3C6OH2-1O2',
        #atoms='C:6 H:13 O:3',
        #thermo=(NASA([300.00, 1404.00],
                    #[-3.31466329E-01,  9.59393057E-02, -7.74923755E-05,
                    #3.28662477E-08, -5.65502054E-12, -3.68344087E+04,
                    #3.72646528E+01]),
                #NASA([1404.00, 5000.00],
                    #[ 2.57582406E+01,  2.85342066E-02, -9.63444902E-06,
                    #1.48272114E-09, -8.54961466E-14, -4.52205016E+04,
                    #-1.00512261E+02])))

#species(name='I3C6OH2-3O2',
        #atoms='C:6 H:13 O:3',
        #thermo=(NASA([300.00, 1396.00],
                    #[-2.73355701E-01,  9.89689418E-02, -8.14906699E-05,
                    #3.46540746E-08, -5.92888167E-12, -3.96872345E+04,
                    #3.50905693E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 2.79344065E+01,  2.62376018E-02, -8.76321384E-06,
                    #1.34003284E-09, -7.69637627E-14, -4.87225734E+04,
                    #-1.13850852E+02])))

#species(name='I3C6OH3-2O2',
        #atoms='C:6 H:13 O:3',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 2.32752743E+00,  8.61661554E-02, -6.37566314E-05,
                    #2.48788540E-08, -4.00510558E-12, -4.05402173E+04,
                    #2.27088927E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 2.61755584E+01,  2.85328264E-02, -9.71230576E-06,
                    #1.50305156E-09, -8.70114594E-14, -4.86437177E+04,
                    #-1.04710323E+02])))

#species(name='I3C6OH3-4O2',
        #atoms='C:6 H:13 O:3',
        #thermo=(NASA([300.00, 1390.00],
                    #[ 2.22922171E+00,  8.26992713E-02, -5.62630047E-05,
                    #1.98991192E-08, -2.92056549E-12, -3.83905415E+04,
                    #2.44543132E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 2.57506751E+01,  2.95026522E-02, -1.01803870E-05,
                    #1.58993329E-09, -9.26266521E-14, -4.67958933E+04,
                    #-1.02595328E+02])))

#species(name='I3C6OH4-3O2',
        #atoms='C:6 H:13 O:3',
        #thermo=(NASA([300.00, 1392.00],
                    #[-1.35981304E-01,  9.57752947E-02, -7.50568154E-05,
                    #3.04562291E-08, -5.01895920E-12, -3.76070383E+04,
                    #3.52710770E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 2.75209276E+01,  2.72087878E-02, -9.22637697E-06,
                    #1.42500357E-09, -8.24040166E-14, -4.67728713E+04,
                    #-1.11789773E+02])))

#species(name='I3C6D1-3O',
        #atoms='C:6 H:11 O:1',
        #thermo=(NASA([300.00, 1386.00],
                    #[ 2.39169991E+00,  6.52692776E-02, -4.09431423E-05,
                    #1.21625236E-08, -1.30140435E-12, -4.26723511E+03,
                    #1.81850542E+01]),
                #NASA([1386.00, 5000.00],
                    #[ 2.18759437E+01,  2.30094771E-02, -7.76165130E-06,
                    #1.19472056E-09, -6.89294471E-14, -1.12728406E+04,
                    #-8.74691903E+01])))

#species(name='I3C6D1-4O',
        #atoms='C:6 H:11 O:1',
        #thermo=(NASA([300.00, 1397.00],
                    #[ 5.49089621E-01,  6.91387773E-02, -4.51069226E-05,
                    #1.45931566E-08, -1.84792638E-12, -9.68829533E+02,
                    #2.86607664E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 2.05688161E+01,  2.46934120E-02, -8.45680194E-06,
                    #1.31416289E-09, -7.62970554E-14, -8.11513775E+03,
                    #-7.96159147E+01])))

#species(name='I3C6D1-5O',
        #atoms='C:6 H:11 O:1',
        #thermo=(NASA([300.00, 1409.00],
                    #[ 6.72333821E-01,  7.10035744E-02, -4.97728343E-05,
                    #1.78022837E-08, -2.55156765E-12, -3.08142909E+03,
                    #2.66267894E+01]),
                #NASA([1409.00, 5000.00],
                    #[ 2.06182101E+01,  2.40963519E-02, -8.12961031E-06,
                    #1.25058211E-09, -7.20923631E-14, -9.89705253E+03,
                    #-8.02479353E+01])))

#species(name='I3C6D1-6O',
        #atoms='C:6 H:11 O:1',
        #thermo=(NASA([300.00, 1397.00],
                    #[ 5.49089621E-01,  6.91387773E-02, -4.51069226E-05,
                    #1.45931566E-08, -1.84792638E-12, -9.68829533E+02,
                    #2.86607664E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 2.05688161E+01,  2.46934120E-02, -8.45680194E-06,
                    #1.31416289E-09, -7.62970554E-14, -8.11513775E+03,
                    #-7.96159147E+01])))

#species(name='I3C6D2-5O',
        #atoms='C:6 H:11 O:1',
        #thermo=(NASA([300.00, 1382.00],
                    #[ 2.50312334E+00,  6.22545288E-02, -3.65011370E-05,
                    #1.04021663E-08, -1.15365453E-12, -4.72066477E+03,
                    #1.82970795E+01]),
                #NASA([1382.00, 5000.00],
                    #[ 2.05302780E+01,  2.53414639E-02, -8.81589318E-06,
                    #1.38428218E-09, -8.09474939E-14, -1.15998372E+04,
                    #-8.05309273E+01])))

#species(name='I3C6D2-6O',
        #atoms='C:6 H:11 O:1',
        #thermo=(NASA([300.00, 1393.00],
                    #[ 2.34049517E+00,  5.91648620E-02, -3.11133275E-05,
                    #6.54458022E-09, -1.79460279E-13, -3.05540951E+03,
                    #2.00985851E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 1.93496842E+01,  2.55658983E-02, -8.72165910E-06,
                    #1.35173889E-09, -7.83323605E-14, -9.54378394E+03,
                    #-7.33725464E+01])))

#species(name='I3C6D3-1O',
        #atoms='C:6 H:11 O:1',
        #thermo=(NASA([300.00, 1395.00],
                    #[ 1.15478404E+00,  6.39087360E-02, -3.73930691E-05,
                    #1.01079087E-08, -9.16994191E-13, -1.52021839E+03,
                    #2.62979518E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 1.95366790E+01,  2.54804255E-02, -8.70821835E-06,
                    #1.35129817E-09, -7.83733565E-14, -8.32442056E+03,
                    #-7.39722317E+01])))

#species(name='I3C6D3-2O',
        #atoms='C:6 H:11 O:1',
        #thermo=(NASA([300.00, 1385.00],
                    #[ 1.30145511E+00,  6.69855598E-02, -4.28022331E-05,
                    #1.39639067E-08, -1.88419183E-12, -3.17995671E+03,
                    #2.45871090E+01]),
                #NASA([1385.00, 5000.00],
                    #[ 2.06755005E+01,  2.52190604E-02, -8.77405195E-06,
                    #1.37780504E-09, -8.05729913E-14, -1.03493492E+04,
                    #-8.08575732E+01])))

#species(name='I3C6-1OOH',
        #atoms='C:6 H:14 O:2',
        #thermo=(NASA([300.00, 1396.00],
                    #[-7.24543823E-01,  9.05529844E-02, -6.65544873E-05,
                    #2.58227499E-08, -4.14558279E-12, -3.30534355E+04,
                    #3.62746161E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 2.43728069E+01,  3.03114693E-02, -1.03785129E-05,
                    #1.61236499E-09, -9.35868378E-14, -4.16422060E+04,
                    #-9.79977448E+01])))

#species(name='I3C6-1O2',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 1.14626042E-01,  8.43331129E-02, -6.09313429E-05,
                    #2.32590137E-08, -3.68241426E-12, -1.63785046E+04,
                    #3.21885716E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 2.33911137E+01,  2.89987094E-02, -9.92019112E-06,
                    #1.54022391E-09, -8.93610686E-14, -2.44022130E+04,
                    #-9.25396843E+01])))

#species(name='I3C6-2OOH',
        #atoms='C:6 H:14 O:2',
        #thermo=(NASA([300.00, 1404.00],
                    #[-2.05581836E-01,  9.21694973E-02, -7.13992813E-05,
                    #2.92481953E-08, -4.90325298E-12, -3.52667717E+04,
                    #3.18900261E+01]),
                #NASA([1404.00, 5000.00],
                    #[ 2.46676901E+01,  2.95355869E-02, -9.99460781E-06,
                    #1.54032657E-09, -8.89018906E-14, -4.34408455E+04,
                    #-1.00066964E+02])))

#species(name='I3C6-2O2',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1404.00],
                    #[ 5.14567703E-01,  8.64551967E-02, -6.64819953E-05,
                    #2.71011693E-08, -4.52934144E-12, -1.85733711E+04,
                    #2.83592435E+01]),
                #NASA([1404.00, 5000.00],
                    #[ 2.37228393E+01,  2.81844565E-02, -9.52188754E-06,
                    #1.46584819E-09, -8.45373612E-14, -2.62200258E+04,
                    #-9.48296458E+01])))

#species(name='I3C6-3OOH',
        #atoms='C:6 H:14 O:2',
        #thermo=(NASA([300.00, 1390.00],
                    #[-5.53721872E-01,  9.32020155E-02, -7.02818816E-05,
                    #2.76333657E-08, -4.45421350E-12, -3.59253714E+04,
                    #3.28701714E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 2.61734302E+01,  2.85635101E-02, -9.73621308E-06,
                    #1.50871089E-09, -8.74357206E-14, -4.49837896E+04,
                    #-1.09886777E+02])))

#species(name='I3C6-3O2',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1389.00],
                    #[ 2.24049554E-01,  8.72017294E-02, -6.48663804E-05,
                    #2.51278564E-08, -3.99288326E-12, -1.92401742E+04,
                    #2.90771994E+01]),
                #NASA([1389.00, 5000.00],
                    #[ 2.53068408E+01,  2.70878034E-02, -9.20985312E-06,
                    #1.42491427E-09, -8.24953782E-14, -2.77935501E+04,
                    #-1.05087429E+02])))

#species(name='I3C6-4OOH',
        #atoms='C:6 H:14 O:2',
        #thermo=(NASA([300.00, 1396.00],
                    #[-7.24543823E-01,  9.05529844E-02, -6.65544873E-05,
                    #2.58227499E-08, -4.14558279E-12, -3.30534355E+04,
                    #3.55851345E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 2.43728069E+01,  3.03114693E-02, -1.03785129E-05,
                    #1.61236499E-09, -9.35868378E-14, -4.16422060E+04,
                    #-9.86872265E+01])))

#species(name='I3C6-4O2',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 1.14626042E-01,  8.43331129E-02, -6.09313429E-05,
                    #2.32590137E-08, -3.68241426E-12, -1.63785046E+04,
                    #3.14990900E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 2.33911137E+01,  2.89987094E-02, -9.92019112E-06,
                    #1.54022391E-09, -8.93610686E-14, -2.44022130E+04,
                    #-9.32291659E+01])))

#species(name='I3C6-1O',
        #atoms='C:6 H:13 O:1',
        #thermo=(NASA([300.00, 1399.00],
                    #[ 1.16195237E-01,  7.72969740E-02, -5.12718445E-05,
                    #1.71947154E-08, -2.30552158E-12, -1.50582118E+04,
                    #3.02502613E+01]),
                #NASA([1399.00, 5000.00],
                    #[ 2.20100002E+01,  2.79337371E-02, -9.52880448E-06,
                    #1.47673466E-09, -8.55706183E-14, -2.28068710E+04,
                    #-8.79084182E+01])))

#species(name='I3C6-2O',
        #atoms='C:6 H:13 O:1',
        #thermo=(NASA([300.00, 1407.00],
                    #[ 3.51969362E-01,  7.87530062E-02, -5.55443844E-05,
                    #2.03210675E-08, -3.02248777E-12, -1.71894993E+04,
                    #2.76853136E+01]),
                #NASA([1407.00, 5000.00],
                    #[ 2.19724992E+01,  2.75061475E-02, -9.28030247E-06,
                    #1.42747141E-09, -8.22796835E-14, -2.45697647E+04,
                    #-8.80743634E+01])))

#species(name='I3C6-3O',
        #atoms='C:6 H:13 O:1',
        #thermo=(NASA([300.00, 1381.00],
                    #[ 3.31322762E+00,  6.48736748E-02, -3.44602363E-05,
                    #8.10414653E-09, -5.78777443E-13, -1.88367096E+04,
                    #1.27836942E+01]),
                #NASA([1381.00, 5000.00],
                    #[ 2.19293395E+01,  2.87976742E-02, -1.00021360E-05,
                    #1.56882249E-09, -9.16674223E-14, -2.61633674E+04,
                    #-9.00284701E+01])))

#species(name='I3C6-4O',
        #atoms='C:6 H:13 O:1',
        #thermo=(NASA([300.00, 1399.00],
                    #[ 1.16195237E-01,  7.72969740E-02, -5.12718445E-05,
                    #1.71947154E-08, -2.30552158E-12, -1.50582118E+04,
                    #2.95607797E+01]),
                #NASA([1399.00, 5000.00],
                    #[ 2.20100002E+01,  2.79337371E-02, -9.52880448E-06,
                    #1.47673466E-09, -8.55706183E-14, -2.28068710E+04,
                    #-8.85978999E+01])))

#species(name='I3C6OOH1-2',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1395.00],
                    #[ 1.22493279E+00,  8.02606879E-02, -5.58800029E-05,
                    #2.03795471E-08, -3.07904823E-12, -9.93464345E+03,
                    #2.93525110E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 2.35603211E+01,  2.85095609E-02, -9.76899687E-06,
                    #1.51845056E-09, -8.81672172E-14, -1.77672303E+04,
                    #-9.08081645E+01])))

#species(name='I3C6OOH1-3',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1391.00],
                    #[ 1.57117723E+00,  7.15764618E-02, -4.04242605E-05,
                    #1.03748596E-08, -8.51492971E-13, -1.09322770E+04,
                    #2.97182709E+01]),
                #NASA([1391.00, 5000.00],
                    #[ 2.19989978E+01,  2.98487056E-02, -1.02347042E-05,
                    #1.59160711E-09, -9.24478389E-14, -1.86324572E+04,
                    #-8.21276699E+01])))

#species(name='I3C6OOH1-4',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1397.00],
                    #[-3.35179913E-01,  8.69835854E-02, -6.53023608E-05,
                    #2.58705369E-08, -4.22769777E-12, -8.33653832E+03,
                    #3.67353416E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 2.37976677E+01,  2.82661727E-02, -9.67598144E-06,
                    #1.50300332E-09, -8.72307022E-14, -1.65101648E+04,
                    #-9.20863073E+01])))

#species(name='I3C6OOH1-5',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1401.00],
                    #[ 9.90050328E-01,  7.77380990E-02, -5.04723512E-05,
                    #1.63809818E-08, -2.09145341E-12, -9.98407572E+03,
                    #3.18348013E+01]),
                #NASA([1401.00, 5000.00],
                    #[ 2.29795289E+01,  2.86810254E-02, -9.75893558E-06,
                    #1.50982849E-09, -8.73847254E-14, -1.78060345E+04,
                    #-8.70037878E+01])))

#species(name='I3C6OOH1-6',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1397.00],
                    #[-3.35179913E-01,  8.69835854E-02, -6.53023608E-05,
                    #2.58705369E-08, -4.22769777E-12, -8.33653832E+03,
                    #3.67353416E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 2.37976677E+01,  2.82661727E-02, -9.67598144E-06,
                    #1.50300332E-09, -8.72307022E-14, -1.65101648E+04,
                    #-9.20863073E+01])))

#species(name='I3C6OOH2-1',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1404.00],
                    #[ 6.63136904E-01,  8.73356588E-02, -6.85938945E-05,
                    #2.84528413E-08, -4.81913285E-12, -1.06738926E+04,
                    #3.00453765E+01]),
                #NASA([1404.00, 5000.00],
                    #[ 2.42643610E+01,  2.73995141E-02, -9.27289207E-06,
                    #1.42922795E-09, -8.24953286E-14, -1.83786115E+04,
                    #-9.49827137E+01])))

#species(name='I3C6OOH2-3',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1399.00],
                    #[ 3.75384063E+00,  7.16661371E-02, -4.64399645E-05,
                    #1.56363960E-08, -2.15998784E-12, -1.34862827E+04,
                    #1.60699189E+01]),
                #NASA([1399.00, 5000.00],
                    #[ 2.28996484E+01,  2.85346376E-02, -9.65875691E-06,
                    #1.48883345E-09, -8.59396295E-14, -2.03077154E+04,
                    #-8.73397646E+01])))

#species(name='I3C6OOH2-4',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1405.00],
                    #[ 1.28044896E-01,  8.88340150E-02, -7.04773586E-05,
                    #2.94949572E-08, -5.02887814E-12, -1.05411087E+04,
                    #3.26115819E+01]),
                #NASA([1405.00, 5000.00],
                    #[ 2.41075259E+01,  2.74764333E-02, -9.28725031E-06,
                    #1.43022267E-09, -8.25034742E-14, -1.83177011E+04,
                    #-9.42475539E+01])))

#species(name='I3C6OOH2-5',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1412.00],
                    #[ 1.22951383E+00,  8.04474374E-02, -5.68739757E-05,
                    #2.07359328E-08, -3.04949078E-12, -1.21506399E+04,
                    #2.87782351E+01]),
                #NASA([1412.00, 5000.00],
                    #[ 2.33466475E+01,  2.77765954E-02, -9.31817590E-06,
                    #1.42779224E-09, -8.20776813E-14, -1.96432918E+04,
                    #-8.94982312E+01])))

#species(name='I3C6OOH2-6',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1405.00],
                    #[ 1.28044896E-01,  8.88340150E-02, -7.04773586E-05,
                    #2.94949572E-08, -5.02887814E-12, -1.05411087E+04,
                    #3.26115819E+01]),
                #NASA([1405.00, 5000.00],
                    #[ 2.41075259E+01,  2.74764333E-02, -9.28725031E-06,
                    #1.43022267E-09, -8.25034742E-14, -1.83177011E+04,
                    #-9.42475539E+01])))

#species(name='I3C6OOH3-1',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1390.00],
                    #[-1.91955084E-01,  8.97368258E-02, -6.91549650E-05,
                    #2.77467647E-08, -4.54963917E-12, -1.12038951E+04,
                    #3.41515577E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 2.56304065E+01,  2.64799895E-02, -9.01891363E-06,
                    #1.39694257E-09, -8.09372652E-14, -1.98683606E+04,
                    #-1.03475727E+02])))

#species(name='I3C6OOH3-2',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1388.00],
                    #[ 1.34099199E+00,  8.31851402E-02, -6.00145036E-05,
                    #2.24357182E-08, -3.44091847E-12, -1.27993569E+04,
                    #2.68834775E+01]),
                #NASA([1388.00, 5000.00],
                    #[ 2.54002540E+01,  2.67142448E-02, -9.10798645E-06,
                    #1.41168285E-09, -8.18286449E-14, -2.11251115E+04,
                    #-1.02232550E+02])))

#species(name='I3C6OOH3-4',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1390.00],
                    #[ 3.67439347E-01,  8.82034770E-02, -6.72916895E-05,
                    #2.67436281E-08, -4.35145103E-12, -1.13422627E+04,
                    #3.07684140E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 2.57523166E+01,  2.64452177E-02, -9.02096601E-06,
                    #1.39863676E-09, -8.10887806E-14, -1.99084005E+04,
                    #-1.04688223E+02])))

#species(name='I3C6OOH4-1',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1397.00],
                    #[-3.35179913E-01,  8.69835854E-02, -6.53023608E-05,
                    #2.58705369E-08, -4.22769777E-12, -8.33653832E+03,
                    #3.67353416E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 2.37976677E+01,  2.82661727E-02, -9.67598144E-06,
                    #1.50300332E-09, -8.72307022E-14, -1.65101648E+04,
                    #-9.20863073E+01])))

#species(name='I3C6OOH4-2',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1401.00],
                    #[ 9.90050328E-01,  7.77380990E-02, -5.04723512E-05,
                    #1.63809818E-08, -2.09145341E-12, -9.98407572E+03,
                    #3.18348013E+01]),
                #NASA([1401.00, 5000.00],
                    #[ 2.29795289E+01,  2.86810254E-02, -9.75893558E-06,
                    #1.50982849E-09, -8.73847254E-14, -1.78060345E+04,
                    #-8.70037878E+01])))

#species(name='I3C6OOH4-3',
        #atoms='C:6 H:13 O:2',
        #thermo=(NASA([300.00, 1390.00],
                    #[ 3.41731432E+00,  6.92105456E-02, -4.03671004E-05,
                    #1.14675186E-08, -1.24115192E-12, -1.12995021E+04,
                    #1.96169011E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 2.25544789E+01,  2.93666822E-02, -1.00646956E-05,
                    #1.56456323E-09, -9.08489270E-14, -1.84877002E+04,
                    #-8.49788912E+01])))

#species(name='I3C6O1-2',
        #atoms='C:6 H:12 O:1',
        #thermo=(NASA([300.00, 1415.00],
                    #[-7.89751491E+00,  1.04701589E-01, -9.15117430E-05,
                    #4.08851458E-08, -7.24517504E-12, -2.09086043E+04,
                    #6.46269059E+01]),
                #NASA([1415.00, 5000.00],
                    #[ 2.10544663E+01,  2.59373105E-02, -8.69210012E-06,
                    #1.33097616E-09, -7.64783994E-14, -2.97280027E+04,
                    #-8.67070861E+01])))

#species(name='I3C6O1-3',
        #atoms='C:6 H:12 O:1',
        #thermo=(NASA([300.00, 1419.00],
                    #[-4.88329121E+00,  8.62637279E-02, -6.29774186E-05,
                    #2.35348837E-08, -3.52106173E-12, -2.33009676E+04,
                    #4.99105272E+01]),
                #NASA([1419.00, 5000.00],
                    #[ 1.92340791E+01,  2.78206067E-02, -9.32344795E-06,
                    #1.42770120E-09, -8.20390808E-14, -3.13340383E+04,
                    #-7.86438979E+01])))

#species(name='I3C6O1-4',
        #atoms='C:6 H:12 O:1',
        #thermo=(NASA([300.00, 1440.00],
                    #[-9.14056492E+00,  9.99577191E-02, -7.98381078E-05,
                    #3.24647588E-08, -5.23122509E-12, -2.97452342E+04,
                    #7.13365438E+01]),
                #NASA([1440.00, 5000.00],
                    #[ 1.87844170E+01,  2.78777936E-02, -9.18550377E-06,
                    #1.39036539E-09, -7.92444475E-14, -3.85236468E+04,
                    #-7.58115066E+01])))

#species(name='I3C6O1-5',
        #atoms='C:6 H:12 O:1',
        #thermo=(NASA([300.00, 1446.00],
                    #[-9.24603343E+00,  1.03066187E-01, -8.62597814E-05,
                    #3.67402281E-08, -6.17342004E-12, -3.14287860E+04,
                    #7.03039674E+01]),
                #NASA([1446.00, 5000.00],
                    #[ 1.90821723E+01,  2.70989530E-02, -8.80622542E-06,
                    #1.32006961E-09, -7.47152757E-14, -4.00275384E+04,
                    #-7.79222176E+01])))

#species(name='I3C6O1-6',
        #atoms='C:6 H:12 O:1',
        #thermo=(NASA([300.00, 1431.00],
                    #[-9.81297769E+00,  1.02303637E-01, -8.07747865E-05,
                    #3.23107609E-08, -5.12332826E-12, -3.17844298E+04,
                    #6.91209894E+01]),
                #NASA([1431.00, 5000.00],
                    #[ 1.94007335E+01,  2.81874917E-02, -9.38100076E-06,
                    #1.42999315E-09, -8.19170036E-14, -4.11164110E+04,
                    #-8.53093898E+01])))

#species(name='I3C6O2-3',
        #atoms='C:6 H:12 O:1',
        #thermo=(NASA([300.00, 1412.00],
                    #[-3.93908827E+00,  8.75947269E-02, -6.61476948E-05,
                    #2.55302342E-08, -3.95474691E-12, -2.43672440E+04,
                    #4.45802330E+01]),
                #NASA([1412.00, 5000.00],
                    #[ 2.11037266E+01,  2.62653465E-02, -8.88659835E-06,
                    #1.36977373E-09, -7.90779249E-14, -3.26643915E+04,
                    #-8.87138060E+01])))

#species(name='I3C6O2-4',
        #atoms='C:6 H:12 O:1',
        #thermo=(NASA([300.00, 1416.00],
                    #[-9.74667180E+00,  1.07738476E-01, -9.28982752E-05,
                    #4.09108034E-08, -7.15241229E-12, -2.17584118E+04,
                    #7.30451489E+01]),
                #NASA([1416.00, 5000.00],
                    #[ 2.02545699E+01,  2.69789397E-02, -9.03872586E-06,
                    #1.38384942E-09, -7.95099255E-14, -3.09657029E+04,
                    #-8.40485763E+01])))

#species(name='I3C6O2-5',
        #atoms='C:6 H:12 O:1',
        #thermo=(NASA([300.00, 1422.00],
                    #[-9.58318418E+00,  1.09606472E-01, -9.71924349E-05,
                    #4.37735952E-08, -7.77220974E-12, -2.30805577E+04,
                    #7.00828495E+01]),
                #NASA([1422.00, 5000.00],
                    #[ 2.06156532E+01,  2.62256134E-02, -8.68372456E-06,
                    #1.31880377E-09, -7.53411021E-14, -3.21073576E+04,
                    #-8.72499388E+01])))

#species(name='I3C6O3-4',
        #atoms='C:6 H:12 O:1',
        #thermo=(NASA([300.00, 1417.00],
                    #[-3.13118477E+00,  8.33583436E-02, -6.16639677E-05,
                    #2.34998348E-08, -3.60202603E-12, -2.32326447E+04,
                    #4.13198573E+01]),
                #NASA([1417.00, 5000.00],
                    #[ 1.99864070E+01,  2.67788716E-02, -8.96822735E-06,
                    #1.37262590E-09, -7.88454641E-14, -3.08872212E+04,
                    #-8.17244886E+01])))

#species(name='I3C6OOH1-2O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1404.00],
                    #[ 1.66484330E+00,  9.79761864E-02, -7.93480883E-05,
                    #3.35829302E-08, -5.76016565E-12, -2.86928331E+04,
                    #2.82080616E+01]),
                #NASA([1404.00, 5000.00],
                    #[ 2.87096781E+01,  2.84155546E-02, -9.65124791E-06,
                    #1.49129209E-09, -8.62340093E-14, -3.74179099E+04,
                    #-1.14724991E+02])))

#species(name='I3C6OOH1-3O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1391.00],
                    #[ 1.25228324E+00,  9.93051896E-02, -7.85782514E-05,
                    #3.21467169E-08, -5.34579565E-12, -2.93426698E+04,
                    #3.01702694E+01]),
                #NASA([1391.00, 5000.00],
                    #[ 3.02575443E+01,  2.74334483E-02, -9.39477715E-06,
                    #1.46051393E-09, -8.48376353E-14, -3.89834935E+04,
                    #-1.24114812E+02])))

#species(name='I3C6OOH1-4O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1397.00],
                    #[ 1.24000348E+00,  9.59197820E-02, -7.37329628E-05,
                    #2.96270075E-08, -4.87729963E-12, -2.64938242E+04,
                    #3.21627683E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 2.84494302E+01,  2.91701229E-02, -1.00296755E-05,
                    #1.56267690E-09, -9.08892305E-14, -3.56351431E+04,
                    #-1.12855572E+02])))

#species(name='I3C6OOH1-5O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1404.00],
                    #[ 1.66484330E+00,  9.79761864E-02, -7.93480883E-05,
                    #3.35829302E-08, -5.76016565E-12, -2.86928331E+04,
                    #2.82080616E+01]),
                #NASA([1404.00, 5000.00],
                    #[ 2.87096781E+01,  2.84155546E-02, -9.65124791E-06,
                    #1.49129209E-09, -8.62340093E-14, -3.74179099E+04,
                    #-1.14724991E+02])))

#species(name='I3C6OOH1-6O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1397.00],
                    #[ 1.24000348E+00,  9.59197820E-02, -7.37329628E-05,
                    #2.96270075E-08, -4.87729963E-12, -2.64938242E+04,
                    #3.21627683E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 2.84494302E+01,  2.91701229E-02, -1.00296755E-05,
                    #1.56267690E-09, -9.08892305E-14, -3.56351431E+04,
                    #-1.12855572E+02])))

#species(name='I3C6OOH2-1O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1404.00],
                    #[ 1.66484330E+00,  9.79761864E-02, -7.93480883E-05,
                    #3.35829302E-08, -5.76016565E-12, -2.86928331E+04,
                    #2.82080616E+01]),
                #NASA([1404.00, 5000.00],
                    #[ 2.87096781E+01,  2.84155546E-02, -9.65124791E-06,
                    #1.49129209E-09, -8.62340093E-14, -3.74179099E+04,
                    #-1.14724991E+02])))

#species(name='I3C6OOH2-3O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 1.62466290E+00,  1.01485466E-01, -8.41815078E-05,
                    #3.60167124E-08, -6.20069770E-12, -3.15314916E+04,
                    #2.64827362E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 3.06593617E+01,  2.65245570E-02, -8.95903300E-06,
                    #1.38001704E-09, -7.96529882E-14, -4.08408826E+04,
                    #-1.26821606E+02])))

#species(name='I3C6OOH2-4O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1404.00],
                    #[ 1.66484330E+00,  9.79761864E-02, -7.93480883E-05,
                    #3.35829302E-08, -5.76016565E-12, -2.86928331E+04,
                    #2.82080616E+01]),
                #NASA([1404.00, 5000.00],
                    #[ 2.87096781E+01,  2.84155546E-02, -9.65124791E-06,
                    #1.49129209E-09, -8.62340093E-14, -3.74179099E+04,
                    #-1.14724991E+02])))

#species(name='I3C6OOH2-5O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1411.00],
                    #[ 2.02259283E+00,  1.00188732E-01, -8.50419442E-05,
                    #3.75072447E-08, -6.62376490E-12, -3.08780301E+04,
                    #2.45978281E+01]),
                #NASA([1411.00, 5000.00],
                    #[ 2.90688538E+01,  2.75062355E-02, -9.20623963E-06,
                    #1.40832691E-09, -8.08621601E-14, -3.92468466E+04,
                    #-1.17160296E+02])))

#species(name='I3C6OOH2-6O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1404.00],
                    #[ 1.66484330E+00,  9.79761864E-02, -7.93480883E-05,
                    #3.35829302E-08, -5.76016565E-12, -2.86928331E+04,
                    #2.82080616E+01]),
                #NASA([1404.00, 5000.00],
                    #[ 2.87096781E+01,  2.84155546E-02, -9.65124791E-06,
                    #1.49129209E-09, -8.62340093E-14, -3.74179099E+04,
                    #-1.14724991E+02])))

#species(name='I3C6OOH3-1O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1391.00],
                    #[ 1.25228324E+00,  9.93051896E-02, -7.85782514E-05,
                    #3.21467169E-08, -5.34579565E-12, -2.93426698E+04,
                    #3.01702694E+01]),
                #NASA([1391.00, 5000.00],
                    #[ 3.02575443E+01,  2.74334483E-02, -9.39477715E-06,
                    #1.46051393E-09, -8.48376353E-14, -3.89834935E+04,
                    #-1.24114812E+02])))

#species(name='I3C6OOH3-2O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 1.62466290E+00,  1.01485466E-01, -8.41815078E-05,
                    #3.60167124E-08, -6.20069770E-12, -3.15314916E+04,
                    #2.64827362E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 3.06593617E+01,  2.65245570E-02, -8.95903300E-06,
                    #1.38001704E-09, -7.96529882E-14, -4.08408826E+04,
                    #-1.26821606E+02])))

#species(name='I3C6OOH3-4O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1391.00],
                    #[ 1.25228324E+00,  9.93051896E-02, -7.85782514E-05,
                    #3.21467169E-08, -5.34579565E-12, -2.93426698E+04,
                    #2.94807877E+01]),
                #NASA([1391.00, 5000.00],
                    #[ 3.02575443E+01,  2.74334483E-02, -9.39477715E-06,
                    #1.46051393E-09, -8.48376353E-14, -3.89834935E+04,
                    #-1.24804293E+02])))

#species(name='I3C6OOH4-1O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1397.00],
                    #[ 1.24000348E+00,  9.59197820E-02, -7.37329628E-05,
                    #2.96270075E-08, -4.87729963E-12, -2.64938242E+04,
                    #3.21627683E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 2.84494302E+01,  2.91701229E-02, -1.00296755E-05,
                    #1.56267690E-09, -9.08892305E-14, -3.56351431E+04,
                    #-1.12855572E+02])))

#species(name='I3C6OOH4-2O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1404.00],
                    #[ 1.66484330E+00,  9.79761864E-02, -7.93480883E-05,
                    #3.35829302E-08, -5.76016565E-12, -2.86928331E+04,
                    #2.82080616E+01]),
                #NASA([1404.00, 5000.00],
                    #[ 2.87096781E+01,  2.84155546E-02, -9.65124791E-06,
                    #1.49129209E-09, -8.62340093E-14, -3.74179099E+04,
                    #-1.14724991E+02])))

#species(name='I3C6OOH4-3O2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1391.00],
                    #[ 1.25228324E+00,  9.93051896E-02, -7.85782514E-05,
                    #3.21467169E-08, -5.34579565E-12, -2.93426698E+04,
                    #2.94807877E+01]),
                #NASA([1391.00, 5000.00],
                    #[ 3.02575443E+01,  2.74334483E-02, -9.39477715E-06,
                    #1.46051393E-09, -8.48376353E-14, -3.89834935E+04,
                    #-1.24804293E+02])))

#species(name='I3C6KET1-2',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1390.00],
                    #[-2.82912675E+00,  1.05746961E-01, -8.93185546E-05,
                    #3.80314024E-08, -6.46447805E-12, -4.36177067E+04,
                    #4.80109930E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 2.94280186E+01,  2.34491474E-02, -8.02934662E-06,
                    #1.24928410E-09, -7.26449555E-14, -5.40282391E+04,
                    #-1.22608960E+02])))

#species(name='I3C6KET1-3',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1390.00],
                    #[ 1.60270460E+00,  8.43290058E-02, -5.95794237E-05,
                    #2.10510279E-08, -2.96764863E-12, -4.65933738E+04,
                    #2.63718975E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 2.71516819E+01,  2.56873176E-02, -8.85309882E-06,
                    #1.38205545E-09, -8.05090091E-14, -5.55062636E+04,
                    #-1.11106420E+02])))

#species(name='I3C6KET1-4',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1395.00],
                    #[ 1.09252138E+00,  8.29807328E-02, -5.72365895E-05,
                    #1.96210184E-08, -2.65043519E-12, -4.36665957E+04,
                    #3.06957756E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 2.58395498E+01,  2.65299680E-02, -9.08262755E-06,
                    #1.41143431E-09, -8.19561373E-14, -5.23098099E+04,
                    #-1.02548212E+02])))

#species(name='I3C6KET1-5',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1406.00],
                    #[ 1.46586206E+00,  8.51561953E-02, -6.28305686E-05,
                    #2.34817503E-08, -3.50227947E-12, -4.58555552E+04,
                    #2.69988329E+01]),
                #NASA([1406.00, 5000.00],
                    #[ 2.62328622E+01,  2.56288484E-02, -8.64901136E-06,
                    #1.33116315E-09, -7.67788989E-14, -5.41618370E+04,
                    #-1.05208044E+02])))

#species(name='I3C6KET1-6',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1395.00],
                    #[ 1.09252138E+00,  8.29807328E-02, -5.72365895E-05,
                    #1.96210184E-08, -2.65043519E-12, -4.36665957E+04,
                    #3.06957756E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 2.58395498E+01,  2.65299680E-02, -9.08262755E-06,
                    #1.41143431E-09, -8.19561373E-14, -5.23098099E+04,
                    #-1.02548212E+02])))

#species(name='I3C6KET2-1',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1392.00],
                    #[-6.91142371E-01,  9.38631620E-02, -7.41130220E-05,
                    #2.97122191E-08, -4.77201054E-12, -4.59543541E+04,
                    #3.94616057E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 2.73168011E+01,  2.44590899E-02, -8.20842160E-06,
                    #1.25988955E-09, -7.25688807E-14, -5.51595705E+04,
                    #-1.09344075E+02])))

#species(name='I3C6KET2-3',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1390.00],
                    #[ 1.08997420E-01,  9.12856130E-02, -7.01013113E-05,
                    #2.73094444E-08, -4.27700575E-12, -4.93352491E+04,
                    #3.35655105E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 2.75610713E+01,  2.47537679E-02, -8.41219217E-06,
                    #1.30166581E-09, -7.53848421E-14, -5.85335510E+04,
                    #-1.12861123E+02])))

#species(name='I3C6KET2-4',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1401.00],
                    #[ 2.98204203E+00,  7.41805228E-02, -4.88804382E-05,
                    #1.65379646E-08, -2.26238890E-12, -4.81593730E+04,
                    #2.22944550E+01]),
                #NASA([1401.00, 5000.00],
                    #[ 2.32384651E+01,  2.81720686E-02, -9.51929113E-06,
                    #1.46573282E-09, -8.45459452E-14, -5.52996306E+04,
                    #-8.69153030E+01])))

#species(name='I3C6KET2-5',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1411.00],
                    #[ 3.98715533E+00,  7.33806979E-02, -5.09360684E-05,
                    #1.86894081E-08, -2.81537364E-12, -5.04324775E+04,
                    #1.57472836E+01]),
                #NASA([1411.00, 5000.00],
                    #[ 2.30823109E+01,  2.77462028E-02, -9.24960609E-06,
                    #1.41090179E-09, -8.08394398E-14, -5.69202530E+04,
                    #-8.63689993E+01])))

#species(name='I3C6KET2-6',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1401.00],
                    #[ 2.98204203E+00,  7.41805228E-02, -4.88804382E-05,
                    #1.65379646E-08, -2.26238890E-12, -4.81593730E+04,
                    #2.22944550E+01]),
                #NASA([1401.00, 5000.00],
                    #[ 2.32384651E+01,  2.81720686E-02, -9.51929113E-06,
                    #1.46573282E-09, -8.45459452E-14, -5.52996306E+04,
                    #-8.69153030E+01])))

#species(name='I3C6KET4-1',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 2.23189890E+00,  7.78228067E-02, -5.14278039E-05,
                    #1.70144281E-08, -2.23416815E-12, -4.43166188E+04,
                    #2.57676230E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 2.47370726E+01,  2.75418831E-02, -9.44410834E-06,
                    #1.46881534E-09, -8.53254255E-14, -5.23310402E+04,
                    #-9.58604025E+01])))

#species(name='I3C6KET4-2',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1388.00],
                    #[ 3.03151256E+00,  7.41348440E-02, -4.15943936E-05,
                    #8.99151177E-09, -1.91225043E-13, -4.64862430E+04,
                    #2.08036999E+01]),
                #NASA([1388.00, 5000.00],
                    #[ 2.68627736E+01,  2.60449817E-02, -9.00157328E-06,
                    #1.40786874E-09, -8.21195703E-14, -5.53675568E+04,
                    #-1.09618827E+02])))

#species(name='I3C6KET4-3',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1389.00],
                    #[-6.17374761E-01,  9.50010289E-02, -7.31048850E-05,
                    #2.82918429E-08, -4.39772540E-12, -4.54997316E+04,
                    #3.62030240E+01]),
                #NASA([1389.00, 5000.00],
                    #[ 2.87492714E+01,  2.46051566E-02, -8.54323127E-06,
                    #1.34050272E-09, -7.83719123E-14, -5.54395562E+04,
                    #-1.20748689E+02])))

#species(name='I3C6D1-3OOH',
        #atoms='C:6 H:12 O:2',
        #thermo=(NASA([300.00, 1392.00],
                    #[-1.47893986E+00,  9.60320023E-02, -8.15255732E-05,
                    #3.54487241E-08, -6.17544373E-12, -2.19412259E+04,
                    #3.85271989E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 2.65232846E+01,  2.31916293E-02, -7.80800799E-06,
                    #1.20095522E-09, -6.92735493E-14, -3.08643316E+04,
                    #-1.09133069E+02])))

#species(name='I3C6D1-4OOH',
        #atoms='C:6 H:12 O:2',
        #thermo=(NASA([300.00, 1395.00],
                    #[-4.16353559E-01,  8.25551514E-02, -6.04613341E-05,
                    #2.32359835E-08, -3.68955967E-12, -1.89333947E+04,
                    #3.53508850E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 2.28453956E+01,  2.71744281E-02, -9.34864157E-06,
                    #1.45703499E-09, -8.47614779E-14, -2.69388112E+04,
                    #-8.92615203E+01])))

#species(name='I3C6D1-5OOH',
        #atoms='C:6 H:12 O:2',
        #thermo=(NASA([300.00, 1403.00],
                    #[ 7.46660873E-02,  8.41752605E-02, -6.52427926E-05,
                    #2.66312411E-08, -4.44637224E-12, -2.11390601E+04,
                    #3.11168442E+01]),
                #NASA([1403.00, 5000.00],
                    #[ 2.31381881E+01,  2.64375662E-02, -8.98624421E-06,
                    #1.38917806E-09, -8.03529549E-14, -2.87548651E+04,
                    #-9.13642552E+01])))

#species(name='I3C6D1-6OOH',
        #atoms='C:6 H:12 O:2',
        #thermo=(NASA([300.00, 1395.00],
                    #[-4.16353559E-01,  8.25551514E-02, -6.04613341E-05,
                    #2.32359835E-08, -3.68955967E-12, -1.89333947E+04,
                    #3.53508850E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 2.28453956E+01,  2.71744281E-02, -9.34864157E-06,
                    #1.45703499E-09, -8.47614779E-14, -2.69388112E+04,
                    #-8.92615203E+01])))

#species(name='I3C6D2-1OOH',
        #atoms='C:6 H:12 O:2',
        #thermo=(NASA([300.00, 1386.00],
                    #[-1.17606214E-01,  8.35419156E-02, -6.18141065E-05,
                    #2.39428476E-08, -3.83917351E-12, -1.92721062E+04,
                    #3.38817914E+01]),
                #NASA([1386.00, 5000.00],
                    #[ 2.40868086E+01,  2.62194412E-02, -9.04615472E-06,
                    #1.41301320E-09, -8.23396293E-14, -2.76508053E+04,
                    #-9.59224566E+01])))

#species(name='I3C6D2-4OOH',
        #atoms='C:6 H:12 O:2',
        #thermo=(NASA([300.00, 1386.00],
                    #[-1.17606214E-01,  8.35419156E-02, -6.18141065E-05,
                    #2.39428476E-08, -3.83917351E-12, -1.92721062E+04,
                    #3.38817914E+01]),
                #NASA([1386.00, 5000.00],
                    #[ 2.40868086E+01,  2.62194412E-02, -9.04615472E-06,
                    #1.41301320E-09, -8.23396293E-14, -2.76508053E+04,
                    #-9.59224566E+01])))

#species(name='I3C6D2-5OOH',
        #atoms='C:6 H:12 O:2',
        #thermo=(NASA([300.00, 1390.00],
                    #[-1.89296207E+00,  9.17142559E-02, -7.19716976E-05,
                    #2.87576324E-08, -4.62504379E-12, -2.09730811E+04,
                    #3.91476057E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 2.55845085E+01,  2.41882057E-02, -8.19405623E-06,
                    #1.26549178E-09, -7.32007888E-14, -3.00940586E+04,
                    #-1.07096532E+02])))

#species(name='I3C6D2-6OOH',
        #atoms='C:6 H:12 O:2',
        #thermo=(NASA([300.00, 1392.00],
                    #[ 1.14992977E+00,  7.39937066E-02, -4.82701402E-05,
                    #1.61642416E-08, -2.21883167E-12, -2.10015474E+04,
                    #2.77221028E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 2.19490094E+01,  2.78653381E-02, -9.57140869E-06,
                    #1.49018750E-09, -8.66263117E-14, -2.85114624E+04,
                    #-8.49221221E+01])))

#species(name='I3C6D3-1OOH',
        #atoms='C:6 H:12 O:2',
        #thermo=(NASA([300.00, 1393.00],
                    #[-5.19930235E-02,  7.90424955E-02, -5.52884427E-05,
                    #2.02823972E-08, -3.09059299E-12, -1.94695832E+04,
                    #3.39511167E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 2.20967919E+01,  2.77789169E-02, -9.55012716E-06,
                    #1.48774467E-09, -8.65193585E-14, -2.72558474E+04,
                    #-8.52485429E+01])))

#species(name='I3C6D3-2OOH',
        #atoms='C:6 H:12 O:2',
        #thermo=(NASA([300.00, 1394.00],
                    #[-2.99006065E+00,  9.64736217E-02, -7.89464181E-05,
                    #3.31058962E-08, -5.59043136E-12, -1.94603019E+04,
                    #4.48568472E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 2.53378148E+01,  2.47067817E-02, -8.43088547E-06,
                    #1.30770963E-09, -7.58477715E-14, -2.86788180E+04,
                    #-1.05199849E+02])))

#species(name='I3C6Q12-3',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1399.00],
                    #[ 4.78123196E+00,  8.32459190E-02, -5.77405860E-05,
                    #2.04671191E-08, -2.93386093E-12, -2.35825037E+04,
                    #1.65454565E+01]),
                #NASA([1399.00, 5000.00],
                    #[ 2.85320737E+01,  2.83579540E-02, -9.68024308E-06,
                    #1.50098903E-09, -8.70104462E-14, -3.18498082E+04,
                    #-1.11153458E+02])))

#species(name='I3C6Q12-4',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1405.00],
                    #[ 1.37038482E+00,  9.99430191E-02, -8.29782779E-05,
                    #3.58432242E-08, -6.24161993E-12, -2.06726792E+04,
                    #3.20436010E+01]),
                #NASA([1405.00, 5000.00],
                    #[ 2.89873692E+01,  2.77338301E-02, -9.41166271E-06,
                    #1.45345438E-09, -8.40136687E-14, -2.94558752E+04,
                    #-1.13481132E+02])))

#species(name='I3C6Q12-5',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1405.00],
                    #[ 1.56961591E+00,  9.58275799E-02, -7.54730674E-05,
                    #3.09479695E-08, -5.15091866E-12, -2.21563860E+04,
                    #3.23160664E+01]),
                #NASA([1405.00, 5000.00],
                    #[ 2.83489526E+01,  2.84110518E-02, -9.67464337E-06,
                    #1.49752576E-09, -8.67009540E-14, -3.09260039E+04,
                    #-1.09705199E+02])))

#species(name='I3C6Q12-6',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1405.00],
                    #[ 1.37038482E+00,  9.99430191E-02, -8.29782779E-05,
                    #3.58432242E-08, -6.24161993E-12, -2.06726792E+04,
                    #3.20436010E+01]),
                #NASA([1405.00, 5000.00],
                    #[ 2.89873692E+01,  2.77338301E-02, -9.41166271E-06,
                    #1.45345438E-09, -8.40136687E-14, -2.94558752E+04,
                    #-1.13481132E+02])))

#species(name='I3C6Q13-2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1390.00],
                    #[ 2.37601132E+00,  9.52480625E-02, -7.36419060E-05,
                    #2.93852267E-08, -4.77563842E-12, -2.29026916E+04,
                    #2.72573128E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 3.03774196E+01,  2.70176946E-02, -9.27474742E-06,
                    #1.44412846E-09, -8.39760616E-14, -3.23253864E+04,
                    #-1.22097672E+02])))

#species(name='I3C6Q13-4',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1392.00],
                    #[ 1.43003762E+00,  1.00162219E-01, -8.07992358E-05,
                    #3.36477830E-08, -5.68161530E-12, -2.14501610E+04,
                    #3.17007231E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 3.06968052E+01,  2.68082435E-02, -9.21478746E-06,
                    #1.43594285E-09, -8.35436769E-14, -3.10979795E+04,
                    #-1.23684905E+02])))

#species(name='I3C6Q13-5',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 2.37496818E+00,  9.17878840E-02, -6.71437221E-05,
                    #2.49627467E-08, -3.74312015E-12, -2.29889884E+04,
                    #2.86420091E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 2.95551263E+01,  2.76225143E-02, -9.46127093E-06,
                    #1.47083295E-09, -8.54301308E-14, -3.22921291E+04,
                    #-1.16980109E+02])))

#species(name='I3C6Q13-6',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1392.00],
                    #[ 8.38274593E-01,  1.01819765E-01, -8.28102681E-05,
                    #3.47113253E-08, -5.88720187E-12, -2.13064735E+04,
                    #3.45478271E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 3.06093481E+01,  2.67792319E-02, -9.18364384E-06,
                    #1.42901233E-09, -8.30604637E-14, -3.10689797E+04,
                    #-1.23350043E+02])))

#species(name='I3C6Q14-2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1390.00],
                    #[ 2.37601132E+00,  9.52480625E-02, -7.36419060E-05,
                    #2.93852267E-08, -4.77563842E-12, -2.29026916E+04,
                    #2.72573128E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 3.03774196E+01,  2.70176946E-02, -9.27474742E-06,
                    #1.44412846E-09, -8.39760616E-14, -3.23253864E+04,
                    #-1.22097672E+02])))

#species(name='I3C6Q14-3',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1384.00],
                    #[ 4.52270781E+00,  8.43869260E-02, -5.83634030E-05,
                    #2.05922665E-08, -2.96007148E-12, -2.42604838E+04,
                    #1.77350562E+01]),
                #NASA([1384.00, 5000.00],
                    #[ 2.94165143E+01,  2.78168681E-02, -9.54676415E-06,
                    #1.48621273E-09, -8.64114120E-14, -3.30657895E+04,
                    #-1.16527822E+02])))

#species(name='I3C6Q14-5',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 2.37496818E+00,  9.17878840E-02, -6.71437221E-05,
                    #2.49627467E-08, -3.74312015E-12, -2.29889884E+04,
                    #2.86420091E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 2.95551263E+01,  2.76225143E-02, -9.46127093E-06,
                    #1.47083295E-09, -8.54301308E-14, -3.22921291E+04,
                    #-1.16980109E+02])))

#species(name='I3C6Q14-6',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1392.00],
                    #[ 8.38274593E-01,  1.01819765E-01, -8.28102681E-05,
                    #3.47113253E-08, -5.88720187E-12, -2.13064735E+04,
                    #3.45478271E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 3.06093481E+01,  2.67792319E-02, -9.18364384E-06,
                    #1.42901233E-09, -8.30604637E-14, -3.10689797E+04,
                    #-1.23350043E+02])))

#species(name='I3C6Q15-2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1401.00],
                    #[ 2.70325052E+00,  9.25107486E-02, -6.85286580E-05,
                    #2.57222618E-08, -3.87718919E-12, -2.22073412E+04,
                    #2.59820882E+01]),
                #NASA([1401.00, 5000.00],
                    #[ 3.00571436E+01,  2.73026180E-02, -9.37131476E-06,
                    #1.45866934E-09, -8.47902649E-14, -3.14846057E+04,
                    #-1.20307671E+02])))

#species(name='I3C6Q15-3',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1399.00],
                    #[ 4.78123196E+00,  8.32459190E-02, -5.77405860E-05,
                    #2.04671191E-08, -2.93386093E-12, -2.35825037E+04,
                    #1.65454565E+01]),
                #NASA([1399.00, 5000.00],
                    #[ 2.85320737E+01,  2.83579540E-02, -9.68024308E-06,
                    #1.50098903E-09, -8.70104462E-14, -3.18498082E+04,
                    #-1.11153458E+02])))

#species(name='I3C6Q15-4',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1405.00],
                    #[ 1.37038482E+00,  9.99430191E-02, -8.29782779E-05,
                    #3.58432242E-08, -6.24161993E-12, -2.06726792E+04,
                    #3.20436010E+01]),
                #NASA([1405.00, 5000.00],
                    #[ 2.89873692E+01,  2.77338301E-02, -9.41166271E-06,
                    #1.45345438E-09, -8.40136687E-14, -2.94558752E+04,
                    #-1.13481132E+02])))

#species(name='I3C6Q15-6',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1403.00],
                    #[ 1.78557090E+00,  9.75502666E-02, -7.62981259E-05,
                    #3.04970338E-08, -4.91054204E-12, -2.07643623E+04,
                    #3.02498477E+01]),
                #NASA([1403.00, 5000.00],
                    #[ 3.03098023E+01,  2.70428933E-02, -9.27137787E-06,
                    #1.44196393E-09, -8.37719147E-14, -3.01956113E+04,
                    #-1.21421993E+02])))

#species(name='I3C6Q16-2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 2.34271407E+00,  9.08891378E-02, -6.50085415E-05,
                    #2.35823625E-08, -3.45710824E-12, -2.00316906E+04,
                    #2.95187734E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 2.93556289E+01,  2.82409561E-02, -9.76997521E-06,
                    #1.52864455E-09, -8.91762512E-14, -2.94217592E+04,
                    #-1.15658038E+02])))

#species(name='I3C6Q16-3',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1393.00],
                    #[ 2.59281166E+00,  8.35932199E-02, -5.37620087E-05,
                    #1.70114791E-08, -2.09457480E-12, -2.10315532E+04,
                    #2.94823592E+01]),
                #NASA([1393.00, 5000.00],
                    #[ 2.71170680E+01,  2.99407131E-02, -1.03108514E-05,
                    #1.60826828E-09, -9.36159425E-14, -2.98884809E+04,
                    #-1.03478487E+02])))

#species(name='I3C6Q16-4',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1398.00],
                    #[ 8.09550239E-01,  9.84817439E-02, -7.79738219E-05,
                    #3.21565836E-08, -5.40397864E-12, -1.84546907E+04,
                    #3.59260515E+01]),
                #NASA([1398.00, 5000.00],
                    #[ 2.88506014E+01,  2.84398897E-02, -9.78573952E-06,
                    #1.52544323E-09, -8.87557954E-14, -2.77396749E+04,
                    #-1.13062886E+02])))

#species(name='I3C6Q23-1',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1398.00],
                    #[ 1.77868275E+00,  1.02437120E-01, -8.65252302E-05,
                    #3.75769881E-08, -6.54618845E-12, -2.36351403E+04,
                    #2.81251586E+01]),
                #NASA([1398.00, 5000.00],
                    #[ 3.11125433E+01,  2.58691602E-02, -8.76450530E-06,
                    #1.35274479E-09, -7.81842485E-14, -3.29586226E+04,
                    #-1.26465133E+02])))

#species(name='I3C6Q23-4',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1398.00],
                    #[ 1.21563069E+00,  1.03988598E-01, -8.84210238E-05,
                    #3.86040706E-08, -6.75032967E-12, -2.34962452E+04,
                    #3.08355007E+01]),
                #NASA([1398.00, 5000.00],
                    #[ 3.09863236E+01,  2.59109944E-02, -8.76558263E-06,
                    #1.35160662E-09, -7.80676888E-14, -3.29170819E+04,
                    #-1.25918315E+02])))

#species(name='I3C6Q23-5',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1396.00],
                    #[ 2.74860107E+00,  9.74316957E-02, -7.92608243E-05,
                    #3.32713397E-08, -5.63506640E-12, -2.50916167E+04,
                    #2.35681443E+01]),
                #NASA([1396.00, 5000.00],
                    #[ 3.07668711E+01,  2.61262854E-02, -8.84615928E-06,
                    #1.36483477E-09, -7.88640360E-14, -3.41773572E+04,
                    #-1.24733741E+02])))

#species(name='I3C6Q23-6',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1398.00],
                    #[ 1.21563069E+00,  1.03988598E-01, -8.84210238E-05,
                    #3.86040706E-08, -6.75032967E-12, -2.34962452E+04,
                    #3.08355007E+01]),
                #NASA([1398.00, 5000.00],
                    #[ 3.09863236E+01,  2.59109944E-02, -8.76558263E-06,
                    #1.35160662E-09, -7.80676888E-14, -3.29170819E+04,
                    #-1.25918315E+02])))

#species(name='I3C6Q24-1',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1403.00],
                    #[ 1.78557090E+00,  9.75502666E-02, -7.62981259E-05,
                    #3.04970338E-08, -4.91054204E-12, -2.07643623E+04,
                    #3.02498477E+01]),
                #NASA([1403.00, 5000.00],
                    #[ 3.03098023E+01,  2.70428933E-02, -9.27137787E-06,
                    #1.44196393E-09, -8.37719147E-14, -3.01956113E+04,
                    #-1.21421993E+02])))

#species(name='I3C6Q24-3',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1399.00],
                    #[ 4.78123196E+00,  8.32459190E-02, -5.77405860E-05,
                    #2.04671191E-08, -2.93386093E-12, -2.35825037E+04,
                    #1.65454565E+01]),
                #NASA([1399.00, 5000.00],
                    #[ 2.85320737E+01,  2.83579540E-02, -9.68024308E-06,
                    #1.50098903E-09, -8.70104462E-14, -3.18498082E+04,
                    #-1.11153458E+02])))

#species(name='I3C6Q24-5',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1405.00],
                    #[ 1.56961591E+00,  9.58275799E-02, -7.54730674E-05,
                    #3.09479695E-08, -5.15091866E-12, -2.21563860E+04,
                    #3.23160664E+01]),
                #NASA([1405.00, 5000.00],
                    #[ 2.83489526E+01,  2.84110518E-02, -9.67464337E-06,
                    #1.49752576E-09, -8.67009540E-14, -3.09260039E+04,
                    #-1.09705199E+02])))

#species(name='I3C6Q24-6',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1405.00],
                    #[ 1.37038482E+00,  9.99430191E-02, -8.29782779E-05,
                    #3.58432242E-08, -6.24161993E-12, -2.06726792E+04,
                    #3.20436010E+01]),
                #NASA([1405.00, 5000.00],
                    #[ 2.89873692E+01,  2.77338301E-02, -9.41166271E-06,
                    #1.45345438E-09, -8.40136687E-14, -2.94558752E+04,
                    #-1.13481132E+02])))

#species(name='I3C6Q25-1',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1412.00],
                    #[ 1.84726055E+00,  1.01575622E-01, -8.80001215E-05,
                    #3.94211008E-08, -7.03684873E-12, -2.28729019E+04,
                    #2.79006010E+01]),
                #NASA([1412.00, 5000.00],
                    #[ 2.92462936E+01,  2.68465762E-02, -8.96082729E-06,
                    #1.36821876E-09, -7.84557172E-14, -3.12305433E+04,
                    #-1.15299252E+02])))

#species(name='I3C6Q25-3',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1408.00],
                    #[ 5.22115054E+00,  8.51327635E-02, -6.28839207E-05,
                    #2.40547346E-08, -3.72820768E-12, -2.57819541E+04,
                    #1.18480191E+01]),
                #NASA([1408.00, 5000.00],
                    #[ 2.89268774E+01,  2.74830286E-02, -9.26093868E-06,
                    #1.42344239E-09, -8.20099471E-14, -3.37073333E+04,
                    #-1.14524284E+02])))

#species(name='I3C6Q25-4',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1412.00],
                    #[ 1.84726055E+00,  1.01575622E-01, -8.80001215E-05,
                    #3.94211008E-08, -7.03684873E-12, -2.28729019E+04,
                    #2.72111194E+01]),
                #NASA([1412.00, 5000.00],
                    #[ 2.92462936E+01,  2.68465762E-02, -8.96082729E-06,
                    #1.36821876E-09, -7.84557172E-14, -3.12305433E+04,
                    #-1.15988733E+02])))

#species(name='I3C6Q34-1',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1392.00],
                    #[ 8.38274593E-01,  1.01819765E-01, -8.28102681E-05,
                    #3.47113253E-08, -5.88720187E-12, -2.13064735E+04,
                    #3.45478271E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 3.06093481E+01,  2.67792319E-02, -9.18364384E-06,
                    #1.42901233E-09, -8.30604637E-14, -3.10689797E+04,
                    #-1.23350043E+02])))

#species(name='I3C6Q34-2',
        #atoms='C:6 H:13 O:4',
        #thermo=(NASA([300.00, 1390.00],
                    #[ 2.37601132E+00,  9.52480625E-02, -7.36419060E-05,
                    #2.93852267E-08, -4.77563842E-12, -2.29026916E+04,
                    #2.72573128E+01]),
                #NASA([1390.00, 5000.00],
                    #[ 3.03774196E+01,  2.70176946E-02, -9.27474742E-06,
                    #1.44412846E-09, -8.39760616E-14, -3.23253864E+04,
                    #-1.22097672E+02])))

#species(name='I3C6O12-3OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1422.00],
                    #[-4.94752752E+00,  1.13990892E-01, -1.05269563E-04,
                    #4.88911513E-08, -8.87967990E-12, -3.41931651E+04,
                    #5.38118714E+01]),
                #NASA([1422.00, 5000.00],
                    #[ 2.63973406E+01,  2.49428309E-02, -8.20157131E-06,
                    #1.23966204E-09, -7.05834291E-14, -4.33049077E+04,
                    #-1.08591982E+02])))

#species(name='I3C6O12-4OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1419.00],
                    #[-4.60865563E+00,  1.06965247E-01, -8.95316178E-05,
                    #3.76760603E-08, -6.27070791E-12, -3.13672811E+04,
                    #5.44774348E+01]),
                #NASA([1419.00, 5000.00],
                    #[ 2.67045981E+01,  2.55737099E-02, -8.61380058E-06,
                    #1.32408678E-09, -7.63062913E-14, -4.12056197E+04,
                    #-1.10413444E+02])))

#species(name='I3C6O12-5OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1428.00],
                    #[-4.12342632E+00,  1.08095970E-01, -9.21896833E-05,
                    #3.92283535E-08, -6.55789888E-12, -3.35648081E+04,
                    #5.03432717E+01]),
                #NASA([1428.00, 5000.00],
                    #[ 2.76444362E+01,  2.43715824E-02, -8.11607374E-06,
                    #1.23822177E-09, -7.09903602E-14, -4.33813302E+04,
                    #-1.16450193E+02])))

#species(name='I3C6O12-6OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1419.00],
                    #[-4.60865563E+00,  1.06965247E-01, -8.95316178E-05,
                    #3.76760603E-08, -6.27070791E-12, -3.13672811E+04,
                    #5.44774348E+01]),
                #NASA([1419.00, 5000.00],
                    #[ 2.67045981E+01,  2.55737099E-02, -8.61380058E-06,
                    #1.32408678E-09, -7.63062913E-14, -4.12056197E+04,
                    #-1.10413444E+02])))

#species(name='I3C6O13-2OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1424.00],
                    #[-3.63134192E+00,  9.86693553E-02, -7.38788568E-05,
                    #2.71741271E-08, -3.88378661E-12, -3.55253638E+04,
                    #4.77200095E+01]),
                #NASA([1424.00, 5000.00],
                    #[ 2.64364809E+01,  2.64057769E-02, -8.94282811E-06,
                    #1.37999933E-09, -7.97517772E-14, -4.55160445E+04,
                    #-1.12626871E+02])))

#species(name='I3C6O13-4OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1415.00],
                    #[-4.04771034E+00,  9.90114227E-02, -7.74281881E-05,
                    #3.09000651E-08, -4.93273017E-12, -3.33687865E+04,
                    #5.12546258E+01]),
                #NASA([1415.00, 5000.00],
                    #[ 2.42841128E+01,  2.80135287E-02, -9.44301925E-06,
                    #1.45196658E-09, -8.36812016E-14, -4.25731473E+04,
                    #-9.89378368E+01])))

#species(name='I3C6O13-5OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1424.00],
                    #[-3.63134192E+00,  9.86693553E-02, -7.38788568E-05,
                    #2.71741271E-08, -3.88378661E-12, -3.55253638E+04,
                    #4.77200095E+01]),
                #NASA([1424.00, 5000.00],
                    #[ 2.64364809E+01,  2.64057769E-02, -8.94282811E-06,
                    #1.37999933E-09, -7.97517772E-14, -4.55160445E+04,
                    #-1.12626871E+02])))

#species(name='I3C6O13-6OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1415.00],
                    #[-4.04771034E+00,  9.90114227E-02, -7.74281881E-05,
                    #3.09000651E-08, -4.93273017E-12, -3.33687865E+04,
                    #5.12546258E+01]),
                #NASA([1415.00, 5000.00],
                    #[ 2.42841128E+01,  2.80135287E-02, -9.44301925E-06,
                    #1.45196658E-09, -8.36812016E-14, -4.25731473E+04,
                    #-9.89378368E+01])))

#species(name='I3C6O14-2OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1414.00],
                    #[-1.00478576E+01,  1.22362907E-01, -1.07064051E-04,
                    #4.72048323E-08, -8.22363539E-12, -4.16467714E+04,
                    #7.90998417E+01]),
                #NASA([1414.00, 5000.00],
                    #[ 2.52388656E+01,  2.77775858E-02, -9.39584783E-06,
                    #1.44824752E-09, -8.36147228E-14, -5.24996752E+04,
                    #-1.05786311E+02])))

#species(name='I3C6O14-3OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1425.00],
                    #[-7.25079629E+00,  1.10380470E-01, -8.71492567E-05,
                    #3.39548220E-08, -5.17955459E-12, -4.28089443E+04,
                    #6.59919714E+01]),
                #NASA([1425.00, 5000.00],
                    #[ 2.64697672E+01,  2.68442334E-02, -9.10447695E-06,
                    #1.40640366E-09, -8.13394632E-14, -5.37616848E+04,
                    #-1.12946074E+02])))

#species(name='I3C6O14-5OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1414.00],
                    #[-1.00478576E+01,  1.22362907E-01, -1.07064051E-04,
                    #4.72048323E-08, -8.22363539E-12, -4.16467714E+04,
                    #7.90998417E+01]),
                #NASA([1414.00, 5000.00],
                    #[ 2.52388656E+01,  2.77775858E-02, -9.39584783E-06,
                    #1.44824752E-09, -8.36147228E-14, -5.24996752E+04,
                    #-1.05786311E+02])))

#species(name='I3C6O14-6OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1411.00],
                    #[-1.03366089E+01,  1.20346033E-01, -1.02766426E-04,
                    #4.44150439E-08, -7.62416432E-12, -3.94806250E+04,
                    #8.23111687E+01]),
                #NASA([1411.00, 5000.00],
                    #[ 2.45559872E+01,  2.86023107E-02, -9.73298877E-06,
                    #1.50624341E-09, -8.72055177E-14, -5.04135491E+04,
                    #-1.01181575E+02])))

#species(name='I3C6O15-2OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1419.00],
                    #[-9.94988101E+00,  1.24453749E-01, -1.11626887E-04,
                    #5.02360028E-08, -8.88369650E-12, -4.33598469E+04,
                    #7.71393391E+01]),
                #NASA([1419.00, 5000.00],
                    #[ 2.55828266E+01,  2.70927490E-02, -9.07547405E-06,
                    #1.38968253E-09, -7.98646439E-14, -5.40501697E+04,
                    #-1.08239859E+02])))

#species(name='I3C6O15-3OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1425.00],
                    #[-8.38087478E+00,  1.19368121E-01, -1.06686216E-04,
                    #4.79658281E-08, -8.46358316E-12, -4.43505809E+04,
                    #6.95458800E+01]),
                #NASA([1425.00, 5000.00],
                    #[ 2.51386704E+01,  2.68450961E-02, -8.85674408E-06,
                    #1.34202514E-09, -7.65538348E-14, -5.43309391E+04,
                    #-1.05027794E+02])))

#species(name='I3C6O15-4OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1416.00],
                    #[-1.03235868E+01,  1.22778618E-01, -1.07905039E-04,
                    #4.78155530E-08, -8.36462744E-12, -4.11791658E+04,
                    #8.07603313E+01]),
                #NASA([1416.00, 5000.00],
                    #[ 2.48373348E+01,  2.79464175E-02, -9.41658500E-06,
                    #1.44762275E-09, -8.34233914E-14, -5.19290058E+04,
                    #-1.03249128E+02])))

#species(name='I3C6O15-6OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1416.00],
                    #[-1.03235868E+01,  1.22778618E-01, -1.07905039E-04,
                    #4.78155530E-08, -8.36462744E-12, -4.11791658E+04,
                    #8.07603313E+01]),
                #NASA([1416.00, 5000.00],
                    #[ 2.48373348E+01,  2.79464175E-02, -9.41658500E-06,
                    #1.44762275E-09, -8.34233914E-14, -5.19290058E+04,
                    #-1.03249128E+02])))

#species(name='I3C6O16-2OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1413.00],
                    #[-1.05162473E+01,  1.24121873E-01, -1.07696329E-04,
                    #4.69950757E-08, -8.10967615E-12, -4.37232903E+04,
                    #7.65732787E+01]),
                #NASA([1413.00, 5000.00],
                    #[ 2.57214656E+01,  2.80049364E-02, -9.52123373E-06,
                    #1.47270866E-09, -8.52368897E-14, -5.49708362E+04,
                    #-1.13657768E+02])))

#species(name='I3C6O16-3OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1417.00],
                    #[-8.10149397E+00,  1.15073547E-01, -9.59707321E-05,
                    #4.03441977E-08, -6.72291801E-12, -4.48417978E+04,
                    #6.36988331E+01]),
                #NASA([1417.00, 5000.00],
                    #[ 2.55198041E+01,  2.79347299E-02, -9.44502839E-06,
                    #1.45550694E-09, -8.40238069E-14, -5.54502543E+04,
                    #-1.13468213E+02])))

#species(name='I3C6O16-4OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1425.00],
                    #[-8.83930941E+00,  1.14540144E-01, -9.45444790E-05,
                    #3.92829144E-08, -6.45232067E-12, -4.18759783E+04,
                    #6.91092553E+01]),
                #NASA([1425.00, 5000.00],
                    #[ 2.44474365E+01,  2.83783818E-02, -9.49868894E-06,
                    #1.45383369E-09, -8.35288626E-14, -5.23471371E+04,
                    #-1.06261617E+02])))

#species(name='I3C6O23-1OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1421.00],
                    #[-2.36832340E+00,  9.74580017E-02, -7.56113605E-05,
                    #2.92504118E-08, -4.45746156E-12, -3.45633413E+04,
                    #4.23992527E+01]),
                #NASA([1421.00, 5000.00],
                    #[ 2.67462710E+01,  2.56215292E-02, -8.64959731E-06,
                    #1.33175431E-09, -7.68390745E-14, -4.40707387E+04,
                    #-1.12235905E+02])))

#species(name='I3C6O23-4OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1421.00],
                    #[-2.36832340E+00,  9.74580017E-02, -7.56113605E-05,
                    #2.92504118E-08, -4.45746156E-12, -3.45633413E+04,
                    #4.23992527E+01]),
                #NASA([1421.00, 5000.00],
                    #[ 2.67462710E+01,  2.56215292E-02, -8.64959731E-06,
                    #1.33175431E-09, -7.68390745E-14, -4.40707387E+04,
                    #-1.12235905E+02])))

#species(name='I3C6O23-5OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1418.00],
                    #[-5.83483410E-01,  9.38782319E-02, -7.30218779E-05,
                    #2.85557675E-08, -4.42448842E-12, -3.69810529E+04,
                    #3.20558800E+01]),
                #NASA([1418.00, 5000.00],
                    #[ 2.70554032E+01,  2.53293620E-02, -8.54066379E-06,
                    #1.31377575E-09, -7.57488585E-14, -4.59909913E+04,
                    #-1.14648911E+02])))

#species(name='I3C6O23-6OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1421.00],
                    #[-2.36832340E+00,  9.74580017E-02, -7.56113605E-05,
                    #2.92504118E-08, -4.45746156E-12, -3.45633413E+04,
                    #4.23992527E+01]),
                #NASA([1421.00, 5000.00],
                    #[ 2.67462710E+01,  2.56215292E-02, -8.64959731E-06,
                    #1.33175431E-09, -7.68390745E-14, -4.40707387E+04,
                    #-1.12235905E+02])))

#species(name='I3C6O24-1OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1415.00],
                    #[-8.19152859E+00,  1.18004430E-01, -1.03918300E-04,
                    #4.62179566E-08, -8.11898212E-12, -3.19569161E+04,
                    #7.08942198E+01]),
                #NASA([1415.00, 5000.00],
                    #[ 2.54564894E+01,  2.70278496E-02, -9.10592259E-06,
                    #1.39970653E-09, -8.06545079E-14, -4.22318654E+04,
                    #-1.05134909E+02])))

#species(name='I3C6O24-3OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1424.00],
                    #[-6.51578749E+00,  1.15569944E-01, -1.04080591E-04,
                    #4.71934953E-08, -8.39422486E-12, -3.50814601E+04,
                    #6.09643066E+01]),
                #NASA([1424.00, 5000.00],
                    #[ 2.57144016E+01,  2.59486883E-02, -8.55039243E-06,
                    #1.29442876E-09, -7.37891093E-14, -4.46217737E+04,
                    #-1.06677091E+02])))

#species(name='I3C6O24-5OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1427.00],
                    #[-6.31214566E+00,  1.12790378E-01, -9.63528638E-05,
                    #4.11151881E-08, -6.89633632E-12, -3.43659897E+04,
                    #6.03098368E+01]),
                #NASA([1427.00, 5000.00],
                    #[ 2.67286622E+01,  2.55448657E-02, -8.51395963E-06,
                    #1.29958989E-09, -7.45329553E-14, -4.45649825E+04,
                    #-1.13117494E+02])))

#species(name='I3C6O24-6OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1415.00],
                    #[-8.19152859E+00,  1.18004430E-01, -1.03918300E-04,
                    #4.62179566E-08, -8.11898212E-12, -3.19569161E+04,
                    #7.08942198E+01]),
                #NASA([1415.00, 5000.00],
                    #[ 2.54564894E+01,  2.70278496E-02, -9.10592259E-06,
                    #1.39970653E-09, -8.06545079E-14, -4.22318654E+04,
                    #-1.05134909E+02])))

#species(name='I3C6O25-1OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1419.00],
                    #[-8.06704251E+00,  1.20032108E-01, -1.08378097E-04,
                    #4.91632931E-08, -8.75539802E-12, -3.32732229E+04,
                    #6.87976406E+01]),
                #NASA([1419.00, 5000.00],
                    #[ 2.58411289E+01,  2.62823335E-02, -8.75949702E-06,
                    #1.33659163E-09, -7.66214517E-14, -4.33903848E+04,
                    #-1.07807121E+02])))

#species(name='I3C6O25-3OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1428.00],
                    #[-6.55401053E+00,  1.18475066E-01, -1.10303331E-04,
                    #5.13660846E-08, -9.31457813E-12, -3.63749706E+04,
                    #5.89167952E+01]),
                #NASA([1428.00, 5000.00],
                    #[ 2.60238777E+01,  2.51109111E-02, -8.14106261E-06,
                    #1.21851255E-09, -6.88975386E-14, -4.57152048E+04,
                    #-1.09505477E+02])))

#species(name='I3C6O25-4OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1419.00],
                    #[-8.06704251E+00,  1.20032108E-01, -1.08378097E-04,
                    #4.91632931E-08, -8.75539802E-12, -3.32732229E+04,
                    #6.81081590E+01]),
                #NASA([1419.00, 5000.00],
                    #[ 2.58411289E+01,  2.62823335E-02, -8.75949702E-06,
                    #1.33659163E-09, -7.66214517E-14, -4.33903848E+04,
                    #-1.08496603E+02])))

#species(name='I3C6O34-1OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1414.00],
                    #[-2.19365284E+00,  9.57573665E-02, -7.57246698E-05,
                    #3.06677771E-08, -4.97570110E-12, -3.33184373E+04,
                    #4.28608483E+01]),
                #NASA([1414.00, 5000.00],
                    #[ 2.50158514E+01,  2.69694918E-02, -9.08271760E-06,
                    #1.39565839E-09, -8.03982034E-14, -4.21075320E+04,
                    #-1.01184336E+02])))

#species(name='I3C6O34-2OOH',
        #atoms='C:6 H:12 O:3',
        #thermo=(NASA([300.00, 1437.00],
                    #[-1.85378717E+00,  9.60969233E-02, -7.30171210E-05,
                    #2.71746730E-08, -3.91468360E-12, -3.54755301E+04,
                    #3.95942211E+01]),
                #NASA([1437.00, 5000.00],
                    #[ 2.75009199E+01,  2.47814716E-02, -8.32166736E-06,
                    #1.27704300E-09, -7.35245468E-14, -4.51175672E+04,
                    #-1.16619334E+02])))

#species(name='IC4H8CHO-I',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1392.00],
                    #[ 6.78430407E-01,  5.47590592E-02, -3.35484208E-05,
                    #9.50304097E-09, -9.12939231E-13, -6.31380501E+03,
                    #2.73016794E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 1.70088018E+01,  1.97463033E-02, -6.68388086E-06,
                    #1.03095251E-09, -5.95584687E-14, -1.22172753E+04,
                    #-6.13790930E+01])))

#species(name='IC5Y3-1J',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1403.00],
                    #[ 2.51943573E+00,  4.61590469E-02, -2.54550286E-05,
                    #6.56306739E-09, -5.53024226E-13, -1.03963466E+04,
                    #1.91325936E+01]),
                #NASA([1403.00, 5000.00],
                    #[ 1.44195565E+01,  2.13745113E-02, -7.11468880E-06,
                    #1.08421089E-09, -6.20821656E-14, -1.48091567E+04,
                    #-4.58099462E+01])))

#species(name='C5H9O2-1',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 2009.00],
                    #[ 5.72019645E+00,  3.38812633E-02, -5.09895590E-06,
                    #-5.35818970E-09,  1.74308203E-12,  2.01705208E+03,
                    #3.62677538E+00]),
                #NASA([2009.00, 5000.00],
                    #[ 1.33770985E+01,  2.58764763E-02, -9.46375867E-06,
                    #1.54116443E-09, -9.25409940E-14, -1.49719105E+03,
                    #-4.08892250E+01])))

#species(name='IC5O34-1J',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1416.00],
                    #[-7.58520151E+00,  8.99166474E-02, -8.33799124E-05,
                    #3.89261590E-08, -7.11907353E-12,  6.38658617E+03,
                    #6.30360362E+01]),
                #NASA([1416.00, 5000.00],
                    #[ 1.72442501E+01,  1.95545494E-02, -6.52245192E-06,
                    #9.95585825E-10, -5.70801192E-14, -8.86533486E+02,
                    #-6.57361056E+01])))

#species(name='C5Y2-3J',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1428.00],
                    #[ 3.62180151E+00,  3.78869197E-02, -9.52099354E-06,
                    #-4.70140858E-09,  2.11456934E-12, -1.46643385E+04,
                    #1.63246571E+01]),
                #NASA([1428.00, 5000.00],
                    #[ 1.53040392E+01,  2.03924168E-02, -6.74389287E-06,
                    #1.02444156E-09, -5.85693457E-14, -1.95765010E+04,
                    #-4.96835091E+01])))

#species(name='C5H9O15-3',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1654.00],
                    #[-7.65918032E+00,  7.51555438E-02, -5.35430991E-05,
                    #1.87653227E-08, -2.55046729E-12, -4.70120353E+03,
                    #6.17593746E+01]),
                #NASA([1654.00, 5000.00],
                    #[ 1.28785030E+01,  2.45625233E-02, -8.33592818E-06,
                    #1.29394798E-09, -7.52420237E-14, -1.10905390E+04,
                    #-4.68798527E+01])))

#species(name='AC5H9O-D',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1397.00],
                    #[ 1.58473487E+00,  5.21376819E-02, -2.99340556E-05,
                    #7.61948758E-09, -5.66969702E-13,  1.31488873E+03,
                    #2.22413284E+01]),
                #NASA([1397.00, 5000.00],
                    #[ 1.68230880E+01,  2.06107677E-02, -7.03818794E-06,
                    #1.09161656E-09, -6.32927008E-14, -4.33693027E+03,
                    #-6.09619458E+01])))

#species(name='IC5D1Y3',
        #atoms='C:5 H:8 O:1',
        #thermo=(NASA([300.00, 1392.00],
                    #[ 1.54292407E+00,  5.05432300E-02, -3.42434122E-05,
                    #1.25252368E-08, -1.95302231E-12, -2.12702785E+04,
                    #1.99855884E+01]),
                #NASA([1392.00, 5000.00],
                    #[ 1.49935813E+01,  1.97788403E-02, -6.78313207E-06,
                    #1.05481230E-09, -6.12614698E-14, -2.60877929E+04,
                    #-5.26202086E+01])))

#species(name='SC4H9CHO',
        #atoms='C:5 H:10 O:1',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 1.54864671E+00,  5.27738215E-02, -2.86549570E-05,
                    #6.79154678E-09, -4.27169992E-13, -3.12991709E+04,
                    #2.20212553E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 1.63021444E+01,  2.30698041E-02, -7.86137450E-06,
                    #1.21738705E-09, -7.05021727E-14, -3.68935591E+04,
                    #-5.88925746E+01])))

#species(name='SC4H9CO',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1394.00],
                    #[ 2.26391239E+00,  4.82108365E-02, -2.59177903E-05,
                    #5.89459018E-09, -2.98061487E-13, -1.27613784E+04,
                    #1.97560888E+01]),
                #NASA([1394.00, 5000.00],
                    #[ 1.59103481E+01,  2.08954660E-02, -7.12058057E-06,
                    #1.10273127E-09, -6.38660845E-14, -1.79403219E+04,
                    #-5.51235141E+01])))

#species(name='SC4H8CHO-1',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1395.00],
                    #[ 1.90536582E+00,  4.93474146E-02, -2.76153915E-05,
                    #6.97838849E-09, -5.41985302E-13, -6.57727764E+03,
                    #2.26336284E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 1.57294793E+01,  2.10330242E-02, -7.16413280E-06,
                    #1.10909618E-09, -6.42183068E-14, -1.17660629E+04,
                    #-5.30038857E+01])))

#species(name='SC4H8CHO-2',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1387.00],
                    #[ 3.81498506E+00,  4.45955336E-02, -2.27390501E-05,
                    #4.77576329E-09, -1.79866775E-13, -1.26864535E+04,
                    #9.75123178E+00]),
                #NASA([1387.00, 5000.00],
                    #[ 1.61231619E+01,  2.07894715E-02, -7.09982327E-06,
                    #1.10107861E-09, -6.38322303E-14, -1.74883191E+04,
                    #-5.81603096E+01])))

#species(name='SC4H8CHO-3',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1517.00],
                    #[ 2.97084586E+00,  4.09072987E-02, -1.35689475E-05,
                    #-2.09797192E-09,  1.49966939E-12, -8.17701585E+03,
                    #1.90042822E+01]),
                #NASA([1517.00, 5000.00],
                    #[ 1.50672347E+01,  2.14197047E-02, -7.26201070E-06,
                    #1.12093960E-09, -6.47770977E-14, -1.31912870E+04,
                    #-4.89644579E+01])))

#species(name='SC4H8CHO-4',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1395.00],
                    #[ 1.90536582E+00,  4.93474146E-02, -2.76153915E-05,
                    #6.97838849E-09, -5.41985302E-13, -6.57727764E+03,
                    #2.26336284E+01]),
                #NASA([1395.00, 5000.00],
                    #[ 1.57294793E+01,  2.10330242E-02, -7.16413280E-06,
                    #1.10909618E-09, -6.42183068E-14, -1.17660629E+04,
                    #-5.30038857E+01])))

#species(name='C5Y3',
        #atoms='C:5 H:10 O:1',
        #thermo=(NASA([300.00, 1421.00],
                    #[ 3.04335984E+00,  4.24811577E-02, -1.12087588E-05,
                    #-5.05051822E-09,  2.35367193E-12, -3.40615638E+04,
                    #1.48305047E+01]),
                #NASA([1421.00, 5000.00],
                    #[ 1.65155931E+01,  2.18966306E-02, -7.26184066E-06,
                    #1.10537530E-09, -6.32910022E-14, -3.96722497E+04,
                    #-6.11243728E+01])))

#species(name='C5Y3-2J',
        #atoms='C:5 H:9 O:1',
        #thermo=(NASA([300.00, 1426.00],
                    #[ 4.26150430E+00,  3.32446421E-02, -1.19632509E-06,
                    #-1.01813061E-08,  3.34232716E-12, -1.47673892E+04,
                    #1.31662371E+01]),
                #NASA([1426.00, 5000.00],
                    #[ 1.52590939E+01,  2.06874083E-02, -6.90185396E-06,
                    #1.05491074E-09, -6.05775562E-14, -1.97701234E+04,
                    #-5.03404450E+01])))

#species(name='I3C6KET1-2O',
        #atoms='C:6 H:11 O:2',
        #thermo=(NASA([300.00, 1387.00],
                    #[ 1.47514508E+00,  7.37368478E-02, -4.75275831E-05,
                    #1.45861545E-08, -1.67541673E-12, -2.69690009E+04,
                    #2.50247479E+01]),
                #NASA([1387.00, 5000.00],
                    #[ 2.43996785E+01,  2.42690001E-02, -8.46948983E-06,
                    #1.33294434E-09, -7.80777993E-14, -3.52998321E+04,
                    #-9.94795816E+01])))

#species(name='I3C6KET1-3O',
        #atoms='C:6 H:11 O:2',
        #thermo=(NASA([300.00, 1684.00],
                    #[ 4.08883205E+00,  6.16936430E-02, -3.23611258E-05,
                    #6.78045374E-09, -2.58965420E-13, -2.92798862E+04,
                    #1.21041822E+01]),
                #NASA([1684.00, 5000.00],
                    #[ 2.04251660E+01,  2.84825575E-02, -1.02006108E-05,
                    #1.63957872E-09, -9.76066080E-14, -3.50770140E+04,
                    #-7.68926591E+01])))

#species(name='I3C6KET1-4O',
        #atoms='C:6 H:11 O:2',
        #thermo=(NASA([300.00, 1402.00],
                    #[ 2.52840003E+00,  6.70964500E-02, -3.86270828E-05,
                    #9.23670076E-09, -4.71797374E-13, -2.57580761E+04,
                    #2.12416721E+01]),
                #NASA([1402.00, 5000.00],
                    #[ 2.31589700E+01,  2.44272013E-02, -8.32837250E-06,
                    #1.29072343E-09, -7.48080311E-14, -3.33339353E+04,
                    #-9.12800070E+01])))

#species(name='I3C6KET1-5O',
        #atoms='C:6 H:11 O:2',
        #thermo=(NASA([300.00, 1421.00],
                    #[ 2.41183155E+00,  7.03279387E-02, -4.53588172E-05,
                    #1.37488631E-08, -1.47061122E-12, -2.78441049E+04,
                    #1.95579807E+01]),
                #NASA([1421.00, 5000.00],
                    #[ 2.33792710E+01,  2.37215024E-02, -7.97370898E-06,
                    #1.22404012E-09, -7.04749814E-14, -3.51963786E+04,
                    #-9.36119976E+01])))

#species(name='I3C6KET1-6O',
        #atoms='C:6 H:11 O:2',
        #thermo=(NASA([300.00, 1402.00],
                    #[ 2.52840003E+00,  6.70964500E-02, -3.86270828E-05,
                    #9.23670076E-09, -4.71797374E-13, -2.57580761E+04,
                    #2.12416721E+01]),
                #NASA([1402.00, 5000.00],
                    #[ 2.31589700E+01,  2.44272013E-02, -8.32837250E-06,
                    #1.29072343E-09, -7.48080311E-14, -3.33339353E+04,
                    #-9.12800070E+01])))

#species(name='I3C6KET2-1O',
        #atoms='C:6 H:11 O:2',
        #thermo=(NASA([300.00, 1681.00],
                    #[ 6.53394028E+00,  5.05671484E-02, -2.04662185E-05,
                    #1.51629257E-09,  5.96583427E-13, -2.97000376E+04,
                    #2.66219510E+00]),
                #NASA([1681.00, 5000.00],
                    #[ 1.83716555E+01,  2.94776735E-02, -1.03680236E-05,
                    #1.64644071E-09, -9.71936322E-14, -3.41888278E+04,
                    #-6.28954907E+01])))

#species(name='I3C6KET2-3O',
        #atoms='C:6 H:11 O:2',
        #thermo=(NASA([300.00, 1379.00],
                    #[ 4.10096125E+00,  5.87754442E-02, -2.64776764E-05,
                    #2.29570121E-09,  9.23350747E-13, -3.16433715E+04,
                    #1.22430748E+01]),
                #NASA([1379.00, 5000.00],
                    #[ 2.24599944E+01,  2.54253000E-02, -8.75952621E-06,
                    #1.36686811E-09, -7.95933727E-14, -3.88982688E+04,
                    #-8.96147641E+01])))

#species(name='I3C6KET2-4O',
        #atoms='C:6 H:11 O:2',
        #thermo=(NASA([300.00, 1418.00],
                    #[ 4.50189394E+00,  5.79939263E-02, -2.99519446E-05,
                    #6.01307546E-09, -6.10184525E-14, -3.02650042E+04,
                    #1.24390596E+01]),
                #NASA([1418.00, 5000.00],
                    #[ 2.05107645E+01,  2.61094947E-02, -8.77868143E-06,
                    #1.34710898E-09, -7.75170005E-14, -3.62982472E+04,
                    #-7.53608067E+01])))

#species(name='I3C6KET2-5O',
        #atoms='C:6 H:11 O:2',
        #thermo=(NASA([300.00, 1402.00],
                    #[ 3.32601505E+00,  6.11076162E-02, -3.30111868E-05,
                    #7.46076408E-09, -3.35635003E-13, -3.20566960E+04,
                    #1.66934584E+01]),
                #NASA([1402.00, 5000.00],
                    #[ 2.05507123E+01,  2.62381301E-02, -8.86019661E-06,
                    #1.36379521E-09, -7.86518816E-14, -3.85101765E+04,
                    #-7.76069702E+01])))

#species(name='I3C6KET2-6O',
        #atoms='C:6 H:11 O:2',
        #thermo=(NASA([300.00, 1418.00],
                    #[ 4.50189394E+00,  5.79939263E-02, -2.99519446E-05,
                    #6.01307546E-09, -6.10184525E-14, -3.02650042E+04,
                    #1.24390596E+01]),
                #NASA([1418.00, 5000.00],
                    #[ 2.05107645E+01,  2.61094947E-02, -8.77868143E-06,
                    #1.34710898E-09, -7.75170005E-14, -3.62982472E+04,
                    #-7.53608067E+01])))

#species(name='I3C6KET4-1O',
        #atoms='C:6 H:11 O:2',
        #thermo=(NASA([300.00, 1678.00],
                    #[ 6.07415745E+00,  4.99570073E-02, -1.60313636E-05,
                    #-1.95829206E-09,  1.37255917E-12, -2.67250333E+04,
                    #5.50215562E+00]),
                #NASA([1678.00, 5000.00],
                    #[ 1.84147452E+01,  3.08132368E-02, -1.11494459E-05,
                    #1.80334629E-09, -1.07789681E-13, -3.16866269E+04,
                    #-6.38709938E+01])))

#species(name='I3C6KET4-2O',
        #atoms='C:6 H:11 O:2',
        #thermo=(NASA([300.00, 1404.00],
                    #[ 2.73918546E+00,  6.86589684E-02, -4.48648325E-05,
                    #1.47084949E-08, -1.90415225E-12, -2.83667867E+04,
                    #1.84225773E+01]),
                #NASA([1404.00, 5000.00],
                    #[ 2.20687296E+01,  2.53046596E-02, -8.61385806E-06,
                    #1.33298744E-09, -7.71600840E-14, -3.52178498E+04,
                    #-8.59554757E+01])))

#species(name='I3C6KET4-3O',
        #atoms='C:6 H:11 O:2',
        #thermo=(NASA([300.00, 1479.00],
                    #[ 3.70417518E+00,  5.81789275E-02, -1.86414586E-05,
                    #-5.18047960E-09,  2.89957058E-12, -2.77945113E+04,
                    #1.45666159E+01]),
                #NASA([1479.00, 5000.00],
                    #[ 2.50319097E+01,  2.41409414E-02, -8.52146180E-06,
                    #1.35152255E-09, -7.95972970E-14, -3.66207850E+04,
                    #-1.05313788E+02])))

##-------------------------------------------------------------------------------
## Reaction data
##-------------------------------------------------------------------------------

## Reaction 1
#reaction('H + O2 <=> O + OH', [1.040000e+14, 0.0, 15286.0])

## Reaction 2
#reaction('O + H2 <=> H + OH', [5.080000e+04, 2.67, 6292.0])

## Reaction 3
#reaction('OH + H2 <=> H + H2O', [4.380000e+13, 0.0, 6990.0])

## Reaction 4
#reaction('O + H2O <=> OH + OH', [6.700000e+07, 1.704, 14986.8])

## Reaction 5
#three_body_reaction('H2 + M <=> H + H + M', [4.577000e+19, -1.4, 104400.0],
                    #efficiencies='H2:2.5 H2O:12.0 CO:1.9 CO2:3.8 HE:0.83 CH4:2.0 C2H6:3.0')

## Reaction 6
#three_body_reaction('O + O + M <=> O2 + M', [6.165000e+15, -0.5, 0.0],
                    #efficiencies='H2:2.5 H2O:12.0 AR:0.83 CO:1.9 CO2:3.8 HE:0.83 CH4:2.0 C2H6:3.0')

## Reaction 7
#three_body_reaction('O + H + M <=> OH + M', [4.714000e+18, -1.0, 0.0],
                    #efficiencies='H2:2.5 H2O:12.0 AR:0.75 CO:1.5 CO2:2.0 HE:0.75 CH4:2.0 C2H6:3.0')

## Reaction 8
#three_body_reaction('H + OH + M <=> H2O + M', [3.500000e+22, -2.0, 0.0],
                    #efficiencies='H2:0.73 H2O:3.65 CH4:2.0 C2H6:3.0 AR:0.38')

## Reaction 9
#falloff_reaction('H + O2 (+ M) <=> HO2 (+ M)',
                #kf=[4.650000e+12, 0.44, 0.0],
                #kf0=[1.737000e+19, -1.23, 0.0],
                #efficiencies='H2:1.3 CO:1.9 CO2:3.8 HE:0.64 H2O:10.0 AR:0.5 CH4:2.0 C2H6:3.0',
                #falloff=Troe(A=0.67, T3=1e-30, T1=1e+30, T2=1e+30))

## Reaction 10
#reaction('HO2 + H <=> OH + OH', [7.079000e+13, 0.0, 295.0])

## Reaction 11
#reaction('H + HO2 <=> H2 + O2', [1.140200e+10, 1.0827, 553.78])

## Reaction 12
#reaction('HO2 + O <=> OH + O2', [3.250000e+13, 0.0, 0.0])

## Reaction 13
#reaction('OH + HO2 <=> H2O + O2', [7.000000e+12, 0.0, -1092.96],
        #options='duplicate')

## Reaction 14
#reaction('OH + HO2 <=> H2O + O2', [4.500000e+14, 0.0, 10929.6],
        #options='duplicate')

## Reaction 15
#reaction('HO2 + HO2 <=> H2O2 + O2', [1.000000e+14, 0.0, 11040.8832],
        #options='duplicate')

## Reaction 16
#reaction('HO2 + HO2 <=> H2O2 + O2', [1.900000e+11, 0.0, -1408.9248],
        #options='duplicate')

## Reaction 17
#falloff_reaction('H2O2 (+ M) <=> OH + OH (+ M)',
                #kf=[2.000000e+12, 0.9, 48749.0],
                #kf0=[2.490000e+24, -2.3, 48749.0],
                #efficiencies='H2O:7.65 CO2:1.6 N2:1.5 O2:1.2 HE:0.65 H2O2:7.7 H2:3.7 CO:2.8',
                #falloff=Troe(A=0.43, T3=1e-30, T1=1e+30))

## Reaction 18
#reaction('H2O2 + H <=> H2O + OH', [2.410000e+13, 0.0, 3970.0])

## Reaction 19
#reaction('H2O2 + H <=> H2 + HO2', [2.150000e+10, 1.0, 6000.0])

## Reaction 20
#reaction('H2O2 + O <=> OH + HO2', [9.550000e+06, 2.0, 3970.0])

## Reaction 21
#reaction('H2O2 + OH <=> H2O + HO2', [1.740000e+12, 0.0, 318.0],
        #options='duplicate')

## Reaction 22
#reaction('H2O2 + OH <=> H2O + HO2', [7.590000e+13, 0.0, 7269.0],
        #options='duplicate')

## Reaction 23
#falloff_reaction('CO + O (+ M) <=> CO2 (+ M)',
                #kf=[1.362000e+10, 0.0, 2384.0],
                #kf0=[1.173000e+24, -2.79, 4191.0],
                #efficiencies='H2:2.0 H2O:12.0 CO:1.75 CO2:3.6 AR:0.7 HE:0.7')

## Reaction 24
#reaction('CO + O2 <=> CO2 + O', [1.119000e+12, 0.0, 47700.0])

## Reaction 25
#reaction('CO + OH <=> CO2 + H', [7.015000e+04, 2.053, -355.7],
        #options='duplicate')

## Reaction 26
#reaction('CO + OH <=> CO2 + H', [5.757000e+12, -0.664, 331.8],
        #options='duplicate')

## Reaction 27
#reaction('CO + HO2 <=> CO2 + OH', [1.570000e+05, 2.18, 17940.0])

## Reaction 28
#three_body_reaction('HCO + M <=> H + CO + M', [5.700000e+11, 0.66, 14870.0],
                    #efficiencies='H2:2.0 H2O:6.0 CO:1.5 CO2:2.0 CH4:2.0 C2H6:3.0')

## Reaction 29
#reaction('HCO + O2 <=> CO + HO2', [7.580000e+12, 0.0, 410.0])

## Reaction 30
#reaction('HCO + H <=> CO + H2', [7.340000e+13, 0.0, 0.0])

## Reaction 31
#reaction('HCO + O <=> CO + OH', [3.020000e+13, 0.0, 0.0])

## Reaction 32
#reaction('HCO + O <=> CO2 + H', [3.000000e+13, 0.0, 0.0])

## Reaction 33
#reaction('HCO + OH <=> CO + H2O', [3.011000e+13, 0.0, 0.0])

## Reaction 34
#reaction('HCO + HO2 => CO2 + H + OH', [3.000000e+13, 0.0, 0.0])

## Reaction 35
#reaction('HCO + HCO => H2 + CO + CO', [3.000000e+12, 0.0, 0.0])

## Reaction 36
#reaction('HCO + CH3 <=> CH4 + CO', [2.650000e+13, 0.0, 0.0])

## Reaction 37
#reaction('CH2O + O2 <=> HCO + HO2', [8.070000e+15, 0.0, 53420.0])

## Reaction 38
#reaction('HCO + O2 <=> O2CHO', [1.200000e+11, 0.0, -1100.0])

## Reaction 39
#reaction('CH2O + O2CHO <=> HCO + HO2CHO', [1.990000e+12, 0.0, 11660.0])

## Reaction 40
#reaction('OCHO + OH <=> HO2CHO', [2.000000e+13, 0.0, 0.0])

## Reaction 41
#reaction('H + CO2 <=> OCHO', [7.500000e+13, 0.0, 29000.0])

## Reaction 42
#reaction('HCO + HCO <=> CH2O + CO', [1.800000e+13, 0.0, 0.0])

## Reaction 43
#three_body_reaction('H + O + M <=> OH* + M', [1.500000e+13, 0.0, 5975.0],
                    #efficiencies='H2:1.0 H2O:6.5 O2:0.4 N2:0.4 AR:0.35')

## Reaction 44
#reaction('OH* + H2O <=> OH + H2O', [5.930000e+12, 0.5, -860.0])

## Reaction 45
#reaction('OH* + H2 <=> OH + H2', [2.950000e+12, 0.5, -444.0])

## Reaction 46
#reaction('OH* + N2 <=> OH + N2', [1.080000e+11, 0.5, -1242.0])

## Reaction 47
#reaction('OH* + OH <=> OH + OH', [6.010000e+12, 0.5, -764.0])

## Reaction 48
#reaction('OH* + H <=> OH + H', [1.310000e+12, 0.5, -167.0])

## Reaction 49
#reaction('OH* + AR <=> OH + AR', [1.690000e+12, 0.0, 4135.0])

## Reaction 50
#reaction('OH* <=> OH', [1.450000e+06, 0.0, 0.0])

## Reaction 51
#reaction('OH* + O2 <=> OH + O2', [2.100000e+12, 0.5, -478.0])

## Reaction 52
#reaction('OH* + CO2 <=> OH + CO2', [2.750000e+12, 0.5, -968.0])

## Reaction 53
#reaction('OH* + CO <=> OH + CO', [3.230000e+12, 0.5, -787.0])

## Reaction 54
#reaction('OH* + CH4 <=> OH + CH4', [3.360000e+12, 0.5, -635.0])

## Reaction 55
#reaction('CH + O2 <=> CO + OH*', [4.040000e+13, 0.0, 0.0])

## Reaction 56
#reaction('C2H + O <=> CO + CH*', [6.200000e+12, 0.0, 0.0])

## Reaction 57
#three_body_reaction('C + H + M <=> CH* + M', [6.000000e+14, 0.0, 6940.0])

## Reaction 58
#reaction('C2H + O2 <=> CO2 + CH*', [2.170000e+10, 0.0, 0.0])

## Reaction 59
#reaction('CH* + AR <=> CH + AR', [4.000000e+11, 0.5, 0.0])

## Reaction 60
#reaction('CH* + H2O <=> CH + H2O', [5.300000e+13, 0.0, 0.0])

## Reaction 61
#reaction('CH* + CO <=> CH + CO', [2.440000e+12, 0.5, 0.0])

## Reaction 62
#reaction('CH* + CO2 <=> CH + CO2', [2.410000e-01, 4.3, -1694.0])

## Reaction 63
#reaction('CH* + O2 <=> CH + O2', [2.480000e+06, 2.14, -1720.0])

## Reaction 64
#reaction('CH* + H2 <=> CH + H2', [1.470000e+14, 0.0, 1361.0])

## Reaction 65
#reaction('CH* + CH4 <=> CH + CH4', [1.730000e+13, 0.0, 167.0])

## Reaction 66
#reaction('CH* <=> CH', [1.860000e+06, 0.0, 0.0])

## Reaction 67
#reaction('CH* + N2 <=> CH + N2', [3.030000e+02, 3.4, -381.0])

## Reaction 68
#falloff_reaction('HCO + H (+ M) <=> CH2O (+ M)',
                #kf=[1.090000e+12, 0.48, -260.0],
                #kf0=[1.350000e+24, -2.57, 1425.0],
                #efficiencies='H2:2.0 H2O:6.0 AR:0.7 CO:1.5 CO2:2.0 CH4:2.0 C2H6:3.0 HE:0.7',
                #falloff=Troe(A=0.7824, T3=271.0, T1=2755.0, T2=6570.0))

## Reaction 69
#falloff_reaction('CO + H2 (+ M) <=> CH2O (+ M)',
                #kf=[4.300000e+07, 1.5, 79600.0],
                #kf0=[5.070000e+27, -3.42, 84348.0],
                #efficiencies='H2:2.0 H2O:6.0 AR:0.7 CO:1.5 CO2:2.0 CH4:2.0 C2H6:3.0 HE:0.7',
                #falloff=Troe(A=0.932, T3=197.0, T1=1540.0, T2=10300.0))

## Reaction 70
#reaction('CH2O + OH <=> HCO + H2O', [7.820000e+07, 1.63, -1055.0])

## Reaction 71
#reaction('CH2O + H <=> HCO + H2', [5.740000e+07, 1.9, 2740.0])

## Reaction 72
#reaction('CH2O + O <=> HCO + OH', [6.260000e+09, 1.15, 2260.0])

## Reaction 73
#reaction('CH2O + CH3 <=> HCO + CH4', [3.830000e+01, 3.36, 4312.0])

## Reaction 74
#reaction('CH2O + HO2 <=> HCO + H2O2', [1.880000e+04, 2.7, 11520.0])

## Reaction 75
#reaction('CH2O + OH <=> HOCH2O', [4.500000e+15, -1.1, 0.0])

## Reaction 76
#reaction('HOCH2O <=> HOCHO + H', [1.000000e+14, 0.0, 14900.0])

## Reaction 77
#reaction('HOCHO <=> CO + H2O', [2.450000e+12, 0.0, 60470.0])

## Reaction 78
#reaction('HOCHO <=> CO2 + H2', [2.950000e+09, 0.0, 48520.0])

## Reaction 79
#reaction('OCHO + HO2 <=> HOCHO + O2', [3.500000e+10, 0.0, -3275.0])

## Reaction 80
#reaction('HOCHO + OH => H2O + CO2 + H', [2.620000e+06, 2.06, 916.0])

## Reaction 81
#reaction('HOCHO + OH => H2O + CO + OH', [1.850000e+07, 1.51, -962.0])

## Reaction 82
#reaction('HOCHO + H => H2 + CO2 + H', [4.240000e+06, 2.1, 4868.0])

## Reaction 83
#reaction('HOCHO + H => H2 + CO + OH', [6.030000e+13, -0.35, 2988.0])

## Reaction 84
#reaction('HOCHO + CH3 => CH4 + CO + OH', [3.900000e-07, 5.8, 2200.0])

## Reaction 85
#reaction('OCHO + H2O2 <=> HOCHO + HO2', [2.400000e+12, 0.0, 10000.0])

## Reaction 86
#reaction('HOCHO + HO2 => H2O2 + CO + OH', [1.000000e+12, 0.0, 11920.0])

## Reaction 87
#reaction('HOCHO + O => CO + OH + OH', [1.770000e+18, -1.9, 2975.0])

## Reaction 88
#reaction('CH2O + OCHO <=> HOCHO + HCO', [5.600000e+12, 0.0, 13600.0])

## Reaction 89
#falloff_reaction('CH3O (+ M) <=> CH2O + H (+ M)',
                #kf=[6.800000e+13, 0.0, 26170.0],
                #kf0=[1.867000e+25, -3.0, 24307.0],
                #efficiencies='H2:2.0 H2O:6.0 CO:1.5 CO2:2.0 CH4:2.0 C2H6:3.0',
                #falloff=Troe(A=0.9, T3=2500.0, T1=1300.0, T2=1e+99))

## Reaction 90
#reaction('CH3O + O2 <=> CH2O + HO2', [4.380000e-19, 9.5, -5501.0])

## Reaction 91
#reaction('CH2O + CH3O <=> CH3OH + HCO', [6.620000e+11, 0.0, 2294.0])

## Reaction 92
#reaction('CH3 + CH3OH <=> CH4 + CH3O', [1.440000e+01, 3.1, 6935.0])

## Reaction 93
#reaction('CH3O + CH3 <=> CH2O + CH4', [1.200000e+13, 0.0, 0.0])

## Reaction 94
#reaction('CH3O + H <=> CH2O + H2', [2.000000e+13, 0.0, 0.0])

## Reaction 95
#reaction('CH3O + HO2 <=> CH2O + H2O2', [3.010000e+11, 0.0, 0.0])

## Reaction 96
#falloff_reaction('CH2O + H (+ M) <=> CH2OH (+ M)',
                #kf=[5.400000e+11, 0.454, 3600.0],
                #kf0=[1.270000e+32, -4.82, 6530.0],
                #efficiencies='H2:2.0 H2O:6.0 CO:1.5 CO2:2.0 CH4:2.0 C2H6:3.0',
                #falloff=Troe(A=0.7187, T3=103.0, T1=1291.0, T2=4160.0))

## Reaction 97
#reaction('CH2OH + O2 <=> CH2O + HO2', [1.510000e+15, -1.0, 0.0],
        #options='duplicate')

## Reaction 98
#reaction('CH2OH + O2 <=> CH2O + HO2', [2.410000e+14, 0.0, 5017.0],
        #options='duplicate')

## Reaction 99
#reaction('CH2OH + H <=> CH2O + H2', [6.000000e+12, 0.0, 0.0])

## Reaction 100
#reaction('CH2OH + HO2 <=> CH2O + H2O2', [1.200000e+13, 0.0, 0.0])

## Reaction 101
#reaction('CH2OH + HCO <=> CH2O + CH2O', [1.800000e+14, 0.0, 0.0])

## Reaction 102
#reaction('CH2OH + CH3O <=> CH2O + CH3OH', [2.400000e+13, 0.0, 0.0])

## Reaction 103
#reaction('CH3OH + HCO <=> CH2OH + CH2O', [9.630000e+03, 2.9, 13110.0])

## Reaction 104
#reaction('OH + CH2OH <=> H2O + CH2O', [2.400000e+13, 0.0, 0.0])

## Reaction 105
#reaction('O + CH2OH <=> OH + CH2O', [4.200000e+13, 0.0, 0.0])

## Reaction 106
#reaction('CH2OH + CH2OH <=> CH2O + CH3OH', [3.000000e+12, 0.0, 0.0])

## Reaction 107
#reaction('CH2OH + HO2 <=> HOCH2O + OH', [1.000000e+13, 0.0, 0.0])

## Reaction 108
#reaction('CH2O + HO2 <=> OCH2O2H', [1.500000e+11, 0.0, 11900.0])

## Reaction 109
#reaction('OCH2O2H <=> HOCH2O2', [3.000000e+11, 0.0, 8600.0])

## Reaction 110
#reaction('HOCH2O2 + HO2 <=> HOCH2O2H + O2', [3.500000e+10, 0.0, -3275.0])

## Reaction 111
#reaction('HOCH2O + OH <=> HOCH2O2H', [1.000000e+13, 0.0, 0.0])

## Reaction 112
#falloff_reaction('CH3OH (+ M) <=> CH3 + OH (+ M)',
                #kf=[2.084000e+18, -0.615, 92540.6],
                #kf0=[1.500000e+43, -6.995, 97992.2],
                #falloff=Troe(A=-0.4748, T3=35580.0, T1=1116.0, T2=9023.0))

## Reaction 113
#falloff_reaction('CH3OH (+ M) <=> CH2(S) + H2O (+ M)',
                #kf=[3.121000e+18, -1.017, 91712.0],
                #kf0=[1.430000e+47, -8.227, 99417.1],
                #falloff=Troe(A=2.545, T3=3290.0, T1=47320.0, T2=47110.0))

## Reaction 114
#falloff_reaction('CH3OH (+ M) <=> CH2OH + H (+ M)',
                #kf=[7.896000e-03, 5.038, 84467.4],
                #kf0=[3.390000e+42, -7.244, 105230.3],
                #falloff=Troe(A=-73.91, T3=37050.0, T1=41500.0, T2=5220.0))

## Reaction 115
#reaction('CH3OH + H <=> CH2OH + H2', [3.070000e+05, 2.55, 5440.0])

## Reaction 116
#reaction('CH3OH + H <=> CH3O + H2', [1.990000e+05, 2.56, 10300.0])

## Reaction 117
#reaction('CH3OH + O <=> CH2OH + OH', [3.880000e+05, 2.5, 3080.0])

## Reaction 118
#reaction('CH3OH + OH <=> CH2OH + H2O', [3.080000e+04, 2.65, -806.7])

## Reaction 119
#reaction('CH3OH + OH <=> CH3O + H2O', [1.500000e+02, 3.03, -763.0])

## Reaction 120
#reaction('CH3OH + O2 <=> CH2OH + HO2', [2.050000e+13, 0.0, 44900.0])

## Reaction 121
#reaction('CH3OH + HO2 <=> CH2OH + H2O2', [1.080000e+04, 2.55, 10530.0])

## Reaction 122
#reaction('CH3OH + CH3 <=> CH2OH + CH4', [3.190000e+01, 3.17, 7172.0])

## Reaction 123
#reaction('CH3O + CH3OH <=> CH2OH + CH3OH', [3.000000e+11, 0.0, 4074.0])

## Reaction 124
#reaction('CH3O + CH3O <=> CH3OH + CH2O', [6.030000e+13, 0.0, 0.0])

## Reaction 125
#falloff_reaction('CH3 + H (+ M) <=> CH4 (+ M)',
                #kf=[1.270000e+16, -0.63, 383.0],
                #kf0=[2.477000e+33, -4.76, 2440.0],
                #efficiencies='H2:2.0 H2O:6.0 AR:0.7 CO:1.5 CO2:2.0 CH4:2.0 C2H6:3.0 HE:0.7',
                #falloff=Troe(A=0.783, T3=74.0, T1=2941.0, T2=6964.0))

## Reaction 126
#reaction('CH4 + H <=> CH3 + H2', [6.140000e+05, 2.5, 9587.0])

## Reaction 127
#reaction('CH4 + OH <=> CH3 + H2O', [5.830000e+04, 2.6, 2190.0])

## Reaction 128
#reaction('CH4 + O <=> CH3 + OH', [1.020000e+09, 1.5, 8600.0])

## Reaction 129
#reaction('CH4 + HO2 <=> CH3 + H2O2', [1.695000e+01, 3.74, 21010.0])

## Reaction 130
#reaction('CH4 + CH2 <=> CH3 + CH3', [2.460000e+06, 2.0, 8270.0])

## Reaction 131
#pdep_arrhenius('CH3 + OH <=> CH2(S) + H2O',
            #[(0.01, 'atm'), 4.936000e+14, -0.669, -445.8],
            #[(0.1, 'atm'), 1.207000e+15, -0.778, -175.6],
            #[(1.0, 'atm'), 5.282000e+17, -1.518, 1772.0],
            #[(10.0, 'atm'), 4.788000e+23, -3.155, 7003.0],
            #[(100.0, 'atm'), 8.433000e+19, -1.962, 8244.0])

## Reaction 132
#pdep_arrhenius('CH3 + OH <=> CH2O + H2',
            #[(0.01, 'atm'), 3.502000e+05, 1.441, -3244.0],
            #[(0.1, 'atm'), 8.854000e+05, 1.327, -2975.0],
            #[(1.0, 'atm'), 1.650000e+07, 0.973, -2010.0],
            #[(10.0, 'atm'), 5.374000e+09, 0.287, 280.0],
            #[(100.0, 'atm'), 9.494000e+18, -2.199, 9769.0])

## Reaction 133
#pdep_arrhenius('CH3 + OH <=> CH2OH + H',
            #[(0.01, 'atm'), 1.621000e+10, 0.965, 3214.0],
            #[(0.1, 'atm'), 1.807000e+10, 0.95, 3247.0],
            #[(1.0, 'atm'), 4.686000e+10, 0.833, 3566.0],
            #[(10.0, 'atm'), 1.525000e+13, 0.134, 5641.0],
            #[(100.0, 'atm'), 3.590000e+14, -0.186, 8601.0])

## Reaction 134
#pdep_arrhenius('CH3 + OH <=> H + CH3O',
            #[(0.01, 'atm'), 1.186000e+09, 1.016, 11940.0],
            #[(0.1, 'atm'), 1.188000e+09, 1.016, 11940.0],
            #[(1.0, 'atm'), 1.230000e+09, 1.011, 11950.0],
            #[(10.0, 'atm'), 1.798000e+09, 0.965, 12060.0],
            #[(100.0, 'atm'), 5.242000e+10, 0.551, 13070.0])

## Reaction 135
#pdep_arrhenius('CH3 + OH <=> HCOH + H2',
            #[(0.01, 'atm'), 8.674000e+08, 0.787, -3046.0],
            #[(0.1, 'atm'), 3.115000e+09, 0.63, -2669.0],
            #[(1.0, 'atm'), 1.557000e+11, 0.156, -1368.0],
            #[(10.0, 'atm'), 1.704000e+21, -2.641, 6412.0],
            #[(100.0, 'atm'), 7.250000e+20, -2.402, 9639.0])

## Reaction 136
#reaction('CH3 + OH <=> CH2 + H2O', [4.293000e+04, 2.568, 3997.8])

## Reaction 137
#reaction('HCOH + OH <=> HCO + H2O', [2.000000e+13, 0.0, 0.0])

## Reaction 138
#reaction('HCOH + H <=> CH2O + H', [2.000000e+14, 0.0, 0.0])

## Reaction 139
#reaction('HCOH + O => CO2 + H + H', [5.000000e+13, 0.0, 0.0])

## Reaction 140
#reaction('HCOH + O => CO + OH + H', [3.000000e+13, 0.0, 0.0])

## Reaction 141
#reaction('HCOH + O2 => CO2 + H + OH', [5.000000e+12, 0.0, 0.0])

## Reaction 142
#reaction('HCOH + O2 <=> CO2 + H2O', [3.000000e+13, 0.0, 0.0])

## Reaction 143
#reaction('CH3 + HO2 <=> CH3O + OH', [1.000000e+12, 0.269, -687.5])

## Reaction 144
#reaction('CH3 + HO2 <=> CH4 + O2', [1.160000e+05, 2.23, -3022.0])

## Reaction 145
#reaction('CH3 + O <=> CH2O + H', [5.540000e+13, 0.05, -136.0])

## Reaction 146
#reaction('CH3 + O2 <=> CH3O + O', [7.546000e+12, 0.0, 28320.0])

## Reaction 147
#reaction('CH3 + O2 <=> CH2O + OH', [2.641000e+00, 3.283, 8105.0])

## Reaction 148
#falloff_reaction('CH3 + O2 (+ M) <=> CH3O2 (+ M)',
                #kf=[7.812000e+09, 0.9, 0.0],
                #kf0=[6.850000e+24, -3.0, 0.0],
                #falloff=Troe(A=0.6, T3=1000.0, T1=70.0, T2=1700.0))

## Reaction 149
#reaction('CH3O2 + CH2O <=> CH3O2H + HCO', [1.990000e+12, 0.0, 11660.0])

## Reaction 150
#reaction('CH4 + CH3O2 <=> CH3 + CH3O2H', [9.600000e-01, 3.77, 17810.0])

## Reaction 151
#reaction('CH3OH + CH3O2 <=> CH2OH + CH3O2H', [1.810000e+12, 0.0, 13710.0])

## Reaction 152
#reaction('CH3O2 + CH3 <=> CH3O + CH3O', [5.080000e+12, 0.0, -1411.0])

## Reaction 153
#reaction('CH3O2 + HO2 <=> CH3O2H + O2', [2.470000e+11, 0.0, -1570.0])

## Reaction 154
#reaction('CH3O2 + H2O2 <=> CH3O2H + HO2', [2.410000e+12, 0.0, 9936.0])

## Reaction 155
#reaction('CH3O2 + CH3O2 => CH2O + CH3OH + O2', [3.110000e+14, -1.61, -1051.0])

## Reaction 156
#reaction('CH3O2 + CH3O2 => O2 + CH3O + CH3O', [1.400000e+16, -1.61, 1860.0])

## Reaction 157
#reaction('CH3O2 + H <=> CH3O + OH', [9.600000e+13, 0.0, 0.0])

## Reaction 158
#reaction('CH3O2 + O <=> CH3O + O2', [3.600000e+13, 0.0, 0.0])

## Reaction 159
#reaction('CH3O2 + OH <=> CH3OH + O2', [6.000000e+13, 0.0, 0.0])

## Reaction 160
#reaction('CH3O2H <=> CH3O + OH', [6.310000e+14, 0.0, 42300.0])

## Reaction 161
#reaction('CH2(S) + N2 <=> CH2 + N2', [1.500000e+13, 0.0, 600.0])

## Reaction 162
#reaction('CH2(S) + AR <=> CH2 + AR', [9.000000e+12, 0.0, 600.0])

## Reaction 163
#reaction('CH2(S) + H <=> CH + H2', [3.000000e+13, 0.0, 0.0])

## Reaction 164
#reaction('CH2(S) + O <=> CO + H2', [1.500000e+13, 0.0, 0.0])

## Reaction 165
#reaction('CH2(S) + O <=> HCO + H', [1.500000e+13, 0.0, 0.0])

## Reaction 166
#reaction('CH2(S) + OH <=> CH2O + H', [3.000000e+13, 0.0, 0.0])

## Reaction 167
#reaction('CH2(S) + H2 <=> CH3 + H', [7.000000e+13, 0.0, 0.0])

## Reaction 168
#reaction('CH2(S) + O2 => H + OH + CO', [2.800000e+13, 0.0, 0.0])

## Reaction 169
#reaction('CH2(S) + O2 <=> CO + H2O', [1.200000e+13, 0.0, 0.0])

## Reaction 170
#reaction('CH2(S) + H2O <=> CH2 + H2O', [3.000000e+13, 0.0, 0.0])

## Reaction 171
#reaction('CH2(S) + CO <=> CH2 + CO', [9.000000e+12, 0.0, 0.0])

## Reaction 172
#reaction('CH2(S) + CO2 <=> CH2 + CO2', [7.000000e+12, 0.0, 0.0])

## Reaction 173
#reaction('CH2(S) + CO2 <=> CH2O + CO', [1.400000e+13, 0.0, 0.0])

## Reaction 174
#falloff_reaction('CH2 + H (+ M) <=> CH3 (+ M)',
                #kf=[2.500000e+16, -0.8, 0.0],
                #kf0=[3.200000e+27, -3.14, 1230.0],
                #efficiencies='H2:2.0 H2O:6.0 AR:0.7 CO:1.5 CO2:2.0 CH4:2.0 C2H6:3.0 HE:0.7',
                #falloff=Troe(A=0.68, T3=78.0, T1=1995.0, T2=5590.0))

## Reaction 175
#reaction('CH2 + O2 <=> HCO + OH', [1.060000e+13, 0.0, 1500.0])

## Reaction 176
#reaction('CH2 + O2 => CO2 + H + H', [2.640000e+12, 0.0, 1500.0])

## Reaction 177
#reaction('CH2 + O => CO + H + H', [5.000000e+13, 0.0, 0.0])

## Reaction 178
#reaction('CH2 + OH <=> CH + H2O', [1.130000e+07, 2.0, 3000.0])

## Reaction 179
#reaction('CH + O2 <=> HCO + O', [3.300000e+13, 0.0, 0.0])

## Reaction 180
#reaction('C + OH <=> CO + H', [5.000000e+13, 0.0, 0.0])

## Reaction 181
#reaction('C + O2 <=> CO + O', [5.000000e+13, 0.0, 0.0])

## Reaction 182
#reaction('CH + H <=> C + H2', [1.100000e+14, 0.0, 0.0])

## Reaction 183
#reaction('CH + O <=> CO + H', [5.700000e+13, 0.0, 0.0])

## Reaction 184
#reaction('CH + OH <=> HCO + H', [3.000000e+13, 0.0, 0.0])

## Reaction 185
#reaction('CH2 + H <=> CH + H2', [3.000000e+13, 0.0, 0.0])

## Reaction 186
#reaction('CH + H2O <=> H + CH2O', [1.774000e+16, -1.22, 23.8])

## Reaction 187
#reaction('CH + CO2 <=> HCO + CO', [1.700000e+12, 0.0, 685.0])

## Reaction 188
#falloff_reaction('CH3 + CH3 (+ M) <=> C2H6 (+ M)',
                #kf=[2.277000e+15, -0.69, 174.9],
                #kf0=[8.054000e+31, -3.75, 981.6],
                #efficiencies='H2O:5.0 CO:2.0 CO2:3.0',
                #falloff=Troe(A=0.0, T3=570.0, T1=1e+30, T2=1e+30))

## Reaction 189
#falloff_reaction('C2H5 + H (+ M) <=> C2H6 (+ M)',
                #kf=[5.210000e+17, -0.99, 1580.0],
                #kf0=[1.990000e+41, -7.08, 6685.0],
                #efficiencies='H2:2.0 H2O:6.0 AR:0.7 CO:1.5 CO2:2.0 CH4:2.0 C2H6:3.0 HE:0.7',
                #falloff=Troe(A=0.842, T3=125.0, T1=2219.0, T2=6882.0))

## Reaction 190
#reaction('C2H6 + H <=> C2H5 + H2', [1.150000e+08, 1.9, 7530.0])

## Reaction 191
#reaction('C2H6 + O <=> C2H5 + OH', [3.550000e+06, 2.4, 5830.0])

## Reaction 192
#reaction('C2H6 + OH <=> C2H5 + H2O', [1.480000e+07, 1.9, 950.0])

## Reaction 193
#reaction('C2H6 + O2 <=> C2H5 + HO2', [6.030000e+13, 0.0, 51870.0])

## Reaction 194
#reaction('C2H6 + CH3 <=> C2H5 + CH4', [5.550000e-04, 4.72, 3231.0])

## Reaction 195
#reaction('C2H6 + HO2 <=> C2H5 + H2O2', [3.460000e+01, 3.61, 16920.0])

## Reaction 196
#reaction('C2H6 + CH3O2 <=> C2H5 + CH3O2H', [1.940000e+01, 3.64, 17100.0])

## Reaction 197
#reaction('C2H6 + CH3O <=> C2H5 + CH3OH', [2.410000e+11, 0.0, 7090.0])

## Reaction 198
#reaction('C2H6 + CH <=> C2H5 + CH2', [1.100000e+14, 0.0, -260.0])

## Reaction 199
#reaction('CH2(S) + C2H6 <=> CH3 + C2H5', [1.200000e+14, 0.0, 0.0])

## Reaction 200
#falloff_reaction('C2H4 + H (+ M) <=> C2H5 (+ M)',
                #kf=[9.569000e+08, 1.463, 1355.0],
                #kf0=[1.419000e+39, -6.642, 5769.0],
                #efficiencies='H2:2.0 H2O:6.0 CH4:2.0 CO:1.5 CO2:2.0 C2H6:3.0 AR:0.7',
                #falloff=Troe(A=-0.569, T3=299.0, T1=-9147.0, T2=152.4))

## Reaction 201
#reaction('H2 + CH3O2 <=> H + CH3O2H', [1.500000e+14, 0.0, 26030.0])

## Reaction 202
#reaction('H2 + C2H5O2 <=> H + C2H5O2H', [1.500000e+14, 0.0, 26030.0])

## Reaction 203
#reaction('C2H4 + C2H4 <=> C2H5 + C2H3', [4.820000e+14, 0.0, 71530.0])

## Reaction 204
#reaction('CH3 + C2H5 <=> CH4 + C2H4', [1.180000e+04, 2.45, -2921.0])

## Reaction 205
#pdep_arrhenius('CH3 + CH3 <=> H + C2H5',
            #[(0.01, 'atm'), 4.740000e+12, 0.105, 10664.3],
            #[(0.1, 'atm'), 2.570000e+13, -0.096, 11406.1],
            #[(1.0, 'atm'), 3.100000e+14, -0.362, 13372.5],
            #[(10.0, 'atm'), 2.150000e+10, 0.885, 13532.5],
            #[(100.0, 'atm'), 1.032000e+02, 3.23, 11236.1])

## Reaction 206
#reaction('C2H5 + H <=> C2H4 + H2', [2.000000e+12, 0.0, 0.0])

## Reaction 207
#reaction('C2H5 + O <=> CH3CHO + H', [1.100000e+14, 0.0, 0.0])

## Reaction 208
#reaction('C2H5 + HO2 <=> C2H5O + OH', [1.100000e+13, 0.0, 0.0])

## Reaction 209
#reaction('CH3O2 + C2H5 <=> CH3O + C2H5O', [8.000000e+12, 0.0, -1000.0])

## Reaction 210
#reaction('C2H5O + O2 <=> CH3CHO + HO2', [4.280000e+10, 0.0, 1097.0])

## Reaction 211
#reaction('CH3 + CH2O <=> C2H5O', [3.000000e+11, 0.0, 6336.0])

## Reaction 212
#reaction('CH3CHO + H <=> C2H5O', [4.610000e+07, 1.71, 7090.0])

## Reaction 213
#reaction('C2H5O2 + CH2O <=> C2H5O2H + HCO', [1.990000e+12, 0.0, 11660.0])

## Reaction 214
#reaction('CH4 + C2H5O2 <=> CH3 + C2H5O2H', [1.810000e+11, 0.0, 18480.0])

## Reaction 215
#reaction('CH3OH + C2H5O2 <=> CH2OH + C2H5O2H', [1.810000e+12, 0.0, 13710.0])

## Reaction 216
#reaction('C2H5O2 + HO2 <=> C2H5O2H + O2', [1.750000e+10, 0.0, -3275.0])

## Reaction 217
#reaction('C2H6 + C2H5O2 <=> C2H5 + C2H5O2H', [8.600000e+00, 3.76, 17200.0])

## Reaction 218
#reaction('C2H5O2H <=> C2H5O + OH', [6.310000e+14, 0.0, 42300.0])

## Reaction 219
#pdep_arrhenius('C2H5 + O2 <=> C2H5O2',
            #[(0.04, 'atm'), 3.398000e+53, -13.9, 9279.0],
            #[(1.0, 'atm'), 9.362000e+59, -15.28, 14240.0],
            #[(10.0, 'atm'), 1.262000e+60, -14.91, 16240.0])

## Reaction 220
#pdep_arrhenius('C2H5 + O2 <=> C2H4O2H',
            #[(0.04, 'atm'), 2.103000e+34, -9.01, 5444.0],
            #[(1.0, 'atm'), 4.884000e+33, -8.31, 7710.0],
            #[(10.0, 'atm'), 1.705000e+45, -11.49, 14590.0])

## Reaction 221
#pdep_arrhenius('C2H5 + O2 <=> C2H4 + HO2',
            #[(0.04, 'atm'), 2.094000e+09, 0.49, -391.4],
            #[(1.0, 'atm'), 1.843000e+07, 1.13, -720.6],
            #[(10.0, 'atm'), 7.561000e+14, -1.01, 4749.0],
            #options='duplicate')

## Reaction 222
#reaction('C2H5 + O2 <=> C2H4 + HO2', [6.609000e+00, 3.51, 14160.0],
        #options='duplicate')

## Reaction 223
#pdep_arrhenius('C2H5 + O2 <=> C2H4O1-2 + OH',
            #[(0.04, 'atm'), 1.303000e+03, 1.93, -502.7],
            #[(1.0, 'atm'), 2.438000e+02, 2.18, -62.5],
            #[(10.0, 'atm'), 4.621000e+09, 0.15, 5409.0])

## Reaction 224
#pdep_arrhenius('C2H5 + O2 <=> CH3CHO + OH',
            #[(0.04, 'atm'), 4.908000e-06, 4.76, 254.3],
            #[(1.0, 'atm'), 6.803000e-02, 3.57, 2643.0],
            #[(10.0, 'atm'), 8.265000e+02, 2.41, 5285.0])

## Reaction 225
#pdep_arrhenius('C2H4O2H <=> C2H5O2',
            #[(0.04, 'atm'), 2.653000e-16, 6.96, 2396.0],
            #[(1.0, 'atm'), 1.064000e+41, -10.1, 26030.0],
            #[(10.0, 'atm'), 1.203000e+36, -8.13, 27020.0])

## Reaction 226
#pdep_arrhenius('C2H5O2 <=> CH3CHO + OH',
            #[(0.04, 'atm'), 1.237000e+35, -9.42, 36360.0],
            #[(1.0, 'atm'), 1.687000e+36, -9.22, 38700.0],
            #[(10.0, 'atm'), 2.520000e+41, -10.2, 43710.0])

## Reaction 227
#pdep_arrhenius('C2H5O2 <=> C2H4 + HO2',
            #[(0.04, 'atm'), 1.782000e+32, -7.1, 32840.0],
            #[(1.0, 'atm'), 2.701000e+37, -8.47, 35840.0],
            #[(10.0, 'atm'), 1.980000e+38, -8.46, 37900.0])

## Reaction 228
#pdep_arrhenius('C2H5O2 <=> C2H4O1-2 + OH',
            #[(0.04, 'atm'), 5.778000e+45, -11.9, 4112.0],
            #[(1.0, 'atm'), 1.916000e+43, -10.75, 42400.0],
            #[(10.0, 'atm'), 3.965000e+43, -10.46, 45580.0])

## Reaction 229
#pdep_arrhenius('C2H4O2H <=> C2H4O1-2 + OH',
            #[(0.04, 'atm'), 8.959000e+38, -9.4, 20660.0],
            #[(1.0, 'atm'), 1.224000e+37, -8.32, 21460.0],
            #[(10.0, 'atm'), 8.848000e+30, -6.08, 20660.0])

## Reaction 230
#pdep_arrhenius('C2H4O2H <=> C2H4 + HO2',
            #[(0.04, 'atm'), 3.918000e+40, -10.2, 22250.0],
            #[(1.0, 'atm'), 6.825000e+40, -9.61, 23840.0],
            #[(10.0, 'atm'), 3.980000e+34, -7.25, 23250.0])

## Reaction 231
#pdep_arrhenius('C2H4O2H <=> CH3CHO + OH',
            #[(0.04, 'atm'), 5.819000e+26, -7.97, 20860.0],
            #[(1.0, 'atm'), 5.520000e+34, -9.88, 26230.0],
            #[(10.0, 'atm'), 1.188000e+34, -9.02, 29210.0])

## Reaction 232
#reaction('C2H4O1-2 <=> CH3 + HCO', [3.630000e+13, 0.0, 57200.0])

## Reaction 233
#reaction('C2H4O1-2 <=> CH3CHO', [7.407000e+12, 0.0, 53800.0])

## Reaction 234
#reaction('C2H4O1-2 + OH <=> C2H3O1-2 + H2O', [1.780000e+13, 0.0, 3610.0])

## Reaction 235
#reaction('C2H4O1-2 + H <=> C2H3O1-2 + H2', [8.000000e+13, 0.0, 9680.0])

## Reaction 236
#reaction('C2H4O1-2 + HO2 <=> C2H3O1-2 + H2O2', [1.130000e+13, 0.0, 30430.0])

## Reaction 237
#reaction('C2H4O1-2 + CH3O2 <=> C2H3O1-2 + CH3O2H', [1.130000e+13, 0.0, 30430.0])

## Reaction 238
#reaction('C2H4O1-2 + C2H5O2 <=> C2H3O1-2 + C2H5O2H', [1.130000e+13, 0.0, 30430.0])

## Reaction 239
#reaction('C2H4O1-2 + CH3 <=> C2H3O1-2 + CH4', [1.070000e+12, 0.0, 11830.0])

## Reaction 240
#reaction('C2H4O1-2 + CH3O <=> C2H3O1-2 + CH3OH', [1.200000e+11, 0.0, 6750.0])

## Reaction 241
#reaction('C2H3O1-2 <=> CH3CO', [8.500000e+14, 0.0, 14000.0])

## Reaction 242
#reaction('C2H3O1-2 <=> CH2CHO', [1.000000e+14, 0.0, 14000.0])

## Reaction 243
#falloff_reaction('CH3CHO (+ M) <=> CH3 + HCO (+ M)',
                #kf=[2.450000e+22, -1.74, 86355.0],
                #kf0=[1.030000e+59, -11.3, 95912.5],
                #falloff=Troe(A=0.00249, T3=718.1, T1=6.089, T2=3780.0))

## Reaction 244
#falloff_reaction('CH3CHO (+ M) <=> CH4 + CO (+ M)',
                #kf=[2.720000e+21, -1.74, 86355.0],
                #kf0=[1.144000e+58, -11.3, 95912.5],
                #falloff=Troe(A=0.00249, T3=718.1, T1=6.089, T2=3780.0))

## Reaction 245
#reaction('CH3CHO + H <=> CH3CO + H2', [1.310000e+05, 2.58, 1220.0])

## Reaction 246
#reaction('CH3CHO + H <=> CH2CHO + H2', [2.720000e+03, 3.1, 5210.0])

## Reaction 247
#reaction('CH3CHO + O <=> CH3CO + OH', [5.940000e+12, 0.0, 1868.0])

## Reaction 248
#reaction('CH3CHO + OH <=> CH3CO + H2O', [3.370000e+12, 0.0, -619.0])

## Reaction 249
#reaction('CH3CHO + O2 <=> CH3CO + HO2', [3.010000e+13, 0.0, 39150.0])

## Reaction 250
#reaction('CH3CHO + CH3 <=> CH3CO + CH4', [7.080000e-04, 4.58, 1966.0])

## Reaction 251
#reaction('CH3CHO + HO2 <=> CH3CO + H2O2', [3.010000e+12, 0.0, 11920.0])

## Reaction 252
#reaction('CH3O2 + CH3CHO <=> CH3O2H + CH3CO', [3.010000e+12, 0.0, 11920.0])

## Reaction 253
#reaction('CH3CHO + CH3CO3 <=> CH3CO + CH3CO3H', [3.010000e+12, 0.0, 11920.0])

## Reaction 254
#reaction('CH3CHO + OH <=> CH3 + HOCHO', [3.000000e+15, -1.076, 0.0])

## Reaction 255
#reaction('CH3CHO + OH <=> CH2CHO + H2O', [1.720000e+05, 2.4, 815.0])

## Reaction 256
#falloff_reaction('CH3CO (+ M) <=> CH3 + CO (+ M)',
                #kf=[1.070000e+12, 0.63, 16900.0],
                #kf0=[5.650000e+18, -0.97, 14600.0],
                #falloff=Troe(A=0.629, T3=8730000000.0, T1=5.52, T2=76000000.0))

## Reaction 257
#reaction('CH3CO + H <=> CH2CO + H2', [2.000000e+13, 0.0, 0.0])

## Reaction 258
#reaction('CH3CO + O <=> CH2CO + OH', [2.000000e+13, 0.0, 0.0])

## Reaction 259
#reaction('CH3CO + CH3 <=> CH2CO + CH4', [5.000000e+13, 0.0, 0.0])

## Reaction 260
#reaction('CH3CO + O2 <=> CH3CO3', [1.200000e+11, 0.0, -1100.0])

## Reaction 261
#reaction('CH3CO3 + HO2 <=> CH3CO3H + O2', [1.750000e+10, 0.0, -3275.0])

## Reaction 262
#reaction('H2O2 + CH3CO3 <=> HO2 + CH3CO3H', [2.410000e+12, 0.0, 9936.0])

## Reaction 263
#reaction('CH4 + CH3CO3 <=> CH3 + CH3CO3H', [1.810000e+11, 0.0, 18480.0])

## Reaction 264
#reaction('CH2O + CH3CO3 <=> HCO + CH3CO3H', [1.990000e+12, 0.0, 11660.0])

## Reaction 265
#reaction('C2H6 + CH3CO3 <=> C2H5 + CH3CO3H', [1.700000e+13, 0.0, 20460.0])

## Reaction 266
#reaction('CH3CO3H <=> CH3CO2 + OH', [5.010000e+14, 0.0, 40150.0])

## Reaction 267
#three_body_reaction('CH3CO2 + M <=> CH3 + CO2 + M', [4.400000e+15, 0.0, 10500.0])

## Reaction 268
#falloff_reaction('CH2CHO (+ M) <=> CH2CO + H (+ M)',
                #kf=[1.430000e+15, -0.15, 45600.0],
                #kf0=[6.000000e+29, -3.8, 43423.9],
                #falloff=Troe(A=0.985, T3=393.0, T1=9800000000.0, T2=5000000000.0))

## Reaction 269
#falloff_reaction('CH2CHO (+ M) <=> CH3 + CO (+ M)',
                #kf=[2.930000e+12, 0.29, 40300.0],
                #kf0=[9.520000e+33, -5.07, 41300.0],
                #falloff=Troe(A=7.13e-17, T3=1150.0, T1=4990000000.0, T2=1790000000.0))

## Reaction 270
#pdep_arrhenius('CH2CHO + O2 <=> O2CH2CHO',
            #[(0.01, 'atm'), 1.580000e+77, -21.9, 19350.0],
            #[(0.1, 'atm'), 3.880000e+69, -18.84, 19240.0],
            #[(1.0, 'atm'), 7.800000e+59, -15.4, 17650.0],
            #[(10.0, 'atm'), 3.050000e+50, -12.2, 15630.0])

## Reaction 271
#pdep_arrhenius('CH2CHO + O2 <=> CH2CO + HO2',
            #[(0.01, 'atm'), 1.880000e+05, 2.37, 23730.0],
            #[(0.1, 'atm'), 1.880000e+05, 2.37, 27370.0],
            #[(1.0, 'atm'), 2.510000e+05, 2.33, 23800.0],
            #[(10.0, 'atm'), 7.050000e+07, 1.63, 25290.0])

## Reaction 272
#pdep_arrhenius('CH2CHO + O2 => CH2O + CO + OH',
            #[(0.01, 'atm'), 2.680000e+17, -1.84, 6530.0],
            #[(0.1, 'atm'), 1.520000e+20, -2.58, 8980.0],
            #[(1.0, 'atm'), 1.650000e+19, -2.22, 10340.0],
            #[(10.0, 'atm'), 8.953000e+13, -0.6, 10120.0])

## Reaction 273
#pdep_arrhenius('CH2CHO + O2 <=> HO2CH2CO',
            #[(0.01, 'atm'), 3.640000e+65, -21.87, 19020.0],
            #[(0.1, 'atm'), 3.640000e+58, -19.0, 19090.0],
            #[(1.0, 'atm'), 6.650000e+48, -15.55, 17460.0],
            #[(10.0, 'atm'), 4.800000e+38, -12.14, 14960.0])

## Reaction 274
#pdep_arrhenius('O2CH2CHO <=> HO2CH2CO',
            #[(0.01, 'atm'), 8.270000e+30, -6.65, 24500.0],
            #[(0.1, 'atm'), 1.730000e+26, -4.99, 23760.0],
            #[(1.0, 'atm'), 9.030000e+19, -2.92, 22170.0],
            #[(10.0, 'atm'), 1.430000e+16, -1.67, 21210.0])

## Reaction 275
#pdep_arrhenius('O2CH2CHO <=> CH2CO + HO2',
            #[(0.01, 'atm'), 2.050000e+40, -13.31, 52150.0],
            #[(0.1, 'atm'), 5.720000e+45, -14.0, 52200.0],
            #[(1.0, 'atm'), 4.160000e+55, -15.76, 55080.0],
            #[(10.0, 'atm'), 1.120000e+61, -16.04, 60010.0])

## Reaction 276
#pdep_arrhenius('HO2CH2CO => CO + CH2O + OH',
            #[(0.01, 'atm'), 2.360000e+17, -2.95, 8100.0],
            #[(0.1, 'atm'), 2.380000e+18, -2.95, 8100.0],
            #[(1.0, 'atm'), 2.510000e+19, -2.95, 8110.0],
            #[(10.0, 'atm'), 4.160000e+20, -3.02, 8240.0])

## Reaction 277
#pdep_arrhenius('HO2CH2CO <=> CH2CO + HO2',
            #[(0.01, 'atm'), 1.120000e+07, -3.76, 21680.0],
            #[(0.1, 'atm'), 1.100000e+08, -3.76, 21680.0],
            #[(1.0, 'atm'), 9.200000e+08, -3.73, 21630.0],
            #[(10.0, 'atm'), 2.090000e+09, -3.55, 21220.0])

## Reaction 278
#falloff_reaction('CH2 + CO (+ M) <=> CH2CO (+ M)',
                #kf=[8.100000e+11, 0.0, 0.0],
                #kf0=[2.690000e+33, -5.11, 7095.0],
                #efficiencies='H2:2.0 H2O:6.0 AR:0.7 CO:1.5 CO2:2.0 CH4:2.0 C2H6:3.0 HE:0.7',
                #falloff=Troe(A=0.5907, T3=275.0, T1=1226.0, T2=5185.0))

## Reaction 279
#falloff_reaction('CH3CO (+ M) <=> CH2CO + H (+ M)',
                #kf=[9.413000e+07, 1.917, 44987.2],
                #kf0=[1.516000e+51, -10.27, 55390.0],
                #falloff=Troe(A=0.6009, T3=8103000000.0, T1=667.7, T2=5000000000.0))

## Reaction 280
#reaction('CH2CO + H <=> HCCO + H2', [1.401000e+15, -0.171, 8783.2])

## Reaction 281
#reaction('CH2CO + H <=> CH3 + CO', [7.704000e+13, -0.171, 4183.2])

## Reaction 282
#reaction('CH2CO + O <=> CH2 + CO2', [1.750000e+12, 0.0, 1350.0])

## Reaction 283
#reaction('CH2CO + O <=> HCCO + OH', [1.000000e+13, 0.0, 8000.0])

## Reaction 284
#reaction('CH2CO + OH <=> HCCO + H2O', [1.000000e+13, 0.0, 2000.0])

## Reaction 285
#reaction('CH2CO + OH <=> CH2OH + CO', [2.000000e+12, 0.0, -1010.0])

## Reaction 286
#reaction('CH2CO + CH3 <=> C2H5 + CO', [4.769000e+04, 2.312, 9468.0])

## Reaction 287
#reaction('CH2(S) + CH2CO <=> C2H4 + CO', [1.600000e+14, 0.0, 0.0])

## Reaction 288
#reaction('HCCO + OH => H2 + CO + CO', [1.000000e+14, 0.0, 0.0])

## Reaction 289
#reaction('HCCO + O => H + CO + CO', [8.000000e+13, 0.0, 0.0])

## Reaction 290
#reaction('HCCO + H <=> CH2(S) + CO', [1.000000e+14, 0.0, 0.0])

## Reaction 291
#reaction('HCCO + O2 => OH + CO + CO', [1.910000e+11, -0.02, 1020.0])

## Reaction 292
#reaction('HCCO + O2 => CO2 + CO + H', [4.780000e+12, -0.142, 1150.0])

## Reaction 293
#three_body_reaction('CH + CO + M <=> HCCO + M', [7.570000e+22, -1.9, 0.0])

## Reaction 294
#reaction('CH + CH2O <=> H + CH2CO', [9.460000e+13, 0.0, -515.0])

## Reaction 295
#reaction('CH + HCCO <=> CO + C2H2', [5.000000e+13, 0.0, 0.0])

## Reaction 296
#falloff_reaction('C2H3 + H (+ M) <=> C2H4 (+ M)',
                #kf=[6.080000e+12, 0.27, 280.0],
                #kf0=[1.400000e+30, -3.86, 3320.0],
                #efficiencies='H2:2.0 H2O:6.0 AR:0.7 CO:1.5 CO2:2.0 CH4:2.0 C2H6:3.0 HE:0.7',
                #falloff=Troe(A=0.782, T3=207.5, T1=2663.0, T2=6095.0))

## Reaction 297
#falloff_reaction('C2H4 (+ M) <=> H2 + H2CC (+ M)',
                #kf=[8.000000e+12, 0.44, 88770.0],
                #kf0=[7.000000e+50, -9.31, 99860.0],
                #efficiencies='H2:2.0 H2O:6.0 CH4:2.0 CO:1.5 CO2:2.0 C2H6:3.0 AR:0.7',
                #falloff=Troe(A=0.7345, T3=180.0, T1=1035.0, T2=5417.0))

## Reaction 298
#reaction('C2H4 + H <=> C2H3 + H2', [5.070000e+07, 1.93, 12950.0])

## Reaction 299
#reaction('C2H4 + O <=> CH3 + HCO', [7.453000e+06, 1.88, 183.0])

## Reaction 300
#reaction('C2H4 + O <=> CH2CHO + H', [6.098000e+06, 1.88, 183.0])

## Reaction 301
#reaction('C2H4 + OH <=> C2H3 + H2O', [2.230000e+04, 2.745, 2215.5])

## Reaction 302
#pdep_arrhenius('C2H4 + OH <=> CH3 + CH2O',
            #[(0.01, 'atm'), 5.350000e+00, 2.92, -1732.7],
            #[(0.025, 'atm'), 3.190000e+01, 2.71, -1172.3],
            #[(0.1, 'atm'), 5.550000e+02, 2.36, -180.8],
            #[(1.0, 'atm'), 1.780000e+05, 1.68, 2060.5],
            #[(10.0, 'atm'), 2.370000e+09, 0.56, 6006.7],
            #[(100.0, 'atm'), 2.760000e+13, -0.5, 11455.1])

## Reaction 303
#pdep_arrhenius('C2H4 + OH <=> CH3CHO + H',
            #[(0.01, 'atm'), 2.370000e-07, 5.3, -2050.6],
            #[(0.025, 'atm'), 8.730000e-05, 4.57, -618.0],
            #[(0.1, 'atm'), 4.030000e-01, 3.54, 1881.7],
            #[(1.0, 'atm'), 2.380000e-02, 3.91, 1722.7],
            #[(10.0, 'atm'), 8.250000e+08, 1.01, 10507.3],
            #[(100.0, 'atm'), 6.800000e+09, 0.81, 13867.3])

## Reaction 304
#pdep_arrhenius('C2H4 + OH <=> C2H3OH + H',
            #[(0.01, 'atm'), 1.040000e+04, 2.6, 4121.0],
            #[(0.025, 'atm'), 1.070000e+04, 2.6, 4129.0],
            #[(0.1, 'atm'), 1.520000e+04, 2.56, 4238.3],
            #[(1.0, 'atm'), 3.190000e+05, 2.19, 5255.6],
            #[(10.0, 'atm'), 1.940000e+08, 1.43, 7828.8],
            #[(100.0, 'atm'), 8.550000e+10, 0.75, 11490.8])

## Reaction 305
#pdep_arrhenius('C2H4 + OH <=> PC2H4OH',
            #[(0.01, 'atm'), 1.740000e+43, -10.461, 7698.7],
            #[(0.025, 'atm'), 3.250000e+37, -8.629, 5214.7],
            #[(0.1, 'atm'), 1.840000e+35, -7.75, 4908.9],
            #[(1.0, 'atm'), 2.560000e+36, -7.752, 6946.1],
            #[(10.0, 'atm'), 3.700000e+33, -6.573, 7605.9],
            #[(100.0, 'atm'), 1.120000e+26, -4.101, 5757.0])

## Reaction 306
#reaction('C2H3OH + O2 <=> CH2CHO + HO2', [5.310000e+11, 0.21, 39830.0])

## Reaction 307
#reaction('C2H3OH + O <=> CH2CHO + OH', [1.875000e+06, 1.9, -860.0])

## Reaction 308
#reaction('C2H3OH + OH <=> CH2CHO + H2O', [3.330000e+09, 1.1, 540.5])

## Reaction 309
#reaction('C2H3OH + CH3 <=> CH2CHO + CH4', [2.030000e-08, 5.9, 1052.0])

## Reaction 310
#reaction('C2H3OH + CH3O2 <=> CH2CHO + CH3O2H', [3.400000e+03, 2.5, 8922.0])

## Reaction 311
#reaction('C2H3OH + H <=> CH2CHO + H2', [1.480000e+03, 3.077, 7220.0])

## Reaction 312
#reaction('C2H3OH + H <=> C2H2OH + H2', [2.470000e+07, 2.03, 15200.0])

## Reaction 313
#reaction('C2H3OH + H <=> PC2H4OH', [3.010000e+08, 1.577, 3670.0])

## Reaction 314
#reaction('C2H3OH + HO2 <=> CH3CHO + HO2', [1.490000e+05, 1.67, 6810.0])

## Reaction 315
#pdep_arrhenius('C2H3OH <=> CH3CHO',
            #[(0.1, 'atm'), 7.420000e+46, -10.56, 67420.0],
            #[(1.0, 'atm'), 4.420000e+42, -9.09, 67069.2],
            #[(100.0, 'atm'), 2.900000e+27, -4.35, 61612.9])

## Reaction 316
#reaction('C2H4 + O2 <=> C2H3 + HO2', [4.220000e+13, 0.0, 57623.1])

## Reaction 317
#reaction('C2H4 + CH3O <=> C2H3 + CH3OH', [1.200000e+11, 0.0, 6750.0])

## Reaction 318
#reaction('C2H4 + CH3O2 <=> C2H3 + CH3O2H', [8.590000e+00, 3.754, 27132.0])

## Reaction 319
#reaction('C2H4 + C2H5O2 <=> C2H3 + C2H5O2H', [8.590000e+00, 3.754, 27132.0])

## Reaction 320
#reaction('C2H4 + CH3CO3 <=> C2H3 + CH3CO3H', [1.130000e+13, 0.0, 30430.0])

## Reaction 321
#reaction('C2H4 + CH3O2 <=> C2H4O1-2 + CH3O', [2.820000e+12, 0.0, 17110.0])

## Reaction 322
#reaction('C2H4 + C2H5O2 <=> C2H4O1-2 + C2H5O', [2.820000e+12, 0.0, 17110.0])

## Reaction 323
#reaction('C2H4 + HO2 <=> C2H4O1-2 + OH', [3.345000e+12, 0.0, 17190.0])

## Reaction 324
#reaction('CH + CH4 <=> C2H4 + H', [6.000000e+13, 0.0, 0.0])

## Reaction 325
#reaction('CH2(S) + CH3 <=> C2H4 + H', [2.000000e+13, 0.0, 0.0])

## Reaction 326
#falloff_reaction('C2H2 + H (+ M) <=> C2H3 (+ M)',
                #kf=[1.710000e+10, 1.266, 2709.0],
                #kf0=[6.346000e+31, -4.664, 3780.0],
                #efficiencies='H2:2.0 H2O:6.0 AR:0.7 CO:1.5 CO2:2.0 CH4:2.0 C2H6:3.0 HE:0.7',
                #falloff=Troe(A=0.788, T3=-10200.0, T1=1e-30))

## Reaction 327
#pdep_arrhenius('C2H3 + O2 <=> C2H3OO',
            #[(0.01, 'atm'), 1.550000e+24, -5.45, 9662.0],
            #[(0.1, 'atm'), 3.480000e+56, -15.01, 19160.0],
            #[(0.316, 'atm'), 1.250000e+64, -16.97, 21290.0],
            #[(1.0, 'atm'), 3.340000e+61, -15.79, 20150.0],
            #[(3.16, 'atm'), 7.340000e+53, -13.11, 17300.0],
            #[(10.0, 'atm'), 4.160000e+48, -11.21, 16000.0],
            #[(31.6, 'atm'), 2.330000e+43, -9.38, 14810.0],
            #[(100.0, 'atm'), 3.410000e+39, -8.04, 14360.0],
            #options='duplicate')

## Reaction 328
#pdep_arrhenius('C2H3 + O2 <=> C2H3OO',
            #[(0.01, 'atm'), 1.780000e-09, 4.15, -4707.0],
            #[(0.1, 'atm'), 2.360000e+22, -4.52, 2839.0],
            #[(0.316, 'atm'), 2.000000e+26, -5.43, 2725.0],
            #[(1.0, 'atm'), 6.130000e+28, -5.89, 3154.0],
            #[(3.16, 'atm'), 2.140000e+29, -5.8, 3520.0],
            #[(10.0, 'atm'), 3.480000e+28, -5.37, 3636.0],
            #[(31.6, 'atm'), 3.320000e+27, -4.95, 3610.0],
            #[(100.0, 'atm'), 1.030000e+27, -4.72, 3680.0],
            #options='duplicate')

## Reaction 329
#pdep_arrhenius('C2H3 + O2 <=> CHCHO + OH',
            #[(0.01, 'atm'), 3.910000e+11, -0.11, 2131.0],
            #[(0.1, 'atm'), 1.130000e+09, 0.55, 46.0],
            #[(0.316, 'atm'), 8.460000e+08, 0.56, 0.7],
            #[(1.0, 'atm'), 2.750000e+14, -1.83, 4.6],
            #[(3.16, 'atm'), 2.580000e+20, -2.84, 7530.0],
            #[(10.0, 'atm'), 9.180000e+14, -2.26, -0.4],
            #[(31.6, 'atm'), 6.110000e+25, -4.21, 13050.0],
            #[(100.0, 'atm'), 1.650000e+30, -5.35, 18430.0],
            #options='duplicate')

## Reaction 330
#pdep_arrhenius('C2H3 + O2 <=> CHCHO + OH',
            #[(0.01, 'atm'), 9.910000e+11, -0.66, -0.6],
            #[(0.1, 'atm'), 6.940000e+14, -1.16, 4542.0],
            #[(0.316, 'atm'), 2.790000e+13, -0.72, 3479.0],
            #[(1.0, 'atm'), 4.990000e+11, -0.14, 1995.0],
            #[(3.16, 'atm'), 2.350000e+10, 0.23, 1573.0],
            #[(10.0, 'atm'), 1.700000e+14, -0.82, 4450.0],
            #[(31.6, 'atm'), 1.420000e+11, 0.05, 3774.0],
            #[(100.0, 'atm'), 3.170000e+11, -0.02, 5338.0],
            #options='duplicate')

## Reaction 331
#pdep_arrhenius('C2H3 + O2 <=> CH2CHO + O',
            #[(0.01, 'atm'), 7.880000e+20, -2.67, 6742.0],
            #[(0.1, 'atm'), 7.720000e+20, -2.67, 6713.0],
            #[(0.316, 'atm'), 9.870000e+20, -2.7, 6724.0],
            #[(1.0, 'atm'), 7.100000e+20, -2.65, 6489.0],
            #[(3.16, 'atm'), 4.500000e+20, -2.53, 6406.0],
            #[(10.0, 'atm'), 1.760000e+23, -3.22, 8697.0],
            #[(31.6, 'atm'), 3.140000e+25, -3.77, 11530.0],
            #[(100.0, 'atm'), 1.020000e+26, -3.8, 13910.0],
            #options='duplicate')

## Reaction 332
#pdep_arrhenius('C2H3 + O2 <=> CH2CHO + O',
            #[(0.01, 'atm'), 1.360000e+10, 0.62, -277.6],
            #[(0.1, 'atm'), 1.420000e+10, 0.62, -247.7],
            #[(0.316, 'atm'), 1.660000e+10, 0.6, -162.5],
            #[(1.0, 'atm'), 2.020000e+10, 0.58, 38.4],
            #[(3.16, 'atm'), 9.750000e+09, 0.67, 248.0],
            #[(10.0, 'atm'), 7.340000e+09, 0.72, 778.1],
            #[(31.6, 'atm'), 1.570000e+09, 0.92, 1219.0],
            #[(100.0, 'atm'), 7.850000e+07, 1.28, 1401.0],
            #options='duplicate')

## Reaction 333
#pdep_arrhenius('C2H3 + O2 <=> C2H2 + HO2',
            #[(0.01, 'atm'), 1.080000e+07, 1.28, 3322.0],
            #[(0.1, 'atm'), 7.750000e+06, 1.33, 3216.0],
            #[(0.316, 'atm'), 1.210000e+07, 1.27, 3311.0],
            #[(1.0, 'atm'), 2.150000e+07, 1.19, 3367.0],
            #[(3.16, 'atm'), 1.130000e+08, 1.0, 3695.0],
            #[(10.0, 'atm'), 1.310000e+11, 0.12, 5872.0],
            #[(31.6, 'atm'), 1.190000e+09, 0.82, 5617.0],
            #[(100.0, 'atm'), 1.060000e+17, -1.45, 12230.0],
            #options='duplicate')

## Reaction 334
#pdep_arrhenius('C2H3 + O2 <=> C2H2 + HO2',
            #[(0.01, 'atm'), 4.760000e+01, 2.75, -796.4],
            #[(0.1, 'atm'), 5.160000e+01, 2.73, -768.3],
            #[(0.316, 'atm'), 5.550000e+01, 2.73, -658.5],
            #[(1.0, 'atm'), 4.600000e+01, 2.76, -492.8],
            #[(3.16, 'atm'), 3.750000e+00, 3.07, -601.0],
            #[(10.0, 'atm'), 5.480000e+00, 3.07, 85.7],
            #[(31.6, 'atm'), 4.470000e+08, 0.0, 955.0],
            #[(100.0, 'atm'), 2.020000e+01, 2.94, 1847.0],
            #options='duplicate')

## Reaction 335
#pdep_arrhenius('C2H3 + O2 <=> CHOCHO + H',
            #[(0.01, 'atm'), 4.790000e+14, -1.03, 912.0],
            #[(0.1, 'atm'), 5.030000e+14, -1.04, 922.5],
            #[(0.316, 'atm'), 6.430000e+14, -1.07, 982.9],
            #[(1.0, 'atm'), 3.730000e+15, -1.29, 1441.0],
            #[(3.16, 'atm'), 2.440000e+18, -2.13, 3234.0],
            #[(10.0, 'atm'), 1.300000e+15, -1.09, 2393.0],
            #[(31.6, 'atm'), 3.570000e+33, -6.5, 14910.0],
            #[(100.0, 'atm'), 3.280000e+31, -5.76, 16250.0],
            #options='duplicate')

## Reaction 336
#pdep_arrhenius('C2H3 + O2 <=> CHOCHO + H',
            #[(0.01, 'atm'), 2.800000e-04, 4.04, -7019.0],
            #[(0.1, 'atm'), 3.450000e-04, 4.01, -6978.0],
            #[(0.316, 'atm'), 9.730000e-04, 3.89, -6768.0],
            #[(1.0, 'atm'), 4.980000e-01, 3.15, -5496.0],
            #[(3.16, 'atm'), 1.340000e+05, 1.67, -2931.0],
            #[(10.0, 'atm'), 4.500000e+15, -3.08, -4836.0],
            #[(31.6, 'atm'), 3.840000e+10, 0.22, 941.3],
            #[(100.0, 'atm'), 2.750000e+08, 0.83, 857.6],
            #options='duplicate')

## Reaction 337
#pdep_arrhenius('C2H3 + O2 <=> CH2CO + OH',
            #[(0.01, 'atm'), 8.660000e+02, 2.41, 6061.0],
            #[(0.1, 'atm'), 8.910000e+02, 2.41, 6078.0],
            #[(0.316, 'atm'), 9.430000e+02, 2.4, 6112.0],
            #[(1.0, 'atm'), 1.060000e+03, 2.39, 6180.0],
            #[(3.16, 'atm'), 1.090000e+03, 2.38, 6179.0],
            #[(10.0, 'atm'), 1.390000e+03, 2.36, 6074.0],
            #[(31.6, 'atm'), 2.490000e+06, 1.42, 8480.0],
            #[(100.0, 'atm'), 1.660000e+10, 0.36, 12010.0],
            #options='duplicate')

## Reaction 338
#pdep_arrhenius('C2H3 + O2 <=> CH2CO + OH',
            #[(0.01, 'atm'), 1.820000e-01, 3.12, 1331.0],
            #[(0.1, 'atm'), 2.070000e-01, 3.11, 1383.0],
            #[(0.316, 'atm'), 2.710000e-01, 3.08, 1496.0],
            #[(1.0, 'atm'), 5.260000e-01, 3.01, 1777.0],
            #[(3.16, 'atm'), 1.370000e+00, 2.9, 2225.0],
            #[(10.0, 'atm'), 4.190000e-01, 2.93, 2052.0],
            #[(31.6, 'atm'), 1.190000e-04, 4.21, 2043.0],
            #[(100.0, 'atm'), 1.300000e-03, 3.97, 3414.0],
            #options='duplicate')

## Reaction 339
#pdep_arrhenius('C2H3 + O2 <=> CH2O + HCO',
            #[(0.01, 'atm'), 2.490000e+36, -7.6, 12640.0],
            #[(0.1, 'atm'), 2.430000e+36, -7.6, 12610.0],
            #[(0.316, 'atm'), 1.950000e+36, -7.57, 12490.0],
            #[(1.0, 'atm'), 2.730000e+35, -7.32, 11820.0],
            #[(3.16, 'atm'), 1.430000e+36, -7.47, 12460.0],
            #[(10.0, 'atm'), 5.180000e+35, -7.2, 13430.0],
            #[(31.6, 'atm'), 3.190000e+20, -2.57, 5578.0],
            #[(100.0, 'atm'), 2.730000e+33, -6.28, 16000.0],
            #options='duplicate')

## Reaction 340
#pdep_arrhenius('C2H3 + O2 <=> CH2O + HCO',
            #[(0.01, 'atm'), 4.540000e+15, -1.28, 515.3],
            #[(0.1, 'atm'), 4.590000e+15, -1.28, 513.0],
            #[(0.316, 'atm'), 4.810000e+15, -1.29, 520.6],
            #[(1.0, 'atm'), 6.080000e+15, -1.31, 645.7],
            #[(3.16, 'atm'), 9.450000e+15, -1.36, 1066.0],
            #[(10.0, 'atm'), 2.560000e+15, -1.18, 1429.0],
            #[(31.6, 'atm'), 1.030000e+69, -19.23, 14760.0],
            #[(100.0, 'atm'), 4.210000e+10, 0.19, 830.6],
            #options='duplicate')

## Reaction 341
#pdep_arrhenius('C2H3 + O2 <=> CH2O + H + CO',
            #[(0.01, 'atm'), 5.820000e+36, -7.6, 12640.0],
            #[(0.1, 'atm'), 5.660000e+36, -7.6, 12610.0],
            #[(0.316, 'atm'), 4.550000e+36, -7.57, 12490.0],
            #[(1.0, 'atm'), 6.360000e+35, -7.32, 11820.0],
            #[(3.16, 'atm'), 3.350000e+36, -7.47, 12460.0],
            #[(10.0, 'atm'), 1.210000e+36, -7.2, 13430.0],
            #[(31.6, 'atm'), 7.430000e+20, -2.57, 5578.0],
            #[(100.0, 'atm'), 6.360000e+33, -6.28, 16000.0],
            #options='duplicate')

## Reaction 342
#pdep_arrhenius('C2H3 + O2 <=> CH2O + H + CO',
            #[(0.01, 'atm'), 1.060000e+16, -1.28, 515.3],
            #[(0.1, 'atm'), 1.070000e+16, -1.28, 513.0],
            #[(0.316, 'atm'), 1.130000e+16, -1.29, 520.6],
            #[(1.0, 'atm'), 1.420000e+16, -1.31, 645.7],
            #[(3.16, 'atm'), 2.200000e+16, -1.36, 1066.0],
            #[(10.0, 'atm'), 5.980000e+15, -1.18, 1429.0],
            #[(31.6, 'atm'), 2.390000e+69, -19.23, 14760.0],
            #[(100.0, 'atm'), 9.810000e+10, 0.19, 830.6],
            #options='duplicate')

## Reaction 343
#pdep_arrhenius('C2H3 + O2 <=> CO + CH3O',
            #[(0.01, 'atm'), 8.190000e+18, -2.66, 3201.0],
            #[(0.1, 'atm'), 4.060000e+14, -1.32, 885.8],
            #[(0.316, 'atm'), 4.340000e+14, -1.33, 900.6],
            #[(1.0, 'atm'), 1.030000e+11, -0.33, -747.8],
            #[(3.16, 'atm'), 1.890000e+12, -3.0, -8995.0],
            #[(10.0, 'atm'), 1.930000e+24, -5.63, 1.8],
            #[(31.6, 'atm'), 1.100000e+18, -2.22, 5178.0],
            #[(100.0, 'atm'), 5.790000e+32, -6.45, 16810.0],
            #options='duplicate')

## Reaction 344
#pdep_arrhenius('C2H3 + O2 <=> CO + CH3O',
            #[(0.01, 'atm'), 1.290000e+09, 0.18, -1717.0],
            #[(0.1, 'atm'), 5.990000e+11, -2.93, -9564.0],
            #[(0.316, 'atm'), 2.910000e+11, -2.93, -10120.0],
            #[(1.0, 'atm'), 5.770000e+21, -3.54, 4772.0],
            #[(3.16, 'atm'), 4.990000e+15, -1.62, 1849.0],
            #[(10.0, 'atm'), 9.330000e+16, -1.96, 3324.0],
            #[(31.6, 'atm'), 1.020000e+72, -20.69, 15860.0],
            #[(100.0, 'atm'), 1.100000e+09, 0.31, 1024.0],
            #options='duplicate')

## Reaction 345
#pdep_arrhenius('C2H3 + O2 <=> CO2 + CH3',
            #[(0.01, 'atm'), 2.370000e+35, -7.76, 12630.0],
            #[(0.1, 'atm'), 1.730000e+35, -7.72, 12520.0],
            #[(0.316, 'atm'), 4.470000e+34, -7.55, 12140.0],
            #[(1.0, 'atm'), 7.250000e+31, -6.7, 10440.0],
            #[(3.16, 'atm'), 3.630000e+35, -7.75, 12830.0],
            #[(10.0, 'atm'), 2.090000e+35, -7.53, 14050.0],
            #[(31.6, 'atm'), 3.840000e+18, -2.44, 5408.0],
            #[(100.0, 'atm'), 1.210000e+32, -6.32, 16190.0],
            #options='duplicate')

## Reaction 346
#pdep_arrhenius('C2H3 + O2 <=> CO2 + CH3',
            #[(0.01, 'atm'), 6.270000e+13, -1.16, 406.3],
            #[(0.1, 'atm'), 6.240000e+13, -1.16, 401.4],
            #[(0.316, 'atm'), 6.120000e+13, -1.16, 397.0],
            #[(1.0, 'atm'), 5.320000e+13, -1.14, 446.7],
            #[(3.16, 'atm'), 1.450000e+14, -1.26, 987.7],
            #[(10.0, 'atm'), 5.020000e+13, -1.11, 1409.0],
            #[(31.6, 'atm'), 1.400000e+70, -20.11, 15430.0],
            #[(100.0, 'atm'), 9.210000e+08, 0.25, 855.3],
            #options='duplicate')

## Reaction 347
#pdep_arrhenius('C2H3OO <=> CHCHO + OH',
            #[(0.01, 'atm'), 3.640000e+49, -12.13, 67420.0],
            #[(0.1, 'atm'), 1.440000e+36, -9.92, 41220.0],
            #[(0.316, 'atm'), 4.180000e+40, -10.53, 43670.0],
            #[(1.0, 'atm'), 3.790000e+46, -10.72, 51900.0],
            #[(3.16, 'atm'), 1.600000e+49, -11.24, 54150.0],
            #[(10.0, 'atm'), 2.380000e+51, -11.64, 56980.0],
            #[(31.6, 'atm'), 2.000000e+54, -12.22, 61840.0],
            #[(100.0, 'atm'), 9.540000e+195, -52.27, 163500.0],
            #options='duplicate')

## Reaction 348
#pdep_arrhenius('C2H3OO <=> CHCHO + OH',
            #[(0.01, 'atm'), 1.170000e+56, -14.81, 60700.0],
            #[(0.1, 'atm'), 2.320000e+40, -9.39, 50420.0],
            #[(0.316, 'atm'), 1.610000e+43, -9.99, 50290.0],
            #[(1.0, 'atm'), 2.330000e+124, -36.77, 70100.0],
            #[(3.16, 'atm'), 1.880000e+103, -29.49, 65410.0],
            #[(10.0, 'atm'), 5.960000e+86, -23.81, 62170.0],
            #[(31.6, 'atm'), 1.510000e+57, -13.94, 55390.0],
            #[(100.0, 'atm'), 1.790000e+34, -6.4, 50000.0],
            #options='duplicate')

## Reaction 349
#pdep_arrhenius('C2H3OO <=> CH2CHO + O',
            #[(0.01, 'atm'), 2.700000e+180, -48.19, 169300.0],
            #[(0.1, 'atm'), 3.900000e+38, -8.69, 42770.0],
            #[(0.316, 'atm'), 4.570000e+47, -11.21, 47050.0],
            #[(1.0, 'atm'), 7.620000e+81, -21.28, 65080.0],
            #[(3.16, 'atm'), 1.860000e+68, -16.83, 60680.0],
            #[(10.0, 'atm'), 2.020000e+55, -12.69, 55840.0],
            #[(31.6, 'atm'), 1.110000e+53, -11.79, 56690.0],
            #[(100.0, 'atm'), 4.300000e+48, -10.31, 56090.0],
            #options='duplicate')

## Reaction 350
#pdep_arrhenius('C2H3OO <=> CH2CHO + O',
            #[(0.01, 'atm'), 1.470000e+30, -6.64, 41110.0],
            #[(0.1, 'atm'), 9.650000e-12, 5.96, 22890.0],
            #[(0.316, 'atm'), 3.950000e+22, -3.71, 36270.0],
            #[(1.0, 'atm'), 2.390000e+33, -6.62, 41280.0],
            #[(3.16, 'atm'), 6.370000e+31, -5.96, 41260.0],
            #[(10.0, 'atm'), 2.130000e+29, -5.1, 40710.0],
            #[(31.6, 'atm'), 4.660000e+27, -4.5, 40530.0],
            #[(100.0, 'atm'), 5.990000e+25, -3.85, 40120.0],
            #options='duplicate')

## Reaction 351
#pdep_arrhenius('C2H3OO <=> CHOCHO + H',
            #[(0.01, 'atm'), 6.410000e+80, -22.2, 51750.0],
            #[(0.1, 'atm'), 3.310000e+65, -17.01, 48090.0],
            #[(0.316, 'atm'), 5.980000e+51, -12.62, 43000.0],
            #[(1.0, 'atm'), 1.480000e+44, -10.12, 40790.0],
            #[(3.16, 'atm'), 1.260000e+59, -14.33, 51390.0],
            #[(10.0, 'atm'), 4.930000e+26, -4.67, 34320.0],
            #[(31.6, 'atm'), 2.060000e+33, -6.38, 39520.0],
            #[(100.0, 'atm'), 1.300000e+32, -5.92, 40660.0],
            #options='duplicate')

## Reaction 352
#pdep_arrhenius('C2H3OO <=> CHOCHO + H',
            #[(0.01, 'atm'), 1.190000e+28, -6.01, 28740.0],
            #[(0.1, 'atm'), 1.400000e+25, -4.8, 28940.0],
            #[(0.316, 'atm'), 2.910000e+20, -3.29, 27550.0],
            #[(1.0, 'atm'), 1.580000e+19, -2.82, 27620.0],
            #[(3.16, 'atm'), 1.930000e+22, -3.54, 29980.0],
            #[(10.0, 'atm'), 7.510000e+29, -5.75, 34490.0],
            #[(31.6, 'atm'), 7.140000e+61, -16.16, 43280.0],
            #[(100.0, 'atm'), 1.140000e+19, -2.56, 29670.0],
            #options='duplicate')

## Reaction 353
#pdep_arrhenius('C2H3OO <=> CH2CO + OH',
            #[(0.01, 'atm'), 1.150000e+47, -12.28, 75330.0],
            #[(0.1, 'atm'), 8.430000e+09, -2.06, 33720.0],
            #[(0.316, 'atm'), 6.060000e+04, 0.17, 34220.0],
            #[(1.0, 'atm'), 1.510000e+19, -3.61, 43060.0],
            #[(3.16, 'atm'), 2.130000e+33, -7.39, 51610.0],
            #[(10.0, 'atm'), 4.440000e+36, -7.99, 54680.0],
            #[(31.6, 'atm'), 1.190000e+37, -7.8, 56460.0],
            #[(100.0, 'atm'), 9.080000e+35, -7.21, 57550.0],
            #options='duplicate')

## Reaction 354
#pdep_arrhenius('C2H3OO <=> CH2CO + OH',
            #[(0.01, 'atm'), 2.310000e+02, -0.73, 25710.0],
            #[(0.1, 'atm'), 1.830000e-23, 7.84, 20190.0],
            #[(0.316, 'atm'), 3.820000e+63, -20.44, 43420.0],
            #[(1.0, 'atm'), 3.180000e+27, -7.76, 37230.0],
            #[(3.16, 'atm'), 2.320000e-05, 3.47, 31560.0],
            #[(10.0, 'atm'), 1.060000e-01, 2.64, 34160.0],
            #[(31.6, 'atm'), 5.620000e+02, 1.7, 36450.0],
            #[(100.0, 'atm'), 1.110000e+07, 0.52, 38670.0],
            #options='duplicate')

## Reaction 355
#pdep_arrhenius('C2H3OO <=> CH2O + HCO',
            #[(0.01, 'atm'), 1.660000e+174, -55.52, 60320.0],
            #[(0.1, 'atm'), 9.030000e+66, -17.25, 48120.0],
            #[(0.316, 'atm'), 1.820000e+43, -9.87, 37960.0],
            #[(1.0, 'atm'), 8.640000e+33, -6.88, 34370.0],
            #[(3.16, 'atm'), 7.290000e+171, -43.53, 191900.0],
            #[(10.0, 'atm'), 1.030000e+32, -6.06, 35500.0],
            #[(31.6, 'atm'), 1.850000e+34, -6.57, 38510.0],
            #[(100.0, 'atm'), 5.700000e+29, -5.19, 36800.0],
            #options='duplicate')

## Reaction 356
#pdep_arrhenius('C2H3OO <=> CH2O + HCO',
            #[(0.01, 'atm'), 2.270000e+35, -7.97, 31280.0],
            #[(0.1, 'atm'), 2.080000e+26, -4.96, 28780.0],
            #[(0.316, 'atm'), 1.450000e+20, -3.08, 26630.0],
            #[(1.0, 'atm'), 1.060000e+130, -39.38, 54700.0],
            #[(3.16, 'atm'), 2.350000e+34, -6.87, 35700.0],
            #[(10.0, 'atm'), 2.180000e+175, -53.78, 68500.0],
            #[(31.6, 'atm'), 1.070000e+185, -54.22, 88990.0],
            #[(100.0, 'atm'), 4.680000e+02, 1.81, 18100.0],
            #options='duplicate')

## Reaction 357
#pdep_arrhenius('C2H3OO <=> CH2O + H + CO',
            #[(0.01, 'atm'), 3.880000e+174, -55.52, 60320.0],
            #[(0.1, 'atm'), 2.110000e+67, -17.25, 48120.0],
            #[(0.316, 'atm'), 4.260000e+43, -9.87, 37960.0],
            #[(1.0, 'atm'), 2.020000e+34, -6.88, 34370.0],
            #[(3.16, 'atm'), 1.700000e+172, -43.53, 191900.0],
            #[(10.0, 'atm'), 2.400000e+32, -6.06, 35500.0],
            #[(31.6, 'atm'), 4.320000e+34, -6.57, 38510.0],
            #[(100.0, 'atm'), 1.330000e+30, -5.19, 36800.0],
            #options='duplicate')

## Reaction 358
#pdep_arrhenius('C2H3OO <=> CH2O + H + CO',
            #[(0.01, 'atm'), 5.290000e+35, -7.97, 31280.0],
            #[(0.1, 'atm'), 4.850000e+26, -4.96, 28780.0],
            #[(0.316, 'atm'), 3.370000e+20, -3.08, 26630.0],
            #[(1.0, 'atm'), 2.460000e+130, -39.38, 54700.0],
            #[(3.16, 'atm'), 5.490000e+34, -6.87, 35700.0],
            #[(10.0, 'atm'), 5.090000e+175, -53.78, 68500.0],
            #[(31.6, 'atm'), 2.490000e+185, -54.22, 88990.0],
            #[(100.0, 'atm'), 1.090000e+03, 1.81, 18100.0],
            #options='duplicate')

## Reaction 359
#pdep_arrhenius('C2H3OO <=> CO + CH3O',
            #[(0.01, 'atm'), 5.200000e+33, -7.92, 31320.0],
            #[(0.1, 'atm'), 1.260000e+98, -27.09, 64060.0],
            #[(0.316, 'atm'), 1.800000e+33, -7.27, 33760.0],
            #[(1.0, 'atm'), 3.830000e+33, -7.2, 35100.0],
            #[(3.16, 'atm'), 1.280000e+79, -19.61, 74870.0],
            #[(10.0, 'atm'), 4.070000e+32, -6.62, 37210.0],
            #[(31.6, 'atm'), 6.860000e+44, -10.04, 47030.0],
            #options='duplicate')

## Reaction 360
#pdep_arrhenius('C2H3OO <=> CO + CH3O',
            #[(0.01, 'atm'), 2.310000e+129, -41.86, 45850.0],
            #[(0.1, 'atm'), 2.420000e+28, -5.99, 30540.0],
            #[(0.316, 'atm'), 8.690000e-50, 16.63, -3900.0],
            #[(1.0, 'atm'), 1.190000e-39, 13.61, -1317.0],
            #[(3.16, 'atm'), 8.800000e+86, -23.08, 61010.0],
            #[(10.0, 'atm'), 1.270000e+03, 1.44, 18660.0],
            #[(31.6, 'atm'), 1.970000e+17, -2.23, 28590.0],
            #[(100.0, 'atm'), 3.250000e+04, 1.694, 23327.6],
            #options='duplicate')

## Reaction 361
#reaction('CH3 + C2H3 <=> CH4 + C2H2', [3.920000e+11, 0.0, 0.0])

## Reaction 362
#reaction('C2H3 + H <=> C2H2 + H2', [1.700000e+14, 0.0, 0.0])

## Reaction 363
#reaction('C2H3 + H <=> H2CC + H2', [6.000000e+13, 0.0, 0.0])

## Reaction 364
#reaction('C2H3 + OH <=> C2H2 + H2O', [3.011000e+13, 0.0, 0.0])

## Reaction 365
#reaction('C2H3 + C2H3 <=> C2H2 + C2H4', [9.600000e+11, 0.0, 0.0])

## Reaction 366
#falloff_reaction('C2H + H (+ M) <=> C2H2 (+ M)',
                #kf=[1.000000e+17, 0.0, 0.0],
                #kf0=[3.750000e+33, -4.8, 1900.0],
                #efficiencies='H2:2.0 H2O:6.0 AR:0.7 CO:1.5 CO2:2.0 CH4:2.0 C2H6:3.0 HE:0.7',
                #falloff=Troe(A=0.646, T3=132.0, T1=1315.0, T2=5566.0))

## Reaction 367
#reaction('C2H + O <=> CH + CO', [5.000000e+13, 0.0, 0.0])

## Reaction 368
#reaction('C2H + OH <=> H + HCCO', [2.000000e+13, 0.0, 0.0])

## Reaction 369
#reaction('C2H + O2 <=> HCO + CO', [5.000000e+13, 0.0, 1500.0])

## Reaction 370
#reaction('C2H + H2 <=> H + C2H2', [4.900000e+05, 2.5, 560.0])

## Reaction 371
#falloff_reaction('C2H2 (+ M) <=> H2CC (+ M)',
                #kf=[8.000000e+14, -0.52, 50750.0],
                #kf0=[2.450000e+15, -0.64, 49700.0],
                #efficiencies='H2:2.0 H2O:6.0 CH4:2.0 CO:1.5 CO2:2.0 C2H6:3.0 C2H2:2.5 C2H4:2.5')

## Reaction 372
#reaction('C2H2 + O <=> CH2 + CO', [7.395000e+08, 1.28, 2472.0])

## Reaction 373
#reaction('C2H2 + O <=> HCCO + H', [2.958000e+09, 1.28, 2472.0])

## Reaction 374
#reaction('C2H2 + HO2 <=> CH2CO + OH', [6.030000e+09, 0.0, 7949.0])

## Reaction 375
#reaction('C2H2 + OH <=> C2H + H2O', [2.632000e+06, 2.14, 17060.0])

## Reaction 376
#pdep_arrhenius('C2H2 + OH <=> HCCOH + H',
            #[(0.01, 'atm'), 2.800000e+05, 2.28, 12420.0],
            #[(0.025, 'atm'), 7.467000e+05, 2.16, 12550.0],
            #[(0.1, 'atm'), 1.776000e+06, 2.04, 12670.0],
            #[(1.0, 'atm'), 2.415000e+06, 2.0, 12710.0],
            #[(10.0, 'atm'), 3.210000e+06, 1.97, 12810.0],
            #[(100.0, 'atm'), 7.347000e+06, 1.89, 13600.0])

## Reaction 377
#pdep_arrhenius('C2H2 + OH <=> CH2CO + H',
            #[(0.01, 'atm'), 1.578000e+03, 2.56, -844.5],
            #[(0.025, 'atm'), 1.518000e+04, 2.28, -292.1],
            #[(0.1, 'atm'), 3.017000e+05, 1.92, 598.1],
            #[(1.0, 'atm'), 7.528000e+06, 1.55, 2106.0],
            #[(10.0, 'atm'), 5.101000e+06, 1.65, 3400.0],
            #[(100.0, 'atm'), 1.457000e+04, 2.45, 4477.0])

## Reaction 378
#pdep_arrhenius('C2H2 + OH <=> CH3 + CO',
            #[(0.01, 'atm'), 4.757000e+05, 1.68, -329.8],
            #[(0.025, 'atm'), 4.372000e+06, 1.4, 226.5],
            #[(0.1, 'atm'), 7.648000e+07, 1.05, 1115.0],
            #[(1.0, 'atm'), 1.277000e+09, 0.73, 2579.0],
            #[(10.0, 'atm'), 4.312000e+08, 0.92, 3736.0],
            #[(100.0, 'atm'), 8.250000e+05, 1.77, 4697.0])

## Reaction 379
#pdep_arrhenius('C2H2 + OH <=> C2H2OH',
            #[(0.01, 'atm'), 3.913000e+32, -7.126, 5824.0],
            #[(0.025, 'atm'), 1.067000e+32, -6.847, 5508.0],
            #[(0.1, 'atm'), 1.646000e+32, -6.717, 5822.0],
            #[(1.0, 'atm'), 1.387000e+31, -6.087, 6348.0],
            #[(10.0, 'atm'), 2.892000e+29, -5.288, 7055.0],
            #[(100.0, 'atm'), 1.367000e+25, -3.754, 6543.0])

## Reaction 380
#reaction('C2H2 + HCO <=> C2H3 + CO', [1.000000e+07, 2.0, 6000.0])

## Reaction 381
#reaction('C2H2 + CH2 <=> C3H3 + H', [1.200000e+13, 0.0, 6620.0])

## Reaction 382
#reaction('C2H2 + CH2(S) <=> C3H3 + H', [2.000000e+13, 0.0, 0.0])

## Reaction 383
#reaction('C2H2 + HCCO <=> C3H3 + CO', [1.000000e+11, 0.0, 3000.0])

## Reaction 384
#reaction('H2CC + H <=> C2H2 + H', [1.000000e+14, 0.0, 0.0])

## Reaction 385
#reaction('H2CC + OH <=> CH2CO + H', [2.000000e+13, 0.0, 0.0])

## Reaction 386
#reaction('H2CC + O2 <=> HCO + HCO', [1.000000e+13, 0.0, 0.0])

## Reaction 387
#reaction('H + HCCOH <=> H + CH2CO', [1.000000e+13, 0.0, 0.0])

## Reaction 388
#pdep_arrhenius('C2H5OH <=> C2H4 + H2O',
            #[(0.001, 'atm'), 3.410000e+59, -14.2, 83672.6],
            #[(0.01, 'atm'), 2.620000e+57, -13.3, 85262.2],
            #[(0.1, 'atm'), 1.650000e+52, -11.5, 84745.6],
            #[(1.0, 'atm'), 5.230000e+43, -8.9, 81506.7],
            #[(10.0, 'atm'), 4.590000e+32, -5.6, 76062.4],
            #[(100.0, 'atm'), 3.840000e+20, -2.06, 69465.5])

## Reaction 389
#pdep_arrhenius('C2H5OH <=> CH3 + CH2OH',
            #[(0.001, 'atm'), 1.200000e+54, -12.9, 100005.7],
            #[(0.01, 'atm'), 5.180000e+59, -14.0, 99906.4],
            #[(0.1, 'atm'), 1.620000e+66, -15.3, 105390.5],
            #[(1.0, 'atm'), 5.550000e+64, -14.5, 106183.0],
            #[(10.0, 'atm'), 1.550000e+58, -12.3, 105768.0],
            #[(100.0, 'atm'), 1.780000e+47, -8.96, 101058.8])

## Reaction 390
#pdep_arrhenius('C2H5OH <=> C2H5 + OH',
            #[(0.001, 'atm'), 8.100000e+46, -11.3, 111053.4],
            #[(0.01, 'atm'), 1.860000e+56, -13.5, 107238.4],
            #[(0.1, 'atm'), 4.650000e+63, -15.0, 109622.8],
            #[(1.0, 'atm'), 4.460000e+65, -14.9, 112345.0],
            #[(10.0, 'atm'), 2.790000e+61, -13.4, 113080.2],
            #[(100.0, 'atm'), 6.170000e+51, -10.3, 109940.7])

## Reaction 391
#reaction('C2H5OH + O2 <=> PC2H4OH + HO2', [2.000000e+13, 0.0, 52800.0])

## Reaction 392
#reaction('C2H5OH + O2 <=> SC2H4OH + HO2', [1.500000e+13, 0.0, 50150.0])

## Reaction 393
#reaction('C2H5OH + H <=> SC2H4OH + H2', [8.790000e+04, 2.68, 2910.0])

## Reaction 394
#reaction('C2H5OH + H <=> PC2H4OH + H2', [5.310000e+04, 2.81, 7490.0])

## Reaction 395
#reaction('C2H5OH + H <=> C2H5O + H2', [9.450000e+02, 3.14, 8701.1])

## Reaction 396
#reaction('C2H5OH + OH <=> SC2H4OH + H2O', [7.520000e+04, 2.49, -1474.1])

## Reaction 397
#reaction('C2H5OH + OH <=> PC2H4OH + H2O', [3.760000e+03, 2.78, -1810.2])

## Reaction 398
#reaction('C2H5OH + OH <=> C2H5O + H2O', [5.810000e-03, 4.28, -3560.0])

## Reaction 399
#reaction('C2H5OH + HO2 <=> SC2H4OH + H2O2', [2.450000e-05, 5.26, 7475.1])

## Reaction 400
#reaction('C2H5OH + HO2 <=> PC2H4OH + H2O2', [3.986000e-02, 4.3, 15333.0])

## Reaction 401
#reaction('C2H5OH + HO2 <=> C2H5O + H2O2', [6.470000e-07, 5.3, 10533.1])

## Reaction 402
#reaction('C2H5OH + CH3O2 <=> SC2H4OH + CH3O2H', [1.225000e-05, 5.26, 7475.1])

## Reaction 403
#reaction('C2H5OH + CH3O2 <=> PC2H4OH + CH3O2H', [1.995000e-02, 4.3, 15333.0])

## Reaction 404
#reaction('C2H5OH + CH3O2 <=> C2H5O + CH3O2H', [3.236000e-07, 5.3, 10533.1])

## Reaction 405
#reaction('C2H5OH + O <=> PC2H4OH + OH', [9.690000e+02, 3.23, 4658.0])

## Reaction 406
#reaction('C2H5OH + O <=> SC2H4OH + OH', [1.450000e+05, 2.47, 876.0])

## Reaction 407
#reaction('C2H5OH + O <=> C2H5O + OH', [1.460000e-03, 4.73, 1727.0])

## Reaction 408
#reaction('C2H5OH + CH3 <=> PC2H4OH + CH4', [3.300000e+02, 3.3, 12290.0])

## Reaction 409
#reaction('C2H5OH + CH3 <=> SC2H4OH + CH4', [1.993000e+01, 3.37, 7634.0])

## Reaction 410
#reaction('C2H5OH + CH3 <=> C2H5O + CH4', [2.035000e+00, 3.57, 7721.0])

## Reaction 411
#reaction('C2H5OH + C2H5 <=> PC2H4OH + C2H6', [5.000000e+10, 0.0, 13400.0])

## Reaction 412
#reaction('C2H5OH + C2H5 <=> SC2H4OH + C2H6', [5.000000e+10, 0.0, 10400.0])

## Reaction 413
#pdep_arrhenius('SC2H4OH <=> CH3CHO + H',
            #[(0.001, 'atm'), 5.690000e+52, -13.38, 45049.0],
            #[(0.01, 'atm'), 3.290000e+56, -14.12, 48129.0],
            #[(0.1, 'atm'), 8.580000e+57, -14.16, 50743.0],
            #[(1.0, 'atm'), 5.360000e+55, -13.15, 51886.0],
            #[(10.0, 'atm'), 1.660000e+48, -10.64, 50297.0],
            #[(20.0, 'atm'), 8.260000e+44, -9.59, 49218.0],
            #[(50.0, 'atm'), 1.010000e+40, -8.06, 47439.0],
            #[(100.0, 'atm'), 1.100000e+36, -6.84, 45899.0])

## Reaction 414
#pdep_arrhenius('SC2H4OH <=> C2H3OH + H',
            #[(0.001, 'atm'), 5.400000e+46, -11.63, 44323.0],
            #[(0.01, 'atm'), 1.210000e+51, -12.55, 47240.0],
            #[(0.1, 'atm'), 2.870000e+54, -13.15, 50702.0],
            #[(1.0, 'atm'), 3.790000e+53, -12.51, 52560.0],
            #[(10.0, 'atm'), 6.330000e+46, -10.2, 51441.0],
            #[(20.0, 'atm'), 3.870000e+43, -9.17, 50440.0],
            #[(50.0, 'atm'), 5.080000e+38, -7.65, 48713.0],
            #[(100.0, 'atm'), 5.120000e+34, -6.41, 47182.0])

## Reaction 415
#pdep_arrhenius('SC2H4OH <=> C2H5O',
            #[(0.001, 'atm'), 5.480000e+45, -11.63, 44328.0],
            #[(0.01, 'atm'), 2.540000e+49, -12.37, 46445.0],
            #[(0.1, 'atm'), 1.650000e+54, -13.4, 50330.0],
            #[(1.0, 'atm'), 1.810000e+55, -13.31, 53132.0],
            #[(10.0, 'atm'), 4.580000e+49, -11.32, 52714.0],
            #[(20.0, 'atm'), 4.110000e+46, -10.33, 51834.0],
            #[(50.0, 'atm'), 6.680000e+41, -8.83, 50202.0],
            #[(100.0, 'atm'), 6.540000e+37, -7.58, 48697.0])

## Reaction 416
#pdep_arrhenius('SC2H4OH <=> PC2H4OH',
            #[(0.001, 'atm'), 2.650000e+36, -8.86, 51019.0],
            #[(0.01, 'atm'), 3.560000e+37, -8.89, 51114.0],
            #[(0.1, 'atm'), 4.140000e+39, -9.19, 51912.0],
            #[(1.0, 'atm'), 5.820000e+44, -10.34, 55296.0],
            #[(10.0, 'atm'), 4.260000e+48, -11.06, 59458.0],
            #[(20.0, 'atm'), 8.840000e+47, -10.74, 59901.0],
            #[(50.0, 'atm'), 2.230000e+45, -9.84, 59604.0],
            #[(100.0, 'atm'), 1.700000e+42, -8.83, 58737.0])

## Reaction 417
#reaction('O2C2H4OH <=> PC2H4OH + O2', [3.900000e+16, -1.0, 30000.0])

## Reaction 418
#reaction('O2C2H4OH => OH + CH2O + CH2O', [1.250000e+11, 0.0, 18900.0])

## Reaction 419
#pdep_arrhenius('SC2H4OH + O2 <=> CH3CHO + HO2',
            #[(0.01, 'atm'), 5.260000e+17, -1.637, 838.0],
            #[(0.1, 'atm'), 5.260000e+17, -1.637, 838.0],
            #[(1.0, 'atm'), 5.280000e+17, -1.638, 839.0],
            #[(10.0, 'atm'), 1.540000e+18, -1.771, 1120.0],
            #[(100.0, 'atm'), 3.780000e+20, -2.429, 3090.0])

## Reaction 420
#pdep_arrhenius('SC2H4OH + O2 <=> C2H3OH + HO2',
            #[(0.01, 'atm'), 5.120000e+02, 2.496, -414.0],
            #[(0.1, 'atm'), 5.330000e+02, 2.49, -402.0],
            #[(1.0, 'atm'), 7.620000e+02, 2.446, -296.0],
            #[(10.0, 'atm'), 8.920000e+03, 2.146, 470.0],
            #[(100.0, 'atm'), 4.380000e+05, 1.699, 2330.0])

## Reaction 421
#pdep_arrhenius('CH3COCH3 <=> CH3CO + CH3',
            #[(0.01, 'atm'), 2.050000e+58, -12.796, 100030.1],
            #[(0.1, 'atm'), 3.300000e+51, -10.574, 98221.2],
            #[(1.0, 'atm'), 1.310000e+42, -7.657, 94660.6],
            #[(10.0, 'atm'), 2.160000e+33, -4.989, 90916.5],
            #[(100.0, 'atm'), 9.400000e+28, -3.669, 89022.8])

## Reaction 422
#reaction('CH3COCH3 + OH <=> CH3COCH2 + H2O', [1.250000e+05, 2.483, 445.0])

## Reaction 423
#reaction('CH3COCH3 + H <=> CH3COCH2 + H2', [9.800000e+05, 2.43, 5160.0])

## Reaction 424
#reaction('CH3COCH3 + O <=> CH3COCH2 + OH', [5.130000e+11, 0.211, 4890.0])

## Reaction 425
#reaction('CH3COCH3 + CH3 <=> CH3COCH2 + CH4', [3.960000e+11, 0.0, 9784.0])

## Reaction 426
#reaction('CH3COCH3 + CH3O <=> CH3COCH2 + CH3OH', [4.340000e+11, 0.0, 6460.0])

## Reaction 427
#reaction('CH3COCH3 + O2 <=> CH3COCH2 + HO2', [6.030000e+13, 0.0, 48500.0])

## Reaction 428
#reaction('CH3COCH3 + HO2 <=> CH3COCH2 + H2O2', [1.700000e+13, 0.0, 20460.0])

## Reaction 429
#reaction('CH3COCH3 + CH3O2 <=> CH3COCH2 + CH3O2H', [1.700000e+13, 0.0, 20460.0])

## Reaction 430
#reaction('CH2CO + CH3 <=> CH3COCH2', [1.760000e+04, 2.48, 6130.0])

## Reaction 431
#reaction('CH3COCH2 + HO2 <=> CH3COCH2O + OH', [2.410000e+13, 0.0, 0.0])

## Reaction 432
#reaction('CH3COCH2 + CH3O2 <=> CH3COCH2O + CH3O', [1.205000e+13, 0.0, 0.0])

## Reaction 433
#reaction('CH3COCH2 + H <=> CH3COCH3', [1.000000e+13, 0.0, 0.0])

## Reaction 434
#reaction('CH3COCH2 + O2 <=> CH3COCH2O2', [1.200000e+11, 0.0, -1100.0])

## Reaction 435
#reaction('CH3COCH3 + CH3COCH2O2 <=> CH3COCH2 + C3KET21', [1.000000e+11, 0.0, 5000.0])

## Reaction 436
#reaction('CH2O + CH3COCH2O2 <=> HCO + C3KET21', [1.288000e+11, 0.0, 9000.0])

## Reaction 437
#reaction('HO2 + CH3COCH2O2 <=> C3KET21 + O2', [1.000000e+12, 0.0, 0.0])

## Reaction 438
#reaction('C2H3 + HCO <=> C2H3CHO', [1.810000e+13, 0.0, 0.0])

## Reaction 439
#reaction('C2H3CHO + H <=> C2H3CO + H2', [1.340000e+13, 0.0, 3300.0])

## Reaction 440
#reaction('C2H3CHO + O <=> C2H3CO + OH', [5.940000e+12, 0.0, 1868.0])

## Reaction 441
#reaction('C2H3CHO + OH <=> C2H3CO + H2O', [9.240000e+06, 1.5, -962.0])

## Reaction 442
#reaction('C2H3CHO + O2 <=> C2H3CO + HO2', [1.005000e+13, 0.0, 40700.0])

## Reaction 443
#reaction('C2H3CHO + HO2 <=> C2H3CO + H2O2', [3.010000e+12, 0.0, 11920.0])

## Reaction 444
#reaction('C2H3CHO + CH3 <=> C2H3CO + CH4', [2.608000e+06, 1.78, 5911.0])

## Reaction 445
#reaction('C2H3CHO + C2H3 <=> C2H3CO + C2H4', [1.740000e+12, 0.0, 8440.0])

## Reaction 446
#reaction('C2H3CHO + CH3O <=> C2H3CO + CH3OH', [1.000000e+12, 0.0, 3300.0])

## Reaction 447
#reaction('C2H3CHO + CH3O2 <=> C2H3CO + CH3O2H', [3.010000e+12, 0.0, 11920.0])

## Reaction 448
#reaction('C2H3 + CO <=> C2H3CO', [1.510000e+11, 0.0, 4810.0])

## Reaction 449
#reaction('C2H5 + HCO <=> C2H5CHO', [1.810000e+13, 0.0, 0.0])

## Reaction 450
#reaction('C2H5CHO + H <=> C2H5CO + H2', [4.000000e+13, 0.0, 4200.0])

## Reaction 451
#reaction('C2H5CHO + O <=> C2H5CO + OH', [5.000000e+12, 0.0, 1790.0])

## Reaction 452
#reaction('C2H5CHO + OH <=> C2H5CO + H2O', [2.690000e+10, 0.76, -340.0])

## Reaction 453
#reaction('C2H5CHO + CH3 <=> C2H5CO + CH4', [2.608000e+06, 1.78, 5911.0])

## Reaction 454
#reaction('C2H5CHO + HO2 <=> C2H5CO + H2O2', [2.800000e+12, 0.0, 13600.0])

## Reaction 455
#reaction('C2H5CHO + CH3O <=> C2H5CO + CH3OH', [1.000000e+12, 0.0, 3300.0])

## Reaction 456
#reaction('C2H5CHO + CH3O2 <=> C2H5CO + CH3O2H', [3.010000e+12, 0.0, 11920.0])

## Reaction 457
#reaction('C2H5CHO + C2H5 <=> C2H5CO + C2H6', [1.000000e+12, 0.0, 8000.0])

## Reaction 458
#reaction('C2H5CHO + C2H5O <=> C2H5CO + C2H5OH', [6.026000e+11, 0.0, 3300.0])

## Reaction 459
#reaction('C2H5CHO + C2H5O2 <=> C2H5CO + C2H5O2H', [3.010000e+12, 0.0, 11920.0])

## Reaction 460
#reaction('C2H5CHO + O2 <=> C2H5CO + HO2', [1.005000e+13, 0.0, 40700.0])

## Reaction 461
#reaction('C2H5CHO + CH3CO3 <=> C2H5CO + CH3CO3H', [3.010000e+12, 0.0, 11920.0])

## Reaction 462
#reaction('C2H5CHO + C2H3 <=> C2H5CO + C2H4', [1.700000e+12, 0.0, 8440.0])

## Reaction 463
#reaction('C2H5 + CO <=> C2H5CO', [1.510000e+11, 0.0, 4810.0])

## Reaction 464
#falloff_reaction('CH3OCH3 (+ M) <=> CH3 + CH3O (+ M)',
                #kf=[2.330000e+19, -0.661, 84139.0],
                #kf0=[1.720000e+59, -11.4, 93295.6],
                #efficiencies='H2:3.0 CO:2.25 CO2:3.0 N2:1.5 H2O:9.0 CH3OCH3:5.0 CH4:3.0 C2H6:4.5',
                #falloff=Troe(A=1.0, T3=1e-30, T1=880.0))

## Reaction 465
#reaction('CH3OCH3 + OH <=> CH3OCH2 + H2O', [9.350000e+05, 2.29, -780.7])

## Reaction 466
#reaction('CH3OCH3 + H <=> CH3OCH2 + H2', [7.721000e+06, 2.09, 3384.0])

## Reaction 467
#reaction('CH3OCH3 + O <=> CH3OCH2 + OH', [7.750000e+08, 1.36, 2250.0])

## Reaction 468
#reaction('CH3OCH3 + HO2 <=> CH3OCH2 + H2O2', [3.170000e-03, 4.64, 10556.0])

## Reaction 469
#reaction('CH3OCH3 + CH3O2 <=> CH3OCH2 + CH3O2H', [1.268000e-03, 4.64, 10556.0])

## Reaction 470
#reaction('CH3OCH3 + CH3 <=> CH3OCH2 + CH4', [7.020000e+00, 3.78, 9687.1])

## Reaction 471
#reaction('CH3OCH3 + O2 <=> CH3OCH2 + HO2', [4.100000e+13, 0.0, 44910.0])

## Reaction 472
#reaction('CH3OCH3 + CH3O <=> CH3OCH2 + CH3OH', [6.020000e+11, 0.0, 4074.0])

## Reaction 473
#reaction('CH3OCH3 + CH3OCH2O2 <=> CH3OCH2 + CH3OCH2O2H', [5.000000e+12, 0.0, 17690.0])

## Reaction 474
#reaction('CH3OCH3 + O2CHO <=> CH3OCH2 + HO2CHO', [4.425000e+04, 2.6, 13910.0])

## Reaction 475
#reaction('CH3OCH3 + OCHO <=> CH3OCH2 + HOCHO', [1.000000e+13, 0.0, 17690.0])

## Reaction 476
#pdep_arrhenius('CH3OCH2 <=> CH3 + CH2O',
            #[(0.01, 'atm'), 7.494000e+23, -4.5152, 25236.1],
            #[(0.1, 'atm'), 6.921000e+28, -5.7271, 27494.9],
            #[(1.0, 'atm'), 4.229000e+29, -5.6103, 28898.3],
            #[(10.0, 'atm'), 6.608000e+27, -4.7073, 29735.2],
            #[(100.0, 'atm'), 2.659000e+29, -4.9358, 31785.5])

## Reaction 477
#pdep_arrhenius('CH3OCH2 + O2 <=> CH3OCH2O2',
            #[(0.001, 'atm'), 1.120000e+18, -3.37, -4294.0],
            #[(0.01, 'atm'), 1.330000e+21, -3.95, -2615.0],
            #[(1.0, 'atm'), 1.130000e+28, -5.24, 4088.0],
            #[(2.0, 'atm'), 3.910000e+27, -5.0, 4512.0],
            #[(10.0, 'atm'), 2.750000e+24, -3.87, 4290.0],
            #[(20.0, 'atm'), 2.970000e+22, -3.23, 3781.0],
            #[(50.0, 'atm'), 5.190000e+19, -2.35, 2908.0],
            #[(100.0, 'atm'), 5.430000e+17, -1.73, 2210.0])

## Reaction 478
#pdep_arrhenius('CH3OCH2 + O2 <=> CH2OCH2O2H',
            #[(0.001, 'atm'), 5.080000e+20, -4.39, 469.0],
            #[(0.01, 'atm'), 5.470000e+23, -4.96, 2183.0],
            #[(1.0, 'atm'), 2.810000e+28, -5.63, 7848.0],
            #[(2.0, 'atm'), 5.190000e+27, -5.33, 8144.0],
            #[(10.0, 'atm'), 9.670000e+24, -4.36, 8417.0],
            #[(20.0, 'atm'), 4.080000e+23, -3.9, 8494.0],
            #[(50.0, 'atm'), 5.080000e+21, -3.28, 8585.0],
            #[(100.0, 'atm'), 1.620000e+20, -2.81, 8619.0])

## Reaction 479
#pdep_arrhenius('CH3OCH2 + O2 <=> CH2O + CH2O + OH',
            #[(0.001, 'atm'), 8.010000e+21, -3.18, 3067.0],
            #[(0.01, 'atm'), 1.730000e+23, -3.55, 4050.0],
            #[(1.0, 'atm'), 2.040000e+31, -5.76, 11594.0],
            #[(2.0, 'atm'), 5.990000e+31, -5.87, 12710.0],
            #[(10.0, 'atm'), 9.390000e+30, -5.59, 14517.0],
            #[(20.0, 'atm'), 1.090000e+30, -5.3, 15051.0],
            #[(50.0, 'atm'), 3.580000e+28, -4.88, 15664.0],
            #[(100.0, 'atm'), 2.410000e+27, -4.55, 16107.0])

## Reaction 480
#pdep_arrhenius('CH3OCH2O2 <=> CH2OCH2O2H',
            #[(0.001, 'atm'), 1.940000e+29, -6.99, 22446.0],
            #[(0.01, 'atm'), 4.070000e+27, -6.16, 21619.0],
            #[(1.0, 'atm'), 2.520000e+25, -4.76, 22691.0],
            #[(2.0, 'atm'), 5.970000e+24, -4.48, 22868.0],
            #[(10.0, 'atm'), 4.440000e+21, -3.38, 22386.0],
            #[(20.0, 'atm'), 4.520000e+19, -2.74, 21803.0],
            #[(50.0, 'atm'), 5.720000e+16, -1.82, 20829.0],
            #[(100.0, 'atm'), 3.700000e+14, -1.13, 20034.0])

## Reaction 481
#pdep_arrhenius('CH3OCH2O2 <=> CH2O + CH2O + OH',
            #[(0.001, 'atm'), 2.060000e+36, -8.32, 33415.0],
            #[(0.01, 'atm'), 2.070000e+39, -8.86, 35842.0],
            #[(1.0, 'atm'), 1.120000e+40, -8.42, 39835.0],
            #[(2.0, 'atm'), 9.720000e+38, -8.04, 39923.0],
            #[(10.0, 'atm'), 6.280000e+35, -6.97, 39900.0],
            #[(20.0, 'atm'), 1.600000e+34, -6.46, 39850.0],
            #[(50.0, 'atm'), 8.320000e+31, -5.75, 39719.0],
            #[(100.0, 'atm'), 1.220000e+30, -5.2, 39549.0])

## Reaction 482
#pdep_arrhenius('CH2OCH2O2H <=> CH2O + CH2O + OH',
            #[(0.001, 'atm'), 1.660000e+23, -4.53, 22243.0],
            #[(0.01, 'atm'), 5.300000e+25, -4.93, 24158.0],
            #[(1.0, 'atm'), 7.810000e+22, -3.5, 23156.0],
            #[(2.0, 'atm'), 4.980000e+22, -3.35, 23062.0],
            #[(10.0, 'atm'), 8.460000e+22, -3.22, 23559.0],
            #[(20.0, 'atm'), 9.090000e+22, -3.14, 23899.0],
            #[(50.0, 'atm'), 4.590000e+22, -2.94, 24262.0],
            #[(100.0, 'atm'), 1.400000e+22, -2.72, 24407.0])

## Reaction 483
#pdep_arrhenius('CH2OCH2O2H + O2 <=> O2CH2OCH2O2H',
            #[(0.001, 'atm'), 9.420000e+12, -1.68, -4998.0],
            #[(0.01, 'atm'), 8.160000e+16, -2.5, -2753.0],
            #[(1.0, 'atm'), 1.060000e+22, -3.3, 3389.0],
            #[(2.0, 'atm'), 3.480000e+20, -2.79, 3131.0],
            #[(10.0, 'atm'), 2.860000e+16, -1.48, 1873.0],
            #[(20.0, 'atm'), 8.550000e+14, -1.01, 1312.0],
            #[(50.0, 'atm'), 2.680000e+13, -0.54, 727.0],
            #[(100.0, 'atm'), 4.870000e+12, -0.32, 428.0])

## Reaction 484
#pdep_arrhenius('CH2OCH2O2H + O2 <=> HO2CH2OCHO + OH',
            #[(0.001, 'atm'), 5.900000e+20, -2.88, 3234.0],
            #[(0.01, 'atm'), 2.060000e+23, -3.59, 5116.0],
            #[(1.0, 'atm'), 4.450000e+29, -5.29, 12791.0],
            #[(2.0, 'atm'), 2.440000e+28, -4.92, 12891.0],
            #[(10.0, 'atm'), 9.420000e+23, -3.68, 12049.0],
            #[(20.0, 'atm'), 1.040000e+22, -3.16, 11505.0],
            #[(50.0, 'atm'), 6.950000e+19, -2.6, 10861.0],
            #[(100.0, 'atm'), 3.960000e+18, -2.31, 10500.0])

## Reaction 485
#pdep_arrhenius('O2CH2OCH2O2H <=> HO2CH2OCHO + OH',
            #[(0.001, 'atm'), 9.050000e+23, -4.88, 18805.0],
            #[(0.01, 'atm'), 6.840000e+26, -5.32, 22533.0],
            #[(1.0, 'atm'), 5.070000e+16, -1.81, 21175.0],
            #[(2.0, 'atm'), 2.660000e+14, -1.11, 20310.0],
            #[(10.0, 'atm'), 1.690000e+10, 0.18, 18604.0],
            #[(20.0, 'atm'), 1.110000e+09, 0.54, 18100.0],
            #[(50.0, 'atm'), 1.070000e+08, 0.84, 17661.0],
            #[(100.0, 'atm'), 3.860000e+07, 0.98, 17467.0])

## Reaction 486
#reaction('HO2CH2OCHO <=> OCH2OCHO + OH', [5.000000e+16, 0.0, 43000.0])

## Reaction 487
#reaction('CH3OCH2 + CH3O <=> CH3OCH3 + CH2O', [2.410000e+13, 0.0, 0.0])

## Reaction 488
#reaction('CH3OCH2 + CH2O <=> CH3OCH3 + HCO', [5.490000e+03, 2.8, 5862.0])

## Reaction 489
#reaction('CH3OCH2 + CH3CHO <=> CH3OCH3 + CH3CO', [1.260000e+12, 0.0, 8499.0])

## Reaction 490
#reaction('CH3OCH2O2 + CH2O <=> CH3OCH2O2H + HCO', [1.000000e+12, 0.0, 11660.0])

## Reaction 491
#reaction('CH3OCH2O2 + CH3CHO <=> CH3OCH2O2H + CH3CO', [2.800000e+12, 0.0, 13600.0])

## Reaction 492
#reaction('CH3OCH2O + OH <=> CH3OCH2O2H', [2.000000e+13, 0.0, 0.0])

## Reaction 493
#reaction('CH3O + CH2O <=> CH3OCH2O', [1.000000e+11, 0.0, 7960.0])

## Reaction 494
#reaction('CH3OCH2O + O2 <=> CH3OCHO + HO2', [4.380000e-19, 9.5, -5501.0])

## Reaction 495
#reaction('CH3OCH2O <=> CH3OCHO + H', [3.300000e+13, 0.004, 26136.3])

## Reaction 496
#reaction('CH2O + OCHO <=> OCH2OCHO', [1.250000e+11, 0.0, 11900.0])

## Reaction 497
#reaction('OCH2OCHO <=> HOCH2OCO', [1.000000e+11, 0.0, 14000.0])

## Reaction 498
#reaction('HOCH2OCO <=> CH2OH + CO2', [1.117000e+17, -1.526, 20771.9])

## Reaction 499
#reaction('HOCH2OCO <=> HOCH2O + CO', [2.920000e+18, -1.965, 19619.0])

## Reaction 500
#reaction('CH3OCH2O2 + CH3OCH2O2 => O2 + CH3OCH2O + CH3OCH2O', [1.547000e+23, -4.5, 0.0])

## Reaction 501
#reaction('CH2OCHO + H <=> CH3OCHO', [1.000000e+14, 0.0, 0.0])

## Reaction 502
#reaction('CH3OCO + H <=> CH3OCHO', [1.000000e+14, 0.0, 0.0])

## Reaction 503
#reaction('CH3O + HCO <=> CH3OCHO', [3.000000e+13, 0.0, 0.0])

## Reaction 504
#reaction('CH3 + OCHO <=> CH3OCHO', [1.000000e+13, 0.0, 0.0])

## Reaction 505
#reaction('CH3OCHO + O2 <=> CH3OCO + HO2', [1.000000e+13, 0.0, 49700.0])

## Reaction 506
#reaction('CH3OCHO + O2 <=> CH2OCHO + HO2', [2.050000e+13, 0.0, 52000.0])

## Reaction 507
#reaction('CH3OCHO + OH <=> CH3OCO + H2O', [1.580000e+07, 1.8, 934.0])

## Reaction 508
#reaction('CH3OCHO + OH <=> CH2OCHO + H2O', [5.270000e+09, 0.97, 1586.0])

## Reaction 509
#reaction('CH3OCHO + HO2 <=> CH3OCO + H2O2', [4.820000e+03, 2.6, 13910.0])

## Reaction 510
#reaction('CH3OCHO + HO2 <=> CH2OCHO + H2O2', [2.380000e+04, 2.55, 16490.0])

## Reaction 511
#reaction('CH3OCHO + O <=> CH3OCO + OH', [2.755000e+05, 2.45, 2830.0])

## Reaction 512
#reaction('CH3OCHO + O <=> CH2OCHO + OH', [9.800000e+05, 2.43, 4750.0])

## Reaction 513
#reaction('CH3OCHO + H <=> CH3OCO + H2', [6.500000e+05, 2.4, 4471.0])

## Reaction 514
#reaction('CH3OCHO + H <=> CH2OCHO + H2', [6.650000e+05, 2.54, 6756.0])

## Reaction 515
#reaction('CH3OCHO + CH3 <=> CH3OCO + CH4', [7.550000e-01, 3.46, 5481.0])

## Reaction 516
#reaction('CH3OCHO + CH3 <=> CH2OCHO + CH4', [4.520000e-01, 3.65, 7154.0])

## Reaction 517
#reaction('CH3OCHO + CH3O <=> CH3OCO + CH3OH', [5.480000e+11, 0.0, 5000.0])

## Reaction 518
#reaction('CH3OCHO + CH3O <=> CH2OCHO + CH3OH', [2.170000e+11, 0.0, 6458.0])

## Reaction 519
#reaction('CH3OCHO + CH3O2 <=> CH3OCO + CH3O2H', [4.820000e+03, 2.6, 13910.0])

## Reaction 520
#reaction('CH3OCHO + CH3O2 <=> CH2OCHO + CH3O2H', [2.380000e+04, 2.55, 16490.0])

## Reaction 521
#reaction('CH3OCHO + HCO <=> CH3OCO + CH2O', [5.400000e+06, 1.9, 17010.0])

## Reaction 522
#reaction('CH3OCHO + HCO <=> CH2OCHO + CH2O', [1.025000e+05, 2.5, 18430.0])

## Reaction 523
#reaction('CH3OCO <=> CH2OCHO', [1.629000e+12, -0.18, 40670.0])

## Reaction 524
#pdep_arrhenius('CH3OCO <=> CH3 + CO2',
            #[(0.05, 'atm'), 9.720000e+12, -1.31, 9416.7],
            #[(1.0, 'atm'), 1.250000e+16, -1.83, 11340.8],
            #[(10.0, 'atm'), 1.040000e+18, -2.1, 12826.9],
            #[(100.0, 'atm'), 8.690000e+17, -1.81, 13656.7])

## Reaction 525
#pdep_arrhenius('CH3OCO <=> CH3O + CO',
            #[(0.05, 'atm'), 1.030000e+03, 1.29, 25401.0],
            #[(1.0, 'atm'), 4.090000e+05, 0.81, 21969.1],
            #[(10.0, 'atm'), 9.020000e+14, -1.72, 21767.5],
            #[(100.0, 'atm'), 2.825000e+22, -3.44, 23592.4])

## Reaction 526
#reaction('CH2O + HCO <=> CH2OCHO', [1.500000e+11, 0.0, 11900.0])

## Reaction 527
#falloff_reaction('C3H8 (+ M) <=> CH3 + C2H5 (+ M)',
                #kf=[1.290000e+37, -5.84, 97380.0],
                #kf0=[5.640000e+74, -15.74, 98714.0],
                #efficiencies='H2:2.0 H2O:6.0 AR:0.7 CO:1.5 CO2:2.0 CH4:2.0 C2H6:3.0 HE:0.7',
                #falloff=Troe(A=0.31, T3=50.0, T1=3000.0, T2=9000.0))

## Reaction 528
#reaction('NC3H7 + H <=> C3H8', [1.000000e+14, 0.0, 0.0])

## Reaction 529
#reaction('IC3H7 + H <=> C3H8', [1.000000e+14, 0.0, 0.0])

## Reaction 530
#reaction('C3H8 + O2 <=> IC3H7 + HO2', [2.000000e+13, 0.0, 49640.0])

## Reaction 531
#reaction('C3H8 + O2 <=> NC3H7 + HO2', [6.000000e+13, 0.0, 52290.0])

## Reaction 532
#reaction('H + C3H8 <=> H2 + IC3H7', [1.300000e+06, 2.4, 4471.0])

## Reaction 533
#reaction('H + C3H8 <=> H2 + NC3H7', [3.490000e+05, 2.69, 6450.0])

## Reaction 534
#reaction('C3H8 + O <=> IC3H7 + OH', [5.490000e+05, 2.5, 3140.0])

## Reaction 535
#reaction('C3H8 + O <=> NC3H7 + OH', [3.710000e+06, 2.4, 5505.0])

## Reaction 536
#reaction('C3H8 + OH <=> NC3H7 + H2O', [1.054000e+10, 0.97, 1586.0])

## Reaction 537
#reaction('C3H8 + OH <=> IC3H7 + H2O', [4.670000e+07, 1.61, -35.0])

## Reaction 538
#reaction('C3H8 + HO2 <=> IC3H7 + H2O2', [6.320000e+01, 3.37, 13720.0])

## Reaction 539
#reaction('C3H8 + HO2 <=> NC3H7 + H2O2', [4.080000e+01, 3.59, 17160.0])

## Reaction 540
#reaction('CH3 + C3H8 <=> CH4 + IC3H7', [6.400000e+04, 2.17, 7520.0])

## Reaction 541
#reaction('CH3 + C3H8 <=> CH4 + NC3H7', [9.040000e-01, 3.65, 7154.0])

## Reaction 542
#reaction('IC3H7 + C3H8 <=> NC3H7 + C3H8', [3.000000e+10, 0.0, 12900.0])

## Reaction 543
#reaction('C2H3 + C3H8 <=> C2H4 + IC3H7', [1.000000e+11, 0.0, 10400.0])

## Reaction 544
#reaction('C2H3 + C3H8 <=> C2H4 + NC3H7', [1.000000e+11, 0.0, 10400.0])

## Reaction 545
#reaction('C2H5 + C3H8 <=> C2H6 + IC3H7', [1.000000e+11, 0.0, 10400.0])

## Reaction 546
#reaction('C2H5 + C3H8 <=> C2H6 + NC3H7', [1.000000e+11, 0.0, 10400.0])

## Reaction 547
#reaction('C3H8 + C3H5-A <=> NC3H7 + C3H6', [7.940000e+11, 0.0, 20500.0])

## Reaction 548
#reaction('C3H8 + C3H5-A <=> IC3H7 + C3H6', [7.940000e+11, 0.0, 16200.0])

## Reaction 549
#reaction('C3H8 + CH3O <=> NC3H7 + CH3OH', [3.000000e+11, 0.0, 7000.0])

## Reaction 550
#reaction('C3H8 + CH3O <=> IC3H7 + CH3OH', [3.000000e+11, 0.0, 7000.0])

## Reaction 551
#reaction('CH3O2 + C3H8 <=> CH3O2H + NC3H7', [1.386000e+00, 3.97, 18280.0])

## Reaction 552
#reaction('CH3O2 + C3H8 <=> CH3O2H + IC3H7', [1.019000e+01, 3.58, 14810.0])

## Reaction 553
#reaction('C2H5O2 + C3H8 <=> C2H5O2H + NC3H7', [1.386000e+00, 3.97, 18280.0])

## Reaction 554
#reaction('C2H5O2 + C3H8 <=> C2H5O2H + IC3H7', [1.019000e+01, 3.58, 14810.0])

## Reaction 555
#reaction('NC3H7O2 + C3H8 <=> NC3H7O2H + NC3H7', [1.700000e+13, 0.0, 20460.0])

## Reaction 556
#reaction('NC3H7O2 + C3H8 <=> NC3H7O2H + IC3H7', [2.000000e+12, 0.0, 17000.0])

## Reaction 557
#reaction('IC3H7O2 + C3H8 <=> IC3H7O2H + NC3H7', [1.700000e+13, 0.0, 20460.0])

## Reaction 558
#reaction('IC3H7O2 + C3H8 <=> IC3H7O2H + IC3H7', [2.000000e+12, 0.0, 17000.0])

## Reaction 559
#reaction('C3H8 + CH3CO3 <=> IC3H7 + CH3CO3H', [2.000000e+12, 0.0, 17000.0])

## Reaction 560
#reaction('C3H8 + CH3CO3 <=> NC3H7 + CH3CO3H', [1.700000e+13, 0.0, 20460.0])

## Reaction 561
#reaction('C3H8 + O2CHO <=> NC3H7 + HO2CHO', [5.520000e+04, 2.55, 16480.0])

## Reaction 562
#reaction('C3H8 + O2CHO <=> IC3H7 + HO2CHO', [1.475000e+04, 2.6, 13910.0])

## Reaction 563
#reaction('IC3H7 + H <=> C2H5 + CH3', [2.000000e+13, 0.0, 0.0])

## Reaction 564
#reaction('IC3H7 + OH <=> C3H6 + H2O', [2.410000e+13, 0.0, 0.0])

## Reaction 565
#reaction('IC3H7 + O <=> CH3COCH3 + H', [4.818000e+13, 0.0, 0.0])

## Reaction 566
#reaction('IC3H7 + O <=> CH3CHO + CH3', [4.818000e+13, 0.0, 0.0])

## Reaction 567
#reaction('NC3H7 + O2 <=> C3H6 + HO2', [3.000000e-19, 0.0, 3000.0])

## Reaction 568
#reaction('C2H5CHO + NC3H7 <=> C2H5CO + C3H8', [1.700000e+12, 0.0, 8440.0])

## Reaction 569
#reaction('C2H5CHO + IC3H7 <=> C2H5CO + C3H8', [1.700000e+12, 0.0, 8440.0])

## Reaction 570
#reaction('C2H5CHO + C3H5-A <=> C2H5CO + C3H6', [1.700000e+12, 0.0, 8440.0])

## Reaction 571
#falloff_reaction('C2H3 + CH3 (+ M) <=> C3H6 (+ M)',
                #kf=[2.500000e+13, 0.0, 0.0],
                #kf0=[4.270000e+58, -11.94, 9769.8],
                #falloff=Troe(A=0.175, T3=1341.0, T1=60000.0, T2=10140.0))

## Reaction 572
#pdep_arrhenius('CH2(S) + C2H4 <=> CC3H6',
            #[(0.01, 'atm'), 1.440000e+51, -13.1, 14200.0],
            #[(0.1, 'atm'), 3.380000e+54, -13.6, 16500.0],
            #[(1.0, 'atm'), 1.350000e+54, -13.0, 18900.0],
            #[(10.0, 'atm'), 2.730000e+47, -10.8, 14200.0],
            #[(100.0, 'atm'), 2.710000e+50, -11.2, 16700.0],
            #options='duplicate')

## Reaction 573
#pdep_arrhenius('CH2(S) + C2H4 <=> CC3H6',
            #[(0.01, 'atm'), 6.160000e+40, -10.5, 5428.1],
            #[(0.1, 'atm'), 6.930000e+41, -10.3, 6188.9],
            #[(1.0, 'atm'), 1.810000e+37, -8.55, 5521.0],
            #[(10.0, 'atm'), 4.260000e+37, -8.32, 4770.2],
            #[(100.0, 'atm'), 4.680000e+35, -7.37, 4689.1],
            #options='duplicate')

## Reaction 574
#pdep_arrhenius('CH2(S) + C2H4 <=> C3H6',
            #[(0.01, 'atm'), 4.820000e+57, -14.3, 17100.0],
            #[(0.1, 'atm'), 3.840000e+59, -14.4, 18400.0],
            #[(1.0, 'atm'), 2.130000e+58, -13.5, 20400.0],
            #[(10.0, 'atm'), 8.480000e+52, -11.6, 20700.0],
            #[(100.0, 'atm'), 6.070000e+47, -9.85, 22100.0],
            #options='duplicate')

## Reaction 575
#pdep_arrhenius('CH2(S) + C2H4 <=> C3H6',
            #[(0.01, 'atm'), 1.150000e+45, -11.1, 6145.2],
            #[(0.1, 'atm'), 1.830000e+45, -10.7, 6638.5],
            #[(1.0, 'atm'), 1.300000e+40, -8.77, 5863.8],
            #[(10.0, 'atm'), 2.270000e+32, -6.14, 4317.9],
            #[(100.0, 'atm'), 1.280000e+24, -3.49, 2529.9],
            #options='duplicate')

## Reaction 576
#pdep_arrhenius('CH2(S) + C2H4 <=> C3H5-A + H',
            #[(0.01, 'atm'), 8.200000e+19, -2.06, 1150.0],
            #[(0.1, 'atm'), 2.270000e+21, -2.44, 2650.0],
            #[(1.0, 'atm'), 4.440000e+35, -6.55, 13900.0],
            #[(10.0, 'atm'), 1.180000e+28, -4.09, 14000.0],
            #[(100.0, 'atm'), 6.510000e+26, -3.58, 18900.0],
            #options='duplicate')

## Reaction 577
#pdep_arrhenius('CH2(S) + C2H4 <=> C3H5-A + H',
            #[(0.01, 'atm'), 1.080000e+07, 1.62, -3174.6],
            #[(0.1, 'atm'), 1.370000e+05, 2.15, -3799.2],
            #[(1.0, 'atm'), 3.890000e+14, -0.42, 1237.6],
            #[(10.0, 'atm'), 2.450000e+10, 0.67, 750.93],
            #[(100.0, 'atm'), 1.810000e+02, 2.97, -746.03],
            #options='duplicate')

## Reaction 578
#pdep_arrhenius('CH2(S) + C2H4 <=> C2H3 + CH3',
            #[(0.01, 'atm'), 1.770000e+19, -1.94, 6790.0],
            #[(0.1, 'atm'), 1.680000e+19, -1.8, 4310.0],
            #[(1.0, 'atm'), 4.160000e+24, -3.19, 9760.0],
            #[(10.0, 'atm'), 7.890000e+24, -3.07, 13900.0],
            #[(100.0, 'atm'), 7.360000e+29, -4.28, 23800.0],
            #options='duplicate')

## Reaction 579
#pdep_arrhenius('CH2(S) + C2H4 <=> C2H3 + CH3',
            #[(0.01, 'atm'), 4.300000e+12, 0.19, -110.41],
            #[(0.1, 'atm'), 2.260000e+11, 0.54, 47.81],
            #[(1.0, 'atm'), 4.920000e+09, 1.02, 599.77],
            #[(10.0, 'atm'), 1.470000e+08, 1.33, 1228.4],
            #[(100.0, 'atm'), 8.110000e+10, 0.55, 5506.5],
            #options='duplicate')

## Reaction 580
#pdep_arrhenius('C2H3 + CH3 <=> C3H5-A + H',
            #[(0.01, 'atm'), 4.120000e+29, -4.95, 8000.0],
            #[(0.1, 'atm'), 4.860000e+30, -5.03, 11300.0],
            #[(1.0, 'atm'), 5.300000e+29, -4.57, 14400.0],
            #[(10.0, 'atm'), 1.320000e+30, -4.54, 19300.0],
            #[(100.0, 'atm'), 5.160000e+28, -4.03, 23800.0],
            #options='duplicate')

## Reaction 581
#pdep_arrhenius('C2H3 + CH3 <=> C3H5-A + H',
            #[(0.01, 'atm'), 5.730000e+15, -0.77, 1195.9],
            #[(0.1, 'atm'), 2.060000e+13, -0.074, 1428.7],
            #[(1.0, 'atm'), 4.480000e+10, 0.6, 1421.6],
            #[(10.0, 'atm'), 4.100000e+06, 1.71, 1056.9],
            #[(100.0, 'atm'), 1.370000e-01, 3.91, -353.55],
            #options='duplicate')

## Reaction 582
#pdep_arrhenius('C3H6 <=> C2H3 + CH3',
            #[(0.01, 'atm'), 1.880000e+78, -18.7, 130000.0],
            #[(0.1, 'atm'), 8.730000e+76, -17.9, 132000.0],
            #[(1.0, 'atm'), 5.800000e+75, -17.2, 134000.0],
            #[(10.0, 'atm'), 8.120000e+71, -15.8, 136000.0],
            #[(100.0, 'atm'), 2.150000e+64, -13.4, 135000.0],
            #options='duplicate')

## Reaction 583
#pdep_arrhenius('C3H6 <=> C2H3 + CH3',
            #[(0.01, 'atm'), 1.690000e+59, -13.6, 113290.0],
            #[(0.1, 'atm'), 2.000000e+60, -13.7, 114890.0],
            #[(1.0, 'atm'), 6.700000e+54, -11.8, 113840.0],
            #[(10.0, 'atm'), 1.060000e+47, -9.27, 111510.0],
            #[(100.0, 'atm'), 7.290000e+38, -6.7, 108740.0],
            #options='duplicate')

## Reaction 584
#pdep_arrhenius('C3H6 <=> C3H5-A + H',
            #[(0.01, 'atm'), 9.160000e+74, -17.6, 120000.0],
            #[(0.1, 'atm'), 1.730000e+70, -16.0, 120000.0],
            #[(1.0, 'atm'), 1.080000e+71, -15.9, 124860.0],
            #[(10.0, 'atm'), 6.400000e+65, -14.2, 125000.0],
            #[(100.0, 'atm'), 8.050000e+56, -11.5, 122000.0],
            #options='duplicate')

## Reaction 585
#pdep_arrhenius('C3H6 <=> C3H5-A + H',
            #[(0.01, 'atm'), 2.980000e+54, -12.3, 101200.0],
            #[(0.1, 'atm'), 1.370000e+43, -8.87, 96365.0],
            #[(1.0, 'atm'), 6.280000e+42, -8.51, 98004.0],
            #[(10.0, 'atm'), 4.730000e+35, -6.26, 95644.0],
            #[(100.0, 'atm'), 4.340000e+28, -4.06, 93114.0],
            #options='duplicate')

## Reaction 586
#pdep_arrhenius('C3H6 <=> CC3H6',
            #[(0.01, 'atm'), 1.260000e+64, -15.6, 95000.0],
            #[(0.1, 'atm'), 3.940000e+67, -16.2, 101000.0],
            #[(1.0, 'atm'), 6.140000e+68, -16.2, 106000.0],
            #[(10.0, 'atm'), 9.700000e+66, -15.3, 109000.0],
            #[(100.0, 'atm'), 1.450000e+62, -13.6, 110000.0],
            #options='duplicate')

## Reaction 587
#pdep_arrhenius('C3H6 <=> CC3H6',
            #[(0.01, 'atm'), 4.840000e+41, -9.62, 79528.0],
            #[(0.1, 'atm'), 8.070000e+44, -10.2, 82671.0],
            #[(1.0, 'atm'), 1.150000e+47, -10.6, 85502.0],
            #[(10.0, 'atm'), 1.330000e+39, -7.98, 83303.0],
            #[(100.0, 'atm'), 4.970000e+31, -5.6, 80987.0],
            #options='duplicate')

## Reaction 588
#pdep_arrhenius('CC3H6 <=> C3H5-A + H',
            #[(0.01, 'atm'), 2.330000e+63, -14.6, 103000.0],
            #[(0.1, 'atm'), 7.030000e+63, -14.4, 107000.0],
            #[(1.0, 'atm'), 5.070000e+64, -14.3, 112000.0],
            #[(10.0, 'atm'), 4.920000e+61, -13.2, 115000.0],
            #[(100.0, 'atm'), 4.820000e+57, -11.7, 118000.0],
            #options='duplicate')

## Reaction 589
#pdep_arrhenius('CC3H6 <=> C3H5-A + H',
            #[(0.01, 'atm'), 1.120000e+40, -8.37, 85836.0],
            #[(0.1, 'atm'), 1.060000e+41, -8.33, 88499.0],
            #[(1.0, 'atm'), 8.230000e+43, -8.88, 92907.0],
            #[(10.0, 'atm'), 1.270000e+39, -7.33, 93401.0],
            #[(100.0, 'atm'), 1.450000e+28, -4.02, 90995.0],
            #options='duplicate')

## Reaction 590
#pdep_arrhenius('CC3H6 <=> C2H3 + CH3',
            #[(0.01, 'atm'), 8.310000e+64, -15.1, 111000.0],
            #[(0.1, 'atm'), 8.290000e+64, -14.7, 114000.0],
            #[(1.0, 'atm'), 1.000000e+70, -15.7, 122000.0],
            #[(10.0, 'atm'), 1.660000e+67, -14.6, 124000.0],
            #[(100.0, 'atm'), 7.870000e+62, -13.1, 127000.0],
            #options='duplicate')

## Reaction 591
#pdep_arrhenius('CC3H6 <=> C2H3 + CH3',
            #[(0.01, 'atm'), 1.510000e+49, -11.0, 99748.0],
            #[(0.1, 'atm'), 1.050000e+45, -9.46, 99275.0],
            #[(1.0, 'atm'), 3.390000e+50, -10.6, 104220.0],
            #[(10.0, 'atm'), 1.760000e+47, -9.43, 104930.0],
            #[(100.0, 'atm'), 1.880000e+39, -6.93, 103980.0],
            #options='duplicate')

## Reaction 592
#pdep_arrhenius('C3H5-T + H <=> C3H6',
            #[(0.01, 'atm'), 4.960000e+60, -15.2, 18000.0],
            #[(0.1, 'atm'), 3.200000e+62, -15.1, 20100.0],
            #[(1.0, 'atm'), 2.310000e+60, -14.0, 21900.0],
            #[(10.0, 'atm'), 3.690000e+54, -12.0, 22100.0],
            #[(100.0, 'atm'), 1.150000e+50, -10.4, 23300.0],
            #options='duplicate')

## Reaction 593
#pdep_arrhenius('C3H5-T + H <=> C3H6',
            #[(0.01, 'atm'), 1.490000e+48, -12.0, 7203.3],
            #[(0.1, 'atm'), 6.760000e+46, -11.1, 7629.9],
            #[(1.0, 'atm'), 1.090000e+40, -8.66, 6447.8],
            #[(10.0, 'atm'), 2.380000e+31, -5.73, 4506.0],
            #[(100.0, 'atm'), 5.690000e+25, -3.83, 3250.4],
            #options='duplicate')

## Reaction 594
#pdep_arrhenius('C3H5-T + H <=> C3H5-A + H',
            #[(0.01, 'atm'), 2.110000e+17, -1.08, 1290.0],
            #[(0.1, 'atm'), 9.050000e+29, -4.91, 8540.0],
            #[(1.0, 'atm'), 2.980000e+30, -4.79, 12000.0],
            #[(10.0, 'atm'), 8.220000e+28, -4.14, 15400.0],
            #[(100.0, 'atm'), 2.280000e+29, -4.12, 20900.0],
            #options='duplicate')

## Reaction 595
#pdep_arrhenius('C3H5-T + H <=> C3H5-A + H',
            #[(0.01, 'atm'), 6.410000e+03, 2.61, -3778.4],
            #[(0.1, 'atm'), 5.190000e+14, -0.3, 1090.4],
            #[(1.0, 'atm'), 8.170000e+11, 0.49, 1184.6],
            #[(10.0, 'atm'), 2.790000e+09, 1.09, 1187.5],
            #[(100.0, 'atm'), 6.750000e+03, 2.7, 373.8],
            #options='duplicate')

## Reaction 596
#pdep_arrhenius('C3H5-T + H <=> C2H3 + CH3',
            #[(0.01, 'atm'), 3.310000e+16, -0.69, 5200.0],
            #[(0.1, 'atm'), 9.040000e+16, -0.81, 4800.0],
            #[(1.0, 'atm'), 2.010000e+24, -2.86, 10900.0],
            #[(10.0, 'atm'), 2.750000e+26, -3.31, 15800.0],
            #[(100.0, 'atm'), 3.150000e+32, -4.83, 26000.0],
            #options='duplicate')

## Reaction 597
#pdep_arrhenius('C3H5-T + H <=> C2H3 + CH3',
            #[(0.01, 'atm'), 8.040000e+13, -0.14, 1150.0],
            #[(0.1, 'atm'), 7.170000e+10, 0.67, 673.8],
            #[(1.0, 'atm'), 9.970000e+08, 1.36, 1596.4],
            #[(10.0, 'atm'), 7.410000e+07, 1.57, 2108.8],
            #[(100.0, 'atm'), 2.700000e+12, 0.32, 6791.8],
            #options='duplicate')

## Reaction 598
#reaction('C3H6 <=> C3H5-S + H', [7.710000e+69, -16.09, 140000.0])

## Reaction 599
#pdep_arrhenius('C3H3 + C3H3 <=> C6H5 + H',
            #[(0.03947, 'atm'), 1.046700e+54, -11.88, 28757.0],
            #[(1.0, 'atm'), 1.697500e+48, -9.977, 36755.0],
            #[(10.0, 'atm'), 3.671200e+26, -3.879, 28963.0])

## Reaction 600
#pdep_arrhenius('C3H3 + C3H3 <=> C6H6',
            #[(0.03947, 'atm'), 1.640000e+66, -15.902, 27529.0],
            #[(1.0, 'atm'), 3.160900e+55, -12.55, 22264.0],
            #[(10.0, 'atm'), 3.888800e+50, -11.01, 20320.0])

## Reaction 601
#pdep_arrhenius('C3H3 + C3H3 <=> FULVENE',
            #[(0.03947, 'atm'), 7.250000e+65, -16.015, 25035.0],
            #[(1.0, 'atm'), 1.379800e+66, -15.66, 28260.0],
            #[(10.0, 'atm'), 1.258400e+56, -12.61, 23515.0])

## Reaction 602
#reaction('FULVENE <=> C6H6', [2.950000e+31, -4.97, 88470.0])

## Reaction 603
#reaction('FULVENE <=> C6H5 + H', [8.511000e+24, -2.505, 113330.0])

## Reaction 604
#reaction('C4H5-N + C2H2 <=> FULVENE + H', [1.740000e+19, -1.86, 12384.0])

## Reaction 605
#reaction('C4H5-I + C2H2 <=> FULVENE + H', [9.738000e+26, -3.76, 21329.0])

## Reaction 606
#reaction('C3H3 + C3H5-A => FULVENE + H + H', [3.260000e+29, -5.397, 3390.0])

## Reaction 607
#reaction('C4H3-N + C2H3 <=> C3H3 + C3H3', [4.000000e+12, 0.0, 0.0])

## Reaction 608
#reaction('C4H5-2 + C2H <=> C3H3 + C3H3', [4.000000e+12, 0.0, 0.0])

## Reaction 609
#reaction('C4H2 + OH <=> CO + C3H3', [2.580000e+19, -2.44, 3034.0])

## Reaction 610
#reaction('C6H6 + H <=> FULVENE + H', [8.430000e+32, -4.95, 51244.0])

## Reaction 611
#reaction('C5H6 + C3H5-A <=> C5H5 + C3H6', [5.500000e+10, 0.0, 5505.0])

## Reaction 612
#reaction('C3H3 + CH <=> C4H3-N + H', [7.000000e+13, 0.0, 0.0])

## Reaction 613
#reaction('C4H4 + OH <=> CH2O + C3H3', [5.000000e+12, 0.0, 0.0])

## Reaction 614
#reaction('C4H5-2 + OH <=> CH2OH + C3H3', [3.000000e+12, 0.0, 0.0])

## Reaction 615
#reaction('C4H5-2 + O <=> CH2O + C3H3', [3.000000e+13, 0.0, 0.0])

## Reaction 616
#pdep_arrhenius('NC3H7 + O2 <=> NC3H7O2',
            #[(0.01, 'atm'), 9.200000e+08, 0.405, -4398.65],
            #[(0.1, 'atm'), 1.450000e+14, -0.984, -1710.8],
            #[(1.0, 'atm'), 2.090000e+13, -0.499, -938.423],
            #[(10.0, 'atm'), 1.150000e+20, -2.42, 2451.26],
            #[(100.0, 'atm'), 2.070000e+16, -1.3, 803.419])

## Reaction 617
#pdep_arrhenius('IC3H7 + O2 <=> IC3H7O2',
            #[(0.01, 'atm'), 7.330000e+05, 1.33, -6345.64],
            #[(0.1, 'atm'), 2.240000e+11, -0.105, -3697.87],
            #[(1.0, 'atm'), 1.540000e+18, -2.02, -498.567],
            #[(10.0, 'atm'), 6.740000e+27, -4.85, 3779.82],
            #[(100.0, 'atm'), 1.670000e+29, -5.15, 5036.45])

## Reaction 618
#reaction('NC3H7O2 <=> C3H6OOH1-2', [4.090000e+08, 1.1, 30100.0])

## Reaction 619
#reaction('NC3H7O2 <=> C3H6OOH1-3', [3.566000e+06, 1.6, 21000.0])

## Reaction 620
#reaction('IC3H7O2 <=> C3H6OOH2-1', [2.917000e+09, 1.1, 33500.0])

## Reaction 621
#reaction('C3H6OOH1-2 <=> C3H6O1-2 + OH', [1.210000e+09, 1.05, 11300.0])

## Reaction 622
#reaction('C3H6OOH1-3 <=> C3H6O1-3 + OH', [2.640000e+09, 0.71, 18500.0])

## Reaction 623
#reaction('C3H6OOH1-2 <=> C3H6 + HO2', [2.340000e+10, 0.77, 15300.0])

## Reaction 624
#reaction('C3H6OOH1-3 => OH + CH2O + C2H4', [2.890000e+09, 1.3, 26700.0])

## Reaction 625
#reaction('C3H6OOH2-1 <=> C2H3OOH + CH3', [8.560000e+10, 0.85, 30700.0])

## Reaction 626
#reaction('C3H6OOH1-2 => C2H4 + CH2O + OH', [1.310000e+33, -7.01, 48120.0])

## Reaction 627
#reaction('C3H6OOH1-2 + O2 <=> C3H6OOH1-2O2', [1.744000e+14, -0.816, -536.5])

## Reaction 628
#reaction('C3H6OOH1-3 + O2 <=> C3H6OOH1-3O2', [3.433000e+16, -1.627, 198.7])

## Reaction 629
#reaction('C3H6OOH2-1 + O2 <=> C3H6OOH2-1O2', [3.433000e+16, -1.627, 198.7])

## Reaction 630
#reaction('C3H6OOH1-2O2 <=> C3KET12 + OH', [6.000000e+11, 0.0, 26400.0])

## Reaction 631
#reaction('C3H6OOH1-3O2 <=> C3KET13 + OH', [7.500000e+10, 0.0, 21400.0])

## Reaction 632
#reaction('C3H6OOH2-1O2 <=> C3KET21 + OH', [3.000000e+11, 0.0, 23850.0])

## Reaction 633
#reaction('C3H6OOH1-2O2 <=> C3H51-2,3OOH', [2.917000e+09, 1.1, 33500.0])

## Reaction 634
#reaction('C3H6OOH1-3O2 <=> C3H52-1,3OOH', [4.009000e+08, 1.1, 30100.0])

## Reaction 635
#reaction('C3H6OOH2-1O2 <=> C3H51-2,3OOH', [3.566000e+06, 1.6, 21000.0])

## Reaction 636
#reaction('C3H51-2,3OOH <=> AC3H5OOH + HO2', [2.560000e+13, -0.49, 17770.0])

## Reaction 637
#reaction('C3H52-1,3OOH <=> AC3H5OOH + HO2', [1.150000e+14, -0.63, 17250.0])

## Reaction 638
#reaction('C3H51-2,3OOH <=> C3H5O1-2OOH-3 + OH', [4.450000e+09, 0.86, 10800.0])

## Reaction 639
#reaction('C3H51-2,3OOH <=> C3H5O1-3OOH-2 + OH', [2.640000e+09, 0.71, 18500.0])

## Reaction 640
#reaction('C3H52-1,3OOH <=> C3H5O1-2OOH-3 + OH', [1.210000e+09, 1.05, 11300.0])

## Reaction 641
#reaction('C3H5O1-2OOH-3 => CH2CHO + CH2O + OH', [1.000000e+16, 0.0, 43000.0])

## Reaction 642
#reaction('C3H5O1-3OOH-2 => CH2CHO + CH2O + OH', [1.000000e+16, 0.0, 43000.0])

## Reaction 643
#reaction('C3KET13 => CH2O + CH2CHO + OH', [1.000000e+16, 0.0, 43000.0])

## Reaction 644
#reaction('C3KET21 => CH2O + CH3CO + OH', [1.000000e+16, 0.0, 43000.0])

## Reaction 645
#reaction('C3H5O + O2 <=> C2H3CHO + HO2', [1.000000e+12, 0.0, 6000.0])

## Reaction 646
#reaction('C2H3OOH <=> CH2CHO + OH', [8.400000e+14, 0.0, 43000.0])

## Reaction 647
#reaction('C3H6O1-2 <=> C2H4 + CH2O', [6.000000e+14, 0.0, 60000.0])

## Reaction 648
#reaction('C3H6O1-2 + OH => CH2O + C2H3 + H2O', [5.000000e+12, 0.0, 0.0])

## Reaction 649
#reaction('C3H6O1-2 + H => CH2O + C2H3 + H2', [2.630000e+07, 2.0, 5000.0])

## Reaction 650
#reaction('C3H6O1-2 + O => CH2O + C2H3 + OH', [8.430000e+13, 0.0, 5200.0])

## Reaction 651
#reaction('C3H6O1-2 + HO2 => CH2O + C2H3 + H2O2', [1.000000e+13, 0.0, 15000.0])

## Reaction 652
#reaction('C3H6O1-2 + CH3O2 => CH2O + C2H3 + CH3O2H', [1.000000e+13, 0.0, 19000.0])

## Reaction 653
#reaction('C3H6O1-2 + CH3 => CH2O + C2H3 + CH4', [2.000000e+11, 0.0, 10000.0])

## Reaction 654
#reaction('C3H6O1-2 + O2 <=> C2H3OCH2 + HO2', [2.000000e+13, 0.0, 52800.0])

## Reaction 655
#reaction('C3H6O1-2 + H <=> C2H3OCH2 + H2', [5.310000e+04, 2.81, 7490.0])

## Reaction 656
#reaction('C3H6O1-2 + OH <=> C2H3OCH2 + H2O', [3.760000e+03, 2.78, -1810.2])

## Reaction 657
#reaction('C3H6O1-2 + HO2 <=> C2H3OCH2 + H2O2', [3.986000e-02, 4.3, 15333.0])

## Reaction 658
#reaction('C3H6O1-2 + CH3O2 <=> C2H3OCH2 + CH3O2H', [1.995000e-02, 4.3, 15333.0])

## Reaction 659
#reaction('C3H6O1-2 + O <=> C2H3OCH2 + OH', [9.690000e+02, 3.23, 4658.0])

## Reaction 660
#reaction('C3H6O1-2 + CH3 <=> C2H3OCH2 + CH4', [3.300000e+02, 3.3, 12290.0])

## Reaction 661
#reaction('C3H6O1-2 + C2H5 <=> C2H3OCH2 + C2H6', [5.000000e+10, 0.0, 13400.0])

## Reaction 662
#reaction('C3H6O1-3 <=> C2H4 + CH2O', [6.000000e+14, 0.0, 60000.0])

## Reaction 663
#reaction('C3H6O1-3 + OH => CH2O + C2H3 + H2O', [5.000000e+12, 0.0, 0.0])

## Reaction 664
#reaction('C3H6O1-3 + O => CH2O + C2H3 + OH', [8.430000e+13, 0.0, 5200.0])

## Reaction 665
#reaction('C3H6O1-3 + H => CH2O + C2H3 + H2', [2.630000e+07, 2.0, 5000.0])

## Reaction 666
#reaction('C3H6O1-3 + CH3O2 => CH2O + C2H3 + CH3O2H', [1.000000e+13, 0.0, 19000.0])

## Reaction 667
#reaction('C3H6O1-3 + HO2 => CH2O + C2H3 + H2O2', [1.000000e+13, 0.0, 15000.0])

## Reaction 668
#reaction('C3H6O1-3 + CH3 => CH2O + C2H3 + CH4', [2.000000e+11, 0.0, 10000.0])

## Reaction 669
#reaction('NC3H7O2 <=> C3H6 + HO2', [6.240000e+08, 1.25, 29600.0])

## Reaction 670
#reaction('C3H6 + O <=> C2H5 + HCO', [7.450000e+06, 1.88, 183.0])

## Reaction 671
#reaction('C3H6 + O => CH2CO + CH3 + H', [3.050000e+06, 1.88, 183.0])

## Reaction 672
#reaction('C3H6 + O => CH3CHCO + H + H', [3.050000e+06, 1.88, 183.0])

## Reaction 673
#reaction('C3H6 + O <=> C3H5-A + OH', [5.240000e+11, 0.7, 5884.0])

## Reaction 674
#reaction('C3H6 + O <=> C3H5-S + OH', [1.200000e+11, 0.7, 8959.1])

## Reaction 675
#reaction('C3H6 + O <=> C3H5-T + OH', [6.030000e+10, 0.7, 7632.0])

## Reaction 676
#reaction('C3H6 + OH <=> C3H5-A + H2O', [4.460000e+06, 2.072, 1050.8])

## Reaction 677
#reaction('C3H6 + OH <=> C3H5-S + H2O', [1.860000e+05, 2.369, 2502.0])

## Reaction 678
#reaction('C3H6 + OH <=> C3H5-T + H2O', [1.800000e+06, 1.979, 2235.2])

## Reaction 679
#reaction('C3H6 + CH3O <=> C3H5-A + CH3OH', [8.400000e+10, 0.0, 2600.0])

## Reaction 680
#reaction('C3H6 + HO2 <=> C3H5-S + H2O2', [9.570000e+02, 3.059, 20798.6])

## Reaction 681
#reaction('C3H6 + HO2 <=> C3H5-T + H2O2', [1.560000e+04, 2.82, 24427.9])

## Reaction 682
#reaction('IC3H7O2 <=> C3H6 + HO2', [1.224000e+09, 1.28, 30000.0])

## Reaction 683
#pdep_arrhenius('C3H6 + HO2 <=> C3H6OOH2-1',
            #[(0.013, 'atm'), 1.310000e+13, -1.84, 8561.0],
            #[(0.9869, 'atm'), 5.930000e+17, -2.61, 11533.0],
            #[(9.87, 'atm'), 8.340000e+24, -4.4, 16440.0],
            #[(98.69, 'atm'), 3.750000e+23, -3.68, 17965.0])

## Reaction 684
#pdep_arrhenius('C3H6 + HO2 <=> C3H6O1-2 + OH',
            #[(0.013, 'atm'), 3.730000e+03, 2.64, 11173.0],
            #[(0.9869, 'atm'), 1.780000e+12, 0.11, 16137.0],
            #[(9.87, 'atm'), 3.900000e+17, -1.4, 20077.0],
            #[(98.69, 'atm'), 1.130000e+19, -1.68, 23587.0])

## Reaction 685
#pdep_arrhenius('C3H6 + HO2 <=> IC3H7 + O2',
            #[(0.013, 'atm'), 1.020000e+07, 1.16, 10273.0],
            #[(0.9869, 'atm'), 1.310000e+20, -2.58, 19078.0],
            #[(9.87, 'atm'), 4.140000e+28, -4.92, 26212.0],
            #[(98.69, 'atm'), 8.870000e+22, -3.09, 26586.0])

## Reaction 686
#pdep_arrhenius('C3H6OOH2-1 <=> C3H6O1-2 + OH',
            #[(0.013, 'atm'), 2.660000e+35, -8.36, 18056.0],
            #[(0.9869, 'atm'), 1.960000e+35, -7.66, 20595.0],
            #[(9.87, 'atm'), 2.240000e+33, -6.75, 21619.0],
            #[(98.69, 'atm'), 8.480000e+26, -4.58, 20278.0])

## Reaction 687
#reaction('C3H6 + HO2 <=> C3H5-A + H2O2', [3.070000e-02, 4.403, 13547.2])

## Reaction 688
#pdep_arrhenius('C3H6 + H <=> NC3H7',
            #[(0.0013, 'atm'), 7.990000e+81, -23.161, 22239.0],
            #[(0.04, 'atm'), 4.240000e+68, -18.427, 19665.0],
            #[(1.0, 'atm'), 1.040000e+49, -11.5, 15359.0],
            #[(10.0, 'atm'), 6.200000e+41, -8.892, 14637.0],
            #options='duplicate')

## Reaction 689
#pdep_arrhenius('C3H6 + H <=> NC3H7',
            #[(0.0013, 'atm'), 1.850000e+26, -5.83, 3865.8],
            #[(0.04, 'atm'), 2.820000e+30, -6.49, 5470.8],
            #[(1.0, 'atm'), 3.780000e+28, -5.57, 5625.1],
            #[(10.0, 'atm'), 1.460000e+25, -4.28, 5247.8],
            #[(100.0, 'atm'), 4.220000e+27, -4.39, 9345.8],
            #options='duplicate')

## Reaction 690
#pdep_arrhenius('C3H6 + H <=> C2H4 + CH3',
            #[(0.0013, 'atm'), 1.540000e+09, 1.35, 2542.0],
            #[(0.04, 'atm'), 7.880000e+10, 0.87, 3599.6],
            #[(1.0, 'atm'), 2.670000e+12, 0.47, 5431.1],
            #[(10.0, 'atm'), 9.250000e+22, -2.6, 12898.0],
            #[(100.0, 'atm'), 1.320000e+23, -2.42, 16500.0],
            #options='duplicate')

## Reaction 691
#pdep_arrhenius('C3H6 + H <=> C2H4 + CH3',
            #[(10.0, 'atm'), 1.240000e+05, 2.52, 3679.1],
            #[(100.0, 'atm'), 2.510000e+03, 2.91, 3980.9],
            #options='duplicate')

## Reaction 692
#pdep_arrhenius('C3H6 + H <=> IC3H7',
            #[(0.0013, 'atm'), 1.350000e+44, -10.68, 8196.4],
            #[(0.04, 'atm'), 2.110000e+57, -14.23, 15147.0],
            #[(1.0, 'atm'), 3.260000e+61, -14.94, 20161.0],
            #[(10.0, 'atm'), 5.300000e+56, -13.12, 20667.0],
            #[(100.0, 'atm'), 1.110000e+50, -10.8, 20202.0],
            #options='duplicate')

## Reaction 693
#pdep_arrhenius('C3H6 + H <=> IC3H7',
            #[(0.0013, 'atm'), 2.170000e+130, -32.58, 136140.0],
            #[(0.04, 'atm'), 2.250000e+29, -5.84, 4241.9],
            #[(1.0, 'atm'), 1.060000e+30, -5.63, 5613.4],
            #[(10.0, 'atm'), 6.110000e+26, -4.44, 5182.3],
            #[(100.0, 'atm'), 2.730000e+23, -3.26, 4597.0],
            #options='duplicate')

## Reaction 694
#pdep_arrhenius('C2H4 + CH3 <=> NC3H7',
            #[(0.0013, 'atm'), 8.670000e+48, -12.54, 18206.0],
            #[(0.04, 'atm'), 1.060000e+49, -12.04, 20001.0],
            #[(1.0, 'atm'), 7.670000e+47, -11.17, 22366.0],
            #[(10.0, 'atm'), 1.810000e+45, -10.03, 23769.0],
            #[(100.0, 'atm'), 2.040000e+40, -8.25, 24214.0],
            #options='duplicate')

## Reaction 695
#pdep_arrhenius('C2H4 + CH3 <=> NC3H7',
            #[(0.0013, 'atm'), 1.120000e+43, -11.3, 13080.0],
            #[(0.04, 'atm'), 7.280000e+39, -9.88, 13164.0],
            #[(1.0, 'atm'), 2.600000e+33, -7.46, 12416.0],
            #[(10.0, 'atm'), 3.850000e+27, -5.38, 11455.0],
            #[(100.0, 'atm'), 1.660000e+21, -3.17, 10241.0],
            #options='duplicate')

## Reaction 696
#reaction('C3H6 + H <=> C3H5-A + H2', [3.644000e+05, 2.455, 4361.2])

## Reaction 697
#reaction('C3H6 + H <=> C3H5-T + H2', [1.498000e+02, 3.381, 8909.5])

## Reaction 698
#reaction('C3H6 + H <=> C3H5-S + H2', [5.101000e+02, 3.234, 12357.0],
        #options='duplicate')

## Reaction 699
#reaction('C3H6 + H <=> C3H5-S + H2', [3.969000e+02, 3.252, 12007.0],
        #options='duplicate')

## Reaction 700
#reaction('C2H4 + CH3 <=> C2H3 + CH4', [9.760000e+02, 2.947, 15148.0],
        #options='duplicate')

## Reaction 701
#reaction('C2H4 + CH3 <=> C2H3 + CH4', [8.130000e-05, 4.417, 8835.8],
        #options='duplicate')

## Reaction 702
#pdep_arrhenius('C3H5-A + HO2 <=> C3H5O + OH',
            #[(0.01, 'atm'), 1.020000e+13, -0.158, -1417.0],
            #[(0.1, 'atm'), 4.980000e+14, -0.642, -349.1],
            #[(1.0, 'atm'), 7.770000e+17, -1.52, 2379.2],
            #[(10.0, 'atm'), 2.930000e+15, -0.684, 3615.3],
            #[(100.0, 'atm'), 1.640000e+04, 2.74, 1144.4])

## Reaction 703
#pdep_arrhenius('C3H5-A + HO2 <=> AC3H5OOH',
            #[(0.01, 'atm'), 1.910000e+31, -7.23, 1336.2],
            #[(0.1, 'atm'), 6.310000e+42, -10.3, 5568.9],
            #[(1.0, 'atm'), 1.030000e+45, -10.6, 7851.5],
            #[(10.0, 'atm'), 2.790000e+37, -7.92, 6497.9],
            #[(100.0, 'atm'), 1.440000e+32, -6.01, 6053.6])

## Reaction 704
#pdep_arrhenius('C3H5-A + HO2 <=> C2H3CHO + H2O',
            #[(0.01, 'atm'), 1.090000e+00, 3.01, -3421.1],
            #[(0.1, 'atm'), 6.350000e+01, 2.5, -2341.4],
            #[(1.0, 'atm'), 6.050000e+05, 1.39, 595.1],
            #[(10.0, 'atm'), 3.100000e+05, 1.59, 2677.6],
            #[(100.0, 'atm'), 5.070000e-05, 4.59, 927.5])

## Reaction 705
#pdep_arrhenius('AC3H5OOH <=> C2H3CHO + H2O',
            #[(0.01, 'atm'), 1.990000e+50, -12.7, 53531.9],
            #[(0.1, 'atm'), 4.720000e+47, -11.5, 54360.9],
            #[(1.0, 'atm'), 1.500000e+40, -8.84, 53179.2],
            #[(10.0, 'atm'), 2.540000e+28, -5.0, 49919.4],
            #[(100.0, 'atm'), 1.480000e+16, -1.12, 45949.3])

## Reaction 706
#pdep_arrhenius('AC3H5OOH <=> C3H5O + OH',
            #[(0.01, 'atm'), 1.490000e+58, -13.9, 54266.9],
            #[(0.1, 'atm'), 1.800000e+54, -12.4, 54193.8],
            #[(1.0, 'atm'), 3.360000e+46, -9.81, 52468.5],
            #[(10.0, 'atm'), 2.390000e+36, -6.54, 49429.0],
            #[(100.0, 'atm'), 1.280000e+27, -3.61, 46333.1])

## Reaction 707
#pdep_arrhenius('C3H5O <=> C2H3 + CH2O',
            #[(0.001, 'atm'), 7.260000e+06, 0.182, 17815.5],
            #[(0.01, 'atm'), 6.970000e+16, -2.5, 20878.7],
            #[(0.1, 'atm'), 6.640000e+23, -4.23, 23565.0],
            #[(1.0, 'atm'), 1.070000e+26, -4.56, 24622.9],
            #[(10.0, 'atm'), 6.500000e+29, -5.37, 26645.0],
            #[(100.0, 'atm'), 4.630000e+31, -5.59, 28915.3],
            #[(1000.0, 'atm'), 8.520000e+25, -3.61, 27863.4])

## Reaction 708
#pdep_arrhenius('C3H5O <=> C2H3OCH2',
            #[(0.001, 'atm'), 3.170000e+20, -4.15, 12121.3],
            #[(0.01, 'atm'), 4.790000e+24, -5.03, 14606.1],
            #[(0.1, 'atm'), 1.900000e+26, -5.16, 16124.4],
            #[(1.0, 'atm'), 1.510000e+28, -5.4, 18165.4],
            #[(10.0, 'atm'), 2.420000e+28, -5.17, 19691.2],
            #[(100.0, 'atm'), 5.570000e+24, -3.86, 19395.2],
            #[(1000.0, 'atm'), 1.350000e+18, -1.73, 17386.5])

## Reaction 709
#pdep_arrhenius('C3H5O <=> CH2CH2CHO',
            #[(0.001, 'atm'), 5.250000e-49, 15.5, -15639.9],
            #[(0.01, 'atm'), 1.460000e-88, 27.6, -35995.0],
            #[(0.1, 'atm'), 4.440000e-22, 8.38, -3819.0],
            #[(1.0, 'atm'), 6.230000e+12, -1.44, 10829.2],
            #[(10.0, 'atm'), 3.480000e+42, -9.91, 25297.9],
            #[(100.0, 'atm'), 1.880000e+38, -8.16, 25974.5],
            #[(1000.0, 'atm'), 1.670000e+21, -2.74, 20337.7])

## Reaction 710
#pdep_arrhenius('C3H5O <=> C2H3CHO + H',
            #[(0.001, 'atm'), 3.000000e+15, -2.31, 14667.9],
            #[(0.01, 'atm'), 1.500000e+22, -3.96, 18283.0],
            #[(0.1, 'atm'), 1.950000e+23, -3.99, 19143.3],
            #[(1.0, 'atm'), 1.150000e+25, -4.24, 20311.2],
            #[(10.0, 'atm'), 1.760000e+28, -4.89, 22765.2],
            #[(100.0, 'atm'), 1.410000e+27, -4.28, 23770.6],
            #[(1000.0, 'atm'), 2.570000e+20, -2.06, 22040.1])

## Reaction 711
#pdep_arrhenius('C3H5O <=> C2H4 + HCO',
            #[(0.001, 'atm'), 6.620000e+16, -2.84, 13197.0],
            #[(0.01, 'atm'), 1.260000e+20, -3.53, 15469.2],
            #[(0.1, 'atm'), 2.130000e+21, -3.64, 16584.5],
            #[(1.0, 'atm'), 1.070000e+24, -4.16, 18985.0],
            #[(10.0, 'atm'), 8.420000e+25, -4.4, 22382.6],
            #[(100.0, 'atm'), 1.860000e+21, -2.73, 23658.8],
            #[(1000.0, 'atm'), 4.750000e+08, 1.14, 20922.5])

## Reaction 712
#pdep_arrhenius('C2H3OCH2 <=> C2H3 + CH2O',
            #[(0.001, 'atm'), 3.300000e+09, -0.638, 19747.8],
            #[(0.01, 'atm'), 3.360000e+21, -3.9, 23945.2],
            #[(0.1, 'atm'), 2.910000e+29, -5.9, 27249.7],
            #[(1.0, 'atm'), 1.830000e+34, -6.94, 30690.4],
            #[(10.0, 'atm'), 9.720000e+33, -6.5, 33002.5],
            #[(100.0, 'atm'), 2.680000e+27, -4.26, 33305.6],
            #[(1000.0, 'atm'), 8.810000e+14, -0.326, 31553.1])

## Reaction 713
#pdep_arrhenius('C2H3OCH2 <=> CH2CH2CHO',
            #[(0.001, 'atm'), 8.010000e-92, 27.8, -37321.2],
            #[(0.01, 'atm'), 7.780000e-11, 3.7, -2766.9],
            #[(0.1, 'atm'), 5.110000e+15, -2.76, 15937.6],
            #[(1.0, 'atm'), 4.480000e+25, -5.2, 21532.2],
            #[(10.0, 'atm'), 3.970000e+34, -7.41, 28116.9],
            #[(100.0, 'atm'), 5.620000e+22, -3.56, 25806.7],
            #[(1000.0, 'atm'), 2.510000e+20, -2.63, 29288.4])

## Reaction 714
#pdep_arrhenius('C2H3OCH2 <=> C2H3CHO + H',
            #[(0.001, 'atm'), 4.930000e+24, -5.05, 20108.4],
            #[(0.01, 'atm'), 2.140000e+28, -5.8, 22219.4],
            #[(0.1, 'atm'), 1.930000e+32, -6.64, 25108.2],
            #[(1.0, 'atm'), 8.600000e+34, -7.11, 28209.1],
            #[(10.0, 'atm'), 2.170000e+34, -6.64, 30647.6],
            #[(100.0, 'atm'), 4.170000e+28, -4.71, 31231.9],
            #[(1000.0, 'atm'), 3.980000e+18, -1.62, 30129.8])

## Reaction 715
#pdep_arrhenius('C2H3OCH2 <=> C2H4 + HCO',
            #[(0.001, 'atm'), 8.230000e+26, -5.84, 19356.9],
            #[(0.01, 'atm'), 1.320000e+29, -6.21, 21293.6],
            #[(0.1, 'atm'), 3.470000e+32, -6.96, 24197.3],
            #[(1.0, 'atm'), 1.440000e+36, -7.76, 28007.8],
            #[(10.0, 'atm'), 9.720000e+37, -8.02, 32394.6],
            #[(100.0, 'atm'), 2.430000e+31, -5.81, 34295.8],
            #[(1000.0, 'atm'), 3.730000e+14, -0.726, 32008.3])

## Reaction 716
#pdep_arrhenius('CH2CH2CHO <=> C2H3 + CH2O',
            #[(0.001, 'atm'), 6.890000e-69, 21.5, 2638.0],
            #[(0.01, 'atm'), 5.340000e-33, 11.1, 16749.1],
            #[(0.1, 'atm'), 6.110000e+26, -6.01, 44116.7],
            #[(1.0, 'atm'), 8.040000e+35, -8.31, 46919.7],
            #[(10.0, 'atm'), 5.520000e+40, -9.19, 50508.7],
            #[(100.0, 'atm'), 5.850000e+35, -7.18, 52038.4],
            #[(1000.0, 'atm'), 1.930000e+19, -1.94, 48440.0])

## Reaction 717
#pdep_arrhenius('CH2CH2CHO <=> C2H3CHO + H',
            #[(0.001, 'atm'), 1.610000e+10, -1.24, 32371.3],
            #[(0.01, 'atm'), 5.840000e+15, -2.61, 32878.4],
            #[(0.1, 'atm'), 3.640000e+23, -4.6, 34275.3],
            #[(1.0, 'atm'), 7.580000e+31, -6.63, 37895.4],
            #[(10.0, 'atm'), 2.860000e+32, -6.3, 39990.7],
            #[(100.0, 'atm'), 1.570000e+23, -3.14, 38011.7],
            #[(1000.0, 'atm'), 4.520000e+12, 0.214, 34570.5])

## Reaction 718
#pdep_arrhenius('CH2CH2CHO <=> C2H4 + HCO',
            #[(0.001, 'atm'), 2.900000e+32, -7.24, 25687.5],
            #[(0.01, 'atm'), 5.300000e+33, -7.28, 27100.6],
            #[(0.1, 'atm'), 2.000000e+35, -7.41, 29027.3],
            #[(1.0, 'atm'), 1.010000e+34, -6.7, 30018.1],
            #[(10.0, 'atm'), 9.760000e+27, -4.63, 28923.9],
            #[(100.0, 'atm'), 2.110000e+19, -1.85, 26239.8],
            #[(1000.0, 'atm'), 1.590000e+13, 0.063, 24086.3])

## Reaction 719
#pdep_arrhenius('C2H3 + CH2O <=> C2H3CHO + H',
            #[(0.001, 'atm'), 2.600000e+04, 2.26, 1510.3],
            #[(0.01, 'atm'), 5.130000e+04, 2.17, 1675.5],
            #[(0.1, 'atm'), 3.990000e+05, 1.91, 2218.3],
            #[(1.0, 'atm'), 1.750000e+07, 1.45, 3428.0],
            #[(10.0, 'atm'), 1.350000e+09, 0.933, 5173.0],
            #[(100.0, 'atm'), 2.240000e+11, 0.357, 8001.3],
            #[(1000.0, 'atm'), 6.010000e+05, 2.09, 7895.6])

## Reaction 720
#pdep_arrhenius('C2H3 + CH2O <=> C2H4 + HCO',
            #[(0.001, 'atm'), 1.110000e+07, 1.09, 1807.2],
            #[(0.01, 'atm'), 2.470000e+07, 0.993, 1994.9],
            #[(0.1, 'atm'), 2.470000e+08, 0.704, 2596.2],
            #[(1.0, 'atm'), 1.420000e+10, 0.209, 3934.2],
            #[(10.0, 'atm'), 3.450000e+13, -0.726, 6944.3],
            #[(100.0, 'atm'), 3.310000e+14, -0.866, 10965.7],
            #[(1000.0, 'atm'), 1.650000e+01, 3.17, 9399.8])

## Reaction 721
#reaction('C3H6 + O2 <=> C3H5-A + HO2', [5.960000e+19, -1.67, 46192.1])

## Reaction 722
#reaction('C3H6 + O2 <=> C3H5-S + HO2', [2.000000e+13, 0.0, 62270.0])

## Reaction 723
#reaction('C3H6 + O2 <=> C3H5-T + HO2', [1.000000e+13, 0.0, 58770.0])

## Reaction 724
#reaction('C3H6 + CH3 <=> C3H5-A + CH4', [2.210000e+00, 3.5, 5675.0])

## Reaction 725
#reaction('C3H6 + CH3 <=> C3H5-S + CH4', [1.348000e+00, 3.5, 12850.0])

## Reaction 726
#reaction('C3H6 + CH3 <=> C3H5-T + CH4', [8.400000e-01, 3.5, 11660.0])

## Reaction 727
#reaction('C3H6 + C2H5 <=> C3H5-A + C2H6', [1.000000e+11, 0.0, 9800.0])

## Reaction 728
#reaction('C3H6 + CH3O2 <=> C3H5-A + CH3O2H', [7.680000e-02, 4.403, 13547.2])

## Reaction 729
#reaction('C3H6 + CH3CO3 <=> C3H5-A + CH3CO3H', [7.680000e-02, 4.403, 13547.2])

## Reaction 730
#reaction('C3H6 + C2H5O2 <=> C3H5-A + C2H5O2H', [7.680000e-02, 4.403, 13547.2])

## Reaction 731
#reaction('C3H6 + NC3H7O2 <=> C3H5-A + NC3H7O2H', [7.680000e-02, 4.403, 13547.2])

## Reaction 732
#reaction('C3H6 + IC3H7O2 <=> C3H5-A + IC3H7O2H', [7.680000e-02, 4.403, 13547.2])

## Reaction 733
#pdep_arrhenius('C3H6 + OH <=> C3H5OH + H',
            #[(0.0013, 'atm'), 2.670000e+13, 0.05, 10611.0],
            #[(0.01, 'atm'), 2.750000e+13, 0.05, 10623.0],
            #[(0.013, 'atm'), 2.870000e+13, 0.04, 10634.0],
            #[(0.025, 'atm'), 1.590000e+14, -0.16, 11125.0],
            #[(0.1, 'atm'), 3.100000e+14, -0.22, 11407.0],
            #[(0.1315, 'atm'), 3.780000e+14, -0.24, 11458.0],
            #[(1.0, 'atm'), 9.150000e+07, 1.42, 10087.0],
            #[(10.0, 'atm'), 3.660000e+05, 2.14, 10410.0],
            #[(100.0, 'atm'), 8.190000e+02, 2.84, 10481.0])

## Reaction 734
#pdep_arrhenius('C3H6 + OH <=> C2H3OH + CH3',
            #[(0.0013, 'atm'), 1.290000e+06, 1.65, 1233.0],
            #[(0.01, 'atm'), 1.820000e+04, 2.1, 1162.0],
            #[(0.013, 'atm'), 2.040000e+03, 2.48, 1128.0],
            #[(0.025, 'atm'), 2.880000e+02, 2.8, 1152.0],
            #[(0.1, 'atm'), 1.400000e+01, 3.21, 1208.0],
            #[(0.1315, 'atm'), 7.710000e+00, 3.29, 1216.0],
            #[(1.0, 'atm'), 1.130000e+04, 2.5, 3238.0],
            #[(10.0, 'atm'), 2.410000e+19, -1.74, 13107.0],
            #[(100.0, 'atm'), 3.300000e-01, 3.7, 3665.0])

## Reaction 735
#pdep_arrhenius('C3H6 + OH <=> IC3H5OH + H',
            #[(0.0013, 'atm'), 2.870000e+00, 2.92, 625.0],
            #[(0.01, 'atm'), 4.840000e-01, 2.98, 704.0],
            #[(0.013, 'atm'), 3.130000e-01, 3.04, 721.0],
            #[(0.025, 'atm'), 9.330000e-03, 3.62, 677.0],
            #[(0.1, 'atm'), 4.640000e-05, 4.48, 687.0],
            #[(0.1315, 'atm'), 2.710000e-05, 4.56, 707.0],
            #[(1.0, 'atm'), 7.650000e-07, 5.05, 874.0],
            #[(10.0, 'atm'), 2.640000e+15, -0.8, 12728.0],
            #[(100.0, 'atm'), 4.870000e-04, 4.32, 4020.0])

## Reaction 736
#pdep_arrhenius('C3H6 + OH <=> SC3H5OH + H',
            #[(0.0013, 'atm'), 3.470000e+06, 1.53, 4288.0],
            #[(0.01, 'atm'), 1.080000e+07, 1.34, 4576.0],
            #[(0.013, 'atm'), 9.760000e+06, 1.33, 4589.0],
            #[(0.025, 'atm'), 5.140000e+06, 1.36, 4594.0],
            #[(0.1, 'atm'), 3.130000e+05, 1.69, 4603.0],
            #[(0.1315, 'atm'), 1.390000e+05, 1.8, 4603.0],
            #[(1.0, 'atm'), 1.030000e+02, 2.83, 4530.0],
            #[(10.0, 'atm'), 3.400000e-02, 3.89, 4390.0],
            #[(100.0, 'atm'), 4.460000e-06, 5.03, 4132.0])

## Reaction 737
#pdep_arrhenius('C3H6 + OH <=> CH3CHO + CH3',
            #[(0.0013, 'atm'), 6.930000e+05, 1.49, -536.0],
            #[(0.01, 'atm'), 5.940000e+03, 2.01, -560.0],
            #[(0.013, 'atm'), 1.100000e+03, 2.22, -680.0],
            #[(0.025, 'atm'), 1.070000e+02, 2.5, -759.0],
            #[(0.1, 'atm'), 7.830000e-01, 3.1, -919.0],
            #[(0.1315, 'atm'), 3.070000e-01, 3.22, -946.0],
            #[(1.0, 'atm'), 3.160000e-04, 4.05, -1144.0],
            #[(10.0, 'atm'), 7.590000e-06, 4.49, -680.0],
            #[(100.0, 'atm'), 5.450000e-05, 4.22, 1141.0])

## Reaction 738
#pdep_arrhenius('C3H6 + OH <=> C3H6OH1-2',
            #[(0.0013, 'atm'), 2.300000e+78, -20.7, 32402.0],
            #[(0.01, 'atm'), 2.740000e+77, -20.0, 33874.0],
            #[(0.013, 'atm'), 1.070000e+76, -19.58, 32874.0],
            #[(0.025, 'atm'), 3.680000e+73, -18.79, 31361.0],
            #[(0.1, 'atm'), 1.040000e+68, -17.01, 27909.0],
            #[(0.1315, 'atm'), 7.230000e+66, -16.64, 27162.0],
            #[(1.0, 'atm'), 1.950000e+59, -14.17, 23079.0],
            #[(10.0, 'atm'), 7.580000e+53, -12.23, 22976.0],
            #[(100.0, 'atm'), 1.430000e+48, -10.23, 23772.0],
            #options='duplicate')

## Reaction 739
#pdep_arrhenius('C3H6 + OH <=> C3H6OH1-2',
            #[(0.0013, 'atm'), 6.410000e+59, -15.84, 11594.0],
            #[(0.01, 'atm'), 7.280000e+59, -15.51, 12898.0],
            #[(0.013, 'atm'), 2.790000e+59, -15.34, 12913.0],
            #[(0.025, 'atm'), 2.650000e+58, -14.93, 12936.0],
            #[(0.1, 'atm'), 1.350000e+56, -14.04, 12945.0],
            #[(0.1315, 'atm'), 3.980000e+55, -13.85, 12887.0],
            #[(1.0, 'atm'), 1.550000e+50, -12.04, 11493.0],
            #[(10.0, 'atm'), 6.410000e+41, -9.35, 8921.0],
            #[(100.0, 'atm'), 2.300000e+32, -6.31, 6088.0],
            #options='duplicate')

## Reaction 740
#pdep_arrhenius('C3H6 + OH <=> C3H6OH2-1',
            #[(0.0013, 'atm'), 2.140000e+59, -15.84, 11594.0],
            #[(0.01, 'atm'), 2.430000e+59, -15.51, 12898.0],
            #[(0.013, 'atm'), 9.300000e+58, -15.34, 12913.0],
            #[(0.025, 'atm'), 8.830000e+57, -14.93, 12936.0],
            #[(0.1, 'atm'), 4.500000e+55, -14.04, 12945.0],
            #[(0.1315, 'atm'), 1.330000e+55, -13.85, 12887.0],
            #[(1.0, 'atm'), 5.180000e+49, -12.04, 11493.0],
            #[(10.0, 'atm'), 2.140000e+41, -9.35, 8921.0],
            #[(100.0, 'atm'), 7.650000e+31, -6.31, 6088.0],
            #options='duplicate')

## Reaction 741
#pdep_arrhenius('C3H6 + OH <=> C3H6OH2-1',
            #[(0.0013, 'atm'), 7.680000e+77, -20.7, 32402.0],
            #[(0.01, 'atm'), 9.130000e+76, -20.0, 33874.0],
            #[(0.013, 'atm'), 3.550000e+75, -19.58, 32874.0],
            #[(0.025, 'atm'), 1.230000e+73, -18.79, 31361.0],
            #[(0.1, 'atm'), 3.450000e+67, -17.01, 27909.0],
            #[(0.1315, 'atm'), 2.410000e+66, -16.64, 27162.0],
            #[(1.0, 'atm'), 6.500000e+58, -14.17, 23079.0],
            #[(10.0, 'atm'), 2.530000e+53, -12.23, 22976.0],
            #[(100.0, 'atm'), 4.780000e+47, -10.23, 23772.0],
            #options='duplicate')

## Reaction 742
#pdep_arrhenius('C3H6OH1-2 + O2 <=> TQJC3H6OH',
            #[(0.001, 'atm'), 1.050000e+114, -33.81, 24741.0],
            #[(0.01, 'atm'), 2.130000e+114, -33.44, 26448.0],
            #[(0.1, 'atm'), 1.620000e+110, -31.75, 26612.0],
            #[(1.0, 'atm'), 6.000000e+101, -28.79, 25197.0],
            #[(10.0, 'atm'), 5.360000e+89, -24.76, 22402.0],
            #[(40.0, 'atm'), 1.550000e+81, -21.95, 20197.0],
            #[(100.0, 'atm'), 1.510000e+75, -20.0, 18578.0],
            #[(200.0, 'atm'), 3.160000e+70, -18.48, 17287.0])

## Reaction 743
#pdep_arrhenius('C3H6OH1-2 + O2 <=> C3H5OH + HO2',
            #[(0.001, 'atm'), 1.210000e+26, -5.09, 5755.0],
            #[(0.01, 'atm'), 1.410000e+31, -6.55, 8781.0],
            #[(0.1, 'atm'), 5.080000e+34, -7.53, 11702.0],
            #[(1.0, 'atm'), 1.670000e+34, -7.27, 13418.0],
            #[(10.0, 'atm'), 2.820000e+28, -5.41, 13318.0],
            #[(40.0, 'atm'), 2.500000e+22, -3.52, 12314.0],
            #[(100.0, 'atm'), 2.820000e+17, -1.99, 11286.0],
            #[(200.0, 'atm'), 1.930000e+13, -0.71, 10340.0],
            #options='duplicate')

## Reaction 744
#pdep_arrhenius('C3H6OH1-2 + O2 <=> C3H5OH + HO2',
            #[(0.001, 'atm'), 2.450000e+21, -4.19, 6837.0],
            #[(0.01, 'atm'), 1.400000e+30, -6.75, 11554.0],
            #[(0.1, 'atm'), 9.110000e+39, -9.56, 17834.0],
            #[(1.0, 'atm'), 5.160000e+42, -10.17, 22412.0],
            #[(10.0, 'atm'), 6.070000e+32, -6.94, 22738.0],
            #[(40.0, 'atm'), 3.880000e+20, -3.14, 20677.0],
            #[(100.0, 'atm'), 3.320000e+10, -0.03, 18552.0],
            #[(200.0, 'atm'), 1.220000e+02, 2.57, 16623.0],
            #options='duplicate')

## Reaction 745
#pdep_arrhenius('C3H6OH1-2 + O2 <=> SC3H5OH + HO2',
            #[(0.001, 'atm'), 1.260000e+25, -4.69, 5755.0],
            #[(0.01, 'atm'), 1.440000e+30, -6.15, 8785.0],
            #[(0.1, 'atm'), 4.780000e+33, -7.11, 11695.0],
            #[(1.0, 'atm'), 1.400000e+33, -6.84, 13395.0],
            #[(10.0, 'atm'), 2.120000e+27, -4.96, 13277.0],
            #[(40.0, 'atm'), 1.780000e+21, -3.07, 12265.0],
            #[(100.0, 'atm'), 1.970000e+16, -1.53, 11234.0],
            #[(200.0, 'atm'), 1.320000e+12, -0.25, 10285.0],
            #options='duplicate')

## Reaction 746
#pdep_arrhenius('C3H6OH1-2 + O2 <=> SC3H5OH + HO2',
            #[(0.001, 'atm'), 3.220000e+23, -4.69, 5341.0],
            #[(0.01, 'atm'), 4.800000e+28, -6.18, 8461.0],
            #[(0.1, 'atm'), 1.710000e+32, -7.16, 11410.0],
            #[(1.0, 'atm'), 1.530000e+32, -7.02, 13378.0],
            #[(10.0, 'atm'), 5.550000e+29, -6.14, 15100.0],
            #[(40.0, 'atm'), 1.720000e+26, -4.97, 15849.0],
            #[(100.0, 'atm'), 4.460000e+21, -3.51, 15644.0],
            #[(200.0, 'atm'), 5.060000e+16, -1.96, 14979.0],
            #options='duplicate')

## Reaction 747
#pdep_arrhenius('C3H6OH1-2 + O2 <=> TQC3H6OI',
            #[(0.001, 'atm'), 6.360000e+104, -33.74, 22390.0],
            #[(0.01, 'atm'), 7.960000e+103, -33.01, 22966.0],
            #[(0.1, 'atm'), 6.990000e+96, -30.48, 20584.0],
            #[(1.0, 'atm'), 1.040000e+88, -27.47, 16629.0],
            #[(10.0, 'atm'), 3.130000e+96, -29.62, 20346.0],
            #[(40.0, 'atm'), 2.790000e+105, -32.04, 24971.0],
            #[(100.0, 'atm'), 4.640000e+109, -33.12, 27657.0],
            #[(200.0, 'atm'), 2.060000e+111, -33.48, 29197.0])

## Reaction 748
#pdep_arrhenius('C3H6OH1-2 + O2 <=> CH3CHO + CH2O + OH',
            #[(0.001, 'atm'), 3.030000e+37, -8.35, 6940.0],
            #[(0.01, 'atm'), 1.070000e+42, -9.64, 9965.0],
            #[(0.1, 'atm'), 9.220000e+43, -10.12, 12427.0],
            #[(1.0, 'atm'), 1.140000e+42, -9.42, 13806.0],
            #[(10.0, 'atm'), 1.800000e+38, -8.13, 15131.0],
            #[(40.0, 'atm'), 1.610000e+34, -6.8, 15691.0],
            #[(100.0, 'atm'), 7.000000e+29, -5.41, 15552.0],
            #[(200.0, 'atm'), 3.740000e+25, -4.06, 15118.0])

## Reaction 749
#pdep_arrhenius('C3H6OH1-2 + O2 <=> QC3H5OHP',
            #[(0.001, 'atm'), 1.180000e+118, -37.6, 26229.0],
            #[(0.01, 'atm'), 1.910000e+131, -40.73, 34079.0],
            #[(0.1, 'atm'), 5.810000e+138, -42.17, 40750.0],
            #[(1.0, 'atm'), 5.050000e+134, -40.19, 43580.0],
            #[(10.0, 'atm'), 2.070000e+117, -34.26, 41516.0],
            #[(40.0, 'atm'), 1.710000e+101, -29.04, 38094.0],
            #[(100.0, 'atm'), 1.250000e+89, -25.15, 35203.0],
            #[(200.0, 'atm'), 2.980000e+79, -22.09, 32802.0])

## Reaction 750
#pdep_arrhenius('C3H6OH1-2 + O2 <=> CY(COC)COH + OH',
            #[(0.001, 'atm'), 2.110000e+18, -2.87, 6870.0],
            #[(0.01, 'atm'), 3.020000e+27, -5.54, 11842.0],
            #[(0.1, 'atm'), 1.360000e+37, -8.29, 18165.0],
            #[(1.0, 'atm'), 1.210000e+39, -8.66, 22517.0],
            #[(10.0, 'atm'), 1.530000e+28, -5.13, 22530.0],
            #[(40.0, 'atm'), 2.720000e+15, -1.16, 20283.0],
            #[(100.0, 'atm'), 1.060000e+05, 2.06, 18044.0],
            #[(200.0, 'atm'), 2.300000e-04, 4.73, 16037.0])

## Reaction 751
#pdep_arrhenius('C3H6OH1-2 + O2 <=> TQC3H5OHI',
            #[(0.001, 'atm'), 1.760000e+53, -19.87, 9019.0],
            #[(0.01, 'atm'), 4.680000e+67, -23.92, 11892.0],
            #[(0.1, 'atm'), 2.070000e+91, -30.58, 17347.0],
            #[(1.0, 'atm'), 2.840000e+100, -32.4, 20041.0],
            #[(10.0, 'atm'), 1.160000e+115, -35.81, 27656.0],
            #[(40.0, 'atm'), 4.870000e+123, -37.83, 33314.0],
            #[(100.0, 'atm'), 7.950000e+124, -37.82, 35683.0],
            #[(200.0, 'atm'), 4.580000e+122, -36.86, 36374.0])

## Reaction 752
#pdep_arrhenius('C3H6OH1-2 + O2 <=> CCY(COC)OH + OH',
            #[(0.001, 'atm'), 1.420000e+32, -6.95, 6210.0],
            #[(0.01, 'atm'), 5.190000e+36, -8.24, 9233.0],
            #[(0.1, 'atm'), 5.770000e+38, -8.76, 11715.0],
            #[(1.0, 'atm'), 2.860000e+36, -7.95, 12823.0],
            #[(10.0, 'atm'), 1.370000e+32, -6.51, 13646.0],
            #[(40.0, 'atm'), 2.150000e+29, -5.56, 14541.0],
            #[(100.0, 'atm'), 1.200000e+26, -4.51, 14778.0],
            #[(200.0, 'atm'), 3.310000e+22, -3.37, 14606.0])

## Reaction 753
#pdep_arrhenius('TQJC3H6OH <=> C3H5OH + HO2',
            #[(0.001, 'atm'), 9.730000e+65, -18.5, 42975.0],
            #[(0.01, 'atm'), 1.040000e+64, -17.25, 44419.0],
            #[(0.1, 'atm'), 7.900000e+59, -15.59, 44504.0],
            #[(1.0, 'atm'), 9.350000e+53, -13.49, 43566.0],
            #[(10.0, 'atm'), 4.130000e+44, -10.39, 41279.0],
            #[(40.0, 'atm'), 6.570000e+38, -8.49, 39745.0],
            #[(100.0, 'atm'), 8.320000e+34, -7.23, 38675.0],
            #[(200.0, 'atm'), 9.460000e+31, -6.28, 37849.0],
            #options='duplicate')

## Reaction 754
#pdep_arrhenius('TQJC3H6OH <=> C3H5OH + HO2',
            #[(0.001, 'atm'), 5.270000e+64, -18.0, 42872.0],
            #[(0.01, 'atm'), 4.820000e+62, -16.74, 44284.0],
            #[(0.1, 'atm'), 3.420000e+58, -15.07, 44348.0],
            #[(1.0, 'atm'), 4.050000e+52, -12.97, 43402.0],
            #[(10.0, 'atm'), 1.930000e+43, -9.88, 41120.0],
            #[(40.0, 'atm'), 3.280000e+37, -7.99, 39593.0],
            #[(100.0, 'atm'), 4.350000e+33, -6.74, 38527.0],
            #[(200.0, 'atm'), 5.130000e+30, -5.79, 37706.0],
            #options='duplicate')

## Reaction 755
#pdep_arrhenius('TQJC3H6OH <=> TQC3H6OI',
            #[(0.001, 'atm'), 3.780000e+50, -12.91, 31539.0],
            #[(0.01, 'atm'), 1.420000e+45, -10.94, 30864.0],
            #[(0.1, 'atm'), 1.500000e+40, -9.21, 29932.0],
            #[(1.0, 'atm'), 3.120000e+35, -7.64, 28864.0],
            #[(10.0, 'atm'), 2.730000e+29, -5.67, 27243.0],
            #[(40.0, 'atm'), 8.880000e+25, -4.54, 26272.0],
            #[(100.0, 'atm'), 4.600000e+23, -3.8, 25622.0],
            #[(200.0, 'atm'), 9.080000e+21, -3.25, 25131.0])

## Reaction 756
#pdep_arrhenius('TQJC3H6OH <=> QC3H5OHP',
            #[(0.001, 'atm'), 9.160000e+62, -18.02, 45297.0],
            #[(0.01, 'atm'), 2.560000e+62, -17.1, 47393.0],
            #[(0.1, 'atm'), 1.250000e+59, -15.61, 47984.0],
            #[(1.0, 'atm'), 1.970000e+53, -13.49, 47281.0],
            #[(10.0, 'atm'), 1.910000e+43, -10.15, 44926.0],
            #[(40.0, 'atm'), 7.680000e+36, -8.05, 43267.0],
            #[(100.0, 'atm'), 3.490000e+32, -6.64, 42089.0],
            #[(200.0, 'atm'), 1.770000e+29, -5.57, 41173.0])

## Reaction 757
#pdep_arrhenius('TQJC3H6OH <=> TQC3H5OHI',
            #[(0.001, 'atm'), 9.570000e+58, -15.99, 38293.0],
            #[(0.01, 'atm'), 7.270000e+54, -14.25, 38593.0],
            #[(0.1, 'atm'), 8.300000e+49, -12.44, 38031.0],
            #[(1.0, 'atm'), 1.900000e+44, -10.51, 36905.0],
            #[(10.0, 'atm'), 2.180000e+36, -7.9, 34865.0],
            #[(40.0, 'atm'), 4.100000e+31, -6.36, 33581.0],
            #[(100.0, 'atm'), 3.000000e+28, -5.35, 32704.0],
            #[(200.0, 'atm'), 1.340000e+26, -4.59, 32035.0])

## Reaction 758
#pdep_arrhenius('TQC3H6OI <=> CH3CHO + CH2O + OH',
            #[(0.001, 'atm'), 8.090000e+38, -9.91, 19096.0],
            #[(0.01, 'atm'), 9.090000e+39, -9.93, 19135.0],
            #[(0.1, 'atm'), 1.890000e+41, -10.02, 19407.0],
            #[(1.0, 'atm'), 7.470000e+23, -4.1, 14658.0],
            #[(10.0, 'atm'), 2.950000e+33, -6.75, 18685.0],
            #[(40.0, 'atm'), 4.240000e+36, -7.56, 20307.0],
            #[(100.0, 'atm'), 7.830000e+36, -7.54, 20747.0],
            #[(200.0, 'atm'), 7.630000e+35, -7.17, 20641.0])

## Reaction 759
#pdep_arrhenius('QC3H5OHP <=> C3H5OH + HO2',
            #[(0.001, 'atm'), 3.290000e+57, -15.64, 28576.0],
            #[(0.01, 'atm'), 9.650000e+58, -15.75, 29927.0],
            #[(0.1, 'atm'), 2.190000e+50, -12.66, 28547.0],
            #[(1.0, 'atm'), 1.670000e+49, -12.05, 29204.0],
            #[(10.0, 'atm'), 7.080000e+40, -9.26, 27188.0],
            #[(40.0, 'atm'), 1.320000e+30, -5.82, 24071.0],
            #[(100.0, 'atm'), 1.150000e+30, -5.8, 24053.0],
            #[(200.0, 'atm'), 1.120000e+30, -5.8, 24050.0])

## Reaction 760
#pdep_arrhenius('QC3H5OHP <=> CY(COC)COH + OH',
            #[(0.001, 'atm'), 1.110000e+51, -12.97, 28497.0],
            #[(0.01, 'atm'), 5.580000e+51, -12.87, 29529.0],
            #[(0.1, 'atm'), 3.030000e+44, -10.28, 28326.0],
            #[(1.0, 'atm'), 1.420000e+44, -9.98, 28986.0],
            #[(10.0, 'atm'), 9.690000e+37, -7.93, 27491.0],
            #[(40.0, 'atm'), 5.900000e+29, -5.3, 25095.0],
            #[(100.0, 'atm'), 5.270000e+29, -5.29, 25081.0],
            #[(200.0, 'atm'), 5.150000e+29, -5.28, 25078.0])

## Reaction 761
#pdep_arrhenius('TQC3H5OHI <=> CCY(COC)OH + OH',
            #[(0.001, 'atm'), 1.100000e+31, -7.21, 14640.0],
            #[(0.01, 'atm'), 1.370000e+32, -7.24, 14716.0],
            #[(0.1, 'atm'), 3.340000e+33, -7.35, 15127.0],
            #[(1.0, 'atm'), 1.860000e+23, -3.68, 12864.0],
            #[(10.0, 'atm'), 9.880000e+31, -6.23, 16040.0],
            #[(40.0, 'atm'), 1.250000e+34, -6.78, 17056.0],
            #[(100.0, 'atm'), 7.730000e+33, -6.67, 17176.0],
            #[(200.0, 'atm'), 2.130000e+33, -6.45, 17170.0])

## Reaction 762
#pdep_arrhenius('TQC3H5OHI <=> SC3H5OH + HO2',
            #[(0.001, 'atm'), 6.010000e+27, -7.27, 14658.0],
            #[(0.01, 'atm'), 7.780000e+28, -7.3, 14733.0],
            #[(0.1, 'atm'), 3.270000e+30, -7.48, 15191.0],
            #[(1.0, 'atm'), 2.810000e+17, -2.82, 12336.0],
            #[(10.0, 'atm'), 1.920000e+26, -5.22, 16146.0],
            #[(40.0, 'atm'), 7.610000e+27, -5.49, 17483.0],
            #[(100.0, 'atm'), 1.910000e+26, -4.86, 17429.0],
            #[(200.0, 'atm'), 2.940000e+24, -4.18, 17285.0])

## Reaction 763
#pdep_arrhenius('C3H6OH2-1 + O2 <=> IQJC3H6OH',
            #[(0.001, 'atm'), 3.340000e+111, -32.67, 25143.0],
            #[(0.01, 'atm'), 4.840000e+107, -31.05, 25460.0],
            #[(0.1, 'atm'), 2.410000e+100, -28.42, 24474.0],
            #[(1.0, 'atm'), 4.550000e+89, -24.78, 22176.0],
            #[(10.0, 'atm'), 1.140000e+76, -20.31, 18721.0],
            #[(40.0, 'atm'), 9.250000e+66, -17.35, 16238.0],
            #[(100.0, 'atm'), 6.200000e+60, -15.36, 14499.0],
            #[(200.0, 'atm'), 1.300000e+56, -13.86, 13159.0])

## Reaction 764
#pdep_arrhenius('C3H6OH2-1 + O2 <=> CHOCOHCH3 + OH',
            #[(0.001, 'atm'), 1.590000e+15, -1.63, 8947.0],
            #[(0.01, 'atm'), 5.070000e+20, -3.24, 11938.0],
            #[(0.1, 'atm'), 5.000000e+25, -4.66, 15251.0],
            #[(1.0, 'atm'), 3.180000e+26, -4.79, 17388.0],
            #[(10.0, 'atm'), 6.000000e+20, -2.95, 17297.0],
            #[(40.0, 'atm'), 1.700000e+14, -0.91, 16099.0],
            #[(100.0, 'atm'), 7.910000e+08, 0.73, 14913.0],
            #[(200.0, 'atm'), 3.030000e+04, 2.08, 13854.0])

## Reaction 765
#pdep_arrhenius('C3H6OH2-1 + O2 <=> IQC3H6OT',
            #[(0.001, 'atm'), 1.120000e+102, -32.4, 23496.0],
            #[(0.01, 'atm'), 4.230000e+95, -30.04, 22067.0],
            #[(0.1, 'atm'), 3.450000e+83, -26.08, 17114.0],
            #[(1.0, 'atm'), 8.160000e+82, -25.61, 16198.0],
            #[(10.0, 'atm'), 1.040000e+97, -29.54, 22648.0],
            #[(40.0, 'atm'), 7.250000e+103, -31.42, 26487.0],
            #[(100.0, 'atm'), 2.480000e+106, -32.06, 28397.0],
            #[(200.0, 'atm'), 1.390000e+107, -32.2, 29446.0])

## Reaction 766
#pdep_arrhenius('C3H6OH2-1 + O2 <=> CH3CHO + CH2O + OH',
            #[(0.001, 'atm'), 4.390000e+48, -11.88, 11603.0],
            #[(0.01, 'atm'), 9.820000e+50, -12.49, 14143.0],
            #[(0.1, 'atm'), 3.430000e+49, -11.94, 15561.0],
            #[(1.0, 'atm'), 2.430000e+45, -10.56, 16415.0],
            #[(10.0, 'atm'), 4.980000e+39, -8.68, 17473.0],
            #[(40.0, 'atm'), 7.990000e+33, -6.83, 17502.0],
            #[(100.0, 'atm'), 1.020000e+29, -5.28, 17165.0],
            #[(200.0, 'atm'), 7.180000e+24, -3.99, 16747.0])

## Reaction 767
#pdep_arrhenius('C3H6OH2-1 + O2 <=> CH3 + CHOCH2OOH',
            #[(0.001, 'atm'), 2.910000e+40, -9.23, 10830.0],
            #[(0.01, 'atm'), 9.660000e+43, -10.2, 13698.0],
            #[(0.1, 'atm'), 1.340000e+44, -10.13, 15661.0],
            #[(1.0, 'atm'), 2.570000e+41, -9.18, 17047.0],
            #[(10.0, 'atm'), 1.820000e+36, -7.46, 18330.0],
            #[(40.0, 'atm'), 8.820000e+29, -5.44, 18205.0],
            #[(100.0, 'atm'), 1.740000e+24, -3.65, 17600.0],
            #[(200.0, 'atm'), 2.160000e+19, -2.12, 16925.0])

## Reaction 768
#pdep_arrhenius('C3H6OH2-1 + O2 <=> IQC3H5OHPJ',
            #[(0.001, 'atm'), 1.030000e+115, -35.13, 25407.0],
            #[(0.01, 'atm'), 1.100000e+129, -38.89, 32891.0],
            #[(0.1, 'atm'), 5.150000e+135, -40.38, 38573.0],
            #[(1.0, 'atm'), 7.270000e+129, -38.13, 39933.0],
            #[(10.0, 'atm'), 4.500000e+112, -32.49, 37045.0],
            #[(40.0, 'atm'), 7.720000e+97, -27.79, 33612.0],
            #[(100.0, 'atm'), 5.170000e+86, -24.26, 30799.0],
            #[(200.0, 'atm'), 4.700000e+77, -21.41, 28440.0])

## Reaction 769
#pdep_arrhenius('C3H6OH2-1 + O2 <=> C2H3OH + CH2O + OH',
            #[(0.001, 'atm'), 5.100000e+24, -4.31, 13009.0],
            #[(0.01, 'atm'), 6.910000e+33, -6.97, 17935.0],
            #[(0.1, 'atm'), 1.740000e+41, -9.03, 23613.0],
            #[(1.0, 'atm'), 2.260000e+36, -7.32, 25633.0],
            #[(10.0, 'atm'), 3.150000e+17, -1.4, 22782.0],
            #[(40.0, 'atm'), 9.350000e+00, 3.71, 19114.0],
            #[(100.0, 'atm'), 5.590000e-12, 7.46, 16156.0],
            #[(200.0, 'atm'), 1.270000e-21, 10.41, 13733.0])

## Reaction 770
#pdep_arrhenius('C3H6OH2-1 + O2 <=> CY(CCOC)OH + OH',
            #[(0.001, 'atm'), 3.060000e+33, -7.24, 11476.0],
            #[(0.01, 'atm'), 6.560000e+42, -9.92, 17197.0],
            #[(0.1, 'atm'), 8.240000e+46, -10.95, 22090.0],
            #[(1.0, 'atm'), 4.480000e+38, -8.22, 23019.0],
            #[(10.0, 'atm'), 1.010000e+18, -1.77, 19496.0],
            #[(40.0, 'atm'), 1.160000e+01, 3.44, 15637.0],
            #[(100.0, 'atm'), 4.850000e-12, 7.23, 12599.0],
            #[(200.0, 'atm'), 8.640000e-22, 10.21, 10126.0])

## Reaction 771
#pdep_arrhenius('IQJC3H6OH <=> CHOCOHCH3 + OH',
            #[(0.001, 'atm'), 3.580000e+71, -20.62, 52656.0],
            #[(0.01, 'atm'), 9.670000e+71, -20.08, 54935.0],
            #[(0.1, 'atm'), 3.390000e+67, -18.18, 55330.0],
            #[(1.0, 'atm'), 7.260000e+58, -15.09, 54016.0],
            #[(10.0, 'atm'), 4.570000e+46, -11.01, 51172.0],
            #[(40.0, 'atm'), 4.310000e+38, -8.38, 49054.0],
            #[(100.0, 'atm'), 2.120000e+33, -6.66, 47587.0],
            #[(200.0, 'atm'), 2.720000e+29, -5.41, 46486.0])

## Reaction 772
#pdep_arrhenius('IQJC3H6OH <=> IQC3H6OT',
            #[(0.001, 'atm'), 1.040000e+45, -11.2, 31755.0],
            #[(0.01, 'atm'), 2.900000e+40, -9.62, 30945.0],
            #[(0.1, 'atm'), 4.850000e+35, -8.01, 29850.0],
            #[(1.0, 'atm'), 2.620000e+30, -6.28, 28498.0],
            #[(10.0, 'atm'), 3.370000e+24, -4.37, 26873.0],
            #[(40.0, 'atm'), 9.750000e+20, -3.23, 25861.0],
            #[(100.0, 'atm'), 5.610000e+18, -2.52, 25208.0],
            #[(200.0, 'atm'), 1.440000e+17, -2.01, 24740.0])

## Reaction 773
#pdep_arrhenius('IQJC3H6OH <=> IQC3H5OHPJ',
            #[(0.001, 'atm'), 1.360000e+50, -12.93, 36743.0],
            #[(0.01, 'atm'), 3.360000e+45, -11.27, 36143.0],
            #[(0.1, 'atm'), 1.130000e+40, -9.41, 34990.0],
            #[(1.0, 'atm'), 5.930000e+33, -7.33, 33438.0],
            #[(10.0, 'atm'), 4.240000e+26, -5.0, 31502.0],
            #[(40.0, 'atm'), 1.970000e+22, -3.61, 30275.0],
            #[(100.0, 'atm'), 3.430000e+19, -2.72, 29477.0],
            #[(200.0, 'atm'), 3.690000e+17, -2.09, 28900.0])

## Reaction 774
#pdep_arrhenius('IQC3H6OT <=> CH3CHO + CH2O + OH',
            #[(0.001, 'atm'), 9.820000e+39, -9.95, 20737.0],
            #[(0.01, 'atm'), 1.190000e+41, -9.97, 20814.0],
            #[(0.1, 'atm'), 2.600000e+42, -10.07, 21229.0],
            #[(1.0, 'atm'), 2.630000e+32, -6.47, 19139.0],
            #[(10.0, 'atm'), 1.320000e+40, -8.68, 22120.0],
            #[(40.0, 'atm'), 4.040000e+41, -9.03, 23000.0],
            #[(100.0, 'atm'), 9.070000e+40, -8.77, 23008.0],
            #[(200.0, 'atm'), 6.700000e+39, -8.38, 22782.0])

## Reaction 775
#pdep_arrhenius('IQC3H6OT <=> CH3 + CHOCH2OOH',
            #[(0.001, 'atm'), 2.040000e+37, -9.27, 20741.0],
            #[(0.01, 'atm'), 2.540000e+38, -9.29, 20821.0],
            #[(0.1, 'atm'), 7.130000e+39, -9.42, 21267.0],
            #[(1.0, 'atm'), 2.950000e+28, -5.34, 18891.0],
            #[(10.0, 'atm'), 1.310000e+36, -7.46, 22106.0],
            #[(40.0, 'atm'), 2.190000e+37, -7.68, 23084.0],
            #[(100.0, 'atm'), 1.390000e+36, -7.22, 23018.0],
            #[(200.0, 'atm'), 2.670000e+34, -6.63, 22672.0])

## Reaction 776
#pdep_arrhenius('IQC3H5OHPJ <=> C2H3OH + CH2O + OH',
            #[(0.001, 'atm'), 4.250000e+16, -3.66, 19364.0],
            #[(0.01, 'atm'), 2.980000e+33, -8.01, 27106.0],
            #[(0.1, 'atm'), 3.460000e+38, -8.83, 31412.0],
            #[(1.0, 'atm'), 7.070000e+28, -5.46, 29633.0],
            #[(10.0, 'atm'), 2.010000e+13, -0.43, 25278.0],
            #[(40.0, 'atm'), 1.430000e+13, -0.38, 25238.0],
            #[(100.0, 'atm'), 1.380000e+13, -0.38, 25234.0],
            #[(200.0, 'atm'), 1.370000e+13, -0.37, 25233.0])

## Reaction 777
#pdep_arrhenius('IQC3H5OHPJ <=> CY(CCOC)OH + OH',
            #[(0.001, 'atm'), 1.120000e+25, -5.59, 19349.0],
            #[(0.01, 'atm'), 7.200000e+32, -7.6, 23093.0],
            #[(0.1, 'atm'), 2.220000e+34, -7.65, 25130.0],
            #[(1.0, 'atm'), 2.400000e+27, -5.28, 23663.0],
            #[(10.0, 'atm'), 2.240000e+17, -2.04, 20810.0],
            #[(40.0, 'atm'), 1.820000e+17, -2.01, 20786.0],
            #[(100.0, 'atm'), 1.780000e+17, -2.01, 20783.0],
            #[(200.0, 'atm'), 1.770000e+17, -2.01, 20782.0])

## Reaction 778
#pdep_arrhenius('IQC3H5OHPJ <=> AC3H5OOH + OH',
            #[(0.001, 'atm'), 6.070000e+13, -4.21, 22478.0],
            #[(0.01, 'atm'), 9.740000e+39, -10.89, 33616.0],
            #[(0.1, 'atm'), 2.180000e+49, -12.65, 40359.0],
            #[(1.0, 'atm'), 3.320000e+37, -8.42, 38539.0],
            #[(10.0, 'atm'), 2.340000e+16, -1.56, 32712.0],
            #[(40.0, 'atm'), 1.440000e+16, -1.49, 32656.0],
            #[(100.0, 'atm'), 1.380000e+16, -1.48, 32651.0],
            #[(200.0, 'atm'), 1.370000e+16, -1.48, 32649.0])

## Reaction 779
#reaction('IQC4H7OHT <=> CH2COHCH2OOH + CH3', [4.950000e+10, 0.83, 27900.0])

## Reaction 780
#reaction('IQC4H8OT <=> CH3COCH3 + CH2O + OH', [3.640000e+09, 1.3, 23700.0])

## Reaction 781
#reaction('IQC4H8OT <=> CH3 + CH3COCOOH', [4.950000e+10, 0.83, 27900.0])

## Reaction 782
#reaction('CY(COC)COH + OH => C2H2OH + CH2O + H2O', [1.260000e+03, 2.97, -2660.6])

## Reaction 783
#reaction('CY(COC)COH + HO2 => C2H2OH + CH2O + H2O2', [1.400000e-05, 5.26, 8267.9])

## Reaction 784
#reaction('CY(COC)COH + OH => C2H3CHO + OH + H2O', [5.660000e+02, 2.93, -4039.4])

## Reaction 785
#reaction('CY(COC)COH + OH => CH2CO + CH2OH + H2O', [2.260000e+03, 2.73, -4688.0])

## Reaction 786
#reaction('CCY(COC)OH + OH => CH3CHCO + OH + H2O', [2.260000e+03, 2.73, -4688.0])

## Reaction 787
#reaction('CY(CCOC)OH + OH => CH2O + C2H2OH + H2O', [5.660000e+02, 2.93, -4039.4])

## Reaction 788
#reaction('CY(CCOC)OH + OH => CH2O + CH2CO + H + H2O', [2.260000e+03, 2.73, -4688.0])

## Reaction 789
#reaction('CY(COC)COH + HO2 => C2H3CHO + OH + H2O2', [1.810000e-01, 3.98, 9056.7])

## Reaction 790
#reaction('CY(COC)COH + HO2 => CH2CO + CH2OH + H2O2', [8.600000e+00, 3.46, 9732.326])

## Reaction 791
#reaction('CCY(COC)OH + HO2 => CH3CHCO + OH + H2O2', [8.600000e+00, 3.46, 9732.326])

## Reaction 792
#reaction('CY(CCOC)OH + HO2 => CH2O + C2H2OH + H2O2', [1.810000e-01, 3.98, 9056.7])

## Reaction 793
#reaction('CY(CCOC)OH + HO2 => CH2O + CH2CO + H + H2O2', [8.600000e+00, 3.46, 9732.326])

## Reaction 794
#reaction('CHOCOHCH3 + OH => CH3CHCO + OH + H2O', [6.132990e+04, 2.65, -4586.4])

## Reaction 795
#reaction('TQC3H5OHI + O2 <=> TQC3H5OHIO2', [5.050000e+12, -0.1, -695.45])

## Reaction 796
#reaction('TQC3H5OHIO2 <=> TQC3H5OHIQ-I', [2.560000e+12, -0.13, 34360.0])

## Reaction 797
#reaction('TQC3H5OHIQ-I <=> HO2CHO + CH3CHO + OH', [5.819000e+05, 2.4, 22790.0])

## Reaction 798
#reaction('TQC3H5OHIQ-I <=> C3KET12 + HO2', [1.829000e+10, 0.79, 15100.0])

## Reaction 799
#reaction('C3KET12 => CH3CHO + HCO + OH', [9.500000e+15, 0.0, 42540.0])

## Reaction 800
#reaction('TQC3H5OHIO2 <=> TQC3H5OHIQ-P', [5.690000e+08, 0.78, 21850.0])

## Reaction 801
#reaction('TQC3H5OHIQ-P <=> C2H3COHOOH + HO2', [1.830000e+10, 0.79, 15100.0])

## Reaction 802
#reaction('TQC3H5OHIQ-P <=> C2H3OOH + HOCHO + OH', [5.380000e+11, 0.07, 24800.0])

## Reaction 803
#reaction('TQC3H5OHIQ-P <=> COHOOHCY(COC) + OH', [2.280000e+08, 1.29, 9890.0])

## Reaction 804
#reaction('TQC3H5OHIQ-P <=> OHCY(COCC)OOH + OH', [4.580000e+15, -1.08, 18440.0])

## Reaction 805
#reaction('C2H3COHOOH => HOCHO + C2H3 + OH', [1.590000e+20, -1.5, 42879.46])

## Reaction 806
#reaction('C2H3OOH => CH2CO + H + OH', [1.590000e+20, -1.5, 42879.46])

## Reaction 807
#reaction('COHOOHCY(COC) => HOCHO + CH2CHO + OH', [1.590000e+20, -1.5, 42879.46])

## Reaction 808
#reaction('OHCY(COCC)OOH => HOCHO + CH2CHO + OH', [1.590000e+20, -1.5, 42879.46])

## Reaction 809
#reaction('TQC3H5OHIO2 <=> TQC3H5OHTO2', [2.956000e+09, 0.04, 16350.0])

## Reaction 810
#reaction('TQC3H5OHTO2 <=> OHCOCOOHCH3 + OH', [1.200000e+10, 0.35, 15700.0])

## Reaction 811
#reaction('OHCOCOOHCH3 => CO + CH3CHO + 2 OH', [1.590000e+20, -1.5, 42879.46])

## Reaction 812
#reaction('IQC3H5OHPJ + O2 <=> IQC3H5OHPJO2', [9.356260e+11, 0.1, -1072.98])

## Reaction 813
#reaction('IQC3H5OHPJO2 <=> IQC3H5OTQ-I', [4.130000e+07, 1.0, 21070.0])

## Reaction 814
#reaction('IQC3H5OTQ-I => OH + CH2O + CHOCH2OOH', [1.200000e+10, 0.35, 15700.0])

## Reaction 815
#reaction('CHOCH2OOH => OH + CH2O + HCO', [1.590000e+20, -1.5, 42879.46])

## Reaction 816
#reaction('IQC3H5OHPJO2 <=> IQC3H5OHQ-SJ', [8.204000e+10, 0.13, 19470.0])

## Reaction 817
#reaction('IQC3H5OHQ-SJ <=> CHOCOHCH2OOH + OH', [1.200000e+10, 0.35, 15700.0])

## Reaction 818
#reaction('CHOCOHCH2OOH => OH + CH2O + H + CHOCHO', [1.590000e+20, -1.5, 42879.46])

## Reaction 819
#reaction('CHOCHO + OH => HCO + CO + H2O', [6.132990e+04, 2.65, -4586.4])

## Reaction 820
#reaction('SC3H5OH + O2 => C2H3CHO + H + HO2', [3.000000e+13, 0.0, 39100.0])

## Reaction 821
#reaction('SC3H5OH + OH => C2H3CHO + H + H2O', [3.100000e+06, 2.0, -298.0])

## Reaction 822
#reaction('SC3H5OH + H => C2H3CHO + H + H2', [1.730000e+05, 2.5, 2492.0])

## Reaction 823
#reaction('SC3H5OH + O => C2H3CHO + H + OH', [1.750000e+12, 0.7, 5884.0])

## Reaction 824
#reaction('SC3H5OH + HO2 => C2H3CHO + H + H2O2', [9.600000e+03, 2.6, 13900.0])

## Reaction 825
#reaction('SC3H5OH + CH3 => C2H3CHO + H + CH4', [2.210000e+00, 3.5, 5675.0])

## Reaction 826
#reaction('SC3H5OH + CH3O2 => C2H3CHO + H + CH3O2H', [9.600000e+03, 2.6, 13900.0])

## Reaction 827
#reaction('SC3H5OH + CH3O => C2H3CHO + H + CH3OH', [8.300000e+10, 0.0, 2600.0])

## Reaction 828
#reaction('SC3H5OH + HO2 <=> C2H5CHO + HO2', [1.490000e+05, 1.67, 6810.0])

## Reaction 829
#reaction('SC3H5OH + HOCHO <=> C2H5CHO + HOCHO', [2.810000e-02, 3.286, -4509.0])

## Reaction 830
#reaction('SC3H5OH <=> C2H5CHO', [8.590000e+11, 0.318, 55900.0])

## Reaction 831
#pdep_arrhenius('C2H2 + CH3 <=> C3H5-A',
            #[(0.1, 'atm'), 8.200000e+53, -13.32, 33200.0],
            #[(1.0, 'atm'), 2.680000e+53, -12.82, 35730.0],
            #[(2.0, 'atm'), 3.640000e+52, -12.46, 36127.0],
            #[(5.0, 'atm'), 1.040000e+51, -11.89, 36476.0],
            #[(10.0, 'atm'), 4.400000e+49, -11.4, 36700.0],
            #[(100.0, 'atm'), 3.800000e+44, -9.63, 37600.0])

## Reaction 832
#pdep_arrhenius('C3H5-A + CH3O2 <=> C3H5O + CH3O',
            #[(0.01, 'atm'), 3.330000e+12, -0.158, -1417.0],
            #[(0.1, 'atm'), 1.660000e+14, -0.642, -349.1],
            #[(1.0, 'atm'), 2.595000e+17, -1.52, 2379.2],
            #[(10.0, 'atm'), 9.780000e+14, -0.684, 3615.3],
            #[(100.0, 'atm'), 5.470000e+03, 2.74, 1144.4])

## Reaction 833
#pdep_arrhenius('C3H5-A + CH3O2 <=> AC4H7OOH',
            #[(0.01, 'atm'), 1.910000e+31, -7.23, 1336.2],
            #[(0.1, 'atm'), 6.310000e+42, -10.3, 5568.9],
            #[(1.0, 'atm'), 1.030000e+45, -10.6, 7851.5],
            #[(10.0, 'atm'), 2.790000e+37, -7.92, 6497.9],
            #[(100.0, 'atm'), 3.400000e+29, -5.28, 4539.8])

## Reaction 834
#pdep_arrhenius('AC4H7OOH <=> C3H5O + CH3O',
            #[(0.01, 'atm'), 1.490000e+58, -13.9, 54266.9],
            #[(0.1, 'atm'), 1.800000e+54, -12.4, 54193.8],
            #[(1.0, 'atm'), 3.360000e+46, -9.81, 52468.5],
            #[(10.0, 'atm'), 2.390000e+36, -6.54, 49429.0],
            #[(100.0, 'atm'), 2.560000e+27, -3.61, 46333.1])

## Reaction 835
#reaction('C3H5-A + H <=> C3H4-A + H2', [1.232000e+03, 3.035, 2582.0])

## Reaction 836
#reaction('C3H5-A + O <=> C2H3CHO + H', [6.000000e+13, 0.0, 0.0])

## Reaction 837
#pdep_arrhenius('C3H5-A + OH => C2H3CHO + H + H',
            #[(0.1, 'atm'), 5.300000e+37, -6.71, 29306.0],
            #[(1.0, 'atm'), 4.200000e+32, -5.16, 30126.0],
            #[(10.0, 'atm'), 1.600000e+20, -1.56, 26330.0])

## Reaction 838
#reaction('C3H5-A + OH <=> C3H4-A + H2O', [6.000000e+12, 0.0, 0.0])

## Reaction 839
#reaction('C3H5-A + CH3 <=> C3H4-A + CH4', [3.000000e+12, -0.32, -131.0])

## Reaction 840
#reaction('C3H5-A + C2H5 <=> C2H6 + C3H4-A', [4.000000e+11, 0.0, 0.0])

## Reaction 841
#reaction('C3H5-A + C2H5 <=> C2H4 + C3H6', [4.000000e+11, 0.0, 0.0])

## Reaction 842
#reaction('C3H5-A + C2H3 <=> C2H4 + C3H4-A', [1.000000e+12, 0.0, 0.0])

## Reaction 843
#pdep_arrhenius('C3H5-A + O2 <=> C3H4-A + HO2',
            #[(1.0, 'atm'), 4.990000e+15, -1.4, 22428.0],
            #[(10.0, 'atm'), 2.180000e+21, -2.85, 30755.0])

## Reaction 844
#pdep_arrhenius('C3H5-A + O2 <=> CH3CO + CH2O',
            #[(1.0, 'atm'), 1.190000e+15, -1.01, 20128.0],
            #[(10.0, 'atm'), 7.140000e+15, -1.21, 21046.0])

## Reaction 845
#pdep_arrhenius('C3H5-A + O2 <=> C2H3CHO + OH',
            #[(1.0, 'atm'), 1.820000e+13, -0.41, 22859.0],
            #[(10.0, 'atm'), 2.470000e+13, -0.45, 23017.0])

## Reaction 846
#reaction('C3H5-A + HCO <=> C3H6 + CO', [6.000000e+13, 0.0, 0.0])

## Reaction 847
#pdep_arrhenius('C3H5-A <=> C3H5-T',
            #[(0.1, 'atm'), 3.900000e+59, -15.42, 75400.0],
            #[(1.0, 'atm'), 7.060000e+56, -14.08, 75868.0],
            #[(2.0, 'atm'), 4.800000e+55, -13.59, 75949.0],
            #[(5.0, 'atm'), 4.860000e+53, -12.81, 75883.0],
            #[(10.0, 'atm'), 6.400000e+51, -12.12, 75700.0],
            #[(100.0, 'atm'), 2.800000e+43, -9.27, 74000.0])

## Reaction 848
#pdep_arrhenius('C3H5-A <=> C3H5-S',
            #[(0.1, 'atm'), 1.300000e+55, -14.53, 73800.0],
            #[(1.0, 'atm'), 5.000000e+51, -13.02, 73300.0],
            #[(10.0, 'atm'), 9.700000e+48, -11.73, 73700.0],
            #[(100.0, 'atm'), 4.860000e+44, -9.84, 73400.0])

## Reaction 849
#reaction('C3H5-S + O2 <=> CH3CHO + HCO', [3.100000e+31, -5.944, 5748.4])

## Reaction 850
#reaction('C3H5-S + O2 <=> CH3CHCHO + O', [5.380000e+18, -2.14, 5142.9])

## Reaction 851
#reaction('C3H5-S + O2 <=> C2H3CHO + OH', [2.700000e+19, -2.14, 5142.9])

## Reaction 852
#reaction('C3H5-S + H <=> C3H4-A + H2', [3.333000e+12, 0.0, 0.0])

## Reaction 853
#reaction('C3H5-S + CH3 <=> C3H4-A + CH4', [1.000000e+11, 0.0, 0.0])

## Reaction 854
#pdep_arrhenius('C2H2 + CH3 <=> C3H5-T',
            #[(0.1, 'atm'), 6.800000e+20, -4.16, 18000.0],
            #[(1.0, 'atm'), 4.990000e+22, -4.39, 18850.0],
            #[(2.0, 'atm'), 6.000000e+23, -4.6, 19571.0],
            #[(5.0, 'atm'), 7.310000e+25, -5.06, 21150.0],
            #[(10.0, 'atm'), 9.300000e+27, -5.55, 22900.0],
            #[(100.0, 'atm'), 3.800000e+36, -7.58, 31300.0])

## Reaction 855
#reaction('C3H5-S + H <=> C3H4-P + H2', [3.340000e+12, 0.0, 0.0])

## Reaction 856
#reaction('C3H5-S + O <=> C2H4 + HCO', [6.000000e+13, 0.0, 0.0])

## Reaction 857
#reaction('C3H5-S + OH => C2H4 + HCO + H', [5.000000e+12, 0.0, 0.0])

## Reaction 858
#reaction('C3H5-S + HO2 => C2H4 + HCO + OH', [2.000000e+13, 0.0, 0.0])

## Reaction 859
#reaction('C3H5-S + HCO <=> C3H6 + CO', [9.000000e+13, 0.0, 0.0])

## Reaction 860
#reaction('C3H5-S + CH3 <=> C3H4-P + CH4', [1.000000e+11, 0.0, 0.0])

## Reaction 861
#pdep_arrhenius('C3H5-T <=> C3H5-S',
            #[(0.1, 'atm'), 1.600000e+44, -12.16, 52200.0],
            #[(1.0, 'atm'), 1.500000e+48, -12.71, 53900.0],
            #[(10.0, 'atm'), 5.100000e+52, -13.37, 57200.0],
            #[(100.0, 'atm'), 5.800000e+51, -12.43, 59200.0])

## Reaction 862
#pdep_arrhenius('C3H4-A + H <=> C3H5-S',
            #[(0.1, 'atm'), 1.100000e+30, -6.52, 15200.0],
            #[(1.0, 'atm'), 5.400000e+29, -6.09, 16300.0],
            #[(10.0, 'atm'), 2.600000e+31, -6.23, 18700.0],
            #[(100.0, 'atm'), 3.200000e+31, -5.88, 21500.0])

## Reaction 863
#reaction('C3H4-A + C3H4-A <=> C3H5-A + C3H3', [5.000000e+14, 0.0, 64746.7])

## Reaction 864
#reaction('C3H5-T + O2 <=> CH3COCH2 + O', [9.860000e+25, -3.751, 11255.4])

## Reaction 865
#reaction('C3H5-T + O2 <=> CH3CO + CH2O', [2.550000e+20, -2.608, 1565.7])

## Reaction 866
#reaction('C3H5-T + O2 <=> C3H4-A + HO2', [3.590000e+10, -0.27, -413.6])

## Reaction 867
#reaction('C3H5-T + H <=> C3H4-P + H2', [3.340000e+12, 0.0, 0.0])

## Reaction 868
#reaction('C3H5-T + CH3 <=> C3H4-P + CH4', [1.000000e+11, 0.0, 0.0])

## Reaction 869
#reaction('C3H5-T + O <=> CH3 + CH2CO', [6.000000e+13, 0.0, 0.0])

## Reaction 870
#reaction('C3H5-T + OH => CH3 + CH2CO + H', [5.000000e+12, 0.0, 0.0])

## Reaction 871
#reaction('C3H5-T + HO2 => CH3 + CH2CO + OH', [2.000000e+13, 0.0, 0.0])

## Reaction 872
#reaction('C3H5-T + HCO <=> C3H6 + CO', [9.000000e+13, 0.0, 0.0])

## Reaction 873
#reaction('C3H4-A + HO2 => CH2CO + CH2 + OH', [4.000000e+12, 0.0, 19000.0])

## Reaction 874
#reaction('C3H4-A + O <=> C2H4 + CO', [2.000000e+07, 1.8, 1000.0])

## Reaction 875
#reaction('C3H4-A + O <=> C2H2 + CH2O', [3.000000e-03, 4.61, -4243.0])

## Reaction 876
#reaction('C3H4-A + CH3 <=> C3H3 + CH4', [1.300000e+12, 0.0, 7700.0])

## Reaction 877
#reaction('C3H4-A + C3H5-A <=> C3H3 + C3H6', [2.000000e+11, 0.0, 7700.0])

## Reaction 878
#reaction('C3H4-A + C2H <=> C2H2 + C3H3', [1.000000e+13, 0.0, 0.0])

## Reaction 879
#pdep_arrhenius('C3H4-P + H <=> C3H5-A',
            #[(0.1, 'atm'), 1.100000e+60, -14.56, 28100.0],
            #[(1.0, 'atm'), 4.910000e+60, -14.37, 31644.0],
            #[(2.0, 'atm'), 3.040000e+60, -14.19, 32642.0],
            #[(5.0, 'atm'), 9.020000e+59, -13.89, 33953.0],
            #[(10.0, 'atm'), 2.200000e+59, -13.61, 34900.0],
            #[(100.0, 'atm'), 1.600000e+55, -12.07, 37500.0])

## Reaction 880
#reaction('C3H4-P + C3H3 <=> C3H4-A + C3H3', [6.140000e+06, 1.74, 10450.0])

## Reaction 881
#reaction('C3H4-P + O <=> HCCO + CH3', [7.300000e+12, 0.0, 2250.0])

## Reaction 882
#reaction('C3H4-P + O <=> C2H4 + CO', [1.000000e+13, 0.0, 2250.0])

## Reaction 883
#reaction('C3H4-P + C2H <=> C2H2 + C3H3', [1.000000e+13, 0.0, 0.0])

## Reaction 884
#reaction('C3H4-P + CH3 <=> C3H3 + CH4', [1.800000e+12, 0.0, 7700.0])

## Reaction 885
#reaction('C2H + CH3 <=> C3H4-P', [8.000000e+13, 0.0, 0.0])

## Reaction 886
#reaction('C3H4-P + HO2 => C2H4 + CO + OH', [3.000000e+12, 0.0, 19000.0])

## Reaction 887
#reaction('C3H4-P + O <=> C2H3 + HCO', [3.200000e+12, 0.0, 2010.0])

## Reaction 888
#reaction('C3H4-P + O <=> C3H3 + OH', [7.650000e+08, 1.5, 8600.0])

## Reaction 889
#reaction('C3H4-P + C2H3 <=> C3H3 + C2H4', [1.000000e+12, 0.0, 7700.0])

## Reaction 890
#reaction('C3H4-P + C3H5-A <=> C3H3 + C3H6', [3.000000e+12, 0.0, 7700.0])

## Reaction 891
#reaction('C3H3 + O <=> CH2O + C2H', [2.000000e+13, 0.0, 0.0])

## Reaction 892
#reaction('C3H3 + HO2 => OH + CO + C2H3', [8.000000e+11, 0.0, 0.0])

## Reaction 893
#reaction('C3H3 + HCO <=> C3H4-A + CO', [2.500000e+13, 0.0, 0.0])

## Reaction 894
#reaction('C3H3 + HCO <=> C3H4-P + CO', [2.500000e+13, 0.0, 0.0])

## Reaction 895
#reaction('C2H5 + C2H <=> C3H3 + CH3', [1.810000e+13, 0.0, 0.0])

## Reaction 896
#reaction('C3H4-A + HO2 => C2H4 + CO + OH', [1.000000e+11, 0.0, 14000.0])

## Reaction 897
#reaction('C3H4-A + HO2 <=> C3H3 + H2O2', [3.580000e-02, 4.17, 9632.8])

## Reaction 898
#reaction('C3H4-A + CH3O2 <=> C3H3 + CH3O2H', [7.161000e-02, 4.17, 9632.8])

## Reaction 899
#reaction('C3H4-P + HO2 <=> C3H3 + H2O2', [9.550000e-02, 4.17, 9632.8])

## Reaction 900
#reaction('C3H4-P + CH3O2 <=> C3H3 + CH3O2H', [9.550000e-02, 4.17, 9632.8])

## Reaction 901
#reaction('C3H4-A + OH <=> CH2CCH2OH', [1.110000e+12, 0.0, -304.0])

## Reaction 902
#reaction('C3H4-A + OH <=> SC3H4OH', [2.220000e+12, 0.0, -304.0])

## Reaction 903
#reaction('CH2CCH2OH <=> CH2O + C2H3', [9.240000e+10, 0.87, 30460.0])

## Reaction 904
#reaction('CH2CCH2OH <=> C2H3CHO + H', [1.110000e+11, 0.48, 36770.0])

## Reaction 905
#reaction('CH2CCH2OH <=> HCO + C2H4', [9.240000e+10, 0.87, 30460.0])

## Reaction 906
#reaction('SC3H4OH <=> CH2CO + CH3', [9.240000e+10, 0.87, 30460.0])

## Reaction 907
#reaction('C3H4-A + H <=> C3H3 + H2', [6.625000e+03, 3.095, 5522.0])

## Reaction 908
#reaction('C3H4-P + OH <=> C3H3 + H2O', [4.940000e+06, 2.027, 1059.6])

## Reaction 909
#reaction('C3H4-A + OH <=> C3H3 + H2O', [1.482000e+05, 2.492, 1807.2])

## Reaction 910
#reaction('C3H4-P + H <=> C3H3 + H2', [3.572000e+04, 2.825, 4821.0])

## Reaction 911
#reaction('C3H4-P + OH <=> PC3H4OH-2', [3.930000e+11, 0.0, 0.0])

## Reaction 912
#reaction('C3H4-P + OH <=> SC3H4OH', [2.360000e+12, 0.0, 0.0])

## Reaction 913
#reaction('PC3H4OH-2 <=> CH3CHCHO', [3.560000e+10, 0.88, 23238.0])

## Reaction 914
#reaction('CH3CHCHO <=> C2H3CHO + H', [4.164000e+12, -0.02, 32410.0])

## Reaction 915
#reaction('CH3CHCHO <=> CH3CHCO + H', [8.328000e+12, -0.02, 32410.0])

## Reaction 916
#reaction('C3H3 + H <=> C3H2 + H2', [2.140000e+05, 2.52, 7453.0])

## Reaction 917
#pdep_arrhenius('C3H3 + H <=> C3H2(S) + H2',
            #[(0.0395, 'atm'), 2.951000e+09, 1.28, 13474.0],
            #[(1.0, 'atm'), 1.097000e+10, 1.13, 13929.0],
            #[(10.0, 'atm'), 3.311000e+13, 0.195, 17579.0])

## Reaction 918
#pdep_arrhenius('C3H3 + H <=> H2CCC(S) + H2',
            #[(0.0395, 'atm'), 2.692000e+09, 1.05, 5371.0],
            #[(1.0, 'atm'), 2.884000e+13, -0.03, 9448.0],
            #[(10.0, 'atm'), 1.000000e+18, -1.23, 15111.0])

## Reaction 919
#pdep_arrhenius('C3H3 + H <=> C3H2C + H2',
            #[(0.0395, 'atm'), 1.072000e+07, 1.37, 15557.0],
            #[(1.0, 'atm'), 1.349000e+07, 1.34, 15560.0],
            #[(10.0, 'atm'), 7.244000e+09, 0.606, 18356.0])

## Reaction 920
#pdep_arrhenius('C3H4-A <=> C3H4-P',
            #[(1.0, 'atm'), 7.762000e+39, -7.8, 78446.0],
            #[(10.0, 'atm'), 4.786000e+48, -10.0, 88685.0])

## Reaction 921
#pdep_arrhenius('CC3H4 <=> C3H4-P',
            #[(0.0395, 'atm'), 2.512000e+50, -11.82, 50914.0],
            #[(1.0, 'atm'), 1.230000e+37, -7.51, 45551.0],
            #[(10.0, 'atm'), 1.660000e+37, -7.24, 48013.0])

## Reaction 922
#pdep_arrhenius('CC3H4 <=> C3H4-A',
            #[(0.0395, 'atm'), 9.772000e+43, -9.97, 56007.0],
            #[(1.0, 'atm'), 2.512000e+26, -4.56, 43922.0],
            #[(10.0, 'atm'), 5.012000e+35, -6.87, 51298.0])

## Reaction 923
#pdep_arrhenius('C3H4-P <=> C3H3 + H',
            #[(1.0, 'atm'), 6.480000e+30, -4.655, 93925.2],
            #[(10.0, 'atm'), 1.210000e+25, -2.787, 92376.1])

## Reaction 924
#pdep_arrhenius('C3H4-A <=> C3H3 + H',
            #[(1.0, 'atm'), 1.320000e+31, -4.749, 92079.5],
            #[(10.0, 'atm'), 3.650000e+25, -2.95, 90624.9])

## Reaction 925
#pdep_arrhenius('C3H3 + H <=> CC3H4',
            #[(0.0395, 'atm'), 8.913000e+112, -28.26, 83611.0],
            #[(1.0, 'atm'), 1.072000e+21, -2.95, 2687.0],
            #[(10.0, 'atm'), 3.236000e+18, -2.05, 2053.0])

## Reaction 926
#three_body_reaction('C3H2(S) + M <=> C3H2 + M', [1.000000e+13, 0.0, 0.0])

## Reaction 927
#reaction('H2CCC(S) + O2 <=> CO2 + C2H2', [1.000000e+13, 0.0, 0.0])

## Reaction 928
#reaction('C3H2(S) + H <=> H2CCC(S) + H', [1.000000e+13, 0.0, 0.0])

## Reaction 929
#reaction('C3H2C + O2 <=> C2H2 + CO2', [1.000000e+13, 0.0, 0.0])

## Reaction 930
#pdep_arrhenius('C3H4-A + H <=> C3H5-A',
            #[(0.001, 'atm'), 2.210000e+61, -15.25, 20076.0],
            #[(0.039, 'atm'), 1.240000e+52, -12.02, 17839.0],
            #[(1.0, 'atm'), 4.670000e+51, -11.45, 21340.0],
            #[(10.0, 'atm'), 3.750000e+48, -10.27, 22511.0],
            #[(100.0, 'atm'), 4.230000e+43, -8.61, 22522.0],
            #options='duplicate')

## Reaction 931
#pdep_arrhenius('C3H4-A + H <=> C3H5-A',
            #[(0.001, 'atm'), 2.800000e+38, -8.67, 8035.0],
            #[(0.039, 'atm'), 9.330000e+36, -8.19, 7462.0],
            #[(1.0, 'atm'), 3.320000e+30, -5.78, 6913.0],
            #[(10.0, 'atm'), 2.290000e+26, -4.32, 6163.0],
            #[(100.0, 'atm'), 4.380000e+21, -2.71, 5187.0],
            #options='duplicate')

## Reaction 932
#pdep_arrhenius('C3H4-A + H <=> C3H4-P + H',
            #[(0.001, 'atm'), 8.490000e+10, 0.89, 2503.0],
            #[(0.039, 'atm'), 1.480000e+13, 0.26, 4103.0],
            #[(1.0, 'atm'), 2.480000e+15, -0.33, 6436.0],
            #[(10.0, 'atm'), 2.350000e+25, -3.23, 13165.0],
            #[(100.0, 'atm'), 1.020000e+24, -2.67, 15552.0],
            #options='duplicate')

## Reaction 933
#pdep_arrhenius('C3H4-A + H <=> C3H4-P + H',
            #[(10.0, 'atm'), 1.790000e+07, 1.98, 4521.0],
            #[(100.0, 'atm'), 4.630000e+04, 2.62, 4466.0],
            #options='duplicate')

## Reaction 934
#pdep_arrhenius('C3H4-A + H <=> CH3 + C2H2',
            #[(0.001, 'atm'), 1.230000e+08, 1.53, 4737.0],
            #[(0.039, 'atm'), 2.720000e+09, 1.2, 6834.0],
            #[(1.0, 'atm'), 1.260000e+20, -1.83, 15003.0],
            #[(10.0, 'atm'), 1.680000e+16, -0.6, 14754.0],
            #[(100.0, 'atm'), 1.370000e+17, -0.79, 17603.0],
            #options='duplicate')

## Reaction 935
#pdep_arrhenius('C3H4-A + H <=> CH3 + C2H2',
            #[(1.0, 'atm'), 1.230000e+04, 2.68, 6335.0],
            #[(10.0, 'atm'), 3.310000e+08, 1.14, 8886.0],
            #[(100.0, 'atm'), 1.280000e+06, 1.71, 9774.0],
            #options='duplicate')

## Reaction 936
#pdep_arrhenius('C3H4-A + H <=> C3H5-T',
            #[(0.001, 'atm'), 6.440000e+102, -27.51, 51768.0],
            #[(0.039, 'atm'), 1.550000e+53, -13.1, 14472.0],
            #[(1.0, 'atm'), 1.900000e+53, -12.59, 16726.0],
            #[(10.0, 'atm'), 7.950000e+51, -11.82, 18286.0],
            #[(100.0, 'atm'), 4.210000e+52, -11.64, 22262.0],
            #options='duplicate')

## Reaction 937
#pdep_arrhenius('C3H4-A + H <=> C3H5-T',
            #[(0.001, 'atm'), 1.100000e+54, -14.29, 10809.0],
            #[(0.039, 'atm'), 9.880000e+44, -11.21, 8212.0],
            #[(1.0, 'atm'), 2.810000e+40, -9.42, 7850.0],
            #[(10.0, 'atm'), 2.600000e+35, -7.57, 7147.0],
            #[(100.0, 'atm'), 9.880000e+29, -5.53, 6581.0],
            #options='duplicate')

## Reaction 938
#pdep_arrhenius('C3H4-P + H <=> C3H5-T',
            #[(0.001, 'atm'), 8.850000e+51, -13.04, 12325.0],
            #[(0.039, 'atm'), 3.170000e+52, -12.69, 14226.0],
            #[(1.0, 'atm'), 2.870000e+53, -12.51, 16853.0],
            #[(10.0, 'atm'), 9.510000e+51, -11.74, 18331.0],
            #[(100.0, 'atm'), 4.510000e+52, -11.58, 22207.0],
            #options='duplicate')

## Reaction 939
#pdep_arrhenius('C3H4-P + H <=> C3H5-T',
            #[(0.001, 'atm'), 1.970000e+46, -11.91, 7456.0],
            #[(0.039, 'atm'), 2.590000e+45, -11.23, 8046.0],
            #[(1.0, 'atm'), 6.930000e+39, -9.11, 7458.0],
            #[(10.0, 'atm'), 6.800000e+34, -7.29, 6722.0],
            #[(100.0, 'atm'), 5.650000e+29, -5.39, 6150.0],
            #options='duplicate')

## Reaction 940
#pdep_arrhenius('C3H4-P + H <=> C3H5-S',
            #[(0.039, 'atm'), 3.380000e+49, -12.75, 14072.0],
            #[(1.0, 'atm'), 1.370000e+51, -12.55, 15428.0],
            #[(10.0, 'atm'), 3.880000e+50, -11.9, 16915.0],
            #[(100.0, 'atm'), 2.170000e+49, -11.1, 18746.0],
            #options='duplicate')

## Reaction 941
#pdep_arrhenius('C3H4-P + H <=> C3H5-S',
            #[(0.001, 'atm'), 1.490000e+38, -10.11, 7458.0],
            #[(0.039, 'atm'), 2.980000e+43, -11.43, 8736.0],
            #[(1.0, 'atm'), 5.750000e+39, -9.51, 8772.0],
            #[(10.0, 'atm'), 4.330000e+40, -9.6, 9401.0],
            #[(100.0, 'atm'), 3.440000e+34, -7.36, 8558.0],
            #options='duplicate')

## Reaction 942
#pdep_arrhenius('C3H4-P + H <=> CH3 + C2H2',
            #[(0.001, 'atm'), 2.440000e+10, 1.04, 3980.0],
            #[(0.039, 'atm'), 3.890000e+10, 0.989, 4114.0],
            #[(1.0, 'atm'), 3.460000e+12, 0.442, 5463.0],
            #[(10.0, 'atm'), 1.720000e+14, -0.01, 7134.0],
            #[(100.0, 'atm'), 1.900000e+15, -0.29, 8306.0])

## Reaction 943
#pdep_arrhenius('CH3 + C2H2 <=> C3H5-S',
            #[(0.001, 'atm'), 1.780000e+42, -10.4, 13647.0],
            #[(0.039, 'atm'), 1.520000e+44, -10.73, 15256.0],
            #[(1.0, 'atm'), 1.190000e+44, -10.19, 18728.0],
            #[(10.0, 'atm'), 6.020000e+43, -9.74, 20561.0],
            #[(100.0, 'atm'), 1.420000e+42, -8.91, 22235.0],
            #options='duplicate')

## Reaction 944
#pdep_arrhenius('CH3 + C2H2 <=> C3H5-S',
            #[(1.0, 'atm'), 8.490000e+35, -8.43, 12356.0],
            #[(10.0, 'atm'), 3.040000e+32, -7.01, 12357.0],
            #[(100.0, 'atm'), 1.690000e+27, -5.07, 11690.0],
            #options='duplicate')

## Reaction 945
#reaction('C3H4-A + O2 <=> C3H3 + HO2', [4.000000e+13, 0.0, 41320.0])

## Reaction 946
#reaction('C3H4-P + O2 <=> C3H3 + HO2', [3.000000e+13, 0.0, 42630.0])

## Reaction 947
#reaction('C3H3 + OH <=> C3H2 + H2O', [2.000000e+13, 0.0, 8000.0])

## Reaction 948
#reaction('C3H3 + OH <=> CH2O + C2H2', [2.000000e+12, 0.0, 0.0])

## Reaction 949
#reaction('C3H3 + OH <=> C2H3 + HCO', [1.000000e+13, 0.0, 0.0])

## Reaction 950
#reaction('C3H3 + OH <=> C2H4 + CO', [1.000000e+13, 0.0, 0.0])

## Reaction 951
#reaction('C3H3 + OH <=> C3H2(S) + H2O', [1.000000e+12, 0.0, 0.0])

## Reaction 952
#reaction('C3H3 + OH <=> H2CCC(S) + H2O', [1.000000e+12, 0.0, 0.0])

## Reaction 953
#reaction('C3H3 + O2 <=> CH2CO + HCO', [1.700000e+05, 1.7, 1500.0])

## Reaction 954
#pdep_arrhenius('C3H3 + HO2 <=> C3H3O + OH',
            #[(0.01, 'atm'), 1.020000e+13, -0.158, -1417.0],
            #[(0.1, 'atm'), 4.980000e+14, -0.642, -349.1],
            #[(1.0, 'atm'), 7.770000e+17, -1.52, 2379.2],
            #[(10.0, 'atm'), 2.930000e+15, -0.684, 3615.3],
            #[(100.0, 'atm'), 1.640000e+04, 2.74, 1144.4])

## Reaction 955
#pdep_arrhenius('C3H3 + HO2 <=> C3H3O2H',
            #[(0.01, 'atm'), 1.910000e+31, -7.23, 1336.2],
            #[(0.1, 'atm'), 6.310000e+42, -10.3, 5568.9],
            #[(1.0, 'atm'), 1.030000e+45, -10.6, 7851.5],
            #[(10.0, 'atm'), 2.790000e+37, -7.92, 6497.9],
            #[(100.0, 'atm'), 4.730000e+25, -4.13, 2923.8])

## Reaction 956
#pdep_arrhenius('C3H3 + HO2 <=> C2HCHO + H2O',
            #[(0.01, 'atm'), 1.090000e+00, 3.01, -3421.1],
            #[(0.1, 'atm'), 6.350000e+01, 2.5, -2341.4],
            #[(1.0, 'atm'), 6.050000e+05, 1.39, 595.1],
            #[(10.0, 'atm'), 3.100000e+05, 1.59, 2677.6],
            #[(100.0, 'atm'), 5.070000e-05, 4.59, 927.5])

## Reaction 957
#pdep_arrhenius('C3H3O2H <=> C2HCHO + H2O',
            #[(0.01, 'atm'), 1.990000e+50, -12.7, 53531.9],
            #[(0.1, 'atm'), 4.720000e+47, -11.5, 54360.9],
            #[(1.0, 'atm'), 1.500000e+40, -8.84, 53179.2],
            #[(10.0, 'atm'), 2.540000e+28, -5.0, 49919.4],
            #[(100.0, 'atm'), 1.480000e+16, -1.12, 45949.3])

## Reaction 958
#pdep_arrhenius('C2H + CH2O <=> C3H3O',
            #[(0.001, 'atm'), 5.925000e-04, 2.609, -4297.3],
            #[(0.01, 'atm'), 5.688000e+06, -0.073, -1234.1],
            #[(0.1, 'atm'), 5.390000e+13, -1.803, 1452.2],
            #[(1.0, 'atm'), 8.733000e+15, -2.074, 2510.1],
            #[(10.0, 'atm'), 5.305000e+19, -2.943, 4532.2],
            #[(100.0, 'atm'), 3.779000e+21, -3.163, 6802.5])

## Reaction 959
#reaction('C2HCHO <=> C2H2 + CO', [2.510000e+14, 0.0, 68000.0])

## Reaction 960
#reaction('C2H + HCO <=> C2HCHO', [1.000000e+13, 0.0, 0.0])

## Reaction 961
#reaction('CH3CHCO + OH <=> C2H5 + CO2', [1.730000e+12, 0.0, -1010.0])

## Reaction 962
#reaction('CH3CHCO + OH <=> SC2H4OH + CO', [2.000000e+12, 0.0, -1010.0])

## Reaction 963
#reaction('CH3CHCO + H <=> C2H5 + CO', [4.400000e+12, 0.0, 1459.0])

## Reaction 964
#reaction('CH3CHCO + O <=> CH3CHO + CO', [3.200000e+12, 0.0, -437.0])

## Reaction 965
#reaction('NC3H7 + HO2 <=> NC3H7O + OH', [7.000000e+12, 0.0, -1000.0])

## Reaction 966
#reaction('IC3H7 + HO2 <=> IC3H7O + OH', [7.000000e+12, 0.0, -1000.0])

## Reaction 967
#reaction('CH3O2 + NC3H7 <=> CH3O + NC3H7O', [7.000000e+12, 0.0, -1000.0])

## Reaction 968
#reaction('CH3O2 + IC3H7 <=> CH3O + IC3H7O', [7.000000e+12, 0.0, -1000.0])

## Reaction 969
#reaction('NC3H7O2 + CH2O <=> NC3H7O2H + HCO', [5.600000e+12, 0.0, 13600.0])

## Reaction 970
#reaction('NC3H7O2 + CH3CHO <=> NC3H7O2H + CH3CO', [2.800000e+12, 0.0, 13600.0])

## Reaction 971
#reaction('IC3H7O2 + CH2O <=> IC3H7O2H + HCO', [5.600000e+12, 0.0, 13600.0])

## Reaction 972
#reaction('IC3H7O2 + CH3CHO <=> IC3H7O2H + CH3CO', [2.800000e+12, 0.0, 13600.0])

## Reaction 973
#reaction('NC3H7O2 + HO2 <=> NC3H7O2H + O2', [1.750000e+10, 0.0, -3275.0])

## Reaction 974
#reaction('IC3H7O2 + HO2 <=> IC3H7O2H + O2', [1.750000e+10, 0.0, -3275.0])

## Reaction 975
#reaction('C2H4 + NC3H7O2 <=> C2H3 + NC3H7O2H', [1.130000e+13, 0.0, 30430.0])

## Reaction 976
#reaction('C2H4 + IC3H7O2 <=> C2H3 + IC3H7O2H', [1.130000e+13, 0.0, 30430.0])

## Reaction 977
#reaction('CH3OH + NC3H7O2 <=> CH2OH + NC3H7O2H', [6.300000e+12, 0.0, 19360.0])

## Reaction 978
#reaction('CH3OH + IC3H7O2 <=> CH2OH + IC3H7O2H', [6.300000e+12, 0.0, 19360.0])

## Reaction 979
#reaction('C2H3CHO + NC3H7O2 <=> C2H3CO + NC3H7O2H', [2.800000e+12, 0.0, 13600.0])

## Reaction 980
#reaction('C2H3CHO + IC3H7O2 <=> C2H3CO + IC3H7O2H', [2.800000e+12, 0.0, 13600.0])

## Reaction 981
#reaction('CH4 + NC3H7O2 <=> CH3 + NC3H7O2H', [1.120000e+13, 0.0, 24640.0])

## Reaction 982
#reaction('CH4 + IC3H7O2 <=> CH3 + IC3H7O2H', [1.120000e+13, 0.0, 24640.0])

## Reaction 983
#reaction('NC3H7O2 + CH3O2 => NC3H7O + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 984
#reaction('IC3H7O2 + CH3O2 => IC3H7O + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 985
#reaction('H2 + NC3H7O2 <=> H + NC3H7O2H', [3.010000e+13, 0.0, 26030.0])

## Reaction 986
#reaction('H2 + IC3H7O2 <=> H + IC3H7O2H', [3.010000e+13, 0.0, 26030.0])

## Reaction 987
#reaction('IC3H7O2 + C2H6 <=> IC3H7O2H + C2H5', [1.700000e+13, 0.0, 20460.0])

## Reaction 988
#reaction('NC3H7O2 + C2H6 <=> NC3H7O2H + C2H5', [1.700000e+13, 0.0, 20460.0])

## Reaction 989
#reaction('IC3H7O2 + C2H5CHO <=> IC3H7O2H + C2H5CO', [2.000000e+11, 0.0, 9500.0])

## Reaction 990
#reaction('NC3H7O2 + C2H5CHO <=> NC3H7O2H + C2H5CO', [2.000000e+11, 0.0, 9500.0])

## Reaction 991
#reaction('IC3H7O2 + CH3CO3 => IC3H7O + CH3CO2 + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 992
#reaction('NC3H7O2 + CH3CO3 => NC3H7O + CH3CO2 + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 993
#reaction('IC3H7O2 + C2H5O2 => IC3H7O + C2H5O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 994
#reaction('NC3H7O2 + C2H5O2 => NC3H7O + C2H5O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 995
#reaction('IC3H7O2 + IC3H7O2 => O2 + IC3H7O + IC3H7O', [1.400000e+16, -1.61, 1860.0])

## Reaction 996
#reaction('NC3H7O2 + NC3H7O2 => O2 + NC3H7O + NC3H7O', [1.400000e+16, -1.61, 1860.0])

## Reaction 997
#reaction('IC3H7O2 + NC3H7O2 => IC3H7O + NC3H7O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 998
#reaction('IC3H7O2 + CH3 <=> IC3H7O + CH3O', [7.000000e+12, 0.0, -1000.0])

## Reaction 999
#reaction('IC3H7O2 + C2H5 <=> IC3H7O + C2H5O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1000
#reaction('IC3H7O2 + IC3H7 <=> IC3H7O + IC3H7O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1001
#reaction('IC3H7O2 + NC3H7 <=> IC3H7O + NC3H7O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1002
#reaction('IC3H7O2 + C3H5-A <=> IC3H7O + C3H5O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1003
#reaction('NC3H7O2 + CH3 <=> NC3H7O + CH3O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1004
#reaction('NC3H7O2 + C2H5 <=> NC3H7O + C2H5O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1005
#reaction('NC3H7O2 + IC3H7 <=> NC3H7O + IC3H7O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1006
#reaction('NC3H7O2 + NC3H7 <=> NC3H7O + NC3H7O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1007
#reaction('NC3H7O2 + C3H5-A <=> NC3H7O + C3H5O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1008
#reaction('NC3H7O2H <=> NC3H7O + OH', [1.500000e+16, 0.0, 42500.0])

## Reaction 1009
#reaction('IC3H7O + OH <=> IC3H7O2H', [1.000000e+15, -0.8, 0.0])

## Reaction 1010
#reaction('C2H5 + CH2O <=> NC3H7O', [1.000000e+11, 0.0, 3496.0])

## Reaction 1011
#reaction('C2H5CHO + H <=> NC3H7O', [4.000000e+12, 0.0, 6260.0])

## Reaction 1012
#reaction('CH3 + CH3CHO <=> IC3H7O', [1.000000e+11, 0.0, 9256.0])

## Reaction 1013
#reaction('CH3COCH3 + H <=> IC3H7O', [2.000000e+12, 0.0, 7270.0])

## Reaction 1014
#reaction('IC3H7O + O2 <=> CH3COCH3 + HO2', [9.090000e+09, 0.0, 390.0])

## Reaction 1015
#falloff_reaction('C4H10 (+ M) <=> C2H5 + C2H5 (+ M)',
                #kf=[1.355000e+37, -6.036, 92929.0],
                #kf0=[4.720000e+18, 0.0, 49578.0],
                #falloff=Troe(A=0.07998, T3=1e-20, T1=32430.0, T2=4858.0))

## Reaction 1016
#falloff_reaction('C4H10 (+ M) <=> NC3H7 + CH3 (+ M)',
                #kf=[6.600000e+52, -10.626, 100330.0],
                #kf0=[5.340000e+17, 0.0, 42959.0],
                #falloff=Troe(A=0.09502, T3=1e-20, T1=5348.0, T2=4326.0))

## Reaction 1017
#pdep_arrhenius('C4H10 <=> PC4H9 + H',
            #[(0.01, 'atm'), 4.450000e+90, -21.91, 140564.0],
            #[(0.1, 'atm'), 4.630000e+76, -17.64, 134669.0],
            #[(1.0, 'atm'), 4.940000e+58, -12.32, 125435.0],
            #[(10.0, 'atm'), 4.800000e+40, -7.06, 115302.0],
            #[(100.0, 'atm'), 1.490000e+27, -3.15, 107323.0])

## Reaction 1018
#pdep_arrhenius('C4H10 <=> SC4H9 + H',
            #[(0.01, 'atm'), 3.100000e+88, -21.24, 136355.0],
            #[(0.1, 'atm'), 4.340000e+73, -16.76, 129590.0],
            #[(1.0, 'atm'), 7.390000e+55, -11.52, 120199.0],
            #[(10.0, 'atm'), 8.520000e+38, -6.58, 110556.0],
            #[(100.0, 'atm'), 5.400000e+26, -3.05, 103313.0])

## Reaction 1019
#reaction('C4H10 + O2 <=> PC4H9 + HO2', [6.000000e+13, 0.0, 52340.0])

## Reaction 1020
#reaction('C4H10 + O2 <=> SC4H9 + HO2', [4.000000e+13, 0.0, 49800.0])

## Reaction 1021
#reaction('C4H10 + C3H5-A <=> PC4H9 + C3H6', [7.940000e+11, 0.0, 20500.0])

## Reaction 1022
#reaction('C4H10 + C3H5-A <=> SC4H9 + C3H6', [3.160000e+11, 0.0, 16400.0])

## Reaction 1023
#reaction('C4H10 + C2H5 <=> PC4H9 + C2H6', [1.580000e+11, 0.0, 12300.0])

## Reaction 1024
#reaction('C4H10 + C2H5 <=> SC4H9 + C2H6', [1.000000e+11, 0.0, 10400.0])

## Reaction 1025
#reaction('C4H10 + C2H3 <=> PC4H9 + C2H4', [1.000000e+12, 0.0, 18000.0])

## Reaction 1026
#reaction('C4H10 + C2H3 <=> SC4H9 + C2H4', [8.000000e+11, 0.0, 16800.0])

## Reaction 1027
#reaction('C4H10 + CH3 <=> PC4H9 + CH4', [9.040000e-01, 3.65, 7154.0])

## Reaction 1028
#reaction('C4H10 + CH3 <=> SC4H9 + CH4', [3.020000e+00, 3.46, 5481.0])

## Reaction 1029
#reaction('C4H10 + H <=> PC4H9 + H2', [3.490000e+05, 2.69, 6450.0])

## Reaction 1030
#reaction('C4H10 + H <=> SC4H9 + H2', [2.600000e+06, 2.4, 4471.0])

## Reaction 1031
#reaction('C4H10 + OH <=> PC4H9 + H2O', [1.054000e+10, 0.97, 1586.0])

## Reaction 1032
#reaction('C4H10 + OH <=> SC4H9 + H2O', [9.340000e+07, 1.61, -35.0])

## Reaction 1033
#reaction('C4H10 + O <=> PC4H9 + OH', [1.130000e+14, 0.0, 7850.0])

## Reaction 1034
#reaction('C4H10 + O <=> SC4H9 + OH', [5.620000e+13, 0.0, 5200.0])

## Reaction 1035
#reaction('C4H10 + HO2 <=> PC4H9 + H2O2', [4.080000e+01, 3.59, 17160.0])

## Reaction 1036
#reaction('C4H10 + HO2 <=> SC4H9 + H2O2', [1.264000e+02, 3.37, 13720.0])

## Reaction 1037
#reaction('C4H10 + CH3O <=> PC4H9 + CH3OH', [3.000000e+11, 0.0, 7000.0])

## Reaction 1038
#reaction('C4H10 + CH3O <=> SC4H9 + CH3OH', [6.000000e+11, 0.0, 7000.0])

## Reaction 1039
#reaction('C4H10 + C2H5O <=> PC4H9 + C2H5OH', [3.000000e+11, 0.0, 7000.0])

## Reaction 1040
#reaction('C4H10 + C2H5O <=> SC4H9 + C2H5OH', [6.000000e+11, 0.0, 7000.0])

## Reaction 1041
#reaction('C4H10 + PC4H9 <=> SC4H9 + C4H10', [1.000000e+11, 0.0, 10400.0])

## Reaction 1042
#reaction('C4H10 + CH3CO3 <=> PC4H9 + CH3CO3H', [1.700000e+13, 0.0, 20460.0])

## Reaction 1043
#reaction('C4H10 + CH3CO3 <=> SC4H9 + CH3CO3H', [1.120000e+13, 0.0, 17700.0])

## Reaction 1044
#reaction('C4H10 + O2CHO <=> PC4H9 + HO2CHO', [1.680000e+13, 0.0, 20440.0])

## Reaction 1045
#reaction('C4H10 + O2CHO <=> SC4H9 + HO2CHO', [1.120000e+13, 0.0, 17690.0])

## Reaction 1046
#reaction('CH3O2 + C4H10 <=> CH3O2H + PC4H9', [1.386000e+00, 3.97, 18280.0])

## Reaction 1047
#reaction('CH3O2 + C4H10 <=> CH3O2H + SC4H9', [2.037000e+01, 3.58, 14810.0])

## Reaction 1048
#reaction('C2H5O2 + C4H10 <=> C2H5O2H + PC4H9', [4.080000e+01, 3.59, 17160.0])

## Reaction 1049
#reaction('C2H5O2 + C4H10 <=> C2H5O2H + SC4H9', [1.264000e+02, 3.37, 13720.0])

## Reaction 1050
#reaction('NC3H7O2 + C4H10 <=> NC3H7O2H + PC4H9', [1.700000e+13, 0.0, 20460.0])

## Reaction 1051
#reaction('NC3H7O2 + C4H10 <=> NC3H7O2H + SC4H9', [1.120000e+13, 0.0, 17700.0])

## Reaction 1052
#reaction('IC3H7O2 + C4H10 <=> IC3H7O2H + PC4H9', [1.700000e+13, 0.0, 20460.0])

## Reaction 1053
#reaction('IC3H7O2 + C4H10 <=> IC3H7O2H + SC4H9', [1.120000e+13, 0.0, 17700.0])

## Reaction 1054
#reaction('PC4H9O2 + C3H8 <=> PC4H9O2H + NC3H7', [1.700000e+13, 0.0, 20460.0])

## Reaction 1055
#reaction('PC4H9O2 + C3H8 <=> PC4H9O2H + IC3H7', [2.000000e+12, 0.0, 17000.0])

## Reaction 1056
#reaction('PC4H9O2 + C4H10 <=> PC4H9O2H + PC4H9', [1.700000e+13, 0.0, 20460.0])

## Reaction 1057
#reaction('PC4H9O2 + C4H10 <=> PC4H9O2H + SC4H9', [1.120000e+13, 0.0, 17700.0])

## Reaction 1058
#reaction('SC4H9O2 + C3H8 <=> SC4H9O2H + NC3H7', [1.700000e+13, 0.0, 20460.0])

## Reaction 1059
#reaction('SC4H9O2 + C3H8 <=> SC4H9O2H + IC3H7', [2.000000e+12, 0.0, 17000.0])

## Reaction 1060
#reaction('SC4H9O2 + C4H10 <=> SC4H9O2H + PC4H9', [1.700000e+13, 0.0, 20460.0])

## Reaction 1061
#reaction('SC4H9O2 + C4H10 <=> SC4H9O2H + SC4H9', [1.120000e+13, 0.0, 17700.0])

## Reaction 1062
#reaction('C4H6 <=> C4H5-I + H', [5.700000e+36, -6.27, 112353.0])

## Reaction 1063
#reaction('C4H6 <=> C4H5-N + H', [5.300000e+44, -8.62, 123608.0])

## Reaction 1064
#reaction('C4H6 <=> C4H4 + H2', [2.500000e+15, 0.0, 94700.0])

## Reaction 1065
#reaction('C4H6 + H <=> C4H5-N + H2', [1.330000e+06, 2.53, 12240.0])

## Reaction 1066
#reaction('C4H6 + H <=> C4H5-I + H2', [6.650000e+05, 2.53, 9240.0])

## Reaction 1067
#pdep_arrhenius('C4H6 + H <=> C2H4 + C2H3',
            #[(1.0, 'atm'), 1.460000e+30, -4.34, 21647.0],
            #[(10.0, 'atm'), 5.450000e+30, -4.51, 21877.0])

## Reaction 1068
#reaction('C4H6 + H <=> C3H4-P + CH3', [2.000000e+12, 0.0, 7000.0])

## Reaction 1069
#reaction('C4H6 + H <=> C3H4-A + CH3', [2.000000e+12, 0.0, 7000.0])

## Reaction 1070
#reaction('C4H6 + O <=> C4H5-N + OH', [7.500000e+06, 1.9, 3740.0])

## Reaction 1071
#reaction('C4H6 + O <=> C4H5-I + OH', [7.500000e+06, 1.9, 3740.0])

## Reaction 1072
#reaction('C4H6 + O <=> C2H2 + C2H4O1-2', [1.000000e+08, 1.45, -860.0])

## Reaction 1073
#reaction('C4H6 + O <=> SC3H5CO + H', [5.000000e+07, 1.45, -860.0])

## Reaction 1074
#reaction('C4H6 + O <=> CH2CHCHCHO + H', [4.500000e+08, 1.45, -860.0])

## Reaction 1075
#reaction('C4H6 + OH <=> C4H5-N + H2O', [6.200000e+06, 2.0, 3430.0])

## Reaction 1076
#reaction('C4H6 + OH <=> C4H5-I + H2O', [3.100000e+06, 2.0, 430.0])

## Reaction 1077
#reaction('C4H6 + OH <=> C2H3CHO + CH3', [1.370000e+12, 0.0, -1040.0])

## Reaction 1078
#reaction('C4H6 + OH <=> C3H5-A + CH2O', [1.370000e+12, 0.0, -1040.0])

## Reaction 1079
#reaction('C4H6 + HO2 <=> C4H6O25 + OH', [1.200000e+12, 0.0, 14000.0])

## Reaction 1080
#reaction('C4H6 + HO2 <=> C2H3CHOCH2 + OH', [4.800000e+12, 0.0, 14000.0])

## Reaction 1081
#reaction('C4H6 + CH3 <=> C4H5-N + CH4', [2.000000e+14, 0.0, 22800.0])

## Reaction 1082
#reaction('C4H6 + CH3 <=> C4H5-I + CH4', [1.000000e+14, 0.0, 19800.0])

## Reaction 1083
#reaction('C4H6 + C2H3 <=> C4H5-N + C2H4', [5.000000e+13, 0.0, 22800.0])

## Reaction 1084
#reaction('C4H6 + C2H3 <=> C4H5-I + C2H4', [2.500000e+13, 0.0, 19800.0])

## Reaction 1085
#reaction('C4H6 + C3H3 <=> C4H5-N + C3H4-A', [1.000000e+13, 0.0, 22500.0])

## Reaction 1086
#reaction('C4H6 + C3H3 <=> C4H5-I + C3H4-A', [5.000000e+12, 0.0, 19500.0])

## Reaction 1087
#reaction('C4H6 + C3H5-A <=> C4H5-N + C3H6', [1.000000e+13, 0.0, 22500.0])

## Reaction 1088
#reaction('C4H6 + C3H5-A <=> C4H5-I + C3H6', [5.000000e+12, 0.0, 19500.0])

## Reaction 1089
#pdep_arrhenius('C2H3 + C2H2 <=> C4H4 + H',
            #[(0.0132, 'atm'), 7.200000e+13, -0.48, 6100.0],
            #[(0.0263, 'atm'), 5.000000e+14, -0.71, 6700.0],
            #[(0.12, 'atm'), 4.600000e+16, -1.25, 8400.0],
            #[(1.0, 'atm'), 2.000000e+18, -1.68, 10600.0],
            #[(10.0, 'atm'), 4.900000e+16, -1.13, 11800.0])

## Reaction 1090
#pdep_arrhenius('C2H3 + C2H2 <=> C4H5-N',
            #[(0.0132, 'atm'), 1.100000e+31, -7.14, 5600.0],
            #[(0.0263, 'atm'), 1.100000e+32, -7.33, 6200.0],
            #[(0.12, 'atm'), 2.400000e+31, -6.95, 5600.0],
            #[(1.0, 'atm'), 9.300000e+38, -8.76, 12000.0],
            #[(10.0, 'atm'), 8.100000e+37, -8.09, 13400.0])

## Reaction 1091
#pdep_arrhenius('C2H3 + C2H2 <=> C4H5-I',
            #[(0.0132, 'atm'), 5.000000e+34, -8.42, 7900.0],
            #[(0.0263, 'atm'), 2.100000e+36, -8.78, 9100.0],
            #[(0.12, 'atm'), 1.000000e+37, -8.77, 9800.0],
            #[(1.0, 'atm'), 1.600000e+46, -10.98, 18600.0],
            #[(10.0, 'atm'), 5.100000e+53, -12.64, 28800.0])

## Reaction 1092
#pdep_arrhenius('C2H3 + C2H3 <=> C4H6',
            #[(0.0263, 'atm'), 7.000000e+57, -13.82, 17629.0],
            #[(0.12, 'atm'), 1.500000e+52, -11.97, 16056.0],
            #[(1.0, 'atm'), 1.500000e+42, -8.84, 12483.0])

## Reaction 1093
#pdep_arrhenius('C2H3 + C2H3 <=> C4H5-I + H',
            #[(0.0263, 'atm'), 1.500000e+30, -4.95, 12958.0],
            #[(0.12, 'atm'), 7.200000e+28, -4.49, 14273.0],
            #[(1.0, 'atm'), 1.200000e+22, -2.44, 13654.0])

## Reaction 1094
#pdep_arrhenius('C2H3 + C2H3 <=> C4H5-N + H',
            #[(0.0263, 'atm'), 1.100000e+24, -3.28, 12395.0],
            #[(0.12, 'atm'), 4.600000e+24, -3.38, 14650.0],
            #[(1.0, 'atm'), 2.400000e+20, -2.04, 15361.0])

## Reaction 1095
#pdep_arrhenius('C4H5-N <=> C4H5-I',
            #[(0.0132, 'atm'), 2.400000e+60, -16.08, 47500.0],
            #[(0.0263, 'atm'), 1.300000e+62, -16.38, 49600.0],
            #[(0.12, 'atm'), 4.900000e+66, -17.26, 55400.0],
            #[(1.0, 'atm'), 1.500000e+67, -16.89, 59100.0],
            #[(10.0, 'atm'), 2.000000e+60, -14.46, 58600.0])

## Reaction 1096
#reaction('C4H5-N + H <=> C4H5-I + H', [3.100000e+26, -3.35, 17423.0])

## Reaction 1097
#reaction('C4H5-N + H <=> C4H4 + H2', [1.500000e+13, 0.0, 0.0])

## Reaction 1098
#reaction('C4H5-N + OH <=> C4H4 + H2O', [2.000000e+12, 0.0, 0.0])

## Reaction 1099
#reaction('C4H5-N + HCO <=> C4H6 + CO', [5.000000e+12, 0.0, 0.0])

## Reaction 1100
#reaction('C4H5-N + HO2 => C2H3 + CH2CO + OH', [6.600000e+12, 0.0, 0.0])

## Reaction 1101
#reaction('C4H5-N + H2O2 <=> C4H6 + HO2', [1.210000e+10, 0.0, -596.0])

## Reaction 1102
#reaction('C4H5-N + HO2 <=> C4H6 + O2', [6.000000e+11, 0.0, 0.0])

## Reaction 1103
#reaction('C4H5-N + O2 <=> CH2CHCHCHO + O', [3.000000e+11, 0.29, 11.0])

## Reaction 1104
#reaction('C4H5-N + O2 <=> HCO + C2H3CHO', [9.200000e+16, -1.39, 1010.0])

## Reaction 1105
#reaction('C4H5-I + H <=> C4H4 + H2', [3.000000e+13, 0.0, 0.0])

## Reaction 1106
#reaction('C4H5-I + H <=> C3H3 + CH3', [2.000000e+13, 0.0, 2000.0])

## Reaction 1107
#reaction('C4H5-I + OH <=> C4H4 + H2O', [4.000000e+12, 0.0, 0.0])

## Reaction 1108
#reaction('C4H5-I + HCO <=> C4H6 + CO', [5.000000e+12, 0.0, 0.0])

## Reaction 1109
#reaction('C4H5-I + HO2 <=> C4H6 + O2', [6.000000e+11, 0.0, 0.0])

## Reaction 1110
#reaction('C4H5-I + HO2 => C2H3 + CH2CO + OH', [6.600000e+12, 0.0, 0.0])

## Reaction 1111
#reaction('C4H5-I + H2O2 <=> C4H6 + HO2', [1.210000e+10, 0.0, -596.0])

## Reaction 1112
#reaction('C4H5-I + O2 <=> CH2CO + CH2CHO', [2.160000e+10, 0.0, 2500.0])

## Reaction 1113
#reaction('C4H5-2 <=> C4H5-I', [1.500000e+67, -16.89, 59100.0])

## Reaction 1114
#reaction('C4H5-2 + H <=> C4H5-I + H', [3.100000e+26, -3.35, 17423.0])

## Reaction 1115
#reaction('C4H5-2 + HO2 => OH + C2H2 + CH3CO', [8.000000e+11, 0.0, 0.0])

## Reaction 1116
#reaction('C4H5-2 + O2 <=> CH3CO + CH2CO', [2.160000e+10, 0.0, 2500.0])

## Reaction 1117
#reaction('C4H612 <=> C4H5-I + H', [4.200000e+15, 0.0, 92600.0])

## Reaction 1118
#reaction('C4H612 + H <=> C4H6 + H', [2.000000e+13, 0.0, 4000.0])

## Reaction 1119
#reaction('C4H612 + H <=> C4H5-I + H2', [1.700000e+05, 2.5, 2490.0])

## Reaction 1120
#reaction('C4H612 + H <=> C3H4-A + CH3', [2.000000e+13, 0.0, 2000.0])

## Reaction 1121
#reaction('C4H612 + H <=> C3H4-P + CH3', [2.000000e+13, 0.0, 2000.0])

## Reaction 1122
#reaction('C4H612 + CH3 <=> C4H5-I + CH4', [7.000000e+13, 0.0, 18500.0])

## Reaction 1123
#reaction('C4H612 + O <=> CH2CO + C2H4', [1.200000e+08, 1.65, 327.0])

## Reaction 1124
#reaction('C4H612 + O <=> C4H5-I + OH', [1.800000e+11, 0.7, 5880.0])

## Reaction 1125
#reaction('C4H612 + OH <=> C4H5-I + H2O', [3.100000e+06, 2.0, -298.0])

## Reaction 1126
#reaction('C4H612 <=> C4H6', [3.000000e+13, 0.0, 65000.0])

## Reaction 1127
#reaction('C4H6-2 <=> C4H6', [3.000000e+13, 0.0, 65000.0])

## Reaction 1128
#reaction('C4H6-2 <=> C4H612', [3.000000e+13, 0.0, 67000.0])

## Reaction 1129
#reaction('C4H6-2 + H <=> C4H612 + H', [2.000000e+13, 0.0, 4000.0])

## Reaction 1130
#reaction('C4H6-2 + H <=> C4H5-2 + H2', [3.400000e+05, 2.5, 2490.0])

## Reaction 1131
#reaction('C4H6-2 + H <=> CH3 + C3H4-P', [2.600000e+05, 2.5, 1000.0])

## Reaction 1132
#reaction('C4H6-2 <=> H + C4H5-2', [5.000000e+15, 0.0, 87300.0])

## Reaction 1133
#reaction('C4H6-2 + CH3 <=> C4H5-2 + CH4', [1.400000e+14, 0.0, 18500.0])

## Reaction 1134
#reaction('C2H3CHOCH2 <=> C4H6O23', [2.000000e+14, 0.0, 50600.0])

## Reaction 1135
#reaction('C4H6O23 <=> SC3H5CHO', [1.950000e+13, 0.0, 49400.0])

## Reaction 1136
#reaction('C4H6O23 <=> C2H4 + CH2CO', [5.750000e+15, 0.0, 69300.0])

## Reaction 1137
#reaction('C4H6O23 <=> C2H2 + C2H4O1-2', [1.000000e+16, 0.0, 75800.0])

## Reaction 1138
#reaction('C4H6O25 <=> C4H4O + H2', [5.300000e+12, 0.0, 48500.0])

## Reaction 1139
#reaction('C4H4O <=> CO + C3H4-P', [1.780000e+15, 0.0, 77500.0])

## Reaction 1140
#reaction('C4H4O <=> C2H2 + CH2CO', [5.010000e+14, 0.0, 77500.0])

## Reaction 1141
#reaction('SC3H5CHO <=> C3H6 + CO', [3.900000e+14, 0.0, 69000.0])

## Reaction 1142
#reaction('SC3H5CHO + H <=> CH2CHCHCHO + H2', [1.700000e+05, 2.5, 2490.0])

## Reaction 1143
#reaction('SC3H5CHO + H <=> CH3 + C2H3CHO', [4.000000e+21, -2.39, 11180.0])

## Reaction 1144
#reaction('SC3H5CHO + H <=> C3H6 + HCO', [4.000000e+21, -2.39, 11180.0])

## Reaction 1145
#reaction('SC3H5CHO + CH3 <=> CH2CHCHCHO + CH4', [2.100000e+00, 3.5, 5675.0])

## Reaction 1146
#reaction('SC3H5CHO + C2H3 <=> CH2CHCHCHO + C2H4', [2.210000e+00, 3.5, 4682.0])

## Reaction 1147
#reaction('SC3H5CHO + C2H3 <=> SC3H5CO + C2H4', [1.110000e+00, 3.5, 4682.0])

## Reaction 1148
#reaction('SC3H5CHO + OH <=> CH2CHCHCHO + H2O', [4.460000e+06, 2.072, 1051.0])

## Reaction 1149
#reaction('SC3H5CHO + OH <=> SC3H5CO + H2O', [3.370000e+12, 0.0, -619.0])

## Reaction 1150
#reaction('SC3H5CHO + O <=> CH2CHCHCHO + OH', [5.240000e+11, 0.7, 5884.0])

## Reaction 1151
#reaction('SC3H5CO + H <=> SC3H5CHO', [1.000000e+14, 0.0, 0.0])

## Reaction 1152
#reaction('CH2CHCHCHO + H <=> SC3H5CHO', [1.000000e+14, 0.0, 0.0])

## Reaction 1153
#pdep_arrhenius('C4H4 + H <=> C4H5-N',
            #[(0.0132, 'atm'), 1.200000e+51, -12.57, 12300.0],
            #[(0.0263, 'atm'), 4.200000e+50, -12.34, 12500.0],
            #[(0.12, 'atm'), 1.100000e+50, -11.94, 13400.0],
            #[(1.0, 'atm'), 1.300000e+51, -11.92, 16500.0],
            #[(10.0, 'atm'), 6.200000e+45, -10.08, 15800.0])

## Reaction 1154
#pdep_arrhenius('C4H4 + H <=> C4H5-I',
            #[(0.0132, 'atm'), 6.100000e+53, -13.19, 14200.0],
            #[(0.0263, 'atm'), 9.600000e+52, -12.85, 14300.0],
            #[(0.12, 'atm'), 2.100000e+52, -12.44, 15500.0],
            #[(1.0, 'atm'), 4.900000e+51, -11.92, 17700.0],
            #[(10.0, 'atm'), 1.500000e+48, -10.58, 18800.0])

## Reaction 1155
#reaction('C4H4 + H <=> C4H3-N + H2', [6.650000e+05, 2.53, 12240.0])

## Reaction 1156
#reaction('C4H4 + H <=> C4H3-I + H2', [3.330000e+05, 2.53, 9240.0])

## Reaction 1157
#reaction('C4H4 + OH <=> C4H3-N + H2O', [3.100000e+07, 2.0, 3430.0])

## Reaction 1158
#reaction('C4H4 + OH <=> C4H3-I + H2O', [1.550000e+07, 2.0, 430.0])

## Reaction 1159
#reaction('C4H4 + O <=> C3H3 + HCO', [6.000000e+08, 1.45, -860.0])

## Reaction 1160
#reaction('C3H3 + HCCO <=> C4H4 + CO', [2.500000e+13, 0.0, 0.0])

## Reaction 1161
#reaction('C3H3 + CH <=> C4H3-I + H', [5.000000e+13, 0.0, 0.0])

## Reaction 1162
#reaction('C3H3 + CH2 <=> C4H4 + H', [5.000000e+13, 0.0, 0.0])

## Reaction 1163
#falloff_reaction('C3H3 + CH3 (+ M) <=> C4H612 (+ M)',
                #kf=[1.500000e+12, 0.0, 0.0],
                #kf0=[2.600000e+57, -11.94, 9770.0],
                #efficiencies='H2:2.0 H2O:6.0 CH4:2.0 CO:1.5 CO2:2.0 C2H6:3.0 AR:0.7',
                #falloff=Troe(A=0.175, T3=1341.0, T1=60000.0, T2=9770.0))

## Reaction 1164
#falloff_reaction('C2H2 + C2H (+ M) <=> C4H3-N (+ M)',
                #kf=[8.300000e+10, 0.899, -363.0],
                #kf0=[1.240000e+31, -4.718, 1871.0],
                #efficiencies='H2:2.0 H2O:6.0 CH4:2.0 CO:1.5 CO2:2.0 C2H6:3.0 C2H2:2.5 C2H4:2.5',
                #falloff=Troe(A=1.0, T3=100.0, T1=5613.0, T2=13390.0))

## Reaction 1165
#reaction('C4H3-N <=> C4H3-I', [4.100000e+43, -9.49, 53000.0])

## Reaction 1166
#reaction('C4H3-N + H <=> C4H3-I + H', [2.500000e+20, -1.67, 10800.0])

## Reaction 1167
#reaction('C4H3-N + H <=> C2H2 + H2CC', [6.300000e+25, -3.34, 10014.0])

## Reaction 1168
#reaction('C4H3-N + H <=> C4H4', [2.000000e+47, -10.26, 13070.0])

## Reaction 1169
#reaction('C4H3-N + H <=> C4H2 + H2', [3.000000e+13, 0.0, 0.0])

## Reaction 1170
#reaction('C4H3-N + OH <=> C4H2 + H2O', [2.000000e+12, 0.0, 0.0])

## Reaction 1171
#falloff_reaction('C2H2 + C2H (+ M) <=> C4H3-I (+ M)',
                #kf=[8.300000e+10, 0.899, -363.0],
                #kf0=[1.240000e+31, -4.718, 1871.0],
                #efficiencies='H2:2.0 H2O:6.0 CH4:2.0 CO:1.5 CO2:2.0 C2H6:3.0 C2H2:2.5 C2H4:2.5',
                #falloff=Troe(A=1.0, T3=100.0, T1=5613.0, T2=13390.0))

## Reaction 1172
#reaction('C4H3-I + H <=> C2H2 + H2CC', [2.800000e+23, -2.55, 10780.0])

## Reaction 1173
#reaction('C4H3-I + H <=> C4H4', [3.400000e+43, -9.01, 12120.0])

## Reaction 1174
#reaction('C4H3-I + H <=> C4H2 + H2', [6.000000e+13, 0.0, 0.0])

## Reaction 1175
#reaction('C4H3-I + OH <=> C4H2 + H2O', [4.000000e+12, 0.0, 0.0])

## Reaction 1176
#reaction('C4H3-I + O2 <=> HCCO + CH2CO', [7.860000e+16, -1.8, 0.0])

## Reaction 1177
#reaction('C4H3-I + CH2 <=> C3H4-A + C2H', [2.000000e+13, 0.0, 0.0])

## Reaction 1178
#reaction('C2H2 + C2H <=> C4H2 + H', [9.600000e+13, 0.0, 0.0])

## Reaction 1179
#reaction('C4H2 + H <=> C4H3-N', [1.100000e+42, -8.72, 15300.0])

## Reaction 1180
#reaction('C4H2 + H <=> C4H3-I', [1.100000e+30, -4.92, 10800.0])

## Reaction 1181
#reaction('C4H2 + OH <=> H2C4O + H', [6.600000e+12, 0.0, -410.0])

## Reaction 1182
#reaction('H2C4O + H <=> C2H2 + HCCO', [5.000000e+13, 0.0, 3000.0])

## Reaction 1183
#reaction('H2C4O + OH <=> CH2CO + HCCO', [1.000000e+07, 2.0, 2000.0])

## Reaction 1184
#falloff_reaction('H2CC + C2H2 (+ M) <=> C4H4 (+ M)',
                #kf=[3.500000e+05, 2.055, -2400.0],
                #kf0=[1.400000e+60, -12.599, 7417.0],
                #efficiencies='H2:2.0 H2O:6.0 CH4:2.0 CO:1.5 CO2:2.0 C2H6:3.0 C2H2:3.0 C2H4:3.0',
                #falloff=Troe(A=0.98, T3=56.0, T1=580.0, T2=4164.0))

## Reaction 1185
#reaction('H2CC + C2H4 <=> C4H6', [1.000000e+12, 0.0, 0.0])

## Reaction 1186
#falloff_reaction('IC4H10 (+ M) <=> CH3 + IC3H7 (+ M)',
                #kf=[2.520000e+31, -4.102, 91495.0],
                #kf0=[2.410000e+19, 0.0, 52576.0],
                #falloff=Troe(A=0.3662, T3=815.3, T1=60.79, T2=1e+20))

## Reaction 1187
#reaction('IC4H10 <=> TC4H9 + H', [2.510000e+98, -23.81, 145300.0])

## Reaction 1188
#reaction('IC4H10 <=> IC4H9 + H', [9.850000e+95, -23.11, 147600.0])

## Reaction 1189
#reaction('IC4H10 + H <=> TC4H9 + H2', [6.020000e+05, 2.4, 2583.0])

## Reaction 1190
#reaction('IC4H10 + H <=> IC4H9 + H2', [1.810000e+06, 2.54, 6756.0])

## Reaction 1191
#reaction('IC4H10 + CH3 <=> TC4H9 + CH4', [9.040000e-01, 3.46, 4598.0])

## Reaction 1192
#reaction('IC4H10 + CH3 <=> IC4H9 + CH4', [1.360000e+00, 3.65, 7154.0])

## Reaction 1193
#reaction('IC4H10 + OH <=> TC4H9 + H2O', [2.925000e+04, 2.531, -1659.0])

## Reaction 1194
#reaction('IC4H10 + OH <=> IC4H9 + H2O', [6.654000e+04, 2.665, -168.9])

## Reaction 1195
#reaction('IC4H10 + C2H5 <=> IC4H9 + C2H6', [1.510000e+12, 0.0, 10400.0])

## Reaction 1196
#reaction('IC4H10 + C2H5 <=> TC4H9 + C2H6', [1.000000e+11, 0.0, 7900.0])

## Reaction 1197
#reaction('IC4H10 + HO2 <=> IC4H9 + H2O2', [6.120000e+01, 3.59, 17160.0])

## Reaction 1198
#reaction('IC4H10 + HO2 <=> TC4H9 + H2O2', [4.332000e+02, 3.01, 12090.0])

## Reaction 1199
#reaction('IC4H10 + O <=> TC4H9 + OH', [1.968000e+05, 2.402, 1150.0])

## Reaction 1200
#reaction('IC4H10 + O <=> IC4H9 + OH', [4.046000e+07, 2.034, 5136.0])

## Reaction 1201
#reaction('IC4H10 + CH3O <=> IC4H9 + CH3OH', [4.800000e+11, 0.0, 7000.0])

## Reaction 1202
#reaction('IC4H10 + CH3O <=> TC4H9 + CH3OH', [1.900000e+10, 0.0, 2800.0])

## Reaction 1203
#reaction('IC4H10 + O2 <=> IC4H9 + HO2', [9.000000e+13, 0.0, 52290.0])

## Reaction 1204
#reaction('IC4H10 + O2 <=> TC4H9 + HO2', [1.000000e+13, 0.0, 48200.0])

## Reaction 1205
#reaction('IC4H10 + CH3O2 <=> IC4H9 + CH3O2H', [2.079000e+00, 3.97, 18280.0])

## Reaction 1206
#reaction('IC4H10 + C2H5O2 <=> IC4H9 + C2H5O2H', [2.550000e+13, 0.0, 20460.0])

## Reaction 1207
#reaction('IC4H10 + CH3CO3 <=> IC4H9 + CH3CO3H', [2.550000e+13, 0.0, 20460.0])

## Reaction 1208
#reaction('IC4H10 + NC3H7O2 <=> IC4H9 + NC3H7O2H', [2.550000e+13, 0.0, 20460.0])

## Reaction 1209
#reaction('IC4H10 + IC3H7O2 <=> IC4H9 + IC3H7O2H', [2.550000e+13, 0.0, 20460.0])

## Reaction 1210
#reaction('IC4H10 + IC4H9O2 <=> IC4H9 + IC4H9O2H', [2.550000e+13, 0.0, 20460.0])

## Reaction 1211
#reaction('IC4H10 + TC4H9O2 <=> IC4H9 + TC4H9O2H', [2.550000e+13, 0.0, 20460.0])

## Reaction 1212
#reaction('IC4H10 + O2CHO <=> IC4H9 + HO2CHO', [2.520000e+13, 0.0, 20440.0])

## Reaction 1213
#reaction('IC4H10 + O2CHO <=> TC4H9 + HO2CHO', [2.800000e+12, 0.0, 16010.0])

## Reaction 1214
#reaction('IC4H10 + SC4H9O2 <=> IC4H9 + SC4H9O2H', [2.250000e+13, 0.0, 20460.0])

## Reaction 1215
#reaction('IC4H10 + SC4H9O2 <=> TC4H9 + SC4H9O2H', [2.800000e+12, 0.0, 16000.0])

## Reaction 1216
#reaction('IC4H10 + PC4H9O2 <=> IC4H9 + PC4H9O2H', [2.250000e+13, 0.0, 20460.0])

## Reaction 1217
#reaction('IC4H10 + PC4H9O2 <=> TC4H9 + PC4H9O2H', [2.800000e+12, 0.0, 16000.0])

## Reaction 1218
#reaction('IC4H10 + CH3O2 <=> TC4H9 + CH3O2H', [1.366000e+02, 3.12, 13190.0])

## Reaction 1219
#reaction('IC4H10 + C2H5O2 <=> TC4H9 + C2H5O2H', [2.800000e+12, 0.0, 16000.0])

## Reaction 1220
#reaction('IC4H10 + CH3CO3 <=> TC4H9 + CH3CO3H', [2.800000e+12, 0.0, 16000.0])

## Reaction 1221
#reaction('IC4H10 + NC3H7O2 <=> TC4H9 + NC3H7O2H', [2.800000e+12, 0.0, 16000.0])

## Reaction 1222
#reaction('IC4H10 + IC3H7O2 <=> TC4H9 + IC3H7O2H', [2.800000e+12, 0.0, 16000.0])

## Reaction 1223
#reaction('IC4H10 + IC4H9O2 <=> TC4H9 + IC4H9O2H', [2.800000e+12, 0.0, 16000.0])

## Reaction 1224
#reaction('IC4H10 + TC4H9O2 <=> TC4H9 + TC4H9O2H', [2.800000e+12, 0.0, 16000.0])

## Reaction 1225
#reaction('IC4H10 + IC4H9 <=> TC4H9 + IC4H10', [2.500000e+10, 0.0, 7900.0])

## Reaction 1226
#reaction('IC4H9 + HO2 <=> IC4H9O + OH', [7.000000e+12, 0.0, -1000.0])

## Reaction 1227
#reaction('TC4H9 + HO2 <=> TC4H9O + OH', [7.000000e+12, 0.0, -1000.0])

## Reaction 1228
#reaction('CH3O2 + IC4H9 <=> CH3O + IC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1229
#reaction('CH3O2 + TC4H9 <=> CH3O + TC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1230
#reaction('IC4H9 <=> TC4H9', [3.560000e+10, 0.88, 34600.0])

## Reaction 1231
#reaction('TC4H9 + O2 <=> IC4H8 + HO2', [8.370000e-01, 3.59, 11960.0])

## Reaction 1232
#reaction('IC4H9 + O2 <=> IC4H8 + HO2', [1.070000e+00, 3.71, 9322.0])

## Reaction 1233
#reaction('NC3H7O2 + IC4H9 <=> NC3H7O + IC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1234
#reaction('NC3H7O2 + TC4H9 <=> NC3H7O + TC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1235
#reaction('NC3H7O2 + IC4H7 <=> NC3H7O + IC4H7O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1236
#reaction('SC4H9O2 + IC4H9 <=> SC4H9O + IC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1237
#reaction('SC4H9O2 + TC4H9 <=> SC4H9O + TC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1238
#reaction('PC4H9O2 + IC4H9 <=> PC4H9O + IC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1239
#reaction('PC4H9O2 + TC4H9 <=> PC4H9O + TC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1240
#reaction('PC4H9O2 + IC4H7 <=> PC4H9O + IC4H7O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1241
#reaction('SC4H9O2 + IC4H7 <=> SC4H9O + IC4H7O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1242
#reaction('IC4H9O2 + C4H10 <=> IC4H9O2H + SC4H9', [1.120000e+13, 0.0, 17700.0])

## Reaction 1243
#reaction('TC4H9O2 + C4H10 <=> TC4H9O2H + SC4H9', [1.120000e+13, 0.0, 17700.0])

## Reaction 1244
#reaction('IC4H9O2 + C4H10 <=> IC4H9O2H + PC4H9', [1.700000e+13, 0.0, 20460.0])

## Reaction 1245
#reaction('TC4H9O2 + C4H10 <=> TC4H9O2H + PC4H9', [1.700000e+13, 0.0, 20460.0])

## Reaction 1246
#reaction('IC3H7O2 + IC4H9 <=> IC3H7O + IC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1247
#reaction('IC3H7O2 + TC4H9 <=> IC3H7O + TC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1248
#reaction('IC3H7O2 + IC4H7 <=> IC3H7O + IC4H7O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1249
#reaction('IC4H9O2 + C3H6 <=> IC4H9O2H + C3H5-A', [5.350000e-02, 4.207, 13288.1])

## Reaction 1250
#reaction('TC4H9O2 + C3H6 <=> TC4H9O2H + C3H5-A', [5.350000e-02, 4.207, 13288.1])

## Reaction 1251
#reaction('IC4H9O2 + IC4H8 <=> IC4H9O2H + IC4H7', [1.400000e+12, 0.0, 14900.0])

## Reaction 1252
#reaction('TC4H9O2 + IC4H8 <=> TC4H9O2H + IC4H7', [1.400000e+12, 0.0, 14900.0])

## Reaction 1253
#reaction('PC4H9O2 + IC4H8 <=> PC4H9O2H + IC4H7', [1.400000e+12, 0.0, 14900.0])

## Reaction 1254
#reaction('SC4H9O2 + IC4H8 <=> SC4H9O2H + IC4H7', [1.400000e+12, 0.0, 14900.0])

## Reaction 1255
#reaction('IC3H7O2 + IC4H8 <=> IC3H7O2H + IC4H7', [1.400000e+12, 0.0, 14900.0])

## Reaction 1256
#reaction('NC3H7O2 + IC4H8 <=> NC3H7O2H + IC4H7', [1.400000e+12, 0.0, 14900.0])

## Reaction 1257
#reaction('CC4H8O + OH => CH2O + C3H5-A + H2O', [5.000000e+12, 0.0, 0.0])

## Reaction 1258
#reaction('CC4H8O + H => CH2O + C3H5-A + H2', [3.510000e+07, 2.0, 5000.0])

## Reaction 1259
#reaction('CC4H8O + O => CH2O + C3H5-A + OH', [1.124000e+14, 0.0, 5200.0])

## Reaction 1260
#reaction('CC4H8O + HO2 => CH2O + C3H5-A + H2O2', [1.000000e+13, 0.0, 15000.0])

## Reaction 1261
#reaction('CC4H8O + CH3O2 => CH2O + C3H5-A + CH3O2H', [1.000000e+13, 0.0, 19000.0])

## Reaction 1262
#reaction('CC4H8O + CH3 => CH2O + C3H5-A + CH4', [2.000000e+11, 0.0, 10000.0])

## Reaction 1263
#reaction('C2H4 + TC4H9O2 <=> C2H3 + TC4H9O2H', [8.590000e+00, 3.754, 27132.0])

## Reaction 1264
#reaction('TC4H9O2 + CH4 <=> TC4H9O2H + CH3', [1.130000e+13, 0.0, 20460.0])

## Reaction 1265
#reaction('H2 + TC4H9O2 <=> H + TC4H9O2H', [3.010000e+13, 0.0, 26030.0])

## Reaction 1266
#reaction('TC4H9O2 + C2H6 <=> TC4H9O2H + C2H5', [1.700000e+13, 0.0, 20460.0])

## Reaction 1267
#reaction('TC4H9O2 + C3H8 <=> TC4H9O2H + IC3H7', [2.000000e+12, 0.0, 17000.0])

## Reaction 1268
#reaction('TC4H9O2 + C3H8 <=> TC4H9O2H + NC3H7', [1.700000e+13, 0.0, 20460.0])

## Reaction 1269
#reaction('TC4H9O2 + CH3OH <=> TC4H9O2H + CH2OH', [6.300000e+12, 0.0, 19360.0])

## Reaction 1270
#reaction('TC4H9O2 + C2H5OH <=> TC4H9O2H + PC2H4OH', [6.300000e+12, 0.0, 19360.0])

## Reaction 1271
#reaction('TC4H9O2 + C2H5OH <=> TC4H9O2H + SC2H4OH', [4.200000e+12, 0.0, 15000.0])

## Reaction 1272
#reaction('IC4H9O2 + CH3CHO <=> IC4H9O2H + CH3CO', [2.800000e+12, 0.0, 13600.0])

## Reaction 1273
#reaction('TC4H9O2 + CH3CHO <=> TC4H9O2H + CH3CO', [2.800000e+12, 0.0, 13600.0])

## Reaction 1274
#reaction('IC4H9O2 + C2H3CHO <=> IC4H9O2H + C2H3CO', [2.800000e+12, 0.0, 13600.0])

## Reaction 1275
#reaction('TC4H9O2 + C2H3CHO <=> TC4H9O2H + C2H3CO', [2.800000e+12, 0.0, 13600.0])

## Reaction 1276
#reaction('IC4H9O2 + C2H5CHO <=> IC4H9O2H + C2H5CO', [2.800000e+12, 0.0, 13600.0])

## Reaction 1277
#reaction('TC4H9O2 + C2H5CHO <=> TC4H9O2H + C2H5CO', [2.800000e+12, 0.0, 13600.0])

## Reaction 1278
#reaction('IC4H9O2 + HO2 <=> IC4H9O2H + O2', [1.750000e+10, 0.0, -3275.0])

## Reaction 1279
#reaction('TC4H9O2 + HO2 <=> TC4H9O2H + O2', [1.750000e+10, 0.0, -3275.0])

## Reaction 1280
#reaction('IC4H9O2 + H2O2 <=> IC4H9O2H + HO2', [2.400000e+12, 0.0, 10000.0])

## Reaction 1281
#reaction('TC4H9O2 + H2O2 <=> TC4H9O2H + HO2', [2.400000e+12, 0.0, 10000.0])

## Reaction 1282
#reaction('IC4H9O2 + CH2O <=> IC4H9O2H + HCO', [1.300000e+11, 0.0, 9000.0])

## Reaction 1283
#reaction('TC4H9O2 + CH2O <=> TC4H9O2H + HCO', [1.300000e+11, 0.0, 9000.0])

## Reaction 1284
#reaction('IC4H9O2 + CH3O2 => IC4H9O + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 1285
#reaction('TC4H9O2 + CH3O2 => TC4H9O + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 1286
#reaction('IC4H9O2 + C2H5O2 => IC4H9O + C2H5O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 1287
#reaction('TC4H9O2 + C2H5O2 => TC4H9O + C2H5O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 1288
#reaction('IC4H9O2 + CH3CO3 => IC4H9O + CH3CO2 + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 1289
#reaction('TC4H9O2 + CH3CO3 => TC4H9O + CH3CO2 + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 1290
#reaction('IC4H9O2 + IC4H9O2 => O2 + IC4H9O + IC4H9O', [1.400000e+16, -1.61, 1860.0])

## Reaction 1291
#reaction('IC4H9O2 + TC4H9O2 => IC4H9O + TC4H9O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 1292
#reaction('TC4H9O2 + TC4H9O2 => O2 + TC4H9O + TC4H9O', [1.400000e+16, -1.61, 1860.0])

## Reaction 1293
#reaction('IC4H9O2 + PC4H9O2 => IC4H9O + PC4H9O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 1294
#reaction('TC4H9O2 + PC4H9O2 => TC4H9O + PC4H9O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 1295
#reaction('IC4H9O2 + SC4H9O2 => IC4H9O + SC4H9O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 1296
#reaction('TC4H9O2 + SC4H9O2 => TC4H9O + SC4H9O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 1297
#reaction('IC4H9O2 + NC3H7O2 => IC4H9O + NC3H7O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 1298
#reaction('TC4H9O2 + NC3H7O2 => TC4H9O + NC3H7O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 1299
#reaction('IC4H9O2 + IC3H7O2 => IC4H9O + IC3H7O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 1300
#reaction('TC4H9O2 + IC3H7O2 => TC4H9O + IC3H7O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 1301
#reaction('IC4H9O2 + HO2 => IC4H9O + OH + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 1302
#reaction('TC4H9O2 + HO2 => TC4H9O + OH + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 1303
#reaction('IC4H9O2 + CH3 <=> IC4H9O + CH3O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1304
#reaction('IC4H9O2 + C2H5 <=> IC4H9O + C2H5O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1305
#reaction('IC4H9O2 + IC3H7 <=> IC4H9O + IC3H7O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1306
#reaction('IC4H9O2 + NC3H7 <=> IC4H9O + NC3H7O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1307
#reaction('IC4H9O2 + PC4H9 <=> IC4H9O + PC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1308
#reaction('IC4H9O2 + SC4H9 <=> IC4H9O + SC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1309
#reaction('IC4H9O2 + IC4H9 <=> IC4H9O + IC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1310
#reaction('IC4H9O2 + TC4H9 <=> IC4H9O + TC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1311
#reaction('IC4H9O2 + C3H5-A <=> IC4H9O + C3H5O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1312
#reaction('IC4H9O2 + C4H71-3 <=> IC4H9O + C4H71-O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1313
#reaction('IC4H9O2 + IC4H7 <=> IC4H9O + IC4H7O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1314
#reaction('TC4H9O2 + CH3 <=> TC4H9O + CH3O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1315
#reaction('TC4H9O2 + C2H5 <=> TC4H9O + C2H5O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1316
#reaction('TC4H9O2 + IC3H7 <=> TC4H9O + IC3H7O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1317
#reaction('TC4H9O2 + NC3H7 <=> TC4H9O + NC3H7O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1318
#reaction('TC4H9O2 + PC4H9 <=> TC4H9O + PC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1319
#reaction('TC4H9O2 + SC4H9 <=> TC4H9O + SC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1320
#reaction('TC4H9O2 + IC4H9 <=> TC4H9O + IC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1321
#reaction('TC4H9O2 + TC4H9 <=> TC4H9O + TC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1322
#reaction('TC4H9O2 + C3H5-A <=> TC4H9O + C3H5O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1323
#reaction('TC4H9O2 + C4H71-3 <=> TC4H9O + C4H71-O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1324
#reaction('IC4H9O2 + C2H4 <=> IC4H9O2H + C2H3', [8.590000e+00, 3.754, 27132.0])

## Reaction 1325
#reaction('IC4H9O2 + CH4 <=> IC4H9O2H + CH3', [1.130000e+13, 0.0, 20460.0])

## Reaction 1326
#reaction('H2 + IC4H9O2 <=> H + IC4H9O2H', [3.010000e+13, 0.0, 26030.0])

## Reaction 1327
#reaction('IC4H9O2 + C2H6 <=> IC4H9O2H + C2H5', [1.700000e+13, 0.0, 20460.0])

## Reaction 1328
#reaction('IC4H9O2 + C3H8 <=> IC4H9O2H + IC3H7', [2.000000e+12, 0.0, 17000.0])

## Reaction 1329
#reaction('IC4H9O2 + C3H8 <=> IC4H9O2H + NC3H7', [1.700000e+13, 0.0, 20460.0])

## Reaction 1330
#reaction('IC4H9O2 + CH3OH <=> IC4H9O2H + CH2OH', [6.300000e+12, 0.0, 19360.0])

## Reaction 1331
#reaction('IC4H9O2 + C2H5OH <=> IC4H9O2H + PC2H4OH', [6.300000e+12, 0.0, 19360.0])

## Reaction 1332
#reaction('IC4H9O2 + C2H5OH <=> IC4H9O2H + SC2H4OH', [4.200000e+12, 0.0, 15000.0])

## Reaction 1333
#reaction('IC4H9O2H <=> IC4H9O + OH', [1.500000e+16, 0.0, 42500.0])

## Reaction 1334
#reaction('TC4H9O2H <=> TC4H9O + OH', [5.950000e+15, 0.0, 42540.0])

## Reaction 1335
#reaction('IC4H9O + HO2 <=> IC3H7CHO + H2O2', [1.000000e+12, 0.0, 0.0])

## Reaction 1336
#reaction('IC4H9O + OH <=> IC3H7CHO + H2O', [1.810000e+13, 0.0, 0.0])

## Reaction 1337
#reaction('IC4H9O + CH3 <=> IC3H7CHO + CH4', [2.400000e+13, 0.0, 0.0])

## Reaction 1338
#reaction('IC4H9O + O <=> IC3H7CHO + OH', [6.000000e+12, 0.0, 0.0])

## Reaction 1339
#reaction('IC4H9O + H <=> IC3H7CHO + H2', [1.990000e+13, 0.0, 0.0])

## Reaction 1340
#reaction('IC3H7CHO + H <=> IC4H9O', [1.000000e+12, 0.0, 5860.0])

## Reaction 1341
#reaction('CH2O + IC3H7 <=> IC4H9O', [5.000000e+10, 0.0, 2330.0])

## Reaction 1342
#reaction('CH3COCH3 + CH3 <=> TC4H9O', [1.500000e+11, 0.0, 11900.0])

## Reaction 1343
#reaction('IC4H9O + O2 <=> IC3H7CHO + HO2', [1.930000e+11, 0.0, 1660.0])

## Reaction 1344
#reaction('TC4H9O + O2 <=> IC4H8O + HO2', [8.100000e+11, 0.0, 4700.0])

## Reaction 1345
#reaction('IC4H8O <=> IC3H7CHO', [4.180000e+13, 0.0, 52720.0])

## Reaction 1346
#reaction('IC4H8O + OH <=> IC3H6CHO + H2O', [7.520000e+04, 2.5, -1474.1])

## Reaction 1347
#reaction('IC4H8O + H <=> IC3H6CHO + H2', [8.790000e+04, 2.7, 2910.0])

## Reaction 1348
#reaction('IC4H8O + HO2 <=> IC3H6CHO + H2O2', [2.450000e-05, 5.3, 7475.1])

## Reaction 1349
#reaction('IC4H8O + CH3O2 <=> IC3H6CHO + CH3O2H', [1.220000e-05, 5.3, 7475.1])

## Reaction 1350
#reaction('IC4H8O + CH3 <=> IC3H6CHO + CH4', [1.990000e+01, 3.4, 7634.0])

## Reaction 1351
#reaction('IC4H8O + O <=> IC3H6CHO + OH', [1.450000e+05, 2.5, 876.0])

## Reaction 1352
#reaction('TC3H6CHO + H <=> IC3H7CHO', [2.000000e+14, 0.0, 0.0])

## Reaction 1353
#reaction('IC3H7 + HCO <=> IC3H7CHO', [1.810000e+13, 0.0, 0.0])

## Reaction 1354
#reaction('IC3H7CHO + HO2 <=> IC3H7CO + H2O2', [3.000000e+12, 0.0, 11920.0])

## Reaction 1355
#reaction('IC3H7CHO + HO2 <=> TC3H6CHO + H2O2', [8.000000e+10, 0.0, 11920.0])

## Reaction 1356
#reaction('IC3H7CHO + CH3 <=> IC3H7CO + CH4', [3.980000e+12, 0.0, 8700.0])

## Reaction 1357
#reaction('IC3H7CHO + O <=> IC3H7CO + OH', [7.180000e+12, 0.0, 1389.0])

## Reaction 1358
#reaction('IC3H7CHO + O2 <=> IC3H7CO + HO2', [4.000000e+13, 0.0, 37600.0])

## Reaction 1359
#reaction('IC3H7CHO + OH <=> IC3H7CO + H2O', [2.690000e+10, 0.76, -340.0])

## Reaction 1360
#reaction('IC3H7CHO + OH <=> TC3H6CHO + H2O', [1.684000e+12, 0.0, -781.0])

## Reaction 1361
#reaction('IC3H7CHO + H <=> IC3H7CO + H2', [2.600000e+12, 0.0, 2600.0])

## Reaction 1362
#reaction('IC3H7CHO + OH <=> IC3H6CHO + H2O', [3.120000e+06, 2.0, -298.0])

## Reaction 1363
#reaction('IC3H7CHO + HO2 <=> IC3H6CHO + H2O2', [2.740000e+04, 2.55, 15500.0])

## Reaction 1364
#reaction('IC3H7CHO + CH3O2 <=> IC3H6CHO + CH3O2H', [4.760000e+04, 2.55, 16490.0])

## Reaction 1365
#reaction('IC3H7 + CO <=> IC3H7CO', [1.500000e+11, 0.0, 4810.0])

## Reaction 1366
#reaction('C2H3CHO + CH3 <=> IC3H6CHO', [1.000000e+11, 0.0, 7800.0])

## Reaction 1367
#reaction('IC4H8OOH-TO2 <=> IC4KETIT + OH', [4.000000e+11, 0.0, 31500.0])

## Reaction 1368
#reaction('IC4H8OOH-TO2 <=> TIC4H7Q2-I', [6.000000e+11, 0.0, 34500.0])

## Reaction 1369
#reaction('CH2O2H <=> CH2O + OH', [9.000000e+14, 0.0, 1500.0])

## Reaction 1370
#reaction('TC3H6CHO + CH2O <=> IC3H7CHO + HCO', [2.520000e+08, 1.9, 18190.0])

## Reaction 1371
#reaction('TC3H6CHO + IC4H8 <=> IC3H7CHO + IC4H7', [4.700000e+02, 3.3, 19840.0])

## Reaction 1372
#reaction('IC3H6CO + OH <=> IC3H7 + CO2', [1.730000e+12, 0.0, -1010.0])

## Reaction 1373
#reaction('TC3H6CHO + OH <=> TC3H6OHCHO', [5.000000e+13, 0.0, 0.0])

## Reaction 1374
#reaction('C3H6OH2-1 + HCO <=> TC3H6OHCHO', [1.810000e+13, 0.0, 0.0])

## Reaction 1375
#reaction('CH3COCH3 + H <=> C3H6OH2-1', [8.000000e+12, 0.0, 9500.0])

## Reaction 1376
#reaction('C2H5CHO + H <=> C3H6OH1-2', [8.000000e+12, 0.0, 9500.0])

## Reaction 1377
#reaction('IC3H5OH + H <=> C3H6OH2-1', [6.250000e+11, 0.51, 4020.0])

## Reaction 1378
#reaction('C3H5-T + OH <=> IC3H5OH', [5.000000e+13, 0.0, 0.0])

## Reaction 1379
#reaction('TC3H6CHO + O2 <=> TC3H6O2CHO', [1.990000e+17, -2.1, 0.0])

## Reaction 1380
#reaction('TC3H6O2CHO <=> IC3H5O2HCHO', [6.000000e+11, 0.0, 29880.0])

## Reaction 1381
#reaction('TC3H6O2CHO <=> TC3H6O2HCO', [1.000000e+11, 0.0, 25750.0])

## Reaction 1382
#reaction('IC3H5CHO + HO2 <=> IC3H5O2HCHO', [2.230000e+11, 0.0, 10600.0])

## Reaction 1383
#reaction('C3H6OH2-1 + O2 <=> CH3COCH3 + HO2', [1.500000e+12, 0.0, 5000.0])

## Reaction 1384
#reaction('C3H6OH1-2 + O2 <=> C2H5CHO + HO2', [1.500000e+12, 0.0, 5000.0])

## Reaction 1385
#reaction('IC3H6CO + OH <=> C3H6OH2-1 + CO', [2.000000e+12, 0.0, -1010.0])

## Reaction 1386
#reaction('TC3H6CHO + HO2 <=> IC3H7CHO + O2', [3.675000e+12, 0.0, 1310.0])

## Reaction 1387
#reaction('TC3H6CHO + CH3 <=> IC3H5CHO + CH4', [3.010000e+12, -0.32, -131.0])

## Reaction 1388
#reaction('TC4H8CHO <=> IC3H5CHO + CH3', [1.000000e+13, 0.0, 26290.0])

## Reaction 1389
#reaction('TC4H8CHO <=> IC4H8 + HCO', [8.520000e+12, 0.0, 20090.0])

## Reaction 1390
#reaction('TC4H8CHO + O2 <=> O2C4H8CHO', [2.000000e+12, 0.0, 0.0])

## Reaction 1391
#reaction('O2C4H8CHO <=> O2HC4H8CO', [2.160000e+11, 0.0, 15360.0])

## Reaction 1392
#reaction('IC4H8O2H-T + CO <=> O2HC4H8CO', [1.500000e+11, 0.0, 4809.0])

## Reaction 1393
#reaction('IC4H7O + IC4H8 <=> IC4H7OH + IC4H7', [2.700000e+11, 0.0, 4000.0])

## Reaction 1394
#reaction('IC4H6OH + HO2 => CH2CCH2OH + CH2O + OH', [1.446000e+13, 0.0, 0.0])

## Reaction 1395
#reaction('IC4H8 + CH2CCH2OH <=> IC4H7 + C3H5OH', [7.940000e+11, 0.0, 20500.0])

## Reaction 1396
#reaction('CH2CCH2OH + H2O2 <=> C3H5OH + HO2', [3.010000e+09, 0.0, 2583.0])

## Reaction 1397
#reaction('C3H5OH + OH <=> CH2CCH2OH + H2O', [5.060000e+12, 0.0, 5960.0])

## Reaction 1398
#reaction('C3H5OH + H <=> CH2CCH2OH + H2', [3.900000e+05, 2.5, 5821.0])

## Reaction 1399
#reaction('C3H5OH + O2 <=> CH2CCH2OH + HO2', [4.000000e+13, 0.0, 60690.0])

## Reaction 1400
#reaction('C3H5OH + CH3 <=> CH2CCH2OH + CH4', [2.400000e+11, 0.0, 8030.0])

## Reaction 1401
#reaction('CH2CCH2OH + CH3 <=> IC4H7OH', [3.000000e+13, 0.0, 0.0])

## Reaction 1402
#reaction('CH2CCH2OH + H <=> C3H5OH', [1.000000e+14, 0.0, 0.0])

## Reaction 1403
#reaction('CH2CCH2OH + O2 => CH2OH + CO + CH2O', [4.335000e+12, 0.0, 0.0])

## Reaction 1404
#reaction('CH2CCH2OH <=> C2H2 + CH2OH', [2.163000e+40, -8.31, 45110.0])

## Reaction 1405
#reaction('C3H5-A + C2H2 <=> C*CC*CCJ', [1.000000e+12, 0.0, 6883.4])

## Reaction 1406
#reaction('C3H5-A + C2H3 => C5H6 + H + H', [1.600000e+35, -14.0, 61137.7])

## Reaction 1407
#reaction('C3H4-A + C3H3 <=> C6H6 + H', [1.400000e+12, 0.0, 9990.4])

## Reaction 1408
#reaction('C4H6 + C2H3 => C6H6 + H2 + H', [5.620000e+11, 0.0, 3240.0])

## Reaction 1409
#reaction('C4H5-N + C2H2 <=> C6H6 + H', [1.600000e+16, -1.33, 5400.0])

## Reaction 1410
#reaction('C4H5-N + C2H3 <=> C6H6 + H2', [1.840000e-13, 7.07, -3611.0])

## Reaction 1411
#reaction('C4H5-2 + C2H2 <=> C6H6 + H', [5.000000e+14, 0.0, 25000.0])

## Reaction 1412
#reaction('C4H5-2 + C2H4 <=> C5H6 + CH3', [5.000000e+14, 0.0, 25000.0])

## Reaction 1413
#reaction('C6H2 + H <=> C6H3', [1.100000e+30, -4.92, 10800.0])

## Reaction 1414
#reaction('C6H3 + H <=> C4H2 + C2H2', [2.800000e+23, -2.55, 10780.0])

## Reaction 1415
#reaction('C6H3 + H <=> L-C6H4', [3.400000e+43, -9.01, 12120.0])

## Reaction 1416
#reaction('C6H3 + H <=> C6H2 + H2', [3.000000e+13, 0.0, 0.0])

## Reaction 1417
#reaction('C6H3 + OH <=> C6H2 + H2O', [4.000000e+12, 0.0, 0.0])

## Reaction 1418
#reaction('L-C6H4 + H <=> C6H5', [1.700000e+78, -19.72, 31400.0])

## Reaction 1419
#reaction('L-C6H4 + H <=> C-C6H4 + H', [1.400000e+54, -11.7, 34500.0])

## Reaction 1420
#reaction('L-C6H4 + H <=> C6H3 + H2', [1.330000e+06, 2.53, 9240.0])

## Reaction 1421
#reaction('L-C6H4 + OH <=> C6H3 + H2O', [3.100000e+06, 2.0, 430.0])

## Reaction 1422
#reaction('C-C6H4 + H <=> C6H5', [2.400000e+60, -13.66, 29500.0])

## Reaction 1423
#reaction('C4H3-N + C2H2 <=> L-C6H4 + H', [2.500000e+14, -0.56, 10600.0])

## Reaction 1424
#reaction('C4H3-N + C2H2 <=> C6H5', [9.600000e+70, -17.77, 31300.0])

## Reaction 1425
#reaction('C4H3-N + C2H2 <=> C-C6H4 + H', [6.900000e+46, -10.01, 30100.0])

## Reaction 1426
#reaction('C4H3-I + CH3 <=> C5H6', [1.000000e+12, 0.0, 0.0])

## Reaction 1427
#reaction('C4H2 + C2H <=> C6H2 + H', [9.600000e+13, 0.0, 0.0])

## Reaction 1428
#reaction('C4H2 + C2H <=> C6H3', [4.500000e+37, -7.68, 7100.0])

## Reaction 1429
#reaction('C4H4 + C2H <=> L-C6H4 + H', [1.200000e+13, 0.0, 0.0])

## Reaction 1430
#falloff_reaction('C6H5 + H (+ M) <=> C6H6 (+ M)',
                #kf=[1.000000e+14, 0.0, 0.0],
                #kf0=[6.600000e+75, -16.3, 7000.0],
                #efficiencies='H2:2.0 H2O:6.0 CH4:2.0 CO:1.5 CO2:2.0',
                #falloff=Troe(A=1.0, T3=0.1, T1=584.9, T2=6113.0))

## Reaction 1431
#reaction('C6H6 + O2 <=> C6H5 + HO2', [6.300000e+13, 0.0, 60000.0])

## Reaction 1432
#reaction('C6H6 + O <=> C6H5O + H', [2.200000e+13, 0.0, 4530.0])

## Reaction 1433
#reaction('C6H6 + O <=> C6H5 + OH', [2.000000e+13, 0.0, 14700.0])

## Reaction 1434
#reaction('C6H6 + H <=> C6H5 + H2', [2.500000e+14, 0.0, 16000.0])

## Reaction 1435
#reaction('C6H6 + CH3 <=> C6H5 + CH4', [7.320000e+12, 0.0, 18920.0])

## Reaction 1436
#reaction('C6H6 + HO2 <=> C6H5 + H2O2', [5.500000e+12, 0.0, 28900.0])

## Reaction 1437
#reaction('C6H6 + OH <=> C6H5 + H2O', [1.200000e+00, 4.1, -301.0])

## Reaction 1438
#reaction('C6H6 + OH <=> C6H5OH + H', [1.320000e+02, 3.25, 5590.0])

## Reaction 1439
#reaction('C6H5 => H + C4H2 + C2H2', [4.300000e+12, 0.62, 77294.0])

## Reaction 1440
#reaction('C6H5 + CH2O <=> C6H6 + HCO', [8.550000e+04, 2.19, 38.0])

## Reaction 1441
#reaction('C6H5 + HCO <=> C6H6 + CO', [8.550000e+04, 2.19, 38.0])

## Reaction 1442
#reaction('C6H5 + HO2 <=> C6H5O + OH', [5.000000e+12, 0.0, 0.0])

## Reaction 1443
#reaction('C6H5 + O2 <=> C6H5O + O', [2.600000e+13, 0.0, 6120.0])

## Reaction 1444
#reaction('C6H5 + O2 <=> O-C6H4O2 + H', [3.000000e+13, 0.0, 8980.0])

## Reaction 1445
#reaction('C6H5 + O2 <=> C6H5OO', [1.860000e+13, -0.22, -711.0])

## Reaction 1446
#reaction('C6H5OO <=> C6H5O + O', [1.270000e+15, -0.246, 38536.0])

## Reaction 1447
#reaction('C6H5O + OH <=> C6H5OOH', [2.000000e+13, 0.0, 0.0])

## Reaction 1448
#reaction('C6H5OO + C6H5OH <=> C6H5OOH + C6H5O', [1.330000e+11, 0.0, 14000.0])

## Reaction 1449
#reaction('C6H5OH + O2 <=> C6H5O + HO2', [1.000000e+13, 0.0, 38800.0])

## Reaction 1450
#reaction('C6H5OH + O <=> OC6H4OH + H', [1.600000e+13, 0.0, 3400.0])

## Reaction 1451
#reaction('C6H5OH + H <=> C6H5O + H2', [1.200000e+14, 0.0, 12400.0])

## Reaction 1452
#reaction('C6H5OH + O <=> C6H5O + OH', [1.300000e+13, 0.0, 2900.0])

## Reaction 1453
#reaction('C6H5OH + OH <=> C6H5O + H2O', [1.400000e+08, 1.4, -960.0])

## Reaction 1454
#reaction('C6H5OH + HO2 <=> C6H5O + H2O2', [1.000000e+12, 0.0, 10000.0])

## Reaction 1455
#reaction('C6H5OH + CH3 <=> C6H5O + CH4', [1.800000e+11, 0.0, 7700.0])

## Reaction 1456
#reaction('C6H5OH + C6H5 <=> C6H5O + C6H6', [4.900000e+12, 0.0, 4400.0])

## Reaction 1457
#reaction('C6H5OH + C3H5-A <=> C6H5O + C3H6', [4.900000e+11, 0.0, 9400.0])

## Reaction 1458
#reaction('C6H5OH + C4H5-I <=> C6H5O + C4H6', [4.900000e+11, 0.0, 9400.0])

## Reaction 1459
#reaction('C6H5OH + H <=> C6H4OH + H2', [1.700000e+14, 0.0, 16000.0])

## Reaction 1460
#reaction('C6H5OH + O <=> C6H4OH + OH', [2.000000e+13, 0.0, 14700.0])

## Reaction 1461
#reaction('C6H5OH + OH <=> C6H4OH + H2O', [1.400000e+13, 0.0, 4600.0])

## Reaction 1462
#reaction('C6H5OH + HO2 <=> C6H4OH + H2O2', [4.000000e+11, 0.0, 28900.0])

## Reaction 1463
#reaction('C6H5OH + CH3 <=> C6H4OH + CH4', [2.000000e+12, 0.0, 15000.0])

## Reaction 1464
#reaction('C6H4OH + O2 <=> OC6H4OH + O', [2.100000e+13, 0.0, 6100.0])

## Reaction 1465
#reaction('C6H4OH + H <=> C6H5OH', [1.000000e+14, 0.0, 0.0])

## Reaction 1466
#reaction('OC6H4OH <=> C5H4OH + CO', [7.400000e+11, 0.0, 43800.0])

## Reaction 1467
#falloff_reaction('C6H5O + H (+ M) <=> C6H5OH (+ M)',
                #kf=[2.000000e+14, 0.0, 0.0],
                #kf0=[1.000000e+94, -21.84, 13880.0],
                #falloff=Troe(A=0.043, T3=304.2, T1=60000.0, T2=5896.0))

## Reaction 1468
#reaction('C6H5OH <=> C5H6 + CO', [4.310000e+15, -0.61, 74115.0])

## Reaction 1469
#reaction('C6H5O <=> CO + C5H5', [2.000000e+11, 0.0, 43900.0])

## Reaction 1470
#reaction('C6H5O + H <=> CO + C5H6', [1.000000e+13, 0.0, 0.0])

## Reaction 1471
#reaction('C6H5O + O <=> C5H5 + CO2', [1.000000e+13, 0.0, 0.0])

## Reaction 1472
#reaction('C6H5O + O <=> OC6H4OH', [2.600000e+10, 0.47, 800.0])

## Reaction 1473
#reaction('C6H5O + HO2 <=> O-OC6H5OJ + OH', [2.000000e+12, 0.0, 0.0])

## Reaction 1474
#reaction('P-C6H4O2 + H <=> P-OC6H5OJ', [4.000000e+12, 0.0, 9740.0])

## Reaction 1475
#reaction('O-C6H4O2 + H <=> O-OC6H5OJ', [4.000000e+12, 0.0, 6960.0])

## Reaction 1476
#reaction('C6H5O + O <=> P-C6H4O2 + H', [4.250000e+13, 0.0, 0.0])

## Reaction 1477
#reaction('C6H5O + O <=> O-C6H4O2 + H', [8.500000e+13, 0.0, 0.0])

## Reaction 1478
#reaction('O-C6H4O2 <=> C5H4O + CO', [1.000000e+12, 0.0, 40000.0])

## Reaction 1479
#reaction('P-C6H4O2 <=> C5H4O + CO', [3.700000e+11, 0.0, 59000.0])

## Reaction 1480
#reaction('P-C6H4O2 + H <=> C5H5O + CO', [2.500000e+13, 0.0, 4700.0])

## Reaction 1481
#reaction('P-C6H4O2 + H <=> P-C6H3O2 + H2', [2.000000e+12, 0.0, 8100.0])

## Reaction 1482
#reaction('P-C6H4O2 + O <=> P-C6H3O2 + OH', [1.400000e+13, 0.0, 14700.0])

## Reaction 1483
#reaction('P-C6H4O2 + OH <=> P-C6H3O2 + H2O', [1.000000e+06, 2.0, 4000.0])

## Reaction 1484
#reaction('P-C6H3O2 + H <=> P-C6H4O2', [1.000000e+14, 0.0, 0.0])

## Reaction 1485
#reaction('P-C6H3O2 + H <=> 2 C2H2 + 2 CO', [1.000000e+14, 0.0, 0.0])

## Reaction 1486
#reaction('P-C6H3O2 + O => C2H2 + HCCO + 2 CO', [1.000000e+14, 0.0, 0.0])

## Reaction 1487
#reaction('P-C6H4O2 + O => 2 CO + C2H2 + CH2CO', [3.000000e+13, 0.0, 5000.0])

## Reaction 1488
#falloff_reaction('C5H5 + H (+ M) <=> C5H6 (+ M)',
                #kf=[2.600000e+14, 0.0, 0.0],
                #kf0=[4.400000e+80, -18.28, 12994.0],
                #efficiencies='CO:1.5 CO2:2.0 H2O:6.0 H2:2.0 CH4:2.0',
                #falloff=Troe(A=0.068, T3=400.7, T1=4136.0, T2=5502.0))

## Reaction 1489
#falloff_reaction('C5H6 (+ M) <=> C3H4-A + C2H2 (+ M)',
                #kf=[3.800000e+17, 0.0, 104000.0],
                #kf0=[1.000000e+98, -22.25, 126321.5],
                #efficiencies='CO:1.5 CO2:2.0 H2O:6.0 H2:2.0 CH4:2.0',
                #falloff=Troe(A=0.1441, T3=5.358, T1=3284.0, T2=6710000000.0))

## Reaction 1490
#reaction('C5H6 + O2 <=> C5H5 + HO2', [4.000000e+13, 0.0, 37150.0])

## Reaction 1491
#reaction('C5H6 + HO2 <=> C5H5 + H2O2', [1.100000e+04, 2.6, 12900.0])

## Reaction 1492
#reaction('C5H6 + OH <=> C5H5 + H2O', [3.080000e+06, 2.0, 0.0])

## Reaction 1493
#reaction('C5H6 + H <=> C5H5 + H2', [7.200000e+13, 0.0, 3500.0])

## Reaction 1494
#reaction('C5H6 + H <=> C2H2 + C3H5-A', [1.548000e+37, -6.18, 32890.0])

## Reaction 1495
#reaction('C5H6 + O <=> C5H5 + OH', [4.800000e+04, 2.71, 1100.0])

## Reaction 1496
#reaction('C5H6 + C2H3 <=> C5H5 + C2H4', [1.200000e-01, 4.0, 0.0])

## Reaction 1497
#reaction('C5H6 + C6H5O <=> C5H5 + C6H5OH', [3.160000e+11, 0.0, 8000.0])

## Reaction 1498
#reaction('C5H6 + CH3 <=> C5H5 + CH4', [1.800000e-01, 4.0, 0.0])

## Reaction 1499
#reaction('C5H6 + C6H5 <=> C5H5 + C6H6', [1.000000e-01, 4.0, 0.0])

## Reaction 1500
#reaction('C5H6 + C5H5 <=> C6H6 + C4H5-N', [5.000000e+09, 0.0, 0.0])

## Reaction 1501
#reaction('C5H5 <=> C3H3 + C2H2', [1.980000e+68, -15.0, 124900.0])

## Reaction 1502
#reaction('C5H5 + O <=> C4H5-N + CO', [3.200000e+13, -0.17, 440.0])

## Reaction 1503
#reaction('C5H5 + O <=> C5H4O + H', [5.800000e+13, -0.02, 20.0])

## Reaction 1504
#reaction('C5H5 + OH <=> C5H5OH', [6.500000e+14, -0.85, -2730.0],
        #options='duplicate')

## Reaction 1505
#reaction('C5H5 + OH <=> C5H5OH', [1.100000e+43, -8.76, 18730.0],
        #options='duplicate')

## Reaction 1506
#reaction('C5H5 + OH <=> C5H5OH', [1.100000e+59, -13.08, 33450.0],
        #options='duplicate')

## Reaction 1507
#reaction('C5H5 + OH <=> C5H4OH + H', [3.500000e+57, -12.18, 48350.0])

## Reaction 1508
#reaction('C5H5 + OH <=> C4H6 + CO', [4.000000e+14, 0.0, 4500.0])

## Reaction 1509
#reaction('C5H5 + HO2 <=> C5H5O + OH', [6.300000e+29, -4.69, 11650.0])

## Reaction 1510
#reaction('C5H5 + O2 <=> C4H4O + HCO', [6.000000e+18, -2.48, 10970.0])

## Reaction 1511
#reaction('C4H4O + H <=> C3H5-A + CO', [6.600000e+13, -0.02, 2740.0],
        #options='duplicate')

## Reaction 1512
#reaction('C4H4O + H <=> C3H5-A + CO', [5.900000e+06, 2.0, 1300.0],
        #options='duplicate')

## Reaction 1513
#reaction('C4H4O + O <=> CH2CHO + HCCO', [3.000000e+08, 1.45, -860.0])

## Reaction 1514
#reaction('C4H4O + OH <=> C3H5-A + CO2', [3.000000e+12, 0.0, 0.0])

## Reaction 1515
#reaction('C3H3 + C2H2 <=> C#CC*CCJ', [4.110000e+72, -18.2, 45400.0])

## Reaction 1516
#reaction('C#CC*CCJ <=> C5H5', [8.000000e+13, 0.0, 34058.3])

## Reaction 1517
#reaction('C#CC*CCJ + H <=> C5H6', [1.000000e+14, 0.0, 0.0])

## Reaction 1518
#reaction('C#CC*CCJ + H <=> C5H6-L', [1.000000e+10, 0.0, 0.0])

## Reaction 1519
#reaction('C5H6-L + O <=> C#CC*CCJ + OH', [1.000000e+10, 0.0, 0.0])

## Reaction 1520
#reaction('C5H6-L + OH <=> C#CC*CCJ + H2O', [1.000000e+10, 0.0, 0.0])

## Reaction 1521
#reaction('C5H5OH + H <=> C5H4OH + H2', [3.200000e+12, 0.0, 0.0])

## Reaction 1522
#reaction('C5H5OH + H <=> C5H5O + H2', [4.000000e+13, 0.0, 6094.0])

## Reaction 1523
#reaction('C5H5OH + OH <=> C5H4OH + H2O', [5.500000e+12, 0.0, 1731.0])

## Reaction 1524
#reaction('C5H5OH + OH <=> C5H5O + H2O', [1.000000e+13, 0.0, 1697.0])

## Reaction 1525
#reaction('C5H5OH + O <=> C5H4OH + OH', [4.700000e+11, 0.0, 0.0])

## Reaction 1526
#reaction('C5H5OH + O <=> C5H5O + OH', [1.000000e+13, 0.0, 4683.0])

## Reaction 1527
#reaction('C5H5OH + HO2 <=> C5H4OH + H2O2', [3.600000e+03, 2.55, 10531.0])

## Reaction 1528
#reaction('C5H5OH + HO2 <=> C5H5O + H2O2', [1.000000e+13, 0.0, 15800.0])

## Reaction 1529
#reaction('C5H5O <=> C5H4O + H', [2.900000e+32, -6.5, 21220.0])

## Reaction 1530
#reaction('C5H5O <=> C4H5-N + CO', [1.100000e+79, -19.62, 66250.0])

## Reaction 1531
#reaction('C5H5O <=> CJ*CC*CC*O', [2.000000e+13, 0.0, 14338.0])

## Reaction 1532
#reaction('CJ*CC*CC*O <=> C*CC*CCJ*O', [4.300000e+11, -1.05, 4118.0])

## Reaction 1533
#reaction('C4H5-N + CO <=> C*CC*CCJ*O', [1.510000e+11, 0.0, 4810.0])

## Reaction 1534
#reaction('CJ*CC*CC*O <=> C2H2 + CJ*CC*O', [3.000000e+13, 0.0, 43710.0])

## Reaction 1535
#reaction('CJ*CC*O <=> C2H3CO', [1.400000e+09, 0.98, 32100.0])

## Reaction 1536
#reaction('C2H2 + HCO <=> CJ*CC*O', [7.770000e+06, 1.41, 7755.0])

## Reaction 1537
#reaction('C5H4OH + H <=> C5H5OH', [1.000000e+14, 0.0, 0.0])

## Reaction 1538
#reaction('C5H4OH + H <=> C5H4O + H2', [2.100000e+13, 0.0, 54000.0])

## Reaction 1539
#reaction('C5H4OH + O2 <=> C5H4O + HO2', [3.000000e+13, 0.0, 5000.0])

## Reaction 1540
#reaction('C5H4O => CO + C2H2 + C2H2', [5.700000e+32, -6.76, 68500.0],
        #options='duplicate')

## Reaction 1541
#reaction('C5H4O => CO + C2H2 + C2H2', [6.200000e+41, -7.87, 98700.0],
        #options='duplicate')

## Reaction 1542
#reaction('C5H4O + H <=> C4H5-N + CO', [2.100000e+61, -13.27, 40810.0])

## Reaction 1543
#reaction('C5H4O + O <=> C4H4 + CO2', [1.000000e+13, 0.0, 2000.0])

## Reaction 1544
#reaction('C5H4O + H <=> C5H3O + H2', [2.000000e+12, 0.0, 8100.0])

## Reaction 1545
#reaction('C5H4O + O <=> C5H3O + OH', [1.400000e+13, 0.0, 1470.0])

## Reaction 1546
#reaction('C5H4O + OH <=> C5H3O + H2O', [1.100000e+08, 1.42, 1450.0])

## Reaction 1547
#reaction('C5H3O + H <=> C5H4O', [1.000000e+14, 0.0, 0.0])

## Reaction 1548
#reaction('C5H3O + O2 => CO2 + C2H2 + HCCO', [9.700000e+58, -13.47, 38180.0])

## Reaction 1549
#reaction('C5H3O => C2H2 + CO + C2H', [2.000000e+13, 0.0, 51000.0])

## Reaction 1550
#reaction('C5H7 <=> C*CCJC*C', [3.200000e+15, 0.0, 39500.0])

## Reaction 1551
#reaction('C5H7 + H <=> C5H6 + H2', [3.600000e+12, 0.0, 0.0])

## Reaction 1552
#reaction('C5H7 + O <=> C5H6 + OH', [1.000000e+13, 0.0, 0.0])

## Reaction 1553
#reaction('C5H7 + OH <=> C5H6 + H2O', [2.400000e+13, 0.0, 0.0])

## Reaction 1554
#reaction('C5H7 + O2 <=> OC5H7O', [8.900000e+24, -3.8, 20000.0])

## Reaction 1555
#reaction('C5H6 + H <=> C5H7', [2.400000e+73, -17.85, 31500.0])

## Reaction 1556
#reaction('C5H6 + H <=> C*CCJC*C', [1.100000e+14, -0.16, 3100.0])

## Reaction 1557
#reaction('C5H6 + O <=> C5H5O + H', [8.900000e+12, -0.15, 590.0],
        #options='duplicate')

## Reaction 1558
#reaction('C5H6 + O <=> C5H5O + H', [5.600000e+12, -0.06, 200.0],
        #options='duplicate')

## Reaction 1559
#reaction('C5H6 + OH <=> C*CCJC*COH', [1.100000e+13, -0.07, 870.0])

## Reaction 1560
#reaction('C5H6 + HO2 <=> C5H7 + O2', [1.300000e+15, -1.07, 9530.0])

## Reaction 1561
#reaction('C5H6 + HCO <=> C5H5 + CH2O', [1.080000e+08, 1.9, 16000.0])

## Reaction 1562
#reaction('C5H6 + C2H3 <=> C6H6 + CH3', [2.100000e+67, -16.08, 42460.0])

## Reaction 1563
#reaction('C5H6 + C4H5-I <=> C5H5 + C4H6', [6.000000e+12, 0.0, 0.0])

## Reaction 1564
#reaction('C*CCJC*C <=> C*CC*CCJ', [5.400000e+11, -0.7, 60.0])

## Reaction 1565
#reaction('C*CC*CCJ + H <=> C5H81-3', [2.300000e+20, -1.6, 3020.0])

## Reaction 1566
#reaction('C5H81-3 + H <=> C4H6 + CH3', [5.200000e+71, -16.38, 51000.0])

## Reaction 1567
#reaction('C5H81-3 + H <=> C*CC*CCJ + H2', [7.000000e+06, 2.0, 5000.0])

## Reaction 1568
#reaction('C5H81-3 + OH <=> C*CC*CCJ + H2O', [7.000000e+06, 2.0, 0.0])

## Reaction 1569
#reaction('C*CCJC*C + O2 <=> C2H3CHO + CH2CHO', [1.200000e+36, -7.25, 33600.0])

## Reaction 1570
#reaction('C*CC*CCJ + H <=> C4H5-N + CH3', [2.900000e+26, -2.18, 36770.0])

## Reaction 1571
#reaction('C*CCJC*C + O <=> C2H3CHO + C2H3', [2.000000e+14, 0.0, 0.0])

## Reaction 1572
#reaction('C*CC*CCJ + OH <=> C*CC*CCOH', [1.500000e+13, 0.0, 0.0])

## Reaction 1573
#reaction('C*CC*CCJ + O2 <=> CH2CHCHCHO + CH2O', [8.200000e+10, 0.18, 9140.0])

## Reaction 1574
#reaction('C*CC*CCOH + H <=> C4H6 + CH2OH', [2.500000e+34, -6.12, 16250.0])

## Reaction 1575
#reaction('OC5H7O + O2 <=> OC4H6O + HOCO', [6.300000e+05, -7.25, 33600.0])

## Reaction 1576
#reaction('C*CCJC*COH + O2 <=> HOC*CC*O + CH2CHO', [1.200000e+36, -7.25, 33600.0])

## Reaction 1577
#reaction('HOC*CC*O + OH <=> HOC*CCJ*O + H2O', [2.690000e+10, 0.76, -340.0])

## Reaction 1578
#reaction('HOC*CC*O + HO2 <=> HOC*CCJ*O + H2O2', [1.000000e+12, 0.0, 11920.0])

## Reaction 1579
#reaction('HOC*CC*O + CH3 <=> HOC*CCJ*O + CH4', [3.980000e+12, 0.0, 8700.0])

## Reaction 1580
#reaction('HOC*CC*O + O <=> HOC*CCJ*O + OH', [7.180000e+12, 0.0, 1389.0])

## Reaction 1581
#reaction('HOC*CC*O + O2 <=> HOC*CCJ*O + HO2', [2.000000e+13, 0.0, 40700.0])

## Reaction 1582
#reaction('HOC*CC*O + H <=> HOC*CCJ*O + H2', [2.600000e+12, 0.0, 2600.0])

## Reaction 1583
#reaction('C2H2OH + CO <=> HOC*CCJ*O', [1.510000e+11, 0.0, 4810.0])

## Reaction 1584
#reaction('OC4H6O + H <=> OC4H5O + H2', [2.300000e+10, 1.05, 3279.0])

## Reaction 1585
#reaction('OC4H6O + OH <=> OC4H5O + H2O', [3.500000e+09, 1.18, -447.0])

## Reaction 1586
#reaction('OC4H5O + O2 <=> O2CCHOOJ + C2H4', [1.600000e+45, -9.92, 20670.0])

## Reaction 1587
#reaction('O2CCHOOJ <=> HOCO + CO2', [3.000000e+13, 0.0, 4000.0])

## Reaction 1588
#reaction('CH2CHCHCHO <=> C3H5-A + CO', [6.100000e+05, 0.92, -1120.0])

## Reaction 1589
#reaction('CH2CHCHCHO + O2 <=> C2H3CHO + HOCO', [1.200000e+36, -7.25, 33600.0])

## Reaction 1590
#pdep_arrhenius('HOCO <=> CO + OH',
            #[(0.001, 'atm'), 1.550000e-08, 2.93, 8768.0],
            #[(0.003, 'atm'), 1.770000e+03, 0.34, 18076.0],
            #[(0.0296, 'atm'), 2.020000e+13, -1.87, 22755.0],
            #[(0.0987, 'atm'), 1.680000e+18, -3.05, 24323.0],
            #[(0.2961, 'atm'), 2.500000e+24, -4.63, 27067.0],
            #[(0.9869, 'atm'), 4.540000e+26, -5.12, 27572.0],
            #[(2.9607, 'atm'), 7.120000e+28, -5.6, 28535.0],
            #[(9.869, 'atm'), 5.480000e+29, -5.7, 28899.0],
            #[(29.607, 'atm'), 9.890000e+31, -6.19, 30518.0],
            #[(98.69, 'atm'), 5.740000e+33, -6.53, 32068.0],
            #[(296.07, 'atm'), 2.610000e+33, -6.29, 32231.0],
            #[(986.9, 'atm'), 6.300000e+32, -5.96, 32470.0])

## Reaction 1591
#pdep_arrhenius('HOCO <=> CO2 + H',
            #[(0.001, 'atm'), 4.758000e+18, -3.817, 17676.0],
            #[(0.003, 'atm'), 2.225000e+20, -4.149, 19037.0],
            #[(0.0099, 'atm'), 7.564000e+21, -4.434, 20325.0],
            #[(0.0296, 'atm'), 9.107000e+24, -5.189, 22419.0],
            #[(0.0987, 'atm'), 3.144000e+29, -6.376, 25233.0],
            #[(0.2961, 'atm'), 1.150000e+32, -7.037, 26662.0],
            #[(0.9869, 'atm'), 1.069000e+36, -8.107, 29064.0],
            #[(2.9607, 'atm'), 2.438000e+36, -8.153, 29336.0],
            #[(9.869, 'atm'), 6.663000e+35, -7.919, 29217.0],
            #[(29.607, 'atm'), 1.723000e+38, -8.506, 31273.0],
            #[(98.69, 'atm'), 3.007000e+41, -9.29, 33966.0],
            #[(296.07, 'atm'), 6.767000e+36, -7.832, 31613.0],
            #[(986.9, 'atm'), 1.897000e+38, -8.047, 34240.0])

## Reaction 1592
#reaction('C6H101-5 + H <=> C6H9-A + H2', [6.752000e+05, 2.36, 207.0])

## Reaction 1593
#reaction('C6H101-5 + CH3 <=> C6H9-A + CH4', [7.380000e+00, 3.31, 4002.0])

## Reaction 1594
#reaction('C6H101-5 + O <=> C6H9-A + OH', [1.320000e+06, 2.43, 1210.0])

## Reaction 1595
#reaction('C6H101-5 + OH <=> C6H9-A + H2O', [5.528000e+04, 2.64, -1919.0])

## Reaction 1596
#reaction('C4H6 + C2H3 <=> C6H9-A', [8.800000e+05, 2.5, 6130.0])

## Reaction 1597
#pdep_arrhenius('C6H101-5 <=> C3H5-A + C3H5-A',
            #[(1.0, 'atm'), 5.070000e+47, -9.7, 72680.0],
            #[(4.0, 'atm'), 4.220000e+39, -7.3, 69390.0],
            #[(10.0, 'atm'), 2.120000e+35, -6.0, 67620.0])

## Reaction 1598
#pdep_arrhenius('C3H5-A + C3H5-A <=> C3H4-A + C3H6',
            #[(1.0, 'atm'), 4.770000e+40, -9.3, 12470.0],
            #[(4.0, 'atm'), 3.970000e+32, -6.8, 9180.0],
            #[(10.0, 'atm'), 1.460000e+28, -5.5, 7410.0])

## Reaction 1599
#reaction('C6H101-5 + O2 <=> C6H9-A + HO2', [4.000000e+14, 0.0, 38890.0])

## Reaction 1600
#reaction('C6H101-5 + H => C3H4-A + C3H5-A + H2', [1.000000e+12, 0.0, 0.0])

## Reaction 1601
#reaction('C6H101-5 + H <=> C3H5-A + C3H6', [1.000000e+13, 0.0, 0.0])

## Reaction 1602
#pdep_arrhenius('CH2 + C3H6 <=> CCYCCC',
            #[(0.01, 'atm'), 1.440000e+51, -13.12, 14153.0],
            #[(0.1, 'atm'), 3.380000e+54, -13.55, 16473.0],
            #[(1.0, 'atm'), 1.350000e+54, -12.97, 18862.0],
            #[(10.0, 'atm'), 2.730000e+47, -10.78, 14232.0],
            #[(100.0, 'atm'), 2.710000e+50, -11.22, 16720.0],
            #options='duplicate')

## Reaction 1603
#pdep_arrhenius('CH2 + C3H6 <=> CCYCCC',
            #[(0.01, 'atm'), 6.160000e+40, -10.5, 5428.1],
            #[(0.1, 'atm'), 6.930000e+41, -10.3, 6188.9],
            #[(1.0, 'atm'), 1.810000e+37, -8.55, 5521.0],
            #[(10.0, 'atm'), 4.260000e+37, -8.32, 4770.2],
            #[(100.0, 'atm'), 4.680000e+35, -7.37, 4689.1],
            #options='duplicate')

## Reaction 1604
#pdep_arrhenius('CH2 + C3H6 <=> IC4H8',
            #[(0.01, 'atm'), 4.820000e+57, -14.34, 17091.0],
            #[(0.1, 'atm'), 3.840000e+59, -14.36, 18427.0],
            #[(1.0, 'atm'), 2.130000e+58, -13.45, 20355.0],
            #[(10.0, 'atm'), 8.480000e+52, -11.63, 20677.0],
            #[(100.0, 'atm'), 6.070000e+47, -9.85, 22055.0],
            #options='duplicate')

## Reaction 1605
#pdep_arrhenius('CH2 + C3H6 <=> IC4H8',
            #[(0.01, 'atm'), 1.150000e+45, -11.13, 6145.2],
            #[(0.1, 'atm'), 1.830000e+45, -10.68, 6638.5],
            #[(1.0, 'atm'), 1.300000e+40, -8.7703, 5863.8],
            #[(10.0, 'atm'), 2.270000e+32, -6.1375, 4317.9],
            #[(100.0, 'atm'), 1.280000e+24, -3.4904, 2529.9],
            #options='duplicate')

## Reaction 1606
#pdep_arrhenius('CH2 + C3H6 <=> IC4H7 + H',
            #[(0.01, 'atm'), 8.200000e+19, -2.06, 1145.2],
            #[(0.1, 'atm'), 2.270000e+21, -2.44, 2648.4],
            #[(1.0, 'atm'), 4.440000e+35, -6.55, 13894.0],
            #[(10.0, 'atm'), 1.180000e+28, -4.09, 14013.0],
            #[(100.0, 'atm'), 6.510000e+26, -3.58, 18927.0],
            #options='duplicate')

## Reaction 1607
#pdep_arrhenius('CH2 + C3H6 <=> IC4H7 + H',
            #[(0.01, 'atm'), 1.080000e+07, 1.62, -3174.6],
            #[(0.1, 'atm'), 1.370000e+05, 2.15, -3799.2],
            #[(1.0, 'atm'), 3.890000e+14, -0.42, 1237.6],
            #[(10.0, 'atm'), 2.450000e+10, 0.67, 750.93],
            #[(100.0, 'atm'), 1.810000e+02, 2.97, -746.03],
            #options='duplicate')

## Reaction 1608
#pdep_arrhenius('CH2 + C3H6 <=> C3H5-T + CH3',
            #[(0.01, 'atm'), 1.770000e+19, -1.94, 6786.8],
            #[(0.1, 'atm'), 1.680000e+19, -1.8, 4310.4],
            #[(1.0, 'atm'), 4.160000e+24, -3.19, 9758.5],
            #[(10.0, 'atm'), 7.890000e+24, -3.07, 13894.0],
            #[(100.0, 'atm'), 7.360000e+29, -4.28, 23849.0],
            #options='duplicate')

## Reaction 1609
#pdep_arrhenius('CH2 + C3H6 <=> C3H5-T + CH3',
            #[(0.01, 'atm'), 4.300000e+12, 0.19, -110.41],
            #[(0.1, 'atm'), 2.260000e+11, 0.54, 47.81],
            #[(1.0, 'atm'), 4.920000e+09, 1.02, 599.77],
            #[(10.0, 'atm'), 1.470000e+08, 1.33, 1228.4],
            #[(100.0, 'atm'), 8.110000e+10, 0.55, 5506.5],
            #options='duplicate')

## Reaction 1610
#pdep_arrhenius('C3H5-T + CH3 <=> IC4H8',
            #[(0.01, 'atm'), 1.930000e+61, -14.91, 19559.0],
            #[(0.1, 'atm'), 7.490000e+58, -13.81, 20442.0],
            #[(1.0, 'atm'), 9.740000e+55, -12.62, 21930.0],
            #[(10.0, 'atm'), 7.700000e+50, -10.88, 22198.0],
            #[(100.0, 'atm'), 2.070000e+48, -9.9, 24315.0],
            #options='duplicate')

## Reaction 1611
#pdep_arrhenius('C3H5-T + CH3 <=> IC4H8',
            #[(0.01, 'atm'), 3.780000e+46, -11.15, 7331.5],
            #[(0.1, 'atm'), 2.860000e+40, -8.97, 6234.4],
            #[(1.0, 'atm'), 7.240000e+33, -6.71, 4848.4],
            #[(10.0, 'atm'), 2.280000e+27, -4.53, 3329.3],
            #[(100.0, 'atm'), 5.400000e+23, -3.31, 2521.8],
            #options='duplicate')

## Reaction 1612
#pdep_arrhenius('C3H5-T + CH3 <=> IC4H7 + H',
            #[(0.01, 'atm'), 4.120000e+29, -4.95, 7995.8],
            #[(0.1, 'atm'), 4.860000e+30, -5.03, 11287.0],
            #[(1.0, 'atm'), 5.300000e+29, -4.57, 14443.0],
            #[(10.0, 'atm'), 1.320000e+30, -4.54, 19255.0],
            #[(100.0, 'atm'), 5.160000e+28, -4.03, 23821.0],
            #options='duplicate')

## Reaction 1613
#pdep_arrhenius('C3H5-T + CH3 <=> IC4H7 + H',
            #[(0.01, 'atm'), 5.730000e+15, -0.77, 1195.9],
            #[(0.1, 'atm'), 2.060000e+13, -0.074, 1428.7],
            #[(1.0, 'atm'), 4.480000e+10, 0.6, 1421.6],
            #[(10.0, 'atm'), 4.100000e+06, 1.71, 1056.9],
            #[(100.0, 'atm'), 1.370000e-01, 3.91, -353.55],
            #options='duplicate')

## Reaction 1614
#pdep_arrhenius('IC4H7 + H <=> IC4H8',
            #[(0.01, 'atm'), 2.000000e+64, -15.02, 23890.0],
            #[(0.1, 'atm'), 1.400000e+61, -13.84, 24846.0],
            #[(1.0, 'atm'), 1.000000e+58, -12.7, 25947.0],
            #[(10.0, 'atm'), 1.680000e+53, -11.13, 25910.0],
            #[(100.0, 'atm'), 6.890000e+44, -8.62, 22842.0],
            #options='duplicate')

## Reaction 1615
#pdep_arrhenius('IC4H7 + H <=> IC4H8',
            #[(0.01, 'atm'), 1.850000e+40, -8.58, 6466.1],
            #[(0.1, 'atm'), 3.370000e+34, -6.64, 5193.0],
            #[(1.0, 'atm'), 9.060000e+29, -5.1, 4171.5],
            #[(10.0, 'atm'), 4.770000e+25, -3.68, 3143.1],
            #[(100.0, 'atm'), 1.100000e+21, -2.16, 1916.5],
            #options='duplicate')

## Reaction 1616
#pdep_arrhenius('IC4H8 <=> CCYCCC',
            #[(0.01, 'atm'), 1.260000e+64, -15.57, 95033.0],
            #[(0.1, 'atm'), 3.940000e+67, -16.19, 100780.0],
            #[(1.0, 'atm'), 6.140000e+68, -16.15, 105890.0],
            #[(10.0, 'atm'), 9.700000e+66, -15.301, 109000.0],
            #[(100.0, 'atm'), 1.450000e+62, -13.64, 110210.0],
            #options='duplicate')

## Reaction 1617
#pdep_arrhenius('IC4H8 <=> CCYCCC',
            #[(0.01, 'atm'), 4.840000e+41, -9.62, 79528.0],
            #[(0.1, 'atm'), 8.070000e+44, -10.21, 82671.0],
            #[(1.0, 'atm'), 1.150000e+47, -10.56, 85502.0],
            #[(10.0, 'atm'), 1.330000e+39, -7.98, 83303.0],
            #[(100.0, 'atm'), 4.970000e+31, -5.6, 80987.0],
            #options='duplicate')

## Reaction 1618
#pdep_arrhenius('CCYCCC <=> IC4H7 + H',
            #[(0.01, 'atm'), 2.330000e+63, -14.55, 103320.0],
            #[(0.1, 'atm'), 7.030000e+63, -14.37, 107160.0],
            #[(1.0, 'atm'), 5.070000e+64, -14.29, 112340.0],
            #[(10.0, 'atm'), 4.920000e+61, -13.15, 114870.0],
            #[(100.0, 'atm'), 4.820000e+57, -11.73, 118250.0],
            #options='duplicate')

## Reaction 1619
#pdep_arrhenius('CCYCCC <=> IC4H7 + H',
            #[(0.01, 'atm'), 1.120000e+40, -8.37, 85836.0],
            #[(0.1, 'atm'), 1.060000e+41, -8.33, 88499.0],
            #[(1.0, 'atm'), 8.230000e+43, -8.88, 92907.0],
            #[(10.0, 'atm'), 1.270000e+39, -7.33, 93401.0],
            #[(100.0, 'atm'), 1.450000e+28, -4.02, 90995.0],
            #options='duplicate')

## Reaction 1620
#pdep_arrhenius('CCYCCC <=> C3H5-T + CH3',
            #[(0.01, 'atm'), 8.310000e+64, -15.14, 111300.0],
            #[(0.1, 'atm'), 8.290000e+64, -14.72, 113700.0],
            #[(1.0, 'atm'), 1.000000e+70, -15.72, 121940.0],
            #[(10.0, 'atm'), 1.660000e+67, -14.59, 124380.0],
            #[(100.0, 'atm'), 7.870000e+62, -13.05, 126800.0],
            #options='duplicate')

## Reaction 1621
#pdep_arrhenius('CCYCCC <=> C3H5-T + CH3',
            #[(0.01, 'atm'), 1.510000e+49, -11.04, 99748.0],
            #[(0.1, 'atm'), 1.050000e+45, -9.46, 99275.0],
            #[(1.0, 'atm'), 3.390000e+50, -10.63, 104220.0],
            #[(10.0, 'atm'), 1.760000e+47, -9.43, 104930.0],
            #[(100.0, 'atm'), 1.880000e+39, -6.93, 103980.0],
            #options='duplicate')

## Reaction 1622
#reaction('IC4H8 <=> IC4H7-I1 + H', [7.710000e+69, -16.09, 140000.0])

## Reaction 1623
#pdep_arrhenius('IC4H8 + H <=> IC4H9',
            #[(0.0013, 'atm'), 7.990000e+81, -23.161, 22239.0],
            #[(0.04, 'atm'), 4.240000e+68, -18.427, 19665.0],
            #[(1.0, 'atm'), 1.040000e+49, -11.5, 15359.0],
            #[(10.0, 'atm'), 6.200000e+41, -8.892, 14637.0],
            #options='duplicate')

## Reaction 1624
#pdep_arrhenius('IC4H8 + H <=> IC4H9',
            #[(0.0013, 'atm'), 1.850000e+26, -5.83, 3865.8],
            #[(0.04, 'atm'), 2.820000e+30, -6.49, 5470.8],
            #[(1.0, 'atm'), 3.780000e+28, -5.57, 5625.1],
            #[(10.0, 'atm'), 1.460000e+25, -4.28, 5247.8],
            #[(100.0, 'atm'), 4.220000e+27, -4.39, 9345.8],
            #options='duplicate')

## Reaction 1625
#pdep_arrhenius('IC4H8 + H <=> C3H6 + CH3',
            #[(0.0013, 'atm'), 1.540000e+09, 1.35, 2542.0],
            #[(0.04, 'atm'), 7.880000e+10, 0.87, 3599.6],
            #[(1.0, 'atm'), 2.670000e+12, 0.47, 5431.1],
            #[(10.0, 'atm'), 9.250000e+22, -2.6, 12898.0],
            #[(100.0, 'atm'), 1.320000e+23, -2.42, 16500.0],
            #options='duplicate')

## Reaction 1626
#pdep_arrhenius('IC4H8 + H <=> C3H6 + CH3',
            #[(10.0, 'atm'), 1.240000e+05, 2.52, 3679.1],
            #[(100.0, 'atm'), 2.510000e+03, 2.91, 3980.9],
            #options='duplicate')

## Reaction 1627
#pdep_arrhenius('IC4H8 + H <=> TC4H9',
            #[(0.0013, 'atm'), 1.350000e+44, -10.68, 8196.4],
            #[(0.04, 'atm'), 2.110000e+57, -14.23, 15147.0],
            #[(1.0, 'atm'), 3.260000e+61, -14.94, 20161.0],
            #[(10.0, 'atm'), 5.300000e+56, -13.12, 20667.0],
            #[(100.0, 'atm'), 1.110000e+50, -10.8, 20202.0],
            #options='duplicate')

## Reaction 1628
#pdep_arrhenius('IC4H8 + H <=> TC4H9',
            #[(0.0013, 'atm'), 2.170000e+130, -32.58, 136140.0],
            #[(0.04, 'atm'), 2.250000e+29, -5.84, 4241.9],
            #[(1.0, 'atm'), 1.060000e+30, -5.63, 5613.4],
            #[(10.0, 'atm'), 6.110000e+26, -4.44, 5182.3],
            #[(100.0, 'atm'), 2.730000e+23, -3.26, 4597.0],
            #options='duplicate')

## Reaction 1629
#pdep_arrhenius('IC4H9 <=> C3H6 + CH3',
            #[(0.1, 'atm'), 3.150000e+41, -9.5, 33486.0],
            #[(1.0, 'atm'), 6.750000e+44, -10.07, 37209.0],
            #[(10.0, 'atm'), 7.790000e+44, -9.7, 39751.0],
            #[(100.0, 'atm'), 3.610000e+39, -7.78, 39583.0])

## Reaction 1630
#reaction('IC4H9 + O2 <=> IC4H9O2', [6.694600e+13, -0.3, -187.2])

## Reaction 1631
#reaction('TC4H9 + O2 <=> TC4H9O2', [6.694600e+13, -0.3, -187.2])

## Reaction 1632
#reaction('IC4H9O2 <=> IC4H8O2H-I', [9.820000e+07, 1.3, 21500.0])

## Reaction 1633
#reaction('IC4H9O2 <=> IC4H8O2H-T', [2.310000e+09, 0.8, 27100.0])

## Reaction 1634
#reaction('TC4H9O2 <=> TC4H8O2H-I', [4.521000e+09, 1.2, 33500.0])

## Reaction 1635
#reaction('IC4H8O2H-I + O2 <=> IC4H8OOH-IO2', [9.353160e+11, 0.1, -1010.88])

## Reaction 1636
#reaction('IC4H8O2H-T + O2 <=> IC4H8OOH-TO2', [5.044820e+12, -0.1, -655.2])

## Reaction 1637
#reaction('TC4H8O2H-I + O2 <=> TC4H8OOH-IO2', [2.284429e+10, 0.5, -786.24])

## Reaction 1638
#reaction('IC4H8O2H-I <=> CC4H8O + OH', [4.470000e+11, 0.0, 21900.0])

## Reaction 1639
#reaction('IC4H8O2H-T <=> IC4H8O + OH', [1.550000e+12, 0.0, 13400.0])

## Reaction 1640
#reaction('IC4H8O2H-I => OH + CH2O + C3H6', [8.451000e+15, -0.68, 29170.0])

## Reaction 1641
#reaction('TC4H8OOH-IO2 <=> TIC4H7Q2-I', [9.820000e+07, 1.3, 21500.0])

## Reaction 1642
#reaction('TIC4H7Q2-I <=> IC4H7OOH + HO2', [1.600000e+12, 0.23, 15200.0])

## Reaction 1643
#reaction('IC4H8OOH-IO2 <=> IC4KETII + OH', [5.000000e+10, 0.0, 21400.0])

## Reaction 1644
#reaction('IC4H8OOH-IO2 <=> IIC4H7Q2-I', [3.750000e+10, 0.0, 24400.0])

## Reaction 1645
#reaction('IC4H8OOH-IO2 <=> IIC4H7Q2-T', [1.000000e+11, 0.0, 29200.0])

## Reaction 1646
#reaction('IC4KETII => CH2O + C2H5CO + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 1647
#reaction('AC3H5OOH + CH2O2H <=> IIC4H7Q2-I', [8.500000e+10, 0.0, 10600.0])

## Reaction 1648
#reaction('IC4H7OOH + HO2 <=> IIC4H7Q2-T', [1.000000e+11, 0.0, 7800.0])

## Reaction 1649
#reaction('IC4H8 + HO2 <=> TC4H9O2', [1.040000e-01, 3.45, 4338.0])

## Reaction 1650
#pdep_arrhenius('IC4H8 + HO2 <=> TC4H8O2H-I',
            #[(0.013, 'atm'), 6.670000e+14, -2.14, 14188.0],
            #[(0.9869, 'atm'), 6.250000e+06, 0.64, 9073.0],
            #[(9.869, 'atm'), 7.960000e+06, 0.82, 8771.0],
            #[(98.69, 'atm'), 2.050000e+13, -0.82, 12919.0])

## Reaction 1651
#pdep_arrhenius('IC4H8 + HO2 <=> IC4H8O + OH',
            #[(0.013, 'atm'), 1.180000e+04, 2.29, 11321.0],
            #[(0.9869, 'atm'), 5.300000e+04, 2.1, 11797.0],
            #[(9.869, 'atm'), 1.470000e+09, 0.83, 14808.0],
            #[(98.69, 'atm'), 1.960000e+17, -1.45, 21195.0])

## Reaction 1652
#pdep_arrhenius('TC4H8O2H-I <=> IC4H8O + OH',
            #[(0.013, 'atm'), 5.230000e+17, -2.97, 8215.0],
            #[(0.9869, 'atm'), 4.390000e+22, -3.9, 11424.0],
            #[(9.869, 'atm'), 4.000000e+25, -4.5, 13952.0],
            #[(98.69, 'atm'), 1.210000e+27, -4.66, 16324.0])

## Reaction 1653
#reaction('IC4H9O2 <=> IC4H8 + HO2', [1.940000e+08, 1.27, 29600.0])

## Reaction 1654
#reaction('IC4H8 + HO2 <=> IC4H8O2H-T', [1.640000e+04, 2.43, 8300.0])

## Reaction 1655
#pdep_arrhenius('IC4H8 + OH <=> IC4H7OH + H',
            #[(0.0013, 'atm'), 2.670000e+13, 0.05, 10611.0],
            #[(0.01, 'atm'), 2.750000e+13, 0.05, 10623.0],
            #[(0.013, 'atm'), 2.870000e+13, 0.04, 10634.0],
            #[(0.025, 'atm'), 1.590000e+14, -0.16, 11125.0],
            #[(0.1, 'atm'), 3.100000e+14, -0.22, 11407.0],
            #[(0.1315, 'atm'), 3.780000e+14, -0.24, 11458.0],
            #[(1.0, 'atm'), 9.150000e+07, 1.42, 10087.0],
            #[(10.0, 'atm'), 3.660000e+05, 2.14, 10410.0],
            #[(100.0, 'atm'), 8.190000e+02, 2.84, 10481.0])

## Reaction 1656
#pdep_arrhenius('IC4H8 + OH <=> SC3H5OH + CH3',
            #[(0.0013, 'atm'), 5.160000e+05, 1.65, 1233.0],
            #[(0.01, 'atm'), 7.280000e+03, 2.1, 1162.0],
            #[(0.013, 'atm'), 8.160000e+02, 2.48, 1128.0],
            #[(0.025, 'atm'), 1.150000e+02, 2.8, 1152.0],
            #[(0.1, 'atm'), 5.600000e+00, 3.21, 1208.0],
            #[(0.1315, 'atm'), 3.080000e+00, 3.29, 1216.0],
            #[(1.0, 'atm'), 4.520000e+03, 2.5, 3238.0],
            #[(10.0, 'atm'), 9.640000e+18, -1.74, 13107.0],
            #[(100.0, 'atm'), 1.320000e-01, 3.7, 3665.0])

## Reaction 1657
#pdep_arrhenius('IC4H8 + OH <=> IC3H5OH + CH3',
            #[(0.0013, 'atm'), 2.060000e+06, 1.65, 1233.0],
            #[(0.01, 'atm'), 2.910000e+04, 2.1, 1162.0],
            #[(0.013, 'atm'), 3.260000e+03, 2.48, 1128.0],
            #[(0.025, 'atm'), 4.610000e+02, 2.8, 1152.0],
            #[(0.1, 'atm'), 2.240000e+01, 3.21, 1208.0],
            #[(0.1315, 'atm'), 1.230000e+01, 3.29, 1216.0],
            #[(1.0, 'atm'), 1.810000e+04, 2.5, 3238.0],
            #[(10.0, 'atm'), 3.860000e+19, -1.74, 13107.0],
            #[(100.0, 'atm'), 5.280000e-01, 3.7, 3665.0])

## Reaction 1658
#pdep_arrhenius('IC4H8 + OH <=> SC4H7OH-I + H',
            #[(0.0013, 'atm'), 2.870000e+00, 2.92, 625.0],
            #[(0.01, 'atm'), 4.840000e-01, 2.98, 704.0],
            #[(0.013, 'atm'), 3.130000e-01, 3.04, 721.0],
            #[(0.025, 'atm'), 9.330000e-03, 3.62, 677.0],
            #[(0.1, 'atm'), 4.640000e-05, 4.48, 687.0],
            #[(0.1315, 'atm'), 2.710000e-05, 4.56, 707.0],
            #[(1.0, 'atm'), 7.650000e-07, 5.05, 874.0],
            #[(10.0, 'atm'), 2.640000e+15, -0.8, 12728.0],
            #[(100.0, 'atm'), 4.870000e-04, 4.32, 4020.0])

## Reaction 1659
#pdep_arrhenius('IC4H8 + OH <=> CH3COCH3 + CH3',
            #[(0.0013, 'atm'), 6.930000e+05, 1.49, -536.0],
            #[(0.01, 'atm'), 5.940000e+03, 2.01, -560.0],
            #[(0.013, 'atm'), 1.100000e+03, 2.22, -680.0],
            #[(0.025, 'atm'), 1.070000e+02, 2.5, -759.0],
            #[(0.1, 'atm'), 7.830000e-01, 3.1, -919.0],
            #[(0.1315, 'atm'), 3.070000e-01, 3.22, -946.0],
            #[(1.0, 'atm'), 3.160000e-04, 4.05, -1144.0],
            #[(10.0, 'atm'), 7.590000e-06, 4.49, -680.0],
            #[(100.0, 'atm'), 5.450000e-05, 4.22, 1141.0])

## Reaction 1660
#pdep_arrhenius('IC4H8 + OH <=> IC4H8OH-IT',
            #[(0.0013, 'atm'), 6.140000e+78, -20.7, 32402.0],
            #[(0.01, 'atm'), 7.300000e+77, -20.0, 33874.0],
            #[(0.013, 'atm'), 2.840000e+76, -19.58, 32874.0],
            #[(0.025, 'atm'), 9.800000e+73, -18.79, 31361.0],
            #[(0.1, 'atm'), 2.760000e+68, -17.01, 27909.0],
            #[(0.1315, 'atm'), 1.930000e+67, -16.64, 27162.0],
            #[(1.0, 'atm'), 5.200000e+59, -14.17, 23079.0],
            #[(10.0, 'atm'), 2.020000e+54, -12.23, 22976.0],
            #[(100.0, 'atm'), 3.820000e+48, -10.23, 23772.0],
            #options='duplicate')

## Reaction 1661
#pdep_arrhenius('IC4H8 + OH <=> IC4H8OH-IT',
            #[(0.0013, 'atm'), 1.710000e+60, -15.84, 11594.0],
            #[(0.01, 'atm'), 1.940000e+60, -15.51, 12898.0],
            #[(0.013, 'atm'), 7.440000e+59, -15.34, 12913.0],
            #[(0.025, 'atm'), 7.060000e+58, -14.93, 12936.0],
            #[(0.1, 'atm'), 3.600000e+56, -14.04, 12945.0],
            #[(0.1315, 'atm'), 1.060000e+56, -13.85, 12887.0],
            #[(1.0, 'atm'), 4.140000e+50, -12.04, 11493.0],
            #[(10.0, 'atm'), 1.710000e+42, -9.35, 8921.0],
            #[(100.0, 'atm'), 6.120000e+32, -6.31, 6088.0],
            #options='duplicate')

## Reaction 1662
#pdep_arrhenius('IC4H8 + OH <=> IC4H8OH-TI',
            #[(0.0013, 'atm'), 1.540000e+78, -20.7, 32402.0],
            #[(0.01, 'atm'), 1.830000e+77, -20.0, 33874.0],
            #[(0.013, 'atm'), 7.100000e+75, -19.58, 32874.0],
            #[(0.025, 'atm'), 2.450000e+73, -18.79, 31361.0],
            #[(0.1, 'atm'), 6.900000e+67, -17.01, 27909.0],
            #[(0.1315, 'atm'), 4.820000e+66, -16.64, 27162.0],
            #[(1.0, 'atm'), 1.300000e+59, -14.17, 23079.0],
            #[(10.0, 'atm'), 5.050000e+53, -12.23, 22976.0],
            #[(100.0, 'atm'), 9.550000e+47, -10.23, 23772.0],
            #options='duplicate')

## Reaction 1663
#pdep_arrhenius('IC4H8 + OH <=> IC4H8OH-TI',
            #[(0.0013, 'atm'), 4.280000e+59, -15.84, 11594.0],
            #[(0.01, 'atm'), 4.850000e+59, -15.51, 12898.0],
            #[(0.013, 'atm'), 1.860000e+59, -15.34, 12913.0],
            #[(0.025, 'atm'), 1.770000e+58, -14.93, 12936.0],
            #[(0.1, 'atm'), 9.000000e+55, -14.04, 12945.0],
            #[(0.1315, 'atm'), 2.660000e+55, -13.85, 12887.0],
            #[(1.0, 'atm'), 1.040000e+50, -12.04, 11493.0],
            #[(10.0, 'atm'), 4.280000e+41, -9.35, 8921.0],
            #[(100.0, 'atm'), 1.530000e+32, -6.31, 6088.0],
            #options='duplicate')

## Reaction 1664
#pdep_arrhenius('IC4H8OH-IT + O2 <=> TQJC4H8OH',
            #[(0.001, 'atm'), 1.050000e+114, -33.81, 24741.0],
            #[(0.01, 'atm'), 2.130000e+114, -33.44, 26448.0],
            #[(0.1, 'atm'), 1.620000e+110, -31.75, 26612.0],
            #[(1.0, 'atm'), 6.000000e+101, -28.79, 25197.0],
            #[(10.0, 'atm'), 5.360000e+89, -24.76, 22402.0],
            #[(40.0, 'atm'), 1.550000e+81, -21.95, 20197.0],
            #[(100.0, 'atm'), 1.510000e+75, -20.0, 18578.0],
            #[(200.0, 'atm'), 3.160000e+70, -18.48, 17287.0])

## Reaction 1665
#pdep_arrhenius('IC4H8OH-IT + O2 <=> IC4H7OH + HO2',
            #[(0.001, 'atm'), 1.210000e+26, -5.09, 5755.0],
            #[(0.01, 'atm'), 1.410000e+31, -6.55, 8781.0],
            #[(0.1, 'atm'), 5.080000e+34, -7.53, 11702.0],
            #[(1.0, 'atm'), 1.670000e+34, -7.27, 13418.0],
            #[(10.0, 'atm'), 2.820000e+28, -5.41, 13318.0],
            #[(40.0, 'atm'), 2.500000e+22, -3.52, 12314.0],
            #[(100.0, 'atm'), 2.820000e+17, -1.99, 11286.0],
            #[(200.0, 'atm'), 1.930000e+13, -0.71, 10340.0],
            #options='duplicate')

## Reaction 1666
#pdep_arrhenius('IC4H8OH-IT + O2 <=> IC4H7OH + HO2',
            #[(0.001, 'atm'), 2.450000e+21, -4.19, 6837.0],
            #[(0.01, 'atm'), 1.400000e+30, -6.75, 11554.0],
            #[(0.1, 'atm'), 9.110000e+39, -9.56, 17834.0],
            #[(1.0, 'atm'), 5.160000e+42, -10.17, 22412.0],
            #[(10.0, 'atm'), 6.070000e+32, -6.94, 22738.0],
            #[(40.0, 'atm'), 3.880000e+20, -3.14, 20677.0],
            #[(100.0, 'atm'), 3.320000e+10, -0.03, 18552.0],
            #[(200.0, 'atm'), 1.220000e+02, 2.57, 16623.0],
            #options='duplicate')

## Reaction 1667
#pdep_arrhenius('IC4H8OH-IT + O2 <=> SC4H7OH-I + HO2',
            #[(0.001, 'atm'), 1.260000e+25, -4.69, 5755.0],
            #[(0.01, 'atm'), 1.440000e+30, -6.15, 8785.0],
            #[(0.1, 'atm'), 4.780000e+33, -7.11, 11695.0],
            #[(1.0, 'atm'), 1.400000e+33, -6.84, 13395.0],
            #[(10.0, 'atm'), 2.120000e+27, -4.96, 13277.0],
            #[(40.0, 'atm'), 1.780000e+21, -3.07, 12265.0],
            #[(100.0, 'atm'), 1.970000e+16, -1.53, 11234.0],
            #[(200.0, 'atm'), 1.320000e+12, -0.25, 10285.0],
            #options='duplicate')

## Reaction 1668
#pdep_arrhenius('IC4H8OH-IT + O2 <=> SC4H7OH-I + HO2',
            #[(0.001, 'atm'), 3.220000e+23, -4.69, 5341.0],
            #[(0.01, 'atm'), 4.800000e+28, -6.18, 8461.0],
            #[(0.1, 'atm'), 1.710000e+32, -7.16, 11410.0],
            #[(1.0, 'atm'), 1.530000e+32, -7.02, 13378.0],
            #[(10.0, 'atm'), 5.550000e+29, -6.14, 15100.0],
            #[(40.0, 'atm'), 1.720000e+26, -4.97, 15849.0],
            #[(100.0, 'atm'), 4.460000e+21, -3.51, 15644.0],
            #[(200.0, 'atm'), 5.060000e+16, -1.96, 14979.0],
            #options='duplicate')

## Reaction 1669
#pdep_arrhenius('IC4H8OH-IT + O2 <=> TQC4H8OI',
            #[(0.001, 'atm'), 6.360000e+104, -33.74, 22390.0],
            #[(0.01, 'atm'), 7.960000e+103, -33.01, 22966.0],
            #[(0.1, 'atm'), 6.990000e+96, -30.48, 20584.0],
            #[(1.0, 'atm'), 1.040000e+88, -27.47, 16629.0],
            #[(10.0, 'atm'), 3.130000e+96, -29.62, 20346.0],
            #[(40.0, 'atm'), 2.790000e+105, -32.04, 24971.0],
            #[(100.0, 'atm'), 4.640000e+109, -33.12, 27657.0],
            #[(200.0, 'atm'), 2.060000e+111, -33.48, 29197.0])

## Reaction 1670
#pdep_arrhenius('IC4H8OH-IT + O2 <=> CH3COCH3 + CH2O + OH',
            #[(0.001, 'atm'), 3.030000e+37, -8.35, 6940.0],
            #[(0.01, 'atm'), 1.070000e+42, -9.64, 9965.0],
            #[(0.1, 'atm'), 9.220000e+43, -10.12, 12427.0],
            #[(1.0, 'atm'), 1.140000e+42, -9.42, 13806.0],
            #[(10.0, 'atm'), 1.800000e+38, -8.13, 15131.0],
            #[(40.0, 'atm'), 1.610000e+34, -6.8, 15691.0],
            #[(100.0, 'atm'), 7.000000e+29, -5.41, 15552.0],
            #[(200.0, 'atm'), 3.740000e+25, -4.06, 15118.0])

## Reaction 1671
#pdep_arrhenius('IC4H8OH-IT + O2 <=> QC4H7OHP',
            #[(0.001, 'atm'), 1.180000e+118, -37.6, 26229.0],
            #[(0.01, 'atm'), 1.910000e+131, -40.73, 34079.0],
            #[(0.1, 'atm'), 5.810000e+138, -42.17, 40750.0],
            #[(1.0, 'atm'), 5.050000e+134, -40.19, 43580.0],
            #[(10.0, 'atm'), 2.070000e+117, -34.26, 41516.0],
            #[(40.0, 'atm'), 1.710000e+101, -29.04, 38094.0],
            #[(100.0, 'atm'), 1.250000e+89, -25.15, 35203.0],
            #[(200.0, 'atm'), 2.980000e+79, -22.09, 32802.0])

## Reaction 1672
#pdep_arrhenius('IC4H8OH-IT + O2 <=> CCY(CCO)COH + OH',
            #[(0.001, 'atm'), 2.110000e+18, -2.87, 6870.0],
            #[(0.01, 'atm'), 3.020000e+27, -5.54, 11842.0],
            #[(0.1, 'atm'), 1.360000e+37, -8.29, 18165.0],
            #[(1.0, 'atm'), 1.210000e+39, -8.66, 22517.0],
            #[(10.0, 'atm'), 1.530000e+28, -5.13, 22530.0],
            #[(40.0, 'atm'), 2.720000e+15, -1.16, 20283.0],
            #[(100.0, 'atm'), 1.060000e+05, 2.06, 18044.0],
            #[(200.0, 'atm'), 2.300000e-04, 4.73, 16037.0])

## Reaction 1673
#pdep_arrhenius('IC4H8OH-IT + O2 <=> TQC4H7OHI',
            #[(0.001, 'atm'), 1.760000e+53, -19.87, 9019.0],
            #[(0.01, 'atm'), 4.680000e+67, -23.92, 11892.0],
            #[(0.1, 'atm'), 2.070000e+91, -30.58, 17347.0],
            #[(1.0, 'atm'), 2.840000e+100, -32.4, 20041.0],
            #[(10.0, 'atm'), 1.160000e+115, -35.81, 27656.0],
            #[(40.0, 'atm'), 4.870000e+123, -37.83, 33314.0],
            #[(100.0, 'atm'), 7.950000e+124, -37.82, 35683.0],
            #[(200.0, 'atm'), 4.580000e+122, -36.86, 36374.0])

## Reaction 1674
#pdep_arrhenius('IC4H8OH-IT + O2 <=> C2CY(COC)OH + OH',
            #[(0.001, 'atm'), 1.420000e+32, -6.95, 6210.0],
            #[(0.01, 'atm'), 5.190000e+36, -8.24, 9233.0],
            #[(0.1, 'atm'), 5.770000e+38, -8.76, 11715.0],
            #[(1.0, 'atm'), 2.860000e+36, -7.95, 12823.0],
            #[(10.0, 'atm'), 1.370000e+32, -6.51, 13646.0],
            #[(40.0, 'atm'), 2.150000e+29, -5.56, 14541.0],
            #[(100.0, 'atm'), 1.200000e+26, -4.51, 14778.0],
            #[(200.0, 'atm'), 3.310000e+22, -3.37, 14606.0])

## Reaction 1675
#pdep_arrhenius('TQJC4H8OH <=> IC4H7OH + HO2',
            #[(0.001, 'atm'), 9.730000e+65, -18.5, 42975.0],
            #[(0.01, 'atm'), 1.040000e+64, -17.25, 44419.0],
            #[(0.1, 'atm'), 7.900000e+59, -15.59, 44504.0],
            #[(1.0, 'atm'), 9.350000e+53, -13.49, 43566.0],
            #[(10.0, 'atm'), 4.130000e+44, -10.39, 41279.0],
            #[(40.0, 'atm'), 6.570000e+38, -8.49, 39745.0],
            #[(100.0, 'atm'), 8.320000e+34, -7.23, 38675.0],
            #[(200.0, 'atm'), 9.460000e+31, -6.28, 37849.0],
            #options='duplicate')

## Reaction 1676
#pdep_arrhenius('TQJC4H8OH <=> IC4H7OH + HO2',
            #[(0.001, 'atm'), 5.270000e+64, -18.0, 42872.0],
            #[(0.01, 'atm'), 4.820000e+62, -16.74, 44284.0],
            #[(0.1, 'atm'), 3.420000e+58, -15.07, 44348.0],
            #[(1.0, 'atm'), 4.050000e+52, -12.97, 43402.0],
            #[(10.0, 'atm'), 1.930000e+43, -9.88, 41120.0],
            #[(40.0, 'atm'), 3.280000e+37, -7.99, 39593.0],
            #[(100.0, 'atm'), 4.350000e+33, -6.74, 38527.0],
            #[(200.0, 'atm'), 5.130000e+30, -5.79, 37706.0],
            #options='duplicate')

## Reaction 1677
#pdep_arrhenius('TQJC4H8OH <=> TQC4H8OI',
            #[(0.001, 'atm'), 3.780000e+50, -12.91, 31539.0],
            #[(0.01, 'atm'), 1.420000e+45, -10.94, 30864.0],
            #[(0.1, 'atm'), 1.500000e+40, -9.21, 29932.0],
            #[(1.0, 'atm'), 3.120000e+35, -7.64, 28864.0],
            #[(10.0, 'atm'), 2.730000e+29, -5.67, 27243.0],
            #[(40.0, 'atm'), 8.880000e+25, -4.54, 26272.0],
            #[(100.0, 'atm'), 4.600000e+23, -3.8, 25622.0],
            #[(200.0, 'atm'), 9.080000e+21, -3.25, 25131.0])

## Reaction 1678
#pdep_arrhenius('TQJC4H8OH <=> QC4H7OHP',
            #[(0.001, 'atm'), 9.160000e+62, -18.02, 45297.0],
            #[(0.01, 'atm'), 2.560000e+62, -17.1, 47393.0],
            #[(0.1, 'atm'), 1.250000e+59, -15.61, 47984.0],
            #[(1.0, 'atm'), 1.970000e+53, -13.49, 47281.0],
            #[(10.0, 'atm'), 1.910000e+43, -10.15, 44926.0],
            #[(40.0, 'atm'), 7.680000e+36, -8.05, 43267.0],
            #[(100.0, 'atm'), 3.490000e+32, -6.64, 42089.0],
            #[(200.0, 'atm'), 1.770000e+29, -5.57, 41173.0])

## Reaction 1679
#pdep_arrhenius('TQJC4H8OH <=> TQC4H7OHI',
            #[(0.001, 'atm'), 9.570000e+58, -15.99, 38293.0],
            #[(0.01, 'atm'), 7.270000e+54, -14.25, 38593.0],
            #[(0.1, 'atm'), 8.300000e+49, -12.44, 38031.0],
            #[(1.0, 'atm'), 1.900000e+44, -10.51, 36905.0],
            #[(10.0, 'atm'), 2.180000e+36, -7.9, 34865.0],
            #[(40.0, 'atm'), 4.100000e+31, -6.36, 33581.0],
            #[(100.0, 'atm'), 3.000000e+28, -5.35, 32704.0],
            #[(200.0, 'atm'), 1.340000e+26, -4.59, 32035.0])

## Reaction 1680
#pdep_arrhenius('TQC4H8OI <=> CH3COCH3 + CH2O + OH',
            #[(0.001, 'atm'), 8.090000e+38, -9.91, 19096.0],
            #[(0.01, 'atm'), 9.090000e+39, -9.93, 19135.0],
            #[(0.1, 'atm'), 1.890000e+41, -10.02, 19407.0],
            #[(1.0, 'atm'), 7.470000e+23, -4.1, 14658.0],
            #[(10.0, 'atm'), 2.950000e+33, -6.75, 18685.0],
            #[(40.0, 'atm'), 4.240000e+36, -7.56, 20307.0],
            #[(100.0, 'atm'), 7.830000e+36, -7.54, 20747.0],
            #[(200.0, 'atm'), 7.630000e+35, -7.17, 20641.0])

## Reaction 1681
#pdep_arrhenius('QC4H7OHP <=> IC4H7OH + HO2',
            #[(0.001, 'atm'), 3.290000e+57, -15.64, 28576.0],
            #[(0.01, 'atm'), 9.650000e+58, -15.75, 29927.0],
            #[(0.1, 'atm'), 2.190000e+50, -12.66, 28547.0],
            #[(1.0, 'atm'), 1.670000e+49, -12.05, 29204.0],
            #[(10.0, 'atm'), 7.080000e+40, -9.26, 27188.0],
            #[(40.0, 'atm'), 1.320000e+30, -5.82, 24071.0],
            #[(100.0, 'atm'), 1.150000e+30, -5.8, 24053.0],
            #[(200.0, 'atm'), 1.120000e+30, -5.8, 24050.0])

## Reaction 1682
#pdep_arrhenius('QC4H7OHP <=> CCY(CCO)COH + OH',
            #[(0.001, 'atm'), 1.110000e+51, -12.97, 28497.0],
            #[(0.01, 'atm'), 5.580000e+51, -12.87, 29529.0],
            #[(0.1, 'atm'), 3.030000e+44, -10.28, 28326.0],
            #[(1.0, 'atm'), 1.420000e+44, -9.98, 28986.0],
            #[(10.0, 'atm'), 9.690000e+37, -7.93, 27491.0],
            #[(40.0, 'atm'), 5.900000e+29, -5.3, 25095.0],
            #[(100.0, 'atm'), 5.270000e+29, -5.29, 25081.0],
            #[(200.0, 'atm'), 5.150000e+29, -5.28, 25078.0])

## Reaction 1683
#pdep_arrhenius('TQC4H7OHI <=> C2CY(COC)OH + OH',
            #[(0.001, 'atm'), 1.100000e+31, -7.21, 14640.0],
            #[(0.01, 'atm'), 1.370000e+32, -7.24, 14716.0],
            #[(0.1, 'atm'), 3.340000e+33, -7.35, 15127.0],
            #[(1.0, 'atm'), 1.860000e+23, -3.68, 12864.0],
            #[(10.0, 'atm'), 9.880000e+31, -6.23, 16040.0],
            #[(40.0, 'atm'), 1.250000e+34, -6.78, 17056.0],
            #[(100.0, 'atm'), 7.730000e+33, -6.67, 17176.0],
            #[(200.0, 'atm'), 2.130000e+33, -6.45, 17170.0])

## Reaction 1684
#pdep_arrhenius('TQC4H7OHI <=> SC4H7OH-I + HO2',
            #[(0.001, 'atm'), 6.010000e+27, -7.27, 14658.0],
            #[(0.01, 'atm'), 7.780000e+28, -7.3, 14733.0],
            #[(0.1, 'atm'), 3.270000e+30, -7.48, 15191.0],
            #[(1.0, 'atm'), 2.810000e+17, -2.82, 12336.0],
            #[(10.0, 'atm'), 1.920000e+26, -5.22, 16146.0],
            #[(40.0, 'atm'), 7.610000e+27, -5.49, 17483.0],
            #[(100.0, 'atm'), 1.910000e+26, -4.86, 17429.0],
            #[(200.0, 'atm'), 2.940000e+24, -4.18, 17285.0])

## Reaction 1685
#pdep_arrhenius('IC4H8OH-TI + O2 <=> IQJC4H8OH',
            #[(0.001, 'atm'), 3.340000e+111, -32.67, 25143.0],
            #[(0.01, 'atm'), 4.840000e+107, -31.05, 25460.0],
            #[(0.1, 'atm'), 2.410000e+100, -28.42, 24474.0],
            #[(1.0, 'atm'), 4.550000e+89, -24.78, 22176.0],
            #[(10.0, 'atm'), 1.140000e+76, -20.31, 18721.0],
            #[(40.0, 'atm'), 9.250000e+66, -17.35, 16238.0],
            #[(100.0, 'atm'), 6.200000e+60, -15.36, 14499.0],
            #[(200.0, 'atm'), 1.300000e+56, -13.86, 13159.0])

## Reaction 1686
#pdep_arrhenius('IC4H8OH-TI + O2 <=> IC3H6OHCHO + OH',
            #[(0.001, 'atm'), 1.590000e+15, -1.63, 8947.0],
            #[(0.01, 'atm'), 5.070000e+20, -3.24, 11938.0],
            #[(0.1, 'atm'), 5.000000e+25, -4.66, 15251.0],
            #[(1.0, 'atm'), 3.180000e+26, -4.79, 17388.0],
            #[(10.0, 'atm'), 6.000000e+20, -2.95, 17297.0],
            #[(40.0, 'atm'), 1.700000e+14, -0.91, 16099.0],
            #[(100.0, 'atm'), 7.910000e+08, 0.73, 14913.0],
            #[(200.0, 'atm'), 3.030000e+04, 2.08, 13854.0])

## Reaction 1687
#pdep_arrhenius('IC4H8OH-TI + O2 <=> IQC4H8OT',
            #[(0.001, 'atm'), 1.120000e+102, -32.4, 23496.0],
            #[(0.01, 'atm'), 4.230000e+95, -30.04, 22067.0],
            #[(0.1, 'atm'), 3.450000e+83, -26.08, 17114.0],
            #[(1.0, 'atm'), 8.160000e+82, -25.61, 16198.0],
            #[(10.0, 'atm'), 1.040000e+97, -29.54, 22648.0],
            #[(40.0, 'atm'), 7.250000e+103, -31.42, 26487.0],
            #[(100.0, 'atm'), 2.480000e+106, -32.06, 28397.0],
            #[(200.0, 'atm'), 1.390000e+107, -32.2, 29446.0])

## Reaction 1688
#pdep_arrhenius('IC4H8OH-TI + O2 <=> CH3COCH3 + CH2O + OH',
            #[(0.001, 'atm'), 4.390000e+48, -11.88, 11603.0],
            #[(0.01, 'atm'), 9.820000e+50, -12.49, 14143.0],
            #[(0.1, 'atm'), 3.430000e+49, -11.94, 15561.0],
            #[(1.0, 'atm'), 2.430000e+45, -10.56, 16415.0],
            #[(10.0, 'atm'), 4.980000e+39, -8.68, 17473.0],
            #[(40.0, 'atm'), 7.990000e+33, -6.83, 17502.0],
            #[(100.0, 'atm'), 1.020000e+29, -5.28, 17165.0],
            #[(200.0, 'atm'), 7.180000e+24, -3.99, 16747.0])

## Reaction 1689
#pdep_arrhenius('IC4H8OH-TI + O2 <=> CH3 + CH3COCOOH',
            #[(0.001, 'atm'), 2.910000e+40, -9.23, 10830.0],
            #[(0.01, 'atm'), 9.660000e+43, -10.2, 13698.0],
            #[(0.1, 'atm'), 1.340000e+44, -10.13, 15661.0],
            #[(1.0, 'atm'), 2.570000e+41, -9.18, 17047.0],
            #[(10.0, 'atm'), 1.820000e+36, -7.46, 18330.0],
            #[(40.0, 'atm'), 8.820000e+29, -5.44, 18205.0],
            #[(100.0, 'atm'), 1.740000e+24, -3.65, 17600.0],
            #[(200.0, 'atm'), 2.160000e+19, -2.12, 16925.0])

## Reaction 1690
#pdep_arrhenius('IC4H8OH-TI + O2 <=> IQC4H7OHT',
            #[(0.001, 'atm'), 1.030000e+115, -35.13, 25407.0],
            #[(0.01, 'atm'), 1.100000e+129, -38.89, 32891.0],
            #[(0.1, 'atm'), 5.150000e+135, -40.38, 38573.0],
            #[(1.0, 'atm'), 7.270000e+129, -38.13, 39933.0],
            #[(10.0, 'atm'), 4.500000e+112, -32.49, 37045.0],
            #[(40.0, 'atm'), 7.720000e+97, -27.79, 33612.0],
            #[(100.0, 'atm'), 5.170000e+86, -24.26, 30799.0],
            #[(200.0, 'atm'), 4.700000e+77, -21.41, 28440.0])

## Reaction 1691
#pdep_arrhenius('IC4H8OH-TI + O2 <=> IC3H5OH + CH2O + OH',
            #[(0.001, 'atm'), 5.100000e+24, -4.31, 13009.0],
            #[(0.01, 'atm'), 6.910000e+33, -6.97, 17935.0],
            #[(0.1, 'atm'), 1.740000e+41, -9.03, 23613.0],
            #[(1.0, 'atm'), 2.260000e+36, -7.32, 25633.0],
            #[(10.0, 'atm'), 3.150000e+17, -1.4, 22782.0],
            #[(40.0, 'atm'), 9.350000e+00, 3.71, 19114.0],
            #[(100.0, 'atm'), 5.590000e-12, 7.46, 16156.0],
            #[(200.0, 'atm'), 1.270000e-21, 10.41, 13733.0])

## Reaction 1692
#pdep_arrhenius('IC4H8OH-TI + O2 <=> CCY(CCOC)OH + OH',
            #[(0.001, 'atm'), 3.060000e+33, -7.24, 11476.0],
            #[(0.01, 'atm'), 6.560000e+42, -9.92, 17197.0],
            #[(0.1, 'atm'), 8.240000e+46, -10.95, 22090.0],
            #[(1.0, 'atm'), 4.480000e+38, -8.22, 23019.0],
            #[(10.0, 'atm'), 1.010000e+18, -1.77, 19496.0],
            #[(40.0, 'atm'), 1.160000e+01, 3.44, 15637.0],
            #[(100.0, 'atm'), 4.850000e-12, 7.23, 12599.0],
            #[(200.0, 'atm'), 8.640000e-22, 10.21, 10126.0])

## Reaction 1693
#pdep_arrhenius('IQJC4H8OH <=> IC3H6OHCHO + OH',
            #[(0.001, 'atm'), 3.580000e+71, -20.62, 52656.0],
            #[(0.01, 'atm'), 9.670000e+71, -20.08, 54935.0],
            #[(0.1, 'atm'), 3.390000e+67, -18.18, 55330.0],
            #[(1.0, 'atm'), 7.260000e+58, -15.09, 54016.0],
            #[(10.0, 'atm'), 4.570000e+46, -11.01, 51172.0],
            #[(40.0, 'atm'), 4.310000e+38, -8.38, 49054.0],
            #[(100.0, 'atm'), 2.120000e+33, -6.66, 47587.0],
            #[(200.0, 'atm'), 2.720000e+29, -5.41, 46486.0])

## Reaction 1694
#pdep_arrhenius('IQJC4H8OH <=> IQC4H8OT',
            #[(0.001, 'atm'), 1.040000e+45, -11.2, 31755.0],
            #[(0.01, 'atm'), 2.900000e+40, -9.62, 30945.0],
            #[(0.1, 'atm'), 4.850000e+35, -8.01, 29850.0],
            #[(1.0, 'atm'), 2.620000e+30, -6.28, 28498.0],
            #[(10.0, 'atm'), 3.370000e+24, -4.37, 26873.0],
            #[(40.0, 'atm'), 9.750000e+20, -3.23, 25861.0],
            #[(100.0, 'atm'), 5.610000e+18, -2.52, 25208.0],
            #[(200.0, 'atm'), 1.440000e+17, -2.01, 24740.0])

## Reaction 1695
#pdep_arrhenius('IQJC4H8OH <=> IQC4H7OHT',
            #[(0.001, 'atm'), 1.360000e+50, -12.93, 36743.0],
            #[(0.01, 'atm'), 3.360000e+45, -11.27, 36143.0],
            #[(0.1, 'atm'), 1.130000e+40, -9.41, 34990.0],
            #[(1.0, 'atm'), 5.930000e+33, -7.33, 33438.0],
            #[(10.0, 'atm'), 4.240000e+26, -5.0, 31502.0],
            #[(40.0, 'atm'), 1.970000e+22, -3.61, 30275.0],
            #[(100.0, 'atm'), 3.430000e+19, -2.72, 29477.0],
            #[(200.0, 'atm'), 3.690000e+17, -2.09, 28900.0])

## Reaction 1696
#pdep_arrhenius('IQC4H7OHT <=> IC3H5OH + CH2O + OH',
            #[(0.001, 'atm'), 4.250000e+16, -3.66, 19364.0],
            #[(0.01, 'atm'), 2.980000e+33, -8.01, 27106.0],
            #[(0.1, 'atm'), 3.460000e+38, -8.83, 31412.0],
            #[(1.0, 'atm'), 7.070000e+28, -5.46, 29633.0],
            #[(10.0, 'atm'), 2.010000e+13, -0.43, 25278.0],
            #[(40.0, 'atm'), 1.430000e+13, -0.38, 25238.0],
            #[(100.0, 'atm'), 1.380000e+13, -0.38, 25234.0],
            #[(200.0, 'atm'), 1.370000e+13, -0.37, 25233.0])

## Reaction 1697
#pdep_arrhenius('IQC4H7OHT <=> CCY(CCOC)OH + OH',
            #[(0.001, 'atm'), 1.120000e+25, -5.59, 19349.0],
            #[(0.01, 'atm'), 7.200000e+32, -7.6, 23093.0],
            #[(0.1, 'atm'), 2.220000e+34, -7.65, 25130.0],
            #[(1.0, 'atm'), 2.400000e+27, -5.28, 23663.0],
            #[(10.0, 'atm'), 2.240000e+17, -2.04, 20810.0],
            #[(40.0, 'atm'), 1.820000e+17, -2.01, 20786.0],
            #[(100.0, 'atm'), 1.780000e+17, -2.01, 20783.0],
            #[(200.0, 'atm'), 1.770000e+17, -2.01, 20782.0])

## Reaction 1698
#pdep_arrhenius('IQC4H7OHT <=> IC4H7OOH + OH',
            #[(0.001, 'atm'), 6.070000e+13, -4.21, 22478.0],
            #[(0.01, 'atm'), 9.740000e+39, -10.89, 33616.0],
            #[(0.1, 'atm'), 2.180000e+49, -12.65, 40359.0],
            #[(1.0, 'atm'), 3.320000e+37, -8.42, 38539.0],
            #[(10.0, 'atm'), 2.340000e+16, -1.56, 32712.0],
            #[(40.0, 'atm'), 1.440000e+16, -1.49, 32656.0],
            #[(100.0, 'atm'), 1.380000e+16, -1.48, 32651.0],
            #[(200.0, 'atm'), 1.370000e+16, -1.48, 32649.0])

## Reaction 1699
#pdep_arrhenius('CH3COCOOH <=> CH3COCH2O + OH',
            #[(0.01, 'atm'), 1.490000e+58, -13.9, 54266.9],
            #[(0.1, 'atm'), 1.800000e+54, -12.4, 54193.8],
            #[(1.0, 'atm'), 3.360000e+46, -9.81, 52468.5],
            #[(10.0, 'atm'), 2.390000e+36, -6.54, 49429.0],
            #[(100.0, 'atm'), 1.280000e+27, -3.61, 46333.1])

## Reaction 1700
#reaction('CH3COCH2O <=> CH3CO + CH2O', [5.872000e+20, -2.4218, 10535.8])

## Reaction 1701
#reaction('C2CY(COC)OH + OH => IC3H6CO + OH + H2O', [2.260000e+03, 2.73, -4688.0])

## Reaction 1702
#reaction('CCY(CCO)COH + OH => IC3H5CHO + OH + H2O', [5.660000e+02, 2.93, -4039.4])

## Reaction 1703
#reaction('CCY(CCOC)OH + OH => CH2O + SC3H4OH + H2O', [5.660000e+02, 2.93, -4039.4])

## Reaction 1704
#reaction('C2CY(COC)OH + HO2 => IC3H6CO + OH + H2O2', [8.600000e+00, 3.46, 9732.326])

## Reaction 1705
#reaction('CCY(CCO)COH + HO2 => IC3H5CHO + OH + H2O2', [1.810000e-01, 3.98, 9056.7])

## Reaction 1706
#reaction('CCY(CCOC)OH + HO2 => CH2O + SC3H4OH + H2O2', [1.810000e-01, 3.98, 9056.7])

## Reaction 1707
#reaction('CCY(CCO)COH + OH => PC3H4OH-2 + CH2O + H2O', [1.260000e+03, 2.97, -2660.6])

## Reaction 1708
#reaction('CCY(CCO)COH + HO2 => PC3H4OH-2 + CH2O + H2O2', [1.400000e-05, 5.26, 8267.9])

## Reaction 1709
#reaction('IC3H6OHCHO + OH => TC3H6OH + CO + H2O', [6.132990e+04, 2.65, -4586.4])

## Reaction 1710
#reaction('IC3H6OHCHO + H => TC3H6OH + CO + H2', [8.660000e+05, 2.3, 1426.0])

## Reaction 1711
#reaction('IC3H6OHCHO + HO2 => TC3H6OH + CO + H2O2', [1.010000e-04, 5.0, 3429.0])

## Reaction 1712
#reaction('IC3H6OHCHO + CH3 => TC3H6OH + CO + CH4', [3.940000e+00, 3.6, 4223.0])

## Reaction 1713
#reaction('TC3H6OH <=> CH3COCH3 + H', [8.981000e+11, 0.271, 32990.0])

## Reaction 1714
#reaction('TC3H6OH <=> IC3H5OH + H', [4.211000e+10, 1.005, 40900.0])

## Reaction 1715
#reaction('TC3H6OH + O2 <=> CH3COCH3 + HO2', [2.230000e+13, 0.0, 0.0])

## Reaction 1716
#reaction('TQC4H7OHIO2 <=> TQC4H7OHIQ-I', [2.560000e+12, -0.13, 34360.0])

## Reaction 1717
#reaction('TQC4H7OHIQ-I <=> HO2CHO + CH3COCH3 + OH', [5.819000e+05, 2.4, 22790.0])

## Reaction 1718
#reaction('TQC4H7OHIQ-I <=> IC4KETIT + HO2', [1.829000e+10, 0.79, 15100.0])

## Reaction 1719
#reaction('IC4KETIT <=> CH3COCH3 + HCO + OH', [9.500000e+15, 0.0, 42540.0])

## Reaction 1720
#reaction('IC4KETIT + OH <=> TC3H6O2HCO + H2O', [6.132995e+04, 2.65419, -4586.39996])

## Reaction 1721
#reaction('IC4KETIT + HO2 <=> TC3H6O2HCO + H2O2', [1.177300e-04, 4.91966, 3684.27444])

## Reaction 1722
#reaction('TC3H6O2HCO => CH3COCH3 + CO + OH', [1.278000e+20, -1.89, 34460.0])

## Reaction 1723
#reaction('TQC4H7OHIO2 <=> TQC4H7OHIQ-P', [5.690000e+08, 0.78, 21850.0])

## Reaction 1724
#reaction('TQC4H7OHIQ-P <=> IC3H5COHQ + HO2', [1.830000e+10, 0.79, 15100.0])

## Reaction 1725
#reaction('TQC4H7OHIQ-P <=> CH2CQCOHQ + CH3', [5.380000e+11, 0.07, 24800.0])

## Reaction 1726
#reaction('TQC4H7OHIQ-P <=> IC3H5Q + HOCHO + OH', [5.380000e+11, 0.07, 24800.0])

## Reaction 1727
#reaction('TQC4H7OHIQ-P <=> COHQCYC(COC) + OH', [2.280000e+08, 1.29, 9890.0])

## Reaction 1728
#reaction('TQC4H7OHIQ-P <=> QCYC(CCOC)OH + OH', [4.580000e+15, -1.08, 18440.0])

## Reaction 1729
#reaction('IC3H5COHQ => HOCHO + C3H5-T + OH', [1.590000e+20, -1.5, 42879.46])

## Reaction 1730
#reaction('CH2CQCOHQ => HOCHO + CH2CO + 2 OH', [1.590000e+20, -1.5, 42879.46])

## Reaction 1731
#reaction('CH2CQCOHQ => HCO + CH2CO + 3 OH', [1.590000e+20, -1.5, 42879.46])

## Reaction 1732
#reaction('IC3H5Q => CH2CO + CH3 + OH', [1.590000e+20, -1.5, 42879.46])

## Reaction 1733
#reaction('COHQCYC(COC) => HOCHO + CH2O + C2H3 + OH', [1.590000e+20, -1.5, 42879.46])

## Reaction 1734
#reaction('QCYC(CCOC)OH => HOCHO + CH2CO + CH3 + OH', [1.590000e+20, -1.5, 42879.46])

## Reaction 1735
#reaction('TQC4H7OHIO2 <=> TQC4H7OHTO2', [2.956000e+09, 0.04, 16350.0])

## Reaction 1736
#reaction('TQC4H7OHTO2 <=> HOCOCQ(CH3)2 + OH', [1.750000e+08, 1.7, 26000.0])

## Reaction 1737
#reaction('HOCOCQ(CH3)2 => CO + CH3COCH3 + 2 OH', [1.590000e+20, -1.5, 42879.46])

## Reaction 1738
#reaction('IQC4H7OHTO2 <=> IQC4H7OHTQ-P', [2.063000e+07, 1.0, 21070.0])

## Reaction 1739
#reaction('IQC4H7OHTQ-P => OH + CH2O + CH2COHCH2OOH', [1.200000e+10, 0.35, 15700.0])

## Reaction 1740
#reaction('CH2COHCH2OOH => OH + CH2O + CH2CO + H', [1.590000e+20, -1.5, 42879.46])

## Reaction 1741
#pdep_arrhenius('IQC4H7OHTQ-P <=> OH + IC4H6Q2-II',
            #[(0.01, 'atm'), 2.530000e+40, -9.91, 32631.4],
            #[(0.1, 'atm'), 2.540000e+37, -8.72, 32909.5],
            #[(1.0, 'atm'), 1.120000e+31, -6.53, 31806.6],
            #[(10.0, 'atm'), 1.980000e+21, -3.34, 29137.8],
            #[(100.0, 'atm'), 1.950000e+10, 0.154, 25612.0])

## Reaction 1742
#reaction('IC4H6Q2-II => OH + CH2O + HO2 + C3H4-A', [1.590000e+20, -1.5, 42879.46])

## Reaction 1743
#reaction('IQC4H7OHTO2 <=> IQC4H8OTQ-I', [4.130000e+07, 1.0, 21070.0])

## Reaction 1744
#reaction('IQC4H8OTQ-I => OH + CH2O + C3KET21', [1.200000e+10, 0.35, 15700.0])

## Reaction 1745
#reaction('C3KET21 => OH + CH2O + CO + CH3', [1.590000e+20, -1.5, 42879.46])

## Reaction 1746
#reaction('IQC4H8OTQ-I <=> CH3 + CO(CH2OOH)2', [2.740000e+13, 0.24, 29830.0])

## Reaction 1747
#reaction('CO(CH2OOH)2 => 2 OH + 2 CH2O + CO', [1.590000e+20, -1.5, 42879.46])

## Reaction 1748
#reaction('IQC4H7OHTO2 <=> CHOC(CH3)OHCH2Q + OH', [1.200000e+10, 0.35, 15700.0])

## Reaction 1749
#reaction('CHOC(CH3)OHCH2Q => OH + CH2O + H + CH3COCHO', [1.590000e+20, -1.5, 42879.46])

## Reaction 1750
#reaction('CH3COCHO + OH => 2 CO + CH3 + H2O', [6.132995e+04, 2.65419, -4586.39996])

## Reaction 1751
#reaction('TQC4H7OHI + O2 <=> TQC4H7OHIO2', [3.490000e+14, -0.816, -536.49])

## Reaction 1752
#reaction('IQC4H7OHT + O2 <=> IQC4H7OHTO2', [6.870000e+16, -1.627, 198.7])

## Reaction 1753
#reaction('SC4H7OH-I + H <=> SC4H7OH-IP + H2', [7.290000e+05, 2.455, 4361.2])

## Reaction 1754
#reaction('SC4H7OH-I + O <=> SC4H7OH-IP + OH', [1.050000e+12, 0.7, 5884.0])

## Reaction 1755
#reaction('SC4H7OH-I + OH <=> SC4H7OH-IP + H2O', [4.395017e+04, 2.67841, -827.103])

## Reaction 1756
#reaction('SC4H7OH-I + HO2 <=> SC4H7OH-IP + H2O2', [2.920000e-01, 4.12, 12802.0])

## Reaction 1757
#reaction('SC4H7OH-I + CH3 <=> SC4H7OH-IP + CH4', [4.420000e+00, 3.5, 5675.0])

## Reaction 1758
#reaction('SC4H7OH-I + CH3O <=> SC4H7OH-IP + CH3OH', [1.680000e+11, 0.0, 2600.0])

## Reaction 1759
#reaction('SC4H7OH-I + CH3O2 <=> SC4H7OH-IP + CH3O2H', [1.928000e+04, 2.6, 13910.0])

## Reaction 1760
#reaction('SC4H7OH-IP + H => CH3 + C3H4-A + OH', [4.020973e+04, 2.51815, 8847.53569])

## Reaction 1761
#reaction('SC4H7OH-I + HO2 <=> IC3H7CHO + HO2', [1.490000e+05, 1.67, 6810.0])

## Reaction 1762
#reaction('SC4H7OH-I + HOCHO <=> IC3H7CHO + HOCHO', [2.810000e-02, 3.286, -4509.0])

## Reaction 1763
#reaction('SC4H7OH-I <=> IC3H7CHO', [8.590000e+11, 0.318, 55900.0])

## Reaction 1764
#reaction('IC4H7OH + OH <=> IC4H6OH + H2O', [1.260000e+03, 2.97, -2660.593])

## Reaction 1765
#reaction('IC4H7OH + HO2 <=> IC4H6OH + H2O2', [1.450000e-05, 5.26, 8267.907])

## Reaction 1766
#reaction('IC4H6OH + H2 <=> IC4H7OH + H', [2.160000e+04, 2.38, 18990.0])

## Reaction 1767
#reaction('IC4H7OH + O2 <=> IC4H6OH + HO2', [6.000000e+13, 0.0, 39900.0])

## Reaction 1768
#reaction('IC4H6OH + CH2O <=> IC4H7OH + HCO', [6.300000e+08, 1.9, 18190.0])

## Reaction 1769
#reaction('IC4H8 + OH <=> IC4H7 + H2O', [5.274024e+04, 2.7, -827.17])

## Reaction 1770
#reaction('IC4H8 + OH <=> IC4H7-I1 + H2O', [1.093009e+04, 2.81477, 1114.19989])

## Reaction 1771
#reaction('IC4H8 + O <=> IC3H7 + HCO', [7.450000e+06, 1.88, 183.0])

## Reaction 1772
#reaction('IC4H8 + O => CH2CO + CH3 + CH3', [3.050000e+06, 1.88, 183.0])

## Reaction 1773
#reaction('IC4H8 + O => IC3H6CO + H + H', [3.050000e+06, 1.88, 183.0])

## Reaction 1774
#reaction('IC4H8 + O <=> IC4H7 + OH', [1.050000e+12, 0.7, 5884.0])

## Reaction 1775
#reaction('IC4H8 + O <=> IC4H7-I1 + OH', [1.200000e+11, 0.7, 8959.1])

## Reaction 1776
#reaction('IC4H8 + CH3 <=> IC4H7 + CH4', [3.200000e+12, 0.0, 10000.0])

## Reaction 1777
#reaction('IC4H8 + CH3 <=> IC4H7-I1 + CH4', [2.000000e+12, 0.0, 15000.0])

## Reaction 1778
#reaction('IC4H8 + O2 <=> IC4H7 + HO2', [3.120000e+13, 0.0, 37450.0])

## Reaction 1779
#reaction('IC4H8 + O2 <=> IC4H7-I1 + HO2', [2.000000e+13, 0.0, 62270.0])

## Reaction 1780
#reaction('IC4H8 + HO2 <=> IC4H7 + H2O2', [2.920000e-01, 4.12, 12802.0])

## Reaction 1781
#reaction('IC4H8 + HO2 <=> IC4H7-I1 + H2O2', [9.703894e+04, 2.54892, 24733.17153])

## Reaction 1782
#pdep_arrhenius('IC4H7 + CH3O2 <=> IC4H7O + CH3O',
            #[(0.01, 'atm'), 3.330000e+12, -0.158, -1417.0],
            #[(0.1, 'atm'), 1.660000e+14, -0.642, -349.1],
            #[(1.0, 'atm'), 2.595000e+17, -1.52, 2379.2],
            #[(10.0, 'atm'), 9.780000e+14, -0.684, 3615.3],
            #[(100.0, 'atm'), 5.470000e+03, 2.74, 1144.4])

## Reaction 1783
#pdep_arrhenius('IC4H7 + CH3O2 <=> IC4H7OOCH3',
            #[(0.01, 'atm'), 1.910000e+31, -7.23, 1336.2],
            #[(0.1, 'atm'), 6.310000e+42, -10.3, 5568.9],
            #[(1.0, 'atm'), 1.030000e+45, -10.6, 7851.5],
            #[(10.0, 'atm'), 2.790000e+37, -7.92, 6497.9],
            #[(100.0, 'atm'), 3.400000e+29, -5.28, 4539.8])

## Reaction 1784
#pdep_arrhenius('IC4H7OOCH3 <=> IC4H7O + CH3O',
            #[(0.01, 'atm'), 1.490000e+58, -13.9, 54266.9],
            #[(0.1, 'atm'), 1.800000e+54, -12.4, 54193.8],
            #[(1.0, 'atm'), 3.360000e+46, -9.81, 52468.5],
            #[(10.0, 'atm'), 2.390000e+36, -6.54, 49429.0],
            #[(100.0, 'atm'), 2.560000e+27, -3.61, 46333.1])

## Reaction 1785
#reaction('IC4H8 + CH3O <=> IC4H7 + CH3OH', [1.680000e+11, 0.0, 2600.0])

## Reaction 1786
#reaction('IC4H8 + H <=> IC4H7 + H2', [7.290000e+05, 2.455, 4361.2])

## Reaction 1787
#reaction('IC4H8 + H <=> IC4H7-I1 + H2', [8.621000e+02, 3.25, 12166.98])

## Reaction 1788
#reaction('C3H4-A + CH3 <=> IC4H7', [4.020970e+04, 2.5, 8847.5])

## Reaction 1789
#pdep_arrhenius('IC4H7 + HO2 <=> IC4H7O + OH',
            #[(0.01, 'atm'), 1.020000e+13, -0.158, -1417.0],
            #[(0.1, 'atm'), 4.980000e+14, -0.642, -349.1],
            #[(1.0, 'atm'), 7.770000e+17, -1.52, 2379.2],
            #[(10.0, 'atm'), 2.930000e+15, -0.684, 3615.3],
            #[(100.0, 'atm'), 1.640000e+04, 2.74, 1144.4])

## Reaction 1790
#pdep_arrhenius('IC4H7 + HO2 <=> IC4H7OOH',
            #[(0.01, 'atm'), 1.910000e+31, -7.23, 1336.2],
            #[(0.1, 'atm'), 6.310000e+42, -10.3, 5568.9],
            #[(1.0, 'atm'), 1.030000e+45, -10.6, 7851.5],
            #[(10.0, 'atm'), 2.790000e+37, -7.92, 6497.9],
            #[(100.0, 'atm'), 1.440000e+32, -6.01, 6053.6])

## Reaction 1791
#pdep_arrhenius('IC4H7 + HO2 <=> IC3H5CHO + H2O',
            #[(0.01, 'atm'), 1.090000e+00, 3.01, -3421.1],
            #[(0.1, 'atm'), 6.350000e+01, 2.5, -2341.4],
            #[(1.0, 'atm'), 6.050000e+05, 1.39, 595.1],
            #[(10.0, 'atm'), 3.100000e+05, 1.59, 2677.6],
            #[(100.0, 'atm'), 5.070000e-05, 4.59, 927.5])

## Reaction 1792
#pdep_arrhenius('IC4H7OOH <=> IC3H5CHO + H2O',
            #[(0.01, 'atm'), 1.990000e+50, -12.7, 53531.9],
            #[(0.1, 'atm'), 4.720000e+47, -11.5, 54360.9],
            #[(1.0, 'atm'), 1.500000e+40, -8.84, 53179.2],
            #[(10.0, 'atm'), 2.540000e+28, -5.0, 49919.4],
            #[(100.0, 'atm'), 1.480000e+16, -1.12, 45949.3])

## Reaction 1793
#pdep_arrhenius('IC4H7OOH <=> IC4H7O + OH',
            #[(0.01, 'atm'), 1.490000e+58, -13.9, 54266.9],
            #[(0.1, 'atm'), 1.800000e+54, -12.4, 54193.8],
            #[(1.0, 'atm'), 3.360000e+46, -9.81, 52468.5],
            #[(10.0, 'atm'), 2.390000e+36, -6.54, 49429.0],
            #[(100.0, 'atm'), 1.280000e+27, -3.61, 46333.1])

## Reaction 1794
#pdep_arrhenius('IC4H7O <=> C3H5-T + CH2O',
            #[(0.001, 'atm'), 7.260000e+06, 0.182, 17815.5],
            #[(0.01, 'atm'), 6.970000e+16, -2.5, 20878.7],
            #[(0.1, 'atm'), 6.640000e+23, -4.23, 23565.0],
            #[(1.0, 'atm'), 1.070000e+26, -4.56, 24622.9],
            #[(10.0, 'atm'), 6.500000e+29, -5.37, 26645.0],
            #[(100.0, 'atm'), 4.630000e+31, -5.59, 28915.3],
            #[(1000.0, 'atm'), 8.520000e+25, -3.61, 27863.4])

## Reaction 1795
#pdep_arrhenius('IC4H7O <=> IC3H5OCH2',
            #[(0.001, 'atm'), 3.170000e+20, -4.15, 12121.3],
            #[(0.01, 'atm'), 4.790000e+24, -5.03, 14606.1],
            #[(0.1, 'atm'), 1.900000e+26, -5.16, 16124.4],
            #[(1.0, 'atm'), 1.510000e+28, -5.4, 18165.4],
            #[(10.0, 'atm'), 2.420000e+28, -5.17, 19691.2],
            #[(100.0, 'atm'), 5.570000e+24, -3.86, 19395.2],
            #[(1000.0, 'atm'), 1.350000e+18, -1.73, 17386.5])

## Reaction 1796
#pdep_arrhenius('IC4H7O <=> IC3H6CHO',
            #[(0.001, 'atm'), 5.250000e-49, 15.5, -15639.9],
            #[(0.01, 'atm'), 1.460000e-88, 27.6, -35995.0],
            #[(0.1, 'atm'), 4.440000e-22, 8.38, -3819.0],
            #[(1.0, 'atm'), 6.230000e+12, -1.44, 10829.2],
            #[(10.0, 'atm'), 3.480000e+42, -9.91, 25297.9],
            #[(100.0, 'atm'), 1.880000e+38, -8.16, 25974.5],
            #[(1000.0, 'atm'), 1.670000e+21, -2.74, 20337.7])

## Reaction 1797
#pdep_arrhenius('IC4H7O <=> IC3H5CHO + H',
            #[(0.001, 'atm'), 3.000000e+15, -2.31, 14667.9],
            #[(0.01, 'atm'), 1.500000e+22, -3.96, 18283.0],
            #[(0.1, 'atm'), 1.950000e+23, -3.99, 19143.3],
            #[(1.0, 'atm'), 1.150000e+25, -4.24, 20311.2],
            #[(10.0, 'atm'), 1.760000e+28, -4.89, 22765.2],
            #[(100.0, 'atm'), 1.410000e+27, -4.28, 23770.6],
            #[(1000.0, 'atm'), 2.570000e+20, -2.06, 22040.1])

## Reaction 1798
#pdep_arrhenius('IC4H7O <=> C3H6 + HCO',
            #[(0.001, 'atm'), 6.620000e+16, -2.84, 13197.0],
            #[(0.01, 'atm'), 1.260000e+20, -3.53, 15469.2],
            #[(0.1, 'atm'), 2.130000e+21, -3.64, 16584.5],
            #[(1.0, 'atm'), 1.070000e+24, -4.16, 18985.0],
            #[(10.0, 'atm'), 8.420000e+25, -4.4, 22382.6],
            #[(100.0, 'atm'), 1.860000e+21, -2.73, 23658.8],
            #[(1000.0, 'atm'), 4.750000e+08, 1.14, 20922.5])

## Reaction 1799
#pdep_arrhenius('IC3H5OCH2 <=> C3H5-T + CH2O',
            #[(0.001, 'atm'), 3.300000e+09, -0.638, 19747.8],
            #[(0.01, 'atm'), 3.360000e+21, -3.9, 23945.2],
            #[(0.1, 'atm'), 2.910000e+29, -5.9, 27249.7],
            #[(1.0, 'atm'), 1.830000e+34, -6.94, 30690.4],
            #[(10.0, 'atm'), 9.720000e+33, -6.5, 33002.5],
            #[(100.0, 'atm'), 2.680000e+27, -4.26, 33305.6],
            #[(1000.0, 'atm'), 8.810000e+14, -0.326, 31553.1])

## Reaction 1800
#pdep_arrhenius('IC3H5OCH2 <=> IC3H6CHO',
            #[(0.001, 'atm'), 8.010000e-92, 27.8, -37321.2],
            #[(0.01, 'atm'), 7.780000e-11, 3.7, -2766.9],
            #[(0.1, 'atm'), 5.110000e+15, -2.76, 15937.6],
            #[(1.0, 'atm'), 4.480000e+25, -5.2, 21532.2],
            #[(10.0, 'atm'), 3.970000e+34, -7.41, 28116.9],
            #[(100.0, 'atm'), 5.620000e+22, -3.56, 25806.7],
            #[(1000.0, 'atm'), 2.510000e+20, -2.63, 29288.4])

## Reaction 1801
#pdep_arrhenius('IC3H5OCH2 <=> IC3H5CHO + H',
            #[(0.001, 'atm'), 4.930000e+24, -5.05, 20108.4],
            #[(0.01, 'atm'), 2.140000e+28, -5.8, 22219.4],
            #[(0.1, 'atm'), 1.930000e+32, -6.64, 25108.2],
            #[(1.0, 'atm'), 8.600000e+34, -7.11, 28209.1],
            #[(10.0, 'atm'), 2.170000e+34, -6.64, 30647.6],
            #[(100.0, 'atm'), 4.170000e+28, -4.71, 31231.9],
            #[(1000.0, 'atm'), 3.980000e+18, -1.62, 30129.8])

## Reaction 1802
#pdep_arrhenius('IC3H5OCH2 <=> C3H6 + HCO',
            #[(0.001, 'atm'), 8.230000e+26, -5.84, 19356.9],
            #[(0.01, 'atm'), 1.320000e+29, -6.21, 21293.6],
            #[(0.1, 'atm'), 3.470000e+32, -6.96, 24197.3],
            #[(1.0, 'atm'), 1.440000e+36, -7.76, 28007.8],
            #[(10.0, 'atm'), 9.720000e+37, -8.02, 32394.6],
            #[(100.0, 'atm'), 2.430000e+31, -5.81, 34295.8],
            #[(1000.0, 'atm'), 3.730000e+14, -0.726, 32008.3])

## Reaction 1803
#pdep_arrhenius('IC3H6CHO <=> C3H5-T + CH2O',
            #[(0.001, 'atm'), 6.890000e-69, 21.5, 2638.0],
            #[(0.01, 'atm'), 5.340000e-33, 11.1, 16749.1],
            #[(0.1, 'atm'), 6.110000e+26, -6.01, 44116.7],
            #[(1.0, 'atm'), 8.040000e+35, -8.31, 46919.7],
            #[(10.0, 'atm'), 5.520000e+40, -9.19, 50508.7],
            #[(100.0, 'atm'), 5.850000e+35, -7.18, 52038.4],
            #[(1000.0, 'atm'), 1.930000e+19, -1.94, 48440.0])

## Reaction 1804
#pdep_arrhenius('IC3H6CHO <=> IC3H5CHO + H',
            #[(0.001, 'atm'), 1.610000e+10, -1.24, 32371.3],
            #[(0.01, 'atm'), 5.840000e+15, -2.61, 32878.4],
            #[(0.1, 'atm'), 3.640000e+23, -4.6, 34275.3],
            #[(1.0, 'atm'), 7.580000e+31, -6.63, 37895.4],
            #[(10.0, 'atm'), 2.860000e+32, -6.3, 39990.7],
            #[(100.0, 'atm'), 1.570000e+23, -3.14, 38011.7],
            #[(1000.0, 'atm'), 4.520000e+12, 0.214, 34570.5])

## Reaction 1805
#pdep_arrhenius('IC3H6CHO <=> C3H6 + HCO',
            #[(0.001, 'atm'), 2.900000e+32, -7.24, 25687.5],
            #[(0.01, 'atm'), 5.300000e+33, -7.28, 27100.6],
            #[(0.1, 'atm'), 2.000000e+35, -7.41, 29027.3],
            #[(1.0, 'atm'), 1.010000e+34, -6.7, 30018.1],
            #[(10.0, 'atm'), 9.760000e+27, -4.63, 28923.9],
            #[(100.0, 'atm'), 2.110000e+19, -1.85, 26239.8],
            #[(1000.0, 'atm'), 1.590000e+13, 0.063, 24086.3])

## Reaction 1806
#pdep_arrhenius('C3H5-T + CH2O <=> IC3H5CHO + H',
            #[(0.001, 'atm'), 2.600000e+04, 2.26, 1510.3],
            #[(0.01, 'atm'), 5.130000e+04, 2.17, 1675.5],
            #[(0.1, 'atm'), 3.990000e+05, 1.91, 2218.3],
            #[(1.0, 'atm'), 1.750000e+07, 1.45, 3428.0],
            #[(10.0, 'atm'), 1.350000e+09, 0.933, 5173.0],
            #[(100.0, 'atm'), 2.240000e+11, 0.357, 8001.3],
            #[(1000.0, 'atm'), 6.010000e+05, 2.09, 7895.6])

## Reaction 1807
#pdep_arrhenius('C3H5-T + CH2O <=> C3H6 + HCO',
            #[(0.001, 'atm'), 1.110000e+07, 1.09, 1807.2],
            #[(0.01, 'atm'), 2.470000e+07, 0.993, 1994.9],
            #[(0.1, 'atm'), 2.470000e+08, 0.704, 2596.2],
            #[(1.0, 'atm'), 1.420000e+10, 0.209, 3934.2],
            #[(10.0, 'atm'), 3.450000e+13, -0.726, 6944.3],
            #[(100.0, 'atm'), 3.310000e+14, -0.866, 10965.7],
            #[(1000.0, 'atm'), 1.650000e+01, 3.17, 9399.8])

## Reaction 1808
#reaction('IC3H5CHO + OH <=> IC3H5CO + H2O', [6.132990e+04, 2.65, -4586.4])

## Reaction 1809
#reaction('IC3H5CHO + O <=> IC3H5CO + OH', [7.180000e+12, 0.0, 1389.0])

## Reaction 1810
#reaction('IC3H5CHO + HO2 <=> IC3H5CO + H2O2', [1.177300e-04, 4.91966, 3684.27444])

## Reaction 1811
#reaction('IC3H5CHO + CH3 <=> IC3H5CO + CH4', [1.248790e+00, 3.63386, 4328.93484])

## Reaction 1812
#reaction('IC3H5CHO + O2 <=> IC3H5CO + HO2', [2.000000e+13, 0.0, 40700.0])

## Reaction 1813
#reaction('IC3H5CO <=> C3H5-T + CO', [1.278000e+20, -1.89, 34460.0])

## Reaction 1814
#reaction('IC3H5CHO + H <=> IC3H5CO + H2', [7.147404e+05, 2.35674, 1577.16272])

## Reaction 1815
#reaction('IC3H5CHO + OH <=> IC3H4CHO-A + H2O', [4.460000e+06, 2.072, 1050.8])

## Reaction 1816
#reaction('IC3H5CHO + O <=> IC3H4CHO-A + OH', [5.240000e+11, 0.7, 5884.0])

## Reaction 1817
#reaction('IC3H5CHO + HO2 <=> IC3H4CHO-A + H2O2', [3.070000e-02, 4.403, 13547.2])

## Reaction 1818
#reaction('IC3H5CHO + CH3 <=> IC3H4CHO-A + CH4', [2.210000e+00, 3.5, 5675.0])

## Reaction 1819
#reaction('IC3H5CHO + O2 <=> IC3H4CHO-A + HO2', [5.960000e+19, -1.67, 46192.1])

## Reaction 1820
#reaction('IC3H5CHO + H <=> IC3H4CHO-A + H2', [3.640000e+05, 2.455, 4361.2])

## Reaction 1821
#reaction('C3H4-A + HCO <=> IC3H4CHO-A', [4.020973e+04, 2.51815, 8847.53569])

## Reaction 1822
#reaction('IC4H7 + O2 <=> IC4H7O2', [1.090000e+10, 0.56725, 2290.0])

## Reaction 1823
#reaction('IC4H7O2 <=> IC4H7O + O', [1.820000e+14, 0.0, 60620.0])

## Reaction 1824
#reaction('IC4H7O2 <=> IC3H5CHO + OH', [1.520000e+09, 1.02524, 39460.0])

## Reaction 1825
#reaction('IC4H7O2 <=> IC4H6OOH-I', [1.410000e+05, 1.83586, 19820.0])

## Reaction 1826
#reaction('IC4H6OOH-I <=> C*CYCCOC + OH', [6.820000e+12, -0.34545, 42140.0])

## Reaction 1827
#reaction('IC4H6OOH-I <=> C3H4-A + CH2O2H', [1.180000e+12, 0.91203, 51390.0])

## Reaction 1828
#reaction('IC4H6OOH-I <=> IC3H5CHO + OH', [2.420000e+09, 0.8739, 54090.0])

## Reaction 1829
#reaction('IC4H7O2 <=> CCYCCOOC-T1', [1.190000e+08, 0.80412, 28020.0])

## Reaction 1830
#reaction('CCYCCOOC-T1 <=> CCYCCOOC-I2', [3.560000e+13, 0.0, 38820.0])

## Reaction 1831
#reaction('CCYCCOOC-I2 <=> CHOIC3H6O', [3.190000e+14, 0.0, 3000.0])

## Reaction 1832
#reaction('CHOIC3H6O <=> CH3CHCHO + CH2O', [4.300000e+12, 0.0, 9780.0])

## Reaction 1833
#reaction('CCYCCOOC-T1 <=> CCY(C2O)CO', [4.560000e+11, 0.92729, 17470.0])

## Reaction 1834
#reaction('CCY(C2O)CO <=> CCYCCO-T1 + CH2O', [1.740000e+13, 0.0, 18150.0])

## Reaction 1835
#reaction('IC4H7O2 <=> C2CYCOOC-I1', [1.070000e+08, 0.89161, 29720.0])

## Reaction 1836
#reaction('C2CYCOOC-I1 <=> IC3H5OOCH2', [4.410000e+13, -0.22618, 18500.0])

## Reaction 1837
#reaction('IC3H5OOCH2 <=> CH3COCH2 + CH2O', [1.410000e+10, 0.0, 1000.0])

## Reaction 1838
#reaction('C2CYCOOC-I1 <=> CCY(C2O)CO', [1.210000e+13, 0.1018, 20320.0])

## Reaction 1839
#reaction('IC4H7O2 + IC4H8 <=> IC4H7 + IC4H7OOH', [1.460000e-01, 4.12, 12802.0])

## Reaction 1840
#pdep_arrhenius('IC4H7O2 + IC4H7 <=> IC4H7O + IC4H7O',
            #[(0.01, 'atm'), 2.550000e+12, -0.158, -1417.0],
            #[(0.1, 'atm'), 1.250000e+14, -0.642, -349.1],
            #[(1.0, 'atm'), 1.940000e+17, -1.52, 2379.2],
            #[(10.0, 'atm'), 7.330000e+14, -0.684, 3615.3],
            #[(100.0, 'atm'), 4.100000e+03, 2.74, 1144.4])

## Reaction 1841
#pdep_arrhenius('IC4H7O2 + IC4H7 <=> IC4H7OOIC4H7',
            #[(0.01, 'atm'), 4.780000e+30, -7.23, 1336.2],
            #[(0.1, 'atm'), 1.580000e+42, -10.3, 5568.9],
            #[(1.0, 'atm'), 2.580000e+44, -10.6, 7851.5],
            #[(10.0, 'atm'), 6.980000e+36, -7.92, 6497.9],
            #[(100.0, 'atm'), 3.600000e+31, -6.01, 6053.6])

## Reaction 1842
#pdep_arrhenius('IC4H7OOIC4H7 <=> IC4H7O + IC4H7O',
            #[(0.01, 'atm'), 3.730000e+57, -13.9, 54266.9],
            #[(0.1, 'atm'), 4.500000e+53, -12.4, 54193.8],
            #[(1.0, 'atm'), 8.400000e+45, -9.81, 52468.5],
            #[(10.0, 'atm'), 5.980000e+35, -6.54, 49429.0],
            #[(100.0, 'atm'), 3.200000e+26, -3.61, 46333.1])

## Reaction 1843
#reaction('TC4H9O2 + IC4H7 <=> TC4H9O + IC4H7O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1844
#reaction('IC4H7O2 + IC4H7O2 <=> IC4H7O + IC4H7O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 1845
#falloff_reaction('IC4H7 + CH3 (+ M) <=> AC5H10 (+ M)',
                #kf=[1.500000e+14, -0.32, -262.3],
                #kf0=[5.865000e+60, -12.81, 6250.0],
                #efficiencies='H2:2.0 H2O:6.0 CH4:2.0 CO:1.5 CO2:2.0 C2H6:3.0 AR:0.7',
                #falloff=Troe(A=0.104, T3=1606.0, T1=60000.0, T2=6118.0))

## Reaction 1846
#pdep_arrhenius('IC4H7 + IC4H7 <=> C3H4-A + AC5H10',
            #[(1.0, 'atm'), 4.100000e+40, -9.3, 12470.0],
            #[(4.0, 'atm'), 3.410000e+32, -6.8, 9180.0],
            #[(10.0, 'atm'), 1.260000e+28, -5.5, 7410.0])

## Reaction 1847
#pdep_arrhenius('IC4H7 + IC4H7 <=> H15DE25DM',
            #[(0.039, 'atm'), 3.270000e+64, -15.935, 20230.0],
            #[(0.078, 'atm'), 7.500000e+59, -14.49, 18600.0],
            #[(0.156, 'atm'), 1.140000e+55, -12.995, 16700.0])

## Reaction 1848
#reaction('AC5H10 + OH <=> AC5H9-A2 + H2O', [3.120000e+06, 2.0, -298.0])

## Reaction 1849
#reaction('AC5H10 + H <=> AC5H9-A2 + H2', [1.730000e+05, 2.5, 2492.0])

## Reaction 1850
#reaction('AC5H10 + CH3 <=> AC5H9-A2 + CH4', [2.210000e+00, 3.5, 5675.0])

## Reaction 1851
#reaction('AC5H10 + O <=> AC5H9-A2 + OH', [3.700000e+05, 2.56, -1130.0])

## Reaction 1852
#reaction('AC5H10 + OH <=> SC4H9 + CH2O', [2.000000e+10, 0.0, 4000.0])

## Reaction 1853
#reaction('AC5H10 + O <=> SC4H9 + HCO', [7.230000e+05, 2.34, -1050.0])

## Reaction 1854
#reaction('AC5H10 + O <=> IC3H7 + CH3CO', [7.230000e+05, 2.34, -1050.0])

## Reaction 1855
#reaction('AC5H10 + O <=> IC4H9 + HCO', [7.230000e+05, 2.34, -1050.0])

## Reaction 1856
#reaction('AC5H10 <=> C3H5-T + C2H5', [8.922000e+24, -2.409, 100500.0])

## Reaction 1857
#reaction('AC5H9-A2 <=> C3H4-A + C2H5', [1.983000e+20, -1.63, 59240.0])

## Reaction 1858
#reaction('AC5H10 + H <=> AC5H9-C + H2', [3.376000e+05, 2.36, 207.0])

## Reaction 1859
#reaction('AC5H10 + H <=> AC5H9-D + H2', [1.950000e+06, 2.4, 4471.0])

## Reaction 1860
#reaction('AC5H10 + OH <=> AC5H9-C + H2O', [2.764000e+04, 2.64, -1919.0])

## Reaction 1861
#reaction('AC5H10 + OH <=> AC5H9-D + H2O', [7.010000e+07, 1.61, -35.0])

## Reaction 1862
#reaction('AC5H10 + CH3 <=> AC5H9-C + CH4', [3.690000e+00, 3.31, 4002.0])

## Reaction 1863
#reaction('AC5H10 + CH3 <=> AC5H9-D + CH4', [2.270000e+00, 3.46, 5481.0])

## Reaction 1864
#reaction('AC5H10 + HO2 <=> AC5H9-A2 + H2O2', [9.639000e+03, 2.6, 13910.0])

## Reaction 1865
#reaction('AC5H10 + HO2 <=> AC5H9-C + H2O2', [4.820000e+03, 2.55, 10530.0])

## Reaction 1866
#reaction('AC5H10 + HO2 <=> AC5H9-D + H2O2', [1.450000e+04, 2.6, 13910.0])

## Reaction 1867
#reaction('AC5H10 + CH3O2 <=> AC5H9-A2 + CH3O2H', [9.639000e+03, 2.6, 13910.0])

## Reaction 1868
#reaction('AC5H10 + CH3O2 <=> AC5H9-C + CH3O2H', [4.820000e+03, 2.55, 10530.0])

## Reaction 1869
#reaction('AC5H10 + CH3O2 <=> AC5H9-D + CH3O2H', [1.450000e+04, 2.6, 13910.0])

## Reaction 1870
#reaction('AC5H10 + CH3O <=> AC5H9-A2 + CH3OH', [9.000000e+01, 2.95, 11990.0])

## Reaction 1871
#reaction('AC5H10 + CH3O <=> AC5H9-C + CH3OH', [4.000000e+01, 2.9, 8609.0])

## Reaction 1872
#reaction('AC5H10 + CH3O <=> AC5H9-D + CH3OH', [2.180000e+11, 0.0, 4571.0])

## Reaction 1873
#reaction('H15DE25DM + H <=> H15DE25DM-S + H2', [1.420000e+06, 2.315165407, 2075.955844])

## Reaction 1874
#reaction('H15DE25DM + H <=> H15DE25DM-A + H2', [7.290000e+05, 2.455, 4361.2])

## Reaction 1875
#reaction('H15DE25DM + HO2 <=> H15DE25DM-S + H2O2', [1.000000e+00, 3.887366896, 9386.357213])

## Reaction 1876
#reaction('H15DE25DM + HO2 <=> H15DE25DM-A + H2O2', [2.920000e-01, 4.12, 12802.0])

## Reaction 1877
#reaction('H15DE25DM + CH3 <=> H15DE25DM-S + CH4', [1.000000e+00, 3.649020008, 3374.05195])

## Reaction 1878
#reaction('H15DE25DM + CH3 <=> H15DE25DM-A + CH4', [4.420000e+00, 3.5, 5675.0])

## Reaction 1879
#reaction('H15DE25DM + O2 <=> H15DE25DM-S + HO2', [4.000000e+14, 0.0, 38890.0])

## Reaction 1880
#reaction('H15DE25DM + O2 <=> H15DE25DM-A + HO2', [1.860000e+09, 1.301, 40939.0])

## Reaction 1881
#reaction('H15DE25DM + O <=> H15DE25DM-S + OH', [1.592430e+11, 0.7, 13.0])

## Reaction 1882
#reaction('H15DE25DM + O <=> H15DE25DM-A + OH', [1.050000e+12, 0.7, 5884.0])

## Reaction 1883
#reaction('H15DE25DM + OH <=> H15DE25DM-S + H2O', [8.950000e+04, 2.636965672, -6.0])

## Reaction 1884
#reaction('H15DE25DM + OH <=> H15DE25DM-A + H2O', [2.530000e+05, 2.46, 729.44])

## Reaction 1885
#reaction('C3H5-T + B13DE2M <=> H15DE25DM-S', [7.070000e+03, 2.48, 6130.0])

## Reaction 1886
#reaction('C3H4-A + AC5H9-D <=> H15DE25DM-A', [2.840000e+04, 2.5, 8847.5])

## Reaction 1887
#reaction('AC5H9-D <=> C3H5-T + C2H4', [1.220000e+12, 0.64, 29370.0])

## Reaction 1888
#reaction('C2H3 + C3H5-T <=> B13DE2M', [8.000000e+12, 0.0, 0.0])

## Reaction 1889
#reaction('H + B13DE2MJ <=> B13DE2M', [1.000000e+14, 0.0, 0.0])

## Reaction 1890
#reaction('B13DE2M + HO2 <=> B13DE2MJ + H2O2', [3.070000e-02, 4.403, 13547.2])

## Reaction 1891
#reaction('B13DE2M + H <=> B13DE2MJ + H2', [3.644000e+05, 2.455, 4361.2])

## Reaction 1892
#reaction('B13DE2M + O2 <=> B13DE2MJ + HO2', [5.960000e+19, -1.67, 46192.1])

## Reaction 1893
#reaction('B13DE2M + CH3 <=> B13DE2MJ + CH4', [2.210000e+00, 3.5, 5675.0])

## Reaction 1894
#reaction('B13DE2M + CH3O2 <=> B13DE2MJ + CH3O2H', [7.680000e-02, 4.403, 13547.2])

## Reaction 1895
#reaction('B13DE2M + OH <=> B13DE2MJ + H2O', [4.460000e+06, 2.072, 1050.8])

## Reaction 1896
#reaction('C2H3 + C3H4-A <=> B13DE2MJ', [9.450000e+02, 2.67, 6850.0])

## Reaction 1897
#reaction('C2H3 + C3H6 <=> CC5H9-A', [9.450000e+02, 2.67, 6850.0])

## Reaction 1898
#reaction('B13DE2M + H <=> AC5H9-D', [2.500000e+11, 0.51, 2620.0])

## Reaction 1899
#reaction('B13DE2M + H <=> AC5H9-C', [4.000000e+13, 0.0, 1300.0])

## Reaction 1900
#reaction('C4H612 + CH3 <=> AC5H9-C', [1.760000e+04, 2.48, 6130.0])

## Reaction 1901
#reaction('B13DE2M + H <=> CC5H9-B', [4.240000e+11, 0.51, 1230.0])

## Reaction 1902
#reaction('B12DE3M + H <=> CC5H9-B', [4.240000e+11, 0.51, 1230.0])

## Reaction 1903
#reaction('B13DE2M + H <=> CC5H9-A', [2.500000e+11, 0.51, 2620.0])

## Reaction 1904
#reaction('AC5H9-C + HO2 <=> AC5H9O-C + OH', [7.000000e+12, 0.0, -1000.0])

## Reaction 1905
#reaction('AC5H9-C + CH3O2 <=> AC5H9O-C + CH3O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1906
#reaction('CH3CHO + C3H5-T <=> AC5H9O-C', [9.450000e+02, 2.67, 6850.0])

## Reaction 1907
#reaction('AC5H9-C + HO2 <=> B2E2M1OJ + OH', [7.000000e+12, 0.0, -1000.0])

## Reaction 1908
#reaction('AC5H9-C + CH3O2 <=> B2E2M1OJ + CH3O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1909
#reaction('CH2O + C4H72-2 <=> B2E2M1OJ', [8.800000e+03, 2.48, 6100.0])

## Reaction 1910
#reaction('C2H3 + CH3COCH3 <=> CC5H9O-B', [9.450000e+02, 2.67, 6850.0])

## Reaction 1911
#reaction('CC5H9-B + HO2 <=> B2E3M1OJ + OH', [7.000000e+12, 0.0, -1000.0])

## Reaction 1912
#reaction('CC5H9-B + CH3O2 <=> B2E3M1OJ + CH3O', [7.000000e+12, 0.0, -1000.0])

## Reaction 1913
#reaction('CH2O + IC4H7-I1 <=> B2E3M1OJ', [8.800000e+03, 2.48, 6100.0])

## Reaction 1914
#reaction('H15DE25DM-A + HO2 <=> H15DE25DM-AO + OH', [7.000000e+12, 0.0, -1000.0])

## Reaction 1915
#reaction('H15DE2M-T + CH2O <=> H15DE25DM-AO', [2.500000e+10, 0.0, 4786.58])

## Reaction 1916
#reaction('IC4H7 + C3H4-A <=> H15DE2M-T', [8.800000e+03, 2.48, 6130.0])

## Reaction 1917
#reaction('H15DE25DM-S + HO2 <=> H15DE25DM-SO + OH', [7.000000e+12, 0.0, -1000.0])

## Reaction 1918
#reaction('IC3H5CHO + IC4H7 <=> H15DE25DM-SO', [2.500000e+10, 0.0, 6329.74])

## Reaction 1919
#reaction('C3H5-T + IC4H7CHO <=> H15DE25DM-SO', [2.500000e+10, 0.0, 15526.42])

## Reaction 1920
#reaction('IC4H7CHO + OH => CH2CHO + C3H4-A + H2O', [5.064000e+07, 2.46, 729.44])

## Reaction 1921
#reaction('IC4H8 + O2CHO <=> IC4H7 + HO2CHO', [1.928000e+04, 2.6, 13910.0])

## Reaction 1922
#pdep_arrhenius('IC4H7 <=> IC4H7-I1',
            #[(0.1, 'atm'), 1.300000e+55, -14.53, 73800.0],
            #[(1.0, 'atm'), 5.000000e+51, -13.02, 73300.0],
            #[(10.0, 'atm'), 9.700000e+48, -11.73, 73700.0],
            #[(100.0, 'atm'), 4.860000e+44, -9.84, 73400.0])

## Reaction 1923
#reaction('IC4H7-I1 + H <=> C3H4-A + CH4', [3.333000e+12, 0.0, 0.0])

## Reaction 1924
#reaction('IC4H7-I1 + H <=> C3H4-P + CH4', [3.340000e+12, 0.0, 0.0])

## Reaction 1925
#reaction('IC4H7-I1 + O <=> C3H6 + HCO', [6.000000e+13, 0.0, 0.0])

## Reaction 1926
#reaction('IC4H7-I1 + OH => C3H6 + HCO + H', [5.000000e+12, 0.0, 0.0])

## Reaction 1927
#reaction('IC4H7-I1 + HO2 => C3H6 + HCO + OH', [2.000000e+13, 0.0, 0.0])

## Reaction 1928
#reaction('IC4H7-I1 + HCO <=> IC4H8 + CO', [9.000000e+13, 0.0, 0.0])

## Reaction 1929
#reaction('IC4H7-I1 + CH3 <=> C3H4-P + C2H6', [1.000000e+11, 0.0, 0.0])

## Reaction 1930
#reaction('IC4H7-I1 + O2 <=> CH3COCH3 + HCO', [3.100000e+31, -5.944, 5748.4])

## Reaction 1931
#reaction('IC4H7-I1 + O2 <=> TC3H6CHO + O', [5.380000e+18, -2.14, 5142.9])

## Reaction 1932
#reaction('IC4H7-I1 + O2 <=> IC3H5CHO + OH', [2.700000e+19, -2.14, 5142.9])

## Reaction 1933
#reaction('IC4H8 + C3H5-A <=> IC4H7 + C3H6', [7.940000e+11, 0.0, 20500.0])

## Reaction 1934
#reaction('IC4H8 + C3H5-S <=> IC4H7 + C3H6', [7.940000e+11, 0.0, 20500.0])

## Reaction 1935
#reaction('IC4H8 + C3H5-T <=> IC4H7 + C3H6', [7.940000e+11, 0.0, 20500.0])

## Reaction 1936
#reaction('IC4H8 + CH3O2 <=> IC4H7 + CH3O2H', [1.540000e-01, 4.403, 13547.2])

## Reaction 1937
#reaction('IC4H8 + CH3CO3 <=> IC4H7 + CH3CO3H', [1.540000e-01, 4.403, 13547.2])

## Reaction 1938
#reaction('IC4H7 + O <=> IC3H5CHO + H', [6.030000e+13, 0.0, 0.0])

## Reaction 1939
#reaction('IC4H7O <=> IC4H6OH', [1.391000e+11, 0.0, 15600.0])

## Reaction 1940
#reaction('IC4H6OH + IC4H8 <=> IC4H7OH + IC4H7', [4.700000e+02, 3.3, 19840.0])

## Reaction 1941
#reaction('IC4H6OH + H <=> IC4H7OH', [1.000000e+14, 0.0, 0.0])

## Reaction 1942
#reaction('C3H4-A + CH2OH <=> IC4H6OH', [1.000000e+11, 0.0, 9200.0])

## Reaction 1943
#reaction('IC4H7O + O2 <=> IC3H5CHO + HO2', [3.000000e+10, 0.0, 1649.0])

## Reaction 1944
#reaction('IC4H7O + HO2 <=> IC3H5CHO + H2O2', [3.000000e+11, 0.0, 0.0])

## Reaction 1945
#reaction('IC4H7O + CH3 <=> IC3H5CHO + CH4', [2.400000e+13, 0.0, 0.0])

## Reaction 1946
#reaction('IC4H7O + O <=> IC3H5CHO + OH', [6.000000e+12, 0.0, 0.0])

## Reaction 1947
#reaction('IC4H7O + OH <=> IC3H5CHO + H2O', [1.810000e+13, 0.0, 0.0])

## Reaction 1948
#reaction('IC4H7O + H <=> IC3H5CHO + H2', [1.990000e+13, 0.0, 0.0])

## Reaction 1949
#reaction('TC3H6OCHO + OH <=> TC3H6CHO + HO2', [2.018000e+17, -1.2, 21010.0])

## Reaction 1950
#reaction('TC3H6OCHO <=> CH3COCH3 + HCO', [3.980000e+13, 0.0, 9700.0])

## Reaction 1951
#reaction('IC3H5CHO + H <=> TC3H6CHO', [1.300000e+13, 0.0, 1200.0])

## Reaction 1952
#reaction('IC3H6CO + H <=> TC3H6CHO', [1.300000e+13, 0.0, 4800.0])

## Reaction 1953
#reaction('TC3H6CHO + H2 <=> IC3H7CHO + H', [2.160000e+05, 2.38, 18990.0])

## Reaction 1954
#reaction('IC4H7O + H <=> IC4H7OH', [4.000000e+13, 0.0, 0.0])

## Reaction 1955
#reaction('IC4H7OH + H <=> IC4H8OH', [1.000000e+13, 0.0, 1200.0])

## Reaction 1956
#reaction('IC4H7O + H2 <=> IC4H7OH + H', [9.050000e+06, 2.0, 17830.0])

## Reaction 1957
#reaction('IC4H7 + OH <=> IC4H7OH', [3.000000e+13, 0.0, 0.0])

## Reaction 1958
#reaction('IC4H7OH + HCO <=> IC4H7O + CH2O', [3.020000e+11, 0.0, 18160.0])

## Reaction 1959
#falloff_reaction('C3H5-A + CH3 (+ M) <=> C4H8-1 (+ M)',
                #kf=[1.000000e+14, -0.32, -262.3],
                #kf0=[3.910000e+60, -12.81, 6250.0],
                #efficiencies='H2:2.0 H2O:6.0 CH4:2.0 CO:1.5 CO2:2.0 C2H6:3.0 AR:0.7',
                #falloff=Troe(A=0.104, T3=1606.0, T1=60000.0, T2=6118.0))

## Reaction 1960
#falloff_reaction('C2H3 + C2H5 (+ M) <=> C4H8-1 (+ M)',
                #kf=[2.500000e+13, 0.0, 0.0],
                #kf0=[4.270000e+58, -11.94, 9769.8],
                #falloff=Troe(A=0.175, T3=1341.0, T1=60000.0, T2=10140.0))

## Reaction 1961
#falloff_reaction('C4H71-4 + H (+ M) <=> C4H8-1 (+ M)',
                #kf=[3.600000e+13, 0.0, 0.0],
                #kf0=[3.010000e+48, -9.32, 5833.6],
                #efficiencies='H2:2.0 H2O:6.0 CH4:2.0 CO:1.5 CO2:2.0 C2H6:3.0 AR:0.7',
                #falloff=Troe(A=0.498, T3=1314.0, T1=1314.0, T2=50000.0))

## Reaction 1962
#falloff_reaction('C4H71-3 + H (+ M) <=> C4H8-1 (+ M)',
                #kf=[1.000000e+14, 0.0, 0.0],
                #kf0=[6.650000e+59, -12.0, 5967.8],
                #efficiencies='H2:2.0 H2O:6.0 CH4:2.0 CO:1.5 CO2:2.0 C2H6:3.0 AR:0.7',
                #falloff=Troe(A=0.02, T3=1097.0, T1=1097.0, T2=6860.0))

## Reaction 1963
#reaction('C4H8-1 <=> C4H71-1 + H', [7.710000e+69, -16.09, 140000.0])

## Reaction 1964
#reaction('C4H8-1 <=> C4H71-2 + H', [5.620000e+71, -16.58, 139300.0])

## Reaction 1965
#reaction('C4H8-1 <=> C4H6 + H2', [2.000000e+13, 0.0, 65500.0])

## Reaction 1966
#falloff_reaction('C3H5-S + CH3 (+ M) <=> C4H8-2 (+ M)',
                #kf=[5.000000e+13, 0.0, 0.0],
                #kf0=[8.540000e+58, -11.94, 9769.8],
                #falloff=Troe(A=0.175, T3=1341.0, T1=60000.0, T2=10140.0))

## Reaction 1967
#falloff_reaction('C4H71-3 + H (+ M) <=> C4H8-2 (+ M)',
                #kf=[2.000000e+14, 0.0, 0.0],
                #kf0=[1.330000e+60, -12.0, 5967.8],
                #efficiencies='H2:2.0 H2O:6.0 CH4:2.0 CO:1.5 CO2:2.0 C2H6:3.0 AR:0.7',
                #falloff=Troe(A=0.02, T3=1097.0, T1=1097.0, T2=6860.0))

## Reaction 1968
#reaction('C4H8-2 <=> C4H72-2 + H', [1.240000e+72, -16.58, 139300.0])

## Reaction 1969
#reaction('C-2-C4H8 <=> C4H8-2', [3.980000e+13, 0.0, 62000.0])

## Reaction 1970
#reaction('C-2-C4H8 <=> C3H5-A + CH3', [3.750000e+66, -15.6, 97300.0])

## Reaction 1971
#falloff_reaction('C3H5-S + CH3 (+ M) <=> C-2-C4H8 (+ M)',
                #kf=[5.000000e+13, 0.0, 0.0],
                #kf0=[8.540000e+58, -11.94, 9769.8],
                #efficiencies='H2O:6.0 CO:1.5 CO2:2.0 CH4:2.0 C2H6:3.0 H2:2.0 C2H2:3.0 C2H4:3.0 AR:0.7',
                #falloff=Troe(A=0.175, T3=1340.6, T1=60000.0, T2=10139.8))

## Reaction 1972
#falloff_reaction('H + C4H71-3 (+ M) <=> C-2-C4H8 (+ M)',
                #kf=[2.000000e+14, 0.0, 0.0],
                #kf0=[1.330000e+60, -12.0, 5967.8],
                #efficiencies='H2:2.0 H2O:6.0 CH4:2.0 CO:1.5 CO2:2.0 C2H6:3.0 AR:0.7',
                #falloff=Troe(A=0.02, T3=1097.0, T1=1097.0, T2=6860.0))

## Reaction 1973
#reaction('C-2-C4H8 <=> C4H6 + H2', [1.000000e+13, 0.0, 65500.0])

## Reaction 1974
#reaction('C4H8-1 + OH <=> C4H71-3 + H2O', [4.040000e+05, 2.2, -437.2])

## Reaction 1975
#reaction('C4H8-1 + OH <=> C4H71-4 + H2O', [4.300000e+06, 2.03, 2621.5])

## Reaction 1976
#reaction('C4H8-1 + OH <=> C4H71-2 + H2O', [3.000000e+06, 1.97, 2845.9])

## Reaction 1977
#reaction('C4H8-1 + OH <=> C4H71-1 + H2O', [6.930000e+06, 1.92, 4959.0])

## Reaction 1978
#reaction('C4H8-2 + OH <=> C4H71-3 + H2O', [2.650000e+14, 0.0, 6269.62])

## Reaction 1979
#reaction('C4H8-2 + OH <=> C4H72-2 + H2O', [3.600000e+06, 1.979, 2235.2])

## Reaction 1980
#reaction('C4H8-1 + HO2 <=> C4H71-3 + H2O2', [3.910000e-01, 3.97, 11702.0])

## Reaction 1981
#reaction('C4H8-2 + HO2 <=> C4H71-3 + H2O2', [3.450000e-01, 4.0, 12103.0])

## Reaction 1982
#reaction('C4H8-1 + HO2 <=> C4H71-4 + H2O2', [4.080000e+01, 3.59, 17160.0])

## Reaction 1983
#reaction('C4H8-1 + HO2 <=> C4H71-1 + H2O2', [9.570000e+02, 3.059, 20798.6])

## Reaction 1984
#reaction('C4H8-1 + HO2 <=> C4H71-2 + H2O2', [1.560000e+04, 2.82, 24427.9])

## Reaction 1985
#reaction('C4H8-2 + HO2 <=> C4H72-2 + H2O2', [3.120000e+04, 2.82, 24427.9])

## Reaction 1986
#reaction('C4H8-1 + H <=> C4H71-4 + H2', [3.840000e+04, 2.87, 6611.0])

## Reaction 1987
#reaction('C4H8-1 + H <=> C4H71-3 + H2', [2.420000e+03, 3.05, 1995.0])

## Reaction 1988
#reaction('C4H8-1 + H <=> C4H71-2 + H2', [2.370000e+04, 2.85, 8917.0])

## Reaction 1989
#reaction('C4H8-1 + H <=> C4H71-1 + H2', [2.630000e+04, 2.83, 12050.0],
        #options='duplicate')

## Reaction 1990
#reaction('C4H8-1 + H <=> C4H71-1 + H2', [2.230000e+04, 2.85, 11710.0],
        #options='duplicate')

## Reaction 1991
#reaction('C4H8-2 + H <=> C4H71-3 + H2', [2.810000e+02, 3.5, 2627.0])

## Reaction 1992
#reaction('C4H8-2 + H <=> C4H72-2 + H2', [8.470000e+04, 2.76, 9304.0])

## Reaction 1993
#reaction('C4H8-1 + O <=> C4H71-3 + OH', [1.750000e+11, 0.7, 5884.0])

## Reaction 1994
#reaction('C4H8-2 + O <=> C4H71-3 + OH', [2.190000e+11, 0.81, 7550.0])

## Reaction 1995
#reaction('C4H8-1 + O <=> C4H71-1 + OH', [1.200000e+11, 0.7, 8959.1])

## Reaction 1996
#reaction('C4H8-1 + O <=> C4H71-2 + OH', [6.030000e+10, 0.7, 7632.0])

## Reaction 1997
#reaction('C4H8-2 + O <=> C4H72-2 + OH', [1.206000e+11, 0.7, 7632.0])

## Reaction 1998
#reaction('C4H8-1 + O <=> C4H71-4 + OH', [1.130000e+14, 0.0, 7850.0])

## Reaction 1999
#reaction('C4H8-1 + O2 <=> C4H71-3 + HO2', [2.000000e+13, 0.0, 37190.0])

## Reaction 2000
#reaction('C4H8-2 + O2 <=> C4H71-3 + HO2', [4.000000e+13, 0.0, 39390.0])

## Reaction 2001
#reaction('C4H8-1 + O2 <=> C4H71-1 + HO2', [2.000000e+13, 0.0, 62270.0])

## Reaction 2002
#reaction('C4H8-1 + O2 <=> C4H71-2 + HO2', [1.000000e+13, 0.0, 58770.0])

## Reaction 2003
#reaction('C4H8-2 + O2 <=> C4H72-2 + HO2', [2.000000e+13, 0.0, 58770.0])

## Reaction 2004
#reaction('C4H8-1 + O2 <=> C4H71-4 + HO2', [6.000000e+13, 0.0, 52340.0])

## Reaction 2005
#reaction('C4H8-1 + CH3 <=> C4H71-3 + CH4', [2.210000e+00, 3.5, 5675.0])

## Reaction 2006
#reaction('C4H8-1 + CH3 <=> C4H71-4 + CH4', [4.520000e-01, 3.65, 7154.0])

## Reaction 2007
#reaction('C4H8-2 + CH3 <=> C4H71-3 + CH4', [7.140000e+00, 3.57, 7642.0])

## Reaction 2008
#reaction('C4H8-1 + CH3 <=> C4H71-1 + CH4', [1.348000e+00, 3.5, 12850.0])

## Reaction 2009
#reaction('C4H8-1 + CH3 <=> C4H71-2 + CH4', [8.400000e-01, 3.5, 11660.0])

## Reaction 2010
#reaction('C4H8-2 + CH3 <=> C4H72-2 + CH4', [1.680000e+00, 3.5, 11660.0])

## Reaction 2011
#reaction('C4H8-1 + CH3O2 <=> C4H71-3 + CH3O2H', [2.700000e+04, 0.7, 5884.0])

## Reaction 2012
#reaction('C4H8-1 + CH3O2 <=> C4H71-4 + CH3O2H', [2.380000e+03, 2.55, 16490.0])

## Reaction 2013
#reaction('C4H8-1 + CH3O <=> C4H71-3 + CH3OH', [4.000000e+01, 2.9, 8609.0])

## Reaction 2014
#reaction('C4H8-1 + CH3O <=> C4H71-4 + CH3OH', [2.170000e+11, 0.0, 6458.0])

## Reaction 2015
#reaction('C4H8-2 + CH3O <=> C4H71-3 + CH3OH', [1.800000e+01, 2.95, 11990.0])

## Reaction 2016
#reaction('C4H8-1 + CH3CO3 <=> C4H71-3 + CH3CO3H', [1.000000e+11, 0.0, 8000.0])

## Reaction 2017
#reaction('C4H8-1 + C3H5-A <=> C4H71-3 + C3H6', [7.900000e+10, 0.0, 12400.0])

## Reaction 2018
#reaction('C4H71-3 + C4H71-3 <=> C4H8-1 + C4H6', [1.600000e+12, 0.0, 0.0])

## Reaction 2019
#reaction('C4H8-1 + C2H5O2 <=> C4H71-3 + C2H5O2H', [1.400000e+12, 0.0, 14900.0])

## Reaction 2020
#reaction('C4H8-1 + NC3H7O2 <=> C4H71-3 + NC3H7O2H', [1.400000e+12, 0.0, 14900.0])

## Reaction 2021
#reaction('C4H8-1 + IC3H7O2 <=> C4H71-3 + IC3H7O2H', [1.400000e+12, 0.0, 14900.0])

## Reaction 2022
#reaction('C4H8-1 + PC4H9O2 <=> C4H71-3 + PC4H9O2H', [1.400000e+12, 0.0, 14900.0])

## Reaction 2023
#reaction('C4H8-1 + SC4H9O2 <=> C4H71-3 + SC4H9O2H', [1.400000e+12, 0.0, 14900.0])

## Reaction 2024
#reaction('C4H8-2 + NC3H7O2 <=> C4H71-3 + NC3H7O2H', [3.200000e+12, 0.0, 14900.0])

## Reaction 2025
#reaction('C4H8-2 + IC3H7O2 <=> C4H71-3 + IC3H7O2H', [3.200000e+12, 0.0, 14900.0])

## Reaction 2026
#reaction('C4H8-2 + PC4H9O2 <=> C4H71-3 + PC4H9O2H', [3.200000e+12, 0.0, 14900.0])

## Reaction 2027
#reaction('C4H8-2 + SC4H9O2 <=> C4H71-3 + SC4H9O2H', [3.200000e+12, 0.0, 14900.0])

## Reaction 2028
#reaction('C4H71-3 + C2H5 <=> C4H8-1 + C2H4', [2.590000e+12, 0.0, -131.0])

## Reaction 2029
#reaction('C4H71-3 + CH3O <=> C4H8-1 + CH2O', [2.410000e+13, 0.0, 0.0])

## Reaction 2030
#reaction('IC4H9O2 + C4H8-1 <=> IC4H9O2H + C4H71-3', [1.400000e+12, 0.0, 14900.0])

## Reaction 2031
#reaction('TC4H9O2 + C4H8-1 <=> TC4H9O2H + C4H71-3', [1.400000e+12, 0.0, 14900.0])

## Reaction 2032
#reaction('IC4H9O2 + C4H8-2 <=> IC4H9O2H + C4H71-3', [1.400000e+12, 0.0, 14900.0])

## Reaction 2033
#reaction('TC4H9O2 + C4H8-2 <=> TC4H9O2H + C4H71-3', [1.400000e+12, 0.0, 14900.0])

## Reaction 2034
#reaction('C4H71-4 + HCO <=> C4H8-1 + CO', [6.000000e+13, 0.0, 0.0])

## Reaction 2035
#reaction('C4H71-4 + C2H4 <=> C4H8-1 + C2H3', [1.000000e+11, 0.0, 15000.0])

## Reaction 2036
#reaction('C4H8-1 + C2H3 <=> C4H71-3 + C2H4', [7.200000e+11, 0.0, 5008.0])

## Reaction 2037
#reaction('C4H8-1 + C2H5 <=> C4H71-3 + C2H6', [1.000000e+11, 0.0, 8300.0])

## Reaction 2038
#reaction('C4H71-4 + C2H6 <=> C4H8-1 + C2H5', [3.000000e+13, 0.0, 12300.01])

## Reaction 2039
#reaction('C4H71-3 + C3H5-A <=> C4H8-1 + C3H4-A', [1.000000e+12, 0.0, 0.0])

## Reaction 2040
#reaction('C4H8-1 + C3H5-S <=> C4H71-3 + C3H6', [8.000000e+10, 0.0, 12400.0])

## Reaction 2041
#reaction('C4H8-1 + C3H5-T <=> C4H71-3 + C3H6', [8.000000e+10, 0.0, 12400.0])

## Reaction 2042
#reaction('C4H71-4 + C4H71-4 <=> C4H6 + C4H8-1', [3.160000e+12, 0.0, 0.0])

## Reaction 2043
#reaction('C4H6 + C4H71-4 <=> C4H5-N + C4H8-1', [6.320000e+02, 3.13, 20010.0])

## Reaction 2044
#reaction('C4H6 + C4H71-4 <=> C4H5-I + C4H8-1', [6.320000e+02, 3.13, 18010.0])

## Reaction 2045
#reaction('C4H71-4 + C4H8-1 <=> C4H8-1 + C4H71-3', [1.000000e+11, 0.0, 8300.0])

## Reaction 2046
#reaction('SC4H9 + C4H71-3 <=> C4H8-2 + C4H8-2', [4.330000e+14, -0.75, -130.0])

## Reaction 2047
#reaction('C4H8-2 + C4H71-3 <=> C4H8-1 + C4H71-3', [3.980000e+13, 0.0, 15800.0])

## Reaction 2048
#reaction('C-2-C4H8 + H <=> C4H71-3 + H2', [4.440000e+04, 2.81, 4414.0])

## Reaction 2049
#reaction('C-2-C4H8 + CH3 <=> C4H71-3 + CH4', [7.140000e+00, 3.57, 7642.0])

## Reaction 2050
#reaction('C-2-C4H8 + HO2 <=> C4H71-3 + H2O2', [5.940000e+04, 2.57, 16140.0])

## Reaction 2051
#reaction('C-2-C4H8 + O <=> C4H71-3 + OH', [2.190000e+11, 0.81, 7550.0])

## Reaction 2052
#reaction('C-2-C4H8 + C4H71-3 <=> C4H8-1 + C4H71-3', [3.980000e+13, 0.0, 15800.0])

## Reaction 2053
#reaction('C-2-C4H8 + C4H71-3 <=> C4H8-2 + C4H71-3', [3.980000e+13, 0.0, 15800.0])

## Reaction 2054
#reaction('C-2-C4H8 + H <=> C4H72-2 + H2', [8.200000e+05, 2.5, 9790.0])

## Reaction 2055
#reaction('C-2-C4H8 + O <=> C4H72-2 + OH', [1.200000e+11, 0.7, 7630.0])

## Reaction 2056
#reaction('C-2-C4H8 + OH <=> C4H72-2 + H2O', [2.200000e+06, 2.0, 1450.0])

## Reaction 2057
#reaction('C-2-C4H8 + CH3 <=> C4H72-2 + CH4', [1.960000e+00, 3.5, 11700.0])

## Reaction 2058
#reaction('C4H71-3 + O <=> C2H3CHO + CH3', [6.030000e+13, 0.0, 0.0])

## Reaction 2059
#reaction('C4H71-3 + CH3O2 <=> C4H71-O + CH3O', [9.640000e+12, 0.0, 0.0])

## Reaction 2060
#reaction('C4H71-3 + PC4H9O2 <=> C4H71-O + PC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 2061
#reaction('C4H71-3 + SC4H9O2 <=> C4H71-O + SC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 2062
#reaction('C3H5-A + C4H71-3 <=> C3H6 + C4H6', [6.310000e+12, 0.0, 0.0])

## Reaction 2063
#reaction('C4H71-3 + O2 <=> C4H6 + HO2', [5.000000e+09, 0.0, 0.0])

## Reaction 2064
#reaction('H + C4H71-3 <=> C4H6 + H2', [3.160000e+13, 0.0, 0.0])

## Reaction 2065
#reaction('C2H5 + C4H71-3 <=> C4H6 + C2H6', [3.980000e+12, 0.0, 0.0])

## Reaction 2066
#reaction('C2H3 + C4H71-3 <=> C2H4 + C4H6', [3.980000e+12, 0.0, 0.0])

## Reaction 2067
#reaction('C4H71-3 + C2H5O2 <=> C4H71-O + C2H5O', [3.800000e+12, 0.0, -1200.0])

## Reaction 2068
#reaction('IC3H7O2 + C4H71-3 <=> IC3H7O + C4H71-O', [3.800000e+12, 0.0, -1200.0])

## Reaction 2069
#reaction('NC3H7O2 + C4H71-3 <=> NC3H7O + C4H71-O', [3.800000e+12, 0.0, -1200.0])

## Reaction 2070
#reaction('C4H71-O <=> C2H3CHO + CH3', [7.940000e+14, 0.0, 19000.0])

## Reaction 2071
#pdep_arrhenius('C4H71-3 + HO2 <=> C4H71-O + OH',
            #[(0.01, 'atm'), 1.020000e+13, -0.158, -1417.0],
            #[(0.1, 'atm'), 4.980000e+14, -0.642, -349.1],
            #[(1.0, 'atm'), 7.770000e+17, -1.52, 2379.2],
            #[(10.0, 'atm'), 2.930000e+15, -0.684, 3615.3],
            #[(100.0, 'atm'), 1.640000e+04, 2.74, 1144.4])

## Reaction 2072
#pdep_arrhenius('C4H71-3 + HO2 <=> C4H71-3OOH',
            #[(0.01, 'atm'), 1.910000e+31, -7.23, 1336.2],
            #[(0.1, 'atm'), 6.310000e+42, -10.3, 5568.9],
            #[(1.0, 'atm'), 1.030000e+45, -10.6, 7851.5],
            #[(10.0, 'atm'), 2.790000e+37, -7.92, 6497.9],
            #[(100.0, 'atm'), 4.730000e+25, -4.13, 2923.78])

## Reaction 2073
#pdep_arrhenius('C4H71-3 + HO2 <=> C2H3COCH3 + H2O',
            #[(0.01, 'atm'), 1.090000e+00, 3.01, -3421.1],
            #[(0.1, 'atm'), 6.350000e+01, 2.5, -2341.4],
            #[(1.0, 'atm'), 6.050000e+05, 1.39, 595.1],
            #[(10.0, 'atm'), 3.100000e+05, 1.59, 2677.6],
            #[(100.0, 'atm'), 5.070000e-05, 4.59, 927.5])

## Reaction 2074
#pdep_arrhenius('C4H71-3OOH <=> C2H3COCH3 + H2O',
            #[(0.01, 'atm'), 1.990000e+50, -12.7, 53531.9],
            #[(0.1, 'atm'), 4.720000e+47, -11.5, 54360.9],
            #[(1.0, 'atm'), 1.500000e+40, -8.84, 53179.2],
            #[(10.0, 'atm'), 2.540000e+28, -5.0, 49919.4],
            #[(100.0, 'atm'), 1.480000e+16, -1.12, 45949.3])

## Reaction 2075
#pdep_arrhenius('C4H71-3OOH <=> C4H71-O + OH',
            #[(0.01, 'atm'), 1.490000e+58, -13.9, 54266.9],
            #[(0.1, 'atm'), 1.800000e+54, -12.4, 54193.8],
            #[(1.0, 'atm'), 3.360000e+46, -9.81, 52468.5],
            #[(10.0, 'atm'), 2.390000e+36, -6.54, 49429.0],
            #[(100.0, 'atm'), 1.280000e+27, -3.61, 46333.1])

## Reaction 2076
#pdep_arrhenius('C4H71-O <=> C2H3 + CH3CHO',
            #[(0.001, 'atm'), 7.260000e+06, 0.182, 17815.5],
            #[(0.01, 'atm'), 6.970000e+16, -2.5, 20878.7],
            #[(0.1, 'atm'), 6.640000e+23, -4.23, 23565.0],
            #[(1.0, 'atm'), 1.070000e+26, -4.56, 24622.9],
            #[(10.0, 'atm'), 6.500000e+29, -5.37, 26645.0],
            #[(100.0, 'atm'), 4.630000e+31, -5.59, 28915.3],
            #[(1000.0, 'atm'), 8.520000e+25, -3.61, 27863.4])

## Reaction 2077
#pdep_arrhenius('C4H71-O <=> C3H5OCH2-1',
            #[(0.001, 'atm'), 3.170000e+20, -4.15, 12121.3],
            #[(0.01, 'atm'), 4.790000e+24, -5.03, 14606.1],
            #[(0.1, 'atm'), 1.900000e+26, -5.16, 16124.4],
            #[(1.0, 'atm'), 1.510000e+28, -5.4, 18165.4],
            #[(10.0, 'atm'), 2.420000e+28, -5.17, 19691.2],
            #[(100.0, 'atm'), 5.570000e+24, -3.86, 19395.2],
            #[(1000.0, 'atm'), 1.350000e+18, -1.73, 17386.5])

## Reaction 2078
#pdep_arrhenius('C4H71-O <=> CH2CH2COCH3',
            #[(0.001, 'atm'), 5.250000e-49, 15.5, -15639.9],
            #[(0.01, 'atm'), 1.460000e-88, 27.6, -35995.0],
            #[(0.1, 'atm'), 4.440000e-22, 8.38, -3819.0],
            #[(1.0, 'atm'), 6.230000e+12, -1.44, 10829.2],
            #[(10.0, 'atm'), 3.480000e+42, -9.91, 25297.9],
            #[(100.0, 'atm'), 1.880000e+38, -8.16, 25974.5],
            #[(1000.0, 'atm'), 1.670000e+21, -2.74, 20337.7])

## Reaction 2079
#pdep_arrhenius('C4H71-O <=> C2H3COCH3 + H',
            #[(0.001, 'atm'), 3.000000e+15, -2.31, 14667.9],
            #[(0.01, 'atm'), 1.500000e+22, -3.96, 18283.0],
            #[(0.1, 'atm'), 1.950000e+23, -3.99, 19143.3],
            #[(1.0, 'atm'), 1.150000e+25, -4.24, 20311.2],
            #[(10.0, 'atm'), 1.760000e+28, -4.89, 22765.2],
            #[(100.0, 'atm'), 1.410000e+27, -4.28, 23770.6],
            #[(1000.0, 'atm'), 2.570000e+20, -2.06, 22040.1])

## Reaction 2080
#pdep_arrhenius('C4H71-O <=> C2H4 + CH3CO',
            #[(0.001, 'atm'), 6.620000e+16, -2.84, 13197.0],
            #[(0.01, 'atm'), 1.260000e+20, -3.53, 15469.2],
            #[(0.1, 'atm'), 2.130000e+21, -3.64, 16584.5],
            #[(1.0, 'atm'), 1.070000e+24, -4.16, 18985.0],
            #[(10.0, 'atm'), 8.420000e+25, -4.4, 22382.6],
            #[(100.0, 'atm'), 1.860000e+21, -2.73, 23658.8],
            #[(1000.0, 'atm'), 4.750000e+08, 1.14, 20922.5])

## Reaction 2081
#pdep_arrhenius('C3H5OCH2-1 <=> C3H5-A + CH2O',
            #[(0.001, 'atm'), 3.300000e+09, -0.638, 19747.8],
            #[(0.01, 'atm'), 3.360000e+21, -3.9, 23945.2],
            #[(0.1, 'atm'), 2.910000e+29, -5.9, 27249.7],
            #[(1.0, 'atm'), 1.830000e+34, -6.94, 30690.4],
            #[(10.0, 'atm'), 9.720000e+33, -6.5, 33002.5],
            #[(100.0, 'atm'), 2.680000e+27, -4.26, 33305.6],
            #[(1000.0, 'atm'), 8.810000e+14, -0.326, 31553.1])

## Reaction 2082
#pdep_arrhenius('C3H5OCH2-1 <=> CH2CH2COCH3',
            #[(0.001, 'atm'), 8.010000e-92, 27.8, -37321.2],
            #[(0.01, 'atm'), 7.780000e-11, 3.7, -2766.9],
            #[(0.1, 'atm'), 5.110000e+15, -2.76, 15937.6],
            #[(1.0, 'atm'), 4.480000e+25, -5.2, 21532.2],
            #[(10.0, 'atm'), 3.970000e+34, -7.41, 28116.9],
            #[(100.0, 'atm'), 5.620000e+22, -3.56, 25806.7],
            #[(1000.0, 'atm'), 2.510000e+20, -2.63, 29288.4])

## Reaction 2083
#pdep_arrhenius('C3H5OCH2-1 <=> C2H3COCH3 + H',
            #[(0.001, 'atm'), 4.930000e+24, -5.05, 20108.4],
            #[(0.01, 'atm'), 2.140000e+28, -5.8, 22219.4],
            #[(0.1, 'atm'), 1.930000e+32, -6.64, 25108.2],
            #[(1.0, 'atm'), 8.600000e+34, -7.11, 28209.1],
            #[(10.0, 'atm'), 2.170000e+34, -6.64, 30647.6],
            #[(100.0, 'atm'), 4.170000e+28, -4.71, 31231.9],
            #[(1000.0, 'atm'), 3.980000e+18, -1.62, 30129.8])

## Reaction 2084
#pdep_arrhenius('C3H5OCH2-1 <=> C3H6 + HCO',
            #[(0.001, 'atm'), 8.230000e+26, -5.84, 19356.9],
            #[(0.01, 'atm'), 1.320000e+29, -6.21, 21293.6],
            #[(0.1, 'atm'), 3.470000e+32, -6.96, 24197.3],
            #[(1.0, 'atm'), 1.440000e+36, -7.76, 28007.8],
            #[(10.0, 'atm'), 9.720000e+37, -8.02, 32394.6],
            #[(100.0, 'atm'), 2.430000e+31, -5.81, 34295.8],
            #[(1000.0, 'atm'), 3.730000e+14, -0.726, 32008.3])

## Reaction 2085
#pdep_arrhenius('CH2CH2COCH3 <=> C2H3 + CH3CHO',
            #[(0.001, 'atm'), 6.890000e-69, 21.5, 2638.0],
            #[(0.01, 'atm'), 5.340000e-33, 11.1, 16749.1],
            #[(0.1, 'atm'), 6.110000e+26, -6.01, 44116.7],
            #[(1.0, 'atm'), 8.040000e+35, -8.31, 46919.7],
            #[(10.0, 'atm'), 5.520000e+40, -9.19, 50508.7],
            #[(100.0, 'atm'), 5.850000e+35, -7.18, 52038.4],
            #[(1000.0, 'atm'), 1.930000e+19, -1.94, 48440.0])

## Reaction 2086
#pdep_arrhenius('CH2CH2COCH3 <=> C2H3COCH3 + H',
            #[(0.001, 'atm'), 1.610000e+10, -1.24, 32371.3],
            #[(0.01, 'atm'), 5.840000e+15, -2.61, 32878.4],
            #[(0.1, 'atm'), 3.640000e+23, -4.6, 34275.3],
            #[(1.0, 'atm'), 7.580000e+31, -6.63, 37895.4],
            #[(10.0, 'atm'), 2.860000e+32, -6.3, 39990.7],
            #[(100.0, 'atm'), 1.570000e+23, -3.14, 38011.7],
            #[(1000.0, 'atm'), 4.520000e+12, 0.214, 34570.5])

## Reaction 2087
#pdep_arrhenius('CH2CH2COCH3 <=> C2H4 + CH3CO',
            #[(0.001, 'atm'), 2.900000e+32, -7.24, 25687.5],
            #[(0.01, 'atm'), 5.300000e+33, -7.28, 27100.6],
            #[(0.1, 'atm'), 2.000000e+35, -7.41, 29027.3],
            #[(1.0, 'atm'), 1.010000e+34, -6.7, 30018.1],
            #[(10.0, 'atm'), 9.760000e+27, -4.63, 28923.9],
            #[(100.0, 'atm'), 2.110000e+19, -1.85, 26239.8],
            #[(1000.0, 'atm'), 1.590000e+13, 0.063, 24086.3])

## Reaction 2088
#pdep_arrhenius('C2H3 + CH3CHO <=> C2H3COCH3 + H',
            #[(0.001, 'atm'), 2.600000e+04, 2.26, 1510.3],
            #[(0.01, 'atm'), 5.130000e+04, 2.17, 1675.5],
            #[(0.1, 'atm'), 3.990000e+05, 1.91, 2218.3],
            #[(1.0, 'atm'), 1.750000e+07, 1.45, 3428.0],
            #[(10.0, 'atm'), 1.350000e+09, 0.933, 5173.0],
            #[(100.0, 'atm'), 2.240000e+11, 0.357, 8001.3],
            #[(1000.0, 'atm'), 6.010000e+05, 2.09, 7895.6])

## Reaction 2089
#pdep_arrhenius('C2H3 + CH3CHO <=> C2H4 + CH3CO',
            #[(0.001, 'atm'), 1.110000e+07, 1.09, 1807.2],
            #[(0.01, 'atm'), 2.470000e+07, 0.993, 1994.9],
            #[(0.1, 'atm'), 2.470000e+08, 0.704, 2596.2],
            #[(1.0, 'atm'), 1.420000e+10, 0.209, 3934.2],
            #[(10.0, 'atm'), 3.450000e+13, -0.726, 6944.3],
            #[(100.0, 'atm'), 3.310000e+14, -0.866, 10965.7],
            #[(1000.0, 'atm'), 1.650000e+01, 3.17, 9399.8])

## Reaction 2090
#falloff_reaction('C4H71-4 (+ M) <=> C4H71-3 (+ M)',
                #kf=[5.750000e+12, 0.0, 30400.0],
                #kf0=[1.170000e+15, 0.0, 8170.0],
                #falloff=Troe(A=0.031, T3=50.0, T1=100000.0, T2=50000.0))

## Reaction 2091
#falloff_reaction('C4H71-3 (+ M) <=> C4H6 + H (+ M)',
                #kf=[4.700000e+08, 1.32, 44697.6],
                #kf0=[4.600000e-37, 15.37, -603.1],
                #efficiencies='H2:2.0 H2O:6.0 CH4:2.0 CO:1.5 CO2:2.0 C2H6:3.0')

## Reaction 2092
#reaction('C4H71-1 <=> C4H6-1 + H', [8.950000e+08, 1.14, 35636.4576])

## Reaction 2093
#reaction('C4H71-1 <=> C2H5 + C2H2', [2.850000e+12, 0.68, 33178.2912])

## Reaction 2094
#reaction('C4H71-2 <=> C4H6-1 + H', [9.790000e+08, 1.2, 38080.7136])

## Reaction 2095
#reaction('C4H71-2 <=> C4H612 + H', [1.990000e+08, 1.56, 37683.2736])

## Reaction 2096
#reaction('C4H71-2 <=> C3H4-A + CH3', [1.230000e+10, 1.19, 32915.9808])

## Reaction 2097
#reaction('C4H71-3 <=> C4H612 + H', [3.950000e+11, 0.93, 61611.1488])

## Reaction 2098
#reaction('C4H71-3 <=> C4H6 + H', [4.265000e+08, 1.95, 47490.1056])

## Reaction 2099
#reaction('C4H71-4 <=> C2H4 + C2H3', [1.420000e+11, 0.99, 38998.8])

## Reaction 2100
#reaction('C4H71-4 <=> C4H6 + H', [1.320000e+05, 2.28, 33245.856])

## Reaction 2101
#reaction('C4H72-2 <=> C4H6-2 + H', [7.870000e+09, 1.16, 36173.0016])

## Reaction 2102
#reaction('C4H72-2 <=> C4H612 + H', [9.540000e+07, 1.81, 38992.8384])

## Reaction 2103
#reaction('C4H72-2 <=> C3H4-P + CH3', [6.300000e+11, 0.93, 34754.1408])

## Reaction 2104
#reaction('C4H71-3 <=> C4H71-4', [5.620000e-12, 7.19, 36200.8224])

## Reaction 2105
#reaction('C4H71-3 <=> C4H72-2', [6.760000e-23, 10.21, 41574.2112])

## Reaction 2106
#reaction('C4H71-3 + HO2 <=> C4H6 + H2O2', [1.000000e+12, 0.0, 0.0])

## Reaction 2107
#reaction('C4H71-3 + OH <=> C4H6 + H2O', [6.000000e+12, 0.0, 0.0])

## Reaction 2108
#reaction('C4H71-3 + OH <=> C4H612 + H2O', [6.000000e+12, 0.0, 0.0])

## Reaction 2109
#reaction('C4H71-3 + O <=> C4H6 + OH', [4.000000e+13, 0.0, 0.0])

## Reaction 2110
#reaction('C4H71-3 + H <=> C4H612 + H2', [1.800000e+13, 0.0, 0.0])

## Reaction 2111
#reaction('C4H71-3 + CH3 <=> C4H6 + CH4', [1.000000e+13, 0.0, 0.0])

## Reaction 2112
#reaction('C4H71-3 + CH3 <=> C4H612 + CH4', [1.000000e+13, 0.0, 0.0])

## Reaction 2113
#reaction('C4H71-3 + C2H3 <=> C4H612 + C2H4', [1.000000e+13, 0.0, 0.0])

## Reaction 2114
#reaction('C4H71-3 + C3H3 <=> C4H6 + C3H4-P', [4.000000e+12, 0.0, 0.0])

## Reaction 2115
#reaction('C4H71-3 + C3H3 <=> C4H6 + C3H4-A', [4.000000e+12, 0.0, 0.0])

## Reaction 2116
#reaction('C4H71-3 + HO2 => C2H3 + CH3CHO + OH', [7.000000e+12, 0.0, -1000.0])

## Reaction 2117
#reaction('C4H71-3 + HO2 => CH2O + C3H5-S + OH', [7.000000e+12, 0.0, -1000.0])

## Reaction 2118
#reaction('C4H71-4 + H <=> C4H6 + H2', [1.800000e+12, 0.0, 0.0])

## Reaction 2119
#reaction('C4H71-4 + O <=> C4H6 + OH', [4.000000e+13, 0.0, 0.0])

## Reaction 2120
#reaction('C4H71-4 + OH <=> C4H6 + H2O', [4.000000e+13, 0.0, 0.0])

## Reaction 2121
#reaction('C4H71-4 + HO2 <=> C4H6 + H2O2', [1.800000e+12, 0.0, 0.0])

## Reaction 2122
#reaction('C4H71-4 + CH3 <=> C4H6 + CH4', [1.100000e+13, 0.0, 0.0])

## Reaction 2123
#reaction('C4H71-4 + C2H3 <=> C4H6 + C2H4', [4.000000e+12, 0.0, 0.0])

## Reaction 2124
#reaction('C4H71-4 + C2H5 <=> C4H6 + C2H6', [4.000000e+12, 0.0, 0.0])

## Reaction 2125
#reaction('C4H71-4 + C3H3 <=> C4H6 + C3H4-P', [4.000000e+12, 0.0, 0.0])

## Reaction 2126
#reaction('C4H71-4 + C3H3 <=> C4H6 + C3H4-A', [4.000000e+12, 0.0, 0.0])

## Reaction 2127
#reaction('C4H71-4 + C3H5-A <=> C4H6 + C3H6', [6.300000e+12, 0.0, 0.0])

## Reaction 2128
#reaction('C4H71-4 + O2 <=> C4H6 + HO2', [1.020000e+12, 0.0, 22654.0])

## Reaction 2129
#reaction('C4H71-4 + HO2 => C3H5-A + CH2O + OH', [1.000000e+12, 0.0, 0.0])

## Reaction 2130
#reaction('C4H71-4 + H <=> CH3 + C3H5-A', [2.000000e+21, -2.0, 11000.0])

## Reaction 2131
#reaction('C4H71-4 + H <=> C4H71-3 + H', [1.000000e+13, 0.0, 0.0])

## Reaction 2132
#reaction('C4H71-2 + O2 <=> C2H5CO + CH2O', [4.600000e+16, -1.39, 1010.0])

## Reaction 2133
#reaction('C4H71-1 + O2 <=> HCO + C2H5CHO', [4.600000e+16, -1.39, 1010.0])

## Reaction 2134
#reaction('C4H71-1 + CH2O <=> C4H8-1 + HCO', [5.420000e+03, 2.81, 5862.0])

## Reaction 2135
#pdep_arrhenius('C4H8-1 + H <=> PC4H9',
            #[(0.001, 'atm'), 3.000000e+15, -3.1, 2932.0],
            #[(0.01, 'atm'), 1.040000e+17, -3.26, 3322.0],
            #[(0.1, 'atm'), 2.510000e+21, -4.2, 5834.0],
            #[(1.0, 'atm'), 1.380000e+31, -6.62, 12947.0],
            #[(10.0, 'atm'), 1.710000e+40, -8.78, 21960.0],
            #[(100.0, 'atm'), 4.040000e+37, -7.54, 26045.0],
            #options='duplicate')

## Reaction 2136
#pdep_arrhenius('C4H8-1 + H <=> PC4H9',
            #[(0.001, 'atm'), 5.090000e+22, -4.75, 1071.0],
            #[(0.01, 'atm'), 1.400000e+24, -4.87, 1398.0],
            #[(0.1, 'atm'), 4.020000e+27, -5.56, 3296.0],
            #[(1.0, 'atm'), 2.590000e+33, -6.88, 7620.0],
            #[(10.0, 'atm'), 1.430000e+36, -7.27, 11386.0],
            #[(100.0, 'atm'), 1.060000e+31, -5.42, 11252.0],
            #options='duplicate')

## Reaction 2137
#pdep_arrhenius('C4H8-1 + H <=> SC4H9',
            #[(0.001, 'atm'), 4.230000e+24, -5.11, -83.0],
            #[(0.01, 'atm'), 6.830000e+26, -5.45, 817.0],
            #[(0.1, 'atm'), 1.100000e+32, -6.62, 4201.0],
            #[(1.0, 'atm'), 3.050000e+37, -7.81, 8973.0],
            #[(10.0, 'atm'), 7.110000e+36, -7.23, 11236.0],
            #[(100.0, 'atm'), 3.120000e+28, -4.47, 9260.0],
            #options='duplicate')

## Reaction 2138
#pdep_arrhenius('C4H8-1 + H <=> SC4H9',
            #[(0.001, 'atm'), 5.370000e+13, -2.69, 2503.0],
            #[(0.01, 'atm'), 1.490000e+15, -2.81, 2822.0],
            #[(0.1, 'atm'), 1.230000e+19, -3.63, 4965.0],
            #[(1.0, 'atm'), 2.940000e+28, -5.96, 11634.0],
            #[(10.0, 'atm'), 2.930000e+38, -8.38, 20940.0],
            #[(100.0, 'atm'), 6.210000e+37, -7.69, 26068.0],
            #options='duplicate')

## Reaction 2139
#pdep_arrhenius('C4H8-1 + H <=> C3H6 + CH3',
            #[(0.001, 'atm'), 1.970000e+11, 0.79, 2353.0],
            #[(0.01, 'atm'), 1.200000e+12, 0.57, 2909.0],
            #[(0.1, 'atm'), 3.450000e+15, -0.4, 5439.0],
            #[(1.0, 'atm'), 1.120000e+22, -2.2, 10693.0],
            #[(10.0, 'atm'), 2.650000e+26, -3.35, 15901.0],
            #[(100.0, 'atm'), 1.900000e+22, -2.04, 17011.0],
            #options='duplicate')

## Reaction 2140
#pdep_arrhenius('C4H8-1 + H <=> C3H6 + CH3',
            #[(0.001, 'atm'), 1.500000e+07, 1.33, 7220.0],
            #[(0.01, 'atm'), 2.790000e+07, 1.26, 7409.0],
            #[(0.1, 'atm'), 2.490000e+09, 0.71, 8799.0],
            #[(1.0, 'atm'), 3.690000e+16, -1.3, 14158.0],
            #[(10.0, 'atm'), 1.430000e+28, -4.49, 24072.0],
            #[(100.0, 'atm'), 4.910000e+33, -5.86, 32795.0],
            #options='duplicate')

## Reaction 2141
#pdep_arrhenius('C4H8-1 + H <=> C4H8-2 + H',
            #[(0.001, 'atm'), 6.940000e+06, 1.87, 3327.0],
            #[(0.01, 'atm'), 2.510000e+07, 1.71, 3718.0],
            #[(0.1, 'atm'), 1.540000e+10, 0.93, 5726.0],
            #[(1.0, 'atm'), 2.750000e+16, -0.81, 10596.0],
            #[(10.0, 'atm'), 7.310000e+21, -2.27, 16205.0],
            #[(100.0, 'atm'), 1.680000e+19, -1.39, 18107.0],
            #options='duplicate')

## Reaction 2142
#pdep_arrhenius('C4H8-1 + H <=> C4H8-2 + H',
            #[(0.001, 'atm'), 1.260000e+03, 2.29, 7722.0],
            #[(0.01, 'atm'), 2.060000e+03, 2.23, 7871.0],
            #[(0.1, 'atm'), 8.650000e+04, 1.77, 9022.0],
            #[(1.0, 'atm'), 2.900000e+11, -0.06, 13845.0],
            #[(10.0, 'atm'), 1.390000e+23, -3.28, 23566.0],
            #[(100.0, 'atm'), 1.140000e+30, -5.06, 32895.0],
            #options='duplicate')

## Reaction 2143
#pdep_arrhenius('C4H8-1 + H <=> C-2-C4H8 + H',
            #[(0.001, 'atm'), 5.470000e+06, 1.88, 3701.0],
            #[(0.01, 'atm'), 1.740000e+07, 1.73, 4053.0],
            #[(0.1, 'atm'), 7.040000e+09, 1.0, 5925.0],
            #[(1.0, 'atm'), 9.570000e+15, -0.71, 10669.0],
            #[(10.0, 'atm'), 4.090000e+21, -2.22, 16344.0],
            #[(100.0, 'atm'), 2.150000e+19, -1.45, 18431.0],
            #options='duplicate')

## Reaction 2144
#pdep_arrhenius('C4H8-1 + H <=> C-2-C4H8 + H',
            #[(0.001, 'atm'), 9.650000e+02, 2.3, 7919.0],
            #[(0.01, 'atm'), 1.530000e+03, 2.24, 8060.0],
            #[(0.1, 'atm'), 5.390000e+04, 1.81, 9155.0],
            #[(1.0, 'atm'), 1.230000e+11, 0.03, 13842.0],
            #[(10.0, 'atm'), 5.740000e+22, -3.2, 23495.0],
            #[(100.0, 'atm'), 9.720000e+29, -5.07, 32955.0],
            #options='duplicate')

## Reaction 2145
#pdep_arrhenius('C4H8-1 + H <=> C2H4 + C2H5',
            #[(0.001, 'atm'), 6.950000e+07, 1.53, 6778.0],
            #[(0.01, 'atm'), 1.445000e+08, 1.45, 7001.0],
            #[(0.1, 'atm'), 2.365000e+10, 0.82, 8581.0],
            #[(1.0, 'atm'), 9.100000e+17, -1.3, 14272.0],
            #[(10.0, 'atm'), 3.095000e+29, -4.47, 24261.0],
            #[(100.0, 'atm'), 1.435000e+34, -5.59, 32564.0],
            #options='duplicate')

## Reaction 2146
#pdep_arrhenius('C4H8-1 + H <=> C2H4 + C2H5',
            #[(0.001, 'atm'), 2.400000e+10, 1.21, 3333.0],
            #[(0.01, 'atm'), 4.415000e+10, 1.14, 3520.0],
            #[(0.1, 'atm'), 2.225000e+12, 0.66, 4743.0],
            #[(1.0, 'atm'), 2.275000e+17, -0.74, 8558.0],
            #[(10.0, 'atm'), 3.335000e+23, -2.42, 14185.0],
            #[(100.0, 'atm'), 1.370000e+24, -2.48, 17694.0],
            #options='duplicate')

## Reaction 2147
#pdep_arrhenius('SC4H9 <=> PC4H9',
            #[(0.001, 'atm'), 3.980000e+35, -10.6, 38129.0],
            #[(0.01, 'atm'), 2.460000e+38, -10.82, 38742.0],
            #[(0.1, 'atm'), 5.050000e+44, -12.01, 42290.0],
            #[(1.0, 'atm'), 1.570000e+53, -13.76, 49365.0],
            #[(10.0, 'atm'), 1.380000e+54, -13.32, 54046.0],
            #[(100.0, 'atm'), 1.460000e+44, -9.84, 52920.0])

## Reaction 2148
#pdep_arrhenius('SC4H9 <=> C3H6 + CH3',
            #[(0.001, 'atm'), 4.110000e+39, -9.54, 33009.0],
            #[(0.01, 'atm'), 5.100000e+43, -10.36, 36601.0],
            #[(0.1, 'atm'), 1.380000e+46, -10.68, 40002.0],
            #[(1.0, 'atm'), 4.640000e+44, -9.86, 41771.0],
            #[(10.0, 'atm'), 4.220000e+37, -7.46, 40615.0],
            #[(100.0, 'atm'), 2.780000e+26, -3.94, 36823.0])

## Reaction 2149
#pdep_arrhenius('SC4H9 <=> C4H8-2 + H',
            #[(0.001, 'atm'), 2.770000e+26, -5.93, 32985.0],
            #[(0.01, 'atm'), 1.110000e+30, -6.65, 34986.0],
            #[(0.1, 'atm'), 7.800000e+35, -7.98, 39295.0],
            #[(1.0, 'atm'), 3.160000e+39, -8.63, 43440.0],
            #[(10.0, 'atm'), 2.900000e+36, -7.36, 44477.0],
            #[(100.0, 'atm'), 7.360000e+26, -4.26, 41744.0])

## Reaction 2150
#pdep_arrhenius('SC4H9 <=> C-2-C4H8 + H',
            #[(0.001, 'atm'), 4.770000e+25, -5.72, 33755.0],
            #[(0.01, 'atm'), 4.430000e+28, -6.27, 35243.0],
            #[(0.1, 'atm'), 2.530000e+34, -7.58, 39268.0],
            #[(1.0, 'atm'), 7.180000e+38, -8.47, 43738.0],
            #[(10.0, 'atm'), 5.920000e+36, -7.48, 45269.0],
            #[(100.0, 'atm'), 4.340000e+27, -4.5, 42823.0])

## Reaction 2151
#pdep_arrhenius('SC4H9 <=> C2H4 + C2H5',
            #[(0.001, 'atm'), 3.650000e+23, -5.3, 38348.0],
            #[(0.01, 'atm'), 1.700000e+25, -5.49, 38840.0],
            #[(0.1, 'atm'), 1.510000e+30, -6.59, 41858.0],
            #[(1.0, 'atm'), 8.290000e+39, -8.99, 49427.0],
            #[(10.0, 'atm'), 6.360000e+46, -10.5, 57624.0],
            #[(100.0, 'atm'), 5.020000e+41, -8.56, 60410.0])

## Reaction 2152
#pdep_arrhenius('PC4H9 <=> C3H6 + CH3',
            #[(0.001, 'atm'), 3.560000e+61, -17.93, 45680.0],
            #[(0.01, 'atm'), 7.380000e+62, -18.02, 45969.0],
            #[(0.1, 'atm'), 6.000000e+63, -17.99, 46385.0],
            #[(1.0, 'atm'), 1.680000e+60, -16.75, 43726.0],
            #[(10.0, 'atm'), 3.000000e+59, -16.29, 42472.0],
            #[(100.0, 'atm'), 2.600000e+60, -16.27, 42427.0])

## Reaction 2153
#pdep_arrhenius('PC4H9 <=> C4H8-2 + H',
            #[(0.001, 'atm'), 2.570000e+61, -18.47, 47278.0],
            #[(0.01, 'atm'), 1.380000e+63, -18.67, 47866.0],
            #[(0.1, 'atm'), 1.270000e+66, -19.22, 49855.0],
            #[(1.0, 'atm'), 3.830000e+61, -17.71, 47012.0],
            #[(10.0, 'atm'), 3.380000e+59, -16.89, 44940.0],
            #[(100.0, 'atm'), 2.600000e+60, -16.86, 44861.0])

## Reaction 2154
#pdep_arrhenius('PC4H9 <=> C-2-C4H8 + H',
            #[(0.001, 'atm'), 1.160000e+62, -18.78, 47553.0],
            #[(0.01, 'atm'), 7.910000e+63, -19.01, 48216.0],
            #[(0.1, 'atm'), 4.320000e+67, -19.77, 50778.0],
            #[(1.0, 'atm'), 1.040000e+63, -18.24, 48049.0],
            #[(10.0, 'atm'), 1.570000e+60, -17.22, 45477.0],
            #[(100.0, 'atm'), 1.100000e+61, -17.17, 45371.0])

## Reaction 2155
#pdep_arrhenius('PC4H9 <=> C2H4 + C2H5',
            #[(0.001, 'atm'), 2.230000e+65, -17.49, 41515.0],
            #[(0.01, 'atm'), 3.900000e+69, -18.39, 44589.0],
            #[(0.1, 'atm'), 4.180000e+73, -19.24, 48099.0],
            #[(1.0, 'atm'), 2.230000e+76, -19.8, 50861.0],
            #[(10.0, 'atm'), 1.780000e+77, -19.99, 51817.0],
            #[(100.0, 'atm'), 2.430000e+77, -20.02, 51957.0])

## Reaction 2156
#pdep_arrhenius('C4H8-1 + OH <=> PC4H8OH-2',
            #[(0.0013, 'atm'), 3.070000e+78, -20.7, 32402.0],
            #[(0.01, 'atm'), 3.650000e+77, -20.0, 33874.0],
            #[(0.013, 'atm'), 1.420000e+76, -19.58, 32874.0],
            #[(0.025, 'atm'), 4.900000e+73, -18.79, 31361.0],
            #[(0.1, 'atm'), 1.380000e+68, -17.01, 27909.0],
            #[(0.1315, 'atm'), 9.640000e+66, -16.64, 27162.0],
            #[(1.0, 'atm'), 2.600000e+59, -14.17, 23079.0],
            #[(10.0, 'atm'), 1.010000e+54, -12.23, 22976.0],
            #[(100.0, 'atm'), 1.910000e+48, -10.23, 23772.0],
            #options='duplicate')

## Reaction 2157
#pdep_arrhenius('C4H8-1 + OH <=> PC4H8OH-2',
            #[(0.0013, 'atm'), 8.550000e+59, -15.84, 11594.0],
            #[(0.01, 'atm'), 9.700000e+59, -15.51, 12898.0],
            #[(0.013, 'atm'), 3.720000e+59, -15.34, 12913.0],
            #[(0.025, 'atm'), 3.530000e+58, -14.93, 12936.0],
            #[(0.1, 'atm'), 1.800000e+56, -14.04, 12945.0],
            #[(0.1315, 'atm'), 5.310000e+55, -13.85, 12887.0],
            #[(1.0, 'atm'), 2.070000e+50, -12.04, 11493.0],
            #[(10.0, 'atm'), 8.550000e+41, -9.35, 8921.0],
            #[(100.0, 'atm'), 3.060000e+32, -6.31, 6088.0],
            #options='duplicate')

## Reaction 2158
#pdep_arrhenius('C4H8-1 + OH <=> SC4H8OH-1',
            #[(0.0013, 'atm'), 3.070000e+78, -20.7, 32402.0],
            #[(0.01, 'atm'), 3.650000e+77, -20.0, 33874.0],
            #[(0.013, 'atm'), 1.420000e+76, -19.58, 32874.0],
            #[(0.025, 'atm'), 4.900000e+73, -18.79, 31361.0],
            #[(0.1, 'atm'), 1.380000e+68, -17.01, 27909.0],
            #[(0.1315, 'atm'), 9.640000e+66, -16.64, 27162.0],
            #[(1.0, 'atm'), 2.600000e+59, -14.17, 23079.0],
            #[(10.0, 'atm'), 1.010000e+54, -12.23, 22976.0],
            #[(100.0, 'atm'), 1.910000e+48, -10.23, 23772.0],
            #options='duplicate')

## Reaction 2159
#pdep_arrhenius('C4H8-1 + OH <=> SC4H8OH-1',
            #[(0.0013, 'atm'), 8.550000e+59, -15.84, 11594.0],
            #[(0.01, 'atm'), 9.700000e+59, -15.51, 12898.0],
            #[(0.013, 'atm'), 3.720000e+59, -15.34, 12913.0],
            #[(0.025, 'atm'), 3.530000e+58, -14.93, 12936.0],
            #[(0.1, 'atm'), 1.800000e+56, -14.04, 12945.0],
            #[(0.1315, 'atm'), 5.310000e+55, -13.85, 12887.0],
            #[(1.0, 'atm'), 2.070000e+50, -12.04, 11493.0],
            #[(10.0, 'atm'), 8.550000e+41, -9.35, 8921.0],
            #[(100.0, 'atm'), 3.060000e+32, -6.31, 6088.0],
            #options='duplicate')

## Reaction 2160
#pdep_arrhenius('C4H8-1 + OH <=> C4H71-3OH + H',
            #[(0.0013, 'atm'), 2.670000e+13, 0.05, 10611.0],
            #[(0.01, 'atm'), 2.750000e+13, 0.05, 10623.0],
            #[(0.013, 'atm'), 2.870000e+13, 0.04, 10634.0],
            #[(0.025, 'atm'), 1.590000e+14, -0.16, 11125.0],
            #[(0.1, 'atm'), 3.100000e+14, -0.22, 11407.0],
            #[(0.1315, 'atm'), 3.780000e+14, -0.24, 11458.0],
            #[(1.0, 'atm'), 9.150000e+07, 1.42, 10087.0],
            #[(10.0, 'atm'), 3.660000e+05, 2.14, 10410.0],
            #[(100.0, 'atm'), 8.190000e+02, 2.84, 10481.0])

## Reaction 2161
#pdep_arrhenius('C4H8-1 + OH <=> C2H3OH + C2H5',
            #[(0.0013, 'atm'), 1.290000e+06, 1.65, 1233.0],
            #[(0.01, 'atm'), 1.820000e+04, 2.1, 1162.0],
            #[(0.013, 'atm'), 2.040000e+03, 2.48, 1128.0],
            #[(0.025, 'atm'), 2.880000e+02, 2.8, 1152.0],
            #[(0.1, 'atm'), 1.400000e+01, 3.21, 1208.0],
            #[(0.1315, 'atm'), 7.710000e+00, 3.29, 1216.0],
            #[(1.0, 'atm'), 1.130000e+04, 2.5, 3238.0],
            #[(10.0, 'atm'), 2.410000e+19, -1.74, 13107.0],
            #[(100.0, 'atm'), 3.300000e-01, 3.7, 3665.0])

## Reaction 2162
#pdep_arrhenius('C4H8-1 + OH <=> C4H71-1OH + H',
            #[(0.0013, 'atm'), 3.470000e+06, 1.53, 4288.0],
            #[(0.01, 'atm'), 1.080000e+07, 1.34, 4576.0],
            #[(0.013, 'atm'), 9.760000e+06, 1.33, 4589.0],
            #[(0.025, 'atm'), 5.140000e+06, 1.36, 4594.0],
            #[(0.1, 'atm'), 3.130000e+05, 1.69, 4603.0],
            #[(0.1315, 'atm'), 1.390000e+05, 1.8, 4603.0],
            #[(1.0, 'atm'), 1.030000e+02, 2.83, 4530.0],
            #[(10.0, 'atm'), 3.400000e-02, -3.89, 4390.0],
            #[(100.0, 'atm'), 4.460000e-06, 5.03, 4132.0])

## Reaction 2163
#pdep_arrhenius('C4H8-1 + OH <=> C4H71-2OH + H',
            #[(0.0013, 'atm'), 2.870000e+00, 2.92, 625.0],
            #[(0.01, 'atm'), 4.840000e-01, 2.98, 704.0],
            #[(0.013, 'atm'), 3.130000e-01, 3.04, 721.0],
            #[(0.025, 'atm'), 9.330000e-03, 3.62, 677.0],
            #[(0.1, 'atm'), 4.640000e-05, 4.48, 687.0],
            #[(0.1315, 'atm'), 2.710000e-05, 4.56, 707.0],
            #[(1.0, 'atm'), 7.650000e-07, 5.05, 874.0],
            #[(10.0, 'atm'), 2.640000e+15, -0.8, 12728.0],
            #[(100.0, 'atm'), 4.870000e-04, 4.32, 4020.0])

## Reaction 2164
#pdep_arrhenius('C4H8-1 + OH <=> CH3COCH3 + CH3',
            #[(0.0013, 'atm'), 6.930000e+05, 1.49, -536.0],
            #[(0.01, 'atm'), 5.940000e+03, 2.01, -560.0],
            #[(0.013, 'atm'), 1.100000e+03, 2.22, -680.0],
            #[(0.025, 'atm'), 1.070000e+02, 2.5, -759.0],
            #[(0.1, 'atm'), 7.830000e-01, 3.1, -919.0],
            #[(0.1315, 'atm'), 3.070000e-01, 3.22, -946.0],
            #[(1.0, 'atm'), 3.160000e-04, 4.05, -1144.0],
            #[(10.0, 'atm'), 7.590000e-06, 4.49, -680.0],
            #[(100.0, 'atm'), 5.450000e-05, 4.22, 1141.0])

## Reaction 2165
#reaction('PC4H8OH-2 + O2 <=> C4H8OH-1O2', [2.000000e+12, 0.0, 0.0])

## Reaction 2166
#reaction('SC4H8OH-3 + O2 <=> C4H8OH-2O2', [2.000000e+12, 0.0, 0.0])

## Reaction 2167
#reaction('C4H8OH-1O2 => C2H5CHO + CH2O + OH', [1.000000e+16, 0.0, 25000.0])

## Reaction 2168
#reaction('C4H8OH-2O2 => OH + CH3CHO + CH3CHO', [1.000000e+16, 0.0, 25000.0])

## Reaction 2169
#pdep_arrhenius('SC4H8OH-3 <=> C4H8-2 + OH',
            #[(0.1, 'atm'), 7.080000e+27, -5.66, 27148.0],
            #[(1.0, 'atm'), 1.720000e+32, -6.54, 31378.0],
            #[(10.0, 'atm'), 1.760000e+32, -6.17, 33789.0],
            #[(100.0, 'atm'), 8.940000e+26, -4.34, 33113.0],
            #options='duplicate')

## Reaction 2170
#pdep_arrhenius('SC4H8OH-3 <=> C-2-C4H8 + OH',
            #[(0.1, 'atm'), 2.910000e+26, -5.3, 27166.0],
            #[(1.0, 'atm'), 2.460000e+31, -6.35, 31545.0],
            #[(10.0, 'atm'), 1.690000e+32, -6.21, 34360.0],
            #[(100.0, 'atm'), 2.670000e+27, -4.52, 33972.0])

## Reaction 2171
#reaction('C-2-C4H8 + OH <=> C4H71-3 + H2O', [2.580000e+14, 0.0, 6571.67])

## Reaction 2172
#pdep_arrhenius('SC4H8OH-3 <=> C4H8-2 + OH',
            #[(0.1, 'atm'), 1.020000e+35, -7.7, 28159.0],
            #[(1.0, 'atm'), 8.010000e+34, -7.29, 30492.0],
            #[(10.0, 'atm'), 3.260000e+30, -5.66, 30608.0],
            #[(100.0, 'atm'), 8.890000e+22, -3.21, 28554.0],
            #options='duplicate')

## Reaction 2173
#reaction('C4H8OOH1-2 <=> C4H8O1-2 + OH', [1.710000e+09, 1.06, 10900.0])

## Reaction 2174
#reaction('C4H8OOH1-3 <=> C4H8O1-3 + OH', [2.590000e+09, 0.69, 16000.0])

## Reaction 2175
#reaction('C4H8OOH1-4 <=> C4H8O1-4 + OH', [1.720000e+08, 0.76, 11100.0])

## Reaction 2176
#reaction('C4H8OOH2-4 <=> C4H8O1-3 + OH', [2.440000e+09, 0.78, 18000.0])

## Reaction 2177
#reaction('C4H8OOH1-3 => OH + CH2O + C3H6', [1.230000e+09, 1.3, 24900.0])

## Reaction 2178
#reaction('C4H8OOH2-4 => OH + CH3CHO + C2H4', [3.080000e+08, 1.5, 23500.0])

## Reaction 2179
#pdep_arrhenius('C4H8-1 + HO2 <=> C4H8OOH2-1',
            #[(0.0133, 'atm'), 9.740000e+05, 0.41, 7570.0],
            #[(1.0, 'atm'), 8.430000e+14, -1.76, 12124.0],
            #[(10.0, 'atm'), 1.820000e+20, -3.03, 16135.0],
            #[(100.0, 'atm'), 8.790000e+17, -2.04, 17111.0])

## Reaction 2180
#pdep_arrhenius('C4H8-1 + HO2 <=> C4H8O1-2 + OH',
            #[(0.0133, 'atm'), 2.960000e+03, 2.45, 12199.0],
            #[(1.0, 'atm'), 1.690000e+10, 0.51, 16606.0],
            #[(10.0, 'atm'), 2.190000e+17, -1.48, 22209.0],
            #[(100.0, 'atm'), 2.730000e+21, -2.51, 27857.0])

## Reaction 2181
#pdep_arrhenius('C4H8OOH2-1 <=> C4H8O1-2 + OH',
            #[(0.0133, 'atm'), 2.750000e+22, -4.39, 12349.0],
            #[(1.0, 'atm'), 2.440000e+29, -5.9, 17573.0],
            #[(10.0, 'atm'), 1.520000e+30, -5.84, 19702.0],
            #[(100.0, 'atm'), 1.930000e+20, -2.66, 16454.0])

## Reaction 2182
#pdep_arrhenius('C4H8-2 + HO2 <=> C4H8OOH2-3',
            #[(0.0133, 'atm'), 2.530000e+12, -1.27, 10098.0],
            #[(1.0, 'atm'), 1.340000e+11, -0.48, 8480.0],
            #[(10.0, 'atm'), 1.930000e+17, -2.04, 12122.0],
            #[(100.0, 'atm'), 4.450000e+19, -2.45, 14812.0])

## Reaction 2183
#pdep_arrhenius('C4H8-2 + HO2 <=> C4H8O2-3 + OH',
            #[(0.0133, 'atm'), 7.620000e+05, 1.92622, 10745.0],
            #[(1.0, 'atm'), 1.790000e+08, 1.25089, 12370.0],
            #[(10.0, 'atm'), 1.740000e+15, -0.74, 17220.0],
            #[(100.0, 'atm'), 8.370000e+20, -2.28314, 22838.0])

## Reaction 2184
#pdep_arrhenius('C4H8OOH2-3 <=> C4H8O2-3 + OH',
            #[(0.0133, 'atm'), 3.020000e+19, -3.51, 9746.0],
            #[(1.0, 'atm'), 6.370000e+24, -4.55, 13480.0],
            #[(10.0, 'atm'), 1.390000e+28, -5.25, 16470.0],
            #[(100.0, 'atm'), 1.540000e+27, -4.69, 17832.0])

## Reaction 2185
#reaction('C4H8-1 + HO2 <=> C4H8OOH1-2', [1.000000e+11, 0.0, 11000.0])

## Reaction 2186
#reaction('C4H8-1 + O <=> NC3H7 + HCO', [3.300000e+08, 1.45, -402.0])

## Reaction 2187
#reaction('C4H8-1 + O <=> C2H5 + CH2CHO', [1.390000e+07, 1.95, -596.0])

## Reaction 2188
#reaction('C4H8-1 + O <=> C3H6 + CH2O', [7.230000e+05, 2.34, -1050.0])

## Reaction 2189
#reaction('C4H8-1 + O <=> CH3CHO + C2H4', [1.250000e+13, 0.0, 850.0])

## Reaction 2190
#reaction('C4H8-1 + OH <=> C2H5CHO + CH3', [1.370000e+12, 0.0, -1040.0])

## Reaction 2191
#reaction('C4H8-1 + OH <=> NC3H7 + CH2O', [1.370000e+12, 0.0, -1040.0])

## Reaction 2192
#reaction('C4H8-2 + O <=> CH3CO + C2H5', [2.170000e+05, 2.34, -1050.0])

## Reaction 2193
#reaction('C4H8-2 + O <=> CH3CHCHO + CH3', [1.810000e+05, 2.34, -1050.0])

## Reaction 2194
#reaction('C4H8-2 + O <=> CH3CHO + C2H4', [3.250000e+05, 2.34, -1050.0])

## Reaction 2195
#reaction('C4H8-2 + OH <=> C2H5 + CH3CHO', [2.740000e+12, 0.0, -1040.0])

## Reaction 2196
#reaction('PC4H9 + O2 <=> C4H8-1 + HO2', [8.370000e-01, 3.59, 11960.0])

## Reaction 2197
#reaction('SC4H9 + O2 <=> C4H8-1 + HO2', [5.350000e-01, 3.71, 9322.0])

## Reaction 2198
#reaction('SC4H9 + O2 <=> C4H8-2 + HO2', [1.070000e+00, 3.71, 9322.0])

## Reaction 2199
#reaction('C4H8-1 + CH3O2 <=> C4H8O1-2 + CH3O', [1.000000e+12, 0.0, 14340.0])

## Reaction 2200
#reaction('C4H8-2 + CH3O2 <=> C4H8O2-3 + CH3O', [5.620000e+11, 0.0, 12310.0])

## Reaction 2201
#reaction('C-2-C4H8 + OH <=> C2H5 + CH3CHO', [2.740000e+12, 0.0, -1040.0])

## Reaction 2202
#reaction('SC4H9 + C4H71-3 <=> C-2-C4H8 + C4H8-2', [4.330000e+14, -0.75, -130.0])

## Reaction 2203
#reaction('SC4H9 + C4H71-3 <=> C-2-C4H8 + C-2-C4H8', [4.330000e+14, -0.75, -130.0])

## Reaction 2204
#reaction('C-2-C4H8 + O2 <=> C4H71-3 + HO2', [4.000000e+13, 0.0, 39390.0])

## Reaction 2205
#reaction('C-2-C4H8 + O <=> CH3CO + C2H5', [2.170000e+05, 2.34, -1050.0])

## Reaction 2206
#reaction('C-2-C4H8 + O <=> CH3CHCHO + CH3', [1.810000e+05, 2.34, -1050.0])

## Reaction 2207
#reaction('C-2-C4H8 + O <=> CH3CHO + C2H4', [3.250000e+05, 2.34, -1050.0])

## Reaction 2208
#pdep_arrhenius('C-2-C4H8 + H <=> C3H6 + CH3',
            #[(0.0013, 'atm'), 3.080000e+09, 1.35, 2542.0],
            #[(0.04, 'atm'), 1.576000e+11, 0.87, 3599.6],
            #[(1.0, 'atm'), 5.340000e+12, 0.47, 5431.1],
            #[(10.0, 'atm'), 1.850000e+23, -2.6, 12898.0],
            #[(100.0, 'atm'), 2.640000e+23, -2.42, 16500.0],
            #options='duplicate')

## Reaction 2209
#pdep_arrhenius('C-2-C4H8 + H <=> C3H6 + CH3',
            #[(10.0, 'atm'), 2.480000e+05, 2.52, 3679.1],
            #[(100.0, 'atm'), 4.020000e+03, 2.91, 3980.9],
            #options='duplicate')

## Reaction 2210
#reaction('PC4H9 + HO2 <=> C4H8-1 + H2O2', [2.410000e+13, 0.0, 0.0])

## Reaction 2211
#reaction('PC4H9 + OH <=> C4H8-1 + H2O', [2.400000e+13, 0.0, 0.0])

## Reaction 2212
#reaction('PC4H9 + H <=> C4H8-1 + H2', [1.250000e+13, 0.0, 0.0])

## Reaction 2213
#reaction('PC4H9 + CH3 <=> C4H8-1 + CH4', [2.000000e+12, 0.0, 0.0])

## Reaction 2214
#reaction('PC4H9 + C2H5 <=> C4H8-1 + C2H6', [1.600000e+12, 0.0, 0.0])

## Reaction 2215
#reaction('PC4H9 + C3H5-A <=> C4H8-1 + C3H6', [1.000000e+12, 0.0, 0.0])

## Reaction 2216
#reaction('PC4H9 + C4H71-3 <=> C4H8-1 + C4H8-1', [1.000000e+12, 0.0, 0.0])

## Reaction 2217
#reaction('PC4H9 + O <=> CH2O + NC3H7', [4.818000e+13, 0.0, 0.0])

## Reaction 2218
#reaction('SC4H9 + O2 <=> C-2-C4H8 + HO2', [1.850000e+16, -1.63, 3420.0])

## Reaction 2219
#reaction('SC4H9 + HO2 <=> C4H8-1 + H2O2', [2.400000e+13, 0.0, 0.0])

## Reaction 2220
#reaction('SC4H9 + HO2 <=> C4H8-2 + H2O2', [2.400000e+13, 0.0, 0.0])

## Reaction 2221
#reaction('SC4H9 + HO2 <=> C-2-C4H8 + H2O2', [2.400000e+13, 0.0, 0.0])

## Reaction 2222
#reaction('SC4H9 + OH <=> C4H8-1 + H2O', [2.400000e+13, 0.0, 0.0])

## Reaction 2223
#reaction('SC4H9 + OH <=> C4H8-2 + H2O', [2.400000e+13, 0.0, 0.0])

## Reaction 2224
#reaction('SC4H9 + OH <=> C-2-C4H8 + H2O', [2.400000e+13, 0.0, 0.0])

## Reaction 2225
#reaction('SC4H9 + H <=> C4H8-1 + H2', [5.000000e+13, 0.0, 0.0])

## Reaction 2226
#reaction('SC4H9 + H <=> C4H8-2 + H2', [5.000000e+13, 0.0, 0.0])

## Reaction 2227
#reaction('SC4H9 + H <=> C-2-C4H8 + H2', [5.000000e+13, 0.0, 0.0])

## Reaction 2228
#reaction('SC4H9 + CH3 <=> C4H8-1 + CH4', [1.000000e+12, 0.0, 0.0])

## Reaction 2229
#reaction('SC4H9 + CH3 <=> C4H8-2 + CH4', [1.000000e+12, 0.0, 0.0])

## Reaction 2230
#reaction('SC4H9 + CH3 <=> C-2-C4H8 + CH4', [1.000000e+12, 0.0, 0.0])

## Reaction 2231
#reaction('SC4H9 + O <=> CH3CHO + C2H5', [4.818000e+13, 0.0, 0.0])

## Reaction 2232
#reaction('C4H8O1-2 + OH => CH2O + C3H5-A + H2O', [5.000000e+12, 0.0, 0.0])

## Reaction 2233
#reaction('C4H8O1-2 + H => CH2O + C3H5-A + H2', [5.000000e+12, 0.0, 0.0])

## Reaction 2234
#reaction('C4H8O1-2 + O => CH2O + C3H5-A + OH', [5.000000e+12, 0.0, 0.0])

## Reaction 2235
#reaction('C4H8O1-2 + HO2 => CH2O + C3H5-A + H2O2', [1.000000e+13, 0.0, 15000.0])

## Reaction 2236
#reaction('C4H8O1-2 + CH3O2 => CH2O + C3H5-A + CH3O2H', [1.000000e+13, 0.0, 19000.0])

## Reaction 2237
#reaction('C4H8O1-2 + CH3 => CH2O + C3H5-A + CH4', [2.000000e+11, 0.0, 10000.0])

## Reaction 2238
#reaction('C4H8O1-3 + OH => CH2O + C3H5-A + H2O', [5.000000e+12, 0.0, 0.0])

## Reaction 2239
#reaction('C4H8O1-3 + H => CH2O + C3H5-A + H2', [5.000000e+12, 0.0, 0.0])

## Reaction 2240
#reaction('C4H8O1-3 + O => CH2O + C3H5-A + OH', [5.000000e+12, 0.0, 0.0])

## Reaction 2241
#reaction('C4H8O1-3 + HO2 => CH2O + C3H5-A + H2O2', [1.000000e+13, 0.0, 15000.0])

## Reaction 2242
#reaction('C4H8O1-3 + CH3O2 => CH2O + C3H5-A + CH3O2H', [1.000000e+13, 0.0, 19000.0])

## Reaction 2243
#reaction('C4H8O1-3 + CH3 => CH2O + C3H5-A + CH4', [2.000000e+11, 0.0, 10000.0])

## Reaction 2244
#reaction('C4H8O1-4 + OH => CH2O + C3H5-A + H2O', [5.000000e+12, 0.0, 0.0])

## Reaction 2245
#reaction('C4H8O1-4 + H => CH2O + C3H5-A + H2', [5.000000e+12, 0.0, 0.0])

## Reaction 2246
#reaction('C4H8O1-4 + O => CH2O + C3H5-A + OH', [5.000000e+12, 0.0, 0.0])

## Reaction 2247
#reaction('C4H8O1-4 + HO2 => CH2O + C3H5-A + H2O2', [1.000000e+13, 0.0, 15000.0])

## Reaction 2248
#reaction('C4H8O1-4 + CH3O2 => CH2O + C3H5-A + CH3O2H', [1.000000e+13, 0.0, 19000.0])

## Reaction 2249
#reaction('C4H8O1-4 + CH3 => CH2O + C3H5-A + CH4', [2.000000e+11, 0.0, 10000.0])

## Reaction 2250
#reaction('C4H8O2-3 + OH => CH2O + C3H5-A + H2O', [5.000000e+12, 0.0, 0.0])

## Reaction 2251
#reaction('C4H8O2-3 + H => CH2O + C3H5-A + H2', [5.000000e+12, 0.0, 0.0])

## Reaction 2252
#reaction('C4H8O2-3 + O => CH2O + C3H5-A + OH', [5.000000e+12, 0.0, 0.0])

## Reaction 2253
#reaction('C4H8O2-3 + HO2 => CH2O + C3H5-A + H2O2', [1.000000e+13, 0.0, 15000.0])

## Reaction 2254
#reaction('C4H8O2-3 + CH3O2 => CH2O + C3H5-A + CH3O2H', [1.000000e+13, 0.0, 19000.0])

## Reaction 2255
#reaction('C4H8O2-3 + CH3 => CH2O + C3H5-A + CH4', [2.000000e+11, 0.0, 10000.0])

## Reaction 2256
#reaction('PC4H9 + O2 <=> PC4H9O2', [6.865000e+16, -1.627, 198.7])

## Reaction 2257
#reaction('SC4H9 + O2 <=> SC4H9O2', [3.487000e+14, -0.816, -536.5])

## Reaction 2258
#reaction('SC4H9O2 + CH2O <=> SC4H9O2H + HCO', [5.600000e+12, 0.0, 13600.0])

## Reaction 2259
#reaction('SC4H9O2 + CH3CHO <=> SC4H9O2H + CH3CO', [2.800000e+12, 0.0, 13600.0])

## Reaction 2260
#reaction('SC4H9O2 + HO2 <=> SC4H9O2H + O2', [1.750000e+10, 0.0, -3275.0])

## Reaction 2261
#reaction('IC3H7O2 + PC4H9 <=> IC3H7O + PC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 2262
#reaction('IC3H7O2 + SC4H9 <=> IC3H7O + SC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 2263
#reaction('NC3H7O2 + PC4H9 <=> NC3H7O + PC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 2264
#reaction('NC3H7O2 + SC4H9 <=> NC3H7O + SC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 2265
#reaction('SC4H9O2 + SC4H9O2 => O2 + SC4H9O + SC4H9O', [1.400000e+16, -1.61, 1860.0])

## Reaction 2266
#reaction('SC4H9O2 + NC3H7O2 => SC4H9O + NC3H7O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 2267
#reaction('SC4H9O2 + IC3H7O2 => SC4H9O + IC3H7O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 2268
#reaction('SC4H9O2 + C2H5O2 => SC4H9O + C2H5O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 2269
#reaction('SC4H9O2 + CH3O2 => SC4H9O + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 2270
#reaction('SC4H9O2 + CH3CO3 => SC4H9O + CH3CO2 + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 2271
#reaction('H2 + PC4H9O2 <=> H + PC4H9O2H', [3.010000e+13, 0.0, 26030.0])

## Reaction 2272
#reaction('H2 + SC4H9O2 <=> H + SC4H9O2H', [3.010000e+13, 0.0, 26030.0])

## Reaction 2273
#reaction('C2H6 + PC4H9O2 <=> C2H5 + PC4H9O2H', [1.700000e+13, 0.0, 20460.0])

## Reaction 2274
#reaction('C2H6 + SC4H9O2 <=> C2H5 + SC4H9O2H', [1.700000e+13, 0.0, 20460.0])

## Reaction 2275
#reaction('PC4H9O2 + C2H5CHO <=> PC4H9O2H + C2H5CO', [2.000000e+11, 0.0, 9500.0])

## Reaction 2276
#reaction('SC4H9O2 + C2H5CHO <=> SC4H9O2H + C2H5CO', [2.000000e+11, 0.0, 9500.0])

## Reaction 2277
#reaction('SC4H9O2 + CH3 <=> SC4H9O + CH3O', [7.000000e+12, 0.0, -1000.0])

## Reaction 2278
#reaction('SC4H9O2 + C2H5 <=> SC4H9O + C2H5O', [7.000000e+12, 0.0, -1000.0])

## Reaction 2279
#reaction('SC4H9O2 + IC3H7 <=> SC4H9O + IC3H7O', [7.000000e+12, 0.0, -1000.0])

## Reaction 2280
#reaction('SC4H9O2 + NC3H7 <=> SC4H9O + NC3H7O', [7.000000e+12, 0.0, -1000.0])

## Reaction 2281
#reaction('SC4H9O2 + PC4H9 <=> SC4H9O + PC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 2282
#reaction('SC4H9O2 + SC4H9 <=> SC4H9O + SC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 2283
#reaction('SC4H9O2 + C3H5-A <=> SC4H9O + C3H5O', [7.000000e+12, 0.0, -1000.0])

## Reaction 2284
#reaction('PC4H9O2 + CH2O <=> PC4H9O2H + HCO', [5.600000e+12, 0.0, 13600.0])

## Reaction 2285
#reaction('PC4H9O2 + CH3CHO <=> PC4H9O2H + CH3CO', [2.800000e+12, 0.0, 13600.0])

## Reaction 2286
#reaction('PC4H9O2 + HO2 <=> PC4H9O2H + O2', [1.750000e+10, 0.0, -3275.0])

## Reaction 2287
#reaction('C3H6 + PC4H9O2 <=> C3H5-A + PC4H9O2H', [5.350000e-02, 4.207, 13288.1])

## Reaction 2288
#reaction('C3H6 + SC4H9O2 <=> C3H5-A + SC4H9O2H', [5.350000e-02, 4.207, 13288.1])

## Reaction 2289
#reaction('C2H4 + PC4H9O2 <=> C2H3 + PC4H9O2H', [1.130000e+13, 0.0, 30430.0])

## Reaction 2290
#reaction('C2H4 + SC4H9O2 <=> C2H3 + SC4H9O2H', [1.130000e+13, 0.0, 30430.0])

## Reaction 2291
#reaction('CH3OH + PC4H9O2 <=> CH2OH + PC4H9O2H', [6.300000e+12, 0.0, 19360.0])

## Reaction 2292
#reaction('CH3OH + SC4H9O2 <=> CH2OH + SC4H9O2H', [6.300000e+12, 0.0, 19360.0])

## Reaction 2293
#reaction('C2H3CHO + PC4H9O2 <=> C2H3CO + PC4H9O2H', [2.800000e+12, 0.0, 13600.0])

## Reaction 2294
#reaction('C2H3CHO + SC4H9O2 <=> C2H3CO + SC4H9O2H', [2.800000e+12, 0.0, 13600.0])

## Reaction 2295
#reaction('CH4 + PC4H9O2 <=> CH3 + PC4H9O2H', [1.120000e+13, 0.0, 24640.0])

## Reaction 2296
#reaction('CH4 + SC4H9O2 <=> CH3 + SC4H9O2H', [1.120000e+13, 0.0, 24640.0])

## Reaction 2297
#reaction('H2O2 + PC4H9O2 <=> HO2 + PC4H9O2H', [2.400000e+12, 0.0, 10000.0])

## Reaction 2298
#reaction('H2O2 + SC4H9O2 <=> HO2 + SC4H9O2H', [2.400000e+12, 0.0, 10000.0])

## Reaction 2299
#reaction('PC4H9O2 + PC4H9O2 => O2 + PC4H9O + PC4H9O', [1.400000e+16, -1.61, 1860.0])

## Reaction 2300
#reaction('PC4H9O2 + SC4H9O2 => PC4H9O + SC4H9O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 2301
#reaction('PC4H9O2 + NC3H7O2 => PC4H9O + NC3H7O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 2302
#reaction('PC4H9O2 + IC3H7O2 => PC4H9O + IC3H7O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 2303
#reaction('PC4H9O2 + C2H5O2 => PC4H9O + C2H5O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 2304
#reaction('PC4H9O2 + CH3O2 => PC4H9O + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 2305
#reaction('PC4H9O2 + CH3CO3 => PC4H9O + CH3CO2 + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 2306
#reaction('PC4H9O2 + CH3 <=> PC4H9O + CH3O', [7.000000e+12, 0.0, -1000.0])

## Reaction 2307
#reaction('PC4H9O2 + C2H5 <=> PC4H9O + C2H5O', [7.000000e+12, 0.0, -1000.0])

## Reaction 2308
#reaction('PC4H9O2 + IC3H7 <=> PC4H9O + IC3H7O', [7.000000e+12, 0.0, -1000.0])

## Reaction 2309
#reaction('PC4H9O2 + NC3H7 <=> PC4H9O + NC3H7O', [7.000000e+12, 0.0, -1000.0])

## Reaction 2310
#reaction('PC4H9O2 + PC4H9 <=> PC4H9O + PC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 2311
#reaction('PC4H9O2 + SC4H9 <=> PC4H9O + SC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 2312
#reaction('PC4H9O2 + C3H5-A <=> PC4H9O + C3H5O', [7.000000e+12, 0.0, -1000.0])

## Reaction 2313
#reaction('PC4H9 + HO2 <=> PC4H9O + OH', [7.000000e+12, 0.0, -1000.0])

## Reaction 2314
#reaction('SC4H9 + HO2 <=> SC4H9O + OH', [7.000000e+12, 0.0, -1000.0])

## Reaction 2315
#reaction('CH3O2 + PC4H9 <=> CH3O + PC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 2316
#reaction('CH3O2 + SC4H9 <=> CH3O + SC4H9O', [7.000000e+12, 0.0, -1000.0])

## Reaction 2317
#reaction('PC4H9O2H <=> PC4H9O + OH', [1.500000e+16, 0.0, 42500.0])

## Reaction 2318
#reaction('SC4H9O + OH <=> SC4H9O2H', [1.000000e+15, -0.8, 0.0])

## Reaction 2319
#reaction('NC3H7 + CH2O <=> PC4H9O', [5.000000e+10, 0.0, 3457.0])

## Reaction 2320
#reaction('CH3 + C2H5CHO <=> SC4H9O', [5.000000e+10, 0.0, 9043.0])

## Reaction 2321
#reaction('C2H5 + CH3CHO <=> SC4H9O', [3.330000e+10, 0.0, 6397.0])

## Reaction 2322
#reaction('PC4H9O2 <=> C4H8OOH1-2', [4.009000e+08, 1.1, 30100.0])

## Reaction 2323
#reaction('PC4H9O2 <=> C4H8OOH1-3', [1.360000e+07, 1.3, 18200.0])

## Reaction 2324
#reaction('PC4H9O2 <=> C4H8OOH1-4', [1.233000e+06, 1.5, 20000.0])

## Reaction 2325
#reaction('SC4H9O2 <=> C4H8OOH2-1', [1.458000e+09, 1.1, 33500.0])

## Reaction 2326
#reaction('SC4H9O2 <=> C4H8OOH2-3', [1.716000e+09, 0.9, 29500.0])

## Reaction 2327
#reaction('SC4H9O2 <=> C4H8OOH2-4', [1.439000e+07, 1.4, 20800.0])

## Reaction 2328
#reaction('PC4H9O2 <=> C4H8-1 + HO2', [6.290000e+07, 1.38, 28900.0])

## Reaction 2329
#reaction('SC4H9O2 <=> C4H8-1 + HO2', [1.710000e+09, 1.0, 30400.0])

## Reaction 2330
#reaction('SC4H9O2 <=> C4H8-2 + HO2', [7.250000e+09, 0.8, 29900.0],
        #options='duplicate')

## Reaction 2331
#reaction('SC4H9O2 <=> C4H8-2 + HO2', [1.700000e+10, 0.67, 30700.0],
        #options='duplicate')

## Reaction 2332
#reaction('C4H8OOH1-2 + O2 <=> C4H8OOH1-2O2', [1.744000e+14, -0.816, -536.5])

## Reaction 2333
#reaction('C4H8OOH1-3 + O2 <=> C4H8OOH1-3O2', [1.744000e+14, -0.816, -536.5])

## Reaction 2334
#reaction('C4H8OOH1-4 + O2 <=> C4H8OOH1-4O2', [3.433000e+16, -1.627, 198.7])

## Reaction 2335
#reaction('C4H8OOH2-1 + O2 <=> C4H8OOH2-1O2', [3.433000e+16, -1.627, 198.7])

## Reaction 2336
#reaction('C4H8OOH2-3 + O2 <=> C4H8OOH2-3O2', [1.744000e+14, -0.816, -536.5])

## Reaction 2337
#reaction('C4H8OOH2-4 + O2 <=> C4H8OOH2-4O2', [3.433000e+16, -1.627, 198.7])

## Reaction 2338
#reaction('C4H8OOH1-2O2 <=> C4H72-1OOH + HO2', [2.540000e+10, 0.804, 30098.5])

## Reaction 2339
#reaction('C4H8OOH1-3O2 <=> C4H71-4OOH + HO2', [5.130000e+09, 1.0, 30400.0])

## Reaction 2340
#reaction('C4H8OOH1-3O2 <=> C4H72-1OOH + HO2', [2.540000e+10, 0.804, 30098.5])

## Reaction 2341
#reaction('C4H8OOH1-4O2 <=> C4H71-4OOH + HO2', [1.440000e+07, 1.38, 28900.0])

## Reaction 2342
#reaction('C4H8OOH2-3O2 <=> C4H71-3OOH + HO2', [5.130000e+09, 1.0, 30400.0])

## Reaction 2343
#reaction('C4H8OOH2-4O2 <=> C4H71-3OOH + HO2', [1.440000e+07, 1.38, 28900.0])

## Reaction 2344
#reaction('C4H8OOH1-2O2 <=> C4H71-3,4OOH', [1.440000e+07, 1.4, 20800.0])

## Reaction 2345
#reaction('C4H8OOH1-2O2 <=> C4H72-3,4OOH', [1.720000e+09, 0.9, 29500.0])

## Reaction 2346
#reaction('C4H8OOH1-2O2 <=> NC4KET12 + OH', [2.440000e+07, 1.6, 27900.0])

## Reaction 2347
#reaction('C4H8OOH1-3O2 <=> C4H71-2,4OOH', [1.460000e+09, 1.1, 33500.0])

## Reaction 2348
#reaction('C4H8OOH1-3O2 <=> C4H72-1,3OOH', [1.720000e+09, 0.9, 29500.0])

## Reaction 2349
#reaction('C4H8OOH1-3O2 <=> NC4KET13 + OH', [1.090000e+04, 2.4, 19900.0])

## Reaction 2350
#reaction('C4H8OOH1-4O2 <=> C4H72-1,4OOH', [1.360000e+07, 1.3, 18200.0],
        #options='duplicate')

## Reaction 2351
#reaction('C4H8OOH1-4O2 <=> C4H72-1,4OOH', [4.010000e+08, 1.1, 30100.0],
        #options='duplicate')

## Reaction 2352
#reaction('C4H8OOH1-4O2 <=> NC4KET14 + OH', [4.800000e+03, 1.7, 16600.0])

## Reaction 2353
#reaction('C4H8OOH2-1O2 <=> C4H72-3,4OOH', [1.720000e+09, 0.9, 29500.0])

## Reaction 2354
#reaction('C4H8OOH2-1O2 <=> C4H71-3,4OOH', [1.230000e+06, 1.5, 20000.0])

## Reaction 2355
#reaction('C4H8OOH2-1O2 <=> NC4KET21 + OH', [2.760000e+08, 1.2, 25700.0])

## Reaction 2356
#reaction('C4H8OOH2-3O2 <=> C4H71-2,3OOH', [1.460000e+09, 1.1, 33500.0],
        #options='duplicate')

## Reaction 2357
#reaction('C4H8OOH2-3O2 <=> C4H71-2,3OOH', [1.440000e+07, 1.4, 20800.0],
        #options='duplicate')

## Reaction 2358
#reaction('C4H8OOH2-3O2 <=> NC4KET23 + OH', [1.750000e+06, 1.7, 26000.0])

## Reaction 2359
#reaction('C4H8OOH2-4O2 <=> C4H71-2,4OOH', [1.230000e+06, 1.5, 20000.0])

## Reaction 2360
#reaction('C4H8OOH2-4O2 <=> C4H72-1,3OOH', [4.010000e+08, 1.1, 30100.0])

## Reaction 2361
#reaction('C4H8OOH2-4O2 <=> NC4KET24 + OH', [5.790000e+01, 2.9, 17000.0])

## Reaction 2362
#reaction('C4H71-3,4OOH <=> C4H7O1-3OOH-4 + OH', [2.440000e+09, 0.78, 18000.0])

## Reaction 2363
#reaction('C4H71-3,4OOH <=> C4H7O1-4OOH-2 + OH', [1.720000e+09, 0.76, 11100.0])

## Reaction 2364
#reaction('C4H72-3,4OOH <=> C4H7O2-3OOH-1 + OH', [6.990000e+09, 0.815, 9788.3])

## Reaction 2365
#reaction('C4H72-3,4OOH <=> C4H7O1-3OOH-2 + OH', [2.590000e+09, 0.69, 16000.0])

## Reaction 2366
#reaction('C4H71-2,4OOH <=> C4H7O1-2OOH-4 + OH', [2.350000e+10, 0.68, 10800.0])

## Reaction 2367
#reaction('C4H71-2,4OOH <=> C4H7O1-4OOH-2 + OH', [1.720000e+09, 0.76, 11100.0])

## Reaction 2368
#reaction('C4H72-1,3OOH <=> C4H7O1-2OOH-3 + OH', [1.710000e+09, 1.06, 10900.0])

## Reaction 2369
#reaction('C4H72-1,3OOH <=> C4H7O2-3OOH-1 + OH', [6.990000e+09, 0.815, 9788.3])

## Reaction 2370
#reaction('C4H72-1,4OOH <=> C4H7O1-2OOH-4 + OH', [1.710000e+09, 1.06, 10900.0])

## Reaction 2371
#reaction('C4H72-1,4OOH <=> C4H7O1-3OOH-4 + OH', [2.590000e+09, 0.69, 16000.0])

## Reaction 2372
#reaction('C4H71-2,3OOH <=> C4H7O1-2OOH-3 + OH', [2.350000e+10, 0.68, 10800.0])

## Reaction 2373
#reaction('C4H71-2,3OOH <=> C4H7O1-3OOH-2 + OH', [2.440000e+09, 0.78, 18000.0])

## Reaction 2374
#reaction('C4H71-3,4OOH => C2H4 + OH + HO2CH2CHO', [1.230000e+09, 1.3, 24900.0])

## Reaction 2375
#reaction('C4H72-3,4OOH <=> C4H72-1OOH + HO2', [2.568000e+11, 0.538, 15324.7])

## Reaction 2376
#reaction('C4H71-2,4OOH <=> C4H71-4OOH + HO2', [2.670000e+11, 0.5, 15800.0])

## Reaction 2377
#reaction('C4H72-1,3OOH <=> C4H72-1OOH + HO2', [2.568000e+11, 0.538, 15324.7])

## Reaction 2378
#reaction('C4H72-1,3OOH <=> C4H71-3OOH + HO2', [6.030000e+09, 0.95, 15200.0])

## Reaction 2379
#reaction('C4H72-1,4OOH <=> C4H71-4OOH + HO2', [2.568000e+11, 0.538, 15324.7])

## Reaction 2380
#reaction('C4H71-2,3OOH <=> C4H71-3OOH + HO2', [6.030000e+09, 0.95, 15200.0])

## Reaction 2381
#reaction('C4H72-1,4OOH <=> C4H72-1OOH + HO2', [2.670000e+11, 0.5, 15800.0])

## Reaction 2382
#reaction('HO2CH2CHO => CH2O + HCO + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 2383
#reaction('C4H72-1OOH => CH2O + C3H5-S + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 2384
#reaction('C4H71-4OOH => CH2O + C3H5-A + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 2385
#reaction('C4H71-3OOH => C2H3CHO + CH3 + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 2386
#reaction('C4H71-3OOH => CH3CHO + C2H3 + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 2387
#reaction('NC4KET12 => C2H5CHO + HCO + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 2388
#reaction('NC4KET13 => CH3CHO + CH2CHO + OH', [1.050000e+16, 0.0, 43000.0])

## Reaction 2389
#reaction('NC4KET14 => CH2CH2CHO + CH2O + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 2390
#reaction('NC4KET21 => CH2O + C2H5CO + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 2391
#reaction('NC4KET23 => CH3CHO + CH3CO + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 2392
#reaction('NC4KET24 => CH2O + CH3COCH2 + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 2393
#reaction('C2H5COCH3 + OH <=> CH2CH2COCH3 + H2O', [7.550000e+09, 0.97, 1586.0])

## Reaction 2394
#reaction('C2H5COCH3 + OH <=> CH3CHCOCH3 + H2O', [8.450000e+11, 0.0, -228.0])

## Reaction 2395
#reaction('C2H5COCH3 + OH <=> C2H5COCH2 + H2O', [5.100000e+11, 0.0, 1192.0])

## Reaction 2396
#reaction('C2H5COCH3 + HO2 <=> CH2CH2COCH3 + H2O2', [2.380000e+04, 2.55, 16490.0])

## Reaction 2397
#reaction('C2H5COCH3 + HO2 <=> CH3CHCOCH3 + H2O2', [2.000000e+11, 0.0, 8698.0])

## Reaction 2398
#reaction('C2H5COCH3 + HO2 <=> C2H5COCH2 + H2O2', [2.380000e+04, 2.55, 14690.0])

## Reaction 2399
#reaction('C2H5COCH3 + O <=> CH2CH2COCH3 + OH', [2.250000e+13, 0.0, 7700.0])

## Reaction 2400
#reaction('C2H5COCH3 + O <=> CH3CHCOCH3 + OH', [3.070000e+13, 0.0, 3400.0])

## Reaction 2401
#reaction('C2H5COCH3 + O <=> C2H5COCH2 + OH', [5.000000e+12, 0.0, 5962.0])

## Reaction 2402
#reaction('C2H5COCH3 + H <=> CH2CH2COCH3 + H2', [9.160000e+06, 2.0, 7700.0])

## Reaction 2403
#reaction('C2H5COCH3 + H <=> CH3CHCOCH3 + H2', [4.460000e+06, 2.0, 3200.0])

## Reaction 2404
#reaction('C2H5COCH3 + H <=> C2H5COCH2 + H2', [9.300000e+12, 0.0, 6357.0])

## Reaction 2405
#reaction('C2H5COCH3 + O2 <=> CH2CH2COCH3 + HO2', [2.050000e+13, 0.0, 51310.0])

## Reaction 2406
#reaction('C2H5COCH3 + O2 <=> CH3CHCOCH3 + HO2', [1.550000e+13, 0.0, 41970.0])

## Reaction 2407
#reaction('C2H5COCH3 + O2 <=> C2H5COCH2 + HO2', [2.050000e+13, 0.0, 49150.0])

## Reaction 2408
#reaction('C2H5COCH3 + CH3 <=> CH2CH2COCH3 + CH4', [3.190000e+01, 3.17, 7172.0])

## Reaction 2409
#reaction('C2H5COCH3 + CH3 <=> CH3CHCOCH3 + CH4', [1.740000e+00, 3.46, 3680.0])

## Reaction 2410
#reaction('C2H5COCH3 + CH3 <=> C2H5COCH2 + CH4', [1.620000e+11, 0.0, 9630.0])

## Reaction 2411
#reaction('C2H5COCH3 + CH3O <=> CH2CH2COCH3 + CH3OH', [2.170000e+11, 0.0, 6460.0])

## Reaction 2412
#reaction('C2H5COCH3 + CH3O <=> CH3CHCOCH3 + CH3OH', [1.450000e+11, 0.0, 2771.0])

## Reaction 2413
#reaction('C2H5COCH3 + CH3O <=> C2H5COCH2 + CH3OH', [2.170000e+11, 0.0, 4660.0])

## Reaction 2414
#reaction('C2H5COCH3 + CH3O2 <=> CH2CH2COCH3 + CH3O2H', [3.010000e+12, 0.0, 19380.0])

## Reaction 2415
#reaction('C2H5COCH3 + CH3O2 <=> CH3CHCOCH3 + CH3O2H', [2.000000e+12, 0.0, 15250.0])

## Reaction 2416
#reaction('C2H5COCH3 + CH3O2 <=> C2H5COCH2 + CH3O2H', [3.010000e+12, 0.0, 17580.0])

## Reaction 2417
#reaction('C2H5COCH3 + C2H3 <=> CH2CH2COCH3 + C2H4', [5.000000e+11, 0.0, 10400.0])

## Reaction 2418
#reaction('C2H5COCH3 + C2H3 <=> CH3CHCOCH3 + C2H4', [3.000000e+11, 0.0, 3400.0])

## Reaction 2419
#reaction('C2H5COCH3 + C2H3 <=> C2H5COCH2 + C2H4', [6.150000e+10, 0.0, 4278.0])

## Reaction 2420
#reaction('C2H5COCH3 + C2H5 <=> CH2CH2COCH3 + C2H6', [5.000000e+10, 0.0, 13400.0])

## Reaction 2421
#reaction('C2H5COCH3 + C2H5 <=> CH3CHCOCH3 + C2H6', [3.000000e+10, 0.0, 8600.0])

## Reaction 2422
#reaction('C2H5COCH3 + C2H5 <=> C2H5COCH2 + C2H6', [5.000000e+10, 0.0, 11600.0])

## Reaction 2423
#reaction('CH3CHCOCH3 + O2 <=> CH3CHOOCOCH3', [1.000000e+11, 0.0, 0.0])

## Reaction 2424
#reaction('CH3CHOOCOCH3 <=> CH2CHOOHCOCH3', [8.900000e+12, 0.0, 29700.0])

## Reaction 2425
#reaction('C2H3COCH3 + HO2 <=> CH2CHOOHCOCH3', [7.000000e+10, 0.0, 7800.0])

## Reaction 2426
#reaction('C2H5COCH2 <=> CH2CO + C2H5', [1.000000e+14, 0.0, 35000.0])

## Reaction 2427
#reaction('C2H3COCH3 + H <=> CH3CHCOCH3', [5.000000e+12, 0.0, 1200.0])

## Reaction 2428
#reaction('CH3CHCO + CH3 <=> CH3CHCOCH3', [1.230000e+11, 0.0, 7800.0])

## Reaction 2429
#reaction('NC3H7CHO + O2 <=> NC3H7CO + HO2', [1.200000e+05, 2.5, 37560.0])

## Reaction 2430
#reaction('NC3H7CHO + OH <=> NC3H7CO + H2O', [2.000000e+06, 1.8, -1300.0])

## Reaction 2431
#reaction('NC3H7CHO + H <=> NC3H7CO + H2', [4.140000e+09, 1.12, 2320.0])

## Reaction 2432
#reaction('NC3H7CHO + O <=> NC3H7CO + OH', [5.940000e+12, 0.0, 1868.0])

## Reaction 2433
#reaction('NC3H7CHO + HO2 <=> NC3H7CO + H2O2', [4.090000e+04, 2.5, 10200.0])

## Reaction 2434
#reaction('NC3H7CHO + CH3 <=> NC3H7CO + CH4', [2.890000e-03, 4.62, 3210.0])

## Reaction 2435
#reaction('NC3H7CHO + CH3O <=> NC3H7CO + CH3OH', [1.000000e+12, 0.0, 3300.0])

## Reaction 2436
#reaction('NC3H7CHO + CH3O2 <=> NC3H7CO + CH3O2H', [4.090000e+04, 2.5, 10200.0])

## Reaction 2437
#reaction('NC3H7CHO + OH <=> C3H6CHO-3 + H2O', [5.280000e+09, 0.97, 1586.0])

## Reaction 2438
#reaction('NC3H7CHO + OH <=> C3H6CHO-2 + H2O', [4.680000e+07, 1.61, -35.0])

## Reaction 2439
#reaction('NC3H7CHO + OH <=> C3H6CHO-1 + H2O', [5.520000e+02, 3.12, -1176.0])

## Reaction 2440
#reaction('NC3H7CHO + HO2 <=> C3H6CHO-3 + H2O2', [2.379000e+04, 2.55, 16490.0])

## Reaction 2441
#reaction('NC3H7CHO + HO2 <=> C3H6CHO-2 + H2O2', [9.640000e+03, 2.6, 13910.0])

## Reaction 2442
#reaction('NC3H7CHO + HO2 <=> C3H6CHO-1 + H2O2', [3.440000e+12, 0.05, 17880.0])

## Reaction 2443
#reaction('NC3H7CHO + CH3O2 <=> C3H6CHO-3 + CH3O2H', [2.379000e+04, 2.55, 16490.0])

## Reaction 2444
#reaction('NC3H7CHO + CH3O2 <=> C3H6CHO-2 + CH3O2H', [9.640000e+03, 2.6, 13910.0])

## Reaction 2445
#reaction('NC3H7CHO + CH3O2 <=> C3H6CHO-1 + CH3O2H', [3.440000e+12, 0.05, 17880.0])

## Reaction 2446
#reaction('NC3H7CO <=> NC3H7 + CO', [1.000000e+11, 0.0, 9600.0])

## Reaction 2447
#reaction('C3H6CHO-3 <=> C2H4 + CH2CHO', [7.400000e+11, 0.0, 21970.0])

## Reaction 2448
#reaction('C2H5CHCO + H <=> C3H6CHO-1', [5.000000e+12, 0.0, 1200.0])

## Reaction 2449
#reaction('C2H3CHO + CH3 <=> C3H6CHO-1', [1.230000e+11, 0.0, 7800.0])

## Reaction 2450
#reaction('SC3H5CHO + H <=> C3H6CHO-2', [5.000000e+12, 0.0, 2900.0])

## Reaction 2451
#reaction('C3H6 + HCO <=> C3H6CHO-2', [1.000000e+11, 0.0, 6000.0])

## Reaction 2452
#reaction('C2H5CHCO + OH <=> NC3H7 + CO2', [3.730000e+12, 0.0, -1010.0])

## Reaction 2453
#reaction('C2H5CHCO + H <=> NC3H7 + CO', [4.400000e+12, 0.0, 1459.0])

## Reaction 2454
#reaction('C2H5CHCO + O <=> C3H6 + CO2', [3.200000e+12, 0.0, -437.0])

## Reaction 2455
#reaction('C3H5-S + CO <=> SC3H5CO', [5.000000e+12, 0.0, 8000.0])

## Reaction 2456
#reaction('SC3H5CHO + HO2 <=> SC3H5CO + H2O2', [1.000000e+12, 0.0, 11920.0])

## Reaction 2457
#reaction('SC3H5CHO + CH3 <=> SC3H5CO + CH4', [3.980000e+12, 0.0, 8700.0])

## Reaction 2458
#reaction('SC3H5CHO + O <=> SC3H5CO + OH', [7.180000e+12, 0.0, 1389.0])

## Reaction 2459
#reaction('SC3H5CHO + O2 <=> SC3H5CO + HO2', [4.000000e+13, 0.0, 37600.0])

## Reaction 2460
#reaction('SC3H5CHO + H <=> SC3H5CO + H2', [2.600000e+12, 0.0, 2600.0])

## Reaction 2461
#reaction('C2H3COCH3 + OH <=> CH3CHO + CH3CO', [1.000000e+11, 0.0, 0.0])

## Reaction 2462
#reaction('C2H3COCH3 + OH => CH2CO + C2H3 + H2O', [5.100000e+11, 0.0, 1192.0])

## Reaction 2463
#reaction('C2H3COCH3 + HO2 => CH2CHO + CH3CO + OH', [6.030000e+09, 0.0, 7949.0])

## Reaction 2464
#reaction('C2H3COCH3 + HO2 => CH2CO + C2H3 + H2O2', [8.500000e+12, 0.0, 20460.0])

## Reaction 2465
#reaction('C2H3COCH3 + CH3O2 => CH2CHO + CH3CO + CH3O', [3.970000e+11, 0.0, 17050.0])

## Reaction 2466
#reaction('C2H3COCH3 + CH3O2 => CH2CO + C2H3 + CH3O2H', [3.010000e+12, 0.0, 17580.0])

## Reaction 2467
#reaction('C4H4 + CH3 <=> C4H3-I + CH4', [5.000000e+13, 0.0, 19800.0])

## Reaction 2468
#reaction('C4H4 + CH3 <=> C4H3-N + CH4', [1.000000e+14, 0.0, 22800.0])

## Reaction 2469
#reaction('C4H5-N + O2 => H + CO + C2H3CHO', [5.190000e+15, -1.26, 3312.62])

## Reaction 2470
#reaction('C4H7O2-1 <=> CH2O + C3H5-S', [7.940000e+14, 0.0, 19000.0])

## Reaction 2471
#reaction('C4H71-3 + HO2 <=> C4H7O2-1 + OH', [9.640000e+12, 0.0, 0.0])

## Reaction 2472
#pdep_arrhenius('NC5H12 <=> NC3H7 + C2H5',
            #[(0.01, 'atm'), 1.180000e+96, -23.17, 126046.0],
            #[(0.1, 'atm'), 1.855000e+89, -20.9, 125630.0],
            #[(1.0, 'atm'), 3.550000e+76, -17.02, 120458.0],
            #[(2.0, 'atm'), 3.475000e+71, -15.54, 117854.0],
            #[(5.0, 'atm'), 2.160000e+64, -13.43, 113815.0],
            #[(10.0, 'atm'), 4.550000e+58, -11.79, 110456.0],
            #[(20.0, 'atm'), 1.020000e+53, -10.15, 106996.0],
            #[(50.0, 'atm'), 9.000000e+45, -8.13, 102538.0])

## Reaction 2473
#pdep_arrhenius('NC5H12 <=> PC4H9 + CH3',
            #[(0.01, 'atm'), 7.090000e+92, -22.46, 126235.0],
            #[(0.1, 'atm'), 4.830000e+86, -20.36, 126331.0],
            #[(1.0, 'atm'), 3.240000e+74, -16.63, 121640.0],
            #[(2.0, 'atm'), 4.030000e+69, -15.18, 119142.0],
            #[(5.0, 'atm'), 3.000000e+62, -13.09, 115199.0],
            #[(10.0, 'atm'), 6.470000e+56, -11.44, 111875.0],
            #[(20.0, 'atm'), 1.360000e+51, -9.8, 108418.0],
            #[(50.0, 'atm'), 9.740000e+43, -7.75, 103923.0])

## Reaction 2474
#pdep_arrhenius('NC5H12 <=> C5H11-1 + H',
            #[(0.01, 'atm'), 1.990000e+74, -17.75, 124798.0],
            #[(0.1, 'atm'), 5.720000e+76, -18.01, 131106.0],
            #[(1.0, 'atm'), 4.420000e+72, -16.46, 132942.0],
            #[(2.0, 'atm'), 4.560000e+69, -15.52, 132148.0],
            #[(5.0, 'atm'), 2.640000e+64, -13.92, 130026.0],
            #[(10.0, 'atm'), 3.800000e+59, -12.48, 127648.0],
            #[(20.0, 'atm'), 1.440000e+54, -10.89, 124707.0],
            #[(50.0, 'atm'), 3.280000e+46, -8.67, 120241.0])

## Reaction 2475
#pdep_arrhenius('NC5H12 <=> C5H11-2 + H',
            #[(0.01, 'atm'), 6.930000e+79, -19.24, 125716.0],
            #[(0.1, 'atm'), 7.690000e+79, -18.83, 130425.0],
            #[(1.0, 'atm'), 2.340000e+73, -16.63, 130412.0],
            #[(2.0, 'atm'), 5.640000e+69, -15.52, 129087.0],
            #[(5.0, 'atm'), 6.770000e+63, -13.74, 126353.0],
            #[(10.0, 'atm'), 4.300000e+58, -12.21, 123622.0],
            #[(20.0, 'atm'), 1.060000e+53, -10.57, 120451.0],
            #[(50.0, 'atm'), 2.500000e+45, -8.36, 115883.0])

## Reaction 2476
#pdep_arrhenius('NC5H12 <=> C5H11-3 + H',
            #[(0.01, 'atm'), 3.490000e+79, -19.24, 125716.0],
            #[(0.1, 'atm'), 3.880000e+79, -18.84, 130425.0],
            #[(1.0, 'atm'), 1.180000e+73, -16.63, 130412.0],
            #[(2.0, 'atm'), 2.840000e+69, -15.52, 129087.0],
            #[(5.0, 'atm'), 3.410000e+63, -13.74, 126353.0],
            #[(10.0, 'atm'), 2.170000e+58, -12.21, 123622.0],
            #[(20.0, 'atm'), 5.320000e+52, -10.57, 120451.0],
            #[(50.0, 'atm'), 1.260000e+45, -8.36, 115883.0])

## Reaction 2477
#reaction('NC5H12 + H <=> C5H11-1 + H2', [3.490000e+05, 2.69, 6450.0])

## Reaction 2478
#reaction('NC5H12 + H <=> C5H11-2 + H2', [2.600000e+06, 2.4, 4471.0])

## Reaction 2479
#reaction('NC5H12 + H <=> C5H11-3 + H2', [1.300000e+06, 2.4, 4471.0])

## Reaction 2480
#reaction('NC5H12 + OH <=> C5H11-1 + H2O', [2.732000e+07, 1.813, 868.4])

## Reaction 2481
#reaction('NC5H12 + OH <=> C5H11-2 + H2O', [1.411000e+10, 0.935, 504.7])

## Reaction 2482
#reaction('NC5H12 + OH <=> C5H11-3 + H2O', [5.721000e+06, 1.811, -1015.5])

## Reaction 2483
#reaction('NC5H12 + O <=> C5H11-1 + OH', [1.130000e+14, 0.0, 7850.0])

## Reaction 2484
#reaction('NC5H12 + O <=> C5H11-2 + OH', [5.620000e+13, 0.0, 5200.0])

## Reaction 2485
#reaction('NC5H12 + O <=> C5H11-3 + OH', [2.810000e+13, 0.0, 5200.0])

## Reaction 2486
#reaction('NC5H12 + CH3 <=> C5H11-1 + CH4', [9.040000e-01, 3.65, 7154.0])

## Reaction 2487
#reaction('NC5H12 + CH3 <=> C5H11-2 + CH4', [1.680000e+05, 2.133, 7574.0])

## Reaction 2488
#reaction('NC5H12 + CH3 <=> C5H11-3 + CH4', [8.400000e+04, 2.133, 7574.0])

## Reaction 2489
#reaction('NC5H12 + HO2 <=> C5H11-1 + H2O2', [4.080000e+01, 3.59, 17160.0])

## Reaction 2490
#reaction('NC5H12 + HO2 <=> C5H11-2 + H2O2', [1.264000e+02, 3.37, 13720.0])

## Reaction 2491
#reaction('NC5H12 + HO2 <=> C5H11-3 + H2O2', [6.320000e+01, 3.37, 13720.0])

## Reaction 2492
#reaction('NC5H12 + CH3O2 <=> C5H11-1 + CH3O2H', [1.386000e+00, 3.97, 18280.0])

## Reaction 2493
#reaction('NC5H12 + CH3O2 <=> C5H11-2 + CH3O2H', [2.037000e+01, 3.58, 14810.0])

## Reaction 2494
#reaction('NC5H12 + CH3O2 <=> C5H11-3 + CH3O2H', [1.019000e+01, 3.58, 14810.0])

## Reaction 2495
#reaction('NC5H12 + C2H5 <=> C5H11-1 + C2H6', [1.000000e+11, 0.0, 13400.0])

## Reaction 2496
#reaction('NC5H12 + C2H5 <=> C5H11-2 + C2H6', [1.000000e+11, 0.0, 10400.0])

## Reaction 2497
#reaction('NC5H12 + C2H5 <=> C5H11-3 + C2H6', [5.000000e+10, 0.0, 10400.0])

## Reaction 2498
#reaction('NC5H12 + C2H3 <=> C5H11-1 + C2H4', [1.000000e+12, 0.0, 18000.0])

## Reaction 2499
#reaction('NC5H12 + C2H3 <=> C5H11-2 + C2H4', [8.000000e+11, 0.0, 16800.0])

## Reaction 2500
#reaction('NC5H12 + C2H3 <=> C5H11-3 + C2H4', [4.000000e+11, 0.0, 16800.0])

## Reaction 2501
#reaction('NC5H12 + C5H11-1 <=> C5H11-2 + NC5H12', [1.000000e+11, 0.0, 10400.0])

## Reaction 2502
#reaction('NC5H12 + C5H11-1 <=> C5H11-3 + NC5H12', [5.000000e+10, 0.0, 10400.0])

## Reaction 2503
#reaction('NC5H12 + C5H11-2 <=> C5H11-3 + NC5H12', [5.000000e+10, 0.0, 12300.0])

## Reaction 2504
#reaction('NC5H12 + O2CHO <=> C5H11-1 + HO2CHO', [1.386000e+00, 3.97, 18280.0])

## Reaction 2505
#reaction('NC5H12 + O2CHO <=> C5H11-2 + HO2CHO', [2.037000e+01, 3.58, 14810.0])

## Reaction 2506
#reaction('NC5H12 + O2CHO <=> C5H11-3 + HO2CHO', [1.019000e+01, 3.58, 14810.0])

## Reaction 2507
#reaction('NC5H12 + CH3O <=> C5H11-1 + CH3OH', [3.000000e+11, 0.0, 7000.0])

## Reaction 2508
#reaction('NC5H12 + CH3O <=> C5H11-2 + CH3OH', [2.200000e+11, 0.0, 5000.0])

## Reaction 2509
#reaction('NC5H12 + CH3O <=> C5H11-3 + CH3OH', [1.100000e+11, 0.0, 5000.0])

## Reaction 2510
#reaction('NC5H12 + O2 <=> C5H11-1 + HO2', [4.200000e+13, 0.0, 52800.0])

## Reaction 2511
#reaction('NC5H12 + O2 <=> C5H11-2 + HO2', [2.800000e+13, 0.0, 50160.0])

## Reaction 2512
#reaction('NC5H12 + O2 <=> C5H11-3 + HO2', [1.400000e+13, 0.0, 50160.0])

## Reaction 2513
#pdep_arrhenius('C5H11-1 <=> C2H4 + NC3H7',
            #[(0.1, 'atm'), 4.410000e+03, 2.192, 18827.0],
            #[(1.0, 'atm'), 8.060000e+20, -2.628, 29232.0],
            #[(10.0, 'atm'), 2.170000e+28, -4.578, 34864.0],
            #[(100.0, 'atm'), 6.470000e+24, -3.383, 34388.0],
            #[(1000.0, 'atm'), 2.340000e+17, -1.123, 31176.0])

## Reaction 2514
#pdep_arrhenius('C5H11-1 <=> H + C5H10-1',
            #[(0.1, 'atm'), 3.240000e-14, 7.022, 15354.0],
            #[(1.0, 'atm'), 5.100000e+12, -0.402, 29991.0],
            #[(10.0, 'atm'), 4.070000e+27, -4.483, 39814.0],
            #[(100.0, 'atm'), 1.020000e+26, -3.794, 40806.0],
            #[(1000.0, 'atm'), 6.520000e+16, -0.987, 36957.0])

## Reaction 2515
#pdep_arrhenius('C5H11-1 <=> C5H11-2',
            #[(0.1, 'atm'), 1.610000e+10, -0.128, 16305.0],
            #[(1.0, 'atm'), 2.880000e+17, -2.073, 21414.0],
            #[(10.0, 'atm'), 5.090000e+18, -2.284, 23337.0],
            #[(100.0, 'atm'), 1.380000e+14, -0.838, 21871.0],
            #[(1000.0, 'atm'), 1.600000e+08, 0.954, 19221.0])

## Reaction 2516
#pdep_arrhenius('C5H11-1 <=> C5H11-3',
            #[(0.1, 'atm'), 6.080000e-15, 6.876, 11901.0],
            #[(1.0, 'atm'), 7.680000e+07, 0.667, 25012.0],
            #[(10.0, 'atm'), 1.260000e+19, -2.35, 32938.0],
            #[(100.0, 'atm'), 1.160000e+16, -1.256, 33072.0],
            #[(1000.0, 'atm'), 6.060000e+06, 1.57, 29120.0])

## Reaction 2517
#pdep_arrhenius('C5H11-2 <=> C5H11-3',
            #[(0.1, 'atm'), 4.800000e-04, 3.725, 28298.0],
            #[(1.0, 'atm'), 3.410000e+22, -3.517, 42976.0],
            #[(10.0, 'atm'), 1.570000e+32, -6.005, 50425.0],
            #[(100.0, 'atm'), 3.120000e+24, -3.514, 48305.0],
            #[(1000.0, 'atm'), 1.590000e+14, -0.404, 43709.0])

## Reaction 2518
#pdep_arrhenius('C5H11-2 <=> C3H6 + C2H5',
            #[(0.1, 'atm'), 8.540000e+25, -4.241, 31303.0],
            #[(1.0, 'atm'), 2.340000e+31, -5.581, 35992.0],
            #[(10.0, 'atm'), 4.630000e+28, -4.592, 36186.0],
            #[(100.0, 'atm'), 3.770000e+20, -2.108, 32927.0],
            #[(1000.0, 'atm'), 3.560000e+14, -0.301, 30124.0])

## Reaction 2519
#pdep_arrhenius('C5H11-2 <=> C5H10-1 + H',
            #[(0.1, 'atm'), 8.260000e+10, -0.118, 29715.0],
            #[(1.0, 'atm'), 6.920000e+26, -4.456, 39997.0],
            #[(10.0, 'atm'), 3.810000e+29, -4.969, 43662.0],
            #[(100.0, 'atm'), 4.220000e+20, -2.16, 40523.0],
            #[(1000.0, 'atm'), 5.050000e+11, 0.526, 36461.0])

## Reaction 2520
#pdep_arrhenius('C5H11-2 <=> C5H10-2 + H',
            #[(0.1, 'atm'), 2.797000e+13, -0.833, 30422.7],
            #[(1.0, 'atm'), 9.922000e+27, -4.775, 39987.0],
            #[(10.0, 'atm'), 8.504000e+29, -5.063, 43150.6],
            #[(100.0, 'atm'), 1.026000e+21, -2.275, 39953.2],
            #[(1000.0, 'atm'), 3.141000e+12, 0.287, 36068.2])

## Reaction 2521
#pdep_arrhenius('C5H11-3 <=> C4H8-1 + CH3',
            #[(0.1, 'atm'), 1.310000e+16, -1.137, 29671.0],
            #[(1.0, 'atm'), 1.850000e+25, -3.643, 35787.0],
            #[(10.0, 'atm'), 1.010000e+24, -3.13, 36363.0],
            #[(100.0, 'atm'), 1.980000e+16, -0.793, 33017.0],
            #[(1000.0, 'atm'), 2.100000e+11, 0.685, 30583.0])

## Reaction 2522
#pdep_arrhenius('C5H11-3 <=> C5H10-2 + H',
            #[(0.1, 'atm'), 5.119000e+06, 1.511, 29138.7],
            #[(1.0, 'atm'), 4.068000e+22, -2.949, 38765.5],
            #[(10.0, 'atm'), 2.298000e+25, -3.569, 41777.2],
            #[(100.0, 'atm'), 8.104000e+16, -0.983, 38322.3],
            #[(1000.0, 'atm'), 1.214000e+10, 1.047, 35008.7])

## Reaction 2523
#reaction('C5H10-1 <=> C2H5 + C3H5-A', [9.864000e+21, -2.086, 75060.0])

## Reaction 2524
#reaction('C5H10-1 + H <=> C5H91-3 + H2', [3.376000e+05, 2.36, 207.0])

## Reaction 2525
#reaction('C5H10-1 + H <=> C5H91-4 + H2', [1.300000e+06, 2.4, 4471.0])

## Reaction 2526
#reaction('C5H10-1 + H <=> C5H91-5 + H2', [6.650000e+05, 2.54, 6756.0])

## Reaction 2527
#reaction('C5H10-1 + O <=> C5H91-3 + OH', [6.600000e+05, 2.43, 1210.0])

## Reaction 2528
#reaction('C5H10-1 + O <=> C5H91-4 + OH', [5.510000e+05, 2.45, 2830.0])

## Reaction 2529
#reaction('C5H10-1 + O <=> C5H91-5 + OH', [9.800000e+05, 2.43, 4750.0])

## Reaction 2530
#reaction('C5H10-1 + OH <=> C5H91-3 + H2O', [2.764000e+04, 2.64, -1919.0])

## Reaction 2531
#reaction('C5H10-1 + OH <=> C5H91-4 + H2O', [4.670000e+07, 1.61, -35.0])

## Reaction 2532
#reaction('C5H10-1 + OH <=> C5H91-5 + H2O', [5.270000e+09, 0.97, 1586.0])

## Reaction 2533
#reaction('C5H10-1 + CH3 <=> C5H91-3 + CH4', [3.690000e+00, 3.31, 4002.0])

## Reaction 2534
#reaction('C5H10-1 + CH3 <=> C5H91-4 + CH4', [1.510000e+00, 3.46, 5481.0])

## Reaction 2535
#reaction('C5H10-1 + CH3 <=> C5H91-5 + CH4', [4.521000e-01, 3.65, 7154.0])

## Reaction 2536
#reaction('C5H10-1 + O2 <=> C5H91-3 + HO2', [2.200000e+12, 0.0, 37220.0])

## Reaction 2537
#reaction('C5H10-1 + O2 <=> C5H91-4 + HO2', [2.000000e+13, 0.0, 49640.0])

## Reaction 2538
#reaction('C5H10-1 + O2 <=> C5H91-5 + HO2', [3.000000e+13, 0.0, 52290.0])

## Reaction 2539
#reaction('C5H10-1 + HO2 <=> C5H91-3 + H2O2', [4.820000e+03, 2.55, 10530.0])

## Reaction 2540
#reaction('C5H10-1 + HO2 <=> C5H91-4 + H2O2', [9.640000e+03, 2.6, 13910.0])

## Reaction 2541
#reaction('C5H10-1 + HO2 <=> C5H91-5 + H2O2', [2.380000e+04, 2.55, 16490.0])

## Reaction 2542
#reaction('C5H10-1 + CH3O2 <=> C5H91-3 + CH3O2H', [4.820000e+03, 2.55, 10530.0])

## Reaction 2543
#reaction('C5H10-1 + CH3O2 <=> C5H91-4 + CH3O2H', [9.640000e+03, 2.6, 13910.0])

## Reaction 2544
#reaction('C5H10-1 + CH3O2 <=> C5H91-5 + CH3O2H', [2.380000e+04, 2.55, 16490.0])

## Reaction 2545
#reaction('C5H10-1 + CH3O <=> C5H91-3 + CH3OH', [4.000000e+01, 2.9, 8609.0])

## Reaction 2546
#reaction('C5H10-1 + CH3O <=> C5H91-4 + CH3OH', [1.450000e+11, 0.0, 4571.0])

## Reaction 2547
#reaction('C5H10-1 + CH3O <=> C5H91-5 + CH3OH', [2.170000e+11, 0.0, 6458.0])

## Reaction 2548
#reaction('C5H10-2 + H <=> C5H91-3 + H2', [1.730000e+05, 2.5, 2492.0])

## Reaction 2549
#reaction('C5H10-2 + H <=> C5H92-4 + H2', [3.376000e+05, 2.36, 207.0])

## Reaction 2550
#reaction('C5H10-2 + H <=> C5H92-5 + H2', [6.651000e+05, 2.54, 6756.0])

## Reaction 2551
#reaction('C5H10-2 + O <=> C5H91-3 + OH', [4.410000e+05, 2.42, 3150.0])

## Reaction 2552
#reaction('C5H10-2 + O <=> C5H92-4 + OH', [9.900000e+05, 2.43, 1210.0])

## Reaction 2553
#reaction('C5H10-2 + O <=> C5H92-5 + OH', [9.800000e+05, 2.43, 4750.0])

## Reaction 2554
#reaction('C5H10-2 + OH <=> C5H91-3 + H2O', [3.120000e+06, 2.0, -298.0])

## Reaction 2555
#reaction('C5H10-2 + OH <=> C5H92-4 + H2O', [2.764000e+04, 2.64, -1919.0])

## Reaction 2556
#reaction('C5H10-2 + OH <=> C5H92-5 + H2O', [5.270000e+09, 0.97, 1586.0])

## Reaction 2557
#reaction('C5H10-2 + CH3 <=> C5H91-3 + CH4', [2.210000e+00, 3.5, 5675.0])

## Reaction 2558
#reaction('C5H10-2 + CH3 <=> C5H92-4 + CH4', [3.690000e+00, 3.31, 4002.0])

## Reaction 2559
#reaction('C5H10-2 + CH3 <=> C5H92-5 + CH4', [4.521000e-01, 3.65, 7154.0])

## Reaction 2560
#reaction('C5H10-2 + O2 <=> C5H91-3 + HO2', [3.300000e+12, 0.0, 39900.0])

## Reaction 2561
#reaction('C5H10-2 + O2 <=> C5H92-4 + HO2', [2.200000e+12, 0.0, 37220.0])

## Reaction 2562
#reaction('C5H10-2 + O2 <=> C5H92-5 + HO2', [3.000000e+13, 0.0, 52290.0])

## Reaction 2563
#reaction('C5H10-2 + HO2 <=> C5H91-3 + H2O2', [9.639000e+03, 2.6, 13910.0])

## Reaction 2564
#reaction('C5H10-2 + HO2 <=> C5H92-4 + H2O2', [4.820000e+03, 2.55, 10530.0])

## Reaction 2565
#reaction('C5H10-2 + HO2 <=> C5H92-5 + H2O2', [2.380000e+04, 2.55, 16490.0])

## Reaction 2566
#reaction('C5H10-2 + CH3O2 <=> C5H91-3 + CH3O2H', [9.639000e+03, 2.6, 13910.0])

## Reaction 2567
#reaction('C5H10-2 + CH3O2 <=> C5H92-4 + CH3O2H', [4.820000e+03, 2.55, 10530.0])

## Reaction 2568
#reaction('C5H10-2 + CH3O2 <=> C5H92-5 + CH3O2H', [2.380000e+04, 2.55, 16490.0])

## Reaction 2569
#reaction('C5H10-2 + CH3O <=> C5H91-3 + CH3OH', [9.000000e+01, 2.95, 11990.0])

## Reaction 2570
#reaction('C5H10-2 + CH3O <=> C5H92-4 + CH3OH', [4.000000e+01, 2.9, 8609.0])

## Reaction 2571
#reaction('C5H10-2 + CH3O <=> C5H92-5 + CH3OH', [2.170000e+11, 0.0, 6458.0])

## Reaction 2572
#reaction('C5H9O1-3 + OH <=> C5H91-3 + HO2', [6.088000e+15, -1.07, 15720.0])

## Reaction 2573
#reaction('C5H9O1-3 + CH3O <=> C5H91-3 + CH3O2', [5.946000e+17, -1.65, 20480.0])

## Reaction 2574
#reaction('C5H9O1-3 + C2H5O <=> C5H91-3 + C2H5O2', [3.893000e+14, -0.72, 18330.0])

## Reaction 2575
#reaction('C5H9O2-4 + OH <=> C5H92-4 + HO2', [7.027000e+15, -1.24, 15890.0])

## Reaction 2576
#reaction('C5H9O2-4 + CH3O <=> C5H92-4 + CH3O2', [6.863000e+17, -1.82, 20650.0])

## Reaction 2577
#reaction('C5H9O2-4 + C2H5O <=> C5H92-4 + C2H5O2', [4.493000e+14, -0.89, 18490.0])

## Reaction 2578
#reaction('C4H6 + CH3 <=> C5H91-3', [1.000000e+11, 0.0, 7800.0])

## Reaction 2579
#reaction('C5H81-3 + H <=> C5H91-3', [2.500000e+11, 0.51, 2620.0])

## Reaction 2580
#reaction('C3H6 + C2H3 <=> C5H91-4', [1.000000e+11, 0.0, 7800.0])

## Reaction 2581
#reaction('C2H4 + C3H5-A <=> C5H91-5', [1.000000e+11, 0.0, 7800.0])

## Reaction 2582
#reaction('C5H81-3 + H <=> C5H92-4', [4.240000e+11, 0.51, 1230.0])

## Reaction 2583
#reaction('C2H4 + C3H5-S <=> C5H92-5', [2.000000e+11, 0.0, 7800.0])

## Reaction 2584
#reaction('C5H9O1-3 <=> C2H3CHO + C2H5', [3.131000e+19, -1.85, 10670.0])

## Reaction 2585
#reaction('C5H9O1-3 <=> C2H5CHO + C2H3', [1.417000e+18, -1.56, 23340.0])

## Reaction 2586
#reaction('C5H81-3 + OH <=> CH2O + C4H71-3', [1.000000e+12, 0.0, 0.0])

## Reaction 2587
#reaction('C5H81-3 + OH <=> C2H3CHO + C2H5', [1.000000e+12, 0.0, 0.0])

## Reaction 2588
#reaction('C5H81-3 + OH <=> CH3CHO + C3H5-S', [1.000000e+12, 0.0, 0.0])

## Reaction 2589
#reaction('C5H9O2-4 <=> SC3H5CHO + CH3', [5.983000e+15, -1.13, 9941.0])

## Reaction 2590
#reaction('C5H9O2-4 <=> CH3CHO + C3H5-S', [1.073000e+22, -2.66, 29650.0])

## Reaction 2591
#falloff_reaction('C5H10-2 (+ M) <=> CH3 + C4H71-3 (+ M)',
                #kf=[6.486000e+19, -1.367, 76320.0],
                #kf0=[1.527000e+104, -24.826, 94800.0],
                #falloff=Troe(A=0.005301, T3=143.7, T1=16770000000000.0, T2=3671.0))

## Reaction 2592
#reaction('C5H11-1 + O2 <=> C5H10-1 + HO2', [8.370000e-01, 3.59, 11960.0])

## Reaction 2593
#reaction('C5H11-2 + O2 <=> C5H10-1 + HO2', [5.350000e-01, 3.71, 9322.0])

## Reaction 2594
#reaction('C5H11-2 + O2 <=> C5H10-2 + HO2', [1.070000e+00, 3.71, 9322.0])

## Reaction 2595
#reaction('C5H11-3 + O2 <=> C5H10-2 + HO2', [2.140000e+00, 3.71, 9322.0])

## Reaction 2596
#reaction('C5H11-1 + HO2 <=> C5H11O-1 + OH', [9.000000e+12, 0.0, -1000.0])

## Reaction 2597
#reaction('C5H11-2 + HO2 <=> C5H11O-2 + OH', [9.000000e+12, 0.0, -1000.0])

## Reaction 2598
#reaction('C5H11-3 + HO2 <=> C5H11O-3 + OH', [9.000000e+12, 0.0, -1000.0])

## Reaction 2599
#reaction('C5H11-1 + CH3O2 <=> C5H11O-1 + CH3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 2600
#reaction('C5H11-2 + CH3O2 <=> C5H11O-2 + CH3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 2601
#reaction('C5H11-3 + CH3O2 <=> C5H11O-3 + CH3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 2602
#reaction('C5H11O2-1 + NC5H12 <=> C5H11O2H-1 + C5H11-1', [1.210000e+13, 0.0, 20430.0])

## Reaction 2603
#reaction('C5H11O2-2 + NC5H12 <=> C5H11O2H-2 + C5H11-1', [1.210000e+13, 0.0, 20430.0])

## Reaction 2604
#reaction('C5H11O2-3 + NC5H12 <=> C5H11O2H-3 + C5H11-1', [1.210000e+13, 0.0, 20430.0])

## Reaction 2605
#reaction('C5H11O2-1 + NC5H12 <=> C5H11O2H-1 + C5H11-2', [8.064000e+12, 0.0, 17700.0])

## Reaction 2606
#reaction('C5H11O2-2 + NC5H12 <=> C5H11O2H-2 + C5H11-2', [8.064000e+12, 0.0, 17700.0])

## Reaction 2607
#reaction('C5H11O2-3 + NC5H12 <=> C5H11O2H-3 + C5H11-2', [8.064000e+12, 0.0, 17700.0])

## Reaction 2608
#reaction('C5H11O2-1 + NC5H12 <=> C5H11O2H-1 + C5H11-3', [4.032000e+12, 0.0, 17700.0])

## Reaction 2609
#reaction('C5H11O2-2 + NC5H12 <=> C5H11O2H-2 + C5H11-3', [4.032000e+12, 0.0, 17700.0])

## Reaction 2610
#reaction('C5H11O2-3 + NC5H12 <=> C5H11O2H-3 + C5H11-3', [4.032000e+12, 0.0, 17700.0])

## Reaction 2611
#reaction('C5H11-1 + C5H11O2-1 <=> C5H11O-1 + C5H11O-1', [9.000000e+12, 0.0, -1000.0])

## Reaction 2612
#reaction('C5H11-1 + C5H11O2-2 <=> C5H11O-1 + C5H11O-2', [9.000000e+12, 0.0, -1000.0])

## Reaction 2613
#reaction('C5H11-1 + C5H11O2-3 <=> C5H11O-1 + C5H11O-3', [9.000000e+12, 0.0, -1000.0])

## Reaction 2614
#reaction('C5H11-2 + C5H11O2-1 <=> C5H11O-2 + C5H11O-1', [9.000000e+12, 0.0, -1000.0])

## Reaction 2615
#reaction('C5H11-2 + C5H11O2-2 <=> C5H11O-2 + C5H11O-2', [9.000000e+12, 0.0, -1000.0])

## Reaction 2616
#reaction('C5H11-2 + C5H11O2-3 <=> C5H11O-2 + C5H11O-3', [9.000000e+12, 0.0, -1000.0])

## Reaction 2617
#reaction('C5H11-3 + C5H11O2-1 <=> C5H11O-3 + C5H11O-1', [9.000000e+12, 0.0, -1000.0])

## Reaction 2618
#reaction('C5H11-3 + C5H11O2-2 <=> C5H11O-3 + C5H11O-2', [9.000000e+12, 0.0, -1000.0])

## Reaction 2619
#reaction('C5H11-3 + C5H11O2-3 <=> C5H11O-3 + C5H11O-3', [9.000000e+12, 0.0, -1000.0])

## Reaction 2620
#reaction('C5H11O2-1 + C5H11O2-2 => O2 + C5H11O-1 + C5H11O-2', [1.400000e+16, -1.61, 1860.0])

## Reaction 2621
#reaction('C5H11O2-1 + C5H11O2-3 => O2 + C5H11O-1 + C5H11O-3', [1.400000e+16, -1.61, 1860.0])

## Reaction 2622
#reaction('C5H11O2-1 + CH3O2 => O2 + C5H11O-1 + CH3O', [1.400000e+16, -1.61, 1860.0])

## Reaction 2623
#reaction('C5H11O2-1 + C5H11O2-1 => O2 + C5H11O-1 + C5H11O-1', [1.400000e+16, -1.61, 1860.0])

## Reaction 2624
#reaction('C5H11O2H-1 + HO2 <=> C5H11O2-1 + H2O2', [2.400000e+12, 0.0, 10000.0])

## Reaction 2625
#reaction('C5H11O2H-1 + O2 <=> C5H11O2-1 + HO2', [3.746000e+13, -0.791, 33620.0])

## Reaction 2626
#reaction('C5H11O2-2 + CH3O2 => O2 + C5H11O-2 + CH3O', [1.400000e+16, -1.61, 1860.0])

## Reaction 2627
#reaction('C5H11O2-2 + C5H11O2-3 => C5H11O-2 + C5H11O-3 + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 2628
#reaction('C5H11O2-2 + C5H11O2-2 => O2 + C5H11O-2 + C5H11O-2', [1.400000e+16, -1.61, 1860.0])

## Reaction 2629
#reaction('C5H11O2H-2 + HO2 <=> C5H11O2-2 + H2O2', [2.400000e+12, 0.0, 10000.0])

## Reaction 2630
#reaction('C5H11O2H-2 + O2 <=> C5H11O2-2 + HO2', [4.380000e+13, -0.812, 33640.0])

## Reaction 2631
#reaction('C5H11O2-3 + CH3O2 => O2 + C5H11O-3 + CH3O', [1.400000e+16, -1.61, 1860.0])

## Reaction 2632
#reaction('C5H11O2-3 + C5H11O2-3 => O2 + C5H11O-3 + C5H11O-3', [1.400000e+16, -1.61, 1860.0])

## Reaction 2633
#reaction('C5H11O2H-3 + HO2 <=> C5H11O2-3 + H2O2', [2.400000e+12, 0.0, 10000.0])

## Reaction 2634
#reaction('C5H11O2H-3 + O2 <=> C5H11O2-3 + HO2', [4.380000e+13, -0.812, 33640.0])

## Reaction 2635
#reaction('C5H11O2H-1 <=> C5H11O-1 + OH', [7.303000e+18, -0.89, 45630.0])

## Reaction 2636
#reaction('C5H11O2H-2 <=> C5H11O-2 + OH', [4.232000e+20, -1.4, 45750.0])

## Reaction 2637
#reaction('C5H11O2H-3 <=> C5H11O-3 + OH', [1.422000e+20, -1.34, 45690.0])

## Reaction 2638
#reaction('CH2O + PC4H9 <=> C5H11O-1', [5.000000e+10, 0.0, 3457.0])

## Reaction 2639
#reaction('CH3CHO + NC3H7 <=> C5H11O-2', [3.330000e+10, 0.0, 6397.0])

## Reaction 2640
#reaction('C2H5 + C2H5CHO <=> C5H11O-3', [1.670000e+10, 0.0, 6193.0])

## Reaction 2641
#reaction('C5H11-1 + O2 <=> C5H11O2-1', [6.865000e+16, -1.627, 198.7])

## Reaction 2642
#reaction('C5H11-2 + O2 <=> C5H11O2-2', [3.487000e+14, -0.816, -536.5])

## Reaction 2643
#reaction('C5H11-3 + O2 <=> C5H11O2-3', [3.487000e+14, -0.816, -536.5])

## Reaction 2644
#reaction('C5H11O2-1 <=> C5H10OOH1-2', [4.009000e+08, 1.1, 30100.0])

## Reaction 2645
#reaction('C5H11O2-1 <=> C5H10OOH1-3', [1.360000e+07, 1.3, 18200.0])

## Reaction 2646
#reaction('C5H11O2-1 <=> C5H10OOH1-4', [3.718000e+06, 1.2, 16600.0])

## Reaction 2647
#reaction('C5H11O2-1 <=> C5H10OOH1-5', [2.941000e+05, 1.5, 19900.0])

## Reaction 2648
#reaction('C5H11O2-2 <=> C5H10OOH2-1', [7.290000e+08, 1.1, 33500.0])

## Reaction 2649
#reaction('C5H11O2-2 <=> C5H10OOH2-3', [1.716000e+09, 0.9, 29500.0])

## Reaction 2650
#reaction('C5H11O2-2 <=> C5H10OOH2-4', [7.033000e+10, 0.2, 18500.0])

## Reaction 2651
#reaction('C5H11O2-2 <=> C5H10OOH2-5', [6.170000e+05, 1.5, 20000.0])

## Reaction 2652
#reaction('C5H11O2-3 <=> C5H10OOH3-1', [2.877000e+07, 1.4, 20800.0])

## Reaction 2653
#reaction('C5H11O2-3 <=> C5H10OOH3-2', [3.432000e+09, 0.9, 29500.0])

## Reaction 2654
#reaction('C5H11O2-1 <=> C5H10-1 + HO2', [3.870000e+07, 1.58, 28500.0])

## Reaction 2655
#reaction('C5H11O2-2 <=> C5H10-1 + HO2', [3.605000e+09, 1.13, 30100.0])

## Reaction 2656
#reaction('C5H11O2-2 <=> C5H10-2 + HO2', [7.710000e+09, 1.01, 29362.0])

## Reaction 2657
#reaction('C5H11O2-3 <=> C5H10-2 + HO2', [7.200000e+10, 0.73, 29883.0])

## Reaction 2658
#reaction('C5H10OOH1-2 <=> C5H10O1-2 + OH', [3.670000e+09, 0.93, 10900.0])

## Reaction 2659
#reaction('C5H10OOH1-3 <=> C5H10O1-3 + OH', [4.225000e+08, 0.9, 15300.0])

## Reaction 2660
#reaction('C5H10OOH1-4 <=> C5H10O1-4 + OH', [2.410000e+07, 0.94, 10000.0])

## Reaction 2661
#reaction('C5H10OOH1-5 <=> C5H10O1-5 + OH', [1.445000e+07, 0.76, 11300.0])

## Reaction 2662
#reaction('C5H10OOH2-1 <=> C5H10O1-2 + OH', [5.800000e+10, 0.43, 10700.0])

## Reaction 2663
#reaction('C5H10OOH2-3 <=> C5H10O2-3 + OH', [2.218000e+09, 0.916, 9389.0])

## Reaction 2664
#reaction('C5H10OOH2-4 <=> C5H10O2-4 + OH', [1.207000e+09, 0.717, 16402.0])

## Reaction 2665
#reaction('C5H10OOH2-5 <=> C5H10O1-4 + OH', [1.240000e+09, 0.35, 11200.0])

## Reaction 2666
#reaction('C5H10OOH3-1 <=> C5H10O1-3 + OH', [1.015000e+10, 0.46, 17700.0])

## Reaction 2667
#reaction('C5H10OOH3-2 <=> C5H10O2-3 + OH', [6.680000e+09, 0.791, 9583.0])

## Reaction 2668
#reaction('C5H10O1-2 + OH => CH2CO + NC3H7 + H2O', [2.260000e+03, 2.73, -4688.6])

## Reaction 2669
#reaction('C5H10O1-3 + OH => C2H4 + C2H5CO + H2O', [2.260000e+03, 2.73, -4688.6])

## Reaction 2670
#reaction('C5H10O1-4 + OH => CH3COCH2 + C2H4 + H2O', [2.260000e+03, 2.73, -4688.6])

## Reaction 2671
#reaction('C5H10O1-5 + OH => CH2CH2CHO + C2H4 + H2O', [2.260000e+03, 2.73, -4688.6])

## Reaction 2672
#reaction('C5H10O2-3 + OH => CH3CHCO + C2H5 + H2O', [2.260000e+03, 2.73, -4688.6])

## Reaction 2673
#reaction('C5H10O2-4 + OH => CH3CO + C3H6 + H2O', [2.260000e+03, 2.73, -4688.6])

## Reaction 2674
#reaction('C5H10O1-2 + OH => C2H3CHO + C2H5 + H2O', [2.260000e+03, 2.73, -4688.6])

## Reaction 2675
#reaction('C5H10O1-3 + OH => HCO + C4H8-1 + H2O', [2.260000e+03, 2.73, -4688.6])

## Reaction 2676
#reaction('C5H10O1-4 + OH => CH2CHO + C3H6 + H2O', [2.260000e+03, 2.73, -4688.6])

## Reaction 2677
#reaction('C5H10O1-5 + OH => CH2O + C4H71-3 + H2O', [2.260000e+03, 2.73, -4688.6])

## Reaction 2678
#reaction('C5H10O2-3 + OH => C2H3COCH3 + CH3 + H2O', [2.260000e+03, 2.73, -4688.6])

## Reaction 2679
#reaction('C5H10O2-4 + OH => CH3CHO + C3H5-S + H2O', [2.260000e+03, 2.73, -4688.6])

## Reaction 2680
#reaction('C5H10O1-2 + HO2 => CH2CO + NC3H7 + H2O2', [8.600000e+00, 3.46, 9733.3])

## Reaction 2681
#reaction('C5H10O1-3 + HO2 => C2H4 + C2H5CO + H2O2', [8.600000e+00, 3.46, 9733.3])

## Reaction 2682
#reaction('C5H10O1-4 + HO2 => CH3COCH2 + C2H4 + H2O2', [8.600000e+00, 3.46, 9733.3])

## Reaction 2683
#reaction('C5H10O1-5 + HO2 => CH2CH2CHO + C2H4 + H2O2', [8.600000e+00, 3.46, 9733.3])

## Reaction 2684
#reaction('C5H10O2-3 + HO2 => CH3CHCO + C2H5 + H2O2', [8.600000e+00, 3.46, 9733.3])

## Reaction 2685
#reaction('C5H10O2-4 + HO2 => CH3CO + C3H6 + H2O2', [8.600000e+00, 3.46, 9733.3])

## Reaction 2686
#reaction('C5H10O1-2 + HO2 => C2H3CHO + C2H5 + H2O2', [8.600000e+00, 3.46, 9733.3])

## Reaction 2687
#reaction('C5H10O1-3 + HO2 => HCO + C4H8-1 + H2O2', [8.600000e+00, 3.46, 9733.3])

## Reaction 2688
#reaction('C5H10O1-4 + HO2 => CH2CHO + C3H6 + H2O2', [8.600000e+00, 3.46, 9733.3])

## Reaction 2689
#reaction('C5H10O1-5 + HO2 => CH2O + C4H71-3 + H2O2', [8.600000e+00, 3.46, 9733.3])

## Reaction 2690
#reaction('C5H10O2-3 + HO2 => C2H3COCH3 + CH3 + H2O2', [8.600000e+00, 3.46, 9733.3])

## Reaction 2691
#reaction('C5H10O2-4 + HO2 => CH3CHO + C3H5-S + H2O2', [8.600000e+00, 3.46, 9733.3])

## Reaction 2692
#reaction('C5H10OOH1-2 <=> C5H10-1 + HO2', [1.040000e+10, 0.94, 15100.0])

## Reaction 2693
#reaction('C5H10OOH2-1 <=> C5H10-1 + HO2', [2.210000e+12, 0.23, 15600.0])

## Reaction 2694
#reaction('C5H10OOH2-3 <=> C5H10-2 + HO2', [4.543000e+10, 0.751, 15100.0])

## Reaction 2695
#reaction('C5H10OOH3-2 <=> C5H10-2 + HO2', [4.066000e+10, 0.678, 15034.0])

## Reaction 2696
#reaction('C5H10OOH1-3 => OH + CH2O + C4H8-1', [9.100000e+08, 1.4, 25500.0])

## Reaction 2697
#reaction('C5H10OOH2-4 => OH + CH3CHO + C3H6', [1.180000e+09, 1.2, 22700.0])

## Reaction 2698
#reaction('C5H10OOH3-1 => OH + C2H5CHO + C2H4', [3.096000e+18, -1.73, 26820.0])

## Reaction 2699
#reaction('C5H10OOH1-2 <=> C2H5 + AC3H5OOH', [2.410000e+10, 1.02, 28200.0])

## Reaction 2700
#reaction('C5H10OOH2-5 <=> C2H4 + C3H6OOH2-1', [6.860000e+13, 0.1, 29000.0])

## Reaction 2701
#reaction('C5H10OOH1-4 <=> C3H6 + C2H4O2H', [5.080000e+11, 0.7, 28700.0])

## Reaction 2702
#reaction('C5H10OOH1-2 + O2 <=> C5H10OOH1-2O2', [1.920000e+14, -0.816, -536.5])

## Reaction 2703
#reaction('C5H10OOH1-3 + O2 <=> C5H10OOH1-3O2', [1.920000e+14, -0.816, -536.5])

## Reaction 2704
#reaction('C5H10OOH1-4 + O2 <=> C5H10OOH1-4O2', [1.920000e+14, -0.816, -536.5])

## Reaction 2705
#reaction('C5H10OOH1-5 + O2 <=> C5H10OOH1-5O2', [3.780000e+16, -1.627, 198.7])

## Reaction 2706
#reaction('C5H10OOH2-1 + O2 <=> C5H10OOH2-1O2', [3.780000e+16, -1.627, 198.7])

## Reaction 2707
#reaction('C5H10OOH2-3 + O2 <=> C5H10OOH2-3O2', [1.920000e+14, -0.816, -536.5])

## Reaction 2708
#reaction('C5H10OOH2-4 + O2 <=> C5H10OOH2-4O2', [1.920000e+14, -0.816, -536.5])

## Reaction 2709
#reaction('C5H10OOH2-5 + O2 <=> C5H10OOH2-5O2', [3.780000e+16, -1.627, 198.7])

## Reaction 2710
#reaction('C5H10OOH3-1 + O2 <=> C5H10OOH3-1O2', [3.780000e+16, -1.627, 198.7])

## Reaction 2711
#reaction('C5H10OOH3-2 + O2 <=> C5H10OOH3-2O2', [1.920000e+14, -0.816, -536.5])

## Reaction 2712
#reaction('C5H10OOH1-2O2 <=> NC5KET12 + OH', [2.441000e+07, 1.6, 27900.0])

## Reaction 2713
#reaction('C5H10OOH1-2O2 <=> C5H93-1,2OOH', [1.716000e+09, 0.9, 29500.0])

## Reaction 2714
#reaction('C5H10OOH1-2O2 <=> C5H92-4,5OOH', [7.033000e+10, 0.2, 18500.0])

## Reaction 2715
#reaction('C5H10OOH1-2O2 <=> C5H91-4,5OOH', [1.233000e+06, 1.5, 20000.0])

## Reaction 2716
#reaction('C5H10OOH1-3O2 <=> NC5KET13 + OH', [1.098000e+04, 2.4, 19900.0])

## Reaction 2717
#reaction('C5H10OOH1-3O2 <=> C5H92-1,3OOH', [3.432000e+09, 0.9, 29500.0])

## Reaction 2718
#reaction('C5H10OOH1-3O2 <=> C5H92-3,5OOH', [3.432000e+09, 0.9, 29500.0])

## Reaction 2719
#reaction('C5H10OOH1-3O2 <=> C5H91-3,5OOH', [2.877000e+07, 1.4, 20800.0])

## Reaction 2720
#reaction('C5H10OOH1-4O2 <=> NC5KET14 + OH', [1.256000e+03, 2.2, 17400.0])

## Reaction 2721
#reaction('C5H10OOH1-4O2 <=> C5H92-1,4OOH', [7.033000e+10, 0.2, 18500.0])

## Reaction 2722
#reaction('C5H10OOH1-4O2 <=> C5H93-1,4OOH', [1.716000e+09, 0.9, 29500.0])

## Reaction 2723
#reaction('C5H10OOH1-4O2 <=> C5H91-2,5OOH', [7.290000e+08, 1.1, 33500.0])

## Reaction 2724
#reaction('C5H10OOH1-5O2 <=> NC5KET15 + OH', [1.115000e+03, 1.8, 16600.0])

## Reaction 2725
#reaction('C5H10OOH1-5O2 <=> C5H92-1,5OOH', [3.718000e+06, 1.2, 16600.0],
        #options='duplicate')

## Reaction 2726
#reaction('C5H10OOH1-5O2 <=> C5H93-1,5OOH', [1.360000e+07, 1.3, 18200.0])

## Reaction 2727
#reaction('C5H10OOH1-5O2 <=> C5H92-1,5OOH', [4.009000e+08, 1.1, 30100.0],
        #options='duplicate')

## Reaction 2728
#reaction('C5H10OOH2-1O2 <=> NC5KET21 + OH', [2.763000e+08, 1.2, 25700.0])

## Reaction 2729
#reaction('C5H10OOH2-1O2 <=> C5H93-1,2OOH', [1.360000e+07, 1.3, 18200.0])

## Reaction 2730
#reaction('C5H10OOH2-1O2 <=> C5H92-4,5OOH', [3.718000e+06, 1.2, 16600.0])

## Reaction 2731
#reaction('C5H10OOH2-1O2 <=> C5H91-4,5OOH', [2.941000e+05, 1.5, 19900.0])

## Reaction 2732
#reaction('C5H10OOH2-3O2 <=> NC5KET23 + OH', [1.748000e+08, 1.7, 26000.0])

## Reaction 2733
#reaction('C5H10OOH2-3O2 <=> C5H91-2,3OOH', [2.877000e+07, 1.4, 20800.0])

## Reaction 2734
#reaction('C5H10OOH2-3O2 <=> C5H92-3,4OOH', [3.432000e+09, 0.9, 29500.0])

## Reaction 2735
#reaction('C5H10OOH2-3O2 <=> C5H91-3,4OOH', [2.877000e+07, 1.4, 20800.0])

## Reaction 2736
#reaction('C5H10OOH2-4O2 <=> NC5KET24 + OH', [1.754000e+02, 3.1, 17500.0])

## Reaction 2737
#reaction('C5H10OOH2-4O2 <=> C5H91-2,4OOH', [1.233000e+06, 1.5, 20000.0],
        #options='duplicate')

## Reaction 2738
#reaction('C5H10OOH2-4O2 <=> C5H93-2,4OOH', [1.716000e+09, 0.9, 29500.0])

## Reaction 2739
#reaction('C5H10OOH2-4O2 <=> C5H91-2,4OOH', [7.290000e+08, 1.1, 33500.0],
        #options='duplicate')

## Reaction 2740
#reaction('C5H10OOH2-5O2 <=> NC5KET25 + OH', [3.768000e+02, 2.2, 15300.0])

## Reaction 2741
#reaction('C5H10OOH2-5O2 <=> C5H91-2,5OOH', [2.941000e+05, 1.5, 19900.0])

## Reaction 2742
#reaction('C5H10OOH2-5O2 <=> C5H93-1,4OOH', [1.360000e+07, 1.3, 18200.0])

## Reaction 2743
#reaction('C5H10OOH2-5O2 <=> C5H92-1,4OOH', [4.009000e+08, 1.1, 30100.0])

## Reaction 2744
#reaction('C5H10OOH3-1O2 <=> NC5KET31 + OH', [5.786000e+01, 2.9, 17000.0])

## Reaction 2745
#reaction('C5H10OOH3-1O2 <=> C5H92-1,3OOH', [4.009000e+08, 1.1, 30100.0])

## Reaction 2746
#reaction('C5H10OOH3-1O2 <=> C5H92-3,5OOH', [3.718000e+06, 1.2, 16600.0])

## Reaction 2747
#reaction('C5H10OOH3-1O2 <=> C5H91-3,5OOH', [2.941000e+05, 1.5, 19900.0])

## Reaction 2748
#reaction('C5H10OOH3-2O2 <=> NC5KET32 + OH', [1.748000e+08, 1.7, 26000.0])

## Reaction 2749
#reaction('C5H10OOH3-2O2 <=> C5H91-2,3OOH', [7.290000e+08, 1.1, 33500.0])

## Reaction 2750
#reaction('C5H10OOH3-2O2 <=> C5H92-3,4OOH', [7.033000e+10, 0.2, 18500.0])

## Reaction 2751
#reaction('C5H10OOH3-2O2 <=> C5H91-3,4OOH', [1.233000e+06, 1.5, 20000.0])

## Reaction 2752
#reaction('C5H91-2,3OOH <=> C5H91-3OOH + HO2', [2.210000e+12, 0.23, 15600.0])

## Reaction 2753
#reaction('C5H91-3OOH => C2H5CHO + C2H3 + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 2754
#reaction('C5H91-3OOH => C2H3CHO + C2H5 + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 2755
#reaction('C5H91-2,4OOH <=> C5H91-4OOH + HO2', [2.210000e+12, 0.23, 15600.0])

## Reaction 2756
#reaction('C5H91-4OOH => CH3CHO + C3H5-A + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 2757
#reaction('C5H91-4OOH => SC3H5CHO + CH3 + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 2758
#reaction('C5H91-2,5OOH <=> C5H91-5OOH + HO2', [2.210000e+12, 0.23, 15600.0])

## Reaction 2759
#reaction('C5H91-5OOH => CH2O + C4H71-4 + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 2760
#reaction('C5H91-3,4OOH => C3KET12 + C2H4 + OH', [3.096000e+18, -1.73, 26820.0])

## Reaction 2761
#reaction('C5H91-3,5OOH => C3KET13 + C2H4 + OH', [3.096000e+18, -1.73, 26820.0])

## Reaction 2762
#reaction('C5H91-4,5OOH => AC3H5OOH + C2H4 + HO2', [6.860000e+13, 0.1, 29000.0])

## Reaction 2763
#reaction('C5H92-1,3OOH <=> C5H91-3OOH + HO2', [1.040000e+10, 0.94, 15100.0])

## Reaction 2764
#reaction('C5H92-1,3OOH <=> C5H92-1OOH + HO2', [4.066000e+10, 0.678, 15034.0])

## Reaction 2765
#reaction('C5H92-1OOH => CH2O + C4H71-1 + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 2766
#reaction('C5H92-1,4OOH <=> C5H91-4OOH + HO2', [1.040000e+10, 0.94, 15100.0])

## Reaction 2767
#reaction('C5H92-1,4OOH => AC3H5OOH + CH3CHO + OH', [2.410000e+10, 1.02, 28200.0])

## Reaction 2768
#reaction('C5H92-1,5OOH <=> C5H91-5OOH + HO2', [1.040000e+10, 0.94, 15100.0])

## Reaction 2769
#reaction('C5H92-1,5OOH <=> AC3H5OOH + C2H4O2H', [2.410000e+10, 1.02, 28200.0])

## Reaction 2770
#reaction('C5H92-3,4OOH <=> C5H92-4OOH + HO2', [4.066000e+10, 0.678, 15034.0])

## Reaction 2771
#reaction('C5H92-4OOH => CH3CHO + C3H5-S + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 2772
#reaction('C5H92-4OOH => SC3H5CHO + CH3 + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 2773
#reaction('C5H92-3,5OOH <=> C5H92-5OOH + HO2', [4.066000e+10, 0.678, 15034.0])

## Reaction 2774
#reaction('C5H92-5OOH => CH2O + C4H71-3 + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 2775
#reaction('C5H92-4,5OOH => HO2CH2CHO + C3H6 + OH', [1.180000e+09, 1.2, 22700.0])

## Reaction 2776
#reaction('C5H93-1,2OOH <=> C5H92-1OOH + HO2', [4.543000e+10, 0.751, 15100.0])

## Reaction 2777
#reaction('C5H93-1,4OOH <=> C5H92-5OOH + HO2', [4.543000e+10, 0.751, 15100.0])

## Reaction 2778
#reaction('C5H93-1,4OOH => CH2O + C4H71-3OOH + OH', [9.100000e+08, 1.4, 25500.0])

## Reaction 2779
#reaction('C5H93-1,5OOH => CH2O + C4H71-4OOH + OH', [1.820000e+09, 1.4, 25500.0])

## Reaction 2780
#reaction('C5H93-2,4OOH <=> C5H92-4OOH + HO2', [9.086000e+10, 0.751, 15100.0])

## Reaction 2781
#reaction('C5H91-2,3OOH <=> C5H9O1-2OOH-3 + OH', [5.800000e+10, 0.43, 10700.0])

## Reaction 2782
#reaction('C5H9O1-2OOH-3 => C2H5CHO + CH2CHO + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 2783
#reaction('C5H91-2,3OOH <=> C5H9O1-3OOH-2 + OH', [1.015000e+10, 0.46, 17700.0])

## Reaction 2784
#reaction('C5H9O1-3OOH-2 => C2H5CHO + CH2CHO + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 2785
#reaction('C5H9O1-3OOH-2 => CH2O + C3H6CHO-1 + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 2786
#reaction('C5H91-2,4OOH <=> C5H9O1-2OOH-4 + OH', [5.800000e+10, 0.43, 10700.0])

## Reaction 2787
#reaction('C5H9O1-2OOH-4 => CH3CHO + C3H5O + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 2788
#reaction('C5H91-2,4OOH <=> C5H9O1-4OOH-2 + OH', [1.240000e+09, 0.35, 11200.0])

## Reaction 2789
#reaction('C5H9O1-4OOH-2 => CH2O + C3H6CHO-2 + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 2790
#reaction('C5H9O1-4OOH-2 => C3H6 + OCH2CHO + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 2791
#reaction('OCH2CHO <=> CH2O + HCO', [2.900000e+12, 0.38, 11080.0])

## Reaction 2792
#reaction('C5H91-2,5OOH <=> C5H9O1-2OOH-5 + OH', [5.800000e+10, 0.43, 10700.0])

## Reaction 2793
#reaction('C5H9O1-2OOH-5 <=> C5H9O1-2O-5 + OH', [4.151000e+19, -1.31, 45700.0])

## Reaction 2794
#reaction('C5H9O1-2O-5 => CH2O + CH2CHO + C2H4', [1.326000e+21, -2.349, 25084.0])

## Reaction 2795
#reaction('C5H91-2,5OOH <=> C5H9O1-5OOH-2 + OH', [1.445000e+07, 0.76, 11300.0])

## Reaction 2796
#reaction('C5H9O1-5OOH-2 => CH2O + C3H6CHO-3 + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 2797
#reaction('C5H9O1-5OOH-2 => C2H4 + CH2OCH2CHO + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 2798
#reaction('CH2OCH2CHO <=> CH2O + CH2CHO', [1.600000e+13, 0.0, 25500.0])

## Reaction 2799
#reaction('C5H91-3,4OOH <=> C5H9O1-3OOH-4 + OH', [1.015000e+10, 0.46, 17700.0])

## Reaction 2800
#reaction('C5H9O1-3OOH-4 => CH3CHO + CH2CH2CHO + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 2801
#reaction('C5H91-3,4OOH <=> C5H9O1-4OOH-3 + OH', [1.240000e+09, 0.35, 11200.0])

## Reaction 2802
#reaction('C5H9O1-4OOH-3 => C2H4 + C2H4OCHO + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 2803
#reaction('C2H4OCHO <=> CH3CHO + HCO', [2.900000e+12, 0.38, 11080.0])

## Reaction 2804
#reaction('C5H91-3,5OOH <=> C5H9O1-3OOH-5 + OH', [1.015000e+10, 0.46, 17700.0])

## Reaction 2805
#reaction('C5H9O1-3OOH-5 => CH2O + C4H7O1-4 + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 2806
#reaction('C4H7O1-4 <=> CH2O + C3H5-A', [1.326000e+21, -2.349, 25084.0])

## Reaction 2807
#reaction('C5H91-3,5OOH <=> C5H9O1-5OOH-3 + OH', [1.445000e+07, 0.76, 11300.0])

## Reaction 2808
#reaction('C5H9O1-5OOH-3 <=> CH2CH2OCH2CH2CHO + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 2809
#reaction('CH2CH2OCH2CH2CHO => CH2O + C2H4 + CH2CHO', [1.600000e+13, 0.0, 25500.0])

## Reaction 2810
#reaction('C5H91-4,5OOH <=> C5H9O1-4OOH-5 + OH', [1.240000e+09, 0.35, 11200.0])

## Reaction 2811
#reaction('C5H9O1-4OOH-5 => CH2O + C3H6CHO-3 + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 2812
#reaction('C5H91-4,5OOH <=> C5H9O1-5OOH-2 + OH', [1.445000e+07, 0.76, 11300.0])

## Reaction 2813
#reaction('C5H92-1,3OOH <=> C5H9O1-2OOH-3 + OH', [3.670000e+09, 0.93, 10900.0])

## Reaction 2814
#reaction('C5H92-1,3OOH <=> C5H9O2-3OOH-1 + OH', [7.040000e+09, 0.784, 9591.0])

## Reaction 2815
#reaction('C5H9O2-3OOH-1 => CH2O + C3H6CHO-1 + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 2816
#reaction('C5H92-1,4OOH <=> C5H9O1-2OOH-4 + OH', [3.670000e+09, 0.93, 10900.0])

## Reaction 2817
#reaction('C5H92-1,4OOH <=> C5H9O2-4OOH-1 + OH', [1.340000e+09, 0.7, 16415.0])

## Reaction 2818
#reaction('C5H9O2-4OOH-1 => CH2O + C3H6CHO-2 + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 2819
#reaction('C5H92-1,5OOH <=> C5H9O1-2OOH-5 + OH', [3.670000e+09, 0.93, 10900.0])

## Reaction 2820
#reaction('C5H92-1,5OOH <=> C5H9O1-4OOH-5 + OH', [2.410000e+07, 0.94, 10000.0])

## Reaction 2821
#reaction('C5H92-3,4OOH <=> C5H9O2-3OOH-4 + OH', [7.040000e+09, 0.784, 9591.0])

## Reaction 2822
#reaction('C5H92-3,4OOH <=> C5H9O2-4OOH-3 + OH', [1.340000e+09, 0.7, 16415.0])

## Reaction 2823
#reaction('C5H9O2-4OOH-3 => CH3CHO + CH3CHCHO + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 2824
#reaction('C5H92-3,5OOH <=> C5H9O2-3OOH-5 + OH', [7.040000e+09, 0.784, 9591.0])

## Reaction 2825
#reaction('C5H9O2-3OOH-5 => CH2O + C4H71-O + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 2826
#reaction('C5H92-3,5OOH <=> C5H9O1-4OOH-3 + OH', [2.410000e+07, 0.94, 10000.0])

## Reaction 2827
#reaction('C5H92-4,5OOH <=> C5H9O2-4OOH-1 + OH', [1.340000e+09, 0.7, 16415.0])

## Reaction 2828
#reaction('C5H92-4,5OOH <=> C5H9O1-4OOH-2 + OH', [2.410000e+07, 0.94, 10000.0])

## Reaction 2829
#reaction('C5H93-1,2OOH <=> C5H9O1-3OOH-2 + OH', [4.225000e+08, 0.9, 15300.0])

## Reaction 2830
#reaction('C5H93-1,2OOH <=> C5H9O2-3OOH-1 + OH', [2.441000e+09, 0.9, 9403.0])

## Reaction 2831
#reaction('C5H93-1,4OOH <=> C5H9O1-3OOH-4 + OH', [4.225000e+08, 0.9, 15300.0])

## Reaction 2832
#reaction('C5H93-1,4OOH <=> C5H9O2-3OOH-5 + OH', [2.441000e+09, 0.9, 9403.0])

## Reaction 2833
#reaction('C5H93-1,5OOH <=> C5H9O1-3OOH-5 + OH', [4.225000e+08, 0.9, 15300.0])

## Reaction 2834
#reaction('C5H93-2,4OOH <=> C5H9O2-3OOH-4 + OH', [4.881000e+09, 0.9, 9403.0])

## Reaction 2835
#reaction('C5H10OOH1-2O2 <=> C5H92-1OOH + HO2', [5.140000e+09, 1.01, 29362.0])

## Reaction 2836
#reaction('C5H10OOH1-3O2 <=> C5H92-1OOH + HO2', [3.600000e+10, 0.73, 29883.0])

## Reaction 2837
#reaction('C5H10OOH1-3O2 <=> C5H92-5OOH + HO2', [3.600000e+10, 0.73, 29883.0])

## Reaction 2838
#reaction('C5H10OOH1-4O2 <=> C5H91-5OOH + HO2', [3.605000e+09, 1.13, 30100.0])

## Reaction 2839
#reaction('C5H10OOH1-4O2 <=> C5H92-5OOH + HO2', [5.140000e+09, 1.01, 29362.0])

## Reaction 2840
#reaction('C5H10OOH1-5O2 <=> C5H91-5OOH + HO2', [3.870000e+07, 1.58, 28500.0])

## Reaction 2841
#reaction('C5H10OOH2-3O2 <=> C5H92-4OOH + HO2', [3.600000e+10, 0.73, 29883.0])

## Reaction 2842
#reaction('C5H10OOH2-4O2 <=> C5H92-4OOH + HO2', [5.140000e+09, 1.01, 29362.0])

## Reaction 2843
#reaction('C5H10OOH2-4O2 <=> C5H91-4OOH + HO2', [3.605000e+09, 1.13, 30100.0])

## Reaction 2844
#reaction('C5H10OOH2-5O2 <=> C5H91-4OOH + HO2', [3.870000e+07, 1.58, 28500.0])

## Reaction 2845
#reaction('C5H10OOH3-1O2 <=> C5H91-3OOH + HO2', [3.870000e+07, 1.58, 28500.0])

## Reaction 2846
#reaction('C5H10OOH3-2O2 <=> C5H91-3OOH + HO2', [3.605000e+09, 1.13, 30100.0])

## Reaction 2847
#reaction('NC5KET12 <=> CHOC4H8O + OH', [1.050000e+16, 0.0, 42600.0])

## Reaction 2848
#reaction('CHOC4H8O <=> NC3H7CHO + HCO', [2.900000e+12, 0.38, 11080.0])

## Reaction 2849
#reaction('NC5KET13 <=> CHOCH2C3H6O + OH', [1.050000e+16, 0.0, 42600.0])

## Reaction 2850
#reaction('CHOCH2C3H6O <=> C2H5CHO + CH2CHO', [2.900000e+12, 0.38, 11080.0])

## Reaction 2851
#reaction('NC5KET14 <=> CHOCH2CH2C2H4O + OH', [1.050000e+16, 0.0, 42600.0])

## Reaction 2852
#reaction('CHOCH2CH2C2H4O <=> CH3CHO + CH2CH2CHO', [2.900000e+12, 0.38, 11080.0])

## Reaction 2853
#reaction('NC5KET15 <=> CHOCH2CH2CH2CH2O + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 2854
#reaction('CHOCH2CH2CH2CH2O <=> CH2O + C3H6CHO-3', [2.900000e+12, 0.38, 11080.0])

## Reaction 2855
#reaction('NC5KET21 <=> C3H7COCH2O + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 2856
#reaction('C3H7COCH2O <=> CH2O + NC3H7CO', [2.900000e+12, 0.38, 11080.0])

## Reaction 2857
#reaction('NC5KET23 <=> CH3COC3H6O + OH', [1.050000e+16, 0.0, 42600.0])

## Reaction 2858
#reaction('CH3COC3H6O <=> C2H5CHO + CH3CO', [2.900000e+12, 0.38, 11080.0])

## Reaction 2859
#reaction('NC5KET24 <=> CH3COCH2C2H4O + OH', [1.050000e+16, 0.0, 42600.0])

## Reaction 2860
#reaction('CH3COCH2C2H4O <=> CH3CHO + CH3COCH2', [2.900000e+12, 0.38, 11080.0])

## Reaction 2861
#reaction('NC5KET25 <=> CH3COCH2CH2CH2O + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 2862
#reaction('CH3COCH2CH2CH2O <=> CH2O + CH2CH2COCH3', [2.900000e+12, 0.38, 11080.0])

## Reaction 2863
#reaction('NC5KET31 <=> C2H5COCH2CH2O + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 2864
#reaction('C2H5COCH2CH2O <=> C2H5COCH2 + CH2O', [2.900000e+12, 0.38, 11080.0])

## Reaction 2865
#reaction('NC5KET32 <=> C2H5COC2H4O + OH', [1.050000e+16, 0.0, 42600.0])

## Reaction 2866
#reaction('C2H5COC2H4O <=> CH3CHO + C2H5CO', [2.900000e+12, 0.38, 11080.0])

## Reaction 2867
#reaction('C5H10-1 + OH <=> C5H10OH-1', [4.750000e+12, 0.0, -782.0])

## Reaction 2868
#reaction('C5H10-2 + OH <=> C5H10OH-2', [4.750000e+12, 0.0, -782.0])

## Reaction 2869
#reaction('C5H10OH-1 + O2 <=> O2C5H10OH-1', [2.000000e+12, 0.0, 0.0])

## Reaction 2870
#reaction('O2C5H10OH-1 => NC3H7CHO + CH2O + OH', [2.500000e+10, 0.0, 18860.0])

## Reaction 2871
#reaction('C5H10OH-2 + O2 <=> O2C5H10OH-2', [2.000000e+12, 0.0, 0.0])

## Reaction 2872
#reaction('O2C5H10OH-2 => C2H5CHO + CH3CHO + OH', [2.500000e+10, 0.0, 18860.0])

## Reaction 2873
#reaction('C5H10OOH1-2O2 + CH3 <=> C5H10OOH1-2O + CH3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 2874
#reaction('C5H10OOH1-3O2 + CH3 <=> C5H10OOH1-3O + CH3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 2875
#reaction('C5H10OOH1-4O2 + CH3 <=> C5H10OOH1-4O + CH3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 2876
#reaction('C5H10OOH1-5O2 + CH3 <=> C5H10OOH1-5O + CH3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 2877
#reaction('C5H10OOH2-1O2 + CH3 <=> C5H10OOH2-1O + CH3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 2878
#reaction('C5H10OOH2-3O2 + CH3 <=> C5H10OOH2-3O + CH3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 2879
#reaction('C5H10OOH2-4O2 + CH3 <=> C5H10OOH2-4O + CH3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 2880
#reaction('C5H10OOH2-5O2 + CH3 <=> C5H10OOH2-5O + CH3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 2881
#reaction('C5H10OOH3-1O2 + CH3 <=> C5H10OOH3-1O + CH3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 2882
#reaction('C5H10OOH3-2O2 + CH3 <=> C5H10OOH3-2O + CH3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 2883
#reaction('C5H10OOH1-2O2 + C2H5 <=> C5H10OOH1-2O + C2H5O', [9.000000e+12, 0.0, -1000.0])

## Reaction 2884
#reaction('C5H10OOH1-3O2 + C2H5 <=> C5H10OOH1-3O + C2H5O', [9.000000e+12, 0.0, -1000.0])

## Reaction 2885
#reaction('C5H10OOH1-4O2 + C2H5 <=> C5H10OOH1-4O + C2H5O', [9.000000e+12, 0.0, -1000.0])

## Reaction 2886
#reaction('C5H10OOH1-5O2 + C2H5 <=> C5H10OOH1-5O + C2H5O', [9.000000e+12, 0.0, -1000.0])

## Reaction 2887
#reaction('C5H10OOH2-1O2 + C2H5 <=> C5H10OOH2-1O + C2H5O', [9.000000e+12, 0.0, -1000.0])

## Reaction 2888
#reaction('C5H10OOH2-3O2 + C2H5 <=> C5H10OOH2-3O + C2H5O', [9.000000e+12, 0.0, -1000.0])

## Reaction 2889
#reaction('C5H10OOH2-4O2 + C2H5 <=> C5H10OOH2-4O + C2H5O', [9.000000e+12, 0.0, -1000.0])

## Reaction 2890
#reaction('C5H10OOH2-5O2 + C2H5 <=> C5H10OOH2-5O + C2H5O', [9.000000e+12, 0.0, -1000.0])

## Reaction 2891
#reaction('C5H10OOH3-1O2 + C2H5 <=> C5H10OOH3-1O + C2H5O', [9.000000e+12, 0.0, -1000.0])

## Reaction 2892
#reaction('C5H10OOH3-2O2 + C2H5 <=> C5H10OOH3-2O + C2H5O', [9.000000e+12, 0.0, -1000.0])

## Reaction 2893
#reaction('C5H10OOH1-2O2 + C5H11-1 <=> C5H10OOH1-2O + C5H11O-1', [9.000000e+12, 0.0, -1000.0])

## Reaction 2894
#reaction('C5H10OOH1-3O2 + C5H11-1 <=> C5H10OOH1-3O + C5H11O-1', [9.000000e+12, 0.0, -1000.0])

## Reaction 2895
#reaction('C5H10OOH1-4O2 + C5H11-1 <=> C5H10OOH1-4O + C5H11O-1', [9.000000e+12, 0.0, -1000.0])

## Reaction 2896
#reaction('C5H10OOH1-5O2 + C5H11-1 <=> C5H10OOH1-5O + C5H11O-1', [9.000000e+12, 0.0, -1000.0])

## Reaction 2897
#reaction('C5H10OOH2-1O2 + C5H11-1 <=> C5H10OOH2-1O + C5H11O-1', [9.000000e+12, 0.0, -1000.0])

## Reaction 2898
#reaction('C5H10OOH2-3O2 + C5H11-1 <=> C5H10OOH2-3O + C5H11O-1', [9.000000e+12, 0.0, -1000.0])

## Reaction 2899
#reaction('C5H10OOH2-4O2 + C5H11-1 <=> C5H10OOH2-4O + C5H11O-1', [9.000000e+12, 0.0, -1000.0])

## Reaction 2900
#reaction('C5H10OOH2-5O2 + C5H11-1 <=> C5H10OOH2-5O + C5H11O-1', [9.000000e+12, 0.0, -1000.0])

## Reaction 2901
#reaction('C5H10OOH3-1O2 + C5H11-1 <=> C5H10OOH3-1O + C5H11O-1', [9.000000e+12, 0.0, -1000.0])

## Reaction 2902
#reaction('C5H10OOH3-2O2 + C5H11-1 <=> C5H10OOH3-2O + C5H11O-1', [9.000000e+12, 0.0, -1000.0])

## Reaction 2903
#reaction('C5H10OOH1-2O2 + C5H11-2 <=> C5H10OOH1-2O + C5H11O-2', [9.000000e+12, 0.0, -1000.0])

## Reaction 2904
#reaction('C5H10OOH1-3O2 + C5H11-2 <=> C5H10OOH1-3O + C5H11O-2', [9.000000e+12, 0.0, -1000.0])

## Reaction 2905
#reaction('C5H10OOH1-4O2 + C5H11-2 <=> C5H10OOH1-4O + C5H11O-2', [9.000000e+12, 0.0, -1000.0])

## Reaction 2906
#reaction('C5H10OOH1-5O2 + C5H11-2 <=> C5H10OOH1-5O + C5H11O-2', [9.000000e+12, 0.0, -1000.0])

## Reaction 2907
#reaction('C5H10OOH2-1O2 + C5H11-2 <=> C5H10OOH2-1O + C5H11O-2', [9.000000e+12, 0.0, -1000.0])

## Reaction 2908
#reaction('C5H10OOH2-3O2 + C5H11-2 <=> C5H10OOH2-3O + C5H11O-2', [9.000000e+12, 0.0, -1000.0])

## Reaction 2909
#reaction('C5H10OOH2-4O2 + C5H11-2 <=> C5H10OOH2-4O + C5H11O-2', [9.000000e+12, 0.0, -1000.0])

## Reaction 2910
#reaction('C5H10OOH2-5O2 + C5H11-2 <=> C5H10OOH2-5O + C5H11O-2', [9.000000e+12, 0.0, -1000.0])

## Reaction 2911
#reaction('C5H10OOH3-1O2 + C5H11-2 <=> C5H10OOH3-1O + C5H11O-2', [9.000000e+12, 0.0, -1000.0])

## Reaction 2912
#reaction('C5H10OOH3-2O2 + C5H11-2 <=> C5H10OOH3-2O + C5H11O-2', [9.000000e+12, 0.0, -1000.0])

## Reaction 2913
#reaction('C5H10OOH1-2O2 + C5H11-3 <=> C5H10OOH1-2O + C5H11O-3', [9.000000e+12, 0.0, -1000.0])

## Reaction 2914
#reaction('C5H10OOH1-3O2 + C5H11-3 <=> C5H10OOH1-3O + C5H11O-3', [9.000000e+12, 0.0, -1000.0])

## Reaction 2915
#reaction('C5H10OOH1-4O2 + C5H11-3 <=> C5H10OOH1-4O + C5H11O-3', [9.000000e+12, 0.0, -1000.0])

## Reaction 2916
#reaction('C5H10OOH1-5O2 + C5H11-3 <=> C5H10OOH1-5O + C5H11O-3', [9.000000e+12, 0.0, -1000.0])

## Reaction 2917
#reaction('C5H10OOH2-1O2 + C5H11-3 <=> C5H10OOH2-1O + C5H11O-3', [9.000000e+12, 0.0, -1000.0])

## Reaction 2918
#reaction('C5H10OOH2-3O2 + C5H11-3 <=> C5H10OOH2-3O + C5H11O-3', [9.000000e+12, 0.0, -1000.0])

## Reaction 2919
#reaction('C5H10OOH2-4O2 + C5H11-3 <=> C5H10OOH2-4O + C5H11O-3', [9.000000e+12, 0.0, -1000.0])

## Reaction 2920
#reaction('C5H10OOH2-5O2 + C5H11-3 <=> C5H10OOH2-5O + C5H11O-3', [9.000000e+12, 0.0, -1000.0])

## Reaction 2921
#reaction('C5H10OOH3-1O2 + C5H11-3 <=> C5H10OOH3-1O + C5H11O-3', [9.000000e+12, 0.0, -1000.0])

## Reaction 2922
#reaction('C5H10OOH3-2O2 + C5H11-3 <=> C5H10OOH3-2O + C5H11O-3', [9.000000e+12, 0.0, -1000.0])

## Reaction 2923
#reaction('C5H10OOH1-2O <=> HO2CH2CHO + NC3H7', [2.900000e+12, 0.4, 11080.0])

## Reaction 2924
#reaction('C5H10OOH1-2O => NC3H7CHO + CH2O + OH', [2.900000e+12, 0.4, 11080.0])

## Reaction 2925
#reaction('C5H10OOH1-3O <=> C3KET13 + C2H5', [2.900000e+12, 0.4, 11080.0])

## Reaction 2926
#reaction('C5H10OOH1-3O => C2H5CHO + C2H4 + HO2', [2.900000e+12, 0.4, 11080.0])

## Reaction 2927
#reaction('C5H10OOH1-4O <=> CH3CHO + C3H6OOH1-3', [2.900000e+12, 0.4, 11080.0])

## Reaction 2928
#reaction('C5H10OOH1-4O <=> NC4KET14 + CH3', [2.900000e+12, 0.4, 11080.0])

## Reaction 2929
#reaction('C5H10OOH1-5O <=> C4H8OOH1-4 + CH2O', [2.900000e+12, 0.4, 11080.0])

## Reaction 2930
#reaction('C5H10OOH2-1O => CH2O + NC3H7CHO + OH', [2.900000e+12, 0.4, 11080.0])

## Reaction 2931
#reaction('C5H10OOH2-3O <=> C3KET12 + C2H5', [2.900000e+12, 0.4, 11080.0])

## Reaction 2932
#reaction('C5H10OOH2-3O => CH3CHO + C2H5CHO + OH', [2.900000e+12, 0.4, 11080.0])

## Reaction 2933
#reaction('C5H10OOH2-4O <=> NC4KET13 + CH3', [2.900000e+12, 0.4, 11080.0])

## Reaction 2934
#reaction('C5H10OOH2-4O <=> C3H6OOH2-1 + CH3CHO', [2.900000e+12, 0.4, 11080.0])

## Reaction 2935
#reaction('C5H10OOH2-5O <=> C4H8OOH2-4 + CH2O', [2.900000e+12, 0.4, 11080.0])

## Reaction 2936
#reaction('C5H10OOH3-1O <=> CH2O + C4H8OOH2-1', [2.900000e+12, 0.4, 11080.0])

## Reaction 2937
#reaction('C5H10OOH3-2O <=> NC4KET12 + CH3', [2.900000e+12, 0.4, 11080.0])

## Reaction 2938
#reaction('C5H10OOH3-2O => CH3CHO + C2H5CHO + OH', [2.900000e+12, 0.4, 11080.0])

## Reaction 2939
#pdep_arrhenius('IC5H12 <=> IC3H7 + C2H5',
            #[(0.01, 'atm'), 4.070000e+95, -23.37, 121614.0],
            #[(0.1, 'atm'), 2.800000e+90, -21.55, 122339.0],
            #[(1.0, 'atm'), 8.580000e+79, -18.27, 119001.0],
            #[(2.0, 'atm'), 3.960000e+75, -16.97, 117015.0],
            #[(5.0, 'atm'), 1.410000e+69, -15.06, 113723.0],
            #[(10.0, 'atm'), 7.290000e+63, -13.51, 110811.0],
            #[(20.0, 'atm'), 2.460000e+58, -11.91, 107645.0],
            #[(50.0, 'atm'), 1.550000e+51, -9.83, 103292.0])

## Reaction 2940
#pdep_arrhenius('IC5H12 <=> SC4H9 + CH3',
            #[(0.01, 'atm'), 1.050000e+93, -22.64, 121641.0],
            #[(0.1, 'atm'), 1.310000e+88, -20.89, 122563.0],
            #[(1.0, 'atm'), 6.800000e+77, -17.67, 119420.0],
            #[(2.0, 'atm'), 3.550000e+73, -16.39, 117484.0],
            #[(5.0, 'atm'), 1.430000e+67, -14.49, 114248.0],
            #[(10.0, 'atm'), 7.810000e+61, -12.94, 111365.0],
            #[(20.0, 'atm'), 2.700000e+56, -11.35, 108218.0],
            #[(50.0, 'atm'), 1.660000e+49, -9.26, 103871.0])

## Reaction 2941
#pdep_arrhenius('IC5H12 <=> IC4H9 + CH3',
            #[(0.01, 'atm'), 2.910000e+88, -21.68, 121666.0],
            #[(0.1, 'atm'), 4.790000e+84, -20.23, 123421.0],
            #[(1.0, 'atm'), 2.450000e+75, -17.28, 121108.0],
            #[(2.0, 'atm'), 2.210000e+71, -16.05, 119391.0],
            #[(5.0, 'atm'), 1.560000e+65, -14.22, 116397.0],
            #[(10.0, 'atm'), 1.120000e+60, -12.7, 113651.0],
            #[(20.0, 'atm'), 4.350000e+54, -11.12, 110591.0],
            #[(50.0, 'atm'), 2.490000e+47, -9.02, 106282.0])

## Reaction 2942
#pdep_arrhenius('IC5H12 <=> AC5H11 + H',
            #[(0.01, 'atm'), 8.010000e+67, -16.14, 118712.0],
            #[(0.1, 'atm'), 1.480000e+72, -16.92, 125626.0],
            #[(1.0, 'atm'), 7.250000e+70, -16.14, 129319.0],
            #[(2.0, 'atm'), 9.570000e+68, -15.5, 129394.0],
            #[(5.0, 'atm'), 2.140000e+65, -14.33, 128592.0],
            #[(10.0, 'atm'), 4.960000e+61, -13.21, 127269.0],
            #[(20.0, 'atm'), 2.300000e+57, -11.91, 125334.0],
            #[(50.0, 'atm'), 5.700000e+50, -9.95, 121933.0])

## Reaction 2943
#pdep_arrhenius('IC5H12 <=> BC5H11 + H',
            #[(0.01, 'atm'), 1.240000e+77, -18.65, 120361.0],
            #[(0.1, 'atm'), 1.430000e+78, -18.53, 125424.0],
            #[(1.0, 'atm'), 2.480000e+73, -16.81, 126657.0],
            #[(2.0, 'atm'), 3.390000e+70, -15.9, 125951.0],
            #[(5.0, 'atm'), 4.900000e+65, -14.41, 124153.0],
            #[(10.0, 'atm'), 1.950000e+61, -13.09, 122150.0],
            #[(20.0, 'atm'), 2.310000e+56, -11.63, 119644.0],
            #[(50.0, 'atm'), 2.030000e+49, -9.56, 115729.0])

## Reaction 2944
#pdep_arrhenius('IC5H12 <=> CC5H11 + H',
            #[(0.01, 'atm'), 7.610000e+73, -17.85, 119817.0],
            #[(0.1, 'atm'), 1.220000e+76, -18.05, 125581.0],
            #[(1.0, 'atm'), 3.070000e+72, -16.65, 127663.0],
            #[(2.0, 'atm'), 8.790000e+69, -15.82, 127218.0],
            #[(5.0, 'atm'), 3.060000e+65, -14.44, 125747.0],
            #[(10.0, 'atm'), 2.110000e+61, -13.18, 123961.0],
            #[(20.0, 'atm'), 3.770000e+56, -11.76, 121634.0],
            #[(50.0, 'atm'), 4.400000e+49, -9.73, 117871.0])

## Reaction 2945
#pdep_arrhenius('IC5H12 <=> DC5H11 + H',
            #[(0.01, 'atm'), 7.960000e+67, -16.35, 118712.0],
            #[(0.1, 'atm'), 1.480000e+72, -17.13, 125626.0],
            #[(1.0, 'atm'), 7.210000e+70, -16.36, 129319.0],
            #[(2.0, 'atm'), 9.510000e+68, -15.71, 129394.0],
            #[(5.0, 'atm'), 2.130000e+65, -14.55, 128592.0],
            #[(10.0, 'atm'), 4.940000e+61, -13.43, 127269.0],
            #[(20.0, 'atm'), 2.290000e+57, -12.12, 125334.0],
            #[(50.0, 'atm'), 5.660000e+50, -10.17, 121933.0])

## Reaction 2946
#reaction('IC5H12 + H <=> AC5H11 + H2', [3.490000e+05, 2.69, 6450.0])

## Reaction 2947
#reaction('IC5H12 + H <=> BC5H11 + H2', [6.020000e+05, 2.4, 2583.0])

## Reaction 2948
#reaction('IC5H12 + H <=> CC5H11 + H2', [1.300000e+06, 2.4, 4471.0])

## Reaction 2949
#reaction('IC5H12 + H <=> DC5H11 + H2', [1.745000e+05, 2.69, 6450.0])

## Reaction 2950
#reaction('IC5H12 + O <=> AC5H11 + OH', [1.130000e+14, 0.0, 7850.0])

## Reaction 2951
#reaction('IC5H12 + O <=> BC5H11 + OH', [3.968000e+05, 2.401, 1150.0])

## Reaction 2952
#reaction('IC5H12 + O <=> CC5H11 + OH', [5.946000e+05, 2.439, 2846.0])

## Reaction 2953
#reaction('IC5H12 + O <=> DC5H11 + OH', [1.046000e+06, 2.424, 4766.0])

## Reaction 2954
#reaction('IC5H12 + OH <=> AC5H11 + H2O', [3.348000e+06, 2.078, 375.6])

## Reaction 2955
#reaction('IC5H12 + OH <=> BC5H11 + H2O', [4.844000e+06, 1.84, -999.6])

## Reaction 2956
#reaction('IC5H12 + OH <=> CC5H11 + H2O', [7.053000e+09, 0.935, 504.7])

## Reaction 2957
#reaction('IC5H12 + OH <=> DC5H11 + H2O', [1.366000e+07, 1.813, 868.4])

## Reaction 2958
#reaction('IC5H12 + CH3 <=> AC5H11 + CH4', [9.040000e-01, 3.65, 7154.0])

## Reaction 2959
#reaction('IC5H12 + CH3 <=> BC5H11 + CH4', [6.010000e-10, 6.36, 893.0])

## Reaction 2960
#reaction('IC5H12 + CH3 <=> CC5H11 + CH4', [8.400000e+04, 2.133, 7574.0])

## Reaction 2961
#reaction('IC5H12 + CH3 <=> DC5H11 + CH4', [4.520000e-01, 3.65, 7154.0])

## Reaction 2962
#reaction('IC5H12 + HO2 <=> AC5H11 + H2O2', [4.080000e+01, 3.59, 17160.0])

## Reaction 2963
#reaction('IC5H12 + HO2 <=> BC5H11 + H2O2', [6.504000e+02, 3.01, 12090.0])

## Reaction 2964
#reaction('IC5H12 + HO2 <=> CC5H11 + H2O2', [6.320000e+01, 3.37, 13720.0])

## Reaction 2965
#reaction('IC5H12 + HO2 <=> DC5H11 + H2O2', [2.040000e+01, 3.59, 17160.0])

## Reaction 2966
#reaction('IC5H12 + CH3O2 <=> AC5H11 + CH3O2H', [1.386000e+00, 3.97, 18280.0])

## Reaction 2967
#reaction('IC5H12 + CH3O2 <=> BC5H11 + CH3O2H', [2.620000e+02, 3.12, 11100.0])

## Reaction 2968
#reaction('IC5H12 + CH3O2 <=> CC5H11 + CH3O2H', [1.019000e+01, 3.58, 14810.0])

## Reaction 2969
#reaction('IC5H12 + CH3O2 <=> DC5H11 + CH3O2H', [6.930000e-01, 3.97, 18280.0])

## Reaction 2970
#reaction('IC5H12 + CH3O <=> AC5H11 + CH3OH', [3.000000e+11, 0.0, 7000.0])

## Reaction 2971
#reaction('IC5H12 + CH3O <=> BC5H11 + CH3OH', [1.900000e+10, 0.0, 2800.0])

## Reaction 2972
#reaction('IC5H12 + CH3O <=> CC5H11 + CH3OH', [1.100000e+11, 0.0, 5000.0])

## Reaction 2973
#reaction('IC5H12 + CH3O <=> DC5H11 + CH3OH', [1.500000e+11, 0.0, 7000.0])

## Reaction 2974
#reaction('IC5H12 + O2 <=> AC5H11 + HO2', [4.200000e+13, 0.0, 52800.0])

## Reaction 2975
#reaction('IC5H12 + O2 <=> BC5H11 + HO2', [7.000000e+12, 0.0, 48000.0])

## Reaction 2976
#reaction('IC5H12 + O2 <=> CC5H11 + HO2', [1.400000e+13, 0.0, 50160.0])

## Reaction 2977
#reaction('IC5H12 + O2 <=> DC5H11 + HO2', [2.100000e+13, 0.0, 52800.0])

## Reaction 2978
#reaction('IC5H12 + C2H5 <=> AC5H11 + C2H6', [1.000000e+11, 0.0, 13400.0])

## Reaction 2979
#reaction('IC5H12 + C2H5 <=> BC5H11 + C2H6', [1.000000e+11, 0.0, 7900.0])

## Reaction 2980
#reaction('IC5H12 + C2H5 <=> CC5H11 + C2H6', [5.000000e+10, 0.0, 10400.0])

## Reaction 2981
#reaction('IC5H12 + C2H5 <=> DC5H11 + C2H6', [5.000000e+10, 0.0, 13400.0])

## Reaction 2982
#reaction('IC5H12 + C2H3 <=> AC5H11 + C2H4', [1.000000e+12, 0.0, 18000.0])

## Reaction 2983
#reaction('IC5H12 + C2H3 <=> BC5H11 + C2H4', [2.000000e+11, 0.0, 14300.0])

## Reaction 2984
#reaction('IC5H12 + C2H3 <=> CC5H11 + C2H4', [4.000000e+11, 0.0, 16800.0])

## Reaction 2985
#reaction('IC5H12 + C2H3 <=> DC5H11 + C2H4', [5.000000e+11, 0.0, 18000.0])

## Reaction 2986
#reaction('IC5H12 + AC5H11 <=> BC5H11 + IC5H12', [2.500000e+10, 0.0, 7900.0])

## Reaction 2987
#reaction('IC5H12 + AC5H11 <=> CC5H11 + IC5H12', [5.000000e+10, 0.0, 10400.0])

## Reaction 2988
#reaction('IC5H12 + AC5H11 <=> DC5H11 + IC5H12', [7.500000e+10, 0.0, 12300.0])

## Reaction 2989
#reaction('IC5H12 + BC5H11 <=> CC5H11 + IC5H12', [5.000000e+10, 0.0, 10400.0])

## Reaction 2990
#reaction('IC5H12 + BC5H11 <=> DC5H11 + IC5H12', [7.500000e+10, 0.0, 12300.0])

## Reaction 2991
#reaction('IC5H12 + CC5H11 <=> DC5H11 + IC5H12', [7.500000e+10, 0.0, 12300.0])

## Reaction 2992
#reaction('IC5H12 + O2CHO <=> AC5H11 + HO2CHO', [1.680000e+13, 0.0, 20440.0])

## Reaction 2993
#reaction('IC5H12 + O2CHO <=> BC5H11 + HO2CHO', [1.500000e+04, 2.5, 12260.0])

## Reaction 2994
#reaction('IC5H12 + O2CHO <=> CC5H11 + HO2CHO', [5.600000e+12, 0.0, 17690.0])

## Reaction 2995
#reaction('IC5H12 + O2CHO <=> DC5H11 + HO2CHO', [8.400000e+12, 0.0, 20440.0])

## Reaction 2996
#reaction('C3H6 + C2H5 <=> AC5H11', [1.420000e+03, 2.67, 6850.0])

## Reaction 2997
#reaction('C4H8-1 + CH3 <=> AC5H11', [1.890000e+03, 2.67, 6850.0])

## Reaction 2998
#reaction('AC5H10 + H <=> AC5H11', [6.250000e+11, 0.51, 2620.0])

## Reaction 2999
#reaction('AC5H11 <=> DC5H11', [3.000000e+11, 0.0, 21100.0])

## Reaction 3000
#reaction('IC4H8 + CH3 <=> BC5H11', [4.400000e+04, 2.48, 6130.0])

## Reaction 3001
#reaction('AC5H10 + H <=> BC5H11', [1.060000e+12, 0.51, 1230.0])

## Reaction 3002
#reaction('BC5H10 + H <=> BC5H11', [6.250000e+11, 0.51, 2620.0])

## Reaction 3003
#reaction('C4H8-2 + CH3 <=> CC5H11', [1.890000e+03, 2.67, 6850.0])

## Reaction 3004
#reaction('BC5H10 + H <=> CC5H11', [2.500000e+11, 0.51, 2620.0])

## Reaction 3005
#reaction('CC5H10 + H <=> CC5H11', [6.250000e+11, 0.51, 2620.0])

## Reaction 3006
#reaction('C2H4 + IC3H7 <=> DC5H11', [8.800000e+03, 2.48, 6130.0])

## Reaction 3007
#reaction('CC5H10 + H <=> DC5H11', [2.500000e+11, 0.51, 2620.0])

## Reaction 3008
#reaction('AC5H11 + O2 <=> AC5H10 + HO2', [8.370000e-01, 3.59, 11960.0])

## Reaction 3009
#reaction('BC5H11 + O2 <=> AC5H10 + HO2', [5.350000e-01, 3.71, 9322.0])

## Reaction 3010
#reaction('BC5H11 + O2 <=> BC5H10 + HO2', [1.070000e+00, 3.71, 9322.0])

## Reaction 3011
#reaction('CC5H11 + O2 <=> BC5H10 + HO2', [1.070000e+00, 3.71, 9322.0])

## Reaction 3012
#reaction('CC5H11 + O2 <=> CC5H10 + HO2', [5.350000e-01, 3.71, 9322.0])

## Reaction 3013
#reaction('DC5H11 + O2 <=> CC5H10 + HO2', [8.370000e-01, 3.59, 11960.0])

## Reaction 3014
#reaction('AC5H11 + HO2 <=> AC5H11O + OH', [9.000000e+12, 0.0, -1000.0])

## Reaction 3015
#reaction('BC5H11 + HO2 <=> BC5H11O + OH', [9.000000e+12, 0.0, -1000.0])

## Reaction 3016
#reaction('CC5H11 + HO2 <=> CC5H11O + OH', [9.000000e+12, 0.0, -1000.0])

## Reaction 3017
#reaction('DC5H11 + HO2 <=> DC5H11O + OH', [9.000000e+12, 0.0, -1000.0])

## Reaction 3018
#reaction('AC5H11 + CH3O2 <=> AC5H11O + CH3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 3019
#reaction('BC5H11 + CH3O2 <=> BC5H11O + CH3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 3020
#reaction('CC5H11 + CH3O2 <=> CC5H11O + CH3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 3021
#reaction('DC5H11 + CH3O2 <=> DC5H11O + CH3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 3022
#reaction('AC5H10 <=> IC4H7 + CH3', [1.900000e+20, -1.582, 75930.0])

## Reaction 3023
#reaction('BC5H10 <=> C4H72-2 + CH3', [1.217000e+23, -1.926, 101400.0])

## Reaction 3024
#reaction('BC5H10 <=> IC4H7 + CH3', [2.610000e+19, -1.017, 79020.0])

## Reaction 3025
#reaction('CC5H10 <=> C4H71-3 + CH3', [1.302000e+21, -1.639, 76140.0])

## Reaction 3026
#reaction('BC5H10 + OH <=> IC3H7 + CH3CHO', [2.000000e+10, 0.0, 4000.0])

## Reaction 3027
#reaction('CC5H10 + OH <=> IC4H9 + CH2O', [2.000000e+10, 0.0, 4000.0])

## Reaction 3028
#reaction('BC5H10 + H <=> AC5H9-C + H2', [3.460000e+05, 2.5, 2492.0])

## Reaction 3029
#reaction('BC5H10 + H <=> CC5H9-B + H2', [1.730000e+05, 2.5, 2492.0])

## Reaction 3030
#reaction('BC5H10 + OH <=> AC5H9-C + H2O', [6.240000e+06, 2.0, -298.0])

## Reaction 3031
#reaction('BC5H10 + OH <=> CC5H9-B + H2O', [3.120000e+06, 2.0, -298.0])

## Reaction 3032
#reaction('BC5H10 + CH3 <=> AC5H9-C + CH4', [4.420000e+00, 3.5, 5675.0])

## Reaction 3033
#reaction('BC5H10 + CH3 <=> CC5H9-B + CH4', [2.210000e+00, 3.5, 5675.0])

## Reaction 3034
#reaction('BC5H10 + HO2 <=> AC5H9-C + H2O2', [1.928000e+04, 2.6, 13910.0])

## Reaction 3035
#reaction('BC5H10 + HO2 <=> CC5H9-B + H2O2', [9.639000e+03, 2.6, 13910.0])

## Reaction 3036
#reaction('BC5H10 + CH3O2 <=> AC5H9-C + CH3O2H', [1.928000e+04, 2.6, 13910.0])

## Reaction 3037
#reaction('BC5H10 + CH3O2 <=> CC5H9-B + CH3O2H', [9.639000e+03, 2.6, 13910.0])

## Reaction 3038
#reaction('BC5H10 + CH3O <=> AC5H9-C + CH3OH', [1.800000e+02, 2.95, 11990.0])

## Reaction 3039
#reaction('BC5H10 + CH3O <=> CC5H9-B + CH3OH', [9.000000e+01, 2.95, 11990.0])

## Reaction 3040
#reaction('CC5H10 + H <=> CC5H9-A + H2', [3.900000e+06, 2.4, 4471.0])

## Reaction 3041
#reaction('CC5H10 + H <=> CC5H9-B + H2', [2.650000e+06, 2.2, 0.0])

## Reaction 3042
#reaction('CC5H10 + OH <=> CC5H9-A + H2O', [1.400000e+08, 1.61, -35.0])

## Reaction 3043
#reaction('CC5H10 + OH <=> CC5H9-B + H2O', [6.140000e+02, 3.2, -3500.0])

## Reaction 3044
#reaction('CC5H10 + CH3 <=> CC5H9-A + CH4', [4.530000e+00, 3.46, 5481.0])

## Reaction 3045
#reaction('CC5H10 + CH3 <=> CC5H9-B + CH4', [4.613000e+00, 3.1, 2330.0])

## Reaction 3046
#reaction('CC5H10 + HO2 <=> CC5H9-A + H2O2', [2.892000e+04, 2.6, 13910.0])

## Reaction 3047
#reaction('CC5H10 + HO2 <=> CC5H9-B + H2O2', [1.810000e+03, 2.5, 7154.0])

## Reaction 3048
#reaction('CC5H10 + CH3O2 <=> CC5H9-A + CH3O2H', [2.892000e+04, 2.6, 13910.0])

## Reaction 3049
#reaction('CC5H10 + CH3O2 <=> CC5H9-B + CH3O2H', [1.810000e+03, 2.5, 7154.0])

## Reaction 3050
#reaction('CC5H10 + CH3O <=> CC5H9-A + CH3OH', [4.350000e+11, 0.0, 4571.0])

## Reaction 3051
#reaction('CC5H10 + CH3O <=> CC5H9-B + CH3OH', [1.000000e+01, 2.85, 5231.0])

## Reaction 3052
#reaction('AC5H9-A2 + HO2 <=> AC5H9O-A2 + OH', [9.640000e+12, 0.0, 0.0])

## Reaction 3053
#reaction('AC5H9-A2 + CH3O2 <=> AC5H9O-A2 + CH3O', [9.640000e+12, 0.0, 0.0])

## Reaction 3054
#reaction('AC5H9-A2 + C2H5O2 <=> AC5H9O-A2 + C2H5O', [9.640000e+12, 0.0, 0.0])

## Reaction 3055
#reaction('AC5H9-C + C2H5O2 <=> AC5H9O-C + C2H5O', [9.640000e+12, 0.0, 0.0])

## Reaction 3056
#reaction('AC5H9-D <=> AC5H9-A2', [1.113000e+12, 0.0, 31700.0])

## Reaction 3057
#reaction('CC5H9-B + HO2 <=> CC5H9O-B + OH', [9.640000e+12, 0.0, 0.0])

## Reaction 3058
#reaction('CC5H9-B + CH3O2 <=> CC5H9O-B + CH3O', [9.640000e+12, 0.0, 0.0])

## Reaction 3059
#reaction('CC5H9-B + C2H5O2 <=> CC5H9O-B + C2H5O', [9.640000e+12, 0.0, 0.0])

## Reaction 3060
#reaction('AC5H9O-A2 <=> C4H71-2 + CH2O', [9.210000e+17, -1.43, 30330.0])

## Reaction 3061
#reaction('AC5H11O2 + IC5H12 <=> AC5H11O2H + AC5H11', [1.210000e+13, 0.0, 20430.0])

## Reaction 3062
#reaction('BC5H11O2 + IC5H12 <=> BC5H11O2H + AC5H11', [1.210000e+13, 0.0, 20430.0])

## Reaction 3063
#reaction('CC5H11O2 + IC5H12 <=> CC5H11O2H + AC5H11', [1.210000e+13, 0.0, 20430.0])

## Reaction 3064
#reaction('DC5H11O2 + IC5H12 <=> DC5H11O2H + AC5H11', [1.210000e+13, 0.0, 20430.0])

## Reaction 3065
#reaction('AC5H11O2 + IC5H12 <=> AC5H11O2H + BC5H11', [2.016000e+12, 0.0, 16000.0])

## Reaction 3066
#reaction('BC5H11O2 + IC5H12 <=> BC5H11O2H + BC5H11', [2.016000e+12, 0.0, 16000.0])

## Reaction 3067
#reaction('CC5H11O2 + IC5H12 <=> CC5H11O2H + BC5H11', [2.016000e+12, 0.0, 16000.0])

## Reaction 3068
#reaction('DC5H11O2 + IC5H12 <=> DC5H11O2H + BC5H11', [2.016000e+12, 0.0, 16000.0])

## Reaction 3069
#reaction('AC5H11O2 + IC5H12 <=> AC5H11O2H + CC5H11', [4.032000e+12, 0.0, 17700.0])

## Reaction 3070
#reaction('BC5H11O2 + IC5H12 <=> BC5H11O2H + CC5H11', [4.032000e+12, 0.0, 17700.0])

## Reaction 3071
#reaction('CC5H11O2 + IC5H12 <=> CC5H11O2H + CC5H11', [4.032000e+12, 0.0, 17700.0])

## Reaction 3072
#reaction('DC5H11O2 + IC5H12 <=> DC5H11O2H + CC5H11', [4.032000e+12, 0.0, 17700.0])

## Reaction 3073
#reaction('AC5H11O2 + IC5H12 <=> AC5H11O2H + DC5H11', [6.048000e+12, 0.0, 20430.0])

## Reaction 3074
#reaction('BC5H11O2 + IC5H12 <=> BC5H11O2H + DC5H11', [6.048000e+12, 0.0, 20430.0])

## Reaction 3075
#reaction('CC5H11O2 + IC5H12 <=> CC5H11O2H + DC5H11', [6.048000e+12, 0.0, 20430.0])

## Reaction 3076
#reaction('DC5H11O2 + IC5H12 <=> DC5H11O2H + DC5H11', [6.048000e+12, 0.0, 20430.0])

## Reaction 3077
#reaction('AC5H11 + AC5H11O2 <=> AC5H11O + AC5H11O', [9.000000e+12, 0.0, -1000.0])

## Reaction 3078
#reaction('AC5H11 + BC5H11O2 <=> AC5H11O + BC5H11O', [9.000000e+12, 0.0, -1000.0])

## Reaction 3079
#reaction('AC5H11 + CC5H11O2 <=> AC5H11O + CC5H11O', [9.000000e+12, 0.0, -1000.0])

## Reaction 3080
#reaction('AC5H11 + DC5H11O2 <=> AC5H11O + DC5H11O', [9.000000e+12, 0.0, -1000.0])

## Reaction 3081
#reaction('BC5H11 + AC5H11O2 <=> BC5H11O + AC5H11O', [9.000000e+12, 0.0, -1000.0])

## Reaction 3082
#reaction('BC5H11 + BC5H11O2 <=> BC5H11O + BC5H11O', [9.000000e+12, 0.0, -1000.0])

## Reaction 3083
#reaction('BC5H11 + CC5H11O2 <=> BC5H11O + CC5H11O', [9.000000e+12, 0.0, -1000.0])

## Reaction 3084
#reaction('BC5H11 + DC5H11O2 <=> BC5H11O + DC5H11O', [9.000000e+12, 0.0, -1000.0])

## Reaction 3085
#reaction('CC5H11 + AC5H11O2 <=> CC5H11O + AC5H11O', [9.000000e+12, 0.0, -1000.0])

## Reaction 3086
#reaction('CC5H11 + BC5H11O2 <=> CC5H11O + BC5H11O', [9.000000e+12, 0.0, -1000.0])

## Reaction 3087
#reaction('CC5H11 + CC5H11O2 <=> CC5H11O + CC5H11O', [9.000000e+12, 0.0, -1000.0])

## Reaction 3088
#reaction('CC5H11 + DC5H11O2 <=> CC5H11O + DC5H11O', [9.000000e+12, 0.0, -1000.0])

## Reaction 3089
#reaction('DC5H11 + AC5H11O2 <=> DC5H11O + AC5H11O', [9.000000e+12, 0.0, -1000.0])

## Reaction 3090
#reaction('DC5H11 + BC5H11O2 <=> DC5H11O + BC5H11O', [9.000000e+12, 0.0, -1000.0])

## Reaction 3091
#reaction('DC5H11 + CC5H11O2 <=> DC5H11O + CC5H11O', [9.000000e+12, 0.0, -1000.0])

## Reaction 3092
#reaction('DC5H11 + DC5H11O2 <=> DC5H11O + DC5H11O', [9.000000e+12, 0.0, -1000.0])

## Reaction 3093
#reaction('AC5H11O2 + HO2 <=> AC5H11O2H + O2', [1.750000e+10, 0.0, -3275.0])

## Reaction 3094
#reaction('AC5H11O2 + H2O2 <=> AC5H11O2H + HO2', [2.400000e+12, 0.0, 10000.0])

## Reaction 3095
#reaction('AC5H11O2 + AC5H11O2 => O2 + AC5H11O + AC5H11O', [1.400000e+16, -1.61, 1860.0])

## Reaction 3096
#reaction('AC5H11O2 + BC5H11O2 => O2 + AC5H11O + BC5H11O', [1.400000e+16, -1.61, 1860.0])

## Reaction 3097
#reaction('AC5H11O2 + CC5H11O2 => O2 + AC5H11O + CC5H11O', [1.400000e+16, -1.61, 1860.0])

## Reaction 3098
#reaction('AC5H11O2 + DC5H11O2 => O2 + AC5H11O + DC5H11O', [1.400000e+16, -1.61, 1860.0])

## Reaction 3099
#reaction('AC5H11O2 + CH3O2 => AC5H11O + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 3100
#reaction('BC5H11O2 + HO2 <=> BC5H11O2H + O2', [1.750000e+10, 0.0, -3275.0])

## Reaction 3101
#reaction('BC5H11O2 + H2O2 <=> BC5H11O2H + HO2', [2.400000e+12, 0.0, 10000.0])

## Reaction 3102
#reaction('BC5H11O2 + BC5H11O2 => O2 + BC5H11O + BC5H11O', [1.400000e+16, -1.61, 1860.0])

## Reaction 3103
#reaction('BC5H11O2 + CC5H11O2 => O2 + BC5H11O + CC5H11O', [1.400000e+16, -1.61, 1860.0])

## Reaction 3104
#reaction('BC5H11O2 + DC5H11O2 => O2 + BC5H11O + DC5H11O', [1.400000e+16, -1.61, 1860.0])

## Reaction 3105
#reaction('BC5H11O2 + CH3O2 => BC5H11O + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 3106
#reaction('CC5H11O2 + HO2 <=> CC5H11O2H + O2', [1.750000e+10, 0.0, -3275.0])

## Reaction 3107
#reaction('CC5H11O2 + H2O2 <=> CC5H11O2H + HO2', [2.400000e+12, 0.0, 10000.0])

## Reaction 3108
#reaction('CC5H11O2 + CC5H11O2 => O2 + CC5H11O + CC5H11O', [1.400000e+16, -1.61, 1860.0])

## Reaction 3109
#reaction('CC5H11O2 + DC5H11O2 => O2 + CC5H11O + DC5H11O', [1.400000e+16, -1.61, 1860.0])

## Reaction 3110
#reaction('CC5H11O2 + CH3O2 => CC5H11O + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 3111
#reaction('DC5H11O2 + HO2 <=> DC5H11O2H + O2', [1.750000e+10, 0.0, -3275.0])

## Reaction 3112
#reaction('DC5H11O2 + H2O2 <=> DC5H11O2H + HO2', [2.400000e+12, 0.0, 10000.0])

## Reaction 3113
#reaction('DC5H11O2 + DC5H11O2 => O2 + DC5H11O + DC5H11O', [1.400000e+16, -1.61, 1860.0])

## Reaction 3114
#reaction('DC5H11O2 + CH3O2 => DC5H11O + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 3115
#reaction('AC5H11O2H <=> AC5H11O + OH', [3.031000e+20, -1.414, 46072.1])

## Reaction 3116
#reaction('BC5H11O2H <=> BC5H11O + OH', [8.618000e+20, -1.641, 45155.6])

## Reaction 3117
#reaction('CC5H11O2H <=> CC5H11O + OH', [1.129000e+21, -1.555, 45777.4])

## Reaction 3118
#reaction('DC5H11O2H <=> DC5H11O + OH', [3.591000e+20, -1.436, 46103.7])

## Reaction 3119
#reaction('CH2O + SC4H9 <=> AC5H11O', [5.000000e+10, 0.0, 2330.0])

## Reaction 3120
#reaction('C2H5 + CH3COCH3 <=> BC5H11O', [8.330000e+09, 0.0, 7240.0])

## Reaction 3121
#reaction('CH3CHO + IC3H7 <=> CC5H11O', [1.665000e+10, 0.0, 5230.0])

## Reaction 3122
#reaction('CH2O + IC4H9 <=> DC5H11O', [5.000000e+10, 0.0, 2330.0])

## Reaction 3123
#reaction('AC5H11 + O2 <=> AC5H11O2', [6.865000e+16, -1.627, 198.7])

## Reaction 3124
#reaction('BC5H11 + O2 <=> BC5H11O2', [9.756000e+11, 0.325, -417.3])

## Reaction 3125
#reaction('CC5H11 + O2 <=> CC5H11O2', [3.487000e+14, -0.816, -536.5])

## Reaction 3126
#reaction('DC5H11 + O2 <=> DC5H11O2', [6.865000e+16, -1.627, 198.7])

## Reaction 3127
#reaction('AC5H11O2 <=> AC5H10OOH-A', [4.910000e+07, 1.3, 21500.0])

## Reaction 3128
#reaction('AC5H11O2 <=> AC5H10OOH-B', [2.309000e+09, 0.8, 27100.0])

## Reaction 3129
#reaction('AC5H11O2 <=> AC5H10OOH-C', [1.360000e+07, 1.3, 18200.0])

## Reaction 3130
#reaction('AC5H11O2 <=> AC5H10OOH-D', [1.233000e+06, 1.5, 20000.0])

## Reaction 3131
#reaction('BC5H11O2 <=> BC5H10OOH-A', [3.014000e+09, 1.2, 33500.0])

## Reaction 3132
#reaction('BC5H11O2 <=> BC5H10OOH-C', [6.513000e+09, 0.7, 30100.0])

## Reaction 3133
#reaction('BC5H11O2 <=> BC5H10OOH-D', [1.504000e+08, 1.1, 21900.0])

## Reaction 3134
#reaction('CC5H11O2 <=> CC5H10OOH-A', [2.877000e+07, 1.4, 20800.0])

## Reaction 3135
#reaction('CC5H11O2 <=> CC5H10OOH-B', [7.449000e+09, 0.6, 27300.0])

## Reaction 3136
#reaction('CC5H11O2 <=> CC5H10OOH-D', [1.458000e+09, 1.1, 33500.0])

## Reaction 3137
#reaction('DC5H11O2 <=> DC5H10OOH-A', [2.467000e+06, 1.5, 20000.0])

## Reaction 3138
#reaction('DC5H11O2 <=> DC5H10OOH-B', [1.231000e+07, 1.2, 15400.0])

## Reaction 3139
#reaction('DC5H11O2 <=> DC5H10OOH-C', [4.009000e+08, 1.1, 30100.0])

## Reaction 3140
#reaction('AC5H11O2 <=> AC5H10 + HO2', [1.260000e+08, 1.32, 28900.0])

## Reaction 3141
#reaction('BC5H11O2 <=> AC5H10 + HO2', [2.448000e+10, 0.89, 29500.0])

## Reaction 3142
#reaction('BC5H11O2 <=> BC5H10 + HO2', [1.124000e+11, 0.58, 29600.0])

## Reaction 3143
#reaction('CC5H11O2 <=> BC5H10 + HO2', [3.500000e+10, 0.71, 30100.0])

## Reaction 3144
#reaction('CC5H11O2 <=> CC5H10 + HO2', [1.476000e+09, 1.21, 29800.0])

## Reaction 3145
#reaction('DC5H11O2 <=> CC5H10 + HO2', [1.518000e+07, 1.66, 28400.0])

## Reaction 3146
#reaction('AC5H10OOH-A <=> A-AC5H10O + OH', [4.100000e+10, 0.36, 17200.0])

## Reaction 3147
#reaction('AC5H10OOH-B <=> A-BC5H10O + OH', [7.190000e+09, 0.88, 9600.0])

## Reaction 3148
#reaction('AC5H10OOH-C <=> A-CC5H10O + OH', [1.701000e+09, 0.734, 14983.0])

## Reaction 3149
#reaction('AC5H10OOH-D <=> A-DC5H10O + OH', [5.200000e+08, 0.67, 10500.0])

## Reaction 3150
#reaction('BC5H10OOH-A <=> A-BC5H10O + OH', [1.050000e+12, 0.15, 9600.0])

## Reaction 3151
#reaction('BC5H10OOH-C <=> B-CC5H10O + OH', [4.270000e+10, 0.52, 8800.0])

## Reaction 3152
#reaction('BC5H10OOH-D <=> B-DC5H10O + OH', [1.030000e+10, 0.55, 16200.0])

## Reaction 3153
#reaction('CC5H10OOH-A <=> A-CC5H10O + OH', [3.792000e+07, 1.517, 16033.0])

## Reaction 3154
#reaction('CC5H10OOH-B <=> B-CC5H10O + OH', [5.360000e+10, 0.54, 8900.0])

## Reaction 3155
#reaction('CC5H10OOH-D <=> C-DC5H10O + OH', [1.530000e+10, 0.64, 10500.0])

## Reaction 3156
#reaction('DC5H10OOH-A <=> A-DC5H10O + OH', [5.200000e+08, 0.67, 10500.0])

## Reaction 3157
#reaction('DC5H10OOH-B <=> B-DC5H10O + OH', [2.850000e+08, 0.92, 13400.0])

## Reaction 3158
#reaction('DC5H10OOH-C <=> C-DC5H10O + OH', [1.120000e+10, 0.82, 10900.0])

## Reaction 3159
#reaction('AC5H10OOH-B <=> AC5H10 + HO2', [8.380000e+09, 0.92, 14500.0])

## Reaction 3160
#reaction('BC5H10OOH-A <=> AC5H10 + HO2', [3.890000e+13, -0.15, 15500.0])

## Reaction 3161
#reaction('BC5H10OOH-C <=> BC5H10 + HO2', [3.010000e+11, 0.35, 15200.0])

## Reaction 3162
#reaction('CC5H10OOH-B <=> BC5H10 + HO2', [9.790000e+12, 0.04, 15400.0])

## Reaction 3163
#reaction('CC5H10OOH-D <=> CC5H10 + HO2', [6.830000e+09, 0.92, 15100.0])

## Reaction 3164
#reaction('DC5H10OOH-C <=> CC5H10 + HO2', [6.240000e+10, 0.72, 15000.0])

## Reaction 3165
#reaction('AC5H10OOH-C => OH + CH2O + C4H8-2', [1.960000e+09, 1.2, 23400.0])

## Reaction 3166
#reaction('AC5H10OOH-A => OH + CH2O + C4H8-1', [8.283000e+13, -0.17, 30090.0])

## Reaction 3167
#reaction('BC5H10OOH-D => OH + CH3COCH3 + C2H4', [1.182000e+20, -2.06, 24210.0])

## Reaction 3168
#reaction('CC5H10OOH-A => OH + CH3CHO + C3H6', [3.530000e+09, 1.2, 21900.0])

## Reaction 3169
#reaction('DC5H10OOH-B => OH + CH2O + IC4H8', [6.270000e+08, 1.4, 23600.0])

## Reaction 3170
#reaction('AC5H10OOH-A + O2 <=> AC5H10OOH-AO2', [3.089000e+16, -1.627, 198.7])

## Reaction 3171
#reaction('AC5H10OOH-B + O2 <=> AC5H10OOH-BO2', [4.390000e+11, 0.325, -417.3])

## Reaction 3172
#reaction('AC5H10OOH-C + O2 <=> AC5H10OOH-CO2', [1.569000e+14, -0.816, -536.5])

## Reaction 3173
#reaction('AC5H10OOH-D + O2 <=> AC5H10OOH-DO2', [3.089000e+16, -1.627, 198.7])

## Reaction 3174
#reaction('BC5H10OOH-A + O2 <=> BC5H10OOH-AO2', [3.089000e+16, -1.627, 198.7])

## Reaction 3175
#reaction('BC5H10OOH-C + O2 <=> BC5H10OOH-CO2', [1.569000e+14, -0.816, -536.5])

## Reaction 3176
#reaction('BC5H10OOH-D + O2 <=> BC5H10OOH-DO2', [3.089000e+16, -1.627, 198.7])

## Reaction 3177
#reaction('CC5H10OOH-A + O2 <=> CC5H10OOH-AO2', [3.089000e+16, -1.627, 198.7])

## Reaction 3178
#reaction('CC5H10OOH-B + O2 <=> CC5H10OOH-BO2', [4.390000e+11, 0.325, -417.3])

## Reaction 3179
#reaction('CC5H10OOH-D + O2 <=> CC5H10OOH-DO2', [3.089000e+16, -1.627, 198.7])

## Reaction 3180
#reaction('DC5H10OOH-A + O2 <=> DC5H10OOH-AO2', [3.089000e+16, -1.627, 198.7])

## Reaction 3181
#reaction('DC5H10OOH-B + O2 <=> DC5H10OOH-BO2', [4.390000e+11, 0.325, -417.3])

## Reaction 3182
#reaction('DC5H10OOH-C + O2 <=> DC5H10OOH-CO2', [1.569000e+14, -0.816, -536.5])

## Reaction 3183
#reaction('AC5H10OOH-AO2 <=> IC5KETAA + OH', [5.188000e+04, 1.9, 18800.0])

## Reaction 3184
#reaction('AC5H10OOH-AO2 <=> C5H9B-A,AOOH', [2.309000e+09, 0.8, 27100.0])

## Reaction 3185
#reaction('AC5H10OOH-AO2 <=> C5H9C-A,AOOH', [1.360000e+07, 1.3, 18200.0])

## Reaction 3186
#reaction('AC5H10OOH-AO2 <=> C5H9D-A,AOOH', [1.233000e+06, 1.5, 20000.0])

## Reaction 3187
#reaction('AC5H10OOH-BO2 <=> IC5KETAB + OH', [2.441000e+07, 1.6, 27900.0])

## Reaction 3188
#reaction('AC5H10OOH-BO2 <=> C5H9A-A,BOOH', [3.014000e+09, 1.2, 33500.0])

## Reaction 3189
#reaction('AC5H10OOH-BO2 <=> C5H9C-A,BOOH', [6.513000e+09, 0.7, 30100.0])

## Reaction 3190
#reaction('AC5H10OOH-BO2 <=> C5H9D-A,BOOH', [1.504000e+08, 1.1, 21900.0])

## Reaction 3191
#reaction('AC5H10OOH-CO2 <=> IC5KETAC + OH', [1.098000e+04, 2.4, 19900.0])

## Reaction 3192
#reaction('AC5H10OOH-CO2 <=> C5H9A-A,COOH', [2.877000e+07, 1.4, 20800.0])

## Reaction 3193
#reaction('AC5H10OOH-CO2 <=> C5H9B-A,COOH', [7.449000e+09, 0.6, 27300.0])

## Reaction 3194
#reaction('AC5H10OOH-CO2 <=> C5H9D-A,COOH', [1.458000e+09, 1.1, 33500.0])

## Reaction 3195
#reaction('AC5H10OOH-DO2 <=> IC5KETAD + OH', [4.766000e+03, 1.7, 16600.0])

## Reaction 3196
#reaction('AC5H10OOH-DO2 <=> C5H9A-A,DOOH', [2.467000e+06, 1.5, 20000.0])

## Reaction 3197
#reaction('AC5H10OOH-DO2 <=> C5H9B-A,DOOH', [1.231000e+07, 1.2, 15400.0])

## Reaction 3198
#reaction('AC5H10OOH-DO2 <=> C5H9C-A,DOOH', [4.009000e+08, 1.1, 30100.0])

## Reaction 3199
#reaction('BC5H10OOH-AO2 <=> C5H9A-A,BOOH', [4.910000e+07, 1.3, 21500.0])

## Reaction 3200
#reaction('BC5H10OOH-AO2 <=> C5H9C-A,BOOH', [1.360000e+07, 1.3, 18200.0])

## Reaction 3201
#reaction('BC5H10OOH-AO2 <=> C5H9D-A,BOOH', [1.233000e+06, 1.5, 20000.0])

## Reaction 3202
#reaction('BC5H10OOH-CO2 <=> C5H9A-B,COOH', [2.877000e+07, 1.4, 20800.0])

## Reaction 3203
#reaction('BC5H10OOH-CO2 <=> C5H9D-B,COOH', [1.458000e+09, 1.1, 33500.0])

## Reaction 3204
#reaction('BC5H10OOH-DO2 <=> C5H9A-B,DOOH', [2.467000e+06, 1.5, 20000.0])

## Reaction 3205
#reaction('BC5H10OOH-DO2 <=> C5H9C-B,DOOH', [4.009000e+08, 1.1, 30100.0])

## Reaction 3206
#reaction('CC5H10OOH-AO2 <=> IC5KETCA + OH', [5.786000e+01, 2.9, 17000.0])

## Reaction 3207
#reaction('CC5H10OOH-AO2 <=> C5H9A-A,COOH', [4.910000e+07, 1.3, 21500.0])

## Reaction 3208
#reaction('CC5H10OOH-AO2 <=> C5H9B-A,COOH', [2.309000e+09, 0.8, 27100.0])

## Reaction 3209
#reaction('CC5H10OOH-AO2 <=> C5H9D-A,COOH', [1.233000e+06, 1.5, 20000.0])

## Reaction 3210
#reaction('CC5H10OOH-BO2 <=> IC5KETCB + OH', [1.747000e+08, 1.7, 26000.0])

## Reaction 3211
#reaction('CC5H10OOH-BO2 <=> C5H9A-B,COOH', [3.014000e+09, 1.2, 33500.0])

## Reaction 3212
#reaction('CC5H10OOH-BO2 <=> C5H9D-B,COOH', [1.504000e+08, 1.1, 21900.0])

## Reaction 3213
#reaction('CC5H10OOH-DO2 <=> IC5KETCD + OH', [2.763000e+08, 1.2, 25700.0])

## Reaction 3214
#reaction('CC5H10OOH-DO2 <=> C5H9A-C,DOOH', [2.467000e+06, 1.5, 20000.0])

## Reaction 3215
#reaction('CC5H10OOH-DO2 <=> C5H9B-C,DOOH', [1.231000e+07, 1.2, 15400.0])

## Reaction 3216
#reaction('DC5H10OOH-AO2 <=> IC5KETDA + OH', [4.766000e+03, 1.7, 16600.0])

## Reaction 3217
#reaction('DC5H10OOH-AO2 <=> C5H9A-A,DOOH', [4.910000e+07, 1.3, 21500.0])

## Reaction 3218
#reaction('DC5H10OOH-AO2 <=> C5H9B-A,DOOH', [2.309000e+09, 0.8, 27100.0])

## Reaction 3219
#reaction('DC5H10OOH-AO2 <=> C5H9C-A,DOOH', [1.360000e+07, 1.3, 18200.0])

## Reaction 3220
#reaction('DC5H10OOH-BO2 <=> IC5KETDB + OH', [1.098000e+04, 2.4, 19900.0])

## Reaction 3221
#reaction('DC5H10OOH-BO2 <=> C5H9A-B,DOOH', [3.014000e+09, 1.2, 33500.0])

## Reaction 3222
#reaction('DC5H10OOH-BO2 <=> C5H9C-B,DOOH', [6.513000e+09, 0.7, 30100.0])

## Reaction 3223
#reaction('DC5H10OOH-CO2 <=> IC5KETDC + OH', [2.441000e+07, 1.6, 27900.0])

## Reaction 3224
#reaction('DC5H10OOH-CO2 <=> C5H9A-C,DOOH', [2.877000e+07, 1.4, 20800.0])

## Reaction 3225
#reaction('DC5H10OOH-CO2 <=> C5H9B-C,DOOH', [7.449000e+09, 0.6, 27300.0])

## Reaction 3226
#reaction('C5H9A-A,BOOH <=> C5H9A-AOOH + HO2', [3.890000e+13, -0.15, 15500.0])

## Reaction 3227
#reaction('C5H9A-AOOH => CH2O + C4H71-2 + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 3228
#reaction('C5H9A-A,COOH => AC4H7OOH + CH2O + OH', [1.978000e+08, 1.521, 21890.0])

## Reaction 3229
#reaction('C5H9A-A,DOOH <=> AC3H5OOH + C2H4O2H', [1.978000e+08, 1.521, 21890.0])

## Reaction 3230
#reaction('C5H9A-B,COOH <=> C5H9A-COOH + HO2', [3.890000e+13, -0.15, 15500.0])

## Reaction 3231
#reaction('C5H9A-COOH => CH3CHO + C3H5-T + OH', [1.050000e+16, 0.0, 42600.0])

## Reaction 3232
#reaction('C5H9A-B,DOOH <=> C5H9A-DOOH + HO2', [3.890000e+13, -0.15, 15500.0])

## Reaction 3233
#reaction('C5H9A-DOOH => CH2O + IC4H7 + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 3234
#reaction('C5H9A-C,DOOH => C3H6 + HO2CH2CHO + OH', [3.530000e+09, 1.2, 21900.0])

## Reaction 3235
#reaction('C5H9B-A,AOOH <=> C5H9A-AOOH + HO2', [1.676000e+10, 0.92, 14500.0])

## Reaction 3236
#reaction('C5H9B-A,COOH <=> C5H9A-COOH + HO2', [8.380000e+09, 0.92, 14500.0])

## Reaction 3237
#reaction('C5H9B-A,COOH <=> C5H9B-AOOH + HO2', [9.790000e+12, 0.04, 15400.0])

## Reaction 3238
#reaction('C5H9B-AOOH => CH2O + C4H72-2 + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 3239
#reaction('C5H9B-A,DOOH <=> C5H9A-DOOH + HO2', [8.380000e+09, 0.92, 14500.0])

## Reaction 3240
#reaction('C5H9B-C,DOOH <=> C5H9B-DOOH + HO2', [9.790000e+12, 0.04, 15400.0])

## Reaction 3241
#reaction('C5H9B-DOOH => CH2O + IC4H7 + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 3242
#reaction('C5H9C-A,AOOH => C4H72-1OOH + CH2O + OH', [1.960000e+09, 1.2, 23400.0])

## Reaction 3243
#reaction('C5H9C-A,BOOH <=> C5H9B-AOOH + HO2', [3.010000e+11, 0.35, 15200.0])

## Reaction 3244
#reaction('C5H9C-A,DOOH <=> C5H9C-AOOH + HO2', [6.240000e+10, 0.72, 15000.0])

## Reaction 3245
#reaction('C5H9C-AOOH => CH2O + C4H71-3 + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 3246
#reaction('C5H9C-B,DOOH <=> C5H9B-DOOH + HO2', [3.010000e+11, 0.35, 15200.0])

## Reaction 3247
#reaction('C5H9C-B,DOOH <=> C5H9C-BOOH + HO2', [6.240000e+10, 0.72, 15000.0])

## Reaction 3248
#reaction('C5H9C-BOOH => CH3COCH3 + C2H3 + OH', [6.000000e+15, 0.0, 42200.0])

## Reaction 3249
#reaction('C5H9D-A,AOOH => AC3H5OOH + C2H4 + HO2', [1.978000e+08, 1.521, 21890.0])

## Reaction 3250
#reaction('C5H9D-A,BOOH => SC3H5OOH + C2H4 + HO2', [1.960000e+09, 1.2, 23400.0])

## Reaction 3251
#reaction('SC3H5OOH => CH2CO + CH3 + OH', [1.050000e+16, 0.0, 42600.0])

## Reaction 3252
#reaction('C5H9D-A,COOH <=> C5H9C-AOOH + HO2', [6.830000e+09, 0.92, 15100.0])

## Reaction 3253
#reaction('C5H9D-B,COOH <=> C5H9C-BOOH + HO2', [6.830000e+09, 0.92, 15100.0])

## Reaction 3254
#reaction('C5H9A-A,BOOH <=> C5H9OA-AOOH-B + OH', [4.100000e+10, 0.36, 17200.0])

## Reaction 3255
#reaction('C5H9OA-AOOH-B => CH2O + C2H5COCH2 + OH', [6.000000e+15, 0.0, 42200.0])

## Reaction 3256
#reaction('C5H9A-A,BOOH <=> C5H9OA-BOOH-A + OH', [1.050000e+12, 0.15, 9600.0])

## Reaction 3257
#reaction('C5H9OA-BOOH-A => CH2O + C2H5COCH2 + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 3258
#reaction('C5H9A-A,COOH <=> C5H9OA-AOOH-C + OH', [4.100000e+10, 0.36, 17200.0])

## Reaction 3259
#reaction('C5H9OA-AOOH-C => CH3CHO + C3H5O + OH', [1.050000e+16, 0.0, 42600.0])

## Reaction 3260
#reaction('C5H9A-A,COOH <=> C5H9OA-COOH-A + OH', [3.792000e+07, 1.517, 16033.0])

## Reaction 3261
#reaction('C5H9OA-COOH-A => CH2O + C4H71-O + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 3262
#reaction('C5H9A-A,DOOH <=> C5H9OA-AOOH-D + OH', [4.100000e+10, 0.36, 17200.0])

## Reaction 3263
#reaction('C5H9OA-AOOH-D => CH2O + AC3H5OCH2 + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 3264
#reaction('AC3H5OCH2 <=> CH2O + C3H5-A', [1.000000e+14, 0.0, 25000.0])

## Reaction 3265
#reaction('C5H9A-A,DOOH <=> C5H9OA-DOOH-A + OH', [5.200000e+08, 0.67, 10500.0])

## Reaction 3266
#reaction('C5H9OA-DOOH-A => CH2O + C4H7O1-4 + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 3267
#reaction('C5H9A-B,COOH <=> C5H9OA-BOOH-C + OH', [1.050000e+12, 0.15, 9600.0])

## Reaction 3268
#reaction('C5H9OA-BOOH-C => CH3CHO + CH3COCH2 + OH', [1.050000e+16, 0.0, 42600.0])

## Reaction 3269
#reaction('C5H9A-B,COOH <=> C5H9OA-COOH-B + OH', [3.792000e+07, 1.517, 16033.0])

## Reaction 3270
#reaction('C5H9OA-COOH-B => CH3CHO + CH3COCH2 + OH', [6.000000e+15, 0.0, 42200.0])

## Reaction 3271
#reaction('C5H9A-B,DOOH <=> C5H9OA-BOOH-D + OH', [1.050000e+12, 0.15, 9600.0])

## Reaction 3272
#reaction('C5H9OA-BOOH-D => CH2O + IC4H7O + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 3273
#reaction('C5H9A-B,DOOH <=> C5H9OA-DOOH-B + OH', [5.200000e+08, 0.67, 10500.0])

## Reaction 3274
#reaction('C5H9OA-DOOH-B <=> CH3COCH2OCH2CH2 + OH', [6.000000e+15, 0.0, 42200.0])

## Reaction 3275
#reaction('CH3COCH2OCH2CH2 => CH2O + C2H4 + CH3CO', [1.600000e+13, 0.0, 25500.0])

## Reaction 3276
#reaction('C5H9A-C,DOOH <=> C5H9OA-COOH-D + OH', [3.792000e+07, 1.517, 16033.0])

## Reaction 3277
#reaction('C5H9OA-COOH-D => CH2O + IC3H6CHO + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 3278
#reaction('C5H9A-C,DOOH <=> C5H9OA-DOOH-C + OH', [5.200000e+08, 0.67, 10500.0])

## Reaction 3279
#reaction('C5H9OA-DOOH-C => CH2O + IC3H6CHO + OH', [1.050000e+16, 0.0, 42600.0])

## Reaction 3280
#reaction('C5H9B-A,AOOH <=> C5H9OA-BOOH-A + OH', [7.190000e+09, 0.88, 9600.0])

## Reaction 3281
#reaction('C5H9B-A,COOH <=> C5H9OA-BOOH-C + OH', [7.190000e+09, 0.88, 9600.0])

## Reaction 3282
#reaction('C5H9B-A,COOH <=> C5H9OB-COOH-A + OH', [5.360000e+10, 0.54, 8900.0])

## Reaction 3283
#reaction('C5H9OB-COOH-A => CH2O + CH3CHCOCH3 + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 3284
#reaction('C5H9B-A,DOOH <=> C5H9OA-BOOH-D + OH', [7.190000e+09, 0.88, 9600.0])

## Reaction 3285
#reaction('C5H9B-A,DOOH <=> C5H9OB-DOOH-A + OH', [2.850000e+08, 0.92, 13400.0])

## Reaction 3286
#reaction('C5H9OB-DOOH-A => CH2O + CH2CH2COCH3 + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 3287
#reaction('C5H9B-C,DOOH <=> C5H9OB-COOH-D + OH', [5.360000e+10, 0.54, 8900.0])

## Reaction 3288
#reaction('C5H9OB-COOH-D => CH2O + TC3H6CHO + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 3289
#reaction('C5H9B-C,DOOH <=> C5H9OB-DOOH-C + OH', [2.850000e+08, 0.92, 13400.0])

## Reaction 3290
#reaction('C5H9OB-DOOH-C => CH3COCH3 + CH2CHO + OH', [1.050000e+16, 0.0, 42600.0])

## Reaction 3291
#reaction('C5H9C-A,AOOH <=> C5H9OA-COOH-A + OH', [1.701000e+09, 0.734, 14983.0])

## Reaction 3292
#reaction('C5H9C-A,BOOH <=> C5H9OA-COOH-B + OH', [1.701000e+09, 0.734, 14983.0])

## Reaction 3293
#reaction('C5H9C-A,BOOH <=> C5H9OB-COOH-A + OH', [4.270000e+10, 0.52, 8800.0])

## Reaction 3294
#reaction('C5H9C-A,DOOH <=> C5H9OA-COOH-D + OH', [1.701000e+09, 0.734, 14983.0])

## Reaction 3295
#reaction('C5H9C-A,DOOH <=> C5H9OC-DOOH-A + OH', [1.120000e+10, 0.82, 10900.0])

## Reaction 3296
#reaction('C5H9OC-DOOH-A => CH2O + C4H7O2-1 + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 3297
#reaction('C5H9C-B,DOOH <=> C5H9OB-COOH-D + OH', [4.270000e+10, 0.52, 8800.0])

## Reaction 3298
#reaction('C5H9C-B,DOOH <=> C5H9OC-DOOH-B + OH', [1.120000e+10, 0.82, 10900.0])

## Reaction 3299
#reaction('C5H9OC-DOOH-B => CH3COCH3 + CH2CHO + OH', [6.000000e+15, 0.0, 42200.0])

## Reaction 3300
#reaction('C5H9D-A,AOOH <=> C5H9OA-DOOH-A + OH', [5.200000e+08, 0.67, 10500.0])

## Reaction 3301
#reaction('C5H9D-A,BOOH <=> C5H9OA-DOOH-B + OH', [5.200000e+08, 0.67, 10500.0])

## Reaction 3302
#reaction('C5H9D-A,BOOH <=> C5H9OB-DOOH-A + OH', [1.030000e+10, 0.55, 16200.0])

## Reaction 3303
#reaction('C5H9D-A,COOH <=> C5H9OA-DOOH-C + OH', [5.200000e+08, 0.67, 10500.0])

## Reaction 3304
#reaction('C5H9D-A,COOH <=> C5H9OC-DOOH-A + OH', [1.530000e+10, 0.64, 10500.0])

## Reaction 3305
#reaction('C5H9D-B,COOH <=> C5H9OB-DOOH-C + OH', [1.030000e+10, 0.55, 16200.0])

## Reaction 3306
#reaction('C5H9D-B,COOH <=> C5H9OC-DOOH-B + OH', [1.530000e+10, 0.64, 10500.0])

## Reaction 3307
#reaction('AC5H10OOH-AO2 <=> C5H9A-AOOH + HO2', [1.260000e+08, 1.32, 28900.0])

## Reaction 3308
#reaction('AC5H10OOH-BO2 <=> C5H9A-AOOH + HO2', [2.448000e+10, 0.89, 29500.0])

## Reaction 3309
#reaction('AC5H10OOH-BO2 <=> C5H9B-AOOH + HO2', [1.124000e+11, 0.58, 29600.0])

## Reaction 3310
#reaction('AC5H10OOH-CO2 <=> C5H9B-AOOH + HO2', [3.500000e+10, 0.71, 30100.0])

## Reaction 3311
#reaction('AC5H10OOH-CO2 <=> C5H9C-AOOH + HO2', [1.476000e+09, 1.21, 29800.0])

## Reaction 3312
#reaction('AC5H10OOH-DO2 <=> C5H9C-AOOH + HO2', [1.518000e+07, 1.66, 28400.0])

## Reaction 3313
#reaction('BC5H10OOH-CO2 <=> C5H9C-BOOH + HO2', [1.476000e+09, 1.21, 29800.0])

## Reaction 3314
#reaction('BC5H10OOH-DO2 <=> C5H9C-BOOH + HO2', [1.518000e+07, 1.66, 28400.0])

## Reaction 3315
#reaction('CC5H10OOH-AO2 <=> C5H9A-COOH + HO2', [1.260000e+08, 1.32, 28900.0])

## Reaction 3316
#reaction('CC5H10OOH-BO2 <=> C5H9A-COOH + HO2', [2.448000e+10, 0.89, 29500.0])

## Reaction 3317
#reaction('DC5H10OOH-AO2 <=> C5H9A-DOOH + HO2', [1.260000e+08, 1.32, 28900.0])

## Reaction 3318
#reaction('DC5H10OOH-BO2 <=> C5H9A-DOOH + HO2', [2.448000e+10, 0.89, 29500.0])

## Reaction 3319
#reaction('DC5H10OOH-BO2 <=> C5H9B-DOOH + HO2', [1.124000e+11, 0.58, 29600.0])

## Reaction 3320
#reaction('DC5H10OOH-CO2 <=> C5H9B-DOOH + HO2', [3.500000e+10, 0.71, 30100.0])

## Reaction 3321
#reaction('IC5KETAA <=> C2H5CH(CH2O)CHO + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 3322
#reaction('C2H5CH(CH2O)CHO <=> CH2O + C3H6CHO-1', [2.900000e+12, 0.38, 11080.0])

## Reaction 3323
#reaction('IC5KETAB <=> C3H5O(CH3)CHO + OH', [6.000000e+15, 0.0, 41200.0])

## Reaction 3324
#reaction('C3H5O(CH3)CHO <=> C2H5COCH3 + HCO', [2.900000e+12, 0.38, 11080.0])

## Reaction 3325
#reaction('IC5KETAC <=> C2H4OCH(CH3)CHO + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 3326
#reaction('C2H4OCH(CH3)CHO <=> CH3CHO + CH3CHCHO', [2.900000e+12, 0.38, 11080.0])

## Reaction 3327
#reaction('IC5KETAD <=> CH2OC3H6CHO + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 3328
#reaction('CH2OC3H6CHO <=> CH2O + IC3H6CHO', [2.900000e+12, 0.38, 11080.0])

## Reaction 3329
#reaction('IC5KETCA <=> CH3COCH(CH3)CH2O + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 3330
#reaction('CH3COCH(CH3)CH2O <=> CH2O + CH3CHCOCH3', [2.900000e+12, 0.38, 11080.0])

## Reaction 3331
#reaction('IC5KETCB <=> CH3CO(CH3)C2H3O + OH', [6.000000e+15, 0.0, 41200.0])

## Reaction 3332
#reaction('CH3CO(CH3)C2H3O <=> CH3COCH3 + CH3CO', [2.900000e+12, 0.38, 11080.0])

## Reaction 3333
#reaction('IC5KETCD <=> CH2OCOCH(CH3)CH3 + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 3334
#reaction('CH2OCOCH(CH3)CH3 <=> CH2O + IC3H7CO', [2.900000e+12, 0.38, 11080.0])

## Reaction 3335
#reaction('IC5KETDA <=> CHOC3H6CH2O + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 3336
#reaction('CHOC3H6CH2O <=> CH2O + C3H6CHO-2', [2.900000e+12, 0.38, 11080.0])

## Reaction 3337
#reaction('IC5KETDB <=> CHOCH2(CH3)C2H3O + OH', [6.000000e+15, 0.0, 41200.0])

## Reaction 3338
#reaction('CHOCH2(CH3)C2H3O <=> CH3COCH3 + CH2CHO', [2.900000e+12, 0.38, 11080.0])

## Reaction 3339
#reaction('IC5KETDC <=> CHOCHOCH(CH3)CH3 + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 3340
#reaction('CHOCHOCH(CH3)CH3 <=> IC3H7CHO + HCO', [2.900000e+12, 0.38, 11080.0])

## Reaction 3341
#reaction('A-AC5H10O + OH => CH2O + C4H71-2 + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 3342
#reaction('A-BC5H10O + OH => IC3H5CHO + CH3 + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 3343
#reaction('A-CC5H10O + OH => C3H6 + CH3CO + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 3344
#reaction('A-DC5H10O + OH => C3H6 + CH2CHO + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 3345
#reaction('B-CC5H10O + OH => IC3H5COCH3 + H + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 3346
#reaction('B-DC5H10O + OH => IC4H8 + HCO + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 3347
#reaction('C-DC5H10O + OH => CH2CO + IC3H7 + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 3348
#reaction('A-AC5H10O + OH => C2H3CHO + C2H5 + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 3349
#reaction('A-BC5H10O + OH => CH2O + C4H72-2 + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 3350
#reaction('A-CC5H10O + OH => C3H5-T + CH3CHO + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 3351
#reaction('A-DC5H10O + OH => CH3CHCHO + C2H4 + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 3352
#reaction('B-CC5H10O + OH => IC3H5CHO + CH3 + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 3353
#reaction('B-DC5H10O + OH => CH3COCH3 + C2H3 + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 3354
#reaction('C-DC5H10O + OH => SC3H5CHO + CH3 + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 3355
#reaction('A-AC5H10O + HO2 => CH2O + C4H71-2 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 3356
#reaction('A-BC5H10O + HO2 => IC3H5CHO + CH3 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 3357
#reaction('A-CC5H10O + HO2 => C3H6 + CH3CO + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 3358
#reaction('A-DC5H10O + HO2 => C3H6 + CH2CHO + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 3359
#reaction('B-CC5H10O + HO2 => IC3H5COCH3 + H + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 3360
#reaction('B-DC5H10O + HO2 => IC4H8 + HCO + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 3361
#reaction('C-DC5H10O + HO2 => CH2CO + IC3H7 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 3362
#reaction('A-AC5H10O + HO2 => C2H3CHO + C2H5 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 3363
#reaction('A-BC5H10O + HO2 => CH2O + C4H72-2 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 3364
#reaction('A-CC5H10O + HO2 => C3H5-T + CH3CHO + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 3365
#reaction('A-DC5H10O + HO2 => CH3CHCHO + C2H4 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 3366
#reaction('B-CC5H10O + HO2 => IC3H5CHO + CH3 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 3367
#reaction('B-DC5H10O + HO2 => CH3COCH3 + C2H3 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 3368
#reaction('C-DC5H10O + HO2 => SC3H5CHO + CH3 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 3369
#reaction('AC5H10 + OH <=> AC5H10OH', [1.000000e+12, 0.0, -1042.0])

## Reaction 3370
#reaction('AC5H10OH + O2 <=> AO2C5H10OH', [2.000000e+12, 0.0, 0.0])

## Reaction 3371
#reaction('AO2C5H10OH => C2H5COCH3 + CH2O + OH', [2.500000e+10, 0.0, 18860.0])

## Reaction 3372
#reaction('BC5H10 + OH <=> BC5H10OH', [1.000000e+12, 0.0, -1042.0])

## Reaction 3373
#reaction('BC5H10OH + O2 <=> BO2C5H10OH', [2.000000e+12, 0.0, 0.0])

## Reaction 3374
#reaction('BO2C5H10OH => CH3COCH3 + CH3CHO + OH', [2.500000e+10, 0.0, 18860.0])

## Reaction 3375
#reaction('CC5H10 + OH <=> CC5H10OH', [1.000000e+12, 0.0, -1042.0])

## Reaction 3376
#reaction('CC5H10OH + O2 <=> CO2C5H10OH', [2.000000e+12, 0.0, 0.0])

## Reaction 3377
#reaction('CO2C5H10OH => IC3H7CHO + CH2O + OH', [2.500000e+10, 0.0, 18860.0])

## Reaction 3378
#reaction('CH3CHCHO + H2 <=> C2H5CHO + H', [2.160000e+05, 2.38, 18990.0])

## Reaction 3379
#reaction('IC3H5COCH3 <=> IC3H5CO + CH3', [1.903000e+22, -1.752, 83625.5])

## Reaction 3380
#reaction('IC3H5COCH3 <=> C3H5-T + CH3CO', [6.070000e+24, -2.21, 96960.0])

## Reaction 3381
#reaction('IC3H5COCH3 + OH <=> IC3H5COCH2 + H2O', [5.100000e+11, 0.0, 1192.0])

## Reaction 3382
#reaction('IC3H5COCH3 + O <=> IC3H5COCH2 + OH', [5.000000e+12, 0.0, 5962.0])

## Reaction 3383
#reaction('IC3H5COCH3 + H <=> IC3H5COCH2 + H2', [9.300000e+12, 0.0, 6357.0])

## Reaction 3384
#reaction('IC3H5COCH3 + CH3 <=> IC3H5COCH2 + CH4', [1.620000e+11, 0.0, 9630.0])

## Reaction 3385
#reaction('IC3H5COCH3 + HO2 <=> IC3H5COCH2 + H2O2', [8.500000e+12, 0.0, 20460.0])

## Reaction 3386
#reaction('IC3H5COCH3 + O2 <=> IC3H5COCH2 + HO2', [6.000000e+13, 0.0, 46000.0])

## Reaction 3387
#reaction('IC3H5COCH2 <=> C3H5-T + CH2CO', [5.972000e+14, 0.188, 51272.0])

## Reaction 3388
#reaction('IC3H5COCH3 + OH <=> AC3H4COCH3 + H2O', [3.120000e+06, 2.0, -298.0])

## Reaction 3389
#reaction('IC3H5COCH3 + O <=> AC3H4COCH3 + OH', [6.030000e+10, 0.7, 7633.0])

## Reaction 3390
#reaction('IC3H5COCH3 + H <=> AC3H4COCH3 + H2', [1.730000e+05, 2.5, 2492.0])

## Reaction 3391
#reaction('IC3H5COCH3 + CH3 <=> AC3H4COCH3 + CH4', [2.210000e+00, 3.5, 5675.0])

## Reaction 3392
#reaction('IC3H5COCH3 + HO2 <=> AC3H4COCH3 + H2O2', [9.640000e+03, 2.6, 13910.0])

## Reaction 3393
#reaction('IC3H5COCH3 + O2 <=> AC3H4COCH3 + HO2', [6.030000e+13, 0.0, 47590.0])

## Reaction 3394
#reaction('AC3H4COCH3 <=> C3H4-A + CH3CO', [1.400000e+13, 0.0, 60000.0])

## Reaction 3395
#pdep_arrhenius('NEOC5H12 <=> TC4H9 + CH3',
            #[(0.01, 'atm'), 3.250000e+78, -18.54, 107427.0],
            #[(0.1, 'atm'), 8.690000e+75, -17.45, 109860.0],
            #[(1.0, 'atm'), 4.530000e+69, -15.35, 109497.0],
            #[(2.0, 'atm'), 7.600000e+66, -14.49, 108643.0],
            #[(5.0, 'atm'), 4.020000e+62, -13.18, 106951.0],
            #[(10.0, 'atm'), 9.500000e+58, -12.1, 105276.0],
            #[(20.0, 'atm'), 1.300000e+55, -10.95, 103321.0],
            #[(50.0, 'atm'), 7.340000e+49, -9.41, 100451.0])

## Reaction 3396
#pdep_arrhenius('NEOC5H12 <=> NEOC5H11 + H',
            #[(0.01, 'atm'), 2.570000e+46, -9.94, 102612.0],
            #[(0.1, 'atm'), 8.980000e+50, -10.81, 109641.0],
            #[(1.0, 'atm'), 1.010000e+52, -10.71, 114764.0],
            #[(2.0, 'atm'), 2.350000e+51, -10.42, 115648.0],
            #[(5.0, 'atm'), 5.540000e+49, -9.83, 116204.0],
            #[(10.0, 'atm'), 7.900000e+47, -9.21, 116115.0],
            #[(20.0, 'atm'), 3.410000e+45, -8.46, 115568.0],
            #[(50.0, 'atm'), 4.840000e+41, -7.28, 114151.0])

## Reaction 3397
#reaction('NEOC5H12 + H <=> NEOC5H11 + H2', [6.980000e+05, 2.69, 6450.0])

## Reaction 3398
#reaction('NEOC5H12 + O <=> NEOC5H11 + OH', [2.260000e+14, 0.0, 7850.0])

## Reaction 3399
#reaction('NEOC5H12 + OH <=> NEOC5H11 + H2O', [6.567000e+07, 1.763, 743.2])

## Reaction 3400
#reaction('NEOC5H12 + CH3 <=> NEOC5H11 + CH4', [1.808000e+00, 3.65, 7154.0])

## Reaction 3401
#reaction('NEOC5H12 + HO2 <=> NEOC5H11 + H2O2', [8.160000e+01, 3.59, 17160.0])

## Reaction 3402
#reaction('NEOC5H12 + CH3O <=> NEOC5H11 + CH3OH', [6.000000e+11, 0.0, 7000.0])

## Reaction 3403
#reaction('NEOC5H12 + O2 <=> NEOC5H11 + HO2', [8.400000e+13, 0.0, 52800.0])

## Reaction 3404
#reaction('NEOC5H12 + C2H5 <=> NEOC5H11 + C2H6', [2.000000e+11, 0.0, 13400.0])

## Reaction 3405
#reaction('NEOC5H12 + C2H3 <=> NEOC5H11 + C2H4', [2.000000e+12, 0.0, 18000.0])

## Reaction 3406
#reaction('NEOC5H12 + O2CHO <=> NEOC5H11 + HO2CHO', [2.772000e+00, 3.97, 18280.0])

## Reaction 3407
#reaction('NEOC5H12 + CH3O2 <=> NEOC5H11 + CH3O2H', [2.772000e+00, 3.97, 18280.0])

## Reaction 3408
#reaction('IC4H6OH + NEOC5H12 <=> IC4H7OH + NEOC5H11', [9.400000e+02, 3.3, 19840.0])

## Reaction 3409
#reaction('IC4H7O + NEOC5H12 <=> IC4H7OH + NEOC5H11', [5.400000e+11, 0.0, 4000.0])

## Reaction 3410
#reaction('TC3H6CHO + NEOC5H12 <=> IC3H7CHO + NEOC5H11', [9.400000e+02, 3.3, 19840.0])

## Reaction 3411
#reaction('NEOC5H12 + CH2CCH2OH <=> NEOC5H11 + C3H5OH', [1.590000e+12, 0.0, 20500.0])

## Reaction 3412
#reaction('IC4H8 + CH3 <=> NEOC5H11', [1.950000e+03, 2.48, 8520.0])

## Reaction 3413
#reaction('NEOC5H11 + HO2 <=> NEOC5H11O + OH', [9.000000e+12, 0.0, -1000.0])

## Reaction 3414
#reaction('NEOC5H11 + CH3O2 <=> NEOC5H11O + CH3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 3415
#reaction('NEOC5H11 + O2 <=> NEOC5H11O2', [6.865000e+16, -1.627, 198.7])

## Reaction 3416
#reaction('NEOC5H11O2 + NEOC5H12 <=> NEOC5H11O2H + NEOC5H11', [2.419000e+13, 0.0, 20430.0])

## Reaction 3417
#reaction('NEOC5H11 + NEOC5H11O2 <=> NEOC5H11O + NEOC5H11O', [9.000000e+12, 0.0, -1000.0])

## Reaction 3418
#reaction('NEOC5H11O2 + HO2 <=> NEOC5H11O2H + O2', [1.750000e+10, 0.0, -3275.0])

## Reaction 3419
#reaction('NEOC5H11O2 + H2O2 <=> NEOC5H11O2H + HO2', [2.400000e+12, 0.0, 10000.0])

## Reaction 3420
#reaction('NEOC5H11O2 + CH3O2 => NEOC5H11O + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 3421
#reaction('NEOC5H11O2 + NEOC5H11O2 => O2 + NEOC5H11O + NEOC5H11O', [1.400000e+16, -1.61, 1860.0])

## Reaction 3422
#reaction('NEOC5H11O2H <=> NEOC5H11O + OH', [2.094000e+17, -0.4, 45110.0])

## Reaction 3423
#reaction('CH2O + TC4H9 <=> NEOC5H11O', [5.000000e+10, 0.0, 2330.0])

## Reaction 3424
#reaction('NEOC5H11O2 <=> NEOC5H10OOH', [1.198000e+09, 1.2, 21600.0])

## Reaction 3425
#reaction('NEOC5H10OOH <=> NEO-C5H10O + OH', [3.710000e+09, 0.7, 15900.0])

## Reaction 3426
#reaction('NEOC5H10OOH => OH + CH2O + IC4H8', [3.640000e+09, 1.3, 23700.0])

## Reaction 3427
#reaction('NEOC5H10OOH <=> IC4H7OOH + CH3', [1.640000e+07, 2.0, 28300.0])

## Reaction 3428
#reaction('NEOC5H10OOH + O2 <=> NEOC5H10OOH-O2', [3.433000e+16, -1.627, 198.7])

## Reaction 3429
#reaction('NEOC5H10OOH-O2 <=> NEOC5H9Q2', [7.987000e+08, 1.2, 21600.0])

## Reaction 3430
#reaction('NEOC5H10OOH-O2 <=> NEOC5KET + OH', [5.188000e+04, 1.9, 18800.0])

## Reaction 3431
#reaction('NEOC5H9Q2 <=> IC4H7OOH + CH2O2H', [3.640000e+09, 1.3, 23700.0])

## Reaction 3432
#reaction('NEOC5H9Q2 <=> IC4H6Q2-II + CH3', [1.640000e+07, 2.0, 28300.0])

## Reaction 3433
#reaction('NEOC5H9Q2 <=> NEOC5H9O-OOH + OH', [3.710000e+09, 0.7, 15900.0])

## Reaction 3434
#reaction('NEOC5KET <=> NEOC5KETOX + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 3435
#reaction('NEOC5KETOX <=> TC3H6CHO + CH2O', [2.475000e+21, -2.54, 15830.0])

## Reaction 3436
#reaction('NEOC5KETOX <=> NEOC5KEJOL', [1.203000e+09, 1.199, 13600.7])

## Reaction 3437
#reaction('NEOC5KEJOL <=> IC4H8OH + CO', [2.273000e+21, -2.34, 11020.0])

## Reaction 3438
#reaction('NEOC5H9O-OOH => C3H5-T + CH2O + CH2O + OH', [1.000000e+16, 0.0, 43000.0])

## Reaction 3439
#reaction('NEO-C5H10O + OH <=> TC4H8CHO + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 3440
#reaction('NEO-C5H10O + OH => IC4H7 + CH2O + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 3441
#reaction('NEO-C5H10O + HO2 <=> TC4H8CHO + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 3442
#reaction('NEO-C5H10O + HO2 => IC4H7 + CH2O + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 3443
#reaction('NEO-C5H10O + H <=> TC4H8CHO + H2', [3.544000e+07, 2.0, 5000.0])

## Reaction 3444
#reaction('NEO-C5H10O + H => IC4H7 + CH2O + H2', [1.330000e+06, 2.54, 6756.0])

## Reaction 3445
#reaction('NEO-C5H10O <=> IC4H8 + CH2O', [3.800000e+15, 0.0, 60700.0])

## Reaction 3446
#pdep_arrhenius('NC6H14 <=> CH3 + C5H11-1',
            #[(0.01, 'atm'), 1.920000e+94, -23.05, 124349.0],
            #[(0.1, 'atm'), 7.100000e+83, -19.75, 121206.0],
            #[(1.0, 'atm'), 3.980000e+68, -15.14, 114401.0],
            #[(2.0, 'atm'), 2.520000e+63, -13.59, 111809.0],
            #[(5.0, 'atm'), 2.510000e+56, -11.52, 108177.0],
            #[(10.0, 'atm'), 1.450000e+51, -9.98, 105372.0],
            #[(20.0, 'atm'), 1.300000e+46, -8.5, 102608.0],
            #[(50.0, 'atm'), 8.780000e+39, -6.7, 99154.0])

## Reaction 3447
#pdep_arrhenius('NC6H14 <=> C2H5 + PC4H9',
            #[(0.01, 'atm'), 3.840000e+96, -23.49, 123675.0],
            #[(0.1, 'atm'), 3.950000e+85, -20.04, 120116.0],
            #[(1.0, 'atm'), 1.160000e+70, -15.35, 113057.0],
            #[(2.0, 'atm'), 6.940000e+64, -13.8, 110428.0],
            #[(5.0, 'atm'), 7.030000e+57, -11.74, 106773.0],
            #[(10.0, 'atm'), 4.380000e+52, -10.21, 103971.0],
            #[(20.0, 'atm'), 4.450000e+47, -8.74, 101224.0],
            #[(50.0, 'atm'), 3.730000e+41, -6.97, 97811.0])

## Reaction 3448
#pdep_arrhenius('NC6H14 <=> NC3H7 + NC3H7',
            #[(0.01, 'atm'), 8.730000e+97, -23.89, 124152.0],
            #[(0.1, 'atm'), 1.930000e+87, -20.53, 120839.0],
            #[(1.0, 'atm'), 8.220000e+71, -15.89, 113924.0],
            #[(2.0, 'atm'), 5.070000e+66, -14.34, 111315.0],
            #[(5.0, 'atm'), 5.060000e+59, -12.27, 107672.0],
            #[(10.0, 'atm'), 3.000000e+54, -10.74, 104866.0],
            #[(20.0, 'atm'), 2.820000e+49, -9.26, 102108.0],
            #[(50.0, 'atm'), 2.070000e+43, -7.47, 98669.0])

## Reaction 3449
#pdep_arrhenius('NC6H14 <=> C6H13-1 + H',
            #[(0.01, 'atm'), 1.140000e+82, -20.07, 128443.0],
            #[(0.1, 'atm'), 4.840000e+78, -18.72, 130156.0],
            #[(1.0, 'atm'), 1.570000e+68, -15.39, 127126.0],
            #[(2.0, 'atm'), 5.740000e+63, -14.04, 125268.0],
            #[(5.0, 'atm'), 1.750000e+57, -12.08, 122267.0],
            #[(10.0, 'atm'), 9.880000e+51, -10.51, 119684.0],
            #[(20.0, 'atm'), 4.440000e+46, -8.93, 116936.0],
            #[(50.0, 'atm'), 5.140000e+39, -6.89, 113232.0])

## Reaction 3450
#pdep_arrhenius('NC6H14 <=> C6H13-2 + H',
            #[(0.01, 'atm'), 7.120000e+85, -21.09, 127902.0],
            #[(0.1, 'atm'), 2.570000e+80, -19.16, 128267.0],
            #[(1.0, 'atm'), 2.050000e+68, -15.39, 124032.0],
            #[(2.0, 'atm'), 3.710000e+63, -13.95, 121912.0],
            #[(5.0, 'atm'), 6.390000e+56, -11.93, 118663.0],
            #[(10.0, 'atm'), 3.070000e+51, -10.35, 115970.0],
            #[(20.0, 'atm'), 1.460000e+46, -8.78, 113177.0],
            #[(50.0, 'atm'), 2.410000e+39, -6.79, 109502.0])

## Reaction 3451
#pdep_arrhenius('NC6H14 <=> C6H13-3 + H',
            #[(0.01, 'atm'), 7.120000e+85, -21.09, 127902.0],
            #[(0.1, 'atm'), 2.570000e+80, -19.16, 128267.0],
            #[(1.0, 'atm'), 2.050000e+68, -15.39, 124032.0],
            #[(2.0, 'atm'), 3.710000e+63, -13.95, 121912.0],
            #[(5.0, 'atm'), 6.390000e+56, -11.93, 118663.0],
            #[(10.0, 'atm'), 3.070000e+51, -10.35, 115970.0],
            #[(20.0, 'atm'), 1.460000e+46, -8.78, 113177.0],
            #[(50.0, 'atm'), 2.410000e+39, -6.79, 109502.0])

## Reaction 3452
#reaction('NC6H14 + H <=> C6H13-1 + H2', [3.490000e+05, 2.7, 6450.0])

## Reaction 3453
#reaction('NC6H14 + H <=> C6H13-2 + H2', [2.600000e+06, 2.4, 4471.0])

## Reaction 3454
#reaction('NC6H14 + H <=> C6H13-3 + H2', [2.600000e+06, 2.4, 4471.0])

## Reaction 3455
#reaction('NC6H14 + O <=> C6H13-1 + OH', [1.130000e+14, 0.0, 7850.0])

## Reaction 3456
#reaction('NC6H14 + O <=> C6H13-2 + OH', [5.620000e+13, 0.0, 5200.0])

## Reaction 3457
#reaction('NC6H14 + O <=> C6H13-3 + OH', [5.620000e+13, 0.0, 5200.0])

## Reaction 3458
#reaction('NC6H14 + OH <=> C6H13-1 + H2O', [2.731000e+07, 1.813, 818.1])

## Reaction 3459
#reaction('NC6H14 + OH <=> C6H13-2 + H2O', [1.410000e+10, 0.935, 475.5])

## Reaction 3460
#reaction('NC6H14 + OH <=> C6H13-3 + H2O', [1.144000e+07, 1.811, -956.6])

## Reaction 3461
#reaction('NC6H14 + O2 <=> C6H13-1 + HO2', [4.200000e+13, 0.0, 52800.0])

## Reaction 3462
#reaction('NC6H14 + O2 <=> C6H13-2 + HO2', [2.800000e+13, 0.0, 50160.0])

## Reaction 3463
#reaction('NC6H14 + O2 <=> C6H13-3 + HO2', [2.800000e+13, 0.0, 50160.0])

## Reaction 3464
#reaction('NC6H14 + HO2 <=> C6H13-1 + H2O2', [4.080000e+01, 3.59, 17160.0])

## Reaction 3465
#reaction('NC6H14 + HO2 <=> C6H13-2 + H2O2', [1.264000e+02, 3.37, 13720.0])

## Reaction 3466
#reaction('NC6H14 + HO2 <=> C6H13-3 + H2O2', [1.264000e+02, 3.37, 13720.0])

## Reaction 3467
#reaction('NC6H14 + CH3 <=> C6H13-1 + CH4', [9.040000e-01, 3.6, 7154.0])

## Reaction 3468
#reaction('NC6H14 + CH3 <=> C6H13-2 + CH4', [1.680000e+05, 2.1, 7574.0])

## Reaction 3469
#reaction('NC6H14 + CH3 <=> C6H13-3 + CH4', [1.680000e+05, 2.1, 7574.0])

## Reaction 3470
#reaction('NC6H14 + CH3O <=> C6H13-1 + CH3OH', [3.000000e+11, 0.0, 7000.0])

## Reaction 3471
#reaction('NC6H14 + CH3O <=> C6H13-2 + CH3OH', [2.200000e+11, 0.0, 5000.0])

## Reaction 3472
#reaction('NC6H14 + CH3O <=> C6H13-3 + CH3OH', [2.200000e+11, 0.0, 5000.0])

## Reaction 3473
#reaction('NC6H14 + CH3O2 <=> C6H13-1 + CH3O2H', [1.440000e+01, 3.7, 16800.0])

## Reaction 3474
#reaction('NC6H14 + CH3O2 <=> C6H13-2 + CH3O2H', [1.050000e+02, 3.49, 13400.0])

## Reaction 3475
#reaction('NC6H14 + CH3O2 <=> C6H13-3 + CH3O2H', [1.050000e+02, 3.49, 13400.0])

## Reaction 3476
#reaction('NC6H14 + O2CHO <=> C6H13-1 + HO2CHO', [1.440000e+01, 3.7, 16800.0])

## Reaction 3477
#reaction('NC6H14 + O2CHO <=> C6H13-2 + HO2CHO', [1.050000e+02, 3.49, 13400.0])

## Reaction 3478
#reaction('NC6H14 + O2CHO <=> C6H13-3 + HO2CHO', [1.050000e+02, 3.49, 13400.0])

## Reaction 3479
#reaction('NC6H14 + C2H3 <=> C6H13-1 + C2H4', [1.000000e+12, 0.0, 18000.0])

## Reaction 3480
#reaction('NC6H14 + C2H3 <=> C6H13-2 + C2H4', [8.000000e+11, 0.0, 16800.0])

## Reaction 3481
#reaction('NC6H14 + C2H3 <=> C6H13-3 + C2H4', [8.000000e+11, 0.0, 16800.0])

## Reaction 3482
#reaction('NC6H14 + C2H5 <=> C6H13-1 + C2H6', [1.000000e+11, 0.0, 13400.0])

## Reaction 3483
#reaction('NC6H14 + C2H5 <=> C6H13-2 + C2H6', [1.000000e+11, 0.0, 10400.0])

## Reaction 3484
#reaction('NC6H14 + C2H5 <=> C6H13-3 + C2H6', [1.000000e+11, 0.0, 10400.0])

## Reaction 3485
#reaction('C6H13-1 <=> C2H4 + PC4H9', [1.023000e+12, 0.3, 27276.3])

## Reaction 3486
#reaction('C6H13-2 <=> C3H6 + NC3H7', [4.467000e+11, 0.57, 28047.3])

## Reaction 3487
#reaction('C6H13-3 <=> C2H5 + C4H8-1', [3.548000e+12, 0.29, 28299.7])

## Reaction 3488
#reaction('C6H13-3 <=> CH3 + C5H10-1', [8.128000e+10, 0.78, 29651.0])

## Reaction 3489
#reaction('C6H13-1 + O2 <=> C6H12-1 + HO2', [8.370000e-01, 3.6, 11960.0])

## Reaction 3490
#reaction('C6H13-2 + O2 <=> C6H12-1 + HO2', [5.350000e-01, 3.7, 9322.0])

## Reaction 3491
#reaction('C6H13-2 + O2 <=> C6H12-2 + HO2', [1.070000e+00, 3.7, 9322.0])

## Reaction 3492
#reaction('C6H13-3 + O2 <=> C6H12-2 + HO2', [1.070000e+00, 3.7, 9322.0])

## Reaction 3493
#reaction('C6H13-3 + O2 <=> C6H12-3 + HO2', [1.070000e+00, 3.7, 9322.0])

## Reaction 3494
#reaction('C6H13-2 <=> C6H13-3', [2.750000e+10, 0.7, 41872.3])

## Reaction 3495
#reaction('C6H13-1 <=> C6H13-2', [1.820000e+02, 2.55, 10961.4])

## Reaction 3496
#reaction('C6H13-1 <=> C6H13-3', [6.918000e+00, 3.2, 16559.3])

## Reaction 3497
#reaction('C6H12-1 + H <=> C6H111-3 + H2', [3.380000e+05, 2.4, 207.0])

## Reaction 3498
#reaction('C6H12-1 + H <=> C6H111-4 + H2', [1.300000e+06, 2.4, 4471.0])

## Reaction 3499
#reaction('C6H12-1 + H <=> C6H111-5 + H2', [1.300000e+06, 2.4, 4471.0])

## Reaction 3500
#reaction('C6H12-1 + H <=> C6H111-6 + H2', [6.650000e+05, 2.5, 6756.0])

## Reaction 3501
#reaction('C6H12-1 + O <=> C6H111-3 + OH', [6.600000e+05, 2.4, 1210.0])

## Reaction 3502
#reaction('C6H12-1 + O <=> C6H111-4 + OH', [5.510000e+05, 2.5, 2830.0])

## Reaction 3503
#reaction('C6H12-1 + O <=> C6H111-5 + OH', [5.510000e+05, 2.5, 2830.0])

## Reaction 3504
#reaction('C6H12-1 + O <=> C6H111-6 + OH', [9.800000e+05, 2.4, 4750.0])

## Reaction 3505
#reaction('C6H12-1 + OH <=> C6H111-3 + H2O', [2.760000e+04, 2.6, -1919.0])

## Reaction 3506
#reaction('C6H12-1 + OH <=> C6H111-4 + H2O', [4.670000e+07, 1.6, -35.0])

## Reaction 3507
#reaction('C6H12-1 + OH <=> C6H111-5 + H2O', [4.670000e+07, 1.6, -35.0])

## Reaction 3508
#reaction('C6H12-1 + OH <=> C6H111-6 + H2O', [5.270000e+09, 1.0, 1586.0])

## Reaction 3509
#reaction('C6H12-1 + O2 <=> C6H111-3 + HO2', [2.200000e+12, 0.0, 37220.0])

## Reaction 3510
#reaction('C6H12-1 + O2 <=> C6H111-4 + HO2', [2.000000e+13, 0.0, 49640.0])

## Reaction 3511
#reaction('C6H12-1 + O2 <=> C6H111-5 + HO2', [2.000000e+13, 0.0, 49640.0])

## Reaction 3512
#reaction('C6H12-1 + O2 <=> C6H111-6 + HO2', [3.000000e+13, 0.0, 52290.0])

## Reaction 3513
#reaction('C6H12-1 + HO2 <=> C6H111-3 + H2O2', [4.820000e+03, 2.5, 10530.0])

## Reaction 3514
#reaction('C6H12-1 + HO2 <=> C6H111-4 + H2O2', [9.640000e+03, 2.6, 13910.0])

## Reaction 3515
#reaction('C6H12-1 + HO2 <=> C6H111-5 + H2O2', [9.640000e+03, 2.6, 13910.0])

## Reaction 3516
#reaction('C6H12-1 + HO2 <=> C6H111-6 + H2O2', [2.380000e+04, 2.5, 16490.0])

## Reaction 3517
#reaction('C6H12-1 + CH3 <=> C6H111-3 + CH4', [3.690000e+00, 3.3, 4002.0])

## Reaction 3518
#reaction('C6H12-1 + CH3 <=> C6H111-4 + CH4', [1.510000e+00, 3.5, 5481.0])

## Reaction 3519
#reaction('C6H12-1 + CH3 <=> C6H111-5 + CH4', [1.510000e+00, 3.5, 5481.0])

## Reaction 3520
#reaction('C6H12-1 + CH3 <=> C6H111-6 + CH4', [4.520000e-01, 3.6, 7154.0])

## Reaction 3521
#reaction('C6H12-1 + CH3O <=> C6H111-3 + CH3OH', [4.000000e+01, 2.9, 8609.0])

## Reaction 3522
#reaction('C6H12-1 + CH3O <=> C6H111-4 + CH3OH', [1.450000e+11, 0.0, 4571.0])

## Reaction 3523
#reaction('C6H12-1 + CH3O <=> C6H111-5 + CH3OH', [1.450000e+11, 0.0, 4571.0])

## Reaction 3524
#reaction('C6H12-1 + CH3O <=> C6H111-6 + CH3OH', [2.170000e+11, 0.0, 6458.0])

## Reaction 3525
#reaction('C6H12-1 + CH3O2 <=> C6H111-3 + CH3O2H', [4.820000e+03, 2.5, 10530.0])

## Reaction 3526
#reaction('C6H12-1 + CH3O2 <=> C6H111-4 + CH3O2H', [9.640000e+03, 2.6, 13910.0])

## Reaction 3527
#reaction('C6H12-1 + CH3O2 <=> C6H111-5 + CH3O2H', [9.640000e+03, 2.6, 13910.0])

## Reaction 3528
#reaction('C6H12-1 + CH3O2 <=> C6H111-6 + CH3O2H', [2.380000e+04, 2.5, 16490.0])

## Reaction 3529
#reaction('C6H12-2 + H <=> C6H111-3 + H2', [1.730000e+05, 2.5, 2492.0])

## Reaction 3530
#reaction('C6H12-2 + H <=> C6H112-4 + H2', [3.380000e+05, 2.4, 207.0])

## Reaction 3531
#reaction('C6H12-2 + H <=> C6H112-5 + H2', [1.300000e+06, 2.4, 4471.0])

## Reaction 3532
#reaction('C6H12-2 + H <=> C6H112-6 + H2', [6.650000e+05, 2.5, 6756.0])

## Reaction 3533
#reaction('C6H12-2 + O <=> C6H111-3 + OH', [4.410000e+05, 2.4, 3150.0])

## Reaction 3534
#reaction('C6H12-2 + O <=> C6H112-4 + OH', [6.600000e+05, 2.4, 1210.0])

## Reaction 3535
#reaction('C6H12-2 + O <=> C6H112-5 + OH', [5.510000e+05, 2.5, 2830.0])

## Reaction 3536
#reaction('C6H12-2 + O <=> C6H112-6 + OH', [9.800000e+05, 2.4, 4750.0])

## Reaction 3537
#reaction('C6H12-2 + OH <=> C6H111-3 + H2O', [3.120000e+06, 2.0, -298.0])

## Reaction 3538
#reaction('C6H12-2 + OH <=> C6H112-4 + H2O', [2.760000e+04, 2.6, -1919.0])

## Reaction 3539
#reaction('C6H12-2 + OH <=> C6H112-5 + H2O', [4.670000e+07, 1.6, -35.0])

## Reaction 3540
#reaction('C6H12-2 + OH <=> C6H112-6 + H2O', [5.270000e+09, 1.0, 1586.0])

## Reaction 3541
#reaction('C6H12-2 + O2 <=> C6H111-3 + HO2', [3.300000e+12, 0.0, 39900.0])

## Reaction 3542
#reaction('C6H12-2 + O2 <=> C6H112-4 + HO2', [2.200000e+12, 0.0, 37220.0])

## Reaction 3543
#reaction('C6H12-2 + O2 <=> C6H112-5 + HO2', [2.000000e+13, 0.0, 49640.0])

## Reaction 3544
#reaction('C6H12-2 + O2 <=> C6H112-6 + HO2', [3.000000e+13, 0.0, 52290.0])

## Reaction 3545
#reaction('C6H12-2 + HO2 <=> C6H111-3 + H2O2', [9.640000e+03, 2.6, 13910.0])

## Reaction 3546
#reaction('C6H12-2 + HO2 <=> C6H112-4 + H2O2', [4.820000e+03, 2.5, 10530.0])

## Reaction 3547
#reaction('C6H12-2 + HO2 <=> C6H112-5 + H2O2', [9.640000e+03, 2.6, 13910.0])

## Reaction 3548
#reaction('C6H12-2 + HO2 <=> C6H112-6 + H2O2', [2.380000e+04, 2.5, 16490.0])

## Reaction 3549
#reaction('C6H12-2 + CH3 <=> C6H111-3 + CH4', [2.210000e+00, 3.5, 5675.0])

## Reaction 3550
#reaction('C6H12-2 + CH3 <=> C6H112-4 + CH4', [3.690000e+00, 3.3, 4002.0])

## Reaction 3551
#reaction('C6H12-2 + CH3 <=> C6H112-5 + CH4', [1.510000e+00, 3.5, 5481.0])

## Reaction 3552
#reaction('C6H12-2 + CH3 <=> C6H112-6 + CH4', [4.520000e-01, 3.6, 7154.0])

## Reaction 3553
#reaction('C6H12-2 + CH3O <=> C6H111-3 + CH3OH', [9.000000e+01, 3.0, 11990.0])

## Reaction 3554
#reaction('C6H12-2 + CH3O <=> C6H112-4 + CH3OH', [4.000000e+01, 2.9, 8609.0])

## Reaction 3555
#reaction('C6H12-2 + CH3O <=> C6H112-5 + CH3OH', [1.450000e+11, 0.0, 4571.0])

## Reaction 3556
#reaction('C6H12-2 + CH3O <=> C6H112-6 + CH3OH', [2.170000e+11, 0.0, 6458.0])

## Reaction 3557
#reaction('C6H12-2 + CH3O2 <=> C6H111-3 + CH3O2H', [9.640000e+03, 2.6, 13910.0])

## Reaction 3558
#reaction('C6H12-2 + CH3O2 <=> C6H112-4 + CH3O2H', [4.820000e+03, 2.5, 10530.0])

## Reaction 3559
#reaction('C6H12-2 + CH3O2 <=> C6H112-5 + CH3O2H', [9.640000e+03, 2.6, 13910.0])

## Reaction 3560
#reaction('C6H12-2 + CH3O2 <=> C6H112-6 + CH3O2H', [2.380000e+04, 2.5, 16490.0])

## Reaction 3561
#reaction('C6H12-3 + H <=> C6H112-4 + H2', [3.380000e+05, 2.4, 207.0])

## Reaction 3562
#reaction('C6H12-3 + H <=> C6H113-1 + H2', [6.650000e+05, 2.5, 6756.0])

## Reaction 3563
#reaction('C6H12-3 + O <=> C6H112-4 + OH', [6.600000e+05, 2.4, 1210.0])

## Reaction 3564
#reaction('C6H12-3 + O <=> C6H113-1 + OH', [9.800000e+05, 2.4, 4750.0])

## Reaction 3565
#reaction('C6H12-3 + OH <=> C6H112-4 + H2O', [2.760000e+04, 2.6, -1919.0])

## Reaction 3566
#reaction('C6H12-3 + OH <=> C6H113-1 + H2O', [5.270000e+09, 1.0, 1586.0])

## Reaction 3567
#reaction('C6H12-3 + O2 <=> C6H112-4 + HO2', [2.200000e+12, 0.0, 37220.0])

## Reaction 3568
#reaction('C6H12-3 + O2 <=> C6H113-1 + HO2', [3.000000e+13, 0.0, 52290.0])

## Reaction 3569
#reaction('C6H12-3 + HO2 <=> C6H112-4 + H2O2', [4.820000e+03, 2.5, 10530.0])

## Reaction 3570
#reaction('C6H12-3 + HO2 <=> C6H113-1 + H2O2', [2.380000e+04, 2.5, 16490.0])

## Reaction 3571
#reaction('C6H12-3 + CH3 <=> C6H112-4 + CH4', [3.690000e+00, 3.3, 4002.0])

## Reaction 3572
#reaction('C6H12-3 + CH3 <=> C6H113-1 + CH4', [4.520000e-01, 3.6, 7154.0])

## Reaction 3573
#reaction('C6H12-3 + CH3O <=> C6H112-4 + CH3OH', [4.000000e+01, 2.9, 8609.0])

## Reaction 3574
#reaction('C6H12-3 + CH3O <=> C6H113-1 + CH3OH', [2.170000e+11, 0.0, 6458.0])

## Reaction 3575
#reaction('C6H12-3 + CH3O2 <=> C6H112-4 + CH3O2H', [4.820000e+03, 2.5, 10530.0])

## Reaction 3576
#reaction('C6H12-3 + CH3O2 <=> C6H113-1 + CH3O2H', [2.380000e+04, 2.5, 16490.0])

## Reaction 3577
#reaction('C6H12-1 + H <=> C6H13-1', [1.200000e+09, 1.3, 1049.2])

## Reaction 3578
#reaction('C6H12-1 + H <=> C6H13-2', [6.610000e+08, 1.4, -677.6])

## Reaction 3579
#reaction('C6H12-2 + H <=> C6H13-2', [3.890000e+08, 1.6, 579.9])

## Reaction 3580
#reaction('C6H12-2 + H <=> C6H13-3', [3.890000e+08, 1.6, 579.9])

## Reaction 3581
#reaction('C6H12-3 + H <=> C6H13-3', [7.780000e+08, 1.6, 579.9])

## Reaction 3582
#reaction('C6H12-1 + OH <=> C6H12OH-1J2', [4.750000e+12, 0.0, -782.0])

## Reaction 3583
#reaction('C6H12-1 + OH <=> C6H12OH-2J1', [4.750000e+12, 0.0, -782.0])

## Reaction 3584
#reaction('C6H12-2 + OH <=> C6H12OH-2J3', [4.750000e+12, 0.0, -782.0])

## Reaction 3585
#reaction('C6H12-2 + OH <=> C6H12OH-3J2', [4.750000e+12, 0.0, -782.0])

## Reaction 3586
#reaction('C6H12-3 + OH <=> C6H12OH-3J4', [4.750000e+12, 0.0, -782.0])

## Reaction 3587
#reaction('C6H12OH-1J2 + O2 <=> C6H12OH-1O2-2', [2.000000e+12, 0.0, 0.0])

## Reaction 3588
#reaction('C6H12OH-2J1 + O2 <=> C6H12OH-2O2-1', [2.000000e+12, 0.0, 0.0])

## Reaction 3589
#reaction('C6H12OH-2J3 + O2 <=> C6H12OH-2O2-3', [2.000000e+12, 0.0, 0.0])

## Reaction 3590
#reaction('C6H12OH-3J2 + O2 <=> C6H12OH-3O2-2', [2.000000e+12, 0.0, 0.0])

## Reaction 3591
#reaction('C6H12OH-3J4 + O2 <=> C6H12OH-3O2-4', [2.000000e+12, 0.0, 0.0])

## Reaction 3592
#reaction('C6H12OH-1O2-2 => CH2O + PC4H9CHO + OH', [2.500000e+10, 0.0, 18860.0])

## Reaction 3593
#reaction('C6H12OH-2O2-1 => CH2O + PC4H9CHO + OH', [2.500000e+10, 0.0, 18860.0])

## Reaction 3594
#reaction('C6H12OH-2O2-3 => CH3CHO + NC3H7CHO + OH', [2.500000e+10, 0.0, 18860.0])

## Reaction 3595
#reaction('C6H12OH-3O2-2 => CH3CHO + NC3H7CHO + OH', [2.500000e+10, 0.0, 18860.0])

## Reaction 3596
#reaction('C6H12OH-3O2-4 => C2H5CHO + C2H5CHO + OH', [2.500000e+10, 0.0, 18860.0])

## Reaction 3597
#reaction('C4H6 + C2H5 <=> C6H111-3', [1.000000e+11, 0.0, 7800.0])

## Reaction 3598
#reaction('C3H6 + C3H5-A <=> C6H111-5', [1.000000e+11, 0.0, 7800.0])

## Reaction 3599
#reaction('C4H71-4 + C2H4 <=> C6H111-6', [1.000000e+11, 0.0, 7800.0])

## Reaction 3600
#reaction('C5H81-3 + CH3 <=> C6H112-4', [1.000000e+11, 0.0, 7800.0])

## Reaction 3601
#reaction('C4H71-3 + C2H4 <=> C6H112-6', [1.000000e+11, 0.0, 7800.0])

## Reaction 3602
#reaction('C6H11-1D3O + OH <=> C6H111-3 + HO2', [6.090000e+15, -1.1, 15720.0])

## Reaction 3603
#reaction('C6H11-1D4O + OH <=> C6H111-4 + HO2', [6.090000e+15, -1.1, 15720.0])

## Reaction 3604
#reaction('C6H11-1D5O + OH <=> C6H111-5 + HO2', [6.090000e+15, -1.1, 15720.0])

## Reaction 3605
#reaction('C6H11-1D6O + OH <=> C6H111-6 + HO2', [6.090000e+15, -1.1, 15720.0])

## Reaction 3606
#reaction('C6H11-2D4O + OH <=> C6H112-4 + HO2', [7.030000e+15, -1.2, 15890.0])

## Reaction 3607
#reaction('C6H11-2D5O + OH <=> C6H112-5 + HO2', [7.030000e+15, -1.2, 15890.0])

## Reaction 3608
#reaction('C6H11-2D6O + OH <=> C6H112-6 + HO2', [7.030000e+15, -1.2, 15890.0])

## Reaction 3609
#reaction('C6H11-3D1O + OH <=> C6H113-1 + HO2', [7.030000e+15, -1.2, 15890.0])

## Reaction 3610
#reaction('C6H11-1D3O + CH3O <=> C6H111-3 + CH3O2', [5.950000e+17, -1.6, 20480.0])

## Reaction 3611
#reaction('C6H11-1D4O + CH3O <=> C6H111-4 + CH3O2', [5.950000e+17, -1.6, 20480.0])

## Reaction 3612
#reaction('C6H11-1D5O + CH3O <=> C6H111-5 + CH3O2', [5.950000e+17, -1.6, 20480.0])

## Reaction 3613
#reaction('C6H11-1D6O + CH3O <=> C6H111-6 + CH3O2', [5.950000e+17, -1.6, 20480.0])

## Reaction 3614
#reaction('C6H11-2D4O + CH3O <=> C6H112-4 + CH3O2', [6.860000e+17, -1.8, 20650.0])

## Reaction 3615
#reaction('C6H11-2D5O + CH3O <=> C6H112-5 + CH3O2', [6.860000e+17, -1.8, 20650.0])

## Reaction 3616
#reaction('C6H11-2D6O + CH3O <=> C6H112-6 + CH3O2', [6.860000e+17, -1.8, 20650.0])

## Reaction 3617
#reaction('C6H11-3D1O + CH3O <=> C6H113-1 + CH3O2', [6.860000e+17, -1.8, 20650.0])

## Reaction 3618
#reaction('C6H11-1D3O + C2H5O <=> C6H111-3 + C2H5O2', [3.890000e+14, -0.7, 18330.0])

## Reaction 3619
#reaction('C6H11-2D1O + C2H5O <=> C6H111-3 + C2H5O2', [3.890000e+14, -0.7, 18330.0])

## Reaction 3620
#reaction('C6H11-1D4O + C2H5O <=> C6H111-4 + C2H5O2', [3.890000e+14, -0.7, 18330.0])

## Reaction 3621
#reaction('C6H11-1D5O + C2H5O <=> C6H111-5 + C2H5O2', [3.890000e+14, -0.7, 18330.0])

## Reaction 3622
#reaction('C6H11-1D6O + C2H5O <=> C6H111-6 + C2H5O2', [3.890000e+14, -0.7, 18330.0])

## Reaction 3623
#reaction('C6H11-2D4O + C2H5O <=> C6H112-4 + C2H5O2', [4.490000e+14, -0.9, 18490.0])

## Reaction 3624
#reaction('C6H11-3D2O + C2H5O <=> C6H112-4 + C2H5O2', [4.490000e+14, -0.9, 18490.0])

## Reaction 3625
#reaction('C6H11-2D5O + C2H5O <=> C6H112-5 + C2H5O2', [4.490000e+14, -0.9, 18490.0])

## Reaction 3626
#reaction('C6H11-2D6O + C2H5O <=> C6H112-6 + C2H5O2', [4.490000e+14, -0.9, 18490.0])

## Reaction 3627
#reaction('C6H11-3D1O + C2H5O <=> C6H113-1 + C2H5O2', [4.490000e+14, -0.9, 18490.0])

## Reaction 3628
#reaction('C6H11-1D3O <=> C2H3CHO + NC3H7', [2.900000e+12, 0.4, 11080.0])

## Reaction 3629
#reaction('C6H11-1D4O <=> C2H5CHO + C3H5-A', [3.130000e+19, -1.9, 10670.0])

## Reaction 3630
#reaction('C6H11-1D5O <=> CH3CHO + C4H71-4', [2.900000e+12, 0.4, 11080.0])

## Reaction 3631
#reaction('C6H11-1D6O <=> CH2O + C5H91-5', [2.900000e+12, 0.4, 11080.0])

## Reaction 3632
#reaction('C6H11-2D4O <=> SC3H5CHO + C2H5', [3.130000e+19, -1.9, 10670.0])

## Reaction 3633
#reaction('C6H11-2D5O <=> CH3CHO + C4H71-3', [3.130000e+19, -1.9, 10670.0])

## Reaction 3634
#reaction('C6H11-2D6O <=> CH2O + C5H92-5', [2.900000e+12, 0.4, 11080.0])

## Reaction 3635
#reaction('C6H11-3D1O <=> CH2O + C5H91-3', [5.980000e+15, -1.1, 9941.0])

## Reaction 3636
#reaction('C6H12-1 <=> NC3H7 + C3H5-A', [1.000000e+16, 0.0, 71000.0])

## Reaction 3637
#reaction('C6H12-2 <=> C2H5 + C4H71-3', [1.000000e+16, 0.0, 71000.0])

## Reaction 3638
#reaction('C6H12-3 <=> CH3 + C5H91-3', [1.000000e+16, 0.0, 71000.0])

## Reaction 3639
#reaction('C6H13-1 + O2 <=> C6H13O2-1', [6.863000e+16, -1.627, 198.7])

## Reaction 3640
#reaction('C6H13-2 + O2 <=> C6H13O2-2', [3.486000e+14, -0.816, -536.5])

## Reaction 3641
#reaction('C6H13-3 + O2 <=> C6H13O2-3', [3.486000e+14, -0.816, -536.5])

## Reaction 3642
#reaction('C6H13-1 + HO2 <=> C6H13O-1 + OH', [9.000000e+12, 0.0, -1000.0])

## Reaction 3643
#reaction('C6H13-2 + HO2 <=> C6H13O-2 + OH', [9.000000e+12, 0.0, -1000.0])

## Reaction 3644
#reaction('C6H13-3 + HO2 <=> C6H13O-3 + OH', [9.000000e+12, 0.0, -1000.0])

## Reaction 3645
#reaction('C6H13-1 + CH3O2 <=> C6H13O-1 + CH3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 3646
#reaction('C6H13-2 + CH3O2 <=> C6H13O-2 + CH3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 3647
#reaction('C6H13-3 + CH3O2 <=> C6H13O-3 + CH3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 3648
#reaction('C6H13-1 + C6H13O2-1 <=> C6H13O-1 + C6H13O-1', [9.000000e+12, 0.0, -1000.0])

## Reaction 3649
#reaction('C6H13-1 + C6H13O2-2 <=> C6H13O-1 + C6H13O-2', [9.000000e+12, 0.0, -1000.0])

## Reaction 3650
#reaction('C6H13-1 + C6H13O2-3 <=> C6H13O-1 + C6H13O-3', [9.000000e+12, 0.0, -1000.0])

## Reaction 3651
#reaction('C6H13-2 + C6H13O2-1 <=> C6H13O-2 + C6H13O-1', [9.000000e+12, 0.0, -1000.0])

## Reaction 3652
#reaction('C6H13-2 + C6H13O2-2 <=> C6H13O-2 + C6H13O-2', [9.000000e+12, 0.0, -1000.0])

## Reaction 3653
#reaction('C6H13-2 + C6H13O2-3 <=> C6H13O-2 + C6H13O-3', [9.000000e+12, 0.0, -1000.0])

## Reaction 3654
#reaction('C6H13-3 + C6H13O2-1 <=> C6H13O-3 + C6H13O-1', [9.000000e+12, 0.0, -1000.0])

## Reaction 3655
#reaction('C6H13-3 + C6H13O2-2 <=> C6H13O-3 + C6H13O-2', [9.000000e+12, 0.0, -1000.0])

## Reaction 3656
#reaction('C6H13-3 + C6H13O2-3 <=> C6H13O-3 + C6H13O-3', [9.000000e+12, 0.0, -1000.0])

## Reaction 3657
#reaction('C6H13O2-1 <=> C6H12OOH1-2', [4.009000e+08, 1.1, 30100.0])

## Reaction 3658
#reaction('C6H13O2-1 <=> C6H12OOH1-3', [1.360000e+07, 1.3, 18200.0])

## Reaction 3659
#reaction('C6H13O2-1 <=> C6H12OOH1-4', [3.718000e+06, 1.2, 16600.0])

## Reaction 3660
#reaction('C6H13O2-1 <=> C6H12OOH1-5', [2.600000e+06, 1.0, 18200.0])

## Reaction 3661
#reaction('C6H13O2-2 <=> C6H12OOH2-1', [2.110000e+11, 0.2, 34200.0])

## Reaction 3662
#reaction('C6H13O2-2 <=> C6H12OOH2-3', [1.716000e+09, 0.9, 29500.0])

## Reaction 3663
#reaction('C6H13O2-2 <=> C6H12OOH2-4', [7.033000e+10, 0.2, 18500.0])

## Reaction 3664
#reaction('C6H13O2-2 <=> C6H12OOH2-5', [1.859000e+06, 1.2, 16600.0])

## Reaction 3665
#reaction('C6H13O2-2 <=> C6H12OOH2-6', [1.471000e+05, 1.5, 19900.0])

## Reaction 3666
#reaction('C6H13O2-3 <=> C6H12OOH3-1', [1.439000e+07, 1.4, 20800.0])

## Reaction 3667
#reaction('C6H13O2-3 <=> C6H12OOH3-2', [1.716000e+09, 0.9, 29500.0])

## Reaction 3668
#reaction('C6H13O2-3 <=> C6H12OOH3-4', [1.716000e+09, 0.9, 29500.0])

## Reaction 3669
#reaction('C6H13O2-3 <=> C6H12OOH3-5', [7.033000e+10, 0.2, 18500.0])

## Reaction 3670
#reaction('C6H13O2-3 <=> C6H12OOH3-6', [1.233000e+06, 1.5, 20000.0])

## Reaction 3671
#reaction('C6H13O2-1 <=> C6H12-1 + HO2', [3.870000e+07, 1.58, 28500.0])

## Reaction 3672
#reaction('C6H13O2-2 <=> C6H12-1 + HO2', [3.605000e+09, 1.13, 30100.0])

## Reaction 3673
#reaction('C6H13O2-2 <=> C6H12-2 + HO2', [5.140000e+09, 1.01, 29362.0])

## Reaction 3674
#reaction('C6H13O2-3 <=> C6H12-2 + HO2', [5.140000e+09, 1.01, 29362.0])

## Reaction 3675
#reaction('C6H13O2-3 <=> C6H12-3 + HO2', [7.710000e+09, 1.01, 29362.0])

## Reaction 3676
#reaction('C6H13OOH-1 + O2 <=> C6H13O2-1 + HO2', [3.750000e+13, -0.8, 33620.0])

## Reaction 3677
#reaction('C6H13OOH-2 + O2 <=> C6H13O2-2 + HO2', [4.380000e+13, -0.8, 33640.0])

## Reaction 3678
#reaction('C6H13OOH-3 + O2 <=> C6H13O2-3 + HO2', [4.380000e+13, -0.8, 33640.0])

## Reaction 3679
#reaction('C6H13OOH-1 + HO2 <=> C6H13O2-1 + H2O2', [2.400000e+12, 0.0, 10000.0])

## Reaction 3680
#reaction('C6H13OOH-2 + HO2 <=> C6H13O2-2 + H2O2', [2.400000e+12, 0.0, 10000.0])

## Reaction 3681
#reaction('C6H13OOH-3 + HO2 <=> C6H13O2-3 + H2O2', [2.400000e+12, 0.0, 10000.0])

## Reaction 3682
#reaction('C6H13O2-1 + NC6H14 <=> C6H13OOH-1 + C6H13-1', [1.210000e+13, 0.0, 20430.0])

## Reaction 3683
#reaction('C6H13O2-1 + NC6H14 <=> C6H13OOH-1 + C6H13-2', [8.060000e+12, 0.0, 17700.0])

## Reaction 3684
#reaction('C6H13O2-1 + NC6H14 <=> C6H13OOH-1 + C6H13-3', [8.060000e+12, 0.0, 17700.0])

## Reaction 3685
#reaction('C6H13O2-2 + NC6H14 <=> C6H13OOH-2 + C6H13-1', [1.210000e+13, 0.0, 20430.0])

## Reaction 3686
#reaction('C6H13O2-2 + NC6H14 <=> C6H13OOH-2 + C6H13-2', [8.060000e+12, 0.0, 17700.0])

## Reaction 3687
#reaction('C6H13O2-2 + NC6H14 <=> C6H13OOH-2 + C6H13-3', [8.060000e+12, 0.0, 17700.0])

## Reaction 3688
#reaction('C6H13O2-3 + NC6H14 <=> C6H13OOH-3 + C6H13-1', [1.210000e+13, 0.0, 20430.0])

## Reaction 3689
#reaction('C6H13O2-3 + NC6H14 <=> C6H13OOH-3 + C6H13-2', [8.060000e+12, 0.0, 17700.0])

## Reaction 3690
#reaction('C6H13O2-3 + NC6H14 <=> C6H13OOH-3 + C6H13-3', [8.060000e+12, 0.0, 17700.0])

## Reaction 3691
#reaction('C6H13O2-1 + CH3O2 => C6H13O-1 + CH3O + O2', [1.400000e+16, -1.6, 1860.0])

## Reaction 3692
#reaction('C6H13O2-2 + CH3O2 => C6H13O-2 + CH3O + O2', [1.400000e+16, -1.6, 1860.0])

## Reaction 3693
#reaction('C6H13O2-3 + CH3O2 => C6H13O-3 + CH3O + O2', [1.400000e+16, -1.6, 1860.0])

## Reaction 3694
#reaction('C6H13O2-1 + C6H13O2-1 => C6H13O-1 + C6H13O-1 + O2', [1.400000e+16, -1.6, 1860.0])

## Reaction 3695
#reaction('C6H13O2-1 + C6H13O2-2 => C6H13O-1 + C6H13O-2 + O2', [1.400000e+16, -1.6, 1860.0])

## Reaction 3696
#reaction('C6H13O2-1 + C6H13O2-3 => C6H13O-1 + C6H13O-3 + O2', [1.400000e+16, -1.6, 1860.0])

## Reaction 3697
#reaction('C6H13O2-2 + C6H13O2-2 => C6H13O-2 + C6H13O-2 + O2', [1.400000e+16, -1.6, 1860.0])

## Reaction 3698
#reaction('C6H13O2-2 + C6H13O2-3 => C6H13O-2 + C6H13O-3 + O2', [1.400000e+16, -1.6, 1860.0])

## Reaction 3699
#reaction('C6H13O2-3 + C6H13O2-3 => C6H13O-3 + C6H13O-3 + O2', [1.400000e+16, -1.6, 1860.0])

## Reaction 3700
#reaction('C6H13OOH-1 <=> C6H13O-1 + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 3701
#reaction('C6H13OOH-2 <=> C6H13O-2 + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 3702
#reaction('C6H13OOH-3 <=> C6H13O-3 + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 3703
#reaction('CH2O + C5H11-1 <=> C6H13O-1', [6.250000e+10, 0.0, 11900.0])

## Reaction 3704
#reaction('CH3CHO + PC4H9 <=> C6H13O-2', [7.500000e+10, 0.0, 11900.0])

## Reaction 3705
#reaction('C2H5CHO + NC3H7 <=> C6H13O-3', [8.500000e+10, 0.0, 11900.0])

## Reaction 3706
#reaction('C6H12OOH1-2 <=> C6H12O1-2 + OH', [2.660000e+12, 0.0, 13376.0])

## Reaction 3707
#reaction('C6H12OOH1-3 <=> C6H12O1-3 + OH', [4.760000e+11, 0.0, 18292.0])

## Reaction 3708
#reaction('C6H12OOH1-4 <=> C6H12O1-4 + OH', [4.160000e+10, 0.0, 11501.0])

## Reaction 3709
#reaction('C6H12OOH1-5 <=> C6H12O1-5 + OH', [7.800000e+09, 0.0, 11137.0])

## Reaction 3710
#reaction('C6H12OOH2-1 <=> C6H12O1-2 + OH', [2.660000e+12, 0.0, 15034.0])

## Reaction 3711
#reaction('C6H12OOH2-3 <=> C6H12O2-3 + OH', [2.660000e+12, 0.0, 13364.0])

## Reaction 3712
#reaction('C6H12OOH2-4 <=> C6H12O2-4 + OH', [4.760000e+11, 0.0, 17284.0])

## Reaction 3713
#reaction('C6H12OOH2-5 <=> C6H12O2-5 + OH', [4.160000e+10, 0.0, 11490.0])

## Reaction 3714
#reaction('C6H12OOH2-6 <=> C6H12O1-5 + OH', [7.800000e+09, 0.0, 11908.0])

## Reaction 3715
#reaction('C6H12OOH3-1 <=> C6H12O1-3 + OH', [4.760000e+11, 0.0, 19006.0])

## Reaction 3716
#reaction('C6H12OOH3-2 <=> C6H12O2-3 + OH', [2.660000e+12, 0.0, 13364.0])

## Reaction 3717
#reaction('C6H12OOH3-4 <=> C6H12O3-4 + OH', [2.660000e+12, 0.0, 13364.0])

## Reaction 3718
#reaction('C6H12OOH3-5 <=> C6H12O2-4 + OH', [4.760000e+11, 0.0, 17284.0])

## Reaction 3719
#reaction('C6H12OOH3-6 <=> C6H12O1-4 + OH', [4.160000e+10, 0.0, 11573.0])

## Reaction 3720
#reaction('C6H12OOH1-2 <=> C6H12-1 + HO2', [5.630000e+09, 0.97, 15000.0])

## Reaction 3721
#reaction('C6H12OOH2-1 <=> C6H12-1 + HO2', [1.780000e+12, 0.24, 15700.0])

## Reaction 3722
#reaction('C6H12OOH2-3 <=> C6H12-2 + HO2', [2.650000e+11, 0.38, 14800.0])

## Reaction 3723
#reaction('C6H12OOH3-2 <=> C6H12-2 + HO2', [3.500000e+11, 0.37, 14900.0])

## Reaction 3724
#reaction('C6H12OOH3-4 <=> C6H12-3 + HO2', [3.500000e+11, 0.37, 14900.0])

## Reaction 3725
#reaction('C6H12OOH1-3 => CH2O + C5H10-1 + OH', [3.081000e+13, 0.0, 27346.5])

## Reaction 3726
#reaction('C6H12OOH2-4 => CH3CHO + C4H8-1 + OH', [3.081000e+13, 0.0, 24888.0])

## Reaction 3727
#reaction('C6H12OOH3-1 => NC3H7CHO + C2H4 + OH', [3.081000e+13, 0.0, 24608.4])

## Reaction 3728
#reaction('C6H12OOH3-5 => C2H5CHO + C3H6 + OH', [3.081000e+13, 0.0, 25045.7])

## Reaction 3729
#reaction('C6H12OOH1-4 <=> C2H4O2H + C4H8-1', [1.392000e+14, 0.0, 28052.7])

## Reaction 3730
#reaction('C6H12OOH2-5 <=> C3H6OOH2-1 + C3H6', [1.392000e+14, 0.0, 28247.6])

## Reaction 3731
#reaction('C6H12OOH3-6 <=> C4H8OOH2-1 + C2H4', [1.392000e+14, 0.0, 28123.8])

## Reaction 3732
#reaction('C6H12OOH1-2 + O2 <=> C6H12OOH1-2O2', [1.740000e+14, -0.816, -536.5])

## Reaction 3733
#reaction('C6H12OOH1-3 + O2 <=> C6H12OOH1-3O2', [1.740000e+14, -0.816, -536.5])

## Reaction 3734
#reaction('C6H12OOH1-4 + O2 <=> C6H12OOH1-4O2', [1.740000e+14, -0.816, -536.5])

## Reaction 3735
#reaction('C6H12OOH1-5 + O2 <=> C6H12OOH1-5O2', [1.740000e+14, -0.816, -536.5])

## Reaction 3736
#reaction('C6H12OOH2-1 + O2 <=> C6H12OOH2-1O2', [3.430000e+16, -1.627, 198.7])

## Reaction 3737
#reaction('C6H12OOH2-3 + O2 <=> C6H12OOH2-3O2', [1.740000e+14, -0.816, -536.5])

## Reaction 3738
#reaction('C6H12OOH2-4 + O2 <=> C6H12OOH2-4O2', [1.740000e+14, -0.816, -536.5])

## Reaction 3739
#reaction('C6H12OOH2-5 + O2 <=> C6H12OOH2-5O2', [1.740000e+14, -0.816, -536.5])

## Reaction 3740
#reaction('C6H12OOH2-6 + O2 <=> C6H12OOH2-6O2', [3.430000e+16, -1.627, 198.7])

## Reaction 3741
#reaction('C6H12OOH3-1 + O2 <=> C6H12OOH3-1O2', [3.430000e+16, -1.627, 198.7])

## Reaction 3742
#reaction('C6H12OOH3-2 + O2 <=> C6H12OOH3-2O2', [1.740000e+14, -0.816, -536.5])

## Reaction 3743
#reaction('C6H12OOH3-4 + O2 <=> C6H12OOH3-4O2', [1.740000e+14, -0.816, -536.5])

## Reaction 3744
#reaction('C6H12OOH3-5 + O2 <=> C6H12OOH3-5O2', [1.740000e+14, -0.816, -536.5])

## Reaction 3745
#reaction('C6H12OOH3-6 + O2 <=> C6H12OOH3-6O2', [3.430000e+16, -1.627, 198.7])

## Reaction 3746
#reaction('C6H12OOH1-2O2 <=> C6H11-2D1OOH + HO2', [5.140000e+09, 1.01, 29362.0])

## Reaction 3747
#reaction('C6H12OOH1-3O2 <=> C6H11-2D1OOH + HO2', [5.140000e+09, 1.01, 29362.0])

## Reaction 3748
#reaction('C6H12OOH1-3O2 <=> C6H11-3D1OOH + HO2', [7.710000e+09, 1.01, 29362.0])

## Reaction 3749
#reaction('C6H12OOH1-4O2 <=> C6H11-3D1OOH + HO2', [7.710000e+09, 1.01, 29362.0])

## Reaction 3750
#reaction('C6H12OOH1-4O2 <=> C6H11-2D6OOH + HO2', [5.140000e+09, 1.01, 29362.0])

## Reaction 3751
#reaction('C6H12OOH1-5O2 <=> C6H11-2D6OOH + HO2', [5.140000e+09, 1.01, 29362.0])

## Reaction 3752
#reaction('C6H12OOH1-5O2 <=> C6H11-1D6OOH + HO2', [3.605000e+09, 1.13, 30100.0])

## Reaction 3753
#reaction('C6H12OOH2-3O2 <=> C6H11-3D2OOH + HO2', [7.710000e+09, 1.01, 29362.0])

## Reaction 3754
#reaction('C6H12OOH2-4O2 <=> C6H11-3D2OOH + HO2', [7.710000e+09, 1.01, 29362.0])

## Reaction 3755
#reaction('C6H12OOH2-4O2 <=> C6H11-2D5OOH + HO2', [5.140000e+09, 1.01, 29362.0])

## Reaction 3756
#reaction('C6H12OOH2-5O2 <=> C6H11-2D5OOH + HO2', [5.140000e+09, 1.01, 29362.0])

## Reaction 3757
#reaction('C6H12OOH2-5O2 <=> C6H11-1D5OOH + HO2', [3.605000e+09, 1.13, 30100.0])

## Reaction 3758
#reaction('C6H12OOH2-6O2 <=> C6H11-1D5OOH + HO2', [3.870000e+07, 1.58, 28500.0])

## Reaction 3759
#reaction('C6H12OOH3-1O2 <=> C6H11-1D3OOH + HO2', [3.870000e+07, 1.58, 28500.0])

## Reaction 3760
#reaction('C6H12OOH3-2O2 <=> C6H11-1D3OOH + HO2', [3.605000e+09, 1.13, 30100.0])

## Reaction 3761
#reaction('C6H12OOH3-4O2 <=> C6H11-2D4OOH + HO2', [5.140000e+09, 1.01, 29362.0])

## Reaction 3762
#reaction('C6H12OOH3-5O2 <=> C6H11-2D4OOH + HO2', [5.140000e+09, 1.01, 29362.0])

## Reaction 3763
#reaction('C6H12OOH3-5O2 <=> C6H11-1D4OOH + HO2', [3.605000e+09, 1.13, 30100.0])

## Reaction 3764
#reaction('C6H12OOH3-6O2 <=> C6H11-1D4OOH + HO2', [3.870000e+07, 1.58, 28500.0])

## Reaction 3765
#reaction('C6H12OOH1-2O2 <=> C6H11Q12-3', [1.716000e+09, 0.9, 29500.0])

## Reaction 3766
#reaction('C6H12OOH1-2O2 <=> C6H11Q12-4', [7.033000e+10, 0.2, 18500.0])

## Reaction 3767
#reaction('C6H12OOH1-2O2 <=> C6H11Q12-5', [1.859000e+06, 1.2, 16600.0])

## Reaction 3768
#reaction('C6H12OOH1-2O2 <=> C6H11Q12-6', [1.470000e+05, 1.5, 19900.0])

## Reaction 3769
#reaction('C6H12OOH1-3O2 <=> C6H11Q13-2', [1.716000e+09, 0.9, 29500.0])

## Reaction 3770
#reaction('C6H12OOH1-3O2 <=> C6H11Q13-4', [1.716000e+09, 0.9, 29500.0])

## Reaction 3771
#reaction('C6H12OOH1-3O2 <=> C6H11Q13-5', [7.033000e+10, 0.2, 18500.0])

## Reaction 3772
#reaction('C6H12OOH1-3O2 <=> C6H11Q13-6', [6.166000e+05, 1.5, 20000.0])

## Reaction 3773
#reaction('C6H12OOH1-4O2 <=> C6H11Q14-2', [7.033000e+10, 0.2, 18500.0])

## Reaction 3774
#reaction('C6H12OOH1-4O2 <=> C6H11Q14-3', [1.716000e+09, 0.9, 29500.0])

## Reaction 3775
#reaction('C6H12OOH1-4O2 <=> C6H11Q14-5', [1.716000e+09, 0.9, 29500.0])

## Reaction 3776
#reaction('C6H12OOH1-4O2 <=> C6H11Q14-6', [1.439000e+07, 1.4, 20800.0])

## Reaction 3777
#reaction('C6H12OOH1-5O2 <=> C6H11Q15-2', [1.859000e+06, 1.2, 16600.0])

## Reaction 3778
#reaction('C6H12OOH1-5O2 <=> C6H11Q15-3', [7.033000e+10, 0.2, 18500.0])

## Reaction 3779
#reaction('C6H12OOH1-5O2 <=> C6H11Q15-4', [1.716000e+09, 0.9, 29500.0])

## Reaction 3780
#reaction('C6H12OOH1-5O2 <=> C6H11Q15-6', [2.110000e+11, 0.2, 34200.0])

## Reaction 3781
#reaction('C6H12OOH2-1O2 <=> C6H11Q12-3', [1.360000e+07, 1.3, 18200.0])

## Reaction 3782
#reaction('C6H12OOH2-1O2 <=> C6H11Q12-4', [3.718000e+06, 1.2, 16600.0])

## Reaction 3783
#reaction('C6H12OOH2-1O2 <=> C6H11Q12-5', [2.600000e+06, 1.0, 18200.0])

## Reaction 3784
#reaction('C6H12OOH2-3O2 <=> C6H11Q23-1', [1.439000e+07, 1.4, 20800.0])

## Reaction 3785
#reaction('C6H12OOH2-3O2 <=> C6H11Q23-4', [1.716000e+09, 0.9, 29500.0])

## Reaction 3786
#reaction('C6H12OOH2-3O2 <=> C6H11Q23-5', [7.033000e+10, 0.2, 18500.0])

## Reaction 3787
#reaction('C6H12OOH2-3O2 <=> C6H11Q23-6', [6.166000e+05, 1.5, 20000.0])

## Reaction 3788
#reaction('C6H12OOH2-4O2 <=> C6H11Q24-1', [6.166000e+05, 1.5, 20000.0])

## Reaction 3789
#reaction('C6H12OOH2-4O2 <=> C6H11Q24-3', [1.716000e+09, 0.9, 29500.0])

## Reaction 3790
#reaction('C6H12OOH2-4O2 <=> C6H11Q24-5', [1.716000e+09, 0.9, 29500.0])

## Reaction 3791
#reaction('C6H12OOH2-4O2 <=> C6H11Q24-6', [1.439000e+07, 1.4, 20800.0])

## Reaction 3792
#reaction('C6H12OOH2-5O2 <=> C6H11Q25-1', [1.470000e+05, 1.5, 19900.0],
        #options='duplicate')

## Reaction 3793
#reaction('C6H12OOH2-5O2 <=> C6H11Q25-3', [7.033000e+10, 0.2, 18500.0],
        #options='duplicate')

## Reaction 3794
#reaction('C6H12OOH2-5O2 <=> C6H11Q25-3', [1.716000e+09, 0.9, 29500.0],
        #options='duplicate')

## Reaction 3795
#reaction('C6H12OOH2-5O2 <=> C6H11Q25-1', [2.110000e+11, 0.2, 34200.0],
        #options='duplicate')

## Reaction 3796
#reaction('C6H12OOH2-6O2 <=> C6H11Q15-4', [3.718000e+06, 1.2, 16600.0])

## Reaction 3797
#reaction('C6H12OOH2-6O2 <=> C6H11Q15-3', [1.360000e+07, 1.3, 18200.0])

## Reaction 3798
#reaction('C6H12OOH2-6O2 <=> C6H11Q15-2', [4.009000e+08, 1.1, 30100.0])

## Reaction 3799
#reaction('C6H12OOH3-1O2 <=> C6H11Q13-2', [4.009000e+08, 1.1, 30100.0])

## Reaction 3800
#reaction('C6H12OOH3-1O2 <=> C6H11Q13-4', [3.718000e+06, 1.2, 16600.0])

## Reaction 3801
#reaction('C6H12OOH3-1O2 <=> C6H11Q13-5', [2.600000e+06, 1.0, 18200.0])

## Reaction 3802
#reaction('C6H12OOH3-2O2 <=> C6H11Q23-1', [2.110000e+11, 0.2, 34200.0])

## Reaction 3803
#reaction('C6H12OOH3-2O2 <=> C6H11Q23-4', [7.033000e+10, 0.2, 18500.0])

## Reaction 3804
#reaction('C6H12OOH3-2O2 <=> C6H11Q23-5', [1.859000e+06, 1.2, 16600.0])

## Reaction 3805
#reaction('C6H12OOH3-2O2 <=> C6H11Q23-6', [1.470000e+05, 1.5, 19900.0])

## Reaction 3806
#reaction('C6H12OOH3-4O2 <=> C6H11Q34-1', [6.166000e+05, 1.5, 20000.0],
        #options='duplicate')

## Reaction 3807
#reaction('C6H12OOH3-4O2 <=> C6H11Q34-2', [7.033000e+10, 0.2, 18500.0],
        #options='duplicate')

## Reaction 3808
#reaction('C6H12OOH3-4O2 <=> C6H11Q34-2', [1.716000e+09, 0.9, 29500.0],
        #options='duplicate')

## Reaction 3809
#reaction('C6H12OOH3-4O2 <=> C6H11Q34-1', [1.439000e+07, 1.4, 20800.0],
        #options='duplicate')

## Reaction 3810
#reaction('C6H12OOH3-5O2 <=> C6H11Q24-6', [1.470000e+05, 1.5, 19900.0])

## Reaction 3811
#reaction('C6H12OOH3-5O2 <=> C6H11Q24-5', [1.859000e+06, 1.2, 16600.0])

## Reaction 3812
#reaction('C6H12OOH3-5O2 <=> C6H11Q24-3', [1.716000e+09, 0.9, 29500.0])

## Reaction 3813
#reaction('C6H12OOH3-5O2 <=> C6H11Q24-1', [2.110000e+11, 0.2, 34200.0])

## Reaction 3814
#reaction('C6H12OOH3-6O2 <=> C6H11Q14-5', [2.600000e+06, 1.0, 18200.0])

## Reaction 3815
#reaction('C6H12OOH3-6O2 <=> C6H11Q14-3', [1.360000e+07, 1.3, 18200.0])

## Reaction 3816
#reaction('C6H12OOH3-6O2 <=> C6H11Q14-2', [4.009000e+08, 1.1, 30100.0])

## Reaction 3817
#reaction('C6H12OOH1-2O2 <=> C6KET12 + OH', [2.440000e+07, 1.6, 27900.0])

## Reaction 3818
#reaction('C6H12OOH1-3O2 <=> C6KET13 + OH', [1.100000e+04, 2.4, 19900.0])

## Reaction 3819
#reaction('C6H12OOH1-4O2 <=> C6KET14 + OH', [1.260000e+03, 2.2, 17400.0])

## Reaction 3820
#reaction('C6H12OOH1-5O2 <=> C6KET15 + OH', [1.110000e+03, 1.8, 16600.0])

## Reaction 3821
#reaction('C6H12OOH2-1O2 <=> C6KET21 + OH', [2.760000e+08, 1.2, 25700.0])

## Reaction 3822
#reaction('C6H12OOH2-3O2 <=> C6KET23 + OH', [1.750000e+08, 1.7, 26000.0])

## Reaction 3823
#reaction('C6H12OOH2-4O2 <=> C6KET24 + OH', [1.750000e+02, 3.1, 17500.0])

## Reaction 3824
#reaction('C6H12OOH2-5O2 <=> C6KET25 + OH', [2.540000e+02, 2.6, 16200.0])

## Reaction 3825
#reaction('C6H12OOH2-6O2 <=> C6KET26 + OH', [2.000000e+03, 1.9, 14900.0])

## Reaction 3826
#reaction('C6H12OOH3-1O2 <=> C6KET31 + OH', [5.790000e+01, 2.9, 17000.0])

## Reaction 3827
#reaction('C6H12OOH3-2O2 <=> C6KET32 + OH', [1.750000e+08, 1.7, 26000.0])

## Reaction 3828
#reaction('C6H12OOH3-4O2 <=> C6KET34 + OH', [1.750000e+08, 1.7, 26000.0])

## Reaction 3829
#reaction('C6H12OOH3-5O2 <=> C6KET35 + OH', [1.750000e+02, 3.1, 17500.0])

## Reaction 3830
#reaction('C6H12OOH3-6O2 <=> C6KET36 + OH', [3.770000e+02, 2.2, 15300.0])

## Reaction 3831
#reaction('C6H11Q12-3 => PC4H8CHO-1 + CH2O + 2 OH', [3.081000e+13, 0.0, 33871.4])

## Reaction 3832
#reaction('C6H11Q12-4 => C4H8-1 + HO2CH2CHO + OH', [3.081000e+13, 0.0, 2316.3])

## Reaction 3833
#reaction('C6H11Q12-5 <=> C3H51-2,3OOH + C3H6', [1.392000e+14, 0.0, 25866.6])

## Reaction 3834
#reaction('C6H11Q13-5 => C3KET13 + C3H6 + OH', [3.081000e+13, 0.0, 24186.1])

## Reaction 3835
#reaction('C6H11Q14-6 => NC4KET14 + C2H4 + OH', [3.081000e+13, 0.0, 24608.4])

## Reaction 3836
#reaction('C6H11Q15-3 => C4H71-4OOH + CH3CHO + OH', [3.081000e+13, 0.0, 24638.9])

## Reaction 3837
#reaction('C6H11Q15-3 => C5H91-4OOH + CH2O + OH', [3.081000e+13, 0.0, 27282.9])

## Reaction 3838
#reaction('C6H11Q23-5 => C3KET12 + C3H6 + OH', [3.081000e+13, 0.0, 23942.2])

## Reaction 3839
#reaction('C6H11Q24-6 => NC4KET13 + C2H4 + OH', [3.081000e+13, 0.0, 24608.4])

## Reaction 3840
#reaction('C6H11Q34-1 => NC4KET12 + C2H4 + OH', [3.081000e+13, 0.0, 26362.9])

## Reaction 3841
#reaction('C6KET12 <=> C6KET12O + OH', [1.050000e+16, 0.0, 42300.0])

## Reaction 3842
#reaction('C6KET13 <=> C6KET13O + OH', [1.050000e+16, 0.0, 42300.0])

## Reaction 3843
#reaction('C6KET14 <=> C6KET14O + OH', [1.050000e+16, 0.0, 42300.0])

## Reaction 3844
#reaction('C6KET15 <=> C6KET15O + OH', [1.050000e+16, 0.0, 42300.0])

## Reaction 3845
#reaction('C6KET21 <=> C6KET21O + OH', [1.500000e+16, 0.0, 42700.0])

## Reaction 3846
#reaction('C6KET23 <=> C6KET23O + OH', [1.050000e+16, 0.0, 42300.0])

## Reaction 3847
#reaction('C6KET24 <=> C6KET24O + OH', [1.050000e+16, 0.0, 42300.0])

## Reaction 3848
#reaction('C6KET25 <=> C6KET25O + OH', [1.050000e+16, 0.0, 42300.0])

## Reaction 3849
#reaction('C6KET26 <=> C6KET26O + OH', [1.500000e+16, 0.0, 42700.0])

## Reaction 3850
#reaction('C6KET31 <=> C6KET31O + OH', [1.500000e+16, 0.0, 42700.0])

## Reaction 3851
#reaction('C6KET32 <=> C6KET32O + OH', [1.050000e+16, 0.0, 42300.0])

## Reaction 3852
#reaction('C6KET34 <=> C6KET34O + OH', [1.050000e+16, 0.0, 42300.0])

## Reaction 3853
#reaction('C6KET35 <=> C6KET35O + OH', [1.050000e+16, 0.0, 42300.0])

## Reaction 3854
#reaction('C6KET36 <=> C6KET36O + OH', [1.500000e+16, 0.0, 42700.0])

## Reaction 3855
#reaction('C6KET12O <=> PC4H9CHO + HCO', [2.900000e+12, 0.4, 11080.0])

## Reaction 3856
#reaction('C6KET13O <=> NC3H7CHO + CH2CHO', [2.900000e+12, 0.4, 11080.0])

## Reaction 3857
#reaction('C6KET14O <=> C2H5CHO + CH2CH2CHO', [2.900000e+12, 0.4, 11080.0])

## Reaction 3858
#reaction('C6KET15O <=> CH3CHO + C3H6CHO-1', [2.900000e+12, 0.4, 11080.0])

## Reaction 3859
#reaction('C6KET21O <=> CH2O + PC4H9CO', [2.900000e+12, 0.4, 11080.0])

## Reaction 3860
#reaction('C6KET23O <=> NC3H7CHO + CH3CO', [2.900000e+12, 0.4, 11080.0])

## Reaction 3861
#reaction('C6KET24O <=> C2H5CHO + CH3COCH2', [2.900000e+12, 0.4, 11080.0])

## Reaction 3862
#reaction('C6KET25O <=> CH3CHO + CH2CH2COCH3', [2.900000e+12, 0.4, 11080.0])

## Reaction 3863
#reaction('C6KET26O <=> CH2O + C5Y2-5J', [2.900000e+12, 0.4, 11080.0])

## Reaction 3864
#reaction('C6KET31O <=> CH2O + C5Y2-1J', [2.900000e+12, 0.4, 11080.0])

## Reaction 3865
#reaction('C6KET32O <=> CH3CHO + NC3H7CO', [2.900000e+12, 0.4, 11080.0])

## Reaction 3866
#reaction('C6KET34O <=> C2H5CHO + C2H5CO', [2.900000e+12, 0.4, 11080.0])

## Reaction 3867
#reaction('C6KET35O <=> CH3CHO + C2H5COCH2', [2.900000e+12, 0.4, 11080.0])

## Reaction 3868
#reaction('C6KET36O <=> CH2O + C5Y3-1J', [2.900000e+12, 0.4, 11080.0])

## Reaction 3869
#reaction('C6H11Q12-3 <=> C6H11-2D1OOH + HO2', [2.650000e+11, 0.38, 14800.0])

## Reaction 3870
#reaction('C6H11Q13-2 <=> C6H11-1D3OOH + HO2', [5.630000e+09, 0.97, 15000.0])

## Reaction 3871
#reaction('C6H11Q13-2 <=> C6H11-2D1OOH + HO2', [2.650000e+11, 0.38, 14800.0])

## Reaction 3872
#reaction('C6H11Q13-4 <=> C6H11-3D1OOH + HO2', [3.500000e+11, 0.37, 14900.0])

## Reaction 3873
#reaction('C6H11Q14-2 <=> C6H11-1D4OOH + HO2', [5.630000e+09, 0.97, 15000.0])

## Reaction 3874
#reaction('C6H11Q14-3 <=> C6H11-3D1OOH + HO2', [3.500000e+11, 0.37, 14900.0])

## Reaction 3875
#reaction('C6H11Q14-5 <=> C6H11-2D6OOH + HO2', [3.500000e+11, 0.37, 14900.0])

## Reaction 3876
#reaction('C6H11Q15-2 <=> C6H11-1D5OOH + HO2', [5.630000e+09, 0.97, 15000.0])

## Reaction 3877
#reaction('C6H11Q15-4 <=> C6H11-2D6OOH + HO2', [2.650000e+11, 0.38, 14800.0])

## Reaction 3878
#reaction('C6H11Q15-6 <=> C6H11-1D6OOH + HO2', [1.780000e+12, 0.24, 15700.0])

## Reaction 3879
#reaction('C6H11Q23-1 <=> C6H11-1D3OOH + HO2', [5.630000e+09, 0.97, 15000.0])

## Reaction 3880
#reaction('C6H11Q23-4 <=> C6H11-3D2OOH + HO2', [3.500000e+11, 0.37, 14900.0])

## Reaction 3881
#reaction('C6H11Q24-1 <=> C6H11-1D4OOH + HO2', [1.780000e+12, 0.24, 15700.0])

## Reaction 3882
#reaction('C6H11Q24-3 <=> C6H11-2D4OOH + HO2', [2.650000e+11, 0.38, 14800.0])

## Reaction 3883
#reaction('C6H11Q24-3 <=> C6H11-3D2OOH + HO2', [3.500000e+11, 0.37, 14900.0])

## Reaction 3884
#reaction('C6H11Q24-5 <=> C6H11-2D5OOH + HO2', [2.650000e+11, 0.38, 14800.0])

## Reaction 3885
#reaction('C6H11Q25-1 <=> C6H11-1D5OOH + HO2', [5.630000e+09, 0.97, 15000.0])

## Reaction 3886
#reaction('C6H11Q25-3 <=> C6H11-2D5OOH + HO2', [2.650000e+11, 0.38, 14800.0])

## Reaction 3887
#reaction('C6H11Q34-2 <=> C6H11-2D4OOH + HO2', [3.500000e+11, 0.37, 14900.0])

## Reaction 3888
#reaction('C6H11-1D3OOH => C2H3CHO + NC3H7 + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 3889
#reaction('C6H11-1D4OOH => AC3H5CHO + C2H5 + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 3890
#reaction('C6H11-1D4OOH => C3H5-A + C2H5CHO + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 3891
#reaction('C6H11-1D5OOH => C4H71-4 + CH3CHO + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 3892
#reaction('C6H11-1D6OOH => C5H91-5 + CH2O + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 3893
#reaction('C6H11-2D1OOH => C5H91-1 + CH2O + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 3894
#reaction('C6H11-2D4OOH => SC3H5CHO + C2H5 + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 3895
#reaction('C6H11-2D4OOH => C3H5-S + C2H5CHO + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 3896
#reaction('C6H11-2D6OOH => C5H92-5 + CH2O + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 3897
#reaction('C6H11-3D1OOH => C5H91-3 + CH2O + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 3898
#reaction('C6H11-3D2OOH => CH3CHO + C4H71-1 + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 3899
#reaction('C6H11Q12-3 <=> C6H11O13-2OOH + OH', [3.170000e+11, 0.0, 18003.0])

## Reaction 3900
#reaction('C6H11Q12-3 <=> C6H11O23-1OOH + OH', [1.780000e+12, 0.0, 13364.0])

## Reaction 3901
#reaction('C6H11Q12-4 <=> C6H11O14-2OOH + OH', [2.770000e+10, 0.0, 11501.0])

## Reaction 3902
#reaction('C6H11Q12-4 <=> C6H11O24-1OOH + OH', [3.170000e+11, 0.0, 17284.0])

## Reaction 3903
#reaction('C6H11Q12-5 <=> C6H11O15-2OOH + OH', [5.200000e+09, 0.0, 11137.0])

## Reaction 3904
#reaction('C6H11Q12-5 <=> C6H11O25-1OOH + OH', [2.770000e+10, 0.0, 11490.0])

## Reaction 3905
#reaction('C6H11Q12-6 <=> C6H11O15-6OOH + OH', [5.200000e+09, 0.0, 11908.0])

## Reaction 3906
#reaction('C6H11Q13-2 <=> C6H11O12-3OOH + OH', [1.780000e+12, 0.0, 13376.0])

## Reaction 3907
#reaction('C6H11Q13-2 <=> C6H11O23-1OOH + OH', [1.780000e+12, 0.0, 13364.0])

## Reaction 3908
#reaction('C6H11Q13-4 <=> C6H11O14-3OOH + OH', [2.770000e+10, 0.0, 11068.0])

## Reaction 3909
#reaction('C6H11Q13-4 <=> C6H11O34-1OOH + OH', [1.780000e+12, 0.0, 13364.0])

## Reaction 3910
#reaction('C6H11Q13-5 <=> C6H11O15-3OOH + OH', [5.200000e+09, 0.0, 11137.0])

## Reaction 3911
#reaction('C6H11Q13-5 <=> C6H11O24-6OOH + OH', [3.170000e+11, 0.0, 17284.0])

## Reaction 3912
#reaction('C6H11Q13-6 <=> C6H11O14-6OOH + OH', [2.770000e+10, 0.0, 11573.0])

## Reaction 3913
#reaction('C6H11Q14-2 <=> C6H11O12-4OOH + OH', [1.780000e+12, 0.0, 13376.0])

## Reaction 3914
#reaction('C6H11Q14-2 <=> C6H11O24-1OOH + OH', [3.170000e+11, 0.0, 16995.0])

## Reaction 3915
#reaction('C6H11Q14-3 <=> C6H11O13-4OOH + OH', [3.170000e+11, 0.0, 18003.0])

## Reaction 3916
#reaction('C6H11Q14-3 <=> C6H11O34-1OOH + OH', [1.780000e+12, 0.0, 13364.0])

## Reaction 3917
#reaction('C6H11Q14-5 <=> C6H11O15-4OOH + OH', [5.200000e+09, 0.0, 10825.0])

## Reaction 3918
#reaction('C6H11Q14-5 <=> C6H11O23-6OOH + OH', [1.780000e+12, 0.0, 13364.0])

## Reaction 3919
#reaction('C6H11Q14-6 <=> C6H11O13-6OOH + OH', [3.170000e+11, 0.0, 19006.0])

## Reaction 3920
#reaction('C6H11Q15-2 <=> C6H11O12-5OOH + OH', [1.780000e+12, 0.0, 13376.0])

## Reaction 3921
#reaction('C6H11Q15-2 <=> C6H11O25-1OOH + OH', [2.770000e+10, 0.0, 11056.0])

## Reaction 3922
#reaction('C6H11Q15-3 <=> C6H11O13-5OOH + OH', [3.170000e+11, 0.0, 18292.0])

## Reaction 3923
#reaction('C6H11Q15-3 <=> C6H11O24-6OOH + OH', [3.170000e+11, 0.0, 17284.0])

## Reaction 3924
#reaction('C6H11Q15-4 <=> C6H11O14-5OOH + OH', [2.770000e+10, 0.0, 11068.0])

## Reaction 3925
#reaction('C6H11Q15-4 <=> C6H11O23-6OOH + OH', [1.780000e+12, 0.0, 13364.0])

## Reaction 3926
#reaction('C6H11Q15-6 <=> C6H11O12-6OOH + OH', [1.780000e+12, 0.0, 15034.0])

## Reaction 3927
#reaction('C6H11Q23-1 <=> C6H11O12-3OOH + OH', [1.780000e+12, 0.0, 15034.0])

## Reaction 3928
#reaction('C6H11Q23-1 <=> C6H11O13-2OOH + OH', [3.170000e+11, 0.0, 19044.0])

## Reaction 3929
#reaction('C6H11Q23-4 <=> C6H11O24-3OOH + OH', [3.170000e+11, 0.0, 16995.0])

## Reaction 3930
#reaction('C6H11Q23-4 <=> C6H11O34-2OOH + OH', [1.780000e+12, 0.0, 13364.0])

## Reaction 3931
#reaction('C6H11Q23-5 <=> C6H11O25-3OOH + OH', [2.770000e+10, 0.0, 11490.0])

## Reaction 3932
#reaction('C6H11Q23-5 <=> C6H11O24-5OOH + OH', [3.170000e+11, 0.0, 17284.0])

## Reaction 3933
#reaction('C6H11Q23-6 <=> C6H11O15-4OOH + OH', [5.200000e+09, 0.0, 11908.0])

## Reaction 3934
#reaction('C6H11Q23-6 <=> C6H11O14-5OOH + OH', [2.770000e+10, 0.0, 11573.0])

## Reaction 3935
#reaction('C6H11Q24-1 <=> C6H11O12-4OOH + OH', [1.780000e+12, 0.0, 15034.0])

## Reaction 3936
#reaction('C6H11Q24-1 <=> C6H11O14-2OOH + OH', [2.770000e+10, 0.0, 12630.0])

## Reaction 3937
#reaction('C6H11Q24-3 <=> C6H11O23-4OOH + OH', [1.780000e+12, 0.0, 13364.0])

## Reaction 3938
#reaction('C6H11Q24-3 <=> C6H11O34-2OOH + OH', [1.780000e+12, 0.0, 13364.0])

## Reaction 3939
#reaction('C6H11Q24-5 <=> C6H11O25-3OOH + OH', [2.770000e+10, 0.0, 11056.0])

## Reaction 3940
#reaction('C6H11Q24-5 <=> C6H11O23-5OOH + OH', [1.780000e+12, 0.0, 13364.0])

## Reaction 3941
#reaction('C6H11Q24-6 <=> C6H11O15-3OOH + OH', [5.200000e+09, 0.0, 11908.0])

## Reaction 3942
#reaction('C6H11Q24-6 <=> C6H11O13-5OOH + OH', [3.170000e+11, 0.0, 19006.0])

## Reaction 3943
#reaction('C6H11Q25-1 <=> C6H11O12-5OOH + OH', [1.780000e+12, 0.0, 15034.0])

## Reaction 3944
#reaction('C6H11Q25-1 <=> C6H11O15-2OOH + OH', [5.200000e+09, 0.0, 11949.0])

## Reaction 3945
#reaction('C6H11Q25-3 <=> C6H11O23-5OOH + OH', [1.780000e+12, 0.0, 13364.0])

## Reaction 3946
#reaction('C6H11Q25-3 <=> C6H11O24-5OOH + OH', [3.170000e+11, 0.0, 16995.0])

## Reaction 3947
#reaction('C6H11Q34-1 <=> C6H11O13-4OOH + OH', [3.170000e+11, 0.0, 19006.0])

## Reaction 3948
#reaction('C6H11Q34-1 <=> C6H11O14-3OOH + OH', [2.770000e+10, 0.0, 11573.0])

## Reaction 3949
#reaction('C6H11Q34-2 <=> C6H11O23-4OOH + OH', [1.780000e+12, 0.0, 13364.0])

## Reaction 3950
#reaction('C6H11Q34-2 <=> C6H11O24-3OOH + OH', [3.170000e+11, 0.0, 16995.0])

## Reaction 3951
#reaction('C6H11O12-3OOH => CH2CHO + NC3H7CHO + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 3952
#reaction('C6H11O12-4OOH => C3H5O + C2H5CHO + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 3953
#reaction('C6H11O12-5OOH => C4H7O12-4 + CH3CHO + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 3954
#reaction('C6H11O12-6OOH => C5H9O12-5 + CH2O + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 3955
#reaction('C6H11O13-2OOH => CH2CHO + NC3H7CHO + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 3956
#reaction('C6H11O13-4OOH => CH2CH2CHO + C2H5CHO + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 3957
#reaction('C6H11O13-5OOH => C4H7O13-4 + CH3CHO + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 3958
#reaction('C6H11O13-6OOH => C5H9O13-5 + CH2O + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 3959
#reaction('C6H11O14-2OOH => CH2O + PC4H8CHO-2 + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 3960
#reaction('C6H11O14-3OOH => CH2CH2CHO + C2H5CHO + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 3961
#reaction('C6H11O14-5OOH => C3H6CHO-1 + CH3CHO + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 3962
#reaction('C6H11O14-6OOH => C5H9O14-5 + CH2O + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 3963
#reaction('C6H11O15-2OOH => CH2O + PC4H8CHO-3 + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 3964
#reaction('C6H11O15-3OOH => C2H4 + C3H6CHO-3O + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 3965
#reaction('C6H11O15-4OOH => C3H6CHO-1 + CH3CHO + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 3966
#reaction('C6H11O15-6OOH => PC4H8CHO-4 + CH2O + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 3967
#reaction('C6H11O23-1OOH => CH2O + PC4H8CHO-1 + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 3968
#reaction('C6H11O23-4OOH => CH3CHCHO + C2H5CHO + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 3969
#reaction('C6H11O23-5OOH => C4H7O23-1 + CH3CHO + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 3970
#reaction('C6H11O23-6OOH => C5H9O23-5 + CH2O + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 3971
#reaction('C6H11O24-1OOH => CH2O + PC4H8CHO-2 + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 3972
#reaction('C6H11O24-3OOH => CH3CHCHO + C2H5CHO + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 3973
#reaction('C6H11O24-5OOH => C3H6CHO-2 + CH3CHO + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 3974
#reaction('C6H11O24-6OOH => C5H9O24-1 + CH2O + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 3975
#reaction('C6H11O25-1OOH => CH2O + PC4H8CHO-3 + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 3976
#reaction('C6H11O25-3OOH => CH3CHO + C3H6CHO-2 + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 3977
#reaction('C6H11O34-1OOH => CH2O + C5H9O23-1 + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 3978
#reaction('C6H11O34-2OOH => CH3CHO + C3H6CHO-3 + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 3979
#reaction('C6H12O1-2 + OH => CH2CO + PC4H9 + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 3980
#reaction('C6H12O1-2 + OH => C2H3CHO + NC3H7 + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 3981
#reaction('C6H12O1-3 + OH => C2H4 + NC3H7CO + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 3982
#reaction('C6H12O1-3 + OH => HCO + C5H10-1 + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 3983
#reaction('C6H12O1-4 + OH => C2H4 + C2H5COCH2 + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 3984
#reaction('C6H12O1-4 + OH => CH2CHO + C4H8-1 + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 3985
#reaction('C6H12O1-5 + OH => C2H4 + CH2CH2COCH3 + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 3986
#reaction('C6H12O1-5 + OH => CH2CH2CHO + C3H6 + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 3987
#reaction('C6H12O2-3 + OH => C2H3COCH3 + C2H5 + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 3988
#reaction('C6H12O2-3 + OH => CH3CHCO + NC3H7 + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 3989
#reaction('C6H12O2-4 + OH => C4H8-1 + CH3CO + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 3990
#reaction('C6H12O2-4 + OH => C2H5CO + C3H6 + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 3991
#reaction('C6H12O2-5 + OH => C3H6 + CH3COCH2 + H2O', [5.000000e+12, 0.0, 0.0])

## Reaction 3992
#reaction('C6H12O3-4 + OH => C2H5 + C2H5CHCO + H2O', [5.000000e+12, 0.0, 0.0])

## Reaction 3993
#reaction('C6H12O1-2 + HO2 => CH2CO + PC4H9 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 3994
#reaction('C6H12O1-2 + HO2 => C2H3CHO + NC3H7 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 3995
#reaction('C6H12O1-3 + HO2 => C2H4 + NC3H7CO + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 3996
#reaction('C6H12O1-3 + HO2 => HCO + C5H10-1 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 3997
#reaction('C6H12O1-4 + HO2 => C2H4 + C2H5COCH2 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 3998
#reaction('C6H12O1-4 + HO2 => CH2CHO + C4H8-1 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 3999
#reaction('C6H12O1-5 + HO2 => C2H4 + CH2CH2COCH3 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 4000
#reaction('C6H12O1-5 + HO2 => CH2CH2CHO + C3H6 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 4001
#reaction('C6H12O2-3 + HO2 => C2H3COCH3 + C2H5 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 4002
#reaction('C6H12O2-3 + HO2 => CH3CHCO + NC3H7 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 4003
#reaction('C6H12O2-4 + HO2 => C4H8-1 + CH3CO + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 4004
#reaction('C6H12O2-4 + HO2 => C2H5CO + C3H6 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 4005
#reaction('C6H12O2-5 + HO2 => C3H6 + CH3COCH2 + H2O2', [1.000000e+13, 0.0, 17700.0])

## Reaction 4006
#reaction('C6H12O3-4 + HO2 => C2H5 + C2H5CHCO + H2O2', [1.000000e+13, 0.0, 17700.0])

## Reaction 4007
#reaction('C2H5 + C2H3CHO <=> PC4H8CHO-1', [1.000000e+11, 0.0, 7800.0])

## Reaction 4008
#reaction('CH3 + AC3H5CHO <=> PC4H8CHO-2', [1.000000e+11, 0.0, 7800.0])

## Reaction 4009
#reaction('C4H8-1 + HCO <=> PC4H8CHO-2', [1.000000e+11, 0.0, 6000.0])

## Reaction 4010
#reaction('C3H6 + CH2CHO <=> PC4H8CHO-3', [1.000000e+11, 0.0, 7800.0])

## Reaction 4011
#reaction('C2H4 + CH2CH2CHO <=> PC4H8CHO-4', [1.000000e+11, 0.0, 7800.0])

## Reaction 4012
#reaction('CH2CHO + C2H4 <=> C4H7O12-4', [1.000000e+11, 0.0, 7800.0])

## Reaction 4013
#reaction('C2H4 + CH2CHO <=> C4H7O13-4', [1.000000e+11, 0.0, 7800.0])

## Reaction 4014
#reaction('CH3 + C2H3CHO <=> C4H7O23-1', [1.000000e+11, 0.0, 7800.0])

## Reaction 4015
#reaction('C3H5O + C2H4 <=> C5H9O12-5', [1.000000e+11, 0.0, 7800.0])

## Reaction 4016
#reaction('CH2CH2CHO + C2H4 <=> C5H9O13-5', [1.000000e+11, 0.0, 7800.0])

## Reaction 4017
#reaction('CH2O + C4H71-4 <=> C5H9O14-5', [1.000000e+11, 0.0, 7800.0])

## Reaction 4018
#reaction('C2H3CHO + C2H5 <=> C5H9O23-1', [1.000000e+11, 0.0, 7800.0])

## Reaction 4019
#reaction('CH3CHCHO + C2H4 <=> C5H9O23-5', [1.000000e+11, 0.0, 7800.0])

## Reaction 4020
#reaction('CH3CHO + C3H5-A <=> C5H9O24-1', [1.000000e+11, 0.0, 7800.0])

## Reaction 4021
#reaction('C3H5-A + HCO <=> AC3H5CHO', [1.810000e+13, 0.0, 0.0])

## Reaction 4022
#reaction('PC4H9CO <=> NC3H7 + CH2CO', [1.000000e+11, 0.0, 9600.0])

## Reaction 4023
#reaction('CH2CO + NC3H7 <=> C5Y2-1J', [1.000000e+11, 0.0, 7800.0])

## Reaction 4024
#reaction('CH3COCH2 + C2H4 <=> C5Y2-5J', [1.000000e+11, 0.0, 7800.0])

## Reaction 4025
#reaction('C2H5CO + C2H4 <=> C5Y3-1J', [1.000000e+11, 0.0, 7800.0])

## Reaction 4026
#reaction('C2H2 + NC3H7 <=> C5H91-1', [2.000000e+11, 0.0, 7800.0])

## Reaction 4027
#reaction('CH2CHO + C3H5-A <=> C5AL-4D', [9.000000e+12, 0.0, 0.0])

## Reaction 4028
#reaction('CH2CHCHCHO + CH3 <=> C5AL-2D', [1.500000e+12, 0.0, 0.0])

## Reaction 4029
#reaction('PC4H9CHO + O2 <=> PC4H9CO + HO2', [1.200000e+05, 2.5, 37560.0])

## Reaction 4030
#reaction('PC4H9CHO + OH <=> PC4H9CO + H2O', [2.000000e+06, 1.8, -1300.0])

## Reaction 4031
#reaction('PC4H9CHO + H <=> PC4H9CO + H2', [4.140000e+09, 1.12, 2320.0])

## Reaction 4032
#reaction('PC4H9CHO + O <=> PC4H9CO + OH', [5.940000e+12, 0.0, 1868.0])

## Reaction 4033
#reaction('PC4H9CHO + HO2 <=> PC4H9CO + H2O2', [4.090000e+04, 2.5, 10200.0])

## Reaction 4034
#reaction('PC4H9CHO + CH3 <=> PC4H9CO + CH4', [2.890000e-03, 4.62, 3210.0])

## Reaction 4035
#reaction('PC4H9CHO + CH3O <=> PC4H9CO + CH3OH', [1.000000e+12, 0.0, 3300.0])

## Reaction 4036
#reaction('PC4H9CHO + CH3O2 <=> PC4H9CO + CH3O2H', [4.090000e+04, 2.5, 10200.0])

## Reaction 4037
#reaction('PC4H9CHO + OH <=> PC4H8CHO-1 + H2O', [5.520000e+02, 3.12, -1176.0])

## Reaction 4038
#reaction('PC4H9CHO + OH <=> PC4H8CHO-2 + H2O', [4.680000e+07, 1.61, -35.0])

## Reaction 4039
#reaction('PC4H9CHO + OH <=> PC4H8CHO-3 + H2O', [4.680000e+07, 1.61, -35.0])

## Reaction 4040
#reaction('PC4H9CHO + OH <=> PC4H8CHO-4 + H2O', [5.280000e+09, 0.97, 1586.0])

## Reaction 4041
#reaction('PC4H9CHO + HO2 <=> PC4H8CHO-1 + H2O2', [3.440000e+12, 0.05, 17880.0])

## Reaction 4042
#reaction('PC4H9CHO + HO2 <=> PC4H8CHO-2 + H2O2', [9.640000e+03, 2.6, 13910.0])

## Reaction 4043
#reaction('PC4H9CHO + HO2 <=> PC4H8CHO-3 + H2O2', [9.640000e+03, 2.6, 13910.0])

## Reaction 4044
#reaction('PC4H9CHO + HO2 <=> PC4H8CHO-4 + H2O2', [2.379000e+04, 2.55, 16490.0])

## Reaction 4045
#reaction('PC4H9CHO + CH3O2 <=> PC4H8CHO-1 + CH3O2H', [3.440000e+12, 0.05, 17880.0])

## Reaction 4046
#reaction('PC4H9CHO + CH3O2 <=> PC4H8CHO-2 + CH3O2H', [9.640000e+03, 2.6, 13910.0])

## Reaction 4047
#reaction('PC4H9CHO + CH3O2 <=> PC4H8CHO-3 + CH3O2H', [9.640000e+03, 2.6, 13910.0])

## Reaction 4048
#reaction('PC4H9CHO + CH3O2 <=> PC4H8CHO-4 + CH3O2H', [2.379000e+04, 2.55, 16490.0])

## Reaction 4049
#reaction('C6H10D13 <=> C2H3 + C4H71-1', [2.500000e+15, 0.0, 99500.0])

## Reaction 4050
#reaction('C6H10D24 <=> C3H5-S + C3H5-S', [2.500000e+15, 0.0, 99500.0])

## Reaction 4051
#reaction('C6H111-3 + O2 <=> C6H10D13 + HO2', [2.000000e+12, 0.0, 5000.0])

## Reaction 4052
#reaction('C6H112-4 + O2 <=> C6H10D24 + HO2', [2.000000e+12, 0.0, 5000.0])

## Reaction 4053
#reaction('C6KET13 => HOCHO + C5Y2', [1.700000e+05, 1.13, 26100.0])

## Reaction 4054
#reaction('C6KET13 => CH3COOH + NC3H7CHO', [5.100000e+05, 1.13, 26100.0])

## Reaction 4055
#reaction('C6KET24 => CH3COOH + C2H5COCH3', [1.700000e+05, 1.13, 26100.0])

## Reaction 4056
#reaction('C6KET35 => C2H5COOH + CH3COCH3', [3.400000e+05, 1.13, 26100.0])

## Reaction 4057
#reaction('CH3COOH <=> CH4 + CO2', [1.500000e+13, 0.0, 70000.0])

## Reaction 4058
#reaction('CH3COOH <=> H2O + CH2CO', [1.400000e+12, 0.0, 65000.0])

## Reaction 4059
#reaction('CH3COOH => CO + OH + CH3', [7.500000e+15, 0.0, 91600.0])

## Reaction 4060
#reaction('CH3COOH => CH3 + CO2 + H', [2.500000e+14, 0.0, 88000.0])

## Reaction 4061
#reaction('C2H5COOH => C2H6 + CO2', [1.500000e+13, 0.0, 70000.0])

## Reaction 4062
#reaction('C2H5COOH => H2O + CH3CHCO', [1.400000e+12, 0.0, 65000.0])

## Reaction 4063
#reaction('C2H5COOH => CO + OH + C2H5', [7.500000e+15, 0.0, 91600.0])

## Reaction 4064
#reaction('C2H5COOH => CO2 + C2H5 + H', [2.500000e+14, 0.0, 88000.0])

## Reaction 4065
#reaction('C6H13O2-1 + CH3 <=> C6H13O-1 + CH3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4066
#reaction('C6H13O2-2 + CH3 <=> C6H13O-2 + CH3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4067
#reaction('C6H13O2-3 + CH3 <=> C6H13O-3 + CH3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4068
#reaction('C6H13O2-1 + C2H5 <=> C6H13O-1 + C2H5O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4069
#reaction('C6H13O2-2 + C2H5 <=> C6H13O-2 + C2H5O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4070
#reaction('C6H13O2-3 + C2H5 <=> C6H13O-3 + C2H5O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4071
#reaction('C6H12OOH1-2O2 + CH3 => CH3O + C6H12OOH1-2O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4072
#reaction('C6H12OOH1-3O2 + CH3 => CH3O + C6H12OOH1-3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4073
#reaction('C6H12OOH1-4O2 + CH3 => CH3O + C6H12OOH1-4O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4074
#reaction('C6H12OOH1-5O2 + CH3 => CH3O + C6H12OOH1-5O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4075
#reaction('C6H12OOH2-1O2 + CH3 => CH3O + C6H12OOH2-1O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4076
#reaction('C6H12OOH2-3O2 + CH3 => CH3O + C6H12OOH2-3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4077
#reaction('C6H12OOH2-4O2 + CH3 => CH3O + C6H12OOH2-4O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4078
#reaction('C6H12OOH2-5O2 + CH3 => CH3O + C6H12OOH2-5O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4079
#reaction('C6H12OOH2-6O2 + CH3 => CH3O + C6H12OOH2-6O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4080
#reaction('C6H12OOH3-1O2 + CH3 => CH3O + C6H12OOH3-1O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4081
#reaction('C6H12OOH3-2O2 + CH3 => CH3O + C6H12OOH3-2O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4082
#reaction('C6H12OOH3-4O2 + CH3 => CH3O + C6H12OOH3-4O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4083
#reaction('C6H12OOH3-5O2 + CH3 => CH3O + C6H12OOH3-5O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4084
#reaction('C6H12OOH3-6O2 + CH3 => CH3O + C6H12OOH3-6O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4085
#reaction('C6H12OOH1-2O => HO2CH2CHO + PC4H9', [2.900000e+12, 0.4, 11080.0])

## Reaction 4086
#reaction('C6H12OOH1-3O => C3KET13 + NC3H7', [2.900000e+12, 0.4, 11080.0])

## Reaction 4087
#reaction('C6H12OOH1-4O => NC4KET14 + C2H5', [2.900000e+12, 0.4, 11080.0])

## Reaction 4088
#reaction('C6H12OOH1-5O => NC5KET15 + CH3', [2.900000e+12, 0.4, 11080.0])

## Reaction 4089
#reaction('C6H12OOH2-1O => CH2O + PC4H9CHO + OH', [2.900000e+12, 0.4, 11080.0])

## Reaction 4090
#reaction('C6H12OOH2-3O => C3KET12 + NC3H7', [2.900000e+12, 0.4, 11080.0])

## Reaction 4091
#reaction('C6H12OOH2-4O => NC4KET13 + C2H5', [2.900000e+12, 0.4, 11080.0])

## Reaction 4092
#reaction('C6H12OOH2-5O => NC5KET14 + CH3', [2.900000e+12, 0.4, 11080.0])

## Reaction 4093
#reaction('C6H12OOH2-6O => CH2O + C5H10OOH2-5', [2.900000e+12, 0.4, 11080.0])

## Reaction 4094
#reaction('C6H12OOH3-1O => CH2O + C5H10OOH2-1', [2.900000e+12, 0.4, 11080.0])

## Reaction 4095
#reaction('C6H12OOH3-2O => CH3 + NC5KET12', [2.900000e+12, 0.4, 11080.0])

## Reaction 4096
#reaction('C6H12OOH3-4O => NC4KET12 + C2H5', [2.900000e+12, 0.4, 11080.0])

## Reaction 4097
#reaction('C6H12OOH3-5O => CH3 + NC5KET13', [2.900000e+12, 0.4, 11080.0])

## Reaction 4098
#reaction('C6H12OOH3-6O => CH2O + C5H10OOH3-1', [2.900000e+12, 0.4, 11080.0])

## Reaction 4099
#reaction('C6H12OOH1-2O => CH2O2H + PC4H9CHO', [2.900000e+12, 0.4, 11080.0])

## Reaction 4100
#reaction('C6H12OOH1-3O => C2H4O2H + NC3H7CHO', [2.900000e+12, 0.4, 11080.0])

## Reaction 4101
#reaction('C6H12OOH1-4O => C3H6OOH1-3 + C2H5CHO', [2.900000e+12, 0.4, 11080.0])

## Reaction 4102
#reaction('C6H12OOH1-5O => C4H8OOH1-4 + CH3CHO', [2.900000e+12, 0.4, 11080.0])

## Reaction 4103
#reaction('C6H12OOH2-3O => CH3CHO + NC3H7CHO + OH', [2.900000e+12, 0.4, 11080.0])

## Reaction 4104
#reaction('C6H12OOH2-4O => C3H6OOH2-1 + C2H5CHO', [2.900000e+12, 0.4, 11080.0])

## Reaction 4105
#reaction('C6H12OOH2-5O => C4H8OOH2-4 + CH3CHO', [2.900000e+12, 0.4, 11080.0])

## Reaction 4106
#reaction('C6H12OOH3-2O => CH3CHO + NC3H7CHO + OH', [2.900000e+12, 0.4, 11080.0])

## Reaction 4107
#reaction('C6H12OOH3-4O => C2H5CHO + C2H5CHO + OH', [2.900000e+12, 0.4, 11080.0])

## Reaction 4108
#reaction('C6H12OOH3-5O => C4H8OOH2-1 + CH3CHO', [2.900000e+12, 0.4, 11080.0])

## Reaction 4109
#reaction('C6H12OOH1-2O2 + C6H13-1 => C6H13O-1 + C6H12OOH1-2O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4110
#reaction('C6H12OOH1-3O2 + C6H13-1 => C6H13O-1 + C6H12OOH1-3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4111
#reaction('C6H12OOH1-4O2 + C6H13-1 => C6H13O-1 + C6H12OOH1-4O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4112
#reaction('C6H12OOH1-5O2 + C6H13-1 => C6H13O-1 + C6H12OOH1-5O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4113
#reaction('C6H12OOH2-1O2 + C6H13-1 => C6H13O-1 + C6H12OOH2-1O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4114
#reaction('C6H12OOH2-3O2 + C6H13-1 => C6H13O-1 + C6H12OOH2-3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4115
#reaction('C6H12OOH2-4O2 + C6H13-1 => C6H13O-1 + C6H12OOH2-4O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4116
#reaction('C6H12OOH2-5O2 + C6H13-1 => C6H13O-1 + C6H12OOH2-5O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4117
#reaction('C6H12OOH2-6O2 + C6H13-1 => C6H13O-1 + C6H12OOH2-6O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4118
#reaction('C6H12OOH3-1O2 + C6H13-1 => C6H13O-1 + C6H12OOH3-1O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4119
#reaction('C6H12OOH3-2O2 + C6H13-1 => C6H13O-1 + C6H12OOH3-2O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4120
#reaction('C6H12OOH3-4O2 + C6H13-1 => C6H13O-1 + C6H12OOH3-4O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4121
#reaction('C6H12OOH3-5O2 + C6H13-1 => C6H13O-1 + C6H12OOH3-5O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4122
#reaction('C6H12OOH3-6O2 + C6H13-1 => C6H13O-1 + C6H12OOH3-6O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4123
#reaction('C6H12OOH1-2O2 + C6H13-2 => C6H13O-2 + C6H12OOH1-2O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4124
#reaction('C6H12OOH1-3O2 + C6H13-2 => C6H13O-2 + C6H12OOH1-3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4125
#reaction('C6H12OOH1-4O2 + C6H13-2 => C6H13O-2 + C6H12OOH1-4O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4126
#reaction('C6H12OOH1-5O2 + C6H13-2 => C6H13O-2 + C6H12OOH1-5O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4127
#reaction('C6H12OOH2-1O2 + C6H13-2 => C6H13O-2 + C6H12OOH2-1O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4128
#reaction('C6H12OOH2-3O2 + C6H13-2 => C6H13O-2 + C6H12OOH2-3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4129
#reaction('C6H12OOH2-4O2 + C6H13-2 => C6H13O-2 + C6H12OOH2-4O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4130
#reaction('C6H12OOH2-5O2 + C6H13-2 => C6H13O-2 + C6H12OOH2-5O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4131
#reaction('C6H12OOH2-6O2 + C6H13-2 => C6H13O-2 + C6H12OOH2-6O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4132
#reaction('C6H12OOH3-1O2 + C6H13-2 => C6H13O-2 + C6H12OOH3-1O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4133
#reaction('C6H12OOH3-2O2 + C6H13-2 => C6H13O-2 + C6H12OOH3-2O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4134
#reaction('C6H12OOH3-4O2 + C6H13-2 => C6H13O-2 + C6H12OOH3-4O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4135
#reaction('C6H12OOH3-5O2 + C6H13-2 => C6H13O-2 + C6H12OOH3-5O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4136
#reaction('C6H12OOH3-6O2 + C6H13-2 => C6H13O-2 + C6H12OOH3-6O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4137
#reaction('C6H12OOH1-2O2 + C6H13-3 => C6H13O-3 + C6H12OOH1-2O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4138
#reaction('C6H12OOH1-3O2 + C6H13-3 => C6H13O-3 + C6H12OOH1-3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4139
#reaction('C6H12OOH1-4O2 + C6H13-3 => C6H13O-3 + C6H12OOH1-4O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4140
#reaction('C6H12OOH1-5O2 + C6H13-3 => C6H13O-3 + C6H12OOH1-5O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4141
#reaction('C6H12OOH2-1O2 + C6H13-3 => C6H13O-3 + C6H12OOH2-1O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4142
#reaction('C6H12OOH2-3O2 + C6H13-3 => C6H13O-3 + C6H12OOH2-3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4143
#reaction('C6H12OOH2-4O2 + C6H13-3 => C6H13O-3 + C6H12OOH2-4O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4144
#reaction('C6H12OOH2-5O2 + C6H13-3 => C6H13O-3 + C6H12OOH2-5O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4145
#reaction('C6H12OOH2-6O2 + C6H13-3 => C6H13O-3 + C6H12OOH2-6O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4146
#reaction('C6H12OOH3-1O2 + C6H13-3 => C6H13O-3 + C6H12OOH3-1O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4147
#reaction('C6H12OOH3-2O2 + C6H13-3 => C6H13O-3 + C6H12OOH3-2O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4148
#reaction('C6H12OOH3-4O2 + C6H13-3 => C6H13O-3 + C6H12OOH3-4O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4149
#reaction('C6H12OOH3-5O2 + C6H13-3 => C6H13O-3 + C6H12OOH3-5O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4150
#reaction('C6H12OOH3-6O2 + C6H13-3 => C6H13O-3 + C6H12OOH3-6O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4151
#reaction('I3C6 <=> I3C6-1 + H', [1.000000e+15, 0.0, 100000.0])

## Reaction 4152
#reaction('I3C6 <=> I3C6-2 + H', [1.000000e+15, 0.0, 100000.0])

## Reaction 4153
#reaction('I3C6 <=> I3C6-3 + H', [1.000000e+15, 0.0, 93300.0])

## Reaction 4154
#reaction('I3C6 <=> I3C6-4 + H', [3.300000e+14, 0.0, 98000.0])

## Reaction 4155
#reaction('I3C6 <=> AC5H11 + CH3', [6.300000e+16, 0.0, 85400.0])

## Reaction 4156
#reaction('I3C6 <=> SC4H9 + C2H5', [5.000000e+16, 0.0, 79480.0])

## Reaction 4157
#reaction('I3C6 <=> C5H11-3 + CH3', [6.700000e+16, 0.0, 83400.0])

## Reaction 4158
#reaction('I3C6 + H <=> I3C6-1 + H2', [3.490000e+05, 2.7, 6450.0])

## Reaction 4159
#reaction('I3C6 + H <=> I3C6-2 + H2', [2.600000e+06, 2.4, 4471.0])

## Reaction 4160
#reaction('I3C6 + H <=> I3C6-3 + H2', [6.020000e+05, 2.4, 2583.0])

## Reaction 4161
#reaction('I3C6 + H <=> I3C6-4 + H2', [1.745000e+05, 2.69, 6450.0])

## Reaction 4162
#reaction('I3C6 + O <=> I3C6-1 + OH', [1.130000e+14, 0.0, 7850.0])

## Reaction 4163
#reaction('I3C6 + O <=> I3C6-2 + OH', [5.620000e+13, 0.0, 5200.0])

## Reaction 4164
#reaction('I3C6 + O <=> I3C6-3 + OH', [3.968000e+05, 2.401, 1150.0])

## Reaction 4165
#reaction('I3C6 + O <=> I3C6-4 + OH', [1.046000e+06, 2.424, 4766.0])

## Reaction 4166
#reaction('I3C6 + OH <=> I3C6-1 + H2O', [1.050000e+10, 1.0, 1586.0])

## Reaction 4167
#reaction('I3C6 + OH <=> I3C6-2 + H2O', [9.340000e+07, 1.6, -35.0])

## Reaction 4168
#reaction('I3C6 + OH <=> I3C6-3 + H2O', [5.730000e+10, 0.51, 63.0])

## Reaction 4169
#reaction('I3C6 + OH <=> I3C6-4 + H2O', [5.270000e+09, 0.97, 1586.0])

## Reaction 4170
#reaction('I3C6 + CH3 <=> I3C6-1 + CH4', [9.040000e-01, 3.6, 7154.0])

## Reaction 4171
#reaction('I3C6 + CH3 <=> I3C6-2 + CH4', [1.680000e+05, 2.1, 7574.0])

## Reaction 4172
#reaction('I3C6 + CH3 <=> I3C6-3 + CH4', [6.010000e-10, 6.36, 893.0])

## Reaction 4173
#reaction('I3C6 + CH3 <=> I3C6-4 + CH4', [4.520000e-01, 3.65, 7154.0])

## Reaction 4174
#reaction('I3C6 + HO2 <=> I3C6-1 + H2O2', [4.080000e+01, 3.6, 17160.0])

## Reaction 4175
#reaction('I3C6 + HO2 <=> I3C6-2 + H2O2', [1.260000e+02, 3.4, 13720.0])

## Reaction 4176
#reaction('I3C6 + HO2 <=> I3C6-3 + H2O2', [6.504000e+02, 3.01, 12090.0])

## Reaction 4177
#reaction('I3C6 + HO2 <=> I3C6-4 + H2O2', [2.040000e+01, 3.59, 17160.0])

## Reaction 4178
#reaction('I3C6 + CH3O2 <=> I3C6-1 + CH3O2H', [1.390000e+00, 4.0, 18280.0])

## Reaction 4179
#reaction('I3C6 + CH3O2 <=> I3C6-2 + CH3O2H', [2.040000e+01, 3.6, 14810.0])

## Reaction 4180
#reaction('I3C6 + CH3O2 <=> I3C6-3 + CH3O2H', [2.620000e+02, 3.12, 11100.0])

## Reaction 4181
#reaction('I3C6 + CH3O2 <=> I3C6-4 + CH3O2H', [6.930000e-01, 3.97, 18280.0])

## Reaction 4182
#reaction('I3C6 + CH3O <=> I3C6-1 + CH3OH', [3.000000e+11, 0.0, 7000.0])

## Reaction 4183
#reaction('I3C6 + CH3O <=> I3C6-2 + CH3OH', [2.200000e+11, 0.0, 5000.0])

## Reaction 4184
#reaction('I3C6 + CH3O <=> I3C6-3 + CH3OH', [1.900000e+10, 0.0, 2800.0])

## Reaction 4185
#reaction('I3C6 + CH3O <=> I3C6-4 + CH3OH', [1.500000e+11, 0.0, 7000.0])

## Reaction 4186
#reaction('I3C6 + O2 <=> I3C6-1 + HO2', [4.200000e+13, 0.0, 52800.0])

## Reaction 4187
#reaction('I3C6 + O2 <=> I3C6-2 + HO2', [2.800000e+13, 0.0, 50160.0])

## Reaction 4188
#reaction('I3C6 + O2 <=> I3C6-3 + HO2', [7.000000e+12, 0.0, 48000.0])

## Reaction 4189
#reaction('I3C6 + O2 <=> I3C6-4 + HO2', [2.100000e+13, 0.0, 52800.0])

## Reaction 4190
#reaction('I3C6 + C2H5 <=> I3C6-1 + C2H6', [1.000000e+11, 0.0, 13400.0])

## Reaction 4191
#reaction('I3C6 + C2H5 <=> I3C6-2 + C2H6', [1.000000e+11, 0.0, 10400.0])

## Reaction 4192
#reaction('I3C6 + C2H5 <=> I3C6-3 + C2H6', [1.000000e+11, 0.0, 7900.0])

## Reaction 4193
#reaction('I3C6 + C2H5 <=> I3C6-4 + C2H6', [5.000000e+10, 0.0, 13400.0])

## Reaction 4194
#reaction('I3C6 + C2H3 <=> I3C6-1 + C2H4', [1.000000e+12, 0.0, 18000.0])

## Reaction 4195
#reaction('I3C6 + C2H3 <=> I3C6-2 + C2H4', [8.000000e+11, 0.0, 16800.0])

## Reaction 4196
#reaction('I3C6 + C2H3 <=> I3C6-3 + C2H4', [2.000000e+11, 0.0, 14300.0])

## Reaction 4197
#reaction('I3C6 + C2H3 <=> I3C6-4 + C2H4', [5.000000e+11, 0.0, 18000.0])

## Reaction 4198
#reaction('I3C6 + O2CHO <=> I3C6-1 + HO2CHO', [1.390000e+00, 4.0, 18280.0])

## Reaction 4199
#reaction('I3C6 + O2CHO <=> I3C6-2 + HO2CHO', [2.040000e+01, 3.6, 14810.0])

## Reaction 4200
#reaction('I3C6 + O2CHO <=> I3C6-3 + HO2CHO', [1.500000e+04, 2.5, 12260.0])

## Reaction 4201
#reaction('I3C6 + O2CHO <=> I3C6-4 + HO2CHO', [8.400000e+12, 0.0, 20440.0])

## Reaction 4202
#reaction('I3C6-1 <=> C2H4 + SC4H9', [1.300000e+13, 0.0, 26200.0])

## Reaction 4203
#reaction('I3C6-2 <=> C5H10-2 + CH3', [1.000000e+13, 0.0, 36000.0])

## Reaction 4204
#reaction('I3C6-2 <=> C4H8-2 + C2H5', [3.300000e+13, 0.0, 32800.0])

## Reaction 4205
#reaction('I3C6-3 <=> AC5H10 + CH3', [1.000000e+13, 0.0, 36100.0])

## Reaction 4206
#reaction('I3C6-4 <=> C4H8-1 + C2H5', [1.000000e+13, 0.0, 32800.0])

## Reaction 4207
#reaction('I3C6-1 + O2 <=> I3C6D1 + HO2', [8.370000e-01, 3.6, 11960.0])

## Reaction 4208
#reaction('I3C6-2 + O2 <=> I3C6D1 + HO2', [5.350000e-01, 3.7, 9322.0])

## Reaction 4209
#reaction('I3C6-2 + O2 <=> I3C6D2 + HO2', [5.350000e-01, 3.7, 9322.0])

## Reaction 4210
#reaction('I3C6-3 + O2 <=> I3C6D2 + HO2', [1.070000e+00, 3.7, 9322.0])

## Reaction 4211
#reaction('I3C6-3 + O2 <=> I3C6D3 + HO2', [5.350000e-01, 3.7, 9322.0])

## Reaction 4212
#reaction('I3C6-4 + O2 <=> I3C6D3 + HO2', [4.185000e-01, 3.6, 11960.0])

## Reaction 4213
#reaction('I3C6-1 <=> I3C6-4', [3.000000e+11, 0.0, 21100.0])

## Reaction 4214
#reaction('I3C6-1 <=> I3C6-2', [2.000000e+11, 0.0, 18100.0])

## Reaction 4215
#reaction('I3C6D1 + OH <=> I3C6D1-3 + H2O', [6.140000e+02, 3.2, -3500.0])

## Reaction 4216
#reaction('I3C6D1 + OH <=> I3C6D1-4 + H2O', [5.270000e+09, 0.97, 1586.0])

## Reaction 4217
#reaction('I3C6D1 + OH <=> I3C6D1-5 + H2O', [4.670000e+07, 1.61, -35.0])

## Reaction 4218
#reaction('I3C6D1 + OH <=> I3C6D1-6 + H2O', [5.270000e+09, 0.97, 1586.0])

## Reaction 4219
#reaction('I3C6D2 + OH <=> I3C6D1-3 + H2O', [3.120000e+06, 2.0, -298.0])

## Reaction 4220
#reaction('I3C6D2 + OH <=> I3C6D3-2 + H2O', [3.120000e+06, 2.0, -298.0])

## Reaction 4221
#reaction('I3C6D2 + OH <=> I3C6D2-5 + H2O', [2.764000e+04, 2.64, -1919.0])

## Reaction 4222
#reaction('I3C6D2 + OH <=> I3C6D2-6 + H2O', [5.270000e+09, 0.97, 1586.0])

## Reaction 4223
#reaction('I3C6D3 + OH <=> I3C6D3-1 + H2O', [5.270000e+09, 0.97, 1586.0])

## Reaction 4224
#reaction('I3C6D3 + OH <=> I3C6D3-2 + H2O', [2.764000e+04, 2.64, -1919.0])

## Reaction 4225
#reaction('I3C6D1 + H <=> I3C6D1-3 + H2', [2.650000e+06, 2.2, 0.0])

## Reaction 4226
#reaction('I3C6D1 + H <=> I3C6D1-4 + H2', [1.745000e+05, 2.69, 6450.0])

## Reaction 4227
#reaction('I3C6D1 + H <=> I3C6D1-5 + H2', [1.300000e+06, 2.4, 4471.0])

## Reaction 4228
#reaction('I3C6D1 + H <=> I3C6D1-6 + H2', [1.745000e+05, 2.69, 6450.0])

## Reaction 4229
#reaction('I3C6D2 + H <=> I3C6D1-3 + H2', [1.730000e+05, 2.5, 2492.0])

## Reaction 4230
#reaction('I3C6D2 + H <=> I3C6D3-2 + H2', [1.730000e+05, 2.5, 2492.0])

## Reaction 4231
#reaction('I3C6D2 + H <=> I3C6D2-5 + H2', [3.376000e+05, 2.36, 207.0])

## Reaction 4232
#reaction('I3C6D2 + H <=> I3C6D2-6 + H2', [1.745000e+05, 2.69, 6450.0])

## Reaction 4233
#reaction('I3C6D3 + H <=> I3C6D3-1 + H2', [1.745000e+05, 2.69, 6450.0])

## Reaction 4234
#reaction('I3C6D3 + H <=> I3C6D3-2 + H2', [3.376000e+05, 2.36, 207.0])

## Reaction 4235
#reaction('I3C6D1 + HO2 <=> I3C6D1-3 + H2O2', [1.810000e+03, 2.5, 7154.0])

## Reaction 4236
#reaction('I3C6D1 + HO2 <=> I3C6D1-4 + H2O2', [2.040000e+01, 3.59, 17160.0])

## Reaction 4237
#reaction('I3C6D1 + HO2 <=> I3C6D1-5 + H2O2', [6.320000e+01, 3.37, 13720.0])

## Reaction 4238
#reaction('I3C6D1 + HO2 <=> I3C6D1-6 + H2O2', [2.040000e+01, 3.59, 17160.0])

## Reaction 4239
#reaction('I3C6D2 + HO2 <=> I3C6D1-3 + H2O2', [9.639000e+03, 2.6, 13910.0])

## Reaction 4240
#reaction('I3C6D2 + HO2 <=> I3C6D3-2 + H2O2', [9.639000e+03, 2.6, 13910.0])

## Reaction 4241
#reaction('I3C6D2 + HO2 <=> I3C6D2-5 + H2O2', [4.820000e+03, 2.55, 10530.0])

## Reaction 4242
#reaction('I3C6D2 + HO2 <=> I3C6D2-6 + H2O2', [2.040000e+01, 3.59, 17160.0])

## Reaction 4243
#reaction('I3C6D3 + HO2 <=> I3C6D3-1 + H2O2', [2.040000e+01, 3.59, 17160.0])

## Reaction 4244
#reaction('I3C6D3 + HO2 <=> I3C6D3-2 + H2O2', [4.820000e+03, 2.55, 10530.0])

## Reaction 4245
#reaction('I3C6D1 + CH3 <=> I3C6D1-3 + CH4', [4.613000e+00, 3.1, 2330.0])

## Reaction 4246
#reaction('I3C6D1 + CH3 <=> I3C6D1-4 + CH4', [4.520000e-01, 3.65, 7154.0])

## Reaction 4247
#reaction('I3C6D1 + CH3 <=> I3C6D1-5 + CH4', [8.400000e+04, 2.133, 7574.0])

## Reaction 4248
#reaction('I3C6D1 + CH3 <=> I3C6D1-6 + CH4', [4.520000e-01, 3.65, 7154.0])

## Reaction 4249
#reaction('I3C6D2 + CH3 <=> I3C6D1-3 + CH4', [2.210000e+00, 3.5, 5675.0])

## Reaction 4250
#reaction('I3C6D2 + CH3 <=> I3C6D3-2 + CH4', [2.210000e+00, 3.5, 5675.0])

## Reaction 4251
#reaction('I3C6D2 + CH3 <=> I3C6D2-5 + CH4', [3.690000e+00, 3.31, 4002.0])

## Reaction 4252
#reaction('I3C6D2 + CH3 <=> I3C6D2-6 + CH4', [4.520000e-01, 3.65, 7154.0])

## Reaction 4253
#reaction('I3C6D3 + CH3 <=> I3C6D3-1 + CH4', [4.520000e-01, 3.65, 7154.0])

## Reaction 4254
#reaction('I3C6D3 + CH3 <=> I3C6D3-2 + CH4', [3.690000e+00, 3.31, 4002.0])

## Reaction 4255
#reaction('I3C6D1 + CH3O <=> I3C6D1-3 + CH3OH', [1.000000e+01, 2.85, 5231.0])

## Reaction 4256
#reaction('I3C6D1 + CH3O <=> I3C6D1-4 + CH3OH', [1.500000e+11, 0.0, 7000.0])

## Reaction 4257
#reaction('I3C6D1 + CH3O <=> I3C6D1-5 + CH3OH', [1.100000e+11, 0.0, 5000.0])

## Reaction 4258
#reaction('I3C6D1 + CH3O <=> I3C6D1-6 + CH3OH', [1.500000e+11, 0.0, 7000.0])

## Reaction 4259
#reaction('I3C6D2 + CH3O <=> I3C6D1-3 + CH3OH', [9.000000e+01, 2.95, 11990.0])

## Reaction 4260
#reaction('I3C6D2 + CH3O <=> I3C6D3-2 + CH3OH', [9.000000e+01, 2.95, 11990.0])

## Reaction 4261
#reaction('I3C6D2 + CH3O <=> I3C6D2-5 + CH3OH', [4.000000e+01, 2.9, 8609.0])

## Reaction 4262
#reaction('I3C6D2 + CH3O <=> I3C6D2-6 + CH3OH', [1.500000e+11, 0.0, 7000.0])

## Reaction 4263
#reaction('I3C6D3 + CH3O <=> I3C6D3-1 + CH3OH', [1.500000e+11, 0.0, 7000.0])

## Reaction 4264
#reaction('I3C6D3 + CH3O <=> I3C6D3-2 + CH3OH', [4.000000e+01, 2.9, 8609.0])

## Reaction 4265
#reaction('I3C6D1 + CH3O2 <=> I3C6D1-3 + CH3O2H', [1.810000e+03, 2.5, 7154.0])

## Reaction 4266
#reaction('I3C6D1 + CH3O2 <=> I3C6D1-4 + CH3O2H', [6.930000e-01, 3.97, 18280.0])

## Reaction 4267
#reaction('I3C6D1 + CH3O2 <=> I3C6D1-5 + CH3O2H', [1.019000e+01, 3.58, 14810.0])

## Reaction 4268
#reaction('I3C6D1 + CH3O2 <=> I3C6D1-6 + CH3O2H', [6.930000e-01, 3.97, 18280.0])

## Reaction 4269
#reaction('I3C6D2 + CH3O2 <=> I3C6D1-3 + CH3O2H', [9.639000e+03, 2.6, 13910.0])

## Reaction 4270
#reaction('I3C6D2 + CH3O2 <=> I3C6D3-2 + CH3O2H', [9.639000e+03, 2.6, 13910.0])

## Reaction 4271
#reaction('I3C6D2 + CH3O2 <=> I3C6D2-5 + CH3O2H', [4.820000e+03, 2.55, 10530.0])

## Reaction 4272
#reaction('I3C6D2 + CH3O2 <=> I3C6D2-6 + CH3O2H', [6.930000e-01, 3.97, 18280.0])

## Reaction 4273
#reaction('I3C6D3 + CH3O2 <=> I3C6D3-1 + CH3O2H', [6.930000e-01, 3.97, 18280.0])

## Reaction 4274
#reaction('I3C6D3 + CH3O2 <=> I3C6D3-2 + CH3O2H', [4.820000e+03, 2.55, 10530.0])

## Reaction 4275
#reaction('I3C6-1 <=> I3C6D1 + H', [4.000000e+12, 0.0, 38600.0])

## Reaction 4276
#reaction('I3C6-2 <=> I3C6D1 + H', [2.000000e+13, 0.0, 40400.0])

## Reaction 4277
#reaction('I3C6-2 <=> I3C6D2 + H', [5.000000e+13, 0.0, 39000.0])

## Reaction 4278
#reaction('I3C6-3 <=> I3C6D2 + H', [5.000000e+12, 0.0, 42600.0])

## Reaction 4279
#reaction('I3C6-3 <=> I3C6D3 + H', [5.000000e+12, 0.0, 43600.0])

## Reaction 4280
#reaction('I3C6-4 <=> I3C6D3 + H', [5.000000e+12, 0.0, 36000.0])

## Reaction 4281
#reaction('I3C6D1 + OH <=> I3C6OH1-2J', [1.000000e+12, 0.0, -1042.0])

## Reaction 4282
#reaction('I3C6D1 + OH <=> I3C6OH2-1J', [1.000000e+12, 0.0, -1042.0])

## Reaction 4283
#reaction('I3C6D2 + OH <=> I3C6OH2-3J', [1.000000e+12, 0.0, -1042.0])

## Reaction 4284
#reaction('I3C6D2 + OH <=> I3C6OH3-2J', [1.000000e+12, 0.0, -1042.0])

## Reaction 4285
#reaction('I3C6D3 + OH <=> I3C6OH3-4J', [1.000000e+12, 0.0, -1042.0])

## Reaction 4286
#reaction('I3C6D3 + OH <=> I3C6OH4-3J', [1.000000e+12, 0.0, -1042.0])

## Reaction 4287
#reaction('I3C6OH1-2J + O2 <=> I3C6OH1-2O2', [2.000000e+12, 0.0, 0.0])

## Reaction 4288
#reaction('I3C6OH2-1J + O2 <=> I3C6OH2-1O2', [2.000000e+12, 0.0, 0.0])

## Reaction 4289
#reaction('I3C6OH2-3J + O2 <=> I3C6OH2-3O2', [2.000000e+12, 0.0, 0.0])

## Reaction 4290
#reaction('I3C6OH3-2J + O2 <=> I3C6OH3-2O2', [2.000000e+12, 0.0, 0.0])

## Reaction 4291
#reaction('I3C6OH3-4J + O2 <=> I3C6OH3-4O2', [2.000000e+12, 0.0, 0.0])

## Reaction 4292
#reaction('I3C6OH4-3J + O2 <=> I3C6OH4-3O2', [2.000000e+12, 0.0, 0.0])

## Reaction 4293
#reaction('I3C6OH1-2O2 => SC4H9CHO + CH2O + OH', [2.500000e+10, 0.0, 18860.0])

## Reaction 4294
#reaction('I3C6OH2-1O2 => SC4H9CHO + CH2O + OH', [2.500000e+10, 0.0, 18860.0])

## Reaction 4295
#reaction('I3C6OH2-3O2 => CH3CHO + C2H5COCH3 + OH', [2.500000e+10, 0.0, 18860.0])

## Reaction 4296
#reaction('I3C6OH3-2O2 => CH3CHO + C2H5COCH3 + OH', [2.500000e+10, 0.0, 18860.0])

## Reaction 4297
#reaction('I3C6OH3-4O2 => C5Y3 + CH2O + OH', [2.500000e+10, 0.0, 18860.0])

## Reaction 4298
#reaction('I3C6OH4-3O2 => C5Y3 + CH2O + OH', [2.500000e+10, 0.0, 18860.0])

## Reaction 4299
#reaction('I3C6D1 + OH <=> CH2O + AC5H11', [2.000000e+10, 0.0, 4000.0])

## Reaction 4300
#reaction('I3C6D1 + OH <=> SC4H9CHO + CH3', [2.000000e+10, 0.0, 4000.0])

## Reaction 4301
#reaction('I3C6D2 + OH <=> CH3CHO + SC4H9', [2.000000e+10, 0.0, 4000.0])

## Reaction 4302
#reaction('I3C6D2 + OH <=> C2H5COCH3 + C2H5', [2.000000e+10, 0.0, 4000.0])

## Reaction 4303
#reaction('I3C6D3 + OH <=> CH2O + C5H11-3', [2.000000e+10, 0.0, 4000.0])

## Reaction 4304
#reaction('I3C6D3 + OH <=> C5Y3 + CH3', [2.000000e+10, 0.0, 4000.0])

## Reaction 4305
#reaction('I3C6D3 + OH <=> C2H5COCH3 + C2H5', [2.000000e+10, 0.0, 4000.0])

## Reaction 4306
#reaction('CH3 + B13DE2M <=> I3C6D1-3', [1.000000e+11, 0.0, 7800.0])

## Reaction 4307
#reaction('C2H5 + C4H6 <=> I3C6D1-4', [1.000000e+11, 0.0, 7800.0])

## Reaction 4308
#reaction('CH3 + C5H81-3 <=> I3C6D1-5', [1.000000e+11, 0.0, 7800.0])

## Reaction 4309
#reaction('C2H4 + C4H71-3 <=> I3C6D1-6', [1.000000e+11, 0.0, 7800.0])

## Reaction 4310
#reaction('C4H72-2 + C2H4 <=> I3C6D2-6', [1.000000e+11, 0.0, 7800.0])

## Reaction 4311
#reaction('C4H71-2 + C2H4 <=> I3C6D3-1', [1.000000e+11, 0.0, 7800.0])

## Reaction 4312
#reaction('I3C6D1-3 + HO2 <=> I3C6D1-3O + OH', [9.640000e+12, 0.0, 0.0])

## Reaction 4313
#reaction('I3C6D1-4 + HO2 <=> I3C6D1-4O + OH', [9.640000e+12, 0.0, 0.0])

## Reaction 4314
#reaction('I3C6D1-5 + HO2 <=> I3C6D1-5O + OH', [9.640000e+12, 0.0, 0.0])

## Reaction 4315
#reaction('I3C6D1-6 + HO2 <=> I3C6D1-6O + OH', [9.640000e+12, 0.0, 0.0])

## Reaction 4316
#reaction('I3C6D2-5 + HO2 <=> I3C6D2-5O + OH', [9.640000e+12, 0.0, 0.0])

## Reaction 4317
#reaction('I3C6D2-6 + HO2 <=> I3C6D2-6O + OH', [9.640000e+12, 0.0, 0.0])

## Reaction 4318
#reaction('I3C6D3-1 + HO2 <=> I3C6D3-1O + OH', [9.640000e+12, 0.0, 0.0])

## Reaction 4319
#reaction('I3C6D3-2 + HO2 <=> I3C6D3-2O + OH', [9.640000e+12, 0.0, 0.0])

## Reaction 4320
#reaction('I3C6D1-3 + CH3O2 <=> I3C6D1-3O + CH3O', [9.640000e+12, 0.0, 0.0])

## Reaction 4321
#reaction('I3C6D1-4 + CH3O2 <=> I3C6D1-4O + CH3O', [9.640000e+12, 0.0, 0.0])

## Reaction 4322
#reaction('I3C6D1-5 + CH3O2 <=> I3C6D1-5O + CH3O', [9.640000e+12, 0.0, 0.0])

## Reaction 4323
#reaction('I3C6D1-6 + CH3O2 <=> I3C6D1-6O + CH3O', [9.640000e+12, 0.0, 0.0])

## Reaction 4324
#reaction('I3C6D2-5 + CH3O2 <=> I3C6D2-5O + CH3O', [9.640000e+12, 0.0, 0.0])

## Reaction 4325
#reaction('I3C6D2-6 + CH3O2 <=> I3C6D2-6O + CH3O', [9.640000e+12, 0.0, 0.0])

## Reaction 4326
#reaction('I3C6D3-1 + CH3O2 <=> I3C6D3-1O + CH3O', [9.640000e+12, 0.0, 0.0])

## Reaction 4327
#reaction('I3C6D3-2 + CH3O2 <=> I3C6D3-2O + CH3O', [9.640000e+12, 0.0, 0.0])

## Reaction 4328
#reaction('I3C6D1-3 + C2H5O2 <=> I3C6D1-3O + C2H5O', [9.640000e+12, 0.0, 0.0])

## Reaction 4329
#reaction('I3C6D1-4 + C2H5O2 <=> I3C6D1-4O + C2H5O', [9.640000e+12, 0.0, 0.0])

## Reaction 4330
#reaction('I3C6D1-5 + C2H5O2 <=> I3C6D1-5O + C2H5O', [9.640000e+12, 0.0, 0.0])

## Reaction 4331
#reaction('I3C6D1-6 + C2H5O2 <=> I3C6D1-6O + C2H5O', [9.640000e+12, 0.0, 0.0])

## Reaction 4332
#reaction('I3C6D2-5 + C2H5O2 <=> I3C6D2-5O + C2H5O', [9.640000e+12, 0.0, 0.0])

## Reaction 4333
#reaction('I3C6D2-6 + C2H5O2 <=> I3C6D2-6O + C2H5O', [9.640000e+12, 0.0, 0.0])

## Reaction 4334
#reaction('I3C6D3-1 + C2H5O2 <=> I3C6D3-1O + C2H5O', [9.640000e+12, 0.0, 0.0])

## Reaction 4335
#reaction('I3C6D3-2 + C2H5O2 <=> I3C6D3-2O + C2H5O', [9.640000e+12, 0.0, 0.0])

## Reaction 4336
#reaction('I3C6D1-3O <=> C2H5COCH3 + C2H3', [7.813000e+13, -0.25, 22330.0])

## Reaction 4337
#reaction('I3C6D1-4O <=> C5H91-3 + CH2O', [4.470000e+15, -1.13, 4140.0])

## Reaction 4338
#reaction('I3C6D1-5O <=> C4H71-3 + CH3CHO', [4.470000e+15, -1.13, 4140.0])

## Reaction 4339
#reaction('CC5H9-A + CH2O <=> I3C6D1-6O', [6.250000e+10, 0.0, 11900.0])

## Reaction 4340
#reaction('I3C6D2-5O <=> C4H72-2 + CH3CHO', [3.240000e+22, -2.59, 30000.0])

## Reaction 4341
#reaction('I3C6D2-6O <=> AC5H9-C + CH2O', [6.880000e+14, -0.89, 7260.0])

## Reaction 4342
#reaction('I3C6D3-1O <=> AC5H9-A2 + CH2O', [6.880000e+14, -0.89, 7260.0])

## Reaction 4343
#reaction('I3C6D3-2O <=> C4H71-2 + CH3CHO', [3.231000e+22, -2.63, 30310.0])

## Reaction 4344
#reaction('I3C6D1 <=> C4H71-3 + C2H5', [1.302000e+21, -1.639, 76140.0])

## Reaction 4345
#reaction('I3C6D1 <=> C5H91-3 + CH3', [1.302000e+21, -1.639, 76140.0])

## Reaction 4346
#reaction('I3C6D2 <=> AC5H9-C + CH3', [1.900000e+20, -1.582, 75930.0])

## Reaction 4347
#reaction('I3C6D3 <=> AC5H9-A2 + CH3', [3.800000e+20, -1.582, 75930.0])

## Reaction 4348
#reaction('I3C6-1 + O2 <=> I3C6-1O2', [6.863000e+16, -1.627, 198.7])

## Reaction 4349
#reaction('I3C6-2 + O2 <=> I3C6-2O2', [3.486000e+14, -0.816, -536.5])

## Reaction 4350
#reaction('I3C6-3 + O2 <=> I3C6-3O2', [9.752000e+11, 0.325, -417.3])

## Reaction 4351
#reaction('I3C6-4 + O2 <=> I3C6-4O2', [6.863000e+16, -1.627, 198.7])

## Reaction 4352
#reaction('I3C6-1 + HO2 <=> I3C6-1O + OH', [9.000000e+12, 0.0, -1000.0])

## Reaction 4353
#reaction('I3C6-2 + HO2 <=> I3C6-2O + OH', [9.000000e+12, 0.0, -1000.0])

## Reaction 4354
#reaction('I3C6-3 + HO2 <=> I3C6-3O + OH', [9.000000e+12, 0.0, -1000.0])

## Reaction 4355
#reaction('I3C6-4 + HO2 <=> I3C6-4O + OH', [9.000000e+12, 0.0, -1000.0])

## Reaction 4356
#reaction('I3C6-1 + CH3O2 <=> I3C6-1O + CH3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4357
#reaction('I3C6-2 + CH3O2 <=> I3C6-2O + CH3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4358
#reaction('I3C6-3 + CH3O2 <=> I3C6-3O + CH3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4359
#reaction('I3C6-4 + CH3O2 <=> I3C6-4O + CH3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4360
#reaction('I3C6-1 + I3C6-1O2 <=> I3C6-1O + I3C6-1O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4361
#reaction('I3C6-2 + I3C6-1O2 <=> I3C6-2O + I3C6-1O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4362
#reaction('I3C6-3 + I3C6-1O2 <=> I3C6-3O + I3C6-1O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4363
#reaction('I3C6-4 + I3C6-1O2 <=> I3C6-4O + I3C6-1O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4364
#reaction('I3C6-1 + I3C6-2O2 <=> I3C6-1O + I3C6-2O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4365
#reaction('I3C6-2 + I3C6-2O2 <=> I3C6-2O + I3C6-2O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4366
#reaction('I3C6-3 + I3C6-2O2 <=> I3C6-3O + I3C6-2O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4367
#reaction('I3C6-4 + I3C6-2O2 <=> I3C6-4O + I3C6-2O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4368
#reaction('I3C6-1 + I3C6-3O2 <=> I3C6-1O + I3C6-3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4369
#reaction('I3C6-2 + I3C6-3O2 <=> I3C6-2O + I3C6-3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4370
#reaction('I3C6-3 + I3C6-3O2 <=> I3C6-3O + I3C6-3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4371
#reaction('I3C6-4 + I3C6-3O2 <=> I3C6-4O + I3C6-3O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4372
#reaction('I3C6-1 + I3C6-4O2 <=> I3C6-1O + I3C6-4O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4373
#reaction('I3C6-2 + I3C6-4O2 <=> I3C6-2O + I3C6-4O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4374
#reaction('I3C6-3 + I3C6-4O2 <=> I3C6-3O + I3C6-4O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4375
#reaction('I3C6-4 + I3C6-4O2 <=> I3C6-4O + I3C6-4O', [9.000000e+12, 0.0, -1000.0])

## Reaction 4376
#reaction('I3C6-1O2 <=> I3C6OOH1-2', [4.009000e+08, 1.1, 30100.0])

## Reaction 4377
#reaction('I3C6-1O2 <=> I3C6OOH1-3', [1.231000e+07, 1.2, 15400.0])

## Reaction 4378
#reaction('I3C6-1O2 <=> I3C6OOH1-4', [1.233000e+06, 1.5, 20000.0])

## Reaction 4379
#reaction('I3C6-1O2 <=> I3C6OOH1-5', [3.718000e+06, 1.2, 16600.0])

## Reaction 4380
#reaction('I3C6-1O2 <=> I3C6OOH1-6', [2.941000e+05, 1.5, 19900.0])

## Reaction 4381
#reaction('I3C6-2O2 <=> I3C6OOH2-1', [2.431000e+08, 1.1, 33500.0])

## Reaction 4382
#reaction('I3C6-2O2 <=> I3C6OOH2-3', [7.449000e+09, 0.6, 27300.0])

## Reaction 4383
#reaction('I3C6-2O2 <=> I3C6OOH2-4', [1.439000e+07, 1.4, 20800.0])

## Reaction 4384
#reaction('I3C6-2O2 <=> I3C6OOH2-5', [7.033000e+10, 0.2, 18500.0])

## Reaction 4385
#reaction('I3C6-2O2 <=> I3C6OOH2-6', [1.233000e+06, 1.5, 20000.0])

## Reaction 4386
#reaction('I3C6-3O2 <=> I3C6OOH3-1', [3.007000e+08, 1.1, 21900.0])

## Reaction 4387
#reaction('I3C6-3O2 <=> I3C6OOH3-2', [1.303000e+10, 0.7, 30100.0])

## Reaction 4388
#reaction('I3C6-3O2 <=> I3C6OOH3-4', [3.014000e+08, 1.2, 33500.0])

## Reaction 4389
#reaction('I3C6-4O2 <=> I3C6OOH4-1', [2.467000e+06, 1.5, 20000.0])

## Reaction 4390
#reaction('I3C6-4O2 <=> I3C6OOH4-2', [2.719000e+07, 1.3, 18200.0])

## Reaction 4391
#reaction('I3C6-4O2 <=> I3C6OOH4-3', [4.618000e+09, 0.8, 27100.0])

## Reaction 4392
#reaction('I3C6D1 + HO2 <=> I3C6-1O2', [1.050000e-01, 3.46, 8900.0])

## Reaction 4393
#reaction('I3C6D1 + HO2 <=> I3C6-2O2', [4.300000e-01, 3.29, 5400.0])

## Reaction 4394
#reaction('I3C6D2 + HO2 <=> I3C6-2O2', [2.160000e-02, 3.68, 8100.0])

## Reaction 4395
#reaction('I3C6D2 + HO2 <=> I3C6-3O2', [7.030000e-02, 3.55, 4800.0])

## Reaction 4396
#reaction('I3C6D3 + HO2 <=> I3C6-3O2', [1.090000e-01, 3.42, 3300.0])

## Reaction 4397
#reaction('I3C6D3 + HO2 <=> I3C6-4O2', [4.980000e-02, 3.53, 11100.0])

## Reaction 4398
#reaction('I3C6-1O2 + HO2 <=> I3C6-1OOH + O2', [1.750000e+10, 0.0, -3275.0])

## Reaction 4399
#reaction('I3C6-2O2 + HO2 <=> I3C6-2OOH + O2', [1.750000e+10, 0.0, -3275.0])

## Reaction 4400
#reaction('I3C6-3O2 + HO2 <=> I3C6-3OOH + O2', [1.750000e+10, 0.0, -3275.0])

## Reaction 4401
#reaction('I3C6-4O2 + HO2 <=> I3C6-4OOH + O2', [1.750000e+10, 0.0, -3275.0])

## Reaction 4402
#reaction('I3C6-1O2 + H2O2 <=> I3C6-1OOH + HO2', [2.400000e+12, 0.0, 10000.0])

## Reaction 4403
#reaction('I3C6-2O2 + H2O2 <=> I3C6-2OOH + HO2', [2.400000e+12, 0.0, 10000.0])

## Reaction 4404
#reaction('I3C6-3O2 + H2O2 <=> I3C6-3OOH + HO2', [2.400000e+12, 0.0, 10000.0])

## Reaction 4405
#reaction('I3C6-4O2 + H2O2 <=> I3C6-4OOH + HO2', [2.400000e+12, 0.0, 10000.0])

## Reaction 4406
#reaction('I3C6-1O2 + I3C6 <=> I3C6-1OOH + I3C6-1', [1.210000e+13, 0.0, 20430.0])

## Reaction 4407
#reaction('I3C6-2O2 + I3C6 <=> I3C6-2OOH + I3C6-1', [1.210000e+13, 0.0, 20430.0])

## Reaction 4408
#reaction('I3C6-3O2 + I3C6 <=> I3C6-3OOH + I3C6-1', [1.210000e+13, 0.0, 20430.0])

## Reaction 4409
#reaction('I3C6-4O2 + I3C6 <=> I3C6-4OOH + I3C6-1', [1.210000e+13, 0.0, 20430.0])

## Reaction 4410
#reaction('I3C6-1O2 + I3C6 <=> I3C6-1OOH + I3C6-2', [2.016000e+12, 0.0, 16000.0])

## Reaction 4411
#reaction('I3C6-2O2 + I3C6 <=> I3C6-2OOH + I3C6-2', [2.016000e+12, 0.0, 16000.0])

## Reaction 4412
#reaction('I3C6-3O2 + I3C6 <=> I3C6-3OOH + I3C6-2', [2.016000e+12, 0.0, 16000.0])

## Reaction 4413
#reaction('I3C6-4O2 + I3C6 <=> I3C6-4OOH + I3C6-2', [2.016000e+12, 0.0, 16000.0])

## Reaction 4414
#reaction('I3C6-1O2 + I3C6 <=> I3C6-1OOH + I3C6-3', [4.032000e+12, 0.0, 17700.0])

## Reaction 4415
#reaction('I3C6-2O2 + I3C6 <=> I3C6-2OOH + I3C6-3', [4.032000e+12, 0.0, 17700.0])

## Reaction 4416
#reaction('I3C6-3O2 + I3C6 <=> I3C6-3OOH + I3C6-3', [4.032000e+12, 0.0, 17700.0])

## Reaction 4417
#reaction('I3C6-4O2 + I3C6 <=> I3C6-4OOH + I3C6-3', [4.032000e+12, 0.0, 17700.0])

## Reaction 4418
#reaction('I3C6-1O2 + I3C6 <=> I3C6-1OOH + I3C6-4', [6.048000e+12, 0.0, 20430.0])

## Reaction 4419
#reaction('I3C6-2O2 + I3C6 <=> I3C6-2OOH + I3C6-4', [6.048000e+12, 0.0, 20430.0])

## Reaction 4420
#reaction('I3C6-3O2 + I3C6 <=> I3C6-3OOH + I3C6-4', [6.048000e+12, 0.0, 20430.0])

## Reaction 4421
#reaction('I3C6-4O2 + I3C6 <=> I3C6-4OOH + I3C6-4', [6.048000e+12, 0.0, 20430.0])

## Reaction 4422
#reaction('I3C6-1O2 + CH3O2 => I3C6-1O + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 4423
#reaction('I3C6-2O2 + CH3O2 => I3C6-2O + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 4424
#reaction('I3C6-3O2 + CH3O2 => I3C6-3O + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 4425
#reaction('I3C6-4O2 + CH3O2 => I3C6-4O + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 4426
#reaction('I3C6-1O2 + I3C6-1O2 => I3C6-1O + I3C6-1O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 4427
#reaction('I3C6-2O2 + I3C6-1O2 => I3C6-2O + I3C6-1O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 4428
#reaction('I3C6-3O2 + I3C6-1O2 => I3C6-3O + I3C6-1O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 4429
#reaction('I3C6-4O2 + I3C6-1O2 => I3C6-4O + I3C6-1O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 4430
#reaction('I3C6-2O2 + I3C6-2O2 => I3C6-2O + I3C6-2O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 4431
#reaction('I3C6-3O2 + I3C6-2O2 => I3C6-3O + I3C6-2O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 4432
#reaction('I3C6-4O2 + I3C6-2O2 => I3C6-4O + I3C6-2O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 4433
#reaction('I3C6-3O2 + I3C6-3O2 => I3C6-3O + I3C6-3O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 4434
#reaction('I3C6-4O2 + I3C6-3O2 => I3C6-4O + I3C6-3O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 4435
#reaction('I3C6-4O2 + I3C6-4O2 => I3C6-4O + I3C6-4O + O2', [1.400000e+16, -1.61, 1860.0])

## Reaction 4436
#reaction('I3C6-1OOH <=> I3C6-1O + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 4437
#reaction('I3C6-2OOH <=> I3C6-2O + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 4438
#reaction('I3C6-3OOH <=> I3C6-3O + OH', [6.000000e+15, 0.0, 41200.0])

## Reaction 4439
#reaction('I3C6-4OOH <=> I3C6-4O + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 4440
#reaction('CH2O + AC5H11 <=> I3C6-1O', [6.250000e+10, 0.0, 11900.0])

## Reaction 4441
#reaction('SC4H9CHO + CH3 <=> I3C6-2O', [7.500000e+10, 0.0, 11900.0])

## Reaction 4442
#reaction('CH3CHO + SC4H9 <=> I3C6-2O', [7.500000e+10, 0.0, 11900.0])

## Reaction 4443
#reaction('C5Y3 + CH3 <=> I3C6-3O', [8.500000e+10, 0.0, 11900.0])

## Reaction 4444
#reaction('C2H5COCH3 + C2H5 <=> I3C6-3O', [8.500000e+10, 0.0, 11900.0])

## Reaction 4445
#reaction('C5H11-3 + CH2O <=> I3C6-4O', [6.250000e+10, 0.0, 11900.0])

## Reaction 4446
#reaction('I3C6OOH1-2 <=> I3C6O1-2 + OH', [1.775000e+12, 0.0, 13376.0])

## Reaction 4447
#reaction('I3C6OOH1-3 <=> I3C6O1-3 + OH', [3.170000e+11, 0.0, 17311.2])

## Reaction 4448
#reaction('I3C6OOH1-4 <=> I3C6O1-4 + OH', [2.770000e+10, 0.0, 10672.1])

## Reaction 4449
#reaction('I3C6OOH1-5 <=> I3C6O1-5 + OH', [2.770000e+10, 0.0, 8676.9])

## Reaction 4450
#reaction('I3C6OOH1-6 <=> I3C6O1-6 + OH', [5.200000e+09, 0.0, 10604.0])

## Reaction 4451
#reaction('I3C6OOH2-1 <=> I3C6O1-2 + OH', [1.775000e+12, 0.0, 15033.9])

## Reaction 4452
#reaction('I3C6OOH2-3 <=> I3C6O2-3 + OH', [1.775000e+12, 0.0, 13037.2])

## Reaction 4453
#reaction('I3C6OOH2-4 <=> I3C6O2-4 + OH', [3.170000e+11, 0.0, 18398.1])

## Reaction 4454
#reaction('I3C6OOH2-5 <=> I3C6O2-5 + OH', [3.170000e+11, 0.0, 17068.0])

## Reaction 4455
#reaction('I3C6OOH2-6 <=> I3C6O1-5 + OH', [2.770000e+10, 0.0, 9748.6])

## Reaction 4456
#reaction('I3C6OOH3-1 <=> I3C6O1-3 + OH', [3.170000e+11, 0.0, 17630.4])

## Reaction 4457
#reaction('I3C6OOH3-2 <=> I3C6O2-3 + OH', [1.775000e+12, 0.0, 12141.7])

## Reaction 4458
#reaction('I3C6OOH3-4 <=> I3C6O3-4 + OH', [1.775000e+12, 0.0, 13811.6])

## Reaction 4459
#reaction('I3C6OOH4-1 <=> I3C6O1-4 + OH', [2.770000e+10, 0.0, 10672.1])

## Reaction 4460
#reaction('I3C6OOH4-2 <=> I3C6O2-4 + OH', [3.170000e+11, 0.0, 17683.6])

## Reaction 4461
#reaction('I3C6OOH4-3 <=> I3C6O3-4 + OH', [1.775000e+12, 0.0, 13049.3])

## Reaction 4462
#reaction('I3C6D1 + HO2 <=> I3C6OOH1-2', [7.910000e+02, 2.78, 9500.0])

## Reaction 4463
#reaction('I3C6D1 + HO2 <=> I3C6OOH2-1', [1.060000e+01, 3.29, 9100.0])

## Reaction 4464
#reaction('I3C6D2 + HO2 <=> I3C6OOH2-3', [1.860000e+02, 2.95, 5400.0])

## Reaction 4465
#reaction('I3C6D2 + HO2 <=> I3C6OOH3-2', [1.720000e-01, 3.7, 4700.0])

## Reaction 4466
#reaction('I3C6D3 + HO2 <=> I3C6OOH3-4', [3.370000e-01, 3.67, 7200.0])

## Reaction 4467
#reaction('I3C6D3 + HO2 <=> I3C6OOH4-3', [1.350000e+03, 2.67, 7900.0])

## Reaction 4468
#reaction('I3C6OOH1-3 => AC5H10 + CH2O + OH', [3.081000e+13, 0.0, 27055.0])

## Reaction 4469
#reaction('I3C6OOH1-4 => C2H4 + C4H8-1 + HO2', [1.392000e+14, 0.0, 30978.0])

## Reaction 4470
#reaction('I3C6OOH1-5 => C2H4 + C4H8-2 + HO2', [1.392000e+14, 0.0, 31359.0])

## Reaction 4471
#reaction('I3C6OOH2-4 => CH3CHO + C4H8-1 + OH', [3.081000e+13, 0.0, 23547.0])

## Reaction 4472
#reaction('I3C6OOH2-5 => CH3CHO + C4H8-2 + OH', [3.081000e+13, 0.0, 23764.0])

## Reaction 4473
#reaction('I3C6OOH2-6 => C2H4 + C4H8-2 + HO2', [1.392000e+14, 0.0, 32233.0])

## Reaction 4474
#reaction('I3C6OOH3-1 => C2H5COCH3 + C2H4 + OH', [3.081000e+13, 0.0, 21905.0])

## Reaction 4475
#reaction('I3C6OOH4-1 => C4H8-1 + C2H4 + HO2', [1.392000e+14, 0.0, 30978.0])

## Reaction 4476
#reaction('I3C6OOH4-2 => C5H10-2 + CH2O + OH', [3.081000e+13, 0.0, 26239.0])

## Reaction 4477
#reaction('I3C6OOH1-2 + O2 <=> I3C6OOH1-2O2', [2.788000e+14, -0.816, -536.5])

## Reaction 4478
#reaction('I3C6OOH1-3 + O2 <=> I3C6OOH1-3O2', [7.802000e+11, 0.325, -417.3])

## Reaction 4479
#reaction('I3C6OOH1-4 + O2 <=> I3C6OOH1-4O2', [5.490000e+16, -1.627, 198.7])

## Reaction 4480
#reaction('I3C6OOH1-5 + O2 <=> I3C6OOH1-5O2', [2.788000e+14, -0.816, -536.5])

## Reaction 4481
#reaction('I3C6OOH1-6 + O2 <=> I3C6OOH1-6O2', [5.490000e+16, -1.627, 198.7])

## Reaction 4482
#reaction('I3C6OOH2-1 + O2 <=> I3C6OOH2-1O2', [5.490000e+16, -1.627, 198.7])

## Reaction 4483
#reaction('I3C6OOH2-3 + O2 <=> I3C6OOH2-3O2', [7.802000e+11, 0.325, -417.3])

## Reaction 4484
#reaction('I3C6OOH2-4 + O2 <=> I3C6OOH2-4O2', [5.490000e+16, -1.627, 198.7])

## Reaction 4485
#reaction('I3C6OOH2-5 + O2 <=> I3C6OOH2-5O2', [2.788000e+14, -0.816, -536.5])

## Reaction 4486
#reaction('I3C6OOH2-6 + O2 <=> I3C6OOH2-6O2', [5.490000e+16, -1.627, 198.7])

## Reaction 4487
#reaction('I3C6OOH3-1 + O2 <=> I3C6OOH3-1O2', [5.490000e+16, -1.627, 198.7])

## Reaction 4488
#reaction('I3C6OOH3-2 + O2 <=> I3C6OOH3-2O2', [2.788000e+14, -0.816, -536.5])

## Reaction 4489
#reaction('I3C6OOH3-4 + O2 <=> I3C6OOH3-4O2', [5.490000e+16, -1.627, 198.7])

## Reaction 4490
#reaction('I3C6OOH4-1 + O2 <=> I3C6OOH4-1O2', [5.490000e+16, -1.627, 198.7])

## Reaction 4491
#reaction('I3C6OOH4-2 + O2 <=> I3C6OOH4-2O2', [2.788000e+14, -0.816, -536.5])

## Reaction 4492
#reaction('I3C6OOH4-3 + O2 <=> I3C6OOH4-3O2', [7.802000e+11, 0.325, -417.3])

## Reaction 4493
#reaction('I3C6OOH1-2O2 <=> I3C6KET1-2 + OH', [2.441000e+07, 1.6, 27900.0])

## Reaction 4494
#reaction('I3C6OOH1-3O2 <=> I3C6KET1-3 + OH', [1.098000e+04, 2.4, 19900.0])

## Reaction 4495
#reaction('I3C6OOH1-4O2 <=> I3C6KET1-4 + OH', [4.766000e+03, 1.7, 16600.0])

## Reaction 4496
#reaction('I3C6OOH1-5O2 <=> I3C6KET1-5 + OH', [1.256000e+03, 2.2, 17400.0])

## Reaction 4497
#reaction('I3C6OOH1-6O2 <=> I3C6KET1-6 + OH', [1.115000e+03, 1.8, 16600.0])

## Reaction 4498
#reaction('I3C6OOH2-1O2 <=> I3C6KET2-1 + OH', [2.763000e+08, 1.2, 25700.0])

## Reaction 4499
#reaction('I3C6OOH2-3O2 <=> I3C6KET2-3 + OH', [1.748000e+08, 1.7, 26000.0])

## Reaction 4500
#reaction('I3C6OOH2-4O2 <=> I3C6KET2-4 + OH', [5.786000e+01, 2.9, 17000.0])

## Reaction 4501
#reaction('I3C6OOH2-5O2 <=> I3C6KET2-5 + OH', [1.754000e+02, 3.1, 17500.0])

## Reaction 4502
#reaction('I3C6OOH2-6O2 <=> I3C6KET2-6 + OH', [3.768000e+02, 2.2, 15300.0])

## Reaction 4503
#reaction('I3C6OOH4-1O2 <=> I3C6KET4-1 + OH', [4.766000e+03, 1.7, 16600.0])

## Reaction 4504
#reaction('I3C6OOH4-2O2 <=> I3C6KET4-2 + OH', [1.098000e+04, 2.4, 19900.0])

## Reaction 4505
#reaction('I3C6OOH4-3O2 <=> I3C6KET4-3 + OH', [2.441000e+07, 1.6, 27900.0])

## Reaction 4506
#reaction('I3C6D1-3OOH + HO2 <=> I3C6OOH3-1O2', [1.050000e-01, 0.0, 18291.7])

## Reaction 4507
#reaction('I3C6D1-3OOH + HO2 <=> I3C6OOH3-2O2', [4.300000e-01, 0.0, 10501.1])

## Reaction 4508
#reaction('I3C6D1-4OOH + HO2 <=> I3C6OOH4-1O2', [1.050000e-01, 0.0, 15033.9])

## Reaction 4509
#reaction('I3C6D1-4OOH + HO2 <=> I3C6OOH4-2O2', [4.300000e-01, 0.0, 13363.9])

## Reaction 4510
#reaction('I3C6D1-5OOH + HO2 <=> I3C6OOH2-5O2', [4.300000e-01, 3.29, 5400.0])

## Reaction 4511
#reaction('I3C6D1-5OOH + HO2 <=> I3C6OOH2-6O2', [1.050000e-01, 3.46, 8900.0])

## Reaction 4512
#reaction('I3C6D1-6OOH + HO2 <=> I3C6OOH1-5O2', [4.300000e-01, 3.29, 5400.0])

## Reaction 4513
#reaction('I3C6D1-6OOH + HO2 <=> I3C6OOH1-6O2', [1.050000e-01, 0.0, 13376.0])

## Reaction 4514
#reaction('I3C6D2-1OOH + HO2 <=> I3C6OOH1-2O2', [2.160000e-02, 3.68, 8100.0])

## Reaction 4515
#reaction('I3C6D2-1OOH + HO2 <=> I3C6OOH1-3O2', [7.030000e-02, 3.55, 4800.0])

## Reaction 4516
#reaction('I3C6D2-4OOH + HO2 <=> I3C6OOH4-2O2', [2.160000e-02, 0.0, 18284.1])

## Reaction 4517
#reaction('I3C6D2-4OOH + HO2 <=> I3C6OOH4-3O2', [7.030000e-02, 0.0, 10489.7])

## Reaction 4518
#reaction('I3C6D2-5OOH + HO2 <=> I3C6OOH2-3O2', [7.030000e-02, 3.55, 4800.0])

## Reaction 4519
#reaction('I3C6D2-5OOH + HO2 <=> I3C6OOH2-5O2', [2.160000e-02, 3.68, 8100.0])

## Reaction 4520
#reaction('I3C6D2-6OOH + HO2 <=> I3C6OOH1-3O2', [7.030000e-02, 3.55, 4800.0])

## Reaction 4521
#reaction('I3C6D2-6OOH + HO2 <=> I3C6OOH1-5O2', [2.160000e-02, 3.68, 8100.0])

## Reaction 4522
#reaction('I3C6D3-1OOH + HO2 <=> I3C6OOH1-3O2', [1.090000e-01, 3.42, 3300.0])

## Reaction 4523
#reaction('I3C6D3-1OOH + HO2 <=> I3C6OOH1-4O2', [4.980000e-02, 3.53, 11100.0])

## Reaction 4524
#reaction('I3C6D3-2OOH + HO2 <=> I3C6OOH2-3O2', [1.090000e-01, 3.42, 3300.0])

## Reaction 4525
#reaction('I3C6D3-2OOH + HO2 <=> I3C6OOH2-4O2', [4.980000e-02, 3.53, 11100.0])

## Reaction 4526
#reaction('I3C6OOH1-2O2 <=> I3C6Q12-3', [7.449000e+09, 0.0, 11128.8])

## Reaction 4527
#reaction('I3C6OOH1-2O2 <=> I3C6Q12-4', [1.439000e+07, 0.0, 19006.2])

## Reaction 4528
#reaction('I3C6OOH1-2O2 <=> I3C6Q12-5', [7.033000e+10, 0.0, 13363.9])

## Reaction 4529
#reaction('I3C6OOH1-2O2 <=> I3C6Q12-6', [1.233000e+06, 0.0, 13363.9])

## Reaction 4530
#reaction('I3C6OOH1-3O2 <=> I3C6Q13-2', [6.513000e+09, 0.0, 18284.1])

## Reaction 4531
#reaction('I3C6OOH1-3O2 <=> I3C6Q13-4', [1.673000e+08, 0.0, 10489.7])

## Reaction 4532
#reaction('I3C6OOH1-3O2 <=> I3C6Q13-5', [6.513000e+09, 0.0, 11907.9])

## Reaction 4533
#reaction('I3C6OOH1-3O2 <=> I3C6Q13-6', [1.504000e+08, 0.0, 11572.8])

## Reaction 4534
#reaction('I3C6OOH1-4O2 <=> I3C6Q14-2', [1.360000e+07, 0.0, 18284.1])

## Reaction 4535
#reaction('I3C6OOH1-4O2 <=> I3C6Q14-3', [2.309000e+09, 0.0, 13363.9])

## Reaction 4536
#reaction('I3C6OOH1-4O2 <=> I3C6Q14-5', [1.360000e+07, 1.3, 18200.0])

## Reaction 4537
#reaction('I3C6OOH1-4O2 <=> I3C6Q14-6', [1.233000e+06, 1.5, 20000.0])

## Reaction 4538
#reaction('I3C6OOH1-5O2 <=> I3C6Q15-2', [7.033000e+10, 0.2, 18500.0])

## Reaction 4539
#reaction('I3C6OOH1-5O2 <=> I3C6Q15-3', [7.449000e+09, 0.6, 27300.0])

## Reaction 4540
#reaction('I3C6OOH1-5O2 <=> I3C6Q15-4', [1.439000e+07, 1.4, 20800.0])

## Reaction 4541
#reaction('I3C6OOH1-5O2 <=> I3C6Q15-6', [2.431000e+08, 1.1, 33500.0])

## Reaction 4542
#reaction('I3C6OOH1-6O2 <=> I3C6Q16-2', [3.718000e+06, 1.2, 16600.0],
        #options='duplicate')

## Reaction 4543
#reaction('I3C6OOH1-6O2 <=> I3C6Q16-3', [1.231000e+07, 1.2, 15400.0])

## Reaction 4544
#reaction('I3C6OOH1-6O2 <=> I3C6Q16-4', [1.233000e+06, 1.5, 20000.0])

## Reaction 4545
#reaction('I3C6OOH1-6O2 <=> I3C6Q16-2', [4.009000e+08, 1.1, 30100.0],
        #options='duplicate')

## Reaction 4546
#reaction('I3C6OOH2-1O2 <=> I3C6Q12-3', [1.231000e+07, 1.2, 15400.0])

## Reaction 4547
#reaction('I3C6OOH2-1O2 <=> I3C6Q12-4', [1.233000e+06, 1.5, 20000.0])

## Reaction 4548
#reaction('I3C6OOH2-1O2 <=> I3C6Q12-5', [3.718000e+06, 1.2, 16600.0])

## Reaction 4549
#reaction('I3C6OOH2-1O2 <=> I3C6Q12-6', [2.941000e+05, 1.5, 19900.0])

## Reaction 4550
#reaction('I3C6OOH2-3O2 <=> I3C6Q23-1', [1.504000e+08, 1.1, 21900.0])

## Reaction 4551
#reaction('I3C6OOH2-3O2 <=> I3C6Q23-4', [1.673000e+08, 1.2, 33500.0])

## Reaction 4552
#reaction('I3C6OOH2-3O2 <=> I3C6Q23-5', [6.513000e+09, 0.7, 30100.0])

## Reaction 4553
#reaction('I3C6OOH2-3O2 <=> I3C6Q23-6', [1.504000e+08, 1.1, 21900.0])

## Reaction 4554
#reaction('I3C6OOH2-4O2 <=> I3C6Q24-1', [1.233000e+06, 1.5, 20000.0])

## Reaction 4555
#reaction('I3C6OOH2-4O2 <=> I3C6Q24-3', [2.309000e+09, 0.8, 27100.0])

## Reaction 4556
#reaction('I3C6OOH2-4O2 <=> I3C6Q24-5', [1.360000e+07, 1.3, 18200.0])

## Reaction 4557
#reaction('I3C6OOH2-4O2 <=> I3C6Q24-6', [1.233000e+06, 1.5, 20000.0])

## Reaction 4558
#reaction('I3C6OOH2-5O2 <=> I3C6Q25-1', [1.233000e+06, 1.5, 20000.0],
        #options='duplicate')

## Reaction 4559
#reaction('I3C6OOH2-5O2 <=> I3C6Q25-3', [7.449000e+09, 0.6, 27300.0])

## Reaction 4560
#reaction('I3C6OOH2-5O2 <=> I3C6Q25-4', [1.439000e+07, 1.4, 20800.0])

## Reaction 4561
#reaction('I3C6OOH2-5O2 <=> I3C6Q25-1', [2.431000e+08, 1.1, 33500.0],
        #options='duplicate')

## Reaction 4562
#reaction('I3C6OOH2-6O2 <=> I3C6Q15-6', [2.941000e+05, 1.5, 19900.0])

## Reaction 4563
#reaction('I3C6OOH2-6O2 <=> I3C6Q15-3', [1.231000e+07, 1.2, 15400.0])

## Reaction 4564
#reaction('I3C6OOH2-6O2 <=> I3C6Q15-4', [1.233000e+06, 1.5, 20000.0])

## Reaction 4565
#reaction('I3C6OOH2-6O2 <=> I3C6Q15-2', [4.009000e+08, 1.1, 30100.0])

## Reaction 4566
#reaction('I3C6OOH3-1O2 <=> I3C6Q13-2', [4.009000e+08, 1.1, 30100.0])

## Reaction 4567
#reaction('I3C6OOH3-1O2 <=> I3C6Q13-4', [1.233000e+06, 1.5, 20000.0])

## Reaction 4568
#reaction('I3C6OOH3-1O2 <=> I3C6Q13-5', [3.718000e+06, 1.2, 16600.0])

## Reaction 4569
#reaction('I3C6OOH3-1O2 <=> I3C6Q13-6', [2.941000e+05, 1.5, 19900.0])

## Reaction 4570
#reaction('I3C6OOH3-2O2 <=> I3C6Q23-1', [2.431000e+08, 1.1, 33500.0])

## Reaction 4571
#reaction('I3C6OOH3-2O2 <=> I3C6Q23-4', [1.439000e+07, 1.4, 20800.0])

## Reaction 4572
#reaction('I3C6OOH3-2O2 <=> I3C6Q23-5', [7.033000e+10, 0.2, 18500.0])

## Reaction 4573
#reaction('I3C6OOH3-2O2 <=> I3C6Q23-6', [1.233000e+06, 1.5, 20000.0])

## Reaction 4574
#reaction('I3C6OOH3-4O2 <=> I3C6Q34-1', [2.467000e+06, 1.5, 20000.0])

## Reaction 4575
#reaction('I3C6OOH3-4O2 <=> I3C6Q34-2', [2.719000e+07, 1.3, 18200.0])

## Reaction 4576
#reaction('I3C6OOH4-1O2 <=> I3C6Q14-2', [4.009000e+08, 1.1, 30100.0])

## Reaction 4577
#reaction('I3C6OOH4-1O2 <=> I3C6Q14-3', [1.231000e+07, 1.2, 15400.0])

## Reaction 4578
#reaction('I3C6OOH4-1O2 <=> I3C6Q14-5', [3.718000e+06, 1.2, 16600.0])

## Reaction 4579
#reaction('I3C6OOH4-1O2 <=> I3C6Q14-6', [2.941000e+05, 1.5, 19900.0])

## Reaction 4580
#reaction('I3C6OOH4-2O2 <=> I3C6Q24-1', [2.431000e+08, 1.1, 33500.0])

## Reaction 4581
#reaction('I3C6OOH4-2O2 <=> I3C6Q24-3', [7.449000e+09, 0.6, 27300.0])

## Reaction 4582
#reaction('I3C6OOH4-2O2 <=> I3C6Q24-5', [7.033000e+10, 0.2, 18500.0])

## Reaction 4583
#reaction('I3C6OOH4-2O2 <=> I3C6Q24-6', [1.233000e+06, 1.5, 20000.0])

## Reaction 4584
#reaction('I3C6OOH4-3O2 <=> I3C6Q34-1', [3.007000e+08, 1.1, 21900.0])

## Reaction 4585
#reaction('I3C6OOH4-3O2 <=> I3C6Q34-2', [1.303000e+10, 0.7, 30100.0])

## Reaction 4586
#reaction('I3C6Q12-4 => HO2CH2CHO + C4H8-1 + OH', [3.081000e+13, 0.0, 2608.0])

## Reaction 4587
#reaction('I3C6Q12-5 => HO2CH2CHO + C4H8-2 + OH', [3.081000e+13, 0.0, 2825.0])

## Reaction 4588
#reaction('I3C6Q13-6 => NC4KET24 + C2H4 + OH', [3.081000e+13, 0.0, 21905.0])

## Reaction 4589
#reaction('I3C6Q14-5 => C5H92-5OOH + CH2O + OH', [3.081000e+13, 0.0, 28873.0])

## Reaction 4590
#reaction('I3C6Q15-4 => C4H71-4OOH + CH3CHO + OH', [3.081000e+13, 0.0, 23298.0])

## Reaction 4591
#reaction('I3C6Q16-3 => C5H9A-DOOH + CH2O + OH', [3.081000e+13, 0.0, 26408.0])

## Reaction 4592
#reaction('I3C6Q16-4 => C4H71-4OOH + CH3CHO + OH', [1.392000e+14, 0.0, -12511.0])

## Reaction 4593
#reaction('I3C6Q23-6 => NC4KET23 + C2H4 + OH', [3.081000e+13, 0.0, 21905.0])

## Reaction 4594
#reaction('I3C6Q24-5 => C4H72-1OOH + CH3CHO + OH', [3.081000e+13, 0.0, 25333.0])

## Reaction 4595
#reaction('I3C6Q24-5 => C5H92-4OOH + CH2O + OH', [3.081000e+13, 0.0, 26175.0])

## Reaction 4596
#reaction('I3C6Q25-4 => C4H71-3OOH + CH3CHO + OH', [3.081000e+13, 0.0, 23478.0])

## Reaction 4597
#reaction('I3C6Q34-1 => NC4KET21 + C2H4 + OH', [3.081000e+13, 0.0, 23787.0])

## Reaction 4598
#reaction('I3C6KET1-2 => I3C6KET1-2O + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 4599
#reaction('I3C6KET1-3 => I3C6KET1-3O + OH', [6.000000e+15, 0.0, 41200.0])

## Reaction 4600
#reaction('I3C6KET1-4 => I3C6KET1-4O + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 4601
#reaction('I3C6KET1-5 => I3C6KET1-5O + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 4602
#reaction('I3C6KET1-6 => I3C6KET1-6O + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 4603
#reaction('I3C6KET2-1 => I3C6KET2-1O + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 4604
#reaction('I3C6KET2-3 => I3C6KET2-3O + OH', [6.000000e+15, 0.0, 41200.0])

## Reaction 4605
#reaction('I3C6KET2-4 => I3C6KET2-4O + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 4606
#reaction('I3C6KET2-5 => I3C6KET2-5O + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 4607
#reaction('I3C6KET2-6 => I3C6KET2-6O + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 4608
#reaction('I3C6KET4-1 => I3C6KET4-1O + OH', [1.500000e+16, 0.0, 42000.0])

## Reaction 4609
#reaction('I3C6KET4-2 => I3C6KET4-2O + OH', [1.050000e+16, 0.0, 41600.0])

## Reaction 4610
#reaction('I3C6KET4-3 => I3C6KET4-3O + OH', [1.200000e+16, 0.0, 41200.0])

## Reaction 4611
#reaction('I3C6KET1-2O => HCO + SC4H9CHO', [2.900000e+12, 0.38, 11080.0])

## Reaction 4612
#reaction('I3C6KET1-3O => CH2CHO + C2H5COCH3', [2.900000e+12, 0.38, 11080.0])

## Reaction 4613
#reaction('I3C6KET1-4O => PC4H8CHO-2 + CH2O', [2.900000e+12, 0.38, 11080.0])

## Reaction 4614
#reaction('I3C6KET1-5O => C3H6CHO-2 + CH3CHO', [2.900000e+12, 0.38, 11080.0])

## Reaction 4615
#reaction('I3C6KET1-6O => IC4H8CHO-I + CH2O', [2.900000e+12, 0.38, 11080.0])

## Reaction 4616
#reaction('I3C6KET2-1O => SC4H9CO + CH2O', [2.900000e+12, 0.38, 11080.0])

## Reaction 4617
#reaction('I3C6KET2-3O => CH3CO + C2H5COCH3', [2.900000e+12, 0.38, 11080.0])

## Reaction 4618
#reaction('I3C6KET2-4O => C5Y2-3J + CH2O', [2.900000e+12, 0.38, 11080.0])

## Reaction 4619
#reaction('I3C6KET2-5O => CH3CHCOCH3 + CH3CHO', [2.900000e+12, 0.38, 11080.0])

## Reaction 4620
#reaction('I3C6KET2-6O => IC5Y3-1J + CH2O', [2.900000e+12, 0.38, 11080.0])

## Reaction 4621
#reaction('I3C6KET4-1O => SC4H8CHO-1 + CH2O', [2.900000e+12, 0.38, 11080.0])

## Reaction 4622
#reaction('I3C6KET4-2O => C3H6CHO-3 + CH3CHO', [2.900000e+12, 0.38, 11080.0])

## Reaction 4623
#reaction('I3C6KET4-3O => HCO + C5Y3', [5.800000e+12, 0.38, 11080.0])

## Reaction 4624
#reaction('I3C6D2-1OOH + HO2 <=> I3C6Q12-3', [1.860000e+02, 2.95, 5400.0])

## Reaction 4625
#reaction('I3C6D2-1OOH + HO2 <=> I3C6Q13-2', [1.720000e-01, 3.7, 4700.0])

## Reaction 4626
#reaction('I3C6D1-3OOH + HO2 <=> I3C6Q13-2', [7.910000e+02, 2.78, 9500.0])

## Reaction 4627
#reaction('I3C6D3-1OOH + HO2 <=> I3C6Q13-4', [3.370000e-01, 3.67, 7200.0])

## Reaction 4628
#reaction('I3C6D2-6OOH + HO2 <=> I3C6Q13-5', [1.720000e-01, 3.7, 4700.0])

## Reaction 4629
#reaction('I3C6D1-4OOH + HO2 <=> I3C6Q14-2', [7.910000e+02, 2.78, 9500.0])

## Reaction 4630
#reaction('I3C6D3-1OOH + HO2 <=> I3C6Q14-3', [1.350000e+03, 2.67, 7900.0])

## Reaction 4631
#reaction('I3C6D1-5OOH + HO2 <=> I3C6Q15-2', [7.910000e+02, 2.78, 9500.0])

## Reaction 4632
#reaction('I3C6D2-6OOH + HO2 <=> I3C6Q15-3', [1.860000e+02, 2.95, 5400.0])

## Reaction 4633
#reaction('I3C6D1-6OOH + HO2 <=> I3C6Q15-6', [1.060000e+01, 3.29, 9100.0])

## Reaction 4634
#reaction('I3C6D1-6OOH + HO2 <=> I3C6Q16-2', [7.910000e+02, 2.78, 9500.0])

## Reaction 4635
#reaction('I3C6D1-3OOH + HO2 <=> I3C6Q23-1', [1.060000e+01, 3.29, 9100.0])

## Reaction 4636
#reaction('I3C6D3-2OOH + HO2 <=> I3C6Q23-4', [3.370000e-01, 3.67, 7200.0])

## Reaction 4637
#reaction('I3C6D2-5OOH + HO2 <=> I3C6Q23-5', [1.720000e-01, 3.7, 4700.0])

## Reaction 4638
#reaction('I3C6D1-4OOH + HO2 <=> I3C6Q24-1', [1.060000e+01, 3.29, 9100.0])

## Reaction 4639
#reaction('I3C6D2-4OOH + HO2 <=> I3C6Q24-3', [1.860000e+02, 2.95, 5400.0])

## Reaction 4640
#reaction('I3C6D3-2OOH + HO2 <=> I3C6Q24-3', [1.350000e+03, 2.67, 7900.0])

## Reaction 4641
#reaction('I3C6D1-5OOH + HO2 <=> I3C6Q25-1', [1.060000e+01, 3.29, 9100.0])

## Reaction 4642
#reaction('I3C6D2-5OOH + HO2 <=> I3C6Q25-3', [1.860000e+02, 2.95, 5400.0])

## Reaction 4643
#reaction('I3C6D2-4OOH + HO2 <=> I3C6Q34-2', [1.720000e-01, 3.7, 4700.0])

## Reaction 4644
#reaction('I3C6D1-3OOH => C2H3COCH3 + C2H5 + OH', [7.180000e+17, -1.2, 28400.0])

## Reaction 4645
#reaction('I3C6D1-3OOH => C2H3 + C2H5COCH3 + OH', [7.180000e+17, -1.2, 28400.0])

## Reaction 4646
#reaction('I3C6D3-1OOH => AC5H9-A2 + CH2O + OH', [6.880000e+14, -0.89, 7260.0])

## Reaction 4647
#reaction('I3C6D2-6OOH => AC5H9-C + CH2O + OH', [6.880000e+14, -0.89, 7260.0])

## Reaction 4648
#reaction('I3C6D1-4OOH => C5H91-3 + CH2O + OH', [4.470000e+15, -1.13, 4140.0])

## Reaction 4649
#reaction('I3C6D1-5OOH => C4H71-3 + CH3CHO + OH', [4.470000e+15, -1.13, 4140.0])

## Reaction 4650
#reaction('I3C6D1-6OOH => CC5H9-A + CH2O + OH', [1.750000e+15, -1.09, 4420.0])

## Reaction 4651
#reaction('I3C6D3-2OOH => C4H71-2 + CH3CHO + OH', [3.231000e+22, -2.63, 30310.0])

## Reaction 4652
#reaction('I3C6D2-5OOH => C4H72-2 + CH3CHO + OH', [3.231000e+22, -2.63, 30310.0])

## Reaction 4653
#reaction('I3C6Q12-3 <=> I3C6O23-1OOH + OH', [1.775000e+12, 0.0, 12069.1])

## Reaction 4654
#reaction('I3C6Q12-3 <=> I3C6O13-2OOH + OH', [3.170000e+11, 0.0, 17189.6])

## Reaction 4655
#reaction('I3C6Q12-4 <=> I3C6O24-1OOH + OH', [3.170000e+11, 0.0, 18398.1])

## Reaction 4656
#reaction('I3C6Q12-4 <=> I3C6O14-2OOH + OH', [2.770000e+10, 0.0, 10672.1])

## Reaction 4657
#reaction('I3C6Q12-5 <=> I3C6O25-1OOH + OH', [3.170000e+11, 0.0, 17068.0])

## Reaction 4658
#reaction('I3C6Q12-5 <=> I3C6O15-2OOH + OH', [2.770000e+10, 0.0, 8676.9])

## Reaction 4659
#reaction('I3C6Q12-6 <=> I3C6O15-6OOH + OH', [2.770000e+10, 0.0, 9748.6])

## Reaction 4660
#reaction('I3C6Q12-6 <=> I3C6O16-2OOH + OH', [5.200000e+09, 0.0, 10604.0])

## Reaction 4661
#reaction('I3C6Q13-2 <=> I3C6O23-1OOH + OH', [1.775000e+12, 0.0, 11173.6])

## Reaction 4662
#reaction('I3C6Q13-2 <=> I3C6O12-3OOH + OH', [1.775000e+12, 0.0, 13376.0])

## Reaction 4663
#reaction('I3C6Q13-4 <=> I3C6O34-1OOH + OH', [1.775000e+12, 0.0, 13811.6])

## Reaction 4664
#reaction('I3C6Q13-4 <=> I3C6O14-3OOH + OH', [2.770000e+10, 0.0, 10729.1])

## Reaction 4665
#reaction('I3C6Q13-5 <=> I3C6O23-6OOH + OH', [1.775000e+12, 0.0, 11633.4])

## Reaction 4666
#reaction('I3C6Q13-5 <=> I3C6O15-3OOH + OH', [2.770000e+10, 0.0, 8676.9])

## Reaction 4667
#reaction('I3C6Q13-6 <=> I3C6O13-6OOH + OH', [3.170000e+11, 0.0, 17630.4])

## Reaction 4668
#reaction('I3C6Q13-6 <=> I3C6O16-3OOH + OH', [5.200000e+09, 0.0, 10604.0])

## Reaction 4669
#reaction('I3C6Q14-2 <=> I3C6O24-1OOH + OH', [3.170000e+11, 0.0, 21522.0])

## Reaction 4670
#reaction('I3C6Q14-2 <=> I3C6O12-4OOH + OH', [1.775000e+12, 0.0, 19947.0])

## Reaction 4671
#reaction('I3C6Q14-3 <=> I3C6O34-1OOH + OH', [1.775000e+12, 0.0, 19620.2])

## Reaction 4672
#reaction('I3C6Q14-3 <=> I3C6O13-4OOH + OH', [3.170000e+11, 0.0, 21316.8])

## Reaction 4673
#reaction('I3C6Q14-5 <=> I3C6O24-6OOH + OH', [3.170000e+11, 0.0, 21810.9])

## Reaction 4674
#reaction('I3C6Q14-5 <=> I3C6O15-4OOH + OH', [2.770000e+10, 0.0, 14867.7])

## Reaction 4675
#reaction('I3C6Q14-6 <=> I3C6O14-6OOH + OH', [2.770000e+10, 0.0, 16862.9])

## Reaction 4676
#reaction('I3C6Q14-6 <=> I3C6O16-4OOH + OH', [5.200000e+09, 0.0, 15057.0])

## Reaction 4677
#reaction('I3C6Q15-2 <=> I3C6O25-1OOH + OH', [3.170000e+11, 0.0, 16779.2])

## Reaction 4678
#reaction('I3C6Q15-2 <=> I3C6O12-5OOH + OH', [1.775000e+12, 0.0, 13376.0])

## Reaction 4679
#reaction('I3C6Q15-3 <=> I3C6O23-6OOH + OH', [1.775000e+12, 0.0, 12069.1])

## Reaction 4680
#reaction('I3C6Q15-3 <=> I3C6O13-5OOH + OH', [3.170000e+11, 0.0, 17189.6])

## Reaction 4681
#reaction('I3C6Q15-4 <=> I3C6O24-6OOH + OH', [3.170000e+11, 0.0, 18398.1])

## Reaction 4682
#reaction('I3C6Q15-4 <=> I3C6O14-5OOH + OH', [2.770000e+10, 0.0, 10672.1])

## Reaction 4683
#reaction('I3C6Q15-6 <=> I3C6O12-6OOH + OH', [1.775000e+12, 0.0, 15033.9])

## Reaction 4684
#reaction('I3C6Q15-6 <=> I3C6O16-2OOH + OH', [5.200000e+09, 0.0, 10645.0])

## Reaction 4685
#reaction('I3C6Q16-2 <=> I3C6O15-6OOH + OH', [2.770000e+10, 0.0, 8243.6])

## Reaction 4686
#reaction('I3C6Q16-2 <=> I3C6O12-6OOH + OH', [1.775000e+12, 0.0, 13376.0])

## Reaction 4687
#reaction('I3C6Q16-3 <=> I3C6O13-6OOH + OH', [3.170000e+11, 0.0, 17311.2])

## Reaction 4688
#reaction('I3C6Q16-4 <=> I3C6O14-6OOH + OH', [2.770000e+10, 0.0, 10672.1])

## Reaction 4689
#reaction('I3C6Q23-1 <=> I3C6O13-2OOH + OH', [3.170000e+11, 0.0, 17668.4])

## Reaction 4690
#reaction('I3C6Q23-1 <=> I3C6O12-3OOH + OH', [1.775000e+12, 0.0, 15033.9])

## Reaction 4691
#reaction('I3C6Q23-4 <=> I3C6O34-2OOH + OH', [1.775000e+12, 0.0, 13751.1])

## Reaction 4692
#reaction('I3C6Q23-4 <=> I3C6O24-3OOH + OH', [3.170000e+11, 0.0, 18398.1])

## Reaction 4693
#reaction('I3C6Q23-5 <=> I3C6O23-5OOH + OH', [1.775000e+12, 0.0, 11173.6])

## Reaction 4694
#reaction('I3C6Q23-5 <=> I3C6O25-3OOH + OH', [3.170000e+11, 0.0, 16779.2])

## Reaction 4695
#reaction('I3C6Q23-6 <=> I3C6O13-5OOH + OH', [3.170000e+11, 0.0, 17630.4])

## Reaction 4696
#reaction('I3C6Q23-6 <=> I3C6O15-3OOH + OH', [2.770000e+10, 0.0, 9748.6])

## Reaction 4697
#reaction('I3C6Q24-1 <=> I3C6O14-2OOH + OH', [2.770000e+10, 0.0, 10729.1])

## Reaction 4698
#reaction('I3C6Q24-1 <=> I3C6O12-4OOH + OH', [1.775000e+12, 0.0, 15033.9])

## Reaction 4699
#reaction('I3C6Q24-3 <=> I3C6O34-2OOH + OH', [1.775000e+12, 0.0, 13049.3])

## Reaction 4700
#reaction('I3C6Q24-3 <=> I3C6O23-4OOH + OH', [1.775000e+12, 0.0, 12069.1])

## Reaction 4701
#reaction('I3C6Q24-5 <=> I3C6O24-5OOH + OH', [3.170000e+11, 0.0, 17683.6])

## Reaction 4702
#reaction('I3C6Q24-5 <=> I3C6O25-4OOH + OH', [3.170000e+11, 0.0, 17068.0])

## Reaction 4703
#reaction('I3C6Q24-6 <=> I3C6O14-5OOH + OH', [2.770000e+10, 0.0, 10672.1])

## Reaction 4704
#reaction('I3C6Q24-6 <=> I3C6O15-4OOH + OH', [2.770000e+10, 0.0, 9748.6])

## Reaction 4705
#reaction('I3C6Q25-1 <=> I3C6O15-2OOH + OH', [2.770000e+10, 0.0, 9748.6])

## Reaction 4706
#reaction('I3C6Q25-1 <=> I3C6O12-5OOH + OH', [1.775000e+12, 0.0, 14973.4])

## Reaction 4707
#reaction('I3C6Q25-3 <=> I3C6O23-5OOH + OH', [1.775000e+12, 0.0, 12069.1])

## Reaction 4708
#reaction('I3C6Q25-4 <=> I3C6O24-5OOH + OH', [3.170000e+11, 0.0, 18398.1])

## Reaction 4709
#reaction('I3C6Q34-1 <=> I3C6O14-3OOH + OH', [2.770000e+10, 0.0, 10672.1])

## Reaction 4710
#reaction('I3C6Q34-1 <=> I3C6O13-4OOH + OH', [3.170000e+11, 0.0, 17630.4])

## Reaction 4711
#reaction('I3C6Q34-2 <=> I3C6O24-3OOH + OH', [3.170000e+11, 0.0, 17394.8])

## Reaction 4712
#reaction('I3C6Q34-2 <=> I3C6O23-4OOH + OH', [1.775000e+12, 0.0, 11173.6])

## Reaction 4713
#reaction('I3C6O12-3OOH => CH2CHO + C2H5COCH3 + OH', [6.000000e+15, 0.0, 42200.0])

## Reaction 4714
#reaction('I3C6O12-4OOH => C5H9O2-1 + CH2O + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 4715
#reaction('I3C6O12-5OOH => CH3CHO + C4H7O2-1 + OH', [1.050000e+16, 0.0, 42600.0])

## Reaction 4716
#reaction('I3C6O12-6OOH => IC5O34-1J + CH2O + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 4717
#reaction('I3C6O13-2OOH => C2H5COCH3 + CH2CHO + OH', [1.050000e+16, 0.0, 42600.0])

## Reaction 4718
#reaction('I3C6O13-4OOH => C5Y3-1J + CH2O + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 4719
#reaction('I3C6O13-5OOH => CH2CH2COCH3 + CH3CHO + OH', [1.050000e+16, 0.0, 42600.0])

## Reaction 4720
#reaction('I3C6O13-6OOH => AC5H9O-D + CH2O + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 4721
#reaction('I3C6O14-2OOH => C4H8-1 + CH2OCHO + OH', [1.050000e+16, 0.0, 42600.0])

## Reaction 4722
#reaction('I3C6O14-3OOH => C5Y3-1J + CH2O + OH', [6.000000e+15, 0.0, 42200.0])

## Reaction 4723
#reaction('I3C6O14-5OOH => C4H7O1-4 + CH3CHO + OH', [1.050000e+16, 0.0, 42600.0])

## Reaction 4724
#reaction('I3C6O14-6OOH => C4H71-4 + 2 CH2O + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 4725
#reaction('I3C6O15-2OOH => SC4H8CHO-3 + CH2O + OH', [1.050000e+16, 0.0, 42600.0])

## Reaction 4726
#reaction('I3C6O15-3OOH => CH2CH2COCH3 + CH3CHO + OH', [6.000000e+15, 0.0, 42200.0])

## Reaction 4727
#reaction('I3C6O15-6OOH => SC4H8CHO-4 + CH2O + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 4728
#reaction('I3C6O16-2OOH => SC4H8CHO-4 + CH2O + OH', [1.050000e+16, 0.0, 42600.0])

## Reaction 4729
#reaction('I3C6O16-4OOH => C5H9O15-3 + CH2O + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 4730
#reaction('I3C6O23-1OOH => SC4H8CHO-2 + CH2O + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 4731
#reaction('I3C6O23-4OOH => C5Y3-2J + CH2O + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 4732
#reaction('I3C6O23-5OOH => CH3CHCOCH3 + CH3CHO + OH', [1.050000e+16, 0.0, 42600.0])

## Reaction 4733
#reaction('I3C6O23-6OOH => AC5H9O-C + CH2O + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 4734
#reaction('I3C6O24-1OOH => SC4H8CHO-1 + CH2O + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 4735
#reaction('I3C6O24-3OOH => CH3CHO + C2H5COCH2 + OH', [6.000000e+15, 0.0, 42200.0])

## Reaction 4736
#reaction('I3C6O24-5OOH => C4H7O2-1 + CH3CHO + OH', [1.050000e+16, 0.0, 42600.0])

## Reaction 4737
#reaction('I3C6O24-6OOH => C4H71-3 + 2 CH2O + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 4738
#reaction('I3C6O25-1OOH => SC4H8CHO-3 + CH2O + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 4739
#reaction('I3C6O25-3OOH => CH3CHCOCH3 + CH3CHO + OH', [1.200000e+16, 0.0, 42200.0])

## Reaction 4740
#reaction('I3C6O25-4OOH => C5H9O2-4 + CH2O + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 4741
#reaction('I3C6O34-1OOH => AC5H9O-A2 + CH2O + OH', [1.500000e+16, 0.0, 43000.0])

## Reaction 4742
#reaction('I3C6O34-2OOH => CH3CHO + C2H5COCH2 + OH', [1.050000e+16, 0.0, 42600.0])

## Reaction 4743
#reaction('I3C6O1-2 + OH => SC3H5CHO + C2H5 + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 4744
#reaction('I3C6O1-2 + OH => SC4H9 + CH2CO + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 4745
#reaction('I3C6O1-3 + OH => AC5H10 + HCO + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 4746
#reaction('I3C6O1-4 + OH => CH2CHO + C4H8-1 + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 4747
#reaction('I3C6O1-4 + OH => C2H4 + C3H6CHO-3 + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 4748
#reaction('I3C6O1-5 + OH => CH2CHO + C4H8-2 + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 4749
#reaction('I3C6O1-5 + OH => CH3CHCOCH3 + C2H4 + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 4750
#reaction('I3C6O1-6 + OH => C2H4 + C3H6CHO-2 + H2O', [5.000000e+12, 0.0, 0.0])

## Reaction 4751
#reaction('I3C6O2-3 + OH => IC5D1Y3 + CH3 + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 4752
#reaction('I3C6O2-4 + OH => CH3CO + C4H8-1 + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 4753
#reaction('I3C6O2-4 + OH => SC3H5CHO + C2H5 + H2O', [2.500000e+12, 0.0, 0.0])

## Reaction 4754
#reaction('I3C6O2-5 + OH => CH3CO + C4H8-2 + H2O', [5.000000e+12, 0.0, 0.0])

## Reaction 4755
#reaction('I3C6O1-2 + HO2 => SC3H5CHO + C2H5 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 4756
#reaction('I3C6O1-2 + HO2 => SC4H9 + CH2CO + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 4757
#reaction('I3C6O1-3 + HO2 => AC5H10 + HCO + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 4758
#reaction('I3C6O1-4 + HO2 => CH2CHO + C4H8-1 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 4759
#reaction('I3C6O1-4 + HO2 => C2H4 + C3H6CHO-3 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 4760
#reaction('I3C6O1-5 + HO2 => CH2CHO + C4H8-2 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 4761
#reaction('I3C6O1-5 + HO2 => CH3CHCOCH3 + C2H4 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 4762
#reaction('I3C6O1-6 + HO2 => C2H4 + C3H6CHO-2 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 4763
#reaction('I3C6O2-3 + HO2 => IC5D1Y3 + CH3 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 4764
#reaction('I3C6O2-4 + HO2 => CH3CO + C4H8-1 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 4765
#reaction('I3C6O2-4 + HO2 => SC3H5CHO + C2H5 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 4766
#reaction('I3C6O2-5 + HO2 => CH3CO + C4H8-2 + H2O2', [5.000000e+12, 0.0, 17700.0])

## Reaction 4767
#reaction('C3H6 + CH2CHO <=> IC4H8CHO-I', [1.000000e+11, 0.0, 7800.0])

## Reaction 4768
#reaction('C3H6 + CH3CO <=> IC5Y3-1J', [1.000000e+11, 0.0, 6000.0])

## Reaction 4769
#reaction('C5H9O2-1 <=> C4H71-1 + CH2O', [4.320000e+16, -1.0, 29500.0])

## Reaction 4770
#reaction('C3H6 + CH2CHO <=> IC5O34-1J', [1.000000e+11, 0.0, 7800.0])

## Reaction 4771
#reaction('C2H3COCH3 + CH3 <=> C5Y2-3J', [1.000000e+11, 0.0, 7800.0])

## Reaction 4772
#reaction('CH2O + C4H71-4 <=> C5H9O15-3', [1.000000e+11, 0.0, 7800.0])

## Reaction 4773
#reaction('AC5H9O-D <=> IC4H7 + CH2O', [6.880000e+14, -0.89, 7260.0])

## Reaction 4774
#reaction('SC4H9CHO + O2 <=> SC4H9CO + HO2', [1.200000e+05, 2.5, 37560.0])

## Reaction 4775
#reaction('SC4H9CHO + OH <=> SC4H9CO + H2O', [2.000000e+06, 1.8, -1300.0])

## Reaction 4776
#reaction('SC4H9CHO + H <=> SC4H9CO + H2', [4.140000e+09, 1.12, 2320.0])

## Reaction 4777
#reaction('SC4H9CHO + O <=> SC4H9CO + OH', [5.940000e+12, 0.0, 1868.0])

## Reaction 4778
#reaction('SC4H9CHO + HO2 <=> SC4H9CO + H2O2', [4.090000e+04, 2.5, 10200.0])

## Reaction 4779
#reaction('SC4H9CHO + CH3 <=> SC4H9CO + CH4', [2.890000e-03, 4.62, 3210.0])

## Reaction 4780
#reaction('SC4H9CHO + CH3O <=> SC4H9CO + CH3OH', [1.000000e+12, 0.0, 3300.0])

## Reaction 4781
#reaction('SC4H9CHO + CH3O2 <=> SC4H9CO + CH3O2H', [4.090000e+04, 2.5, 10200.0])

## Reaction 4782
#reaction('SC4H9CHO + OH <=> SC4H8CHO-1 + H2O', [5.280000e+09, 0.97, 1586.0])

## Reaction 4783
#reaction('SC4H9CHO + OH <=> SC4H8CHO-2 + H2O', [5.520000e+02, 3.12, -1176.0])

## Reaction 4784
#reaction('SC4H9CHO + OH <=> SC4H8CHO-3 + H2O', [4.680000e+07, 1.61, -35.0])

## Reaction 4785
#reaction('SC4H9CHO + OH <=> SC4H8CHO-4 + H2O', [5.280000e+09, 0.97, 1586.0])

## Reaction 4786
#reaction('SC4H9CHO + HO2 <=> SC4H8CHO-1 + H2O2', [2.379000e+04, 2.55, 16490.0])

## Reaction 4787
#reaction('SC4H9CHO + HO2 <=> SC4H8CHO-2 + H2O2', [3.440000e+12, 0.05, 17880.0])

## Reaction 4788
#reaction('SC4H9CHO + HO2 <=> SC4H8CHO-3 + H2O2', [9.640000e+03, 2.6, 13910.0])

## Reaction 4789
#reaction('SC4H9CHO + HO2 <=> SC4H8CHO-4 + H2O2', [2.379000e+04, 2.55, 16490.0])

## Reaction 4790
#reaction('SC4H9CHO + CH3O2 <=> SC4H8CHO-1 + CH3O2H', [2.379000e+04, 2.55, 16490.0])

## Reaction 4791
#reaction('SC4H9CHO + CH3O2 <=> SC4H8CHO-2 + CH3O2H', [3.440000e+12, 0.05, 17880.0])

## Reaction 4792
#reaction('SC4H9CHO + CH3O2 <=> SC4H8CHO-3 + CH3O2H', [9.640000e+03, 2.6, 13910.0])

## Reaction 4793
#reaction('SC4H9CHO + CH3O2 <=> SC4H8CHO-4 + CH3O2H', [2.379000e+04, 2.55, 16490.0])

## Reaction 4794
#reaction('SC4H9 + HCO <=> SC4H9CHO', [1.810000e+13, 0.0, 0.0])

## Reaction 4795
#reaction('SC4H9CO <=> SC4H9 + CO', [1.000000e+11, 0.0, 9600.0])

## Reaction 4796
#reaction('C2H3CHO + C2H5 <=> SC4H8CHO-1', [1.000000e+11, 0.0, 6000.0])

## Reaction 4797
#reaction('HCO + C4H8-1 <=> SC4H8CHO-1', [1.000000e+11, 0.0, 7800.0])

## Reaction 4798
#reaction('CH3 + IC3H5CHO <=> SC4H8CHO-2', [1.000000e+11, 0.0, 7800.0])

## Reaction 4799
#reaction('HCO + C4H8-2 <=> SC4H8CHO-3', [1.000000e+11, 0.0, 6000.0])

## Reaction 4800
#reaction('SC3H5CHO + CH3 <=> SC4H8CHO-3', [1.000000e+11, 0.0, 7800.0])

## Reaction 4801
#reaction('C2H4 + CH3CHCHO <=> SC4H8CHO-4', [1.000000e+11, 0.0, 7800.0])

## Reaction 4802
#reaction('C5Y3 + OH <=> C5Y3-1J + H2O', [5.280000e+09, 0.97, 1586.0])

## Reaction 4803
#reaction('C5Y3 + OH <=> C5Y3-2J + H2O', [5.520000e+02, 3.12, -1176.0])

## Reaction 4804
#reaction('C5Y3 + HO2 <=> C5Y3-1J + H2O2', [2.379000e+04, 2.55, 16490.0])

## Reaction 4805
#reaction('C5Y3 + HO2 <=> C5Y3-2J + H2O2', [3.440000e+12, 0.05, 17880.0])

## Reaction 4806
#reaction('C5Y3 + CH3O2 <=> C5Y3-1J + CH3O2H', [2.379000e+04, 2.55, 16490.0])

## Reaction 4807
#reaction('C5Y3 + CH3O2 <=> C5Y3-2J + CH3O2H', [3.440000e+12, 0.05, 17880.0])

## Reaction 4808
#reaction('CH3CHCO + C2H5 <=> C5Y3-2J', [1.000000e+11, 0.0, 7800.0])

    #""")





    # ================================================================
    # ================================================================
    # ================================================================
    # ================================================================

    fichier_data = open('C7_Mehl.cti', 'w')
    fichier_data.write("""#-------------------------------------------------------------------------------
# Mehl et al. 2011  https://doi.org/10.1016/j.proci.2010.05.027
#-------------------------------------------------------------------------------

#n-heptane mechanism ver. 3.1 2012-03-30
#M. Mehl, W. J. Pitz, C. K. Westbrook and H. J. Curran, "Kinetic Modeling of Gasoline Surrogate Components and Mixtures under Engine Conditions," Proc. Combust. #Inst.  33 (1) (2011) 193-200.
#LLNL-MI-536391
#March 2012: Multiplied the A-factors of the molecular elimination of HO2 from heptyl-O2 by a factor of 2. The previous version reported lower values doe to a #clerical error in the uploaded file.
#n-heptane mechanism ver. 3.0 2009-12-11


units(length='cm', time='s', quantity='mol', act_energy='cal/mol')

ideal_gas(name='gas',
        elements="C H N O Ar He",
        species=\"""H               H2              O               O2              OH
                    H2O             N2              HO2             H2O2            AR
                    CO              CO2             CH2O            HCO             HO2CHO
                    O2CHO           HOCHO           OCHO            HOCH2O2H        HOCH2O2
                    OCH2O2H         HOCH2O          CH3OH           CH2OH           CH3O
                    CH3O2H          CH3O2           CH4             CH3             CH2
                    CH2(S)          C               CH              C2H6            C2H5
                    C2H4            C2H3            C2H2            C2H             CH3CHO
                    CH3CO           CH2CHO          CH2CO           HCCO            HCCOH
                    CH3CO3H         CH3CO3          CH3CO2          C2H5OH          C2H5O
                    PC2H4OH         SC2H4OH         O2C2H4OH        C2H5O2H         C2H5O2
                    C2H4O2H         C2H4O1-2        C2H3O1-2        CH3COCH3        CH3COCH2
                    CH3COCH2O2      CH3COCH2O2H     CH3COCH2O       C2H3CHO         C2H3CO
                    C2H5CHO         C2H5CO          CH3OCH3         CH3OCH2         CH3OCH2O2
                    CH2OCH2O2H      CH3OCH2O2H      CH3OCH2O        O2CH2OCH2O2H    HO2CH2OCHO
                    OCH2OCHO        HOCH2OCO        CH3OCHO         CH3OCO          CH2OCHO
                    HE              C3H8            IC3H7           NC3H7           C3H6
                    C3H5-A          C3H5-S          C3H5-T          C3H4-P          C3H4-A
                    C3H3            C3H2            C3H5O           C3H6OOH1-2      C3H6OOH1-3
                    C3H6OOH2-1      C3H6OOH1-2O2    C3H6OOH1-3O2    C3H6OOH2-1O2    C3H6OOH2-2
                    NC3H7O2H        IC3H7O2H        NC3H7O2         IC3H7O2         NC3H7O
                    IC3H7O          C3H6O1-2        C3H6O1-3        C3KET12         C3KET13
                    C3KET21         C3H51-2,3OOH    C3H52-1,3OOH    C3H6OH          HOC3H6O2
                    CH3CHCO         AC3H5OOH        C2H3OOH         C4H10           C4H8-1
                    C4H8-2          PC4H9           SC4H9           C4H71-1         C4H71-2
                    C4H71-3         C4H71-4         C4H72-2         C4H6            PC4H9O2H
                    SC4H9O2H        PC4H9O2         SC4H9O2         PC4H9O          SC4H9O
                    C4H7O           C4H8O1-2        C4H8O1-3        C4H8O1-4        C4H8O2-3
                    PC4H8OH         SC4H8OH         C4H8OH-1O2      C4H8OH-2O2      C4H8OOH1-1
                    C4H8OOH1-2      C4H8OOH1-3      C4H8OOH1-4      C4H8OOH2-1      C4H8OOH2-2
                    C4H8OOH2-3      C4H8OOH2-4      C4H8OOH1-2O2    C4H8OOH1-3O2    C4H8OOH1-4O2
                    C4H8OOH2-1O2    C4H8OOH2-3O2    C4H8OOH2-4O2    NC4KET12        NC4KET13
                    NC4KET14        NC4KET21        NC4KET23        NC4KET24        C2H5COCH3
                    C2H5COCH2       CH2CH2COCH3     CH3CHCOCH3      C2H3COCH3       CH3CHOOCOCH3
                    CH2CHOOHCOCH3   NC3H7CHO        NC3H7CO         C3H6CHO-1       C3H6CHO-2
                    C3H6CHO-3       C2H5CHCO        SC3H5CHO        SC3H5CO         CH2CH2CHO
                    IC4H10          IC4H9           TC4H9           IC4H8           IC4H7
                    TC4H9O2         IC4H9O2         TC4H8O2H-I      IC4H8O2H-I      IC4H8O2H-T
                    IC4H8O          CC4H8O          IC4H9O          TC4H9O          IC4H9O2H
                    TC4H9O2H        IC4H7O          IC4H8OH         IO2C4H8OH       IC3H7CHO
                    TC3H6CHO        IC3H7CO         IC3H6CHO        TC4H8OOH-IO2    IC4H8OOH-IO2
                    IC4H8OOH-TO2    IC4KETII        IC4KETIT        IC4H7OH         IC4H6OH
                    IC3H5CHO        IC3H5CO         TC3H6OCHO       IC3H6CO         IC4H7OOH
                    TC3H6OHCHO      TC3H6OH         IC3H5OH         TC3H6O2CHO      TC3H6O2HCO
                    IC3H5O2HCHO     CH2CCH2OH       TC4H8CHO        O2C4H8CHO       O2HC4H8CO
                    C3H5OH          TIC4H7Q2-I      IIC4H7Q2-T      IIC4H7Q2-I      CH2O2H
                    NC5H12          C5H11-1         C5H11-2         C5H11-3         C5H10-1
                    C5H10-2         C5H81-3         C5H91-3         C5H91-4         C5H91-5
                    C5H92-4         C5H92-5         C5H9O1-3        C5H9O2-4        C5H11O2H-1
                    C5H11O2H-2      C5H11O2H-3      C5H11O2-1       C5H11O2-2       C5H11O2-3
                    C5H11O-1        C5H11O-2        C5H11O-3        C5H10OOH1-2     C5H10OOH1-3
                    C5H10OOH1-4     C5H10OOH1-5     C5H10OOH2-1     C5H10OOH2-3     C5H10OOH2-4
                    C5H10OOH2-5     C5H10OOH3-2     C5H10OOH3-1     C5H10O1-2       C5H10O1-3
                    C5H10O1-4       C5H10O1-5       C5H10O2-3       C5H10O2-4       C5H10OOH1-2O2
                    C5H10OOH1-3O2   C5H10OOH1-4O2   C5H10OOH1-5O2   C5H10OOH2-1O2   C5H10OOH2-3O2
                    C5H10OOH2-4O2   C5H10OOH2-5O2   C5H10OOH3-1O2   C5H10OOH3-2O2   NC5KET12
                    NC5KET13        NC5KET14        NC5KET15        NC5KET21        NC5KET23
                    NC5KET24        NC5KET25        NC5KET31        NC5KET32        C5H10OH-1
                    C5H10OH-2       O2C5H10OH-1     O2C5H10OH-2     NC6H14          C6H13-1
                    C6H13-2         C6H13-3         C6H12-1         C6H12-2         C6H12-3
                    C6H111-3        C6H111-4        C6H111-5        C6H111-6        C6H112-4
                    C6H112-5        C6H112-6        C6H113-1        C6H11O1-3       C6H11O2-4
                    C6H12O1-2       C6H12O1-3       C6H12O1-4       C6H12O1-5       C6H12O2-3
                    C6H12O2-4       C6H12O2-5       C6H12O3-4       C6H13O2H-1      C6H13O2H-2
                    C6H13O2H-3      C6H13O2-1       C6H13O2-2       C6H13O2-3       C6H12OOH1-2
                    C6H12OOH1-3     C6H12OOH1-4     C6H12OOH1-5     C6H12OOH2-1     C6H12OOH2-3
                    C6H12OOH2-4     C6H12OOH2-5     C6H12OOH2-6     C6H12OOH3-1     C6H12OOH3-2
                    C6H12OOH3-4     C6H12OOH3-5     C6H12OOH3-6     C6H12OOH1-2O2   C6H12OOH1-3O2
                    C6H12OOH1-4O2   C6H12OOH1-5O2   C6H12OOH2-1O2   C6H12OOH2-3O2   C6H12OOH2-4O2
                    C6H12OOH2-5O2   C6H12OOH2-6O2   C6H12OOH3-1O2   C6H12OOH3-2O2   C6H12OOH3-4O2
                    C6H12OOH3-5O2   C6H12OOH3-6O2   NC6KET12        NC6KET13        NC6KET14
                    NC6KET15        NC6KET21        NC6KET23        NC6KET24        NC6KET25
                    NC6KET26        NC6KET31        NC6KET32        NC6KET34        NC6KET35
                    NC6KET36        C6H13O-1        C6H13O-2        C6H13O-3        C6H12OH-1
                    C6H12OH-2       C6H12OH-3       O2C6H12OH-1     O2C6H12OH-2     O2C6H12OH-3
                    NC4H9CHO        NC4H9CO         C4H8CHO-1       C4H8CHO-2       C4H8CHO-3
                    C4H8CHO-4       AC3H5CHO        AC3H5CO         C2H3CHCHO       NC3H7COCH3
                    NC3H7COCH2      C3H6COCH3-1     C3H6COCH3-2     C3H6COCH3-3     C2H5COC2H5
                    C2H5COC2H4P     C2H5COC2H4S     C2H5COC2H3      PC2H4COC2H3     SC2H4COC2H3
                    NC7H16          C7H15-1         C7H15-2         C7H15-3         C7H15-4
                    C7H14-1         C7H14-2         C7H14-3         C7H131-3        C7H131-4
                    C7H131-5        C7H131-6        C7H131-7        C7H132-4        C7H132-5
                    C7H132-6        C7H132-7        C7H133-1        C7H133-5        C7H133-6
                    C7H133-7        C7H13O1-3       C7H13O2-4       C7H13O3-5       C7H15O2-1
                    C7H15O2-2       C7H15O2-3       C7H15O2-4       C7H15O2H-1      C7H15O2H-2
                    C7H15O2H-3      C7H15O2H-4      C7H15O-1        C7H15O-2        C7H15O-3
                    C7H15O-4        C7H14OOH1-2     C7H14OOH1-3     C7H14OOH1-4     C7H14OOH1-5
                    C7H14OOH2-1     C7H14OOH2-3     C7H14OOH2-4     C7H14OOH2-5     C7H14OOH2-6
                    C7H14OOH3-1     C7H14OOH3-2     C7H14OOH3-4     C7H14OOH3-5     C7H14OOH3-6
                    C7H14OOH3-7     C7H14OOH4-1     C7H14OOH4-2     C7H14OOH4-3     C7H14OOH1-2O2
                    C7H14OOH1-3O2   C7H14OOH1-4O2   C7H14OOH1-5O2   C7H14OOH2-1O2   C7H14OOH2-3O2
                    C7H14OOH2-4O2   C7H14OOH2-5O2   C7H14OOH2-6O2   C7H14OOH3-1O2   C7H14OOH3-2O2
                    C7H14OOH3-4O2   C7H14OOH3-5O2   C7H14OOH3-6O2   C7H14OOH3-7O2   C7H14OOH4-1O2
                    C7H14OOH4-2O2   C7H14OOH4-3O2   C7H14O1-2       C7H14O1-3       C7H14O1-4
                    C7H14O1-5       C7H14O2-3       C7H14O2-4       C7H14O2-5       C7H14O2-6
                    C7H14O3-4       C7H14O3-5       NC7KET12        NC7KET13        NC7KET14
                    NC7KET15        NC7KET21        NC7KET23        NC7KET24        NC7KET25
                    NC7KET26        NC7KET31        NC7KET32        NC7KET34        NC7KET35
                    NC7KET36        NC7KET37        NC7KET41        NC7KET42        NC7KET43
                    C7H14OH-1       C7H14OH-2       C7H14OH-3       O2C7H14OH-1     O2C7H14OH-2
                    O2C7H14OH-3     NC5H11CHO       NC5H11CO        C5H10CHO-1      C5H10CHO-2
                    C5H10CHO-3      C5H10CHO-4      C5H10CHO-5      C4H7CHO1-4      C4H7CO1-4
                    C4H6CHO1-43     C4H6CHO1-44     NC4H9COCH3      NC4H9COCH2      C4H8COCH3-1
                    C4H8COCH3-2     C4H8COCH3-3     C4H8COCH3-4     C4H7OOH1-4      C4H7O1-4
                    C5H9OOH1-4      C5H9O1-4        C5H9OOH1-5      C5H9O1-5        C6H11OOH1-4
                    C6H11O1-4       C6H11OOH1-5     C6H11O1-5       C5H91-1         C4H7CHO1-1
                    C4H7CO1-1       C4H6CHO1-14     C4H6CHO1-13     NC3H7COC2H5     C3H6COC2H5-1
                    C3H6COC2H5-2    C3H6COC2H5-3    NC3H7COC2H4P    NC3H7COC2H4S    CHCHCHO
                    C6H101-3        C6H101-4        C6H101-5        C6H102-4        ETES1
                    MVOX            VTHF            EDHF            ETC6            KES1
                    C6H111O2H-3     C6H111O2H-4     C6H111O2H-5     C6H111O2H-6     C6H112O2H-4
                    C6H112O2H-5     C6H112O2H-6     C6H113O2H-2     C6H113O2H-1     NC6D1KET34
                    NC6D1KET35      NC6D1KET36      NC6D1KET43      NC6D1KET45      NC6D1KET46
                    NC6D1KET53      NC6D1KET54      NC6D1KET56      NC6D1KET63      NC6D1KET64
                    NC6D1KET65      NC6D2KET45      NC6D2KET46      NC6D2KET54      NC6D2KET56
                    NC6D2KET64      NC6D2KET65      NC6D3KET12      NC6D3KET21      C6H112O2-1
                    C6H111O2-3      C6H111O2-4      C6H111O2-5      C6H111O2-6      C6H112O2-4
                    C6H112O2-5      C6H112O2-6      C6H113O2-1      C6H113O2-2      C6H101OOH3-4
                    C6H101OOH3-5    C6H101OOH3-6    C6H101OOH4-3    C6H101OOH4-5    C6H101OOH4-6
                    C6H101OOH5-3    C6H101OOH5-4    C6H101OOH5-6    C6H101OOH6-3    C6H101OOH6-4
                    C6H101OOH6-5    C6H101OOH3-4O2  C6H101OOH3-5O2  C6H101OOH3-6O2  C6H101OOH4-3O2
                    C6H101OOH4-5O2  C6H101OOH4-6O2  C6H101OOH5-3O2  C6H101OOH5-4O2  C6H101OOH5-6O2
                    C6H101OOH6-3O2  C6H101OOH6-4O2  C6H101OOH6-5O2  C6H102OOH5-4    C6H103OOH1-2
                    C6H102OOH4-5    C6H102OOH4-6    C6H102OOH5-6    C6H102OOH6-4    C6H102OOH6-5
                    C6H103OOH2-1    C6H102OOH4-5O2  C6H102OOH4-6O2  C6H102OOH5-4O2  C6H102OOH5-6O2
                    C6H102OOH6-4O2  C6H102OOH6-5O2  C6H103OOH2-1O2  C6H103OOH1-2O2  C2H3COC3H7
                    C5H81-4         C5H91O2-3       C5H91O2-4       C5H91O2-5       C5H81OOH3-4
                    C5H81OOH3-5     C5H81OOH4-3     C5H81OOH4-5     C5H81OOH5-4     C5H81OOH5-3
                    C5H92O2-1       C5H81OOH3-4O2   C5H81OOH3-5O2   C5H81OOH4-3O2   C5H81OOH4-5O2
                    C5H81OOH5-3O2   C5H81OOH5-4O2   NC5D1KET34      NC5D1KET35      NC5D1KET43
                    NC5D1KET45      NC5D1KET53      NC5D1KET54      CY3C5H8O\""",
        reactions='all',
        transport='Mix',
        initial_state=state(temperature=300.0, pressure=OneAtm))

#-------------------------------------------------------------------------------
# Species data
#-------------------------------------------------------------------------------

species(name='H',
        atoms='H:1',
        thermo=(NASA([300.00, 1000.00],
                    [ 2.50000000E+00,  0.00000000E+00,  0.00000000E+00,
                    0.00000000E+00,  0.00000000E+00,  2.54716300E+04,
                    -4.60117600E-01]),
                NASA([1000.00, 5000.00],
                    [ 2.50000000E+00,  0.00000000E+00,  0.00000000E+00,
                    0.00000000E+00,  0.00000000E+00,  2.54716300E+04,
                    -4.60117600E-01])),
        transport=gas_transport(geom='atom',
                                diam=2.05,
                                well_depth=145.0),
        note='120186')

species(name='H2',
        atoms='H:2',
        thermo=(NASA([300.00, 1000.00],
                    [ 3.29812400E+00,  8.24944200E-04, -8.14301500E-07,
                    -9.47543400E-11,  4.13487200E-13, -1.01252100E+03,
                    -3.29409400E+00]),
                NASA([1000.00, 5000.00],
                    [ 2.99142300E+00,  7.00064400E-04, -5.63382900E-08,
                    -9.23157800E-12,  1.58275200E-15, -8.35034000E+02,
                    -1.35511000E+00])),
        transport=gas_transport(geom='linear',
                                diam=2.92,
                                well_depth=38.0,
                                polar=0.79,
                                rot_relax=280.0),
        note='121286')

species(name='O',
        atoms='O:1',
        thermo=(NASA([300.00, 1000.00],
                    [ 2.94642900E+00, -1.63816600E-03,  2.42103200E-06,
                    -1.60284300E-09,  3.89069600E-13,  2.91476400E+04,
                    2.96399500E+00]),
                NASA([1000.00, 5000.00],
                    [ 2.54206000E+00, -2.75506200E-05, -3.10280300E-09,
                    4.55106700E-12, -4.36805200E-16,  2.92308000E+04,
                    4.92030800E+00])),
        transport=gas_transport(geom='atom',
                                diam=2.75,
                                well_depth=80.0),
        note='120186')

species(name='O2',
        atoms='O:2',
        thermo=(NASA([300.00, 1000.00],
                    [ 3.21293600E+00,  1.12748600E-03, -5.75615000E-07,
                    1.31387700E-09, -8.76855400E-13, -1.00524900E+03,
                    6.03473800E+00]),
                NASA([1000.00, 5000.00],
                    [ 3.69757800E+00,  6.13519700E-04, -1.25884200E-07,
                    1.77528100E-11, -1.13643500E-15, -1.23393000E+03,
                    3.18916600E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.458,
                                well_depth=107.4,
                                polar=1.6,
                                rot_relax=3.8),
        note='121386')

species(name='OH',
        atoms='H:1 O:1',
        thermo=(NASA([300.00, 1710.00],
                    [ 3.41896226E+00,  3.19255801E-04, -3.08292717E-07,
                    3.64407494E-10, -1.00195479E-13,  3.45264448E+03,
                    2.54433372E+00]),
                NASA([1710.00, 5000.00],
                    [ 2.85376040E+00,  1.02994334E-03, -2.32666477E-07,
                    1.93750704E-11, -3.15759847E-16,  3.69949720E+03,
                    5.78756825E+00])),
        transport=gas_transport(geom='linear',
                                diam=2.75,
                                well_depth=80.0),
        note='7/13/0RUCIC')

species(name='H2O',
        atoms='H:2 O:1',
        thermo=(NASA([300.00, 1000.00],
                    [ 3.38684200E+00,  3.47498200E-03, -6.35469600E-06,
                    6.96858100E-09, -2.50658800E-12, -3.02081100E+04,
                    2.59023300E+00]),
                NASA([1000.00, 5000.00],
                    [ 2.67214600E+00,  3.05629300E-03, -8.73026000E-07,
                    1.20099600E-10, -6.39161800E-15, -2.98992100E+04,
                    6.86281700E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=2.605,
                                well_depth=572.4,
                                dipole=1.844,
                                rot_relax=4.0),
        note='20387')

species(name='N2',
        atoms='N:2',
        thermo=(NASA([300.00, 1000.00],
                    [ 3.29867700E+00,  1.40824000E-03, -3.96322200E-06,
                    5.64151500E-09, -2.44485500E-12, -1.02090000E+03,
                    3.95037200E+00]),
                NASA([1000.00, 5000.00],
                    [ 2.92664000E+00,  1.48797700E-03, -5.68476100E-07,
                    1.00970400E-10, -6.75335100E-15, -9.22797700E+02,
                    5.98052800E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.621,
                                well_depth=97.53,
                                polar=1.76,
                                rot_relax=4.0),
        note='121286')

species(name='HO2',
        atoms='H:1 O:2',
        thermo=(NASA([200.00, 1000.00],
                    [ 4.30179801E+00, -4.74912051E-03,  2.11582891E-05,
                    -2.42763894E-08,  9.29225124E-12,  2.94808040E+02,
                    3.71666245E+00]),
                NASA([1000.00, 3500.00],
                    [ 4.01721090E+00,  2.23982013E-03, -6.33658150E-07,
                    1.14246370E-10, -1.07908535E-14,  1.11856713E+02,
                    3.78510215E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.458,
                                well_depth=107.4,
                                rot_relax=1.0),
        note='L5/89')

species(name='H2O2',
        atoms='H:2 O:2',
        thermo=(NASA([300.00, 1000.00],
                    [ 3.38875400E+00,  6.56922600E-03, -1.48501300E-07,
                    -4.62580600E-09,  2.47151500E-12, -1.76631500E+04,
                    6.78536300E+00]),
                NASA([1000.00, 5000.00],
                    [ 4.57316700E+00,  4.33613600E-03, -1.47468900E-06,
                    2.34890400E-10, -1.43165400E-14, -1.80069600E+04,
                    5.01137000E-01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.458,
                                well_depth=107.4,
                                rot_relax=3.8),
        note='120186')

species(name='AR',
        atoms='Ar:1',
        thermo=(NASA([300.00, 1000.00],
                    [ 2.50000000E+00,  0.00000000E+00,  0.00000000E+00,
                    0.00000000E+00,  0.00000000E+00, -7.45375000E+02,
                    4.36600100E+00]),
                NASA([1000.00, 5000.00],
                    [ 2.50000000E+00,  0.00000000E+00,  0.00000000E+00,
                    0.00000000E+00,  0.00000000E+00, -7.45375000E+02,
                    4.36600100E+00])),
        transport=gas_transport(geom='atom',
                                diam=3.33,
                                well_depth=136.5),
        note='120186')

species(name='CO',
        atoms='C:1 O:1',
        thermo=(NASA([300.00, 1429.00],
                    [ 3.19036352E+00,  8.94419972E-04, -3.24927563E-08,
                    -1.04599967E-10,  2.41965693E-14, -1.42869054E+04,
                    5.33277914E+00]),
                NASA([1429.00, 5000.00],
                    [ 3.11216890E+00,  1.15948283E-03, -3.38480362E-07,
                    4.41403098E-11, -2.12862228E-15, -1.42718539E+04,
                    5.71725177E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.65,
                                well_depth=98.1,
                                polar=1.95,
                                rot_relax=1.8),
        note='29/11/04')

species(name='CO2',
        atoms='C:1 O:2',
        thermo=(NASA([300.00, 1380.00],
                    [ 2.57930490E+00,  8.24684987E-03, -6.42716047E-06,
                    2.54637024E-09, -4.12030443E-13, -4.84162830E+04,
                    8.81141041E+00]),
                NASA([1380.00, 5000.00],
                    [ 5.18953018E+00,  2.06006476E-03, -7.33575324E-07,
                    1.17004374E-10, -6.91729215E-15, -4.93178953E+04,
                    -5.18289303E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.763,
                                well_depth=244.0,
                                polar=2.65,
                                rot_relax=2.1),
        note='29/11/04')

species(name='CH2O',
        atoms='H:2 C:1 O:1',
        thermo=(NASA([300.00, 1486.00],
                    [ 3.00754197E+00,  3.04729496E-03,  5.25109246E-06,
                    -5.12019281E-09,  1.27133795E-12, -1.41188397E+04,
                    8.10120233E+00]),
                NASA([1486.00, 5000.00],
                    [ 4.02068394E+00,  5.09903417E-03, -1.76430480E-06,
                    2.76025879E-10, -1.60998042E-14, -1.49287258E+04,
                    1.06525547E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.59,
                                well_depth=498.0,
                                rot_relax=2.0),
        note='6/29/04RUSCI')

species(name='HCO',
        atoms='H:1 C:1 O:1',
        thermo=(NASA([300.00, 1690.00],
                    [ 3.81049965E+00,  8.13269825E-04,  3.13164701E-06,
                    -2.39478268E-09,  5.06894554E-13,  4.03859901E+03,
                    4.94843165E+00]),
                NASA([1690.00, 5000.00],
                    [ 3.44148164E+00,  3.52157719E-03, -1.24136118E-06,
                    1.97328644E-10, -1.16538616E-14,  3.97409684E+03,
                    6.24593456E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.59,
                                well_depth=498.0),
        note='29/11/04')

species(name='HO2CHO',
        atoms='H:2 C:1 O:3',
        thermo=(NASA([300.00, 1378.00],
                    [ 2.42464726E+00,  2.19706380E-02, -1.68705546E-05,
                    6.25612194E-09, -9.11645843E-13, -3.54828006E+04,
                    1.75027796E+01]),
                NASA([1378.00, 5000.00],
                    [ 9.87503878E+00,  4.64663708E-03, -1.67230522E-06,
                    2.68624413E-10, -1.59595232E-14, -3.80502496E+04,
                    -2.24939155E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='6/26/95THERM')

species(name='O2CHO',
        atoms='H:1 C:1 O:3',
        thermo=(NASA([300.00, 1368.00],
                    [ 3.96059309E+00,  1.06002279E-02, -5.25713351E-06,
                    1.01716726E-09, -2.87487602E-14, -1.73599383E+04,
                    1.17807483E+01]),
                NASA([1368.00, 5000.00],
                    [ 7.24075139E+00,  4.63312951E-03, -1.63693995E-06,
                    2.59706693E-10, -1.52964699E-14, -1.87027618E+04,
                    -6.49547212E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='6/26/95THERM')

species(name='HOCHO',
        atoms='H:2 C:1 O:2',
        thermo=(NASA([300.00, 1419.00],
                    [ 1.28069021E+00,  1.52887758E-02, -5.64150476E-06,
                    -1.22968799E-09,  8.14273233E-13, -4.64347524E+04,
                    1.83142081E+01]),
                NASA([1419.00, 5000.00],
                    [ 7.67134440E+00,  4.43426373E-03, -1.60867436E-06,
                    2.59770020E-10, -1.54899704E-14, -4.89708660E+04,
                    -1.73024418E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='9/27/5THERM')

species(name='OCHO',
        atoms='H:1 C:1 O:2',
        thermo=(NASA([300.00, 1412.00],
                    [ 1.42991854E+00,  1.22803284E-02, -5.28461096E-06,
                    -3.73026551E-10,  5.07953571E-13, -1.62042814E+04,
                    1.80411779E+01]),
                NASA([1412.00, 5000.00],
                    [ 6.49815560E+00,  3.22319012E-03, -1.17099595E-06,
                    1.89272918E-10, -1.12937000E-14, -1.81679042E+04,
                    -1.00412613E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.59,
                                well_depth=498.0,
                                rot_relax=2.0),
        note='9/27/5THERM')

species(name='HOCH2O2H',
        atoms='H:4 C:1 O:3',
        thermo=(NASA([300.00, 1422.00],
                    [ 1.85716693E+00,  3.23153132E-02, -2.69928902E-05,
                    1.11694484E-08, -1.81284103E-12, -4.00314471E+04,
                    1.90917729E+01]),
                NASA([1422.00, 5000.00],
                    [ 1.16303827E+01,  7.15133688E-03, -2.39035030E-06,
                    3.65772791E-10, -2.10199524E-14, -4.31079242E+04,
                    -3.24276725E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.626,
                                well_depth=481.8,
                                dipole=1.7,
                                rot_relax=1.0),
        note='4/9/98THERM')

species(name='HOCH2O2',
        atoms='H:3 C:1 O:3',
        thermo=(NASA([300.00, 1412.00],
                    [ 2.85441621E+00,  2.33663535E-02, -1.88115990E-05,
                    7.96709515E-09, -1.36346618E-12, -2.29866196E+04,
                    1.51730565E+01]),
                NASA([1412.00, 5000.00],
                    [ 9.04545938E+00,  7.15223373E-03, -2.37005676E-06,
                    3.60083481E-10, -2.05750228E-14, -2.49414886E+04,
                    -1.74210530E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.626,
                                well_depth=481.8,
                                dipole=1.7,
                                rot_relax=1.0),
        note='4/9/98THERM')

species(name='OCH2O2H',
        atoms='H:3 C:1 O:3',
        thermo=(NASA([300.00, 1420.00],
                    [ 1.93823075E+00,  3.01465730E-02, -2.61053152E-05,
                    1.09463562E-08, -1.78312692E-12, -1.38166625E+04,
                    1.85042002E+01]),
                NASA([1420.00, 5000.00],
                    [ 1.15398246E+01,  5.34291432E-03, -1.81878917E-06,
                    2.81968625E-10, -1.63584348E-14, -1.68237489E+04,
                    -3.20700633E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.626,
                                well_depth=481.8,
                                dipole=1.7,
                                rot_relax=1.0),
        note='4/9/98THERM')

species(name='HOCH2O',
        atoms='H:3 C:1 O:2',
        thermo=(NASA([300.00, 1452.00],
                    [ 4.11183145E+00,  7.53850697E-03,  3.77337370E-06,
                    -5.38746005E-09,  1.45615887E-12, -2.28023001E+04,
                    7.46807254E+00]),
                NASA([1452.00, 5000.00],
                    [ 6.39521515E+00,  7.43673043E-03, -2.50422354E-06,
                    3.84879712E-10, -2.21778689E-14, -2.41108840E+04,
                    -6.63865583E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.41,
                                well_depth=470.6,
                                rot_relax=1.5),
        note='2/16/99THERM')

species(name='CH3OH',
        atoms='H:4 C:1 O:1',
        thermo=(NASA([300.00, 1985.00],
                    [ 1.82176844E+00,  1.23326420E-02, -2.73742177E-06,
                    -1.07387318E-09,  4.23644837E-13, -2.52222207E+04,
                    1.48104300E+01]),
                NASA([1985.00, 5000.00],
                    [ 3.92867513E+00,  9.81626074E-03, -3.40615476E-06,
                    5.35510197E-10, -3.13776029E-14, -2.61454175E+04,
                    2.69464483E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.626,
                                well_depth=481.8,
                                rot_relax=1.0),
        note='2/25/4THERM')

species(name='CH2OH',
        atoms='H:3 C:1 O:1',
        thermo=(NASA([300.00, 1399.00],
                    [ 3.05674228E+00,  1.19335634E-02, -8.72501303E-06,
                    3.82781054E-09, -7.22887951E-13, -2.83140190E+03,
                    8.98878133E+00]),
                NASA([1399.00, 5000.00],
                    [ 5.41875913E+00,  5.66185244E-03, -1.87471136E-06,
                    2.84441938E-10, -1.62299798E-14, -3.61475540E+03,
                    -3.49277963E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.69,
                                well_depth=417.0,
                                dipole=1.7,
                                rot_relax=2.0),
        note='2/25/4THERM')

species(name='CH3O',
        atoms='H:3 C:1 O:1',
        thermo=(NASA([300.00, 1509.00],
                    [ 2.23058023E+00,  8.53178586E-03,  1.02166624E-06,
                    -3.41046916E-09,  9.94691038E-13,  9.45939708E+02,
                    1.28377569E+01]),
                NASA([1509.00, 5000.00],
                    [ 4.64787019E+00,  6.90830683E-03, -2.34404776E-06,
                    3.61994570E-10, -2.09253541E-14, -2.99208881E+02,
                    -1.57740193E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.69,
                                well_depth=417.0,
                                dipole=1.7,
                                rot_relax=2.0),
        note='8/9/4THERM')

species(name='CH3O2H',
        atoms='H:4 C:1 O:2',
        thermo=(NASA([300.00, 1367.00],
                    [ 2.83880024E+00,  1.86096249E-02, -8.48165412E-06,
                    1.00387451E-09,  1.71612429E-13, -1.74033753E+04,
                    1.16092433E+01]),
                NASA([1367.00, 5000.00],
                    [ 8.80409289E+00,  8.09427218E-03, -2.85843274E-06,
                    4.53369754E-10, -2.66980707E-14, -1.98512174E+04,
                    -2.17000591E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.626,
                                well_depth=481.8,
                                rot_relax=1.0),
        note='1/14/5THERM')

species(name='CH3O2',
        atoms='H:3 C:1 O:2',
        thermo=(NASA([300.00, 1365.00],
                    [ 3.80497590E+00,  9.80784660E-03, -3.90940624E-07,
                    -2.23072602E-09,  6.43310820E-13, -4.55625796E+02,
                    7.81789100E+00]),
                NASA([1365.00, 5000.00],
                    [ 6.34718801E+00,  7.92089358E-03, -2.76601913E-06,
                    4.35360631E-10, -2.54984762E-14, -1.83436055E+03,
                    -7.42552545E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.626,
                                well_depth=481.8,
                                rot_relax=1.0),
        note='1/14/5THERM')

species(name='CH4',
        atoms='H:4 C:1',
        thermo=(NASA([300.00, 1462.00],
                    [ 3.72113020E+00, -2.50293289E-03,  1.90246534E-05,
                    -1.46871253E-08,  3.43791152E-12, -1.01424099E+04,
                    1.22776596E+00]),
                NASA([1462.00, 5000.00],
                    [ 4.09617653E+00,  7.44330845E-03, -2.63871900E-06,
                    4.19577604E-10, -2.47508050E-14, -1.13835704E+04,
                    -4.67561383E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.746,
                                well_depth=141.4,
                                polar=2.6,
                                rot_relax=13.0),
        note='29/11/04')

species(name='CH3',
        atoms='H:3 C:1',
        thermo=(NASA([300.00, 1389.00],
                    [ 3.43858162E+00,  4.07752664E-03,  3.19830994E-07,
                    -9.47669390E-10,  2.21828166E-13,  1.63164018E+04,
                    2.52807406E+00]),
                NASA([1389.00, 5000.00],
                    [ 3.51281376E+00,  5.11412613E-03, -1.67632050E-06,
                    2.52495174E-10, -1.43302923E-14,  1.61238027E+04,
                    1.62436112E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.8,
                                well_depth=144.0))

species(name='CH2',
        atoms='H:2 C:1',
        thermo=(NASA([250.00, 1000.00],
                    [ 3.76223700E+00,  1.15981900E-03,  2.48958500E-07,
                    8.80083600E-10, -7.33243500E-13,  4.53679100E+04,
                    1.71257800E+00]),
                NASA([1000.00, 4000.00],
                    [ 3.63640800E+00,  1.93305700E-03, -1.68701600E-07,
                    -1.00989900E-10,  1.80825600E-14,  4.53413400E+04,
                    2.15656100E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.8,
                                well_depth=144.0),
        note='120186')

species(name='CH2(S)',
        atoms='H:2 C:1',
        thermo=(NASA([300.00, 1000.00],
                    [ 3.97126500E+00, -1.69908900E-04,  1.02536900E-06,
                    2.49255100E-09, -1.98126600E-12,  4.98936800E+04,
                    5.75320700E-02]),
                NASA([1000.00, 4000.00],
                    [ 3.55288900E+00,  2.06678800E-03, -1.91411600E-07,
                    -1.10467300E-10,  2.02135000E-14,  4.98497500E+04,
                    1.68657000E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.8,
                                well_depth=144.0),
        note='31287')

species(name='C',
        atoms='C:1',
        thermo=(NASA([300.00, 1000.00],
                    [ 2.49858500E+00,  8.08577700E-05, -2.69769700E-07,
                    3.04072900E-10, -1.10665200E-13,  8.54587800E+04,
                    4.75345900E+00]),
                NASA([1000.00, 5000.00],
                    [ 2.60208700E+00, -1.78708100E-04,  9.08704100E-08,
                    -1.14993300E-11,  3.31084400E-16,  8.54215400E+04,
                    4.19517700E+00])),
        transport=gas_transport(geom='atom',
                                diam=3.298,
                                well_depth=71.4))

species(name='CH',
        atoms='H:1 C:1',
        thermo=(NASA([300.00, 1000.00],
                    [ 3.20020200E+00,  2.07287600E-03, -5.13443100E-06,
                    5.73389000E-09, -1.95553300E-12,  7.04525900E+04,
                    3.33158800E+00]),
                NASA([1000.00, 5000.00],
                    [ 2.19622300E+00,  2.34038100E-03, -7.05820100E-07,
                    9.00758200E-11, -3.85504000E-15,  7.08672300E+04,
                    9.17837300E+00])),
        transport=gas_transport(geom='linear',
                                diam=2.75,
                                well_depth=80.0),
        note='121286')

species(name='C2H6',
        atoms='H:6 C:2',
        thermo=(NASA([300.00, 1383.00],
                    [ 4.78623203E-02,  2.40569127E-02, -1.15155912E-05,
                    2.48666238E-09, -1.78343944E-13, -1.10923014E+04,
                    2.06544071E+01]),
                NASA([1383.00, 5000.00],
                    [ 6.05972630E+00,  1.30382837E-02, -4.48103942E-06,
                    6.97762095E-10, -4.05606353E-14, -1.35751226E+04,
                    -1.28608001E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.35,
                                well_depth=247.5,
                                rot_relax=1.5),
        note='8/4/4THERM')

species(name='C2H5',
        atoms='H:5 C:2',
        thermo=(NASA([300.00, 1387.00],
                    [ 1.32730217E+00,  1.76656753E-02, -6.14926558E-06,
                    -3.01143466E-10,  4.38617775E-13,  1.34284028E+04,
                    1.71789216E+01]),
                NASA([1387.00, 5000.00],
                    [ 5.88784390E+00,  1.03076793E-02, -3.46844396E-06,
                    5.32499257E-10, -3.06512651E-14,  1.15065499E+04,
                    -8.49651771E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.35,
                                well_depth=247.5,
                                rot_relax=1.5),
        note='8/4/4THERM')

species(name='C2H4',
        atoms='H:4 C:2',
        thermo=(NASA([300.00, 1395.00],
                    [ 2.33879687E-01,  1.96334647E-02, -1.16833214E-05,
                    3.64246453E-09, -4.77442715E-13,  5.46489338E+03,
                    1.97084228E+01]),
                NASA([1395.00, 5000.00],
                    [ 5.22176372E+00,  8.96137303E-03, -3.04868886E-06,
                    4.71465524E-10, -2.72739592E-14,  3.60389679E+03,
                    -7.47789234E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.496,
                                well_depth=238.4,
                                rot_relax=1.5),
        note='10/4/5THERM')

species(name='C2H3',
        atoms='H:3 C:2',
        thermo=(NASA([300.00, 1395.00],
                    [ 1.25329724E+00,  1.56258370E-02, -1.07803879E-05,
                    4.18054634E-09, -7.01360362E-13,  3.50734773E+04,
                    1.71341661E+01]),
                NASA([1395.00, 5000.00],
                    [ 5.07331248E+00,  6.58316278E-03, -2.23762924E-06,
                    3.45803379E-10, -1.99940490E-14,  3.37234748E+04,
                    -3.39792712E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.721,
                                well_depth=265.3,
                                rot_relax=1.0),
        note='10/4/5THERM')

species(name='C2H2',
        atoms='H:2 C:2',
        thermo=(NASA([300.00, 1407.00],
                    [ 2.06742667E+00,  1.46568506E-02, -1.52947063E-05,
                    8.30965761E-09, -1.72932175E-12,  2.59578589E+04,
                    8.62758672E+00]),
                NASA([1407.00, 5000.00],
                    [ 4.98265164E+00,  4.25992930E-03, -1.37483523E-06,
                    2.04717964E-10, -1.15191740E-14,  2.52697118E+04,
                    -5.81321385E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.721,
                                well_depth=265.3,
                                rot_relax=2.5),
        note='10/4/5THERM')

species(name='C2H',
        atoms='H:1 C:2',
        thermo=(NASA([300.00, 1361.00],
                    [ 4.27708350E+00,  1.74130076E-03,  7.31180206E-07,
                    -9.71827901E-10,  2.36956727E-13,  6.63125955E+04,
                    3.16341916E-01]),
                NASA([1361.00, 5000.00],
                    [ 4.66893822E+00,  2.01284720E-03, -7.01828634E-07,
                    1.10347228E-10, -6.45790465E-15,  6.60310696E+04,
                    -2.25424693E+00])),
        transport=gas_transport(geom='linear',
                                diam=3.721,
                                well_depth=265.3,
                                rot_relax=2.5),
        note='10/4/5THERM')

species(name='CH3CHO',
        atoms='H:4 C:2 O:1',
        thermo=(NASA([300.00, 1377.00],
                    [ 1.77060035E+00,  1.84475161E-02, -7.24138162E-06,
                    2.34364561E-10,  3.35543891E-13, -2.18078850E+04,
                    1.65023437E+01]),
                NASA([1377.00, 5000.00],
                    [ 6.98518866E+00,  9.67897787E-03, -3.31841954E-06,
                    5.16025901E-10, -2.99725903E-14, -2.39807279E+04,
                    -1.27484852E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='8/10/4THERM')

species(name='CH3CO',
        atoms='H:3 C:2 O:1',
        thermo=(NASA([300.00, 1371.00],
                    [ 2.52884150E+00,  1.37152173E-02, -4.28607476E-06,
                    -7.71684278E-10,  4.83836380E-13, -3.02546532E+03,
                    1.40340315E+01]),
                NASA([1371.00, 5000.00],
                    [ 6.56682466E+00,  7.55308669E-03, -2.59966784E-06,
                    4.05334994E-10, -2.35876164E-14, -4.76690401E+03,
                    -8.83301965E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='8/10/4THERM')

species(name='CH2CHO',
        atoms='H:3 C:2 O:1',
        thermo=(NASA([300.00, 1388.00],
                    [ 1.47616956E+00,  2.08974186E-02, -1.50123552E-05,
                    5.62967382E-09, -8.76624470E-13, -4.82050510E+02,
                    1.63756465E+01]),
                NASA([1388.00, 5000.00],
                    [ 7.54145790E+00,  6.82297125E-03, -2.36938607E-06,
                    3.71633517E-10, -2.17160048E-14, -2.61437239E+03,
                    -1.62602766E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='8/10/4THERM')

species(name='CH2CO',
        atoms='H:2 C:2 O:1',
        thermo=(NASA([300.00, 1000.00],
                    [ 2.97497100E+00,  1.21187100E-02, -2.34504600E-06,
                    -6.46668500E-09,  3.90564900E-12, -7.63263700E+03,
                    8.67355300E+00]),
                NASA([1000.00, 5000.00],
                    [ 6.03881700E+00,  5.80484000E-03, -1.92095400E-06,
                    2.79448500E-10, -1.45886800E-14, -8.58340200E+03,
                    -7.65758100E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='121686')

species(name='HCCO',
        atoms='H:1 C:2 O:1',
        thermo=(NASA([300.00, 1000.00],
                    [ 5.04796500E+00,  4.45347800E-03,  2.26828300E-07,
                    -1.48209500E-09,  2.25074200E-13,  1.96589200E+04,
                    4.81843900E-01]),
                NASA([1000.00, 4000.00],
                    [ 6.75807300E+00,  2.00040000E-03, -2.02760700E-07,
                    -1.04113200E-10,  1.96516500E-14,  1.90151300E+04,
                    -9.07126200E+00])),
        transport=gas_transport(geom='nonlinear',
                                diam=2.5,
                                well_depth=150.0,
                                rot_relax=1.0),
        note='32387')

species(name='HCCOH',
        atoms='H:2 C:2 O:1',
        thermo=(NASA([300.00, 1000.00],
                    [ 3.89946500E+00,  9.70107500E-03, -3.11930900E-07,
                    -5.53773200E-09,  2.46573200E-12,  8.70119000E+03,
                    4.49187500E+00]),
                NASA([1000.00, 4000.00],
                    [ 7.32832400E+00,  3.33641600E-03, -3.02470500E-07,
                    -1.78110600E-10,  3.24516800E-14,  7.59825800E+03,
                    -1.40121400E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='32387')

species(name='CH3CO3H',
        atoms='H:4 C:2 O:3',
        thermo=(NASA([300.00, 1391.00],
                    [ 2.24135876E+00,  3.37963514E-02, -2.53887482E-05,
                    9.67583587E-09, -1.49266157E-12, -4.24677831E+04,
                    1.70668133E+01]),
                NASA([1391.00, 5000.00],
                    [ 1.25060485E+01,  9.47789695E-03, -3.30402246E-06,
                    5.19630793E-10, -3.04233568E-14, -4.59856703E+04,
                    -3.79195947E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='6/26/95THERM')

species(name='CH3CO3',
        atoms='H:3 C:2 O:3',
        thermo=(NASA([300.00, 1391.00],
                    [ 3.60373432E+00,  2.70080341E-02, -2.08293438E-05,
                    8.50541104E-09, -1.43846110E-12, -2.34205171E+04,
                    1.12014914E+01]),
                NASA([1391.00, 5000.00],
                    [ 1.12522498E+01,  8.33652672E-03, -2.89014530E-06,
                    4.52781734E-10, -2.64354456E-14, -2.60238584E+04,
                    -2.96370457E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='4/3/0THERM')

species(name='CH3CO2',
        atoms='H:3 C:2 O:2',
        thermo=(NASA([300.00, 1395.00],
                    [ 1.37440768E+00,  2.49115604E-02, -1.74308894E-05,
                    6.24799508E-09, -9.09516835E-13, -2.72330150E+04,
                    1.81405454E+01]),
                NASA([1395.00, 5000.00],
                    [ 8.54059736E+00,  8.32951214E-03, -2.84722010E-06,
                    4.41927196E-10, -2.56373394E-14, -2.97290678E+04,
                    -2.03883545E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='2/14/95THERM')

species(name='C2H5OH',
        atoms='H:6 C:2 O:1',
        thermo=(NASA([300.00, 1395.00],
                    [ 1.79106094E-01,  3.09060081E-02, -1.93596697E-05,
                    6.31830859E-09, -8.53166875E-13, -2.95670605E+04,
                    2.44715919E+01]),
                NASA([1395.00, 5000.00],
                    [ 8.31742137E+00,  1.29603294E-02, -4.39286497E-06,
                    6.77734995E-10, -3.91448231E-14, -3.25219715E+04,
                    -1.96502783E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.41,
                                well_depth=470.6,
                                rot_relax=1.5),
        note='8/9/4THERM')

species(name='C2H5O',
        atoms='H:5 C:2 O:1',
        thermo=(NASA([300.00, 1393.00],
                    [ 2.87429022E-01,  2.86500918E-02, -1.83857008E-05,
                    6.03096179E-09, -8.04562643E-13, -3.35717377E+03,
                    2.37513898E+01]),
                NASA([1393.00, 5000.00],
                    [ 8.23717244E+00,  1.10885879E-02, -3.78808372E-06,
                    5.87613473E-10, -3.40712889E-14, -6.22948597E+03,
                    -1.93190543E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.41,
                                well_depth=470.6,
                                rot_relax=1.5),
        note='8/9/4THERM')

species(name='PC2H4OH',
        atoms='H:5 C:2 O:1',
        thermo=(NASA([300.00, 1396.00],
                    [ 9.90023426E-01,  2.61564272E-02, -1.65928300E-05,
                    5.50525131E-09, -7.57520667E-13, -4.96032283E+03,
                    2.28675716E+01]),
                NASA([1396.00, 5000.00],
                    [ 7.88509591E+00,  1.08287795E-02, -3.66829683E-06,
                    5.65734300E-10, -3.26673351E-14, -7.44974622E+03,
                    -1.44673770E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.41,
                                well_depth=470.6,
                                rot_relax=1.5),
        note='8/9/4THERM')

species(name='SC2H4OH',
        atoms='H:5 C:2 O:1',
        thermo=(NASA([300.00, 1397.00],
                    [ 1.61103852E+00,  2.63511867E-02, -1.82941500E-05,
                    6.91243160E-09, -1.10384057E-12, -8.71494725E+03,
                    1.80114775E+01]),
                NASA([1397.00, 5000.00],
                    [ 8.34980768E+00,  1.03718068E-02, -3.49683481E-06,
                    5.37524748E-10, -3.09661005E-14, -1.10556188E+04,
                    -1.81354112E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.41,
                                well_depth=470.6,
                                rot_relax=1.5),
        note='8/9/4THERM')

species(name='O2C2H4OH',
        atoms='H:5 C:2 O:3',
        thermo=(NASA([300.00, 1392.00],
                    [ 4.11839445E+00,  2.72240632E-02, -1.60824430E-05,
                    5.17033408E-09, -7.31610168E-13, -2.30857785E+04,
                    1.28482112E+01]),
                NASA([1392.00, 5000.00],
                    [ 1.07432659E+01,  1.30957787E-02, -4.45370088E-06,
                    6.88548738E-10, -3.98230113E-14, -2.55911274E+04,
                    -2.33254953E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='2/14/95THERM')

species(name='C2H5O2H',
        atoms='H:6 C:2 O:2',
        thermo=(NASA([300.00, 1391.00],
                    [ 1.57329011E+00,  3.52379996E-02, -2.53203993E-05,
                    9.56802476E-09, -1.48167375E-12, -2.15278368E+04,
                    1.90472032E+01]),
                NASA([1391.00, 5000.00],
                    [ 1.12305737E+01,  1.20482120E-02, -3.96730201E-06,
                    6.00754632E-10, -3.42657803E-14, -2.47977531E+04,
                    -3.25607232E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.41,
                                well_depth=470.6,
                                rot_relax=1.5),
        note='1/14/5THERM')

species(name='C2H5O2',
        atoms='H:5 C:2 O:2',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.58630333E+00,  2.61836362E-02, -1.68306193E-05,
                    6.08749295E-09, -9.57302040E-13, -4.58588992E+03,
                    1.50486289E+01]),
                NASA([1387.00, 5000.00],
                    [ 8.77641329E+00,  1.18767163E-02, -3.87683050E-06,
                    5.83210754E-10, -3.31006724E-14, -6.78748703E+03,
                    -1.83119972E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.41,
                                well_depth=470.6,
                                rot_relax=1.5),
        note='1/14/5THERM')

species(name='C2H4O2H',
        atoms='H:5 C:2 O:2',
        thermo=(NASA([300.00, 1397.00],
                    [ 3.46916874E+00,  2.71188626E-02, -2.08022550E-05,
                    8.44284845E-09, -1.40756215E-12,  3.89688270E+03,
                    1.43400726E+01]),
                NASA([1397.00, 5000.00],
                    [ 1.05228954E+01,  9.48091381E-03, -3.55727763E-06,
                    6.41445994E-10, -4.21232247E-14,  1.55718322E+03,
                    -2.31413632E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.41,
                                well_depth=470.6,
                                rot_relax=1.5),
        note='1/14/5THERM')

species(name='C2H4O1-2',
        atoms='H:4 C:2 O:1',
        thermo=(NASA([300.00, 1399.00],
                    [-1.44361790E+00,  2.91837107E-02, -1.96805054E-05,
                    6.67472575E-09, -9.02139342E-13, -7.43433507E+03,
                    2.95212453E+01]),
                NASA([1399.00, 5000.00],
                    [ 6.99485763E+00,  1.00772459E-02, -3.45220932E-06,
                    5.36571216E-10, -3.11563543E-14, -1.04091344E+04,
                    -1.59857976E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='9/1/0THERM')

species(name='C2H3O1-2',
        atoms='H:3 C:2 O:1',
        thermo=(NASA([300.00, 1492.00],
                    [-1.62965122E+00,  2.93455486E-02, -2.43737550E-05,
                    1.00522325E-08, -1.61259036E-12,  1.52459425E+04,
                    3.22782741E+01]),
                NASA([1492.00, 5000.00],
                    [ 6.88486471E+00,  6.94720501E-03, -2.23214698E-06,
                    3.32190767E-10, -1.87124555E-14,  1.26442200E+04,
                    -1.23842570E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.97,
                                well_depth=436.0,
                                rot_relax=2.0),
        note='4/3/0THERM')

species(name='CH3COCH3',
        atoms='H:6 C:3 O:1',
        thermo=(NASA([300.00, 1382.00],
                    [ 1.24527408E+00,  2.99760255E-02, -1.40026661E-05,
                    2.16453512E-09,  1.27637295E-13, -2.78348727E+04,
                    2.03682615E+01]),
                NASA([1382.00, 5000.00],
                    [ 9.62674379E+00,  1.45519245E-02, -4.97749457E-06,
                    7.72794591E-10, -4.48367165E-14, -3.11862263E+04,
                    -2.61613449E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.86,
                                well_depth=435.5,
                                rot_relax=1.0),
        note='8/30/4THERM')

species(name='CH3COCH2',
        atoms='H:5 C:3 O:1',
        thermo=(NASA([300.00, 1388.00],
                    [ 1.22337251E+00,  3.24546742E-02, -2.13542518E-05,
                    6.96777735E-09, -8.99160299E-13, -6.59419324E+03,
                    2.05537233E+01]),
                NASA([1388.00, 5000.00],
                    [ 1.08892477E+01,  1.11540675E-02, -3.85516785E-06,
                    6.02834048E-10, -3.51533449E-14, -1.00741464E+04,
                    -3.18043322E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.86,
                                well_depth=435.5,
                                rot_relax=1.0),
        note='8/30/4THERM')

species(name='CH3COCH2O2',
        atoms='H:5 C:3 O:3',
        thermo=(NASA([300.00, 1379.00],
                    [ 5.95535468E+00,  2.70255205E-02, -1.37385031E-05,
                    3.53735851E-09, -4.03922557E-13, -2.06679464E+04,
                    5.21436049E+00]),
                NASA([1379.00, 5000.00],
                    [ 1.27690342E+01,  1.42554828E-02, -4.92821461E-06,
                    7.70448921E-10, -4.49110534E-14, -2.34798669E+04,
                    -3.27155799E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.429,
                                well_depth=502.115),
        note='2/14/95')

species(name='CH3COCH2O2H',
        atoms='H:6 C:3 O:3',
        thermo=(NASA([300.00, 1381.00],
                    [ 4.94789761E+00,  3.60474432E-02, -2.21719933E-05,
                    6.98296874E-09, -9.21269260E-13, -3.88687178E+04,
                    8.49926130E+00]),
                NASA([1381.00, 5000.00],
                    [ 1.52372810E+01,  1.44114651E-02, -5.01290009E-06,
                    7.87071229E-10, -4.60225784E-14, -4.27564444E+04,
                    -4.77383784E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.449,
                                well_depth=505.403),
        note='2/14/95')

species(name='CH3COCH2O',
        atoms='H:5 C:3 O:2',
        thermo=(NASA([300.00, 1370.00],
                    [ 3.72927689E+00,  2.63943697E-02, -1.09796486E-05,
                    8.58185303E-10,  3.39474590E-13, -1.91551565E+04,
                    1.18505335E+01]),
                NASA([1370.00, 5000.00],
                    [ 1.14637586E+01,  1.32124342E-02, -4.56580495E-06,
                    7.13897538E-10, -4.16281047E-14, -2.23833017E+04,
                    -3.15127868E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.086,
                                well_depth=447.639),
        note='4/3/0THERM')

species(name='C2H3CHO',
        atoms='H:4 C:3 O:1',
        thermo=(NASA([300.00, 1393.00],
                    [ 2.92355162E-01,  3.54321417E-02, -2.94936324E-05,
                    1.28100124E-08, -2.26144108E-12, -1.16521584E+04,
                    2.28878280E+01]),
                NASA([1393.00, 5000.00],
                    [ 1.04184959E+01,  9.48963321E-03, -3.29310529E-06,
                    5.16279203E-10, -3.01587291E-14, -1.49630281E+04,
                    -3.07235061E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.958,
                                well_depth=428.8,
                                dipole=2.9,
                                rot_relax=1.0),
        note='6/26/95THERM')

species(name='C2H3CO',
        atoms='H:3 C:3 O:1',
        thermo=(NASA([300.00, 1402.00],
                    [ 1.36242013E+00,  3.15273972E-02, -3.00218935E-05,
                    1.48167112E-08, -2.87971530E-12,  4.25770215E+03,
                    1.72626546E+01]),
                NASA([1402.00, 5000.00],
                    [ 9.37467676E+00,  7.91296900E-03, -2.67198280E-06,
                    4.11115430E-10, -2.36978981E-14,  1.92969514E+03,
                    -2.40892696E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.12,
                                well_depth=443.2,
                                rot_relax=1.0),
        note='4/3/0THERM')

species(name='C2H5CHO',
        atoms='H:6 C:3 O:1',
        thermo=(NASA([300.00, 1375.00],
                    [ 5.93107530E+00,  8.87943855E-03,  2.03764325E-05,
                    -1.80149348E-08,  4.14085655E-12, -2.51377751E+04,
                    -5.87019974E-01]),
                NASA([1375.00, 5000.00],
                    [ 8.87216223E+00,  2.01710762E-02, -8.06487647E-06,
                    1.37617784E-09, -8.48922325E-14, -2.83848527E+04,
                    -2.35069163E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.662,
                                well_depth=435.2,
                                dipole=2.7,
                                rot_relax=1.0),
        note='8/10/4THERM')

species(name='C2H5CO',
        atoms='H:5 C:3 O:1',
        thermo=(NASA([300.00, 1362.00],
                    [ 8.35352246E+00, -4.08739727E-03,  3.64217954E-05,
                    -2.72457713E-08,  6.06440266E-12, -6.58577307E+03,
                    -1.05948346E+01]),
                NASA([1362.00, 5000.00],
                    [ 1.00147418E+01,  1.61527961E-02, -6.62976122E-06,
                    1.15290667E-09, -7.21263458E-14, -1.00430305E+04,
                    -2.88570933E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.82,
                                well_depth=424.6,
                                rot_relax=1.0),
        note='8/10/4THERM')

species(name='CH3OCH3',
        atoms='H:6 C:2 O:1',
        thermo=(NASA([300.00, 1368.00],
                    [ 1.50763450E+00,  2.39914228E-02, -8.68910500E-06,
                    -9.66835762E-11,  4.89319361E-13, -2.32810894E+04,
                    1.67317297E+01]),
                NASA([1368.00, 5000.00],
                    [ 8.27745836E+00,  1.32135539E-02, -4.53264362E-06,
                    7.05316507E-10, -4.09933283E-14, -2.61982700E+04,
                    -2.15190894E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.037,
                                well_depth=395.0,
                                dipole=1.3,
                                rot_relax=1.0),
        note='7/20/98THERM')

species(name='CH3OCH2',
        atoms='H:5 C:2 O:1',
        thermo=(NASA([300.00, 1376.00],
                    [ 2.91327415E+00,  2.03364659E-02, -9.59712342E-06,
                    2.07478525E-09, -1.71343362E-13, -1.18844240E+03,
                    1.16066817E+01]),
                NASA([1376.00, 5000.00],
                    [ 8.17137842E+00,  1.10086181E-02, -3.82352277E-06,
                    5.99637202E-10, -3.50317513E-14, -3.41941605E+03,
                    -1.78650856E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.037,
                                well_depth=395.0,
                                dipole=1.3,
                                rot_relax=1.0),
        note='7/20/98THERM')

species(name='CH3OCH2O2',
        atoms='H:5 C:2 O:3',
        thermo=(NASA([300.00, 1389.00],
                    [ 2.21029612E+00,  3.68877454E-02, -2.82561555E-05,
                    1.15730533E-08, -1.97130470E-12, -1.94940940E+04,
                    1.91463601E+01]),
                NASA([1389.00, 5000.00],
                    [ 1.24249729E+01,  1.18705986E-02, -4.07906532E-06,
                    6.35310809E-10, -3.69427867E-14, -2.29679238E+04,
                    -3.53740145E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.037,
                                well_depth=395.0,
                                dipole=1.3,
                                rot_relax=1.0),
        note='7/20/98THERM')

species(name='CH2OCH2O2H',
        atoms='H:5 C:2 O:3',
        thermo=(NASA([300.00, 1393.00],
                    [ 2.52895507E+00,  4.24128290E-02, -3.73406386E-05,
                    1.66639333E-08, -2.96443312E-12, -1.44293306E+04,
                    1.76899251E+01]),
                NASA([1393.00, 5000.00],
                    [ 1.51191783E+01,  9.23718883E-03, -3.19127505E-06,
                    4.99114678E-10, -2.91162488E-14, -1.84114867E+04,
                    -4.85706618E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.037,
                                well_depth=395.0,
                                dipole=1.3,
                                rot_relax=1.0),
        note='7/20/98THERM')

species(name='CH3OCH2O2H',
        atoms='H:6 C:2 O:3',
        thermo=(NASA([300.00, 1392.00],
                    [ 1.19855761E+00,  4.59060764E-02, -3.66252420E-05,
                    1.49318970E-08, -2.46057445E-12, -3.65363161E+04,
                    2.31339904E+01]),
                NASA([1392.00, 5000.00],
                    [ 1.49370964E+01,  1.19465829E-02, -4.12746359E-06,
                    6.45422590E-10, -3.76427939E-14, -4.11001068E+04,
                    -4.99552737E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.037,
                                well_depth=395.0,
                                dipole=1.3,
                                rot_relax=1.0),
        note='7/20/98THERM')

species(name='CH3OCH2O',
        atoms='H:5 C:2 O:2',
        thermo=(NASA([300.00, 2012.00],
                    [ 3.25889339E+00,  2.22146359E-02, -7.78556340E-06,
                    -2.41484158E-10,  4.51914496E-13, -1.92377212E+04,
                    1.23680069E+01]),
                NASA([2012.00, 5000.00],
                    [ 8.60261845E+00,  1.35772195E-02, -4.84661602E-06,
                    7.77766193E-10, -4.62633624E-14, -2.13762444E+04,
                    -1.75775023E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.037,
                                well_depth=395.0,
                                dipole=1.3,
                                rot_relax=1.0),
        note='2/9/96THERM')

species(name='O2CH2OCH2O2H',
        atoms='H:5 C:2 O:5',
        thermo=(NASA([300.00, 1402.00],
                    [ 1.99640551E+00,  5.83226232E-02, -5.53259778E-05,
                    2.59810540E-08, -4.77141005E-12, -3.27628742E+04,
                    2.44215005E+01]),
                NASA([1402.00, 5000.00],
                    [ 1.92038046E+01,  1.04394841E-02, -3.60582939E-06,
                    5.63792843E-10, -3.28807214E-14, -3.79207055E+04,
                    -6.51847273E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.037,
                                well_depth=395.0,
                                dipole=1.3,
                                rot_relax=1.0),
        note='7/20/98TRM')

species(name='HO2CH2OCHO',
        atoms='H:4 C:2 O:4',
        thermo=(NASA([300.00, 1387.00],
                    [ 3.47935703E+00,  4.02952392E-02, -3.30109296E-05,
                    1.34360117E-08, -2.18601580E-12, -5.80629934E+04,
                    1.52521392E+01]),
                NASA([1387.00, 5000.00],
                    [ 1.64584298E+01,  8.52683511E-03, -3.04113500E-06,
                    4.85596908E-10, -2.87316334E-14, -6.23959608E+04,
                    -5.38924139E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.037,
                                well_depth=395.0,
                                dipole=1.3,
                                rot_relax=1.0),
        note='8/31/99THERM')

species(name='OCH2OCHO',
        atoms='H:3 C:2 O:3',
        thermo=(NASA([300.00, 1475.00],
                    [ 5.19690837E+00,  1.58839723E-02,  3.53540547E-07,
                    -6.10456923E-09,  1.94661801E-12, -4.02242792E+04,
                    6.11645828E+00]),
                NASA([1475.00, 5000.00],
                    [ 1.20233916E+01,  8.11262659E-03, -2.91356462E-06,
                    4.67340384E-10, -2.77375525E-14, -4.33647231E+04,
                    -3.33691809E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.037,
                                well_depth=395.0,
                                dipole=1.3,
                                rot_relax=1.0),
        note='4/9/98THERM')

species(name='HOCH2OCO',
        atoms='H:3 C:2 O:3',
        thermo=(NASA([300.00, 1603.00],
                    [ 6.08180801E+00,  1.28768359E-02,  2.04419418E-06,
                    -6.10154921E-09,  1.79820559E-12, -4.39526183E+04,
                    2.54054449E+00]),
                NASA([1603.00, 5000.00],
                    [ 1.13737391E+01,  8.17663898E-03, -2.92034021E-06,
                    4.66695616E-10, -2.76276823E-14, -4.65575743E+04,
                    -2.86035265E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.037,
                                well_depth=395.0,
                                dipole=1.3,
                                rot_relax=1.0),
        note='8/31/99THERM')

species(name='CH3OCHO',
        atoms='H:4 C:2 O:2',
        thermo=(NASA([300.00, 1452.00],
                    [ 2.43526200E+00,  1.81948325E-02,  1.91236590E-06,
                    -8.44248945E-09,  2.61758910E-12, -4.46514732E+04,
                    1.49522265E+01]),
                NASA([1452.00, 5000.00],
                    [ 1.06569185E+01,  9.38112723E-03, -3.37677206E-06,
                    5.42455596E-10, -3.22295330E-14, -4.84651758E+04,
                    -3.27579443E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.037,
                                well_depth=395.0,
                                dipole=1.3,
                                rot_relax=1.0),
        note='4/15/8THERM')

species(name='CH3OCO',
        atoms='H:3 C:2 O:2',
        thermo=(NASA([300.00, 1423.00],
                    [ 2.83313145E+00,  1.53447505E-02,  1.89583962E-06,
                    -7.70200413E-09,  2.41564410E-12, -2.13431832E+04,
                    1.39524183E+01]),
                NASA([1423.00, 5000.00],
                    [ 1.02053021E+01,  7.12269668E-03, -2.57594255E-06,
                    4.15104358E-10, -2.47168227E-14, -2.46870635E+04,
                    -2.86407262E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.037,
                                well_depth=395.0,
                                dipole=1.3,
                                rot_relax=1.0),
        note='4/15/8THERM')

species(name='CH2OCHO',
        atoms='H:3 C:2 O:2',
        thermo=(NASA([300.00, 1442.00],
                    [ 2.31031671E+00,  1.80474065E-02, -2.71519637E-06,
                    -4.60918579E-09,  1.70037078E-12, -2.02910878E+04,
                    1.71549722E+01]),
                NASA([1442.00, 5000.00],
                    [ 1.00960096E+01,  7.19887066E-03, -2.59813465E-06,
                    4.18110812E-10, -2.48723387E-14, -2.36389018E+04,
                    -2.71144175E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.037,
                                well_depth=395.0,
                                dipole=1.3,
                                rot_relax=1.0),
        note='4/15/8THERM')

species(name='HE',
        atoms='He:1',
        thermo=(NASA([300.00, 1000.00],
                    [ 2.50000000E+00,  0.00000000E+00,  0.00000000E+00,
                    0.00000000E+00,  0.00000000E+00, -7.45375000E+02,
                    9.15348800E-01]),
                NASA([1000.00, 5000.00],
                    [ 2.50000000E+00,  0.00000000E+00,  0.00000000E+00,
                    0.00000000E+00,  0.00000000E+00, -7.45375000E+02,
                    9.15348900E-01])),
        transport=gas_transport(geom='atom',
                                diam=2.576,
                                well_depth=10.2),
        note='120186')

species(name='C3H8',
        atoms='H:8 C:3',
        thermo=(NASA([300.00, 1387.00],
                    [-3.08230087E-01,  3.65331801E-02, -1.97610926E-05,
                    5.15283344E-09, -5.06336649E-13, -1.39792198E+04,
                    2.40478733E+01]),
                NASA([1387.00, 5000.00],
                    [ 9.31450083E+00,  1.73577154E-02, -5.94222386E-06,
                    9.22939762E-10, -5.35576604E-14, -1.77234234E+04,
                    -2.89241517E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.81,
                                well_depth=303.4,
                                rot_relax=1.0),
        note='8/4/4THERM')

species(name='IC3H7',
        atoms='H:7 C:3',
        thermo=(NASA([300.00, 1373.00],
                    [ 1.63417589E+00,  2.40171372E-02, -4.72808067E-06,
                    -3.24354603E-09,  1.23539044E-12,  9.20752889E+03,
                    1.83848082E+01]),
                NASA([1373.00, 5000.00],
                    [ 8.14705217E+00,  1.58727106E-02, -5.44611541E-06,
                    8.47207689E-10, -4.92178752E-14,  6.18073367E+03,
                    -1.91980850E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.81,
                                well_depth=303.4,
                                rot_relax=1.0),
        note='8/4/4THERM')

species(name='NC3H7',
        atoms='H:7 C:3',
        thermo=(NASA([300.00, 1386.00],
                    [ 4.74365329E-01,  3.19190367E-02, -1.71909780E-05,
                    4.45927887E-09, -4.36675398E-13,  1.06314565E+04,
                    2.32670929E+01]),
                NASA([1386.00, 5000.00],
                    [ 8.88635047E+00,  1.52273279E-02, -5.21905860E-06,
                    8.11259208E-10, -4.71028059E-14,  7.34671584E+03,
                    -2.30727791E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.81,
                                well_depth=303.4,
                                rot_relax=1.0),
        note='8/4/4THERM')

species(name='C3H6',
        atoms='H:6 C:3',
        thermo=(NASA([300.00, 1388.00],
                    [ 3.94615444E-01,  2.89107662E-02, -1.54886808E-05,
                    3.88814209E-09, -3.37890352E-13,  1.17760132E+03,
                    2.19003736E+01]),
                NASA([1388.00, 5000.00],
                    [ 8.01595958E+00,  1.37023634E-02, -4.66249733E-06,
                    7.21254402E-10, -4.17370126E-14, -1.76749303E+03,
                    -2.00160668E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.14,
                                well_depth=307.8,
                                rot_relax=1.0),
        note='10/23/5THERM')

species(name='C3H5-A',
        atoms='H:5 C:3',
        thermo=(NASA([300.00, 1397.00],
                    [-5.29131958E-01,  3.34559100E-02, -2.53401027E-05,
                    1.02865754E-08, -1.73258340E-12,  1.94941423E+04,
                    2.46172315E+01]),
                NASA([1397.00, 5000.00],
                    [ 8.45883958E+00,  1.12695483E-02, -3.83792864E-06,
                    5.94059119E-10, -3.43918030E-14,  1.64683289E+04,
                    -2.32704266E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.22,
                                well_depth=316.0,
                                rot_relax=1.0),
        note='10/23/5THERM')

species(name='C3H5-S',
        atoms='H:5 C:3',
        thermo=(NASA([300.00, 1390.00],
                    [ 1.32807335E+00,  2.53107914E-02, -1.51530439E-05,
                    4.74345565E-09, -6.24666084E-13,  3.07981100E+04,
                    1.83328787E+01]),
                NASA([1390.00, 5000.00],
                    [ 7.88765879E+00,  1.13012591E-02, -3.84213130E-06,
                    5.93982677E-10, -3.43567175E-14,  2.83484913E+04,
                    -1.74291589E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.22,
                                well_depth=316.0,
                                rot_relax=1.0),
        note='10/23/5THERM')

species(name='C3H5-T',
        atoms='H:5 C:3',
        thermo=(NASA([300.00, 1382.00],
                    [ 2.17916644E+00,  2.03826623E-02, -7.91413834E-06,
                    4.76906187E-10,  2.70398536E-13,  2.96002535E+04,
                    1.48785684E+01]),
                NASA([1382.00, 5000.00],
                    [ 7.37492443E+00,  1.17510061E-02, -4.00021283E-06,
                    6.18947395E-10, -3.58215018E-14,  2.73982108E+04,
                    -1.43478655E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.22,
                                well_depth=316.0,
                                rot_relax=1.0),
        note='10/23/5THERM')

species(name='C3H4-P',
        atoms='H:4 C:3',
        thermo=(NASA([300.00, 1400.00],
                    [ 3.02973000E+00,  1.49896100E-02, -1.39850000E-06,
                    -3.96961900E-09,  1.38821700E-12,  2.14840800E+04,
                    8.00459400E+00]),
                NASA([1400.00, 4000.00],
                    [ 9.76810200E+00,  5.21915100E-03, -3.75314000E-07,
                    -2.99219100E-10,  5.10787800E-14,  1.86027700E+04,
                    -3.02067800E+01])),
        transport=gas_transport(geom='linear',
                                diam=4.29,
                                well_depth=324.8,
                                rot_relax=1.0),
        note='101993')

species(name='C3H4-A',
        atoms='H:4 C:3',
        thermo=(NASA([300.00, 1400.00],
                    [ 2.53983100E+00,  1.63343700E-02, -1.76495000E-06,
                    -4.64736500E-09,  1.72913100E-12,  2.25124300E+04,
                    9.93570200E+00]),
                NASA([1400.00, 4000.00],
                    [ 9.77625600E+00,  5.30213800E-03, -3.70111800E-07,
                    -3.02638600E-10,  5.08958100E-14,  1.95497200E+04,
                    -3.07706100E+01])),
        transport=gas_transport(geom='linear',
                                diam=4.29,
                                well_depth=324.8,
                                rot_relax=1.0),
        note='101993')

species(name='C3H3',
        atoms='H:3 C:3',
        thermo=(NASA([300.00, 1000.00],
                    [ 4.75420000E+00,  1.10802800E-02,  2.79332300E-07,
                    -5.47921200E-09,  1.94962900E-12,  3.98888300E+04,
                    5.85454900E-01]),
                NASA([1000.00, 4000.00],
                    [ 8.83104700E+00,  4.35719500E-03, -4.10906700E-07,
                    -2.36872300E-10,  4.37652000E-14,  3.84742000E+04,
                    -2.17791900E+01])),
        transport=gas_transport(geom='linear',
                                diam=4.29,
                                well_depth=324.8,
                                rot_relax=1.0),
        note='82489')

species(name='C3H2',
        atoms='H:2 C:3',
        thermo=(NASA([150.00, 1000.00],
                    [ 3.16671400E+00,  2.48257200E-02, -4.59163700E-05,
                    4.26801900E-08, -1.48215200E-11,  6.35042100E+04,
                    8.86944600E+00]),
                NASA([1000.00, 4000.00],
                    [ 7.67098100E+00,  2.74874900E-03, -4.37094300E-07,
                    -6.45559900E-11,  1.66388700E-14,  6.25972200E+04,
                    -1.23689000E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.1,
                                well_depth=209.0,
                                rot_relax=1.0),
        note='102193')

species(name='C3H5O',
        atoms='H:5 C:3 O:1',
        thermo=(NASA([300.00, 1380.00],
                    [ 1.19822582E+00,  3.05579837E-02, -1.80630276E-05,
                    4.86150033E-09, -4.19854562E-13,  9.58217784E+03,
                    2.15566221E+01]),
                NASA([1380.00, 5000.00],
                    [ 1.02551752E+01,  1.14983720E-02, -3.84645659E-06,
                    5.88910346E-10, -3.38557923E-14,  6.26560810E+03,
                    -2.77655042E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.82,
                                well_depth=411.0,
                                rot_relax=1.0),
        note='7/20/95THERM')

species(name='C3H6OOH1-2',
        atoms='H:7 C:3 O:2',
        thermo=(NASA([300.00, 1402.00],
                    [ 2.87774562E+00,  3.74166999E-02, -2.36058063E-05,
                    7.79930860E-09, -1.06042562E-12, -7.82368119E+02,
                    1.93941340E+01]),
                NASA([1402.00, 5000.00],
                    [ 1.24605763E+01,  1.58889526E-02, -5.32742106E-06,
                    8.15818791E-10, -4.68723250E-14, -4.20305196E+03,
                    -3.23858947E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.662,
                                well_depth=435.2,
                                dipole=2.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C3H6OOH1-3',
        atoms='H:7 C:3 O:2',
        thermo=(NASA([300.00, 1384.00],
                    [ 1.88331465E+00,  4.40156051E-02, -3.07858462E-05,
                    1.13214862E-08, -1.75323184E-12, -1.68779168E+02,
                    2.14040621E+01]),
                NASA([1384.00, 5000.00],
                    [ 1.46881564E+01,  1.49941200E-02, -5.24056505E-06,
                    8.25462711E-10, -4.83758952E-14, -4.77342863E+03,
                    -4.77984492E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.662,
                                well_depth=435.2,
                                dipole=2.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C3H6OOH2-1',
        atoms='H:7 C:3 O:2',
        thermo=(NASA([300.00, 1407.00],
                    [ 2.09193950E+00,  4.69220394E-02, -3.90280831E-05,
                    1.72381453E-08, -3.07968979E-12, -1.89377918E+03,
                    2.00178282E+01]),
                NASA([1407.00, 5000.00],
                    [ 1.42163221E+01,  1.43382450E-02, -4.78004477E-06,
                    7.29133134E-10, -4.17761973E-14, -5.67381620E+03,
                    -4.35770997E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.662,
                                well_depth=435.2,
                                dipole=2.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C3H6OOH1-2O2',
        atoms='H:7 C:3 O:4',
        thermo=(NASA([300.00, 1386.00],
                    [ 2.65196584E+00,  5.74638149E-02, -4.72190867E-05,
                    2.05591557E-08, -3.68787387E-12, -2.08829371E+04,
                    2.01547955E+01]),
                NASA([1386.00, 5000.00],
                    [ 1.91759159E+01,  1.59857013E-02, -5.61306378E-06,
                    8.86880495E-10, -5.20877040E-14, -2.64412115E+04,
                    -6.77512936E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.662,
                                well_depth=435.2,
                                dipole=2.7,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C3H6OOH1-3O2',
        atoms='H:7 C:3 O:4',
        thermo=(NASA([300.00, 1382.00],
                    [ 3.14864588E+00,  5.33542571E-02, -4.08330611E-05,
                    1.67542220E-08, -2.89288326E-12, -1.85473645E+04,
                    1.86305431E+01]),
                NASA([1382.00, 5000.00],
                    [ 1.85916698E+01,  1.65328553E-02, -5.81343626E-06,
                    9.19396939E-10, -5.40320702E-14, -2.39598698E+04,
                    -6.42544402E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.662,
                                well_depth=435.2,
                                dipole=2.7,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C3H6OOH2-1O2',
        atoms='H:7 C:3 O:4',
        thermo=(NASA([300.00, 1386.00],
                    [ 2.65196584E+00,  5.74638149E-02, -4.72190867E-05,
                    2.05591557E-08, -3.68787387E-12, -2.08829371E+04,
                    2.01547955E+01]),
                NASA([1386.00, 5000.00],
                    [ 1.91759159E+01,  1.59857013E-02, -5.61306378E-06,
                    8.86880495E-10, -5.20877040E-14, -2.64412115E+04,
                    -6.77512936E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.662,
                                well_depth=435.2,
                                dipole=2.7,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C3H6OOH2-2',
        atoms='H:7 C:3 O:2',
        thermo=(NASA([300.00, 1407.00],
                    [ 2.09193950E+00,  4.69220394E-02, -3.90280831E-05,
                    1.72381453E-08, -3.07968979E-12, -1.89377918E+03,
                    2.00178282E+01]),
                NASA([1407.00, 5000.00],
                    [ 1.42163221E+01,  1.43382450E-02, -4.78004477E-06,
                    7.29133134E-10, -4.17761973E-14, -5.67381620E+03,
                    -4.35770997E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.662,
                                well_depth=435.2,
                                dipole=2.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC3H7O2H',
        atoms='H:8 C:3 O:2',
        thermo=(NASA([300.00, 1385.00],
                    [ 1.13366696E+00,  4.84919904E-02, -3.31801457E-05,
                    1.19308278E-08, -1.80905372E-12, -2.47845493E+04,
                    2.27256088E+01]),
                NASA([1385.00, 5000.00],
                    [ 1.50895501E+01,  1.71714707E-02, -5.98498879E-06,
                    9.40964412E-10, -5.50728381E-14, -2.98352445E+04,
                    -5.28090166E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.997,
                                well_depth=481.5,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='IC3H7O2H',
        atoms='H:8 C:3 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 5.19265570E-01,  5.32111228E-02, -4.05156892E-05,
                    1.63346713E-08, -2.73751233E-12, -2.71048486E+04,
                    2.40815065E+01]),
                NASA([1389.00, 5000.00],
                    [ 1.57046391E+01,  1.65924516E-02, -5.77250934E-06,
                    9.06453070E-10, -5.30084111E-14, -3.23268564E+04,
                    -5.71735156E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.036,
                                well_depth=459.5,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC3H7O2',
        atoms='H:7 C:3 O:2',
        thermo=(NASA([300.00, 1384.00],
                    [ 2.10731492E+00,  3.96164986E-02, -2.49491599E-05,
                    8.59450300E-09, -1.31240330E-12, -7.93745567E+03,
                    1.89082748E+01]),
                NASA([1384.00, 5000.00],
                    [ 1.26327059E+01,  1.69910726E-02, -5.88866873E-06,
                    9.22195396E-10, -5.38230675E-14, -1.19194652E+04,
                    -3.85349297E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.997,
                                well_depth=481.5,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='IC3H7O2',
        atoms='H:7 C:3 O:2',
        thermo=(NASA([300.00, 1388.00],
                    [ 1.49941639E+00,  4.43081205E-02, -3.22414456E-05,
                    1.29687136E-08, -2.23370569E-12, -1.02587980E+04,
                    2.02336490E+01]),
                NASA([1388.00, 5000.00],
                    [ 1.32493493E+01,  1.64082190E-02, -5.67432062E-06,
                    8.87336340E-10, -5.17361535E-14, -1.44109855E+04,
                    -4.29066213E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.036,
                                well_depth=459.5,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC3H7O',
        atoms='H:7 C:3 O:1',
        thermo=(NASA([300.00, 1398.00],
                    [-2.16869993E-01,  4.14789744E-02, -2.68724173E-05,
                    8.76685704E-09, -1.13947550E-12, -6.21358042E+03,
                    2.78942641E+01]),
                NASA([1398.00, 5000.00],
                    [ 1.14771726E+01,  1.54936722E-02, -5.29459014E-06,
                    8.21476970E-10, -4.76382658E-14, -1.03998532E+04,
                    -3.53681150E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.997,
                                well_depth=481.5,
                                dipole=1.7,
                                rot_relax=1.0),
        note='8/9/4THERM')

species(name='IC3H7O',
        atoms='H:7 C:3 O:1',
        thermo=(NASA([300.00, 1393.00],
                    [-8.18260771E-01,  4.68807147E-02, -3.74369377E-05,
                    1.55917264E-08, -2.64308171E-12, -8.15153216E+03,
                    2.79483196E+01]),
                NASA([1393.00, 5000.00],
                    [ 1.23135031E+01,  1.38062606E-02, -4.91585531E-06,
                    8.21075632E-10, -5.07493001E-14, -1.24717023E+04,
                    -4.17270532E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.036,
                                well_depth=459.5,
                                dipole=1.7,
                                rot_relax=1.0),
        note='8/9/4THERM')

species(name='C3H6O1-2',
        atoms='H:6 C:3 O:1',
        thermo=(NASA([300.00, 1379.00],
                    [-1.21988153E+00,  4.28578772E-02, -3.17530249E-05,
                    1.21763736E-08, -1.94154303E-12, -1.22953761E+04,
                    2.88467826E+01]),
                NASA([1379.00, 5000.00],
                    [ 1.19825289E+01,  1.23964696E-02, -4.39113813E-06,
                    6.97911182E-10, -4.11585127E-14, -1.69739189E+04,
                    -4.22894035E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.968,
                                well_depth=403.6,
                                dipole=2.0,
                                rot_relax=1.0),
        note='1/22/95THERM')

species(name='C3H6O1-3',
        atoms='H:6 C:3 O:1',
        thermo=(NASA([300.00, 2035.00],
                    [-2.75717996E+00,  4.26371602E-02, -2.56831231E-05,
                    6.62981208E-09, -5.31193325E-13, -1.00990375E+04,
                    3.67181804E+01]),
                NASA([2035.00, 5000.00],
                    [ 1.02363953E+01,  1.50514705E-02, -5.63636534E-06,
                    9.32463455E-10, -5.66102441E-14, -1.46212029E+04,
                    -3.36724732E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.968,
                                well_depth=403.6,
                                dipole=2.0,
                                rot_relax=1.0),
        note='1/22/95THERM')

species(name='C3KET12',
        atoms='H:6 C:3 O:3',
        thermo=(NASA([300.00, 1388.00],
                    [ 1.10507238E+00,  5.27396706E-02, -4.31805774E-05,
                    1.81618292E-08, -3.10535568E-12, -3.63627536E+04,
                    2.54111636E+01]),
                NASA([1388.00, 5000.00],
                    [ 1.70756225E+01,  1.31013491E-02, -4.61949408E-06,
                    7.31991329E-10, -4.30788679E-14, -4.17008637E+04,
                    -5.95778952E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.009,
                                well_depth=464.2,
                                dipole=2.6,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C3KET13',
        atoms='H:6 C:3 O:3',
        thermo=(NASA([300.00, 1379.00],
                    [ 3.55241022E+00,  4.18720270E-02, -2.94550370E-05,
                    1.09982900E-08, -1.75045977E-12, -3.48902671E+04,
                    1.42082894E+01]),
                NASA([1379.00, 5000.00],
                    [ 1.58927479E+01,  1.40990923E-02, -4.96118851E-06,
                    7.84992198E-10, -4.61488928E-14, -3.93774829E+04,
                    -5.26049341E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.009,
                                well_depth=464.2,
                                dipole=2.6,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C3KET21',
        atoms='H:6 C:3 O:3',
        thermo=(NASA([300.00, 1371.00],
                    [ 4.55686367E+00,  3.57076837E-02, -1.94712054E-05,
                    4.70695431E-09, -3.69753807E-13, -3.86710975E+04,
                    9.97761694E+00]),
                NASA([1371.00, 5000.00],
                    [ 1.56377776E+01,  1.44059342E-02, -5.08808082E-06,
                    8.07076119E-10, -4.75295650E-14, -4.30657975E+04,
                    -5.13105869E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.009,
                                well_depth=464.2,
                                dipole=2.6,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C3H51-2,3OOH',
        atoms='H:7 C:3 O:4',
        thermo=(NASA([300.00, 1386.00],
                    [ 2.55619708E+00,  6.13504487E-02, -5.23205391E-05,
                    2.28208029E-08, -4.02231508E-12, -1.31353414E+04,
                    2.21043799E+01]),
                NASA([1386.00, 5000.00],
                    [ 2.12378169E+01,  1.39519596E-02, -4.94539222E-06,
                    7.86381389E-10, -4.63925564E-14, -1.92864584E+04,
                    -7.69636561E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.997,
                                well_depth=481.5,
                                dipole=1.7,
                                rot_relax=1.0),
        note='8/26/3THRM')

species(name='C3H52-1,3OOH',
        atoms='H:7 C:3 O:4',
        thermo=(NASA([300.00, 1379.00],
                    [ 4.12253742E+00,  5.19553611E-02, -3.83733727E-05,
                    1.45851637E-08, -2.29820536E-12, -1.22759164E+04,
                    1.48367359E+01]),
                NASA([1379.00, 5000.00],
                    [ 2.02817964E+01,  1.48155431E-02, -5.25503386E-06,
                    8.35963453E-10, -4.93308915E-14, -1.80085066E+04,
                    -7.22688262E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.997,
                                well_depth=481.5,
                                dipole=1.7,
                                rot_relax=1.0),
        note='8/26/3THRM')

species(name='C3H6OH',
        atoms='H:7 C:3 O:1',
        thermo=(NASA([300.00, 1674.00],
                    [ 1.20494302E+00,  3.30857885E-02, -1.63893637E-05,
                    3.18103918E-09, -6.84229288E-14, -9.12961720E+03,
                    2.47155202E+01]),
                NASA([1674.00, 5000.00],
                    [ 9.31287816E+00,  1.67579212E-02, -5.75555480E-06,
                    9.00584362E-10, -5.26566836E-14, -1.20169635E+04,
                    -1.95011064E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.82,
                                well_depth=487.9,
                                rot_relax=1.0),
        note='1/7/98THERM')

species(name='HOC3H6O2',
        atoms='H:7 C:3 O:3',
        thermo=(NASA([300.00, 1397.00],
                    [ 3.06954289E+00,  4.43105640E-02, -3.18636260E-05,
                    1.26409287E-08, -2.12751901E-12, -2.75893676E+04,
                    1.83891218E+01]),
                NASA([1397.00, 5000.00],
                    [ 1.42691004E+01,  1.71837825E-02, -5.83419536E-06,
                    9.00994191E-10, -5.20716210E-14, -3.14501750E+04,
                    -4.15295035E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.82,
                                well_depth=487.9,
                                rot_relax=1.0),
        note='1/7/98THERM')

species(name='CH3CHCO',
        atoms='H:4 C:3 O:1',
        thermo=(NASA([300.00, 1400.00],
                    [ 1.48380119E+00,  3.22203013E-02, -2.70250033E-05,
                    1.20499164E-08, -2.18365931E-12, -1.15276540E+04,
                    1.71552068E+01]),
                NASA([1400.00, 5000.00],
                    [ 1.00219123E+01,  9.56966300E-03, -3.26221644E-06,
                    5.05231706E-10, -2.92593257E-14, -1.42482738E+04,
                    -2.77829973E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.12,
                                well_depth=443.2,
                                rot_relax=1.0),
        note='03/03/95THERM')

species(name='AC3H5OOH',
        atoms='H:6 C:3 O:2',
        thermo=(NASA([300.00, 1394.00],
                    [ 2.43934647E+00,  4.02070638E-02, -2.95322679E-05,
                    1.14715600E-08, -1.85170511E-12, -9.43679906E+03,
                    1.70549969E+01]),
                NASA([1394.00, 5000.00],
                    [ 1.36837693E+01,  1.33968049E-02, -4.61533631E-06,
                    7.19988958E-10, -4.19109988E-14, -1.33165248E+04,
                    -4.31904193E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.997,
                                well_depth=481.5,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/20/98THERM')

species(name='C2H3OOH',
        atoms='H:4 C:2 O:2',
        thermo=(NASA([300.00, 1397.00],
                    [ 1.35644398E+00,  3.37002447E-02, -2.75988500E-05,
                    1.14222854E-08, -1.89488886E-12, -5.49996692E+03,
                    1.98354466E+01]),
                NASA([1397.00, 5000.00],
                    [ 1.15749951E+01,  8.09909174E-03, -2.81808668E-06,
                    4.42697954E-10, -2.58998042E-14, -8.84852664E+03,
                    -3.43859117E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.958,
                                well_depth=428.8,
                                dipole=2.9,
                                rot_relax=1.0),
        note='4/18/8THERM')

species(name='C4H10',
        atoms='H:10 C:4',
        thermo=(NASA([300.00, 1392.00],
                    [-4.55756824E-01,  4.80323389E-02, -2.65497552E-05,
                    6.92544700E-09, -6.38317504E-13, -1.68960904E+04,
                    2.64870966E+01]),
                NASA([1392.00, 5000.00],
                    [ 1.24940183E+01,  2.17726258E-02, -7.44272215E-06,
                    1.15487023E-09, -6.69712949E-14, -2.18403437E+04,
                    -4.45558921E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.206,
                                well_depth=350.9,
                                rot_relax=1.0),
        note='8/4/4THERM')

species(name='C4H8-1',
        atoms='H:8 C:4',
        thermo=(NASA([300.00, 1391.00],
                    [-8.33377133E-01,  4.54141123E-02, -2.97609030E-05,
                    1.03518354E-08, -1.51975733E-12, -1.49161811E+03,
                    2.94327905E+01]),
                NASA([1391.00, 5000.00],
                    [ 1.13456574E+01,  1.80843428E-02, -6.17276514E-06,
                    9.56925815E-10, -5.54586212E-14, -5.88659450E+03,
                    -3.64627206E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.65,
                                well_depth=355.0,
                                rot_relax=1.0),
        note='6/29/4THERM')

species(name='C4H8-2',
        atoms='H:8 C:4',
        thermo=(NASA([300.00, 1384.00],
                    [ 6.69213599E-01,  3.76839941E-02, -1.85689683E-05,
                    3.69571062E-09, -1.03095209E-13, -3.11675402E+03,
                    2.14868169E+01]),
                NASA([1384.00, 5000.00],
                    [ 1.07864376E+01,  1.84629194E-02, -6.28289490E-06,
                    9.72052191E-10, -5.62579166E-14, -7.11649782E+03,
                    -3.44892763E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.65,
                                well_depth=355.0,
                                rot_relax=1.0),
        note='6/29/4THERM')

species(name='PC4H9',
        atoms='H:9 C:4',
        thermo=(NASA([300.00, 1391.00],
                    [ 3.20730933E-01,  4.34654454E-02, -2.40584970E-05,
                    6.28245308E-09, -5.80113166E-13,  7.71490893E+03,
                    2.57301085E+01]),
                NASA([1391.00, 5000.00],
                    [ 1.20779744E+01,  1.96264778E-02, -6.71302199E-06,
                    1.04206424E-09, -6.04469282E-14,  3.22550473E+03,
                    -3.87719384E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.24,
                                well_depth=352.0,
                                rot_relax=1.0),
        note='8/4/4THERM')

species(name='SC4H9',
        atoms='H:9 C:4',
        thermo=(NASA([300.00, 1381.00],
                    [ 8.49159986E-01,  3.82085320E-02, -1.49626797E-05,
                    2.04499211E-10,  8.24254437E-13,  6.38832956E+03,
                    2.44466606E+01]),
                NASA([1381.00, 5000.00],
                    [ 1.16934304E+01,  1.96402287E-02, -6.65306517E-06,
                    1.02631895E-09, -5.92826294E-14,  1.96382429E+03,
                    -3.61626672E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.24,
                                well_depth=352.0,
                                rot_relax=1.0),
        note='8/4/4THERM')

species(name='C4H71-1',
        atoms='H:7 C:4',
        thermo=(NASA([300.00, 1393.00],
                    [ 1.19943124E-01,  4.15804401E-02, -2.89254150E-05,
                    1.08174803E-08, -1.70622296E-12,  2.80390734E+04,
                    2.58581923E+01]),
                NASA([1393.00, 5000.00],
                    [ 1.12162042E+01,  1.56658034E-02, -5.34224661E-06,
                    8.27633431E-10, -4.79428797E-14,  2.41404271E+04,
                    -3.38119007E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.65,
                                well_depth=355.0,
                                rot_relax=1.0),
        note='4/2/97THERM')

species(name='C4H71-2',
        atoms='H:7 C:4',
        thermo=(NASA([300.00, 1389.00],
                    [ 9.64124370E-01,  3.66781354E-02, -2.17105448E-05,
                    6.55880219E-09, -8.11791145E-13,  2.68423270E+04,
                    2.24365633E+01]),
                NASA([1389.00, 5000.00],
                    [ 1.07092407E+01,  1.61102825E-02, -5.49846133E-06,
                    8.52304018E-10, -4.93905022E-14,  2.31874323E+04,
                    -3.07648054E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.65,
                                well_depth=355.0,
                                rot_relax=1.0),
        note='4/2/97THERM')

species(name='C4H71-3',
        atoms='H:7 C:4',
        thermo=(NASA([300.00, 1392.00],
                    [-1.12475112E+00,  4.34882854E-02, -2.87750454E-05,
                    9.88046971E-09, -1.40120341E-12,  1.54420160E+04,
                    2.97651156E+01]),
                NASA([1392.00, 5000.00],
                    [ 1.09470170E+01,  1.63724387E-02, -5.59896200E-06,
                    8.69147693E-10, -5.04214389E-14,  1.11181741E+04,
                    -3.54908400E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.65,
                                well_depth=355.0,
                                rot_relax=1.0),
        note='4/2/97THERM')

species(name='C4H71-4',
        atoms='H:7 C:4',
        thermo=(NASA([300.00, 1392.00],
                    [-4.09581101E-02,  4.06024409E-02, -2.67339769E-05,
                    9.28995844E-09, -1.35368138E-12,  2.30309664E+04,
                    2.79985374E+01]),
                NASA([1392.00, 5000.00],
                    [ 1.09215027E+01,  1.59294073E-02, -5.43642246E-06,
                    8.42695797E-10, -4.88353299E-14,  1.90921144E+04,
                    -3.12721325E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.65,
                                well_depth=355.0,
                                rot_relax=1.0),
        note='4/2/97THERM')

species(name='C4H72-2',
        atoms='H:7 C:4',
        thermo=(NASA([300.00, 1379.00],
                    [ 2.43874352E+00,  2.92295980E-02, -1.10907624E-05,
                    3.35089369E-10,  4.95669524E-13,  2.50864168E+04,
                    1.52219548E+01]),
                NASA([1379.00, 5000.00],
                    [ 1.01554133E+01,  1.65013248E-02, -5.61672328E-06,
                    8.69103603E-10, -5.03037094E-14,  2.18242102E+04,
                    -2.81884706E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.65,
                                well_depth=355.0,
                                rot_relax=1.0),
        note='4/2/97THERM')

species(name='C4H6',
        atoms='H:6 C:4',
        thermo=(NASA([300.00, 1398.00],
                    [-1.43095121E+00,  4.78706062E-02, -4.15446800E-05,
                    1.91549552E-08, -3.57158507E-12,  1.17551314E+04,
                    2.90825833E+01]),
                NASA([1398.00, 5000.00],
                    [ 1.11633789E+01,  1.37163965E-02, -4.69715783E-06,
                    7.29693836E-10, -4.23486203E-14,  7.79039770E+03,
                    -3.69847949E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.72,
                                well_depth=357.0,
                                rot_relax=1.0),
        note='4/4/0THERM')

species(name='PC4H9O2H',
        atoms='H:10 C:4 O:2',
        thermo=(NASA([300.00, 1387.00],
                    [ 6.34644286E-01,  6.20007657E-02, -4.34477749E-05,
                    1.61407616E-08, -2.53121507E-12, -2.76345805E+04,
                    2.67026090E+01]),
                NASA([1387.00, 5000.00],
                    [ 1.82687454E+01,  2.16940079E-02, -7.54629828E-06,
                    1.18484806E-09, -6.92818498E-14, -3.39441633E+04,
                    -6.84816300E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='SC4H9O2H',
        atoms='H:10 C:4 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [ 3.40587657E-01,  6.52078069E-02, -4.84840460E-05,
                    1.91011830E-08, -3.13833073E-12, -3.00007960E+04,
                    2.72835451E+01]),
                NASA([1390.00, 5000.00],
                    [ 1.88587916E+01,  2.11200987E-02, -7.33205568E-06,
                    1.14969842E-09, -6.71656991E-14, -3.64236058E+04,
                    -7.20026458E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='PC4H9O2',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1385.00],
                    [ 1.94363650E+00,  5.15513163E-02, -3.28284400E-05,
                    1.13064860E-08, -1.70118606E-12, -1.08358103E+04,
                    2.13503149E+01]),
                NASA([1385.00, 5000.00],
                    [ 1.57845448E+01,  2.15210910E-02, -7.44909017E-06,
                    1.16558071E-09, -6.79885609E-14, -1.60146054E+04,
                    -5.40388525E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='SC4H9O2',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 1.32689044E+00,  5.62785583E-02, -4.01717786E-05,
                    1.57120967E-08, -2.62948443E-12, -1.31557057E+04,
                    2.34069659E+01]),
                NASA([1389.00, 5000.00],
                    [ 1.64031135E+01,  2.09361006E-02, -7.23393011E-06,
                    1.13058996E-09, -6.58938667E-14, -1.85074517E+04,
                    -5.77331636E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='PC4H9O',
        atoms='H:9 C:4 O:1',
        thermo=(NASA([300.00, 1403.00],
                    [-4.99964924E-01,  5.37157310E-02, -3.44426650E-05,
                    1.08145957E-08, -1.29600044E-12, -9.11644218E+03,
                    3.09183423E+01]),
                NASA([1403.00, 5000.00],
                    [ 1.49315588E+01,  1.95927054E-02, -6.66958265E-06,
                    1.03222635E-09, -5.97583630E-14, -1.46178979E+04,
                    -5.25561919E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='8/9/4THERM')

species(name='SC4H9O',
        atoms='H:9 C:4 O:1',
        thermo=(NASA([300.00, 1679.00],
                    [-4.09000900E-01,  5.38573427E-02, -3.39105695E-05,
                    1.01023834E-08, -1.11268025E-12, -1.11109627E+04,
                    2.88554742E+01]),
                NASA([1679.00, 5000.00],
                    [ 1.43323395E+01,  2.04542365E-02, -7.12271896E-06,
                    1.12545447E-09, -6.62697853E-14, -1.60006806E+04,
                    -5.02030895E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='8/9/4THERM')

species(name='C4H7O',
        atoms='H:7 C:4 O:1',
        thermo=(NASA([300.00, 1395.00],
                    [-1.60619192E+00,  5.58562682E-02, -4.35595767E-05,
                    1.70589279E-08, -2.65635180E-12,  4.85090326E+03,
                    3.47112559E+01]),
                NASA([1395.00, 5000.00],
                    [ 1.53137780E+01,  1.43427017E-02, -4.81625517E-06,
                    7.39574839E-10, -4.26140814E-14, -7.29342884E+02,
                    -5.52937859E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='4/3/0THERM')

species(name='C4H8O1-2',
        atoms='H:8 C:4 O:1',
        thermo=(NASA([300.00, 1399.00],
                    [-2.42033073E+00,  5.79308508E-02, -4.30236499E-05,
                    1.66679793E-08, -2.64714350E-12, -1.53950880E+04,
                    3.66425242E+01]),
                NASA([1399.00, 5000.00],
                    [ 1.39197290E+01,  1.85551492E-02, -6.36014179E-06,
                    9.88844645E-10, -5.74274779E-14, -2.09452548E+04,
                    -5.06788231E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='4/3/0THERM')

species(name='C4H8O1-3',
        atoms='H:8 C:4 O:1',
        thermo=(NASA([300.00, 1371.00],
                    [-2.53690104E+00,  5.43995707E-02, -3.43390305E-05,
                    1.01079922E-08, -1.10262736E-12, -1.52980680E+04,
                    3.67400719E+01]),
                NASA([1371.00, 5000.00],
                    [ 1.54227092E+01,  1.70211052E-02, -6.06347951E-06,
                    9.67354762E-10, -5.71992419E-14, -2.20194174E+04,
                    -6.13871862E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='1/22/95THERM')

species(name='C4H8O1-4',
        atoms='H:8 C:4 O:1',
        thermo=(NASA([300.00, 1361.00],
                    [-3.28505561E+00,  5.04800902E-02, -2.51998984E-05,
                    3.65743744E-09,  3.94863161E-13, -2.30067422E+04,
                    4.19349253E+01]),
                NASA([1361.00, 5000.00],
                    [ 1.42360731E+01,  1.81176105E-02, -6.46263809E-06,
                    1.03194372E-09, -6.10557331E-14, -2.99478670E+04,
                    -5.52081041E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='1/22/95THERM')

species(name='C4H8O2-3',
        atoms='H:8 C:4 O:1',
        thermo=(NASA([300.00, 1384.00],
                    [-2.98701197E+00,  6.32678880E-02, -5.20855777E-05,
                    2.24064364E-08, -3.95416618E-12, -1.71957196E+04,
                    3.66920390E+01]),
                NASA([1384.00, 5000.00],
                    [ 1.58263859E+01,  1.64400938E-02, -5.80680311E-06,
                    9.21146000E-10, -5.42511002E-14, -2.35334573E+04,
                    -6.34844562E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='1/22/95THERM')

species(name='PC4H8OH',
        atoms='H:9 C:4 O:1',
        thermo=(NASA([300.00, 1503.00],
                    [ 1.20329680E+00,  4.40218297E-02, -2.12295754E-05,
                    3.03714993E-09,  3.72027554E-13, -1.18622440E+04,
                    2.65515175E+01]),
                NASA([1503.00, 5000.00],
                    [ 1.40357398E+01,  1.94173155E-02, -6.50230141E-06,
                    9.95481694E-10, -5.72034525E-14, -1.67631276E+04,
                    -4.41162528E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.356,
                                well_depth=502.2,
                                dipole=1.8,
                                rot_relax=1.0),
        note='2/12/9THERM')

species(name='SC4H8OH',
        atoms='H:9 C:4 O:1',
        thermo=(NASA([300.00, 1392.00],
                    [-1.04565044E+00,  5.72423420E-02, -4.43109142E-05,
                    1.80944672E-08, -3.02843329E-12, -1.28718032E+04,
                    3.48383489E+01]),
                NASA([1392.00, 5000.00],
                    [ 1.48421985E+01,  1.78435665E-02, -6.16709050E-06,
                    1.00443458E-09, -6.10697983E-14, -1.81714637E+04,
                    -4.96976895E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.356,
                                well_depth=502.2,
                                dipole=1.8,
                                rot_relax=1.0),
        note='2/12/9THERM')

species(name='C4H8OH-1O2',
        atoms='H:9 C:4 O:3',
        thermo=(NASA([300.00, 1396.00],
                    [ 2.88497398E+00,  5.63287929E-02, -3.98403503E-05,
                    1.53891643E-08, -2.51940787E-12, -3.05246721E+04,
                    2.09293355E+01]),
                NASA([1396.00, 5000.00],
                    [ 1.74383247E+01,  2.16778876E-02, -7.37772628E-06,
                    1.14128811E-09, -6.60391451E-14, -3.55892620E+04,
                    -5.71247140E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.356,
                                well_depth=502.2,
                                dipole=1.8,
                                rot_relax=1.0),
        note='6/26/95THERM')

species(name='C4H8OH-2O2',
        atoms='H:9 C:4 O:3',
        thermo=(NASA([300.00, 1399.00],
                    [ 2.10003504E+00,  6.17528218E-02, -4.77238541E-05,
                    1.99145784E-08, -3.44416275E-12, -3.26698755E+04,
                    2.30901271E+01]),
                NASA([1399.00, 5000.00],
                    [ 1.82942871E+01,  2.09395389E-02, -7.12096934E-06,
                    1.10103094E-09, -6.36889768E-14, -3.80505855E+04,
                    -6.29154326E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.356,
                                well_depth=502.2,
                                dipole=1.8,
                                rot_relax=1.0),
        note='6/26/95THERM')

species(name='C4H8OOH1-1',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1379.00],
                    [ 4.36087539E+00,  4.46153162E-02, -2.14784588E-05,
                    3.14619957E-09,  3.03152649E-13, -6.89950326E+03,
                    1.24757248E+01]),
                NASA([1379.00, 5000.00],
                    [ 1.80477436E+01,  1.92137176E-02, -6.66503008E-06,
                    1.04468117E-09, -6.10170737E-14, -1.23069570E+04,
                    -6.33580990E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='3/27/97THERM')

species(name='C4H8OOH1-2',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1393.00],
                    [ 1.51698234E+00,  5.44584331E-02, -3.84524319E-05,
                    1.48145523E-08, -2.42809434E-12, -3.49728423E+03,
                    2.64916926E+01]),
                NASA([1393.00, 5000.00],
                    [ 1.58327906E+01,  2.07220436E-02, -7.09869028E-06,
                    1.10302045E-09, -6.40251984E-14, -8.53034314E+03,
                    -5.04399828E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C4H8OOH1-3',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1377.00],
                    [ 1.94106276E+00,  5.18789351E-02, -3.10411683E-05,
                    8.63568881E-09, -8.42841994E-13, -4.34315962E+03,
                    2.40230471E+01]),
                NASA([1377.00, 5000.00],
                    [ 1.76442170E+01,  1.91706536E-02, -6.57168641E-06,
                    1.02246571E-09, -5.94304735E-14, -1.01859280E+04,
                    -6.17115813E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C4H8OOH1-4',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1386.00],
                    [ 1.45017084E+00,  5.72418991E-02, -4.06431218E-05,
                    1.52821793E-08, -2.42123218E-12, -3.02905522E+03,
                    2.50737890E+01]),
                NASA([1386.00, 5000.00],
                    [ 1.78623700E+01,  1.95212501E-02, -6.80351173E-06,
                    1.06960529E-09, -6.26000540E-14, -8.87923008E+03,
                    -6.34393359E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C4H8OOH2-1',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [ 1.01711681E+00,  6.10560866E-02, -4.65789185E-05,
                    1.88024138E-08, -3.15265575E-12, -5.37391474E+03,
                    2.63059691E+01]),
                NASA([1390.00, 5000.00],
                    [ 1.84555673E+01,  1.89543719E-02, -6.59381457E-06,
                    1.03538370E-09, -6.05467101E-14, -1.13637100E+04,
                    -6.69839138E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C4H8OOH2-2',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1383.00],
                    [ 4.40329034E+00,  4.30478989E-02, -2.14076129E-05,
                    4.28267268E-09, -1.27804885E-13, -9.79432029E+03,
                    1.30514046E+01]),
                NASA([1383.00, 5000.00],
                    [ 1.63837700E+01,  2.04422768E-02, -7.04524557E-06,
                    1.09927552E-09, -6.39967086E-14, -1.45616229E+04,
                    -5.33147431E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='3/27/97THERM')

species(name='C4H8OOH2-3',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1387.00],
                    [ 1.45436266E+00,  5.63121286E-02, -3.84161425E-05,
                    1.34229297E-08, -1.92382914E-12, -6.68916116E+03,
                    2.54247554E+01]),
                NASA([1387.00, 5000.00],
                    [ 1.79338472E+01,  1.92153623E-02, -6.64486961E-06,
                    1.03940007E-09, -6.06239862E-14, -1.25767913E+04,
                    -6.36194089E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C4H8OOH2-4',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [ 1.01711681E+00,  6.10560866E-02, -4.65789185E-05,
                    1.88024138E-08, -3.15265575E-12, -5.37391474E+03,
                    2.63059691E+01]),
                NASA([1390.00, 5000.00],
                    [ 1.84555673E+01,  1.89543719E-02, -6.59381457E-06,
                    1.03538370E-09, -6.05467101E-14, -1.13637100E+04,
                    -6.69839138E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C4H8OOH1-2O2',
        atoms='H:9 C:4 O:4',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.22400728E+00,  7.04994620E-02, -5.66978827E-05,
                    2.42627751E-08, -4.29715459E-12, -2.37391490E+04,
                    2.38371533E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.23244015E+01,  2.05474775E-02, -7.19076348E-06,
                    1.13361536E-09, -6.64744383E-14, -3.05468277E+04,
                    -8.32666070E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C4H8OOH1-3O2',
        atoms='H:9 C:4 O:4',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.22400728E+00,  7.04994620E-02, -5.66978827E-05,
                    2.42627751E-08, -4.29715459E-12, -2.37391490E+04,
                    2.38371533E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.23244015E+01,  2.05474775E-02, -7.19076348E-06,
                    1.13361536E-09, -6.64744383E-14, -3.05468277E+04,
                    -8.32666070E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C4H8OOH1-4O2',
        atoms='H:9 C:4 O:4',
        thermo=(NASA([300.00, 1383.00],
                    [ 3.12351717E+00,  6.46725908E-02, -4.78357914E-05,
                    1.89656817E-08, -3.18025828E-12, -2.14664300E+04,
                    2.04322494E+01]),
                NASA([1383.00, 5000.00],
                    [ 2.17116470E+01,  2.11216172E-02, -7.40101448E-06,
                    1.16772959E-09, -6.85139846E-14, -2.80483124E+04,
                    -7.95885249E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C4H8OOH2-1O2',
        atoms='H:9 C:4 O:4',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.22400728E+00,  7.04994620E-02, -5.66978827E-05,
                    2.42627751E-08, -4.29715459E-12, -2.37391490E+04,
                    2.38371533E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.23244015E+01,  2.05474775E-02, -7.19076348E-06,
                    1.13361536E-09, -6.64744383E-14, -3.05468277E+04,
                    -8.32666070E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C4H8OOH2-3O2',
        atoms='H:9 C:4 O:4',
        thermo=(NASA([300.00, 1389.00],
                    [ 2.19248006E+00,  7.36763430E-02, -6.24428300E-05,
                    2.79371307E-08, -5.10126013E-12, -2.61773538E+04,
                    2.29568752E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.29951044E+01,  1.99238919E-02, -6.96326765E-06,
                    1.09677832E-09, -6.42747281E-14, -3.30095408E+04,
                    -8.71461958E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C4H8OOH2-4O2',
        atoms='H:9 C:4 O:4',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.22400728E+00,  7.04994620E-02, -5.66978827E-05,
                    2.42627751E-08, -4.29715459E-12, -2.37391490E+04,
                    2.38371533E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.23244015E+01,  2.05474775E-02, -7.19076348E-06,
                    1.13361536E-09, -6.64744383E-14, -3.05468277E+04,
                    -8.32666070E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='NC4KET12',
        atoms='H:8 C:4 O:3',
        thermo=(NASA([300.00, 1388.00],
                    [ 8.05387149E-01,  6.52375119E-02, -5.18373651E-05,
                    2.13454896E-08, -3.59860255E-12, -3.92395961E+04,
                    2.84908472E+01]),
                NASA([1388.00, 5000.00],
                    [ 2.02458485E+01,  1.76440360E-02, -6.19054598E-06,
                    9.77688024E-10, -5.74053258E-14, -4.58141511E+04,
                    -7.52145443E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.778,
                                well_depth=476.0,
                                dipole=2.6,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC4KET13',
        atoms='H:8 C:4 O:3',
        thermo=(NASA([300.00, 1386.00],
                    [ 2.74883461E+00,  5.86936745E-02, -4.49605895E-05,
                    1.83200130E-08, -3.11765369E-12, -4.01065878E+04,
                    1.88072090E+01]),
                NASA([1386.00, 5000.00],
                    [ 1.96430808E+01,  1.80940566E-02, -6.33063232E-06,
                    9.97860399E-10, -5.85076458E-14, -4.59588851E+04,
                    -7.16905094E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.778,
                                well_depth=476.0,
                                dipole=2.6,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC4KET14',
        atoms='H:8 C:4 O:3',
        thermo=(NASA([300.00, 1382.00],
                    [ 3.40797147E+00,  5.37386096E-02, -3.72613233E-05,
                    1.36862855E-08, -2.13798779E-12, -3.77919233E+04,
                    1.65580190E+01]),
                NASA([1382.00, 5000.00],
                    [ 1.90283822E+01,  1.86604765E-02, -6.53616227E-06,
                    1.03102482E-09, -6.04833256E-14, -4.34678787E+04,
                    -6.80210373E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.778,
                                well_depth=476.0,
                                dipole=2.6,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC4KET21',
        atoms='H:8 C:4 O:3',
        thermo=(NASA([300.00, 1376.00],
                    [ 4.69785973E+00,  4.83859623E-02, -3.05576939E-05,
                    1.02705744E-08, -1.51713483E-12, -4.18665528E+04,
                    1.06715264E+01]),
                NASA([1376.00, 5000.00],
                    [ 1.86231473E+01,  1.90969401E-02, -6.70745998E-06,
                    1.05994690E-09, -6.22564248E-14, -4.71929933E+04,
                    -6.55477892E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.778,
                                well_depth=476.0,
                                dipole=2.6,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC4KET23',
        atoms='H:8 C:4 O:3',
        thermo=(NASA([300.00, 1387.00],
                    [ 4.50575961E-01,  6.48862527E-02, -5.04446464E-05,
                    2.03085220E-08, -3.35952497E-12, -4.28779359E+04,
                    3.00153676E+01]),
                NASA([1387.00, 5000.00],
                    [ 1.99512969E+01,  1.79558596E-02, -6.31230878E-06,
                    9.98217386E-10, -5.86636973E-14, -4.95585495E+04,
                    -7.43071328E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.778,
                                well_depth=476.0,
                                dipole=2.6,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC4KET24',
        atoms='H:8 C:4 O:3',
        thermo=(NASA([300.00, 1380.00],
                    [ 2.93007283E+00,  5.39296076E-02, -3.66164919E-05,
                    1.30880106E-08, -1.99245456E-12, -4.14118815E+04,
                    1.86567511E+01]),
                NASA([1380.00, 5000.00],
                    [ 1.87707201E+01,  1.89490823E-02, -6.65182145E-06,
                    1.05080860E-09, -6.17069393E-14, -4.72319802E+04,
                    -6.73335107E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.778,
                                well_depth=476.0,
                                dipole=2.6,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C2H5COCH3',
        atoms='H:8 C:4 O:1',
        thermo=(NASA([300.00, 1381.00],
                    [ 1.60664724E+00,  4.13951332E-02, -2.23585444E-05,
                    5.78456048E-09, -5.69530338E-13, -3.07586009E+04,
                    2.02429671E+01]),
                NASA([1381.00, 5000.00],
                    [ 1.29004080E+01,  1.92211770E-02, -6.63870986E-06,
                    1.03733411E-09, -6.04509514E-14, -3.52062426E+04,
                    -4.20803599E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.413,
                                well_depth=454.0,
                                dipole=3.3,
                                rot_relax=1.0),
        note='4/3/0THERM')

species(name='C2H5COCH2',
        atoms='H:7 C:4 O:1',
        thermo=(NASA([300.00, 1383.00],
                    [ 1.54013856E+00,  4.39486258E-02, -2.97002421E-05,
                    1.05495313E-08, -1.58598769E-12, -9.50796505E+03,
                    1.99706641E+01]),
                NASA([1383.00, 5000.00],
                    [ 1.42098738E+01,  1.57866459E-02, -5.50529183E-06,
                    8.65870540E-10, -5.06913329E-14, -1.41284951E+04,
                    -4.87132911E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.413,
                                well_depth=454.0,
                                dipole=3.3,
                                rot_relax=1.0),
        note='4/3/0THERM')

species(name='CH2CH2COCH3',
        atoms='H:7 C:4 O:1',
        thermo=(NASA([300.00, 1380.00],
                    [ 2.40255609E+00,  3.67294268E-02, -1.97316510E-05,
                    5.07323216E-09, -4.99655275E-13, -6.15006886E+03,
                    1.93993386E+01]),
                NASA([1380.00, 5000.00],
                    [ 1.24694368E+01,  1.71022143E-02, -5.92156726E-06,
                    9.26816806E-10, -5.40730504E-14, -1.01378242E+04,
                    -3.62186375E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.04,
                                well_depth=440.584),
        note='6/21/95THER')

species(name='CH3CHCOCH3',
        atoms='H:7 C:4 O:1',
        thermo=(NASA([300.00, 1384.00],
                    [ 8.12941079E-01,  4.29256944E-02, -2.69230252E-05,
                    8.59326807E-09, -1.13188129E-12, -1.05247481E+04,
                    2.32952685E+01]),
                NASA([1384.00, 5000.00],
                    [ 1.31388032E+01,  1.66091073E-02, -5.76924215E-06,
                    9.04978165E-10, -5.28826527E-14, -1.51162098E+04,
                    -4.38876580E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.04,
                                well_depth=440.584),
        note='4/3/0THERM')

species(name='C2H3COCH3',
        atoms='H:6 C:4 O:1',
        thermo=(NASA([300.00, 1386.00],
                    [ 2.45578501E-01,  4.26432049E-02, -2.91126822E-05,
                    1.03478392E-08, -1.53551381E-12, -1.70305379E+04,
                    2.64430799E+01]),
                NASA([1386.00, 5000.00],
                    [ 1.25571995E+01,  1.49672645E-02, -5.20015351E-06,
                    8.15864365E-10, -4.76824406E-14, -2.14622958E+04,
                    -4.01434299E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.413,
                                well_depth=454.0,
                                dipole=3.3,
                                rot_relax=1.0),
        note='6/19/95THERM')

species(name='CH3CHOOCOCH3',
        atoms='H:7 C:4 O:3',
        thermo=(NASA([300.00, 1393.00],
                    [ 1.59405312E+00,  5.74729487E-02, -4.65071459E-05,
                    2.01816805E-08, -3.61175600E-12, -2.53445178E+04,
                    2.60857480E+01]),
                NASA([1393.00, 5000.00],
                    [ 1.72170832E+01,  1.76740453E-02, -6.10387321E-06,
                    9.53711693E-10, -5.55757409E-14, -3.05280370E+04,
                    -5.67959657E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.698,
                                well_depth=546.562),
        note='6/27/95')

species(name='CH2CHOOHCOCH3',
        atoms='H:7 C:4 O:3',
        thermo=(NASA([300.00, 1394.00],
                    [ 1.55204769E+00,  6.08201262E-02, -5.04704383E-05,
                    2.16347587E-08, -3.75674081E-12, -1.89563540E+04,
                    2.71011896E+01]),
                NASA([1394.00, 5000.00],
                    [ 1.93158016E+01,  1.56592602E-02, -5.45395604E-06,
                    8.57160086E-10, -5.01582885E-14, -2.47752201E+04,
                    -6.70296581E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.698,
                                well_depth=546.562),
        note='6/27/95')

species(name='NC3H7CHO',
        atoms='H:8 C:4 O:1',
        thermo=(NASA([300.00, 1378.00],
                    [ 1.87415959E+00,  4.19240315E-02, -2.35148779E-05,
                    6.26913673E-09, -6.09443908E-13, -2.71032194E+04,
                    1.91568574E+01]),
                NASA([1378.00, 5000.00],
                    [ 1.35988068E+01,  1.81652474E-02, -6.17844458E-06,
                    9.55980208E-10, -5.53442958E-14, -3.15845348E+04,
                    -4.51790228E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.009,
                                well_depth=464.2,
                                dipole=2.6,
                                rot_relax=1.0),
        note='9/27/95THERM')

species(name='NC3H7CO',
        atoms='H:7 C:4 O:1',
        thermo=(NASA([300.00, 1380.00],
                    [ 2.67256826E+00,  3.71198825E-02, -2.06862859E-05,
                    5.48873888E-09, -5.35864183E-13, -8.58050888E+03,
                    1.64848950E+01]),
                NASA([1380.00, 5000.00],
                    [ 1.30026331E+01,  1.63104877E-02, -5.57642899E-06,
                    8.65670629E-10, -5.02255667E-14, -1.25523385E+04,
                    -4.02608515E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.009,
                                well_depth=464.2,
                                dipole=2.6,
                                rot_relax=1.0),
        note='9/27/95THERM')

species(name='C3H6CHO-1',
        atoms='H:7 C:4 O:1',
        thermo=(NASA([300.00, 1379.00],
                    [ 2.67672303E+00,  3.73064128E-02, -2.11281405E-05,
                    5.80472681E-09, -6.09688236E-13, -2.49714183E+03,
                    1.75750933E+01]),
                NASA([1379.00, 5000.00],
                    [ 1.30322954E+01,  1.62418373E-02, -5.54388124E-06,
                    8.59723685E-10, -4.98459726E-14, -6.45915975E+03,
                    -3.92399021E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.009,
                                well_depth=464.2,
                                dipole=2.6,
                                rot_relax=1.0),
        note='9/27/95THERM')

species(name='C3H6CHO-2',
        atoms='H:7 C:4 O:1',
        thermo=(NASA([300.00, 1682.00],
                    [ 2.95067531E+00,  3.34223079E-02, -1.45356815E-05,
                    1.67282048E-09,  2.62011555E-13, -3.79034324E+03,
                    1.74324072E+01]),
                NASA([1682.00, 5000.00],
                    [ 1.11942816E+01,  1.81806772E-02, -6.35916662E-06,
                    1.00727333E-09, -5.93943618E-14, -6.86826460E+03,
                    -2.80298956E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.009,
                                well_depth=464.2,
                                dipole=2.6,
                                rot_relax=1.0),
        note='11/15/95THERM')

species(name='C3H6CHO-3',
        atoms='H:7 C:4 O:1',
        thermo=(NASA([300.00, 1387.00],
                    [ 1.35168659E+00,  4.23714083E-02, -2.69123192E-05,
                    8.70132567E-09, -1.15287455E-12, -6.91166610E+03,
                    2.09381509E+01]),
                NASA([1387.00, 5000.00],
                    [ 1.34301738E+01,  1.62250792E-02, -5.60631713E-06,
                    8.76356273E-10, -5.10864241E-14, -1.13561098E+04,
                    -4.47371262E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.009,
                                well_depth=464.2,
                                dipole=2.6,
                                rot_relax=1.0),
        note='11/15/95THERM')

species(name='C2H5CHCO',
        atoms='H:6 C:4 O:1',
        thermo=(NASA([300.00, 1400.00],
                    [ 2.76600655E-01,  4.82966593E-02, -4.02376928E-05,
                    1.76564254E-08, -3.14365357E-12, -1.41717351E+04,
                    2.47118671E+01]),
                NASA([1400.00, 5000.00],
                    [ 1.34100537E+01,  1.38766678E-02, -4.74130143E-06,
                    7.35504188E-10, -4.26457752E-14, -1.83790837E+04,
                    -4.45295352E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.016,
                                well_depth=436.95),
        note='9/27/95THERM')

species(name='SC3H5CHO',
        atoms='H:6 C:4 O:1',
        thermo=(NASA([300.00, 1390.00],
                    [ 4.35795171E-01,  4.48719314E-02, -3.36582931E-05,
                    1.33066870E-08, -2.17839128E-12, -1.60394651E+04,
                    2.37597452E+01]),
                NASA([1390.00, 5000.00],
                    [ 1.31695904E+01,  1.42484434E-02, -4.90843998E-06,
                    7.65789041E-10, -4.45834896E-14, -2.04032613E+04,
                    -4.43673205E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.009,
                                well_depth=464.2,
                                dipole=2.6,
                                rot_relax=1.0),
        note='11/15/95THER')

species(name='SC3H5CO',
        atoms='H:5 C:4 O:1',
        thermo=(NASA([300.00, 1392.00],
                    [ 1.74191343E+00,  3.97229536E-02, -3.20061901E-05,
                    1.38227925E-08, -2.46272017E-12, -6.64428100E+02,
                    1.70762023E+01]),
                NASA([1392.00, 5000.00],
                    [ 1.25514754E+01,  1.22521948E-02, -4.22382101E-06,
                    6.59184896E-10, -3.83818826E-14, -4.25349795E+03,
                    -4.02864145E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.009,
                                well_depth=464.2,
                                dipole=2.6,
                                rot_relax=1.0),
        note='11/15/95THERM')

species(name='CH2CH2CHO',
        atoms='H:5 C:3 O:1',
        thermo=(NASA([300.00, 1363.00],
                    [ 8.36272659E+00, -3.90712606E-03,  3.59751257E-05,
                    -2.69209237E-08,  5.98920261E-12, -7.55243244E+02,
                    -9.53140102E+00]),
                NASA([1363.00, 5000.00],
                    [ 9.96855598E+00,  1.61917946E-02, -6.64161740E-06,
                    1.15444520E-09, -7.21981790E-14, -4.16450354E+03,
                    -2.73971573E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.82,
                                well_depth=424.6,
                                rot_relax=1.0),
        note='8/10/4THERM')

species(name='IC4H10',
        atoms='H:10 C:4',
        thermo=(NASA([300.00, 1387.00],
                    [-9.72807331E-01,  5.08660928E-02, -3.12355006E-05,
                    1.02057756E-08, -1.44168766E-12, -1.78542290E+04,
                    2.69572545E+01]),
                NASA([1387.00, 5000.00],
                    [ 1.25375828E+01,  2.20249881E-02, -7.59339274E-06,
                    1.18498137E-09, -6.89902468E-14, -2.29421635E+04,
                    -4.67741623E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.208,
                                well_depth=335.7,
                                dipole=0.1,
                                rot_relax=1.0),
        note='8/4/4THERM')

species(name='IC4H9',
        atoms='H:9 C:4',
        thermo=(NASA([300.00, 1386.00],
                    [-2.21457835E-01,  4.63756324E-02, -2.88282920E-05,
                    9.60200646E-09, -1.39021034E-12,  6.76153637E+03,
                    2.64801220E+01]),
                NASA([1386.00, 5000.00],
                    [ 1.21276930E+01,  1.98689494E-02, -6.85937004E-06,
                    1.07141528E-09, -6.24184609E-14,  2.11952051E+03,
                    -4.08727278E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.24,
                                well_depth=352.0,
                                rot_relax=1.0),
        note='8/4/4THERM')

species(name='TC4H9',
        atoms='H:9 C:4',
        thermo=(NASA([300.00, 1372.00],
                    [ 3.04300181E+00,  2.91276259E-02, -4.01386786E-06,
                    -4.90273974E-09,  1.64014891E-12,  4.04780980E+03,
                    1.21127351E+01]),
                NASA([1372.00, 5000.00],
                    [ 1.05855083E+01,  2.11892055E-02, -7.31668360E-06,
                    1.14296173E-09, -6.65898835E-14,  3.18311189E+02,
                    -3.20671258E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.24,
                                well_depth=352.0,
                                rot_relax=1.0),
        note='8/4/4THERM')

species(name='IC4H8',
        atoms='H:8 C:4',
        thermo=(NASA([300.00, 1388.00],
                    [ 9.38433173E-01,  3.90547287E-02, -2.16437148E-05,
                    5.87267077E-09, -6.14435479E-13, -3.95452013E+03,
                    1.98337802E+01]),
                NASA([1388.00, 5000.00],
                    [ 1.12258330E+01,  1.81795798E-02, -6.20348592E-06,
                    9.61444458E-10, -5.57088057E-14, -7.90617899E+03,
                    -3.66411888E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.089,
                                well_depth=344.5,
                                dipole=0.5,
                                rot_relax=1.0),
        note='10/23/5THERM')

species(name='IC4H7',
        atoms='H:7 C:4',
        thermo=(NASA([300.00, 1393.00],
                    [-7.20881697E-04,  4.36495730E-02, -3.16385877E-05,
                    1.23984983E-08, -2.04378360E-12,  1.43654373E+04,
                    2.33234340E+01]),
                NASA([1393.00, 5000.00],
                    [ 1.16382753E+01,  1.57681299E-02, -5.38538858E-06,
                    8.35172927E-10, -4.84141083E-14,  1.03408230E+04,
                    -3.90259890E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.65,
                                well_depth=355.0,
                                rot_relax=1.0),
        note='10/23/5THERM')

species(name='TC4H9O2',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1388.00],
                    [ 1.08742583E+00,  5.82780709E-02, -4.33292697E-05,
                    1.76892535E-08, -3.06769606E-12, -1.49456640E+04,
                    2.01871963E+01]),
                NASA([1388.00, 5000.00],
                    [ 1.67061556E+01,  2.07327780E-02, -7.17596169E-06,
                    1.12282473E-09, -6.54941386E-14, -2.04046924E+04,
                    -6.35558608E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.43,
                                well_depth=502.252),
        note='7/19/0THERM')

species(name='IC4H9O2',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1387.00],
                    [ 1.21434293E+00,  5.45388311E-02, -3.67001593E-05,
                    1.34131042E-08, -2.11741793E-12, -1.18482450E+04,
                    2.34153048E+01]),
                NASA([1387.00, 5000.00],
                    [ 1.59741221E+01,  2.13534740E-02, -7.39001105E-06,
                    1.15624411E-09, -6.74408046E-14, -1.72329304E+04,
                    -5.65302409E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.43,
                                well_depth=502.252),
        note='7/19/0THERM')

species(name='TC4H8O2H-I',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1405.00],
                    [ 1.57469723E+00,  6.04648659E-02, -4.84991814E-05,
                    2.05236181E-08, -3.52974722E-12, -6.54190901E+03,
                    2.16852700E+01]),
                NASA([1405.00, 5000.00],
                    [ 1.78060253E+01,  1.84952716E-02, -6.21268423E-06,
                    9.52753962E-10, -5.48010330E-14, -1.17555816E+04,
                    -6.40180236E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.356,
                                well_depth=502.2,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='IC4H8O2H-I',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1388.00],
                    [ 9.94784793E-01,  5.89212240E-02, -4.25202225E-05,
                    1.61370574E-08, -2.55904902E-12, -4.08029057E+03,
                    2.58950880E+01]),
                NASA([1388.00, 5000.00],
                    [ 1.80246456E+01,  1.93668264E-02, -6.74676496E-06,
                    1.06040129E-09, -6.20506445E-14, -1.00858977E+04,
                    -6.57629693E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.356,
                                well_depth=502.2,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='IC4H8O2H-T',
        atoms='H:9 C:4 O:2',
        thermo=(NASA([300.00, 1397.00],
                    [ 3.77633001E+00,  4.50889269E-02, -2.62423875E-05,
                    7.74736900E-09, -9.11520393E-13, -5.98295842E+03,
                    1.53503399E+01]),
                NASA([1397.00, 5000.00],
                    [ 1.52395012E+01,  2.06496983E-02, -6.94791927E-06,
                    1.06651097E-09, -6.13782754E-14, -1.02271199E+04,
                    -4.70908585E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.356,
                                well_depth=502.2,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='IC4H8O',
        atoms='H:8 C:4 O:1',
        thermo=(NASA([300.00, 1399.00],
                    [-2.97373741E+00,  6.24618637E-02, -5.04348211E-05,
                    2.13345360E-08, -3.67382824E-12, -1.73274318E+04,
                    3.62336362E+01]),
                NASA([1399.00, 5000.00],
                    [ 1.44624830E+01,  1.80562860E-02, -6.18008791E-06,
                    9.59941341E-10, -5.57132297E-14, -2.30230616E+04,
                    -5.61190165E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.063,
                                well_depth=444.197),
        note='7/19/95THERM')

species(name='CC4H8O',
        atoms='H:8 C:4 O:1',
        thermo=(NASA([300.00, 1401.00],
                    [-5.27768230E+00,  6.52482242E-02, -5.00209030E-05,
                    1.97452179E-08, -3.15605813E-12, -1.37706522E+04,
                    4.78500688E+01]),
                NASA([1401.00, 5000.00],
                    [ 1.35131704E+01,  1.91666441E-02, -6.53902359E-06,
                    1.01366570E-09, -5.87546671E-14, -2.00311488E+04,
                    -5.22146967E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.063,
                                well_depth=444.197),
        note='6/29/95THERM')

species(name='IC4H9O',
        atoms='H:9 C:4 O:1',
        thermo=(NASA([300.00, 1393.00],
                    [-7.66695776E-01,  5.55173852E-02, -3.81796473E-05,
                    1.37862085E-08, -2.07312751E-12, -1.01111245E+04,
                    3.06331234E+01]),
                NASA([1393.00, 5000.00],
                    [ 1.46765729E+01,  2.00942091E-02, -6.90362097E-06,
                    1.07495878E-09, -6.24925271E-14, -1.55789914E+04,
                    -5.26039549E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='8/9/4THERM')

species(name='TC4H9O',
        atoms='H:9 C:4 O:1',
        thermo=(NASA([300.00, 1397.00],
                    [-5.32084074E-01,  5.71613469E-02, -4.20177228E-05,
                    1.63745425E-08, -2.64124700E-12, -1.34963439E+04,
                    2.62776957E+01]),
                NASA([1397.00, 5000.00],
                    [ 1.50819361E+01,  1.94454128E-02, -6.61333610E-06,
                    1.02279160E-09, -5.91796928E-14, -1.88111456E+04,
                    -5.71658947E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='8/9/4THERM')

species(name='IC4H9O2H',
        atoms='H:10 C:4 O:2',
        thermo=(NASA([300.00, 1388.00],
                    [ 2.44181899E-01,  6.34027841E-02, -4.49162379E-05,
                    1.67406877E-08, -2.61216478E-12, -2.86959808E+04,
                    2.72156229E+01]),
                NASA([1388.00, 5000.00],
                    [ 1.84308794E+01,  2.15337606E-02, -7.48623965E-06,
                    1.17499240E-09, -6.86890788E-14, -3.51483277E+04,
                    -7.08030846E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.45,
                                well_depth=505.54),
        note='7/19/0THERM')

species(name='TC4H9O2H',
        atoms='H:10 C:4 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [ 1.14505932E-01,  6.71507854E-02, -5.15578410E-05,
                    2.10256365E-08, -3.56467024E-12, -3.17928725E+04,
                    2.40011030E+01]),
                NASA([1390.00, 5000.00],
                    [ 1.91617041E+01,  2.09157249E-02, -7.27346821E-06,
                    1.14180989E-09, -6.67577135E-14, -3.83201687E+04,
                    -7.78230874E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.45,
                                well_depth=505.54),
        note='7/19/0THERM')

species(name='IC4H7O',
        atoms='H:7 C:4 O:1',
        thermo=(NASA([300.00, 1386.00],
                    [ 1.74700687E+00,  4.07783436E-02, -2.44750243E-05,
                    7.06502958E-09, -7.51570589E-13,  4.86979233E+03,
                    1.94535999E+01]),
                NASA([1386.00, 5000.00],
                    [ 1.33457615E+01,  1.61218588E-02, -5.44376403E-06,
                    8.38199374E-10, -4.83608280E-14,  6.11443644E+02,
                    -4.36818838E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='4/3/0THERM')

species(name='IC4H8OH',
        atoms='H:9 C:4 O:1',
        thermo=(NASA([300.00, 1376.00],
                    [ 3.29612707E+00,  3.47649647E-02, -1.02505618E-05,
                    -2.04641931E-09,  1.18879408E-12, -1.45627247E+04,
                    1.58606320E+01]),
                NASA([1376.00, 5000.00],
                    [ 1.25605997E+01,  2.10637488E-02, -7.15019648E-06,
                    1.10439262E-09, -6.38428695E-14, -1.86203249E+04,
                    -3.67889430E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='2/14/95THERM')

species(name='IO2C4H8OH',
        atoms='H:9 C:4 O:3',
        thermo=(NASA([300.00, 1395.00],
                    [ 2.88826008E+00,  5.80363124E-02, -4.26690874E-05,
                    1.71759332E-08, -2.91842186E-12, -3.33881503E+04,
                    1.83381149E+01]),
                NASA([1395.00, 5000.00],
                    [ 1.79798208E+01,  2.13452115E-02, -7.29108198E-06,
                    1.13068846E-09, -6.55402236E-14, -3.85769917E+04,
                    -6.23553522E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=496.0,
                                rot_relax=1.0),
        note='8/23/95THERM')

species(name='IC3H7CHO',
        atoms='H:8 C:4 O:1',
        thermo=(NASA([300.00, 1391.00],
                    [-2.73021382E-01,  4.89696307E-02, -3.12770049E-05,
                    1.00052945E-08, -1.27512074E-12, -2.76054737E+04,
                    2.83451139E+01]),
                NASA([1391.00, 5000.00],
                    [ 1.37501656E+01,  1.83126722E-02, -6.28572629E-06,
                    9.78250756E-10, -5.68538653E-14, -3.26936771E+04,
                    -4.77270548E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.352,
                                well_depth=436.4,
                                rot_relax=1.0),
        note='2/22/96THERM')

species(name='TC3H6CHO',
        atoms='H:7 C:4 O:1',
        thermo=(NASA([300.00, 1389.00],
                    [ 1.87052762E+00,  4.14869677E-02, -2.66815701E-05,
                    9.01531610E-09, -1.27870633E-12, -8.97730744E+03,
                    1.66174178E+01]),
                NASA([1389.00, 5000.00],
                    [ 1.31013047E+01,  1.66391865E-02, -5.68457623E-06,
                    8.81808351E-10, -5.11290161E-14, -1.30638647E+04,
                    -4.42705813E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.352,
                                well_depth=436.4,
                                rot_relax=1.0),
        note='2/22/96THERM')

species(name='IC3H7CO',
        atoms='H:7 C:4 O:1',
        thermo=(NASA([300.00, 1390.00],
                    [ 5.03452639E-01,  4.41607510E-02, -2.82139091E-05,
                    8.93548675E-09, -1.11327422E-12, -9.07755468E+03,
                    2.61991461E+01]),
                NASA([1390.00, 5000.00],
                    [ 1.33305736E+01,  1.61873930E-02, -5.56711402E-06,
                    8.67575951E-10, -5.04696549E-14, -1.37307001E+04,
                    -4.33958746E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.352,
                                well_depth=436.4,
                                rot_relax=1.0),
        note='2/22/96THERM')

species(name='IC3H6CHO',
        atoms='H:7 C:4 O:1',
        thermo=(NASA([300.00, 1390.00],
                    [ 5.21481767E-01,  4.43114357E-02, -2.86617314E-05,
                    9.30319894E-09, -1.20761563E-12, -2.99677086E+03,
                    2.68182130E+01]),
                NASA([1390.00, 5000.00],
                    [ 1.33102250E+01,  1.62097959E-02, -5.57575891E-06,
                    8.69003718E-10, -5.05554202E-14, -7.62177931E+03,
                    -4.25050854E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.352,
                                well_depth=436.4,
                                rot_relax=1.0),
        note='2/22/96THERM')

species(name='TC4H8OOH-IO2',
        atoms='H:9 C:4 O:4',
        thermo=(NASA([300.00, 1386.00],
                    [ 2.17782096E+00,  7.17785118E-02, -5.88836046E-05,
                    2.56759171E-08, -4.61678643E-12, -2.55622378E+04,
                    2.07846231E+01]),
                NASA([1386.00, 5000.00],
                    [ 2.26382889E+01,  2.03246143E-02, -7.12390770E-06,
                    1.12425028E-09, -6.59732993E-14, -3.24406877E+04,
                    -8.80387843E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.009,
                                well_depth=600.078),
        note='7/19/0TRM')

species(name='IC4H8OOH-IO2',
        atoms='H:9 C:4 O:4',
        thermo=(NASA([300.00, 1385.00],
                    [ 2.39424426E+00,  6.76572549E-02, -5.17083682E-05,
                    2.10796041E-08, -3.59960373E-12, -2.24787495E+04,
                    2.25029839E+01]),
                NASA([1385.00, 5000.00],
                    [ 2.18969581E+01,  2.09637874E-02, -7.34664900E-06,
                    1.15926594E-09, -6.80225413E-14, -2.92664889E+04,
                    -8.20540807E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.009,
                                well_depth=600.078),
        note='7/19/0TRM')

species(name='IC4H8OOH-TO2',
        atoms='H:9 C:4 O:4',
        thermo=(NASA([300.00, 1386.00],
                    [ 2.17782096E+00,  7.17785118E-02, -5.88836046E-05,
                    2.56759171E-08, -4.61678643E-12, -2.55622378E+04,
                    2.07846231E+01]),
                NASA([1386.00, 5000.00],
                    [ 2.26382889E+01,  2.03246143E-02, -7.12390770E-06,
                    1.12425028E-09, -6.59732993E-14, -3.24406877E+04,
                    -8.80387843E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.009,
                                well_depth=600.078),
        note='7/19/0TRM')

species(name='IC4KETII',
        atoms='H:8 C:4 O:3',
        thermo=(NASA([300.00, 1387.00],
                    [ 1.15501614E+00,  6.10622345E-02, -4.49711323E-05,
                    1.70514654E-08, -2.65948602E-12, -3.82747956E+04,
                    2.69612235E+01]),
                NASA([1387.00, 5000.00],
                    [ 1.95143059E+01,  1.82377395E-02, -6.38908606E-06,
                    1.00801571E-09, -5.91440350E-14, -4.46884836E+04,
                    -7.17167584E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.716,
                                well_depth=549.655),
        note='7/19/0THERM')

species(name='IC4KETIT',
        atoms='H:8 C:4 O:3',
        thermo=(NASA([300.00, 1388.00],
                    [ 1.14243741E+00,  6.33840797E-02, -4.73084738E-05,
                    1.77145373E-08, -2.67265475E-12, -4.09366796E+04,
                    2.34844867E+01]),
                NASA([1388.00, 5000.00],
                    [ 2.09369850E+01,  1.71090955E-02, -6.01892169E-06,
                    9.52353863E-10, -5.59926176E-14, -4.77819819E+04,
                    -8.27717611E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.716,
                                well_depth=549.655),
        note='7/19/0THERM')

species(name='IC4H7OH',
        atoms='H:8 C:4 O:1',
        thermo=(NASA([300.00, 1384.00],
                    [ 1.69099899E+00,  4.27168891E-02, -2.49281695E-05,
                    7.00961522E-09, -7.23262828E-13, -2.14512334E+04,
                    1.99500833E+01]),
                NASA([1384.00, 5000.00],
                    [ 1.35043419E+01,  1.78646968E-02, -5.99304371E-06,
                    9.18717641E-10, -5.28435302E-14, -2.58255688E+04,
                    -4.44645715E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.356,
                                well_depth=502.2,
                                dipole=1.8,
                                rot_relax=1.0),
        note='4/8/97THERM')

species(name='IC4H6OH',
        atoms='H:7 C:4 O:1',
        thermo=(NASA([300.00, 1390.00],
                    [ 8.63371227E-01,  4.68711282E-02, -3.43580339E-05,
                    1.33031052E-08, -2.13914975E-12, -3.14948305E+03,
                    2.29075523E+01]),
                NASA([1390.00, 5000.00],
                    [ 1.40310926E+01,  1.55317541E-02, -5.32754961E-06,
                    8.28785902E-10, -4.81545257E-14, -7.69378228E+03,
                    -4.76555306E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.356,
                                well_depth=502.2,
                                dipole=1.8,
                                rot_relax=1.0),
        note='8/19/95THERM')

species(name='IC3H5CHO',
        atoms='H:6 C:4 O:1',
        thermo=(NASA([300.00, 1396.00],
                    [ 6.27183793E-01,  4.66780254E-02, -3.74430631E-05,
                    1.58330542E-08, -2.73952155E-12, -1.57203117E+04,
                    2.16034294E+01]),
                NASA([1396.00, 5000.00],
                    [ 1.36203958E+01,  1.37917192E-02, -4.73370118E-06,
                    7.36655226E-10, -4.20097974E-14, -2.00025274E+04,
                    -4.73184531E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.352,
                                well_depth=436.4,
                                rot_relax=1.0),
        note='7/19/95THERM')

species(name='IC3H5CO',
        atoms='H:5 C:4 O:1',
        thermo=(NASA([300.00, 1397.00],
                    [ 1.85097069E+00,  4.18855846E-02, -3.62553731E-05,
                    1.65690659E-08, -3.05850846E-12,  1.70381441E+02,
                    1.53014433E+01]),
                NASA([1397.00, 5000.00],
                    [ 1.30667437E+01,  1.16704244E-02, -3.99106523E-06,
                    6.19498148E-10, -3.59348249E-14, -3.36519344E+03,
                    -4.35803090E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.352,
                                well_depth=436.4,
                                rot_relax=1.0),
        note='4/3/0THERM')

species(name='TC3H6OCHO',
        atoms='H:7 C:4 O:2',
        thermo=(NASA([300.00, 1394.00],
                    [ 3.70830259E-01,  5.38475661E-02, -3.82477565E-05,
                    1.32882237E-08, -1.79228730E-12, -2.18391262E+04,
                    2.58142112E+01]),
                NASA([1394.00, 5000.00],
                    [ 1.70371287E+01,  1.54400645E-02, -5.28332886E-06,
                    8.21085347E-10, -4.76898429E-14, -2.75871941E+04,
                    -6.37271230E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.352,
                                well_depth=436.4,
                                rot_relax=1.0),
        note='8/25/95THERM')

species(name='IC3H6CO',
        atoms='H:6 C:4 O:1',
        thermo=(NASA([300.00, 1397.00],
                    [ 2.28039055E+00,  4.17016989E-02, -3.25089661E-05,
                    1.37243419E-08, -2.40573132E-12, -1.63939712E+04,
                    1.38187714E+01]),
                NASA([1397.00, 5000.00],
                    [ 1.32548232E+01,  1.40142787E-02, -4.78910215E-06,
                    7.42924342E-10, -4.30737566E-14, -2.00529779E+04,
                    -4.44810221E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.352,
                                well_depth=436.4,
                                rot_relax=1.0),
        note='03/03/95THERM')

species(name='IC4H7OOH',
        atoms='H:8 C:4 O:2',
        thermo=(NASA([300.00, 1392.00],
                    [ 2.99117402E+00,  5.03349278E-02, -3.56280061E-05,
                    1.33952154E-08, -2.11053409E-12, -1.51095046E+04,
                    1.54537413E+01]),
                NASA([1392.00, 5000.00],
                    [ 1.69234564E+01,  1.78396769E-02, -6.14273279E-06,
                    9.57895028E-10, -5.57438304E-14, -2.00040686E+04,
                    -5.94746070E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.41,
                                well_depth=498.949),
        note='7/19/95THERM')

species(name='TC3H6OHCHO',
        atoms='H:8 C:4 O:2',
        thermo=(NASA([300.00, 1396.00],
                    [ 2.10033500E-01,  5.63086284E-02, -3.95158669E-05,
                    1.37780428E-08, -1.89068805E-12, -4.80402638E+04,
                    2.67836070E+01]),
                NASA([1396.00, 5000.00],
                    [ 1.70788996E+01,  1.74005554E-02, -5.93008112E-06,
                    9.18860909E-10, -5.32512140E-14, -5.38730427E+04,
                    -6.38638754E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.352,
                                well_depth=436.4,
                                rot_relax=1.0),
        note='8/1/95THERM')

species(name='TC3H6OH',
        atoms='H:7 C:3 O:1',
        thermo=(NASA([300.00, 1392.00],
                    [ 1.09670360E+00,  3.80727565E-02, -2.75022497E-05,
                    1.07477493E-08, -1.74895773E-12, -1.40764487E+04,
                    2.22475799E+01]),
                NASA([1392.00, 5000.00],
                    [ 1.12222277E+01,  1.36444398E-02, -4.51406709E-06,
                    7.10523275E-10, -4.22690392E-14, -1.75350136E+04,
                    -3.18911926E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.036,
                                well_depth=459.5,
                                dipole=1.7,
                                rot_relax=1.0),
        note='8/9/4THERM')

species(name='IC3H5OH',
        atoms='H:6 C:3 O:1',
        thermo=(NASA([300.00, 1374.00],
                    [ 1.58376391E+00,  3.16215366E-02, -1.73664942E-05,
                    4.18927663E-09, -2.79899620E-13, -2.12643496E+04,
                    1.88313766E+01]),
                NASA([1374.00, 5000.00],
                    [ 1.07381025E+01,  1.31698194E-02, -4.41529622E-06,
                    6.77009837E-10, -3.89608901E-14, -2.47298321E+04,
                    -3.13634050E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.036,
                                well_depth=459.5,
                                dipole=1.7,
                                rot_relax=1.0),
        note='8/1/95THERM')

species(name='TC3H6O2CHO',
        atoms='H:7 C:4 O:3',
        thermo=(NASA([300.00, 1386.00],
                    [ 2.17883383E+00,  5.41595832E-02, -3.83435886E-05,
                    1.38308104E-08, -2.04190147E-12, -2.27394154E+04,
                    2.00751264E+01]),
                NASA([1386.00, 5000.00],
                    [ 1.85534443E+01,  1.68774389E-02, -5.90752965E-06,
                    9.31518085E-10, -5.46345187E-14, -2.85447191E+04,
                    -6.82486667E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.352,
                                well_depth=436.4,
                                rot_relax=1.0),
        note='8/2/95THERM')

species(name='TC3H6O2HCO',
        atoms='H:7 C:4 O:3',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.03864428E+00,  5.80421003E-02, -4.32123528E-05,
                    1.58792094E-08, -2.32209543E-12, -2.24284673E+04,
                    2.03680990E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.06472678E+01,  1.48526500E-02, -5.25104875E-06,
                    8.33619219E-10, -4.91256069E-14, -2.88719869E+04,
                    -7.95951389E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.352,
                                well_depth=436.4,
                                rot_relax=1.0),
        note='8/2/95THERM')

species(name='IC3H5O2HCHO',
        atoms='H:7 C:4 O:3',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.05984770E+00,  5.82331716E-02, -4.37672100E-05,
                    1.63249918E-08, -2.43462051E-12, -1.63496250E+04,
                    2.13687921E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.06288832E+01,  1.48625539E-02, -5.25305276E-06,
                    8.33772951E-10, -4.91277401E-14, -2.27589076E+04,
                    -7.82962888E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.352,
                                well_depth=436.4,
                                rot_relax=1.0),
        note='8/2/95THERM')

species(name='CH2CCH2OH',
        atoms='H:5 C:3 O:1',
        thermo=(NASA([300.00, 1372.00],
                    [ 2.88422544E+00,  2.42428071E-02, -1.14152268E-05,
                    1.71775334E-09,  1.42177454E-13,  1.17935615E+04,
                    1.52102335E+01]),
                NASA([1372.00, 5000.00],
                    [ 9.70702027E+00,  1.13972660E-02, -3.77993962E-06,
                    5.75209277E-10, -3.29229125E-14,  9.13212884E+03,
                    -2.25012933E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.997,
                                well_depth=481.5,
                                dipole=1.7,
                                rot_relax=1.0),
        note='9/8/95THERM')

species(name='TC4H8CHO',
        atoms='H:9 C:5 O:1',
        thermo=(NASA([300.00, 1397.00],
                    [-9.58078294E-01,  6.42003258E-02, -4.70776827E-05,
                    1.75737698E-08, -2.64896151E-12, -6.86582501E+03,
                    3.33781112E+01]),
                NASA([1397.00, 5000.00],
                    [ 1.79663933E+01,  1.94207117E-02, -6.67409451E-06,
                    1.03969221E-09, -6.04702651E-14, -1.33368585E+04,
                    -6.79819424E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.348,
                                well_depth=489.084),
        note='9/7/95THERM')

species(name='O2C4H8CHO',
        atoms='H:9 C:5 O:3',
        thermo=(NASA([300.00, 1395.00],
                    [ 1.91847699E+00,  6.67245869E-02, -4.80871046E-05,
                    1.78588690E-08, -2.71163880E-12, -2.49837984E+04,
                    2.38577867E+01]),
                NASA([1395.00, 5000.00],
                    [ 2.12629904E+01,  2.14072282E-02, -7.38342949E-06,
                    1.15281523E-09, -6.71508438E-14, -3.16854524E+04,
                    -7.99828703E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.942,
                                well_depth=588.531),
        note='9/7/95THERM')

species(name='O2HC4H8CO',
        atoms='H:9 C:5 O:3',
        thermo=(NASA([300.00, 1394.00],
                    [ 1.82607262E+00,  6.93466111E-02, -4.93125140E-05,
                    1.69848340E-08, -2.26117657E-12, -2.46578311E+04,
                    2.41167544E+01]),
                NASA([1394.00, 5000.00],
                    [ 2.38219630E+01,  1.91411448E-02, -6.67919154E-06,
                    1.05127303E-09, -6.15876805E-14, -3.23093973E+04,
                    -9.42580755E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.942,
                                well_depth=588.531),
        note='9/7/95THERM')

species(name='C3H5OH',
        atoms='H:6 C:3 O:1',
        thermo=(NASA([300.00, 1388.00],
                    [ 1.08899314E+00,  3.28481769E-02, -1.91322648E-05,
                    5.21635136E-09, -4.82394566E-13, -1.66282463E+04,
                    2.22754535E+01]),
                NASA([1388.00, 5000.00],
                    [ 1.02818222E+01,  1.34248012E-02, -4.47039448E-06,
                    6.81924195E-10, -3.90899553E-14, -1.99960114E+04,
                    -2.77736680E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.997,
                                well_depth=481.5,
                                dipole=1.7,
                                rot_relax=1.0),
        note='9/8/95THERM')

species(name='TIC4H7Q2-I',
        atoms='H:9 C:4 O:4',
        thermo=(NASA([300.00, 1400.00],
                    [ 4.48426361E+00,  6.61225007E-02, -5.27349018E-05,
                    2.18215585E-08, -3.66788946E-12, -1.98906586E+04,
                    1.26719614E+01]),
                NASA([1400.00, 5000.00],
                    [ 2.33848631E+01,  1.87070035E-02, -6.44021945E-06,
                    1.00428123E-09, -5.84468189E-14, -2.61180902E+04,
                    -8.76610135E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.009,
                                well_depth=600.078),
        note='5/6/96THeRM')

species(name='IIC4H7Q2-T',
        atoms='H:9 C:4 O:4',
        thermo=(NASA([300.00, 1377.00],
                    [ 8.16274487E+00,  4.34463050E-02, -1.76972456E-05,
                    4.88790666E-10,  9.03915465E-13, -1.96501749E+04,
                    2.62067299E-01]),
                NASA([1377.00, 5000.00],
                    [ 2.15070321E+01,  2.05359839E-02, -7.12383399E-06,
                    1.11655053E-09, -6.52112103E-14, -2.51117508E+04,
                    -7.43379783E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.009,
                                well_depth=600.078),
        note='7/15/96THeRM')

species(name='IIC4H7Q2-I',
        atoms='H:9 C:4 O:4',
        thermo=(NASA([300.00, 1394.00],
                    [ 4.93055661E+00,  6.05819201E-02, -4.23665566E-05,
                    1.49122008E-08, -2.10978665E-12, -1.68415495E+04,
                    1.36228018E+01]),
                NASA([1394.00, 5000.00],
                    [ 2.30500244E+01,  1.92149194E-02, -6.66622576E-06,
                    1.04495725E-09, -6.10370520E-14, -2.32086881E+04,
                    -8.39949885E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.009,
                                well_depth=600.078),
        note='7/15/96THeRM')

species(name='CH2O2H',
        atoms='H:3 C:1 O:2',
        thermo=(NASA([300.00, 1357.00],
                    [ 4.47228333E+00,  1.33401095E-02, -5.92919725E-06,
                    4.44481025E-10,  2.12699899E-13,  5.67413711E+03,
                    4.72608208E+00]),
                NASA([1357.00, 5000.00],
                    [ 9.10784249E+00,  5.27260434E-03, -1.88170543E-06,
                    3.00561364E-10, -1.77865959E-14,  3.77440183E+03,
                    -2.11741044E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=3.626,
                                well_depth=481.8,
                                rot_relax=1.0),
        note='1/14/5THERM')

species(name='NC5H12',
        atoms='H:12 C:5',
        thermo=(NASA([300.00, 1396.00],
                    [-1.67881131E+00,  6.41231428E-02, -3.96458285E-05,
                    1.23445155E-08, -1.53325385E-12, -1.96474971E+04,
                    3.39348804E+01]),
                NASA([1396.00, 5000.00],
                    [ 1.59247008E+01,  2.59703101E-02, -8.86821278E-06,
                    1.37514707E-09, -7.97105419E-14, -2.60824218E+04,
                    -6.16997349E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.591,
                                well_depth=391.7,
                                rot_relax=1.0),
        note='8/4/4THERM')

species(name='C5H11-1',
        atoms='H:11 C:5',
        thermo=(NASA([300.00, 1396.00],
                    [-9.94216435E-01,  5.99860809E-02, -3.78318587E-05,
                    1.21310895E-08, -1.57042863E-12,  4.97697538E+03,
                    3.35998032E+01]),
                NASA([1396.00, 5000.00],
                    [ 1.54992372E+01,  2.38273176E-02, -8.13845161E-06,
                    1.26220569E-09, -7.31730066E-14, -1.00988975E+03,
                    -5.58563592E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='8/4/4THERM')

species(name='C5H11-2',
        atoms='H:11 C:5',
        thermo=(NASA([300.00, 1391.00],
                    [ 7.25187788E-01,  4.97343569E-02, -2.18806076E-05,
                    2.09007834E-09,  6.62752863E-13,  3.46418370E+03,
                    2.67485348E+01]),
                NASA([1391.00, 5000.00],
                    [ 1.48744992E+01,  2.40593799E-02, -8.15587903E-06,
                    1.25867109E-09, -7.27223868E-14, -2.14733945E+03,
                    -5.17907576E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='8/4/4THERM')

species(name='C5H11-3',
        atoms='H:11 C:5',
        thermo=(NASA([300.00, 1391.00],
                    [ 7.25187788E-01,  4.97343569E-02, -2.18806076E-05,
                    2.09007834E-09,  6.62752863E-13,  3.46418370E+03,
                    2.60590532E+01]),
                NASA([1391.00, 5000.00],
                    [ 1.48744992E+01,  2.40593799E-02, -8.15587903E-06,
                    1.25867109E-09, -7.27223868E-14, -2.14733945E+03,
                    -5.24802393E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='8/4/4THERM')

species(name='C5H10-1',
        atoms='H:10 C:5',
        thermo=(NASA([300.00, 1392.00],
                    [-1.06223481E+00,  5.74218294E-02, -3.74486890E-05,
                    1.27364989E-08, -1.79609789E-12, -4.46546666E+03,
                    3.22739790E+01]),
                NASA([1392.00, 5000.00],
                    [ 1.45851539E+01,  2.24072471E-02, -7.63348025E-06,
                    1.18188966E-09, -6.84385139E-14, -1.00898205E+04,
                    -5.23683936E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.489,
                                well_depth=386.2,
                                dipole=0.4,
                                rot_relax=1.0),
        note='4/7/97THERM')

species(name='C5H10-2',
        atoms='H:10 C:5',
        thermo=(NASA([300.00, 1389.00],
                    [-5.41560551E-01,  5.39629918E-02, -3.23508738E-05,
                    9.77416037E-09, -1.18534668E-12, -5.98606169E+03,
                    2.97142748E+01]),
                NASA([1389.00, 5000.00],
                    [ 1.41109267E+01,  2.28348272E-02, -7.78626835E-06,
                    1.20627491E-09, -6.98795983E-14, -1.14336507E+04,
                    -5.01601163E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.458,
                                well_depth=396.8,
                                rot_relax=1.0),
        note='4/7/97THERM')

species(name='C5H81-3',
        atoms='H:8 C:5',
        thermo=(NASA([300.00, 1396.00],
                    [-2.32617433E+00,  6.38696042E-02, -5.57052591E-05,
                    2.58049480E-08, -4.83405924E-12,  6.55183321E+03,
                    3.42036513E+01]),
                NASA([1396.00, 5000.00],
                    [ 1.45574490E+01,  1.80939266E-02, -6.21671892E-06,
                    9.67920254E-10, -5.62629326E-14,  1.22365462E+03,
                    -5.43889766E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.2,
                                well_depth=408.0,
                                rot_relax=1.0),
        note='3/27/97THERM')

species(name='C5H91-3',
        atoms='H:9 C:5',
        thermo=(NASA([300.00, 1392.00],
                    [-1.38013950E+00,  5.57608487E-02, -3.70143928E-05,
                    1.26883901E-08, -1.78538835E-12,  1.25589824E+04,
                    3.26441304E+01]),
                NASA([1392.00, 5000.00],
                    [ 1.41860454E+01,  2.07128899E-02, -7.06960617E-06,
                    1.09607133E-09, -6.35322208E-14,  7.00496135E+03,
                    -5.14501773E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.458,
                                well_depth=396.8,
                                rot_relax=1.0),
        note='4/7/97THERM')

species(name='C5H91-4',
        atoms='H:9 C:5',
        thermo=(NASA([300.00, 1379.00],
                    [ 2.07302224E-01,  4.75322572E-02, -2.55232300E-05,
                    5.70570808E-09, -2.53926602E-13,  1.88300100E+04,
                    2.97527890E+01]),
                NASA([1379.00, 5000.00],
                    [ 1.39904314E+01,  1.99962562E-02, -6.73105937E-06,
                    1.03447598E-09, -5.96148983E-14,  1.36075123E+04,
                    -4.58753046E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.458,
                                well_depth=396.8,
                                rot_relax=1.0),
        note='4/7/97THERM')

species(name='C5H91-5',
        atoms='H:9 C:5',
        thermo=(NASA([300.00, 1392.00],
                    [-3.00352280E-01,  5.29029897E-02, -3.50199571E-05,
                    1.21280172E-08, -1.74459793E-12,  2.01482388E+04,
                    3.08938514E+01]),
                NASA([1392.00, 5000.00],
                    [ 1.41604486E+01,  2.02710046E-02, -6.90765357E-06,
                    1.06972742E-09, -6.19529514E-14,  1.49795090E+04,
                    -4.72301324E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.458,
                                well_depth=396.8,
                                rot_relax=1.0),
        note='4/7/97THERM')

species(name='C5H92-4',
        atoms='H:9 C:5',
        thermo=(NASA([300.00, 1389.00],
                    [-8.61323386E-01,  5.23106875E-02, -3.19282664E-05,
                    9.73236622E-09, -1.17585595E-12,  1.10386471E+04,
                    3.00928924E+01]),
                NASA([1389.00, 5000.00],
                    [ 1.37128284E+01,  2.11391857E-02, -7.22186145E-06,
                    1.12036442E-09, -6.49675872E-14,  5.66079366E+03,
                    -4.92475603E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.458,
                                well_depth=396.8,
                                rot_relax=1.0),
        note='4/7/97THERM')

species(name='C5H92-5',
        atoms='H:9 C:5',
        thermo=(NASA([300.00, 1389.00],
                    [ 2.16350550E-01,  4.94614176E-02, -2.99438792E-05,
                    9.17728978E-09, -1.13613754E-12,  1.86282306E+04,
                    2.83575279E+01]),
                NASA([1389.00, 5000.00],
                    [ 1.36885460E+01,  2.06966684E-02, -7.05980383E-06,
                    1.09401634E-09, -6.33885833E-14,  1.36345574E+04,
                    -4.50306538E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.458,
                                well_depth=396.8,
                                rot_relax=1.0),
        note='4/7/97THERM')

species(name='C5H9O1-3',
        atoms='H:9 C:5 O:1',
        thermo=(NASA([300.00, 1393.00],
                    [-2.22532588E+00,  6.97271843E-02, -5.42285799E-05,
                    2.13591167E-08, -3.36511599E-12,  1.91826163E+03,
                    3.99711270E+01]),
                NASA([1393.00, 5000.00],
                    [ 1.86270165E+01,  1.85189638E-02, -6.21111137E-06,
                    9.52916260E-10, -5.48704082E-14, -4.97349374E+03,
                    -7.09678919E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='4/7/97THERM')

species(name='C5H9O2-4',
        atoms='H:9 C:5 O:1',
        thermo=(NASA([300.00, 1387.00],
                    [-1.66449167E+00,  6.51525448E-02, -4.57159226E-05,
                    1.55903830E-08, -2.04144207E-12,  4.10326271E+02,
                    3.84890543E+01]),
                NASA([1387.00, 5000.00],
                    [ 1.86104516E+01,  1.87100216E-02, -6.31627652E-06,
                    9.73434030E-10, -5.62316246E-14, -6.59575918E+03,
                    -7.05163465E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='4/7/97THERM')

species(name='C5H11O2H-1',
        atoms='H:12 C:5 O:2',
        thermo=(NASA([300.00, 1387.00],
                    [ 4.74596161E-01,  7.38808098E-02, -5.12216645E-05,
                    1.87797149E-08, -2.90291870E-12, -3.05323779E+04,
                    2.91362231E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.14180725E+01,  2.62251472E-02, -9.10691067E-06,
                    1.42824507E-09, -8.34473635E-14, -3.80401323E+04,
                    -8.39746734E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C5H11O2H-2',
        atoms='H:12 C:5 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [-2.68431172E-01,  7.90975970E-02, -5.89751277E-05,
                    2.33029166E-08, -3.83794655E-12, -3.28329870E+04,
                    3.17836191E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.21712264E+01,  2.55540337E-02, -8.86857346E-06,
                    1.39034021E-09, -8.12122040E-14, -4.06027396E+04,
                    -8.84830959E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C5H11O2H-3',
        atoms='H:12 C:5 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [-2.68431172E-01,  7.90975970E-02, -5.89751277E-05,
                    2.33029166E-08, -3.83794655E-12, -3.28329870E+04,
                    3.17836191E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.21712264E+01,  2.55540337E-02, -8.86857346E-06,
                    1.39034021E-09, -8.12122040E-14, -4.06027396E+04,
                    -8.84830959E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C5H11O2-1',
        atoms='H:11 C:5 O:2',
        thermo=(NASA([300.00, 1386.00],
                    [ 1.39788038E+00,  6.52788893E-02, -4.34224032E-05,
                    1.57151743E-08, -2.46631324E-12, -1.36791222E+04,
                    2.55411824E+01]),
                NASA([1386.00, 5000.00],
                    [ 1.89702536E+01,  2.60355552E-02, -9.00706834E-06,
                    1.40888868E-09, -8.21618381E-14, -2.01253683E+04,
                    -6.97477616E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C5H11O2-2',
        atoms='H:11 C:5 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 7.10786402E-01,  7.03521821E-02, -5.11565299E-05,
                    2.03005197E-08, -3.42470585E-12, -1.59903126E+04,
                    2.79113132E+01]),
                NASA([1389.00, 5000.00],
                    [ 1.96655539E+01,  2.53946113E-02, -8.77477373E-06,
                    1.37145913E-09, -7.99350015E-14, -2.26521255E+04,
                    -7.38934310E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C5H11O2-3',
        atoms='H:11 C:5 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 7.10786402E-01,  7.03521821E-02, -5.11565299E-05,
                    2.03005197E-08, -3.42470585E-12, -1.59903126E+04,
                    2.79113132E+01]),
                NASA([1389.00, 5000.00],
                    [ 1.96655539E+01,  2.53946113E-02, -8.77477373E-06,
                    1.37145913E-09, -7.99350015E-14, -2.26521255E+04,
                    -7.38934310E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C5H11O-1',
        atoms='H:11 C:5 O:1',
        thermo=(NASA([300.00, 1391.00],
                    [-9.02925873E-02,  6.34116550E-02, -4.09268668E-05,
                    1.36982820E-08, -1.89646466E-12, -1.20333690E+04,
                    3.05672313E+01]),
                NASA([1391.00, 5000.00],
                    [ 1.73225378E+01,  2.48217999E-02, -8.48159842E-06,
                    1.31587304E-09, -7.63050620E-14, -1.83356880E+04,
                    -6.37681611E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='4/7/97THERM')

species(name='C5H11O-2',
        atoms='H:11 C:5 O:1',
        thermo=(NASA([300.00, 1396.00],
                    [-4.94527725E-01,  6.73149844E-02, -4.72324734E-05,
                    1.75936532E-08, -2.73651943E-12, -1.42372243E+04,
                    3.16370708E+01]),
                NASA([1396.00, 5000.00],
                    [ 1.78183150E+01,  2.44541516E-02, -8.36505611E-06,
                    1.29863054E-09, -7.53356869E-14, -2.06350810E+04,
                    -6.67615213E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='4/7/97THERM')

species(name='C5H11O-3',
        atoms='H:11 C:5 O:1',
        thermo=(NASA([300.00, 1396.00],
                    [-4.94527725E-01,  6.73149844E-02, -4.72324734E-05,
                    1.75936532E-08, -2.73651943E-12, -1.42372243E+04,
                    3.09475892E+01]),
                NASA([1396.00, 5000.00],
                    [ 1.78183150E+01,  2.44541516E-02, -8.36505611E-06,
                    1.29863054E-09, -7.53356869E-14, -2.06350810E+04,
                    -6.74510029E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='4/7/97THERM')

species(name='C5H10OOH1-2',
        atoms='H:11 C:5 O:2',
        thermo=(NASA([300.00, 1379.00],
                    [ 1.66304681E+00,  6.42511734E-02, -3.94124754E-05,
                    1.15348101E-08, -1.25128422E-12, -7.22288006E+03,
                    2.70063739E+01]),
                NASA([1379.00, 5000.00],
                    [ 2.09225281E+01,  2.34863057E-02, -8.03644420E-06,
                    1.24880724E-09, -7.25227776E-14, -1.43245680E+04,
                    -7.79152904E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THRM')

species(name='C5H10OOH1-3',
        atoms='H:11 C:5 O:2',
        thermo=(NASA([300.00, 1379.00],
                    [ 1.66304681E+00,  6.42511734E-02, -3.94124754E-05,
                    1.15348101E-08, -1.25128422E-12, -7.22288006E+03,
                    2.70063739E+01]),
                NASA([1379.00, 5000.00],
                    [ 2.09225281E+01,  2.34863057E-02, -8.03644420E-06,
                    1.24880724E-09, -7.25227776E-14, -1.43245680E+04,
                    -7.79152904E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THRM')

species(name='C5H10OOH1-4',
        atoms='H:11 C:5 O:2',
        thermo=(NASA([300.00, 1379.00],
                    [ 1.66304681E+00,  6.42511734E-02, -3.94124754E-05,
                    1.15348101E-08, -1.25128422E-12, -7.22288006E+03,
                    2.70063739E+01]),
                NASA([1379.00, 5000.00],
                    [ 2.09225281E+01,  2.34863057E-02, -8.03644420E-06,
                    1.24880724E-09, -7.25227776E-14, -1.43245680E+04,
                    -7.79152904E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THRM')

species(name='C5H10OOH1-5',
        atoms='H:11 C:5 O:2',
        thermo=(NASA([300.00, 1387.00],
                    [ 1.26957212E+00,  6.92172253E-02, -4.85626715E-05,
                    1.80144791E-08, -2.81415382E-12, -5.92384799E+03,
                    2.76017751E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.10114488E+01,  2.40561226E-02, -8.36616955E-06,
                    1.31339875E-09, -7.67917194E-14, -1.29757849E+04,
                    -7.89329903E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.698,
                                well_depth=546.691),
        note='7/19/0THRM')

species(name='C5H10OOH2-1',
        atoms='H:11 C:5 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [ 5.18861756E-01,  7.44934918E-02, -5.64622871E-05,
                    2.26488673E-08, -3.77612026E-12, -8.22372586E+03,
                    3.02858118E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.17441074E+01,  2.33958311E-02, -8.13002591E-06,
                    1.27566943E-09, -7.45598238E-14, -1.55260698E+04,
                    -8.33092449E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THRM')

species(name='C5H10OOH2-3',
        atoms='H:11 C:5 O:2',
        thermo=(NASA([300.00, 1388.00],
                    [ 1.27092209E+00,  6.82958985E-02, -4.63421243E-05,
                    1.61403755E-08, -2.30884220E-12, -9.58339948E+03,
                    2.79668173E+01]),
                NASA([1388.00, 5000.00],
                    [ 2.10899451E+01,  2.37168023E-02, -8.19006128E-06,
                    1.27986065E-09, -7.45974063E-14, -1.66702092E+04,
                    -7.91383677E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THRM')

species(name='C5H10OOH2-4',
        atoms='H:11 C:5 O:2',
        thermo=(NASA([300.00, 1388.00],
                    [ 1.27092209E+00,  6.82958985E-02, -4.63421243E-05,
                    1.61403755E-08, -2.30884220E-12, -9.58339948E+03,
                    2.79668173E+01]),
                NASA([1388.00, 5000.00],
                    [ 2.10899451E+01,  2.37168023E-02, -8.19006128E-06,
                    1.27986065E-09, -7.45974063E-14, -1.66702092E+04,
                    -7.91383677E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THRM')

species(name='C5H10OOH2-5',
        atoms='H:11 C:5 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [ 5.18861756E-01,  7.44934918E-02, -5.64622871E-05,
                    2.26488673E-08, -3.77612026E-12, -8.22372586E+03,
                    3.02858118E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.17441074E+01,  2.33958311E-02, -8.13002591E-06,
                    1.27566943E-09, -7.45598238E-14, -1.55260698E+04,
                    -8.33092449E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THRM')

species(name='C5H10OOH3-2',
        atoms='H:11 C:5 O:2',
        thermo=(NASA([300.00, 1388.00],
                    [ 1.27092209E+00,  6.82958985E-02, -4.63421243E-05,
                    1.61403755E-08, -2.30884220E-12, -9.58339948E+03,
                    2.79668173E+01]),
                NASA([1388.00, 5000.00],
                    [ 2.10899451E+01,  2.37168023E-02, -8.19006128E-06,
                    1.27986065E-09, -7.45974063E-14, -1.66702092E+04,
                    -7.91383677E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THRM')

species(name='C5H10OOH3-1',
        atoms='H:11 C:5 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [ 5.18861756E-01,  7.44934918E-02, -5.64622871E-05,
                    2.26488673E-08, -3.77612026E-12, -8.22372586E+03,
                    3.02858118E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.17441074E+01,  2.33958311E-02, -8.13002591E-06,
                    1.27566943E-09, -7.45598238E-14, -1.55260698E+04,
                    -8.33092449E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THRM')

species(name='C5H10O1-2',
        atoms='H:10 C:5 O:1',
        thermo=(NASA([300.00, 1382.00],
                    [-2.50915891E+00,  7.02820905E-02, -5.23078154E-05,
                    2.05330738E-08, -3.37999472E-12, -1.79263879E+04,
                    3.83402738E+01]),
                NASA([1382.00, 5000.00],
                    [ 1.83450823E+01,  2.15894287E-02, -7.59914917E-06,
                    1.20265688E-09, -7.07149348E-14, -2.52918816E+04,
                    -7.38748127E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='1/22/95THERM')

species(name='C5H10O1-3',
        atoms='H:10 C:5 O:1',
        thermo=(NASA([300.00, 1370.00],
                    [-2.70067147E+00,  6.51234466E-02, -3.97863209E-05,
                    1.12983558E-08, -1.18141056E-12, -1.81611984E+04,
                    3.87589334E+01]),
                NASA([1370.00, 5000.00],
                    [ 1.85051134E+01,  2.17909650E-02, -7.74803799E-06,
                    1.23453810E-09, -7.29327086E-14, -2.62151881E+04,
                    -7.74524221E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='1/22/95THERM')

species(name='C5H10O1-4',
        atoms='H:10 C:5 O:1',
        thermo=(NASA([300.00, 1371.00],
                    [-3.94205066E+00,  6.68440102E-02, -4.09940878E-05,
                    1.17684771E-08, -1.26029870E-12, -2.80943137E+04,
                    4.58617150E+01]),
                NASA([1371.00, 5000.00],
                    [ 1.77074517E+01,  2.24725523E-02, -7.98404192E-06,
                    1.27146816E-09, -7.50866945E-14, -3.63066445E+04,
                    -7.27392560E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='1/22/95THERM')

species(name='C5H10O1-5',
        atoms='H:10 C:5 O:1',
        thermo=(NASA([300.00, 1457.00],
                    [-3.02904503E+00,  6.01320166E-02, -2.12072490E-05,
                    -4.91313274E-09,  3.07577547E-12, -2.91028862E+04,
                    3.64698785E+01]),
                NASA([1457.00, 5000.00],
                    [ 2.13684109E+01,  1.97321411E-02, -7.12842160E-06,
                    1.14789888E-09, -6.83162787E-14, -3.89834904E+04,
                    -1.00032176E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='1/22/95THERM')

species(name='C5H10O2-3',
        atoms='H:10 C:5 O:1',
        thermo=(NASA([300.00, 1385.00],
                    [-3.03891874E+00,  7.46406721E-02, -5.91275018E-05,
                    2.46307903E-08, -4.24312686E-12, -2.01088998E+04,
                    4.00106761E+01]),
                NASA([1385.00, 5000.00],
                    [ 1.89379846E+01,  2.10434849E-02, -7.40107577E-06,
                    1.17069402E-09, -6.88106356E-14, -2.76228814E+04,
                    -7.74017673E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='1/22/95THERM')

species(name='C5H10O2-4',
        atoms='H:10 C:5 O:1',
        thermo=(NASA([300.00, 1376.00],
                    [-3.50702064E+00,  7.13707278E-02, -4.99483546E-05,
                    1.76515898E-08, -2.56653255E-12, -2.03198734E+04,
                    4.22612571E+01]),
                NASA([1376.00, 5000.00],
                    [ 1.91186278E+01,  2.12169126E-02, -7.53764833E-06,
                    1.20035780E-09, -7.08866585E-14, -2.85216189E+04,
                    -8.03408390E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='1/22/95THERM')

species(name='C5H10OOH1-2O2',
        atoms='H:11 C:5 O:4',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.53072834E+00,  8.12167401E-02, -6.33054798E-05,
                    2.63900700E-08, -4.58739612E-12, -2.67337637E+04,
                    2.39105049E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.55478814E+01,  2.50385111E-02, -8.74243938E-06,
                    1.37611416E-09, -8.06073182E-14, -3.46345151E+04,
                    -9.91022020E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0TM')

species(name='C5H10OOH1-3O2',
        atoms='H:11 C:5 O:4',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.53072834E+00,  8.12167401E-02, -6.33054798E-05,
                    2.63900700E-08, -4.58739612E-12, -2.67337637E+04,
                    2.39105049E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.55478814E+01,  2.50385111E-02, -8.74243938E-06,
                    1.37611416E-09, -8.06073182E-14, -3.46345151E+04,
                    -9.91022020E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0TM')

species(name='C5H10OOH1-4O2',
        atoms='H:11 C:5 O:4',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.53072834E+00,  8.12167401E-02, -6.33054798E-05,
                    2.63900700E-08, -4.58739612E-12, -2.67337637E+04,
                    2.39105049E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.55478814E+01,  2.50385111E-02, -8.74243938E-06,
                    1.37611416E-09, -8.06073182E-14, -3.46345151E+04,
                    -9.91022020E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0TM')

species(name='C5H10OOH1-5O2',
        atoms='H:11 C:5 O:4',
        thermo=(NASA([300.00, 1385.00],
                    [ 2.81997434E+00,  7.72690504E-02, -5.67282509E-05,
                    2.23236523E-08, -3.71501383E-12, -2.43447483E+04,
                    2.35135259E+01]),
                NASA([1385.00, 5000.00],
                    [ 2.48666576E+01,  2.56688838E-02, -8.97169324E-06,
                    1.41315614E-09, -8.28158711E-14, -3.21479218E+04,
                    -9.51211738E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0TM')

species(name='C5H10OOH2-1O2',
        atoms='H:11 C:5 O:4',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.53072834E+00,  8.12167401E-02, -6.33054798E-05,
                    2.63900700E-08, -4.58739612E-12, -2.67337637E+04,
                    2.39105049E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.55478814E+01,  2.50385111E-02, -8.74243938E-06,
                    1.37611416E-09, -8.06073182E-14, -3.46345151E+04,
                    -9.91022020E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0TM')

species(name='C5H10OOH2-3O2',
        atoms='H:11 C:5 O:4',
        thermo=(NASA([300.00, 1389.00],
                    [ 2.08935173E+00,  8.57156954E-02, -7.04986902E-05,
                    3.07497216E-08, -5.51132533E-12, -2.90970365E+04,
                    2.50297599E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.62850175E+01,  2.43678422E-02, -8.50101950E-06,
                    1.33736519E-09, -7.83075048E-14, -3.71537284E+04,
                    -1.03435286E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0TM')

species(name='C5H10OOH2-4O2',
        atoms='H:11 C:5 O:4',
        thermo=(NASA([300.00, 1389.00],
                    [ 2.08935173E+00,  8.57156954E-02, -7.04986902E-05,
                    3.07497216E-08, -5.51132533E-12, -2.90970365E+04,
                    2.50297599E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.62850175E+01,  2.43678422E-02, -8.50101950E-06,
                    1.33736519E-09, -7.83075048E-14, -3.71537284E+04,
                    -1.03435286E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0TM')

species(name='C5H10OOH2-5O2',
        atoms='H:11 C:5 O:4',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.53072834E+00,  8.12167401E-02, -6.33054798E-05,
                    2.63900700E-08, -4.58739612E-12, -2.67337637E+04,
                    2.39105049E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.55478814E+01,  2.50385111E-02, -8.74243938E-06,
                    1.37611416E-09, -8.06073182E-14, -3.46345151E+04,
                    -9.91022020E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0TM')

species(name='C5H10OOH3-1O2',
        atoms='H:11 C:5 O:4',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.53072834E+00,  8.12167401E-02, -6.33054798E-05,
                    2.63900700E-08, -4.58739612E-12, -2.67337637E+04,
                    2.39105049E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.55478814E+01,  2.50385111E-02, -8.74243938E-06,
                    1.37611416E-09, -8.06073182E-14, -3.46345151E+04,
                    -9.91022020E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0TM')

species(name='C5H10OOH3-2O2',
        atoms='H:11 C:5 O:4',
        thermo=(NASA([300.00, 1389.00],
                    [ 2.08935173E+00,  8.57156954E-02, -7.04986902E-05,
                    3.07497216E-08, -5.51132533E-12, -2.90970365E+04,
                    2.50297599E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.62850175E+01,  2.43678422E-02, -8.50101950E-06,
                    1.33736519E-09, -7.83075048E-14, -3.71537284E+04,
                    -1.03435286E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0TM')

species(name='NC5KET12',
        atoms='H:10 C:5 O:3',
        thermo=(NASA([300.00, 1389.00],
                    [ 5.64845658E-01,  7.74889690E-02, -6.00818427E-05,
                    2.42432817E-08, -4.02329880E-12, -4.21262314E+04,
                    3.12910774E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.34427049E+01,  2.21540692E-02, -7.74817880E-06,
                    1.22107095E-09, -7.15883644E-14, -4.99350593E+04,
                    -9.09970812E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.174,
                                well_depth=494.8,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC5KET13',
        atoms='H:10 C:5 O:3',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.53945153E+00,  7.08645319E-02, -5.32266350E-05,
                    2.12906293E-08, -3.56676536E-12, -4.29988725E+04,
                    2.14543054E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.27897549E+01,  2.26387719E-02, -7.89832354E-06,
                    1.24260309E-09, -7.27609759E-14, -5.00526193E+04,
                    -8.71676166E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.174,
                                well_depth=494.8,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC5KET14',
        atoms='H:10 C:5 O:3',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.53945153E+00,  7.08645319E-02, -5.32266350E-05,
                    2.12906293E-08, -3.56676536E-12, -4.29988725E+04,
                    2.14543054E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.27897549E+01,  2.26387719E-02, -7.89832354E-06,
                    1.24260309E-09, -7.27609759E-14, -5.00526193E+04,
                    -8.71676166E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.174,
                                well_depth=494.8,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC5KET15',
        atoms='H:10 C:5 O:3',
        thermo=(NASA([300.00, 1383.00],
                    [ 3.22565144E+00,  6.57620962E-02, -4.52752542E-05,
                    1.64810866E-08, -2.54467139E-12, -4.06876266E+04,
                    1.90851234E+01]),
                NASA([1383.00, 5000.00],
                    [ 2.21776239E+01,  2.31939864E-02, -8.09797899E-06,
                    1.27464047E-09, -7.46625107E-14, -4.75611599E+04,
                    -8.35091033E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.174,
                                well_depth=494.8,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC5KET21',
        atoms='H:10 C:5 O:3',
        thermo=(NASA([300.00, 1379.00],
                    [ 4.56687437E+00,  6.01978739E-02, -3.82935508E-05,
                    1.29220662E-08, -1.89791227E-12, -4.47703435E+04,
                    1.29580830E+01]),
                NASA([1379.00, 5000.00],
                    [ 2.17534680E+01,  2.36666119E-02, -8.28610759E-06,
                    1.30663620E-09, -7.66324076E-14, -5.12815467E+04,
                    -8.09340957E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.174,
                                well_depth=494.8,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC5KET23',
        atoms='H:10 C:5 O:3',
        thermo=(NASA([300.00, 1387.00],
                    [ 1.09304151E-01,  7.75997760E-02, -5.94124810E-05,
                    2.36744270E-08, -3.89094338E-12, -4.57495597E+04,
                    3.32799795E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.31409808E+01,  2.24801534E-02, -7.87662916E-06,
                    1.24282516E-09, -7.29252729E-14, -5.36784029E+04,
                    -9.00538236E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.174,
                                well_depth=494.8,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC5KET24',
        atoms='H:10 C:5 O:3',
        thermo=(NASA([300.00, 1385.00],
                    [ 2.19333641E+00,  7.04050260E-02, -5.15856312E-05,
                    2.00741570E-08, -3.28568948E-12, -4.66368524E+04,
                    2.29520202E+01]),
                NASA([1385.00, 5000.00],
                    [ 2.25126387E+01,  2.29495597E-02, -8.02286486E-06,
                    1.26389975E-09, -7.40777420E-14, -5.38122195E+04,
                    -8.63788737E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.174,
                                well_depth=494.8,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC5KET25',
        atoms='H:10 C:5 O:3',
        thermo=(NASA([300.00, 1382.00],
                    [ 2.76270425E+00,  6.58724877E-02, -4.45144578E-05,
                    1.58213340E-08, -2.38800145E-12, -4.43093682E+04,
                    2.11180678E+01]),
                NASA([1382.00, 5000.00],
                    [ 2.19099644E+01,  2.35018200E-02, -8.22261497E-06,
                    1.29606701E-09, -7.59914647E-14, -5.13242006E+04,
                    -8.27708903E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.174,
                                well_depth=494.8,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC5KET31',
        atoms='H:10 C:5 O:3',
        thermo=(NASA([300.00, 1382.00],
                    [ 3.14481499E+00,  6.64310592E-02, -4.75232545E-05,
                    1.85573234E-08, -3.11995924E-12, -4.46228757E+04,
                    1.89760116E+01]),
                NASA([1382.00, 5000.00],
                    [ 2.17723599E+01,  2.36116706E-02, -8.25813864E-06,
                    1.30129886E-09, -7.62813000E-14, -5.13543364E+04,
                    -8.16385880E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.174,
                                well_depth=494.8,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC5KET32',
        atoms='H:10 C:5 O:3',
        thermo=(NASA([300.00, 1387.00],
                    [ 7.24497110E-01,  7.71622935E-02, -6.10369902E-05,
                    2.55909432E-08, -4.44696298E-12, -4.60989830E+04,
                    3.00522534E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.29536744E+01,  2.26192649E-02, -7.91922638E-06,
                    1.24883440E-09, -7.32467701E-14, -5.36792159E+04,
                    -8.86117968E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.174,
                                well_depth=494.8,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C5H10OH-1',
        atoms='H:11 C:5 O:1',
        thermo=(NASA([300.00, 1371.00],
                    [ 9.51873649E-01,  5.58286144E-02, -2.87750953E-05,
                    5.54684921E-09,  1.10578801E-14, -1.48168651E+04,
                    2.93877119E+01]),
                NASA([1371.00, 5000.00],
                    [ 1.75642279E+01,  2.33577469E-02, -7.81503517E-06,
                    1.19674654E-09, -6.88114207E-14, -2.11605652E+04,
                    -6.19821034E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='1/14/99THERM')

species(name='C5H10OH-2',
        atoms='H:11 C:5 O:1',
        thermo=(NASA([300.00, 1390.00],
                    [ 4.68530877E-01,  6.05595171E-02, -3.75900684E-05,
                    1.16488887E-08, -1.41506375E-12, -1.70158316E+04,
                    3.00652481E+01]),
                NASA([1390.00, 5000.00],
                    [ 1.72742775E+01,  2.39409708E-02, -8.07388576E-06,
                    1.24188426E-09, -7.15926603E-14, -2.31054055E+04,
                    -6.11106423E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='1/14/99THERM')

species(name='O2C5H10OH-1',
        atoms='H:11 C:5 O:3',
        thermo=(NASA([300.00, 1395.00],
                    [ 2.70461132E+00,  6.83289096E-02, -4.77921008E-05,
                    1.81238720E-08, -2.90871344E-12, -3.44822614E+04,
                    2.22673007E+01]),
                NASA([1395.00, 5000.00],
                    [ 2.06059936E+01,  2.61750874E-02, -8.92270421E-06,
                    1.38184638E-09, -8.00235870E-14, -4.07496333E+04,
                    -7.38943495E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653),
        note='1/14/99THRM')

species(name='O2C5H10OH-2',
        atoms='H:11 C:5 O:3',
        thermo=(NASA([300.00, 1398.00],
                    [ 1.89616952E+00,  7.37918569E-02, -5.55488679E-05,
                    2.24979959E-08, -3.79045894E-12, -3.66230283E+04,
                    2.45460172E+01]),
                NASA([1398.00, 5000.00],
                    [ 2.15259930E+01,  2.54039694E-02, -8.65952886E-06,
                    1.34110440E-09, -7.76663740E-14, -4.32473477E+04,
                    -8.00796638E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653),
        note='1/14/99THRM')

species(name='NC6H14',
        atoms='H:14 C:6',
        thermo=(NASA([300.00, 1390.00],
                    [-9.69606184E-01,  7.29085608E-02, -4.38853919E-05,
                    1.32312807E-08, -1.58437423E-12, -2.27803862E+04,
                    3.23069798E+01]),
                NASA([1390.00, 5000.00],
                    [ 1.89634117E+01,  3.04480204E-02, -1.03794829E-05,
                    1.60775457E-09, -9.31269728E-14, -3.01628739E+04,
                    -7.62839355E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.946,
                                well_depth=427.4,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C6H13-1',
        atoms='H:13 C:6',
        thermo=(NASA([300.00, 1390.00],
                    [-2.04871465E-01,  6.83801272E-02, -4.14447912E-05,
                    1.26155802E-08, -1.53120058E-12,  1.83280393E+03,
                    3.16075093E+01]),
                NASA([1390.00, 5000.00],
                    [ 1.85385470E+01,  2.83107962E-02, -9.65307246E-06,
                    1.49547585E-09, -8.66336064E-14, -5.09299041E+03,
                    -7.04490943E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.349,
                                well_depth=489.224),
        note='7/19/0THERM')

species(name='C6H13-2',
        atoms='H:13 C:6',
        thermo=(NASA([300.00, 1380.00],
                    [ 2.29560149E-01,  6.33327323E-02, -3.24135431E-05,
                    6.46387687E-09, -9.61420427E-14,  5.25639156E+02,
                    3.08006138E+01]),
                NASA([1380.00, 5000.00],
                    [ 1.83687363E+01,  2.80268110E-02, -9.47032396E-06,
                    1.45888527E-09, -8.42002461E-14, -6.46093974E+03,
                    -6.90934018E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.349,
                                well_depth=489.224),
        note='7/19/0THERM')

species(name='C6H13-3',
        atoms='H:13 C:6',
        thermo=(NASA([300.00, 1380.00],
                    [ 2.29560149E-01,  6.33327323E-02, -3.24135431E-05,
                    6.46387687E-09, -9.61420427E-14,  5.25639156E+02,
                    3.08006138E+01]),
                NASA([1380.00, 5000.00],
                    [ 1.83687363E+01,  2.80268110E-02, -9.47032396E-06,
                    1.45888527E-09, -8.42002461E-14, -6.46093974E+03,
                    -6.90934018E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.349,
                                well_depth=489.224),
        note='7/19/0THERM')

species(name='C6H12-1',
        atoms='H:12 C:6',
        thermo=(NASA([300.00, 1392.00],
                    [-1.35275205E+00,  6.98655426E-02, -4.59408022E-05,
                    1.56967343E-08, -2.21296175E-12, -7.34368617E+03,
                    3.53120691E+01]),
                NASA([1392.00, 5000.00],
                    [ 1.78337529E+01,  2.67377658E-02, -9.10036773E-06,
                    1.40819768E-09, -8.15124244E-14, -1.42062860E+04,
                    -6.83818851E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.328,
                                well_depth=485.857),
        note='5/14/3THERM')

species(name='C6H12-2',
        atoms='H:12 C:6',
        thermo=(NASA([300.00, 1390.00],
                    [-8.37810520E-01,  6.64334712E-02, -4.08801772E-05,
                    1.27558015E-08, -1.60665165E-12, -8.86347329E+03,
                    3.27785200E+01]),
                NASA([1390.00, 5000.00],
                    [ 1.73616148E+01,  2.71634902E-02, -9.25249439E-06,
                    1.43247725E-09, -8.29472521E-14, -1.55509547E+04,
                    -6.61857343E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.328,
                                well_depth=485.857),
        note='5/14/3THERM')

species(name='C6H12-3',
        atoms='H:12 C:6',
        thermo=(NASA([300.00, 1392.00],
                    [-1.74471739E+00,  7.02087492E-02, -4.60836866E-05,
                    1.58225609E-08, -2.26113570E-12, -8.63510107E+03,
                    3.65273503E+01]),
                NASA([1392.00, 5000.00],
                    [ 1.74392115E+01,  2.72012909E-02, -9.28739405E-06,
                    1.44011755E-09, -8.34782517E-14, -1.55311520E+04,
                    -6.72316177E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.328,
                                well_depth=485.857),
        note='5/14/3THERM')

species(name='C6H111-3',
        atoms='H:11 C:6',
        thermo=(NASA([300.00, 1392.00],
                    [-1.68400759E+00,  6.82608096E-02, -4.55829418E-05,
                    1.56920477E-08, -2.21118204E-12,  9.68283708E+03,
                    3.57445494E+01]),
                NASA([1392.00, 5000.00],
                    [ 1.74385677E+01,  2.50390443E-02, -8.53480314E-06,
                    1.32209951E-09, -7.65892920E-14,  2.88667463E+03,
                    -6.74868036E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.307,
                                well_depth=482.473),
        note='5/14/3THERM')

species(name='C6H111-4',
        atoms='H:11 C:6',
        thermo=(NASA([300.00, 1381.00],
                    [-6.66672195E-02,  5.98990444E-02, -3.38938138E-05,
                    8.58922359E-09, -6.54329636E-13,  1.59493460E+04,
                    3.27147756E+01]),
                NASA([1381.00, 5000.00],
                    [ 1.72621286E+01,  2.42963321E-02, -8.18563605E-06,
                    1.25871870E-09, -7.25641092E-14,  9.47965134E+03,
                    -6.20229885E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.307,
                                well_depth=482.473),
        note='5/14/3THERM')

species(name='C6H111-5',
        atoms='H:11 C:6',
        thermo=(NASA([300.00, 1381.00],
                    [-6.66672195E-02,  5.98990444E-02, -3.38938138E-05,
                    8.58922359E-09, -6.54329636E-13,  1.59493460E+04,
                    3.27147756E+01]),
                NASA([1381.00, 5000.00],
                    [ 1.72621286E+01,  2.42963321E-02, -8.18563605E-06,
                    1.25871870E-09, -7.25641092E-14,  9.47965134E+03,
                    -6.20229885E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.307,
                                well_depth=482.473),
        note='5/14/3THERM')

species(name='C6H111-6',
        atoms='H:11 C:6',
        thermo=(NASA([300.00, 1392.00],
                    [-6.17484828E-01,  6.54674092E-02, -4.36872486E-05,
                    1.51936842E-08, -2.18414423E-12,  1.72739396E+04,
                    3.40544957E+01]),
                NASA([1392.00, 5000.00],
                    [ 1.74129892E+01,  2.45984185E-02, -8.37352752E-06,
                    1.29588349E-09, -7.50182758E-14,  1.08613574E+04,
                    -6.32669814E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.307,
                                well_depth=482.473),
        note='5/14/3THERM')

species(name='C6H112-4',
        atoms='H:11 C:6',
        thermo=(NASA([300.00, 1389.00],
                    [-1.16992307E+00,  6.48326433E-02, -4.05262795E-05,
                    1.27524726E-08, -1.60498435E-12,  8.16316429E+03,
                    3.32148832E+01]),
                NASA([1389.00, 5000.00],
                    [ 1.69675593E+01,  2.54635235E-02, -8.68644122E-06,
                    1.34629700E-09, -7.80191162E-14,  1.54157079E+03,
                    -6.52971106E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.307,
                                well_depth=482.473),
        note='5/14/3THERM')

species(name='C6H112-5',
        atoms='H:11 C:6',
        thermo=(NASA([300.00, 1377.00],
                    [ 4.63606027E-01,  5.63992000E-02, -2.87342339E-05,
                    5.59014813E-09, -3.57957530E-14,  1.44272349E+04,
                    3.01101704E+01]),
                NASA([1377.00, 5000.00],
                    [ 1.67818090E+01,  2.47356833E-02, -8.34368101E-06,
                    1.28402955E-09, -7.40627547E-14,  8.13809667E+03,
                    -5.97810800E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.307,
                                well_depth=482.473),
        note='5/14/3THERM')

species(name='C6H112-6',
        atoms='H:11 C:6',
        thermo=(NASA([300.00, 1390.00],
                    [-1.06006806E-01,  6.20508067E-02, -3.86460745E-05,
                    1.22632255E-08, -1.57992990E-12,  1.57546498E+04,
                    3.15418771E+01]),
                NASA([1390.00, 5000.00],
                    [ 1.69431052E+01,  2.50224953E-02, -8.52515322E-06,
                    1.32009254E-09, -7.64493285E-14,  9.51557475E+03,
                    -6.10792812E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.307,
                                well_depth=482.473),
        note='5/14/3THERM')

species(name='C6H113-1',
        atoms='H:11 C:6',
        thermo=(NASA([300.00, 1392.00],
                    [-9.77623834E-01,  6.56664597E-02, -4.36212067E-05,
                    1.51941737E-08, -2.20544154E-12,  1.59778322E+04,
                    3.58177060E+01]),
                NASA([1392.00, 5000.00],
                    [ 1.70133131E+01,  2.50666461E-02, -8.56223877E-06,
                    1.32807194E-09, -7.69999372E-14,  9.53859681E+03,
                    -6.13920940E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.307,
                                well_depth=482.473),
        note='5/14/3THERM')

species(name='C6H11O1-3',
        atoms='H:11 C:6 O:1',
        thermo=(NASA([300.00, 1393.00],
                    [-2.54423388E+00,  8.15467714E-02, -6.04734167E-05,
                    2.23330400E-08, -3.25460865E-12, -9.39930947E+02,
                    4.32736817E+01]),
                NASA([1393.00, 5000.00],
                    [ 2.22920437E+01,  2.24441825E-02, -7.53070994E-06,
                    1.15575450E-09, -6.65675174E-14, -9.30566725E+03,
                    -8.94778368E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.674,
                                well_depth=541.5,
                                dipole=1.8,
                                rot_relax=1.0),
        note='3/28/97THERM')

species(name='C6H11O2-4',
        atoms='H:11 C:6 O:1',
        thermo=(NASA([300.00, 1388.00],
                    [-1.97396384E+00,  7.78641252E-02, -5.51856115E-05,
                    1.93459318E-08, -2.65553891E-12, -2.46698943E+03,
                    4.04899267E+01]),
                NASA([1388.00, 5000.00],
                    [ 2.17536453E+01,  2.29477789E-02, -7.71427788E-06,
                    1.18540388E-09, -6.83332601E-14, -1.06296518E+04,
                    -8.69112396E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.674,
                                well_depth=541.5,
                                dipole=1.8,
                                rot_relax=1.0),
        note='3/28/97THERM')

species(name='C6H12O1-2',
        atoms='H:12 C:6 O:1',
        thermo=(NASA([300.00, 1390.00],
                    [-2.60027118E+00,  8.25140327E-02, -6.19868951E-05,
                    2.52986585E-08, -4.35196287E-12, -2.08474661E+04,
                    4.03432783E+01]),
                NASA([1390.00, 5000.00],
                    [ 1.98237675E+01,  2.83076305E-02, -9.79700725E-06,
                    1.53290291E-09, -8.94136512E-14, -2.86156250E+04,
                    -7.97147353E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.643,
                                well_depth=537.467),
        note='1/22/95THERM')

species(name='C6H12O1-3',
        atoms='H:12 C:6 O:1',
        thermo=(NASA([300.00, 1381.00],
                    [-3.67890943E+00,  8.18734230E-02, -5.65067852E-05,
                    2.03434434E-08, -3.05338137E-12, -2.09644147E+04,
                    4.54357891E+01]),
                NASA([1381.00, 5000.00],
                    [ 2.00943213E+01,  2.80281041E-02, -9.61073019E-06,
                    1.49563556E-09, -8.69482515E-14, -2.94623491E+04,
                    -8.29672080E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.643,
                                well_depth=537.467),
        note='1/22/95THERM')

species(name='C6H12O1-4',
        atoms='H:12 C:6 O:1',
        thermo=(NASA([300.00, 1385.00],
                    [-4.98759747E+00,  8.36690743E-02, -5.83931131E-05,
                    2.19383204E-08, -3.51302429E-12, -3.08768470E+04,
                    5.22311756E+01]),
                NASA([1385.00, 5000.00],
                    [ 1.81937451E+01,  3.05132082E-02, -1.05538621E-05,
                    1.65081301E-09, -9.62756740E-14, -3.91861539E+04,
                    -7.28892157E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.643,
                                well_depth=537.467),
        note='1/22/95THERM')

species(name='C6H12O1-5',
        atoms='H:12 C:6 O:1',
        thermo=(NASA([300.00, 1369.00],
                    [-4.89881418E+00,  8.17946607E-02, -4.67358620E-05,
                    1.15229724E-08, -8.38133459E-13, -3.39937629E+04,
                    4.60903286E+01]),
                NASA([1369.00, 5000.00],
                    [ 2.18696492E+01,  2.89644808E-02, -1.02843212E-05,
                    1.63714319E-09, -9.66552991E-14, -4.43407005E+04,
                    -1.01256960E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.643,
                                well_depth=537.467),
        note='1/22/95THERM')

species(name='C6H12O2-3',
        atoms='H:12 C:6 O:1',
        thermo=(NASA([300.00, 1392.00],
                    [-3.24328996E+00,  8.73751851E-02, -6.95553645E-05,
                    2.98574354E-08, -5.31644810E-12, -2.30127954E+04,
                    4.25385508E+01]),
                NASA([1392.00, 5000.00],
                    [ 2.04400807E+01,  2.77275705E-02, -9.58422558E-06,
                    1.49835476E-09, -8.73476387E-14, -3.09571014E+04,
                    -8.33771795E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.643,
                                well_depth=537.467),
        note='1/22/95THERM')

species(name='C6H12O2-4',
        atoms='H:12 C:6 O:1',
        thermo=(NASA([300.00, 1388.00],
                    [-3.75894138E+00,  8.46044396E-02, -6.18101303E-05,
                    2.42424607E-08, -4.01441141E-12, -2.32216305E+04,
                    4.49636341E+01]),
                NASA([1388.00, 5000.00],
                    [ 1.99501923E+01,  2.88253331E-02, -1.00217930E-05,
                    1.57295064E-09, -9.19508743E-14, -3.15674310E+04,
                    -8.24846912E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.643,
                                well_depth=537.467),
        note='1/22/95THERM')

species(name='C6H12O2-5',
        atoms='H:12 C:6 O:1',
        thermo=(NASA([300.00, 1388.00],
                    [-5.21438722E+00,  8.67847640E-02, -6.36083019E-05,
                    2.52625812E-08, -4.25063214E-12, -3.31071531E+04,
                    5.24812152E+01]),
                NASA([1388.00, 5000.00],
                    [ 1.86455632E+01,  3.02705842E-02, -1.05000692E-05,
                    1.64542415E-09, -9.60802316E-14, -4.14905135E+04,
                    -7.56801716E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.643,
                                well_depth=537.467),
        note='1/22/95THERM')

species(name='C6H12O3-4',
        atoms='H:12 C:6 O:1',
        thermo=(NASA([300.00, 1392.00],
                    [-3.24328996E+00,  8.73751851E-02, -6.95553645E-05,
                    2.98574354E-08, -5.31644810E-12, -2.30127954E+04,
                    4.25385508E+01]),
                NASA([1392.00, 5000.00],
                    [ 2.04400807E+01,  2.77275705E-02, -9.58422558E-06,
                    1.49835476E-09, -8.73476387E-14, -3.09571014E+04,
                    -8.33771795E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.643,
                                well_depth=537.467),
        note='1/22/95THERM')

species(name='C6H13O2H-1',
        atoms='H:14 C:6 O:2',
        thermo=(NASA([300.00, 1388.00],
                    [ 4.28827381E-01,  8.52153162E-02, -5.81549717E-05,
                    2.08798889E-08, -3.15227267E-12, -3.34464197E+04,
                    3.10485287E+01]),
                NASA([1388.00, 5000.00],
                    [ 2.45609177E+01,  3.07471465E-02, -1.06611615E-05,
                    1.67031752E-09, -9.75224221E-14, -4.21305786E+04,
                    -9.94217981E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585),
        note='7/19/0THERM')

species(name='C6H13O2H-2',
        atoms='H:14 C:6 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [-4.38853727E-01,  9.10055843E-02, -6.66797289E-05,
                    2.58472566E-08, -4.18105592E-12, -3.57293720E+04,
                    3.35762294E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.53703550E+01,  3.00455578E-02, -1.04162589E-05,
                    1.63180782E-09, -9.52693153E-14, -4.47221575E+04,
                    -1.04960674E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585),
        note='7/19/0THERM')

species(name='C6H13O2H-3',
        atoms='H:14 C:6 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [-4.38853727E-01,  9.10055843E-02, -6.66797289E-05,
                    2.58472566E-08, -4.18105592E-12, -3.57293720E+04,
                    3.35762294E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.53703550E+01,  3.00455578E-02, -1.04162589E-05,
                    1.63180782E-09, -9.52693153E-14, -4.47221575E+04,
                    -1.04960674E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585),
        note='7/19/0THERM')

species(name='C6H13O2-1',
        atoms='H:13 C:6 O:2',
        thermo=(NASA([300.00, 1387.00],
                    [ 1.24283154E+00,  7.71349324E-02, -5.11549390E-05,
                    1.83242032E-08, -2.83058645E-12, -1.65776531E+04,
                    2.79518801E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.21224820E+01,  3.05609710E-02, -1.05650154E-05,
                    1.65180006E-09, -9.62963346E-14, -2.42219722E+04,
                    -8.52560175E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653),
        note='7/19/0THERM')

species(name='C6H13O2-2',
        atoms='H:13 C:6 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [ 5.25316289E-01,  8.23287654E-02, -5.89792005E-05,
                    2.29248698E-08, -3.78641136E-12, -1.88843662E+04,
                    2.97737842E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.28578467E+01,  2.98910703E-02, -1.03239520E-05,
                    1.61313659E-09, -9.40033512E-14, -2.67682999E+04,
                    -9.03306141E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653),
        note='7/19/0THERM')

species(name='C6H13O2-3',
        atoms='H:13 C:6 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [ 5.25316289E-01,  8.23287654E-02, -5.89792005E-05,
                    2.29248698E-08, -3.78641136E-12, -1.88843662E+04,
                    2.97737842E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.28578467E+01,  2.98910703E-02, -1.03239520E-05,
                    1.61313659E-09, -9.40033512E-14, -2.67682999E+04,
                    -9.03306141E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653),
        note='7/19/0THERM')

species(name='C6H12OOH1-2',
        atoms='H:13 C:6 O:2',
        thermo=(NASA([300.00, 1383.00],
                    [ 1.46739689E+00,  7.63926681E-02, -4.77256697E-05,
                    1.46093613E-08, -1.73635234E-12, -1.01177260E+04,
                    2.95852272E+01]),
                NASA([1383.00, 5000.00],
                    [ 2.39324013E+01,  2.82540690E-02, -9.70102532E-06,
                    1.51047138E-09, -8.78256178E-14, -1.83659389E+04,
                    -9.26286863E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653),
        note='7/19/0THERM')

species(name='C6H12OOH1-3',
        atoms='H:13 C:6 O:2',
        thermo=(NASA([300.00, 1383.00],
                    [ 1.46739689E+00,  7.63926681E-02, -4.77256697E-05,
                    1.46093613E-08, -1.73635234E-12, -1.01177260E+04,
                    2.95852272E+01]),
                NASA([1383.00, 5000.00],
                    [ 2.39324013E+01,  2.82540690E-02, -9.70102532E-06,
                    1.51047138E-09, -8.78256178E-14, -1.83659389E+04,
                    -9.26286863E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653),
        note='7/19/0THERM')

species(name='C6H12OOH1-4',
        atoms='H:13 C:6 O:2',
        thermo=(NASA([300.00, 1383.00],
                    [ 1.46739689E+00,  7.63926681E-02, -4.77256697E-05,
                    1.46093613E-08, -1.73635234E-12, -1.01177260E+04,
                    2.95852272E+01]),
                NASA([1383.00, 5000.00],
                    [ 2.39324013E+01,  2.82540690E-02, -9.70102532E-06,
                    1.51047138E-09, -8.78256178E-14, -1.83659389E+04,
                    -9.26286863E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653),
        note='7/19/0THERM')

species(name='C6H12OOH1-5',
        atoms='H:13 C:6 O:2',
        thermo=(NASA([300.00, 1383.00],
                    [ 1.46739689E+00,  7.63926681E-02, -4.77256697E-05,
                    1.46093613E-08, -1.73635234E-12, -1.01177260E+04,
                    2.95852272E+01]),
                NASA([1383.00, 5000.00],
                    [ 2.39324013E+01,  2.82540690E-02, -9.70102532E-06,
                    1.51047138E-09, -8.78256178E-14, -1.83659389E+04,
                    -9.26286863E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653),
        note='7/19/0THERM')

species(name='C6H12OOH2-1',
        atoms='H:13 C:6 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [ 3.27957994E-01,  8.64935702E-02, -6.43067781E-05,
                    2.52816700E-08, -4.13905950E-12, -1.11170270E+04,
                    3.21731341E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.49425195E+01,  2.78897045E-02, -9.67889017E-06,
                    1.51735809E-09, -8.86312636E-14, -1.96455693E+04,
                    -9.97839086E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653),
        note='7/19/0THERM')

species(name='C6H12OOH2-3',
        atoms='H:13 C:6 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 1.01849347E+00,  8.05626618E-02, -5.46523533E-05,
                    1.90948643E-08, -2.74735588E-12, -1.24666956E+04,
                    3.01431792E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.42386096E+01,  2.82611912E-02, -9.75785456E-06,
                    1.52463492E-09, -8.88530032E-14, -2.07681322E+04,
                    -9.53255329E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653),
        note='7/19/0THERM')

species(name='C6H12OOH2-4',
        atoms='H:13 C:6 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 1.01849347E+00,  8.05626618E-02, -5.46523533E-05,
                    1.90948643E-08, -2.74735588E-12, -1.24666956E+04,
                    3.01431792E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.42386096E+01,  2.82611912E-02, -9.75785456E-06,
                    1.52463492E-09, -8.88530032E-14, -2.07681322E+04,
                    -9.53255329E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653),
        note='7/19/0THERM')

species(name='C6H12OOH2-5',
        atoms='H:13 C:6 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 1.01849347E+00,  8.05626618E-02, -5.46523533E-05,
                    1.90948643E-08, -2.74735588E-12, -1.24666956E+04,
                    3.01431792E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.42386096E+01,  2.82611912E-02, -9.75785456E-06,
                    1.52463492E-09, -8.88530032E-14, -2.07681322E+04,
                    -9.53255329E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653),
        note='7/19/0THERM')

species(name='C6H12OOH2-6',
        atoms='H:13 C:6 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [ 3.27957994E-01,  8.64935702E-02, -6.43067781E-05,
                    2.52816700E-08, -4.13905950E-12, -1.11170270E+04,
                    3.21731341E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.49425195E+01,  2.78897045E-02, -9.67889017E-06,
                    1.51735809E-09, -8.86312636E-14, -1.96455693E+04,
                    -9.97839086E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653),
        note='7/19/0THERM')

species(name='C6H12OOH3-1',
        atoms='H:13 C:6 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [ 3.27957994E-01,  8.64935702E-02, -6.43067781E-05,
                    2.52816700E-08, -4.13905950E-12, -1.11170270E+04,
                    3.21731341E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.49425195E+01,  2.78897045E-02, -9.67889017E-06,
                    1.51735809E-09, -8.86312636E-14, -1.96455693E+04,
                    -9.97839086E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653),
        note='7/19/0THERM')

species(name='C6H12OOH3-2',
        atoms='H:13 C:6 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 1.01849347E+00,  8.05626618E-02, -5.46523533E-05,
                    1.90948643E-08, -2.74735588E-12, -1.24666956E+04,
                    3.01431792E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.42386096E+01,  2.82611912E-02, -9.75785456E-06,
                    1.52463492E-09, -8.88530032E-14, -2.07681322E+04,
                    -9.53255329E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653),
        note='7/19/0THERM')

species(name='C6H12OOH3-4',
        atoms='H:13 C:6 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 1.01849347E+00,  8.05626618E-02, -5.46523533E-05,
                    1.90948643E-08, -2.74735588E-12, -1.24666956E+04,
                    3.01431792E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.42386096E+01,  2.82611912E-02, -9.75785456E-06,
                    1.52463492E-09, -8.88530032E-14, -2.07681322E+04,
                    -9.53255329E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653),
        note='7/19/0THERM')

species(name='C6H12OOH3-5',
        atoms='H:13 C:6 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 1.01849347E+00,  8.05626618E-02, -5.46523533E-05,
                    1.90948643E-08, -2.74735588E-12, -1.24666956E+04,
                    3.01431792E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.42386096E+01,  2.82611912E-02, -9.75785456E-06,
                    1.52463492E-09, -8.88530032E-14, -2.07681322E+04,
                    -9.53255329E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653),
        note='7/19/0THERM')

species(name='C6H12OOH3-6',
        atoms='H:13 C:6 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [ 3.27957994E-01,  8.64935702E-02, -6.43067781E-05,
                    2.52816700E-08, -4.13905950E-12, -1.11170270E+04,
                    3.21731341E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.49425195E+01,  2.78897045E-02, -9.67889017E-06,
                    1.51735809E-09, -8.86312636E-14, -1.96455693E+04,
                    -9.97839086E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653),
        note='7/19/0THERM')

species(name='C6H12OOH1-2O2',
        atoms='H:13 C:6 O:4',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.19542973E+00,  9.38208455E-02, -7.19541247E-05,
                    2.94849979E-08, -5.04751943E-12, -2.96046234E+04,
                    2.71617731E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.87780675E+01,  2.95251779E-02, -1.02931692E-05,
                    1.61854521E-09, -9.47401775E-14, -3.87743825E+04,
                    -1.15090906E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149),
        note='7/19/0TRM')

species(name='C6H12OOH1-3O2',
        atoms='H:13 C:6 O:4',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.19542973E+00,  9.38208455E-02, -7.19541247E-05,
                    2.94849979E-08, -5.04751943E-12, -2.96046234E+04,
                    2.71617731E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.87780675E+01,  2.95251779E-02, -1.02931692E-05,
                    1.61854521E-09, -9.47401775E-14, -3.87743825E+04,
                    -1.15090906E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149),
        note='7/19/0TRM')

species(name='C6H12OOH1-4O2',
        atoms='H:13 C:6 O:4',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.19542973E+00,  9.38208455E-02, -7.19541247E-05,
                    2.94849979E-08, -5.04751943E-12, -2.96046234E+04,
                    2.71617731E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.87780675E+01,  2.95251779E-02, -1.02931692E-05,
                    1.61854521E-09, -9.47401775E-14, -3.87743825E+04,
                    -1.15090906E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149),
        note='7/19/0TRM')

species(name='C6H12OOH1-5O2',
        atoms='H:13 C:6 O:4',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.19542973E+00,  9.38208455E-02, -7.19541247E-05,
                    2.94849979E-08, -5.04751943E-12, -2.96046234E+04,
                    2.71617731E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.87780675E+01,  2.95251779E-02, -1.02931692E-05,
                    1.61854521E-09, -9.47401775E-14, -3.87743825E+04,
                    -1.15090906E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149),
        note='7/19/0TRM')

species(name='C6H12OOH2-1O2',
        atoms='H:13 C:6 O:4',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.19542973E+00,  9.38208455E-02, -7.19541247E-05,
                    2.94849979E-08, -5.04751943E-12, -2.96046234E+04,
                    2.71617731E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.87780675E+01,  2.95251779E-02, -1.02931692E-05,
                    1.61854521E-09, -9.47401775E-14, -3.87743825E+04,
                    -1.15090906E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149),
        note='7/19/0TRM')

species(name='C6H12OOH2-3O2',
        atoms='H:13 C:6 O:4',
        thermo=(NASA([300.00, 1389.00],
                    [ 1.71726092E+00,  9.83623459E-02, -7.90354964E-05,
                    3.37038035E-08, -5.93018486E-12, -3.19595468E+04,
                    2.84741624E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.95548919E+01,  2.88270019E-02, -1.00437732E-05,
                    1.57871998E-09, -9.23848804E-14, -4.13185786E+04,
                    -1.19673306E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149),
        note='7/19/0TRM')

species(name='C6H12OOH2-4O2',
        atoms='H:13 C:6 O:4',
        thermo=(NASA([300.00, 1389.00],
                    [ 1.71726092E+00,  9.83623459E-02, -7.90354964E-05,
                    3.37038035E-08, -5.93018486E-12, -3.19595468E+04,
                    2.84741624E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.95548919E+01,  2.88270019E-02, -1.00437732E-05,
                    1.57871998E-09, -9.23848804E-14, -4.13185786E+04,
                    -1.19673306E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149),
        note='7/19/0TRM')

species(name='C6H12OOH2-5O2',
        atoms='H:13 C:6 O:4',
        thermo=(NASA([300.00, 1389.00],
                    [ 1.71726092E+00,  9.83623459E-02, -7.90354964E-05,
                    3.37038035E-08, -5.93018486E-12, -3.19595468E+04,
                    2.84741624E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.95548919E+01,  2.88270019E-02, -1.00437732E-05,
                    1.57871998E-09, -9.23848804E-14, -4.13185786E+04,
                    -1.19673306E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149),
        note='7/19/0TRM')

species(name='C6H12OOH2-6O2',
        atoms='H:13 C:6 O:4',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.19542973E+00,  9.38208455E-02, -7.19541247E-05,
                    2.94849979E-08, -5.04751943E-12, -2.96046234E+04,
                    2.71617731E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.87780675E+01,  2.95251779E-02, -1.02931692E-05,
                    1.61854521E-09, -9.47401775E-14, -3.87743825E+04,
                    -1.15090906E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149),
        note='7/19/0TRM')

species(name='C6H12OOH3-1O2',
        atoms='H:13 C:6 O:4',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.19542973E+00,  9.38208455E-02, -7.19541247E-05,
                    2.94849979E-08, -5.04751943E-12, -2.96046234E+04,
                    2.71617731E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.87780675E+01,  2.95251779E-02, -1.02931692E-05,
                    1.61854521E-09, -9.47401775E-14, -3.87743825E+04,
                    -1.15090906E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149),
        note='7/19/0TRM')

species(name='C6H12OOH3-2O2',
        atoms='H:13 C:6 O:4',
        thermo=(NASA([300.00, 1389.00],
                    [ 1.71726092E+00,  9.83623459E-02, -7.90354964E-05,
                    3.37038035E-08, -5.93018486E-12, -3.19595468E+04,
                    2.84741624E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.95548919E+01,  2.88270019E-02, -1.00437732E-05,
                    1.57871998E-09, -9.23848804E-14, -4.13185786E+04,
                    -1.19673306E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149),
        note='7/19/0TRM')

species(name='C6H12OOH3-4O2',
        atoms='H:13 C:6 O:4',
        thermo=(NASA([300.00, 1389.00],
                    [ 1.71726092E+00,  9.83623459E-02, -7.90354964E-05,
                    3.37038035E-08, -5.93018486E-12, -3.19595468E+04,
                    2.84741624E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.95548919E+01,  2.88270019E-02, -1.00437732E-05,
                    1.57871998E-09, -9.23848804E-14, -4.13185786E+04,
                    -1.19673306E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149),
        note='7/19/0TRM')

species(name='C6H12OOH3-5O2',
        atoms='H:13 C:6 O:4',
        thermo=(NASA([300.00, 1389.00],
                    [ 1.71726092E+00,  9.83623459E-02, -7.90354964E-05,
                    3.37038035E-08, -5.93018486E-12, -3.19595468E+04,
                    2.84741624E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.95548919E+01,  2.88270019E-02, -1.00437732E-05,
                    1.57871998E-09, -9.23848804E-14, -4.13185786E+04,
                    -1.19673306E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149),
        note='7/19/0TRM')

species(name='C6H12OOH3-6O2',
        atoms='H:13 C:6 O:4',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.19542973E+00,  9.38208455E-02, -7.19541247E-05,
                    2.94849979E-08, -5.04751943E-12, -2.96046234E+04,
                    2.71617731E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.87780675E+01,  2.95251779E-02, -1.02931692E-05,
                    1.61854521E-09, -9.47401775E-14, -3.87743825E+04,
                    -1.15090906E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149),
        note='7/19/0TRM')

species(name='NC6KET12',
        atoms='H:12 C:6 O:3',
        thermo=(NASA([300.00, 1431.00],
                    [ 3.38273518E+00,  6.51971936E-02, -4.27848450E-05,
                    1.55394912E-08, -2.34685173E-12, -4.20179622E+04,
                    2.71536226E+01]),
                NASA([1431.00, 5000.00],
                    [ 1.76692622E+01,  2.98180137E-02, -9.43179071E-06,
                    1.38583919E-09, -7.72738616E-14, -4.67026645E+04,
                    -4.87302952E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC6KET13',
        atoms='H:12 C:6 O:3',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.36496601E+00,  8.28625471E-02, -6.12300153E-05,
                    2.40951600E-08, -3.97857713E-12, -4.58958895E+04,
                    2.39438215E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.59328855E+01,  2.71799865E-02, -9.46340472E-06,
                    1.48679538E-09, -8.69765690E-14, -5.41442590E+04,
                    -1.02621676E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC6KET14',
        atoms='H:12 C:6 O:3',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.36496601E+00,  8.28625471E-02, -6.12300153E-05,
                    2.40951600E-08, -3.97857713E-12, -4.58958895E+04,
                    2.39438215E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.59328855E+01,  2.71799865E-02, -9.46340472E-06,
                    1.48679538E-09, -8.69765690E-14, -5.41442590E+04,
                    -1.02621676E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC6KET15',
        atoms='H:12 C:6 O:3',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.36496601E+00,  8.28625471E-02, -6.12300153E-05,
                    2.40951600E-08, -3.97857713E-12, -4.58958895E+04,
                    2.39438215E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.59328855E+01,  2.71799865E-02, -9.46340472E-06,
                    1.48679538E-09, -8.69765690E-14, -5.41442590E+04,
                    -1.02621676E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC6KET21',
        atoms='H:12 C:6 O:3',
        thermo=(NASA([300.00, 1380.00],
                    [ 4.39221532E+00,  7.22001454E-02, -4.62914048E-05,
                    1.57162916E-08, -2.30615915E-12, -4.76675282E+04,
                    1.54471571E+01]),
                NASA([1380.00, 5000.00],
                    [ 2.48977627E+01,  2.82099062E-02, -9.85249224E-06,
                    1.55108568E-09, -9.08649219E-14, -5.53731276E+04,
                    -9.63947605E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC6KET23',
        atoms='H:12 C:6 O:3',
        thermo=(NASA([300.00, 1388.00],
                    [-1.35622858E-01,  8.98798610E-02, -6.76978860E-05,
                    2.65961001E-08, -4.32068174E-12, -4.86358432E+04,
                    3.60982792E+01]),
                NASA([1388.00, 5000.00],
                    [ 2.63430506E+01,  2.69861544E-02, -9.43297025E-06,
                    1.48601772E-09, -8.70977117E-14, -5.78013235E+04,
                    -1.05866313E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC6KET24',
        atoms='H:12 C:6 O:3',
        thermo=(NASA([300.00, 1386.00],
                    [ 1.73253220E+00,  8.37365688E-02, -6.15876454E-05,
                    2.41188610E-08, -3.97119823E-12, -4.94924385E+04,
                    2.67536906E+01]),
                NASA([1386.00, 5000.00],
                    [ 2.56847445E+01,  2.74782488E-02, -9.58627575E-06,
                    1.50810536E-09, -8.83051336E-14, -5.79176568E+04,
                    -1.02009176E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC6KET25',
        atoms='H:12 C:6 O:3',
        thermo=(NASA([300.00, 1386.00],
                    [ 1.73253220E+00,  8.37365688E-02, -6.15876454E-05,
                    2.41188610E-08, -3.97119823E-12, -4.94924385E+04,
                    2.67536906E+01]),
                NASA([1386.00, 5000.00],
                    [ 2.56847445E+01,  2.74782488E-02, -9.58627575E-06,
                    1.50810536E-09, -8.83051336E-14, -5.79176568E+04,
                    -1.02009176E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC6KET26',
        atoms='H:12 C:6 O:3',
        thermo=(NASA([300.00, 1383.00],
                    [ 2.58934810E+00,  7.78460026E-02, -5.24513152E-05,
                    1.85689713E-08, -2.78455147E-12, -4.72061155E+04,
                    2.36059697E+01]),
                NASA([1383.00, 5000.00],
                    [ 2.50574425E+01,  2.80369155E-02, -9.78501982E-06,
                    1.53978046E-09, -9.01766406E-14, -5.54174830E+04,
                    -9.82500096E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC6KET31',
        atoms='H:12 C:6 O:3',
        thermo=(NASA([300.00, 1383.00],
                    [ 2.96919433E+00,  7.84436519E-02, -5.55361775E-05,
                    2.13589050E-08, -3.52935948E-12, -4.75201133E+04,
                    2.14680944E+01]),
                NASA([1383.00, 5000.00],
                    [ 2.49194999E+01,  2.81507231E-02, -9.82266516E-06,
                    1.54541788E-09, -9.04929606E-14, -5.54463958E+04,
                    -9.71141228E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC6KET32',
        atoms='H:12 C:6 O:3',
        thermo=(NASA([300.00, 1387.00],
                    [ 5.64144594E-01,  8.90854292E-02, -6.89132035E-05,
                    2.83151935E-08, -4.84139667E-12, -4.89978587E+04,
                    3.24785934E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.60960142E+01,  2.71703309E-02, -9.48970254E-06,
                    1.49407399E-09, -8.75315577E-14, -5.77729470E+04,
                    -1.04067264E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC6KET34',
        atoms='H:12 C:6 O:3',
        thermo=(NASA([300.00, 1387.00],
                    [ 5.64144594E-01,  8.90854292E-02, -6.89132035E-05,
                    2.83151935E-08, -4.84139667E-12, -4.89978587E+04,
                    3.24785934E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.60960142E+01,  2.71703309E-02, -9.48970254E-06,
                    1.49407399E-09, -8.75315577E-14, -5.77729470E+04,
                    -1.04067264E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC6KET35',
        atoms='H:12 C:6 O:3',
        thermo=(NASA([300.00, 1385.00],
                    [ 2.47172709E+00,  8.26409167E-02, -6.21054242E-05,
                    2.53063842E-08, -4.36121469E-12, -4.98579950E+04,
                    2.29726065E+01]),
                NASA([1385.00, 5000.00],
                    [ 2.55125763E+01,  2.76136115E-02, -9.62969344E-06,
                    1.51447067E-09, -8.86568371E-14, -5.79323622E+04,
                    -1.00670318E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC6KET36',
        atoms='H:12 C:6 O:3',
        thermo=(NASA([300.00, 1383.00],
                    [ 2.96919433E+00,  7.84436519E-02, -5.55361775E-05,
                    2.13589050E-08, -3.52935948E-12, -4.75201133E+04,
                    2.14680944E+01]),
                NASA([1383.00, 5000.00],
                    [ 2.49194999E+01,  2.81507231E-02, -9.82266516E-06,
                    1.54541788E-09, -9.04929606E-14, -5.54463958E+04,
                    -9.71141228E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C6H13O-1',
        atoms='H:13 C:6 O:1',
        thermo=(NASA([300.00, 1391.00],
                    [-2.75164321E-01,  7.54319868E-02, -4.89209083E-05,
                    1.64641255E-08, -2.29328813E-12, -1.49282944E+04,
                    3.31091672E+01]),
                NASA([1391.00, 5000.00],
                    [ 2.04847348E+01,  2.93216823E-02, -1.00270389E-05,
                    1.55645319E-09, -9.02886762E-14, -2.24313312E+04,
                    -7.93220600E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.662,
                                well_depth=540.597),
        note='4/7/97THERM')

species(name='C6H13O-2',
        atoms='H:13 C:6 O:1',
        thermo=(NASA([300.00, 1396.00],
                    [-7.81250097E-01,  7.97868977E-02, -5.58105384E-05,
                    2.06261124E-08, -3.17214287E-12, -1.71172031E+04,
                    3.46483507E+01]),
                NASA([1396.00, 5000.00],
                    [ 2.10569902E+01,  2.87986753E-02, -9.83767726E-06,
                    1.52590233E-09, -8.84677299E-14, -2.47452394E+04,
                    -8.27139640E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.662,
                                well_depth=540.597),
        note='4/7/97THERM')

species(name='C6H13O-3',
        atoms='H:13 C:6 O:1',
        thermo=(NASA([300.00, 1396.00],
                    [-7.81250097E-01,  7.97868977E-02, -5.58105384E-05,
                    2.06261124E-08, -3.17214287E-12, -1.71172031E+04,
                    3.46483507E+01]),
                NASA([1396.00, 5000.00],
                    [ 2.10569902E+01,  2.87986753E-02, -9.83767726E-06,
                    1.52590233E-09, -8.84677299E-14, -2.47452394E+04,
                    -8.27139640E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.662,
                                well_depth=540.597),
        note='4/7/97THERM')

species(name='C6H12OH-1',
        atoms='H:13 C:6 O:1',
        thermo=(NASA([300.00, 1374.00],
                    [ 7.30467050E-01,  6.80553890E-02, -3.71632261E-05,
                    8.47842648E-09, -3.99384332E-13, -1.78478288E+04,
                    3.20923365E+01]),
                NASA([1374.00, 5000.00],
                    [ 2.08351245E+01,  2.75160242E-02, -9.19212642E-06,
                    1.40613856E-09, -8.07910065E-14, -2.54006048E+04,
                    -7.80442288E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.674,
                                well_depth=541.5,
                                dipole=1.8,
                                rot_relax=1.0),
        note='6/25/99THERM')

species(name='C6H12OH-2',
        atoms='H:13 C:6 O:1',
        thermo=(NASA([300.00, 1390.00],
                    [ 2.03282633E-01,  7.28935457E-02, -4.59200360E-05,
                    1.45115010E-08, -1.81146964E-12, -2.00036461E+04,
                    3.36751650E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.05388853E+01,  2.82512269E-02, -9.53280388E-06,
                    1.46688381E-09, -8.45889133E-14, -2.73356962E+04,
                    -7.65278460E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.674,
                                well_depth=541.5,
                                dipole=1.8,
                                rot_relax=1.0),
        note='6/25/99THERM')

species(name='C6H12OH-3',
        atoms='H:13 C:6 O:1',
        thermo=(NASA([300.00, 1390.00],
                    [ 2.03282633E-01,  7.28935457E-02, -4.59200360E-05,
                    1.45115010E-08, -1.81146964E-12, -2.00036461E+04,
                    3.36751650E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.05388853E+01,  2.82512269E-02, -9.53280388E-06,
                    1.46688381E-09, -8.45889133E-14, -2.73356962E+04,
                    -7.65278460E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.674,
                                well_depth=541.5,
                                dipole=1.8,
                                rot_relax=1.0),
        note='6/25/99THERM')

species(name='O2C6H12OH-1',
        atoms='H:13 C:6 O:3',
        thermo=(NASA([300.00, 1395.00],
                    [ 2.52250523E+00,  8.03363569E-02, -5.57536493E-05,
                    2.08636775E-08, -3.29895205E-12, -3.74833645E+04,
                    2.47960934E+01]),
                NASA([1395.00, 5000.00],
                    [ 2.37745808E+01,  3.06705946E-02, -1.04669074E-05,
                    1.62226472E-09, -9.39991386E-14, -4.49540557E+04,
                    -8.94862826E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='6/25/99THRM')

species(name='O2C6H12OH-2',
        atoms='H:13 C:6 O:3',
        thermo=(NASA([300.00, 1397.00],
                    [ 1.59446253E+00,  8.62865587E-02, -6.42669164E-05,
                    2.57424275E-08, -4.29953824E-12, -3.95691578E+04,
                    2.83298549E+01]),
                NASA([1397.00, 5000.00],
                    [ 2.46632042E+01,  2.99541594E-02, -1.02286047E-05,
                    1.58601587E-09, -9.19277798E-14, -4.74127816E+04,
                    -9.48234723E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='6/25/99THRM')

species(name='O2C6H12OH-3',
        atoms='H:13 C:6 O:3',
        thermo=(NASA([300.00, 1397.00],
                    [ 1.59446253E+00,  8.62865587E-02, -6.42669164E-05,
                    2.57424275E-08, -4.29953824E-12, -3.95691578E+04,
                    2.83298549E+01]),
                NASA([1397.00, 5000.00],
                    [ 2.46632042E+01,  2.99541594E-02, -1.02286047E-05,
                    1.58601587E-09, -9.19277798E-14, -4.74127816E+04,
                    -9.48234723E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='6/25/99THRM')

species(name='NC4H9CHO',
        atoms='H:10 C:5 O:1',
        thermo=(NASA([300.00, 1381.00],
                    [ 1.59663472E+00,  5.43541416E-02, -3.21020651E-05,
                    9.35773559E-09, -1.06688932E-12, -2.99841025E+04,
                    2.21281498E+01]),
                NASA([1381.00, 5000.00],
                    [ 1.67965163E+01,  2.25684519E-02, -7.67631588E-06,
                    1.18769369E-09, -6.87545554E-14, -3.56826080E+04,
                    -6.09063312E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.778,
                                well_depth=476.0,
                                dipole=2.6,
                                rot_relax=1.0),
        note='2/29/96THERM')

species(name='NC4H9CO',
        atoms='H:9 C:5 O:1',
        thermo=(NASA([300.00, 1382.00],
                    [ 2.43530238E+00,  4.93682376E-02, -2.89883082E-05,
                    8.42863812E-09, -9.68287172E-13, -1.14675541E+04,
                    1.92695879E+01]),
                NASA([1382.00, 5000.00],
                    [ 1.61782939E+01,  2.07991920E-02, -7.11787892E-06,
                    1.10559975E-09, -6.41697187E-14, -1.66537752E+04,
                    -5.58943579E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.778,
                                well_depth=476.0,
                                dipole=2.6,
                                rot_relax=1.0),
        note='2/29/96THERM')

species(name='C4H8CHO-1',
        atoms='H:9 C:5 O:1',
        thermo=(NASA([300.00, 1382.00],
                    [ 2.41842832E+00,  4.96272127E-02, -2.94995966E-05,
                    8.76068653E-09, -1.04084476E-12, -5.38055616E+03,
                    2.04609422E+01]),
                NASA([1382.00, 5000.00],
                    [ 1.62387385E+01,  2.06785503E-02, -7.06216576E-06,
                    1.09553003E-09, -6.35307382E-14, -1.05717872E+04,
                    -5.50446953E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.778,
                                well_depth=476.0,
                                dipole=2.6,
                                rot_relax=1.0),
        note='2/29/96THERM')

species(name='C4H8CHO-2',
        atoms='H:9 C:5 O:1',
        thermo=(NASA([300.00, 2014.00],
                    [ 2.67005026E+00,  4.58193092E-02, -2.29410393E-05,
                    4.53761436E-09, -1.26593772E-13, -6.67002712E+03,
                    2.04249941E+01]),
                NASA([2014.00, 5000.00],
                    [ 1.41407667E+01,  2.28357886E-02, -7.96904210E-06,
                    1.26040379E-09, -7.42466834E-14, -1.07960087E+04,
                    -4.22099585E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.778,
                                well_depth=476.0,
                                dipole=2.6,
                                rot_relax=1.0),
        note='2/29/96THERM')

species(name='C4H8CHO-3',
        atoms='H:9 C:5 O:1',
        thermo=(NASA([300.00, 2014.00],
                    [ 2.67005026E+00,  4.58193092E-02, -2.29410393E-05,
                    4.53761436E-09, -1.26593772E-13, -6.67002712E+03,
                    2.04249941E+01]),
                NASA([2014.00, 5000.00],
                    [ 1.41407667E+01,  2.28357886E-02, -7.96904210E-06,
                    1.26040379E-09, -7.42466834E-14, -1.07960087E+04,
                    -4.22099585E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.778,
                                well_depth=476.0,
                                dipole=2.6,
                                rot_relax=1.0),
        note='2/29/96THERM')

species(name='C4H8CHO-4',
        atoms='H:9 C:5 O:1',
        thermo=(NASA([300.00, 1388.00],
                    [ 1.16494271E+00,  5.43945461E-02, -3.49000792E-05,
                    1.14582711E-08, -1.54697375E-12, -9.80619729E+03,
                    2.34897110E+01]),
                NASA([1388.00, 5000.00],
                    [ 1.65969615E+01,  2.07196999E-02, -7.14969212E-06,
                    1.11659188E-09, -6.50492039E-14, -1.54539499E+04,
                    -6.03180192E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.778,
                                well_depth=476.0,
                                dipole=2.6,
                                rot_relax=1.0),
        note='2/29/96THERM')

species(name='AC3H5CHO',
        atoms='H:6 C:4 O:1',
        thermo=(NASA([300.00, 1390.00],
                    [ 1.59653121E-01,  4.25761347E-02, -2.91900593E-05,
                    1.04044255E-08, -1.53682904E-12, -1.07859656E+04,
                    2.82087323E+01]),
                NASA([1390.00, 5000.00],
                    [ 1.22811161E+01,  1.50167397E-02, -5.17605303E-06,
                    8.07796972E-10, -4.70379360E-14, -1.50985283E+04,
                    -3.72055710E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.009,
                                well_depth=464.2,
                                dipole=2.6,
                                rot_relax=1.0),
        note='2/29/96THERM')

species(name='AC3H5CO',
        atoms='H:5 C:4 O:1',
        thermo=(NASA([300.00, 1389.00],
                    [ 9.34300027E-01,  3.77751824E-02, -2.61384581E-05,
                    9.34041036E-09, -1.37598294E-12,  7.74223639E+03,
                    2.56636231E+01]),
                NASA([1389.00, 5000.00],
                    [ 1.18646590E+01,  1.28852832E-02, -4.45460888E-06,
                    6.96613045E-10, -4.06215490E-14,  3.86344452E+03,
                    -3.32989999E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.009,
                                well_depth=464.2,
                                dipole=2.6,
                                rot_relax=1.0),
        note='2/29/96THERM')

species(name='C2H3CHCHO',
        atoms='H:5 C:4 O:1',
        thermo=(NASA([300.00, 1382.00],
                    [-8.77992983E-03,  3.98630509E-02, -2.66130850E-05,
                    8.74644509E-09, -1.14128759E-12,  6.22475038E+03,
                    2.79591664E+01]),
                NASA([1382.00, 5000.00],
                    [ 1.24074059E+01,  1.27319711E-02, -4.47199396E-06,
                    7.06819922E-10, -4.15245536E-14,  1.71706585E+03,
                    -3.94039464E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.009,
                                well_depth=464.2,
                                dipole=2.6,
                                rot_relax=1.0),
        note='2/29/96THERM')

species(name='NC3H7COCH3',
        atoms='H:10 C:5 O:1',
        thermo=(NASA([300.00, 1383.00],
                    [ 1.24037361E+00,  5.42494047E-02, -3.14825332E-05,
                    9.14492321E-09, -1.07665566E-12, -3.36276472E+04,
                    2.36150068E+01]),
                NASA([1383.00, 5000.00],
                    [ 1.61841564E+01,  2.35245881E-02, -8.09675055E-06,
                    1.26230690E-09, -7.34487839E-14, -3.93375474E+04,
                    -5.82980615E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.735,
                                well_depth=477.0,
                                dipole=2.5,
                                rot_relax=1.0),
        note='2/22/96THERM')

species(name='NC3H7COCH2',
        atoms='H:9 C:5 O:1',
        thermo=(NASA([300.00, 1389.00],
                    [ 9.58299271E-01,  5.68162532E-02, -3.99112781E-05,
                    1.52671514E-08, -2.49221047E-12, -1.23062238E+04,
                    2.34112884E+01]),
                NASA([1389.00, 5000.00],
                    [ 1.61502419E+01,  2.14093466E-02, -7.36059614E-06,
                    1.14656816E-09, -6.66712722E-14, -1.76968561E+04,
                    -5.83864884E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.348,
                                well_depth=489.084),
        note='2/22/96THERM')

species(name='C3H6COCH3-1',
        atoms='H:9 C:5 O:1',
        thermo=(NASA([300.00, 1381.00],
                    [ 2.00033269E+00,  4.96074565E-02, -2.87470600E-05,
                    8.26666830E-09, -9.51074212E-13, -9.01010845E+03,
                    2.22761791E+01]),
                NASA([1381.00, 5000.00],
                    [ 1.58052948E+01,  2.12787457E-02, -7.31842496E-06,
                    1.14049355E-09, -6.63446381E-14, -1.42827587E+04,
                    -5.34016678E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.348,
                                well_depth=489.084),
        note='2/22/96THRM')

species(name='C3H6COCH3-2',
        atoms='H:9 C:5 O:1',
        thermo=(NASA([300.00, 2017.00],
                    [ 2.20585184E+00,  4.59605697E-02, -2.23175553E-05,
                    4.03645332E-09, -1.91585252E-14, -1.02918446E+04,
                    2.24556304E+01]),
                NASA([2017.00, 5000.00],
                    [ 1.36994763E+01,  2.33997602E-02, -8.20838083E-06,
                    1.30268499E-09, -7.69151198E-14, -1.44773517E+04,
                    -4.04820065E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.348,
                                well_depth=489.084),
        note='2/22/96THRM')

species(name='C3H6COCH3-3',
        atoms='H:9 C:5 O:1',
        thermo=(NASA([300.00, 1389.00],
                    [ 6.36829393E-01,  5.50047719E-02, -3.53220538E-05,
                    1.18237807E-08, -1.65550023E-12, -1.34221768E+04,
                    2.57839238E+01]),
                NASA([1389.00, 5000.00],
                    [ 1.57976789E+01,  2.17091054E-02, -7.46487769E-06,
                    1.16300253E-09, -6.76366022E-14, -1.89670489E+04,
                    -5.65047695E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.348,
                                well_depth=489.084),
        note='2/22/96THRM')

species(name='C2H5COC2H5',
        atoms='H:10 C:5 O:1',
        thermo=(NASA([300.00, 1376.00],
                    [ 1.59094611E+00,  5.45541559E-02, -3.33407641E-05,
                    1.05931387E-08, -1.42002914E-12, -3.39282740E+04,
                    2.09983080E+01]),
                NASA([1376.00, 5000.00],
                    [ 1.67131896E+01,  2.24769805E-02, -7.61830162E-06,
                    1.17658492E-09, -6.80456888E-14, -3.96020446E+04,
                    -6.15362004E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.714,
                                well_depth=476.7,
                                dipole=2.7,
                                rot_relax=1.0),
        note='2/22/96THERM')

species(name='C2H5COC2H4P',
        atoms='H:9 C:5 O:1',
        thermo=(NASA([300.00, 1376.00],
                    [ 2.39449461E+00,  4.98731839E-02, -3.07394387E-05,
                    9.95080260E-09, -1.37570832E-12, -9.32123954E+03,
                    2.01162807E+01]),
                NASA([1376.00, 5000.00],
                    [ 1.62000827E+01,  2.05106647E-02, -6.96977982E-06,
                    1.07826230E-09, -6.24324406E-14, -1.45067221E+04,
                    -5.52272610E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.348,
                                well_depth=489.084),
        note='2/22/96THRM')

species(name='C2H5COC2H4S',
        atoms='H:9 C:5 O:1',
        thermo=(NASA([300.00, 1385.00],
                    [ 1.10728106E+00,  5.48902970E-02, -3.67695939E-05,
                    1.32867426E-08, -2.06966690E-12, -1.37435363E+04,
                    2.32804315E+01]),
                NASA([1385.00, 5000.00],
                    [ 1.61904470E+01,  2.12169293E-02, -7.35412919E-06,
                    1.15187377E-09, -6.72380864E-14, -1.92650763E+04,
                    -5.84937265E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.348,
                                well_depth=489.084),
        note='2/22/96THRM')

species(name='C2H5COC2H3',
        atoms='H:8 C:5 O:1',
        thermo=(NASA([300.00, 1387.00],
                    [ 4.89126432E-01,  5.49951823E-02, -3.96989101E-05,
                    1.55700373E-08, -2.60070818E-12, -2.02458499E+04,
                    2.59411142E+01]),
                NASA([1387.00, 5000.00],
                    [ 1.55916947E+01,  1.96028302E-02, -6.79737285E-06,
                    1.06496015E-09, -6.21767189E-14, -2.55988351E+04,
                    -5.53302818E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.714,
                                well_depth=476.7,
                                dipole=2.7,
                                rot_relax=1.0),
        note='2/22/96THERM')

species(name='PC2H4COC2H3',
        atoms='H:7 C:5 O:1',
        thermo=(NASA([300.00, 1404.00],
                    [ 7.62539679E-01,  5.35805655E-02, -4.36465022E-05,
                    1.99232699E-08, -3.73335386E-12,  4.42260800E+03,
                    2.66604751E+01]),
                NASA([1404.00, 5000.00],
                    [ 1.26732841E+01,  2.02366192E-02, -6.61089203E-06,
                    9.93878746E-10, -5.63457733E-14,  7.74011225E+02,
                    -3.54543594E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.714,
                                well_depth=476.7,
                                dipole=2.7,
                                rot_relax=1.0),
        note='2/22/96THRM')

species(name='SC2H4COC2H3',
        atoms='H:7 C:5 O:1',
        thermo=(NASA([300.00, 1389.00],
                    [-1.60114035E-01,  5.57954071E-02, -4.31353298E-05,
                    1.76457444E-08, -2.99384261E-12, -3.12060218E+01,
                    2.83440609E+01]),
                NASA([1389.00, 5000.00],
                    [ 1.58074213E+01,  1.69721022E-02, -5.91317608E-06,
                    9.29449082E-10, -5.43903161E-14, -5.49230389E+03,
                    -5.69867205E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.714,
                                well_depth=476.7,
                                dipole=2.7,
                                rot_relax=1.0),
        note='2/22/96THRM')

species(name='NC7H16',
        atoms='H:16 C:7',
        thermo=(NASA([300.00, 1391.00],
                    [-1.26836187E+00,  8.54355820E-02, -5.25346786E-05,
                    1.62945721E-08, -2.02394925E-12, -2.56586565E+04,
                    3.53732912E+01]),
                NASA([1391.00, 5000.00],
                    [ 2.22148969E+01,  3.47675750E-02, -1.18407129E-05,
                    1.83298478E-09, -1.06130266E-13, -3.42760081E+04,
                    -9.23040196E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.253,
                                well_depth=459.6,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H15-1',
        atoms='H:15 C:7',
        thermo=(NASA([300.00, 1390.00],
                    [-4.99570406E-01,  8.08826467E-02, -5.00532754E-05,
                    1.56549308E-08, -1.96616227E-12, -1.04590223E+03,
                    3.46564011E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.17940709E+01,  3.26280243E-02, -1.11138244E-05,
                    1.72067148E-09, -9.96366999E-14, -9.20938221E+03,
                    -8.64954311E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.253,
                                well_depth=459.6,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H15-2',
        atoms='H:15 C:7',
        thermo=(NASA([300.00, 1382.00],
                    [-3.79155767E-02,  7.56726570E-02, -4.07473634E-05,
                    9.32678943E-09, -4.92360745E-13, -2.35605303E+03,
                    3.37321506E+01]),
                NASA([1382.00, 5000.00],
                    [ 2.16368842E+01,  3.23324804E-02, -1.09273807E-05,
                    1.68357060E-09, -9.71774091E-14, -1.05873616E+04,
                    -8.52209653E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.253,
                                well_depth=459.6,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H15-3',
        atoms='H:15 C:7',
        thermo=(NASA([300.00, 1382.00],
                    [-3.79155767E-02,  7.56726570E-02, -4.07473634E-05,
                    9.32678943E-09, -4.92360745E-13, -2.35605303E+03,
                    3.37321506E+01]),
                NASA([1382.00, 5000.00],
                    [ 2.16368842E+01,  3.23324804E-02, -1.09273807E-05,
                    1.68357060E-09, -9.71774091E-14, -1.05873616E+04,
                    -8.52209653E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.253,
                                well_depth=459.6,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H15-4',
        atoms='H:15 C:7',
        thermo=(NASA([300.00, 1382.00],
                    [-3.79155767E-02,  7.56726570E-02, -4.07473634E-05,
                    9.32678943E-09, -4.92360745E-13, -2.35605303E+03,
                    3.30426690E+01]),
                NASA([1382.00, 5000.00],
                    [ 2.16368842E+01,  3.23324804E-02, -1.09273807E-05,
                    1.68357060E-09, -9.71774091E-14, -1.05873616E+04,
                    -8.59104470E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.253,
                                well_depth=459.6,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14-1',
        atoms='H:14 C:7',
        thermo=(NASA([300.00, 1392.00],
                    [-1.67720549E+00,  8.24611601E-02, -5.46504108E-05,
                    1.87862303E-08, -2.65737983E-12, -1.02168601E+04,
                    3.85068032E+01]),
                NASA([1392.00, 5000.00],
                    [ 2.10898039E+01,  3.10607878E-02, -1.05644793E-05,
                    1.63405780E-09, -9.45598219E-14, -1.83260065E+04,
                    -8.44391108E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.173,
                                well_depth=457.8,
                                dipole=0.3,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14-2',
        atoms='H:14 C:7',
        thermo=(NASA([300.00, 1390.00],
                    [-1.16533279E+00,  7.90439806E-02, -4.96101666E-05,
                    1.58569009E-08, -2.05346433E-12, -1.17362359E+04,
                    3.59871070E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.06192047E+01,  3.14852991E-02, -1.07162057E-05,
                    1.65827662E-09, -9.59911785E-14, -1.96713162E+04,
                    -8.22519387E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.173,
                                well_depth=457.8,
                                dipole=0.3,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14-3',
        atoms='H:14 C:7',
        thermo=(NASA([300.00, 1392.00],
                    [-2.03026994E+00,  8.26324377E-02, -5.45514471E-05,
                    1.87705822E-08, -2.67571220E-12, -1.15141029E+04,
                    4.02316266E+01]),
                NASA([1392.00, 5000.00],
                    [ 2.06822750E+01,  3.15388629E-02, -1.07571215E-05,
                    1.66690260E-09, -9.65810393E-14, -1.96450604E+04,
                    -8.25235002E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.173,
                                well_depth=457.8,
                                dipole=0.3,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H131-3',
        atoms='H:13 C:7',
        thermo=(NASA([300.00, 1392.00],
                    [-2.01707658E+00,  8.08915559E-02, -5.43383904E-05,
                    1.88066108E-08, -2.66059253E-12,  6.66507961E+03,
                    3.89797235E+01]),
                NASA([1392.00, 5000.00],
                    [ 2.06977799E+01,  2.93585249E-02, -9.99754351E-06,
                    1.54773305E-09, -8.96230815E-14, -1.38049514E+03,
                    -8.35627043E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.173,
                                well_depth=457.8,
                                dipole=0.3,
                                rot_relax=1.0),
        note='6/24/99THERM')

species(name='C7H131-4',
        atoms='H:13 C:7',
        thermo=(NASA([300.00, 1383.00],
                    [-3.37870973E-01,  7.22549356E-02, -4.22499105E-05,
                    1.14646497E-08, -1.05306924E-12,  1.29222803E+04,
                    3.56636773E+01]),
                NASA([1383.00, 5000.00],
                    [ 2.05321062E+01,  2.85988588E-02, -9.64125534E-06,
                    1.48314293E-09, -8.55245930E-14,  5.20664966E+03,
                    -7.81606882E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.173,
                                well_depth=457.8,
                                dipole=0.3,
                                rot_relax=1.0),
        note='6/24/99THERM')

species(name='C7H131-5',
        atoms='H:13 C:7',
        thermo=(NASA([300.00, 1383.00],
                    [-3.37870973E-01,  7.22549356E-02, -4.22499105E-05,
                    1.14646497E-08, -1.05306924E-12,  1.29222803E+04,
                    3.56636773E+01]),
                NASA([1383.00, 5000.00],
                    [ 2.05321062E+01,  2.85988588E-02, -9.64125534E-06,
                    1.48314293E-09, -8.55245930E-14,  5.20664966E+03,
                    -7.81606882E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.173,
                                well_depth=457.8,
                                dipole=0.3,
                                rot_relax=1.0),
        note='6/24/99THERM')

species(name='C7H131-6',
        atoms='H:13 C:7',
        thermo=(NASA([300.00, 1383.00],
                    [-3.37870973E-01,  7.22549356E-02, -4.22499105E-05,
                    1.14646497E-08, -1.05306924E-12,  1.29222803E+04,
                    3.56636773E+01]),
                NASA([1383.00, 5000.00],
                    [ 2.05321062E+01,  2.85988588E-02, -9.64125534E-06,
                    1.48314293E-09, -8.55245930E-14,  5.20664966E+03,
                    -7.81606882E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.173,
                                well_depth=457.8,
                                dipole=0.3,
                                rot_relax=1.0),
        note='6/24/99THERM')

species(name='C7H131-7',
        atoms='H:13 C:7',
        thermo=(NASA([300.00, 1392.00],
                    [-8.56743712E-01,  7.77195832E-02, -5.19282134E-05,
                    1.80146398E-08, -2.57342002E-12,  1.42410455E+04,
                    3.68476570E+01]),
                NASA([1392.00, 5000.00],
                    [ 2.06754243E+01,  2.89092236E-02, -9.83222685E-06,
                    1.52079763E-09, -8.80072255E-14,  6.59360193E+03,
                    -7.93564689E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.173,
                                well_depth=457.8,
                                dipole=0.3,
                                rot_relax=1.0),
        note='6/24/99THERM')

species(name='C7H132-4',
        atoms='H:13 C:7',
        thermo=(NASA([300.00, 1390.00],
                    [-1.50555526E+00,  7.74758940E-02, -4.92982536E-05,
                    1.58761889E-08, -2.05624822E-12,  5.16084218E+03,
                    3.64615905E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.02283723E+01,  2.97818393E-02, -1.01488201E-05,
                    1.57187813E-09, -9.10500113E-14, -2.71119859E+03,
                    -8.13824175E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.173,
                                well_depth=457.8,
                                dipole=0.3,
                                rot_relax=1.0),
        note='6/24/99THERM')

species(name='C7H132-5',
        atoms='H:13 C:7',
        thermo=(NASA([300.00, 1379.00],
                    [ 1.92192199E-01,  6.87557473E-02, -3.70909926E-05,
                    8.46587268E-09, -4.34596882E-13,  1.14153061E+04,
                    3.30601003E+01]),
                NASA([1379.00, 5000.00],
                    [ 2.00522952E+01,  2.90375466E-02, -9.79903704E-06,
                    1.50841035E-09, -8.70206514E-14,  3.87993522E+03,
                    -7.59217636E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.173,
                                well_depth=457.8,
                                dipole=0.3,
                                rot_relax=1.0),
        note='6/24/99THERM')

species(name='C7H132-6',
        atoms='H:13 C:7',
        thermo=(NASA([300.00, 1379.00],
                    [ 1.92192199E-01,  6.87557473E-02, -3.70909926E-05,
                    8.46587268E-09, -4.34596882E-13,  1.14153061E+04,
                    3.30601003E+01]),
                NASA([1379.00, 5000.00],
                    [ 2.00522952E+01,  2.90375466E-02, -9.79903704E-06,
                    1.50841035E-09, -8.70206514E-14,  3.87993522E+03,
                    -7.59217636E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.173,
                                well_depth=457.8,
                                dipole=0.3,
                                rot_relax=1.0),
        note='6/24/99THERM')

species(name='C7H132-7',
        atoms='H:13 C:7',
        thermo=(NASA([300.00, 1390.00],
                    [-4.54324169E-01,  7.47544837E-02, -4.75103113E-05,
                    1.54448401E-08, -2.04403706E-12,  1.27541308E+04,
                    3.48464580E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.02038608E+01,  2.93419067E-02, -9.98810855E-06,
                    1.54578179E-09, -8.94871756E-14,  5.26290572E+03,
                    -7.71644519E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.173,
                                well_depth=457.8,
                                dipole=0.3,
                                rot_relax=1.0),
        note='6/24/99THERM')

species(name='C7H133-1',
        atoms='H:13 C:7',
        thermo=(NASA([300.00, 1392.00],
                    [-1.28816625E+00,  7.82039641E-02, -5.22552162E-05,
                    1.82427873E-08, -2.64174309E-12,  1.29867502E+04,
                    3.89474921E+01]),
                NASA([1392.00, 5000.00],
                    [ 2.02595343E+01,  2.94019199E-02, -1.00312524E-05,
                    1.55475333E-09, -9.00969885E-14,  5.30761197E+03,
                    -7.73922048E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.173,
                                well_depth=457.8,
                                dipole=0.3,
                                rot_relax=1.0),
        note='6/24/99THERM')

species(name='C7H133-5',
        atoms='H:13 C:7',
        thermo=(NASA([300.00, 1392.00],
                    [-2.36092106E+00,  8.10269808E-02, -5.41952361E-05,
                    1.87683817E-08, -2.67470772E-12,  5.39652999E+03,
                    4.06609319E+01]),
                NASA([1392.00, 5000.00],
                    [ 2.02852000E+01,  2.98429179E-02, -1.01927262E-05,
                    1.58100534E-09, -9.16702393E-14, -2.66699385E+03,
                    -8.16175392E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.173,
                                well_depth=457.8,
                                dipole=0.3,
                                rot_relax=1.0),
        note='6/24/99THERM')

species(name='C7H133-6',
        atoms='H:13 C:7',
        thermo=(NASA([300.00, 1384.00],
                    [-7.98785387E-01,  7.29104436E-02, -4.28677284E-05,
                    1.18818540E-08, -1.16315170E-12,  1.16713874E+04,
                    3.78868290E+01]),
                NASA([1384.00, 5000.00],
                    [ 2.00865333E+01,  2.91250010E-02, -9.85196347E-06,
                    1.51881375E-09, -8.77061669E-14,  3.93905131E+03,
                    -7.60211796E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.173,
                                well_depth=457.8,
                                dipole=0.3,
                                rot_relax=1.0),
        note='6/24/99THERM')

species(name='C7H133-7',
        atoms='H:13 C:7',
        thermo=(NASA([300.00, 1392.00],
                    [-1.28816625E+00,  7.82039641E-02, -5.22552162E-05,
                    1.82427873E-08, -2.64174309E-12,  1.29867502E+04,
                    3.89474921E+01]),
                NASA([1392.00, 5000.00],
                    [ 2.02595343E+01,  2.94019199E-02, -1.00312524E-05,
                    1.55475333E-09, -9.00969885E-14,  5.30761197E+03,
                    -7.73922048E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.173,
                                well_depth=457.8,
                                dipole=0.3,
                                rot_relax=1.0),
        note='6/24/99THERM')

species(name='C7H13O1-3',
        atoms='H:13 C:7 O:1',
        thermo=(NASA([300.00, 1392.00],
                    [-2.97186844E+00,  9.50411096E-02, -7.14210586E-05,
                    2.71211636E-08, -4.11210980E-12, -3.69998874E+03,
                    4.68736292E+01]),
                NASA([1392.00, 5000.00],
                    [ 2.53923888E+01,  2.67015459E-02, -8.93162555E-06,
                    1.36776722E-09, -7.86546152E-14, -1.31992107E+04,
                    -1.04485728E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='5/14/3THERM')

species(name='C7H13O2-4',
        atoms='H:13 C:7 O:1',
        thermo=(NASA([300.00, 1389.00],
                    [-2.38155227E+00,  9.12476056E-02, -6.58712429E-05,
                    2.39221243E-08, -3.45819168E-12, -5.23001862E+03,
                    4.39988302E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.48888884E+01,  2.71728188E-02, -9.10354489E-06,
                    1.39555380E-09, -8.03103189E-14, -1.45377934E+04,
                    -1.02121928E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='5/14/3THERM')

species(name='C7H13O3-5',
        atoms='H:13 C:7 O:1',
        thermo=(NASA([300.00, 1393.00],
                    [-4.04764701E+00,  9.94962377E-02, -7.94282837E-05,
                    3.26919568E-08, -5.42561898E-12, -5.01665573E+03,
                    5.17524548E+01]),
                NASA([1393.00, 5000.00],
                    [ 2.46614185E+01,  2.71531970E-02, -9.04971128E-06,
                    1.38224414E-09, -7.93365017E-14, -1.43669878E+04,
                    -1.00409922E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='3/28/97THERM')

species(name='C7H15O2-1',
        atoms='H:15 C:7 O:2',
        thermo=(NASA([300.00, 1388.00],
                    [ 1.20888552E+00,  8.84061706E-02, -5.79861260E-05,
                    2.03574689E-08, -3.06460336E-12, -1.92918765E+04,
                    2.98117097E+01]),
                NASA([1388.00, 5000.00],
                    [ 2.52656556E+01,  3.50803587E-02, -1.21178801E-05,
                    1.89360806E-09, -1.10354125E-13, -2.81113392E+04,
                    -1.00704794E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H15O2-2',
        atoms='H:15 C:7 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [ 3.55252917E-01,  9.42381007E-02, -6.66955106E-05,
                    2.54795583E-08, -4.13211105E-12, -2.15795254E+04,
                    3.29433435E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.60535640E+01,  3.43831688E-02, -1.18713247E-05,
                    1.85450241E-09, -1.08052399E-13, -3.06842095E+04,
                    -1.05407563E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H15O2-3',
        atoms='H:15 C:7 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [ 3.55252917E-01,  9.42381007E-02, -6.66955106E-05,
                    2.54795583E-08, -4.13211105E-12, -2.15795254E+04,
                    3.29433435E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.60535640E+01,  3.43831688E-02, -1.18713247E-05,
                    1.85450241E-09, -1.08052399E-13, -3.06842095E+04,
                    -1.05407563E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H15O2-4',
        atoms='H:15 C:7 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [ 3.55252917E-01,  9.42381007E-02, -6.66955106E-05,
                    2.54795583E-08, -4.13211105E-12, -2.15795254E+04,
                    3.22538619E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.60535640E+01,  3.43831688E-02, -1.18713247E-05,
                    1.85450241E-09, -1.08052399E-13, -3.06842095E+04,
                    -1.06097045E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H15O2H-1',
        atoms='H:16 C:7 O:2',
        thermo=(NASA([300.00, 1388.00],
                    [ 2.69178904E-01,  9.71211877E-02, -6.59765921E-05,
                    2.35449071E-08, -3.52849016E-12, -3.63451119E+04,
                    3.34740655E+01]),
                NASA([1388.00, 5000.00],
                    [ 2.77185802E+01,  3.52624458E-02, -1.22144164E-05,
                    1.91237244E-09, -1.11602067E-13, -4.62267320E+04,
                    -1.14955409E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H15O2H-2',
        atoms='H:16 C:7 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [-6.02100521E-01,  1.02885015E-01, -7.43422840E-05,
                    2.83646038E-08, -4.51796063E-12, -3.86269481E+04,
                    3.67137899E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.85706023E+01,  3.45345478E-02, -1.19627321E-05,
                    1.87305198E-09, -1.09312061E-13, -4.88414901E+04,
                    -1.20064377E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H15O2H-3',
        atoms='H:16 C:7 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [-6.02100521E-01,  1.02885015E-01, -7.43422840E-05,
                    2.83646038E-08, -4.51796063E-12, -3.86269481E+04,
                    3.67137899E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.85706023E+01,  3.45345478E-02, -1.19627321E-05,
                    1.87305198E-09, -1.09312061E-13, -4.88414901E+04,
                    -1.20064377E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H15O2H-4',
        atoms='H:16 C:7 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [-6.02100521E-01,  1.02885015E-01, -7.43422840E-05,
                    2.83646038E-08, -4.51796063E-12, -3.86269481E+04,
                    3.60243083E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.85706023E+01,  3.45345478E-02, -1.19627321E-05,
                    1.87305198E-09, -1.09312061E-13, -4.88414901E+04,
                    -1.20753858E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H15O-1',
        atoms='H:15 C:7 O:1',
        thermo=(NASA([300.00, 1391.00],
                    [-4.59189934E-01,  8.74464647E-02, -5.69015135E-05,
                    1.92195908E-08, -2.68753465E-12, -1.78233113E+04,
                    3.56475170E+01]),
                NASA([1391.00, 5000.00],
                    [ 2.36489937E+01,  3.38195865E-02, -1.15717686E-05,
                    1.79692111E-09, -1.04265757E-13, -2.65280214E+04,
                    -9.48882293E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='4/7/97THERM')

species(name='C7H15O-2',
        atoms='H:15 C:7 O:1',
        thermo=(NASA([300.00, 1395.00],
                    [-1.09087925E+00,  9.23217022E-02, -6.44477835E-05,
                    2.36808474E-08, -3.61067567E-12, -1.99926675E+04,
                    3.77738325E+01]),
                NASA([1395.00, 5000.00],
                    [ 2.43070968E+01,  3.31267815E-02, -1.13034023E-05,
                    1.75198875E-09, -1.01526945E-13, -2.88623481E+04,
                    -9.87360860E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='4/7/97THERM')

species(name='C7H15O-3',
        atoms='H:15 C:7 O:1',
        thermo=(NASA([300.00, 1395.00],
                    [-1.09087925E+00,  9.23217022E-02, -6.44477835E-05,
                    2.36808474E-08, -3.61067567E-12, -1.99926675E+04,
                    3.77738325E+01]),
                NASA([1395.00, 5000.00],
                    [ 2.43070968E+01,  3.31267815E-02, -1.13034023E-05,
                    1.75198875E-09, -1.01526945E-13, -2.88623481E+04,
                    -9.87360860E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='4/7/97THERM')

species(name='C7H15O-4',
        atoms='H:15 C:7 O:1',
        thermo=(NASA([300.00, 1395.00],
                    [-1.09087925E+00,  9.23217022E-02, -6.44477835E-05,
                    2.36808474E-08, -3.61067567E-12, -1.99926675E+04,
                    3.70843509E+01]),
                NASA([1395.00, 5000.00],
                    [ 2.43070968E+01,  3.31267815E-02, -1.13034023E-05,
                    1.75198875E-09, -1.01526945E-13, -2.88623481E+04,
                    -9.94255677E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='4/7/97THERM')

species(name='C7H14OOH1-2',
        atoms='H:15 C:7 O:2',
        thermo=(NASA([300.00, 1383.00],
                    [ 1.11146206E+00,  8.91595392E-02, -5.67352382E-05,
                    1.78890094E-08, -2.22260206E-12, -1.29868314E+04,
                    3.29200052E+01]),
                NASA([1383.00, 5000.00],
                    [ 2.72290222E+01,  3.24921901E-02, -1.11259758E-05,
                    1.72928878E-09, -1.00428326E-13, -2.24977070E+04,
                    -1.08901784E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14OOH1-3',
        atoms='H:15 C:7 O:2',
        thermo=(NASA([300.00, 1383.00],
                    [ 1.11146206E+00,  8.91595392E-02, -5.67352382E-05,
                    1.78890094E-08, -2.22260206E-12, -1.29868314E+04,
                    3.29200052E+01]),
                NASA([1383.00, 5000.00],
                    [ 2.72290222E+01,  3.24921901E-02, -1.11259758E-05,
                    1.72928878E-09, -1.00428326E-13, -2.24977070E+04,
                    -1.08901784E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14OOH1-4',
        atoms='H:15 C:7 O:2',
        thermo=(NASA([300.00, 1383.00],
                    [ 1.11146206E+00,  8.91595392E-02, -5.67352382E-05,
                    1.78890094E-08, -2.22260206E-12, -1.29868314E+04,
                    3.29200052E+01]),
                NASA([1383.00, 5000.00],
                    [ 2.72290222E+01,  3.24921901E-02, -1.11259758E-05,
                    1.72928878E-09, -1.00428326E-13, -2.24977070E+04,
                    -1.08901784E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14OOH1-5',
        atoms='H:15 C:7 O:2',
        thermo=(NASA([300.00, 1383.00],
                    [ 1.11146206E+00,  8.91595392E-02, -5.67352382E-05,
                    1.78890094E-08, -2.22260206E-12, -1.29868314E+04,
                    3.29200052E+01]),
                NASA([1383.00, 5000.00],
                    [ 2.72290222E+01,  3.24921901E-02, -1.11259758E-05,
                    1.72928878E-09, -1.00428326E-13, -2.24977070E+04,
                    -1.08901784E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14OOH2-1',
        atoms='H:15 C:7 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [ 1.51527018E-01,  9.84303098E-02, -7.20555130E-05,
                    2.78535130E-08, -4.48821028E-12, -1.40125600E+04,
                    3.53720992E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.81418167E+01,  3.23810222E-02, -1.12264895E-05,
                    1.75881065E-09, -1.02687430E-13, -2.37649208E+04,
                    -1.14883305E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14OOH2-3',
        atoms='H:15 C:7 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 8.49786005E-01,  9.25107196E-02, -6.25359834E-05,
                    2.17880390E-08, -3.12720398E-12, -1.53640679E+04,
                    3.32997487E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.74034778E+01,  3.27527893E-02, -1.12991770E-05,
                    1.76444649E-09, -1.02786945E-13, -2.48636651E+04,
                    -1.10201978E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14OOH2-4',
        atoms='H:15 C:7 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 8.49786005E-01,  9.25107196E-02, -6.25359834E-05,
                    2.17880390E-08, -3.12720398E-12, -1.53640679E+04,
                    3.32997487E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.74034778E+01,  3.27527893E-02, -1.12991770E-05,
                    1.76444649E-09, -1.02786945E-13, -2.48636651E+04,
                    -1.10201978E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14OOH2-5',
        atoms='H:15 C:7 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 8.49786005E-01,  9.25107196E-02, -6.25359834E-05,
                    2.17880390E-08, -3.12720398E-12, -1.53640679E+04,
                    3.32997487E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.74034778E+01,  3.27527893E-02, -1.12991770E-05,
                    1.76444649E-09, -1.02786945E-13, -2.48636651E+04,
                    -1.10201978E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14OOH2-6',
        atoms='H:15 C:7 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 8.49786005E-01,  9.25107196E-02, -6.25359834E-05,
                    2.17880390E-08, -3.12720398E-12, -1.53640679E+04,
                    3.32997487E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.74034778E+01,  3.27527893E-02, -1.12991770E-05,
                    1.76444649E-09, -1.02786945E-13, -2.48636651E+04,
                    -1.10201978E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14OOH3-1',
        atoms='H:15 C:7 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [ 1.51527018E-01,  9.84303098E-02, -7.20555130E-05,
                    2.78535130E-08, -4.48821028E-12, -1.40125600E+04,
                    3.53720992E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.81418167E+01,  3.23810222E-02, -1.12264895E-05,
                    1.75881065E-09, -1.02687430E-13, -2.37649208E+04,
                    -1.14883305E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14OOH3-2',
        atoms='H:15 C:7 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 8.49786005E-01,  9.25107196E-02, -6.25359834E-05,
                    2.17880390E-08, -3.12720398E-12, -1.53640679E+04,
                    3.32997487E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.74034778E+01,  3.27527893E-02, -1.12991770E-05,
                    1.76444649E-09, -1.02786945E-13, -2.48636651E+04,
                    -1.10201978E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14OOH3-4',
        atoms='H:15 C:7 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 8.49786005E-01,  9.25107196E-02, -6.25359834E-05,
                    2.17880390E-08, -3.12720398E-12, -1.53640679E+04,
                    3.32997487E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.74034778E+01,  3.27527893E-02, -1.12991770E-05,
                    1.76444649E-09, -1.02786945E-13, -2.48636651E+04,
                    -1.10201978E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14OOH3-5',
        atoms='H:15 C:7 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 8.49786005E-01,  9.25107196E-02, -6.25359834E-05,
                    2.17880390E-08, -3.12720398E-12, -1.53640679E+04,
                    3.32997487E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.74034778E+01,  3.27527893E-02, -1.12991770E-05,
                    1.76444649E-09, -1.02786945E-13, -2.48636651E+04,
                    -1.10201978E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14OOH3-6',
        atoms='H:15 C:7 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 8.49786005E-01,  9.25107196E-02, -6.25359834E-05,
                    2.17880390E-08, -3.12720398E-12, -1.53640679E+04,
                    3.32997487E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.74034778E+01,  3.27527893E-02, -1.12991770E-05,
                    1.76444649E-09, -1.02786945E-13, -2.48636651E+04,
                    -1.10201978E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14OOH3-7',
        atoms='H:15 C:7 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [ 1.51527018E-01,  9.84303098E-02, -7.20555130E-05,
                    2.78535130E-08, -4.48821028E-12, -1.40125600E+04,
                    3.53720992E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.81418167E+01,  3.23810222E-02, -1.12264895E-05,
                    1.75881065E-09, -1.02687430E-13, -2.37649208E+04,
                    -1.14883305E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14OOH4-1',
        atoms='H:15 C:7 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [ 1.51527018E-01,  9.84303098E-02, -7.20555130E-05,
                    2.78535130E-08, -4.48821028E-12, -1.40125600E+04,
                    3.46826176E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.81418167E+01,  3.23810222E-02, -1.12264895E-05,
                    1.75881065E-09, -1.02687430E-13, -2.37649208E+04,
                    -1.15572787E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14OOH4-2',
        atoms='H:15 C:7 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 8.49786005E-01,  9.25107196E-02, -6.25359834E-05,
                    2.17880390E-08, -3.12720398E-12, -1.53640679E+04,
                    3.26102671E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.74034778E+01,  3.27527893E-02, -1.12991770E-05,
                    1.76444649E-09, -1.02786945E-13, -2.48636651E+04,
                    -1.10891460E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14OOH4-3',
        atoms='H:15 C:7 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 8.49786005E-01,  9.25107196E-02, -6.25359834E-05,
                    2.17880390E-08, -3.12720398E-12, -1.53640679E+04,
                    3.26102671E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.74034778E+01,  3.27527893E-02, -1.12991770E-05,
                    1.76444649E-09, -1.02786945E-13, -2.48636651E+04,
                    -1.10891460E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14OOH1-2O2',
        atoms='H:15 C:7 O:4',
        thermo=(NASA([300.00, 1388.00],
                    [ 1.89718004E+00,  1.06229796E-01, -8.02770218E-05,
                    3.23649093E-08, -5.45850472E-12, -3.22790229E+04,
                    3.02474612E+01]),
                NASA([1388.00, 5000.00],
                    [ 3.20144349E+01,  3.40055996E-02, -1.18415856E-05,
                    1.86060347E-09, -1.08851012E-13, -4.27171859E+04,
                    -1.31118243E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=7.229,
                                well_depth=600.6,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C7H14OOH1-3O2',
        atoms='H:15 C:7 O:4',
        thermo=(NASA([300.00, 1388.00],
                    [ 1.89718004E+00,  1.06229796E-01, -8.02770218E-05,
                    3.23649093E-08, -5.45850472E-12, -3.22790229E+04,
                    3.02474612E+01]),
                NASA([1388.00, 5000.00],
                    [ 3.20144349E+01,  3.40055996E-02, -1.18415856E-05,
                    1.86060347E-09, -1.08851012E-13, -4.27171859E+04,
                    -1.31118243E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=7.229,
                                well_depth=600.6,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C7H14OOH1-4O2',
        atoms='H:15 C:7 O:4',
        thermo=(NASA([300.00, 1388.00],
                    [ 1.89718004E+00,  1.06229796E-01, -8.02770218E-05,
                    3.23649093E-08, -5.45850472E-12, -3.22790229E+04,
                    3.02474612E+01]),
                NASA([1388.00, 5000.00],
                    [ 3.20144349E+01,  3.40055996E-02, -1.18415856E-05,
                    1.86060347E-09, -1.08851012E-13, -4.27171859E+04,
                    -1.31118243E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=7.229,
                                well_depth=600.6,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C7H14OOH1-5O2',
        atoms='H:15 C:7 O:4',
        thermo=(NASA([300.00, 1388.00],
                    [ 1.89718004E+00,  1.06229796E-01, -8.02770218E-05,
                    3.23649093E-08, -5.45850472E-12, -3.22790229E+04,
                    3.02474612E+01]),
                NASA([1388.00, 5000.00],
                    [ 3.20144349E+01,  3.40055996E-02, -1.18415856E-05,
                    1.86060347E-09, -1.08851012E-13, -4.27171859E+04,
                    -1.31118243E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=7.229,
                                well_depth=600.6,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C7H14OOH2-1O2',
        atoms='H:15 C:7 O:4',
        thermo=(NASA([300.00, 1388.00],
                    [ 1.89718004E+00,  1.06229796E-01, -8.02770218E-05,
                    3.23649093E-08, -5.45850472E-12, -3.22790229E+04,
                    3.02474612E+01]),
                NASA([1388.00, 5000.00],
                    [ 3.20144349E+01,  3.40055996E-02, -1.18415856E-05,
                    1.86060347E-09, -1.08851012E-13, -4.27171859E+04,
                    -1.31118243E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=7.229,
                                well_depth=600.6,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C7H14OOH2-3O2',
        atoms='H:15 C:7 O:4',
        thermo=(NASA([300.00, 1390.00],
                    [ 1.40912479E+00,  1.10720834E-01, -8.71178075E-05,
                    3.63662749E-08, -6.28350689E-12, -3.46304788E+04,
                    3.16223383E+01]),
                NASA([1390.00, 5000.00],
                    [ 3.28358296E+01,  3.32767285E-02, -1.15833285E-05,
                    1.81958638E-09, -1.06434419E-13, -4.52873003E+04,
                    -1.35975710E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=7.229,
                                well_depth=600.6,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C7H14OOH2-4O2',
        atoms='H:15 C:7 O:4',
        thermo=(NASA([300.00, 1390.00],
                    [ 1.40912479E+00,  1.10720834E-01, -8.71178075E-05,
                    3.63662749E-08, -6.28350689E-12, -3.46304788E+04,
                    3.16223383E+01]),
                NASA([1390.00, 5000.00],
                    [ 3.28358296E+01,  3.32767285E-02, -1.15833285E-05,
                    1.81958638E-09, -1.06434419E-13, -4.52873003E+04,
                    -1.35975710E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=7.229,
                                well_depth=600.6,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C7H14OOH2-5O2',
        atoms='H:15 C:7 O:4',
        thermo=(NASA([300.00, 1390.00],
                    [ 1.40912479E+00,  1.10720834E-01, -8.71178075E-05,
                    3.63662749E-08, -6.28350689E-12, -3.46304788E+04,
                    3.16223383E+01]),
                NASA([1390.00, 5000.00],
                    [ 3.28358296E+01,  3.32767285E-02, -1.15833285E-05,
                    1.81958638E-09, -1.06434419E-13, -4.52873003E+04,
                    -1.35975710E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=7.229,
                                well_depth=600.6,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C7H14OOH2-6O2',
        atoms='H:15 C:7 O:4',
        thermo=(NASA([300.00, 1390.00],
                    [ 1.40912479E+00,  1.10720834E-01, -8.71178075E-05,
                    3.63662749E-08, -6.28350689E-12, -3.46304788E+04,
                    3.16223383E+01]),
                NASA([1390.00, 5000.00],
                    [ 3.28358296E+01,  3.32767285E-02, -1.15833285E-05,
                    1.81958638E-09, -1.06434419E-13, -4.52873003E+04,
                    -1.35975710E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=7.229,
                                well_depth=600.6,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C7H14OOH3-1O2',
        atoms='H:15 C:7 O:4',
        thermo=(NASA([300.00, 1388.00],
                    [ 1.89718004E+00,  1.06229796E-01, -8.02770218E-05,
                    3.23649093E-08, -5.45850472E-12, -3.22790229E+04,
                    3.02474612E+01]),
                NASA([1388.00, 5000.00],
                    [ 3.20144349E+01,  3.40055996E-02, -1.18415856E-05,
                    1.86060347E-09, -1.08851012E-13, -4.27171859E+04,
                    -1.31118243E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=7.229,
                                well_depth=600.6,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C7H14OOH3-2O2',
        atoms='H:15 C:7 O:4',
        thermo=(NASA([300.00, 1390.00],
                    [ 1.40912479E+00,  1.10720834E-01, -8.71178075E-05,
                    3.63662749E-08, -6.28350689E-12, -3.46304788E+04,
                    3.16223383E+01]),
                NASA([1390.00, 5000.00],
                    [ 3.28358296E+01,  3.32767285E-02, -1.15833285E-05,
                    1.81958638E-09, -1.06434419E-13, -4.52873003E+04,
                    -1.35975710E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=7.229,
                                well_depth=600.6,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C7H14OOH3-4O2',
        atoms='H:15 C:7 O:4',
        thermo=(NASA([300.00, 1390.00],
                    [ 1.40912479E+00,  1.10720834E-01, -8.71178075E-05,
                    3.63662749E-08, -6.28350689E-12, -3.46304788E+04,
                    3.16223383E+01]),
                NASA([1390.00, 5000.00],
                    [ 3.28358296E+01,  3.32767285E-02, -1.15833285E-05,
                    1.81958638E-09, -1.06434419E-13, -4.52873003E+04,
                    -1.35975710E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=7.229,
                                well_depth=600.6,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C7H14OOH3-5O2',
        atoms='H:15 C:7 O:4',
        thermo=(NASA([300.00, 1390.00],
                    [ 1.40912479E+00,  1.10720834E-01, -8.71178075E-05,
                    3.63662749E-08, -6.28350689E-12, -3.46304788E+04,
                    3.16223383E+01]),
                NASA([1390.00, 5000.00],
                    [ 3.28358296E+01,  3.32767285E-02, -1.15833285E-05,
                    1.81958638E-09, -1.06434419E-13, -4.52873003E+04,
                    -1.35975710E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=7.229,
                                well_depth=600.6,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C7H14OOH3-6O2',
        atoms='H:15 C:7 O:4',
        thermo=(NASA([300.00, 1390.00],
                    [ 1.40912479E+00,  1.10720834E-01, -8.71178075E-05,
                    3.63662749E-08, -6.28350689E-12, -3.46304788E+04,
                    3.16223383E+01]),
                NASA([1390.00, 5000.00],
                    [ 3.28358296E+01,  3.32767285E-02, -1.15833285E-05,
                    1.81958638E-09, -1.06434419E-13, -4.52873003E+04,
                    -1.35975710E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=7.229,
                                well_depth=600.6,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C7H14OOH3-7O2',
        atoms='H:15 C:7 O:4',
        thermo=(NASA([300.00, 1388.00],
                    [ 1.89718004E+00,  1.06229796E-01, -8.02770218E-05,
                    3.23649093E-08, -5.45850472E-12, -3.22790229E+04,
                    3.02474612E+01]),
                NASA([1388.00, 5000.00],
                    [ 3.20144349E+01,  3.40055996E-02, -1.18415856E-05,
                    1.86060347E-09, -1.08851012E-13, -4.27171859E+04,
                    -1.31118243E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=7.229,
                                well_depth=600.6,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C7H14OOH4-1O2',
        atoms='H:15 C:7 O:4',
        thermo=(NASA([300.00, 1388.00],
                    [ 1.89718004E+00,  1.06229796E-01, -8.02770218E-05,
                    3.23649093E-08, -5.45850472E-12, -3.22790229E+04,
                    3.02474612E+01]),
                NASA([1388.00, 5000.00],
                    [ 3.20144349E+01,  3.40055996E-02, -1.18415856E-05,
                    1.86060347E-09, -1.08851012E-13, -4.27171859E+04,
                    -1.31118243E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=7.229,
                                well_depth=600.6,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C7H14OOH4-2O2',
        atoms='H:15 C:7 O:4',
        thermo=(NASA([300.00, 1390.00],
                    [ 1.40912479E+00,  1.10720834E-01, -8.71178075E-05,
                    3.63662749E-08, -6.28350689E-12, -3.46304788E+04,
                    3.16223383E+01]),
                NASA([1390.00, 5000.00],
                    [ 3.28358296E+01,  3.32767285E-02, -1.15833285E-05,
                    1.81958638E-09, -1.06434419E-13, -4.52873003E+04,
                    -1.35975710E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=7.229,
                                well_depth=600.6,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C7H14OOH4-3O2',
        atoms='H:15 C:7 O:4',
        thermo=(NASA([300.00, 1390.00],
                    [ 1.40912479E+00,  1.10720834E-01, -8.71178075E-05,
                    3.63662749E-08, -6.28350689E-12, -3.46304788E+04,
                    3.16223383E+01]),
                NASA([1390.00, 5000.00],
                    [ 3.28358296E+01,  3.32767285E-02, -1.15833285E-05,
                    1.81958638E-09, -1.06434419E-13, -4.52873003E+04,
                    -1.35975710E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=7.229,
                                well_depth=600.6,
                                dipole=1.8,
                                rot_relax=1.0),
        note='7/19/0TRM')

species(name='C7H14O1-2',
        atoms='H:14 C:7 O:1',
        thermo=(NASA([300.00, 1394.00],
                    [-3.65270763E+00,  9.70546060E-02, -7.14482297E-05,
                    2.77781632E-08, -4.48128551E-12, -2.39799254E+04,
                    4.80801402E+01]),
                NASA([1394.00, 5000.00],
                    [ 2.35147269E+01,  3.21674287E-02, -1.10803549E-05,
                    1.72835023E-09, -1.00601041E-13, -3.33352322E+04,
                    -9.74257302E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.297,
                                well_depth=511.5,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14O1-3',
        atoms='H:14 C:7 O:1',
        thermo=(NASA([300.00, 1395.00],
                    [-5.50036450E+00,  1.00155088E-01, -7.29935857E-05,
                    2.79344270E-08, -4.42245696E-12, -2.44292287E+04,
                    5.64768482E+01]),
                NASA([1395.00, 5000.00],
                    [ 2.26918916E+01,  3.32510472E-02, -1.14457765E-05,
                    1.78455535E-09, -1.03841112E-13, -3.41591419E+04,
                    -9.46346563E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.297,
                                well_depth=511.5,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14O1-4',
        atoms='H:14 C:7 O:1',
        thermo=(NASA([300.00, 1398.00],
                    [-6.56017746E+00,  1.00694520E-01, -7.10466567E-05,
                    2.59142094E-08, -3.86849843E-12, -3.42312152E+04,
                    6.12629813E+01]),
                NASA([1398.00, 5000.00],
                    [ 2.20530468E+01,  3.40541230E-02, -1.16873054E-05,
                    1.81862696E-09, -1.05680270E-13, -4.41822326E+04,
                    -9.24677192E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.297,
                                well_depth=511.5,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14O1-5',
        atoms='H:14 C:7 O:1',
        thermo=(NASA([300.00, 1394.00],
                    [-7.14726345E+00,  1.02606685E-01, -7.12373028E-05,
                    2.52354809E-08, -3.63484271E-12, -3.62821855E+04,
                    5.93513033E+01]),
                NASA([1394.00, 5000.00],
                    [ 2.27681049E+01,  3.42097414E-02, -1.18172110E-05,
                    1.84695360E-09, -1.07658833E-13, -4.68159558E+04,
                    -1.01833212E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.297,
                                well_depth=511.5,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14O2-3',
        atoms='H:14 C:7 O:1',
        thermo=(NASA([300.00, 1397.00],
                    [-4.35014619E+00,  1.01435718E-01, -7.70541034E-05,
                    3.06059134E-08, -4.99101265E-12, -2.61218264E+04,
                    5.06557622E+01]),
                NASA([1397.00, 5000.00],
                    [ 2.44676947E+01,  3.13508938E-02, -1.07981336E-05,
                    1.68432141E-09, -9.80409528E-14, -3.58726267E+04,
                    -1.03164928E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.297,
                                well_depth=511.5,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14O2-4',
        atoms='H:14 C:7 O:1',
        thermo=(NASA([300.00, 1399.00],
                    [-5.39477206E+00,  1.01846577E-01, -7.60046236E-05,
                    2.96538020E-08, -4.74854370E-12, -2.67118295E+04,
                    5.51731319E+01]),
                NASA([1399.00, 5000.00],
                    [ 2.32692731E+01,  3.25585391E-02, -1.11624906E-05,
                    1.73573658E-09, -1.00813215E-13, -3.64349664E+04,
                    -9.79457278E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.297,
                                well_depth=511.5,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14O2-5',
        atoms='H:14 C:7 O:1',
        thermo=(NASA([300.00, 1399.00],
                    [-7.24295043E+00,  1.05687256E-01, -7.87952366E-05,
                    3.05899677E-08, -4.86103477E-12, -3.63891669E+04,
                    6.36525527E+01]),
                NASA([1399.00, 5000.00],
                    [ 2.26707471E+01,  3.34803352E-02, -1.14785395E-05,
                    1.78491428E-09, -1.03671828E-13, -4.65307741E+04,
                    -9.61517004E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.297,
                                well_depth=511.5,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14O2-6',
        atoms='H:14 C:7 O:1',
        thermo=(NASA([300.00, 1396.00],
                    [-7.82017858E+00,  1.07564943E-01, -7.89396005E-05,
                    2.98830915E-08, -4.62102018E-12, -3.84419001E+04,
                    6.16931674E+01]),
                NASA([1396.00, 5000.00],
                    [ 2.33833361E+01,  3.36382957E-02, -1.16092281E-05,
                    1.81335441E-09, -1.05656460E-13, -4.91623271E+04,
                    -1.05500759E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.297,
                                well_depth=511.5,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14O3-4',
        atoms='H:14 C:7 O:1',
        thermo=(NASA([300.00, 1397.00],
                    [-4.35014619E+00,  1.01435718E-01, -7.70541034E-05,
                    3.06059134E-08, -4.99101265E-12, -2.61218264E+04,
                    5.06557622E+01]),
                NASA([1397.00, 5000.00],
                    [ 2.44676947E+01,  3.13508938E-02, -1.07981336E-05,
                    1.68432141E-09, -9.80409528E-14, -3.58726267E+04,
                    -1.03164928E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.297,
                                well_depth=511.5,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14O3-5',
        atoms='H:14 C:7 O:1',
        thermo=(NASA([300.00, 1399.00],
                    [-5.39477206E+00,  1.01846577E-01, -7.60046236E-05,
                    2.96538020E-08, -4.74854370E-12, -2.67118295E+04,
                    5.51731319E+01]),
                NASA([1399.00, 5000.00],
                    [ 2.32692731E+01,  3.25585391E-02, -1.11624906E-05,
                    1.73573658E-09, -1.00813215E-13, -3.64349664E+04,
                    -9.79457278E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.297,
                                well_depth=511.5,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC7KET12',
        atoms='H:14 C:7 O:3',
        thermo=(NASA([300.00, 1387.00],
                    [-1.01240540E+00,  1.08525210E-01, -8.59990884E-05,
                    3.49272961E-08, -5.73789850E-12, -4.66314955E+04,
                    4.09175088E+01]),
                NASA([1387.00, 5000.00],
                    [ 3.14039860E+01,  2.85485198E-02, -9.69585695E-06,
                    1.50017009E-09, -8.68909491E-14, -5.74044717E+04,
                    -1.31576613E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.506,
                                well_depth=581.3,
                                dipole=2.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC7KET13',
        atoms='H:14 C:7 O:3',
        thermo=(NASA([300.00, 1388.00],
                    [ 2.21629328E+00,  9.47373762E-02, -6.90473651E-05,
                    2.67820400E-08, -4.36392181E-12, -4.87965510E+04,
                    2.63157133E+01]),
                NASA([1388.00, 5000.00],
                    [ 2.90744507E+01,  3.17177458E-02, -1.10262646E-05,
                    1.73053608E-09, -1.01161699E-13, -5.82345477E+04,
                    -1.18064324E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.506,
                                well_depth=581.3,
                                dipole=2.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC7KET14',
        atoms='H:14 C:7 O:3',
        thermo=(NASA([300.00, 1388.00],
                    [ 2.21629328E+00,  9.47373762E-02, -6.90473651E-05,
                    2.67820400E-08, -4.36392181E-12, -4.87965510E+04,
                    2.63157133E+01]),
                NASA([1388.00, 5000.00],
                    [ 2.90744507E+01,  3.17177458E-02, -1.10262646E-05,
                    1.73053608E-09, -1.01161699E-13, -5.82345477E+04,
                    -1.18064324E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.506,
                                well_depth=581.3,
                                dipole=2.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC7KET15',
        atoms='H:14 C:7 O:3',
        thermo=(NASA([300.00, 1388.00],
                    [ 2.21629328E+00,  9.47373762E-02, -6.90473651E-05,
                    2.67820400E-08, -4.36392181E-12, -4.87965510E+04,
                    2.63157133E+01]),
                NASA([1388.00, 5000.00],
                    [ 2.90744507E+01,  3.17177458E-02, -1.10262646E-05,
                    1.73053608E-09, -1.01161699E-13, -5.82345477E+04,
                    -1.18064324E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.506,
                                well_depth=581.3,
                                dipole=2.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC7KET21',
        atoms='H:14 C:7 O:3',
        thermo=(NASA([300.00, 1385.00],
                    [-9.64230115E-01,  1.08158991E-01, -8.67775775E-05,
                    3.61807459E-08, -6.14661796E-12, -4.82404885E+04,
                    4.13251144E+01]),
                NASA([1385.00, 5000.00],
                    [ 3.07082732E+01,  2.92465419E-02, -9.95807579E-06,
                    1.54315782E-09, -8.94718150E-14, -5.87615592E+04,
                    -1.27058904E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.506,
                                well_depth=581.3,
                                dipole=2.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC7KET23',
        atoms='H:14 C:7 O:3',
        thermo=(NASA([300.00, 1388.00],
                    [-1.45570671E+00,  1.08772384E-01, -8.59757426E-05,
                    3.51095315E-08, -5.83436499E-12, -5.02602030E+04,
                    4.28185431E+01]),
                NASA([1388.00, 5000.00],
                    [ 3.05981614E+01,  2.97258855E-02, -1.01972578E-05,
                    1.58732411E-09, -9.22942417E-14, -6.09638920E+04,
                    -1.27837876E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.506,
                                well_depth=581.3,
                                dipole=2.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC7KET24',
        atoms='H:14 C:7 O:3',
        thermo=(NASA([300.00, 1387.00],
                    [ 1.52936692E+00,  9.58173466E-02, -6.96688520E-05,
                    2.69540382E-08, -4.38728126E-12, -5.23839542E+04,
                    2.93851334E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.88330388E+01,  3.20168096E-02, -1.11508456E-05,
                    1.75226159E-09, -1.02520451E-13, -6.20142474E+04,
                    -1.17499704E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.506,
                                well_depth=581.3,
                                dipole=2.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC7KET25',
        atoms='H:14 C:7 O:3',
        thermo=(NASA([300.00, 1387.00],
                    [ 1.52936692E+00,  9.58173466E-02, -6.96688520E-05,
                    2.69540382E-08, -4.38728126E-12, -5.23839542E+04,
                    2.93851334E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.88330388E+01,  3.20168096E-02, -1.11508456E-05,
                    1.75226159E-09, -1.02520451E-13, -6.20142474E+04,
                    -1.17499704E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.506,
                                well_depth=581.3,
                                dipole=2.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC7KET26',
        atoms='H:14 C:7 O:3',
        thermo=(NASA([300.00, 1387.00],
                    [ 1.52936692E+00,  9.58173466E-02, -6.96688520E-05,
                    2.69540382E-08, -4.38728126E-12, -5.23839542E+04,
                    2.93851334E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.88330388E+01,  3.20168096E-02, -1.11508456E-05,
                    1.75226159E-09, -1.02520451E-13, -6.20142474E+04,
                    -1.17499704E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.506,
                                well_depth=581.3,
                                dipole=2.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC7KET31',
        atoms='H:14 C:7 O:3',
        thermo=(NASA([300.00, 1384.00],
                    [ 2.77522939E+00,  9.05308437E-02, -6.36476024E-05,
                    2.42107778E-08, -3.94769763E-12, -5.04144221E+04,
                    2.40464052E+01]),
                NASA([1384.00, 5000.00],
                    [ 2.80739929E+01,  3.26743194E-02, -1.13798765E-05,
                    1.78819290E-09, -1.04618326E-13, -5.95401535E+04,
                    -1.12628648E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.506,
                                well_depth=581.3,
                                dipole=2.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC7KET32',
        atoms='H:14 C:7 O:3',
        thermo=(NASA([300.00, 1385.00],
                    [-9.94514343E-01,  1.08762100E-01, -8.76137740E-05,
                    3.66478182E-08, -6.24244946E-12, -5.05819016E+04,
                    4.03445313E+01]),
                NASA([1385.00, 5000.00],
                    [ 3.09165578E+01,  2.91258044E-02, -9.92861588E-06,
                    1.53981152E-09, -8.93271341E-14, -6.11708347E+04,
                    -1.29264593E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.506,
                                well_depth=581.3,
                                dipole=2.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC7KET34',
        atoms='H:14 C:7 O:3',
        thermo=(NASA([300.00, 1385.00],
                    [-9.94514343E-01,  1.08762100E-01, -8.76137740E-05,
                    3.66478182E-08, -6.24244946E-12, -5.05819016E+04,
                    4.03445313E+01]),
                NASA([1385.00, 5000.00],
                    [ 3.09165578E+01,  2.91258044E-02, -9.92861588E-06,
                    1.53981152E-09, -8.93271341E-14, -6.11708347E+04,
                    -1.29264593E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.506,
                                well_depth=581.3,
                                dipole=2.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC7KET35',
        atoms='H:14 C:7 O:3',
        thermo=(NASA([300.00, 1387.00],
                    [ 1.52936692E+00,  9.58173466E-02, -6.96688520E-05,
                    2.69540382E-08, -4.38728126E-12, -5.23839542E+04,
                    2.93851334E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.88330388E+01,  3.20168096E-02, -1.11508456E-05,
                    1.75226159E-09, -1.02520451E-13, -6.20142474E+04,
                    -1.17499704E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.506,
                                well_depth=581.3,
                                dipole=2.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC7KET36',
        atoms='H:14 C:7 O:3',
        thermo=(NASA([300.00, 1387.00],
                    [ 1.52936692E+00,  9.58173466E-02, -6.96688520E-05,
                    2.69540382E-08, -4.38728126E-12, -5.23839542E+04,
                    2.93851334E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.88330388E+01,  3.20168096E-02, -1.11508456E-05,
                    1.75226159E-09, -1.02520451E-13, -6.20142474E+04,
                    -1.17499704E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.506,
                                well_depth=581.3,
                                dipole=2.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC7KET37',
        atoms='H:14 C:7 O:3',
        thermo=(NASA([300.00, 1384.00],
                    [ 2.77522939E+00,  9.05308437E-02, -6.36476024E-05,
                    2.42107778E-08, -3.94769763E-12, -5.04144221E+04,
                    2.40464052E+01]),
                NASA([1384.00, 5000.00],
                    [ 2.80739929E+01,  3.26743194E-02, -1.13798765E-05,
                    1.78819290E-09, -1.04618326E-13, -5.95401535E+04,
                    -1.12628648E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.506,
                                well_depth=581.3,
                                dipole=2.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC7KET41',
        atoms='H:14 C:7 O:3',
        thermo=(NASA([300.00, 1384.00],
                    [ 2.77522939E+00,  9.05308437E-02, -6.36476024E-05,
                    2.42107778E-08, -3.94769763E-12, -5.04144221E+04,
                    2.40464052E+01]),
                NASA([1384.00, 5000.00],
                    [ 2.80739929E+01,  3.26743194E-02, -1.13798765E-05,
                    1.78819290E-09, -1.04618326E-13, -5.95401535E+04,
                    -1.12628648E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.506,
                                well_depth=581.3,
                                dipole=2.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC7KET42',
        atoms='H:14 C:7 O:3',
        thermo=(NASA([300.00, 1386.00],
                    [ 2.32269738E+00,  9.45160149E-02, -6.99057179E-05,
                    2.79772916E-08, -4.74244431E-12, -5.27586830E+04,
                    2.53458269E+01]),
                NASA([1386.00, 5000.00],
                    [ 2.86572867E+01,  3.21567908E-02, -1.11961570E-05,
                    1.75895000E-09, -1.02891960E-13, -6.20252988E+04,
                    -1.16135256E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.506,
                                well_depth=581.3,
                                dipole=2.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='NC7KET43',
        atoms='H:14 C:7 O:3',
        thermo=(NASA([300.00, 1386.00],
                    [ 2.32269738E+00,  9.45160149E-02, -6.99057179E-05,
                    2.79772916E-08, -4.74244431E-12, -5.27586830E+04,
                    2.53458269E+01]),
                NASA([1386.00, 5000.00],
                    [ 2.86572867E+01,  3.21567908E-02, -1.11961570E-05,
                    1.75895000E-09, -1.02891960E-13, -6.20252988E+04,
                    -1.16135256E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.506,
                                well_depth=581.3,
                                dipole=2.0,
                                rot_relax=1.0),
        note='7/19/0THERM')

species(name='C7H14OH-1',
        atoms='H:15 C:7 O:1',
        thermo=(NASA([300.00, 1376.00],
                    [ 4.58848577E-01,  8.04446771E-02, -4.57034176E-05,
                    1.14685073E-08, -8.17742081E-13, -2.07641264E+04,
                    3.50408521E+01]),
                NASA([1376.00, 5000.00],
                    [ 2.41265675E+01,  3.16637088E-02, -1.05671043E-05,
                    1.61536314E-09, -9.27674815E-14, -2.95485867E+04,
                    -9.42377369E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='6/25/99THERM')

species(name='C7H14OH-2',
        atoms='H:15 C:7 O:1',
        thermo=(NASA([300.00, 1390.00],
                    [-6.17451781E-02,  8.52264839E-02, -5.42481745E-05,
                    1.73728776E-08, -2.20758857E-12, -2.29210338E+04,
                    3.65945984E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.38036547E+01,  3.25612587E-02, -1.09916297E-05,
                    1.69186772E-09, -9.75842145E-14, -3.14956023E+04,
                    -9.26354590E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='6/25/99THERM')

species(name='C7H14OH-3',
        atoms='H:15 C:7 O:1',
        thermo=(NASA([300.00, 1390.00],
                    [-6.17451781E-02,  8.52264839E-02, -5.42481745E-05,
                    1.73728776E-08, -2.20758857E-12, -2.29210338E+04,
                    3.65945984E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.38036547E+01,  3.25612587E-02, -1.09916297E-05,
                    1.69186772E-09, -9.75842145E-14, -3.14956023E+04,
                    -9.26354590E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='6/25/99THERM')

species(name='O2C7H14OH-1',
        atoms='H:15 C:7 O:3',
        thermo=(NASA([300.00, 1395.00],
                    [ 2.34060326E+00,  9.23428863E-02, -6.37138459E-05,
                    2.36026902E-08, -3.68902757E-12, -4.04140401E+04,
                    2.73239434E+01]),
                NASA([1395.00, 5000.00],
                    [ 2.69431812E+01,  3.51661203E-02, -1.20111248E-05,
                    1.86268617E-09, -1.07974911E-13, -4.90880404E+04,
                    -1.05078319E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149),
        note='6/25/99THRM')

species(name='O2C7H14OH-2',
        atoms='H:15 C:7 O:3',
        thermo=(NASA([300.00, 1396.00],
                    [ 1.43173891E+00,  9.81681859E-02, -7.19259300E-05,
                    2.82501348E-08, -4.63278447E-12, -4.25022305E+04,
                    3.07742185E+01]),
                NASA([1396.00, 5000.00],
                    [ 2.78711248E+01,  3.44298705E-02, -1.17690678E-05,
                    1.82617826E-09, -1.05901891E-13, -5.15690694E+04,
                    -1.10657303E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149),
        note='6/25/99THRM')

species(name='O2C7H14OH-3',
        atoms='H:15 C:7 O:3',
        thermo=(NASA([300.00, 1396.00],
                    [ 1.43173891E+00,  9.81681859E-02, -7.19259300E-05,
                    2.82501348E-08, -4.63278447E-12, -4.25022305E+04,
                    3.07742185E+01]),
                NASA([1396.00, 5000.00],
                    [ 2.78711248E+01,  3.44298705E-02, -1.17690678E-05,
                    1.82617826E-09, -1.05901891E-13, -5.15690694E+04,
                    -1.10657303E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149),
        note='6/25/99THRM')

species(name='NC5H11CHO',
        atoms='H:12 C:6 O:1',
        thermo=(NASA([300.00, 1384.00],
                    [ 1.37517192E+00,  6.65669689E-02, -4.04423050E-05,
                    1.23836270E-08, -1.52905857E-12, -3.28740986E+04,
                    2.48343934E+01]),
                NASA([1384.00, 5000.00],
                    [ 1.98891043E+01,  2.71869340E-02, -9.27391515E-06,
                    1.43744158E-09, -8.33090761E-14, -3.97523444E+04,
                    -7.60741671E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.009,
                                well_depth=498.6,
                                dipole=2.0,
                                rot_relax=1.0),
        note='2/29/96THERM')

species(name='NC5H11CO',
        atoms='H:11 C:6 O:1',
        thermo=(NASA([300.00, 1383.00],
                    [ 2.14479069E+00,  6.17863563E-02, -3.74134690E-05,
                    1.13283795E-08, -1.36917698E-12, -1.43451172E+04,
                    2.23128045E+01]),
                NASA([1383.00, 5000.00],
                    [ 1.94783812E+01,  2.50466029E-02, -8.54861346E-06,
                    1.32557944E-09, -7.68503296E-14, -2.07923937E+04,
                    -7.21995578E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.009,
                                well_depth=498.6,
                                dipole=2.0,
                                rot_relax=1.0),
        note='2/29/96THERM')

species(name='C5H10CHO-1',
        atoms='H:11 C:6 O:1',
        thermo=(NASA([300.00, 1383.00],
                    [ 2.15377345E+00,  6.19684203E-02, -3.78904723E-05,
                    1.17022892E-08, -1.46276094E-12, -8.26278569E+03,
                    2.33779159E+01]),
                NASA([1383.00, 5000.00],
                    [ 1.94701427E+01,  2.50490154E-02, -8.54838759E-06,
                    1.32543025E-09, -7.68369090E-14, -1.46878327E+04,
                    -7.09736169E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.625,
                                well_depth=534.323),
        note='2/29/96THERM')

species(name='C5H10CHO-2',
        atoms='H:11 C:6 O:1',
        thermo=(NASA([300.00, 1371.00],
                    [ 4.16999842E-01,  6.37354262E-02, -3.53661717E-05,
                    7.90793428E-09, -2.83991537E-13, -1.00633239E+04,
                    3.30278698E+01]),
                NASA([1371.00, 5000.00],
                    [ 2.01227062E+01,  2.38797102E-02, -8.02929509E-06,
                    1.23389172E-09, -7.11284276E-14, -1.74161761E+04,
                    -7.48168825E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.625,
                                well_depth=534.323),
        note='2/29/96THERM')

species(name='C5H10CHO-3',
        atoms='H:11 C:6 O:1',
        thermo=(NASA([300.00, 1371.00],
                    [ 4.16999842E-01,  6.37354262E-02, -3.53661717E-05,
                    7.90793428E-09, -2.83991537E-13, -1.00633239E+04,
                    3.30278698E+01]),
                NASA([1371.00, 5000.00],
                    [ 2.01227062E+01,  2.38797102E-02, -8.02929509E-06,
                    1.23389172E-09, -7.11284276E-14, -1.74161761E+04,
                    -7.48168825E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.625,
                                well_depth=534.323),
        note='2/29/96THERM')

species(name='C5H10CHO-4',
        atoms='H:11 C:6 O:1',
        thermo=(NASA([300.00, 1371.00],
                    [ 4.16999842E-01,  6.37354262E-02, -3.53661717E-05,
                    7.90793428E-09, -2.83991537E-13, -1.00633239E+04,
                    3.30278698E+01]),
                NASA([1371.00, 5000.00],
                    [ 2.01227062E+01,  2.38797102E-02, -8.02929509E-06,
                    1.23389172E-09, -7.11284276E-14, -1.74161761E+04,
                    -7.48168825E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.625,
                                well_depth=534.323),
        note='2/29/96THERM')

species(name='C5H10CHO-5',
        atoms='H:11 C:6 O:1',
        thermo=(NASA([300.00, 1395.00],
                    [-1.26556708E+00,  7.36551339E-02, -5.04263913E-05,
                    1.74734247E-08, -2.44116609E-12, -1.31838221E+04,
                    3.68187516E+01]),
                NASA([1395.00, 5000.00],
                    [ 2.02669246E+01,  2.47914857E-02, -8.54953747E-06,
                    1.33477274E-09, -7.77453379E-14, -2.07889535E+04,
                    -7.93060747E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.625,
                                well_depth=534.323),
        note='2/29/96THERM')

species(name='C4H7CHO1-4',
        atoms='H:8 C:5 O:1',
        thermo=(NASA([300.00, 1379.00],
                    [ 1.30688261E+00,  5.07911858E-02, -3.31509980E-05,
                    1.10391295E-08, -1.49578391E-12, -1.45595753E+04,
                    2.40234219E+01]),
                NASA([1379.00, 5000.00],
                    [ 1.59299730E+01,  1.84432887E-02, -6.21997362E-06,
                    9.57537306E-10, -5.52580126E-14, -1.98311071E+04,
                    -5.51771737E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.778,
                                well_depth=476.0,
                                dipole=2.6,
                                rot_relax=1.0),
        note='5/6/96THERM')

species(name='C4H7CO1-4',
        atoms='H:7 C:5 O:1',
        thermo=(NASA([300.00, 1378.00],
                    [ 2.09347005E+00,  4.59484008E-02, -3.00651488E-05,
                    9.99306335E-09, -1.34709811E-12,  3.96659370E+03,
                    2.14210959E+01]),
                NASA([1378.00, 5000.00],
                    [ 1.54637001E+01,  1.64158685E-02, -5.54672481E-06,
                    8.55055295E-10, -4.93928750E-14, -8.55858619E+02,
                    -5.10065564E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.778,
                                well_depth=476.0,
                                dipole=2.6,
                                rot_relax=1.0),
        note='5/6/96THERM')

species(name='C4H6CHO1-43',
        atoms='H:7 C:5 O:1',
        thermo=(NASA([300.00, 1382.00],
                    [ 1.13728412E+00,  4.85677415E-02, -3.21395914E-05,
                    1.08714802E-08, -1.51624448E-12,  2.44080824E+03,
                    2.36917499E+01]),
                NASA([1382.00, 5000.00],
                    [ 1.54178855E+01,  1.71672536E-02, -5.95214751E-06,
                    9.32713368E-10, -5.44705237E-14, -2.75958461E+03,
                    -5.37764772E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.778,
                                well_depth=476.0,
                                dipole=2.6,
                                rot_relax=1.0),
        note='5/6/96THRM')

species(name='C4H6CHO1-44',
        atoms='H:7 C:5 O:1',
        thermo=(NASA([300.00, 1390.00],
                    [ 7.50037894E-01,  5.14805714E-02, -3.71071056E-05,
                    1.40009579E-08, -2.19006703E-12,  5.63542979E+03,
                    2.59488465E+01]),
                NASA([1390.00, 5000.00],
                    [ 1.54721595E+01,  1.70164879E-02, -5.87513255E-06,
                    9.17941359E-10, -5.34945363E-14,  4.99122207E+02,
                    -5.31445505E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.778,
                                well_depth=476.0,
                                dipole=2.6,
                                rot_relax=1.0),
        note='5/6/96THRM')

species(name='NC4H9COCH3',
        atoms='H:12 C:6 O:1',
        thermo=(NASA([300.00, 1384.00],
                    [ 1.04489132E+00,  6.62869241E-02, -3.94849702E-05,
                    1.19083422E-08, -1.47157105E-12, -3.65200950E+04,
                    2.62126637E+01]),
                NASA([1384.00, 5000.00],
                    [ 1.93396071E+01,  2.80294129E-02, -9.64367685E-06,
                    1.50309259E-09, -8.74431874E-14, -4.34308987E+04,
                    -7.38144199E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.02,
                                well_depth=503.7,
                                dipole=2.0,
                                rot_relax=1.0),
        note='2/22/96THERM')

species(name='NC4H9COCH2',
        atoms='H:11 C:6 O:1',
        thermo=(NASA([300.00, 1387.00],
                    [ 6.57240675E-01,  6.91820222E-02, -4.79956182E-05,
                    1.78114857E-08, -2.80456388E-12, -1.51806661E+04,
                    2.65180216E+01]),
                NASA([1387.00, 5000.00],
                    [ 1.98575871E+01,  2.52406370E-02, -8.72522623E-06,
                    1.36422047E-09, -7.95374883E-14, -2.20531958E+04,
                    -7.71140999E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.625,
                                well_depth=534.323),
        note='2/22/96THERM')

species(name='C4H8COCH3-1',
        atoms='H:11 C:6 O:1',
        thermo=(NASA([300.00, 1384.00],
                    [ 1.83079991E+00,  6.16869509E-02, -3.69733811E-05,
                    1.12792095E-08, -1.42170745E-12, -1.19106373E+04,
                    2.47213709E+01]),
                NASA([1384.00, 5000.00],
                    [ 1.89006813E+01,  2.59288818E-02, -8.93525037E-06,
                    1.39416390E-09, -8.11664734E-14, -1.83595544E+04,
                    -6.85988070E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.625,
                                well_depth=534.323),
        note='2/22/96THRM')

species(name='C4H8COCH3-2',
        atoms='H:11 C:6 O:1',
        thermo=(NASA([300.00, 1365.00],
                    [ 2.32442247E+00,  5.60360489E-02, -2.62897892E-05,
                    3.55349257E-09,  4.62560403E-13, -1.32203258E+04,
                    2.36967419E+01]),
                NASA([1365.00, 5000.00],
                    [ 1.92472410E+01,  2.47394219E-02, -8.34917480E-06,
                    1.28606717E-09, -7.42515416E-14, -1.98964461E+04,
                    -7.00732125E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.625,
                                well_depth=534.323),
        note='2/22/96THRM')

species(name='C4H8COCH3-3',
        atoms='H:11 C:6 O:1',
        thermo=(NASA([300.00, 1365.00],
                    [ 2.32442247E+00,  5.60360489E-02, -2.62897892E-05,
                    3.55349257E-09,  4.62560403E-13, -1.32203258E+04,
                    2.36967419E+01]),
                NASA([1365.00, 5000.00],
                    [ 1.92472410E+01,  2.47394219E-02, -8.34917480E-06,
                    1.28606717E-09, -7.42515416E-14, -1.98964461E+04,
                    -7.00732125E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.625,
                                well_depth=534.323),
        note='2/22/96THRM')

species(name='C4H8COCH3-4',
        atoms='H:11 C:6 O:1',
        thermo=(NASA([300.00, 1387.00],
                    [ 4.64936924E-01,  6.68009275E-02, -4.26203243E-05,
                    1.39372398E-08, -1.88407814E-12, -1.63159454E+04,
                    2.82932537E+01]),
                NASA([1387.00, 5000.00],
                    [ 1.94544114E+01,  2.56278187E-02, -8.86916402E-06,
                    1.38780777E-09, -8.09574343E-14, -2.33097365E+04,
                    -7.49578200E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.625,
                                well_depth=534.323),
        note='2/22/96THRM')

species(name='C4H7OOH1-4',
        atoms='H:8 C:4 O:2',
        thermo=(NASA([300.00, 1393.00],
                    [ 1.79804670E+00,  5.23302551E-02, -3.64638246E-05,
                    1.31971888E-08, -1.97165993E-12, -1.32846748E+04,
                    2.25862038E+01]),
                NASA([1393.00, 5000.00],
                    [ 1.66978343E+01,  1.80578399E-02, -6.22476839E-06,
                    9.71479434E-10, -5.65686551E-14, -1.85344720E+04,
                    -5.76609547E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='4/1/96THERM')

species(name='C4H7O1-4',
        atoms='H:7 C:4 O:1',
        thermo=(NASA([300.00, 1393.00],
                    [-2.92647947E-01,  4.83149119E-02, -3.48133220E-05,
                    1.31921468E-08, -2.06485553E-12,  6.29978883E+03,
                    3.03629978E+01]),
                NASA([1393.00, 5000.00],
                    [ 1.31417783E+01,  1.64166051E-02, -5.56949199E-06,
                    8.60114769E-10, -4.97230006E-14,  1.67967537E+03,
                    -4.16160087E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.778,
                                well_depth=476.0,
                                dipole=2.6,
                                rot_relax=1.0),
        note='4/1/96THERM')

species(name='C5H9OOH1-4',
        atoms='H:10 C:5 O:2',
        thermo=(NASA([300.00, 1397.00],
                    [ 7.67782330E-01,  6.93519317E-02, -5.21650213E-05,
                    2.06237355E-08, -3.35844189E-12, -1.78318101E+04,
                    2.80373555E+01]),
                NASA([1397.00, 5000.00],
                    [ 2.02179950E+01,  2.21511744E-02, -7.60711244E-06,
                    1.18420679E-09, -6.88331594E-14, -2.44297406E+04,
                    -7.58287543E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.68,
                                well_depth=543.585),
        note='3/27/97THERM')

species(name='C5H9O1-4',
        atoms='H:9 C:5 O:1',
        thermo=(NASA([300.00, 1398.00],
                    [-8.53018802E-01,  6.41880432E-02, -4.91931508E-05,
                    2.00188214E-08, -3.35963237E-12,  1.19509504E+03,
                    3.38333616E+01]),
                NASA([1398.00, 5000.00],
                    [ 1.66809382E+01,  2.07591828E-02, -7.09289768E-06,
                    1.10031617E-09, -6.37992567E-14, -4.67568455E+03,
                    -5.95056483E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.348,
                                well_depth=489.084),
        note='3/28/97THERM')

species(name='C5H9OOH1-5',
        atoms='H:10 C:5 O:2',
        thermo=(NASA([300.00, 1393.00],
                    [ 1.65166493E+00,  6.41855434E-02, -4.42152137E-05,
                    1.58117718E-08, -2.33469388E-12, -1.61856012E+04,
                    2.49485316E+01]),
                NASA([1393.00, 5000.00],
                    [ 1.98575596E+01,  2.25572969E-02, -7.76951451E-06,
                    1.21189491E-09, -7.05404248E-14, -2.26275637E+04,
                    -7.31968346E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.68,
                                well_depth=543.585),
        note='3/27/97THERM')

species(name='C5H9O1-5',
        atoms='H:9 C:5 O:1',
        thermo=(NASA([300.00, 1394.00],
                    [-6.08379884E-01,  6.09139191E-02, -4.36395120E-05,
                    1.64966933E-08, -2.58694444E-12,  3.42450558E+03,
                    3.35104649E+01]),
                NASA([1394.00, 5000.00],
                    [ 1.62564642E+01,  2.10561226E-02, -7.18323965E-06,
                    1.11331434E-09, -6.45162851E-14, -2.41003213E+03,
                    -5.69404193E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.348,
                                well_depth=489.084),
        note='3/28/97THERM')

species(name='C6H11OOH1-4',
        atoms='H:12 C:6 O:2',
        thermo=(NASA([300.00, 1397.00],
                    [ 6.08693415E-01,  8.12533376E-02, -5.99544347E-05,
                    2.32306975E-08, -3.71302106E-12, -2.07306507E+04,
                    3.04600742E+01]),
                NASA([1397.00, 5000.00],
                    [ 2.33879230E+01,  2.66182814E-02, -9.13553672E-06,
                    1.42154300E-09, -8.26043195E-14, -2.85211117E+04,
                    -9.14098376E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='3/27/97THRM')

species(name='C6H11O1-4',
        atoms='H:11 C:6 O:1',
        thermo=(NASA([300.00, 1397.00],
                    [-1.17611678E+00,  7.67820021E-02, -5.79189808E-05,
                    2.31283510E-08, -3.81018248E-12, -1.67830272E+03,
                    3.70214234E+01]),
                NASA([1397.00, 5000.00],
                    [ 1.99304018E+01,  2.50878998E-02, -8.55881248E-06,
                    1.32642733E-09, -7.68596330E-14, -8.79248519E+03,
                    -7.55242129E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.674,
                                well_depth=541.5,
                                dipole=1.8,
                                rot_relax=1.0),
        note='3/28/97THERM')

species(name='C6H11OOH1-5',
        atoms='H:12 C:6 O:2',
        thermo=(NASA([300.00, 1397.00],
                    [ 6.08693415E-01,  8.12533376E-02, -5.99544347E-05,
                    2.32306975E-08, -3.71302106E-12, -2.07306507E+04,
                    3.04600742E+01]),
                NASA([1397.00, 5000.00],
                    [ 2.33879230E+01,  2.66182814E-02, -9.13553672E-06,
                    1.42154300E-09, -8.26043195E-14, -2.85211117E+04,
                    -9.14098376E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.317,
                                well_depth=561.0,
                                dipole=1.7,
                                rot_relax=1.0),
        note='3/27/97THRM')

species(name='C6H11O1-5',
        atoms='H:11 C:6 O:1',
        thermo=(NASA([300.00, 1397.00],
                    [-1.17611678E+00,  7.67820021E-02, -5.79189808E-05,
                    2.31283510E-08, -3.81018248E-12, -1.67830272E+03,
                    3.70214234E+01]),
                NASA([1397.00, 5000.00],
                    [ 1.99304018E+01,  2.50878998E-02, -8.55881248E-06,
                    1.32642733E-09, -7.68596330E-14, -8.79248519E+03,
                    -7.55242129E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.674,
                                well_depth=541.5,
                                dipole=1.8,
                                rot_relax=1.0),
        note='3/28/97THERM')

species(name='C5H91-1',
        atoms='H:9 C:5',
        thermo=(NASA([300.00, 1393.00],
                    [-1.68848603E-01,  5.40218667E-02, -3.74233337E-05,
                    1.37866431E-08, -2.12590460E-12,  2.51604761E+04,
                    2.88873177E+01]),
                NASA([1393.00, 5000.00],
                    [ 1.44574345E+01,  2.00072707E-02, -6.81378895E-06,
                    1.05474728E-09, -6.10665440E-14,  2.00271156E+04,
                    -4.97833624E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.458,
                                well_depth=396.8,
                                rot_relax=1.0),
        note='5/14/3THERM')

species(name='C4H7CHO1-1',
        atoms='H:8 C:5 O:1',
        thermo=(NASA([300.00, 1393.00],
                    [-8.96311479E-01,  6.16639930E-02, -4.81456435E-05,
                    1.98018086E-08, -3.35010063E-12, -1.86685108E+04,
                    3.18639054E+01]),
                NASA([1393.00, 5000.00],
                    [ 1.65147810E+01,  1.86176385E-02, -6.41423546E-06,
                    1.00074444E-09, -5.82623258E-14, -2.45132738E+04,
                    -6.08626797E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.778,
                                well_depth=476.0,
                                dipole=2.6,
                                rot_relax=1.0),
        note='2/22/96THERM')

species(name='C4H7CO1-1',
        atoms='H:7 C:5 O:1',
        thermo=(NASA([300.00, 1394.00],
                    [ 4.61430370E-01,  5.62940371E-02, -4.61644186E-05,
                    2.00772412E-08, -3.57227308E-12, -3.30150514E+03,
                    2.49394919E+01]),
                NASA([1394.00, 5000.00],
                    [ 1.59313061E+01,  1.65161742E-02, -5.67719248E-06,
                    8.84327967E-10, -5.14250392E-14, -8.36367262E+03,
                    -5.69425937E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.778,
                                well_depth=476.0,
                                dipole=2.6,
                                rot_relax=1.0),
        note='2/22/96THERM')

species(name='C4H6CHO1-14',
        atoms='H:7 C:5 O:1',
        thermo=(NASA([300.00, 1393.00],
                    [-9.37897914E-02,  5.69789312E-02, -4.55153233E-05,
                    1.91029934E-08, -3.28540832E-12,  5.93885740E+03,
                    3.02936035E+01]),
                NASA([1393.00, 5000.00],
                    [ 1.60679011E+01,  1.65205734E-02, -5.70629506E-06,
                    8.91819150E-10, -5.19829213E-14,  5.61658536E+02,
                    -5.56047479E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.306,
                                well_depth=482.331),
        note='2/22/96TRM')

species(name='C4H6CHO1-13',
        atoms='H:7 C:5 O:1',
        thermo=(NASA([300.00, 1388.00],
                    [-8.68106067E-01,  5.81156028E-02, -4.44075851E-05,
                    1.74858999E-08, -2.82132259E-12, -1.68831658E+03,
                    3.06969623E+01]),
                NASA([1388.00, 5000.00],
                    [ 1.65957737E+01,  1.64087623E-02, -5.74432894E-06,
                    9.05904509E-10, -5.31381430E-14, -7.68324332E+03,
                    -6.28091961E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.306,
                                well_depth=482.331),
        note='2/22/96TRM')

species(name='NC3H7COC2H5',
        atoms='H:12 C:6 O:1',
        thermo=(NASA([300.00, 1378.00],
                    [ 1.28300489E+00,  6.70849881E-02, -4.19984982E-05,
                    1.36804265E-08, -1.86987155E-12, -3.68039430E+04,
                    2.48054841E+01]),
                NASA([1378.00, 5000.00],
                    [ 1.99627893E+01,  2.68078451E-02, -9.08527269E-06,
                    1.40288330E-09, -8.11176345E-14, -4.37215719E+04,
                    -7.68702768E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.643,
                                well_depth=537.467),
        note='2/22/96THRM')

species(name='C3H6COC2H5-1',
        atoms='H:11 C:6 O:1',
        thermo=(NASA([300.00, 1380.00],
                    [ 2.08958703E+00,  6.24878115E-02, -3.97257078E-05,
                    1.33752941E-08, -1.92193795E-12, -1.21992557E+04,
                    2.32035207E+01]),
                NASA([1380.00, 5000.00],
                    [ 1.92391148E+01,  2.51659147E-02, -8.57508083E-06,
                    1.32847255E-09, -7.69774808E-14, -1.85404890E+04,
                    -7.00617026E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.625,
                                well_depth=534.323),
        note='2/22/96TRM')

species(name='C3H6COC2H5-2',
        atoms='H:11 C:6 O:1',
        thermo=(NASA([300.00, 1369.00],
                    [ 2.83974334E+00,  5.59471569E-02, -2.81954889E-05,
                    5.56103221E-09, -1.03163981E-13, -1.35531569E+04,
                    2.09458274E+01]),
                NASA([1369.00, 5000.00],
                    [ 1.91083648E+01,  2.49004438E-02, -8.41246692E-06,
                    1.29661383E-09, -7.48882792E-14, -1.99334270E+04,
                    -6.89565390E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.625,
                                well_depth=534.323),
        note='2/22/96TRM')

species(name='C3H6COC2H5-3',
        atoms='H:11 C:6 O:1',
        thermo=(NASA([300.00, 1386.00],
                    [ 8.23830406E-01,  6.74626897E-02, -4.57147359E-05,
                    1.66707586E-08, -2.60416740E-12, -1.66263615E+04,
                    2.62565174E+01]),
                NASA([1386.00, 5000.00],
                    [ 1.93409644E+01,  2.56927090E-02, -8.88365538E-06,
                    1.38917272E-09, -8.09982255E-14, -2.33416177E+04,
                    -7.39480531E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.625,
                                well_depth=534.323),
        note='2/22/96TRM')

species(name='NC3H7COC2H4P',
        atoms='H:11 C:6 O:1',
        thermo=(NASA([300.00, 1380.00],
                    [ 2.08958703E+00,  6.24878115E-02, -3.97257078E-05,
                    1.33752941E-08, -1.92193795E-12, -1.21992557E+04,
                    2.32035207E+01]),
                NASA([1380.00, 5000.00],
                    [ 1.92391148E+01,  2.51659147E-02, -8.57508083E-06,
                    1.32847255E-09, -7.69774808E-14, -1.85404890E+04,
                    -7.00617026E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.625,
                                well_depth=534.323),
        note='2/22/96TRM')

species(name='NC3H7COC2H4S',
        atoms='H:11 C:6 O:1',
        thermo=(NASA([300.00, 1386.00],
                    [ 8.23830406E-01,  6.74626897E-02, -4.57147359E-05,
                    1.66707586E-08, -2.60416740E-12, -1.66263615E+04,
                    2.62565174E+01]),
                NASA([1386.00, 5000.00],
                    [ 1.93409644E+01,  2.56927090E-02, -8.88365538E-06,
                    1.38917272E-09, -8.09982255E-14, -2.33416177E+04,
                    -7.39480531E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.625,
                                well_depth=534.323),
        note='2/22/96TRM')

species(name='CHCHCHO',
        atoms='H:3 C:3 O:1',
        thermo=(NASA([300.00, 1394.00],
                    [ 1.20905268E+00,  3.19125216E-02, -2.93443761E-05,
                    1.38077482E-08, -2.58328113E-12,  1.79711576E+04,
                    1.93989644E+01]),
                NASA([1394.00, 5000.00],
                    [ 1.02666402E+01,  7.09612931E-03, -2.47272731E-06,
                    3.88739735E-10, -2.27523413E-14,  1.51653800E+04,
                    -2.79927436E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=4.12,
                                well_depth=443.2,
                                rot_relax=1.0),
        note='5/14/3')

species(name='C6H101-3',
        atoms='H:10 C:6',
        thermo=(NASA([300.00, 1396.00],
                    [-2.38507596E+00,  7.30622234E-02, -5.86808335E-05,
                    2.52424911E-08, -4.46820100E-12,  4.80385112E+03,
                    3.76104460E+01]),
                NASA([1396.00, 5000.00],
                    [ 1.72274136E+01,  2.28968868E-02, -7.84176920E-06,
                    1.21837034E-09, -7.07190802E-14, -1.64978530E+03,
                    -6.63069861E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.51,
                                well_depth=426.3,
                                rot_relax=1.0))

species(name='C6H101-4',
        atoms='H:10 C:6',
        thermo=(NASA([300.00, 1396.00],
                    [-2.38507596E+00,  7.30622234E-02, -5.86808335E-05,
                    2.52424911E-08, -4.46820100E-12,  4.80385112E+03,
                    3.76104460E+01]),
                NASA([1396.00, 5000.00],
                    [ 1.72274136E+01,  2.28968868E-02, -7.84176920E-06,
                    1.21837034E-09, -7.07190802E-14, -1.64978530E+03,
                    -6.63069861E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.51,
                                well_depth=426.3,
                                rot_relax=1.0))

species(name='C6H101-5',
        atoms='H:10 C:6',
        thermo=(NASA([300.00, 1396.00],
                    [-2.38507596E+00,  7.30622234E-02, -5.86808335E-05,
                    2.52424911E-08, -4.46820100E-12,  4.80385112E+03,
                    3.76104460E+01]),
                NASA([1396.00, 5000.00],
                    [ 1.72274136E+01,  2.28968868E-02, -7.84176920E-06,
                    1.21837034E-09, -7.07190802E-14, -1.64978530E+03,
                    -6.63069861E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.51,
                                well_depth=426.3,
                                rot_relax=1.0))

species(name='C6H102-4',
        atoms='H:10 C:6',
        thermo=(NASA([300.00, 1396.00],
                    [-2.38507596E+00,  7.30622234E-02, -5.86808335E-05,
                    2.52424911E-08, -4.46820100E-12,  4.80385112E+03,
                    3.76104460E+01]),
                NASA([1396.00, 5000.00],
                    [ 1.72274136E+01,  2.28968868E-02, -7.84176920E-06,
                    1.21837034E-09, -7.07190802E-14, -1.64978530E+03,
                    -6.63069861E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.51,
                                well_depth=426.3,
                                rot_relax=1.0))

species(name='ETES1',
        atoms='H:10 C:6 O:1',
        thermo=(NASA([300.00, 1401.00],
                    [-5.96655202E+00,  8.90553981E-02, -7.35045234E-05,
                    3.14932080E-08, -5.45546649E-12, -7.32232871E+03,
                    5.55446829E+01]),
                NASA([1401.00, 5000.00],
                    [ 1.91884587E+01,  2.42677667E-02, -8.33134636E-06,
                    1.29673623E-09, -7.53674415E-14, -1.54455993E+04,
                    -7.73983087E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.643,
                                well_depth=537.467))

species(name='MVOX',
        atoms='H:10 C:6 O:1',
        thermo=(NASA([300.00, 1401.00],
                    [-5.96655202E+00,  8.90553981E-02, -7.35045234E-05,
                    3.14932080E-08, -5.45546649E-12, -7.32232871E+03,
                    5.55446829E+01]),
                NASA([1401.00, 5000.00],
                    [ 1.91884587E+01,  2.42677667E-02, -8.33134636E-06,
                    1.29673623E-09, -7.53674415E-14, -1.54455993E+04,
                    -7.73983087E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.643,
                                well_depth=537.467))

species(name='VTHF',
        atoms='H:10 C:6 O:1',
        thermo=(NASA([300.00, 1401.00],
                    [-5.96655202E+00,  8.90553981E-02, -7.35045234E-05,
                    3.14932080E-08, -5.45546649E-12, -7.32232871E+03,
                    5.55446829E+01]),
                NASA([1401.00, 5000.00],
                    [ 1.91884587E+01,  2.42677667E-02, -8.33134636E-06,
                    1.29673623E-09, -7.53674415E-14, -1.54455993E+04,
                    -7.73983087E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.643,
                                well_depth=537.467))

species(name='EDHF',
        atoms='H:10 C:6 O:1',
        thermo=(NASA([300.00, 1401.00],
                    [-5.96655202E+00,  8.90553981E-02, -7.35045234E-05,
                    3.14932080E-08, -5.45546649E-12, -7.32232871E+03,
                    5.55446829E+01]),
                NASA([1401.00, 5000.00],
                    [ 1.91884587E+01,  2.42677667E-02, -8.33134636E-06,
                    1.29673623E-09, -7.53674415E-14, -1.54455993E+04,
                    -7.73983087E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.643,
                                well_depth=537.467))

species(name='ETC6',
        atoms='H:12 C:6 O:1',
        thermo=(NASA([300.00, 1394.00],
                    [-3.56246496E+00,  8.55235442E-02, -6.41582453E-05,
                    2.54389942E-08, -4.17794820E-12, -2.05700039E+04,
                    4.59025070E+01]),
                NASA([1394.00, 5000.00],
                    [ 2.03993533E+01,  2.76439546E-02, -9.53039676E-06,
                    1.48744527E-09, -8.66140106E-14, -2.87586050E+04,
                    -8.22078661E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.643,
                                well_depth=537.467))

species(name='KES1',
        atoms='H:10 C:6 O:1',
        thermo=(NASA([300.00, 1388.00],
                    [ 4.46467269E-01,  6.64586762E-02, -4.70073659E-05,
                    1.79961137E-08, -2.93748963E-12, -2.30628492E+04,
                    2.77541300E+01]),
                NASA([1388.00, 5000.00],
                    [ 1.86909693E+01,  2.41939465E-02, -8.38233377E-06,
                    1.31252224E-09, -7.65989643E-14, -2.95692017E+04,
                    -7.05828250E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.643,
                                well_depth=537.467))

species(name='C6H111O2H-3',
        atoms='H:12 C:6 O:2',
        thermo=(NASA([300.00, 1392.00],
                    [-7.97246592E-01,  8.84990873E-02, -6.99517134E-05,
                    2.92202963E-08, -5.03185345E-12, -1.95195453E+04,
                    3.59633634E+01]),
                NASA([1392.00, 5000.00],
                    [ 2.42792322E+01,  2.63321024E-02, -9.14194501E-06,
                    1.43356642E-09, -8.37527455E-14, -2.79597127E+04,
                    -9.75927336E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='C6H111O2H-4',
        atoms='H:12 C:6 O:2',
        thermo=(NASA([300.00, 1392.00],
                    [-6.71334306E-01,  8.72726302E-02, -6.78106371E-05,
                    2.77977889E-08, -4.70536630E-12, -2.00372603E+04,
                    3.57689620E+01]),
                NASA([1392.00, 5000.00],
                    [ 2.41598315E+01,  2.64141743E-02, -9.16604821E-06,
                    1.43689654E-09, -8.39293545E-14, -2.84539273E+04,
                    -9.67105334E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='C6H111O2H-5',
        atoms='H:12 C:6 O:2',
        thermo=(NASA([300.00, 1392.00],
                    [-6.71334306E-01,  8.72726302E-02, -6.78106371E-05,
                    2.77977889E-08, -4.70536630E-12, -2.00372603E+04,
                    3.57689620E+01]),
                NASA([1392.00, 5000.00],
                    [ 2.41598315E+01,  2.64141743E-02, -9.16604821E-06,
                    1.43689654E-09, -8.39293545E-14, -2.84539273E+04,
                    -9.67105334E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='C6H111O2H-6',
        atoms='H:12 C:6 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 3.11486807E-01,  8.10135140E-02, -5.85042940E-05,
                    2.22954450E-08, -3.54988216E-12, -1.80250145E+04,
                    3.27265962E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.34008570E+01,  2.70661621E-02, -9.39242872E-06,
                    1.47237442E-09, -8.60004431E-14, -2.61302043E+04,
                    -9.14219690E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='C6H112O2H-4',
        atoms='H:12 C:6 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [-7.53695219E-01,  8.73573674E-02, -6.83958497E-05,
                    2.84775641E-08, -4.91362774E-12, -2.09736202E+04,
                    3.55730105E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.38605691E+01,  2.67263131E-02, -9.28634310E-06,
                    1.45697133E-09, -8.51501903E-14, -2.93238663E+04,
                    -9.57046301E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='C6H112O2H-5',
        atoms='H:12 C:6 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [-9.77786271E-02,  8.36152385E-02, -6.24511953E-05,
                    2.46774273E-08, -4.05915948E-12, -2.15666972E+04,
                    3.29610143E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.36804698E+01,  2.68319366E-02, -9.31237637E-06,
                    1.45995807E-09, -8.52807436E-14, -2.97893970E+04,
                    -9.44537469E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='C6H112O2H-6',
        atoms='H:12 C:6 O:2',
        thermo=(NASA([300.00, 1388.00],
                    [ 8.02060614E-01,  7.77041021E-02, -5.36127485E-05,
                    1.94323872E-08, -2.95453196E-12, -1.95415277E+04,
                    3.03063194E+01]),
                NASA([1388.00, 5000.00],
                    [ 2.29488499E+01,  2.74438680E-02, -9.52133333E-06,
                    1.49235013E-09, -8.71576434E-14, -2.74756907E+04,
                    -8.93200393E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='C6H113O2H-2',
        atoms='H:12 C:6 O:2',
        thermo=(NASA([300.00, 1391.00],
                    [-1.26739750E+00,  8.91469732E-02, -7.04153407E-05,
                    2.94208766E-08, -5.07121703E-12, -2.06975087E+04,
                    3.81839670E+01]),
                NASA([1391.00, 5000.00],
                    [ 2.39772718E+01,  2.66035589E-02, -9.23895225E-06,
                    1.44906345E-09, -8.46695750E-14, -2.92031812E+04,
                    -9.62904636E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='C6H113O2H-1',
        atoms='H:12 C:6 O:2',
        thermo=(NASA([300.00, 1387.00],
                    [-1.31133949E-01,  8.13310493E-02, -5.80494135E-05,
                    2.16450938E-08, -3.36477854E-12, -1.92042452E+04,
                    3.48562905E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.36308952E+01,  2.66617050E-02, -9.30222928E-06,
                    1.46355920E-09, -8.57041772E-14, -2.76274166E+04,
                    -9.32063164E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='NC6D1KET34',
        atoms='H:10 C:6 O:3',
        thermo=(NASA([300.00, 1390.00],
                    [-8.79937693E-01,  9.06618501E-02, -7.60486297E-05,
                    3.30211493E-08, -5.81958510E-12, -3.47639182E+04,
                    3.78353180E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.57983094E+01,  2.28116397E-02, -8.00057802E-06,
                    1.26322558E-09, -7.41574060E-14, -4.35413400E+04,
                    -1.03594800E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0))

species(name='NC6D1KET35',
        atoms='H:10 C:6 O:3',
        thermo=(NASA([300.00, 1388.00],
                    [ 8.59032690E-01,  8.49726292E-02, -7.03646845E-05,
                    3.07049154E-08, -5.49124822E-12, -3.58150700E+04,
                    2.95315219E+01]),
                NASA([1388.00, 5000.00],
                    [ 2.52251243E+01,  2.32468040E-02, -8.13790138E-06,
                    1.28321972E-09, -7.52598084E-14, -4.39222295E+04,
                    -9.98386061E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0))

species(name='NC6D1KET36',
        atoms='H:10 C:6 O:3',
        thermo=(NASA([300.00, 1386.00],
                    [ 1.64544868E+00,  7.94444833E-02, -6.17833446E-05,
                    2.55121988E-08, -4.38629485E-12, -3.37709382E+04,
                    2.74198430E+01]),
                NASA([1386.00, 5000.00],
                    [ 2.46079059E+01,  2.38193350E-02, -8.34639116E-06,
                    1.31693263E-09, -7.72708520E-14, -4.16806989E+04,
                    -9.54262876E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0))

species(name='NC6D1KET43',
        atoms='H:10 C:6 O:3',
        thermo=(NASA([300.00, 1386.00],
                    [ 1.07756600E+00,  8.15204297E-02, -6.41161371E-05,
                    2.66231177E-08, -4.58072562E-12, -3.33626466E+04,
                    3.03700873E+01]),
                NASA([1386.00, 5000.00],
                    [ 2.48134658E+01,  2.36679402E-02, -8.29973712E-06,
                    1.31026548E-09, -7.69086532E-14, -4.14850247E+04,
                    -9.64574729E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0))

species(name='NC6D1KET45',
        atoms='H:10 C:6 O:3',
        thermo=(NASA([300.00, 1390.00],
                    [-1.36114624E+00,  9.04259145E-02, -7.50375115E-05,
                    3.23037734E-08, -5.65763026E-12, -3.21505116E+04,
                    4.20073825E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.51943125E+01,  2.33015181E-02, -8.16424277E-06,
                    1.28820260E-09, -7.55881804E-14, -4.09300018E+04,
                    -9.89201697E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0))

species(name='NC6D1KET46',
        atoms='H:10 C:6 O:3',
        thermo=(NASA([300.00, 1386.00],
                    [ 1.20615803E+00,  7.90495420E-02, -6.05529017E-05,
                    2.46670926E-08, -4.19757300E-12, -3.11658629E+04,
                    3.13918111E+01]),
                NASA([1386.00, 5000.00],
                    [ 2.40046171E+01,  2.43109061E-02, -8.51114087E-06,
                    1.34213064E-09, -7.87165658E-14, -3.90699585E+04,
                    -9.07535094E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0))

species(name='NC6D1KET53',
        atoms='H:10 C:6 O:3',
        thermo=(NASA([300.00, 1387.00],
                    [ 1.24868954E+00,  8.16249635E-02, -6.49725180E-05,
                    2.73590198E-08, -4.76412072E-12, -3.33627981E+04,
                    2.91095742E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.47812957E+01,  2.36294207E-02, -8.27119017E-06,
                    1.30416539E-09, -7.64852162E-14, -4.13548778E+04,
                    -9.64100793E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0))

species(name='NC6D1KET54',
        atoms='H:10 C:6 O:3',
        thermo=(NASA([300.00, 1389.00],
                    [-2.01058658E-01,  8.54688737E-02, -6.79446523E-05,
                    2.80458376E-08, -4.74057650E-12, -3.28539887E+04,
                    3.64494167E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.50907419E+01,  2.33708456E-02, -8.18394462E-06,
                    1.29083225E-09, -7.57230478E-14, -4.13898543E+04,
                    -9.84066457E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0))

species(name='NC6D1KET56',
        atoms='H:10 C:6 O:3',
        thermo=(NASA([300.00, 1381.00],
                    [ 4.01869706E+00,  6.91654483E-02, -4.83997299E-05,
                    1.82279228E-08, -2.94649828E-12, -3.13379792E+04,
                    1.75263729E+01]),
                NASA([1381.00, 5000.00],
                    [ 2.37618794E+01,  2.45055834E-02, -8.57519967E-06,
                    1.35176226E-09, -7.92610624E-14, -3.85143810E+04,
                    -8.93197157E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0))

species(name='NC6D1KET63',
        atoms='H:10 C:6 O:3',
        thermo=(NASA([300.00, 1388.00],
                    [ 1.70060879E+00,  8.15598144E-02, -6.58032934E-05,
                    2.80678696E-08, -4.93297720E-12, -2.97393488E+04,
                    2.71300311E+01]),
                NASA([1388.00, 5000.00],
                    [ 2.50462110E+01,  2.33305307E-02, -8.15110144E-06,
                    1.28360076E-09, -7.52125954E-14, -3.75927321E+04,
                    -9.71357410E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0))

species(name='NC6D1KET64',
        atoms='H:10 C:6 O:3',
        thermo=(NASA([300.00, 1389.00],
                    [ 2.15202567E+00,  7.90190793E-02, -6.20735181E-05,
                    2.58128239E-08, -4.44363386E-12, -3.03069622E+04,
                    2.54105115E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.47760251E+01,  2.35073590E-02, -8.20009414E-06,
                    1.28997010E-09, -7.55308784E-14, -3.80021544E+04,
                    -9.53273681E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0))

species(name='NC6D1KET65',
        atoms='H:10 C:6 O:3',
        thermo=(NASA([300.00, 1391.00],
                    [-1.48870749E-01,  8.73133093E-02, -7.16904116E-05,
                    3.05825140E-08, -5.31542687E-12, -2.91739013E+04,
                    3.62934571E+01]),
                NASA([1391.00, 5000.00],
                    [ 2.53952898E+01,  2.30473334E-02, -8.05738668E-06,
                    1.26947036E-09, -7.44126533E-14, -3.76433550E+04,
                    -9.93640862E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0))

species(name='NC6D2KET45',
        atoms='H:10 C:6 O:3',
        thermo=(NASA([300.00, 1389.00],
                    [-3.44302412E-01,  8.71469094E-02, -7.08606252E-05,
                    2.99912727E-08, -5.19127430E-12, -3.62870364E+04,
                    3.52066066E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.53327329E+01,  2.32200480E-02, -8.14427749E-06,
                    1.28594516E-09, -7.54918410E-14, -4.48853627E+04,
                    -1.01425531E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0))

species(name='NC6D2KET46',
        atoms='H:10 C:6 O:3',
        thermo=(NASA([300.00, 1438.00],
                    [ 2.04214382E+00,  6.74637140E-02, -5.76467507E-05,
                    2.75403606E-08, -5.24364424E-12, -3.07999075E+04,
                    2.31801511E+01]),
                NASA([1438.00, 5000.00],
                    [ 1.47972337E+01,  2.63787617E-02, -7.98793527E-06,
                    1.13433394E-09, -6.15921181E-14, -3.40290783E+04,
                    -4.11916841E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0))

species(name='NC6D2KET54',
        atoms='H:10 C:6 O:3',
        thermo=(NASA([300.00, 1381.00],
                    [ 3.12404345E+00,  6.95391264E-02, -4.65433054E-05,
                    1.62779600E-08, -2.41180094E-12, -3.19606832E+04,
                    2.21368632E+01]),
                NASA([1381.00, 5000.00],
                    [ 2.35243117E+01,  2.47724587E-02, -8.68258702E-06,
                    1.37019527E-09, -8.04043024E-14, -3.94701680E+04,
                    -8.86810443E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0))

species(name='NC6D2KET56',
        atoms='H:10 C:6 O:3',
        thermo=(NASA([300.00, 1381.00],
                    [ 3.12404345E+00,  6.95391264E-02, -4.65433054E-05,
                    1.62779600E-08, -2.41180094E-12, -3.19606832E+04,
                    2.21368632E+01]),
                NASA([1381.00, 5000.00],
                    [ 2.35243117E+01,  2.47724587E-02, -8.68258702E-06,
                    1.37019527E-09, -8.04043024E-14, -3.94701680E+04,
                    -8.86810443E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0))

species(name='NC6D2KET64',
        atoms='H:10 C:6 O:3',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.24971473E+00,  7.79953716E-02, -6.05507316E-05,
                    2.50027046E-08, -4.29772077E-12, -3.12647870E+04,
                    2.44367570E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.45802947E+01,  2.37403050E-02, -8.29551027E-06,
                    1.30645583E-09, -7.65559206E-14, -3.89364374E+04,
                    -9.49639669E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0))

species(name='NC6D2KET65',
        atoms='H:10 C:6 O:3',
        thermo=(NASA([300.00, 1389.00],
                    [ 4.55809320E-01,  8.34466520E-02, -6.59538465E-05,
                    2.72040849E-08, -4.60895493E-12, -3.07073699E+04,
                    3.33547639E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.49192976E+01,  2.34620561E-02, -8.20268998E-06,
                    1.29237940E-09, -7.57556255E-14, -3.89849838E+04,
                    -9.71351584E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0))

species(name='NC6D3KET12',
        atoms='H:10 C:6 O:3',
        thermo=(NASA([300.00, 1388.00],
                    [ 4.63076404E-01,  8.25462584E-02, -6.39189002E-05,
                    2.57434455E-08, -4.26673272E-12, -3.06739450E+04,
                    3.37910731E+01]),
                NASA([1388.00, 5000.00],
                    [ 2.49365998E+01,  2.35127695E-02, -8.23530308E-06,
                    1.29909614E-09, -7.62141294E-14, -3.90472215E+04,
                    -9.70899266E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0))

species(name='NC6D3KET21',
        atoms='H:10 C:6 O:3',
        thermo=(NASA([300.00, 1383.00],
                    [ 2.56357701E+00,  7.40872652E-02, -5.34164996E-05,
                    2.03460526E-08, -3.26583985E-12, -3.35688920E+04,
                    2.29823962E+01]),
                NASA([1383.00, 5000.00],
                    [ 2.42955905E+01,  2.41356158E-02, -8.46727186E-06,
                    1.33706691E-09, -7.84956336E-14, -4.13270251E+04,
                    -9.42415282E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.548,
                                well_depth=467.0,
                                dipole=1.8,
                                rot_relax=1.0))

species(name='C6H112O2-1',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1391.00],
                    [ 1.72342389E-01,  7.96365546E-02, -6.17325710E-05,
                    2.58868673E-08, -4.53504168E-12, -1.31317497E+03,
                    3.28552164E+01]),
                NASA([1391.00, 5000.00],
                    [ 2.18247003E+01,  2.61477191E-02, -9.04382408E-06,
                    1.41447358E-09, -8.24824437E-14, -8.68620099E+03,
                    -8.26421791E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653))

species(name='C6H111O2-3',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1391.00],
                    [ 1.72342389E-01,  7.96365546E-02, -6.17325710E-05,
                    2.58868673E-08, -4.53504168E-12, -1.31317497E+03,
                    3.28552164E+01]),
                NASA([1391.00, 5000.00],
                    [ 2.18247003E+01,  2.61477191E-02, -9.04382408E-06,
                    1.41447358E-09, -8.24824437E-14, -8.68620099E+03,
                    -8.26421791E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653))

species(name='C6H111O2-4',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1391.00],
                    [ 2.77529293E-01,  7.86289898E-02, -6.01042683E-05,
                    2.48490451E-08, -4.30167349E-12, -1.83062566E+03,
                    3.27344165E+01]),
                NASA([1391.00, 5000.00],
                    [ 2.16615043E+01,  2.62498570E-02, -9.07088524E-06,
                    1.41783876E-09, -8.26433939E-14, -9.15007553E+03,
                    -8.14781346E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653))

species(name='C6H111O2-5',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1391.00],
                    [ 2.77529293E-01,  7.86289898E-02, -6.01042683E-05,
                    2.48490451E-08, -4.30167349E-12, -1.83062566E+03,
                    3.27344165E+01]),
                NASA([1391.00, 5000.00],
                    [ 2.16615043E+01,  2.62498570E-02, -9.07088524E-06,
                    1.41783876E-09, -8.26433939E-14, -9.15007553E+03,
                    -8.14781346E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653))

species(name='C6H111O2-6',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1388.00],
                    [ 1.27980301E+00,  7.21524639E-02, -5.02889230E-05,
                    1.89682658E-08, -3.05539055E-12,  1.81691439E+02,
                    2.96253760E+01]),
                NASA([1388.00, 5000.00],
                    [ 2.09441967E+01,  2.68874173E-02, -9.29709904E-06,
                    1.45380527E-09, -8.47641892E-14, -6.85709066E+03,
                    -7.64620284E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653))

species(name='C6H112O2-4',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 7.09241510E-01,  7.61167882E-02, -5.65441627E-05,
                    2.28626264E-08, -3.90914815E-12, -2.83646907E+03,
                    3.02206800E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.13545199E+01,  2.65647098E-02, -9.19147506E-06,
                    1.43790977E-09, -8.38625367E-14, -1.00291675E+04,
                    -8.04485096E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653))

species(name='C6H112O2-5',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 8.61603309E-01,  7.49248603E-02, -5.46729053E-05,
                    2.16848199E-08, -3.64611710E-12, -3.36167444E+03,
                    2.98764440E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.11830704E+01,  2.66691844E-02, -9.21827673E-06,
                    1.44112100E-09, -8.40098554E-14, -1.04865774E+04,
                    -7.92293261E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653))

species(name='C6H112O2-6',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1387.00],
                    [ 1.83281227E+00,  6.85685322E-02, -4.50094020E-05,
                    1.58888653E-08, -2.41748671E-12, -1.34425559E+03,
                    2.69157982E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.04743827E+01,  2.73032958E-02, -9.44419009E-06,
                    1.47713579E-09, -8.61373934E-14, -8.19960133E+03,
                    -7.42677120E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653))

species(name='C6H113O2-1',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1385.00],
                    [ 2.82240306E+00,  6.44791002E-02, -3.94590946E-05,
                    1.27102060E-08, -1.75876839E-12, -1.48617135E+03,
                    2.20302771E+01]),
                NASA([1385.00, 5000.00],
                    [ 2.03167587E+01,  2.74247868E-02, -9.48325138E-06,
                    1.48292781E-09, -8.64620193E-14, -8.09917620E+03,
                    -7.35400368E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653))

species(name='C6H113O2-2',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [-3.02934052E-01,  8.03127668E-02, -6.22466191E-05,
                    2.61216993E-08, -4.58234209E-12, -2.49048769E+03,
                    3.50985318E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.15210498E+01,  2.64199917E-02, -9.14096761E-06,
                    1.42997611E-09, -8.33989347E-14, -9.92850970E+03,
                    -8.13291479E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.943,
                                well_depth=588.653))

species(name='C6H101OOH3-4',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 9.52498186E-02,  8.03618403E-02, -6.06668109E-05,
                    2.37467937E-08, -3.80814694E-12,  3.83105496E+03,
                    3.51713411E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.35865321E+01,  2.39558178E-02, -8.23516169E-06,
                    1.28335285E-09, -7.46678485E-14, -4.17838818E+03,
                    -9.04585438E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='C6H101OOH3-5',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 9.52498186E-02,  8.03618403E-02, -6.06668109E-05,
                    2.37467937E-08, -3.80814694E-12,  3.83105496E+03,
                    3.51713411E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.35865321E+01,  2.39558178E-02, -8.23516169E-06,
                    1.28335285E-09, -7.46678485E-14, -4.17838818E+03,
                    -9.04585438E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='C6H101OOH3-6',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1391.00],
                    [-4.56627752E-02,  8.40204422E-02, -6.75654515E-05,
                    2.86242771E-08, -4.98055850E-12,  5.09507304E+03,
                    3.46344213E+01]),
                NASA([1391.00, 5000.00],
                    [ 2.38720815E+01,  2.41660837E-02, -8.40279243E-06,
                    1.31902436E-09, -7.71170592E-14, -2.89705656E+03,
                    -9.25485494E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='C6H101OOH4-3',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1392.00],
                    [-9.10297866E-01,  8.52583006E-02, -6.69192636E-05,
                    2.74969957E-08, -4.64326012E-12, -3.02468665E+03,
                    3.57766300E+01]),
                NASA([1392.00, 5000.00],
                    [ 2.37243515E+01,  2.47401711E-02, -8.60684147E-06,
                    1.35155980E-09, -7.90410535E-14, -1.13413060E+04,
                    -9.55723076E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='C6H101OOH4-5',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 6.31714497E-01,  7.74646638E-02, -5.64802642E-05,
                    2.12731200E-08, -3.28310530E-12,  3.24910406E+03,
                    3.30550855E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.32488323E+01,  2.42528100E-02, -8.33845278E-06,
                    1.29945460E-09, -7.56011681E-14, -4.56843145E+03,
                    -8.82821109E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='C6H101OOH4-6',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1392.00],
                    [ 1.32873065E-01,  8.25961968E-02, -6.51872823E-05,
                    2.70764415E-08, -4.62921664E-12,  4.56860069E+03,
                    3.41908856E+01]),
                NASA([1392.00, 5000.00],
                    [ 2.37355148E+01,  2.42568015E-02, -8.42850745E-06,
                    1.32245876E-09, -7.72936360E-14, -3.38002357E+03,
                    -9.15558475E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='C6H101OOH5-3',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1392.00],
                    [-9.10297866E-01,  8.52583006E-02, -6.69192636E-05,
                    2.74969957E-08, -4.64326012E-12, -3.02468665E+03,
                    3.57766300E+01]),
                NASA([1392.00, 5000.00],
                    [ 2.37243515E+01,  2.47401711E-02, -8.60684147E-06,
                    1.35155980E-09, -7.90410535E-14, -1.13413060E+04,
                    -9.55723076E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='C6H101OOH5-4',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [ 6.31714497E-01,  7.74646638E-02, -5.64802642E-05,
                    2.12731200E-08, -3.28310530E-12,  3.24910406E+03,
                    3.30550855E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.32488323E+01,  2.42528100E-02, -8.33845278E-06,
                    1.29945460E-09, -7.56011681E-14, -4.56843145E+03,
                    -8.82821109E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='C6H101OOH5-6',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1392.00],
                    [ 1.32873065E-01,  8.25961968E-02, -6.51872823E-05,
                    2.70764415E-08, -4.62921664E-12,  4.56980353E+03,
                    3.41908856E+01]),
                NASA([1392.00, 5000.00],
                    [ 2.37355148E+01,  2.42568015E-02, -8.42850745E-06,
                    1.32245876E-09, -7.72936360E-14, -3.37882072E+03,
                    -9.15558475E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='C6H101OOH6-3',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1389.00],
                    [-6.08388291E-04,  7.93277090E-02, -5.80291325E-05,
                    2.22216224E-08, -3.53369351E-12, -1.00198828E+03,
                    3.30695740E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.30069994E+01,  2.53677322E-02, -8.82729766E-06,
                    1.38638017E-09, -8.10848797E-14, -9.03881549E+03,
                    -9.05352219E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='C6H101OOH6-4',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1380.00],
                    [ 1.12488828E+00,  7.30055728E-02, -4.89258020E-05,
                    1.62744046E-08, -2.13927970E-12,  5.34182053E+03,
                    3.23392943E+01]),
                NASA([1380.00, 5000.00],
                    [ 2.31648127E+01,  2.39156984E-02, -8.14392966E-06,
                    1.26186543E-09, -7.31481869E-14, -2.51231039E+03,
                    -8.68162634E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='C6H101OOH6-5',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1380.00],
                    [ 1.12488828E+00,  7.30055728E-02, -4.89258020E-05,
                    1.62744046E-08, -2.13927970E-12,  5.34182053E+03,
                    3.23392943E+01]),
                NASA([1380.00, 5000.00],
                    [ 2.31648127E+01,  2.39156984E-02, -8.14392966E-06,
                    1.26186543E-09, -7.31481869E-14, -2.51231039E+03,
                    -8.68162634E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='C6H101OOH3-4O2',
        atoms='H:11 C:6 O:4',
        thermo=(NASA([300.00, 1390.00],
                    [ 1.02229563E+00,  9.72289536E-02, -8.42572926E-05,
                    3.82500314E-08, -7.03511167E-12, -1.41603382E+04,
                    3.16908172E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.84866924E+01,  2.50953833E-02, -8.76367432E-06,
                    1.37964073E-09, -8.08226365E-14, -2.30411917E+04,
                    -1.13219092E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149))

species(name='C6H101OOH3-5O2',
        atoms='H:11 C:6 O:4',
        thermo=(NASA([300.00, 1390.00],
                    [ 1.02229563E+00,  9.72289536E-02, -8.42572926E-05,
                    3.82500314E-08, -7.03511167E-12, -1.41603382E+04,
                    3.16908172E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.84866924E+01,  2.50953833E-02, -8.76367432E-06,
                    1.37964073E-09, -8.08226365E-14, -2.30411917E+04,
                    -1.13219092E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149))

species(name='C6H101OOH3-6O2',
        atoms='H:11 C:6 O:4',
        thermo=(NASA([300.00, 1389.00],
                    [ 1.21543622E+00,  9.32821348E-02, -7.74588187E-05,
                    3.39724888E-08, -6.10567238E-12, -1.19955805E+04,
                    3.25594137E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.77417137E+01,  2.57731343E-02, -9.00775991E-06,
                    1.41883954E-09, -8.31503458E-14, -2.07971148E+04,
                    -1.08180759E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149))

species(name='C6H101OOH4-3O2',
        atoms='H:11 C:6 O:4',
        thermo=(NASA([300.00, 1390.00],
                    [ 1.02229563E+00,  9.72289536E-02, -8.42572926E-05,
                    3.82500314E-08, -7.03511167E-12, -1.41603382E+04,
                    3.16908172E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.84866924E+01,  2.50953833E-02, -8.76367432E-06,
                    1.37964073E-09, -8.08226365E-14, -2.30411917E+04,
                    -1.13219092E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149))

species(name='C6H101OOH4-5O2',
        atoms='H:11 C:6 O:4',
        thermo=(NASA([300.00, 1391.00],
                    [ 1.29384625E+00,  9.53423407E-02, -8.12997536E-05,
                    3.63927238E-08, -6.62212513E-12, -1.46985283E+04,
                    3.08320369E+01]),
                NASA([1391.00, 5000.00],
                    [ 2.82642586E+01,  2.52451223E-02, -8.80643745E-06,
                    1.38537282E-09, -8.11177141E-14, -2.34838646E+04,
                    -1.11717074E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149))

species(name='C6H101OOH4-6O2',
        atoms='H:11 C:6 O:4',
        thermo=(NASA([300.00, 1389.00],
                    [ 2.04641152E+00,  8.98933326E-02, -7.28891049E-05,
                    3.13362370E-08, -5.55187478E-12, -1.26465661E+04,
                    2.88960825E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.75912179E+01,  2.58612736E-02, -9.02901410E-06,
                    1.42117498E-09, -8.32456690E-14, -2.12205426E+04,
                    -1.06984128E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149))

species(name='C6H101OOH5-3O2',
        atoms='H:11 C:6 O:4',
        thermo=(NASA([300.00, 1390.00],
                    [ 1.02229563E+00,  9.72289536E-02, -8.42572926E-05,
                    3.82500314E-08, -7.03511167E-12, -1.41603382E+04,
                    3.16908172E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.84866924E+01,  2.50953833E-02, -8.76367432E-06,
                    1.37964073E-09, -8.08226365E-14, -2.30411917E+04,
                    -1.13219092E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149))

species(name='C6H101OOH5-4O2',
        atoms='H:11 C:6 O:4',
        thermo=(NASA([300.00, 1391.00],
                    [ 1.29384625E+00,  9.53423407E-02, -8.12997536E-05,
                    3.63927238E-08, -6.62212513E-12, -1.46985283E+04,
                    3.08320369E+01]),
                NASA([1391.00, 5000.00],
                    [ 2.82642586E+01,  2.52451223E-02, -8.80643745E-06,
                    1.38537282E-09, -8.11177141E-14, -2.34838646E+04,
                    -1.11717074E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149))

species(name='C6H101OOH5-6O2',
        atoms='H:11 C:6 O:4',
        thermo=(NASA([300.00, 1389.00],
                    [ 2.04641152E+00,  8.98933326E-02, -7.28891049E-05,
                    3.13362370E-08, -5.55187478E-12, -1.26465661E+04,
                    2.88960825E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.75912179E+01,  2.58612736E-02, -9.02901410E-06,
                    1.42117498E-09, -8.32456690E-14, -2.12205426E+04,
                    -1.06984128E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149))

species(name='C6H101OOH6-3O2',
        atoms='H:11 C:6 O:4',
        thermo=(NASA([300.00, 1389.00],
                    [ 1.21543622E+00,  9.32821348E-02, -7.74588187E-05,
                    3.39724888E-08, -6.10567238E-12, -1.19955805E+04,
                    3.25594137E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.77417137E+01,  2.57731343E-02, -9.00775991E-06,
                    1.41883954E-09, -8.31503458E-14, -2.07971148E+04,
                    -1.08180759E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149))

species(name='C6H101OOH6-4O2',
        atoms='H:11 C:6 O:4',
        thermo=(NASA([300.00, 1389.00],
                    [ 2.04641152E+00,  8.98933326E-02, -7.28891049E-05,
                    3.13362370E-08, -5.55187478E-12, -1.26465661E+04,
                    2.88960825E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.75912179E+01,  2.58612736E-02, -9.02901410E-06,
                    1.42117498E-09, -8.32456690E-14, -2.12205426E+04,
                    -1.06984128E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149))

species(name='C6H101OOH6-5O2',
        atoms='H:11 C:6 O:4',
        thermo=(NASA([300.00, 1389.00],
                    [ 2.04641152E+00,  8.98933326E-02, -7.28891049E-05,
                    3.13362370E-08, -5.55187478E-12, -1.26465661E+04,
                    2.88960825E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.75912179E+01,  2.58612736E-02, -9.02901410E-06,
                    1.42117498E-09, -8.32456690E-14, -2.12205426E+04,
                    -1.06984128E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149))

species(name='C6H102OOH5-4',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [-4.02203129E-01,  8.18778797E-02, -6.19069792E-05,
                    2.45652582E-08, -4.03514422E-12, -4.54420456E+03,
                    3.32729123E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.32752685E+01,  2.51409461E-02, -8.74931155E-06,
                    1.37423314E-09, -8.03786732E-14, -1.26933360E+04,
                    -9.35010174E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='C6H103OOH1-2',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1386.00],
                    [ 1.54280561E+00,  7.16314208E-02, -4.71541806E-05,
                    1.59384080E-08, -2.23271419E-12, -2.66822299E+03,
                    2.54748806E+01]),
                NASA([1386.00, 5000.00],
                    [ 2.23753433E+01,  2.59192338E-02, -9.02065197E-06,
                    1.41686858E-09, -8.28719949E-14, -1.02813685E+04,
                    -8.75969572E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='C6H102OOH4-5',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1386.00],
                    [ 5.82697759E-01,  7.69938922E-02, -5.55576336E-05,
                    2.06590703E-08, -3.14739197E-12,  2.31656586E+03,
                    3.27785094E+01]),
                NASA([1386.00, 5000.00],
                    [ 2.32398122E+01,  2.41563455E-02, -8.28585057E-06,
                    1.28951635E-09, -7.49613636E-14, -5.56292849E+03,
                    -8.89426926E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='C6H102OOH4-6',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [ 4.98679078E-02,  8.26265919E-02, -6.56231370E-05,
                    2.76373931E-08, -4.80771067E-12,  3.63374932E+03,
                    3.47019993E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.34491400E+01,  2.45596126E-02, -8.54602502E-06,
                    1.34215244E-09, -7.84946077E-14, -4.25919499E+03,
                    -8.99400439E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='C6H102OOH5-6',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1390.00],
                    [ 7.03137830E-01,  7.89566248E-02, -5.98556399E-05,
                    2.39724683E-08, -3.98633224E-12,  3.03956769E+03,
                    3.20891098E+01]),
                NASA([1390.00, 5000.00],
                    [ 2.32562311E+01,  2.46734129E-02, -8.57419421E-06,
                    1.34539437E-09, -7.86366380E-14, -4.71495052E+03,
                    -8.86067004E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='C6H102OOH6-4',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1385.00],
                    [ 5.46105318E-01,  7.55506610E-02, -5.20172914E-05,
                    1.83934696E-08, -2.69131116E-12, -2.52256179E+03,
                    3.04271050E+01]),
                NASA([1385.00, 5000.00],
                    [ 2.30773623E+01,  2.50917259E-02, -8.77976000E-06,
                    1.38404736E-09, -8.11587377E-14, -1.06363474E+04,
                    -9.14736269E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='C6H102OOH6-5',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1375.00],
                    [ 1.64691737E+00,  6.94744642E-02, -4.35533729E-05,
                    1.30171015E-08, -1.44134516E-12,  3.82106194E+03,
                    2.97878034E+01]),
                NASA([1375.00, 5000.00],
                    [ 2.28661984E+01,  2.40568017E-02, -8.17267637E-06,
                    1.26464679E-09, -7.32521694E-14, -3.92285482E+03,
                    -8.55823037E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='C6H103OOH2-1',
        atoms='H:11 C:6 O:2',
        thermo=(NASA([300.00, 1391.00],
                    [-5.12700845E-01,  8.46518233E-02, -6.80011968E-05,
                    2.88057929E-08, -5.01540641E-12,  3.91670434E+03,
                    3.75351732E+01]),
                NASA([1391.00, 5000.00],
                    [ 2.35706798E+01,  2.44362818E-02, -8.49919874E-06,
                    1.33441074E-09, -7.80267805E-14, -4.14069669E+03,
                    -9.05552440E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.96,
                                well_depth=591.585))

species(name='C6H102OOH4-5O2',
        atoms='H:11 C:6 O:4',
        thermo=(NASA([300.00, 1389.00],
                    [ 1.58140767E+00,  9.35665623E-02, -7.88770135E-05,
                    3.51187133E-08, -6.38734634E-12, -1.56854097E+04,
                    2.89658218E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.79894624E+01,  2.55285153E-02, -8.91538113E-06,
                    1.40355248E-09, -8.22240440E-14, -2.43730959E+04,
                    -1.10866625E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149))

species(name='C6H102OOH4-6O2',
        atoms='H:11 C:6 O:4',
        thermo=(NASA([300.00, 1387.00],
                    [ 1.77730962E+00,  8.96793550E-02, -7.21579412E-05,
                    3.08782424E-08, -5.46384052E-12, -1.35246967E+04,
                    2.98041253E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.72773170E+01,  2.61796496E-02, -9.15059792E-06,
                    1.44140203E-09, -8.44746905E-14, -2.21409666E+04,
                    -1.06012850E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149))

species(name='C6H102OOH5-4O2',
        atoms='H:11 C:6 O:4',
        thermo=(NASA([300.00, 1389.00],
                    [ 1.58140767E+00,  9.35665623E-02, -7.88770135E-05,
                    3.51187133E-08, -6.38734634E-12, -1.56854097E+04,
                    2.89658218E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.79894624E+01,  2.55285153E-02, -8.91538113E-06,
                    1.40355248E-09, -8.22240440E-14, -2.43730959E+04,
                    -1.10866625E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149))

species(name='C6H102OOH5-6O2',
        atoms='H:11 C:6 O:4',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.61043954E+00,  8.62028229E-02, -6.74604084E-05,
                    2.81730962E-08, -4.89693847E-12, -1.41722218E+04,
                    2.61483734E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.70950096E+01,  2.62942321E-02, -9.18082428E-06,
                    1.44512064E-09, -8.46497937E-14, -2.25534667E+04,
                    -1.04639988E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149))

species(name='C6H102OOH6-4O2',
        atoms='H:11 C:6 O:4',
        thermo=(NASA([300.00, 1387.00],
                    [ 1.77730962E+00,  8.96793550E-02, -7.21579412E-05,
                    3.08782424E-08, -5.46384052E-12, -1.35246967E+04,
                    2.98041253E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.72773170E+01,  2.61796496E-02, -9.15059792E-06,
                    1.44140203E-09, -8.44746905E-14, -2.21409666E+04,
                    -1.06012850E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149))

species(name='C6H102OOH6-5O2',
        atoms='H:11 C:6 O:4',
        thermo=(NASA([300.00, 1387.00],
                    [ 2.61043954E+00,  8.62028229E-02, -6.74604084E-05,
                    2.81730962E-08, -4.89693847E-12, -1.41722218E+04,
                    2.61483734E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.70950096E+01,  2.62942321E-02, -9.18082428E-06,
                    1.44512064E-09, -8.46497937E-14, -2.25534667E+04,
                    -1.04639988E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149))

species(name='C6H103OOH2-1O2',
        atoms='H:11 C:6 O:4',
        thermo=(NASA([300.00, 1389.00],
                    [ 7.41773416E-01,  9.40159244E-02, -7.80995580E-05,
                    3.42988822E-08, -6.17488044E-12, -1.31761876E+04,
                    3.47824369E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.74462755E+01,  2.60361877E-02, -9.10124714E-06,
                    1.43372336E-09, -8.40289184E-14, -2.20407968E+04,
                    -1.06906811E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149))

species(name='C6H103OOH1-2O2',
        atoms='H:11 C:6 O:4',
        thermo=(NASA([300.00, 1389.00],
                    [ 7.41773416E-01,  9.40159244E-02, -7.80995580E-05,
                    3.42988822E-08, -6.17488044E-12, -1.31761876E+04,
                    3.47824369E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.74462755E+01,  2.60361877E-02, -9.10124714E-06,
                    1.43372336E-09, -8.40289184E-14, -2.20407968E+04,
                    -1.06906811E+02])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.436,
                                well_depth=677.149))

species(name='C2H3COC3H7',
        atoms='H:10 C:6 O:1',
        thermo=(NASA([300.00, 1388.00],
                    [ 4.46467269E-01,  6.64586762E-02, -4.70073659E-05,
                    1.79961137E-08, -2.93748963E-12, -2.30628492E+04,
                    2.77541300E+01]),
                NASA([1388.00, 5000.00],
                    [ 1.86909693E+01,  2.41939465E-02, -8.38233377E-06,
                    1.31252224E-09, -7.65989643E-14, -2.95692017E+04,
                    -7.05828250E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.674,
                                well_depth=541.5,
                                dipole=1.8,
                                rot_relax=1.0))

species(name='C5H81-4',
        atoms='H:8 C:5',
        thermo=(NASA([300.00, 1393.00],
                    [-1.64766327E+00,  5.54190112E-02, -4.12156069E-05,
                    1.62877209E-08, -2.66483052E-12,  1.11493804E+04,
                    3.48094470E+01]),
                NASA([1393.00, 5000.00],
                    [ 1.35352242E+01,  1.85170786E-02, -6.26984814E-06,
                    9.66983255E-10, -5.58485033E-14,  6.00118884E+03,
                    -4.62552894E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.489,
                                well_depth=386.2,
                                dipole=0.4,
                                rot_relax=1.0),
        note='3/4/8THERM')

species(name='C5H91O2-3',
        atoms='H:9 C:5 O:2',
        thermo=(NASA([300.00, 1391.00],
                    [ 3.43181024E-01,  6.76700566E-02, -5.38079915E-05,
                    2.31484493E-08, -4.14073100E-12,  2.25231466E+02,
                    2.96899952E+01]),
                NASA([1391.00, 5000.00],
                    [ 1.86702210E+01,  2.16216768E-02, -7.48512415E-06,
                    1.17139079E-09, -6.83358511E-14, -5.94893165E+03,
                    -6.78134047E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='2/27/9THERM')

species(name='C5H91O2-4',
        atoms='H:9 C:5 O:2',
        thermo=(NASA([300.00, 1391.00],
                    [ 8.29161679E-01,  6.49730353E-02, -4.98486874E-05,
                    2.07535028E-08, -3.62109673E-12, -3.48709260E+02,
                    2.78111162E+01]),
                NASA([1391.00, 5000.00],
                    [ 1.84021744E+01,  2.17923554E-02, -7.53113607E-06,
                    1.17721317E-09, -6.86192272E-14, -6.36049709E+03,
                    -6.60160803E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='2/27/9THERM')

species(name='C5H91O2-5',
        atoms='H:9 C:5 O:2',
        thermo=(NASA([300.00, 1388.00],
                    [ 1.45241685E+00,  6.01856302E-02, -4.23740334E-05,
                    1.62426968E-08, -2.66522777E-12,  1.71964569E+03,
                    2.50683944E+01]),
                NASA([1388.00, 5000.00],
                    [ 1.77874223E+01,  2.23672091E-02, -7.74125889E-06,
                    1.21125611E-09, -7.06521652E-14, -4.11946666E+03,
                    -6.30035459E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='2/27/9THERM')

species(name='C5H81OOH3-4',
        atoms='H:9 C:5 O:2',
        thermo=(NASA([300.00, 1387.00],
                    [ 3.21512682E-01,  6.81363037E-02, -5.23290322E-05,
                    2.07388630E-08, -3.35294073E-12,  6.71931786E+03,
                    3.24418970E+01]),
                NASA([1387.00, 5000.00],
                    [ 2.04736244E+01,  1.93718338E-02, -6.65285950E-06,
                    1.03631449E-09, -6.02831329E-14, -1.02391566E+02,
                    -7.51780148E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='THERM')

species(name='C5H81OOH3-5',
        atoms='H:9 C:5 O:2',
        thermo=(NASA([300.00, 1391.00],
                    [ 1.22655285E-01,  7.20652015E-02, -5.96557159E-05,
                    2.58930340E-08, -4.58741292E-12,  7.99306548E+03,
                    3.28640779E+01]),
                NASA([1391.00, 5000.00],
                    [ 2.07195327E+01,  1.96364025E-02, -6.84241764E-06,
                    1.07563823E-09, -6.29511640E-14,  1.19896467E+03,
                    -7.63467720E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='THERM')

species(name='C5H81OOH4-3',
        atoms='H:9 C:5 O:2',
        thermo=(NASA([300.00, 1395.00],
                    [-1.09905666E+00,  7.85076363E-02, -6.81368662E-05,
                    3.06237682E-08, -5.54283245E-12,  1.12681194E+03,
                    3.47817523E+01]),
                NASA([1395.00, 5000.00],
                    [ 2.13003585E+01,  1.95078183E-02, -6.78481644E-06,
                    1.06526973E-09, -6.22918022E-14, -6.03638639E+03,
                    -8.32373187E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='THERM')

species(name='C5H81OOH4-5',
        atoms='H:9 C:5 O:2',
        thermo=(NASA([300.00, 1392.00],
                    [ 3.38002912E-01,  7.05163186E-02, -5.72050085E-05,
                    2.43500517E-08, -4.24439952E-12,  7.46051091E+03,
                    3.22447136E+01]),
                NASA([1392.00, 5000.00],
                    [ 2.05400792E+01,  1.97587133E-02, -6.87784366E-06,
                    1.08045541E-09, -6.32025760E-14,  7.38073415E+02,
                    -7.50954151E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='THERM')

species(name='C5H81OOH5-4',
        atoms='H:9 C:5 O:2',
        thermo=(NASA([300.00, 1378.00],
                    [ 1.40984227E+00,  6.05508825E-02, -4.03214237E-05,
                    1.31746719E-08, -1.67859779E-12,  8.22059330E+03,
                    2.86434933E+01]),
                NASA([1378.00, 5000.00],
                    [ 1.99797637E+01,  1.94783902E-02, -6.63044969E-06,
                    1.02741023E-09, -5.95687008E-14,  1.58234802E+03,
                    -7.18412080E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='THERM')

species(name='C5H81OOH5-3',
        atoms='H:9 C:5 O:2',
        thermo=(NASA([300.00, 1392.00],
                    [-4.80015195E-01,  7.37366908E-02, -6.06933626E-05,
                    2.61394273E-08, -4.59439636E-12,  3.19591805E+03,
                    3.20592964E+01]),
                NASA([1392.00, 5000.00],
                    [ 2.06841270E+01,  2.00877308E-02, -6.99755273E-06,
                    1.09981315E-09, -6.43576906E-14, -3.79637676E+03,
                    -8.02196830E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='THERM')

species(name='C5H92O2-1',
        atoms='H:9 C:5 O:2',
        thermo=(NASA([300.00, 1382.00],
                    [ 3.26704287E+00,  5.11396085E-02, -2.90627615E-05,
                    8.29183575E-09, -9.77238694E-13,  6.87723123E+02,
                    1.62365492E+01]),
                NASA([1382.00, 5000.00],
                    [ 1.73081531E+01,  2.28361563E-02, -7.91701528E-06,
                    1.24017112E-09, -7.23964243E-14, -4.77473659E+03,
                    -6.09976106E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='2/27/9THERM')

species(name='C5H81OOH3-4O2',
        atoms='H:9 C:5 O:4',
        thermo=(NASA([300.00, 1391.00],
                    [ 7.00675697E-01,  8.61763030E-02, -7.69681261E-05,
                    3.56494726E-08, -6.63419204E-12, -1.25103846E+04,
                    3.28746076E+01]),
                NASA([1391.00, 5000.00],
                    [ 2.51899540E+01,  2.06629888E-02, -7.23143253E-06,
                    1.14009288E-09, -6.68582448E-14, -2.02995332E+04,
                    -9.58973374E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='THERM')

species(name='C5H81OOH3-5O2',
        atoms='H:9 C:5 O:4',
        thermo=(NASA([300.00, 1388.00],
                    [ 1.50852696E+00,  8.07768050E-02, -6.87240343E-05,
                    3.07180782E-08, -5.59438096E-12, -1.04758678E+04,
                    2.88278214E+01]),
                NASA([1388.00, 5000.00],
                    [ 2.45930237E+01,  2.12211773E-02, -7.43558644E-06,
                    1.17319038E-09, -6.88358634E-14, -1.80568519E+04,
                    -9.33693862E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='THERM')

species(name='C5H81OOH4-3O2',
        atoms='H:9 C:5 O:4',
        thermo=(NASA([300.00, 1391.00],
                    [ 7.00675697E-01,  8.61763030E-02, -7.69681261E-05,
                    3.56494726E-08, -6.63419204E-12, -1.25103846E+04,
                    3.28746076E+01]),
                NASA([1391.00, 5000.00],
                    [ 2.51899540E+01,  2.06629888E-02, -7.23143253E-06,
                    1.14009288E-09, -6.68582448E-14, -2.02995332E+04,
                    -9.58973374E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='THERM')

species(name='C5H81OOH4-5O2',
        atoms='H:9 C:5 O:4',
        thermo=(NASA([300.00, 1389.00],
                    [ 1.76455657E+00,  7.89923440E-02, -6.60490017E-05,
                    2.90938793E-08, -5.24124922E-12, -1.10119071E+04,
                    2.80364294E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.43180086E+01,  2.14111651E-02, -7.49110481E-06,
                    1.18078372E-09, -6.92336785E-14, -1.84737378E+04,
                    -9.15550695E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='THERM')

species(name='C5H81OOH5-3O2',
        atoms='H:9 C:5 O:4',
        thermo=(NASA([300.00, 1388.00],
                    [ 1.50852696E+00,  8.07768050E-02, -6.87240343E-05,
                    3.07180782E-08, -5.59438096E-12, -1.04758678E+04,
                    2.88278214E+01]),
                NASA([1388.00, 5000.00],
                    [ 2.45930237E+01,  2.12211773E-02, -7.43558644E-06,
                    1.17319038E-09, -6.88358634E-14, -1.80568519E+04,
                    -9.33693862E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='THERM')

species(name='C5H81OOH5-4O2',
        atoms='H:9 C:5 O:4',
        thermo=(NASA([300.00, 1389.00],
                    [ 1.76455657E+00,  7.89923440E-02, -6.60490017E-05,
                    2.90938793E-08, -5.24124922E-12, -1.10119071E+04,
                    2.80364294E+01]),
                NASA([1389.00, 5000.00],
                    [ 2.43180086E+01,  2.14111651E-02, -7.49110481E-06,
                    1.18078372E-09, -6.92336785E-14, -1.84737378E+04,
                    -9.15550695E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='THERM')

species(name='NC5D1KET34',
        atoms='H:8 C:5 O:3',
        thermo=(NASA([300.00, 1385.00],
                    [ 1.82097767E+00,  6.74699547E-02, -5.38264037E-05,
                    2.27270486E-08, -3.97590535E-12, -3.08745570E+04,
                    2.42292350E+01]),
                NASA([1385.00, 5000.00],
                    [ 2.14757522E+01,  1.92472750E-02, -6.76598303E-06,
                    1.06986594E-09, -6.28685877E-14, -3.75872106E+04,
                    -8.07109370E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.174,
                                well_depth=494.8,
                                dipole=1.8,
                                rot_relax=1.0),
        note='3/2/9THERM')

species(name='NC5D1KET35',
        atoms='H:8 C:5 O:3',
        thermo=(NASA([300.00, 1385.00],
                    [ 1.82097767E+00,  6.74699547E-02, -5.38264037E-05,
                    2.27270486E-08, -3.97590535E-12, -3.08745570E+04,
                    2.42292350E+01]),
                NASA([1385.00, 5000.00],
                    [ 2.14757522E+01,  1.92472750E-02, -6.76598303E-06,
                    1.06986594E-09, -6.28685877E-14, -3.75872106E+04,
                    -8.07109370E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.174,
                                well_depth=494.8,
                                dipole=1.8,
                                rot_relax=1.0),
        note='3/2/9THERM')

species(name='NC5D1KET43',
        atoms='H:8 C:5 O:3',
        thermo=(NASA([300.00, 1385.00],
                    [ 1.82097767E+00,  6.74699547E-02, -5.38264037E-05,
                    2.27270486E-08, -3.97590535E-12, -3.08745570E+04,
                    2.42292350E+01]),
                NASA([1385.00, 5000.00],
                    [ 2.14757522E+01,  1.92472750E-02, -6.76598303E-06,
                    1.06986594E-09, -6.28685877E-14, -3.75872106E+04,
                    -8.07109370E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.174,
                                well_depth=494.8,
                                dipole=1.8,
                                rot_relax=1.0),
        note='3/2/9THERM')

species(name='NC5D1KET45',
        atoms='H:8 C:5 O:3',
        thermo=(NASA([300.00, 1385.00],
                    [ 1.82097767E+00,  6.74699547E-02, -5.38264037E-05,
                    2.27270486E-08, -3.97590535E-12, -3.08745570E+04,
                    2.42292350E+01]),
                NASA([1385.00, 5000.00],
                    [ 2.14757522E+01,  1.92472750E-02, -6.76598303E-06,
                    1.06986594E-09, -6.28685877E-14, -3.75872106E+04,
                    -8.07109370E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.174,
                                well_depth=494.8,
                                dipole=1.8,
                                rot_relax=1.0),
        note='3/2/9THERM')

species(name='NC5D1KET53',
        atoms='H:8 C:5 O:3',
        thermo=(NASA([300.00, 1385.00],
                    [ 1.82097767E+00,  6.74699547E-02, -5.38264037E-05,
                    2.27270486E-08, -3.97590535E-12, -3.08745570E+04,
                    2.42292350E+01]),
                NASA([1385.00, 5000.00],
                    [ 2.14757522E+01,  1.92472750E-02, -6.76598303E-06,
                    1.06986594E-09, -6.28685877E-14, -3.75872106E+04,
                    -8.07109370E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.174,
                                well_depth=494.8,
                                dipole=1.8,
                                rot_relax=1.0),
        note='3/2/9THERM')

species(name='NC5D1KET54',
        atoms='H:8 C:5 O:3',
        thermo=(NASA([300.00, 1385.00],
                    [ 1.82097767E+00,  6.74699547E-02, -5.38264037E-05,
                    2.27270486E-08, -3.97590535E-12, -3.08745570E+04,
                    2.42292350E+01]),
                NASA([1385.00, 5000.00],
                    [ 2.14757522E+01,  1.92472750E-02, -6.76598303E-06,
                    1.06986594E-09, -6.28685877E-14, -3.75872106E+04,
                    -8.07109370E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=6.174,
                                well_depth=494.8,
                                dipole=1.8,
                                rot_relax=1.0),
        note='3/2/9THERM')

species(name='CY3C5H8O',
        atoms='H:8 C:5 O:1',
        thermo=(NASA([300.00, 1402.00],
                    [-4.02699181E+00,  7.41038879E-02, -6.41370347E-05,
                    2.86307024E-08, -5.12072635E-12, -3.95617677E+03,
                    4.44028774E+01]),
                NASA([1402.00, 5000.00],
                    [ 1.67912845E+01,  1.86881223E-02, -6.41180595E-06,
                    9.97554346E-10, -5.79620662E-14, -1.05005341E+04,
                    -6.49852651E+01])),
        transport=gas_transport(geom='nonlinear',
                                diam=5.664,
                                well_depth=523.2,
                                dipole=1.7,
                                rot_relax=1.0),
        note='3/2/9THERM')

#-------------------------------------------------------------------------------
# Reaction data
#-------------------------------------------------------------------------------

# Reaction 1
reaction('H + O2 => O + OH', [3.547000e+15, -0.406, 16600.0])

# Reaction 2
reaction('O + OH => H + O2', [1.027000e+13, -0.015, -133.0])

# Reaction 3
reaction('O + H2 => H + OH', [5.080000e+04, 2.67, 6292.0])

# Reaction 4
reaction('H + OH => O + H2', [2.637000e+04, 2.651, 4880.0])

# Reaction 5
reaction('OH + H2 => H + H2O', [2.160000e+08, 1.51, 3430.0])

# Reaction 6
reaction('H + H2O => OH + H2', [2.290000e+09, 1.404, 18320.0])

# Reaction 7
reaction('O + H2O => OH + OH', [2.970000e+06, 2.02, 13400.0])

# Reaction 8
reaction('OH + OH => O + H2O', [1.454000e+05, 2.107, -2904.0])

# Reaction 9
three_body_reaction('H2 + M => H + H + M', [4.577000e+19, -1.4, 104400.0],
                    efficiencies='H2:2.5 H2O:12.0 CO2:3.8 CO:1.9')

# Reaction 10
three_body_reaction('H + H + M => H2 + M', [1.145000e+20, -1.676, 820.0],
                    efficiencies='H2:2.5 H2O:12.0 CO2:3.8 CO:1.9')

# Reaction 11
three_body_reaction('O2 + M => O + O + M', [4.420000e+17, -0.634, 118900.0],
                    efficiencies='CO2:3.8 CO:1.9 H2:2.5 H2O:12.0 AR:0.83 CH4:2.0 C2H6:3.0 HE:0.83')

# Reaction 12
three_body_reaction('O + O + M => O2 + M', [6.165000e+15, -0.5, 0.0],
                    efficiencies='CO2:3.8 CO:1.9 H2:2.5 H2O:12.0 AR:0.83 CH4:2.0 C2H6:3.0 HE:0.83')

# Reaction 13
three_body_reaction('OH + M => O + H + M', [9.780000e+17, -0.743, 102100.0],
                    efficiencies='CO2:2.0 CO:1.5 H2:2.5 H2O:12.0 AR:0.75 CH4:2.0 C2H6:3.0 HE:0.75')

# Reaction 14
three_body_reaction('O + H + M => OH + M', [4.714000e+18, -1.0, 0.0],
                    efficiencies='CO2:2.0 CO:1.5 H2:2.5 H2O:12.0 AR:0.75 CH4:2.0 C2H6:3.0 HE:0.75')

# Reaction 15
three_body_reaction('H2O + M => H + OH + M', [1.907000e+23, -1.83, 118500.0],
                    efficiencies='H2:0.73 H2O:12.0 AR:0.38 CH4:2.0 C2H6:3.0 HE:0.38')

# Reaction 16
three_body_reaction('H + OH + M => H2O + M', [4.500000e+22, -2.0, 0.0],
                    efficiencies='H2:0.73 H2O:12.0 AR:0.38 CH4:2.0 C2H6:3.0 HE:0.38')

# Reaction 17
falloff_reaction('H + O2 (+ M) <=> HO2 (+ M)',
                kf=[1.475000e+12, 0.6, 0.0],
                kf0=[3.482000e+16, -0.411, -1115.0],
                efficiencies='CO2:3.8 CO:1.9 H2:1.3 H2O:14.0 AR:0.67 CH4:2.0 C2H6:3.0 HE:0.67',
                falloff=Troe(A=0.5, T3=1e-30, T1=1e+30, T2=10000000000.0))

# Reaction 18
reaction('HO2 + H => H2 + O2', [1.660000e+13, 0.0, 823.0])

# Reaction 19
reaction('H2 + O2 => HO2 + H', [3.166000e+12, 0.348, 55510.0])

# Reaction 20
reaction('HO2 + H => OH + OH', [7.079000e+13, 0.0, 295.0])

# Reaction 21
reaction('OH + OH => HO2 + H', [2.028000e+10, 0.72, 36840.0])

# Reaction 22
reaction('HO2 + O => OH + O2', [3.250000e+13, 0.0, 0.0])

# Reaction 23
reaction('OH + O2 => HO2 + O', [3.217000e+12, 0.329, 53280.0])

# Reaction 24
reaction('HO2 + OH => H2O + O2', [1.973000e+10, 0.962, -328.4])

# Reaction 25
reaction('H2O + O2 => HO2 + OH', [3.989000e+10, 1.204, 69250.0])

# Reaction 26
reaction('H2O2 + O2 => HO2 + HO2', [1.136000e+16, -0.347, 49730.0],
        options='duplicate')

# Reaction 27
reaction('HO2 + HO2 => H2O2 + O2', [1.030000e+14, 0.0, 11040.0],
        options='duplicate')

# Reaction 28
reaction('H2O2 + O2 => HO2 + HO2', [2.141000e+13, -0.347, 37280.0],
        options='duplicate')

# Reaction 29
reaction('HO2 + HO2 => H2O2 + O2', [1.940000e+11, 0.0, -1409.0],
        options='duplicate')

# Reaction 30
falloff_reaction('H2O2 (+ M) <=> OH + OH (+ M)',
                kf=[2.951000e+14, 0.0, 48430.0],
                kf0=[1.202000e+17, 0.0, 45500.0],
                efficiencies='CO2:3.8 CO:1.9 H2:2.5 H2O:12.0 AR:0.64 CH4:2.0 C2H6:3.0 HE:0.64',
                falloff=Troe(A=0.5, T3=1e-30, T1=1e+30, T2=10000000000.0))

# Reaction 31
reaction('H2O2 + H => H2O + OH', [2.410000e+13, 0.0, 3970.0])

# Reaction 32
reaction('H2O + OH => H2O2 + H', [1.265000e+08, 1.31, 71410.0])

# Reaction 33
reaction('H2O2 + H => H2 + HO2', [2.150000e+10, 1.0, 6000.0])

# Reaction 34
reaction('H2 + HO2 => H2O2 + H', [3.716000e+07, 1.695, 22000.0])

# Reaction 35
reaction('H2O2 + O => OH + HO2', [9.550000e+06, 2.0, 3970.0])

# Reaction 36
reaction('OH + HO2 => H2O2 + O', [8.568000e+03, 2.676, 18560.0])

# Reaction 37
reaction('H2O2 + OH => H2O + HO2', [2.000000e+12, 0.0, 427.2],
        options='duplicate')

# Reaction 38
reaction('H2O + HO2 => H2O2 + OH', [3.665000e+10, 0.589, 31320.0],
        options='duplicate')

# Reaction 39
reaction('H2O2 + OH => H2O + HO2', [1.700000e+18, 0.0, 29410.0],
        options='duplicate')

# Reaction 40
reaction('H2O + HO2 => H2O2 + OH', [3.115000e+16, 0.589, 60300.0],
        options='duplicate')

# Reaction 41
falloff_reaction('CO + O (+ M) <=> CO2 (+ M)',
                kf=[1.800000e+10, 0.0, 2384.0],
                kf0=[1.350000e+24, -2.788, 4191.0],
                efficiencies='CO2:3.5 CO:1.5 H2:2.0 H2O:6.0 AR:0.5 CH4:2.0 C2H6:3.0 O2:6.0 HE:0.5')

# Reaction 42
reaction('CO + O2 => CO2 + O', [1.050000e+12, 0.0, 42540.0])

# Reaction 43
reaction('CO2 + O => CO + O2', [8.035000e+15, -0.8, 51230.0])

# Reaction 44
reaction('CO + OH => CO2 + H', [2.230000e+05, 1.89, -1158.0])

# Reaction 45
reaction('CO2 + H => CO + OH', [5.896000e+11, 0.699, 24260.0])

# Reaction 46
reaction('CO + HO2 => CO2 + OH', [3.010000e+13, 0.0, 23000.0])

# Reaction 47
reaction('CO2 + OH => CO + HO2', [2.280000e+16, -0.47, 84970.0])

# Reaction 48
three_body_reaction('HCO + M => H + CO + M', [4.750000e+11, 0.66, 14870.0],
                    efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:12.0 CH4:2.0 C2H6:3.0')

# Reaction 49
three_body_reaction('H + CO + M => HCO + M', [3.582000e+10, 1.041, -457.3],
                    efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:12.0 CH4:2.0 C2H6:3.0')

# Reaction 50
reaction('HCO + O2 => CO + HO2', [7.580000e+12, 0.0, 410.0])

# Reaction 51
reaction('CO + HO2 => HCO + O2', [1.198000e+12, 0.309, 33950.0])

# Reaction 52
reaction('HCO + H => CO + H2', [7.340000e+13, 0.0, 0.0])

# Reaction 53
reaction('CO + H2 => HCO + H', [2.212000e+12, 0.656, 88230.0])

# Reaction 54
reaction('HCO + O => CO + OH', [3.020000e+13, 0.0, 0.0])

# Reaction 55
reaction('CO + OH => HCO + O', [4.725000e+11, 0.638, 86820.0])

# Reaction 56
reaction('HCO + O => CO2 + H', [3.000000e+13, 0.0, 0.0])

# Reaction 57
reaction('CO2 + H => HCO + O', [1.241000e+18, -0.553, 112200.0])

# Reaction 58
reaction('HCO + OH => CO + H2O', [1.020000e+14, 0.0, 0.0])

# Reaction 59
reaction('CO + H2O => HCO + OH', [3.259000e+13, 0.551, 103100.0])

# Reaction 60
reaction('HCO + CH3 => CH4 + CO', [2.650000e+13, 0.0, 0.0])

# Reaction 61
reaction('CH4 + CO => HCO + CH3', [7.286000e+14, 0.211, 89770.0])

# Reaction 62
reaction('HCO + HO2 => CH2O + O2', [2.499000e+14, -0.061, 13920.0])

# Reaction 63
reaction('CH2O + O2 => HCO + HO2', [8.070000e+15, 0.0, 53420.0])

# Reaction 64
reaction('HCO + HO2 => CO2 + H + OH', [3.000000e+13, 0.0, 0.0])

# Reaction 65
reaction('CO2 + H + OH => HCO + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 66
reaction('O2CHO => HCO + O2', [9.959000e+15, -1.126, 41000.0])

# Reaction 67
reaction('HCO + O2 => O2CHO', [1.200000e+11, 0.0, -1100.0])

# Reaction 68
reaction('CH2O + O2CHO => HCO + HO2CHO', [1.990000e+12, 0.0, 11660.0])

# Reaction 69
reaction('HCO + HO2CHO => CH2O + O2CHO', [3.908000e+14, -0.909, 11810.0])

# Reaction 70
reaction('HO2CHO => OCHO + OH', [5.010000e+14, 0.0, 40150.0])

# Reaction 71
reaction('OCHO + OH => HO2CHO', [3.856000e+08, 1.532, -6372.0])

# Reaction 72
three_body_reaction('OCHO + M => H + CO2 + M', [5.318000e+14, -0.353, 17580.0])

# Reaction 73
three_body_reaction('H + CO2 + M => OCHO + M', [7.500000e+13, 0.0, 29000.0])

# Reaction 74
reaction('CH2O + CO => HCO + HCO', [9.186000e+13, 0.37, 73040.0])

# Reaction 75
reaction('HCO + HCO => CH2O + CO', [1.800000e+13, 0.0, 0.0])

# Reaction 76
reaction('HCO + HCO => H2 + CO + CO', [3.000000e+12, 0.0, 0.0])

# Reaction 77
reaction('H2 + CO + CO => HCO + HCO', [0.000000e+00, 0.0, 0.0])

# Reaction 78
falloff_reaction('HCO + H (+ M) <=> CH2O (+ M)',
                kf=[1.090000e+12, 0.48, -260.0],
                kf0=[1.350000e+24, -2.57, 1425.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0 HE:0.7',
                falloff=Troe(A=0.7824, T3=271.0, T1=2755.0, T2=6570.0))

# Reaction 79
falloff_reaction('CO + H2 (+ M) <=> CH2O (+ M)',
                kf=[4.300000e+07, 1.5, 79600.0],
                kf0=[5.070000e+27, -3.42, 84348.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0 HE:0.7',
                falloff=Troe(A=0.932, T3=197.0, T1=1540.0, T2=10300.0))

# Reaction 80
reaction('CH2O + OH => HCO + H2O', [7.820000e+07, 1.63, -1055.0])

# Reaction 81
reaction('HCO + H2O => CH2O + OH', [4.896000e+06, 1.811, 29030.0])

# Reaction 82
reaction('CH2O + H => HCO + H2', [5.740000e+07, 1.9, 2740.0])

# Reaction 83
reaction('HCO + H2 => CH2O + H', [3.390000e+05, 2.187, 17930.0])

# Reaction 84
reaction('CH2O + O => HCO + OH', [6.260000e+09, 1.15, 2260.0])

# Reaction 85
reaction('HCO + OH => CH2O + O', [1.919000e+07, 1.418, 16040.0])

# Reaction 86
reaction('CH2O + CH3 => HCO + CH4', [3.830000e+01, 3.36, 4312.0])

# Reaction 87
reaction('HCO + CH4 => CH2O + CH3', [2.063000e+02, 3.201, 21040.0])

# Reaction 88
reaction('CH2O + HO2 => HCO + H2O2', [7.100000e-03, 4.517, 6580.0])

# Reaction 89
reaction('HCO + H2O2 => CH2O + HO2', [2.426000e-02, 4.108, 5769.0])

# Reaction 90
reaction('HOCH2O => CH2O + OH', [2.056000e+21, -2.336, 25730.0])

# Reaction 91
reaction('CH2O + OH => HOCH2O', [4.500000e+15, -1.1, 0.0])

# Reaction 92
reaction('HOCH2O => HOCHO + H', [1.000000e+14, 0.0, 14900.0])

# Reaction 93
reaction('HOCHO + H => HOCH2O', [1.123000e+15, -0.295, 11500.0])

# Reaction 94
reaction('HOCHO => CO + H2O', [2.450000e+12, 0.0, 60470.0])

# Reaction 95
reaction('CO + H2O => HOCHO', [2.255000e+03, 2.093, 52890.0])

# Reaction 96
reaction('HOCHO => CO2 + H2', [2.950000e+09, 0.0, 48520.0])

# Reaction 97
reaction('CO2 + H2 => HOCHO', [6.772000e+05, 1.008, 51470.0])

# Reaction 98
reaction('HOCHO => HCO + OH', [3.471000e+22, -1.542, 110700.0])

# Reaction 99
reaction('HCO + OH => HOCHO', [1.000000e+14, 0.0, 0.0])

# Reaction 100
reaction('HOCHO + O2 => OCHO + HO2', [4.101000e+12, -0.308, 59880.0])

# Reaction 101
reaction('OCHO + HO2 => HOCHO + O2', [3.500000e+10, 0.0, -3275.0])

# Reaction 102
reaction('HOCHO + OH => H2O + CO2 + H', [2.620000e+06, 2.06, 916.0])

# Reaction 103
reaction('H2O + CO2 + H => HOCHO + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 104
reaction('HOCHO + OH => H2O + CO + OH', [1.850000e+07, 1.51, -962.0])

# Reaction 105
reaction('H2O + CO + OH => HOCHO + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 106
reaction('HOCHO + H => H2 + CO2 + H', [4.240000e+06, 2.1, 4868.0])

# Reaction 107
reaction('H2 + CO2 + H => HOCHO + H', [0.000000e+00, 0.0, 0.0])

# Reaction 108
reaction('HOCHO + H => H2 + CO + OH', [6.030000e+13, -0.35, 2988.0])

# Reaction 109
reaction('H2 + CO + OH => HOCHO + H', [0.000000e+00, 0.0, 0.0])

# Reaction 110
reaction('HOCHO + CH3 => CH4 + CO + OH', [3.900000e-07, 5.8, 2200.0])

# Reaction 111
reaction('CH4 + CO + OH => HOCHO + CH3', [0.000000e+00, 0.0, 0.0])

# Reaction 112
reaction('HOCHO + HO2 => OCHO + H2O2', [2.549000e+12, 0.04, 34470.0])

# Reaction 113
reaction('OCHO + H2O2 => HOCHO + HO2', [2.400000e+12, 0.0, 10000.0])

# Reaction 114
reaction('HOCHO + HO2 => H2O2 + CO + OH', [1.000000e+12, 0.0, 11920.0])

# Reaction 115
reaction('H2O2 + CO + OH => HOCHO + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 116
reaction('HOCHO + O => CO + OH + OH', [1.770000e+18, -1.9, 2975.0])

# Reaction 117
reaction('CO + OH + OH => HOCHO + O', [0.000000e+00, 0.0, 0.0])

# Reaction 118
reaction('HOCHO + HCO => CH2O + OCHO', [8.584000e+11, 0.04, 26750.0])

# Reaction 119
reaction('CH2O + OCHO => HOCHO + HCO', [5.600000e+12, 0.0, 13600.0])

# Reaction 120
falloff_reaction('CH3O (+ M) <=> CH2O + H (+ M)',
                kf=[6.800000e+13, 0.0, 26170.0],
                kf0=[1.867000e+25, -3.0, 24307.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.9, T3=2500.0, T1=1300.0, T2=1e+99))

# Reaction 121
reaction('CH3O + O2 => CH2O + HO2', [4.380000e-19, 9.5, -5501.0])

# Reaction 122
reaction('CH2O + HO2 => CH3O + O2', [1.416000e-20, 9.816, 21080.0])

# Reaction 123
reaction('CH2O + CH3O => CH3OH + HCO', [6.620000e+11, 0.0, 2294.0])

# Reaction 124
reaction('CH3OH + HCO => CH2O + CH3O', [8.393000e+10, 0.074, 17710.0])

# Reaction 125
reaction('CH4 + CH3O => CH3 + CH3OH', [6.119000e+02, 2.867, 8248.0])

# Reaction 126
reaction('CH3 + CH3OH => CH4 + CH3O', [1.440000e+01, 3.1, 6935.0])

# Reaction 127
reaction('CH3O + CH3 => CH2O + CH4', [1.200000e+13, 0.0, 0.0])

# Reaction 128
reaction('CH2O + CH4 => CH3O + CH3', [6.749000e+13, 0.218, 82810.0])

# Reaction 129
reaction('CH3O + H => CH2O + H2', [2.000000e+13, 0.0, 0.0])

# Reaction 130
reaction('CH2O + H2 => CH3O + H', [1.233000e+11, 0.664, 81270.0])

# Reaction 131
reaction('CH3O + HO2 => CH2O + H2O2', [3.010000e+11, 0.0, 0.0])

# Reaction 132
reaction('CH2O + H2O2 => CH3O + HO2', [1.074000e+12, -0.031, 65270.0])

# Reaction 133
falloff_reaction('CH2O + H (+ M) <=> CH2OH (+ M)',
                kf=[5.400000e+11, 0.454, 3600.0],
                kf0=[1.270000e+32, -4.82, 6530.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 CH4:2.0 C2H6:3.0',
                falloff=Troe(A=0.7187, T3=103.0, T1=1291.0, T2=4160.0))

# Reaction 134
reaction('CH2OH + O2 => CH2O + HO2', [1.510000e+15, -1.0, 0.0],
        options='duplicate')

# Reaction 135
reaction('CH2O + HO2 => CH2OH + O2', [1.975000e+14, -0.58, 20060.0],
        options='duplicate')

# Reaction 136
reaction('CH2OH + O2 => CH2O + HO2', [2.410000e+14, 0.0, 5017.0],
        options='duplicate')

# Reaction 137
reaction('CH2O + HO2 => CH2OH + O2', [3.152000e+13, 0.42, 25080.0],
        options='duplicate')

# Reaction 138
reaction('CH2OH + H => CH2O + H2', [6.000000e+12, 0.0, 0.0])

# Reaction 139
reaction('CH2O + H2 => CH2OH + H', [1.497000e+11, 0.768, 74750.0])

# Reaction 140
reaction('CH2OH + HO2 => CH2O + H2O2', [1.200000e+13, 0.0, 0.0])

# Reaction 141
reaction('CH2O + H2O2 => CH2OH + HO2', [1.732000e+14, 0.073, 58750.0])

# Reaction 142
reaction('CH2OH + HCO => CH2O + CH2O', [1.800000e+14, 0.0, 0.0])

# Reaction 143
reaction('CH2O + CH2O => CH2OH + HCO', [7.602000e+14, 0.481, 59560.0])

# Reaction 144
reaction('CH2OH + CH3O => CH2O + CH3OH', [2.400000e+13, 0.0, 0.0])

# Reaction 145
reaction('CH2O + CH3OH => CH2OH + CH3O', [1.285000e+13, 0.555, 74980.0])

# Reaction 146
reaction('CH2OH + CH2O => CH3OH + HCO', [1.878000e+04, 2.722, 4208.0])

# Reaction 147
reaction('CH3OH + HCO => CH2OH + CH2O', [9.630000e+03, 2.9, 13110.0])

# Reaction 148
reaction('OH + CH2OH => H2O + CH2O', [2.400000e+13, 0.0, 0.0])

# Reaction 149
reaction('H2O + CH2O => OH + CH2OH', [6.347000e+12, 0.662, 89640.0])

# Reaction 150
reaction('O + CH2OH => OH + CH2O', [4.200000e+13, 0.0, 0.0])

# Reaction 151
reaction('OH + CH2O => O + CH2OH', [5.438000e+11, 0.749, 73340.0])

# Reaction 152
reaction('CH2O + CH3OH => CH2OH + CH2OH', [6.498000e+12, 0.659, 68460.0])

# Reaction 153
reaction('CH2OH + CH2OH => CH2O + CH3OH', [3.000000e+12, 0.0, 0.0])

# Reaction 154
reaction('CH2OH + HO2 => HOCH2O + OH', [1.000000e+13, 0.0, 0.0])

# Reaction 155
reaction('HOCH2O + OH => CH2OH + HO2', [8.169000e+13, -0.024, 33470.0])

# Reaction 156
reaction('OCH2O2H => CH2O + HO2', [1.278000e+18, -1.8, 10460.0])

# Reaction 157
reaction('CH2O + HO2 => OCH2O2H', [1.500000e+11, 0.0, 11900.0])

# Reaction 158
reaction('OCH2O2H => HOCH2O2', [3.000000e+11, 0.0, 8600.0])

# Reaction 159
reaction('HOCH2O2 => OCH2O2H', [4.241000e+08, 0.95, 26200.0])

# Reaction 160
reaction('HOCH2O2 + HO2 => HOCH2O2H + O2', [3.500000e+10, 0.0, -3275.0])

# Reaction 161
reaction('HOCH2O2H + O2 => HOCH2O2 + HO2', [1.046000e+14, -0.84, 34870.0])

# Reaction 162
reaction('HOCH2O2H => HOCH2O + OH', [1.023000e+21, -1.92, 42490.0])

# Reaction 163
reaction('HOCH2O + OH => HOCH2O2H', [1.000000e+13, 0.0, 0.0])

# Reaction 164
falloff_reaction('CH3OH (+ M) <=> CH3 + OH (+ M)',
                kf=[1.900000e+16, 0.0, 91730.0],
                kf0=[2.950000e+44, -7.35, 95460.0],
                falloff=Troe(A=0.414, T3=279.0, T1=5459.0, T2=10000000000.0))

# Reaction 165
falloff_reaction('CH3OH (+ M) <=> CH2OH + H (+ M)',
                kf=[2.690000e+16, -0.08, 98940.0],
                kf0=[2.340000e+40, -6.33, 103100.0],
                falloff=Troe(A=0.773, T3=693.0, T1=5333.0, T2=10000000000.0))

# Reaction 166
reaction('CH3OH + H => CH3O + H2', [3.600000e+12, 0.0, 6095.0])

# Reaction 167
reaction('CH3O + H2 => CH3OH + H', [1.677000e+11, 0.212, 5868.0])

# Reaction 168
reaction('CH3OH + H => CH2OH + H2', [1.204000e+06, 2.4, 2583.0])

# Reaction 169
reaction('CH2OH + H2 => CH3OH + H', [1.386000e+04, 2.509, 8871.0])

# Reaction 170
reaction('CH3OH + O => CH2OH + OH', [3.880000e+05, 2.5, 3080.0])

# Reaction 171
reaction('CH2OH + OH => CH3OH + O', [2.319000e+03, 2.59, 7956.0])

# Reaction 172
reaction('CH3OH + OH => CH3O + H2O', [5.130000e+05, 2.13, 2450.0])

# Reaction 173
reaction('CH3O + H2O => CH3OH + OH', [2.534000e+05, 2.237, 17120.0])

# Reaction 174
reaction('CH3OH + OH => CH2OH + H2O', [1.440000e+06, 2.0, -839.0])

# Reaction 175
reaction('CH2OH + H2O => CH3OH + OH', [1.758000e+05, 2.003, 20340.0])

# Reaction 176
reaction('CH3OH + O2 => CH2OH + HO2', [2.050000e+13, 0.0, 44900.0])

# Reaction 177
reaction('CH2OH + HO2 => CH3OH + O2', [1.238000e+12, -0.239, -3501.0])

# Reaction 178
reaction('CH3OH + HO2 => CH2OH + H2O2', [1.080000e+04, 2.55, 10530.0])

# Reaction 179
reaction('CH2OH + H2O2 => CH3OH + HO2', [7.195000e+04, 1.963, 819.0])

# Reaction 180
reaction('CH3OH + CH3 => CH2OH + CH4', [3.190000e+01, 3.17, 7172.0])

# Reaction 181
reaction('CH2OH + CH4 => CH3OH + CH3', [3.351000e+02, 2.833, 15000.0])

# Reaction 182
reaction('CH3O + CH3OH => CH2OH + CH3OH', [3.000000e+11, 0.0, 4074.0])

# Reaction 183
reaction('CH2OH + CH3OH => CH3O + CH3OH', [7.416000e+10, -0.104, 10590.0])

# Reaction 184
reaction('CH3OH + CH2O => CH3O + CH3O', [7.981000e+12, 0.452, 81490.0])

# Reaction 185
reaction('CH3O + CH3O => CH3OH + CH2O', [6.030000e+13, 0.0, 0.0])

# Reaction 186
falloff_reaction('CH3 + H (+ M) <=> CH4 (+ M)',
                kf=[1.270000e+16, -0.6, 383.0],
                kf0=[2.477000e+33, -4.76, 2444.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0 HE:0.7',
                falloff=Troe(A=0.783, T3=74.0, T1=2940.0, T2=6960.0))

# Reaction 187
reaction('CH4 + H => CH3 + H2', [6.140000e+05, 2.5, 9587.0])

# Reaction 188
reaction('CH3 + H2 => CH4 + H', [6.730000e+02, 2.946, 8047.0])

# Reaction 189
reaction('CH4 + OH => CH3 + H2O', [5.830000e+04, 2.6, 2190.0])

# Reaction 190
reaction('CH3 + H2O => CH4 + OH', [6.776000e+02, 2.94, 15540.0])

# Reaction 191
reaction('CH4 + O => CH3 + OH', [1.020000e+09, 1.5, 8600.0])

# Reaction 192
reaction('CH3 + OH => CH4 + O', [5.804000e+05, 1.927, 5648.0])

# Reaction 193
reaction('CH4 + HO2 => CH3 + H2O2', [1.130000e+01, 3.74, 21010.0])

# Reaction 194
reaction('CH3 + H2O2 => CH4 + HO2', [7.166000e+00, 3.491, 3468.0])

# Reaction 195
reaction('CH4 + CH2 => CH3 + CH3', [2.460000e+06, 2.0, 8270.0])

# Reaction 196
reaction('CH3 + CH3 => CH4 + CH2', [1.736000e+06, 1.868, 12980.0])

# Reaction 197
reaction('CH3 + OH => CH2O + H2', [8.000000e+09, 0.5, -1755.0])

# Reaction 198
reaction('CH2O + H2 => CH3 + OH', [1.066000e+12, 0.322, 68210.0])

# Reaction 199
reaction('CH3 + OH => CH2(S) + H2O', [4.508000e+17, -1.34, 1417.0])

# Reaction 200
reaction('CH2(S) + H2O => CH3 + OH', [1.654000e+16, -0.855, 1039.0])

# Reaction 201
reaction('CH3 + OH => CH3O + H', [6.943000e+07, 1.343, 11200.0])

# Reaction 202
reaction('CH3O + H => CH3 + OH', [1.500000e+12, 0.5, -110.0])

# Reaction 203
reaction('CH3 + OH => CH2OH + H', [3.090000e+07, 1.596, 4506.0])

# Reaction 204
reaction('CH2OH + H => CH3 + OH', [1.650000e+11, 0.65, -284.0])

# Reaction 205
reaction('CH3 + OH => CH2 + H2O', [5.600000e+07, 1.6, 5420.0])

# Reaction 206
reaction('CH2 + H2O => CH3 + OH', [9.224000e+05, 2.072, 14060.0])

# Reaction 207
reaction('CH3 + HO2 => CH3O + OH', [1.000000e+12, 0.269, -687.5])

# Reaction 208
reaction('CH3O + OH => CH3 + HO2', [6.190000e+12, 0.147, 24550.0])

# Reaction 209
reaction('CH3 + HO2 => CH4 + O2', [1.160000e+05, 2.23, -3022.0])

# Reaction 210
reaction('CH4 + O2 => CH3 + HO2', [2.018000e+07, 2.132, 53210.0])

# Reaction 211
reaction('CH3 + O => CH2O + H', [5.540000e+13, 0.05, -136.0])

# Reaction 212
reaction('CH2O + H => CH3 + O', [3.830000e+15, -0.147, 68410.0])

# Reaction 213
reaction('CH3 + O2 => CH3O + O', [7.546000e+12, 0.0, 28320.0])

# Reaction 214
reaction('CH3O + O => CH3 + O2', [4.718000e+14, -0.451, 288.0])

# Reaction 215
reaction('CH3 + O2 => CH2O + OH', [2.641000e+00, 3.283, 8105.0])

# Reaction 216
reaction('CH2O + OH => CH3 + O2', [5.285000e-01, 3.477, 59920.0])

# Reaction 217
falloff_reaction('CH3 + O2 (+ M) <=> CH3O2 (+ M)',
                kf=[7.812000e+09, 0.9, 0.0],
                kf0=[6.850000e+24, -3.0, 0.0],
                falloff=Troe(A=0.6, T3=1000.0, T1=70.0, T2=1700.0))

# Reaction 218
reaction('CH3O2 + CH2O => CH3O2H + HCO', [1.990000e+12, 0.0, 11660.0])

# Reaction 219
reaction('CH3O2H + HCO => CH3O2 + CH2O', [1.323000e+14, -0.853, 9259.0])

# Reaction 220
reaction('CH4 + CH3O2 => CH3 + CH3O2H', [1.810000e+11, 0.0, 18480.0])

# Reaction 221
reaction('CH3 + CH3O2H => CH4 + CH3O2', [2.233000e+12, -0.694, -655.0])

# Reaction 222
reaction('CH3OH + CH3O2 => CH2OH + CH3O2H', [1.810000e+12, 0.0, 13710.0])

# Reaction 223
reaction('CH2OH + CH3O2H => CH3OH + CH3O2', [2.346000e+14, -1.031, 2404.0])

# Reaction 224
reaction('CH3O2 + CH3 => CH3O + CH3O', [5.080000e+12, 0.0, -1411.0])

# Reaction 225
reaction('CH3O + CH3O => CH3O2 + CH3', [1.967000e+12, 0.176, 28070.0])

# Reaction 226
reaction('CH3O2 + HO2 => CH3O2H + O2', [2.470000e+11, 0.0, -1570.0])

# Reaction 227
reaction('CH3O2H + O2 => CH3O2 + HO2', [5.302000e+14, -0.792, 35520.0])

# Reaction 228
reaction('CH3O2 + CH3O2 => CH2O + CH3OH + O2', [3.110000e+14, -1.61, -1051.0])

# Reaction 229
reaction('CH2O + CH3OH + O2 => CH3O2 + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 230
reaction('CH3O2 + CH3O2 => O2 + CH3O + CH3O', [1.400000e+16, -1.61, 1860.0])

# Reaction 231
reaction('O2 + CH3O + CH3O => CH3O2 + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 232
reaction('CH3O2 + H => CH3O + OH', [9.600000e+13, 0.0, 0.0])

# Reaction 233
reaction('CH3O + OH => CH3O2 + H', [1.720000e+09, 1.019, 40780.0])

# Reaction 234
reaction('CH3O2 + O => CH3O + O2', [3.600000e+13, 0.0, 0.0])

# Reaction 235
reaction('CH3O + O2 => CH3O2 + O', [2.229000e+11, 0.628, 57520.0])

# Reaction 236
reaction('CH3O2 + OH => CH3OH + O2', [6.000000e+13, 0.0, 0.0])

# Reaction 237
reaction('CH3OH + O2 => CH3O2 + OH', [1.536000e+13, 0.434, 59160.0])

# Reaction 238
reaction('CH3O2H => CH3O + OH', [6.310000e+14, 0.0, 42300.0])

# Reaction 239
reaction('CH3O + OH => CH3O2H', [2.514000e+06, 1.883, -2875.0])

# Reaction 240
reaction('CH2(S) => CH2', [1.000000e+13, 0.0, 0.0])

# Reaction 241
reaction('CH2 => CH2(S)', [4.488000e+12, -0.013, 9020.0])

# Reaction 242
reaction('CH2(S) + CH4 => CH3 + CH3', [1.600000e+13, 0.0, -570.0])

# Reaction 243
reaction('CH3 + CH3 => CH2(S) + CH4', [5.067000e+12, -0.145, 13160.0])

# Reaction 244
reaction('CH2(S) + O2 => CO + OH + H', [7.000000e+13, 0.0, 0.0])

# Reaction 245
reaction('CO + OH + H => CH2(S) + O2', [0.000000e+00, 0.0, 0.0])

# Reaction 246
reaction('CH2(S) + H2 => CH3 + H', [7.000000e+13, 0.0, 0.0])

# Reaction 247
reaction('CH3 + H => CH2(S) + H2', [2.022000e+16, -0.591, 15270.0])

# Reaction 248
reaction('CH2(S) + H => CH2 + H', [3.000000e+13, 0.0, 0.0])

# Reaction 249
reaction('CH2 + H => CH2(S) + H', [1.346000e+13, -0.013, 9020.0])

# Reaction 250
reaction('CH2(S) + H => CH + H2', [3.000000e+13, 0.0, 0.0])

# Reaction 251
reaction('CH + H2 => CH2(S) + H', [6.948000e+13, -0.253, 12480.0])

# Reaction 252
reaction('CH2(S) + O => CO + H + H', [3.000000e+13, 0.0, 0.0])

# Reaction 253
reaction('CO + H + H => CH2(S) + O', [0.000000e+00, 0.0, 0.0])

# Reaction 254
reaction('CH2(S) + OH => CH2O + H', [3.000000e+13, 0.0, 0.0])

# Reaction 255
reaction('CH2O + H => CH2(S) + OH', [1.154000e+18, -0.77, 85230.0])

# Reaction 256
reaction('CH2(S) + CO2 => CH2O + CO', [3.000000e+12, 0.0, 0.0])

# Reaction 257
reaction('CH2O + CO => CH2(S) + CO2', [4.366000e+10, 0.421, 59810.0])

# Reaction 258
falloff_reaction('CH2 + H (+ M) <=> CH3 (+ M)',
                kf=[2.500000e+16, -0.8, 0.0],
                kf0=[3.200000e+27, -3.14, 1230.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0 HE:0.7',
                falloff=Troe(A=0.68, T3=78.0, T1=1995.0, T2=5590.0))

# Reaction 259
reaction('CH2 + O2 => CH2O + O', [2.400000e+12, 0.0, 1500.0])

# Reaction 260
reaction('CH2O + O => CH2 + O2', [5.955000e+14, -0.365, 60980.0])

# Reaction 261
reaction('CH2 + O2 => CO2 + H + H', [5.800000e+12, 0.0, 1500.0])

# Reaction 262
reaction('CO2 + H + H => CH2 + O2', [0.000000e+00, 0.0, 0.0])

# Reaction 263
reaction('CH2 + O2 => CO + OH + H', [5.000000e+12, 0.0, 1500.0])

# Reaction 264
reaction('CO + OH + H => CH2 + O2', [0.000000e+00, 0.0, 0.0])

# Reaction 265
reaction('CH2 + O => CO + H + H', [5.000000e+13, 0.0, 0.0])

# Reaction 266
reaction('CO + H + H => CH2 + O', [0.000000e+00, 0.0, 0.0])

# Reaction 267
reaction('CH2 + H => CH + H2', [1.000000e+18, -1.56, 0.0],
        options='duplicate')

# Reaction 268
reaction('CH + H2 => CH2 + H', [5.160000e+18, -1.8, 3460.0],
        options='duplicate')

# Reaction 269
reaction('CH2 + OH => CH + H2O', [1.130000e+07, 2.0, 3000.0])

# Reaction 270
reaction('CH + H2O => CH2 + OH', [6.183000e+08, 1.655, 21350.0])

# Reaction 271
reaction('CH + O2 => HCO + O', [3.300000e+13, 0.0, 0.0])

# Reaction 272
reaction('HCO + O => CH + O2', [9.371000e+12, 0.161, 71210.0])

# Reaction 273
reaction('C + OH => CO + H', [5.000000e+13, 0.0, 0.0])

# Reaction 274
reaction('CO + H => C + OH', [1.356000e+15, 0.0, 154300.0])

# Reaction 275
reaction('C + O2 => CO + O', [5.000000e+13, 0.0, 0.0])

# Reaction 276
reaction('CO + O => C + O2', [1.070000e+14, 0.0, 138200.0])

# Reaction 277
reaction('CH + H => C + H2', [5.000000e+13, 0.0, 0.0])

# Reaction 278
reaction('C + H2 => CH + H', [2.043000e+14, 0.0, 23820.0])

# Reaction 279
reaction('CH + O => CO + H', [5.700000e+13, 0.0, 0.0])

# Reaction 280
reaction('CO + H => CH + O', [2.774000e+15, 0.0, 176000.0])

# Reaction 281
reaction('CH + OH => HCO + H', [3.000000e+13, 0.0, 0.0])

# Reaction 282
reaction('HCO + H => CH + OH', [5.069000e+14, 0.0, 88110.0])

# Reaction 283
reaction('CH2 + H => CH + H2', [2.700000e+11, 0.67, 25700.0],
        options='duplicate')

# Reaction 284
reaction('CH + H2 => CH2 + H', [1.897000e+11, 0.67, 28730.0],
        options='duplicate')

# Reaction 285
reaction('CH + H2O => H + CH2O', [1.713000e+13, 0.0, -755.0])

# Reaction 286
reaction('H + CH2O => CH + H2O', [8.372000e+14, 0.0, 57520.0])

# Reaction 287
reaction('CH + CO2 => HCO + CO', [1.700000e+12, 0.0, 685.0])

# Reaction 288
reaction('HCO + CO => CH + CO2', [2.565000e+11, 0.0, 66460.0])

# Reaction 289
falloff_reaction('CH3 + CH3 (+ M) <=> C2H6 (+ M)',
                kf=[9.214000e+16, -1.17, 635.8],
                kf0=[1.135000e+36, -5.246, 1705.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0 HE:0.7',
                falloff=Troe(A=0.405, T3=1120.0, T1=69.6, T2=10000000000.0))

# Reaction 290
falloff_reaction('C2H5 + H (+ M) <=> C2H6 (+ M)',
                kf=[5.210000e+17, -0.99, 1580.0],
                kf0=[1.990000e+41, -7.08, 6685.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0 HE:0.7',
                falloff=Troe(A=0.842, T3=125.0, T1=2219.0, T2=6882.0))

# Reaction 291
reaction('C2H6 + H => C2H5 + H2', [1.150000e+08, 1.9, 7530.0])

# Reaction 292
reaction('C2H5 + H2 => C2H6 + H', [1.062000e+04, 2.582, 9760.0])

# Reaction 293
reaction('C2H6 + O => C2H5 + OH', [3.550000e+06, 2.4, 5830.0])

# Reaction 294
reaction('C2H5 + OH => C2H6 + O', [1.702000e+02, 3.063, 6648.0])

# Reaction 295
reaction('C2H6 + OH => C2H5 + H2O', [1.480000e+07, 1.9, 950.0])

# Reaction 296
reaction('C2H5 + H2O => C2H6 + OH', [1.450000e+04, 2.476, 18070.0])

# Reaction 297
reaction('C2H6 + O2 => C2H5 + HO2', [6.030000e+13, 0.0, 51870.0])

# Reaction 298
reaction('C2H5 + HO2 => C2H6 + O2', [2.921000e+10, 0.334, -593.0])

# Reaction 299
reaction('C2H6 + CH3 => C2H5 + CH4', [1.510000e-07, 6.0, 6047.0])

# Reaction 300
reaction('C2H5 + CH4 => C2H6 + CH3', [1.273000e-08, 6.236, 9817.0])

# Reaction 301
reaction('C2H6 + HO2 => C2H5 + H2O2', [3.460000e+01, 3.61, 16920.0])

# Reaction 302
reaction('C2H5 + H2O2 => C2H6 + HO2', [1.849000e+00, 3.597, 3151.0])

# Reaction 303
reaction('C2H6 + CH3O2 => C2H5 + CH3O2H', [1.940000e+01, 3.64, 17100.0])

# Reaction 304
reaction('C2H5 + CH3O2H => C2H6 + CH3O2', [2.017000e+01, 3.182, 1734.0])

# Reaction 305
reaction('C2H6 + CH3O => C2H5 + CH3OH', [2.410000e+11, 0.0, 7090.0])

# Reaction 306
reaction('C2H5 + CH3OH => C2H6 + CH3O', [4.779000e+08, 0.469, 9547.0])

# Reaction 307
reaction('C2H6 + CH => C2H5 + CH2', [1.100000e+14, 0.0, -260.0])

# Reaction 308
reaction('C2H5 + CH2 => C2H6 + CH', [1.969000e+09, 0.921, -1490.0])

# Reaction 309
reaction('CH2(S) + C2H6 => CH3 + C2H5', [1.200000e+14, 0.0, 0.0])

# Reaction 310
reaction('CH3 + C2H5 => CH2(S) + C2H6', [3.203000e+12, 0.091, 17500.0])

# Reaction 311
falloff_reaction('C2H4 + H (+ M) <=> C2H5 (+ M)',
                kf=[1.081000e+12, 0.454, 1822.0],
                kf0=[1.200000e+42, -7.62, 6970.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0 HE:0.7',
                falloff=Troe(A=0.975, T3=210.0, T1=984.0, T2=4374.0))

# Reaction 312
reaction('H2 + CH3O2 => H + CH3O2H', [1.500000e+14, 0.0, 26030.0])

# Reaction 313
reaction('H + CH3O2H => H2 + CH3O2', [1.688000e+18, -1.14, 8434.0])

# Reaction 314
reaction('H2 + C2H5O2 => H + C2H5O2H', [1.500000e+14, 0.0, 26030.0])

# Reaction 315
reaction('H + C2H5O2H => H2 + C2H5O2', [1.691000e+18, -1.14, 8438.0])

# Reaction 316
reaction('C2H5 + C2H3 => C2H4 + C2H4', [6.859000e+11, 0.11, -4300.0])

# Reaction 317
reaction('C2H4 + C2H4 => C2H5 + C2H3', [4.820000e+14, 0.0, 71530.0])

# Reaction 318
reaction('CH3 + C2H5 => CH4 + C2H4', [1.180000e+04, 2.45, -2921.0])

# Reaction 319
reaction('CH4 + C2H4 => CH3 + C2H5', [2.390000e+06, 2.4, 66690.0])

# Reaction 320
reaction('C2H5 + H => CH3 + CH3', [9.690000e+13, 0.0, 220.0])

# Reaction 321
reaction('CH3 + CH3 => C2H5 + H', [2.029000e+09, 1.028, 10510.0])

# Reaction 322
reaction('C2H5 + H => C2H4 + H2', [2.000000e+12, 0.0, 0.0])

# Reaction 323
reaction('C2H4 + H2 => C2H5 + H', [4.440000e+11, 0.396, 68070.0])

# Reaction 324
reaction('C2H5 + O => CH3CHO + H', [1.100000e+14, 0.0, 0.0])

# Reaction 325
reaction('CH3CHO + H => C2H5 + O', [1.033000e+17, -0.5, 77420.0])

# Reaction 326
reaction('C2H5 + HO2 => C2H5O + OH', [1.100000e+13, 0.0, 0.0])

# Reaction 327
reaction('C2H5O + OH => C2H5 + HO2', [9.680000e+15, -0.723, 27650.0])

# Reaction 328
reaction('CH3O2 + C2H5 => CH3O + C2H5O', [8.000000e+12, 0.0, -1000.0])

# Reaction 329
reaction('CH3O + C2H5O => CH3O2 + C2H5', [4.404000e+14, -0.425, 30890.0])

# Reaction 330
reaction('C2H5O + O2 => CH3CHO + HO2', [4.280000e+10, 0.0, 1097.0])

# Reaction 331
reaction('CH3CHO + HO2 => C2H5O + O2', [1.322000e+08, 0.615, 34130.0])

# Reaction 332
reaction('C2H5O => CH3 + CH2O', [1.321000e+20, -2.018, 20750.0])

# Reaction 333
reaction('CH3 + CH2O => C2H5O', [3.000000e+11, 0.0, 6336.0])

# Reaction 334
reaction('C2H5O => CH3CHO + H', [5.428000e+15, -0.687, 22230.0])

# Reaction 335
reaction('CH3CHO + H => C2H5O', [8.000000e+12, 0.0, 6400.0])

# Reaction 336
reaction('C2H5O2 => C2H5 + O2', [1.312000e+62, -14.784, 49180.0])

# Reaction 337
reaction('C2H5 + O2 => C2H5O2', [2.876000e+56, -13.82, 14620.0])

# Reaction 338
reaction('C2H5O2 + CH2O => C2H5O2H + HCO', [1.990000e+12, 0.0, 11660.0])

# Reaction 339
reaction('C2H5O2H + HCO => C2H5O2 + CH2O', [1.325000e+14, -0.853, 9263.0])

# Reaction 340
reaction('CH4 + C2H5O2 => CH3 + C2H5O2H', [1.810000e+11, 0.0, 18480.0])

# Reaction 341
reaction('CH3 + C2H5O2H => CH4 + C2H5O2', [2.237000e+12, -0.694, -651.0])

# Reaction 342
reaction('CH3OH + C2H5O2 => CH2OH + C2H5O2H', [1.810000e+12, 0.0, 13710.0])

# Reaction 343
reaction('CH2OH + C2H5O2H => CH3OH + C2H5O2', [2.350000e+14, -1.031, 2408.0])

# Reaction 344
reaction('C2H5O2 + HO2 => C2H5O2H + O2', [1.750000e+10, 0.0, -3275.0])

# Reaction 345
reaction('C2H5O2H + O2 => C2H5O2 + HO2', [3.763000e+13, -0.792, 33820.0])

# Reaction 346
reaction('C2H6 + C2H5O2 => C2H5 + C2H5O2H', [8.600000e+00, 3.76, 17200.0])

# Reaction 347
reaction('C2H5 + C2H5O2H => C2H6 + C2H5O2', [8.957000e+00, 3.302, 1838.0])

# Reaction 348
reaction('C2H5O2H => C2H5O + OH', [6.310000e+14, 0.0, 42300.0])

# Reaction 349
reaction('C2H5O + OH => C2H5O2H', [5.661000e+08, 1.033, -1705.0])

# Reaction 350
reaction('C2H4O2H => C2H5 + O2', [4.374000e+47, -12.115, 31020.0])

# Reaction 351
reaction('C2H5 + O2 => C2H4O2H', [1.814000e+45, -11.5, 14600.0])

# Reaction 352
reaction('C2H5 + O2 => C2H4 + HO2', [7.561000e+14, -1.01, 4749.0],
        options='duplicate')

# Reaction 353
reaction('C2H4 + HO2 => C2H5 + O2', [8.802000e+14, -0.962, 18130.0],
        options='duplicate')

# Reaction 354
reaction('C2H5 + O2 => C2H4 + HO2', [4.000000e-01, 3.88, 13620.0],
        options='duplicate')

# Reaction 355
reaction('C2H4 + HO2 => C2H5 + O2', [4.656000e-01, 3.928, 27000.0],
        options='duplicate')

# Reaction 356
reaction('C2H5 + O2 => C2H4O1-2 + OH', [1.626000e+11, -0.31, 6150.0])

# Reaction 357
reaction('C2H4O1-2 + OH => C2H5 + O2', [3.633000e+13, -0.626, 39840.0])

# Reaction 358
reaction('C2H5 + O2 => CH3CHO + OH', [8.265000e+02, 2.41, 5285.0])

# Reaction 359
reaction('CH3CHO + OH => C2H5 + O2', [2.247000e+03, 2.301, 65970.0])

# Reaction 360
reaction('C2H5O2 => C2H4O2H', [2.276000e+39, -8.479, 45170.0])

# Reaction 361
reaction('C2H4O2H => C2H5O2', [1.203000e+36, -8.13, 27020.0])

# Reaction 362
reaction('C2H5O2 => CH3CHO + OH', [2.520000e+41, -10.2, 43710.0])

# Reaction 363
reaction('CH3CHO + OH => C2H5O2', [1.502000e+36, -9.345, 69840.0])

# Reaction 364
reaction('C2H5O2 => C2H4 + HO2', [1.815000e+38, -8.45, 37890.0])

# Reaction 365
reaction('C2H4 + HO2 => C2H5O2', [4.632000e+32, -7.438, 16700.0])

# Reaction 366
reaction('C2H5O2 => C2H4O1-2 + OH', [4.000000e+43, -10.46, 45580.0])

# Reaction 367
reaction('C2H4O1-2 + OH => C2H5O2', [1.959000e+40, -9.812, 44710.0])

# Reaction 368
reaction('C2H4O2H => C2H4O1-2 + OH', [8.848000e+30, -6.08, 20660.0])

# Reaction 369
reaction('C2H4O1-2 + OH => C2H4O2H', [8.199000e+30, -5.781, 37930.0])

# Reaction 370
reaction('C2H4O2H => C2H4 + HO2', [3.980000e+34, -7.25, 23250.0])

# Reaction 371
reaction('C2H4 + HO2 => C2H4O2H', [1.922000e+32, -6.587, 20210.0])

# Reaction 372
reaction('C2H4O2H => CH3CHO + OH', [1.188000e+34, -9.02, 29210.0])

# Reaction 373
reaction('CH3CHO + OH => C2H4O2H', [1.339000e+32, -8.514, 73480.0])

# Reaction 374
reaction('C2H4O1-2 => CH3 + HCO', [3.630000e+13, 0.0, 57200.0])

# Reaction 375
reaction('CH3 + HCO => C2H4O1-2', [1.006000e+04, 1.549, -2750.0])

# Reaction 376
reaction('C2H4O1-2 => CH3CHO', [7.407000e+12, 0.0, 53800.0])

# Reaction 377
reaction('CH3CHO => C2H4O1-2', [9.013000e+10, 0.207, 80800.0])

# Reaction 378
reaction('C2H4O1-2 + OH => C2H3O1-2 + H2O', [1.780000e+13, 0.0, 3610.0])

# Reaction 379
reaction('C2H3O1-2 + H2O => C2H4O1-2 + OH', [1.347000e+10, 0.693, 24740.0])

# Reaction 380
reaction('C2H4O1-2 + H => C2H3O1-2 + H2', [8.000000e+13, 0.0, 9680.0])

# Reaction 381
reaction('C2H3O1-2 + H2 => C2H4O1-2 + H', [5.710000e+09, 0.799, 15920.0])

# Reaction 382
reaction('C2H4O1-2 + HO2 => C2H3O1-2 + H2O2', [1.130000e+13, 0.0, 30430.0])

# Reaction 383
reaction('C2H3O1-2 + H2O2 => C2H4O1-2 + HO2', [4.666000e+11, 0.104, 20670.0])

# Reaction 384
reaction('C2H4O1-2 + CH3O2 => C2H3O1-2 + CH3O2H', [1.130000e+13, 0.0, 30430.0])

# Reaction 385
reaction('C2H3O1-2 + CH3O2H => C2H4O1-2 + CH3O2', [9.078000e+12, -0.341, 19070.0])

# Reaction 386
reaction('C2H4O1-2 + C2H5O2 => C2H3O1-2 + C2H5O2H', [1.130000e+13, 0.0, 30430.0])

# Reaction 387
reaction('C2H3O1-2 + C2H5O2H => C2H4O1-2 + C2H5O2', [9.093000e+12, -0.341, 19080.0])

# Reaction 388
reaction('C2H4O1-2 + CH3 => C2H3O1-2 + CH4', [1.070000e+12, 0.0, 11830.0])

# Reaction 389
reaction('C2H3O1-2 + CH4 => C2H4O1-2 + CH3', [6.967000e+10, 0.353, 19610.0])

# Reaction 390
reaction('C2H4O1-2 + CH3O => C2H3O1-2 + CH3OH', [1.200000e+11, 0.0, 6750.0])

# Reaction 391
reaction('C2H3O1-2 + CH3OH => C2H4O1-2 + CH3O', [1.839000e+08, 0.586, 13220.0])

# Reaction 392
reaction('C2H3O1-2 => CH3CO', [8.500000e+14, 0.0, 14000.0])

# Reaction 393
reaction('CH3CO => C2H3O1-2', [1.002000e+14, 0.041, 48710.0])

# Reaction 394
reaction('C2H3O1-2 => CH2CHO', [1.000000e+14, 0.0, 14000.0])

# Reaction 395
reaction('CH2CHO => C2H3O1-2', [1.245000e+15, -0.375, 44010.0])

# Reaction 396
reaction('CH3CHO => CH3 + HCO', [7.687000e+20, -1.342, 86950.0])

# Reaction 397
reaction('CH3 + HCO => CH3CHO', [1.750000e+13, 0.0, 0.0])

# Reaction 398
reaction('CH3CHO + H => CH3CO + H2', [2.370000e+13, 0.0, 3642.0])

# Reaction 399
reaction('CH3CO + H2 => CH3CHO + H', [1.639000e+10, 0.633, 17600.0])

# Reaction 400
reaction('CH3CHO + O => CH3CO + OH', [5.940000e+12, 0.0, 1868.0])

# Reaction 401
reaction('CH3CO + OH => CH3CHO + O', [2.133000e+09, 0.614, 14410.0])

# Reaction 402
reaction('CH3CHO + OH => CH3CO + H2O', [3.370000e+12, 0.0, -619.0])

# Reaction 403
reaction('CH3CO + H2O => CH3CHO + OH', [2.472000e+10, 0.527, 28230.0])

# Reaction 404
reaction('CH3CHO + O2 => CH3CO + HO2', [3.010000e+13, 0.0, 39150.0])

# Reaction 405
reaction('CH3CO + HO2 => CH3CHO + O2', [1.092000e+11, 0.285, -1588.0])

# Reaction 406
reaction('CH3CHO + CH3 => CH3CO + CH4', [7.080000e-04, 4.58, 1966.0])

# Reaction 407
reaction('CH3CO + CH4 => CH3CHO + CH3', [4.468000e-04, 4.767, 17460.0])

# Reaction 408
reaction('CH3CHO + HO2 => CH3CO + H2O2', [3.010000e+12, 0.0, 11920.0])

# Reaction 409
reaction('CH3CO + H2O2 => CH3CHO + HO2', [1.205000e+12, -0.062, 9877.0])

# Reaction 410
reaction('CH3O2 + CH3CHO => CH3O2H + CH3CO', [3.010000e+12, 0.0, 11920.0])

# Reaction 411
reaction('CH3O2H + CH3CO => CH3O2 + CH3CHO', [2.344000e+13, -0.507, 8282.0])

# Reaction 412
reaction('CH3CHO + CH3CO3 => CH3CO + CH3CO3H', [3.010000e+12, 0.0, 11920.0])

# Reaction 413
reaction('CH3CO + CH3CO3H => CH3CHO + CH3CO3', [1.922000e+12, -0.01, 12650.0])

# Reaction 414
reaction('CH3CHO + OH => CH3 + HOCHO', [3.000000e+15, -1.076, 0.0])

# Reaction 415
reaction('CH3 + HOCHO => CH3CHO + OH', [2.371000e+16, -1.277, 23750.0])

# Reaction 416
reaction('CH3CHO + OH => CH2CHO + H2O', [1.720000e+05, 2.4, 815.0])

# Reaction 417
reaction('CH2CHO + H2O => CH3CHO + OH', [1.332000e+05, 2.511, 24950.0])

# Reaction 418
falloff_reaction('CH3CO (+ M) <=> CH3 + CO (+ M)',
                kf=[3.000000e+12, 0.0, 16720.0],
                kf0=[1.200000e+15, 0.0, 12518.0])

# Reaction 419
reaction('CH3CO + H => CH2CO + H2', [2.000000e+13, 0.0, 0.0])

# Reaction 420
reaction('CH2CO + H2 => CH3CO + H', [1.037000e+13, 0.201, 60560.0])

# Reaction 421
reaction('CH3CO + O => CH2CO + OH', [2.000000e+13, 0.0, 0.0])

# Reaction 422
reaction('CH2CO + OH => CH3CO + O', [5.381000e+12, 0.182, 59140.0])

# Reaction 423
reaction('CH3CO + CH3 => CH2CO + CH4', [5.000000e+13, 0.0, 0.0])

# Reaction 424
reaction('CH2CO + CH4 => CH3CO + CH3', [2.364000e+16, -0.245, 62100.0])

# Reaction 425
reaction('CH3CO3 => CH3CO + O2', [6.863000e+19, -1.949, 38530.0])

# Reaction 426
reaction('CH3CO + O2 => CH3CO3', [1.200000e+11, 0.0, -1100.0])

# Reaction 427
reaction('CH3CO3 + HO2 => CH3CO3H + O2', [1.750000e+10, 0.0, -3275.0])

# Reaction 428
reaction('CH3CO3H + O2 => CH3CO3 + HO2', [3.080000e+12, -0.294, 38180.0])

# Reaction 429
reaction('H2O2 + CH3CO3 => HO2 + CH3CO3H', [2.410000e+12, 0.0, 9936.0])

# Reaction 430
reaction('HO2 + CH3CO3H => H2O2 + CH3CO3', [3.845000e+12, 0.053, 12710.0])

# Reaction 431
reaction('CH4 + CH3CO3 => CH3 + CH3CO3H', [1.810000e+11, 0.0, 18480.0])

# Reaction 432
reaction('CH3 + CH3CO3H => CH4 + CH3CO3', [1.831000e+11, -0.196, 3711.0])

# Reaction 433
reaction('CH2O + CH3CO3 => HCO + CH3CO3H', [1.990000e+12, 0.0, 11660.0])

# Reaction 434
reaction('HCO + CH3CO3H => CH2O + CH3CO3', [1.085000e+13, -0.356, 13620.0])

# Reaction 435
reaction('C2H6 + CH3CO3 => C2H5 + CH3CO3H', [1.700000e+13, 0.0, 20460.0])

# Reaction 436
reaction('C2H5 + CH3CO3H => C2H6 + CH3CO3', [1.450000e+12, 0.04, 9460.0])

# Reaction 437
reaction('CH3CO3H => CH3CO2 + OH', [5.010000e+14, 0.0, 40150.0])

# Reaction 438
reaction('CH3CO2 + OH => CH3CO3H', [3.618000e+07, 1.761, 1338.0])

# Reaction 439
three_body_reaction('CH3CO2 + M => CH3 + CO2 + M', [4.400000e+15, 0.0, 10500.0])

# Reaction 440
three_body_reaction('CH3 + CO2 + M => CH3CO2 + M', [4.548000e+08, 1.378, 17520.0])

# Reaction 441
reaction('CH2CHO => CH2CO + H', [4.071000e+15, -0.342, 50600.0])

# Reaction 442
reaction('CH2CO + H => CH2CHO', [5.000000e+13, 0.0, 12300.0])

# Reaction 443
reaction('CH2CHO + O2 => CH2O + CO + OH', [8.950000e+13, -0.6, 10120.0])

# Reaction 444
reaction('CH2O + CO + OH => CH2CHO + O2', [0.000000e+00, 0.0, 0.0])

# Reaction 445
falloff_reaction('CH2 + CO (+ M) <=> CH2CO (+ M)',
                kf=[8.100000e+11, 0.0, 0.0],
                kf0=[2.690000e+33, -5.11, 7095.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0 HE:0.7',
                falloff=Troe(A=0.5907, T3=275.0, T1=1226.0, T2=5185.0))

# Reaction 446
reaction('CH2CO + H => CH3 + CO', [1.100000e+13, 0.0, 3400.0])

# Reaction 447
reaction('CH3 + CO => CH2CO + H', [2.400000e+12, 0.0, 40200.0])

# Reaction 448
reaction('CH2CO + H => HCCO + H2', [2.000000e+14, 0.0, 8000.0])

# Reaction 449
reaction('HCCO + H2 => CH2CO + H', [1.434000e+11, 0.47, 4520.0])

# Reaction 450
reaction('CH2CO + O => CH2 + CO2', [1.750000e+12, 0.0, 1350.0])

# Reaction 451
reaction('CH2 + CO2 => CH2CO + O', [2.854000e+09, 0.809, 49440.0])

# Reaction 452
reaction('CH2CO + O => HCCO + OH', [1.000000e+13, 0.0, 8000.0])

# Reaction 453
reaction('HCCO + OH => CH2CO + O', [3.723000e+09, 0.452, 3108.0])

# Reaction 454
reaction('CH2CO + OH => HCCO + H2O', [1.000000e+13, 0.0, 2000.0])

# Reaction 455
reaction('HCCO + H2O => CH2CO + OH', [7.604000e+10, 0.365, 13410.0])

# Reaction 456
reaction('CH2CO + OH => CH2OH + CO', [2.000000e+12, 0.0, -1010.0])

# Reaction 457
reaction('CH2OH + CO => CH2CO + OH', [8.170000e+09, 0.494, 24530.0])

# Reaction 458
reaction('CH2(S) + CH2CO => C2H4 + CO', [1.600000e+14, 0.0, 0.0])

# Reaction 459
reaction('C2H4 + CO => CH2(S) + CH2CO', [3.750000e+14, 0.217, 103400.0])

# Reaction 460
reaction('HCCO + OH => H2 + CO + CO', [1.000000e+14, 0.0, 0.0])

# Reaction 461
reaction('H2 + CO + CO => HCCO + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 462
reaction('H + HCCO => CH2(S) + CO', [1.100000e+13, 0.0, 0.0])

# Reaction 463
reaction('CH2(S) + CO => H + HCCO', [4.061000e+07, 1.561, 18540.0])

# Reaction 464
reaction('HCCO + O => H + CO + CO', [8.000000e+13, 0.0, 0.0])

# Reaction 465
reaction('H + CO + CO => HCCO + O', [0.000000e+00, 0.0, 0.0])

# Reaction 466
reaction('HCCO + O2 => OH + CO + CO', [4.200000e+10, 0.0, 850.0])

# Reaction 467
reaction('OH + CO + CO => HCCO + O2', [0.000000e+00, 0.0, 0.0])

# Reaction 468
three_body_reaction('HCCO + M => CH + CO + M', [6.500000e+15, 0.0, 58820.0])

# Reaction 469
three_body_reaction('CH + CO + M => HCCO + M', [1.391000e+11, 1.033, -13720.0])

# Reaction 470
reaction('CH + CH2O => H + CH2CO', [9.460000e+13, 0.0, -515.0])

# Reaction 471
reaction('H + CH2CO => CH + CH2O', [1.623000e+15, 0.0, 69060.0])

# Reaction 472
reaction('CH + HCCO => CO + C2H2', [5.000000e+13, 0.0, 0.0])

# Reaction 473
reaction('CO + C2H2 => CH + HCCO', [1.721000e+17, 0.0, 164600.0])

# Reaction 474
falloff_reaction('C2H3 + H (+ M) <=> C2H4 (+ M)',
                kf=[1.360000e+14, 0.173, 660.0],
                kf0=[1.400000e+30, -3.86, 3320.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0 HE:0.7',
                falloff=Troe(A=0.782, T3=207.5, T1=2663.0, T2=6095.0))

# Reaction 475
falloff_reaction('C2H4 (+ M) <=> C2H2 + H2 (+ M)',
                kf=[8.000000e+12, 0.44, 88770.0],
                kf0=[1.580000e+51, -9.3, 97800.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0 HE:0.7',
                falloff=Troe(A=0.735, T3=180.0, T1=1035.0, T2=5417.0))

# Reaction 476
reaction('C2H4 + H => C2H3 + H2', [5.070000e+07, 1.93, 12950.0])

# Reaction 477
reaction('C2H3 + H2 => C2H4 + H', [1.602000e+04, 2.436, 5190.0])

# Reaction 478
reaction('C2H4 + O => CH3 + HCO', [8.564000e+06, 1.88, 183.0])

# Reaction 479
reaction('CH3 + HCO => C2H4 + O', [3.297000e+02, 2.602, 26140.0])

# Reaction 480
reaction('C2H4 + O => CH2CHO + H', [4.986000e+06, 1.88, 183.0])

# Reaction 481
reaction('CH2CHO + H => C2H4 + O', [1.541000e+09, 1.201, 18780.0])

# Reaction 482
reaction('C2H4 + OH => C2H3 + H2O', [1.800000e+06, 2.0, 2500.0])

# Reaction 483
reaction('C2H3 + H2O => C2H4 + OH', [6.029000e+03, 2.4, 9632.0])

# Reaction 484
reaction('C2H4 + CH3 => C2H3 + CH4', [6.620000e+00, 3.7, 9500.0])

# Reaction 485
reaction('C2H3 + CH4 => C2H4 + CH3', [1.908000e+00, 3.76, 3280.0])

# Reaction 486
reaction('C2H4 + O2 => C2H3 + HO2', [4.000000e+13, 0.0, 58200.0])

# Reaction 487
reaction('C2H3 + HO2 => C2H4 + O2', [6.626000e+10, 0.158, -4249.0])

# Reaction 488
reaction('C2H4 + CH3O => C2H3 + CH3OH', [1.200000e+11, 0.0, 6750.0])

# Reaction 489
reaction('C2H3 + CH3OH => C2H4 + CH3O', [8.138000e+08, 0.293, -783.0])

# Reaction 490
reaction('C2H4 + CH3O2 => C2H3 + CH3O2H', [2.230000e+12, 0.0, 17190.0])

# Reaction 491
reaction('C2H3 + CH3O2H => C2H4 + CH3O2', [7.929000e+12, -0.634, -8167.0])

# Reaction 492
reaction('C2H4 + C2H5O2 => C2H3 + C2H5O2H', [2.230000e+12, 0.0, 17190.0])

# Reaction 493
reaction('C2H3 + C2H5O2H => C2H4 + C2H5O2', [7.943000e+12, -0.634, -8163.0])

# Reaction 494
reaction('C2H4 + CH3CO3 => C2H3 + CH3CO3H', [1.130000e+13, 0.0, 30430.0])

# Reaction 495
reaction('C2H3 + CH3CO3H => C2H4 + CH3CO3', [3.295000e+12, -0.136, 9440.0])

# Reaction 496
reaction('C2H4 + CH3O2 => C2H4O1-2 + CH3O', [2.820000e+12, 0.0, 17110.0])

# Reaction 497
reaction('C2H4O1-2 + CH3O => C2H4 + CH3O2', [3.385000e+13, -0.065, 41660.0])

# Reaction 498
reaction('C2H4 + C2H5O2 => C2H4O1-2 + C2H5O', [2.820000e+12, 0.0, 17110.0])

# Reaction 499
reaction('C2H4O1-2 + C2H5O => C2H4 + C2H5O2', [7.638000e+15, -0.916, 42830.0])

# Reaction 500
reaction('C2H4 + HO2 => C2H4O1-2 + OH', [2.230000e+12, 0.0, 17190.0])

# Reaction 501
reaction('C2H4O1-2 + OH => C2H4 + HO2', [4.280000e+14, -0.364, 37500.0])

# Reaction 502
reaction('CH + CH4 => C2H4 + H', [6.000000e+13, 0.0, 0.0])

# Reaction 503
reaction('C2H4 + H => CH + CH4', [3.573000e+14, 0.0, 55480.0])

# Reaction 504
reaction('CH2(S) + CH3 => C2H4 + H', [2.000000e+13, 0.0, 0.0])

# Reaction 505
reaction('C2H4 + H => CH2(S) + CH3', [6.128000e+19, -1.223, 73050.0])

# Reaction 506
falloff_reaction('C2H2 + H (+ M) <=> C2H3 (+ M)',
                kf=[5.600000e+12, 0.0, 2400.0],
                kf0=[3.800000e+40, -7.27, 7220.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0 HE:0.7',
                falloff=Troe(A=0.751, T3=98.5, T1=1302.0, T2=4167.0))

# Reaction 507
reaction('C2H3 + O2 => C2H2 + HO2', [2.120000e-06, 6.0, 9484.0])

# Reaction 508
reaction('C2H2 + HO2 => C2H3 + O2', [1.087000e-05, 5.905, 24030.0])

# Reaction 509
reaction('C2H3 + O2 => CH2O + HCO', [8.500000e+28, -5.312, 6500.0])

# Reaction 510
reaction('CH2O + HCO => C2H3 + O2', [3.994000e+27, -4.883, 93450.0])

# Reaction 511
reaction('C2H3 + O2 => CH2CHO + O', [5.500000e+14, -0.611, 5260.0])

# Reaction 512
reaction('CH2CHO + O => C2H3 + O2', [3.000000e+18, -1.386, 16300.0])

# Reaction 513
reaction('CH3 + C2H3 => CH4 + C2H2', [3.920000e+11, 0.0, 0.0])

# Reaction 514
reaction('CH4 + C2H2 => CH3 + C2H3', [3.497000e+14, -0.193, 70780.0])

# Reaction 515
reaction('C2H3 + H => C2H2 + H2', [9.640000e+13, 0.0, 0.0])

# Reaction 516
reaction('C2H2 + H2 => C2H3 + H', [9.427000e+13, 0.253, 69240.0])

# Reaction 517
reaction('C2H3 + OH => C2H2 + H2O', [5.000000e+12, 0.0, 0.0])

# Reaction 518
reaction('C2H2 + H2O => C2H3 + OH', [5.184000e+13, 0.147, 84130.0])

# Reaction 519
falloff_reaction('C2H + H (+ M) <=> C2H2 (+ M)',
                kf=[1.000000e+17, 0.0, 0.0],
                kf0=[3.750000e+33, -4.8, 1900.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0 HE:0.7',
                falloff=Troe(A=0.646, T3=132.0, T1=1315.0, T2=5566.0))

# Reaction 520
reaction('C2H2 + O2 => HCCO + OH', [2.000000e+08, 1.5, 30100.0])

# Reaction 521
reaction('HCCO + OH => C2H2 + O2', [2.039000e+06, 1.541, 32270.0])

# Reaction 522
reaction('O + C2H2 => C2H + OH', [4.600000e+19, -1.4, 28950.0])

# Reaction 523
reaction('C2H + OH => O + C2H2', [3.023000e+15, -0.604, -1782.0])

# Reaction 524
reaction('C2H2 + O => CH2 + CO', [6.940000e+06, 2.0, 1900.0])

# Reaction 525
reaction('CH2 + CO => C2H2 + O', [4.050000e+01, 3.198, 48360.0])

# Reaction 526
reaction('C2H2 + O => HCCO + H', [1.350000e+07, 2.0, 1900.0])

# Reaction 527
reaction('HCCO + H => C2H2 + O', [4.755000e+07, 1.65, 20800.0])

# Reaction 528
reaction('C2H2 + OH => C2H + H2O', [3.370000e+07, 2.0, 14000.0])

# Reaction 529
reaction('C2H + H2O => C2H2 + OH', [4.524000e+04, 2.709, -428.0])

# Reaction 530
reaction('C2H2 + OH => CH2CO + H', [3.236000e+13, 0.0, 12000.0])

# Reaction 531
reaction('CH2CO + H => C2H2 + OH', [3.061000e+17, -0.802, 35790.0])

# Reaction 532
reaction('C2H2 + OH => CH3 + CO', [4.830000e-04, 4.0, -2000.0])

# Reaction 533
reaction('CH3 + CO => C2H2 + OH', [3.495000e-06, 4.638, 52120.0])

# Reaction 534
reaction('OH + C2H2 => H + HCCOH', [5.040000e+05, 2.3, 13500.0])

# Reaction 535
reaction('H + HCCOH => OH + C2H2', [3.852000e+09, 1.436, 4382.0])

# Reaction 536
reaction('H + HCCOH => H + CH2CO', [1.000000e+13, 0.0, 0.0])

# Reaction 537
reaction('H + CH2CO => H + HCCOH', [9.299000e+13, -0.29, 31110.0])

# Reaction 538
falloff_reaction('C2H5OH (+ M) <=> CH2OH + CH3 (+ M)',
                kf=[2.000000e+23, -1.68, 96400.0],
                kf0=[3.110000e+85, -18.84, 113100.0],
                efficiencies='H2:2.0 H2O:5.0 CO2:3.0 CO:2.0',
                falloff=Troe(A=0.5, T3=550.0, T1=825.0, T2=6100.0))

# Reaction 539
falloff_reaction('C2H5OH (+ M) <=> C2H5 + OH (+ M)',
                kf=[2.400000e+23, -1.62, 99540.0],
                kf0=[5.110000e+85, -18.8, 118770.0],
                efficiencies='H2:2.0 H2O:5.0 CO2:3.0 CO:2.0',
                falloff=Troe(A=0.5, T3=650.0, T1=800.0, T2=1e+15))

# Reaction 540
falloff_reaction('C2H5OH (+ M) <=> C2H4 + H2O (+ M)',
                kf=[1.320000e+05, 2.52, 60660.0],
                kf0=[3.090000e+55, -10.92, 62644.0],
                efficiencies='H2O:5.0',
                falloff=Troe(A=0.897, T3=10000000000.0, T1=1.0, T2=5000000000.0))

# Reaction 541
falloff_reaction('C2H5OH (+ M) <=> CH3CHO + H2 (+ M)',
                kf=[7.240000e+11, 0.095, 91010.0],
                kf0=[4.460000e+87, -19.42, 115580.0],
                efficiencies='H2O:5.0',
                falloff=Troe(A=0.9, T3=900.0, T1=1100.0, T2=3500.0))

# Reaction 542
reaction('C2H5OH + O2 => PC2H4OH + HO2', [2.000000e+13, 0.0, 52800.0])

# Reaction 543
reaction('PC2H4OH + HO2 => C2H5OH + O2', [2.192000e+10, 0.278, 443.0])

# Reaction 544
reaction('C2H5OH + O2 => SC2H4OH + HO2', [1.500000e+13, 0.0, 50150.0])

# Reaction 545
reaction('SC2H4OH + HO2 => C2H5OH + O2', [1.946000e+11, 0.089, 4879.0])

# Reaction 546
reaction('C2H5OH + OH => PC2H4OH + H2O', [1.810000e+11, 0.4, 717.0])

# Reaction 547
reaction('PC2H4OH + H2O => C2H5OH + OH', [4.012000e+08, 0.92, 17940.0])

# Reaction 548
reaction('C2H5OH + OH => SC2H4OH + H2O', [5.560000e+10, 0.5, -380.0])

# Reaction 549
reaction('SC2H4OH + H2O => C2H5OH + OH', [1.458000e+09, 0.831, 23930.0])

# Reaction 550
reaction('C2H5OH + OH => C2H5O + H2O', [1.500000e+10, 0.8, 2534.0])

# Reaction 551
reaction('C2H5O + H2O => C2H5OH + OH', [7.320000e+09, 0.906, 17210.0])

# Reaction 552
reaction('C2H5OH + H => PC2H4OH + H2', [1.880000e+03, 3.2, 7150.0])

# Reaction 553
reaction('PC2H4OH + H2 => C2H5OH + H', [3.930000e-01, 3.826, 9484.0])

# Reaction 554
reaction('C2H5OH + H => SC2H4OH + H2', [1.790000e+05, 2.53, 3420.0])

# Reaction 555
reaction('SC2H4OH + H2 => C2H5OH + H', [4.429000e+02, 2.967, 12840.0])

# Reaction 556
reaction('C2H5OH + H => C2H5O + H2', [5.360000e+04, 2.53, 4405.0])

# Reaction 557
reaction('C2H5O + H2 => C2H5OH + H', [2.467000e+03, 2.742, 4188.0])

# Reaction 558
reaction('C2H5OH + HO2 => PC2H4OH + H2O2', [2.379000e+04, 2.55, 16490.0])

# Reaction 559
reaction('PC2H4OH + H2O2 => C2H5OH + HO2', [2.877000e+03, 2.481, 2827.0])

# Reaction 560
reaction('C2H5OH + HO2 => SC2H4OH + H2O2', [6.000000e+12, 0.0, 16000.0])

# Reaction 561
reaction('SC2H4OH + H2O2 => C2H5OH + HO2', [8.589000e+12, -0.258, 9419.0])

# Reaction 562
reaction('C2H5OH + HO2 => C2H5O + H2O2', [2.500000e+12, 0.0, 24000.0])

# Reaction 563
reaction('C2H5O + H2O2 => C2H5OH + HO2', [6.658000e+13, -0.483, 7782.0])

# Reaction 564
reaction('C2H5OH + CH3O2 => PC2H4OH + CH3O2H', [1.230000e+04, 2.55, 15750.0])

# Reaction 565
reaction('PC2H4OH + CH3O2H => C2H5OH + CH3O2', [2.894000e+04, 2.036, 488.0])

# Reaction 566
reaction('C2H5OH + CH3O2 => SC2H4OH + CH3O2H', [8.200000e+03, 2.55, 10750.0])

# Reaction 567
reaction('SC2H4OH + CH3O2H => C2H5OH + CH3O2', [2.284000e+05, 1.847, 2574.0])

# Reaction 568
reaction('C2H5OH + CH3O2 => C2H5O + CH3O2H', [2.500000e+12, 0.0, 24000.0])

# Reaction 569
reaction('C2H5O + CH3O2H => C2H5OH + CH3O2', [1.295000e+15, -0.927, 6187.0])

# Reaction 570
reaction('C2H5OH + O => PC2H4OH + OH', [9.690000e+02, 3.23, 4658.0])

# Reaction 571
reaction('PC2H4OH + OH => C2H5OH + O', [1.052000e-01, 3.837, 5580.0])

# Reaction 572
reaction('C2H5OH + O => SC2H4OH + OH', [1.450000e+05, 2.47, 876.0])

# Reaction 573
reaction('SC2H4OH + OH => C2H5OH + O', [1.862000e+02, 2.888, 8884.0])

# Reaction 574
reaction('C2H5OH + O => C2H5O + OH', [1.460000e-03, 4.73, 1727.0])

# Reaction 575
reaction('C2H5O + OH => C2H5OH + O', [3.488000e-05, 4.924, 98.0])

# Reaction 576
reaction('C2H5OH + CH3 => PC2H4OH + CH4', [3.300000e+02, 3.3, 12290.0])

# Reaction 577
reaction('PC2H4OH + CH4 => C2H5OH + CH3', [6.294000e+01, 3.48, 16160.0])

# Reaction 578
reaction('C2H5OH + CH3 => SC2H4OH + CH4', [1.993000e+01, 3.37, 7634.0])

# Reaction 579
reaction('SC2H4OH + CH4 => C2H5OH + CH3', [4.498000e+01, 3.361, 18590.0])

# Reaction 580
reaction('C2H5OH + CH3 => C2H5O + CH4', [2.035000e+00, 3.57, 7721.0])

# Reaction 581
reaction('C2H5O + CH4 => C2H5OH + CH3', [8.545000e+01, 3.336, 9044.0])

# Reaction 582
reaction('C2H5OH + C2H5 => PC2H4OH + C2H6', [5.000000e+10, 0.0, 13400.0])

# Reaction 583
reaction('PC2H4OH + C2H6 => C2H5OH + C2H5', [6.995000e+10, 0.0, 26990.0])

# Reaction 584
reaction('C2H5OH + C2H5 => SC2H4OH + C2H6', [5.000000e+10, 0.0, 10400.0])

# Reaction 585
reaction('SC2H4OH + C2H6 => C2H5OH + C2H5', [6.995000e+10, 0.0, 2.399])

# Reaction 586
reaction('PC2H4OH => C2H4 + OH', [1.047000e+25, -3.99, 30390.0])

# Reaction 587
reaction('C2H4 + OH => PC2H4OH', [4.170000e+20, -2.84, 1240.0])

# Reaction 588
three_body_reaction('SC2H4OH + M => CH3CHO + H + M', [1.000000e+14, 0.0, 25000.0])

# Reaction 589
three_body_reaction('CH3CHO + H + M => SC2H4OH + M', [2.742000e+12, 0.462, -470.0])

# Reaction 590
reaction('O2C2H4OH => PC2H4OH + O2', [3.900000e+16, -1.0, 30000.0])

# Reaction 591
reaction('PC2H4OH + O2 => O2C2H4OH', [1.200000e+11, 0.0, -1100.0])

# Reaction 592
reaction('O2C2H4OH => OH + CH2O + CH2O', [3.125000e+09, 0.0, 18900.0])

# Reaction 593
reaction('OH + CH2O + CH2O => O2C2H4OH', [0.000000e+00, 0.0, 0.0])

# Reaction 594
reaction('SC2H4OH + O2 => CH3CHO + HO2', [3.810000e+06, 2.0, 1641.0])

# Reaction 595
reaction('CH3CHO + HO2 => SC2H4OH + O2', [2.190000e+05, 2.39, 25040.0])

# Reaction 596
falloff_reaction('CH3COCH3 (+ M) <=> CH3CO + CH3 (+ M)',
                kf=[7.108000e+21, -1.57, 84680.0],
                kf0=[7.013000e+89, -20.38, 107150.0],
                falloff=Troe(A=0.863, T3=10000000000.0, T1=416.4, T2=3290000000.0))

# Reaction 597
reaction('CH3COCH3 + OH => CH3COCH2 + H2O', [1.250000e+05, 2.483, 445.0])

# Reaction 598
reaction('CH3COCH2 + H2O => CH3COCH3 + OH', [8.620000e+04, 2.322, 24710.0])

# Reaction 599
reaction('CH3COCH3 + H => CH3COCH2 + H2', [9.800000e+05, 2.43, 5160.0])

# Reaction 600
reaction('CH3COCH2 + H2 => CH3COCH3 + H', [6.374000e+04, 2.375, 14530.0])

# Reaction 601
reaction('CH3COCH3 + O => CH3COCH2 + OH', [5.130000e+11, 0.211, 4890.0])

# Reaction 602
reaction('CH3COCH2 + OH => CH3COCH3 + O', [1.732000e+10, 0.137, 12850.0])

# Reaction 603
reaction('CH3COCH3 + CH3 => CH3COCH2 + CH4', [3.960000e+11, 0.0, 9784.0])

# Reaction 604
reaction('CH3COCH2 + CH4 => CH3COCH3 + CH3', [2.350000e+13, -0.501, 20690.0])

# Reaction 605
reaction('CH3COCH3 + CH3O => CH3COCH2 + CH3OH', [4.340000e+11, 0.0, 6460.0])

# Reaction 606
reaction('CH3COCH2 + CH3OH => CH3COCH3 + CH3O', [6.060000e+11, -0.268, 16060.0])

# Reaction 607
reaction('CH3COCH3 + O2 => CH3COCH2 + HO2', [6.030000e+13, 0.0, 48500.0])

# Reaction 608
reaction('CH3COCH2 + HO2 => CH3COCH3 + O2', [2.057000e+13, -0.403, 3181.0])

# Reaction 609
reaction('CH3COCH3 + HO2 => CH3COCH2 + H2O2', [1.700000e+13, 0.0, 20460.0])

# Reaction 610
reaction('CH3COCH2 + H2O2 => CH3COCH3 + HO2', [6.397000e+14, -0.75, 13830.0])

# Reaction 611
reaction('CH3COCH3 + CH3O2 => CH3COCH2 + CH3O2H', [1.700000e+13, 0.0, 20460.0])

# Reaction 612
reaction('CH3COCH2 + CH3O2H => CH3COCH3 + CH3O2', [1.245000e+16, -1.195, 12230.0])

# Reaction 613
reaction('CH3COCH2 => CH2CO + CH3', [1.000000e+14, 0.0, 31000.0])

# Reaction 614
reaction('CH2CO + CH3 => CH3COCH2', [1.000000e+11, 0.0, 6000.0])

# Reaction 615
reaction('CH3COCH2O2 => CH3COCH2 + O2', [2.021000e+15, -0.956, 24460.0])

# Reaction 616
reaction('CH3COCH2 + O2 => CH3COCH2O2', [1.200000e+11, 0.0, -1100.0])

# Reaction 617
reaction('CH3COCH3 + CH3COCH2O2 => CH3COCH2 + CH3COCH2O2H', [1.000000e+11, 0.0, 5000.0])

# Reaction 618
reaction('CH3COCH2 + CH3COCH2O2H => CH3COCH3 + CH3COCH2O2', [1.995000e+10, 0.0, 10000.0])

# Reaction 619
reaction('CH2O + CH3COCH2O2 => HCO + CH3COCH2O2H', [1.288000e+11, 0.0, 9000.0])

# Reaction 620
reaction('HCO + CH3COCH2O2H => CH2O + CH3COCH2O2', [2.512000e+10, 0.0, 10100.0])

# Reaction 621
reaction('HO2 + CH3COCH2O2 => CH3COCH2O2H + O2', [1.000000e+12, 0.0, 0.0])

# Reaction 622
reaction('CH3COCH2O2H + O2 => HO2 + CH3COCH2O2', [0.000000e+00, 0.0, 0.0])

# Reaction 623
reaction('CH3COCH2O2H => CH3COCH2O + OH', [1.000000e+16, 0.0, 43000.0])

# Reaction 624
reaction('CH3COCH2O + OH => CH3COCH2O2H', [4.242000e+08, 1.74, -4342.0])

# Reaction 625
reaction('CH3COCH2O => CH3CO + CH2O', [3.732000e+20, -2.176, 17260.0])

# Reaction 626
reaction('CH3CO + CH2O => CH3COCH2O', [1.000000e+11, 0.0, 11900.0])

# Reaction 627
reaction('C2H3CHO => C2H3 + HCO', [2.003000e+24, -2.135, 103400.0])

# Reaction 628
reaction('C2H3 + HCO => C2H3CHO', [1.810000e+13, 0.0, 0.0])

# Reaction 629
reaction('C2H3CHO + H => C2H3CO + H2', [1.340000e+13, 0.0, 3300.0])

# Reaction 630
reaction('C2H3CO + H2 => C2H3CHO + H', [3.311000e+10, 0.613, 22680.0])

# Reaction 631
reaction('C2H3CHO + O => C2H3CO + OH', [5.940000e+12, 0.0, 1868.0])

# Reaction 632
reaction('C2H3CO + OH => C2H3CHO + O', [7.618000e+09, 0.594, 19840.0])

# Reaction 633
reaction('C2H3CHO + OH => C2H3CO + H2O', [9.240000e+06, 1.5, -962.0])

# Reaction 634
reaction('C2H3CO + H2O => C2H3CHO + OH', [2.420000e+05, 2.007, 33310.0])

# Reaction 635
reaction('C2H3CHO + O2 => C2H3CO + HO2', [1.005000e+13, 0.0, 40700.0])

# Reaction 636
reaction('C2H3CO + HO2 => C2H3CHO + O2', [1.302000e+11, 0.265, 5391.0])

# Reaction 637
reaction('C2H3CHO + HO2 => C2H3CO + H2O2', [3.010000e+12, 0.0, 11920.0])

# Reaction 638
reaction('C2H3CO + H2O2 => C2H3CHO + HO2', [4.303000e+12, -0.082, 15300.0])

# Reaction 639
reaction('C2H3CHO + CH3 => C2H3CO + CH4', [2.608000e+06, 1.78, 5911.0])

# Reaction 640
reaction('C2H3CO + CH4 => C2H3CHO + CH3', [5.878000e+06, 1.947, 26830.0])

# Reaction 641
reaction('C2H3CHO + C2H3 => C2H3CO + C2H4', [1.740000e+12, 0.0, 8440.0])

# Reaction 642
reaction('C2H3CO + C2H4 => C2H3CHO + C2H3', [1.000000e+13, 0.0, 28000.0])

# Reaction 643
reaction('C2H3CHO + CH3O => C2H3CO + CH3OH', [1.000000e+12, 0.0, 3300.0])

# Reaction 644
reaction('C2H3CO + CH3OH => C2H3CHO + CH3O', [5.304000e+10, 0.401, 22910.0])

# Reaction 645
reaction('C2H3CHO + CH3O2 => C2H3CO + CH3O2H', [3.010000e+12, 0.0, 11920.0])

# Reaction 646
reaction('C2H3CO + CH3O2H => C2H3CHO + CH3O2', [8.371000e+13, -0.527, 13710.0])

# Reaction 647
reaction('C2H3CO => C2H3 + CO', [1.370000e+21, -2.179, 39410.0])

# Reaction 648
reaction('C2H3 + CO => C2H3CO', [1.510000e+11, 0.0, 4810.0])

# Reaction 649
reaction('C2H5CHO => C2H5 + HCO', [1.496000e+27, -3.205, 87040.0])

# Reaction 650
reaction('C2H5 + HCO => C2H5CHO', [1.810000e+13, 0.0, 0.0])

# Reaction 651
reaction('C2H5CHO + H => C2H5CO + H2', [4.000000e+13, 0.0, 4200.0])

# Reaction 652
reaction('C2H5CO + H2 => C2H5CHO + H', [2.377000e+10, 0.654, 18130.0])

# Reaction 653
reaction('C2H5CHO + O => C2H5CO + OH', [5.000000e+12, 0.0, 1790.0])

# Reaction 654
reaction('C2H5CO + OH => C2H5CHO + O', [1.542000e+09, 0.636, 14310.0])

# Reaction 655
reaction('C2H5CHO + OH => C2H5CO + H2O', [2.690000e+10, 0.76, -340.0])

# Reaction 656
reaction('C2H5CO + H2O => C2H5CHO + OH', [1.695000e+08, 1.308, 28480.0])

# Reaction 657
reaction('C2H5CHO + CH3 => C2H5CO + CH4', [2.608000e+06, 1.78, 5911.0])

# Reaction 658
reaction('C2H5CO + CH4 => C2H5CHO + CH3', [1.414000e+06, 1.988, 21380.0])

# Reaction 659
reaction('C2H5CHO + HO2 => C2H5CO + H2O2', [2.800000e+12, 0.0, 13600.0])

# Reaction 660
reaction('C2H5CO + H2O2 => C2H5CHO + HO2', [9.626000e+11, -0.041, 11530.0])

# Reaction 661
reaction('C2H5CHO + CH3O => C2H5CO + CH3OH', [1.000000e+12, 0.0, 3300.0])

# Reaction 662
reaction('C2H5CO + CH3OH => C2H5CHO + CH3O', [1.276000e+10, 0.442, 17460.0])

# Reaction 663
reaction('C2H5CHO + CH3O2 => C2H5CO + CH3O2H', [3.010000e+12, 0.0, 11920.0])

# Reaction 664
reaction('C2H5CO + CH3O2H => C2H5CHO + CH3O2', [2.013000e+13, -0.485, 8260.0])

# Reaction 665
reaction('C2H5CHO + C2H5 => C2H5CO + C2H6', [1.000000e+12, 0.0, 8000.0])

# Reaction 666
reaction('C2H5CO + C2H6 => C2H5CHO + C2H5', [6.432000e+12, -0.028, 19700.0])

# Reaction 667
reaction('C2H5CHO + C2H5O => C2H5CO + C2H5OH', [6.026000e+11, 0.0, 3300.0])

# Reaction 668
reaction('C2H5CO + C2H5OH => C2H5CHO + C2H5O', [3.020000e+11, 0.0, 18160.0])

# Reaction 669
reaction('C2H5CHO + C2H5O2 => C2H5CO + C2H5O2H', [3.010000e+12, 0.0, 11920.0])

# Reaction 670
reaction('C2H5CO + C2H5O2H => C2H5CHO + C2H5O2', [2.017000e+13, -0.486, 8264.0])

# Reaction 671
reaction('C2H5CHO + O2 => C2H5CO + HO2', [1.005000e+13, 0.0, 40700.0])

# Reaction 672
reaction('C2H5CO + HO2 => C2H5CHO + O2', [3.131000e+10, 0.306, -58.0])

# Reaction 673
reaction('C2H5CHO + CH3CO3 => C2H5CO + CH3CO3H', [3.010000e+12, 0.0, 11920.0])

# Reaction 674
reaction('C2H5CO + CH3CO3H => C2H5CHO + CH3CO3', [1.651000e+12, 0.012, 12630.0])

# Reaction 675
reaction('C2H5CHO + C2H3 => C2H5CO + C2H4', [1.700000e+12, 0.0, 8440.0])

# Reaction 676
reaction('C2H5CO + C2H4 => C2H5CHO + C2H3', [3.198000e+12, 0.148, 30130.0])

# Reaction 677
reaction('C2H5CO => C2H5 + CO', [2.460000e+23, -3.208, 17550.0])

# Reaction 678
reaction('C2H5 + CO => C2H5CO', [1.510000e+11, 0.0, 4810.0])

# Reaction 679
falloff_reaction('CH3OCH3 (+ M) <=> CH3 + CH3O (+ M)',
                kf=[7.250000e+21, -0.94, 80250.0],
                kf0=[3.500000e+60, -11.56, 101000.0],
                falloff=Troe(A=0.183, T3=1.3, T1=13000.0, T2=6710000000.0))

# Reaction 680
reaction('CH3OCH3 + OH => CH3OCH2 + H2O', [6.324000e+06, 2.0, -651.7])

# Reaction 681
reaction('CH3OCH2 + H2O => CH3OCH3 + OH', [7.853000e+04, 2.236, 21210.0])

# Reaction 682
reaction('CH3OCH3 + H => CH3OCH2 + H2', [7.721000e+06, 2.09, 3384.0])

# Reaction 683
reaction('CH3OCH2 + H2 => CH3OCH3 + H', [9.042000e+03, 2.432, 10360.0])

# Reaction 684
reaction('CH3OCH3 + O => CH3OCH2 + OH', [7.750000e+08, 1.36, 2250.0])

# Reaction 685
reaction('CH3OCH2 + OH => CH3OCH3 + O', [4.712000e+05, 1.683, 7810.0])

# Reaction 686
reaction('CH3OCH3 + HO2 => CH3OCH2 + H2O2', [1.680000e+13, 0.0, 17690.0])

# Reaction 687
reaction('CH3OCH2 + H2O2 => CH3OCH3 + HO2', [1.138000e+13, -0.353, 8657.0])

# Reaction 688
reaction('CH3OCH3 + CH3O2 => CH3OCH2 + CH3O2H', [1.680000e+13, 0.0, 17690.0])

# Reaction 689
reaction('CH3OCH2 + CH3O2H => CH3OCH3 + CH3O2', [2.215000e+14, -0.798, 7062.0])

# Reaction 690
reaction('CH3OCH3 + CH3 => CH3OCH2 + CH4', [1.445000e-06, 5.73, 5700.0])

# Reaction 691
reaction('CH3OCH2 + CH4 => CH3OCH3 + CH3', [1.544000e-06, 5.626, 14210.0])

# Reaction 692
reaction('CH3OCH3 + O2 => CH3OCH2 + HO2', [4.100000e+13, 0.0, 44910.0])

# Reaction 693
reaction('CH3OCH2 + HO2 => CH3OCH3 + O2', [2.518000e+11, -0.006, -2806.0])

# Reaction 694
reaction('CH3OCH3 + CH3O => CH3OCH2 + CH3OH', [6.020000e+11, 0.0, 4074.0])

# Reaction 695
reaction('CH3OCH2 + CH3OH => CH3OCH3 + CH3O', [7.383000e+10, -0.27, 10260.0])

# Reaction 696
reaction('CH3OCH3 + CH3OCH2O2 => CH3OCH2 + CH3OCH2O2H', [5.000000e+12, 0.0, 17690.0])

# Reaction 697
reaction('CH3OCH2 + CH3OCH2O2H => CH3OCH3 + CH3OCH2O2', [6.428000e+13, -0.794, 7258.0])

# Reaction 698
reaction('CH3OCH3 + O2CHO => CH3OCH2 + HO2CHO', [4.425000e+04, 2.6, 13910.0])

# Reaction 699
reaction('CH3OCH2 + HO2CHO => CH3OCH3 + O2CHO', [1.723000e+06, 1.746, 5832.0])

# Reaction 700
reaction('CH3OCH3 + OCHO => CH3OCH2 + HOCHO', [1.000000e+13, 0.0, 17690.0])

# Reaction 701
reaction('CH3OCH2 + HOCHO => CH3OCH3 + OCHO', [7.195000e+12, -0.314, 33130.0])

# Reaction 702
reaction('CH3OCH2 => CH2O + CH3', [1.600000e+13, 0.0, 25500.0])

# Reaction 703
reaction('CH2O + CH3 => CH3OCH2', [2.601000e+05, 1.879, 16670.0])

# Reaction 704
reaction('CH3OCH2 + CH3O => CH3OCH3 + CH2O', [2.410000e+13, 0.0, 0.0])

# Reaction 705
reaction('CH3OCH3 + CH2O => CH3OCH2 + CH3O', [1.250000e+14, 0.32, 78540.0])

# Reaction 706
reaction('CH3OCH2 + CH2O => CH3OCH3 + HCO', [5.490000e+03, 2.8, 5862.0])

# Reaction 707
reaction('CH3OCH3 + HCO => CH3OCH2 + CH2O', [2.768000e+04, 2.745, 14080.0])

# Reaction 708
reaction('CH3OCH2 + CH3CHO => CH3OCH3 + CH3CO', [1.260000e+12, 0.0, 8499.0])

# Reaction 709
reaction('CH3OCH3 + CH3CO => CH3OCH2 + CH3CHO', [7.746000e+11, 0.28, 16980.0])

# Reaction 710
reaction('CH3OCH2O2 => CH3OCH2 + O2', [4.439000e+19, -1.594, 36240.0])

# Reaction 711
reaction('CH3OCH2 + O2 => CH3OCH2O2', [2.000000e+12, 0.0, 0.0])

# Reaction 712
reaction('CH3OCH2O2 + CH2O => CH3OCH2O2H + HCO', [1.000000e+12, 0.0, 11660.0])

# Reaction 713
reaction('CH3OCH2O2H + HCO => CH3OCH2O2 + CH2O', [6.482000e+13, -0.849, 9455.0])

# Reaction 714
reaction('CH3OCH2O2 + CH3CHO => CH3OCH2O2H + CH3CO', [2.800000e+12, 0.0, 13600.0])

# Reaction 715
reaction('CH3OCH2O2H + CH3CO => CH3OCH2O2 + CH3CHO', [1.000000e+12, 0.0, 10000.0])

# Reaction 716
reaction('CH3OCH2O2 + CH3OCH2O2 => O2 + CH3OCH2O + CH3OCH2O', [2.210000e+23, -4.5, 0.0])

# Reaction 717
reaction('O2 + CH3OCH2O + CH3OCH2O => CH3OCH2O2 + CH3OCH2O2', [0.000000e+00, 0.0, 0.0])

# Reaction 718
reaction('CH3OCH2O2H => CH3OCH2O + OH', [2.106000e+22, -2.124, 43830.0])

# Reaction 719
reaction('CH3OCH2O + OH => CH3OCH2O2H', [2.000000e+13, 0.0, 0.0])

# Reaction 720
reaction('CH3OCH2O => CH3O + CH2O', [4.384000e+19, -2.014, 25190.0])

# Reaction 721
reaction('CH3O + CH2O => CH3OCH2O', [1.000000e+11, 0.0, 11900.0])

# Reaction 722
reaction('CH3OCH2O + O2 => CH3OCHO + HO2', [5.000000e+10, 0.0, 500.0])

# Reaction 723
reaction('CH3OCHO + HO2 => CH3OCH2O + O2', [1.086000e+10, -0.02, 46480.0])

# Reaction 724
reaction('CH3OCH2O => CH3OCHO + H', [6.060000e+12, 0.056, 8218.0])

# Reaction 725
reaction('CH3OCHO + H => CH3OCH2O', [1.000000e+13, 0.0, 7838.0])

# Reaction 726
reaction('CH3OCH2O2 => CH2OCH2O2H', [6.000000e+10, 0.0, 21580.0])

# Reaction 727
reaction('CH2OCH2O2H => CH3OCH2O2', [1.248000e+12, -0.765, 11120.0])

# Reaction 728
reaction('CH2OCH2O2H => OH + CH2O + CH2O', [1.500000e+13, 0.0, 20760.0])

# Reaction 729
reaction('OH + CH2O + CH2O => CH2OCH2O2H', [0.000000e+00, 0.0, 0.0])

# Reaction 730
reaction('O2CH2OCH2O2H => CH2OCH2O2H + O2', [1.924000e+19, -1.622, 36270.0])

# Reaction 731
reaction('CH2OCH2O2H + O2 => O2CH2OCH2O2H', [7.000000e+11, 0.0, 0.0])

# Reaction 732
reaction('O2CH2OCH2O2H => HO2CH2OCHO + OH', [4.000000e+10, 0.0, 18580.0])

# Reaction 733
reaction('HO2CH2OCHO + OH => O2CH2OCH2O2H', [5.041000e+03, 1.416, 59640.0])

# Reaction 734
reaction('HO2CH2OCHO => OCH2OCHO + OH', [2.000000e+16, 0.0, 40500.0])

# Reaction 735
reaction('OCH2OCHO + OH => HO2CH2OCHO', [1.183000e+08, 1.934, -3952.0])

# Reaction 736
reaction('OCH2OCHO => CH2O + OCHO', [2.898000e+19, -2.201, 31850.0])

# Reaction 737
reaction('CH2O + OCHO => OCH2OCHO', [1.250000e+11, 0.0, 11900.0])

# Reaction 738
reaction('OCH2OCHO => HOCH2OCO', [1.000000e+11, 0.0, 14000.0])

# Reaction 739
reaction('HOCH2OCO => OCH2OCHO', [1.568000e+09, 0.487, 20670.0])

# Reaction 740
reaction('HOCH2OCO => HOCH2O + CO', [2.238000e+19, -2.021, 19690.0])

# Reaction 741
reaction('HOCH2O + CO => HOCH2OCO', [1.500000e+11, 0.0, 4800.0])

# Reaction 742
reaction('HOCH2OCO => CH2OH + CO2', [2.413000e+17, -1.574, 22120.0])

# Reaction 743
reaction('CH2OH + CO2 => HOCH2OCO', [1.500000e+11, 0.0, 35720.0])

# Reaction 744
reaction('CH3OCHO => CH2OCHO + H', [8.241000e+19, -1.15, 102500.0])

# Reaction 745
reaction('CH2OCHO + H => CH3OCHO', [1.000000e+14, 0.0, 0.0])

# Reaction 746
reaction('CH3OCHO => CH3OCO + H', [1.325000e+19, -1.0, 100100.0])

# Reaction 747
reaction('CH3OCO + H => CH3OCHO', [1.000000e+14, 0.0, 0.0])

# Reaction 748
falloff_reaction('CH3OCHO (+ M) <=> CH3OH + CO (+ M)',
                kf=[1.000000e+14, 0.0, 62500.0],
                kf0=[6.143000e+60, -12.07, 75400.0],
                falloff=Troe(A=0.78, T3=8280000000.0, T1=438.9, T2=670000000.0))

# Reaction 749
reaction('CH3OCHO => CH3O + HCO', [5.374000e+16, -0.01, 97090.0])

# Reaction 750
reaction('CH3O + HCO => CH3OCHO', [3.000000e+13, 0.0, 0.0])

# Reaction 751
reaction('CH3OCHO => CH3 + OCHO', [3.214000e+17, -0.53, 79970.0])

# Reaction 752
reaction('CH3 + OCHO => CH3OCHO', [1.000000e+13, 0.0, 0.0])

# Reaction 753
reaction('CH3OCHO + O2 => CH3OCO + HO2', [1.000000e+13, 0.0, 49700.0])

# Reaction 754
reaction('CH3OCO + HO2 => CH3OCHO + O2', [2.979000e+10, 0.329, 0.9998])

# Reaction 755
reaction('CH3OCHO + O2 => CH2OCHO + HO2', [2.050000e+13, 0.0, 52000.0])

# Reaction 756
reaction('CH2OCHO + HO2 => CH3OCHO + O2', [2.143000e+10, 0.356, 331.0])

# Reaction 757
reaction('CH3OCHO + OH => CH3OCO + H2O', [1.580000e+07, 1.8, 934.0])

# Reaction 758
reaction('CH3OCO + H2O => CH3OCHO + OH', [9.519000e+04, 2.371, 20820.0])

# Reaction 759
reaction('CH3OCHO + OH => CH2OCHO + H2O', [5.270000e+09, 0.97, 1586.0])

# Reaction 760
reaction('CH2OCHO + H2O => CH3OCHO + OH', [1.114000e+07, 1.568, 19500.0])

# Reaction 761
reaction('CH3OCHO + HO2 => CH3OCO + H2O2', [4.820000e+03, 2.6, 13910.0])

# Reaction 762
reaction('CH3OCO + H2O2 => CH3OCHO + HO2', [1.585000e+03, 2.581, 2899.0])

# Reaction 763
reaction('CH3OCHO + HO2 => CH2OCHO + H2O2', [2.380000e+04, 2.55, 16490.0])

# Reaction 764
reaction('CH2OCHO + H2O2 => CH3OCHO + HO2', [2.746000e+03, 2.558, 3513.0])

# Reaction 765
reaction('CH3OCHO + O => CH3OCO + OH', [2.755000e+05, 2.45, 2830.0])

# Reaction 766
reaction('CH3OCO + OH => CH3OCHO + O', [8.126000e+01, 3.108, 6408.0])

# Reaction 767
reaction('CH3OCHO + O => CH2OCHO + OH', [9.800000e+05, 2.43, 4750.0])

# Reaction 768
reaction('CH2OCHO + OH => CH3OCHO + O', [1.014000e+02, 3.115, 6358.0])

# Reaction 769
reaction('CH3OCHO + H => CH3OCO + H2', [6.500000e+05, 2.4, 4471.0])

# Reaction 770
reaction('CH3OCO + H2 => CH3OCHO + H', [3.693000e+02, 3.076, 9461.0])

# Reaction 771
reaction('CH3OCHO + H => CH2OCHO + H2', [6.650000e+05, 2.54, 6756.0])

# Reaction 772
reaction('CH2OCHO + H2 => CH3OCHO + H', [1.326000e+02, 3.243, 9776.0])

# Reaction 773
reaction('CH3OCHO + CH3 => CH3OCO + CH4', [7.550000e-01, 3.46, 5481.0])

# Reaction 774
reaction('CH3OCO + CH4 => CH3OCHO + CH3', [3.914000e-01, 3.691, 12010.0])

# Reaction 775
reaction('CH3OCHO + CH3 => CH2OCHO + CH4', [4.520000e-01, 3.65, 7154.0])

# Reaction 776
reaction('CH2OCHO + CH4 => CH3OCHO + CH3', [8.222000e-02, 3.908, 11710.0])

# Reaction 777
reaction('CH3OCHO + CH3O => CH3OCO + CH3OH', [5.480000e+11, 0.0, 5000.0])

# Reaction 778
reaction('CH3OCO + CH3OH => CH3OCHO + CH3O', [6.685000e+09, 0.464, 10220.0])

# Reaction 779
reaction('CH3OCHO + CH3O => CH2OCHO + CH3OH', [2.170000e+11, 0.0, 6458.0])

# Reaction 780
reaction('CH2OCHO + CH3OH => CH3OCHO + CH3O', [9.289000e+08, 0.491, 9705.0])

# Reaction 781
reaction('CH3OCHO + CH3O2 => CH3OCO + CH3O2H', [4.820000e+03, 2.6, 13910.0])

# Reaction 782
reaction('CH3OCO + CH3O2H => CH3OCHO + CH3O2', [3.083000e+04, 2.137, 1304.0])

# Reaction 783
reaction('CH3OCHO + CH3O2 => CH2OCHO + CH3O2H', [2.380000e+04, 2.55, 16490.0])

# Reaction 784
reaction('CH2OCHO + CH3O2H => CH3OCHO + CH3O2', [5.342000e+04, 2.114, 1918.0])

# Reaction 785
reaction('CH3OCHO + HCO => CH3OCO + CH2O', [5.400000e+06, 1.9, 17010.0])

# Reaction 786
reaction('CH3OCO + CH2O => CH3OCHO + HCO', [5.196000e+05, 2.29, 6806.0])

# Reaction 787
reaction('CH3OCHO + HCO => CH2OCHO + CH2O', [1.025000e+05, 2.5, 18430.0])

# Reaction 788
reaction('CH2OCHO + CH2O => CH3OCHO + HCO', [3.461000e+03, 2.917, 6261.0])

# Reaction 789
reaction('CH2OCHO => CH3OCO', [2.620000e+11, -0.03, 38180.0])

# Reaction 790
reaction('CH3OCO => CH2OCHO', [1.629000e+12, -0.18, 40670.0])

# Reaction 791
reaction('CH3OCO => CH3 + CO2', [3.590000e+14, -0.172, 16010.0])

# Reaction 792
reaction('CH3 + CO2 => CH3OCO', [4.760000e+07, 1.54, 34700.0])

# Reaction 793
reaction('CH3OCO => CH3O + CO', [1.431000e+15, -0.041, 23770.0])

# Reaction 794
reaction('CH3O + CO => CH3OCO', [1.550000e+06, 2.02, 5730.0])

# Reaction 795
reaction('CH2OCHO => CH2O + HCO', [4.661000e+12, 0.12, 27440.0])

# Reaction 796
reaction('CH2O + HCO => CH2OCHO', [1.500000e+11, 0.0, 11900.0])

# Reaction 797
falloff_reaction('C3H8 (+ M) <=> CH3 + C2H5 (+ M)',
                kf=[1.290000e+37, -5.84, 97380.0],
                kf0=[5.640000e+74, -15.74, 98714.0],
                efficiencies='CO2:2.0 CO:1.5 H2:2.0 H2O:6.0 AR:0.7 CH4:2.0 C2H6:3.0 HE:0.7',
                falloff=Troe(A=0.31, T3=50.0, T1=3000.0, T2=9000.0))

# Reaction 798
reaction('C3H8 => NC3H7 + H', [3.750000e+17, -0.357, 101200.0])

# Reaction 799
reaction('NC3H7 + H => C3H8', [1.000000e+14, 0.0, 0.0])

# Reaction 800
reaction('C3H8 => IC3H7 + H', [2.377000e+18, -0.671, 98680.0])

# Reaction 801
reaction('IC3H7 + H => C3H8', [1.000000e+14, 0.0, 0.0])

# Reaction 802
reaction('C3H8 + O2 => IC3H7 + HO2', [2.000000e+13, 0.0, 49640.0])

# Reaction 803
reaction('IC3H7 + HO2 => C3H8 + O2', [1.764000e+09, 0.599, -169.0])

# Reaction 804
reaction('C3H8 + O2 => NC3H7 + HO2', [6.000000e+13, 0.0, 52290.0])

# Reaction 805
reaction('NC3H7 + HO2 => C3H8 + O2', [3.354000e+10, 0.285, -59.0])

# Reaction 806
reaction('H + C3H8 => H2 + IC3H7', [1.300000e+06, 2.4, 4471.0])

# Reaction 807
reaction('H2 + IC3H7 => H + C3H8', [2.186000e+01, 3.347, 9351.0])

# Reaction 808
reaction('H + C3H8 => H2 + NC3H7', [1.330000e+06, 2.54, 6756.0])

# Reaction 809
reaction('H2 + NC3H7 => H + C3H8', [1.418000e+02, 3.173, 9096.0])

# Reaction 810
reaction('C3H8 + O => IC3H7 + OH', [5.490000e+05, 2.5, 3140.0])

# Reaction 811
reaction('IC3H7 + OH => C3H8 + O', [4.793000e+00, 3.428, 6608.0])

# Reaction 812
reaction('C3H8 + O => NC3H7 + OH', [3.710000e+06, 2.4, 5505.0])

# Reaction 813
reaction('NC3H7 + OH => C3H8 + O', [2.053000e+02, 3.014, 6433.0])

# Reaction 814
reaction('C3H8 + OH => NC3H7 + H2O', [1.054000e+10, 0.97, 1586.0])

# Reaction 815
reaction('NC3H7 + H2O => C3H8 + OH', [1.191000e+07, 1.497, 18820.0])

# Reaction 816
reaction('C3H8 + OH => IC3H7 + H2O', [4.670000e+07, 1.61, -35.0])

# Reaction 817
reaction('IC3H7 + H2O => C3H8 + OH', [8.327000e+03, 2.451, 19740.0])

# Reaction 818
reaction('C3H8 + HO2 => IC3H7 + H2O2', [5.880000e+04, 2.5, 14860.0])

# Reaction 819
reaction('IC3H7 + H2O2 => C3H8 + HO2', [5.721000e+02, 2.752, 3742.0])

# Reaction 820
reaction('C3H8 + HO2 => NC3H7 + H2O2', [8.100000e+04, 2.5, 16690.0])

# Reaction 821
reaction('NC3H7 + H2O2 => C3H8 + HO2', [4.995000e+03, 2.438, 3030.0])

# Reaction 822
reaction('CH3 + C3H8 => CH4 + IC3H7', [6.400000e+04, 2.17, 7520.0])

# Reaction 823
reaction('CH4 + IC3H7 => CH3 + C3H8', [9.819000e+02, 2.671, 13940.0])

# Reaction 824
reaction('CH3 + C3H8 => CH4 + NC3H7', [9.040000e-01, 3.65, 7154.0])

# Reaction 825
reaction('CH4 + NC3H7 => CH3 + C3H8', [8.791000e-02, 3.837, 11030.0])

# Reaction 826
reaction('IC3H7 + C3H8 => NC3H7 + C3H8', [3.000000e+10, 0.0, 12900.0])

# Reaction 827
reaction('NC3H7 + C3H8 => IC3H7 + C3H8', [3.000000e+10, 0.0, 12900.0])

# Reaction 828
reaction('C2H3 + C3H8 => C2H4 + IC3H7', [1.000000e+11, 0.0, 10400.0])

# Reaction 829
reaction('C2H4 + IC3H7 => C2H3 + C3H8', [1.310000e+11, 0.0, 17800.0])

# Reaction 830
reaction('C2H3 + C3H8 => C2H4 + NC3H7', [1.000000e+11, 0.0, 10400.0])

# Reaction 831
reaction('C2H4 + NC3H7 => C2H3 + C3H8', [1.310000e+11, 0.0, 17800.0])

# Reaction 832
reaction('C2H5 + C3H8 => C2H6 + IC3H7', [1.000000e+11, 0.0, 10400.0])

# Reaction 833
reaction('C2H6 + IC3H7 => C2H5 + C3H8', [3.630000e+10, 0.0, 9934.0])

# Reaction 834
reaction('C2H5 + C3H8 => C2H6 + NC3H7', [1.000000e+11, 0.0, 10400.0])

# Reaction 835
reaction('C2H6 + NC3H7 => C2H5 + C3H8', [3.630000e+10, 0.0, 9934.0])

# Reaction 836
reaction('C3H8 + C3H5-A => NC3H7 + C3H6', [7.940000e+11, 0.0, 20500.0])

# Reaction 837
reaction('NC3H7 + C3H6 => C3H8 + C3H5-A', [5.372000e+16, -1.33, 13400.0])

# Reaction 838
reaction('C3H8 + C3H5-A => IC3H7 + C3H6', [7.940000e+11, 0.0, 16200.0])

# Reaction 839
reaction('IC3H7 + C3H6 => C3H8 + C3H5-A', [5.372000e+16, -1.33, 9095.0])

# Reaction 840
reaction('C3H8 + CH3O => NC3H7 + CH3OH', [3.000000e+11, 0.0, 7000.0])

# Reaction 841
reaction('NC3H7 + CH3OH => C3H8 + CH3O', [1.220000e+10, 0.0, 9182.0])

# Reaction 842
reaction('C3H8 + CH3O => IC3H7 + CH3OH', [3.000000e+11, 0.0, 7000.0])

# Reaction 843
reaction('IC3H7 + CH3OH => C3H8 + CH3O', [1.220000e+10, 0.0, 9182.0])

# Reaction 844
reaction('CH3O2 + C3H8 => CH3O2H + NC3H7', [8.100000e+04, 2.5, 16690.0])

# Reaction 845
reaction('CH3O2H + NC3H7 => CH3O2 + C3H8', [9.718000e+04, 1.993, 1435.0])

# Reaction 846
reaction('CH3O2 + C3H8 => CH3O2H + IC3H7', [5.880000e+04, 2.5, 14860.0])

# Reaction 847
reaction('CH3O2H + IC3H7 => CH3O2 + C3H8', [1.113000e+04, 2.307, 2147.0])

# Reaction 848
reaction('C2H5O2 + C3H8 => C2H5O2H + NC3H7', [8.100000e+04, 2.5, 16690.0])

# Reaction 849
reaction('C2H5O2H + NC3H7 => C2H5O2 + C3H8', [9.735000e+04, 1.993, 1439.0])

# Reaction 850
reaction('C2H5O2 + C3H8 => C2H5O2H + IC3H7', [5.880000e+04, 2.5, 14860.0])

# Reaction 851
reaction('C2H5O2H + IC3H7 => C2H5O2 + C3H8', [1.115000e+04, 2.307, 2151.0])

# Reaction 852
reaction('NC3H7O2 + C3H8 => NC3H7O2H + NC3H7', [1.700000e+13, 0.0, 20460.0])

# Reaction 853
reaction('NC3H7O2H + NC3H7 => NC3H7O2 + C3H8', [2.086000e+13, -0.51, 5000.0])

# Reaction 854
reaction('NC3H7O2 + C3H8 => NC3H7O2H + IC3H7', [2.000000e+12, 0.0, 17000.0])

# Reaction 855
reaction('NC3H7O2H + IC3H7 => NC3H7O2 + C3H8', [3.871000e+11, -0.196, 4080.0])

# Reaction 856
reaction('IC3H7O2 + C3H8 => IC3H7O2H + NC3H7', [1.700000e+13, 0.0, 20460.0])

# Reaction 857
reaction('IC3H7O2H + NC3H7 => IC3H7O2 + C3H8', [2.093000e+13, -0.511, 5000.0])

# Reaction 858
reaction('IC3H7O2 + C3H8 => IC3H7O2H + IC3H7', [2.000000e+12, 0.0, 17000.0])

# Reaction 859
reaction('IC3H7O2H + IC3H7 => IC3H7O2 + C3H8', [3.885000e+11, -0.197, 4080.0])

# Reaction 860
reaction('C3H8 + CH3CO3 => IC3H7 + CH3CO3H', [2.000000e+12, 0.0, 17000.0])

# Reaction 861
reaction('IC3H7 + CH3CO3H => C3H8 + CH3CO3', [3.104000e+10, 0.305, 8650.0])

# Reaction 862
reaction('C3H8 + CH3CO3 => NC3H7 + CH3CO3H', [1.700000e+13, 0.0, 20460.0])

# Reaction 863
reaction('NC3H7 + CH3CO3H => C3H8 + CH3CO3', [1.673000e+12, -0.009, 9570.0])

# Reaction 864
reaction('C3H8 + O2CHO => NC3H7 + HO2CHO', [5.520000e+04, 2.55, 16480.0])

# Reaction 865
reaction('NC3H7 + HO2CHO => C3H8 + O2CHO', [1.187000e-08, 5.54, -1920.0])

# Reaction 866
reaction('C3H8 + O2CHO => IC3H7 + HO2CHO', [1.475000e+04, 2.6, 13910.0])

# Reaction 867
reaction('IC3H7 + HO2CHO => C3H8 + O2CHO', [7.838000e-06, 4.65, -30.0])

# Reaction 868
reaction('IC3H7 => H + C3H6', [6.919000e+13, -0.025, 37690.0])

# Reaction 869
reaction('H + C3H6 => IC3H7', [2.640000e+13, 0.0, 2160.0])

# Reaction 870
reaction('IC3H7 + H => C2H5 + CH3', [2.000000e+13, 0.0, 0.0])

# Reaction 871
reaction('C2H5 + CH3 => IC3H7 + H', [4.344000e+07, 1.176, 8620.0])

# Reaction 872
reaction('IC3H7 + O2 => C3H6 + HO2', [4.500000e-19, 0.0, 5020.0])

# Reaction 873
reaction('C3H6 + HO2 => IC3H7 + O2', [2.000000e-19, 0.0, 17500.0])

# Reaction 874
reaction('IC3H7 + OH => C3H6 + H2O', [2.410000e+13, 0.0, 0.0])

# Reaction 875
reaction('C3H6 + H2O => IC3H7 + OH', [2.985000e+12, 0.57, 83820.0])

# Reaction 876
reaction('IC3H7 + O => CH3COCH3 + H', [4.818000e+13, 0.0, 0.0])

# Reaction 877
reaction('CH3COCH3 + H => IC3H7 + O', [1.293000e+16, -0.19, 79380.0])

# Reaction 878
reaction('IC3H7 + O => CH3CHO + CH3', [4.818000e+13, 0.0, 0.0])

# Reaction 879
reaction('CH3CHO + CH3 => IC3H7 + O', [1.279000e+11, 0.8, 86480.0])

# Reaction 880
reaction('NC3H7 => CH3 + C2H4', [9.970000e+40, -8.6, 41430.0])

# Reaction 881
reaction('CH3 + C2H4 => NC3H7', [1.898000e+34, -6.99, 17100.0])

# Reaction 882
reaction('NC3H7 => H + C3H6', [8.780000e+39, -8.1, 46580.0])

# Reaction 883
reaction('H + C3H6 => NC3H7', [2.070000e+37, -7.39, 12020.0])

# Reaction 884
reaction('NC3H7 + O2 => C3H6 + HO2', [3.000000e-19, 0.0, 3000.0])

# Reaction 885
reaction('C3H6 + HO2 => NC3H7 + O2', [2.000000e-19, 0.0, 17500.0])

# Reaction 886
reaction('C2H5CHO + NC3H7 => C2H5CO + C3H8', [1.700000e+12, 0.0, 8440.0])

# Reaction 887
reaction('C2H5CO + C3H8 => C2H5CHO + NC3H7', [1.900000e+14, 0.0, 18790.0])

# Reaction 888
reaction('C2H5CHO + IC3H7 => C2H5CO + C3H8', [1.700000e+12, 0.0, 8440.0])

# Reaction 889
reaction('C2H5CO + C3H8 => C2H5CHO + IC3H7', [1.900000e+14, 0.0, 18790.0])

# Reaction 890
reaction('C2H5CHO + C3H5-A => C2H5CO + C3H6', [1.700000e+12, 0.0, 8440.0])

# Reaction 891
reaction('C2H5CO + C3H6 => C2H5CHO + C3H5-A', [1.000000e+13, 0.0, 28000.0])

# Reaction 892
reaction('C3H6 => C2H3 + CH3', [2.730000e+62, -13.28, 123200.0])

# Reaction 893
reaction('C2H3 + CH3 => C3H6', [6.822000e+53, -11.779, 20550.0])

# Reaction 894
reaction('C3H6 => C3H5-A + H', [2.010000e+61, -13.26, 118500.0])

# Reaction 895
reaction('C3H5-A + H => C3H6', [2.041000e+61, -13.52, 30610.0])

# Reaction 896
reaction('C3H6 => C3H5-S + H', [7.710000e+69, -16.09, 140000.0])

# Reaction 897
reaction('C3H5-S + H => C3H6', [2.551000e+67, -15.867, 28690.0])

# Reaction 898
reaction('C3H6 => C3H5-T + H', [5.620000e+71, -16.58, 139300.0])

# Reaction 899
reaction('C3H5-T + H => C3H6', [4.260000e+68, -16.164, 30080.0])

# Reaction 900
reaction('C3H6 + O => C2H5 + HCO', [1.580000e+07, 1.76, -1216.0])

# Reaction 901
reaction('C2H5 + HCO => C3H6 + O', [9.188000e+01, 2.725, 23110.0])

# Reaction 902
reaction('C3H6 + O => CH2CO + CH3 + H', [2.500000e+07, 1.76, 76.0])

# Reaction 903
reaction('CH2CO + CH3 + H => C3H6 + O', [0.000000e+00, 0.0, 0.0])

# Reaction 904
reaction('C3H6 + O => CH3CHCO + H + H', [2.500000e+07, 1.76, 76.0])

# Reaction 905
reaction('CH3CHCO + H + H => C3H6 + O', [0.000000e+00, 0.0, 0.0])

# Reaction 906
reaction('C3H6 + O => C3H5-A + OH', [5.240000e+11, 0.7, 5884.0])

# Reaction 907
reaction('C3H5-A + OH => C3H6 + O', [1.104000e+11, 0.697, 20150.0])

# Reaction 908
reaction('C3H6 + O => C3H5-S + OH', [1.200000e+11, 0.7, 8959.0])

# Reaction 909
reaction('C3H5-S + OH => C3H6 + O', [8.239000e+07, 1.18, -207.0])

# Reaction 910
reaction('C3H6 + O => C3H5-T + OH', [6.030000e+10, 0.7, 7632.0])

# Reaction 911
reaction('C3H5-T + OH => C3H6 + O', [9.483000e+06, 1.373, 576.0])

# Reaction 912
reaction('C3H6 + OH => C3H5-A + H2O', [3.120000e+06, 2.0, -298.0])

# Reaction 913
reaction('C3H5-A + H2O => C3H6 + OH', [1.343000e+07, 1.909, 30270.0])

# Reaction 914
reaction('C3H6 + OH => C3H5-S + H2O', [2.110000e+06, 2.0, 2778.0])

# Reaction 915
reaction('C3H5-S + H2O => C3H6 + OH', [2.959000e+04, 2.393, 9916.0])

# Reaction 916
reaction('C3H6 + OH => C3H5-T + H2O', [1.110000e+06, 2.0, 1451.0])

# Reaction 917
reaction('C3H5-T + H2O => C3H6 + OH', [3.565000e+03, 2.586, 10700.0])

# Reaction 918
reaction('C3H6 + HO2 => C3H5-A + H2O2', [2.700000e+04, 2.5, 12340.0])

# Reaction 919
reaction('C3H5-A + H2O2 => C3H6 + HO2', [6.341000e+06, 1.82, 12010.0])

# Reaction 920
reaction('C3H6 + HO2 => C3H5-S + H2O2', [1.800000e+04, 2.5, 27620.0])

# Reaction 921
reaction('C3H5-S + H2O2 => C3H6 + HO2', [1.377000e+04, 2.304, 3864.0])

# Reaction 922
reaction('C3H6 + HO2 => C3H5-T + H2O2', [9.000000e+03, 2.5, 23590.0])

# Reaction 923
reaction('C3H5-T + H2O2 => C3H6 + HO2', [1.577000e+03, 2.497, 1941.0])

# Reaction 924
reaction('C3H6 + H => C3H5-A + H2', [1.730000e+05, 2.5, 2492.0])

# Reaction 925
reaction('C3H5-A + H2 => C3H6 + H', [7.023000e+04, 2.515, 18170.0])

# Reaction 926
reaction('C3H6 + H => C3H5-S + H2', [8.040000e+05, 2.5, 12280.0])

# Reaction 927
reaction('C3H5-S + H2 => C3H6 + H', [1.063000e+03, 2.999, 4526.0])

# Reaction 928
reaction('C3H6 + H => C3H5-T + H2', [4.050000e+05, 2.5, 9794.0])

# Reaction 929
reaction('C3H5-T + H2 => C3H6 + H', [1.227000e+02, 3.192, 4150.0])

# Reaction 930
reaction('C3H6 + H => C2H4 + CH3', [2.300000e+13, 0.0, 2547.0])

# Reaction 931
reaction('C2H4 + CH3 => C3H6 + H', [7.272000e+07, 1.271, 11200.0])

# Reaction 932
reaction('C3H6 + O2 => C3H5-A + HO2', [4.000000e+12, 0.0, 39900.0])

# Reaction 933
reaction('C3H5-A + HO2 => C3H6 + O2', [8.514000e+12, -0.333, 887.0])

# Reaction 934
reaction('C3H6 + O2 => C3H5-S + HO2', [2.000000e+12, 0.0, 62900.0])

# Reaction 935
reaction('C3H5-S + HO2 => C3H6 + O2', [1.387000e+10, 0.151, 459.0])

# Reaction 936
reaction('C3H6 + O2 => C3H5-T + HO2', [1.400000e+12, 0.0, 60700.0])

# Reaction 937
reaction('C3H5-T + HO2 => C3H6 + O2', [2.224000e+09, 0.344, 369.0])

# Reaction 938
reaction('C3H6 + CH3 => C3H5-A + CH4', [2.210000e+00, 3.5, 5675.0])

# Reaction 939
reaction('C3H5-A + CH4 => C3H6 + CH3', [8.184000e+02, 3.07, 22890.0])

# Reaction 940
reaction('C3H6 + CH3 => C3H5-S + CH4', [1.348000e+00, 3.5, 12850.0])

# Reaction 941
reaction('C3H5-S + CH4 => C3H6 + CH3', [1.626000e+00, 3.553, 6635.0])

# Reaction 942
reaction('C3H6 + CH3 => C3H5-T + CH4', [8.400000e-01, 3.5, 11660.0])

# Reaction 943
reaction('C3H5-T + CH4 => C3H6 + CH3', [2.322000e-01, 3.746, 7552.0])

# Reaction 944
reaction('C3H6 + C2H5 => C3H5-A + C2H6', [1.000000e+11, 0.0, 9800.0])

# Reaction 945
reaction('C3H5-A + C2H6 => C3H6 + C2H5', [5.369000e+05, 1.33, 16440.0])

# Reaction 946
reaction('C3H6 + CH3CO3 => C3H5-A + CH3CO3H', [3.240000e+11, 0.0, 14900.0])

# Reaction 947
reaction('C3H5-A + CH3CO3H => C3H6 + CH3CO3', [2.000000e+10, 0.0, 15000.0])

# Reaction 948
reaction('C3H6 + CH3O2 => C3H5-A + CH3O2H', [3.240000e+11, 0.0, 14900.0])

# Reaction 949
reaction('C3H5-A + CH3O2H => C3H6 + CH3O2', [2.000000e+10, 0.0, 15000.0])

# Reaction 950
reaction('C3H6 + HO2 => C3H6O1-2 + OH', [1.290000e+12, 0.0, 14900.0])

# Reaction 951
reaction('C3H6O1-2 + OH => C3H6 + HO2', [1.000000e-10, 0.0, 0.0])

# Reaction 952
reaction('C3H6 + C2H5O2 => C3H5-A + C2H5O2H', [3.240000e+11, 0.0, 14900.0])

# Reaction 953
reaction('C3H5-A + C2H5O2H => C3H6 + C2H5O2', [2.000000e+10, 0.0, 15000.0])

# Reaction 954
reaction('C3H6 + NC3H7O2 => C3H5-A + NC3H7O2H', [3.240000e+11, 0.0, 14900.0])

# Reaction 955
reaction('C3H5-A + NC3H7O2H => C3H6 + NC3H7O2', [2.000000e+10, 0.0, 15000.0])

# Reaction 956
reaction('C3H6 + IC3H7O2 => C3H5-A + IC3H7O2H', [3.240000e+11, 0.0, 14900.0])

# Reaction 957
reaction('C3H5-A + IC3H7O2H => C3H6 + IC3H7O2', [2.000000e+10, 0.0, 15000.0])

# Reaction 958
reaction('C3H6OH => C3H6 + OH', [4.343000e+15, -0.805, 27900.0])

# Reaction 959
reaction('C3H6 + OH => C3H6OH', [9.930000e+11, 0.0, -960.0])

# Reaction 960
reaction('HOC3H6O2 => C3H6OH + O2', [2.873000e+19, -1.897, 34290.0])

# Reaction 961
reaction('C3H6OH + O2 => HOC3H6O2', [1.200000e+11, 0.0, -1100.0])

# Reaction 962
reaction('HOC3H6O2 => CH3CHO + CH2O + OH', [1.250000e+10, 0.0, 18900.0])

# Reaction 963
reaction('CH3CHO + CH2O + OH => HOC3H6O2', [0.000000e+00, 0.0, 0.0])

# Reaction 964
reaction('C3H5-A => C2H2 + CH3', [2.397000e+48, -9.9, 82080.0])

# Reaction 965
reaction('C2H2 + CH3 => C3H5-A', [2.610000e+46, -9.82, 36950.0])

# Reaction 966
reaction('C3H5-A => C3H4-A + H', [4.194000e+13, 0.216, 61930.0])

# Reaction 967
reaction('C3H4-A + H => C3H5-A', [2.400000e+11, 0.69, 3007.0])

# Reaction 968
reaction('C3H5-A + HO2 => C3H5O + OH', [7.000000e+12, 0.0, -1000.0])

# Reaction 969
reaction('C3H5O + OH => C3H5-A + HO2', [1.605000e+12, 0.06, 11660.0])

# Reaction 970
reaction('C3H5-A + CH3O2 => C3H5O + CH3O', [7.000000e+12, 0.0, -1000.0])

# Reaction 971
reaction('C3H5O + CH3O => C3H5-A + CH3O2', [1.990000e+15, -0.74, 17020.0])

# Reaction 972
reaction('C3H5-A + H => C3H4-A + H2', [1.232000e+03, 3.035, 2582.0])

# Reaction 973
reaction('C3H4-A + H2 => C3H5-A + H', [2.818000e+00, 3.784, 47220.0])

# Reaction 974
reaction('C3H5-A + CH3 => C3H4-A + CH4', [1.000000e+11, 0.0, 0.0])

# Reaction 975
reaction('C3H4-A + CH4 => C3H5-A + CH3', [4.921000e+12, 0.05, 47780.0])

# Reaction 976
reaction('C3H5-A + C2H5 => C2H6 + C3H4-A', [4.000000e+11, 0.0, 0.0])

# Reaction 977
reaction('C2H6 + C3H4-A => C3H5-A + C2H5', [1.802000e+12, 0.05, 40330.0])

# Reaction 978
reaction('C3H5-A + C2H5 => C2H4 + C3H6', [4.000000e+11, 0.0, 0.0])

# Reaction 979
reaction('C2H4 + C3H6 => C3H5-A + C2H5', [6.937000e+16, -1.33, 52800.0])

# Reaction 980
reaction('C3H5-A + C2H3 => C2H4 + C3H4-A', [1.000000e+12, 0.0, 0.0])

# Reaction 981
reaction('C2H4 + C3H4-A => C3H5-A + C2H3', [1.624000e+13, 0.05, 48190.0])

# Reaction 982
reaction('C3H4-A + C3H6 => C3H5-A + C3H5-A', [4.749000e+08, 0.734, 28700.0])

# Reaction 983
reaction('C3H5-A + C3H5-A => C3H4-A + C3H6', [8.430000e+10, 0.0, -262.0])

# Reaction 984
reaction('C3H5-A + O2 => C3H4-A + HO2', [2.180000e+21, -2.85, 30760.0])

# Reaction 985
reaction('C3H4-A + HO2 => C3H5-A + O2', [2.614000e+19, -2.449, 20710.0])

# Reaction 986
reaction('C3H5-A + O2 => CH2CHO + CH2O', [7.140000e+15, -1.21, 21050.0])

# Reaction 987
reaction('CH2CHO + CH2O => C3H5-A + O2', [4.944000e+16, -1.4, 88620.0])

# Reaction 988
reaction('C3H5-A + O2 => C2H3CHO + OH', [2.470000e+13, -0.44, 23020.0])

# Reaction 989
reaction('C2H3CHO + OH => C3H5-A + O2', [1.989000e+13, -0.609, 75140.0])

# Reaction 990
reaction('C3H5-A + O2 => C2H2 + CH2O + OH', [9.720000e+29, -5.71, 21450.0])

# Reaction 991
reaction('C2H2 + CH2O + OH => C3H5-A + O2', [0.000000e+00, 0.0, 0.0])

# Reaction 992
reaction('C3H5-S => C2H2 + CH3', [9.598000e+39, -8.17, 42030.0])

# Reaction 993
reaction('C2H2 + CH3 => C3H5-S', [1.610000e+40, -8.58, 20330.0])

# Reaction 994
reaction('C3H5-S => C3H4-P + H', [4.187000e+15, -0.79, 37480.0])

# Reaction 995
reaction('C3H4-P + H => C3H5-S', [5.800000e+12, 0.0, 3100.0])

# Reaction 996
reaction('C3H5-S + O2 => CH3CHO + HCO', [4.335000e+12, 0.0, 0.0])

# Reaction 997
reaction('CH3CHO + HCO => C3H5-S + O2', [1.611000e+17, -1.27, 96530.0])

# Reaction 998
reaction('C3H5-S + H => C3H4-A + H2', [3.333000e+12, 0.0, 0.0])

# Reaction 999
reaction('C3H4-A + H2 => C3H5-S + H', [7.977000e+12, 0.11, 68860.0])

# Reaction 1000
reaction('C3H5-S + CH3 => C3H4-A + CH4', [1.000000e+11, 0.0, 0.0])

# Reaction 1001
reaction('C3H4-A + CH4 => C3H5-S + CH3', [6.253000e+12, 0.11, 69340.0])

# Reaction 1002
reaction('C3H5-T => C2H2 + CH3', [2.163000e+40, -8.31, 45110.0])

# Reaction 1003
reaction('C2H2 + CH3 => C3H5-T', [1.610000e+40, -8.58, 20330.0])

# Reaction 1004
reaction('C3H5-T => C3H4-A + H', [3.508000e+14, -0.44, 40890.0])

# Reaction 1005
reaction('C3H4-A + H => C3H5-T', [8.500000e+12, 0.0, 2000.0])

# Reaction 1006
reaction('C3H5-T => C3H4-P + H', [1.075000e+15, -0.6, 38490.0])

# Reaction 1007
reaction('C3H4-P + H => C3H5-T', [6.500000e+12, 0.0, 2000.0])

# Reaction 1008
reaction('C3H5-T + O2 => C3H4-A + HO2', [1.890000e+30, -5.59, 15540.0])

# Reaction 1009
reaction('C3H4-A + HO2 => C3H5-T + O2', [3.037000e+31, -5.865, 26810.0])

# Reaction 1010
reaction('C3H5-T + O2 => CH3COCH2 + O', [3.810000e+17, -1.36, 5580.0])

# Reaction 1011
reaction('CH3COCH2 + O => C3H5-T + O2', [2.000000e+11, 0.0, 17500.0])

# Reaction 1012
reaction('C3H5-T + O2 => CH2O + CH3CO', [3.710000e+25, -3.96, 7043.0])

# Reaction 1013
reaction('CH2O + CH3CO => C3H5-T + O2', [1.872000e+27, -4.43, 101200.0])

# Reaction 1014
reaction('C3H5-T + H => C3H4-P + H2', [3.333000e+12, 0.0, 0.0])

# Reaction 1015
reaction('C3H4-P + H2 => C3H5-T + H', [2.138000e+16, -0.88, 71050.0])

# Reaction 1016
reaction('C3H5-T + CH3 => C3H4-P + CH4', [1.000000e+11, 0.0, 0.0])

# Reaction 1017
reaction('C3H4-P + CH4 => C3H5-T + CH3', [1.676000e+16, -0.88, 71530.0])

# Reaction 1018
three_body_reaction('C3H4-A + M => C3H3 + H + M', [1.143000e+17, 0.0, 70000.0])

# Reaction 1019
three_body_reaction('C3H3 + H + M => C3H4-A + M', [1.798000e+15, -0.38, 10610.0])

# Reaction 1020
reaction('C3H4-A => C3H4-P', [1.202000e+15, 0.0, 92400.0])

# Reaction 1021
reaction('C3H4-P => C3H4-A', [3.222000e+18, -0.99, 96590.0])

# Reaction 1022
reaction('C3H4-A + O2 => C3H3 + HO2', [4.000000e+13, 0.0, 39160.0])

# Reaction 1023
reaction('C3H3 + HO2 => C3H4-A + O2', [3.170000e+11, -0.086, 311.0])

# Reaction 1024
reaction('C3H4-A + HO2 => CH2CO + CH2 + OH', [4.000000e+12, 0.0, 19000.0])

# Reaction 1025
reaction('CH2CO + CH2 + OH => C3H4-A + HO2', [1.000000e+00, 0.0, 0.0])

# Reaction 1026
reaction('C3H4-A + OH => CH2CO + CH3', [3.120000e+12, 0.0, -397.0])

# Reaction 1027
reaction('CH2CO + CH3 => C3H4-A + OH', [1.806000e+17, -1.38, 36070.0])

# Reaction 1028
reaction('C3H4-A + OH => C3H3 + H2O', [1.000000e+07, 2.0, 1000.0])

# Reaction 1029
reaction('C3H3 + H2O => C3H4-A + OH', [1.602000e+05, 2.157, 31730.0])

# Reaction 1030
reaction('C3H4-A + O => C2H4 + CO', [7.800000e+12, 0.0, 1600.0])

# Reaction 1031
reaction('C2H4 + CO => C3H4-A + O', [3.269000e+08, 1.252, 121900.0])

# Reaction 1032
reaction('C3H4-A + O => C2H2 + CH2O', [3.000000e-03, 4.61, -4243.0])

# Reaction 1033
reaction('C2H2 + CH2O => C3H4-A + O', [2.320000e+02, 3.23, 81190.0])

# Reaction 1034
reaction('C3H4-A + H => C3H3 + H2', [2.000000e+07, 2.0, 5000.0])

# Reaction 1035
reaction('C3H3 + H2 => C3H4-A + H', [3.022000e+04, 2.262, 20840.0])

# Reaction 1036
reaction('C3H4-A + CH3 => C3H3 + CH4', [3.670000e-02, 4.01, 6830.0])

# Reaction 1037
reaction('C3H3 + CH4 => C3H4-A + CH3', [5.060000e-02, 3.826, 24210.0])

# Reaction 1038
reaction('C3H4-A + C3H5-A => C3H3 + C3H6', [2.000000e+11, 0.0, 7700.0])

# Reaction 1039
reaction('C3H3 + C3H6 => C3H4-A + C3H5-A', [2.644000e+19, -2.71, 42140.0])

# Reaction 1040
reaction('C3H4-A + C2H => C3H3 + C2H2', [1.000000e+13, 0.0, 0.0])

# Reaction 1041
reaction('C3H3 + C2H2 => C3H4-A + C2H', [1.420000e+16, -1.38, 53820.0])

# Reaction 1042
three_body_reaction('C3H4-P + M => C3H3 + H + M', [1.143000e+17, 0.0, 70000.0])

# Reaction 1043
three_body_reaction('C3H3 + H + M => C3H4-P + M', [6.708000e+11, 0.61, 6420.0])

# Reaction 1044
reaction('C3H4-P => C2H + CH3', [4.200000e+16, 0.0, 100000.0])

# Reaction 1045
reaction('C2H + CH3 => C3H4-P', [1.018000e+12, 0.61, -1600.0])

# Reaction 1046
reaction('C3H4-P + O2 => HCCO + OH + CH2', [1.000000e+07, 1.5, 30100.0])

# Reaction 1047
reaction('HCCO + OH + CH2 => C3H4-P + O2', [1.000000e+00, 0.0, 0.0])

# Reaction 1048
reaction('C3H4-P + O2 => C3H3 + HO2', [2.000000e+13, 0.0, 41600.0])

# Reaction 1049
reaction('C3H3 + HO2 => C3H4-P + O2', [6.371000e+11, -0.208, 1021.0])

# Reaction 1050
reaction('C3H4-P + HO2 => C2H4 + CO + OH', [3.000000e+12, 0.0, 19000.0])

# Reaction 1051
reaction('C2H4 + CO + OH => C3H4-P + HO2', [1.000000e+00, 0.0, 0.0])

# Reaction 1052
reaction('C3H4-P + OH => C3H3 + H2O', [1.000000e+07, 2.0, 1000.0])

# Reaction 1053
reaction('C3H3 + H2O => C3H4-P + OH', [6.441000e+05, 2.034, 30000.0])

# Reaction 1054
reaction('C3H4-P + OH => CH2CO + CH3', [5.000000e-04, 4.5, -1000.0])

# Reaction 1055
reaction('CH2CO + CH3 => C3H4-P + OH', [1.079000e-02, 4.11, 31280.0])

# Reaction 1056
reaction('C3H4-P + O => C2H3 + HCO', [3.200000e+12, 0.0, 2010.0])

# Reaction 1057
reaction('C2H3 + HCO => C3H4-P + O', [2.548000e+12, -0.39, 32350.0])

# Reaction 1058
reaction('C3H4-P + O => HCCO + CH3', [9.600000e+08, 1.0, 0.0])

# Reaction 1059
reaction('HCCO + CH3 => C3H4-P + O', [1.430000e+04, 1.793, 26990.0])

# Reaction 1060
reaction('C3H4-P + O => HCCO + CH2 + H', [3.200000e-19, 0.0, 2010.0])

# Reaction 1061
reaction('HCCO + CH2 + H => C3H4-P + O', [1.000000e-30, 0.0, 0.0])

# Reaction 1062
reaction('C3H4-P + O => C3H3 + OH', [7.650000e+08, 1.5, 8600.0])

# Reaction 1063
reaction('C3H3 + OH => C3H4-P + O', [2.177000e+08, 1.31, 22470.0])

# Reaction 1064
reaction('C3H4-P + H => C3H3 + H2', [2.000000e+07, 2.0, 5000.0])

# Reaction 1065
reaction('C3H3 + H2 => C3H4-P + H', [1.215000e+05, 2.14, 19110.0])

# Reaction 1066
reaction('C3H4-P + CH3 => C3H3 + CH4', [1.500000e+00, 3.5, 5600.0])

# Reaction 1067
reaction('C3H3 + CH4 => C3H4-P + CH3', [8.313000e+00, 3.195, 21250.0])

# Reaction 1068
reaction('C3H4-P + C2H => C3H3 + C2H2', [1.000000e+12, 0.0, 0.0])

# Reaction 1069
reaction('C3H3 + C2H2 => C3H4-P + C2H', [5.297000e+11, -0.39, 49630.0])

# Reaction 1070
reaction('C3H4-P + C2H3 => C3H3 + C2H4', [1.000000e+12, 0.0, 7700.0])

# Reaction 1071
reaction('C3H3 + C2H4 => C3H4-P + C2H3', [9.541000e+11, -0.39, 52450.0])

# Reaction 1072
reaction('C3H4-P + C3H5-A => C3H3 + C3H6', [1.000000e+12, 0.0, 7700.0])

# Reaction 1073
reaction('C3H3 + C3H6 => C3H4-P + C3H5-A', [4.931000e+16, -1.73, 37950.0])

# Reaction 1074
reaction('C3H3 + O => CH2O + C2H', [1.000000e+13, 0.0, 0.0])

# Reaction 1075
reaction('CH2O + C2H => C3H3 + O', [5.446000e+14, 0.0, 31610.0])

# Reaction 1076
reaction('C3H3 + OH => C3H2 + H2O', [1.000000e+13, 0.0, 0.0])

# Reaction 1077
reaction('C3H2 + H2O => C3H3 + OH', [1.343000e+15, 0.0, 15680.0])

# Reaction 1078
reaction('C3H3 + O2 => CH2CO + HCO', [3.010000e+10, 0.0, 2870.0])

# Reaction 1079
reaction('CH2CO + HCO => C3H3 + O2', [4.881000e+11, 0.0, 59470.0])

# Reaction 1080
reaction('C3H3 + CH3 => C2H5 + C2H', [4.299000e+15, -0.79, 45630.0])

# Reaction 1081
reaction('C2H5 + C2H => C3H3 + CH3', [1.810000e+13, 0.0, 0.0])

# Reaction 1082
reaction('C3H2 + O2 => HCO + HCCO', [5.000000e+13, 0.0, 0.0])

# Reaction 1083
reaction('HCO + HCCO => C3H2 + O2', [2.326000e+14, -0.214, 77190.0])

# Reaction 1084
reaction('C3H4-A + HO2 => C2H4 + CO + OH', [1.000000e+12, 0.0, 14000.0])

# Reaction 1085
reaction('C2H4 + CO + OH => C3H4-A + HO2', [1.000000e+00, 0.0, 0.0])

# Reaction 1086
reaction('C3H4-A + HO2 => C3H3 + H2O2', [3.000000e+13, 0.0, 14000.0])

# Reaction 1087
reaction('C3H3 + H2O2 => C3H4-A + HO2', [1.551000e+16, -1.38, 44000.0])

# Reaction 1088
reaction('C2H2 + CH3 => C3H4-P + H', [4.229000e+08, 1.143, 12090.0])

# Reaction 1089
reaction('C3H4-P + H => C2H2 + CH3', [1.000000e+14, 0.0, 4000.0])

# Reaction 1090
reaction('C2H2 + CH3 => C3H4-A + H', [6.740000e+19, -2.08, 31590.0])

# Reaction 1091
reaction('C3H4-A + H => C2H2 + CH3', [6.407000e+25, -3.345, 21770.0])

# Reaction 1092
reaction('C3H3 + H => C3H2 + H2', [5.000000e+13, 0.0, 0.0])

# Reaction 1093
reaction('C3H2 + H2 => C3H3 + H', [5.999000e+07, 1.365, 4110.0])

# Reaction 1094
reaction('C3H2 + OH => C2H2 + HCO', [5.000000e+13, 0.0, 0.0])

# Reaction 1095
reaction('C2H2 + HCO => C3H2 + OH', [2.282000e+16, -0.254, 75020.0])

# Reaction 1096
reaction('C3H2 + O2 => HCCO + CO + H', [5.000000e+13, 0.0, 0.0])

# Reaction 1097
reaction('HCCO + CO + H => C3H2 + O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1098
reaction('CH3CHCO + OH => C2H5 + CO2', [1.730000e+12, 0.0, -1010.0])

# Reaction 1099
reaction('C2H5 + CO2 => CH3CHCO + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 1100
reaction('CH3CHCO + OH => SC2H4OH + CO', [2.000000e+12, 0.0, -1010.0])

# Reaction 1101
reaction('SC2H4OH + CO => CH3CHCO + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 1102
reaction('CH3CHCO + H => C2H5 + CO', [4.400000e+12, 0.0, 1459.0])

# Reaction 1103
reaction('C2H5 + CO => CH3CHCO + H', [0.000000e+00, 0.0, 0.0])

# Reaction 1104
reaction('CH3CHCO + O => CH3CHO + CO', [3.200000e+12, 0.0, -437.0])

# Reaction 1105
reaction('CH3CHO + CO => CH3CHCO + O', [0.000000e+00, 0.0, 0.0])

# Reaction 1106
reaction('NC3H7 + HO2 => NC3H7O + OH', [7.000000e+12, 0.0, -1000.0])

# Reaction 1107
reaction('NC3H7O + OH => NC3H7 + HO2', [6.220000e+15, -0.692, 25310.0])

# Reaction 1108
reaction('IC3H7 + HO2 => IC3H7O + OH', [7.000000e+12, 0.0, -1000.0])

# Reaction 1109
reaction('IC3H7O + OH => IC3H7 + HO2', [1.051000e+16, -0.557, 27320.0])

# Reaction 1110
reaction('CH3O2 + NC3H7 => CH3O + NC3H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1111
reaction('CH3O + NC3H7O => CH3O2 + NC3H7', [3.890000e+14, -0.394, 29550.0])

# Reaction 1112
reaction('CH3O2 + IC3H7 => CH3O + IC3H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1113
reaction('CH3O + IC3H7O => CH3O2 + IC3H7', [6.573000e+14, -0.258, 31560.0])

# Reaction 1114
reaction('NC3H7O2 => NC3H7 + O2', [2.400000e+20, -1.616, 35960.0])

# Reaction 1115
reaction('NC3H7 + O2 => NC3H7O2', [4.520000e+12, 0.0, 0.0])

# Reaction 1116
reaction('IC3H7O2 => IC3H7 + O2', [3.132000e+22, -2.167, 38160.0])

# Reaction 1117
reaction('IC3H7 + O2 => IC3H7O2', [7.540000e+12, 0.0, 0.0])

# Reaction 1118
reaction('NC3H7O2 + CH2O => NC3H7O2H + HCO', [5.600000e+12, 0.0, 13600.0])

# Reaction 1119
reaction('NC3H7O2H + HCO => NC3H7O2 + CH2O', [8.000000e+11, 0.0, 10000.0])

# Reaction 1120
reaction('NC3H7O2 + CH3CHO => NC3H7O2H + CH3CO', [2.800000e+12, 0.0, 13600.0])

# Reaction 1121
reaction('NC3H7O2H + CH3CO => NC3H7O2 + CH3CHO', [1.000000e+12, 0.0, 10000.0])

# Reaction 1122
reaction('IC3H7O2 + CH2O => IC3H7O2H + HCO', [5.600000e+12, 0.0, 13600.0])

# Reaction 1123
reaction('IC3H7O2H + HCO => IC3H7O2 + CH2O', [8.000000e+11, 0.0, 10000.0])

# Reaction 1124
reaction('IC3H7O2 + CH3CHO => IC3H7O2H + CH3CO', [2.800000e+12, 0.0, 13600.0])

# Reaction 1125
reaction('IC3H7O2H + CH3CO => IC3H7O2 + CH3CHO', [1.000000e+12, 0.0, 10000.0])

# Reaction 1126
reaction('NC3H7O2 + HO2 => NC3H7O2H + O2', [1.750000e+10, 0.0, -3275.0])

# Reaction 1127
reaction('NC3H7O2H + O2 => NC3H7O2 + HO2', [3.841000e+13, -0.795, 33610.0])

# Reaction 1128
reaction('IC3H7O2 + HO2 => IC3H7O2H + O2', [1.750000e+10, 0.0, -3275.0])

# Reaction 1129
reaction('IC3H7O2H + O2 => IC3H7O2 + HO2', [3.855000e+13, -0.796, 33610.0])

# Reaction 1130
reaction('C2H4 + NC3H7O2 => C2H3 + NC3H7O2H', [1.130000e+13, 0.0, 30430.0])

# Reaction 1131
reaction('C2H3 + NC3H7O2H => C2H4 + NC3H7O2', [3.000000e+12, 0.0, 11500.0])

# Reaction 1132
reaction('C2H4 + IC3H7O2 => C2H3 + IC3H7O2H', [1.130000e+13, 0.0, 30430.0])

# Reaction 1133
reaction('C2H3 + IC3H7O2H => C2H4 + IC3H7O2', [3.000000e+12, 0.0, 11500.0])

# Reaction 1134
reaction('CH3OH + NC3H7O2 => CH2OH + NC3H7O2H', [6.300000e+12, 0.0, 19360.0])

# Reaction 1135
reaction('CH2OH + NC3H7O2H => CH3OH + NC3H7O2', [1.000000e+09, 0.0, 10000.0])

# Reaction 1136
reaction('CH3OH + IC3H7O2 => CH2OH + IC3H7O2H', [6.300000e+12, 0.0, 19360.0])

# Reaction 1137
reaction('CH2OH + IC3H7O2H => CH3OH + IC3H7O2', [1.000000e+09, 0.0, 10000.0])

# Reaction 1138
reaction('C2H3CHO + NC3H7O2 => C2H3CO + NC3H7O2H', [2.800000e+12, 0.0, 13600.0])

# Reaction 1139
reaction('C2H3CO + NC3H7O2H => C2H3CHO + NC3H7O2', [1.000000e+12, 0.0, 10000.0])

# Reaction 1140
reaction('C2H3CHO + IC3H7O2 => C2H3CO + IC3H7O2H', [2.800000e+12, 0.0, 13600.0])

# Reaction 1141
reaction('C2H3CO + IC3H7O2H => C2H3CHO + IC3H7O2', [1.000000e+12, 0.0, 10000.0])

# Reaction 1142
reaction('CH4 + NC3H7O2 => CH3 + NC3H7O2H', [1.120000e+13, 0.0, 24640.0])

# Reaction 1143
reaction('CH3 + NC3H7O2H => CH4 + NC3H7O2', [7.430000e+11, 0.0, 5500.0])

# Reaction 1144
reaction('CH4 + IC3H7O2 => CH3 + IC3H7O2H', [1.120000e+13, 0.0, 24640.0])

# Reaction 1145
reaction('CH3 + IC3H7O2H => CH4 + IC3H7O2', [7.430000e+11, 0.0, 5500.0])

# Reaction 1146
reaction('NC3H7O2 + CH3O2 => NC3H7O + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1147
reaction('NC3H7O + CH3O + O2 => NC3H7O2 + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1148
reaction('IC3H7O2 + CH3O2 => IC3H7O + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1149
reaction('IC3H7O + CH3O + O2 => IC3H7O2 + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1150
reaction('H2 + NC3H7O2 => H + NC3H7O2H', [3.010000e+13, 0.0, 26030.0])

# Reaction 1151
reaction('H + NC3H7O2H => H2 + NC3H7O2', [4.800000e+13, 0.0, 7950.0])

# Reaction 1152
reaction('H2 + IC3H7O2 => H + IC3H7O2H', [3.010000e+13, 0.0, 26030.0])

# Reaction 1153
reaction('H + IC3H7O2H => H2 + IC3H7O2', [4.800000e+13, 0.0, 7950.0])

# Reaction 1154
reaction('IC3H7O2 + C2H6 => IC3H7O2H + C2H5', [1.700000e+13, 0.0, 20460.0])

# Reaction 1155
reaction('IC3H7O2H + C2H5 => IC3H7O2 + C2H6', [5.000000e+11, 0.0, 6500.0])

# Reaction 1156
reaction('NC3H7O2 + C2H6 => NC3H7O2H + C2H5', [1.700000e+13, 0.0, 20460.0])

# Reaction 1157
reaction('NC3H7O2H + C2H5 => NC3H7O2 + C2H6', [5.000000e+11, 0.0, 6500.0])

# Reaction 1158
reaction('IC3H7O2 + C2H5CHO => IC3H7O2H + C2H5CO', [2.000000e+11, 0.0, 9500.0])

# Reaction 1159
reaction('IC3H7O2H + C2H5CO => IC3H7O2 + C2H5CHO', [5.000000e+09, 0.0, 10000.0])

# Reaction 1160
reaction('NC3H7O2 + C2H5CHO => NC3H7O2H + C2H5CO', [2.000000e+11, 0.0, 9500.0])

# Reaction 1161
reaction('NC3H7O2H + C2H5CO => NC3H7O2 + C2H5CHO', [5.000000e+09, 0.0, 10000.0])

# Reaction 1162
reaction('IC3H7O2 + CH3CO3 => IC3H7O + CH3CO2 + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1163
reaction('IC3H7O + CH3CO2 + O2 => IC3H7O2 + CH3CO3', [0.000000e+00, 0.0, 0.0])

# Reaction 1164
reaction('NC3H7O2 + CH3CO3 => NC3H7O + CH3CO2 + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1165
reaction('NC3H7O + CH3CO2 + O2 => NC3H7O2 + CH3CO3', [0.000000e+00, 0.0, 0.0])

# Reaction 1166
reaction('IC3H7O2 + C2H5O2 => IC3H7O + C2H5O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1167
reaction('IC3H7O + C2H5O + O2 => IC3H7O2 + C2H5O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1168
reaction('NC3H7O2 + C2H5O2 => NC3H7O + C2H5O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1169
reaction('NC3H7O + C2H5O + O2 => NC3H7O2 + C2H5O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1170
reaction('IC3H7O2 + IC3H7O2 => O2 + IC3H7O + IC3H7O', [1.400000e+16, -1.61, 1860.0])

# Reaction 1171
reaction('O2 + IC3H7O + IC3H7O => IC3H7O2 + IC3H7O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1172
reaction('NC3H7O2 + NC3H7O2 => O2 + NC3H7O + NC3H7O', [1.400000e+16, -1.61, 1860.0])

# Reaction 1173
reaction('O2 + NC3H7O + NC3H7O => NC3H7O2 + NC3H7O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1174
reaction('IC3H7O2 + NC3H7O2 => IC3H7O + NC3H7O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1175
reaction('IC3H7O + NC3H7O + O2 => IC3H7O2 + NC3H7O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1176
reaction('IC3H7O2 + CH3 => IC3H7O + CH3O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1177
reaction('IC3H7O + CH3O => IC3H7O2 + CH3', [1.145000e+11, 0.695, 26720.0])

# Reaction 1178
reaction('IC3H7O2 + C2H5 => IC3H7O + C2H5O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1179
reaction('IC3H7O + C2H5O => IC3H7O2 + C2H5', [1.628000e+13, 0.094, 29130.0])

# Reaction 1180
reaction('IC3H7O2 + IC3H7 => IC3H7O + IC3H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1181
reaction('IC3H7O + IC3H7O => IC3H7O2 + IC3H7', [2.778000e+13, 0.261, 29800.0])

# Reaction 1182
reaction('IC3H7O2 + NC3H7 => IC3H7O + NC3H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1183
reaction('IC3H7O + NC3H7O => IC3H7O2 + NC3H7', [1.644000e+13, 0.125, 27800.0])

# Reaction 1184
reaction('IC3H7O2 + C3H5-A => IC3H7O + C3H5O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1185
reaction('IC3H7O + C3H5O => IC3H7O2 + C3H5-A', [4.242000e+09, 0.877, 14140.0])

# Reaction 1186
reaction('NC3H7O2 + CH3 => NC3H7O + CH3O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1187
reaction('NC3H7O + CH3O => NC3H7O2 + CH3', [5.303000e+12, 0.009, 26920.0])

# Reaction 1188
reaction('NC3H7O2 + C2H5 => NC3H7O + C2H5O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1189
reaction('NC3H7O + C2H5O => NC3H7O2 + C2H5', [7.540000e+14, -0.592, 29330.0])

# Reaction 1190
reaction('NC3H7O2 + IC3H7 => NC3H7O + IC3H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1191
reaction('NC3H7O + IC3H7O => NC3H7O2 + IC3H7', [1.286000e+15, -0.425, 30000.0])

# Reaction 1192
reaction('NC3H7O2 + NC3H7 => NC3H7O + NC3H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1193
reaction('NC3H7O + NC3H7O => NC3H7O2 + NC3H7', [7.612000e+14, -0.561, 28000.0])

# Reaction 1194
reaction('NC3H7O2 + C3H5-A => NC3H7O + C3H5O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1195
reaction('NC3H7O + C3H5O => NC3H7O2 + C3H5-A', [1.964000e+11, 0.191, 14340.0])

# Reaction 1196
reaction('NC3H7O2H => NC3H7O + OH', [1.500000e+16, 0.0, 42500.0])

# Reaction 1197
reaction('NC3H7O + OH => NC3H7O2H', [1.143000e+08, 1.719, -4034.0])

# Reaction 1198
reaction('IC3H7O2H => IC3H7O + OH', [9.450000e+15, 0.0, 42600.0])

# Reaction 1199
reaction('IC3H7O + OH => IC3H7O2H', [1.550000e+06, 2.406, -4132.0])

# Reaction 1200
reaction('NC3H7O => C2H5 + CH2O', [2.716000e+21, -2.449, 15700.0])

# Reaction 1201
reaction('C2H5 + CH2O => NC3H7O', [1.000000e+11, 0.0, 3496.0])

# Reaction 1202
reaction('NC3H7O => C2H5CHO + H', [8.899000e+10, 0.746, 19800.0])

# Reaction 1203
reaction('C2H5CHO + H => NC3H7O', [4.000000e+12, 0.0, 6260.0])

# Reaction 1204
reaction('IC3H7O => CH3 + CH3CHO', [5.328000e+19, -1.696, 17140.0])

# Reaction 1205
reaction('CH3 + CH3CHO => IC3H7O', [1.000000e+11, 0.0, 9256.0])

# Reaction 1206
reaction('IC3H7O => CH3COCH3 + H', [8.663000e+14, -0.483, 20080.0])

# Reaction 1207
reaction('CH3COCH3 + H => IC3H7O', [2.000000e+12, 0.0, 7270.0])

# Reaction 1208
reaction('IC3H7O + O2 => CH3COCH3 + HO2', [9.090000e+09, 0.0, 390.0])

# Reaction 1209
reaction('CH3COCH3 + HO2 => IC3H7O + O2', [1.000000e+11, 0.0, 32000.0])

# Reaction 1210
reaction('NC3H7O2 => C3H6OOH1-2', [6.000000e+11, 0.0, 26850.0])

# Reaction 1211
reaction('C3H6OOH1-2 => NC3H7O2', [1.117000e+08, 0.583, 11720.0])

# Reaction 1212
reaction('NC3H7O2 => C3H6OOH1-3', [1.125000e+11, 0.0, 24400.0])

# Reaction 1213
reaction('C3H6OOH1-3 => NC3H7O2', [2.716000e+11, -0.507, 8936.0])

# Reaction 1214
reaction('IC3H7O2 => C3H6OOH2-1', [1.800000e+12, 0.0, 29400.0])

# Reaction 1215
reaction('C3H6OOH2-1 => IC3H7O2', [1.122000e+10, 0.119, 11810.0])

# Reaction 1216
reaction('IC3H7O2 => C3H6OOH2-2', [1.230000e+35, -6.96, 48880.0])

# Reaction 1217
reaction('C3H6OOH2-2 => IC3H7O2', [2.384000e+34, -7.06, 44940.0])

# Reaction 1218
reaction('C3H6OOH1-2 => C3H6O1-2 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 1219
reaction('C3H6O1-2 + OH => C3H6OOH1-2', [1.150000e+11, 0.49, 38370.0])

# Reaction 1220
reaction('C3H6OOH1-3 => C3H6O1-3 + OH', [7.500000e+10, 0.0, 15250.0])

# Reaction 1221
reaction('C3H6O1-3 + OH => C3H6OOH1-3', [1.186000e+06, 1.765, 28710.0])

# Reaction 1222
reaction('C3H6OOH2-1 => C3H6O1-2 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 1223
reaction('C3H6O1-2 + OH => C3H6OOH2-1', [2.780000e+08, 1.191, 36090.0])

# Reaction 1224
reaction('C3H6OOH1-2 => C3H6 + HO2', [7.834000e+15, -1.3, 15950.0])

# Reaction 1225
reaction('C3H6 + HO2 => C3H6OOH1-2', [1.000000e+11, 0.0, 11000.0])

# Reaction 1226
reaction('C3H6OOH2-1 => C3H6 + HO2', [3.239000e+18, -2.0, 18970.0])

# Reaction 1227
reaction('C3H6 + HO2 => C3H6OOH2-1', [1.000000e+11, 0.0, 11750.0])

# Reaction 1228
reaction('C3H6OOH1-3 => OH + CH2O + C2H4', [3.035000e+15, -0.79, 27400.0])

# Reaction 1229
reaction('OH + CH2O + C2H4 => C3H6OOH1-3', [0.000000e+00, 0.0, 0.0])

# Reaction 1230
reaction('C3H6OOH2-1 => C2H3OOH + CH3', [6.540000e+27, -5.14, 38320.0])

# Reaction 1231
reaction('C2H3OOH + CH3 => C3H6OOH2-1', [4.460000e+22, -4.24, 10630.0])

# Reaction 1232
reaction('C3H6OOH1-2 => C2H4 + CH2O + OH', [1.310000e+33, -7.01, 48120.0])

# Reaction 1233
reaction('C2H4 + CH2O + OH => C3H6OOH1-2', [0.000000e+00, 0.0, 0.0])

# Reaction 1234
reaction('C3H6OOH2-2 => CH3COCH3 + OH', [9.000000e+14, 0.0, 1500.0])

# Reaction 1235
reaction('CH3COCH3 + OH => C3H6OOH2-2', [1.021000e+14, 0.31, 36750.0])

# Reaction 1236
reaction('C3H6OOH1-2O2 => C3H6OOH1-2 + O2', [2.391000e+25, -2.945, 40100.0])

# Reaction 1237
reaction('C3H6OOH1-2 + O2 => C3H6OOH1-2O2', [5.000000e+12, 0.0, 0.0])

# Reaction 1238
reaction('C3H6OOH1-3O2 => C3H6OOH1-3 + O2', [2.853000e+20, -1.626, 35690.0])

# Reaction 1239
reaction('C3H6OOH1-3 + O2 => C3H6OOH1-3O2', [4.520000e+12, 0.0, 0.0])

# Reaction 1240
reaction('C3H6OOH2-1O2 => C3H6OOH2-1 + O2', [5.227000e+22, -2.244, 37820.0])

# Reaction 1241
reaction('C3H6OOH2-1 + O2 => C3H6OOH2-1O2', [4.520000e+12, 0.0, 0.0])

# Reaction 1242
reaction('C3H6OOH1-2O2 => C3KET12 + OH', [6.000000e+11, 0.0, 26400.0])

# Reaction 1243
reaction('C3KET12 + OH => C3H6OOH1-2O2', [9.249000e+04, 1.329, 48920.0])

# Reaction 1244
reaction('C3H6OOH1-3O2 => C3KET13 + OH', [7.500000e+10, 0.0, 21400.0])

# Reaction 1245
reaction('C3KET13 + OH => C3H6OOH1-3O2', [4.101000e+03, 1.496, 44740.0])

# Reaction 1246
reaction('C3H6OOH2-1O2 => C3KET21 + OH', [3.000000e+11, 0.0, 23850.0])

# Reaction 1247
reaction('C3KET21 + OH => C3H6OOH2-1O2', [1.397000e+03, 1.834, 49750.0])

# Reaction 1248
reaction('C3H6OOH2-1O2 => C3H51-2,3OOH', [1.125000e+11, 0.0, 24400.0])

# Reaction 1249
reaction('C3H51-2,3OOH => C3H6OOH2-1O2', [2.391000e+11, -0.499, 8920.0])

# Reaction 1250
reaction('C3H6OOH1-2O2 => C3H51-2,3OOH', [9.000000e+11, 0.0, 29400.0])

# Reaction 1251
reaction('C3H51-2,3OOH => C3H6OOH1-2O2', [1.913000e+12, -0.499, 13920.0])

# Reaction 1252
reaction('C3H51-2,3OOH => AC3H5OOH + HO2', [2.560000e+13, -0.49, 17770.0])

# Reaction 1253
reaction('AC3H5OOH + HO2 => C3H51-2,3OOH', [3.180000e+15, -1.16, 12040.0])

# Reaction 1254
reaction('C3H52-1,3OOH => C3H6OOH1-3O2', [1.255000e+12, -0.36, 13940.0])

# Reaction 1255
reaction('C3H6OOH1-3O2 => C3H52-1,3OOH', [6.000000e+11, 0.0, 26850.0])

# Reaction 1256
reaction('C3H52-1,3OOH => AC3H5OOH + HO2', [1.150000e+14, -0.63, 17250.0])

# Reaction 1257
reaction('AC3H5OOH + HO2 => C3H52-1,3OOH', [1.564000e+11, 0.12, 10200.0])

# Reaction 1258
reaction('C3KET12 => CH3CHO + HCO + OH', [9.450000e+15, 0.0, 43000.0])

# Reaction 1259
reaction('CH3CHO + HCO + OH => C3KET12', [0.000000e+00, 0.0, 0.0])

# Reaction 1260
reaction('C3KET13 => CH2O + CH2CHO + OH', [1.000000e+16, 0.0, 43000.0])

# Reaction 1261
reaction('CH2O + CH2CHO + OH => C3KET13', [0.000000e+00, 0.0, 0.0])

# Reaction 1262
reaction('C3KET21 => CH2O + CH3CO + OH', [1.000000e+16, 0.0, 43000.0])

# Reaction 1263
reaction('CH2O + CH3CO + OH => C3KET21', [0.000000e+00, 0.0, 0.0])

# Reaction 1264
reaction('AC3H5OOH => C3H5O + OH', [3.880000e+19, -1.46, 45370.0])

# Reaction 1265
reaction('C3H5O + OH => AC3H5OOH', [2.000000e+13, 0.0, 0.0])

# Reaction 1266
reaction('C3H5O => C2H3CHO + H', [1.000000e+14, 0.0, 29100.0])

# Reaction 1267
reaction('C2H3CHO + H => C3H5O', [1.676000e+14, -0.156, 19690.0])

# Reaction 1268
reaction('C3H5O => C2H3 + CH2O', [1.464000e+20, -1.968, 35090.0])

# Reaction 1269
reaction('C2H3 + CH2O => C3H5O', [1.500000e+11, 0.0, 10600.0])

# Reaction 1270
reaction('C3H5O + O2 => C2H3CHO + HO2', [1.000000e+12, 0.0, 6000.0])

# Reaction 1271
reaction('C2H3CHO + HO2 => C3H5O + O2', [1.288000e+11, 0.0, 32000.0])

# Reaction 1272
reaction('C2H3OOH => CH2CHO + OH', [8.400000e+14, 0.0, 43000.0])

# Reaction 1273
reaction('CH2CHO + OH => C2H3OOH', [1.000000e+11, 0.0, 0.0])

# Reaction 1274
reaction('C3H6O1-2 => C2H4 + CH2O', [6.000000e+14, 0.0, 60000.0])

# Reaction 1275
reaction('C2H4 + CH2O => C3H6O1-2', [2.970000e+11, 1.0, 31080.0])

# Reaction 1276
reaction('C3H6O1-2 + OH => CH2O + C2H3 + H2O', [5.000000e+12, 0.0, 0.0])

# Reaction 1277
reaction('CH2O + C2H3 + H2O => C3H6O1-2 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 1278
reaction('C3H6O1-2 + H => CH2O + C2H3 + H2', [2.630000e+07, 2.0, 5000.0])

# Reaction 1279
reaction('CH2O + C2H3 + H2 => C3H6O1-2 + H', [0.000000e+00, 0.0, 0.0])

# Reaction 1280
reaction('C3H6O1-2 + O => CH2O + C2H3 + OH', [8.430000e+13, 0.0, 5200.0])

# Reaction 1281
reaction('CH2O + C2H3 + OH => C3H6O1-2 + O', [0.000000e+00, 0.0, 0.0])

# Reaction 1282
reaction('C3H6O1-2 + HO2 => CH2O + C2H3 + H2O2', [1.000000e+13, 0.0, 15000.0])

# Reaction 1283
reaction('CH2O + C2H3 + H2O2 => C3H6O1-2 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 1284
reaction('C3H6O1-2 + CH3O2 => CH2O + C2H3 + CH3O2H', [1.000000e+13, 0.0, 19000.0])

# Reaction 1285
reaction('CH2O + C2H3 + CH3O2H => C3H6O1-2 + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1286
reaction('C3H6O1-2 + CH3 => CH2O + C2H3 + CH4', [2.000000e+11, 0.0, 10000.0])

# Reaction 1287
reaction('CH2O + C2H3 + CH4 => C3H6O1-2 + CH3', [0.000000e+00, 0.0, 0.0])

# Reaction 1288
reaction('C3H6O1-3 => C2H4 + CH2O', [6.000000e+14, 0.0, 60000.0])

# Reaction 1289
reaction('C2H4 + CH2O => C3H6O1-3', [2.970000e+11, 0.0, 31080.0])

# Reaction 1290
reaction('C3H6O1-3 + OH => CH2O + C2H3 + H2O', [5.000000e+12, 0.0, 0.0])

# Reaction 1291
reaction('CH2O + C2H3 + H2O => C3H6O1-3 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 1292
reaction('C3H6O1-3 + O => CH2O + C2H3 + OH', [8.430000e+13, 0.0, 5200.0])

# Reaction 1293
reaction('CH2O + C2H3 + OH => C3H6O1-3 + O', [0.000000e+00, 0.0, 0.0])

# Reaction 1294
reaction('C3H6O1-3 + H => CH2O + C2H3 + H2', [2.630000e+07, 2.0, 5000.0])

# Reaction 1295
reaction('CH2O + C2H3 + H2 => C3H6O1-3 + H', [0.000000e+00, 0.0, 0.0])

# Reaction 1296
reaction('C3H6O1-3 + CH3O2 => CH2O + C2H3 + CH3O2H', [1.000000e+13, 0.0, 19000.0])

# Reaction 1297
reaction('CH2O + C2H3 + CH3O2H => C3H6O1-3 + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1298
reaction('C3H6O1-3 + HO2 => CH2O + C2H3 + H2O2', [1.000000e+13, 0.0, 15000.0])

# Reaction 1299
reaction('CH2O + C2H3 + H2O2 => C3H6O1-3 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 1300
reaction('C3H6O1-3 + CH3 => CH2O + C2H3 + CH4', [2.000000e+11, 0.0, 10000.0])

# Reaction 1301
reaction('CH2O + C2H3 + CH4 => C3H6O1-3 + CH3', [0.000000e+00, 0.0, 0.0])

# Reaction 1302
reaction('IC3H7O2 => C3H6 + HO2', [1.015000e+43, -9.409, 41490.0])

# Reaction 1303
reaction('C3H6 + HO2 => IC3H7O2', [1.954000e+33, -7.289, 16670.0])

# Reaction 1304
reaction('NC3H7O2 => C3H6 + HO2', [5.044000e+38, -8.112, 40490.0])

# Reaction 1305
reaction('C3H6 + HO2 => NC3H7O2', [1.198000e+30, -6.229, 20420.0])

# Reaction 1306
falloff_reaction('C4H10 (+ M) <=> C2H5 + C2H5 (+ M)',
                kf=[2.720000e+15, 0.0, 75610.0],
                kf0=[4.720000e+18, 0.0, 49576.0],
                falloff=Troe(A=0.72, T3=1500.0, T1=1e-10, T2=10000000000.0))

# Reaction 1307
falloff_reaction('C4H10 (+ M) <=> NC3H7 + CH3 (+ M)',
                kf=[4.280000e+14, 0.0, 69900.0],
                kf0=[5.340000e+17, 0.0, 42959.0],
                falloff=Troe(A=0.72, T3=1500.0, T1=1e-10, T2=10000000000.0))

# Reaction 1308
reaction('C4H10 => PC4H9 + H', [1.344000e+17, -0.356, 101200.0])

# Reaction 1309
reaction('PC4H9 + H => C4H10', [3.610000e+13, 0.0, 0.0])

# Reaction 1310
reaction('C4H10 => SC4H9 + H', [1.975000e+18, -0.694, 98720.0])

# Reaction 1311
reaction('SC4H9 + H => C4H10', [3.610000e+13, 0.0, 0.0])

# Reaction 1312
reaction('C4H10 + O2 => PC4H9 + HO2', [6.000000e+13, 0.0, 52340.0])

# Reaction 1313
reaction('PC4H9 + HO2 => C4H10 + O2', [3.377000e+10, 0.284, -19.0])

# Reaction 1314
reaction('C4H10 + O2 => SC4H9 + HO2', [4.000000e+13, 0.0, 49800.0])

# Reaction 1315
reaction('SC4H9 + HO2 => C4H10 + O2', [1.532000e+09, 0.622, -49.0])

# Reaction 1316
reaction('C4H10 + C3H5-A => PC4H9 + C3H6', [7.940000e+11, 0.0, 20500.0])

# Reaction 1317
reaction('PC4H9 + C3H6 => C4H10 + C3H5-A', [1.000000e+12, 0.0, 20000.0])

# Reaction 1318
reaction('C4H10 + C3H5-A => SC4H9 + C3H6', [3.160000e+11, 0.0, 16400.0])

# Reaction 1319
reaction('SC4H9 + C3H6 => C4H10 + C3H5-A', [1.000000e+12, 0.0, 20000.0])

# Reaction 1320
reaction('C4H10 + C2H5 => PC4H9 + C2H6', [1.580000e+11, 0.0, 12300.0])

# Reaction 1321
reaction('PC4H9 + C2H6 => C4H10 + C2H5', [3.560000e+10, 0.0, 12920.0])

# Reaction 1322
reaction('C4H10 + C2H5 => SC4H9 + C2H6', [1.000000e+11, 0.0, 10400.0])

# Reaction 1323
reaction('SC4H9 + C2H6 => C4H10 + C2H5', [7.120000e+10, 0.0, 9917.0])

# Reaction 1324
reaction('C4H10 + C2H3 => PC4H9 + C2H4', [1.000000e+12, 0.0, 18000.0])

# Reaction 1325
reaction('PC4H9 + C2H4 => C4H10 + C2H3', [2.570000e+12, 0.0, 25380.0])

# Reaction 1326
reaction('C4H10 + C2H3 => SC4H9 + C2H4', [8.000000e+11, 0.0, 16800.0])

# Reaction 1327
reaction('SC4H9 + C2H4 => C4H10 + C2H3', [2.050000e+12, 0.0, 24180.0])

# Reaction 1328
reaction('C4H10 + CH3 => PC4H9 + CH4', [9.040000e-01, 3.65, 7154.0])

# Reaction 1329
reaction('PC4H9 + CH4 => C4H10 + CH3', [8.853000e-02, 3.836, 11020.0])

# Reaction 1330
reaction('C4H10 + CH3 => SC4H9 + CH4', [3.020000e+00, 3.46, 5481.0])

# Reaction 1331
reaction('SC4H9 + CH4 => C4H10 + CH3', [2.013000e-02, 3.984, 11860.0])

# Reaction 1332
reaction('C4H10 + H => PC4H9 + H2', [1.880000e+05, 2.75, 6280.0])

# Reaction 1333
reaction('PC4H9 + H2 => C4H10 + H', [2.018000e+01, 3.382, 8610.0])

# Reaction 1334
reaction('C4H10 + H => SC4H9 + H2', [2.600000e+06, 2.4, 4471.0])

# Reaction 1335
reaction('SC4H9 + H2 => C4H10 + H', [1.900000e+01, 3.37, 9311.0])

# Reaction 1336
reaction('C4H10 + OH => PC4H9 + H2O', [1.054000e+10, 0.97, 1586.0])

# Reaction 1337
reaction('PC4H9 + H2O => C4H10 + OH', [1.200000e+07, 1.496, 18810.0])

# Reaction 1338
reaction('C4H10 + OH => SC4H9 + H2O', [9.340000e+07, 1.61, -35.0])

# Reaction 1339
reaction('SC4H9 + H2O => C4H10 + OH', [7.235000e+03, 2.474, 19700.0])

# Reaction 1340
reaction('C4H10 + O => PC4H9 + OH', [1.130000e+14, 0.0, 7850.0])

# Reaction 1341
reaction('PC4H9 + OH => C4H10 + O', [1.480000e+13, 0.0, 12240.0])

# Reaction 1342
reaction('C4H10 + O => SC4H9 + OH', [5.620000e+13, 0.0, 5200.0])

# Reaction 1343
reaction('SC4H9 + OH => C4H10 + O', [7.350000e+12, 0.0, 9590.0])

# Reaction 1344
reaction('C4H10 + HO2 => PC4H9 + H2O2', [4.080000e+01, 3.59, 17160.0])

# Reaction 1345
reaction('PC4H9 + H2O2 => C4H10 + HO2', [2.534000e+00, 3.527, 3490.0])

# Reaction 1346
reaction('C4H10 + HO2 => SC4H9 + H2O2', [1.264000e+02, 3.37, 13720.0])

# Reaction 1347
reaction('SC4H9 + H2O2 => C4H10 + HO2', [5.343000e-01, 3.645, 2558.0])

# Reaction 1348
reaction('C4H10 + CH3O => PC4H9 + CH3OH', [3.000000e+11, 0.0, 7000.0])

# Reaction 1349
reaction('PC4H9 + CH3OH => C4H10 + CH3O', [1.220000e+10, 0.0, 50000.0])

# Reaction 1350
reaction('C4H10 + CH3O => SC4H9 + CH3OH', [6.000000e+11, 0.0, 7000.0])

# Reaction 1351
reaction('SC4H9 + CH3OH => C4H10 + CH3O', [2.441000e+10, 0.0, 50000.0])

# Reaction 1352
reaction('C4H10 + C2H5O => PC4H9 + C2H5OH', [3.000000e+11, 0.0, 7000.0])

# Reaction 1353
reaction('PC4H9 + C2H5OH => C4H10 + C2H5O', [1.000000e+10, 0.0, 9000.0])

# Reaction 1354
reaction('C4H10 + C2H5O => SC4H9 + C2H5OH', [6.000000e+11, 0.0, 7000.0])

# Reaction 1355
reaction('SC4H9 + C2H5OH => C4H10 + C2H5O', [1.000000e+10, 0.0, 9000.0])

# Reaction 1356
reaction('C4H10 + PC4H9 => SC4H9 + C4H10', [1.000000e+11, 0.0, 10400.0])

# Reaction 1357
reaction('SC4H9 + C4H10 => C4H10 + PC4H9', [1.500000e+11, 0.0, 12300.0])

# Reaction 1358
reaction('C4H10 + CH3CO3 => PC4H9 + CH3CO3H', [1.700000e+13, 0.0, 20460.0])

# Reaction 1359
reaction('PC4H9 + CH3CO3H => C4H10 + CH3CO3', [5.000000e+11, 0.0, 6500.0])

# Reaction 1360
reaction('C4H10 + CH3CO3 => SC4H9 + CH3CO3H', [1.120000e+13, 0.0, 17700.0])

# Reaction 1361
reaction('SC4H9 + CH3CO3H => C4H10 + CH3CO3', [5.000000e+11, 0.0, 6500.0])

# Reaction 1362
reaction('C4H10 + O2CHO => PC4H9 + HO2CHO', [1.680000e+13, 0.0, 20440.0])

# Reaction 1363
reaction('PC4H9 + HO2CHO => C4H10 + O2CHO', [3.680000e+00, 2.99, 2505.0])

# Reaction 1364
reaction('C4H10 + O2CHO => SC4H9 + HO2CHO', [1.120000e+13, 0.0, 17690.0])

# Reaction 1365
reaction('SC4H9 + HO2CHO => C4H10 + O2CHO', [7.595000e+03, 2.06, 4266.0])

# Reaction 1366
reaction('CH3O2 + C4H10 => CH3O2H + PC4H9', [1.386000e+00, 3.97, 18280.0])

# Reaction 1367
reaction('CH3O2H + PC4H9 => CH3O2 + C4H10', [1.675000e+00, 3.462, 3016.0])

# Reaction 1368
reaction('CH3O2 + C4H10 => CH3O2H + SC4H9', [2.037000e+01, 3.58, 14810.0])

# Reaction 1369
reaction('CH3O2H + SC4H9 => CH3O2 + C4H10', [1.675000e+00, 3.41, 2050.0])

# Reaction 1370
reaction('C2H5O2 + C4H10 => C2H5O2H + PC4H9', [4.080000e+01, 3.59, 17160.0])

# Reaction 1371
reaction('C2H5O2H + PC4H9 => C2H5O2 + C4H10', [4.938000e+01, 3.082, 1899.0])

# Reaction 1372
reaction('C2H5O2 + C4H10 => C2H5O2H + SC4H9', [1.264000e+02, 3.37, 13720.0])

# Reaction 1373
reaction('C2H5O2H + SC4H9 => C2H5O2 + C4H10', [1.041000e+01, 3.2, 967.0])

# Reaction 1374
reaction('NC3H7O2 + C4H10 => NC3H7O2H + PC4H9', [1.700000e+13, 0.0, 20460.0])

# Reaction 1375
reaction('NC3H7O2H + PC4H9 => NC3H7O2 + C4H10', [5.000000e+11, 0.0, 6500.0])

# Reaction 1376
reaction('NC3H7O2 + C4H10 => NC3H7O2H + SC4H9', [1.120000e+13, 0.0, 17700.0])

# Reaction 1377
reaction('NC3H7O2H + SC4H9 => NC3H7O2 + C4H10', [5.000000e+11, 0.0, 6500.0])

# Reaction 1378
reaction('IC3H7O2 + C4H10 => IC3H7O2H + PC4H9', [1.700000e+13, 0.0, 20460.0])

# Reaction 1379
reaction('IC3H7O2H + PC4H9 => IC3H7O2 + C4H10', [5.000000e+11, 0.0, 6500.0])

# Reaction 1380
reaction('IC3H7O2 + C4H10 => IC3H7O2H + SC4H9', [1.120000e+13, 0.0, 17700.0])

# Reaction 1381
reaction('IC3H7O2H + SC4H9 => IC3H7O2 + C4H10', [5.000000e+11, 0.0, 6500.0])

# Reaction 1382
reaction('PC4H9O2 + C3H8 => PC4H9O2H + NC3H7', [1.700000e+13, 0.0, 20460.0])

# Reaction 1383
reaction('PC4H9O2H + NC3H7 => PC4H9O2 + C3H8', [5.000000e+11, 0.0, 6500.0])

# Reaction 1384
reaction('PC4H9O2 + C3H8 => PC4H9O2H + IC3H7', [2.000000e+12, 0.0, 17000.0])

# Reaction 1385
reaction('PC4H9O2H + IC3H7 => PC4H9O2 + C3H8', [5.000000e+11, 0.0, 6500.0])

# Reaction 1386
reaction('PC4H9O2 + C4H10 => PC4H9O2H + PC4H9', [1.700000e+13, 0.0, 20460.0])

# Reaction 1387
reaction('PC4H9O2H + PC4H9 => PC4H9O2 + C4H10', [5.000000e+11, 0.0, 6500.0])

# Reaction 1388
reaction('PC4H9O2 + C4H10 => PC4H9O2H + SC4H9', [1.120000e+13, 0.0, 17700.0])

# Reaction 1389
reaction('PC4H9O2H + SC4H9 => PC4H9O2 + C4H10', [5.000000e+11, 0.0, 6500.0])

# Reaction 1390
reaction('SC4H9O2 + C3H8 => SC4H9O2H + NC3H7', [1.700000e+13, 0.0, 20460.0])

# Reaction 1391
reaction('SC4H9O2H + NC3H7 => SC4H9O2 + C3H8', [5.000000e+11, 0.0, 6500.0])

# Reaction 1392
reaction('SC4H9O2 + C3H8 => SC4H9O2H + IC3H7', [2.000000e+12, 0.0, 17000.0])

# Reaction 1393
reaction('SC4H9O2H + IC3H7 => SC4H9O2 + C3H8', [5.000000e+11, 0.0, 6500.0])

# Reaction 1394
reaction('SC4H9O2 + C4H10 => SC4H9O2H + PC4H9', [1.700000e+13, 0.0, 20460.0])

# Reaction 1395
reaction('SC4H9O2H + PC4H9 => SC4H9O2 + C4H10', [5.000000e+11, 0.0, 6500.0])

# Reaction 1396
reaction('SC4H9O2 + C4H10 => SC4H9O2H + SC4H9', [1.120000e+13, 0.0, 17700.0])

# Reaction 1397
reaction('SC4H9O2H + SC4H9 => SC4H9O2 + C4H10', [5.000000e+11, 0.0, 6500.0])

# Reaction 1398
reaction('PC4H9 => C2H5 + C2H4', [3.504000e+12, 0.463, 29470.0])

# Reaction 1399
reaction('C2H5 + C2H4 => PC4H9', [1.320000e+04, 2.48, 6130.0])

# Reaction 1400
reaction('SC4H9 => C3H6 + CH3', [4.803000e+10, 1.044, 30350.0])

# Reaction 1401
reaction('C3H6 + CH3 => SC4H9', [1.760000e+04, 2.48, 6130.0])

# Reaction 1402
reaction('PC4H9 => C4H8-1 + H', [2.622000e+12, 0.253, 35700.0])

# Reaction 1403
reaction('C4H8-1 + H => PC4H9', [2.500000e+11, 0.51, 2620.0])

# Reaction 1404
reaction('SC4H9 => C4H8-2 + H', [2.844000e+11, 0.337, 35520.0])

# Reaction 1405
reaction('C4H8-2 + H => SC4H9', [2.500000e+11, 0.51, 2620.0])

# Reaction 1406
reaction('SC4H9 => C4H8-1 + H', [3.027000e+11, 0.591, 36820.0])

# Reaction 1407
reaction('C4H8-1 + H => SC4H9', [4.240000e+11, 0.51, 1230.0])

# Reaction 1408
reaction('PC4H9 + O2 => C4H8-1 + HO2', [2.000000e-18, 0.0, 5000.0])

# Reaction 1409
reaction('C4H8-1 + HO2 => PC4H9 + O2', [2.000000e-19, 0.0, 17500.0])

# Reaction 1410
reaction('SC4H9 + O2 => C4H8-1 + HO2', [2.000000e-18, 0.0, 5000.0])

# Reaction 1411
reaction('C4H8-1 + HO2 => SC4H9 + O2', [2.000000e-19, 0.0, 17500.0])

# Reaction 1412
reaction('SC4H9 + O2 => C4H8-2 + HO2', [2.000000e-18, 0.0, 5000.0])

# Reaction 1413
reaction('C4H8-2 + HO2 => SC4H9 + O2', [2.000000e-19, 0.0, 17500.0])

# Reaction 1414
reaction('C4H8-1 => C3H5-A + CH3', [5.081000e+19, -1.256, 76510.0])

# Reaction 1415
reaction('C3H5-A + CH3 => C4H8-1', [1.350000e+13, 0.0, 0.0])

# Reaction 1416
reaction('C4H8-1 => C2H3 + C2H5', [2.882000e+23, -1.99, 101600.0])

# Reaction 1417
reaction('C2H3 + C2H5 => C4H8-1', [9.000000e+12, 0.0, 0.0])

# Reaction 1418
reaction('C4H8-1 => H + C4H71-3', [3.724000e+14, -0.111, 85200.0])

# Reaction 1419
reaction('H + C4H71-3 => C4H8-1', [5.000000e+13, 0.0, 0.0])

# Reaction 1420
reaction('C4H8-1 + O2 => C4H71-3 + HO2', [2.000000e+13, 0.0, 37190.0])

# Reaction 1421
reaction('C4H71-3 + HO2 => C4H8-1 + O2', [4.653000e+12, 0.07, -168.0])

# Reaction 1422
reaction('C4H8-1 + H => C4H71-1 + H2', [7.810000e+05, 2.5, 12290.0])

# Reaction 1423
reaction('C4H71-1 + H2 => C4H8-1 + H', [2.213000e+05, 2.36, 6469.0])

# Reaction 1424
reaction('C4H8-1 + H => C4H71-2 + H2', [3.900000e+05, 2.5, 5821.0])

# Reaction 1425
reaction('C4H71-2 + H2 => C4H8-1 + H', [2.558000e+04, 2.55, 2125.0])

# Reaction 1426
reaction('C4H8-1 + H => C4H71-3 + H2', [3.376000e+05, 2.36, 207.0])

# Reaction 1427
reaction('C4H71-3 + H2 => C4H8-1 + H', [4.323000e+06, 2.1, 20330.0])

# Reaction 1428
reaction('C4H8-1 + H => C4H71-4 + H2', [6.651000e+05, 2.54, 6756.0])

# Reaction 1429
reaction('C4H71-4 + H2 => C4H8-1 + H', [3.045000e+04, 2.54, 11030.0])

# Reaction 1430
reaction('C4H8-1 + OH => C4H71-1 + H2O', [2.140000e+06, 2.0, 2778.0])

# Reaction 1431
reaction('C4H71-1 + H2O => C4H8-1 + OH', [2.625000e+06, 1.86, 12120.0])

# Reaction 1432
reaction('C4H8-1 + OH => C4H71-2 + H2O', [2.220000e+06, 2.0, 1451.0])

# Reaction 1433
reaction('C4H71-2 + H2O => C4H8-1 + OH', [6.304000e+05, 2.05, 12910.0])

# Reaction 1434
reaction('C4H8-1 + OH => C4H71-3 + H2O', [2.764000e+04, 2.64, -1919.0])

# Reaction 1435
reaction('C4H71-3 + H2O => C4H8-1 + OH', [1.532000e+06, 2.38, 33360.0])

# Reaction 1436
reaction('C4H8-1 + OH => C4H71-4 + H2O', [5.270000e+09, 0.97, 1586.0])

# Reaction 1437
reaction('C4H71-4 + H2O => C4H8-1 + OH', [1.044000e+09, 0.97, 21010.0])

# Reaction 1438
reaction('C4H8-1 + CH3 => C4H71-3 + CH4', [3.690000e+00, 3.31, 4002.0])

# Reaction 1439
reaction('C4H71-3 + CH4 => C4H8-1 + CH3', [1.234000e+03, 3.05, 24610.0])

# Reaction 1440
reaction('C4H8-1 + CH3 => C4H71-4 + CH4', [4.520000e-01, 3.65, 7154.0])

# Reaction 1441
reaction('C4H71-4 + CH4 => C4H8-1 + CH3', [5.405000e-01, 3.65, 11910.0])

# Reaction 1442
reaction('C4H8-1 + HO2 => C4H71-3 + H2O2', [4.820000e+03, 2.55, 10530.0])

# Reaction 1443
reaction('C4H71-3 + H2O2 => C4H8-1 + HO2', [1.586000e+06, 1.96, 14350.0])

# Reaction 1444
reaction('C4H8-1 + HO2 => C4H71-4 + H2O2', [2.380000e+03, 2.55, 16490.0])

# Reaction 1445
reaction('C4H71-4 + H2O2 => C4H8-1 + HO2', [2.800000e+03, 2.22, 4460.0])

# Reaction 1446
reaction('C4H8-1 + CH3O2 => C4H71-3 + CH3O2H', [4.820000e+03, 2.55, 10530.0])

# Reaction 1447
reaction('C4H71-3 + CH3O2H => C4H8-1 + CH3O2', [3.303000e+06, 1.79, 11330.0])

# Reaction 1448
reaction('C4H8-1 + CH3O2 => C4H71-4 + CH3O2H', [2.380000e+03, 2.55, 16490.0])

# Reaction 1449
reaction('C4H71-4 + CH3O2H => C4H8-1 + CH3O2', [5.831000e+03, 2.04, 1440.0])

# Reaction 1450
reaction('C4H8-1 + CH3O => C4H71-3 + CH3OH', [4.000000e+01, 2.9, 8609.0])

# Reaction 1451
reaction('C4H71-3 + CH3OH => C4H8-1 + CH3O', [2.470000e+02, 2.67, 27000.0])

# Reaction 1452
reaction('C4H8-1 + CH3O => C4H71-4 + CH3OH', [2.170000e+11, 0.0, 6458.0])

# Reaction 1453
reaction('C4H71-4 + CH3OH => C4H8-1 + CH3O', [4.789000e+09, 0.02, 9002.0])

# Reaction 1454
reaction('C4H8-1 + CH3CO3 => C4H71-3 + CH3CO3H', [1.000000e+11, 0.0, 8000.0])

# Reaction 1455
reaction('C4H71-3 + CH3CO3H => C4H8-1 + CH3CO3', [2.000000e+10, 0.0, 10000.0])

# Reaction 1456
reaction('C4H8-1 + C3H5-A => C4H71-3 + C3H6', [7.900000e+10, 0.0, 12400.0])

# Reaction 1457
reaction('C4H71-3 + C3H6 => C4H8-1 + C3H5-A', [1.000000e+11, 0.0, 17500.0])

# Reaction 1458
reaction('C4H8-1 + C4H6 => C4H71-3 + C4H71-3', [2.350000e+12, 0.0, 46720.0])

# Reaction 1459
reaction('C4H71-3 + C4H71-3 => C4H8-1 + C4H6', [1.600000e+12, 0.0, 0.0])

# Reaction 1460
reaction('C4H8-1 + C2H5O2 => C4H71-3 + C2H5O2H', [1.400000e+12, 0.0, 14900.0])

# Reaction 1461
reaction('C4H71-3 + C2H5O2H => C4H8-1 + C2H5O2', [3.160000e+11, 0.0, 13000.0])

# Reaction 1462
reaction('C4H8-1 + NC3H7O2 => C4H71-3 + NC3H7O2H', [1.400000e+12, 0.0, 14900.0])

# Reaction 1463
reaction('C4H71-3 + NC3H7O2H => C4H8-1 + NC3H7O2', [3.160000e+11, 0.0, 13000.0])

# Reaction 1464
reaction('C4H8-1 + IC3H7O2 => C4H71-3 + IC3H7O2H', [1.400000e+12, 0.0, 14900.0])

# Reaction 1465
reaction('C4H71-3 + IC3H7O2H => C4H8-1 + IC3H7O2', [3.160000e+11, 0.0, 13000.0])

# Reaction 1466
reaction('C4H8-1 + PC4H9O2 => C4H71-3 + PC4H9O2H', [1.400000e+12, 0.0, 14900.0])

# Reaction 1467
reaction('C4H71-3 + PC4H9O2H => C4H8-1 + PC4H9O2', [3.160000e+11, 0.0, 13000.0])

# Reaction 1468
reaction('C4H8-1 + SC4H9O2 => C4H71-3 + SC4H9O2H', [1.400000e+12, 0.0, 14900.0])

# Reaction 1469
reaction('C4H71-3 + SC4H9O2H => C4H8-1 + SC4H9O2', [3.160000e+11, 0.0, 13000.0])

# Reaction 1470
reaction('C4H8-1 + CH3O2 => C4H8O1-2 + CH3O', [1.000000e+12, 0.0, 14340.0])

# Reaction 1471
reaction('C4H8O1-2 + CH3O => C4H8-1 + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1472
reaction('C4H8-2 => H + C4H71-3', [2.337000e+14, 0.143, 87890.0])

# Reaction 1473
reaction('H + C4H71-3 => C4H8-2', [5.000000e+13, 0.0, 0.0])

# Reaction 1474
reaction('C4H8-2 + O2 => C4H71-3 + HO2', [4.000000e+13, 0.0, 39390.0])

# Reaction 1475
reaction('C4H71-3 + HO2 => C4H8-2 + O2', [1.350000e+13, -0.18, -924.0])

# Reaction 1476
reaction('C4H8-2 + H => C4H71-3 + H2', [3.460000e+05, 2.5, 2492.0])

# Reaction 1477
reaction('C4H71-3 + H2 => C4H8-2 + H', [6.428000e+06, 1.99, 19660.0])

# Reaction 1478
reaction('C4H8-2 + OH => C4H71-3 + H2O', [6.240000e+06, 2.0, -298.0])

# Reaction 1479
reaction('C4H71-3 + H2O => C4H8-2 + OH', [5.019000e+08, 1.49, 32020.0])

# Reaction 1480
reaction('C4H8-2 + CH3 => C4H71-3 + CH4', [4.420000e+00, 3.5, 5675.0])

# Reaction 1481
reaction('C4H71-3 + CH4 => C4H8-2 + CH3', [2.145000e+03, 2.99, 23320.0])

# Reaction 1482
reaction('C4H8-2 + HO2 => C4H71-3 + H2O2', [1.928000e+04, 2.6, 13910.0])

# Reaction 1483
reaction('C4H71-3 + H2O2 => C4H8-2 + HO2', [9.205000e+06, 1.77, 14770.0])

# Reaction 1484
reaction('C4H8-2 + CH3O2 => C4H71-3 + CH3O2H', [1.928000e+04, 2.6, 13910.0])

# Reaction 1485
reaction('C4H71-3 + CH3O2H => C4H8-2 + CH3O2', [1.917000e+07, 1.59, 11750.0])

# Reaction 1486
reaction('C4H8-2 + CH3O => C4H71-3 + CH3OH', [1.800000e+01, 2.95, 11990.0])

# Reaction 1487
reaction('C4H71-3 + CH3OH => C4H8-2 + CH3O', [1.612000e+02, 2.47, 27420.0])

# Reaction 1488
reaction('C4H8-2 + C2H5O2 => C4H71-3 + C2H5O2H', [3.200000e+12, 0.0, 14900.0])

# Reaction 1489
reaction('C4H71-3 + C2H5O2H => C4H8-2 + C2H5O2', [1.585000e+11, 0.0, 14700.0])

# Reaction 1490
reaction('C4H8-2 + NC3H7O2 => C4H71-3 + NC3H7O2H', [3.200000e+12, 0.0, 14900.0])

# Reaction 1491
reaction('C4H71-3 + NC3H7O2H => C4H8-2 + NC3H7O2', [1.585000e+11, 0.0, 14700.0])

# Reaction 1492
reaction('C4H8-2 + IC3H7O2 => C4H71-3 + IC3H7O2H', [3.200000e+12, 0.0, 14900.0])

# Reaction 1493
reaction('C4H71-3 + IC3H7O2H => C4H8-2 + IC3H7O2', [1.585000e+11, 0.0, 14700.0])

# Reaction 1494
reaction('C4H8-2 + PC4H9O2 => C4H71-3 + PC4H9O2H', [3.200000e+12, 0.0, 14900.0])

# Reaction 1495
reaction('C4H71-3 + PC4H9O2H => C4H8-2 + PC4H9O2', [1.585000e+11, 0.0, 14700.0])

# Reaction 1496
reaction('C4H8-2 + SC4H9O2 => C4H71-3 + SC4H9O2H', [3.200000e+12, 0.0, 14900.0])

# Reaction 1497
reaction('C4H71-3 + SC4H9O2H => C4H8-2 + SC4H9O2', [1.585000e+11, 0.0, 14700.0])

# Reaction 1498
reaction('C4H8-1 + HO2 => C4H8O1-2 + OH', [1.000000e+12, 0.0, 14340.0])

# Reaction 1499
reaction('C4H8O1-2 + OH => C4H8-1 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 1500
reaction('C4H8-2 + HO2 => C4H8O2-3 + OH', [5.620000e+11, 0.0, 12310.0])

# Reaction 1501
reaction('C4H8O2-3 + OH => C4H8-2 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 1502
reaction('C4H8-2 + CH3O2 => C4H8O2-3 + CH3O', [5.620000e+11, 0.0, 12310.0])

# Reaction 1503
reaction('C4H8O2-3 + CH3O => C4H8-2 + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1504
reaction('PC4H8OH => C4H8-1 + OH', [1.079000e+16, -0.699, 28090.0])

# Reaction 1505
reaction('C4H8-1 + OH => PC4H8OH', [4.750000e+12, 0.0, -782.0])

# Reaction 1506
reaction('SC4H8OH => C4H8-2 + OH', [3.379000e+17, -1.253, 29920.0])

# Reaction 1507
reaction('C4H8-2 + OH => SC4H8OH', [4.750000e+12, 0.0, -782.0])

# Reaction 1508
reaction('C4H8OH-1O2 => PC4H8OH + O2', [6.753000e+20, -1.944, 35520.0])

# Reaction 1509
reaction('PC4H8OH + O2 => C4H8OH-1O2', [2.000000e+12, 0.0, 0.0])

# Reaction 1510
reaction('C4H8OH-2O2 => SC4H8OH + O2', [7.686000e+20, -1.968, 35510.0])

# Reaction 1511
reaction('SC4H8OH + O2 => C4H8OH-2O2', [2.000000e+12, 0.0, 0.0])

# Reaction 1512
reaction('C4H8OH-1O2 => C2H5CHO + CH2O + OH', [1.000000e+16, 0.0, 25000.0])

# Reaction 1513
reaction('C2H5CHO + CH2O + OH => C4H8OH-1O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1514
reaction('C4H8OH-2O2 => OH + CH3CHO + CH3CHO', [1.000000e+16, 0.0, 25000.0])

# Reaction 1515
reaction('OH + CH3CHO + CH3CHO => C4H8OH-2O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1516
reaction('C4H71-1 => C2H2 + C2H5', [1.066000e+15, -0.56, 30320.0])

# Reaction 1517
reaction('C2H2 + C2H5 => C4H71-1', [2.000000e+11, 0.0, 7800.0])

# Reaction 1518
reaction('C4H71-2 => C3H4-A + CH3', [9.589000e+14, -0.71, 31260.0])

# Reaction 1519
reaction('C3H4-A + CH3 => C4H71-2', [2.000000e+11, 0.0, 7800.0])

# Reaction 1520
reaction('C4H71-4 => C2H4 + C2H3', [8.766000e+12, -0.22, 36290.0])

# Reaction 1521
reaction('C2H4 + C2H3 => C4H71-4', [2.000000e+11, 0.0, 7800.0])

# Reaction 1522
reaction('C4H72-2 => C3H4-P + CH3', [6.326000e+10, 0.52, 30020.0])

# Reaction 1523
reaction('C3H4-P + CH3 => C4H72-2', [1.000000e+11, 0.0, 7800.0])

# Reaction 1524
reaction('C4H71-3 => C4H6 + H', [1.200000e+14, 0.0, 49300.0])

# Reaction 1525
reaction('C4H6 + H => C4H71-3', [4.000000e+13, 0.0, 1300.0])

# Reaction 1526
reaction('C4H71-3 + C2H5 => C4H8-1 + C2H4', [2.590000e+12, 0.0, -131.0])

# Reaction 1527
reaction('C4H8-1 + C2H4 => C4H71-3 + C2H5', [1.149000e+13, 0.06, 49440.0])

# Reaction 1528
reaction('C4H71-3 + CH3O => C4H8-1 + CH2O', [2.410000e+13, 0.0, 0.0])

# Reaction 1529
reaction('C4H8-1 + CH2O => C4H71-3 + CH3O', [2.482000e+12, 0.28, 66330.0])

# Reaction 1530
reaction('C4H71-3 + O => C2H3CHO + CH3', [6.030000e+13, 0.0, 0.0])

# Reaction 1531
reaction('C2H3CHO + CH3 => C4H71-3 + O', [3.385000e+15, -0.78, 81630.0])

# Reaction 1532
reaction('C4H71-3 + HO2 => C4H7O + OH', [9.640000e+12, 0.0, 0.0])

# Reaction 1533
reaction('C4H7O + OH => C4H71-3 + HO2', [7.290000e+15, -1.09, 15530.0])

# Reaction 1534
reaction('C4H71-3 + CH3O2 => C4H7O + CH3O', [9.640000e+12, 0.0, 0.0])

# Reaction 1535
reaction('C4H7O + CH3O => C4H71-3 + CH3O2', [7.120000e+17, -1.67, 20290.0])

# Reaction 1536
reaction('C3H5-A + C4H71-3 => C3H6 + C4H6', [6.310000e+12, 0.0, 0.0])

# Reaction 1537
reaction('C3H6 + C4H6 => C3H5-A + C4H71-3', [1.000000e+10, 0.0, 50000.0])

# Reaction 1538
reaction('C4H71-3 + O2 => C4H6 + HO2', [1.000000e+09, 0.0, 0.0])

# Reaction 1539
reaction('C4H6 + HO2 => C4H71-3 + O2', [1.000000e+11, 0.0, 17000.0])

# Reaction 1540
reaction('H + C4H71-3 => C4H6 + H2', [3.160000e+13, 0.0, 0.0])

# Reaction 1541
reaction('C4H6 + H2 => H + C4H71-3', [1.066000e+13, 0.0, 56810.0])

# Reaction 1542
reaction('C2H5 + C4H71-3 => C4H6 + C2H6', [3.980000e+12, 0.0, 0.0])

# Reaction 1543
reaction('C4H6 + C2H6 => C2H5 + C4H71-3', [3.211000e+12, 0.0, 49840.0])

# Reaction 1544
reaction('C2H3 + C4H71-3 => C2H4 + C4H6', [3.980000e+12, 0.0, 0.0])

# Reaction 1545
reaction('C2H4 + C4H6 => C2H3 + C4H71-3', [1.157000e+13, 0.0, 57710.0])

# Reaction 1546
reaction('C4H71-3 + C2H5O2 => C4H7O + C2H5O', [3.800000e+12, 0.0, -1200.0])

# Reaction 1547
reaction('C4H7O + C2H5O => C4H71-3 + C2H5O2', [2.000000e+10, 0.0, 0.0])

# Reaction 1548
reaction('IC3H7O2 + C4H71-3 => IC3H7O + C4H7O', [3.800000e+12, 0.0, -1200.0])

# Reaction 1549
reaction('IC3H7O + C4H7O => IC3H7O2 + C4H71-3', [2.000000e+10, 0.0, 0.0])

# Reaction 1550
reaction('NC3H7O2 + C4H71-3 => NC3H7O + C4H7O', [3.800000e+12, 0.0, -1200.0])

# Reaction 1551
reaction('NC3H7O + C4H7O => NC3H7O2 + C4H71-3', [2.000000e+10, 0.0, 0.0])

# Reaction 1552
reaction('C4H7O => CH3CHO + C2H3', [7.940000e+14, 0.0, 19000.0])

# Reaction 1553
reaction('CH3CHO + C2H3 => C4H7O', [1.000000e+10, 0.0, 20000.0])

# Reaction 1554
reaction('C4H7O => C2H3CHO + CH3', [7.940000e+14, 0.0, 19000.0])

# Reaction 1555
reaction('C2H3CHO + CH3 => C4H7O', [1.000000e+10, 0.0, 20000.0])

# Reaction 1556
reaction('C4H6 => C2H3 + C2H3', [4.027000e+19, -1.0, 98150.0])

# Reaction 1557
reaction('C2H3 + C2H3 => C4H6', [1.260000e+13, 0.0, 0.0])

# Reaction 1558
reaction('C4H6 + OH => C2H5 + CH2CO', [1.000000e+12, 0.0, 0.0])

# Reaction 1559
reaction('C2H5 + CH2CO => C4H6 + OH', [3.730000e+12, 0.0, 30020.0])

# Reaction 1560
reaction('C4H6 + OH => CH2O + C3H5-A', [1.000000e+12, 0.0, 0.0])

# Reaction 1561
reaction('CH2O + C3H5-A => C4H6 + OH', [3.501000e+06, 0.0, 71060.0])

# Reaction 1562
reaction('C4H6 + OH => C2H3 + CH3CHO', [1.000000e+12, 0.0, 0.0])

# Reaction 1563
reaction('C2H3 + CH3CHO => C4H6 + OH', [5.437000e+11, 0.0, 18550.0])

# Reaction 1564
reaction('C4H6 + O => C2H4 + CH2CO', [1.000000e+12, 0.0, 0.0])

# Reaction 1565
reaction('C2H4 + CH2CO => C4H6 + O', [6.377000e+11, 0.0, 94340.0])

# Reaction 1566
reaction('C4H6 + O => CH2O + C3H4-A', [1.000000e+12, 0.0, 0.0])

# Reaction 1567
reaction('CH2O + C3H4-A => C4H6 + O', [1.075000e+12, 0.0, 79050.0])

# Reaction 1568
reaction('C2H3 + C2H4 => C4H6 + H', [5.000000e+11, 0.0, 7300.0])

# Reaction 1569
reaction('C4H6 + H => C2H3 + C2H4', [1.000000e+13, 0.0, 4700.0])

# Reaction 1570
reaction('C4H8O1-2 + OH => CH2O + C3H5-A + H2O', [5.000000e+12, 0.0, 0.0])

# Reaction 1571
reaction('CH2O + C3H5-A + H2O => C4H8O1-2 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 1572
reaction('C4H8O1-2 + H => CH2O + C3H5-A + H2', [5.000000e+12, 0.0, 0.0])

# Reaction 1573
reaction('CH2O + C3H5-A + H2 => C4H8O1-2 + H', [0.000000e+00, 0.0, 0.0])

# Reaction 1574
reaction('C4H8O1-2 + O => CH2O + C3H5-A + OH', [5.000000e+12, 0.0, 0.0])

# Reaction 1575
reaction('CH2O + C3H5-A + OH => C4H8O1-2 + O', [0.000000e+00, 0.0, 0.0])

# Reaction 1576
reaction('C4H8O1-2 + HO2 => CH2O + C3H5-A + H2O2', [1.000000e+13, 0.0, 15000.0])

# Reaction 1577
reaction('CH2O + C3H5-A + H2O2 => C4H8O1-2 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 1578
reaction('C4H8O1-2 + CH3O2 => CH2O + C3H5-A + CH3O2H', [1.000000e+13, 0.0, 19000.0])

# Reaction 1579
reaction('CH2O + C3H5-A + CH3O2H => C4H8O1-2 + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1580
reaction('C4H8O1-2 + CH3 => CH2O + C3H5-A + CH4', [2.000000e+11, 0.0, 10000.0])

# Reaction 1581
reaction('CH2O + C3H5-A + CH4 => C4H8O1-2 + CH3', [0.000000e+00, 0.0, 0.0])

# Reaction 1582
reaction('C4H8O1-3 + OH => CH2O + C3H5-A + H2O', [5.000000e+12, 0.0, 0.0])

# Reaction 1583
reaction('CH2O + C3H5-A + H2O => C4H8O1-3 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 1584
reaction('C4H8O1-3 + H => CH2O + C3H5-A + H2', [5.000000e+12, 0.0, 0.0])

# Reaction 1585
reaction('CH2O + C3H5-A + H2 => C4H8O1-3 + H', [0.000000e+00, 0.0, 0.0])

# Reaction 1586
reaction('C4H8O1-3 + O => CH2O + C3H5-A + OH', [5.000000e+12, 0.0, 0.0])

# Reaction 1587
reaction('CH2O + C3H5-A + OH => C4H8O1-3 + O', [0.000000e+00, 0.0, 0.0])

# Reaction 1588
reaction('C4H8O1-3 + HO2 => CH2O + C3H5-A + H2O2', [1.000000e+13, 0.0, 15000.0])

# Reaction 1589
reaction('CH2O + C3H5-A + H2O2 => C4H8O1-3 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 1590
reaction('C4H8O1-3 + CH3O2 => CH2O + C3H5-A + CH3O2H', [1.000000e+13, 0.0, 19000.0])

# Reaction 1591
reaction('CH2O + C3H5-A + CH3O2H => C4H8O1-3 + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1592
reaction('C4H8O1-3 + CH3 => CH2O + C3H5-A + CH4', [2.000000e+11, 0.0, 10000.0])

# Reaction 1593
reaction('CH2O + C3H5-A + CH4 => C4H8O1-3 + CH3', [0.000000e+00, 0.0, 0.0])

# Reaction 1594
reaction('C4H8O1-4 + OH => CH2O + C3H5-A + H2O', [5.000000e+12, 0.0, 0.0])

# Reaction 1595
reaction('CH2O + C3H5-A + H2O => C4H8O1-4 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 1596
reaction('C4H8O1-4 + H => CH2O + C3H5-A + H2', [5.000000e+12, 0.0, 0.0])

# Reaction 1597
reaction('CH2O + C3H5-A + H2 => C4H8O1-4 + H', [0.000000e+00, 0.0, 0.0])

# Reaction 1598
reaction('C4H8O1-4 + O => CH2O + C3H5-A + OH', [5.000000e+12, 0.0, 0.0])

# Reaction 1599
reaction('CH2O + C3H5-A + OH => C4H8O1-4 + O', [0.000000e+00, 0.0, 0.0])

# Reaction 1600
reaction('C4H8O1-4 + HO2 => CH2O + C3H5-A + H2O2', [1.000000e+13, 0.0, 15000.0])

# Reaction 1601
reaction('CH2O + C3H5-A + H2O2 => C4H8O1-4 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 1602
reaction('C4H8O1-4 + CH3O2 => CH2O + C3H5-A + CH3O2H', [1.000000e+13, 0.0, 19000.0])

# Reaction 1603
reaction('CH2O + C3H5-A + CH3O2H => C4H8O1-4 + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1604
reaction('C4H8O1-4 + CH3 => CH2O + C3H5-A + CH4', [2.000000e+11, 0.0, 10000.0])

# Reaction 1605
reaction('CH2O + C3H5-A + CH4 => C4H8O1-4 + CH3', [0.000000e+00, 0.0, 0.0])

# Reaction 1606
reaction('C4H8O2-3 + OH => CH2O + C3H5-A + H2O', [5.000000e+12, 0.0, 0.0])

# Reaction 1607
reaction('CH2O + C3H5-A + H2O => C4H8O2-3 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 1608
reaction('C4H8O2-3 + H => CH2O + C3H5-A + H2', [5.000000e+12, 0.0, 0.0])

# Reaction 1609
reaction('CH2O + C3H5-A + H2 => C4H8O2-3 + H', [0.000000e+00, 0.0, 0.0])

# Reaction 1610
reaction('C4H8O2-3 + O => CH2O + C3H5-A + OH', [5.000000e+12, 0.0, 0.0])

# Reaction 1611
reaction('CH2O + C3H5-A + OH => C4H8O2-3 + O', [0.000000e+00, 0.0, 0.0])

# Reaction 1612
reaction('C4H8O2-3 + HO2 => CH2O + C3H5-A + H2O2', [1.000000e+13, 0.0, 15000.0])

# Reaction 1613
reaction('CH2O + C3H5-A + H2O2 => C4H8O2-3 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 1614
reaction('C4H8O2-3 + CH3O2 => CH2O + C3H5-A + CH3O2H', [1.000000e+13, 0.0, 19000.0])

# Reaction 1615
reaction('CH2O + C3H5-A + CH3O2H => C4H8O2-3 + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1616
reaction('C4H8O2-3 + CH3 => CH2O + C3H5-A + CH4', [2.000000e+11, 0.0, 10000.0])

# Reaction 1617
reaction('CH2O + C3H5-A + CH4 => C4H8O2-3 + CH3', [0.000000e+00, 0.0, 0.0])

# Reaction 1618
reaction('PC4H9O2 => PC4H9 + O2', [2.849000e+20, -1.642, 35930.0])

# Reaction 1619
reaction('PC4H9 + O2 => PC4H9O2', [4.520000e+12, 0.0, 0.0])

# Reaction 1620
reaction('SC4H9O2 => SC4H9 + O2', [4.329000e+22, -2.216, 38160.0])

# Reaction 1621
reaction('SC4H9 + O2 => SC4H9O2', [7.540000e+12, 0.0, 0.0])

# Reaction 1622
reaction('SC4H9O2 + CH2O => SC4H9O2H + HCO', [5.600000e+12, 0.0, 13600.0])

# Reaction 1623
reaction('SC4H9O2H + HCO => SC4H9O2 + CH2O', [8.000000e+11, 0.0, 10000.0])

# Reaction 1624
reaction('SC4H9O2 + CH3CHO => SC4H9O2H + CH3CO', [2.800000e+12, 0.0, 13600.0])

# Reaction 1625
reaction('SC4H9O2H + CH3CO => SC4H9O2 + CH3CHO', [1.000000e+12, 0.0, 10000.0])

# Reaction 1626
reaction('SC4H9O2 + HO2 => SC4H9O2H + O2', [1.750000e+10, 0.0, -3275.0])

# Reaction 1627
reaction('SC4H9O2H + O2 => SC4H9O2 + HO2', [3.861000e+13, -0.796, 33620.0])

# Reaction 1628
reaction('IC3H7O2 + PC4H9 => IC3H7O + PC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1629
reaction('IC3H7O + PC4H9O => IC3H7O2 + PC4H9', [2.281000e+13, 0.075, 27480.0])

# Reaction 1630
reaction('IC3H7O2 + SC4H9 => IC3H7O + SC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1631
reaction('IC3H7O + SC4H9O => IC3H7O2 + SC4H9', [3.942000e+15, -0.417, 29190.0])

# Reaction 1632
reaction('NC3H7O2 + PC4H9 => NC3H7O + PC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1633
reaction('NC3H7O + PC4H9O => NC3H7O2 + PC4H9', [1.056000e+15, -0.611, 27680.0])

# Reaction 1634
reaction('NC3H7O2 + SC4H9 => NC3H7O + SC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1635
reaction('NC3H7O + SC4H9O => NC3H7O2 + SC4H9', [1.825000e+17, -1.103, 29390.0])

# Reaction 1636
reaction('SC4H9O2 + SC4H9O2 => O2 + SC4H9O + SC4H9O', [1.400000e+16, -1.61, 1860.0])

# Reaction 1637
reaction('O2 + SC4H9O + SC4H9O => SC4H9O2 + SC4H9O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1638
reaction('SC4H9O2 + NC3H7O2 => SC4H9O + NC3H7O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1639
reaction('SC4H9O + NC3H7O + O2 => SC4H9O2 + NC3H7O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1640
reaction('SC4H9O2 + IC3H7O2 => SC4H9O + IC3H7O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1641
reaction('SC4H9O + IC3H7O + O2 => SC4H9O2 + IC3H7O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1642
reaction('SC4H9O2 + C2H5O2 => SC4H9O + C2H5O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1643
reaction('SC4H9O + C2H5O + O2 => SC4H9O2 + C2H5O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1644
reaction('SC4H9O2 + CH3O2 => SC4H9O + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1645
reaction('SC4H9O + CH3O + O2 => SC4H9O2 + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1646
reaction('SC4H9O2 + CH3CO3 => SC4H9O + CH3CO2 + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1647
reaction('SC4H9O + CH3CO2 + O2 => SC4H9O2 + CH3CO3', [0.000000e+00, 0.0, 0.0])

# Reaction 1648
reaction('PC4H9O2 + HO2 => PC4H9O + OH + O2', [1.400000e-14, -1.61, 1860.0])

# Reaction 1649
reaction('PC4H9O + OH + O2 => PC4H9O2 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 1650
reaction('SC4H9O2 + HO2 => SC4H9O + OH + O2', [1.400000e-14, -1.61, 1860.0])

# Reaction 1651
reaction('SC4H9O + OH + O2 => SC4H9O2 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 1652
reaction('H2 + PC4H9O2 => H + PC4H9O2H', [3.010000e+13, 0.0, 26030.0])

# Reaction 1653
reaction('H + PC4H9O2H => H2 + PC4H9O2', [4.800000e+13, 0.0, 7950.0])

# Reaction 1654
reaction('H2 + SC4H9O2 => H + SC4H9O2H', [3.010000e+13, 0.0, 26030.0])

# Reaction 1655
reaction('H + SC4H9O2H => H2 + SC4H9O2', [4.800000e+13, 0.0, 7950.0])

# Reaction 1656
reaction('C2H6 + PC4H9O2 => C2H5 + PC4H9O2H', [1.700000e+13, 0.0, 20460.0])

# Reaction 1657
reaction('C2H5 + PC4H9O2H => C2H6 + PC4H9O2', [5.000000e+11, 0.0, 6500.0])

# Reaction 1658
reaction('C2H6 + SC4H9O2 => C2H5 + SC4H9O2H', [1.700000e+13, 0.0, 20460.0])

# Reaction 1659
reaction('C2H5 + SC4H9O2H => C2H6 + SC4H9O2', [5.000000e+11, 0.0, 6500.0])

# Reaction 1660
reaction('PC4H9O2 + C2H5CHO => PC4H9O2H + C2H5CO', [2.000000e+11, 0.0, 9500.0])

# Reaction 1661
reaction('PC4H9O2H + C2H5CO => PC4H9O2 + C2H5CHO', [5.000000e+09, 0.0, 10000.0])

# Reaction 1662
reaction('SC4H9O2 + C2H5CHO => SC4H9O2H + C2H5CO', [2.000000e+11, 0.0, 9500.0])

# Reaction 1663
reaction('SC4H9O2H + C2H5CO => SC4H9O2 + C2H5CHO', [5.000000e+09, 0.0, 10000.0])

# Reaction 1664
reaction('SC4H9O2 + CH3 => SC4H9O + CH3O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1665
reaction('SC4H9O + CH3O => SC4H9O2 + CH3', [1.176000e+13, 0.067, 26110.0])

# Reaction 1666
reaction('SC4H9O2 + C2H5 => SC4H9O + C2H5O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1667
reaction('SC4H9O + C2H5O => SC4H9O2 + C2H5', [1.672000e+15, -0.534, 28520.0])

# Reaction 1668
reaction('SC4H9O2 + IC3H7 => SC4H9O + IC3H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1669
reaction('SC4H9O + IC3H7O => SC4H9O2 + IC3H7', [2.852000e+15, -0.367, 29190.0])

# Reaction 1670
reaction('SC4H9O2 + NC3H7 => SC4H9O + NC3H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1671
reaction('SC4H9O + NC3H7O => SC4H9O2 + NC3H7', [1.688000e+15, -0.503, 27190.0])

# Reaction 1672
reaction('SC4H9O2 + PC4H9 => SC4H9O + PC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1673
reaction('SC4H9O + PC4H9O => SC4H9O2 + PC4H9', [2.342000e+15, -0.553, 26870.0])

# Reaction 1674
reaction('SC4H9O2 + SC4H9 => SC4H9O + SC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1675
reaction('SC4H9O + SC4H9O => SC4H9O2 + SC4H9', [4.048000e+17, -1.045, 28580.0])

# Reaction 1676
reaction('SC4H9O2 + C3H5-A => SC4H9O + C3H5O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1677
reaction('SC4H9O + C3H5O => SC4H9O2 + C3H5-A', [4.356000e+11, 0.249, 13530.0])

# Reaction 1678
reaction('PC4H9O2 + CH2O => PC4H9O2H + HCO', [5.600000e+12, 0.0, 13600.0])

# Reaction 1679
reaction('PC4H9O2H + HCO => PC4H9O2 + CH2O', [8.000000e+11, 0.0, 10000.0])

# Reaction 1680
reaction('PC4H9O2 + CH3CHO => PC4H9O2H + CH3CO', [2.800000e+12, 0.0, 13600.0])

# Reaction 1681
reaction('PC4H9O2H + CH3CO => PC4H9O2 + CH3CHO', [1.000000e+12, 0.0, 10000.0])

# Reaction 1682
reaction('PC4H9O2 + HO2 => PC4H9O2H + O2', [1.750000e+10, 0.0, -3275.0])

# Reaction 1683
reaction('PC4H9O2H + O2 => PC4H9O2 + HO2', [4.357000e+13, -0.813, 33630.0])

# Reaction 1684
reaction('C3H6 + PC4H9O2 => C3H5-A + PC4H9O2H', [3.240000e+11, 0.0, 14900.0])

# Reaction 1685
reaction('C3H5-A + PC4H9O2H => C3H6 + PC4H9O2', [2.000000e+10, 0.0, 15000.0])

# Reaction 1686
reaction('C3H6 + SC4H9O2 => C3H5-A + SC4H9O2H', [3.240000e+11, 0.0, 14900.0])

# Reaction 1687
reaction('C3H5-A + SC4H9O2H => C3H6 + SC4H9O2', [2.000000e+10, 0.0, 15000.0])

# Reaction 1688
reaction('C2H4 + PC4H9O2 => C2H3 + PC4H9O2H', [1.130000e+13, 0.0, 30430.0])

# Reaction 1689
reaction('C2H3 + PC4H9O2H => C2H4 + PC4H9O2', [3.000000e+12, 0.0, 11500.0])

# Reaction 1690
reaction('C2H4 + SC4H9O2 => C2H3 + SC4H9O2H', [1.130000e+13, 0.0, 30430.0])

# Reaction 1691
reaction('C2H3 + SC4H9O2H => C2H4 + SC4H9O2', [3.000000e+12, 0.0, 11500.0])

# Reaction 1692
reaction('CH3OH + PC4H9O2 => CH2OH + PC4H9O2H', [6.300000e+12, 0.0, 19360.0])

# Reaction 1693
reaction('CH2OH + PC4H9O2H => CH3OH + PC4H9O2', [1.000000e+09, 0.0, 10000.0])

# Reaction 1694
reaction('CH3OH + SC4H9O2 => CH2OH + SC4H9O2H', [6.300000e+12, 0.0, 19360.0])

# Reaction 1695
reaction('CH2OH + SC4H9O2H => CH3OH + SC4H9O2', [1.000000e+09, 0.0, 10000.0])

# Reaction 1696
reaction('C2H3CHO + PC4H9O2 => C2H3CO + PC4H9O2H', [2.800000e+12, 0.0, 13600.0])

# Reaction 1697
reaction('C2H3CO + PC4H9O2H => C2H3CHO + PC4H9O2', [1.000000e+12, 0.0, 10000.0])

# Reaction 1698
reaction('C2H3CHO + SC4H9O2 => C2H3CO + SC4H9O2H', [2.800000e+12, 0.0, 13600.0])

# Reaction 1699
reaction('C2H3CO + SC4H9O2H => C2H3CHO + SC4H9O2', [1.000000e+12, 0.0, 10000.0])

# Reaction 1700
reaction('CH4 + PC4H9O2 => CH3 + PC4H9O2H', [1.120000e+13, 0.0, 24640.0])

# Reaction 1701
reaction('CH3 + PC4H9O2H => CH4 + PC4H9O2', [7.430000e+11, 0.0, 5500.0])

# Reaction 1702
reaction('CH4 + SC4H9O2 => CH3 + SC4H9O2H', [1.120000e+13, 0.0, 24640.0])

# Reaction 1703
reaction('CH3 + SC4H9O2H => CH4 + SC4H9O2', [7.430000e+11, 0.0, 5500.0])

# Reaction 1704
reaction('C4H71-3 + PC4H9O2 => C4H7O + PC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1705
reaction('C4H7O + PC4H9O => C4H71-3 + PC4H9O2', [1.186000e+14, -0.763, 16100.0])

# Reaction 1706
reaction('C4H71-3 + SC4H9O2 => C4H7O + SC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1707
reaction('C4H7O + SC4H9O => C4H71-3 + SC4H9O2', [2.251000e+14, -0.68, 15580.0])

# Reaction 1708
reaction('H2O2 + PC4H9O2 => HO2 + PC4H9O2H', [2.400000e+12, 0.0, 10000.0])

# Reaction 1709
reaction('HO2 + PC4H9O2H => H2O2 + PC4H9O2', [2.400000e+12, 0.0, 10000.0])

# Reaction 1710
reaction('H2O2 + SC4H9O2 => HO2 + SC4H9O2H', [2.400000e+12, 0.0, 10000.0])

# Reaction 1711
reaction('HO2 + SC4H9O2H => H2O2 + SC4H9O2', [2.400000e+12, 0.0, 10000.0])

# Reaction 1712
reaction('PC4H9O2 + PC4H9O2 => O2 + PC4H9O + PC4H9O', [1.400000e+16, -1.61, 1860.0])

# Reaction 1713
reaction('O2 + PC4H9O + PC4H9O => PC4H9O2 + PC4H9O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1714
reaction('PC4H9O2 + SC4H9O2 => PC4H9O + SC4H9O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1715
reaction('PC4H9O + SC4H9O + O2 => PC4H9O2 + SC4H9O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1716
reaction('PC4H9O2 + NC3H7O2 => PC4H9O + NC3H7O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1717
reaction('PC4H9O + NC3H7O + O2 => PC4H9O2 + NC3H7O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1718
reaction('PC4H9O2 + IC3H7O2 => PC4H9O + IC3H7O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1719
reaction('PC4H9O + IC3H7O + O2 => PC4H9O2 + IC3H7O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1720
reaction('PC4H9O2 + C2H5O2 => PC4H9O + C2H5O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1721
reaction('PC4H9O + C2H5O + O2 => PC4H9O2 + C2H5O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1722
reaction('PC4H9O2 + CH3O2 => PC4H9O + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1723
reaction('PC4H9O + CH3O + O2 => PC4H9O2 + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1724
reaction('PC4H9O2 + CH3CO3 => PC4H9O + CH3CO2 + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 1725
reaction('PC4H9O + CH3CO2 + O2 => PC4H9O2 + CH3CO3', [0.000000e+00, 0.0, 0.0])

# Reaction 1726
reaction('PC4H9O2 + CH3 => PC4H9O + CH3O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1727
reaction('PC4H9O + CH3O => PC4H9O2 + CH3', [6.199000e+12, -0.015, 26630.0])

# Reaction 1728
reaction('PC4H9O2 + C2H5 => PC4H9O + C2H5O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1729
reaction('PC4H9O + C2H5O => PC4H9O2 + C2H5', [8.813000e+14, -0.617, 29040.0])

# Reaction 1730
reaction('PC4H9O2 + IC3H7 => PC4H9O + IC3H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1731
reaction('PC4H9O + IC3H7O => PC4H9O2 + IC3H7', [1.503000e+15, -0.45, 29710.0])

# Reaction 1732
reaction('PC4H9O2 + NC3H7 => PC4H9O + NC3H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1733
reaction('PC4H9O + NC3H7O => PC4H9O2 + NC3H7', [8.898000e+14, -0.586, 27710.0])

# Reaction 1734
reaction('PC4H9O2 + PC4H9 => PC4H9O + PC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1735
reaction('PC4H9O + PC4H9O => PC4H9O2 + PC4H9', [1.235000e+15, -0.636, 27390.0])

# Reaction 1736
reaction('PC4H9O2 + SC4H9 => PC4H9O + SC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1737
reaction('PC4H9O + SC4H9O => PC4H9O2 + SC4H9', [2.134000e+17, -1.128, 29100.0])

# Reaction 1738
reaction('PC4H9O2 + C3H5-A => PC4H9O + C3H5O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1739
reaction('PC4H9O + C3H5O => PC4H9O2 + C3H5-A', [2.296000e+11, 0.167, 14050.0])

# Reaction 1740
reaction('PC4H9 + HO2 => PC4H9O + OH', [7.000000e+12, 0.0, -1000.0])

# Reaction 1741
reaction('PC4H9O + OH => PC4H9 + HO2', [8.630000e+15, -0.743, 25000.0])

# Reaction 1742
reaction('SC4H9 + HO2 => SC4H9O + OH', [7.000000e+12, 0.0, -1000.0])

# Reaction 1743
reaction('SC4H9O + OH => SC4H9 + HO2', [1.491000e+18, -1.234, 26710.0])

# Reaction 1744
reaction('CH3O2 + PC4H9 => CH3O + PC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1745
reaction('CH3O + PC4H9O => CH3O2 + PC4H9', [5.398000e+14, -0.444, 29240.0])

# Reaction 1746
reaction('CH3O2 + SC4H9 => CH3O + SC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 1747
reaction('CH3O + SC4H9O => CH3O2 + SC4H9', [9.329000e+16, -0.936, 30950.0])

# Reaction 1748
reaction('PC4H9O2H => PC4H9O + OH', [1.500000e+16, 0.0, 42500.0])

# Reaction 1749
reaction('PC4H9O + OH => PC4H9O2H', [1.178000e+08, 1.712, -4342.0])

# Reaction 1750
reaction('SC4H9O2H => SC4H9O + OH', [9.450000e+15, 0.0, 41600.0])

# Reaction 1751
reaction('SC4H9O + OH => SC4H9O2H', [1.589000e+08, 1.778, -5752.0])

# Reaction 1752
reaction('PC4H9O => NC3H7 + CH2O', [1.558000e+21, -2.444, 15230.0])

# Reaction 1753
reaction('NC3H7 + CH2O => PC4H9O', [5.000000e+10, 0.0, 3457.0])

# Reaction 1754
reaction('SC4H9O => CH3 + C2H5CHO', [4.381000e+16, -0.893, 15200.0])

# Reaction 1755
reaction('CH3 + C2H5CHO => SC4H9O', [5.000000e+10, 0.0, 9043.0])

# Reaction 1756
reaction('SC4H9O => C2H5 + CH3CHO', [5.489000e+22, -2.757, 12650.0])

# Reaction 1757
reaction('C2H5 + CH3CHO => SC4H9O', [3.330000e+10, 0.0, 6397.0])

# Reaction 1758
reaction('PC4H9O2 => C4H8OOH1-2', [2.000000e+11, 0.0, 26850.0])

# Reaction 1759
reaction('C4H8OOH1-2 => PC4H9O2', [5.597000e+08, 0.339, 11970.0])

# Reaction 1760
reaction('PC4H9O2 => C4H8OOH1-3', [2.500000e+10, 0.0, 20850.0])

# Reaction 1761
reaction('C4H8OOH1-3 => PC4H9O2', [3.231000e+09, -0.136, 7871.0])

# Reaction 1762
reaction('PC4H9O2 => C4H8OOH1-4', [4.688000e+09, 0.0, 22350.0])

# Reaction 1763
reaction('C4H8OOH1-4 => PC4H9O2', [1.269000e+10, -0.523, 6900.0])

# Reaction 1764
reaction('SC4H9O2 => C4H8OOH2-1', [3.000000e+11, 0.0, 29400.0])

# Reaction 1765
reaction('C4H8OOH2-1 => SC4H9O2', [7.442000e+11, -0.511, 13940.0])

# Reaction 1766
reaction('SC4H9O2 => C4H8OOH2-3', [2.000000e+11, 0.0, 26850.0])

# Reaction 1767
reaction('C4H8OOH2-3 => SC4H9O2', [3.994000e+10, -0.196, 13930.0])

# Reaction 1768
reaction('SC4H9O2 => C4H8OOH2-4', [3.750000e+10, 0.0, 24400.0])

# Reaction 1769
reaction('C4H8OOH2-4 => SC4H9O2', [9.302000e+10, -0.511, 8944.0])

# Reaction 1770
reaction('PC4H9O2 => C4H8-1 + HO2', [5.044000e+38, -8.11, 40490.0])

# Reaction 1771
reaction('C4H8-1 + HO2 => PC4H9O2', [1.599000e+30, -6.283, 20350.0])

# Reaction 1772
reaction('SC4H9O2 => C4H8-1 + HO2', [5.075000e+42, -9.41, 41490.0])

# Reaction 1773
reaction('C4H8-1 + HO2 => SC4H9O2', [2.595000e+33, -7.347, 16610.0])

# Reaction 1774
reaction('SC4H9O2 => C4H8-2 + HO2', [5.044000e+38, -8.11, 40490.0])

# Reaction 1775
reaction('C4H8-2 + HO2 => SC4H9O2', [1.618000e+29, -5.793, 18300.0])

# Reaction 1776
reaction('C4H8OOH1-2 => C4H8-1 + HO2', [8.828000e+16, -1.488, 16260.0])

# Reaction 1777
reaction('C4H8-1 + HO2 => C4H8OOH1-2', [1.000000e+11, 0.0, 11000.0])

# Reaction 1778
reaction('C4H8OOH2-1 => C4H8-1 + HO2', [4.851000e+20, -2.574, 21180.0])

# Reaction 1779
reaction('C4H8-1 + HO2 => C4H8OOH2-1', [1.000000e+11, 0.0, 11750.0])

# Reaction 1780
reaction('C4H8OOH2-3 => C4H8-2 + HO2', [6.224000e+19, -2.513, 21020.0])

# Reaction 1781
reaction('C4H8-2 + HO2 => C4H8OOH2-3', [1.000000e+11, 0.0, 11750.0])

# Reaction 1782
reaction('C4H8OOH1-2 => C4H8O1-2 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 1783
reaction('C4H8O1-2 + OH => C4H8OOH1-2', [0.000000e+00, 0.0, 0.0])

# Reaction 1784
reaction('C4H8OOH1-3 => C4H8O1-3 + OH', [7.500000e+10, 0.0, 15250.0])

# Reaction 1785
reaction('C4H8O1-3 + OH => C4H8OOH1-3', [0.000000e+00, 0.0, 0.0])

# Reaction 1786
reaction('C4H8OOH1-4 => C4H8O1-4 + OH', [9.375000e+09, 0.0, 6000.0])

# Reaction 1787
reaction('C4H8O1-4 + OH => C4H8OOH1-4', [0.000000e+00, 0.0, 0.0])

# Reaction 1788
reaction('C4H8OOH2-1 => C4H8O1-2 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 1789
reaction('C4H8O1-2 + OH => C4H8OOH2-1', [0.000000e+00, 0.0, 0.0])

# Reaction 1790
reaction('C4H8OOH2-3 => C4H8O2-3 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 1791
reaction('C4H8O2-3 + OH => C4H8OOH2-3', [0.000000e+00, 0.0, 0.0])

# Reaction 1792
reaction('C4H8OOH2-4 => C4H8O1-3 + OH', [7.500000e+10, 0.0, 15250.0])

# Reaction 1793
reaction('C4H8O1-3 + OH => C4H8OOH2-4', [0.000000e+00, 0.0, 0.0])

# Reaction 1794
reaction('C4H8OOH1-1 => NC3H7CHO + OH', [9.000000e+14, 0.0, 1500.0])

# Reaction 1795
reaction('NC3H7CHO + OH => C4H8OOH1-1', [1.733000e+08, 1.89, 33470.0])

# Reaction 1796
reaction('C4H8OOH2-2 => C2H5COCH3 + OH', [9.000000e+14, 0.0, 1500.0])

# Reaction 1797
reaction('C2H5COCH3 + OH => C4H8OOH2-2', [1.990000e+10, 1.34, 35510.0])

# Reaction 1798
reaction('C4H8OOH1-3 => OH + CH2O + C3H6', [6.635000e+13, -0.16, 29900.0])

# Reaction 1799
reaction('OH + CH2O + C3H6 => C4H8OOH1-3', [0.000000e+00, 0.0, 0.0])

# Reaction 1800
reaction('C4H8OOH2-4 => OH + CH3CHO + C2H4', [1.945000e+18, -1.63, 26790.0])

# Reaction 1801
reaction('OH + CH3CHO + C2H4 => C4H8OOH2-4', [0.000000e+00, 0.0, 0.0])

# Reaction 1802
reaction('C4H8OOH1-2O2 => C4H8OOH1-2 + O2', [2.586000e+24, -2.709, 39860.0])

# Reaction 1803
reaction('C4H8OOH1-2 + O2 => C4H8OOH1-2O2', [7.540000e+12, 0.0, 0.0])

# Reaction 1804
reaction('C4H8OOH1-3O2 => C4H8OOH1-3 + O2', [5.601000e+22, -2.234, 37960.0])

# Reaction 1805
reaction('C4H8OOH1-3 + O2 => C4H8OOH1-3O2', [7.540000e+12, 0.0, 0.0])

# Reaction 1806
reaction('C4H8OOH1-4O2 => C4H8OOH1-4 + O2', [2.569000e+20, -1.611, 35680.0])

# Reaction 1807
reaction('C4H8OOH1-4 + O2 => C4H8OOH1-4O2', [4.520000e+12, 0.0, 0.0])

# Reaction 1808
reaction('C4H8OOH2-1O2 => C4H8OOH2-1 + O2', [2.821000e+20, -1.622, 35700.0])

# Reaction 1809
reaction('C4H8OOH2-1 + O2 => C4H8OOH2-1O2', [4.520000e+12, 0.0, 0.0])

# Reaction 1810
reaction('C4H8OOH2-3O2 => C4H8OOH2-3 + O2', [4.516000e+22, -2.218, 37880.0])

# Reaction 1811
reaction('C4H8OOH2-3 + O2 => C4H8OOH2-3O2', [7.540000e+12, 0.0, 0.0])

# Reaction 1812
reaction('C4H8OOH2-4O2 => C4H8OOH2-4 + O2', [2.821000e+20, -1.622, 35700.0])

# Reaction 1813
reaction('C4H8OOH2-4 + O2 => C4H8OOH2-4O2', [4.520000e+12, 0.0, 0.0])

# Reaction 1814
reaction('C4H8OOH1-2O2 => NC4KET12 + OH', [2.000000e+11, 0.0, 26400.0])

# Reaction 1815
reaction('NC4KET12 + OH => C4H8OOH1-2O2', [3.199000e+04, 1.323, 48930.0])

# Reaction 1816
reaction('C4H8OOH1-3O2 => NC4KET13 + OH', [2.500000e+10, 0.0, 21400.0])

# Reaction 1817
reaction('NC4KET13 + OH => C4H8OOH1-3O2', [1.435000e+03, 1.486, 44740.0])

# Reaction 1818
reaction('C4H8OOH1-4O2 => NC4KET14 + OH', [3.125000e+09, 0.0, 19350.0])

# Reaction 1819
reaction('NC4KET14 + OH => C4H8OOH1-4O2', [1.726000e+02, 1.494, 42690.0])

# Reaction 1820
reaction('C4H8OOH2-1O2 => NC4KET21 + OH', [1.000000e+11, 0.0, 23850.0])

# Reaction 1821
reaction('NC4KET21 + OH => C4H8OOH2-1O2', [3.626000e+02, 1.835, 50050.0])

# Reaction 1822
reaction('C4H8OOH2-3O2 => NC4KET23 + OH', [1.000000e+11, 0.0, 23850.0])

# Reaction 1823
reaction('NC4KET23 + OH => C4H8OOH2-3O2', [1.739000e+03, 1.731, 49130.0])

# Reaction 1824
reaction('C4H8OOH2-4O2 => NC4KET24 + OH', [1.250000e+10, 0.0, 17850.0])

# Reaction 1825
reaction('NC4KET24 + OH => C4H8OOH2-4O2', [1.021000e+02, 1.843, 43920.0])

# Reaction 1826
reaction('NC4KET12 => C2H5CHO + HCO + OH', [1.050000e+16, 0.0, 41600.0])

# Reaction 1827
reaction('C2H5CHO + HCO + OH => NC4KET12', [0.000000e+00, 0.0, 0.0])

# Reaction 1828
reaction('NC4KET13 => CH3CHO + CH2CHO + OH', [1.050000e+16, 0.0, 41600.0])

# Reaction 1829
reaction('CH3CHO + CH2CHO + OH => NC4KET13', [0.000000e+00, 0.0, 0.0])

# Reaction 1830
reaction('NC4KET14 => CH2CH2CHO + CH2O + OH', [1.500000e+16, 0.0, 42000.0])

# Reaction 1831
reaction('CH2CH2CHO + CH2O + OH => NC4KET14', [0.000000e+00, 0.0, 0.0])

# Reaction 1832
reaction('NC4KET21 => CH2O + C2H5CO + OH', [1.500000e+16, 0.0, 42000.0])

# Reaction 1833
reaction('CH2O + C2H5CO + OH => NC4KET21', [0.000000e+00, 0.0, 0.0])

# Reaction 1834
reaction('NC4KET23 => CH3CHO + CH3CO + OH', [1.050000e+16, 0.0, 41600.0])

# Reaction 1835
reaction('CH3CHO + CH3CO + OH => NC4KET23', [0.000000e+00, 0.0, 0.0])

# Reaction 1836
reaction('NC4KET24 => CH2O + CH3COCH2 + OH', [1.500000e+16, 0.0, 42000.0])

# Reaction 1837
reaction('CH2O + CH3COCH2 + OH => NC4KET24', [0.000000e+00, 0.0, 0.0])

# Reaction 1838
reaction('C2H5COCH3 + OH => CH2CH2COCH3 + H2O', [7.550000e+09, 0.97, 1586.0])

# Reaction 1839
reaction('CH2CH2COCH3 + H2O => C2H5COCH3 + OH', [1.527000e+09, 0.96, 21020.0])

# Reaction 1840
reaction('C2H5COCH3 + OH => CH3CHCOCH3 + H2O', [8.450000e+11, 0.0, -228.0])

# Reaction 1841
reaction('CH3CHCOCH3 + H2O => C2H5COCH3 + OH', [1.586000e+13, -0.23, 28620.0])

# Reaction 1842
reaction('C2H5COCH3 + OH => C2H5COCH2 + H2O', [5.100000e+11, 0.0, 1192.0])

# Reaction 1843
reaction('C2H5COCH2 + H2O => C2H5COCH3 + OH', [6.628000e+13, -0.7, 27670.0])

# Reaction 1844
reaction('C2H5COCH3 + HO2 => CH2CH2COCH3 + H2O2', [2.380000e+04, 2.55, 16490.0])

# Reaction 1845
reaction('CH2CH2COCH3 + H2O2 => C2H5COCH3 + HO2', [2.857000e+04, 2.22, 4459.0])

# Reaction 1846
reaction('C2H5COCH3 + HO2 => CH3CHCOCH3 + H2O2', [2.000000e+11, 0.0, 8698.0])

# Reaction 1847
reaction('CH3CHCOCH3 + H2O2 => C2H5COCH3 + HO2', [2.229000e+13, -0.55, 6080.0])

# Reaction 1848
reaction('C2H5COCH3 + HO2 => C2H5COCH2 + H2O2', [2.380000e+04, 2.55, 14690.0])

# Reaction 1849
reaction('C2H5COCH2 + H2O2 => C2H5COCH3 + HO2', [1.836000e+07, 1.52, 9702.0])

# Reaction 1850
reaction('C2H5COCH3 + O => CH2CH2COCH3 + OH', [2.250000e+13, 0.0, 7700.0])

# Reaction 1851
reaction('CH2CH2COCH3 + OH => C2H5COCH3 + O', [4.616000e+11, -0.01, 9882.0])

# Reaction 1852
reaction('C2H5COCH3 + O => CH3CHCOCH3 + OH', [3.070000e+13, 0.0, 3400.0])

# Reaction 1853
reaction('CH3CHCOCH3 + OH => C2H5COCH3 + O', [5.847000e+13, -0.23, 15000.0])

# Reaction 1854
reaction('C2H5COCH3 + O => C2H5COCH2 + OH', [5.000000e+12, 0.0, 5962.0])

# Reaction 1855
reaction('C2H5COCH2 + OH => C2H5COCH3 + O', [6.592000e+13, -0.7, 15190.0])

# Reaction 1856
reaction('C2H5COCH3 + H => CH2CH2COCH3 + H2', [9.160000e+06, 2.0, 7700.0])

# Reaction 1857
reaction('CH2CH2COCH3 + H2 => C2H5COCH3 + H', [4.279000e+05, 1.99, 11980.0])

# Reaction 1858
reaction('C2H5COCH3 + H => CH3CHCOCH3 + H2', [4.460000e+06, 2.0, 3200.0])

# Reaction 1859
reaction('CH3CHCOCH3 + H2 => C2H5COCH3 + H', [1.934000e+07, 1.77, 16890.0])

# Reaction 1860
reaction('C2H5COCH3 + H => C2H5COCH2 + H2', [9.300000e+12, 0.0, 6357.0])

# Reaction 1861
reaction('C2H5COCH2 + H2 => C2H5COCH3 + H', [2.792000e+14, -0.7, 17680.0])

# Reaction 1862
reaction('C2H5COCH3 + O2 => CH2CH2COCH3 + HO2', [2.050000e+13, 0.0, 51310.0])

# Reaction 1863
reaction('CH2CH2COCH3 + HO2 => C2H5COCH3 + O2', [1.740000e+10, 0.32, -1895.0])

# Reaction 1864
reaction('C2H5COCH3 + O2 => CH3CHCOCH3 + HO2', [1.550000e+13, 0.0, 41970.0])

# Reaction 1865
reaction('CH3CHCOCH3 + HO2 => C2H5COCH3 + O2', [1.221000e+12, 0.1, -1822.0])

# Reaction 1866
reaction('C2H5COCH3 + O2 => C2H5COCH2 + HO2', [2.050000e+13, 0.0, 49150.0])

# Reaction 1867
reaction('C2H5COCH2 + HO2 => C2H5COCH3 + O2', [1.118000e+13, -0.37, 2988.0])

# Reaction 1868
reaction('C2H5COCH3 + CH3 => CH2CH2COCH3 + CH4', [3.190000e+01, 3.17, 7172.0])

# Reaction 1869
reaction('CH2CH2COCH3 + CH4 => C2H5COCH3 + CH3', [3.893000e+01, 3.16, 11930.0])

# Reaction 1870
reaction('C2H5COCH3 + CH3 => CH3CHCOCH3 + CH4', [1.740000e+00, 3.46, 3680.0])

# Reaction 1871
reaction('CH3CHCOCH3 + CH4 => C2H5COCH3 + CH3', [1.971000e+02, 3.23, 17850.0])

# Reaction 1872
reaction('C2H5COCH3 + CH3 => C2H5COCH2 + CH4', [1.620000e+11, 0.0, 9630.0])

# Reaction 1873
reaction('C2H5COCH2 + CH4 => C2H5COCH3 + CH3', [1.270000e+14, -0.7, 21430.0])

# Reaction 1874
reaction('C2H5COCH3 + CH3O => CH2CH2COCH3 + CH3OH', [2.170000e+11, 0.0, 6460.0])

# Reaction 1875
reaction('CH2CH2COCH3 + CH3OH => C2H5COCH3 + CH3O', [4.887000e+09, 0.02, 9007.0])

# Reaction 1876
reaction('C2H5COCH3 + CH3O => CH3CHCOCH3 + CH3OH', [1.450000e+11, 0.0, 2771.0])

# Reaction 1877
reaction('CH3CHCOCH3 + CH3OH => C2H5COCH3 + CH3O', [3.031000e+11, -0.2, 14730.0])

# Reaction 1878
reaction('C2H5COCH3 + CH3O => C2H5COCH2 + CH3OH', [2.170000e+11, 0.0, 4660.0])

# Reaction 1879
reaction('C2H5COCH2 + CH3OH => C2H5COCH3 + CH3O', [3.140000e+12, -0.68, 14250.0])

# Reaction 1880
reaction('C2H5COCH3 + CH3O2 => CH2CH2COCH3 + CH3O2H', [3.010000e+12, 0.0, 19380.0])

# Reaction 1881
reaction('CH2CH2COCH3 + CH3O2H => C2H5COCH3 + CH3O2', [7.525000e+12, -0.51, 4329.0])

# Reaction 1882
reaction('C2H5COCH3 + CH3O2 => CH3CHCOCH3 + CH3O2H', [2.000000e+12, 0.0, 15250.0])

# Reaction 1883
reaction('CH3CHCOCH3 + CH3O2H => C2H5COCH3 + CH3O2', [4.641000e+14, -0.73, 9612.0])

# Reaction 1884
reaction('C2H5COCH3 + CH3O2 => C2H5COCH2 + CH3O2H', [3.010000e+12, 0.0, 17580.0])

# Reaction 1885
reaction('C2H5COCH2 + CH3O2H => C2H5COCH3 + CH3O2', [4.836000e+15, -1.2, 9572.0])

# Reaction 1886
reaction('C2H5COCH3 + C2H3 => CH2CH2COCH3 + C2H4', [5.000000e+11, 0.0, 10400.0])

# Reaction 1887
reaction('CH2CH2COCH3 + C2H4 => C2H5COCH3 + C2H3', [3.436000e+08, 0.82, 14030.0])

# Reaction 1888
reaction('C2H5COCH3 + C2H3 => CH3CHCOCH3 + C2H4', [3.000000e+11, 0.0, 3400.0])

# Reaction 1889
reaction('CH3CHCOCH3 + C2H4 => C2H5COCH3 + C2H3', [1.914000e+10, 0.6, 16440.0])

# Reaction 1890
reaction('C2H5COCH3 + C2H3 => C2H5COCH2 + C2H4', [6.150000e+10, 0.0, 4278.0])

# Reaction 1891
reaction('C2H5COCH2 + C2H4 => C2H5COCH3 + C2H3', [2.716000e+10, 0.13, 14950.0])

# Reaction 1892
reaction('C2H5COCH3 + C2H5 => CH2CH2COCH3 + C2H6', [5.000000e+10, 0.0, 13400.0])

# Reaction 1893
reaction('CH2CH2COCH3 + C2H6 => C2H5COCH3 + C2H5', [9.549000e+12, -0.57, 13990.0])

# Reaction 1894
reaction('C2H5COCH3 + C2H5 => CH3CHCOCH3 + C2H6', [3.000000e+10, 0.0, 8600.0])

# Reaction 1895
reaction('CH3CHCOCH3 + C2H6 => C2H5COCH3 + C2H5', [5.319000e+14, -0.79, 18600.0])

# Reaction 1896
reaction('C2H5COCH3 + C2H5 => C2H5COCH2 + C2H6', [5.000000e+10, 0.0, 11600.0])

# Reaction 1897
reaction('C2H5COCH2 + C2H6 => C2H5COCH3 + C2H5', [6.136000e+15, -1.26, 19230.0])

# Reaction 1898
reaction('CH3CHOOCOCH3 => CH3CHCOCH3 + O2', [1.372000e+17, -1.69, 28460.0])

# Reaction 1899
reaction('CH3CHCOCH3 + O2 => CH3CHOOCOCH3', [1.000000e+11, 0.0, 0.0])

# Reaction 1900
reaction('CH3CHOOCOCH3 => CH2CHOOHCOCH3', [8.900000e+12, 0.0, 29700.0])

# Reaction 1901
reaction('CH2CHOOHCOCH3 => CH3CHOOCOCH3', [4.703000e+13, -0.52, 16960.0])

# Reaction 1902
reaction('CH2CHOOHCOCH3 => C2H3COCH3 + HO2', [2.026000e+19, -2.35, 14130.0])

# Reaction 1903
reaction('C2H3COCH3 + HO2 => CH2CHOOHCOCH3', [7.000000e+10, 0.0, 7800.0])

# Reaction 1904
reaction('CH2CH2CHO => C2H4 + HCO', [3.127000e+13, -0.52, 24590.0])

# Reaction 1905
reaction('C2H4 + HCO => CH2CH2CHO', [1.500000e+11, 0.0, 8300.0])

# Reaction 1906
reaction('CH2CH2COCH3 => C2H4 + CH3CO', [1.000000e+14, 0.0, 18000.0])

# Reaction 1907
reaction('C2H4 + CH3CO => CH2CH2COCH3', [1.000000e+11, 0.0, 0.0])

# Reaction 1908
reaction('C2H5COCH2 => CH2CO + C2H5', [1.000000e+14, 0.0, 35000.0])

# Reaction 1909
reaction('CH2CO + C2H5 => C2H5COCH2', [1.000000e+11, 0.0, 0.0])

# Reaction 1910
reaction('CH3CHCOCH3 => C2H3COCH3 + H', [3.417000e+16, -0.82, 41770.0])

# Reaction 1911
reaction('C2H3COCH3 + H => CH3CHCOCH3', [5.000000e+12, 0.0, 1200.0])

# Reaction 1912
reaction('CH3CHCOCH3 => CH3CHCO + CH3', [1.406000e+15, -0.44, 38340.0])

# Reaction 1913
reaction('CH3CHCO + CH3 => CH3CHCOCH3', [1.230000e+11, 0.0, 7800.0])

# Reaction 1914
reaction('NC3H7CHO + O2 => NC3H7CO + HO2', [1.200000e+05, 2.5, 37560.0])

# Reaction 1915
reaction('NC3H7CO + HO2 => NC3H7CHO + O2', [1.000000e+07, 0.5, 4000.0])

# Reaction 1916
reaction('NC3H7CHO + OH => NC3H7CO + H2O', [2.000000e+06, 1.8, -1300.0])

# Reaction 1917
reaction('NC3H7CO + H2O => NC3H7CHO + OH', [1.553000e+04, 2.32, 28050.0])

# Reaction 1918
reaction('NC3H7CHO + H => NC3H7CO + H2', [4.140000e+09, 1.12, 2320.0])

# Reaction 1919
reaction('NC3H7CO + H2 => NC3H7CHO + H', [3.030000e+06, 1.746, 16780.0])

# Reaction 1920
reaction('NC3H7CHO + O => NC3H7CO + OH', [5.940000e+12, 0.0, 1868.0])

# Reaction 1921
reaction('NC3H7CO + OH => NC3H7CHO + O', [2.258000e+09, 0.607, 14920.0])

# Reaction 1922
reaction('NC3H7CHO + HO2 => NC3H7CO + H2O2', [4.090000e+04, 2.5, 10200.0])

# Reaction 1923
reaction('NC3H7CO + H2O2 => NC3H7CHO + HO2', [1.733000e+04, 2.431, 8662.0])

# Reaction 1924
reaction('NC3H7CHO + CH3 => NC3H7CO + CH4', [2.890000e-03, 4.62, 3210.0])

# Reaction 1925
reaction('NC3H7CO + CH4 => NC3H7CHO + CH3', [1.930000e-03, 4.8, 19210.0])

# Reaction 1926
reaction('NC3H7CHO + CH3O => NC3H7CO + CH3OH', [1.000000e+12, 0.0, 3300.0])

# Reaction 1927
reaction('NC3H7CO + CH3OH => NC3H7CHO + CH3O', [1.572000e+10, 0.413, 17990.0])

# Reaction 1928
reaction('NC3H7CHO + CH3O2 => NC3H7CO + CH3O2H', [4.090000e+04, 2.5, 10200.0])

# Reaction 1929
reaction('NC3H7CO + CH3O2H => NC3H7CHO + CH3O2', [3.371000e+05, 1.986, 7067.0])

# Reaction 1930
reaction('NC3H7CHO + OH => C3H6CHO-1 + H2O', [5.280000e+09, 0.97, 1586.0])

# Reaction 1931
reaction('C3H6CHO-1 + H2O => NC3H7CHO + OH', [1.230000e+07, 1.494, 18820.0])

# Reaction 1932
reaction('NC3H7CHO + OH => C3H6CHO-2 + H2O', [4.680000e+07, 1.61, -35.0])

# Reaction 1933
reaction('C3H6CHO-2 + H2O => NC3H7CHO + OH', [9.102000e+03, 2.444, 19740.0])

# Reaction 1934
reaction('NC3H7CHO + OH => C3H6CHO-3 + H2O', [5.520000e+02, 3.12, -1176.0])

# Reaction 1935
reaction('C3H6CHO-3 + H2O => NC3H7CHO + OH', [1.428000e+02, 3.4, 25510.0])

# Reaction 1936
reaction('NC3H7CHO + HO2 => C3H6CHO-1 + H2O2', [2.379000e+04, 2.55, 16490.0])

# Reaction 1937
reaction('C3H6CHO-1 + H2O2 => NC3H7CHO + HO2', [3.025000e+03, 2.485, 2834.0])

# Reaction 1938
reaction('NC3H7CHO + HO2 => C3H6CHO-2 + H2O2', [9.640000e+03, 2.6, 13910.0])

# Reaction 1939
reaction('C3H6CHO-2 + H2O2 => NC3H7CHO + HO2', [1.023000e+02, 2.845, 2791.0])

# Reaction 1940
reaction('NC3H7CHO + HO2 => C3H6CHO-3 + H2O2', [3.440000e+12, 0.05, 17880.0])

# Reaction 1941
reaction('C3H6CHO-3 + H2O2 => NC3H7CHO + HO2', [4.855000e+13, -0.259, 13660.0])

# Reaction 1942
reaction('NC3H7CHO + CH3O2 => C3H6CHO-1 + CH3O2H', [2.379000e+04, 2.55, 16490.0])

# Reaction 1943
reaction('C3H6CHO-1 + CH3O2H => NC3H7CHO + CH3O2', [5.885000e+04, 2.04, 1239.0])

# Reaction 1944
reaction('NC3H7CHO + CH3O2 => C3H6CHO-2 + CH3O2H', [9.640000e+03, 2.6, 13910.0])

# Reaction 1945
reaction('C3H6CHO-2 + CH3O2H => NC3H7CHO + CH3O2', [1.990000e+03, 2.4, 1196.0])

# Reaction 1946
reaction('NC3H7CHO + CH3O2 => C3H6CHO-3 + CH3O2H', [3.440000e+12, 0.05, 17880.0])

# Reaction 1947
reaction('C3H6CHO-3 + CH3O2H => NC3H7CHO + CH3O2', [9.445000e+14, -0.704, 12070.0])

# Reaction 1948
reaction('NC3H7CO => NC3H7 + CO', [1.000000e+11, 0.0, 9600.0])

# Reaction 1949
reaction('NC3H7 + CO => NC3H7CO', [2.193000e+03, 1.763, -1100.0])

# Reaction 1950
reaction('C3H6CHO-1 => C2H4 + CH2CHO', [7.400000e+11, 0.0, 21970.0])

# Reaction 1951
reaction('C2H4 + CH2CHO => C3H6CHO-1', [2.110000e+11, 0.0, 7350.0])

# Reaction 1952
reaction('C3H6CHO-3 => C2H5CHCO + H', [8.431000e+15, -0.6, 40400.0])

# Reaction 1953
reaction('C2H5CHCO + H => C3H6CHO-3', [5.000000e+12, 0.0, 1200.0])

# Reaction 1954
reaction('C3H6CHO-3 => C2H3CHO + CH3', [3.174000e+14, -0.39, 29900.0])

# Reaction 1955
reaction('C2H3CHO + CH3 => C3H6CHO-3', [1.230000e+11, 0.0, 7800.0])

# Reaction 1956
reaction('C3H6CHO-2 => SC3H5CHO + H', [4.947000e+12, -0.15, 31300.0])

# Reaction 1957
reaction('SC3H5CHO + H => C3H6CHO-2', [5.000000e+12, 0.0, 2900.0])

# Reaction 1958
reaction('C3H6CHO-2 => C3H6 + HCO', [8.249000e+12, -0.18, 21900.0])

# Reaction 1959
reaction('C3H6 + HCO => C3H6CHO-2', [1.000000e+11, 0.0, 6000.0])

# Reaction 1960
reaction('C2H5CHCO + OH => NC3H7 + CO2', [3.730000e+12, 0.0, -1010.0])

# Reaction 1961
reaction('NC3H7 + CO2 => C2H5CHCO + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 1962
reaction('C2H5CHCO + H => NC3H7 + CO', [4.400000e+12, 0.0, 1459.0])

# Reaction 1963
reaction('NC3H7 + CO => C2H5CHCO + H', [0.000000e+00, 0.0, 0.0])

# Reaction 1964
reaction('C2H5CHCO + O => C3H6 + CO2', [3.200000e+12, 0.0, -437.0])

# Reaction 1965
reaction('C3H6 + CO2 => C2H5CHCO + O', [0.000000e+00, 0.0, 0.0])

# Reaction 1966
reaction('SC3H5CHO + OH => SC3H5CO + H2O', [2.690000e+10, 0.76, -340.0])

# Reaction 1967
reaction('SC3H5CO + H2O => SC3H5CHO + OH', [4.831000e+10, 0.77, 37090.0])

# Reaction 1968
reaction('SC3H5CO => C3H5-S + CO', [8.600000e+15, 0.0, 23000.0])

# Reaction 1969
reaction('C3H5-S + CO => SC3H5CO', [1.000000e+11, 0.0, 6000.0])

# Reaction 1970
reaction('SC3H5CHO + HO2 => SC3H5CO + H2O2', [1.000000e+12, 0.0, 11920.0])

# Reaction 1971
reaction('SC3H5CO + H2O2 => SC3H5CHO + HO2', [1.066000e+13, -0.32, 17890.0])

# Reaction 1972
reaction('SC3H5CHO + CH3 => SC3H5CO + CH4', [3.980000e+12, 0.0, 8700.0])

# Reaction 1973
reaction('SC3H5CO + CH4 => SC3H5CHO + CH3', [4.313000e+13, 0.01, 31460.0])

# Reaction 1974
reaction('SC3H5CHO + O => SC3H5CO + OH', [7.180000e+12, 0.0, 1389.0])

# Reaction 1975
reaction('SC3H5CO + OH => SC3H5CHO + O', [1.308000e+12, 0.01, 21570.0])

# Reaction 1976
reaction('SC3H5CHO + O2 => SC3H5CO + HO2', [4.000000e+13, 0.0, 37600.0])

# Reaction 1977
reaction('SC3H5CO + HO2 => SC3H5CHO + O2', [3.014000e+11, 0.34, 2394.0])

# Reaction 1978
reaction('SC3H5CHO + H => SC3H5CO + H2', [2.600000e+12, 0.0, 2600.0])

# Reaction 1979
reaction('SC3H5CO + H2 => SC3H5CHO + H', [1.079000e+12, 0.01, 24880.0])

# Reaction 1980
reaction('C2H3COCH3 + OH => CH3CHO + CH3CO', [1.000000e+11, 0.0, 0.0])

# Reaction 1981
reaction('CH3CHO + CH3CO => C2H3COCH3 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 1982
reaction('C2H3COCH3 + OH => CH2CO + C2H3 + H2O', [5.100000e+11, 0.0, 1192.0])

# Reaction 1983
reaction('CH2CO + C2H3 + H2O => C2H3COCH3 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 1984
reaction('C2H3COCH3 + HO2 => CH2CHO + CH3CO + OH', [6.030000e+09, 0.0, 7949.0])

# Reaction 1985
reaction('CH2CHO + CH3CO + OH => C2H3COCH3 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 1986
reaction('C2H3COCH3 + HO2 => CH2CO + C2H3 + H2O2', [8.500000e+12, 0.0, 20460.0])

# Reaction 1987
reaction('CH2CO + C2H3 + H2O2 => C2H3COCH3 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 1988
reaction('C2H3COCH3 + CH3O2 => CH2CHO + CH3CO + CH3O', [3.970000e+11, 0.0, 17050.0])

# Reaction 1989
reaction('CH2CHO + CH3CO + CH3O => C2H3COCH3 + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1990
reaction('C2H3COCH3 + CH3O2 => CH2CO + C2H3 + CH3O2H', [3.010000e+12, 0.0, 17580.0])

# Reaction 1991
reaction('CH2CO + C2H3 + CH3O2H => C2H3COCH3 + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 1992
falloff_reaction('IC4H10 (+ M) <=> CH3 + IC3H7 (+ M)',
                kf=[4.830000e+16, 0.0, 79900.0],
                kf0=[2.410000e+19, 0.0, 52576.0],
                falloff=Troe(A=0.25, T3=750.0, T1=1e-10, T2=10000000000.0))

# Reaction 1993
reaction('IC4H10 => TC4H9 + H', [2.510000e+98, -23.81, 145300.0])

# Reaction 1994
reaction('TC4H9 + H => IC4H10', [1.131000e+93, -22.873, 48390.0])

# Reaction 1995
reaction('IC4H10 => IC4H9 + H', [9.850000e+95, -23.11, 147600.0])

# Reaction 1996
reaction('IC4H9 + H => IC4H10', [2.248000e+92, -22.752, 46360.0])

# Reaction 1997
reaction('IC4H10 + H => TC4H9 + H2', [6.020000e+05, 2.4, 2583.0])

# Reaction 1998
reaction('TC4H9 + H2 => IC4H10 + H', [1.084000e+00, 3.613, 9213.0])

# Reaction 1999
reaction('IC4H10 + H => IC4H9 + H2', [1.810000e+06, 2.54, 6756.0])

# Reaction 2000
reaction('IC4H9 + H2 => IC4H10 + H', [1.651000e+02, 3.174, 9096.0])

# Reaction 2001
reaction('IC4H10 + CH3 => TC4H9 + CH4', [9.040000e-01, 3.46, 4598.0])

# Reaction 2002
reaction('TC4H9 + CH4 => IC4H10 + CH3', [1.485000e-03, 4.227, 12770.0])

# Reaction 2003
reaction('IC4H10 + CH3 => IC4H9 + CH4', [1.360000e+00, 3.65, 7154.0])

# Reaction 2004
reaction('IC4H9 + CH4 => IC4H10 + CH3', [1.132000e-01, 3.838, 11030.0])

# Reaction 2005
reaction('IC4H10 + OH => TC4H9 + H2O', [2.925000e+04, 2.531, -1659.0])

# Reaction 2006
reaction('TC4H9 + H2O => IC4H10 + OH', [5.584000e-01, 3.638, 19860.0])

# Reaction 2007
reaction('IC4H10 + OH => IC4H9 + H2O', [6.654000e+04, 2.665, -168.9])

# Reaction 2008
reaction('IC4H9 + H2O => IC4H10 + OH', [6.437000e+01, 3.193, 17060.0])

# Reaction 2009
reaction('IC4H10 + C2H5 => IC4H9 + C2H6', [1.510000e+12, 0.0, 10400.0])

# Reaction 2010
reaction('IC4H9 + C2H6 => IC4H10 + C2H5', [3.200000e+11, 0.0, 12300.0])

# Reaction 2011
reaction('IC4H10 + C2H5 => TC4H9 + C2H6', [1.000000e+11, 0.0, 7900.0])

# Reaction 2012
reaction('TC4H9 + C2H6 => IC4H10 + C2H5', [3.000000e+11, 0.0, 21000.0])

# Reaction 2013
reaction('IC4H10 + HO2 => IC4H9 + H2O2', [6.120000e+01, 3.59, 17160.0])

# Reaction 2014
reaction('IC4H9 + H2O2 => IC4H10 + HO2', [3.231000e+00, 3.529, 3500.0])

# Reaction 2015
reaction('IC4H10 + HO2 => TC4H9 + H2O2', [4.332000e+02, 3.01, 12090.0])

# Reaction 2016
reaction('TC4H9 + H2O2 => IC4H10 + HO2', [4.513000e-01, 3.528, 2723.0])

# Reaction 2017
reaction('IC4H10 + O => TC4H9 + OH', [1.968000e+05, 2.402, 1150.0])

# Reaction 2018
reaction('TC4H9 + OH => IC4H10 + O', [1.839000e-01, 3.596, 6368.0])

# Reaction 2019
reaction('IC4H10 + O => IC4H9 + OH', [4.046000e+07, 2.034, 5136.0])

# Reaction 2020
reaction('IC4H9 + OH => IC4H10 + O', [1.916000e+03, 2.649, 6064.0])

# Reaction 2021
reaction('IC4H10 + CH3O => IC4H9 + CH3OH', [4.800000e+11, 0.0, 7000.0])

# Reaction 2022
reaction('IC4H9 + CH3OH => IC4H10 + CH3O', [9.402000e+08, 0.421, 9567.0])

# Reaction 2023
reaction('IC4H10 + CH3O => TC4H9 + CH3OH', [1.900000e+10, 0.0, 2800.0])

# Reaction 2024
reaction('TC4H9 + CH3OH => IC4H10 + CH3O', [7.344000e+05, 1.0, 9657.0])

# Reaction 2025
reaction('IC4H10 + O2 => IC4H9 + HO2', [9.000000e+13, 0.0, 52290.0])

# Reaction 2026
reaction('IC4H9 + HO2 => IC4H10 + O2', [4.306000e+10, 0.286, -59.0])

# Reaction 2027
reaction('IC4H10 + O2 => TC4H9 + HO2', [1.000000e+13, 0.0, 48200.0])

# Reaction 2028
reaction('TC4H9 + HO2 => IC4H10 + O2', [9.442000e+07, 0.865, 141.0])

# Reaction 2029
reaction('IC4H10 + CH3O2 => IC4H9 + CH3O2H', [2.079000e+00, 3.97, 18280.0])

# Reaction 2030
reaction('IC4H9 + CH3O2H => IC4H10 + CH3O2', [2.135000e+00, 3.464, 3026.0])

# Reaction 2031
reaction('IC4H10 + C2H5O2 => IC4H9 + C2H5O2H', [2.550000e+13, 0.0, 20460.0])

# Reaction 2032
reaction('IC4H9 + C2H5O2H => IC4H10 + C2H5O2', [2.623000e+13, -0.506, 5208.0])

# Reaction 2033
reaction('IC4H10 + CH3CO3 => IC4H9 + CH3CO3H', [2.550000e+13, 0.0, 20460.0])

# Reaction 2034
reaction('IC4H9 + CH3CO3H => IC4H10 + CH3CO3', [2.147000e+12, -0.009, 9570.0])

# Reaction 2035
reaction('IC4H10 + NC3H7O2 => IC4H9 + NC3H7O2H', [2.550000e+13, 0.0, 20460.0])

# Reaction 2036
reaction('IC4H9 + NC3H7O2H => IC4H10 + NC3H7O2', [2.678000e+13, -0.509, 5000.0])

# Reaction 2037
reaction('IC4H10 + IC3H7O2 => IC4H9 + IC3H7O2H', [2.550000e+13, 0.0, 20460.0])

# Reaction 2038
reaction('IC4H9 + IC3H7O2H => IC4H10 + IC3H7O2', [2.688000e+13, -0.51, 5000.0])

# Reaction 2039
reaction('IC4H10 + IC4H9O2 => IC4H9 + IC4H9O2H', [2.550000e+13, 0.0, 20460.0])

# Reaction 2040
reaction('IC4H9 + IC4H9O2H => IC4H10 + IC4H9O2', [2.679000e+13, -0.509, 5010.0])

# Reaction 2041
reaction('IC4H10 + TC4H9O2 => IC4H9 + TC4H9O2H', [2.550000e+13, 0.0, 20460.0])

# Reaction 2042
reaction('IC4H9 + TC4H9O2H => IC4H10 + TC4H9O2', [2.684000e+13, -0.51, 5010.0])

# Reaction 2043
reaction('IC4H10 + O2CHO => IC4H9 + HO2CHO', [2.520000e+13, 0.0, 20440.0])

# Reaction 2044
reaction('IC4H9 + HO2CHO => IC4H10 + O2CHO', [1.006000e+00, 3.06, 2375.0])

# Reaction 2045
reaction('IC4H10 + O2CHO => TC4H9 + HO2CHO', [2.800000e+12, 0.0, 16010.0])

# Reaction 2046
reaction('TC4H9 + HO2CHO => IC4H10 + O2CHO', [1.893000e+01, 2.72, 4678.0])

# Reaction 2047
reaction('IC4H10 + SC4H9O2 => IC4H9 + SC4H9O2H', [2.250000e+13, 0.0, 20460.0])

# Reaction 2048
reaction('IC4H9 + SC4H9O2H => IC4H10 + SC4H9O2', [5.000000e+11, 0.0, 6500.0])

# Reaction 2049
reaction('IC4H10 + SC4H9O2 => TC4H9 + SC4H9O2H', [2.800000e+12, 0.0, 16000.0])

# Reaction 2050
reaction('TC4H9 + SC4H9O2H => IC4H10 + SC4H9O2', [5.000000e+11, 0.0, 6500.0])

# Reaction 2051
reaction('IC4H10 + PC4H9O2 => IC4H9 + PC4H9O2H', [2.250000e+13, 0.0, 20460.0])

# Reaction 2052
reaction('IC4H9 + PC4H9O2H => IC4H10 + PC4H9O2', [5.000000e+11, 0.0, 6500.0])

# Reaction 2053
reaction('IC4H10 + PC4H9O2 => TC4H9 + PC4H9O2H', [2.800000e+12, 0.0, 16000.0])

# Reaction 2054
reaction('TC4H9 + PC4H9O2H => IC4H10 + PC4H9O2', [5.000000e+11, 0.0, 6500.0])

# Reaction 2055
reaction('IC4H10 + CH3O2 => TC4H9 + CH3O2H', [1.366000e+02, 3.12, 13190.0])

# Reaction 2056
reaction('TC4H9 + CH3O2H => IC4H10 + CH3O2', [2.769000e+00, 3.193, 2228.0])

# Reaction 2057
reaction('IC4H10 + C2H5O2 => TC4H9 + C2H5O2H', [2.800000e+12, 0.0, 16000.0])

# Reaction 2058
reaction('TC4H9 + C2H5O2H => IC4H10 + C2H5O2', [1.000000e+11, 0.0, 10000.0])

# Reaction 2059
reaction('IC4H10 + CH3CO3 => TC4H9 + CH3CO3H', [2.800000e+12, 0.0, 16000.0])

# Reaction 2060
reaction('TC4H9 + CH3CO3H => IC4H10 + CH3CO3', [1.000000e+11, 0.0, 10000.0])

# Reaction 2061
reaction('IC4H10 + NC3H7O2 => TC4H9 + NC3H7O2H', [2.800000e+12, 0.0, 16000.0])

# Reaction 2062
reaction('TC4H9 + NC3H7O2H => IC4H10 + NC3H7O2', [1.000000e+11, 0.0, 10000.0])

# Reaction 2063
reaction('IC4H10 + IC3H7O2 => TC4H9 + IC3H7O2H', [2.800000e+12, 0.0, 16000.0])

# Reaction 2064
reaction('TC4H9 + IC3H7O2H => IC4H10 + IC3H7O2', [1.000000e+11, 0.0, 10000.0])

# Reaction 2065
reaction('IC4H10 + IC4H9O2 => TC4H9 + IC4H9O2H', [2.800000e+12, 0.0, 16000.0])

# Reaction 2066
reaction('TC4H9 + IC4H9O2H => IC4H10 + IC4H9O2', [1.000000e+11, 0.0, 10000.0])

# Reaction 2067
reaction('IC4H10 + TC4H9O2 => TC4H9 + TC4H9O2H', [2.800000e+12, 0.0, 16000.0])

# Reaction 2068
reaction('TC4H9 + TC4H9O2H => IC4H10 + TC4H9O2', [1.000000e+11, 0.0, 10000.0])

# Reaction 2069
reaction('IC4H10 + IC4H9 => TC4H9 + IC4H10', [2.500000e+10, 0.0, 7900.0])

# Reaction 2070
reaction('TC4H9 + IC4H10 => IC4H10 + IC4H9', [2.250000e+11, 0.0, 12300.0])

# Reaction 2071
reaction('IC4H9 + HO2 => IC4H9O + OH', [7.000000e+12, 0.0, -1000.0])

# Reaction 2072
reaction('IC4H9O + OH => IC4H9 + HO2', [3.712000e+15, -0.658, 26540.0])

# Reaction 2073
reaction('TC4H9 + HO2 => TC4H9O + OH', [7.000000e+12, 0.0, -1000.0])

# Reaction 2074
reaction('TC4H9O + OH => TC4H9 + HO2', [4.083000e+18, -1.329, 28650.0])

# Reaction 2075
reaction('CH3O2 + IC4H9 => CH3O + IC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2076
reaction('CH3O + IC4H9O => CH3O2 + IC4H9', [2.322000e+14, -0.359, 30780.0])

# Reaction 2077
reaction('CH3O2 + TC4H9 => CH3O + TC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2078
reaction('CH3O + TC4H9O => CH3O2 + TC4H9', [2.554000e+17, -1.03, 32890.0])

# Reaction 2079
reaction('IC4H9 => IC4H8 + H', [3.371000e+13, 0.124, 33660.0])

# Reaction 2080
reaction('IC4H8 + H => IC4H9', [6.250000e+11, 0.51, 2620.0])

# Reaction 2081
reaction('IC4H9 => C3H6 + CH3', [9.504000e+11, 0.773, 30700.0])

# Reaction 2082
reaction('C3H6 + CH3 => IC4H9', [1.890000e+03, 2.67, 6850.0])

# Reaction 2083
reaction('TC4H9 => H + IC4H8', [1.128000e+12, 0.703, 36560.0])

# Reaction 2084
reaction('H + IC4H8 => TC4H9', [1.060000e+12, 0.51, 1230.0])

# Reaction 2085
reaction('TC4H9 + O2 => IC4H8 + HO2', [8.370000e-01, 3.59, 11960.0])

# Reaction 2086
reaction('IC4H8 + HO2 => TC4H9 + O2', [1.648000e+00, 3.325, 25500.0])

# Reaction 2087
reaction('IC4H9 + O2 => IC4H8 + HO2', [1.070000e+00, 3.71, 9322.0])

# Reaction 2088
reaction('IC4H8 + HO2 => IC4H9 + O2', [4.158000e-02, 4.024, 27150.0])

# Reaction 2089
reaction('NC3H7O2 + IC4H9 => NC3H7O + IC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2090
reaction('NC3H7O + IC4H9O => NC3H7O2 + IC4H9', [4.543000e+14, -0.526, 29220.0])

# Reaction 2091
reaction('NC3H7O2 + TC4H9 => NC3H7O + TC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2092
reaction('NC3H7O + TC4H9O => NC3H7O2 + TC4H9', [4.997000e+17, -1.197, 31330.0])

# Reaction 2093
reaction('NC3H7O2 + IC4H7 => NC3H7O + IC4H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2094
reaction('NC3H7O + IC4H7O => NC3H7O2 + IC4H7', [4.183000e+11, 0.182, 13500.0])

# Reaction 2095
reaction('SC4H9O2 + IC4H9 => SC4H9O + IC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2096
reaction('SC4H9O + IC4H9O => SC4H9O2 + IC4H9', [1.007000e+15, -0.468, 28410.0])

# Reaction 2097
reaction('SC4H9O2 + TC4H9 => SC4H9O + TC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2098
reaction('SC4H9O + TC4H9O => SC4H9O2 + TC4H9', [1.108000e+18, -1.139, 30520.0])

# Reaction 2099
reaction('PC4H9O2 + IC4H9 => PC4H9O + IC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2100
reaction('PC4H9O + IC4H9O => PC4H9O2 + IC4H9', [5.310000e+14, -0.551, 28930.0])

# Reaction 2101
reaction('PC4H9O2 + TC4H9 => PC4H9O + TC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2102
reaction('PC4H9O + TC4H9O => PC4H9O2 + TC4H9', [5.841000e+17, -1.222, 31040.0])

# Reaction 2103
reaction('PC4H9O2 + IC4H7 => PC4H9O + IC4H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2104
reaction('PC4H9O + IC4H7O => PC4H9O2 + IC4H7', [4.890000e+11, 0.157, 13210.0])

# Reaction 2105
reaction('SC4H9O2 + IC4H7 => SC4H9O + IC4H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2106
reaction('SC4H9O + IC4H7O => SC4H9O2 + IC4H7', [9.277000e+11, 0.24, 12690.0])

# Reaction 2107
reaction('IC4H9O2 => IC4H9 + O2', [6.640000e+19, -1.575, 36080.0])

# Reaction 2108
reaction('IC4H9 + O2 => IC4H9O2', [2.260000e+12, 0.0, 0.0])

# Reaction 2109
reaction('TC4H9O2 => TC4H9 + O2', [3.331000e+24, -2.472, 37870.0])

# Reaction 2110
reaction('TC4H9 + O2 => TC4H9O2', [1.410000e+13, 0.0, 0.0])

# Reaction 2111
reaction('IC4H9O2 + C4H10 => IC4H9O2H + SC4H9', [1.120000e+13, 0.0, 17700.0])

# Reaction 2112
reaction('IC4H9O2H + SC4H9 => IC4H9O2 + C4H10', [5.000000e+11, 0.0, 6500.0])

# Reaction 2113
reaction('TC4H9O2 + C4H10 => TC4H9O2H + SC4H9', [1.120000e+13, 0.0, 17700.0])

# Reaction 2114
reaction('TC4H9O2H + SC4H9 => TC4H9O2 + C4H10', [5.000000e+11, 0.0, 6500.0])

# Reaction 2115
reaction('IC4H9O2 + C4H10 => IC4H9O2H + PC4H9', [1.700000e+13, 0.0, 20460.0])

# Reaction 2116
reaction('IC4H9O2H + PC4H9 => IC4H9O2 + C4H10', [5.000000e+11, 0.0, 6500.0])

# Reaction 2117
reaction('TC4H9O2 + C4H10 => TC4H9O2H + PC4H9', [1.700000e+13, 0.0, 20460.0])

# Reaction 2118
reaction('TC4H9O2H + PC4H9 => TC4H9O2 + C4H10', [5.000000e+11, 0.0, 6500.0])

# Reaction 2119
reaction('IC3H7O2 + IC4H9 => IC3H7O + IC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2120
reaction('IC3H7O + IC4H9O => IC3H7O2 + IC4H9', [9.811000e+12, 0.16, 29020.0])

# Reaction 2121
reaction('IC3H7O2 + TC4H9 => IC3H7O + TC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2122
reaction('IC3H7O + TC4H9O => IC3H7O2 + TC4H9', [1.079000e+16, -0.511, 31130.0])

# Reaction 2123
reaction('IC3H7O2 + IC4H7 => IC3H7O + IC4H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2124
reaction('IC3H7O + IC4H7O => IC3H7O2 + IC4H7', [9.035000e+09, 0.868, 13300.0])

# Reaction 2125
reaction('IC4H9O2 + C3H6 => IC4H9O2H + C3H5-A', [3.240000e+11, 0.0, 14900.0])

# Reaction 2126
reaction('IC4H9O2H + C3H5-A => IC4H9O2 + C3H6', [2.000000e+10, 0.0, 15000.0])

# Reaction 2127
reaction('TC4H9O2 + C3H6 => TC4H9O2H + C3H5-A', [3.240000e+11, 0.0, 14900.0])

# Reaction 2128
reaction('TC4H9O2H + C3H5-A => TC4H9O2 + C3H6', [2.000000e+10, 0.0, 15000.0])

# Reaction 2129
reaction('IC4H9O2 + IC4H8 => IC4H9O2H + IC4H7', [1.400000e+12, 0.0, 14900.0])

# Reaction 2130
reaction('IC4H9O2H + IC4H7 => IC4H9O2 + IC4H8', [3.160000e+11, 0.0, 13000.0])

# Reaction 2131
reaction('TC4H9O2 + IC4H8 => TC4H9O2H + IC4H7', [1.400000e+12, 0.0, 14900.0])

# Reaction 2132
reaction('TC4H9O2H + IC4H7 => TC4H9O2 + IC4H8', [3.160000e+11, 0.0, 13000.0])

# Reaction 2133
reaction('PC4H9O2 + IC4H8 => PC4H9O2H + IC4H7', [1.400000e+12, 0.0, 14900.0])

# Reaction 2134
reaction('PC4H9O2H + IC4H7 => PC4H9O2 + IC4H8', [3.160000e+11, 0.0, 13000.0])

# Reaction 2135
reaction('SC4H9O2 + IC4H8 => SC4H9O2H + IC4H7', [1.400000e+12, 0.0, 14900.0])

# Reaction 2136
reaction('SC4H9O2H + IC4H7 => SC4H9O2 + IC4H8', [3.160000e+11, 0.0, 13000.0])

# Reaction 2137
reaction('IC3H7O2 + IC4H8 => IC3H7O2H + IC4H7', [1.400000e+12, 0.0, 14900.0])

# Reaction 2138
reaction('IC3H7O2H + IC4H7 => IC3H7O2 + IC4H8', [3.160000e+11, 0.0, 13000.0])

# Reaction 2139
reaction('NC3H7O2 + IC4H8 => NC3H7O2H + IC4H7', [1.400000e+12, 0.0, 14900.0])

# Reaction 2140
reaction('NC3H7O2H + IC4H7 => NC3H7O2 + IC4H8', [3.160000e+11, 0.0, 13000.0])

# Reaction 2141
reaction('IC4H9O2 + C4H8-1 => IC4H9O2H + C4H71-3', [1.400000e+12, 0.0, 14900.0])

# Reaction 2142
reaction('IC4H9O2H + C4H71-3 => IC4H9O2 + C4H8-1', [3.160000e+11, 0.0, 13000.0])

# Reaction 2143
reaction('TC4H9O2 + C4H8-1 => TC4H9O2H + C4H71-3', [1.400000e+12, 0.0, 14900.0])

# Reaction 2144
reaction('TC4H9O2H + C4H71-3 => TC4H9O2 + C4H8-1', [3.160000e+11, 0.0, 13000.0])

# Reaction 2145
reaction('IC4H9O2 + C4H8-2 => IC4H9O2H + C4H71-3', [1.400000e+12, 0.0, 14900.0])

# Reaction 2146
reaction('IC4H9O2H + C4H71-3 => IC4H9O2 + C4H8-2', [3.160000e+11, 0.0, 13000.0])

# Reaction 2147
reaction('TC4H9O2 + C4H8-2 => TC4H9O2H + C4H71-3', [1.400000e+12, 0.0, 14900.0])

# Reaction 2148
reaction('TC4H9O2H + C4H71-3 => TC4H9O2 + C4H8-2', [3.160000e+11, 0.0, 13000.0])

# Reaction 2149
reaction('CC4H8O + OH => CH2O + C3H5-A + H2O', [5.000000e+12, 0.0, 0.0])

# Reaction 2150
reaction('CH2O + C3H5-A + H2O => CC4H8O + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 2151
reaction('CC4H8O + H => CH2O + C3H5-A + H2', [3.510000e+07, 2.0, 5000.0])

# Reaction 2152
reaction('CH2O + C3H5-A + H2 => CC4H8O + H', [0.000000e+00, 0.0, 0.0])

# Reaction 2153
reaction('CC4H8O + O => CH2O + C3H5-A + OH', [1.124000e+14, 0.0, 5200.0])

# Reaction 2154
reaction('CH2O + C3H5-A + OH => CC4H8O + O', [0.000000e+00, 0.0, 0.0])

# Reaction 2155
reaction('CC4H8O + HO2 => CH2O + C3H5-A + H2O2', [1.000000e+13, 0.0, 15000.0])

# Reaction 2156
reaction('CH2O + C3H5-A + H2O2 => CC4H8O + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 2157
reaction('CC4H8O + CH3O2 => CH2O + C3H5-A + CH3O2H', [1.000000e+13, 0.0, 19000.0])

# Reaction 2158
reaction('CH2O + C3H5-A + CH3O2H => CC4H8O + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 2159
reaction('CC4H8O + CH3 => CH2O + C3H5-A + CH4', [2.000000e+11, 0.0, 10000.0])

# Reaction 2160
reaction('CH2O + C3H5-A + CH4 => CC4H8O + CH3', [0.000000e+00, 0.0, 0.0])

# Reaction 2161
reaction('C2H4 + TC4H9O2 => C2H3 + TC4H9O2H', [7.000000e+11, 0.0, 17110.0])

# Reaction 2162
reaction('C2H3 + TC4H9O2H => C2H4 + TC4H9O2', [1.000000e+11, 0.0, 10000.0])

# Reaction 2163
reaction('TC4H9O2 + CH4 => TC4H9O2H + CH3', [1.130000e+13, 0.0, 20460.0])

# Reaction 2164
reaction('TC4H9O2H + CH3 => TC4H9O2 + CH4', [7.500000e+08, 0.0, 1280.0])

# Reaction 2165
reaction('H2 + TC4H9O2 => H + TC4H9O2H', [3.010000e+13, 0.0, 26030.0])

# Reaction 2166
reaction('H + TC4H9O2H => H2 + TC4H9O2', [4.800000e+13, 0.0, 7950.0])

# Reaction 2167
reaction('TC4H9O2 + C2H6 => TC4H9O2H + C2H5', [1.700000e+13, 0.0, 20460.0])

# Reaction 2168
reaction('TC4H9O2H + C2H5 => TC4H9O2 + C2H6', [5.000000e+11, 0.0, 6500.0])

# Reaction 2169
reaction('TC4H9O2 + C3H8 => TC4H9O2H + IC3H7', [2.000000e+12, 0.0, 17000.0])

# Reaction 2170
reaction('TC4H9O2H + IC3H7 => TC4H9O2 + C3H8', [5.000000e+11, 0.0, 6500.0])

# Reaction 2171
reaction('TC4H9O2 + C3H8 => TC4H9O2H + NC3H7', [1.700000e+13, 0.0, 20460.0])

# Reaction 2172
reaction('TC4H9O2H + NC3H7 => TC4H9O2 + C3H8', [5.000000e+11, 0.0, 6500.0])

# Reaction 2173
reaction('TC4H9O2 + CH3OH => TC4H9O2H + CH2OH', [6.300000e+12, 0.0, 19360.0])

# Reaction 2174
reaction('TC4H9O2H + CH2OH => TC4H9O2 + CH3OH', [1.000000e+09, 0.0, 10000.0])

# Reaction 2175
reaction('TC4H9O2 + C2H5OH => TC4H9O2H + PC2H4OH', [6.300000e+12, 0.0, 19360.0])

# Reaction 2176
reaction('TC4H9O2H + PC2H4OH => TC4H9O2 + C2H5OH', [3.061000e+12, 0.0, 22100.0])

# Reaction 2177
reaction('TC4H9O2 + C2H5OH => TC4H9O2H + SC2H4OH', [4.200000e+12, 0.0, 15000.0])

# Reaction 2178
reaction('TC4H9O2H + SC2H4OH => TC4H9O2 + C2H5OH', [2.040000e+12, 0.0, 17740.0])

# Reaction 2179
reaction('IC4H9O2 + CH3CHO => IC4H9O2H + CH3CO', [2.800000e+12, 0.0, 13600.0])

# Reaction 2180
reaction('IC4H9O2H + CH3CO => IC4H9O2 + CH3CHO', [1.000000e+12, 0.0, 10000.0])

# Reaction 2181
reaction('TC4H9O2 + CH3CHO => TC4H9O2H + CH3CO', [2.800000e+12, 0.0, 13600.0])

# Reaction 2182
reaction('TC4H9O2H + CH3CO => TC4H9O2 + CH3CHO', [1.000000e+12, 0.0, 10000.0])

# Reaction 2183
reaction('IC4H9O2 + C2H3CHO => IC4H9O2H + C2H3CO', [2.800000e+12, 0.0, 13600.0])

# Reaction 2184
reaction('IC4H9O2H + C2H3CO => IC4H9O2 + C2H3CHO', [1.000000e+12, 0.0, 10000.0])

# Reaction 2185
reaction('TC4H9O2 + C2H3CHO => TC4H9O2H + C2H3CO', [2.800000e+12, 0.0, 13600.0])

# Reaction 2186
reaction('TC4H9O2H + C2H3CO => TC4H9O2 + C2H3CHO', [1.000000e+12, 0.0, 10000.0])

# Reaction 2187
reaction('IC4H9O2 + C2H5CHO => IC4H9O2H + C2H5CO', [2.800000e+12, 0.0, 13600.0])

# Reaction 2188
reaction('IC4H9O2H + C2H5CO => IC4H9O2 + C2H5CHO', [1.000000e+12, 0.0, 10000.0])

# Reaction 2189
reaction('TC4H9O2 + C2H5CHO => TC4H9O2H + C2H5CO', [2.800000e+12, 0.0, 13600.0])

# Reaction 2190
reaction('TC4H9O2H + C2H5CO => TC4H9O2 + C2H5CHO', [1.000000e+12, 0.0, 10000.0])

# Reaction 2191
reaction('IC4H9O2 + HO2 => IC4H9O2H + O2', [1.750000e+10, 0.0, -3275.0])

# Reaction 2192
reaction('IC4H9O2H + O2 => IC4H9O2 + HO2', [3.843000e+13, -0.795, 33620.0])

# Reaction 2193
reaction('TC4H9O2 + HO2 => TC4H9O2H + O2', [1.750000e+10, 0.0, -3275.0])

# Reaction 2194
reaction('TC4H9O2H + O2 => TC4H9O2 + HO2', [3.850000e+13, -0.795, 33620.0])

# Reaction 2195
reaction('IC4H9O2 + H2O2 => IC4H9O2H + HO2', [2.400000e+12, 0.0, 10000.0])

# Reaction 2196
reaction('IC4H9O2H + HO2 => IC4H9O2 + H2O2', [2.400000e+12, 0.0, 10000.0])

# Reaction 2197
reaction('TC4H9O2 + H2O2 => TC4H9O2H + HO2', [2.400000e+12, 0.0, 10000.0])

# Reaction 2198
reaction('TC4H9O2H + HO2 => TC4H9O2 + H2O2', [2.400000e+12, 0.0, 10000.0])

# Reaction 2199
reaction('IC4H9O2 + CH2O => IC4H9O2H + HCO', [1.300000e+11, 0.0, 9000.0])

# Reaction 2200
reaction('IC4H9O2H + HCO => IC4H9O2 + CH2O', [2.500000e+10, 0.0, 10100.0])

# Reaction 2201
reaction('TC4H9O2 + CH2O => TC4H9O2H + HCO', [1.300000e+11, 0.0, 9000.0])

# Reaction 2202
reaction('TC4H9O2H + HCO => TC4H9O2 + CH2O', [2.500000e+10, 0.0, 10100.0])

# Reaction 2203
reaction('IC4H9O2 + CH3O2 => IC4H9O + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 2204
reaction('IC4H9O + CH3O + O2 => IC4H9O2 + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 2205
reaction('TC4H9O2 + CH3O2 => TC4H9O + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 2206
reaction('TC4H9O + CH3O + O2 => TC4H9O2 + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 2207
reaction('IC4H9O2 + C2H5O2 => IC4H9O + C2H5O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 2208
reaction('IC4H9O + C2H5O + O2 => IC4H9O2 + C2H5O2', [0.000000e+00, 0.0, 0.0])

# Reaction 2209
reaction('TC4H9O2 + C2H5O2 => TC4H9O + C2H5O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 2210
reaction('TC4H9O + C2H5O + O2 => TC4H9O2 + C2H5O2', [0.000000e+00, 0.0, 0.0])

# Reaction 2211
reaction('IC4H9O2 + CH3CO3 => IC4H9O + CH3CO2 + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 2212
reaction('IC4H9O + CH3CO2 + O2 => IC4H9O2 + CH3CO3', [0.000000e+00, 0.0, 0.0])

# Reaction 2213
reaction('TC4H9O2 + CH3CO3 => TC4H9O + CH3CO2 + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 2214
reaction('TC4H9O + CH3CO2 + O2 => TC4H9O2 + CH3CO3', [0.000000e+00, 0.0, 0.0])

# Reaction 2215
reaction('IC4H9O2 + IC4H9O2 => O2 + IC4H9O + IC4H9O', [1.400000e+16, -1.61, 1860.0])

# Reaction 2216
reaction('O2 + IC4H9O + IC4H9O => IC4H9O2 + IC4H9O2', [0.000000e+00, 0.0, 0.0])

# Reaction 2217
reaction('IC4H9O2 + TC4H9O2 => IC4H9O + TC4H9O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 2218
reaction('IC4H9O + TC4H9O + O2 => IC4H9O2 + TC4H9O2', [0.000000e+00, 0.0, 0.0])

# Reaction 2219
reaction('TC4H9O2 + TC4H9O2 => O2 + TC4H9O + TC4H9O', [1.400000e+16, -1.61, 1860.0])

# Reaction 2220
reaction('O2 + TC4H9O + TC4H9O => TC4H9O2 + TC4H9O2', [0.000000e+00, 0.0, 0.0])

# Reaction 2221
reaction('IC4H9O2 + PC4H9O2 => IC4H9O + PC4H9O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 2222
reaction('IC4H9O + PC4H9O + O2 => IC4H9O2 + PC4H9O2', [0.000000e+00, 0.0, 0.0])

# Reaction 2223
reaction('TC4H9O2 + PC4H9O2 => TC4H9O + PC4H9O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 2224
reaction('TC4H9O + PC4H9O + O2 => TC4H9O2 + PC4H9O2', [0.000000e+00, 0.0, 0.0])

# Reaction 2225
reaction('IC4H9O2 + SC4H9O2 => IC4H9O + SC4H9O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 2226
reaction('IC4H9O + SC4H9O + O2 => IC4H9O2 + SC4H9O2', [0.000000e+00, 0.0, 0.0])

# Reaction 2227
reaction('TC4H9O2 + SC4H9O2 => TC4H9O + SC4H9O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 2228
reaction('TC4H9O + SC4H9O + O2 => TC4H9O2 + SC4H9O2', [0.000000e+00, 0.0, 0.0])

# Reaction 2229
reaction('IC4H9O2 + NC3H7O2 => IC4H9O + NC3H7O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 2230
reaction('IC4H9O + NC3H7O + O2 => IC4H9O2 + NC3H7O2', [0.000000e+00, 0.0, 0.0])

# Reaction 2231
reaction('TC4H9O2 + NC3H7O2 => TC4H9O + NC3H7O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 2232
reaction('TC4H9O + NC3H7O + O2 => TC4H9O2 + NC3H7O2', [0.000000e+00, 0.0, 0.0])

# Reaction 2233
reaction('IC4H9O2 + IC3H7O2 => IC4H9O + IC3H7O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 2234
reaction('IC4H9O + IC3H7O + O2 => IC4H9O2 + IC3H7O2', [0.000000e+00, 0.0, 0.0])

# Reaction 2235
reaction('TC4H9O2 + IC3H7O2 => TC4H9O + IC3H7O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 2236
reaction('TC4H9O + IC3H7O + O2 => TC4H9O2 + IC3H7O2', [0.000000e+00, 0.0, 0.0])

# Reaction 2237
reaction('IC4H9O2 + HO2 => IC4H9O + OH + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 2238
reaction('IC4H9O + OH + O2 => IC4H9O2 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 2239
reaction('TC4H9O2 + HO2 => TC4H9O + OH + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 2240
reaction('TC4H9O + OH + O2 => TC4H9O2 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 2241
reaction('IC4H9O2 + CH3 => IC4H9O + CH3O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2242
reaction('IC4H9O + CH3O => IC4H9O2 + CH3', [5.720000e+12, 0.003, 28020.0])

# Reaction 2243
reaction('IC4H9O2 + C2H5 => IC4H9O + C2H5O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2244
reaction('IC4H9O + C2H5O => IC4H9O2 + C2H5', [8.133000e+14, -0.599, 30430.0])

# Reaction 2245
reaction('IC4H9O2 + IC3H7 => IC4H9O + IC3H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2246
reaction('IC4H9O + IC3H7O => IC4H9O2 + IC3H7', [1.387000e+15, -0.432, 31100.0])

# Reaction 2247
reaction('IC4H9O2 + NC3H7 => IC4H9O + NC3H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2248
reaction('IC4H9O + NC3H7O => IC4H9O2 + NC3H7', [8.212000e+14, -0.568, 29100.0])

# Reaction 2249
reaction('IC4H9O2 + PC4H9 => IC4H9O + PC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2250
reaction('IC4H9O + PC4H9O => IC4H9O2 + PC4H9', [1.139000e+15, -0.618, 28780.0])

# Reaction 2251
reaction('IC4H9O2 + SC4H9 => IC4H9O + SC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2252
reaction('IC4H9O + SC4H9O => IC4H9O2 + SC4H9', [1.969000e+17, -1.11, 30490.0])

# Reaction 2253
reaction('IC4H9O2 + IC4H9 => IC4H9O + IC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2254
reaction('IC4H9O + IC4H9O => IC4H9O2 + IC4H9', [4.900000e+14, -0.533, 30320.0])

# Reaction 2255
reaction('IC4H9O2 + TC4H9 => IC4H9O + TC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2256
reaction('IC4H9O + TC4H9O => IC4H9O2 + TC4H9', [5.390000e+17, -1.204, 32430.0])

# Reaction 2257
reaction('IC4H9O2 + C3H5-A => IC4H9O + C3H5O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2258
reaction('IC4H9O + C3H5O => IC4H9O2 + C3H5-A', [2.119000e+11, 0.184, 15440.0])

# Reaction 2259
reaction('IC4H9O2 + C4H71-3 => IC4H9O + C4H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2260
reaction('IC4H9O + C4H7O => IC4H9O2 + C4H71-3', [1.095000e+14, -0.745, 17490.0])

# Reaction 2261
reaction('IC4H9O2 + IC4H7 => IC4H9O + IC4H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2262
reaction('IC4H9O + IC4H7O => IC4H9O2 + IC4H7', [4.512000e+11, 0.175, 14600.0])

# Reaction 2263
reaction('TC4H9O2 + CH3 => TC4H9O + CH3O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2264
reaction('TC4H9O + CH3O => TC4H9O2 + CH3', [7.824000e+11, 0.229, 28340.0])

# Reaction 2265
reaction('TC4H9O2 + C2H5 => TC4H9O + C2H5O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2266
reaction('TC4H9O + C2H5O => TC4H9O2 + C2H5', [1.112000e+14, -0.372, 30750.0])

# Reaction 2267
reaction('TC4H9O2 + IC3H7 => TC4H9O + IC3H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2268
reaction('TC4H9O + IC3H7O => TC4H9O2 + IC3H7', [1.898000e+14, -0.206, 31420.0])

# Reaction 2269
reaction('TC4H9O2 + NC3H7 => TC4H9O + NC3H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2270
reaction('TC4H9O + NC3H7O => TC4H9O2 + NC3H7', [1.123000e+14, -0.341, 29420.0])

# Reaction 2271
reaction('TC4H9O2 + PC4H9 => TC4H9O + PC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2272
reaction('TC4H9O + PC4H9O => TC4H9O2 + PC4H9', [1.558000e+14, -0.392, 29100.0])

# Reaction 2273
reaction('TC4H9O2 + SC4H9 => TC4H9O + SC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2274
reaction('TC4H9O + SC4H9O => TC4H9O2 + SC4H9', [2.693000e+16, -0.883, 30810.0])

# Reaction 2275
reaction('TC4H9O2 + IC4H9 => TC4H9O + IC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2276
reaction('TC4H9O + IC4H9O => TC4H9O2 + IC4H9', [6.703000e+13, -0.307, 30640.0])

# Reaction 2277
reaction('TC4H9O2 + TC4H9 => TC4H9O + TC4H9O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2278
reaction('TC4H9O + TC4H9O => TC4H9O2 + TC4H9', [7.373000e+16, -0.978, 32750.0])

# Reaction 2279
reaction('TC4H9O2 + C3H5-A => TC4H9O + C3H5O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2280
reaction('TC4H9O + C3H5O => TC4H9O2 + C3H5-A', [2.898000e+10, 0.411, 15760.0])

# Reaction 2281
reaction('TC4H9O2 + C4H71-3 => TC4H9O + C4H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2282
reaction('TC4H9O + C4H7O => TC4H9O2 + C4H71-3', [1.497000e+13, -0.519, 17810.0])

# Reaction 2283
reaction('TC4H9O2 + IC4H7 => TC4H9O + IC4H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2284
reaction('TC4H9O + IC4H7O => TC4H9O2 + IC4H7', [6.172000e+10, 0.401, 14920.0])

# Reaction 2285
reaction('IC4H9O2 + C2H4 => IC4H9O2H + C2H3', [2.000000e+11, 0.0, 6000.0])

# Reaction 2286
reaction('IC4H9O2H + C2H3 => IC4H9O2 + C2H4', [2.000000e+10, 0.0, 8000.0])

# Reaction 2287
reaction('IC4H9O2 + CH4 => IC4H9O2H + CH3', [1.130000e+13, 0.0, 20460.0])

# Reaction 2288
reaction('IC4H9O2H + CH3 => IC4H9O2 + CH4', [7.500000e+08, 0.0, 1280.0])

# Reaction 2289
reaction('H2 + IC4H9O2 => H + IC4H9O2H', [3.010000e+13, 0.0, 26030.0])

# Reaction 2290
reaction('H + IC4H9O2H => H2 + IC4H9O2', [4.800000e+13, 0.0, 7950.0])

# Reaction 2291
reaction('IC4H9O2 + C2H6 => IC4H9O2H + C2H5', [1.700000e+13, 0.0, 20460.0])

# Reaction 2292
reaction('IC4H9O2H + C2H5 => IC4H9O2 + C2H6', [5.000000e+11, 0.0, 6500.0])

# Reaction 2293
reaction('IC4H9O2 + C3H8 => IC4H9O2H + IC3H7', [2.000000e+12, 0.0, 17000.0])

# Reaction 2294
reaction('IC4H9O2H + IC3H7 => IC4H9O2 + C3H8', [5.000000e+11, 0.0, 6500.0])

# Reaction 2295
reaction('IC4H9O2 + C3H8 => IC4H9O2H + NC3H7', [1.700000e+13, 0.0, 20460.0])

# Reaction 2296
reaction('IC4H9O2H + NC3H7 => IC4H9O2 + C3H8', [5.000000e+11, 0.0, 6500.0])

# Reaction 2297
reaction('IC4H9O2 + CH3OH => IC4H9O2H + CH2OH', [6.300000e+12, 0.0, 19360.0])

# Reaction 2298
reaction('IC4H9O2H + CH2OH => IC4H9O2 + CH3OH', [1.000000e+09, 0.0, 10000.0])

# Reaction 2299
reaction('IC4H9O2 + C2H5OH => IC4H9O2H + PC2H4OH', [6.300000e+12, 0.0, 19360.0])

# Reaction 2300
reaction('IC4H9O2H + PC2H4OH => IC4H9O2 + C2H5OH', [3.061000e+12, 0.0, 22100.0])

# Reaction 2301
reaction('IC4H9O2 + C2H5OH => IC4H9O2H + SC2H4OH', [4.200000e+12, 0.0, 15000.0])

# Reaction 2302
reaction('IC4H9O2H + SC2H4OH => IC4H9O2 + C2H5OH', [2.040000e+12, 0.0, 17740.0])

# Reaction 2303
reaction('IC4H9O2H => IC4H9O + OH', [1.500000e+16, 0.0, 42500.0])

# Reaction 2304
reaction('IC4H9O + OH => IC4H9O2H', [1.233000e+08, 1.712, -2942.0])

# Reaction 2305
reaction('TC4H9O2H => TC4H9O + OH', [5.950000e+15, 0.0, 42540.0])

# Reaction 2306
reaction('TC4H9O + OH => TC4H9O2H', [6.677000e+06, 1.939, -2582.0])

# Reaction 2307
reaction('IC4H9O + HO2 => IC3H7CHO + H2O2', [1.000000e+12, 0.0, 0.0])

# Reaction 2308
reaction('IC3H7CHO + H2O2 => IC4H9O + HO2', [1.380000e+13, -0.25, 72130.0])

# Reaction 2309
reaction('IC4H9O + OH => IC3H7CHO + H2O', [1.810000e+13, 0.0, 0.0])

# Reaction 2310
reaction('IC3H7CHO + H2O => IC4H9O + OH', [4.209000e+13, 0.08, 103600.0])

# Reaction 2311
reaction('IC4H9O + CH3 => IC3H7CHO + CH4', [2.400000e+13, 0.0, 0.0])

# Reaction 2312
reaction('IC3H7CHO + CH4 => IC4H9O + CH3', [3.368000e+14, 0.08, 88920.0])

# Reaction 2313
reaction('IC4H9O + O => IC3H7CHO + OH', [6.000000e+12, 0.0, 0.0])

# Reaction 2314
reaction('IC3H7CHO + OH => IC4H9O + O', [1.415000e+12, 0.08, 86350.0])

# Reaction 2315
reaction('IC4H9O + H => IC3H7CHO + H2', [1.990000e+13, 0.0, 0.0])

# Reaction 2316
reaction('IC3H7CHO + H2 => IC4H9O + H', [1.069000e+13, 0.08, 88440.0])

# Reaction 2317
reaction('IC4H9O => IC3H7CHO + H', [4.000000e+14, 0.0, 21500.0])

# Reaction 2318
reaction('IC3H7CHO + H => IC4H9O', [1.139000e+10, 1.08, 2500.0])

# Reaction 2319
reaction('IC4H9O => CH2O + IC3H7', [2.000000e+14, 0.0, 17500.0])

# Reaction 2320
reaction('CH2O + IC3H7 => IC4H9O', [1.877000e+02, 2.796, 4591.0])

# Reaction 2321
reaction('TC4H9O => CH3COCH3 + CH3', [9.558000e+22, -2.548, 18650.0])

# Reaction 2322
reaction('CH3COCH3 + CH3 => TC4H9O', [1.500000e+11, 0.0, 11900.0])

# Reaction 2323
reaction('IC4H9O + O2 => IC3H7CHO + HO2', [1.930000e+11, 0.0, 1660.0])

# Reaction 2324
reaction('IC3H7CHO + HO2 => IC4H9O + O2', [7.026000e+08, 0.564, 33200.0])

# Reaction 2325
reaction('TC4H9O + O2 => IC4H8O + HO2', [8.100000e+11, 0.0, 4700.0])

# Reaction 2326
reaction('IC4H8O + HO2 => TC4H9O + O2', [1.000000e+11, 0.0, 32000.0])

# Reaction 2327
reaction('IC4H8O => IC3H7CHO', [4.180000e+13, 0.0, 52720.0])

# Reaction 2328
reaction('IC3H7CHO => IC4H8O', [1.392000e+10, 0.55, 72050.0])

# Reaction 2329
reaction('IC4H8O + OH => IC3H6CHO + H2O', [1.250000e+12, 0.0, 0.0])

# Reaction 2330
reaction('IC3H6CHO + H2O => IC4H8O + OH', [9.609000e+05, 1.069, 36560.0])

# Reaction 2331
reaction('IC4H8O + H => IC3H6CHO + H2', [1.250000e+12, 0.0, 0.0])

# Reaction 2332
reaction('IC3H6CHO + H2 => IC4H8O + H', [9.063000e+04, 1.175, 21670.0])

# Reaction 2333
reaction('IC4H8O + HO2 => IC3H6CHO + H2O2', [2.500000e+12, 0.0, 15000.0])

# Reaction 2334
reaction('IC3H6CHO + H2O2 => IC4H8O + HO2', [1.049000e+08, 0.48, 20660.0])

# Reaction 2335
reaction('IC4H8O + CH3O2 => IC3H6CHO + CH3O2H', [2.500000e+12, 0.0, 19000.0])

# Reaction 2336
reaction('IC3H6CHO + CH3O2H => IC4H8O + CH3O2', [2.040000e+09, 0.036, 23070.0])

# Reaction 2337
reaction('IC4H8O + CH3 => IC3H6CHO + CH4', [5.000000e+10, 0.0, 10000.0])

# Reaction 2338
reaction('IC3H6CHO + CH4 => IC4H8O + CH3', [3.307000e+06, 0.729, 33210.0])

# Reaction 2339
reaction('IC4H8O + O => IC3H6CHO + OH', [1.250000e+12, 0.0, 0.0])

# Reaction 2340
reaction('IC3H6CHO + OH => IC4H8O + O', [4.704000e+04, 1.156, 20250.0])

# Reaction 2341
reaction('IC3H7CHO => TC3H6CHO + H', [2.304000e+18, -0.91, 92000.0])

# Reaction 2342
reaction('TC3H6CHO + H => IC3H7CHO', [2.000000e+14, 0.0, 0.0])

# Reaction 2343
reaction('IC3H7CHO => IC3H7 + HCO', [1.129000e+17, -0.03, 79760.0])

# Reaction 2344
reaction('IC3H7 + HCO => IC3H7CHO', [1.810000e+13, 0.0, 0.0])

# Reaction 2345
reaction('IC3H7CHO + HO2 => IC3H7CO + H2O2', [3.000000e+12, 0.0, 11920.0])

# Reaction 2346
reaction('IC3H7CO + H2O2 => IC3H7CHO + HO2', [7.987000e+11, -0.062, 10360.0])

# Reaction 2347
reaction('IC3H7CHO + HO2 => TC3H6CHO + H2O2', [8.000000e+10, 0.0, 11920.0])

# Reaction 2348
reaction('TC3H6CHO + H2O2 => IC3H7CHO + HO2', [3.366000e+12, -0.42, 11050.0])

# Reaction 2349
reaction('IC3H7CHO + CH3 => IC3H7CO + CH4', [3.980000e+12, 0.0, 8700.0])

# Reaction 2350
reaction('IC3H7CO + CH4 => IC3H7CHO + CH3', [2.325000e+13, -0.06, 25630.0])

# Reaction 2351
reaction('IC3H7CHO + O => IC3H7CO + OH', [7.180000e+12, 0.0, 1389.0])

# Reaction 2352
reaction('IC3H7CO + OH => IC3H7CHO + O', [7.052000e+11, -0.06, 15740.0])

# Reaction 2353
reaction('IC3H7CHO + O2 => IC3H7CO + HO2', [4.000000e+13, 0.0, 37600.0])

# Reaction 2354
reaction('IC3H7CO + HO2 => IC3H7CHO + O2', [1.625000e+11, 0.27, -3432.0])

# Reaction 2355
reaction('IC3H7CHO + OH => IC3H7CO + H2O', [2.690000e+10, 0.76, -340.0])

# Reaction 2356
reaction('IC3H7CO + H2O => IC3H7CHO + OH', [1.164000e+10, 0.75, 31200.0])

# Reaction 2357
reaction('IC3H7CHO + OH => TC3H6CHO + H2O', [1.684000e+12, 0.0, -781.0])

# Reaction 2358
reaction('TC3H6CHO + H2O => IC3H7CHO + OH', [1.194000e+13, -0.09, 29810.0])

# Reaction 2359
reaction('IC3H7CHO + H => IC3H7CO + H2', [2.600000e+12, 0.0, 2600.0])

# Reaction 2360
reaction('IC3H7CO + H2 => IC3H7CHO + H', [1.196000e+09, 0.633, 17040.0])

# Reaction 2361
reaction('IC3H7CHO + OH => IC3H6CHO + H2O', [3.120000e+06, 2.0, -298.0])

# Reaction 2362
reaction('IC3H6CHO + H2O => IC3H7CHO + OH', [6.388000e+05, 1.99, 19130.0])

# Reaction 2363
reaction('IC3H7CHO + HO2 => IC3H6CHO + H2O2', [2.740000e+04, 2.55, 15500.0])

# Reaction 2364
reaction('IC3H6CHO + H2O2 => IC3H7CHO + HO2', [3.330000e+04, 2.21, 3468.0])

# Reaction 2365
reaction('IC3H7CHO + CH3O2 => IC3H6CHO + CH3O2H', [4.760000e+04, 2.55, 16490.0])

# Reaction 2366
reaction('IC3H6CHO + CH3O2H => IC3H7CHO + CH3O2', [2.377000e+05, 2.04, 3742.0])

# Reaction 2367
reaction('IC3H7CO => IC3H7 + CO', [2.869000e+20, -2.194, 14970.0])

# Reaction 2368
reaction('IC3H7 + CO => IC3H7CO', [1.500000e+11, 0.0, 4810.0])

# Reaction 2369
reaction('IC3H6CHO => C3H6 + HCO', [1.031000e+15, -0.62, 23170.0])

# Reaction 2370
reaction('C3H6 + HCO => IC3H6CHO', [1.000000e+11, 0.0, 7800.0])

# Reaction 2371
reaction('IC3H6CHO => C2H3CHO + CH3', [2.425000e+13, -0.27, 22470.0])

# Reaction 2372
reaction('C2H3CHO + CH3 => IC3H6CHO', [1.000000e+11, 0.0, 7800.0])

# Reaction 2373
reaction('IC4H8OH => IC4H8 + OH', [9.232000e+14, -0.562, 28050.0])

# Reaction 2374
reaction('IC4H8 + OH => IC4H8OH', [9.930000e+11, 0.0, -960.0])

# Reaction 2375
reaction('IO2C4H8OH => IC4H8OH + O2', [1.917000e+21, -2.347, 35790.0])

# Reaction 2376
reaction('IC4H8OH + O2 => IO2C4H8OH', [1.200000e+11, 0.0, -1100.0])

# Reaction 2377
reaction('IO2C4H8OH => CH3COCH3 + CH2O + OH', [1.250000e+10, 0.0, 18900.0])

# Reaction 2378
reaction('CH3COCH3 + CH2O + OH => IO2C4H8OH', [0.000000e+00, 0.0, 0.0])

# Reaction 2379
reaction('IC4H9O2 => IC4H8O2H-I', [7.500000e+10, 0.0, 24400.0])

# Reaction 2380
reaction('IC4H8O2H-I => IC4H9O2', [1.815000e+11, -0.507, 8946.0])

# Reaction 2381
reaction('TC4H9O2 => TC4H8O2H-I', [9.000000e+11, 0.0, 34500.0])

# Reaction 2382
reaction('TC4H8O2H-I => TC4H9O2', [2.027000e+09, 0.123, 16940.0])

# Reaction 2383
reaction('IC4H9O2 => IC4H8O2H-T', [1.000000e+11, 0.0, 29200.0])

# Reaction 2384
reaction('IC4H8O2H-T => IC4H9O2', [5.085000e+06, 0.78, 15880.0])

# Reaction 2385
reaction('IC4H9O2 => IC4H8 + HO2', [2.265000e+35, -7.22, 39490.0])

# Reaction 2386
reaction('IC4H8 + HO2 => IC4H9O2', [2.996000e+26, -5.331, 21240.0])

# Reaction 2387
reaction('TC4H9O2 => IC4H8 + HO2', [7.612000e+42, -9.41, 41490.0])

# Reaction 2388
reaction('IC4H8 + HO2 => TC4H9O2', [6.344000e+31, -7.203, 17160.0])

# Reaction 2389
reaction('IC4H8OOH-IO2 => IC4H8O2H-I + O2', [1.440000e+20, -1.627, 35690.0])

# Reaction 2390
reaction('IC4H8O2H-I + O2 => IC4H8OOH-IO2', [2.260000e+12, 0.0, 0.0])

# Reaction 2391
reaction('TC4H8OOH-IO2 => TC4H8O2H-I + O2', [5.167000e+22, -2.257, 37800.0])

# Reaction 2392
reaction('TC4H8O2H-I + O2 => TC4H8OOH-IO2', [2.260000e+12, 0.0, 0.0])

# Reaction 2393
reaction('IC4H8OOH-TO2 => IC4H8O2H-T + O2', [2.266000e+27, -3.233, 39640.0])

# Reaction 2394
reaction('IC4H8O2H-T + O2 => IC4H8OOH-TO2', [1.410000e+13, 0.0, 0.0])

# Reaction 2395
reaction('IC4H8OOH-IO2 => IC4KETII + OH', [5.000000e+10, 0.0, 21400.0])

# Reaction 2396
reaction('IC4KETII + OH => IC4H8OOH-IO2', [1.986000e+03, 1.455, 44420.0])

# Reaction 2397
reaction('IC4H8OOH-TO2 => IC4KETIT + OH', [4.000000e+11, 0.0, 31500.0])

# Reaction 2398
reaction('IC4KETIT + OH => IC4H8OOH-TO2', [1.910000e+05, 1.24, 53830.0])

# Reaction 2399
reaction('TC4H8OOH-IO2 => TIC4H7Q2-I', [7.500000e+10, 0.0, 24400.0])

# Reaction 2400
reaction('TIC4H7Q2-I => TC4H8OOH-IO2', [6.360000e+08, 0.091, 11720.0])

# Reaction 2401
reaction('TIC4H7Q2-I => IC4H7OOH + HO2', [1.940000e+20, -2.19, 22590.0])

# Reaction 2402
reaction('IC4H7OOH + HO2 => TIC4H7Q2-I', [1.000000e+11, 0.0, 10600.0])

# Reaction 2403
reaction('IC4H8OOH-IO2 => IIC4H7Q2-I', [3.750000e+10, 0.0, 24400.0])

# Reaction 2404
reaction('IIC4H7Q2-I => IC4H8OOH-IO2', [8.986000e+08, -0.088, 11900.0])

# Reaction 2405
reaction('IC4H8OOH-IO2 => IIC4H7Q2-T', [1.000000e+11, 0.0, 29200.0])

# Reaction 2406
reaction('IIC4H7Q2-T => IC4H8OOH-IO2', [9.448000e+07, 0.406, 21200.0])

# Reaction 2407
reaction('IC4H8OOH-TO2 => TIC4H7Q2-I', [6.000000e+11, 0.0, 34500.0])

# Reaction 2408
reaction('TIC4H7Q2-I => IC4H8OOH-TO2', [5.088000e+09, 0.091, 21820.0])

# Reaction 2409
reaction('IIC4H7Q2-I => AC3H5OOH + CH2O2H', [7.631000e+19, -1.74, 38310.0])

# Reaction 2410
reaction('AC3H5OOH + CH2O2H => IIC4H7Q2-I', [8.500000e+10, 0.0, 10600.0])

# Reaction 2411
reaction('IIC4H7Q2-T => IC4H7OOH + HO2', [4.076000e+17, -1.56, 18390.0])

# Reaction 2412
reaction('IC4H7OOH + HO2 => IIC4H7Q2-T', [1.000000e+11, 0.0, 7800.0])

# Reaction 2413
reaction('CH2O2H => CH2O + OH', [9.000000e+14, 0.0, 1500.0])

# Reaction 2414
reaction('CH2O + OH => CH2O2H', [1.869000e+11, 1.1, 36840.0])

# Reaction 2415
reaction('IC4KETII => CH2O + C2H5CO + OH', [1.500000e+16, 0.0, 42000.0])

# Reaction 2416
reaction('CH2O + C2H5CO + OH => IC4KETII', [0.000000e+00, 0.0, 0.0])

# Reaction 2417
reaction('IC4KETIT => CH3COCH3 + HCO + OH', [9.500000e+15, 0.0, 42540.0])

# Reaction 2418
reaction('CH3COCH3 + HCO + OH => IC4KETIT', [0.000000e+00, 0.0, 0.0])

# Reaction 2419
reaction('TC4H8O2H-I => IC4H8 + HO2', [1.073000e+20, -2.085, 19390.0])

# Reaction 2420
reaction('IC4H8 + HO2 => TC4H8O2H-I', [3.970000e+11, 0.0, 12620.0])

# Reaction 2421
reaction('IC4H8O2H-T => IC4H8 + HO2', [1.526000e+16, -1.109, 17560.0])

# Reaction 2422
reaction('IC4H8 + HO2 => IC4H8O2H-T', [3.970000e+11, 0.0, 12620.0])

# Reaction 2423
reaction('IC4H8O2H-I => CC4H8O + OH', [2.000000e+11, 0.0, 19500.0])

# Reaction 2424
reaction('CC4H8O + OH => IC4H8O2H-I', [0.000000e+00, 0.0, 0.0])

# Reaction 2425
reaction('IC4H8O2H-T => IC4H8O + OH', [1.380000e+12, 0.0, 14800.0])

# Reaction 2426
reaction('IC4H8O + OH => IC4H8O2H-T', [0.000000e+00, 0.0, 0.0])

# Reaction 2427
reaction('TC4H8O2H-I => IC4H8O + OH', [4.000000e+12, 0.0, 17000.0])

# Reaction 2428
reaction('IC4H8O + OH => TC4H8O2H-I', [0.000000e+00, 0.0, 0.0])

# Reaction 2429
reaction('IC4H8O2H-I => OH + CH2O + C3H6', [8.451000e+15, -0.68, 29170.0])

# Reaction 2430
reaction('OH + CH2O + C3H6 => IC4H8O2H-I', [0.000000e+00, 0.0, 0.0])

# Reaction 2431
reaction('IC4H8 => C3H5-T + CH3', [1.920000e+66, -14.22, 128100.0])

# Reaction 2432
reaction('C3H5-T + CH3 => IC4H8', [1.561000e+56, -12.293, 26100.0])

# Reaction 2433
reaction('IC4H8 => IC4H7 + H', [3.070000e+55, -11.49, 114300.0])

# Reaction 2434
reaction('IC4H7 + H => IC4H8', [1.428000e+55, -11.738, 26400.0])

# Reaction 2435
reaction('IC4H8 + H => C3H6 + CH3', [5.680000e+33, -5.72, 20000.0])

# Reaction 2436
reaction('C3H6 + CH3 => IC4H8 + H', [6.093000e+26, -4.209, 27200.0])

# Reaction 2437
reaction('IC4H8 + H => IC4H7 + H2', [3.400000e+05, 2.5, 2492.0])

# Reaction 2438
reaction('IC4H7 + H2 => IC4H8 + H', [6.320000e+04, 2.528, 18160.0])

# Reaction 2439
reaction('IC4H8 + O => CH2CO + CH3 + CH3', [3.330000e+07, 1.76, 76.0])

# Reaction 2440
reaction('CH2CO + CH3 + CH3 => IC4H8 + O', [0.000000e+00, 0.0, 0.0])

# Reaction 2441
reaction('IC4H8 + O => IC3H6CO + H + H', [1.660000e+07, 1.76, 76.0])

# Reaction 2442
reaction('IC3H6CO + H + H => IC4H8 + O', [0.000000e+00, 0.0, 0.0])

# Reaction 2443
reaction('IC4H8 + O => IC4H7 + OH', [1.206000e+11, 0.7, 7633.0])

# Reaction 2444
reaction('IC4H7 + OH => IC4H8 + O', [1.164000e+10, 0.709, 21890.0])

# Reaction 2445
reaction('IC4H8 + CH3 => IC4H7 + CH4', [4.420000e+00, 3.5, 5675.0])

# Reaction 2446
reaction('IC4H7 + CH4 => IC4H8 + CH3', [7.495000e+02, 3.082, 22890.0])

# Reaction 2447
reaction('IC4H8 + HO2 => IC4H7 + H2O2', [1.928000e+04, 2.6, 13910.0])

# Reaction 2448
reaction('IC4H7 + H2O2 => IC4H8 + HO2', [2.073000e+06, 1.933, 13580.0])

# Reaction 2449
reaction('IC4H8 + O2CHO => IC4H7 + HO2CHO', [1.928000e+04, 2.6, 13910.0])

# Reaction 2450
reaction('IC4H7 + HO2CHO => IC4H8 + O2CHO', [6.514000e-07, 4.9, -3468.0])

# Reaction 2451
reaction('IC4H8 + O2 => IC4H7 + HO2', [6.000000e+12, 0.0, 39900.0])

# Reaction 2452
reaction('IC4H7 + HO2 => IC4H8 + O2', [5.848000e+12, -0.32, 883.0])

# Reaction 2453
reaction('IC4H8 + C3H5-A => IC4H7 + C3H6', [7.940000e+11, 0.0, 20500.0])

# Reaction 2454
reaction('IC4H7 + C3H6 => IC4H8 + C3H5-A', [4.400000e+20, -1.33, 60610.0])

# Reaction 2455
reaction('IC4H8 + C3H5-S => IC4H7 + C3H6', [7.940000e+11, 0.0, 20500.0])

# Reaction 2456
reaction('IC4H7 + C3H6 => IC4H8 + C3H5-S', [5.592000e+20, -1.27, 82170.0])

# Reaction 2457
reaction('IC4H8 + C3H5-T => IC4H7 + C3H6', [7.940000e+11, 0.0, 20500.0])

# Reaction 2458
reaction('IC4H7 + C3H6 => IC4H8 + C3H5-T', [5.592000e+20, -1.27, 80170.0])

# Reaction 2459
reaction('IC4H8 + OH => IC4H7 + H2O', [5.200000e+06, 2.0, -298.0])

# Reaction 2460
reaction('IC4H7 + H2O => IC4H8 + OH', [1.025000e+07, 1.922, 30270.0])

# Reaction 2461
reaction('IC4H8 + O => IC3H7 + HCO', [1.580000e+07, 1.76, -1216.0])

# Reaction 2462
reaction('IC3H7 + HCO => IC4H8 + O', [4.538000e+00, 3.06, 21690.0])

# Reaction 2463
reaction('IC4H8 + CH3O2 => IC4H7 + CH3O2H', [1.928000e+04, 2.6, 13910.0])

# Reaction 2464
reaction('IC4H7 + CH3O2H => IC4H8 + CH3O2', [4.034000e+07, 1.488, 11990.0])

# Reaction 2465
reaction('IC4H8 + HO2 => IC4H8O + OH', [1.290000e+12, 0.0, 13340.0])

# Reaction 2466
reaction('IC4H8O + OH => IC4H8 + HO2', [1.000000e+12, 0.0, 7500.0])

# Reaction 2467
reaction('IC4H7 + O2 => IC3H5CHO + OH', [2.470000e+13, -0.45, 23020.0])

# Reaction 2468
reaction('IC3H5CHO + OH => IC4H7 + O2', [3.372000e+13, -0.577, 73010.0])

# Reaction 2469
reaction('IC4H7 + O2 => CH3COCH2 + CH2O', [7.140000e+15, -1.21, 21050.0])

# Reaction 2470
reaction('CH3COCH2 + CH2O => IC4H7 + O2', [1.700000e+12, -0.407, 88250.0])

# Reaction 2471
reaction('IC4H7 + O2 => C3H4-A + CH2O + OH', [7.290000e+29, -5.71, 21450.0])

# Reaction 2472
reaction('C3H4-A + CH2O + OH => IC4H7 + O2', [0.000000e+00, 0.0, 0.0])

# Reaction 2473
reaction('IC4H7 + O => IC3H5CHO + H', [6.030000e+13, 0.0, 0.0])

# Reaction 2474
reaction('IC3H5CHO + H => IC4H7 + O', [2.844000e+16, -0.519, 66730.0])

# Reaction 2475
reaction('IC4H7 => C3H4-A + CH3', [1.230000e+47, -9.74, 74260.0])

# Reaction 2476
reaction('C3H4-A + CH3 => IC4H7', [1.649000e+38, -7.768, 22540.0])

# Reaction 2477
reaction('CH3O2 + IC4H7 => CH3O + IC4H7O', [7.000000e+12, 0.0, -1000.0])

# Reaction 2478
reaction('CH3O + IC4H7O => CH3O2 + IC4H7', [2.138000e+11, 0.349, 15060.0])

# Reaction 2479
reaction('IC4H7 + HO2 => IC4H7O + OH', [7.000000e+12, 0.0, -1000.0])

# Reaction 2480
reaction('IC4H7O + OH => IC4H7 + HO2', [3.418000e+12, 0.05, 10820.0])

# Reaction 2481
reaction('IC4H7O => C3H5-T + CH2O', [2.925000e+21, -2.391, 35590.0])

# Reaction 2482
reaction('C3H5-T + CH2O => IC4H7O', [1.000000e+11, 0.0, 12600.0])

# Reaction 2483
reaction('IC4H7O => IC4H6OH', [1.391000e+11, 0.0, 15600.0])

# Reaction 2484
reaction('IC4H6OH => IC4H7O', [4.233000e+11, -0.164, 31670.0])

# Reaction 2485
reaction('IC4H7O => IC3H5CHO + H', [5.000000e+13, 0.0, 29100.0])

# Reaction 2486
reaction('IC3H5CHO + H => IC4H7O', [6.670000e+13, -0.105, 18410.0])

# Reaction 2487
reaction('IC4H6OH + H2 => IC4H7OH + H', [2.160000e+04, 2.38, 18990.0])

# Reaction 2488
reaction('IC4H7OH + H => IC4H6OH + H2', [5.614000e+02, 2.98, 1399.0])

# Reaction 2489
reaction('IC4H6OH + HO2 => IC4H7OH + O2', [5.570000e+13, -0.315, 862.0])

# Reaction 2490
reaction('IC4H7OH + O2 => IC4H6OH + HO2', [6.000000e+13, 0.0, 39900.0])

# Reaction 2491
reaction('IC4H6OH + CH2O => IC4H7OH + HCO', [6.300000e+08, 1.9, 18190.0])

# Reaction 2492
reaction('IC4H7OH + HCO => IC4H6OH + CH2O', [2.101000e+07, 2.153, 17730.0])

# Reaction 2493
reaction('IC4H6OH + IC4H8 => IC4H7OH + IC4H7', [4.700000e+02, 3.3, 19840.0])

# Reaction 2494
reaction('IC4H7OH + IC4H7 => IC4H6OH + IC4H8', [2.814000e-01, 3.9, 6521.0])

# Reaction 2495
reaction('IC4H7OH => IC4H6OH + H', [4.902000e+16, -0.4, 89850.0])

# Reaction 2496
reaction('IC4H6OH + H => IC4H7OH', [1.000000e+14, 0.0, 0.0])

# Reaction 2497
reaction('IC4H7OH + HO2 => IC4H6OH + H2O2', [7.644000e+03, 2.712, 13930.0])

# Reaction 2498
reaction('IC4H6OH + H2O2 => IC4H7OH + HO2', [7.830000e+05, 2.05, 13580.0])

# Reaction 2499
reaction('IC4H6OH => C3H4-A + CH2OH', [7.244000e+19, -1.859, 57050.0])

# Reaction 2500
reaction('C3H4-A + CH2OH => IC4H6OH', [1.000000e+11, 0.0, 9200.0])

# Reaction 2501
reaction('IC4H7O + O2 => IC3H5CHO + HO2', [3.000000e+10, 0.0, 1649.0])

# Reaction 2502
reaction('IC3H5CHO + HO2 => IC4H7O + O2', [6.312000e+10, -0.14, 38980.0])

# Reaction 2503
reaction('IC4H7O + HO2 => IC3H5CHO + H2O2', [3.000000e+11, 0.0, 0.0])

# Reaction 2504
reaction('IC3H5CHO + H2O2 => IC4H7O + HO2', [8.930000e+14, -0.8, 78500.0])

# Reaction 2505
reaction('IC4H7O + CH3 => IC3H5CHO + CH4', [2.400000e+13, 0.0, 0.0])

# Reaction 2506
reaction('IC3H5CHO + CH4 => IC4H7O + CH3', [7.261000e+16, -0.47, 95290.0])

# Reaction 2507
reaction('IC4H7O + O => IC3H5CHO + OH', [6.000000e+12, 0.0, 0.0])

# Reaction 2508
reaction('IC3H5CHO + OH => IC4H7O + O', [3.052000e+14, -0.47, 92720.0])

# Reaction 2509
reaction('IC4H7O + OH => IC3H5CHO + H2O', [1.810000e+13, 0.0, 0.0])

# Reaction 2510
reaction('IC3H5CHO + H2O => IC4H7O + OH', [9.076000e+15, -0.47, 110000.0])

# Reaction 2511
reaction('IC4H7O + H => IC3H5CHO + H2', [1.990000e+13, 0.0, 0.0])

# Reaction 2512
reaction('IC3H5CHO + H2 => IC4H7O + H', [2.305000e+15, -0.47, 94810.0])

# Reaction 2513
reaction('IC3H5CHO + OH => IC3H5CO + H2O', [2.690000e+10, 0.76, -340.0])

# Reaction 2514
reaction('IC3H5CO + H2O => IC3H5CHO + OH', [4.400000e+10, 0.78, 36080.0])

# Reaction 2515
reaction('IC3H5CHO + HO2 => IC3H5CO + H2O2', [1.000000e+12, 0.0, 11920.0])

# Reaction 2516
reaction('IC3H5CO + H2O2 => IC3H5CHO + HO2', [9.709000e+12, -0.31, 16880.0])

# Reaction 2517
reaction('IC3H5CHO + CH3 => IC3H5CO + CH4', [3.980000e+12, 0.0, 8700.0])

# Reaction 2518
reaction('IC3H5CO + CH4 => IC3H5CHO + CH3', [3.928000e+13, 0.02, 30450.0])

# Reaction 2519
reaction('IC3H5CHO + O => IC3H5CO + OH', [7.180000e+12, 0.0, 1389.0])

# Reaction 2520
reaction('IC3H5CO + OH => IC3H5CHO + O', [1.191000e+12, 0.02, 20560.0])

# Reaction 2521
reaction('IC3H5CHO + O2 => IC3H5CO + HO2', [2.000000e+13, 0.0, 40700.0])

# Reaction 2522
reaction('IC3H5CO + HO2 => IC3H5CHO + O2', [1.824000e+11, 0.311, 5337.0])

# Reaction 2523
reaction('IC3H5CHO + H => IC3H5CO + H2', [2.600000e+12, 0.0, 2600.0])

# Reaction 2524
reaction('IC3H5CO + H2 => IC3H5CHO + H', [9.822000e+11, 0.02, 23870.0])

# Reaction 2525
reaction('IC3H5CO => C3H5-T + CO', [1.278000e+20, -1.89, 34460.0])

# Reaction 2526
reaction('C3H5-T + CO => IC3H5CO', [1.510000e+11, 0.0, 4809.0])

# Reaction 2527
reaction('TC3H6CHO + HO2 => TC3H6OCHO + OH', [9.640000e+12, 0.0, 0.0])

# Reaction 2528
reaction('TC3H6OCHO + OH => TC3H6CHO + HO2', [2.018000e+17, -1.2, 21010.0])

# Reaction 2529
reaction('TC3H6OCHO => CH3COCH3 + HCO', [3.980000e+13, 0.0, 9700.0])

# Reaction 2530
reaction('CH3COCH3 + HCO => TC3H6OCHO', [2.173000e+08, 0.8, 14240.0])

# Reaction 2531
reaction('TC3H6CHO => IC3H5CHO + H', [1.325000e+14, 0.01, 39340.0])

# Reaction 2532
reaction('IC3H5CHO + H => TC3H6CHO', [1.300000e+13, 0.0, 1200.0])

# Reaction 2533
reaction('TC3H6CHO => IC3H6CO + H', [4.086000e+14, -0.072, 42410.0])

# Reaction 2534
reaction('IC3H6CO + H => TC3H6CHO', [1.300000e+13, 0.0, 4800.0])

# Reaction 2535
reaction('TC3H6CHO + H2 => IC3H7CHO + H', [2.160000e+05, 2.38, 18990.0])

# Reaction 2536
reaction('IC3H7CHO + H => TC3H6CHO + H2', [1.319000e+05, 2.47, 3550.0])

# Reaction 2537
reaction('IC4H7OOH => IC4H7O + OH', [6.400000e+15, 0.0, 45550.0])

# Reaction 2538
reaction('IC4H7O + OH => IC4H7OOH', [1.000000e+11, 0.0, 0.0])

# Reaction 2539
reaction('IC4H7OH => IC4H7O + H', [5.969000e+16, -0.56, 105900.0])

# Reaction 2540
reaction('IC4H7O + H => IC4H7OH', [4.000000e+13, 0.0, 0.0])

# Reaction 2541
reaction('IC4H8OH => IC4H7OH + H', [1.708000e+12, 0.277, 38850.0])

# Reaction 2542
reaction('IC4H7OH + H => IC4H8OH', [1.000000e+13, 0.0, 1200.0])

# Reaction 2543
reaction('IC4H7O + H2 => IC4H7OH + H', [9.050000e+06, 2.0, 17830.0])

# Reaction 2544
reaction('IC4H7OH + H => IC4H7O + H2', [7.160000e+05, 2.44, 16310.0])

# Reaction 2545
reaction('IC4H7OH => IC4H7 + OH', [7.310000e+16, -0.41, 79700.0])

# Reaction 2546
reaction('IC4H7 + OH => IC4H7OH', [3.000000e+13, 0.0, 0.0])

# Reaction 2547
reaction('IC4H7O + CH2O => IC4H7OH + HCO', [1.150000e+11, 0.0, 1280.0])

# Reaction 2548
reaction('IC4H7OH + HCO => IC4H7O + CH2O', [3.020000e+11, 0.0, 18160.0])

# Reaction 2549
reaction('TC3H6CHO + CH2O => IC3H7CHO + HCO', [2.520000e+08, 1.9, 18190.0])

# Reaction 2550
reaction('IC3H7CHO + HCO => TC3H6CHO + CH2O', [1.229000e+07, 1.99, 17420.0])

# Reaction 2551
reaction('TC3H6CHO + IC4H8 => IC3H7CHO + IC4H7', [4.700000e+02, 3.3, 19840.0])

# Reaction 2552
reaction('IC3H7CHO + IC4H7 => TC3H6CHO + IC4H8', [6.613000e+00, 3.39, 8672.0])

# Reaction 2553
reaction('IC3H6CO + OH => IC3H7 + CO2', [1.730000e+12, 0.0, -1010.0])

# Reaction 2554
reaction('IC3H7 + CO2 => IC3H6CO + OH', [2.577000e+14, -0.43, 55480.0])

# Reaction 2555
reaction('TC3H6OHCHO => TC3H6CHO + OH', [9.990000e+20, -1.46, 87480.0])

# Reaction 2556
reaction('TC3H6CHO + OH => TC3H6OHCHO', [5.000000e+13, 0.0, 0.0])

# Reaction 2557
reaction('TC3H6OHCHO => TC3H6OH + HCO', [5.164000e+23, -1.9, 76850.0])

# Reaction 2558
reaction('TC3H6OH + HCO => TC3H6OHCHO', [1.810000e+13, 0.0, 0.0])

# Reaction 2559
reaction('TC3H6OH => CH3COCH3 + H', [5.000000e+13, 0.0, 21860.0])

# Reaction 2560
reaction('CH3COCH3 + H => TC3H6OH', [1.000000e+12, 0.0, 0.0])

# Reaction 2561
reaction('TC3H6OH => IC3H5OH + H', [6.196000e+15, -0.66, 40340.0])

# Reaction 2562
reaction('IC3H5OH + H => TC3H6OH', [1.300000e+13, 0.0, 1560.0])

# Reaction 2563
reaction('IC3H5OH => C3H5-T + OH', [7.369000e+19, -0.94, 109100.0])

# Reaction 2564
reaction('C3H5-T + OH => IC3H5OH', [5.000000e+13, 0.0, 0.0])

# Reaction 2565
reaction('TC3H6O2CHO => TC3H6CHO + O2', [2.458000e+25, -4.065, 27080.0])

# Reaction 2566
reaction('TC3H6CHO + O2 => TC3H6O2CHO', [1.990000e+17, -2.1, 0.0])

# Reaction 2567
reaction('TC3H6O2CHO => IC3H5O2HCHO', [6.000000e+11, 0.0, 29880.0])

# Reaction 2568
reaction('IC3H5O2HCHO => TC3H6O2CHO', [3.014000e+12, -0.516, 17110.0])

# Reaction 2569
reaction('TC3H6O2CHO => TC3H6O2HCO', [1.000000e+11, 0.0, 25750.0])

# Reaction 2570
reaction('TC3H6O2HCO => TC3H6O2CHO', [1.628000e+12, -0.516, 25100.0])

# Reaction 2571
reaction('IC3H5O2HCHO => IC3H5CHO + HO2', [8.944000e+20, -2.44, 15030.0])

# Reaction 2572
reaction('IC3H5CHO + HO2 => IC3H5O2HCHO', [2.230000e+11, 0.0, 10600.0])

# Reaction 2573
reaction('TC3H6O2HCO => CH3COCH3 + CO + OH', [4.244000e+18, -1.43, 4800.0])

# Reaction 2574
reaction('CH3COCH3 + CO + OH => TC3H6O2HCO', [0.000000e+00, 0.0, 0.0])

# Reaction 2575
reaction('TC3H6OH + O2 => CH3COCH3 + HO2', [2.230000e+13, 0.0, 0.0])

# Reaction 2576
reaction('CH3COCH3 + HO2 => TC3H6OH + O2', [6.529000e+17, -1.19, 25610.0])

# Reaction 2577
reaction('IC3H6CO + OH => TC3H6OH + CO', [2.000000e+12, 0.0, -1010.0])

# Reaction 2578
reaction('TC3H6OH + CO => IC3H6CO + OH', [1.009000e+09, 0.9, 31540.0])

# Reaction 2579
reaction('TC3H6CHO + O2 => IC3H5CHO + HO2', [2.725000e-19, 0.0, 7240.0])

# Reaction 2580
reaction('IC3H5CHO + HO2 => TC3H6CHO + O2', [1.390000e+11, -0.2, 17310.0])

# Reaction 2581
reaction('TC3H6CHO + O2 => CH3COCH3 + CO + OH', [3.620000e-20, 0.0, 0.0])

# Reaction 2582
reaction('CH3COCH3 + CO + OH => TC3H6CHO + O2', [0.000000e+00, 0.0, 0.0])

# Reaction 2583
reaction('TC3H6CHO + HO2 => IC3H7CHO + O2', [3.675000e+12, 0.0, 1310.0])

# Reaction 2584
reaction('IC3H7CHO + O2 => TC3H6CHO + HO2', [1.236000e+14, -0.24, 43350.0])

# Reaction 2585
reaction('TC3H6CHO + CH3 => IC3H5CHO + CH4', [3.010000e+12, -0.32, -131.0])

# Reaction 2586
reaction('IC3H5CHO + CH4 => TC3H6CHO + CH3', [2.207000e+15, -0.85, 67900.0])

# Reaction 2587
reaction('TC4H8CHO => IC3H5CHO + CH3', [1.000000e+13, 0.0, 26290.0])

# Reaction 2588
reaction('IC3H5CHO + CH3 => TC4H8CHO', [2.230000e+11, 0.0, 10600.0])

# Reaction 2589
reaction('TC4H8CHO => IC4H8 + HCO', [8.520000e+12, 0.0, 20090.0])

# Reaction 2590
reaction('IC4H8 + HCO => TC4H8CHO', [1.000000e+11, 0.0, 6000.0])

# Reaction 2591
reaction('O2C4H8CHO => TC4H8CHO + O2', [1.519000e+19, -1.44, 34510.0])

# Reaction 2592
reaction('TC4H8CHO + O2 => O2C4H8CHO', [2.000000e+12, 0.0, 0.0])

# Reaction 2593
reaction('O2C4H8CHO => O2HC4H8CO', [2.160000e+11, 0.0, 15360.0])

# Reaction 2594
reaction('O2HC4H8CO => O2C4H8CHO', [1.173000e+13, -0.68, 14880.0])

# Reaction 2595
reaction('O2HC4H8CO => IC4H8O2H-T + CO', [3.299000e+22, -2.72, 11760.0])

# Reaction 2596
reaction('IC4H8O2H-T + CO => O2HC4H8CO', [1.500000e+11, 0.0, 4809.0])

# Reaction 2597
reaction('IC4H7O + IC4H8 => IC4H7OH + IC4H7', [2.700000e+11, 0.0, 4000.0])

# Reaction 2598
reaction('IC4H7OH + IC4H7 => IC4H7O + IC4H8', [1.000000e+10, 0.0, 9000.0])

# Reaction 2599
reaction('IC4H6OH + HO2 => CH2CCH2OH + CH2O + OH', [1.446000e+13, 0.0, 0.0])

# Reaction 2600
reaction('CH2CCH2OH + CH2O + OH => IC4H6OH + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 2601
reaction('IC4H8 + CH2CCH2OH => IC4H7 + C3H5OH', [7.940000e+11, 0.0, 20500.0])

# Reaction 2602
reaction('IC4H7 + C3H5OH => IC4H8 + CH2CCH2OH', [2.750000e+11, -0.05, 28470.0])

# Reaction 2603
reaction('C3H5OH + HO2 => CH2CCH2OH + H2O2', [1.761000e+09, 0.28, 22590.0])

# Reaction 2604
reaction('CH2CCH2OH + H2O2 => C3H5OH + HO2', [3.010000e+09, 0.0, 2583.0])

# Reaction 2605
reaction('C3H5OH + OH => CH2CCH2OH + H2O', [5.060000e+12, 0.0, 5960.0])

# Reaction 2606
reaction('CH2CCH2OH + H2O => C3H5OH + OH', [1.457000e+12, 0.05, 17410.0])

# Reaction 2607
reaction('C3H5OH + H => CH2CCH2OH + H2', [3.900000e+05, 2.5, 5821.0])

# Reaction 2608
reaction('CH2CCH2OH + H2 => C3H5OH + H', [2.594000e+04, 2.55, 2121.0])

# Reaction 2609
reaction('C3H5OH + O2 => CH2CCH2OH + HO2', [4.000000e+13, 0.0, 60690.0])

# Reaction 2610
reaction('CH2CCH2OH + HO2 => C3H5OH + O2', [4.833000e+10, 0.38, -492.0])

# Reaction 2611
reaction('C3H5OH + CH3 => CH2CCH2OH + CH4', [2.400000e+11, 0.0, 8030.0])

# Reaction 2612
reaction('CH2CCH2OH + CH4 => C3H5OH + CH3', [4.170000e+11, 0.05, 4810.0])

# Reaction 2613
reaction('IC4H7OH => CH2CCH2OH + CH3', [1.247000e+20, -0.98, 98570.0])

# Reaction 2614
reaction('CH2CCH2OH + CH3 => IC4H7OH', [3.000000e+13, 0.0, 0.0])

# Reaction 2615
reaction('C3H5OH => CH2CCH2OH + H', [2.835000e+19, -1.05, 111100.0])

# Reaction 2616
reaction('CH2CCH2OH + H => C3H5OH', [1.000000e+14, 0.0, 0.0])

# Reaction 2617
reaction('CH2CCH2OH + O2 => CH2OH + CO + CH2O', [4.335000e+12, 0.0, 0.0])

# Reaction 2618
reaction('CH2OH + CO + CH2O => CH2CCH2OH + O2', [0.000000e+00, 0.0, 0.0])

# Reaction 2619
reaction('CH2CCH2OH => C2H2 + CH2OH', [2.163000e+40, -8.31, 45110.0])

# Reaction 2620
reaction('C2H2 + CH2OH => CH2CCH2OH', [1.610000e+40, -8.58, 20330.0])

# Reaction 2621
reaction('CH2CCH2OH => C3H4-A + OH', [6.697000e+16, -1.11, 42580.0])

# Reaction 2622
reaction('C3H4-A + OH => CH2CCH2OH', [8.500000e+12, 0.0, 2000.0])

# Reaction 2623
reaction('NC5H12 => C5H11-1 + H', [1.375000e+17, -0.36, 101200.0])

# Reaction 2624
reaction('C5H11-1 + H => NC5H12', [3.610000e+13, 0.0, 0.0])

# Reaction 2625
reaction('NC5H12 => C5H11-2 + H', [3.325000e+18, -0.763, 98800.0])

# Reaction 2626
reaction('C5H11-2 + H => NC5H12', [3.610000e+13, 0.0, 0.0])

# Reaction 2627
reaction('NC5H12 => C5H11-3 + H', [1.602000e+18, -0.758, 98790.0])

# Reaction 2628
reaction('C5H11-3 + H => NC5H12', [3.610000e+13, 0.0, 0.0])

# Reaction 2629
reaction('NC5H12 => CH3 + PC4H9', [6.101000e+22, -1.862, 89430.0])

# Reaction 2630
reaction('CH3 + PC4H9 => NC5H12', [4.000000e+12, 0.0, -596.0])

# Reaction 2631
reaction('NC5H12 => NC3H7 + C2H5', [3.082000e+24, -2.269, 88440.0])

# Reaction 2632
reaction('NC3H7 + C2H5 => NC5H12', [4.000000e+12, 0.0, -596.0])

# Reaction 2633
reaction('NC5H12 + H => C5H11-1 + H2', [1.880000e+05, 2.75, 6280.0])

# Reaction 2634
reaction('C5H11-1 + H2 => NC5H12 + H', [1.972000e+01, 3.386, 8610.0])

# Reaction 2635
reaction('NC5H12 + H => C5H11-2 + H2', [2.600000e+06, 2.4, 4471.0])

# Reaction 2636
reaction('C5H11-2 + H2 => NC5H12 + H', [1.128000e+01, 3.439, 9236.0])

# Reaction 2637
reaction('NC5H12 + H => C5H11-3 + H2', [1.300000e+06, 2.4, 4471.0])

# Reaction 2638
reaction('C5H11-3 + H2 => NC5H12 + H', [1.171000e+01, 3.434, 9244.0])

# Reaction 2639
reaction('NC5H12 + OH => C5H11-1 + H2O', [1.054000e+10, 0.97, 1590.0])

# Reaction 2640
reaction('C5H11-1 + H2O => NC5H12 + OH', [1.172000e+07, 1.5, 18810.0])

# Reaction 2641
reaction('NC5H12 + OH => C5H11-2 + H2O', [9.340000e+07, 1.61, -35.0])

# Reaction 2642
reaction('C5H11-2 + H2O => NC5H12 + OH', [4.297000e+03, 2.543, 19620.0])

# Reaction 2643
reaction('NC5H12 + OH => C5H11-3 + H2O', [4.670000e+07, 1.61, -35.0])

# Reaction 2644
reaction('C5H11-3 + H2O => NC5H12 + OH', [4.459000e+03, 2.538, 19630.0])

# Reaction 2645
reaction('NC5H12 + O => C5H11-1 + OH', [2.092000e+06, 2.424, 4766.0])

# Reaction 2646
reaction('C5H11-1 + OH => NC5H12 + O', [1.139000e+02, 3.041, 5684.0])

# Reaction 2647
reaction('NC5H12 + O => C5H11-2 + OH', [1.189000e+06, 2.439, 2846.0])

# Reaction 2648
reaction('C5H11-2 + OH => NC5H12 + O', [2.678000e+00, 3.459, 6199.0])

# Reaction 2649
reaction('NC5H12 + O => C5H11-3 + OH', [5.946000e+05, 2.439, 2846.0])

# Reaction 2650
reaction('C5H11-3 + OH => NC5H12 + O', [2.779000e+00, 3.454, 6207.0])

# Reaction 2651
reaction('NC5H12 + CH3 => C5H11-1 + CH4', [9.040000e-01, 3.65, 7154.0])

# Reaction 2652
reaction('C5H11-1 + CH4 => NC5H12 + CH3', [8.653000e-02, 3.84, 11020.0])

# Reaction 2653
reaction('NC5H12 + CH3 => C5H11-2 + CH4', [1.680000e+05, 2.133, 7574.0])

# Reaction 2654
reaction('C5H11-2 + CH4 => NC5H12 + CH3', [6.650000e+02, 2.726, 13880.0])

# Reaction 2655
reaction('NC5H12 + CH3 => C5H11-3 + CH4', [8.400000e+04, 2.133, 7574.0])

# Reaction 2656
reaction('C5H11-3 + CH4 => NC5H12 + CH3', [6.901000e+02, 2.721, 13890.0])

# Reaction 2657
reaction('NC5H12 + HO2 => C5H11-1 + H2O2', [4.080000e+01, 3.59, 17160.0])

# Reaction 2658
reaction('C5H11-1 + H2O2 => NC5H12 + HO2', [2.477000e+00, 3.531, 3490.0])

# Reaction 2659
reaction('NC5H12 + HO2 => C5H11-2 + H2O2', [1.264000e+02, 3.37, 13720.0])

# Reaction 2660
reaction('C5H11-2 + H2O2 => NC5H12 + HO2', [3.173000e-01, 3.714, 2483.0])

# Reaction 2661
reaction('NC5H12 + HO2 => C5H11-3 + H2O2', [6.320000e+01, 3.37, 13720.0])

# Reaction 2662
reaction('C5H11-3 + H2O2 => NC5H12 + HO2', [3.293000e-01, 3.709, 2491.0])

# Reaction 2663
reaction('NC5H12 + CH3O2 => C5H11-1 + CH3O2H', [4.080000e+01, 3.59, 17160.0])

# Reaction 2664
reaction('C5H11-1 + CH3O2H => NC5H12 + CH3O2', [4.818000e+01, 3.086, 1895.0])

# Reaction 2665
reaction('NC5H12 + CH3O2 => C5H11-2 + CH3O2H', [1.264000e+02, 3.37, 13720.0])

# Reaction 2666
reaction('C5H11-2 + CH3O2H => NC5H12 + CH3O2', [6.173000e+00, 3.269, 888.0])

# Reaction 2667
reaction('NC5H12 + CH3O2 => C5H11-3 + CH3O2H', [6.320000e+01, 3.37, 13720.0])

# Reaction 2668
reaction('C5H11-3 + CH3O2H => NC5H12 + CH3O2', [6.406000e+00, 3.264, 896.0])

# Reaction 2669
reaction('NC5H12 + C2H5 => C5H11-1 + C2H6', [1.000000e+11, 0.0, 13400.0])

# Reaction 2670
reaction('C5H11-1 + C2H6 => NC5H12 + C2H5', [3.200000e+11, 0.0, 12300.0])

# Reaction 2671
reaction('NC5H12 + C2H5 => C5H11-2 + C2H6', [1.000000e+11, 0.0, 10400.0])

# Reaction 2672
reaction('C5H11-2 + C2H6 => NC5H12 + C2H5', [1.000000e+11, 0.0, 12900.0])

# Reaction 2673
reaction('NC5H12 + C2H5 => C5H11-3 + C2H6', [5.000000e+10, 0.0, 10400.0])

# Reaction 2674
reaction('C5H11-3 + C2H6 => NC5H12 + C2H5', [1.000000e+11, 0.0, 12900.0])

# Reaction 2675
reaction('NC5H12 + C2H3 => C5H11-1 + C2H4', [1.000000e+12, 0.0, 18000.0])

# Reaction 2676
reaction('C5H11-1 + C2H4 => NC5H12 + C2H3', [2.600000e+12, 0.0, 25400.0])

# Reaction 2677
reaction('NC5H12 + C2H3 => C5H11-2 + C2H4', [8.000000e+11, 0.0, 16800.0])

# Reaction 2678
reaction('C5H11-2 + C2H4 => NC5H12 + C2H3', [2.000000e+12, 0.0, 24200.0])

# Reaction 2679
reaction('NC5H12 + C2H3 => C5H11-3 + C2H4', [4.000000e+11, 0.0, 16800.0])

# Reaction 2680
reaction('C5H11-3 + C2H4 => NC5H12 + C2H3', [2.000000e+12, 0.0, 24200.0])

# Reaction 2681
reaction('NC5H12 + C5H11-1 => C5H11-2 + NC5H12', [1.000000e+11, 0.0, 10400.0])

# Reaction 2682
reaction('C5H11-2 + NC5H12 => NC5H12 + C5H11-1', [1.000000e+11, 0.0, 12900.0])

# Reaction 2683
reaction('NC5H12 + C5H11-1 => C5H11-3 + NC5H12', [5.000000e+10, 0.0, 10400.0])

# Reaction 2684
reaction('C5H11-3 + NC5H12 => NC5H12 + C5H11-1', [1.000000e+11, 0.0, 12900.0])

# Reaction 2685
reaction('NC5H12 + C5H11-2 => C5H11-3 + NC5H12', [5.000000e+10, 0.0, 12300.0])

# Reaction 2686
reaction('C5H11-3 + NC5H12 => NC5H12 + C5H11-2', [1.000000e+11, 0.0, 12300.0])

# Reaction 2687
reaction('NC5H12 + O2CHO => C5H11-1 + HO2CHO', [1.680000e+13, 0.0, 20440.0])

# Reaction 2688
reaction('C5H11-1 + HO2CHO => NC5H12 + O2CHO', [1.368000e+00, 3.11, 2295.0])

# Reaction 2689
reaction('NC5H12 + O2CHO => C5H11-2 + HO2CHO', [1.120000e+13, 0.0, 17690.0])

# Reaction 2690
reaction('C5H11-2 + HO2CHO => NC5H12 + O2CHO', [4.306000e+03, 2.13, 4147.0])

# Reaction 2691
reaction('NC5H12 + O2CHO => C5H11-3 + HO2CHO', [5.600000e+12, 0.0, 17690.0])

# Reaction 2692
reaction('C5H11-3 + HO2CHO => NC5H12 + O2CHO', [4.318000e+03, 2.13, 4149.0])

# Reaction 2693
reaction('NC5H12 + CH3O => C5H11-1 + CH3OH', [3.000000e+11, 0.0, 7000.0])

# Reaction 2694
reaction('C5H11-1 + CH3OH => NC5H12 + CH3O', [1.200000e+10, 0.0, 9200.0])

# Reaction 2695
reaction('NC5H12 + CH3O => C5H11-2 + CH3OH', [2.200000e+11, 0.0, 5000.0])

# Reaction 2696
reaction('C5H11-2 + CH3OH => NC5H12 + CH3O', [9.000000e+09, 0.0, 7200.0])

# Reaction 2697
reaction('NC5H12 + CH3O => C5H11-3 + CH3OH', [1.100000e+11, 0.0, 5000.0])

# Reaction 2698
reaction('C5H11-3 + CH3OH => NC5H12 + CH3O', [9.000000e+09, 0.0, 7200.0])

# Reaction 2699
reaction('NC5H12 + O2 => C5H11-1 + HO2', [4.200000e+13, 0.0, 52800.0])

# Reaction 2700
reaction('C5H11-1 + HO2 => NC5H12 + O2', [2.311000e+10, 0.288, 446.0])

# Reaction 2701
reaction('NC5H12 + O2 => C5H11-2 + HO2', [2.800000e+13, 0.0, 50160.0])

# Reaction 2702
reaction('C5H11-2 + HO2 => NC5H12 + O2', [6.370000e+08, 0.691, 231.0])

# Reaction 2703
reaction('NC5H12 + O2 => C5H11-3 + HO2', [1.400000e+13, 0.0, 50160.0])

# Reaction 2704
reaction('C5H11-3 + HO2 => NC5H12 + O2', [6.611000e+08, 0.686, 239.0])

# Reaction 2705
reaction('C5H11-1 => C2H4 + NC3H7', [3.205000e+12, 0.451, 29430.0])

# Reaction 2706
reaction('C2H4 + NC3H7 => C5H11-1', [8.800000e+03, 2.48, 6130.0])

# Reaction 2707
reaction('C5H11-1 => H + C5H10-1', [3.354000e+11, 0.608, 35640.0])

# Reaction 2708
reaction('H + C5H10-1 => C5H11-1', [2.500000e+11, 0.51, 2620.0])

# Reaction 2709
reaction('C5H11-1 => C5H11-2', [3.875000e+09, 0.353, 19760.0])

# Reaction 2710
reaction('C5H11-2 => C5H11-1', [1.602000e+08, 0.756, 22200.0])

# Reaction 2711
reaction('C5H11-2 => C3H6 + C2H5', [1.223000e+12, 0.635, 29360.0])

# Reaction 2712
reaction('C3H6 + C2H5 => C5H11-2', [8.800000e+03, 2.48, 6130.0])

# Reaction 2713
reaction('C5H11-2 => C5H10-1 + H', [2.353000e+10, 1.011, 36680.0])

# Reaction 2714
reaction('C5H10-1 + H => C5H11-2', [4.240000e+11, 0.51, 1230.0])

# Reaction 2715
reaction('C5H11-2 => C5H10-2 + H', [3.988000e+11, 0.41, 35220.0])

# Reaction 2716
reaction('C5H10-2 + H => C5H11-2', [2.500000e+11, 0.51, 2620.0])

# Reaction 2717
reaction('C5H11-3 => C4H8-1 + CH3', [6.343000e+10, 1.119, 30460.0])

# Reaction 2718
reaction('C4H8-1 + CH3 => C5H11-3', [1.760000e+04, 2.48, 6130.0])

# Reaction 2719
reaction('C5H11-3 => C5H10-2 + H', [8.277000e+11, 0.405, 35230.0])

# Reaction 2720
reaction('C5H10-2 + H => C5H11-3', [2.500000e+11, 0.51, 2620.0])

# Reaction 2721
reaction('C5H10-1 => C2H5 + C3H5-A', [9.864000e+21, -2.086, 75060.0])

# Reaction 2722
reaction('C2H5 + C3H5-A => C5H10-1', [4.000000e+12, 0.0, -596.0])

# Reaction 2723
reaction('C5H10-1 + H => C5H91-3 + H2', [3.376000e+05, 2.36, 207.0])

# Reaction 2724
reaction('C5H91-3 + H2 => C5H10-1 + H', [3.096000e+05, 2.53, 20320.0])

# Reaction 2725
reaction('C5H10-1 + H => C5H91-4 + H2', [1.300000e+06, 2.4, 4471.0])

# Reaction 2726
reaction('C5H91-4 + H2 => C5H10-1 + H', [2.777000e+02, 3.17, 11240.0])

# Reaction 2727
reaction('C5H10-1 + H => C5H91-5 + H2', [6.650000e+05, 2.54, 6756.0])

# Reaction 2728
reaction('C5H91-5 + H2 => C5H10-1 + H', [2.193000e+03, 2.96, 11020.0])

# Reaction 2729
reaction('C5H10-1 + O => C5H91-3 + OH', [6.600000e+05, 2.43, 1210.0])

# Reaction 2730
reaction('C5H91-3 + OH => C5H10-1 + O', [2.658000e+05, 2.6, 19230.0])

# Reaction 2731
reaction('C5H10-1 + O => C5H91-4 + OH', [5.510000e+05, 2.45, 2830.0])

# Reaction 2732
reaction('C5H91-4 + OH => C5H10-1 + O', [5.169000e+01, 3.22, 7505.0])

# Reaction 2733
reaction('C5H10-1 + O => C5H91-5 + OH', [9.800000e+05, 2.43, 4750.0])

# Reaction 2734
reaction('C5H91-5 + OH => C5H10-1 + O', [1.419000e+03, 2.85, 6915.0])

# Reaction 2735
reaction('C5H10-1 + OH => C5H91-3 + H2O', [2.764000e+04, 2.64, -1919.0])

# Reaction 2736
reaction('C5H91-3 + H2O => C5H10-1 + OH', [1.097000e+05, 2.81, 33350.0])

# Reaction 2737
reaction('C5H10-1 + OH => C5H91-4 + H2O', [4.670000e+07, 1.61, -35.0])

# Reaction 2738
reaction('C5H91-4 + H2O => C5H10-1 + OH', [4.319000e+04, 2.38, 21890.0])

# Reaction 2739
reaction('C5H10-1 + OH => C5H91-5 + H2O', [5.270000e+09, 0.97, 1586.0])

# Reaction 2740
reaction('C5H91-5 + H2O => C5H10-1 + OH', [7.524000e+07, 1.39, 21000.0])

# Reaction 2741
reaction('C5H10-1 + CH3 => C5H91-3 + CH4', [3.690000e+00, 3.31, 4002.0])

# Reaction 2742
reaction('C5H91-3 + CH4 => C5H10-1 + CH3', [8.840000e+01, 3.48, 24590.0])

# Reaction 2743
reaction('C5H10-1 + CH3 => C5H91-4 + CH4', [1.510000e+00, 3.46, 5481.0])

# Reaction 2744
reaction('C5H91-4 + CH4 => C5H10-1 + CH3', [8.426000e-03, 4.23, 12730.0])

# Reaction 2745
reaction('C5H10-1 + CH3 => C5H91-5 + CH4', [4.521000e-01, 3.65, 7154.0])

# Reaction 2746
reaction('C5H91-5 + CH4 => C5H10-1 + CH3', [3.895000e-02, 4.07, 11890.0])

# Reaction 2747
reaction('C5H10-1 + O2 => C5H91-3 + HO2', [2.200000e+12, 0.0, 37220.0])

# Reaction 2748
reaction('C5H91-3 + HO2 => C5H10-1 + O2', [3.665000e+10, 0.49, -152.0])

# Reaction 2749
reaction('C5H10-1 + O2 => C5H91-4 + HO2', [2.000000e+13, 0.0, 49640.0])

# Reaction 2750
reaction('C5H91-4 + HO2 => C5H10-1 + O2', [7.760000e+07, 1.09, -1072.0])

# Reaction 2751
reaction('C5H10-1 + O2 => C5H91-5 + HO2', [3.000000e+13, 0.0, 52290.0])

# Reaction 2752
reaction('C5H91-5 + HO2 => C5H10-1 + O2', [1.797000e+09, 0.75, -932.0])

# Reaction 2753
reaction('C5H10-1 + HO2 => C5H91-3 + H2O2', [4.820000e+03, 2.55, 10530.0])

# Reaction 2754
reaction('C5H91-3 + H2O2 => C5H10-1 + HO2', [1.136000e+05, 2.39, 14330.0])

# Reaction 2755
reaction('C5H10-1 + HO2 => C5H91-4 + H2O2', [9.640000e+03, 2.6, 13910.0])

# Reaction 2756
reaction('C5H91-4 + H2O2 => C5H10-1 + HO2', [5.292000e+01, 3.04, 4372.0])

# Reaction 2757
reaction('C5H10-1 + HO2 => C5H91-5 + H2O2', [2.380000e+04, 2.55, 16490.0])

# Reaction 2758
reaction('C5H91-5 + H2O2 => C5H10-1 + HO2', [2.017000e+03, 2.64, 4446.0])

# Reaction 2759
reaction('C5H10-1 + CH3O2 => C5H91-3 + CH3O2H', [4.820000e+03, 2.55, 10530.0])

# Reaction 2760
reaction('C5H91-3 + CH3O2H => C5H10-1 + CH3O2', [2.366000e+05, 2.21, 11310.0])

# Reaction 2761
reaction('C5H10-1 + CH3O2 => C5H91-4 + CH3O2H', [9.640000e+03, 2.6, 13910.0])

# Reaction 2762
reaction('C5H91-4 + CH3O2H => C5H10-1 + CH3O2', [1.102000e+02, 2.86, 1352.0])

# Reaction 2763
reaction('C5H10-1 + CH3O2 => C5H91-5 + CH3O2H', [2.380000e+04, 2.55, 16490.0])

# Reaction 2764
reaction('C5H91-5 + CH3O2H => C5H10-1 + CH3O2', [4.200000e+03, 2.47, 1426.0])

# Reaction 2765
reaction('C5H10-1 + CH3O => C5H91-3 + CH3OH', [4.000000e+01, 2.9, 8609.0])

# Reaction 2766
reaction('C5H91-3 + CH3OH => C5H10-1 + CH3O', [1.769000e+01, 3.09, 26990.0])

# Reaction 2767
reaction('C5H10-1 + CH3O => C5H91-4 + CH3OH', [1.450000e+11, 0.0, 4571.0])

# Reaction 2768
reaction('C5H91-4 + CH3OH => C5H10-1 + CH3O', [1.493000e+07, 0.79, 9611.0])

# Reaction 2769
reaction('C5H10-1 + CH3O => C5H91-5 + CH3OH', [2.170000e+11, 0.0, 6458.0])

# Reaction 2770
reaction('C5H91-5 + CH3OH => C5H10-1 + CH3O', [3.450000e+08, 0.45, 8988.0])

# Reaction 2771
reaction('C5H10-2 + H => C5H91-3 + H2', [1.730000e+05, 2.5, 2492.0])

# Reaction 2772
reaction('C5H91-3 + H2 => C5H10-2 + H', [4.561000e+06, 2.06, 19750.0])

# Reaction 2773
reaction('C5H10-2 + H => C5H92-4 + H2', [3.376000e+05, 2.36, 207.0])

# Reaction 2774
reaction('C5H92-4 + H2 => C5H10-2 + H', [4.361000e+06, 2.1, 20330.0])

# Reaction 2775
reaction('C5H10-2 + H => C5H92-5 + H2', [6.651000e+05, 2.54, 6756.0])

# Reaction 2776
reaction('C5H92-5 + H2 => C5H10-2 + H', [3.085000e+04, 2.54, 11030.0])

# Reaction 2777
reaction('C5H10-2 + O => C5H91-3 + OH', [4.410000e+05, 2.42, 3150.0])

# Reaction 2778
reaction('C5H91-3 + OH => C5H10-2 + O', [5.107000e+06, 1.98, 18310.0])

# Reaction 2779
reaction('C5H10-2 + O => C5H92-4 + OH', [9.900000e+05, 2.43, 1210.0])

# Reaction 2780
reaction('C5H92-4 + OH => C5H10-2 + O', [5.617000e+06, 2.17, 19240.0])

# Reaction 2781
reaction('C5H10-2 + O => C5H92-5 + OH', [9.800000e+05, 2.43, 4750.0])

# Reaction 2782
reaction('C5H92-5 + OH => C5H10-2 + O', [1.996000e+04, 2.42, 6931.0])

# Reaction 2783
reaction('C5H10-2 + OH => C5H91-3 + H2O', [3.120000e+06, 2.0, -298.0])

# Reaction 2784
reaction('C5H91-3 + H2O => C5H10-2 + OH', [3.562000e+08, 1.56, 32110.0])

# Reaction 2785
reaction('C5H10-2 + OH => C5H92-4 + H2O', [2.764000e+04, 2.64, -1919.0])

# Reaction 2786
reaction('C5H92-4 + H2O => C5H10-2 + OH', [1.546000e+06, 2.38, 33360.0])

# Reaction 2787
reaction('C5H10-2 + OH => C5H92-5 + H2O', [5.270000e+09, 0.97, 1586.0])

# Reaction 2788
reaction('C5H92-5 + H2O => C5H10-2 + OH', [1.058000e+09, 0.96, 21020.0])

# Reaction 2789
reaction('C5H10-2 + CH3 => C5H91-3 + CH4', [2.210000e+00, 3.5, 5675.0])

# Reaction 2790
reaction('C5H91-3 + CH4 => C5H10-2 + CH3', [1.522000e+03, 3.06, 23410.0])

# Reaction 2791
reaction('C5H10-2 + CH3 => C5H92-4 + CH4', [3.690000e+00, 3.31, 4002.0])

# Reaction 2792
reaction('C5H92-4 + CH4 => C5H10-2 + CH3', [1.245000e+03, 3.05, 24610.0])

# Reaction 2793
reaction('C5H10-2 + CH3 => C5H92-5 + CH4', [4.521000e-01, 3.65, 7154.0])

# Reaction 2794
reaction('C5H92-5 + CH4 => C5H10-2 + CH3', [5.478000e-01, 3.64, 11910.0])

# Reaction 2795
reaction('C5H10-2 + O2 => C5H91-3 + HO2', [3.300000e+12, 0.0, 39900.0])

# Reaction 2796
reaction('C5H91-3 + HO2 => C5H10-2 + O2', [1.581000e+12, -0.11, -326.0])

# Reaction 2797
reaction('C5H10-2 + O2 => C5H92-4 + HO2', [2.200000e+12, 0.0, 37220.0])

# Reaction 2798
reaction('C5H92-4 + HO2 => C5H10-2 + O2', [5.163000e+11, 0.07, -136.0])

# Reaction 2799
reaction('C5H10-2 + O2 => C5H92-5 + HO2', [3.000000e+13, 0.0, 52290.0])

# Reaction 2800
reaction('C5H92-5 + HO2 => C5H10-2 + O2', [2.528000e+10, 0.32, -916.0])

# Reaction 2801
reaction('C5H10-2 + HO2 => C5H91-3 + H2O2', [9.639000e+03, 2.6, 13910.0])

# Reaction 2802
reaction('C5H91-3 + H2O2 => C5H10-2 + HO2', [6.532000e+06, 1.84, 14860.0])

# Reaction 2803
reaction('C5H10-2 + HO2 => C5H92-4 + H2O2', [4.820000e+03, 2.55, 10530.0])

# Reaction 2804
reaction('C5H92-4 + H2O2 => C5H10-2 + HO2', [1.600000e+06, 1.96, 14350.0])

# Reaction 2805
reaction('C5H10-2 + HO2 => C5H92-5 + H2O2', [2.380000e+04, 2.55, 16490.0])

# Reaction 2806
reaction('C5H92-5 + H2O2 => C5H10-2 + HO2', [2.837000e+04, 2.22, 4462.0])

# Reaction 2807
reaction('C5H10-2 + CH3O2 => C5H91-3 + CH3O2H', [9.639000e+03, 2.6, 13910.0])

# Reaction 2808
reaction('C5H91-3 + CH3O2H => C5H10-2 + CH3O2', [1.360000e+07, 1.66, 11840.0])

# Reaction 2809
reaction('C5H10-2 + CH3O2 => C5H92-4 + CH3O2H', [4.820000e+03, 2.55, 10530.0])

# Reaction 2810
reaction('C5H92-4 + CH3O2H => C5H10-2 + CH3O2', [3.332000e+06, 1.79, 11330.0])

# Reaction 2811
reaction('C5H10-2 + CH3O2 => C5H92-5 + CH3O2H', [2.380000e+04, 2.55, 16490.0])

# Reaction 2812
reaction('C5H92-5 + CH3O2H => C5H10-2 + CH3O2', [5.908000e+04, 2.04, 1442.0])

# Reaction 2813
reaction('C5H10-2 + CH3O => C5H91-3 + CH3OH', [9.000000e+01, 2.95, 11990.0])

# Reaction 2814
reaction('C5H91-3 + CH3OH => C5H10-2 + CH3O', [1.144000e+03, 2.54, 27510.0])

# Reaction 2815
reaction('C5H10-2 + CH3O => C5H92-4 + CH3OH', [4.000000e+01, 2.9, 8609.0])

# Reaction 2816
reaction('C5H92-4 + CH3OH => C5H10-2 + CH3O', [2.491000e+02, 2.67, 27010.0])

# Reaction 2817
reaction('C5H10-2 + CH3O => C5H92-5 + CH3OH', [2.170000e+11, 0.0, 6458.0])

# Reaction 2818
reaction('C5H92-5 + CH3OH => C5H10-2 + CH3O', [4.853000e+09, 0.02, 9004.0])

# Reaction 2819
reaction('C5H91-3 + HO2 => C5H9O1-3 + OH', [9.640000e+12, 0.0, 0.0])

# Reaction 2820
reaction('C5H9O1-3 + OH => C5H91-3 + HO2', [6.088000e+15, -1.07, 15720.0])

# Reaction 2821
reaction('C5H91-3 + CH3O2 => C5H9O1-3 + CH3O', [9.640000e+12, 0.0, 0.0])

# Reaction 2822
reaction('C5H9O1-3 + CH3O => C5H91-3 + CH3O2', [5.946000e+17, -1.65, 20480.0])

# Reaction 2823
reaction('C5H91-3 + C2H5O2 => C5H9O1-3 + C2H5O', [9.640000e+12, 0.0, 0.0])

# Reaction 2824
reaction('C5H9O1-3 + C2H5O => C5H91-3 + C2H5O2', [3.893000e+14, -0.72, 18330.0])

# Reaction 2825
reaction('C5H92-4 + HO2 => C5H9O2-4 + OH', [9.640000e+12, 0.0, 0.0])

# Reaction 2826
reaction('C5H9O2-4 + OH => C5H92-4 + HO2', [7.027000e+15, -1.24, 15890.0])

# Reaction 2827
reaction('C5H92-4 + CH3O2 => C5H9O2-4 + CH3O', [9.640000e+12, 0.0, 0.0])

# Reaction 2828
reaction('C5H9O2-4 + CH3O => C5H92-4 + CH3O2', [6.863000e+17, -1.82, 20650.0])

# Reaction 2829
reaction('C5H92-4 + C2H5O2 => C5H9O2-4 + C2H5O', [9.640000e+12, 0.0, 0.0])

# Reaction 2830
reaction('C5H9O2-4 + C2H5O => C5H92-4 + C2H5O2', [4.493000e+14, -0.89, 18490.0])

# Reaction 2831
reaction('C5H91-3 => C4H6 + CH3', [7.554000e+14, -0.52, 38520.0])

# Reaction 2832
reaction('C4H6 + CH3 => C5H91-3', [1.000000e+11, 0.0, 7800.0])

# Reaction 2833
reaction('C5H91-3 => C5H81-3 + H', [4.733000e+11, 0.636, 42640.0])

# Reaction 2834
reaction('C5H81-3 + H => C5H91-3', [2.500000e+11, 0.51, 2620.0])

# Reaction 2835
reaction('C5H91-4 => C3H6 + C2H3', [5.814000e+11, 0.17, 35850.0])

# Reaction 2836
reaction('C3H6 + C2H3 => C5H91-4', [1.000000e+11, 0.0, 7800.0])

# Reaction 2837
reaction('C5H91-5 => C2H4 + C3H5-A', [7.157000e+16, -1.42, 17750.0])

# Reaction 2838
reaction('C2H4 + C3H5-A => C5H91-5', [1.000000e+11, 0.0, 7800.0])

# Reaction 2839
reaction('C5H92-4 => C5H81-3 + H', [2.015000e+15, -0.34, 46030.0])

# Reaction 2840
reaction('C5H81-3 + H => C5H92-4', [1.000000e+13, 0.0, 1200.0])

# Reaction 2841
reaction('C5H92-5 => C2H4 + C3H5-S', [8.899000e+16, -1.18, 42180.0])

# Reaction 2842
reaction('C2H4 + C3H5-S => C5H92-5', [2.000000e+11, 0.0, 7800.0])

# Reaction 2843
reaction('C5H9O1-3 => C2H3CHO + C2H5', [3.131000e+19, -1.85, 10670.0])

# Reaction 2844
reaction('C2H3CHO + C2H5 => C5H9O1-3', [1.000000e+11, 0.0, 9600.0])

# Reaction 2845
reaction('C5H9O1-3 => C2H5CHO + C2H3', [1.417000e+18, -1.56, 23340.0])

# Reaction 2846
reaction('C2H5CHO + C2H3 => C5H9O1-3', [1.000000e+11, 0.0, 9600.0])

# Reaction 2847
reaction('C5H81-3 + OH => CH2O + C4H71-3', [1.000000e+12, 0.0, 0.0])

# Reaction 2848
reaction('CH2O + C4H71-3 => C5H81-3 + OH', [1.000000e+11, 0.0, 15000.0])

# Reaction 2849
reaction('C5H81-3 + OH => C2H3CHO + C2H5', [1.000000e+12, 0.0, 0.0])

# Reaction 2850
reaction('C2H3CHO + C2H5 => C5H81-3 + OH', [1.000000e+11, 0.0, 15000.0])

# Reaction 2851
reaction('C5H81-3 + OH => CH3CHO + C3H5-S', [1.000000e+12, 0.0, 0.0])

# Reaction 2852
reaction('CH3CHO + C3H5-S => C5H81-3 + OH', [1.000000e+11, 0.0, 15000.0])

# Reaction 2853
reaction('C5H9O2-4 => SC3H5CHO + CH3', [5.983000e+15, -1.13, 9941.0])

# Reaction 2854
reaction('SC3H5CHO + CH3 => C5H9O2-4', [1.000000e+11, 0.0, 9600.0])

# Reaction 2855
reaction('C5H9O2-4 => CH3CHO + C3H5-S', [1.073000e+22, -2.66, 29650.0])

# Reaction 2856
reaction('CH3CHO + C3H5-S => C5H9O2-4', [1.000000e+11, 0.0, 9600.0])

# Reaction 2857
reaction('C5H10-2 => CH3 + C4H71-3', [6.486000e+19, -1.367, 76320.0])

# Reaction 2858
reaction('CH3 + C4H71-3 => C5H10-2', [8.000000e+12, 0.0, -596.0])

# Reaction 2859
reaction('C5H11-1 + O2 => C5H10-1 + HO2', [8.370000e-01, 3.59, 11960.0])

# Reaction 2860
reaction('C5H10-1 + HO2 => C5H11-1 + O2', [1.307000e+00, 3.42, 27810.0])

# Reaction 2861
reaction('C5H11-2 + O2 => C5H10-1 + HO2', [5.350000e-01, 3.71, 9322.0])

# Reaction 2862
reaction('C5H10-1 + HO2 => C5H11-2 + O2', [2.021000e+01, 3.137, 22740.0])

# Reaction 2863
reaction('C5H11-2 + O2 => C5H10-2 + HO2', [1.070000e+00, 3.71, 9322.0])

# Reaction 2864
reaction('C5H10-2 + HO2 => C5H11-2 + O2', [1.406000e+00, 3.738, 25590.0])

# Reaction 2865
reaction('C5H11-3 + O2 => C5H10-2 + HO2', [2.140000e+00, 3.71, 9322.0])

# Reaction 2866
reaction('C5H10-2 + HO2 => C5H11-3 + O2', [1.355000e+00, 3.743, 25580.0])

# Reaction 2867
reaction('C5H11-1 + HO2 => C5H11O-1 + OH', [9.000000e+12, 0.0, -1000.0])

# Reaction 2868
reaction('C5H11O-1 + OH => C5H11-1 + HO2', [1.944000e+15, -0.479, 26210.0])

# Reaction 2869
reaction('C5H11-2 + HO2 => C5H11O-2 + OH', [9.000000e+12, 0.0, -1000.0])

# Reaction 2870
reaction('C5H11O-2 + OH => C5H11-2 + HO2', [3.143000e+17, -1.132, 28220.0])

# Reaction 2871
reaction('C5H11-3 + HO2 => C5H11O-3 + OH', [9.000000e+12, 0.0, -1000.0])

# Reaction 2872
reaction('C5H11O-3 + OH => C5H11-3 + HO2', [3.019000e+17, -1.127, 28210.0])

# Reaction 2873
reaction('C5H11-1 + CH3O2 => C5H11O-1 + CH3O', [9.000000e+12, 0.0, -1000.0])

# Reaction 2874
reaction('C5H11O-1 + CH3O => C5H11-1 + CH3O2', [1.216000e+14, -0.18, 30450.0])

# Reaction 2875
reaction('C5H11-2 + CH3O2 => C5H11O-2 + CH3O', [9.000000e+12, 0.0, -1000.0])

# Reaction 2876
reaction('C5H11O-2 + CH3O => C5H11-2 + CH3O2', [1.966000e+16, -0.833, 32460.0])

# Reaction 2877
reaction('C5H11-3 + CH3O2 => C5H11O-3 + CH3O', [9.000000e+12, 0.0, -1000.0])

# Reaction 2878
reaction('C5H11O-3 + CH3O => C5H11-3 + CH3O2', [1.889000e+16, -0.828, 32450.0])

# Reaction 2879
reaction('C5H11O2-1 + NC5H12 => C5H11O2H-1 + C5H11-1', [1.210000e+13, 0.0, 20430.0])

# Reaction 2880
reaction('C5H11O2H-1 + C5H11-1 => C5H11O2-1 + NC5H12', [1.440000e+10, 0.0, 15000.0])

# Reaction 2881
reaction('C5H11O2-2 + NC5H12 => C5H11O2H-2 + C5H11-1', [1.210000e+13, 0.0, 20430.0])

# Reaction 2882
reaction('C5H11O2H-2 + C5H11-1 => C5H11O2-2 + NC5H12', [1.440000e+10, 0.0, 15000.0])

# Reaction 2883
reaction('C5H11O2-3 + NC5H12 => C5H11O2H-3 + C5H11-1', [1.210000e+13, 0.0, 20430.0])

# Reaction 2884
reaction('C5H11O2H-3 + C5H11-1 => C5H11O2-3 + NC5H12', [1.440000e+10, 0.0, 15000.0])

# Reaction 2885
reaction('C5H11O2-1 + NC5H12 => C5H11O2H-1 + C5H11-2', [8.064000e+12, 0.0, 17700.0])

# Reaction 2886
reaction('C5H11O2H-1 + C5H11-2 => C5H11O2-1 + NC5H12', [1.440000e+10, 0.0, 15000.0])

# Reaction 2887
reaction('C5H11O2-2 + NC5H12 => C5H11O2H-2 + C5H11-2', [8.064000e+12, 0.0, 17700.0])

# Reaction 2888
reaction('C5H11O2H-2 + C5H11-2 => C5H11O2-2 + NC5H12', [1.440000e+10, 0.0, 15000.0])

# Reaction 2889
reaction('C5H11O2-3 + NC5H12 => C5H11O2H-3 + C5H11-2', [8.064000e+12, 0.0, 17700.0])

# Reaction 2890
reaction('C5H11O2H-3 + C5H11-2 => C5H11O2-3 + NC5H12', [1.440000e+10, 0.0, 15000.0])

# Reaction 2891
reaction('C5H11O2-1 + NC5H12 => C5H11O2H-1 + C5H11-3', [4.032000e+12, 0.0, 17700.0])

# Reaction 2892
reaction('C5H11O2H-1 + C5H11-3 => C5H11O2-1 + NC5H12', [1.440000e+10, 0.0, 15000.0])

# Reaction 2893
reaction('C5H11O2-2 + NC5H12 => C5H11O2H-2 + C5H11-3', [4.032000e+12, 0.0, 17700.0])

# Reaction 2894
reaction('C5H11O2H-2 + C5H11-3 => C5H11O2-2 + NC5H12', [1.440000e+10, 0.0, 15000.0])

# Reaction 2895
reaction('C5H11O2-3 + NC5H12 => C5H11O2H-3 + C5H11-3', [4.032000e+12, 0.0, 17700.0])

# Reaction 2896
reaction('C5H11O2H-3 + C5H11-3 => C5H11O2-3 + NC5H12', [1.440000e+10, 0.0, 15000.0])

# Reaction 2897
reaction('C5H11-1 + C5H11O2-1 => C5H11O-1 + C5H11O-1', [9.000000e+12, 0.0, -1000.0])

# Reaction 2898
reaction('C5H11O-1 + C5H11O-1 => C5H11-1 + C5H11O2-1', [5.940000e+13, -0.13, 29910.0])

# Reaction 2899
reaction('C5H11-1 + C5H11O2-2 => C5H11O-1 + C5H11O-2', [9.000000e+12, 0.0, -1000.0])

# Reaction 2900
reaction('C5H11O-1 + C5H11O-2 => C5H11-1 + C5H11O2-2', [5.285000e+13, -0.116, 29600.0])

# Reaction 2901
reaction('C5H11-1 + C5H11O2-3 => C5H11O-1 + C5H11O-3', [9.000000e+12, 0.0, -1000.0])

# Reaction 2902
reaction('C5H11O-1 + C5H11O-3 => C5H11-1 + C5H11O2-3', [1.053000e+14, -0.116, 29600.0])

# Reaction 2903
reaction('C5H11-2 + C5H11O2-1 => C5H11O-2 + C5H11O-1', [9.000000e+12, 0.0, -1000.0])

# Reaction 2904
reaction('C5H11O-2 + C5H11O-1 => C5H11-2 + C5H11O2-1', [9.604000e+15, -0.783, 31920.0])

# Reaction 2905
reaction('C5H11-2 + C5H11O2-2 => C5H11O-2 + C5H11O-2', [9.000000e+12, 0.0, -1000.0])

# Reaction 2906
reaction('C5H11O-2 + C5H11O-2 => C5H11-2 + C5H11O2-2', [8.544000e+15, -0.769, 31620.0])

# Reaction 2907
reaction('C5H11-2 + C5H11O2-3 => C5H11O-2 + C5H11O-3', [9.000000e+12, 0.0, -1000.0])

# Reaction 2908
reaction('C5H11O-2 + C5H11O-3 => C5H11-2 + C5H11O2-3', [1.703000e+16, -0.769, 31620.0])

# Reaction 2909
reaction('C5H11-3 + C5H11O2-1 => C5H11O-3 + C5H11O-1', [9.000000e+12, 0.0, -1000.0])

# Reaction 2910
reaction('C5H11O-3 + C5H11O-1 => C5H11-3 + C5H11O2-1', [9.225000e+15, -0.778, 31920.0])

# Reaction 2911
reaction('C5H11-3 + C5H11O2-2 => C5H11O-3 + C5H11O-2', [9.000000e+12, 0.0, -1000.0])

# Reaction 2912
reaction('C5H11O-3 + C5H11O-2 => C5H11-3 + C5H11O2-2', [8.206000e+15, -0.764, 31610.0])

# Reaction 2913
reaction('C5H11-3 + C5H11O2-3 => C5H11O-3 + C5H11O-3', [9.000000e+12, 0.0, -1000.0])

# Reaction 2914
reaction('C5H11O-3 + C5H11O-3 => C5H11-3 + C5H11O2-3', [1.636000e+16, -0.764, 31610.0])

# Reaction 2915
reaction('C5H11O2-1 + C5H11O2-2 => O2 + C5H11O-1 + C5H11O-2', [1.400000e+16, -1.61, 1860.0])

# Reaction 2916
reaction('O2 + C5H11O-1 + C5H11O-2 => C5H11O2-1 + C5H11O2-2', [0.000000e+00, 0.0, 0.0])

# Reaction 2917
reaction('C5H11O2-1 + C5H11O2-3 => O2 + C5H11O-1 + C5H11O-3', [1.400000e+16, -1.61, 1860.0])

# Reaction 2918
reaction('O2 + C5H11O-1 + C5H11O-3 => C5H11O2-1 + C5H11O2-3', [0.000000e+00, 0.0, 0.0])

# Reaction 2919
reaction('C5H11O2-1 + CH3O2 => O2 + C5H11O-1 + CH3O', [1.400000e+16, -1.61, 1860.0])

# Reaction 2920
reaction('O2 + C5H11O-1 + CH3O => C5H11O2-1 + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 2921
reaction('C5H11O2-1 + C5H11O2-1 => O2 + C5H11O-1 + C5H11O-1', [1.400000e+16, -1.61, 1860.0])

# Reaction 2922
reaction('O2 + C5H11O-1 + C5H11O-1 => C5H11O2-1 + C5H11O2-1', [0.000000e+00, 0.0, 0.0])

# Reaction 2923
reaction('H2O2 + C5H11O2-1 => HO2 + C5H11O2H-1', [2.400000e+12, 0.0, 10000.0])

# Reaction 2924
reaction('HO2 + C5H11O2H-1 => H2O2 + C5H11O2-1', [2.400000e+12, 0.0, 10000.0])

# Reaction 2925
reaction('C5H11O2-1 + HO2 => C5H11O2H-1 + O2', [1.750000e+10, 0.0, -3275.0])

# Reaction 2926
reaction('C5H11O2H-1 + O2 => C5H11O2-1 + HO2', [3.746000e+13, -0.791, 33620.0])

# Reaction 2927
reaction('C5H11O2-2 + CH3O2 => O2 + C5H11O-2 + CH3O', [1.400000e+16, -1.61, 1860.0])

# Reaction 2928
reaction('O2 + C5H11O-2 + CH3O => C5H11O2-2 + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 2929
reaction('C5H11O2-2 + C5H11O2-3 => C5H11O-2 + C5H11O-3 + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 2930
reaction('C5H11O-2 + C5H11O-3 + O2 => C5H11O2-2 + C5H11O2-3', [0.000000e+00, 0.0, 0.0])

# Reaction 2931
reaction('C5H11O2-2 + C5H11O2-2 => O2 + C5H11O-2 + C5H11O-2', [1.400000e+16, -1.61, 1860.0])

# Reaction 2932
reaction('O2 + C5H11O-2 + C5H11O-2 => C5H11O2-2 + C5H11O2-2', [0.000000e+00, 0.0, 0.0])

# Reaction 2933
reaction('H2O2 + C5H11O2-2 => HO2 + C5H11O2H-2', [2.400000e+12, 0.0, 10000.0])

# Reaction 2934
reaction('HO2 + C5H11O2H-2 => H2O2 + C5H11O2-2', [2.400000e+12, 0.0, 10000.0])

# Reaction 2935
reaction('C5H11O2-2 + HO2 => C5H11O2H-2 + O2', [1.750000e+10, 0.0, -3275.0])

# Reaction 2936
reaction('C5H11O2H-2 + O2 => C5H11O2-2 + HO2', [4.380000e+13, -0.812, 33640.0])

# Reaction 2937
reaction('C5H11O2-3 + CH3O2 => O2 + C5H11O-3 + CH3O', [1.400000e+16, -1.61, 1860.0])

# Reaction 2938
reaction('O2 + C5H11O-3 + CH3O => C5H11O2-3 + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 2939
reaction('C5H11O2-3 + C5H11O2-3 => O2 + C5H11O-3 + C5H11O-3', [1.400000e+16, -1.61, 1860.0])

# Reaction 2940
reaction('O2 + C5H11O-3 + C5H11O-3 => C5H11O2-3 + C5H11O2-3', [0.000000e+00, 0.0, 0.0])

# Reaction 2941
reaction('H2O2 + C5H11O2-3 => HO2 + C5H11O2H-3', [2.400000e+12, 0.0, 10000.0])

# Reaction 2942
reaction('HO2 + C5H11O2H-3 => H2O2 + C5H11O2-3', [2.400000e+12, 0.0, 10000.0])

# Reaction 2943
reaction('C5H11O2-3 + HO2 => C5H11O2H-3 + O2', [1.750000e+10, 0.0, -3275.0])

# Reaction 2944
reaction('C5H11O2H-3 + O2 => C5H11O2-3 + HO2', [4.380000e+13, -0.812, 33640.0])

# Reaction 2945
reaction('C5H11O2H-1 => C5H11O-1 + OH', [1.500000e+16, 0.0, 42500.0])

# Reaction 2946
reaction('C5H11O-1 + OH => C5H11O2H-1', [2.927000e+07, 1.932, -3022.0])

# Reaction 2947
reaction('C5H11O2H-2 => C5H11O-2 + OH', [9.450000e+15, 0.0, 41600.0])

# Reaction 2948
reaction('C5H11O-2 + OH => C5H11O2H-2', [1.403000e+07, 1.968, -4252.0])

# Reaction 2949
reaction('C5H11O2H-3 => C5H11O-3 + OH', [9.450000e+15, 0.0, 41600.0])

# Reaction 2950
reaction('C5H11O-3 + OH => C5H11O2H-3', [2.797000e+07, 1.968, -4252.0])

# Reaction 2951
reaction('C5H11O-1 => CH2O + PC4H9', [5.661000e+20, -2.247, 24960.0])

# Reaction 2952
reaction('CH2O + PC4H9 => C5H11O-1', [6.250000e+10, 0.0, 11900.0])

# Reaction 2953
reaction('C5H11O-2 => CH3CHO + NC3H7', [1.689000e+22, -2.601, 19550.0])

# Reaction 2954
reaction('CH3CHO + NC3H7 => C5H11O-2', [7.500000e+10, 0.0, 11900.0])

# Reaction 2955
reaction('C5H11O-3 => C2H5 + C2H5CHO', [1.239000e+18, -1.199, 18590.0])

# Reaction 2956
reaction('C2H5 + C2H5CHO => C5H11O-3', [8.500000e+10, 0.0, 11900.0])

# Reaction 2957
reaction('C5H11O2-1 => C5H11-1 + O2', [2.338000e+20, -1.62, 35830.0])

# Reaction 2958
reaction('C5H11-1 + O2 => C5H11O2-1', [4.520000e+12, 0.0, 0.0])

# Reaction 2959
reaction('C5H11O2-2 => C5H11-2 + O2', [7.087000e+22, -2.287, 38150.0])

# Reaction 2960
reaction('C5H11-2 + O2 => C5H11O2-2', [7.540000e+12, 0.0, 0.0])

# Reaction 2961
reaction('C5H11O2-3 => C5H11-3 + O2', [3.415000e+22, -2.282, 38150.0])

# Reaction 2962
reaction('C5H11-3 + O2 => C5H11O2-3', [7.540000e+12, 0.0, 0.0])

# Reaction 2963
reaction('C5H11O2-1 => C5H10OOH1-2', [2.000000e+11, 0.0, 26850.0])

# Reaction 2964
reaction('C5H10OOH1-2 => C5H11O2-1', [2.083000e+10, -0.106, 13850.0])

# Reaction 2965
reaction('C5H11O2-1 => C5H10OOH1-3', [2.500000e+10, 0.0, 20850.0])

# Reaction 2966
reaction('C5H10OOH1-3 => C5H11O2-1', [2.604000e+09, -0.106, 7850.0])

# Reaction 2967
reaction('C5H11O2-1 => C5H10OOH1-4', [3.125000e+09, 0.0, 19050.0])

# Reaction 2968
reaction('C5H10OOH1-4 => C5H11O2-1', [3.255000e+08, -0.106, 6050.0])

# Reaction 2969
reaction('C5H11O2-1 => C5H10OOH1-5', [5.860000e+08, 0.0, 25550.0])

# Reaction 2970
reaction('C5H10OOH1-5 => C5H11O2-1', [1.379000e+09, -0.503, 10090.0])

# Reaction 2971
reaction('C5H11O2-2 => C5H10OOH2-1', [3.000000e+11, 0.0, 29400.0])

# Reaction 2972
reaction('C5H10OOH2-1 => C5H11O2-2', [7.790000e+11, -0.517, 13960.0])

# Reaction 2973
reaction('C5H11O2-2 => C5H10OOH2-3', [2.000000e+11, 0.0, 26850.0])

# Reaction 2974
reaction('C5H10OOH2-3 => C5H11O2-2', [2.698000e+10, -0.14, 13900.0])

# Reaction 2975
reaction('C5H11O2-2 => C5H10OOH2-4', [2.500000e+10, 0.0, 20850.0])

# Reaction 2976
reaction('C5H10OOH2-4 => C5H11O2-2', [3.372000e+09, -0.14, 7900.0])

# Reaction 2977
reaction('C5H11O2-2 => C5H10OOH2-5', [4.688000e+09, 0.0, 22350.0])

# Reaction 2978
reaction('C5H10OOH2-5 => C5H11O2-2', [1.217000e+10, -0.517, 6910.0])

# Reaction 2979
reaction('C5H11O2-3 => C5H10OOH3-1', [7.500000e+10, 0.0, 24400.0])

# Reaction 2980
reaction('C5H10OOH3-1 => C5H11O2-3', [1.948000e+11, -0.517, 8960.0])

# Reaction 2981
reaction('C5H11O2-3 => C5H10OOH3-2', [4.000000e+11, 0.0, 26850.0])

# Reaction 2982
reaction('C5H10OOH3-2 => C5H11O2-3', [5.396000e+10, -0.14, 13900.0])

# Reaction 2983
reaction('C5H11O2-1 => C5H10-1 + HO2', [5.044000e+38, -8.11, 40490.0])

# Reaction 2984
reaction('C5H10-1 + HO2 => C5H11O2-1', [1.523000e+31, -6.66, 20510.0])

# Reaction 2985
reaction('C5H11O2-2 => C5H10-1 + HO2', [5.075000e+42, -9.41, 41490.0])

# Reaction 2986
reaction('C5H10-1 + HO2 => C5H11O2-2', [2.039000e+34, -7.696, 16750.0])

# Reaction 2987
reaction('C5H11O2-2 => C5H10-2 + HO2', [5.044000e+38, -8.11, 40490.0])

# Reaction 2988
reaction('C5H10-2 + HO2 => C5H11O2-2', [7.050000e+28, -5.795, 18600.0])

# Reaction 2989
reaction('C5H11O2-3 => C5H10-2 + HO2', [1.009000e+39, -8.11, 40490.0])

# Reaction 2990
reaction('C5H10-2 + HO2 => C5H11O2-3', [1.410000e+29, -5.795, 18600.0])

# Reaction 2991
reaction('C5H10OOH1-2 => C5H10O1-2 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 2992
reaction('C5H10O1-2 + OH => C5H10OOH1-2', [0.000000e+00, 0.0, 0.0])

# Reaction 2993
reaction('C5H10OOH1-3 => C5H10O1-3 + OH', [7.500000e+10, 0.0, 15250.0])

# Reaction 2994
reaction('C5H10O1-3 + OH => C5H10OOH1-3', [0.000000e+00, 0.0, 0.0])

# Reaction 2995
reaction('C5H10OOH1-4 => C5H10O1-4 + OH', [9.375000e+09, 0.0, 6000.0])

# Reaction 2996
reaction('C5H10O1-4 + OH => C5H10OOH1-4', [0.000000e+00, 0.0, 0.0])

# Reaction 2997
reaction('C5H10OOH1-5 => C5H10O1-5 + OH', [1.172000e+09, 0.0, 1800.0])

# Reaction 2998
reaction('C5H10O1-5 + OH => C5H10OOH1-5', [0.000000e+00, 0.0, 0.0])

# Reaction 2999
reaction('C5H10OOH2-1 => C5H10O1-2 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 3000
reaction('C5H10O1-2 + OH => C5H10OOH2-1', [0.000000e+00, 0.0, 0.0])

# Reaction 3001
reaction('C5H10OOH2-3 => C5H10O2-3 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 3002
reaction('C5H10O2-3 + OH => C5H10OOH2-3', [0.000000e+00, 0.0, 0.0])

# Reaction 3003
reaction('C5H10OOH2-4 => C5H10O2-4 + OH', [7.500000e+10, 0.0, 15250.0])

# Reaction 3004
reaction('C5H10O2-4 + OH => C5H10OOH2-4', [0.000000e+00, 0.0, 0.0])

# Reaction 3005
reaction('C5H10OOH2-5 => C5H10O1-4 + OH', [9.375000e+09, 0.0, 6000.0])

# Reaction 3006
reaction('C5H10O1-4 + OH => C5H10OOH2-5', [0.000000e+00, 0.0, 0.0])

# Reaction 3007
reaction('C5H10OOH3-2 => C5H10O2-3 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 3008
reaction('C5H10O2-3 + OH => C5H10OOH3-2', [0.000000e+00, 0.0, 0.0])

# Reaction 3009
reaction('C5H10OOH3-1 => C5H10O1-3 + OH', [7.500000e+10, 0.0, 15250.0])

# Reaction 3010
reaction('C5H10O1-3 + OH => C5H10OOH3-1', [0.000000e+00, 0.0, 0.0])

# Reaction 3011
reaction('C5H10O1-2 + OH => CH2CO + NC3H7 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 3012
reaction('CH2CO + NC3H7 + H2O => C5H10O1-2 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3013
reaction('C5H10O1-3 + OH => C2H4 + C2H5CO + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 3014
reaction('C2H4 + C2H5CO + H2O => C5H10O1-3 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3015
reaction('C5H10O1-4 + OH => CH3COCH2 + C2H4 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 3016
reaction('CH3COCH2 + C2H4 + H2O => C5H10O1-4 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3017
reaction('C5H10O1-5 + OH => CH2CH2CHO + C2H4 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 3018
reaction('CH2CH2CHO + C2H4 + H2O => C5H10O1-5 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3019
reaction('C5H10O2-3 + OH => CH3CHCO + C2H5 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 3020
reaction('CH3CHCO + C2H5 + H2O => C5H10O2-3 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3021
reaction('C5H10O2-4 + OH => CH3CO + C3H6 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 3022
reaction('CH3CO + C3H6 + H2O => C5H10O2-4 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3023
reaction('C5H10O1-2 + OH => C2H3CHO + C2H5 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 3024
reaction('C2H3CHO + C2H5 + H2O => C5H10O1-2 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3025
reaction('C5H10O1-3 + OH => HCO + C4H8-1 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 3026
reaction('HCO + C4H8-1 + H2O => C5H10O1-3 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3027
reaction('C5H10O1-4 + OH => CH2CHO + C3H6 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 3028
reaction('CH2CHO + C3H6 + H2O => C5H10O1-4 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3029
reaction('C5H10O1-5 + OH => CH2O + C4H71-3 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 3030
reaction('CH2O + C4H71-3 + H2O => C5H10O1-5 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3031
reaction('C5H10O2-3 + OH => C2H3COCH3 + CH3 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 3032
reaction('C2H3COCH3 + CH3 + H2O => C5H10O2-3 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3033
reaction('C5H10O2-4 + OH => CH3CHO + C3H5-S + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 3034
reaction('CH3CHO + C3H5-S + H2O => C5H10O2-4 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3035
reaction('C5H10O1-2 + HO2 => CH2CO + NC3H7 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 3036
reaction('CH2CO + NC3H7 + H2O2 => C5H10O1-2 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3037
reaction('C5H10O1-3 + HO2 => C2H4 + C2H5CO + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 3038
reaction('C2H4 + C2H5CO + H2O2 => C5H10O1-3 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3039
reaction('C5H10O1-4 + HO2 => CH3COCH2 + C2H4 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 3040
reaction('CH3COCH2 + C2H4 + H2O2 => C5H10O1-4 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3041
reaction('C5H10O1-5 + HO2 => CH2CH2CHO + C2H4 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 3042
reaction('CH2CH2CHO + C2H4 + H2O2 => C5H10O1-5 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3043
reaction('C5H10O2-3 + HO2 => CH3CHCO + C2H5 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 3044
reaction('CH3CHCO + C2H5 + H2O2 => C5H10O2-3 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3045
reaction('C5H10O2-4 + HO2 => CH3CO + C3H6 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 3046
reaction('CH3CO + C3H6 + H2O2 => C5H10O2-4 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3047
reaction('C5H10O1-2 + HO2 => C2H3CHO + C2H5 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 3048
reaction('C2H3CHO + C2H5 + H2O2 => C5H10O1-2 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3049
reaction('C5H10O1-3 + HO2 => HCO + C4H8-1 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 3050
reaction('HCO + C4H8-1 + H2O2 => C5H10O1-3 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3051
reaction('C5H10O1-4 + HO2 => CH2CHO + C3H6 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 3052
reaction('CH2CHO + C3H6 + H2O2 => C5H10O1-4 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3053
reaction('C5H10O1-5 + HO2 => CH2O + C4H71-3 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 3054
reaction('CH2O + C4H71-3 + H2O2 => C5H10O1-5 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3055
reaction('C5H10O2-3 + HO2 => C2H3COCH3 + CH3 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 3056
reaction('C2H3COCH3 + CH3 + H2O2 => C5H10O2-3 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3057
reaction('C5H10O2-4 + HO2 => CH3CHO + C3H5-S + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 3058
reaction('CH3CHO + C3H5-S + H2O2 => C5H10O2-4 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3059
reaction('C5H10OOH1-2 => C5H10-1 + HO2', [3.449000e+17, -1.556, 17980.0])

# Reaction 3060
reaction('C5H10-1 + HO2 => C5H10OOH1-2', [1.000000e+11, 0.0, 11000.0])

# Reaction 3061
reaction('C5H10OOH2-1 => C5H10-1 + HO2', [6.462000e+19, -2.231, 21050.0])

# Reaction 3062
reaction('C5H10-1 + HO2 => C5H10OOH2-1', [1.000000e+11, 0.0, 11750.0])

# Reaction 3063
reaction('C5H10OOH2-3 => C5H10-2 + HO2', [9.651000e+19, -2.455, 20680.0])

# Reaction 3064
reaction('C5H10-2 + HO2 => C5H10OOH2-3', [1.000000e+11, 0.0, 11750.0])

# Reaction 3065
reaction('C5H10OOH3-2 => C5H10-2 + HO2', [9.651000e+19, -2.455, 20680.0])

# Reaction 3066
reaction('C5H10-2 + HO2 => C5H10OOH3-2', [1.000000e+11, 0.0, 11750.0])

# Reaction 3067
reaction('C5H10OOH1-3 => OH + CH2O + C4H8-1', [8.283000e+13, -0.17, 30090.0])

# Reaction 3068
reaction('OH + CH2O + C4H8-1 => C5H10OOH1-3', [0.000000e+00, 0.0, 0.0])

# Reaction 3069
reaction('C5H10OOH2-4 => OH + CH3CHO + C3H6', [8.312000e+17, -1.4, 27170.0])

# Reaction 3070
reaction('OH + CH3CHO + C3H6 => C5H10OOH2-4', [0.000000e+00, 0.0, 0.0])

# Reaction 3071
reaction('C5H10OOH3-1 => OH + C2H5CHO + C2H4', [3.096000e+18, -1.73, 26820.0])

# Reaction 3072
reaction('OH + C2H5CHO + C2H4 => C5H10OOH3-1', [0.000000e+00, 0.0, 0.0])

# Reaction 3073
reaction('C5H10OOH1-2O2 => C5H10OOH1-2 + O2', [8.039000e+22, -2.295, 37970.0])

# Reaction 3074
reaction('C5H10OOH1-2 + O2 => C5H10OOH1-2O2', [7.540000e+12, 0.0, 0.0])

# Reaction 3075
reaction('C5H10OOH1-3O2 => C5H10OOH1-3 + O2', [8.039000e+22, -2.295, 37970.0])

# Reaction 3076
reaction('C5H10OOH1-3 + O2 => C5H10OOH1-3O2', [7.540000e+12, 0.0, 0.0])

# Reaction 3077
reaction('C5H10OOH1-4O2 => C5H10OOH1-4 + O2', [8.039000e+22, -2.295, 37970.0])

# Reaction 3078
reaction('C5H10OOH1-4 + O2 => C5H10OOH1-4O2', [7.540000e+12, 0.0, 0.0])

# Reaction 3079
reaction('C5H10OOH1-5O2 => C5H10OOH1-5 + O2', [2.777000e+20, -1.623, 35690.0])

# Reaction 3080
reaction('C5H10OOH1-5 + O2 => C5H10OOH1-5O2', [4.520000e+12, 0.0, 0.0])

# Reaction 3081
reaction('C5H10OOH2-1O2 => C5H10OOH2-1 + O2', [2.572000e+20, -1.62, 35650.0])

# Reaction 3082
reaction('C5H10OOH2-1 + O2 => C5H10OOH2-1O2', [4.520000e+12, 0.0, 0.0])

# Reaction 3083
reaction('C5H10OOH2-3O2 => C5H10OOH2-3 + O2', [7.251000e+22, -2.29, 37910.0])

# Reaction 3084
reaction('C5H10OOH2-3 + O2 => C5H10OOH2-3O2', [7.540000e+12, 0.0, 0.0])

# Reaction 3085
reaction('C5H10OOH2-4O2 => C5H10OOH2-4 + O2', [7.251000e+22, -2.29, 37910.0])

# Reaction 3086
reaction('C5H10OOH2-4 + O2 => C5H10OOH2-4O2', [7.540000e+12, 0.0, 0.0])

# Reaction 3087
reaction('C5H10OOH2-5O2 => C5H10OOH2-5 + O2', [2.572000e+20, -1.62, 35650.0])

# Reaction 3088
reaction('C5H10OOH2-5 + O2 => C5H10OOH2-5O2', [4.520000e+12, 0.0, 0.0])

# Reaction 3089
reaction('C5H10OOH3-1O2 => C5H10OOH3-1 + O2', [2.572000e+20, -1.62, 35650.0])

# Reaction 3090
reaction('C5H10OOH3-1 + O2 => C5H10OOH3-1O2', [4.520000e+12, 0.0, 0.0])

# Reaction 3091
reaction('C5H10OOH3-2O2 => C5H10OOH3-2 + O2', [7.251000e+22, -2.29, 37910.0])

# Reaction 3092
reaction('C5H10OOH3-2 + O2 => C5H10OOH3-2O2', [7.540000e+12, 0.0, 0.0])

# Reaction 3093
reaction('C5H10OOH1-2O2 => NC5KET12 + OH', [2.000000e+11, 0.0, 26400.0])

# Reaction 3094
reaction('NC5KET12 + OH => C5H10OOH1-2O2', [2.780000e+04, 1.354, 48950.0])

# Reaction 3095
reaction('C5H10OOH1-3O2 => NC5KET13 + OH', [2.500000e+10, 0.0, 21400.0])

# Reaction 3096
reaction('NC5KET13 + OH => C5H10OOH1-3O2', [1.109000e+03, 1.533, 44740.0])

# Reaction 3097
reaction('C5H10OOH1-4O2 => NC5KET14 + OH', [3.125000e+09, 0.0, 19350.0])

# Reaction 3098
reaction('NC5KET14 + OH => C5H10OOH1-4O2', [1.386000e+02, 1.533, 42690.0])

# Reaction 3099
reaction('C5H10OOH1-5O2 => NC5KET15 + OH', [3.907000e+08, 0.0, 22550.0])

# Reaction 3100
reaction('NC5KET15 + OH => C5H10OOH1-5O2', [2.011000e+01, 1.504, 45880.0])

# Reaction 3101
reaction('C5H10OOH2-1O2 => NC5KET21 + OH', [1.000000e+11, 0.0, 23850.0])

# Reaction 3102
reaction('NC5KET21 + OH => C5H10OOH2-1O2', [2.785000e+02, 1.883, 50050.0])

# Reaction 3103
reaction('C5H10OOH2-3O2 => NC5KET23 + OH', [1.000000e+11, 0.0, 23850.0])

# Reaction 3104
reaction('NC5KET23 + OH => C5H10OOH2-3O2', [1.346000e+03, 1.773, 49120.0])

# Reaction 3105
reaction('C5H10OOH2-4O2 => NC5KET24 + OH', [1.250000e+10, 0.0, 17850.0])

# Reaction 3106
reaction('NC5KET24 + OH => C5H10OOH2-4O2', [5.624000e+01, 1.946, 43910.0])

# Reaction 3107
reaction('C5H10OOH2-5O2 => NC5KET25 + OH', [1.563000e+09, 0.0, 16050.0])

# Reaction 3108
reaction('NC5KET25 + OH => C5H10OOH2-5O2', [9.716000e+00, 1.893, 42120.0])

# Reaction 3109
reaction('C5H10OOH3-1O2 => NC5KET31 + OH', [1.250000e+10, 0.0, 17850.0])

# Reaction 3110
reaction('NC5KET31 + OH => C5H10OOH3-1O2', [6.548000e+01, 1.882, 44220.0])

# Reaction 3111
reaction('C5H10OOH3-2O2 => NC5KET32 + OH', [1.000000e+11, 0.0, 23850.0])

# Reaction 3112
reaction('NC5KET32 + OH => C5H10OOH3-2O2', [9.884000e+02, 1.781, 49390.0])

# Reaction 3113
reaction('NC5KET12 => NC3H7CHO + HCO + OH', [1.050000e+16, 0.0, 41600.0])

# Reaction 3114
reaction('NC3H7CHO + HCO + OH => NC5KET12', [0.000000e+00, 0.0, 0.0])

# Reaction 3115
reaction('NC5KET13 => C2H5CHO + CH2CHO + OH', [1.050000e+16, 0.0, 41600.0])

# Reaction 3116
reaction('C2H5CHO + CH2CHO + OH => NC5KET13', [0.000000e+00, 0.0, 0.0])

# Reaction 3117
reaction('NC5KET14 => CH3CHO + CH2CH2CHO + OH', [1.050000e+16, 0.0, 41600.0])

# Reaction 3118
reaction('CH3CHO + CH2CH2CHO + OH => NC5KET14', [0.000000e+00, 0.0, 0.0])

# Reaction 3119
reaction('NC5KET15 => CH2O + C3H6CHO-1 + OH', [1.500000e+16, 0.0, 42000.0])

# Reaction 3120
reaction('CH2O + C3H6CHO-1 + OH => NC5KET15', [0.000000e+00, 0.0, 0.0])

# Reaction 3121
reaction('NC5KET21 => CH2O + NC3H7CO + OH', [1.500000e+16, 0.0, 42000.0])

# Reaction 3122
reaction('CH2O + NC3H7CO + OH => NC5KET21', [0.000000e+00, 0.0, 0.0])

# Reaction 3123
reaction('NC5KET23 => C2H5CHO + CH3CO + OH', [1.500000e+16, 0.0, 42000.0])

# Reaction 3124
reaction('C2H5CHO + CH3CO + OH => NC5KET23', [0.000000e+00, 0.0, 0.0])

# Reaction 3125
reaction('NC5KET24 => CH3CHO + CH3COCH2 + OH', [1.500000e+16, 0.0, 42000.0])

# Reaction 3126
reaction('CH3CHO + CH3COCH2 + OH => NC5KET24', [0.000000e+00, 0.0, 0.0])

# Reaction 3127
reaction('NC5KET25 => CH2O + CH2CH2COCH3 + OH', [1.500000e+16, 0.0, 42000.0])

# Reaction 3128
reaction('CH2O + CH2CH2COCH3 + OH => NC5KET25', [0.000000e+00, 0.0, 0.0])

# Reaction 3129
reaction('NC5KET31 => CH2O + C2H5COCH2 + OH', [1.500000e+16, 0.0, 42000.0])

# Reaction 3130
reaction('CH2O + C2H5COCH2 + OH => NC5KET31', [0.000000e+00, 0.0, 0.0])

# Reaction 3131
reaction('NC5KET32 => CH3CHO + C2H5CO + OH', [1.050000e+16, 0.0, 41600.0])

# Reaction 3132
reaction('CH3CHO + C2H5CO + OH => NC5KET32', [0.000000e+00, 0.0, 0.0])

# Reaction 3133
reaction('C5H10OH-1 => C5H10-1 + OH', [1.500000e+13, 0.0, 25830.0])

# Reaction 3134
reaction('C5H10-1 + OH => C5H10OH-1', [4.750000e+12, 0.0, -782.0])

# Reaction 3135
reaction('C5H10OH-2 => C5H10-2 + OH', [1.500000e+13, 0.0, 25830.0])

# Reaction 3136
reaction('C5H10-2 + OH => C5H10OH-2', [4.750000e+12, 0.0, -782.0])

# Reaction 3137
reaction('O2C5H10OH-1 => C5H10OH-1 + O2', [1.058000e+21, -1.84, 37750.0])

# Reaction 3138
reaction('C5H10OH-1 + O2 => O2C5H10OH-1', [2.000000e+12, 0.0, 0.0])

# Reaction 3139
reaction('O2C5H10OH-1 => NC3H7CHO + CH2O + OH', [2.500000e+10, 0.0, 18860.0])

# Reaction 3140
reaction('NC3H7CHO + CH2O + OH => O2C5H10OH-1', [0.000000e+00, 0.0, 0.0])

# Reaction 3141
reaction('O2C5H10OH-2 => C5H10OH-2 + O2', [3.385000e+21, -2.01, 37870.0])

# Reaction 3142
reaction('C5H10OH-2 + O2 => O2C5H10OH-2', [2.000000e+12, 0.0, 0.0])

# Reaction 3143
reaction('O2C5H10OH-2 => C2H5CHO + CH3CHO + OH', [2.500000e+10, 0.0, 18860.0])

# Reaction 3144
reaction('C2H5CHO + CH3CHO + OH => O2C5H10OH-2', [0.000000e+00, 0.0, 0.0])

# Reaction 3145
reaction('NC6H14 => C5H11-1 + CH3', [1.608000e+22, -1.61, 89350.0])

# Reaction 3146
reaction('C5H11-1 + CH3 => NC6H14', [4.000000e+12, 0.0, -596.0])

# Reaction 3147
reaction('NC6H14 => NC3H7 + NC3H7', [2.169000e+24, -2.19, 87840.0])

# Reaction 3148
reaction('NC3H7 + NC3H7 => NC6H14', [4.000000e+12, 0.0, -596.0])

# Reaction 3149
reaction('NC6H14 => PC4H9 + C2H5', [1.617000e+24, -2.21, 88580.0])

# Reaction 3150
reaction('PC4H9 + C2H5 => NC6H14', [4.000000e+12, 0.0, -596.0])

# Reaction 3151
reaction('NC6H14 => C6H13-1 + H', [1.356000e+17, -0.36, 101200.0])

# Reaction 3152
reaction('C6H13-1 + H => NC6H14', [3.610000e+13, 0.0, 0.0])

# Reaction 3153
reaction('NC6H14 => C6H13-2 + H', [2.088000e+18, -0.7, 98710.0])

# Reaction 3154
reaction('C6H13-2 + H => NC6H14', [3.610000e+13, 0.0, 0.0])

# Reaction 3155
reaction('NC6H14 => C6H13-3 + H', [2.088000e+18, -0.7, 98710.0])

# Reaction 3156
reaction('C6H13-3 + H => NC6H14', [3.610000e+13, 0.0, 0.0])

# Reaction 3157
reaction('NC6H14 + H => C6H13-1 + H2', [1.880000e+05, 2.75, 6280.0])

# Reaction 3158
reaction('C6H13-1 + H2 => NC6H14 + H', [1.998000e+01, 3.38, 8615.0])

# Reaction 3159
reaction('NC6H14 + H => C6H13-2 + H2', [2.600000e+06, 2.4, 4471.0])

# Reaction 3160
reaction('C6H13-2 + H2 => NC6H14 + H', [1.795000e+01, 3.38, 9318.0])

# Reaction 3161
reaction('NC6H14 + H => C6H13-3 + H2', [2.600000e+06, 2.4, 4471.0])

# Reaction 3162
reaction('C6H13-3 + H2 => NC6H14 + H', [1.795000e+01, 3.38, 9318.0])

# Reaction 3163
reaction('NC6H14 + O => C6H13-1 + OH', [2.092000e+06, 2.42, 4766.0])

# Reaction 3164
reaction('C6H13-1 + OH => NC6H14 + O', [1.167000e+02, 3.04, 5689.0])

# Reaction 3165
reaction('NC6H14 + O => C6H13-2 + OH', [1.189000e+06, 2.44, 2846.0])

# Reaction 3166
reaction('C6H13-2 + OH => NC6H14 + O', [4.308000e+00, 3.4, 6281.0])

# Reaction 3167
reaction('NC6H14 + O => C6H13-3 + OH', [1.189000e+06, 2.44, 2846.0])

# Reaction 3168
reaction('C6H13-3 + OH => NC6H14 + O', [4.308000e+00, 3.4, 6281.0])

# Reaction 3169
reaction('NC6H14 + OH => C6H13-1 + H2O', [2.570000e+07, 1.8, 954.0])

# Reaction 3170
reaction('C6H13-1 + H2O => NC6H14 + OH', [2.906000e+04, 2.33, 18180.0])

# Reaction 3171
reaction('NC6H14 + OH => C6H13-2 + H2O', [4.900000e+06, 2.0, -596.0])

# Reaction 3172
reaction('C6H13-2 + H2O => NC6H14 + OH', [3.599000e+02, 2.87, 19140.0])

# Reaction 3173
reaction('NC6H14 + OH => C6H13-3 + H2O', [4.900000e+06, 2.0, -596.0])

# Reaction 3174
reaction('C6H13-3 + H2O => NC6H14 + OH', [3.599000e+02, 2.87, 19140.0])

# Reaction 3175
reaction('NC6H14 + CH3 => C6H13-1 + CH4', [9.040000e-01, 3.65, 7154.0])

# Reaction 3176
reaction('C6H13-1 + CH4 => NC6H14 + CH3', [8.770000e-02, 3.84, 11030.0])

# Reaction 3177
reaction('NC6H14 + CH3 => C6H13-3 + CH4', [1.680000e+05, 2.13, 7574.0])

# Reaction 3178
reaction('C6H13-3 + CH4 => NC6H14 + CH3', [1.058000e+03, 2.67, 13960.0])

# Reaction 3179
reaction('NC6H14 + CH3 => C6H13-2 + CH4', [1.680000e+05, 2.13, 7574.0])

# Reaction 3180
reaction('C6H13-2 + CH4 => NC6H14 + CH3', [1.058000e+03, 2.67, 13960.0])

# Reaction 3181
reaction('NC6H14 + HO2 => C6H13-1 + H2O2', [8.100000e+04, 2.5, 16690.0])

# Reaction 3182
reaction('C6H13-1 + H2O2 => NC6H14 + HO2', [4.983000e+03, 2.44, 3025.0])

# Reaction 3183
reaction('NC6H14 + HO2 => C6H13-2 + H2O2', [1.176000e+05, 2.5, 14860.0])

# Reaction 3184
reaction('C6H13-2 + H2O2 => NC6H14 + HO2', [4.699000e+02, 2.78, 3709.0])

# Reaction 3185
reaction('NC6H14 + HO2 => C6H13-3 + H2O2', [1.176000e+05, 2.5, 14860.0])

# Reaction 3186
reaction('C6H13-3 + H2O2 => NC6H14 + HO2', [4.699000e+02, 2.78, 3709.0])

# Reaction 3187
reaction('NC6H14 + CH3O => C6H13-1 + CH3OH', [3.160000e+11, 0.0, 7000.0])

# Reaction 3188
reaction('C6H13-1 + CH3OH => NC6H14 + CH3O', [1.200000e+10, 0.0, 9200.0])

# Reaction 3189
reaction('NC6H14 + CH3O => C6H13-2 + CH3OH', [2.190000e+11, 0.0, 5000.0])

# Reaction 3190
reaction('C6H13-2 + CH3OH => NC6H14 + CH3O', [8.910000e+09, 0.0, 7200.0])

# Reaction 3191
reaction('NC6H14 + CH3O => C6H13-3 + CH3OH', [2.190000e+11, 0.0, 5000.0])

# Reaction 3192
reaction('C6H13-3 + CH3OH => NC6H14 + CH3O', [8.910000e+09, 0.0, 7200.0])

# Reaction 3193
reaction('NC6H14 + O2 => C6H13-1 + HO2', [4.200000e+13, 0.0, 52800.0])

# Reaction 3194
reaction('C6H13-1 + HO2 => NC6H14 + O2', [2.342000e+10, 0.29, 451.0])

# Reaction 3195
reaction('NC6H14 + O2 => C6H13-2 + HO2', [2.800000e+13, 0.0, 50160.0])

# Reaction 3196
reaction('C6H13-2 + HO2 => NC6H14 + O2', [1.014000e+09, 0.63, 313.0])

# Reaction 3197
reaction('NC6H14 + O2 => C6H13-3 + HO2', [2.800000e+13, 0.0, 50160.0])

# Reaction 3198
reaction('C6H13-3 + HO2 => NC6H14 + O2', [1.014000e+09, 0.63, 313.0])

# Reaction 3199
reaction('NC6H14 + C2H5 => C6H13-1 + C2H6', [1.000000e+11, 0.0, 13400.0])

# Reaction 3200
reaction('C6H13-1 + C2H6 => NC6H14 + C2H5', [3.240000e+11, 0.0, 12300.0])

# Reaction 3201
reaction('NC6H14 + C2H5 => C6H13-2 + C2H6', [1.000000e+11, 0.0, 10400.0])

# Reaction 3202
reaction('C6H13-2 + C2H6 => NC6H14 + C2H5', [1.000000e+11, 0.0, 12900.0])

# Reaction 3203
reaction('NC6H14 + C2H5 => C6H13-3 + C2H6', [1.000000e+11, 0.0, 10400.0])

# Reaction 3204
reaction('C6H13-3 + C2H6 => NC6H14 + C2H5', [1.000000e+11, 0.0, 12900.0])

# Reaction 3205
reaction('NC6H14 + C2H3 => C6H13-1 + C2H4', [1.000000e+12, 0.0, 18000.0])

# Reaction 3206
reaction('C6H13-1 + C2H4 => NC6H14 + C2H3', [2.570000e+12, 0.0, 25400.0])

# Reaction 3207
reaction('NC6H14 + C2H3 => C6H13-2 + C2H4', [8.000000e+11, 0.0, 16800.0])

# Reaction 3208
reaction('C6H13-2 + C2H4 => NC6H14 + C2H3', [2.000000e+12, 0.0, 24200.0])

# Reaction 3209
reaction('NC6H14 + C2H3 => C6H13-3 + C2H4', [8.000000e+11, 0.0, 16800.0])

# Reaction 3210
reaction('C6H13-3 + C2H4 => NC6H14 + C2H3', [2.000000e+12, 0.0, 24200.0])

# Reaction 3211
reaction('NC6H14 + CH3O2 => C6H13-1 + CH3O2H', [8.100000e+04, 2.5, 16690.0])

# Reaction 3212
reaction('C6H13-1 + CH3O2H => NC6H14 + CH3O2', [9.695000e+04, 2.0, 1430.0])

# Reaction 3213
reaction('NC6H14 + CH3O2 => C6H13-2 + CH3O2H', [1.176000e+05, 2.5, 14860.0])

# Reaction 3214
reaction('C6H13-2 + CH3O2H => NC6H14 + CH3O2', [9.141000e+03, 2.34, 2114.0])

# Reaction 3215
reaction('NC6H14 + CH3O2 => C6H13-3 + CH3O2H', [1.176000e+05, 2.5, 14860.0])

# Reaction 3216
reaction('C6H13-3 + CH3O2H => NC6H14 + CH3O2', [9.141000e+03, 2.34, 2114.0])

# Reaction 3217
reaction('NC6H14 + O2CHO => C6H13-1 + HO2CHO', [1.680000e+13, 0.0, 20440.0])

# Reaction 3218
reaction('C6H13-1 + HO2CHO => NC6H14 + O2CHO', [5.692000e+02, 2.3, 3060.0])

# Reaction 3219
reaction('NC6H14 + O2CHO => C6H13-2 + HO2CHO', [1.120000e+13, 0.0, 17690.0])

# Reaction 3220
reaction('C6H13-2 + HO2CHO => NC6H14 + O2CHO', [2.464000e+01, 2.64, 2823.0])

# Reaction 3221
reaction('NC6H14 + O2CHO => C6H13-3 + HO2CHO', [1.120000e+13, 0.0, 17690.0])

# Reaction 3222
reaction('C6H13-3 + HO2CHO => NC6H14 + O2CHO', [2.464000e+01, 2.64, 2823.0])

# Reaction 3223
reaction('C6H13O2-1 + NC6H14 => C6H13O2H-1 + C6H13-1', [1.210000e+13, 0.0, 20430.0])

# Reaction 3224
reaction('C6H13O2H-1 + C6H13-1 => C6H13O2-1 + NC6H14', [1.440000e+10, 0.0, 15000.0])

# Reaction 3225
reaction('C6H13O2-2 + NC6H14 => C6H13O2H-2 + C6H13-1', [1.210000e+13, 0.0, 20430.0])

# Reaction 3226
reaction('C6H13O2H-2 + C6H13-1 => C6H13O2-2 + NC6H14', [1.440000e+10, 0.0, 15000.0])

# Reaction 3227
reaction('C6H13O2-3 + NC6H14 => C6H13O2H-3 + C6H13-1', [1.210000e+13, 0.0, 20430.0])

# Reaction 3228
reaction('C6H13O2H-3 + C6H13-1 => C6H13O2-3 + NC6H14', [1.440000e+10, 0.0, 15000.0])

# Reaction 3229
reaction('C6H13O2-1 + NC6H14 => C6H13O2H-1 + C6H13-2', [8.064000e+12, 0.0, 17700.0])

# Reaction 3230
reaction('C6H13O2H-1 + C6H13-2 => C6H13O2-1 + NC6H14', [1.440000e+10, 0.0, 15000.0])

# Reaction 3231
reaction('C6H13O2-2 + NC6H14 => C6H13O2H-2 + C6H13-2', [8.064000e+12, 0.0, 17700.0])

# Reaction 3232
reaction('C6H13O2H-2 + C6H13-2 => C6H13O2-2 + NC6H14', [1.440000e+10, 0.0, 15000.0])

# Reaction 3233
reaction('C6H13O2-3 + NC6H14 => C6H13O2H-3 + C6H13-2', [8.064000e+12, 0.0, 17700.0])

# Reaction 3234
reaction('C6H13O2H-3 + C6H13-2 => C6H13O2-3 + NC6H14', [1.440000e+10, 0.0, 15000.0])

# Reaction 3235
reaction('C6H13O2-1 + NC6H14 => C6H13O2H-1 + C6H13-3', [8.064000e+12, 0.0, 17700.0])

# Reaction 3236
reaction('C6H13O2H-1 + C6H13-3 => C6H13O2-1 + NC6H14', [1.440000e+10, 0.0, 15000.0])

# Reaction 3237
reaction('C6H13O2-2 + NC6H14 => C6H13O2H-2 + C6H13-3', [8.064000e+12, 0.0, 17700.0])

# Reaction 3238
reaction('C6H13O2H-2 + C6H13-3 => C6H13O2-2 + NC6H14', [1.440000e+10, 0.0, 15000.0])

# Reaction 3239
reaction('C6H13O2-3 + NC6H14 => C6H13O2H-3 + C6H13-3', [8.064000e+12, 0.0, 17700.0])

# Reaction 3240
reaction('C6H13O2H-3 + C6H13-3 => C6H13O2-3 + NC6H14', [1.440000e+10, 0.0, 15000.0])

# Reaction 3241
reaction('NC6H14 + C6H13-1 => NC6H14 + C6H13-2', [1.000000e+11, 0.0, 10400.0])

# Reaction 3242
reaction('NC6H14 + C6H13-2 => NC6H14 + C6H13-1', [1.500000e+11, 0.0, 12300.0])

# Reaction 3243
reaction('NC6H14 + C6H13-1 => NC6H14 + C6H13-3', [1.000000e+11, 0.0, 10400.0])

# Reaction 3244
reaction('NC6H14 + C6H13-3 => NC6H14 + C6H13-1', [1.500000e+11, 0.0, 12300.0])

# Reaction 3245
reaction('NC6H14 + C6H13-2 => NC6H14 + C6H13-3', [1.000000e+11, 0.0, 10400.0])

# Reaction 3246
reaction('NC6H14 + C6H13-3 => NC6H14 + C6H13-2', [1.000000e+11, 0.0, 10400.0])

# Reaction 3247
reaction('C6H13-1 + HO2 => C6H13O-1 + OH', [7.000000e+12, 0.0, -1000.0])

# Reaction 3248
reaction('C6H13O-1 + OH => C6H13-1 + HO2', [3.266000e+15, -0.56, 26150.0])

# Reaction 3249
reaction('C6H13-2 + HO2 => C6H13O-2 + OH', [7.000000e+12, 0.0, -1000.0])

# Reaction 3250
reaction('C6H13O-2 + OH => C6H13-2 + HO2', [3.073000e+17, -1.15, 28070.0])

# Reaction 3251
reaction('C6H13-3 + HO2 => C6H13O-3 + OH', [7.000000e+12, 0.0, -1000.0])

# Reaction 3252
reaction('C6H13O-3 + OH => C6H13-3 + HO2', [3.073000e+17, -1.15, 28070.0])

# Reaction 3253
reaction('C6H13-1 + CH3O2 => C6H13O-1 + CH3O', [7.000000e+12, 0.0, -1000.0])

# Reaction 3254
reaction('C6H13O-1 + CH3O => C6H13-1 + CH3O2', [2.043000e+14, -0.26, 30390.0])

# Reaction 3255
reaction('C6H13-2 + CH3O2 => C6H13O-2 + CH3O', [7.000000e+12, 0.0, -1000.0])

# Reaction 3256
reaction('C6H13O-2 + CH3O => C6H13-2 + CH3O2', [1.922000e+16, -0.85, 32310.0])

# Reaction 3257
reaction('C6H13-3 + CH3O2 => C6H13O-3 + CH3O', [7.000000e+12, 0.0, -1000.0])

# Reaction 3258
reaction('C6H13O-3 + CH3O => C6H13-3 + CH3O2', [1.922000e+16, -0.85, 32310.0])

# Reaction 3259
reaction('C6H13-1 + O2 => C6H12-1 + HO2', [3.000000e-19, 0.0, 3000.0])

# Reaction 3260
reaction('C6H12-1 + HO2 => C6H13-1 + O2', [2.000000e-19, 0.0, 17500.0])

# Reaction 3261
reaction('C6H13-2 + O2 => C6H12-1 + HO2', [4.500000e-19, 0.0, 5020.0])

# Reaction 3262
reaction('C6H12-1 + HO2 => C6H13-2 + O2', [2.000000e-19, 0.0, 17500.0])

# Reaction 3263
reaction('C6H13-2 + O2 => C6H12-2 + HO2', [3.000000e-19, 0.0, 3000.0])

# Reaction 3264
reaction('C6H12-2 + HO2 => C6H13-2 + O2', [2.000000e-19, 0.0, 17500.0])

# Reaction 3265
reaction('C6H13-3 + O2 => C6H12-2 + HO2', [3.000000e-19, 0.0, 3000.0])

# Reaction 3266
reaction('C6H12-2 + HO2 => C6H13-3 + O2', [2.000000e-19, 0.0, 17500.0])

# Reaction 3267
reaction('C6H13-3 + O2 => C6H12-3 + HO2', [3.000000e-19, 0.0, 3000.0])

# Reaction 3268
reaction('C6H12-3 + HO2 => C6H13-3 + O2', [2.000000e-19, 0.0, 17500.0])

# Reaction 3269
reaction('C6H13-1 => C2H4 + PC4H9', [6.391000e+19, -1.97, 30640.0])

# Reaction 3270
reaction('C2H4 + PC4H9 => C6H13-1', [3.300000e+11, 0.0, 7200.0])

# Reaction 3271
reaction('C6H13-1 => C6H12-1 + H', [9.621000e+13, -0.26, 36000.0])

# Reaction 3272
reaction('C6H12-1 + H => C6H13-1', [1.000000e+13, 0.0, 2900.0])

# Reaction 3273
reaction('C6H13-2 => C3H6 + NC3H7', [1.827000e+19, -1.8, 30170.0])

# Reaction 3274
reaction('C3H6 + NC3H7 => C6H13-2', [1.500000e+11, 0.0, 7200.0])

# Reaction 3275
reaction('C6H13-2 => C6H12-1 + H', [6.248000e+12, 0.09, 36820.0])

# Reaction 3276
reaction('C6H12-1 + H => C6H13-2', [1.000000e+13, 0.0, 1200.0])

# Reaction 3277
reaction('C6H13-2 => C6H12-2 + H', [1.271000e+13, -0.09, 35650.0])

# Reaction 3278
reaction('C6H12-2 + H => C6H13-2', [1.000000e+13, 0.0, 2900.0])

# Reaction 3279
reaction('C6H13-3 => C4H8-1 + C2H5', [1.202000e+19, -1.76, 30450.0])

# Reaction 3280
reaction('C4H8-1 + C2H5 => C6H13-3', [1.500000e+11, 0.0, 7200.0])

# Reaction 3281
reaction('C6H13-3 => C5H10-1 + CH3', [1.671000e+16, -0.93, 31480.0])

# Reaction 3282
reaction('C5H10-1 + CH3 => C6H13-3', [1.500000e+11, 0.0, 7200.0])

# Reaction 3283
reaction('C6H13-3 => C6H12-2 + H', [1.271000e+13, -0.09, 35650.0])

# Reaction 3284
reaction('C6H12-2 + H => C6H13-3', [1.000000e+13, 0.0, 2900.0])

# Reaction 3285
reaction('C6H13-3 => C6H12-3 + H', [4.659000e+12, -0.02, 35740.0])

# Reaction 3286
reaction('C6H12-3 + H => C6H13-3', [1.000000e+13, 0.0, 2900.0])

# Reaction 3287
reaction('C6H13-1 => C6H13-2', [5.478000e+08, 1.62, 38760.0])

# Reaction 3288
reaction('C6H13-2 => C6H13-1', [3.558000e+07, 1.97, 41270.0])

# Reaction 3289
reaction('C6H13-1 => C6H13-3', [1.495000e+09, 0.97, 33760.0])

# Reaction 3290
reaction('C6H13-3 => C6H13-1', [9.709000e+07, 1.31, 36270.0])

# Reaction 3291
reaction('C6H12-1 => NC3H7 + C3H5-A', [1.000000e+16, 0.0, 71000.0])

# Reaction 3292
reaction('NC3H7 + C3H5-A => C6H12-1', [1.000000e+13, 0.0, 0.0])

# Reaction 3293
reaction('C6H12-2 => C2H5 + C4H71-3', [1.000000e+16, 0.0, 71000.0])

# Reaction 3294
reaction('C2H5 + C4H71-3 => C6H12-2', [1.000000e+13, 0.0, 0.0])

# Reaction 3295
reaction('C6H12-3 => CH3 + C5H91-3', [1.000000e+16, 0.0, 71000.0])

# Reaction 3296
reaction('CH3 + C5H91-3 => C6H12-3', [1.000000e+13, 0.0, 0.0])

# Reaction 3297
reaction('C6H12-1 + OH => C5H11-1 + CH2O', [1.000000e+11, 0.0, -4000.0])

# Reaction 3298
reaction('C5H11-1 + CH2O => C6H12-1 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3299
reaction('C6H12-2 + OH => PC4H9 + CH3CHO', [1.000000e+11, 0.0, -4000.0])

# Reaction 3300
reaction('PC4H9 + CH3CHO => C6H12-2 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3301
reaction('C6H12-3 + OH => PC4H9 + CH3CHO', [1.000000e+11, 0.0, -4000.0])

# Reaction 3302
reaction('PC4H9 + CH3CHO => C6H12-3 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3303
reaction('C6H12-1 + O => C5H11-1 + HCO', [1.000000e+11, 0.0, -1050.0])

# Reaction 3304
reaction('C5H11-1 + HCO => C6H12-1 + O', [0.000000e+00, 0.0, 0.0])

# Reaction 3305
reaction('C6H12-2 + O => PC4H9 + CH3CO', [1.000000e+11, 0.0, -1050.0])

# Reaction 3306
reaction('PC4H9 + CH3CO => C6H12-2 + O', [0.000000e+00, 0.0, 0.0])

# Reaction 3307
reaction('C6H12-1 + H => C6H111-3 + H2', [3.376000e+05, 2.36, 207.0])

# Reaction 3308
reaction('C6H111-3 + H2 => C6H12-1 + H', [4.378000e+06, 2.1, 20330.0])

# Reaction 3309
reaction('C6H12-1 + H => C6H111-4 + H2', [1.300000e+06, 2.4, 4471.0])

# Reaction 3310
reaction('C6H111-4 + H2 => C6H12-1 + H', [3.896000e+03, 2.74, 11250.0])

# Reaction 3311
reaction('C6H12-1 + H => C6H111-5 + H2', [1.300000e+06, 2.4, 4471.0])

# Reaction 3312
reaction('C6H111-5 + H2 => C6H12-1 + H', [3.896000e+03, 2.74, 11250.0])

# Reaction 3313
reaction('C6H12-1 + H => C6H111-6 + H2', [6.650000e+05, 2.54, 6756.0])

# Reaction 3314
reaction('C6H111-6 + H2 => C6H12-1 + H', [3.120000e+04, 2.54, 11030.0])

# Reaction 3315
reaction('C6H12-1 + OH => C6H111-3 + H2O', [2.764000e+04, 2.64, -1919.0])

# Reaction 3316
reaction('C6H111-3 + H2O => C6H12-1 + OH', [1.552000e+06, 2.38, 33360.0])

# Reaction 3317
reaction('C6H12-1 + OH => C6H111-4 + H2O', [4.670000e+07, 1.61, -35.0])

# Reaction 3318
reaction('C6H111-4 + H2O => C6H12-1 + OH', [6.060000e+05, 1.95, 21900.0])

# Reaction 3319
reaction('C6H12-1 + OH => C6H111-5 + H2O', [4.670000e+07, 1.61, -35.0])

# Reaction 3320
reaction('C6H111-5 + H2O => C6H12-1 + OH', [6.060000e+05, 1.95, 21900.0])

# Reaction 3321
reaction('C6H12-1 + OH => C6H111-6 + H2O', [5.270000e+09, 0.97, 1586.0])

# Reaction 3322
reaction('C6H111-6 + H2O => C6H12-1 + OH', [1.071000e+09, 0.97, 21010.0])

# Reaction 3323
reaction('C6H12-1 + CH3 => C6H111-3 + CH4', [3.690000e+00, 3.31, 4002.0])

# Reaction 3324
reaction('C6H111-3 + CH4 => C6H12-1 + CH3', [1.250000e+03, 3.05, 24600.0])

# Reaction 3325
reaction('C6H12-1 + CH3 => C6H111-4 + CH4', [1.510000e+00, 3.46, 5481.0])

# Reaction 3326
reaction('C6H111-4 + CH4 => C6H12-1 + CH3', [1.182000e-01, 3.8, 12740.0])

# Reaction 3327
reaction('C6H12-1 + CH3 => C6H111-5 + CH4', [1.510000e+00, 3.46, 5481.0])

# Reaction 3328
reaction('C6H111-5 + CH4 => C6H12-1 + CH3', [1.182000e-01, 3.8, 12740.0])

# Reaction 3329
reaction('C6H12-1 + CH3 => C6H111-6 + CH4', [4.521000e-01, 3.65, 7154.0])

# Reaction 3330
reaction('C6H111-6 + CH4 => C6H12-1 + CH3', [5.541000e-01, 3.65, 11900.0])

# Reaction 3331
reaction('C6H12-1 + HO2 <=> C6H111-3 + H2O2', [9.640000e+03, 2.55, 10530.0])

# Reaction 3332
reaction('C6H12-1 + HO2 <=> C6H111-4 + H2O2', [9.640000e+03, 2.6, 13910.0])

# Reaction 3333
reaction('C6H12-1 + HO2 <=> C6H111-5 + H2O2', [9.640000e+03, 2.6, 13910.0])

# Reaction 3334
reaction('C6H12-1 + HO2 <=> C6H111-6 + H2O2', [2.380000e+04, 2.55, 16490.0])

# Reaction 3335
reaction('C6H12-1 + CH3O2 => C6H111-3 + CH3O2H', [4.820000e+03, 2.55, 10530.0])

# Reaction 3336
reaction('C6H111-3 + CH3O2H => C6H12-1 + CH3O2', [3.345000e+06, 1.79, 11320.0])

# Reaction 3337
reaction('C6H12-1 + CH3O2 => C6H111-4 + CH3O2H', [9.640000e+03, 2.6, 13910.0])

# Reaction 3338
reaction('C6H111-4 + CH3O2H => C6H12-1 + CH3O2', [1.546000e+03, 2.44, 1362.0])

# Reaction 3339
reaction('C6H12-1 + CH3O2 => C6H111-5 + CH3O2H', [9.640000e+03, 2.6, 13910.0])

# Reaction 3340
reaction('C6H111-5 + CH3O2H => C6H12-1 + CH3O2', [1.546000e+03, 2.44, 1362.0])

# Reaction 3341
reaction('C6H12-1 + CH3O2 => C6H111-6 + CH3O2H', [2.380000e+04, 2.55, 16490.0])

# Reaction 3342
reaction('C6H111-6 + CH3O2H => C6H12-1 + CH3O2', [5.976000e+04, 2.04, 1436.0])

# Reaction 3343
reaction('C6H12-1 + CH3O => C6H111-3 + CH3OH', [4.000000e+01, 2.9, 8609.0])

# Reaction 3344
reaction('C6H111-3 + CH3OH => C6H12-1 + CH3O', [2.501000e+02, 2.67, 27000.0])

# Reaction 3345
reaction('C6H12-1 + CH3O => C6H111-4 + CH3OH', [1.450000e+11, 0.0, 4571.0])

# Reaction 3346
reaction('C6H111-4 + CH3OH => C6H12-1 + CH3O', [2.095000e+08, 0.37, 9621.0])

# Reaction 3347
reaction('C6H12-1 + CH3O => C6H111-5 + CH3OH', [1.450000e+11, 0.0, 4571.0])

# Reaction 3348
reaction('C6H111-5 + CH3OH => C6H12-1 + CH3O', [2.095000e+08, 0.37, 9621.0])

# Reaction 3349
reaction('C6H12-1 + CH3O => C6H111-6 + CH3OH', [2.170000e+11, 0.0, 6458.0])

# Reaction 3350
reaction('C6H111-6 + CH3OH => C6H12-1 + CH3O', [4.908000e+09, 0.02, 8998.0])

# Reaction 3351
reaction('C6H12-2 + H => C6H111-3 + H2', [1.730000e+05, 2.5, 2492.0])

# Reaction 3352
reaction('C6H111-3 + H2 => C6H12-2 + H', [4.563000e+06, 2.06, 19740.0])

# Reaction 3353
reaction('C6H12-2 + H => C6H112-4 + H2', [3.376000e+05, 2.36, 207.0])

# Reaction 3354
reaction('C6H112-4 + H2 => C6H12-2 + H', [4.384000e+06, 2.1, 20330.0])

# Reaction 3355
reaction('C6H12-2 + H => C6H112-5 + H2', [1.300000e+06, 2.4, 4471.0])

# Reaction 3356
reaction('C6H112-5 + H2 => C6H12-2 + H', [3.921000e+03, 2.74, 11250.0])

# Reaction 3357
reaction('C6H12-2 + H => C6H112-6 + H2', [6.651000e+05, 2.54, 6756.0])

# Reaction 3358
reaction('C6H112-6 + H2 => C6H12-2 + H', [3.122000e+04, 2.53, 11030.0])

# Reaction 3359
reaction('C6H12-2 + OH => C6H111-3 + H2O', [3.120000e+06, 2.0, -298.0])

# Reaction 3360
reaction('C6H111-3 + H2O => C6H12-2 + OH', [3.563000e+08, 1.56, 32110.0])

# Reaction 3361
reaction('C6H12-2 + OH => C6H112-4 + H2O', [2.764000e+04, 2.64, -1919.0])

# Reaction 3362
reaction('C6H112-4 + H2O => C6H12-2 + OH', [1.554000e+06, 2.38, 33360.0])

# Reaction 3363
reaction('C6H12-2 + OH => C6H112-5 + H2O', [4.670000e+05, 1.61, -35.0])

# Reaction 3364
reaction('C6H112-5 + H2O => C6H12-2 + OH', [6.099000e+03, 1.95, 21900.0])

# Reaction 3365
reaction('C6H12-2 + OH => C6H112-6 + H2O', [5.270000e+09, 0.97, 1586.0])

# Reaction 3366
reaction('C6H112-6 + H2O => C6H12-2 + OH', [1.071000e+09, 0.96, 21010.0])

# Reaction 3367
reaction('C6H12-2 + CH3 => C6H111-3 + CH4', [2.210000e+00, 3.5, 5675.0])

# Reaction 3368
reaction('C6H111-3 + CH4 => C6H12-2 + CH3', [1.523000e+03, 3.06, 23410.0])

# Reaction 3369
reaction('C6H12-2 + CH3 => C6H112-4 + CH4', [3.690000e+00, 3.31, 4002.0])

# Reaction 3370
reaction('C6H112-4 + CH4 => C6H12-2 + CH3', [1.252000e+03, 3.05, 24600.0])

# Reaction 3371
reaction('C6H12-2 + CH3 => C6H112-5 + CH4', [1.510000e+00, 3.46, 5481.0])

# Reaction 3372
reaction('C6H112-5 + CH4 => C6H12-2 + CH3', [1.190000e-01, 3.8, 12740.0])

# Reaction 3373
reaction('C6H12-2 + CH3 => C6H112-6 + CH4', [4.521000e-01, 3.65, 7154.0])

# Reaction 3374
reaction('C6H112-6 + CH4 => C6H12-2 + CH3', [5.543000e-01, 3.64, 11900.0])

# Reaction 3375
reaction('C6H12-2 + HO2 => C6H111-3 + H2O2', [9.639000e+03, 2.6, 13910.0])

# Reaction 3376
reaction('C6H111-3 + H2O2 => C6H12-2 + HO2', [6.534000e+06, 1.84, 14850.0])

# Reaction 3377
reaction('C6H12-2 + HO2 => C6H112-4 + H2O2', [4.820000e+03, 2.55, 10530.0])

# Reaction 3378
reaction('C6H112-4 + H2O2 => C6H12-2 + HO2', [1.609000e+06, 1.96, 14340.0])

# Reaction 3379
reaction('C6H12-2 + HO2 => C6H112-5 + H2O2', [9.640000e+03, 2.6, 13910.0])

# Reaction 3380
reaction('C6H112-5 + H2O2 => C6H12-2 + HO2', [7.473000e+02, 2.61, 4382.0])

# Reaction 3381
reaction('C6H12-2 + HO2 => C6H112-6 + H2O2', [2.380000e+04, 2.55, 16490.0])

# Reaction 3382
reaction('C6H112-6 + H2O2 => C6H12-2 + HO2', [2.871000e+04, 2.22, 4456.0])

# Reaction 3383
reaction('C6H12-2 + CH3O2 => C6H111-3 + CH3O2H', [9.639000e+03, 2.6, 13910.0])

# Reaction 3384
reaction('C6H111-3 + CH3O2H => C6H12-2 + CH3O2', [1.361000e+07, 1.66, 11830.0])

# Reaction 3385
reaction('C6H12-2 + CH3O2 => C6H112-4 + CH3O2H', [4.820000e+03, 2.55, 10530.0])

# Reaction 3386
reaction('C6H112-4 + CH3O2H => C6H12-2 + CH3O2', [3.350000e+06, 1.79, 11320.0])

# Reaction 3387
reaction('C6H12-2 + CH3O2 => C6H112-5 + CH3O2H', [9.640000e+03, 2.6, 13910.0])

# Reaction 3388
reaction('C6H112-5 + CH3O2H => C6H12-2 + CH3O2', [1.556000e+03, 2.44, 1362.0])

# Reaction 3389
reaction('C6H12-2 + CH3O2 => C6H112-6 + CH3O2H', [2.380000e+04, 2.55, 16490.0])

# Reaction 3390
reaction('C6H112-6 + CH3O2H => C6H12-2 + CH3O2', [5.978000e+04, 2.04, 1436.0])

# Reaction 3391
reaction('C6H12-2 + CH3O => C6H111-3 + CH3OH', [9.000000e+01, 2.95, 11990.0])

# Reaction 3392
reaction('C6H111-3 + CH3OH => C6H12-2 + CH3O', [1.144000e+03, 2.54, 27510.0])

# Reaction 3393
reaction('C6H12-2 + CH3O => C6H112-4 + CH3OH', [4.000000e+01, 2.9, 8609.0])

# Reaction 3394
reaction('C6H112-4 + CH3OH => C6H12-2 + CH3O', [2.504000e+02, 2.66, 27000.0])

# Reaction 3395
reaction('C6H12-2 + CH3O => C6H112-5 + CH3OH', [1.450000e+11, 0.0, 4571.0])

# Reaction 3396
reaction('C6H112-5 + CH3OH => C6H12-2 + CH3O', [2.109000e+08, 0.37, 9621.0])

# Reaction 3397
reaction('C6H12-2 + CH3O => C6H112-6 + CH3OH', [2.170000e+11, 0.0, 6458.0])

# Reaction 3398
reaction('C6H112-6 + CH3OH => C6H12-2 + CH3O', [4.910000e+09, 0.02, 8998.0])

# Reaction 3399
reaction('C6H12-3 + H => C6H113-1 + H2', [1.330000e+06, 2.54, 6756.0])

# Reaction 3400
reaction('C6H113-1 + H2 => C6H12-3 + H', [3.078000e+04, 2.54, 11030.0])

# Reaction 3401
reaction('C6H12-3 + H => C6H112-4 + H2', [6.752000e+05, 2.36, 207.0])

# Reaction 3402
reaction('C6H112-4 + H2 => C6H12-3 + H', [3.215000e+06, 2.17, 20420.0])

# Reaction 3403
reaction('C6H12-3 + OH => C6H113-1 + H2O', [1.054000e+10, 0.97, 1586.0])

# Reaction 3404
reaction('C6H113-1 + H2O => C6H12-3 + OH', [1.056000e+09, 0.97, 21010.0])

# Reaction 3405
reaction('C6H12-3 + OH => C6H112-4 + H2O', [5.528000e+04, 2.64, -1919.0])

# Reaction 3406
reaction('C6H112-4 + H2O => C6H12-3 + OH', [1.140000e+06, 2.45, 33450.0])

# Reaction 3407
reaction('C6H12-3 + CH3 => C6H113-1 + CH4', [9.042000e-01, 3.65, 7154.0])

# Reaction 3408
reaction('C6H113-1 + CH4 => C6H12-3 + CH3', [5.467000e-01, 3.65, 11900.0])

# Reaction 3409
reaction('C6H12-3 + CH3 => C6H112-4 + CH4', [7.380000e+00, 3.31, 4002.0])

# Reaction 3410
reaction('C6H112-4 + CH4 => C6H12-3 + CH3', [9.179000e+02, 3.12, 24690.0])

# Reaction 3411
reaction('C6H12-3 + HO2 => C6H113-1 + H2O2', [4.760000e+04, 2.55, 16490.0])

# Reaction 3412
reaction('C6H113-1 + H2O2 => C6H12-3 + HO2', [2.832000e+04, 2.22, 4456.0])

# Reaction 3413
reaction('C6H12-3 + HO2 => C6H112-4 + H2O2', [9.640000e+03, 2.55, 10530.0])

# Reaction 3414
reaction('C6H112-4 + H2O2 => C6H12-3 + HO2', [1.180000e+06, 2.03, 14430.0])

# Reaction 3415
reaction('C6H12-3 + CH3O2 => C6H113-1 + CH3O2H', [4.760000e+04, 2.55, 16490.0])

# Reaction 3416
reaction('C6H113-1 + CH3O2H => C6H12-3 + CH3O2', [5.896000e+04, 2.04, 1436.0])

# Reaction 3417
reaction('C6H12-3 + CH3O2 => C6H112-4 + CH3O2H', [9.640000e+03, 2.55, 10530.0])

# Reaction 3418
reaction('C6H112-4 + CH3O2H => C6H12-3 + CH3O2', [2.456000e+06, 1.85, 11410.0])

# Reaction 3419
reaction('C6H12-3 + CH3O => C6H113-1 + CH3OH', [4.340000e+11, 0.0, 6458.0])

# Reaction 3420
reaction('C6H113-1 + CH3OH => C6H12-3 + CH3O', [4.843000e+09, 0.02, 8998.0])

# Reaction 3421
reaction('C6H12-3 + CH3O => C6H112-4 + CH3OH', [8.000000e+01, 2.9, 8609.0])

# Reaction 3422
reaction('C6H112-4 + CH3OH => C6H12-3 + CH3O', [1.836000e+02, 2.73, 27090.0])

# Reaction 3423
reaction('C6H111-3 + HO2 => C6H11O1-3 + OH', [9.640000e+12, 0.0, 0.0])

# Reaction 3424
reaction('C6H11O1-3 + OH => C6H111-3 + HO2', [1.070000e+16, -1.14, 15800.0])

# Reaction 3425
reaction('C6H111-3 + CH3O2 => C6H11O1-3 + CH3O', [9.640000e+12, 0.0, 0.0])

# Reaction 3426
reaction('C6H11O1-3 + CH3O => C6H111-3 + CH3O2', [1.045000e+18, -1.72, 20570.0])

# Reaction 3427
reaction('C6H111-3 + C2H5O2 => C6H11O1-3 + C2H5O', [9.640000e+12, 0.0, 0.0])

# Reaction 3428
reaction('C6H11O1-3 + C2H5O => C6H111-3 + C2H5O2', [6.844000e+14, -0.8, 18410.0])

# Reaction 3429
reaction('C6H111-6 => C6H111-3', [1.113000e+12, 0.0, 31700.0])

# Reaction 3430
reaction('C6H111-3 => C6H111-6', [3.076000e+14, -0.25, 47550.0])

# Reaction 3431
reaction('C6H112-4 + HO2 => C6H11O2-4 + OH', [9.640000e+12, 0.0, 0.0])

# Reaction 3432
reaction('C6H11O2-4 + OH => C6H112-4 + HO2', [9.704000e+15, -1.13, 15790.0])

# Reaction 3433
reaction('C6H112-4 + CH3O2 => C6H11O2-4 + CH3O', [9.640000e+12, 0.0, 0.0])

# Reaction 3434
reaction('C6H11O2-4 + CH3O => C6H112-4 + CH3O2', [9.478000e+17, -1.71, 20560.0])

# Reaction 3435
reaction('C6H112-4 + C2H5O2 => C6H11O2-4 + C2H5O', [9.640000e+12, 0.0, 0.0])

# Reaction 3436
reaction('C6H11O2-4 + C2H5O => C6H112-4 + C2H5O2', [6.204000e+14, -0.78, 18400.0])

# Reaction 3437
reaction('C6H11O1-3 => C2H3CHO + NC3H7', [1.427000e+20, -2.04, 11230.0])

# Reaction 3438
reaction('C2H3CHO + NC3H7 => C6H11O1-3', [1.000000e+11, 0.0, 9600.0])

# Reaction 3439
reaction('C6H11O1-3 => NC3H7CHO + C2H3', [2.400000e+18, -1.63, 23410.0])

# Reaction 3440
reaction('NC3H7CHO + C2H3 => C6H11O1-3', [1.000000e+11, 0.0, 9600.0])

# Reaction 3441
reaction('C6H11O2-4 => SC3H5CHO + C2H5', [5.449000e+19, -1.92, 10760.0])

# Reaction 3442
reaction('SC3H5CHO + C2H5 => C6H11O2-4', [1.000000e+11, 0.0, 9600.0])

# Reaction 3443
reaction('C6H11O2-4 => C2H5CHO + C3H5-S', [2.346000e+22, -2.58, 29310.0])

# Reaction 3444
reaction('C2H5CHO + C3H5-S => C6H11O2-4', [1.000000e+11, 0.0, 9600.0])

# Reaction 3445
reaction('C6H13O2-1 => C6H13-1 + O2', [5.150000e+20, -1.71, 35790.0])

# Reaction 3446
reaction('C6H13-1 + O2 => C6H13O2-1', [4.520000e+12, 0.0, 0.0])

# Reaction 3447
reaction('C6H13O2-2 => C6H13-2 + O2', [2.180000e+23, -2.33, 38040.0])

# Reaction 3448
reaction('C6H13-2 + O2 => C6H13O2-2', [7.540000e+12, 0.0, 0.0])

# Reaction 3449
reaction('C6H13O2-3 => C6H13-3 + O2', [2.180000e+23, -2.33, 38040.0])

# Reaction 3450
reaction('C6H13-3 + O2 => C6H13O2-3', [7.540000e+12, 0.0, 0.0])

# Reaction 3451
reaction('C6H13-1 + C6H13O2-1 => C6H13O-1 + C6H13O-1', [7.000000e+12, 0.0, -1000.0])

# Reaction 3452
reaction('C6H13O-1 + C6H13O-1 => C6H13-1 + C6H13O2-1', [9.781000e+13, -0.21, 29840.0])

# Reaction 3453
reaction('C6H13-1 + C6H13O2-2 => C6H13O-1 + C6H13O-2', [7.000000e+12, 0.0, -1000.0])

# Reaction 3454
reaction('C6H13O-1 + C6H13O-2 => C6H13-1 + C6H13O2-2', [3.628000e+13, -0.17, 29500.0])

# Reaction 3455
reaction('C6H13-1 + C6H13O2-3 => C6H13O-1 + C6H13O-3', [7.000000e+12, 0.0, -1000.0])

# Reaction 3456
reaction('C6H13O-1 + C6H13O-3 => C6H13-1 + C6H13O2-3', [3.628000e+13, -0.17, 29500.0])

# Reaction 3457
reaction('C6H13-2 + C6H13O2-1 => C6H13O-2 + C6H13O-1', [7.000000e+12, 0.0, -1000.0])

# Reaction 3458
reaction('C6H13O-2 + C6H13O-1 => C6H13-2 + C6H13O2-1', [9.205000e+15, -0.79, 31760.0])

# Reaction 3459
reaction('C6H13-2 + C6H13O2-2 => C6H13O-2 + C6H13O-2', [7.000000e+12, 0.0, -1000.0])

# Reaction 3460
reaction('C6H13O-2 + C6H13O-2 => C6H13-2 + C6H13O2-2', [3.414000e+15, -0.76, 31420.0])

# Reaction 3461
reaction('C6H13-2 + C6H13O2-3 => C6H13O-2 + C6H13O-3', [7.000000e+12, 0.0, -1000.0])

# Reaction 3462
reaction('C6H13O-2 + C6H13O-3 => C6H13-2 + C6H13O2-3', [3.414000e+15, -0.76, 31420.0])

# Reaction 3463
reaction('C6H13-3 + C6H13O2-1 => C6H13O-3 + C6H13O-1', [7.000000e+12, 0.0, -1000.0])

# Reaction 3464
reaction('C6H13O-3 + C6H13O-1 => C6H13-3 + C6H13O2-1', [9.205000e+15, -0.79, 31760.0])

# Reaction 3465
reaction('C6H13-3 + C6H13O2-2 => C6H13O-3 + C6H13O-2', [7.000000e+12, 0.0, -1000.0])

# Reaction 3466
reaction('C6H13O-3 + C6H13O-2 => C6H13-3 + C6H13O2-2', [3.414000e+15, -0.76, 31420.0])

# Reaction 3467
reaction('C6H13-3 + C6H13O2-3 => C6H13O-3 + C6H13O-3', [7.000000e+12, 0.0, -1000.0])

# Reaction 3468
reaction('C6H13O-3 + C6H13O-3 => C6H13-3 + C6H13O2-3', [3.414000e+15, -0.76, 31420.0])

# Reaction 3469
reaction('C6H13O2-1 => C6H12-1 + HO2', [5.044000e+38, -8.11, 40490.0])

# Reaction 3470
reaction('C6H12-1 + HO2 => C6H13O2-1', [9.638000e+29, -6.22, 20470.0])

# Reaction 3471
reaction('C6H13O2-2 => C6H12-1 + HO2', [5.075000e+42, -9.41, 41490.0])

# Reaction 3472
reaction('C6H12-1 + HO2 => C6H13O2-2', [5.885000e+32, -7.24, 16700.0])

# Reaction 3473
reaction('C6H13O2-2 => C6H12-2 + HO2', [5.044000e+38, -8.11, 40490.0])

# Reaction 3474
reaction('C6H12-2 + HO2 => C6H13O2-2', [2.876000e+28, -5.77, 18570.0])

# Reaction 3475
reaction('C6H13O2-3 => C6H12-3 + HO2', [1.009000e+39, -8.11, 40490.0])

# Reaction 3476
reaction('C6H12-3 + HO2 => C6H13O2-3', [1.569000e+29, -5.83, 18480.0])

# Reaction 3477
reaction('C6H13O2-1 => C6H12OOH1-2', [2.000000e+11, 0.0, 26850.0])

# Reaction 3478
reaction('C6H12OOH1-2 => C6H13O2-1', [2.402000e+10, -0.12, 13860.0])

# Reaction 3479
reaction('C6H13O2-1 => C6H12OOH1-3', [2.500000e+10, 0.0, 20850.0])

# Reaction 3480
reaction('C6H12OOH1-3 => C6H13O2-1', [3.003000e+09, -0.12, 7860.0])

# Reaction 3481
reaction('C6H13O2-1 => C6H12OOH1-4', [3.125000e+09, 0.0, 19050.0])

# Reaction 3482
reaction('C6H12OOH1-4 => C6H13O2-1', [3.754000e+08, -0.12, 6060.0])

# Reaction 3483
reaction('C6H13O2-1 => C6H12OOH1-5', [3.906000e+08, 0.0, 22050.0])

# Reaction 3484
reaction('C6H12OOH1-5 => C6H13O2-1', [4.692000e+07, -0.12, 9060.0])

# Reaction 3485
reaction('C6H13O2-2 => C6H12OOH2-1', [3.000000e+11, 0.0, 29400.0])

# Reaction 3486
reaction('C6H12OOH2-1 => C6H13O2-2', [7.948000e+11, -0.52, 13950.0])

# Reaction 3487
reaction('C6H13O2-2 => C6H12OOH2-3', [2.000000e+11, 0.0, 26850.0])

# Reaction 3488
reaction('C6H12OOH2-3 => C6H13O2-2', [2.578000e+10, -0.13, 13880.0])

# Reaction 3489
reaction('C6H13O2-2 => C6H12OOH2-4', [2.500000e+10, 0.0, 20850.0])

# Reaction 3490
reaction('C6H12OOH2-4 => C6H13O2-2', [3.223000e+09, -0.13, 7880.0])

# Reaction 3491
reaction('C6H13O2-2 => C6H12OOH2-5', [3.125000e+09, 0.0, 19050.0])

# Reaction 3492
reaction('C6H12OOH2-5 => C6H13O2-2', [4.028000e+08, -0.13, 6080.0])

# Reaction 3493
reaction('C6H13O2-2 => C6H12OOH2-6', [5.860000e+08, 0.0, 25550.0])

# Reaction 3494
reaction('C6H12OOH2-6 => C6H13O2-2', [1.553000e+09, -0.52, 10100.0])

# Reaction 3495
reaction('C6H13O2-3 => C6H12OOH3-1', [3.750000e+10, 0.0, 24400.0])

# Reaction 3496
reaction('C6H12OOH3-1 => C6H13O2-3', [9.935000e+10, -0.52, 8950.0])

# Reaction 3497
reaction('C6H13O2-3 => C6H12OOH3-2', [2.000000e+11, 0.0, 26850.0])

# Reaction 3498
reaction('C6H12OOH3-2 => C6H13O2-3', [2.578000e+10, -0.13, 13880.0])

# Reaction 3499
reaction('C6H13O2-3 => C6H12OOH3-4', [2.000000e+11, 0.0, 26850.0])

# Reaction 3500
reaction('C6H12OOH3-4 => C6H13O2-3', [2.578000e+10, -0.13, 13880.0])

# Reaction 3501
reaction('C6H13O2-3 => C6H12OOH3-5', [2.500000e+10, 0.0, 20850.0])

# Reaction 3502
reaction('C6H12OOH3-5 => C6H13O2-3', [3.223000e+09, -0.13, 7880.0])

# Reaction 3503
reaction('C6H13O2-3 => C6H12OOH3-6', [4.688000e+09, 0.0, 22350.0])

# Reaction 3504
reaction('C6H12OOH3-6 => C6H13O2-3', [1.242000e+10, -0.52, 6900.0])

# Reaction 3505
reaction('C6H13O2-1 + HO2 => C6H13O2H-1 + O2', [1.750000e+10, 0.0, -3275.0])

# Reaction 3506
reaction('C6H13O2H-1 + O2 => C6H13O2-1 + HO2', [3.538000e+13, -0.78, 33600.0])

# Reaction 3507
reaction('C6H13O2-2 + HO2 => C6H13O2H-2 + O2', [1.750000e+10, 0.0, -3275.0])

# Reaction 3508
reaction('C6H13O2H-2 + O2 => C6H13O2-2 + HO2', [4.447000e+13, -0.81, 33640.0])

# Reaction 3509
reaction('C6H13O2-3 + HO2 => C6H13O2H-3 + O2', [1.750000e+10, 0.0, -3275.0])

# Reaction 3510
reaction('C6H13O2H-3 + O2 => C6H13O2-3 + HO2', [4.447000e+13, -0.81, 33640.0])

# Reaction 3511
reaction('C6H13O2-1 + H2O2 => C6H13O2H-1 + HO2', [2.400000e+12, 0.0, 10000.0])

# Reaction 3512
reaction('C6H13O2H-1 + HO2 => C6H13O2-1 + H2O2', [2.400000e+12, 0.0, 10000.0])

# Reaction 3513
reaction('C6H13O2-2 + H2O2 => C6H13O2H-2 + HO2', [2.400000e+12, 0.0, 10000.0])

# Reaction 3514
reaction('C6H13O2H-2 + HO2 => C6H13O2-2 + H2O2', [2.400000e+12, 0.0, 10000.0])

# Reaction 3515
reaction('C6H13O2-3 + H2O2 => C6H13O2H-3 + HO2', [2.400000e+12, 0.0, 10000.0])

# Reaction 3516
reaction('C6H13O2H-3 + HO2 => C6H13O2-3 + H2O2', [2.400000e+12, 0.0, 10000.0])

# Reaction 3517
reaction('C6H13O2-1 + CH3O2 => C6H13O-1 + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 3518
reaction('C6H13O-1 + CH3O + O2 => C6H13O2-1 + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 3519
reaction('C6H13O2-2 + CH3O2 => C6H13O-2 + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 3520
reaction('C6H13O-2 + CH3O + O2 => C6H13O2-2 + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 3521
reaction('C6H13O2-3 + CH3O2 => C6H13O-3 + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 3522
reaction('C6H13O-3 + CH3O + O2 => C6H13O2-3 + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 3523
reaction('C6H13O2-1 + C6H13O2-1 => O2 + C6H13O-1 + C6H13O-1', [1.400000e+16, -1.61, 1860.0])

# Reaction 3524
reaction('O2 + C6H13O-1 + C6H13O-1 => C6H13O2-1 + C6H13O2-1', [0.000000e+00, 0.0, 0.0])

# Reaction 3525
reaction('C6H13O2-1 + C6H13O2-2 => O2 + C6H13O-1 + C6H13O-2', [1.400000e+16, -1.61, 1860.0])

# Reaction 3526
reaction('O2 + C6H13O-1 + C6H13O-2 => C6H13O2-1 + C6H13O2-2', [0.000000e+00, 0.0, 0.0])

# Reaction 3527
reaction('C6H13O2-1 + C6H13O2-3 => O2 + C6H13O-1 + C6H13O-3', [1.400000e+16, -1.61, 1860.0])

# Reaction 3528
reaction('O2 + C6H13O-1 + C6H13O-3 => C6H13O2-1 + C6H13O2-3', [0.000000e+00, 0.0, 0.0])

# Reaction 3529
reaction('C6H13O2-2 + C6H13O2-2 => O2 + C6H13O-2 + C6H13O-2', [1.400000e+16, -1.61, 1860.0])

# Reaction 3530
reaction('O2 + C6H13O-2 + C6H13O-2 => C6H13O2-2 + C6H13O2-2', [0.000000e+00, 0.0, 0.0])

# Reaction 3531
reaction('C6H13O2-2 + C6H13O2-3 => O2 + C6H13O-2 + C6H13O-3', [1.400000e+16, -1.61, 1860.0])

# Reaction 3532
reaction('O2 + C6H13O-2 + C6H13O-3 => C6H13O2-2 + C6H13O2-3', [0.000000e+00, 0.0, 0.0])

# Reaction 3533
reaction('C6H13O2-3 + C6H13O2-3 => O2 + C6H13O-3 + C6H13O-3', [1.400000e+16, -1.61, 1860.0])

# Reaction 3534
reaction('O2 + C6H13O-3 + C6H13O-3 => C6H13O2-3 + C6H13O2-3', [0.000000e+00, 0.0, 0.0])

# Reaction 3535
reaction('C6H13O2H-1 => C6H13O-1 + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 3536
reaction('C6H13O-1 + OH => C6H13O2H-1', [2.025000e+07, 1.93, -6512.0])

# Reaction 3537
reaction('C6H13O2H-2 => C6H13O-2 + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 3538
reaction('C6H13O-2 + OH => C6H13O2H-2', [5.977000e+06, 2.0, -6892.0])

# Reaction 3539
reaction('C6H13O2H-3 => C6H13O-3 + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 3540
reaction('C6H13O-3 + OH => C6H13O2H-3', [5.977000e+06, 2.0, -6892.0])

# Reaction 3541
reaction('C6H13O-1 => C5H11-1 + CH2O', [5.226000e+20, -2.08, 24830.0])

# Reaction 3542
reaction('C5H11-1 + CH2O => C6H13O-1', [1.000000e+11, 0.0, 11900.0])

# Reaction 3543
reaction('C6H13O-2 => PC4H9 + CH3CHO', [2.365000e+22, -2.61, 19620.0])

# Reaction 3544
reaction('PC4H9 + CH3CHO => C6H13O-2', [1.000000e+11, 0.0, 11900.0])

# Reaction 3545
reaction('C6H13O-3 => C2H5CHO + NC3H7', [8.045000e+17, -1.17, 18170.0])

# Reaction 3546
reaction('C2H5CHO + NC3H7 => C6H13O-3', [1.000000e+11, 0.0, 11900.0])

# Reaction 3547
reaction('C6H12OOH1-2 => C6H12O1-2 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 3548
reaction('C6H12O1-2 + OH => C6H12OOH1-2', [0.000000e+00, 0.0, 0.0])

# Reaction 3549
reaction('C6H12OOH1-3 => C6H12O1-3 + OH', [7.500000e+10, 0.0, 15250.0])

# Reaction 3550
reaction('C6H12O1-3 + OH => C6H12OOH1-3', [0.000000e+00, 0.0, 0.0])

# Reaction 3551
reaction('C6H12OOH1-4 => C6H12O1-4 + OH', [9.375000e+09, 0.0, 7000.0])

# Reaction 3552
reaction('C6H12O1-4 + OH => C6H12OOH1-4', [0.000000e+00, 0.0, 0.0])

# Reaction 3553
reaction('C6H12OOH1-5 => C6H12O1-5 + OH', [1.172000e+09, 0.0, 1800.0])

# Reaction 3554
reaction('C6H12O1-5 + OH => C6H12OOH1-5', [0.000000e+00, 0.0, 0.0])

# Reaction 3555
reaction('C6H12OOH2-1 => C6H12O1-2 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 3556
reaction('C6H12O1-2 + OH => C6H12OOH2-1', [0.000000e+00, 0.0, 0.0])

# Reaction 3557
reaction('C6H12OOH2-3 => C6H12O2-3 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 3558
reaction('C6H12O2-3 + OH => C6H12OOH2-3', [0.000000e+00, 0.0, 0.0])

# Reaction 3559
reaction('C6H12OOH2-4 => C6H12O2-4 + OH', [7.500000e+10, 0.0, 15250.0])

# Reaction 3560
reaction('C6H12O2-4 + OH => C6H12OOH2-4', [0.000000e+00, 0.0, 0.0])

# Reaction 3561
reaction('C6H12OOH2-5 => C6H12O2-5 + OH', [9.375000e+09, 0.0, 7000.0])

# Reaction 3562
reaction('C6H12O2-5 + OH => C6H12OOH2-5', [0.000000e+00, 0.0, 0.0])

# Reaction 3563
reaction('C6H12OOH2-6 => C6H12O1-5 + OH', [1.172000e+09, 0.0, 1800.0])

# Reaction 3564
reaction('C6H12O1-5 + OH => C6H12OOH2-6', [0.000000e+00, 0.0, 0.0])

# Reaction 3565
reaction('C6H12OOH3-2 => C6H12O2-3 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 3566
reaction('C6H12O2-3 + OH => C6H12OOH3-2', [0.000000e+00, 0.0, 0.0])

# Reaction 3567
reaction('C6H12OOH3-4 => C6H12O3-4 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 3568
reaction('C6H12O3-4 + OH => C6H12OOH3-4', [0.000000e+00, 0.0, 0.0])

# Reaction 3569
reaction('C6H12OOH3-1 => C6H12O1-3 + OH', [7.500000e+10, 0.0, 15250.0])

# Reaction 3570
reaction('C6H12O1-3 + OH => C6H12OOH3-1', [0.000000e+00, 0.0, 0.0])

# Reaction 3571
reaction('C6H12OOH3-5 => C6H12O2-4 + OH', [7.500000e+10, 0.0, 15250.0])

# Reaction 3572
reaction('C6H12O2-4 + OH => C6H12OOH3-5', [0.000000e+00, 0.0, 0.0])

# Reaction 3573
reaction('C6H12OOH3-6 => C6H12O1-4 + OH', [9.375000e+09, 0.0, 7000.0])

# Reaction 3574
reaction('C6H12O1-4 + OH => C6H12OOH3-6', [0.000000e+00, 0.0, 0.0])

# Reaction 3575
reaction('C6H12OOH1-3 => OH + CH2O + C5H10-1', [7.703000e+13, -0.16, 30090.0])

# Reaction 3576
reaction('OH + CH2O + C5H10-1 => C6H12OOH1-3', [0.000000e+00, 0.0, 0.0])

# Reaction 3577
reaction('C6H12OOH2-4 => OH + CH3CHO + C4H8-1', [2.933000e+18, -1.7, 24080.0])

# Reaction 3578
reaction('OH + CH3CHO + C4H8-1 => C6H12OOH2-4', [0.000000e+00, 0.0, 0.0])

# Reaction 3579
reaction('C6H12OOH3-5 => OH + C2H5CHO + C3H6', [1.171000e+17, -1.31, 28880.0])

# Reaction 3580
reaction('OH + C2H5CHO + C3H6 => C6H12OOH3-5', [0.000000e+00, 0.0, 0.0])

# Reaction 3581
reaction('C6H12OOH3-1 => OH + CH3CHO + C2H4 + C2H4', [7.326000e+18, -1.74, 27420.0])

# Reaction 3582
reaction('OH + CH3CHO + C2H4 + C2H4 => C6H12OOH3-1', [0.000000e+00, 0.0, 0.0])

# Reaction 3583
reaction('C6H12OOH1-2O2 => C6H12OOH1-2 + O2', [8.868000e+22, -2.31, 37980.0])

# Reaction 3584
reaction('C6H12OOH1-2 + O2 => C6H12OOH1-2O2', [7.540000e+12, 0.0, 0.0])

# Reaction 3585
reaction('C6H12OOH1-3O2 => C6H12OOH1-3 + O2', [8.868000e+22, -2.31, 37980.0])

# Reaction 3586
reaction('C6H12OOH1-3 + O2 => C6H12OOH1-3O2', [7.540000e+12, 0.0, 0.0])

# Reaction 3587
reaction('C6H12OOH1-4O2 => C6H12OOH1-4 + O2', [8.868000e+22, -2.31, 37980.0])

# Reaction 3588
reaction('C6H12OOH1-4 + O2 => C6H12OOH1-4O2', [7.540000e+12, 0.0, 0.0])

# Reaction 3589
reaction('C6H12OOH1-5O2 => C6H12OOH1-5 + O2', [8.868000e+22, -2.31, 37980.0])

# Reaction 3590
reaction('C6H12OOH1-5 + O2 => C6H12OOH1-5O2', [7.540000e+12, 0.0, 0.0])

# Reaction 3591
reaction('C6H12OOH2-1O2 => C6H12OOH2-1 + O2', [1.463000e+20, -1.64, 35670.0])

# Reaction 3592
reaction('C6H12OOH2-1 + O2 => C6H12OOH2-1O2', [4.520000e+12, 0.0, 0.0])

# Reaction 3593
reaction('C6H12OOH2-3O2 => C6H12OOH2-3 + O2', [4.792000e+22, -2.33, 37960.0])

# Reaction 3594
reaction('C6H12OOH2-3 + O2 => C6H12OOH2-3O2', [7.540000e+12, 0.0, 0.0])

# Reaction 3595
reaction('C6H12OOH2-4O2 => C6H12OOH2-4 + O2', [4.792000e+22, -2.33, 37960.0])

# Reaction 3596
reaction('C6H12OOH2-4 + O2 => C6H12OOH2-4O2', [7.540000e+12, 0.0, 0.0])

# Reaction 3597
reaction('C6H12OOH2-5O2 => C6H12OOH2-5 + O2', [4.792000e+22, -2.33, 37960.0])

# Reaction 3598
reaction('C6H12OOH2-5 + O2 => C6H12OOH2-5O2', [7.540000e+12, 0.0, 0.0])

# Reaction 3599
reaction('C6H12OOH2-6O2 => C6H12OOH2-6 + O2', [1.463000e+20, -1.64, 35670.0])

# Reaction 3600
reaction('C6H12OOH2-6 + O2 => C6H12OOH2-6O2', [4.520000e+12, 0.0, 0.0])

# Reaction 3601
reaction('C6H12OOH3-1O2 => C6H12OOH3-1 + O2', [1.463000e+20, -1.64, 35670.0])

# Reaction 3602
reaction('C6H12OOH3-1 + O2 => C6H12OOH3-1O2', [4.520000e+12, 0.0, 0.0])

# Reaction 3603
reaction('C6H12OOH3-2O2 => C6H12OOH3-2 + O2', [4.792000e+22, -2.33, 37960.0])

# Reaction 3604
reaction('C6H12OOH3-2 + O2 => C6H12OOH3-2O2', [7.540000e+12, 0.0, 0.0])

# Reaction 3605
reaction('C6H12OOH3-4O2 => C6H12OOH3-4 + O2', [4.792000e+22, -2.33, 37960.0])

# Reaction 3606
reaction('C6H12OOH3-4 + O2 => C6H12OOH3-4O2', [7.540000e+12, 0.0, 0.0])

# Reaction 3607
reaction('C6H12OOH3-5O2 => C6H12OOH3-5 + O2', [4.792000e+22, -2.33, 37960.0])

# Reaction 3608
reaction('C6H12OOH3-5 + O2 => C6H12OOH3-5O2', [7.540000e+12, 0.0, 0.0])

# Reaction 3609
reaction('C6H12OOH3-6O2 => C6H12OOH3-6 + O2', [1.463000e+20, -1.64, 35670.0])

# Reaction 3610
reaction('C6H12OOH3-6 + O2 => C6H12OOH3-6O2', [4.520000e+12, 0.0, 0.0])

# Reaction 3611
reaction('C6H12OOH1-2O2 => NC6KET12 + OH', [2.000000e+11, 0.0, 26400.0])

# Reaction 3612
reaction('NC6KET12 + OH => C6H12OOH1-2O2', [2.607000e-12, 6.14, 39700.0])

# Reaction 3613
reaction('C6H12OOH1-3O2 => NC6KET13 + OH', [2.500000e+10, 0.0, 21400.0])

# Reaction 3614
reaction('NC6KET13 + OH => C6H12OOH1-3O2', [8.825000e+02, 1.56, 44710.0])

# Reaction 3615
reaction('C6H12OOH1-4O2 => NC6KET14 + OH', [3.125000e+09, 0.0, 19350.0])

# Reaction 3616
reaction('NC6KET14 + OH => C6H12OOH1-4O2', [1.103000e+02, 1.56, 42660.0])

# Reaction 3617
reaction('C6H12OOH1-5O2 => NC6KET15 + OH', [3.906000e+08, 0.0, 22550.0])

# Reaction 3618
reaction('NC6KET15 + OH => C6H12OOH1-5O2', [1.379000e+01, 1.56, 45860.0])

# Reaction 3619
reaction('C6H12OOH2-1O2 => NC6KET21 + OH', [1.000000e+11, 0.0, 23850.0])

# Reaction 3620
reaction('NC6KET21 + OH => C6H12OOH2-1O2', [2.253000e+02, 1.91, 50030.0])

# Reaction 3621
reaction('C6H12OOH2-3O2 => NC6KET23 + OH', [1.000000e+11, 0.0, 23850.0])

# Reaction 3622
reaction('NC6KET23 + OH => C6H12OOH2-3O2', [1.156000e+03, 1.79, 49090.0])

# Reaction 3623
reaction('C6H12OOH2-4O2 => NC6KET24 + OH', [1.250000e+10, 0.0, 17850.0])

# Reaction 3624
reaction('NC6KET24 + OH => C6H12OOH2-4O2', [4.618000e+01, 1.97, 43880.0])

# Reaction 3625
reaction('C6H12OOH2-5O2 => NC6KET25 + OH', [1.563000e+09, 0.0, 16050.0])

# Reaction 3626
reaction('NC6KET25 + OH => C6H12OOH2-5O2', [5.774000e+00, 1.97, 42080.0])

# Reaction 3627
reaction('C6H12OOH2-6O2 => NC6KET26 + OH', [3.906000e+08, 0.0, 22550.0])

# Reaction 3628
reaction('NC6KET26 + OH => C6H12OOH2-6O2', [1.939000e+00, 1.92, 48590.0])

# Reaction 3629
reaction('C6H12OOH3-1O2 => NC6KET31 + OH', [1.250000e+10, 0.0, 17850.0])

# Reaction 3630
reaction('NC6KET31 + OH => C6H12OOH3-1O2', [5.308000e+01, 1.91, 44190.0])

# Reaction 3631
reaction('C6H12OOH3-2O2 => NC6KET32 + OH', [1.000000e+11, 0.0, 23850.0])

# Reaction 3632
reaction('NC6KET32 + OH => C6H12OOH3-2O2', [7.384000e+02, 1.82, 49350.0])

# Reaction 3633
reaction('C6H12OOH3-4O2 => NC6KET34 + OH', [1.000000e+11, 0.0, 23850.0])

# Reaction 3634
reaction('NC6KET34 + OH => C6H12OOH3-4O2', [7.384000e+02, 1.82, 49350.0])

# Reaction 3635
reaction('C6H12OOH3-5O2 => NC6KET35 + OH', [1.250000e+10, 0.0, 17850.0])

# Reaction 3636
reaction('NC6KET35 + OH => C6H12OOH3-5O2', [3.503000e+01, 1.98, 44160.0])

# Reaction 3637
reaction('C6H12OOH3-6O2 => NC6KET36 + OH', [1.563000e+09, 0.0, 16050.0])

# Reaction 3638
reaction('NC6KET36 + OH => C6H12OOH3-6O2', [6.638000e+00, 1.91, 42390.0])

# Reaction 3639
reaction('NC6KET12 => NC4H9CHO + HCO + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 3640
reaction('NC4H9CHO + HCO + OH => NC6KET12', [0.000000e+00, 0.0, 0.0])

# Reaction 3641
reaction('NC6KET13 => NC3H7CHO + CH2CHO + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 3642
reaction('NC3H7CHO + CH2CHO + OH => NC6KET13', [0.000000e+00, 0.0, 0.0])

# Reaction 3643
reaction('NC6KET14 => C2H5CHO + CH2CH2CHO + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 3644
reaction('C2H5CHO + CH2CH2CHO + OH => NC6KET14', [0.000000e+00, 0.0, 0.0])

# Reaction 3645
reaction('NC6KET15 => CH3CHO + C3H6CHO-1 + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 3646
reaction('CH3CHO + C3H6CHO-1 + OH => NC6KET15', [0.000000e+00, 0.0, 0.0])

# Reaction 3647
reaction('NC6KET21 => CH2O + NC4H9CO + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 3648
reaction('CH2O + NC4H9CO + OH => NC6KET21', [0.000000e+00, 0.0, 0.0])

# Reaction 3649
reaction('NC6KET23 => NC3H7CHO + CH3CO + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 3650
reaction('NC3H7CHO + CH3CO + OH => NC6KET23', [0.000000e+00, 0.0, 0.0])

# Reaction 3651
reaction('NC6KET24 => C2H5CHO + CH3COCH2 + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 3652
reaction('C2H5CHO + CH3COCH2 + OH => NC6KET24', [0.000000e+00, 0.0, 0.0])

# Reaction 3653
reaction('NC6KET25 => CH3CHO + CH2CH2COCH3 + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 3654
reaction('CH3CHO + CH2CH2COCH3 + OH => NC6KET25', [0.000000e+00, 0.0, 0.0])

# Reaction 3655
reaction('NC6KET26 => CH2O + C3H6COCH3-1 + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 3656
reaction('CH2O + C3H6COCH3-1 + OH => NC6KET26', [0.000000e+00, 0.0, 0.0])

# Reaction 3657
reaction('NC6KET31 => CH2O + NC3H7COCH2 + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 3658
reaction('CH2O + NC3H7COCH2 + OH => NC6KET31', [0.000000e+00, 0.0, 0.0])

# Reaction 3659
reaction('NC6KET32 => CH3CHO + NC3H7CO + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 3660
reaction('CH3CHO + NC3H7CO + OH => NC6KET32', [0.000000e+00, 0.0, 0.0])

# Reaction 3661
reaction('NC6KET34 => C2H5CHO + C2H5CO + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 3662
reaction('C2H5CHO + C2H5CO + OH => NC6KET34', [0.000000e+00, 0.0, 0.0])

# Reaction 3663
reaction('NC6KET35 => CH3CHO + C2H5COCH2 + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 3664
reaction('CH3CHO + C2H5COCH2 + OH => NC6KET35', [0.000000e+00, 0.0, 0.0])

# Reaction 3665
reaction('NC6KET36 => CH2O + C2H5COC2H4P + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 3666
reaction('CH2O + C2H5COC2H4P + OH => NC6KET36', [0.000000e+00, 0.0, 0.0])

# Reaction 3667
reaction('C6H12O1-2 + OH => C2H3CHO + NC3H7 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 3668
reaction('C2H3CHO + NC3H7 + H2O => C6H12O1-2 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3669
reaction('C6H12O1-3 + OH => C5H10-1 + HCO + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 3670
reaction('C5H10-1 + HCO + H2O => C6H12O1-3 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3671
reaction('C6H12O1-4 + OH => C4H8-1 + CH2CHO + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 3672
reaction('C4H8-1 + CH2CHO + H2O => C6H12O1-4 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3673
reaction('C6H12O1-5 + OH => C3H6 + CH2CH2CHO + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 3674
reaction('C3H6 + CH2CH2CHO + H2O => C6H12O1-5 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3675
reaction('C6H12O2-3 + OH => C2H3COCH3 + C2H5 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 3676
reaction('C2H3COCH3 + C2H5 + H2O => C6H12O2-3 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3677
reaction('C6H12O2-4 + OH => C4H8-1 + CH3CO + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 3678
reaction('C4H8-1 + CH3CO + H2O => C6H12O2-4 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3679
reaction('C6H12O2-5 + OH => C3H6 + CH3COCH2 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 3680
reaction('C3H6 + CH3COCH2 + H2O => C6H12O2-5 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3681
reaction('C6H12O3-4 + OH => C2H5COC2H3 + CH3 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 3682
reaction('C2H5COC2H3 + CH3 + H2O => C6H12O3-4 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3683
reaction('C6H12O1-2 + OH => CH2CO + PC4H9 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 3684
reaction('CH2CO + PC4H9 + H2O => C6H12O1-2 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3685
reaction('C6H12O1-3 + OH => C2H4 + NC3H7CO + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 3686
reaction('C2H4 + NC3H7CO + H2O => C6H12O1-3 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3687
reaction('C6H12O1-4 + OH => C2H4 + C2H5COCH2 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 3688
reaction('C2H4 + C2H5COCH2 + H2O => C6H12O1-4 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3689
reaction('C6H12O1-5 + OH => C2H4 + CH2CH2COCH3 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 3690
reaction('C2H4 + CH2CH2COCH3 + H2O => C6H12O1-5 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3691
reaction('C6H12O2-3 + OH => CH3CHCO + NC3H7 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 3692
reaction('CH3CHCO + NC3H7 + H2O => C6H12O2-3 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3693
reaction('C6H12O2-4 + OH => C3H6 + C2H5CO + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 3694
reaction('C3H6 + C2H5CO + H2O => C6H12O2-4 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3695
reaction('C6H12O2-5 + OH => CH3CHO + C4H71-3 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 3696
reaction('CH3CHO + C4H71-3 + H2O => C6H12O2-5 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3697
reaction('C6H12O3-4 + OH => C2H5CHO + C3H5-S + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 3698
reaction('C2H5CHO + C3H5-S + H2O => C6H12O3-4 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 3699
reaction('C6H12O1-2 + HO2 => C2H3CHO + NC3H7 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 3700
reaction('C2H3CHO + NC3H7 + H2O2 => C6H12O1-2 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3701
reaction('C6H12O1-3 + HO2 => C5H10-1 + HCO + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 3702
reaction('C5H10-1 + HCO + H2O2 => C6H12O1-3 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3703
reaction('C6H12O1-4 + HO2 => C4H8-1 + CH2CHO + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 3704
reaction('C4H8-1 + CH2CHO + H2O2 => C6H12O1-4 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3705
reaction('C6H12O1-5 + HO2 => C3H6 + CH2CH2CHO + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 3706
reaction('C3H6 + CH2CH2CHO + H2O2 => C6H12O1-5 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3707
reaction('C6H12O2-3 + HO2 => C2H3COCH3 + C2H5 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 3708
reaction('C2H3COCH3 + C2H5 + H2O2 => C6H12O2-3 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3709
reaction('C6H12O2-4 + HO2 => C4H8-1 + CH3CO + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 3710
reaction('C4H8-1 + CH3CO + H2O2 => C6H12O2-4 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3711
reaction('C6H12O2-5 + HO2 => C3H6 + CH3COCH2 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 3712
reaction('C3H6 + CH3COCH2 + H2O2 => C6H12O2-5 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3713
reaction('C6H12O3-4 + HO2 => C2H5COC2H3 + CH3 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 3714
reaction('C2H5COC2H3 + CH3 + H2O2 => C6H12O3-4 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3715
reaction('C6H12O1-2 + HO2 => CH2CO + PC4H9 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 3716
reaction('CH2CO + PC4H9 + H2O2 => C6H12O1-2 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3717
reaction('C6H12O1-3 + HO2 => C2H4 + NC3H7CO + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 3718
reaction('C2H4 + NC3H7CO + H2O2 => C6H12O1-3 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3719
reaction('C6H12O1-4 + HO2 => C2H4 + C2H5COCH2 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 3720
reaction('C2H4 + C2H5COCH2 + H2O2 => C6H12O1-4 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3721
reaction('C6H12O1-5 + HO2 => C2H4 + CH2CH2COCH3 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 3722
reaction('C2H4 + CH2CH2COCH3 + H2O2 => C6H12O1-5 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3723
reaction('C6H12O2-3 + HO2 => CH3CHCO + NC3H7 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 3724
reaction('CH3CHCO + NC3H7 + H2O2 => C6H12O2-3 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3725
reaction('C6H12O2-4 + HO2 => C3H6 + C2H5CO + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 3726
reaction('C3H6 + C2H5CO + H2O2 => C6H12O2-4 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3727
reaction('C6H12O2-5 + HO2 => CH3CHO + C4H71-3 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 3728
reaction('CH3CHO + C4H71-3 + H2O2 => C6H12O2-5 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3729
reaction('C6H12O3-4 + HO2 => C2H5CHO + C3H5-S + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 3730
reaction('C2H5CHO + C3H5-S + H2O2 => C6H12O3-4 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3731
reaction('C6H12-1 + OH <=> C6H12OH-1', [1.000000e+13, 0.0, 0.0])

# Reaction 3732
reaction('C6H12OH-1 + O2 <=> O2C6H12OH-1', [6.000000e+12, 0.0, 0.0])

# Reaction 3733
reaction('O2C6H12OH-1 <=> NC4H9CHO + CH2O + OH', [1.000000e+12, 0.0, 27800.0])

# Reaction 3734
reaction('C6H12OH-2 => C6H12-2 + OH', [1.065000e+16, -1.0, 29330.0])

# Reaction 3735
reaction('C6H12-2 + OH => C6H12OH-2', [1.000000e+12, 0.0, -1042.0])

# Reaction 3736
reaction('O2C6H12OH-2 => C6H12OH-2 + O2', [3.816000e+21, -2.02, 37830.0])

# Reaction 3737
reaction('C6H12OH-2 + O2 => O2C6H12OH-2', [2.000000e+12, 0.0, 0.0])

# Reaction 3738
reaction('O2C6H12OH-2 => NC3H7CHO + CH3CHO + OH', [2.500000e+10, 0.0, 18860.0])

# Reaction 3739
reaction('NC3H7CHO + CH3CHO + OH => O2C6H12OH-2', [0.000000e+00, 0.0, 0.0])

# Reaction 3740
reaction('C6H12OH-3 => C6H12-3 + OH', [3.903000e+15, -0.93, 29420.0])

# Reaction 3741
reaction('C6H12-3 + OH => C6H12OH-3', [1.000000e+12, 0.0, -1042.0])

# Reaction 3742
reaction('O2C6H12OH-3 => C6H12OH-3 + O2', [3.816000e+21, -2.02, 37830.0])

# Reaction 3743
reaction('C6H12OH-3 + O2 => O2C6H12OH-3', [2.000000e+12, 0.0, 0.0])

# Reaction 3744
reaction('O2C6H12OH-3 => OH + C2H5CHO + C2H5CHO', [2.500000e+10, 0.0, 18860.0])

# Reaction 3745
reaction('OH + C2H5CHO + C2H5CHO => O2C6H12OH-3', [0.000000e+00, 0.0, 0.0])

# Reaction 3746
reaction('NC4H9CHO + O2 => NC4H9CO + HO2', [2.000000e+13, 0.5, 42200.0])

# Reaction 3747
reaction('NC4H9CO + HO2 => NC4H9CHO + O2', [1.000000e+07, 0.0, 40000.0])

# Reaction 3748
reaction('NC4H9CHO + OH => NC4H9CO + H2O', [2.690000e+10, 0.76, -340.0])

# Reaction 3749
reaction('NC4H9CO + H2O => NC4H9CHO + OH', [2.143000e+10, 0.73, 31240.0])

# Reaction 3750
reaction('NC4H9CHO + H => NC4H9CO + H2', [4.000000e+13, 0.0, 4200.0])

# Reaction 3751
reaction('NC4H9CO + H2 => NC4H9CHO + H', [1.800000e+13, 0.0, 24000.0])

# Reaction 3752
reaction('NC4H9CHO + O => NC4H9CO + OH', [5.000000e+12, 0.0, 1790.0])

# Reaction 3753
reaction('NC4H9CO + OH => NC4H9CHO + O', [1.000000e+12, 0.0, 19000.0])

# Reaction 3754
reaction('NC4H9CHO + HO2 => NC4H9CO + H2O2', [2.800000e+12, 0.0, 13600.0])

# Reaction 3755
reaction('NC4H9CO + H2O2 => NC4H9CHO + HO2', [1.000000e+12, 0.0, 10000.0])

# Reaction 3756
reaction('NC4H9CHO + CH3 => NC4H9CO + CH4', [1.700000e+12, 0.0, 8440.0])

# Reaction 3757
reaction('NC4H9CO + CH4 => NC4H9CHO + CH3', [1.500000e+13, 0.0, 28000.0])

# Reaction 3758
reaction('NC4H9CHO + CH3O => NC4H9CO + CH3OH', [1.150000e+11, 0.0, 1280.0])

# Reaction 3759
reaction('NC4H9CO + CH3OH => NC4H9CHO + CH3O', [3.000000e+11, 0.0, 18000.0])

# Reaction 3760
reaction('NC4H9CHO + CH3O2 => NC4H9CO + CH3O2H', [1.000000e+12, 0.0, 9500.0])

# Reaction 3761
reaction('NC4H9CO + CH3O2H => NC4H9CHO + CH3O2', [2.500000e+10, 0.0, 10000.0])

# Reaction 3762
reaction('NC4H9CHO + OH => C4H8CHO-1 + H2O', [5.270000e+09, 0.97, 1586.0])

# Reaction 3763
reaction('C4H8CHO-1 + H2O => NC4H9CHO + OH', [1.233000e+09, 0.95, 21040.0])

# Reaction 3764
reaction('NC4H9CHO + OH => C4H8CHO-2 + H2O', [4.670000e+07, 1.61, -35.0])

# Reaction 3765
reaction('C4H8CHO-2 + H2O => NC4H9CHO + OH', [6.837000e+05, 1.93, 21920.0])

# Reaction 3766
reaction('NC4H9CHO + OH => C4H8CHO-3 + H2O', [4.670000e+07, 1.61, -35.0])

# Reaction 3767
reaction('C4H8CHO-3 + H2O => NC4H9CHO + OH', [6.837000e+05, 1.93, 21920.0])

# Reaction 3768
reaction('NC4H9CHO + OH => C4H8CHO-4 + H2O', [4.670000e+07, 1.61, -35.0])

# Reaction 3769
reaction('C4H8CHO-4 + H2O => NC4H9CHO + OH', [1.202000e+09, 1.34, 28870.0])

# Reaction 3770
reaction('NC4H9CO => PC4H9 + CO', [1.000000e+11, 0.0, 9600.0])

# Reaction 3771
reaction('PC4H9 + CO => NC4H9CO', [1.000000e+11, 0.0, 0.0])

# Reaction 3772
reaction('C4H8CHO-1 => C2H4 + CH2CH2CHO', [5.984000e+18, -1.6, 30430.0])

# Reaction 3773
reaction('C2H4 + CH2CH2CHO => C4H8CHO-1', [2.500000e+11, 0.0, 7800.0])

# Reaction 3774
reaction('C4H8CHO-2 => C3H6 + CH2CHO', [2.984000e+14, -0.76, 23320.0])

# Reaction 3775
reaction('C3H6 + CH2CHO => C4H8CHO-2', [1.000000e+11, 0.0, 7800.0])

# Reaction 3776
reaction('C4H8CHO-3 => C4H8-1 + HCO', [4.797000e+14, -0.72, 24350.0])

# Reaction 3777
reaction('C4H8-1 + HCO => C4H8CHO-3', [1.000000e+11, 0.0, 7800.0])

# Reaction 3778
reaction('C4H8CHO-3 => AC3H5CHO + CH3', [3.643000e+13, -0.36, 30330.0])

# Reaction 3779
reaction('AC3H5CHO + CH3 => C4H8CHO-3', [1.000000e+11, 0.0, 7800.0])

# Reaction 3780
reaction('C4H8CHO-4 => C2H3CHO + C2H5', [1.863000e+18, -1.3, 30830.0])

# Reaction 3781
reaction('C2H3CHO + C2H5 => C4H8CHO-4', [1.000000e+11, 0.0, 7800.0])

# Reaction 3782
reaction('AC3H5CHO => C3H5-A + HCO', [1.813000e+19, -1.08, 68480.0])

# Reaction 3783
reaction('C3H5-A + HCO => AC3H5CHO', [1.000000e+13, 0.0, 0.0])

# Reaction 3784
reaction('AC3H5CHO + OH => AC3H5CO + H2O', [2.690000e+10, 0.76, -340.0])

# Reaction 3785
reaction('AC3H5CO + H2O => AC3H5CHO + OH', [1.741000e+10, 0.76, 31200.0])

# Reaction 3786
reaction('AC3H5CHO + OH => C2H3CHCHO + H2O', [3.120000e+06, 2.0, -298.0])

# Reaction 3787
reaction('C2H3CHCHO + H2O => AC3H5CHO + OH', [1.246000e+08, 1.79, 34930.0])

# Reaction 3788
reaction('AC3H5CHO + HO2 => AC3H5CO + H2O2', [3.000000e+12, 0.0, 11920.0])

# Reaction 3789
reaction('AC3H5CO + H2O2 => AC3H5CHO + HO2', [1.152000e+13, -0.33, 11990.0])

# Reaction 3790
reaction('AC3H5CHO + HO2 => C2H3CHCHO + H2O2', [9.630000e+03, 2.6, 13910.0])

# Reaction 3791
reaction('C2H3CHCHO + H2O2 => AC3H5CHO + HO2', [2.283000e+06, 2.06, 17670.0])

# Reaction 3792
reaction('AC3H5CHO + CH3O2 => AC3H5CO + CH3O2H', [1.000000e+12, 0.0, 11920.0])

# Reaction 3793
reaction('AC3H5CO + CH3O2H => AC3H5CHO + CH3O2', [1.579000e+13, -0.51, 11270.0])

# Reaction 3794
reaction('AC3H5CHO + CH3O2 => C2H3CHCHO + CH3O2H', [1.990000e+12, 0.0, 17050.0])

# Reaction 3795
reaction('C2H3CHCHO + CH3O2H => AC3H5CHO + CH3O2', [1.939000e+15, -0.72, 20090.0])

# Reaction 3796
reaction('AC3H5CO => C3H5-A + CO', [6.199000e+15, -1.09, -330.0])

# Reaction 3797
reaction('C3H5-A + CO => AC3H5CO', [1.500000e+11, 0.0, 4810.0])

# Reaction 3798
reaction('C2H3CHCHO + HO2 => C2H3CHO + HCO + OH', [8.910000e+12, 0.0, 0.0])

# Reaction 3799
reaction('C2H3CHO + HCO + OH => C2H3CHCHO + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 3800
reaction('NC3H7COCH3 + OH => C3H6COCH3-1 + H2O', [2.065000e+07, 1.73, 753.0])

# Reaction 3801
reaction('C3H6COCH3-1 + H2O => NC3H7COCH3 + OH', [3.604000e+06, 1.75, 20170.0])

# Reaction 3802
reaction('NC3H7COCH3 + OH => C3H6COCH3-2 + H2O', [3.615000e+07, 1.64, -247.0])

# Reaction 3803
reaction('C3H6COCH3-2 + H2O => NC3H7COCH3 + OH', [3.668000e+05, 2.01, 21660.0])

# Reaction 3804
reaction('NC3H7COCH3 + OH => C3H6COCH3-3 + H2O', [8.450000e+11, 0.0, -228.0])

# Reaction 3805
reaction('C3H6COCH3-3 + H2O => NC3H7COCH3 + OH', [2.368000e+13, -0.28, 28700.0])

# Reaction 3806
reaction('NC3H7COCH3 + OH => NC3H7COCH2 + H2O', [5.100000e+11, 0.0, 1192.0])

# Reaction 3807
reaction('NC3H7COCH2 + H2O => NC3H7COCH3 + OH', [9.286000e+13, -0.74, 27730.0])

# Reaction 3808
reaction('NC3H7COCH3 + HO2 => C3H6COCH3-1 + H2O2', [2.380000e+04, 2.55, 16490.0])

# Reaction 3809
reaction('C3H6COCH3-1 + H2O2 => NC3H7COCH3 + HO2', [2.465000e+04, 2.24, 4442.0])

# Reaction 3810
reaction('NC3H7COCH3 + HO2 => C3H6COCH3-2 + H2O2', [5.600000e+12, 0.0, 17700.0])

# Reaction 3811
reaction('C3H6COCH3-2 + H2O2 => NC3H7COCH3 + HO2', [3.373000e+11, 0.05, 8142.0])

# Reaction 3812
reaction('NC3H7COCH3 + HO2 => C3H6COCH3-3 + H2O2', [2.000000e+11, 0.0, 8698.0])

# Reaction 3813
reaction('C3H6COCH3-3 + H2O2 => NC3H7COCH3 + HO2', [3.327000e+13, -0.6, 6160.0])

# Reaction 3814
reaction('NC3H7COCH3 + HO2 => NC3H7COCH2 + H2O2', [2.380000e+04, 2.55, 14690.0])

# Reaction 3815
reaction('NC3H7COCH2 + H2O2 => NC3H7COCH3 + HO2', [2.572000e+07, 1.48, 9762.0])

# Reaction 3816
reaction('NC3H7COCH3 + CH3O2 => C3H6COCH3-1 + CH3O2H', [3.010000e+12, 0.0, 19380.0])

# Reaction 3817
reaction('C3H6COCH3-1 + CH3O2H => NC3H7COCH3 + CH3O2', [1.281000e+13, -0.49, 6607.0])

# Reaction 3818
reaction('NC3H7COCH3 + CH3O2 => C3H6COCH3-2 + CH3O2H', [1.990000e+12, 0.0, 17050.0])

# Reaction 3819
reaction('C3H6COCH3-2 + CH3O2H => NC3H7COCH3 + CH3O2', [4.926000e+11, -0.13, 6774.0])

# Reaction 3820
reaction('NC3H7COCH3 + CH3O2 => C3H6COCH3-3 + CH3O2H', [2.000000e+12, 0.0, 15250.0])

# Reaction 3821
reaction('C3H6COCH3-3 + CH3O2H => NC3H7COCH3 + CH3O2', [1.367000e+15, -0.78, 11990.0])

# Reaction 3822
reaction('NC3H7COCH3 + CH3O2 => NC3H7COCH2 + CH3O2H', [3.010000e+12, 0.0, 17580.0])

# Reaction 3823
reaction('NC3H7COCH2 + CH3O2H => NC3H7COCH3 + CH3O2', [1.337000e+16, -1.25, 11930.0])

# Reaction 3824
reaction('C3H6COCH3-1 => C2H4 + CH3COCH2', [2.904000e+16, -1.21, 27000.0])

# Reaction 3825
reaction('C2H4 + CH3COCH2 => C3H6COCH3-1', [1.000000e+11, 0.0, 8600.0])

# Reaction 3826
reaction('C3H6COCH3-2 => C3H6 + CH3CO', [2.719000e+16, -1.05, 25590.0])

# Reaction 3827
reaction('C3H6 + CH3CO => C3H6COCH3-2', [1.000000e+11, 0.0, 7800.0])

# Reaction 3828
reaction('C3H6COCH3-3 => C2H3COCH3 + CH3', [9.615000e+15, -0.75, 32390.0])

# Reaction 3829
reaction('C2H3COCH3 + CH3 => C3H6COCH3-3', [1.000000e+11, 0.0, 7800.0])

# Reaction 3830
reaction('NC3H7COCH2 => NC3H7 + CH2CO', [1.226000e+18, -1.4, 43450.0])

# Reaction 3831
reaction('NC3H7 + CH2CO => NC3H7COCH2', [1.000000e+11, 0.0, 11600.0])

# Reaction 3832
reaction('C2H5COC2H5 + OH => C2H5COC2H4P + H2O', [1.054000e+10, 0.97, 1586.0])

# Reaction 3833
reaction('C2H5COC2H4P + H2O => C2H5COC2H5 + OH', [1.194000e+09, 0.95, 21030.0])

# Reaction 3834
reaction('C2H5COC2H5 + OH => C2H5COC2H4S + H2O', [1.690000e+12, 0.0, -228.0])

# Reaction 3835
reaction('C2H5COC2H4S + H2O => C2H5COC2H5 + OH', [3.296000e+13, -0.32, 28740.0])

# Reaction 3836
reaction('C2H5COC2H5 + HO2 => C2H5COC2H4P + H2O2', [4.760000e+04, 2.55, 16490.0])

# Reaction 3837
reaction('C2H5COC2H4P + H2O2 => C2H5COC2H5 + HO2', [3.201000e+04, 2.2, 4472.0])

# Reaction 3838
reaction('C2H5COC2H5 + HO2 => C2H5COC2H4S + H2O2', [4.000000e+11, 0.0, 8698.0])

# Reaction 3839
reaction('C2H5COC2H4S + H2O2 => C2H5COC2H5 + HO2', [4.631000e+13, -0.65, 6200.0])

# Reaction 3840
reaction('C2H5COC2H5 + O2 => C2H5COC2H4P + HO2', [4.100000e+13, 0.0, 51310.0])

# Reaction 3841
reaction('C2H5COC2H4P + HO2 => C2H5COC2H5 + O2', [1.949000e+10, 0.31, -1882.0])

# Reaction 3842
reaction('C2H5COC2H5 + O2 => C2H5COC2H4S + HO2', [3.100000e+13, 0.0, 41970.0])

# Reaction 3843
reaction('C2H5COC2H4S + HO2 => C2H5COC2H5 + O2', [2.537000e+12, 0.0, -1702.0])

# Reaction 3844
reaction('C2H5COC2H5 + H => C2H5COC2H4P + H2', [1.832000e+07, 2.0, 7700.0])

# Reaction 3845
reaction('C2H5COC2H4P + H2 => C2H5COC2H5 + H', [4.794000e+05, 1.98, 11990.0])

# Reaction 3846
reaction('C2H5COC2H5 + H => C2H5COC2H4S + H2', [8.960000e+06, 2.0, 3200.0])

# Reaction 3847
reaction('C2H5COC2H4S + H2 => C2H5COC2H5 + H', [4.036000e+07, 1.68, 17010.0])

# Reaction 3848
reaction('C2H5COC2H5 + C2H3 => C2H5COC2H4P + C2H4', [1.000000e+12, 0.0, 10400.0])

# Reaction 3849
reaction('C2H5COC2H4P + C2H4 => C2H5COC2H5 + C2H3', [3.850000e+08, 0.81, 14040.0])

# Reaction 3850
reaction('C2H5COC2H5 + C2H3 => C2H5COC2H4S + C2H4', [6.000000e+11, 0.0, 5600.0])

# Reaction 3851
reaction('C2H5COC2H4S + C2H4 => C2H5COC2H5 + C2H3', [3.976000e+10, 0.5, 18760.0])

# Reaction 3852
reaction('C2H5COC2H5 + C2H5 => C2H5COC2H4P + C2H6', [1.000000e+11, 0.0, 13400.0])

# Reaction 3853
reaction('C2H5COC2H4P + C2H6 => C2H5COC2H5 + C2H5', [1.070000e+13, -0.58, 14000.0])

# Reaction 3854
reaction('C2H5COC2H5 + C2H5 => C2H5COC2H4S + C2H6', [6.000000e+10, 0.0, 8600.0])

# Reaction 3855
reaction('C2H5COC2H4S + C2H6 => C2H5COC2H5 + C2H5', [1.105000e+15, -0.89, 18720.0])

# Reaction 3856
reaction('C2H5COC2H5 + CH3O => C2H5COC2H4P + CH3OH', [4.340000e+11, 0.0, 6460.0])

# Reaction 3857
reaction('C2H5COC2H4P + CH3OH => C2H5COC2H5 + CH3O', [5.475000e+09, 0.0, 9020.0])

# Reaction 3858
reaction('C2H5COC2H5 + CH3O => C2H5COC2H4S + CH3OH', [2.900000e+11, 0.0, 2771.0])

# Reaction 3859
reaction('C2H5COC2H4S + CH3OH => C2H5COC2H5 + CH3O', [6.298000e+11, -0.3, 14850.0])

# Reaction 3860
reaction('C2H5COC2H5 + CH3O2 => C2H5COC2H4P + CH3O2H', [6.020000e+12, 0.0, 19380.0])

# Reaction 3861
reaction('C2H5COC2H4P + CH3O2H => C2H5COC2H5 + CH3O2', [1.664000e+13, -0.52, 6642.0])

# Reaction 3862
reaction('C2H5COC2H5 + CH3O2 => C2H5COC2H4S + CH3O2H', [4.000000e+12, 0.0, 15250.0])

# Reaction 3863
reaction('C2H5COC2H4S + CH3O2H => C2H5COC2H5 + CH3O2', [1.903000e+15, -0.83, 12030.0])

# Reaction 3864
reaction('C2H5COC2H4P => C2H5CO + C2H4', [1.769000e+17, -1.46, 29540.0])

# Reaction 3865
reaction('C2H5CO + C2H4 => C2H5COC2H4P', [8.000000e+10, 0.0, 11300.0])

# Reaction 3866
reaction('C2H5COC2H4S => C2H5COC2H3 + H', [8.704000e+16, -0.82, 42130.0])

# Reaction 3867
reaction('C2H5COC2H3 + H => C2H5COC2H4S', [1.300000e+13, 0.0, 1560.0])

# Reaction 3868
reaction('C2H5COC2H3 + OH => C2H5COCH2 + CH2O', [1.000000e+10, 0.0, 0.0])

# Reaction 3869
reaction('C2H5COCH2 + CH2O => C2H5COC2H3 + OH', [1.673000e+11, -0.39, 18220.0])

# Reaction 3870
reaction('C2H5COC2H3 + OH => PC2H4COC2H3 + H2O', [7.550000e+09, 0.97, 1586.0])

# Reaction 3871
reaction('PC2H4COC2H3 + H2O => C2H5COC2H3 + OH', [1.285000e+08, 2.53, 20210.0])

# Reaction 3872
reaction('C2H5COC2H3 + OH => SC2H4COC2H3 + H2O', [8.450000e+11, 0.0, -228.0])

# Reaction 3873
reaction('SC2H4COC2H3 + H2O => C2H5COC2H3 + OH', [1.350000e+13, -0.2, 28600.0])

# Reaction 3874
reaction('C2H5COC2H3 + HO2 => C2H5CO + CH2CHO + OH', [6.030000e+09, 0.0, 7949.0])

# Reaction 3875
reaction('C2H5CO + CH2CHO + OH => C2H5COC2H3 + HO2', [1.000000e+00, 0.0, 0.0])

# Reaction 3876
reaction('C2H5COC2H3 + HO2 => PC2H4COC2H3 + H2O2', [2.380000e+04, 2.55, 16490.0])

# Reaction 3877
reaction('PC2H4COC2H3 + H2O2 => C2H5COC2H3 + HO2', [2.405000e+03, 3.78, 3652.0])

# Reaction 3878
reaction('C2H5COC2H3 + HO2 => SC2H4COC2H3 + H2O2', [2.000000e+11, 0.0, 8698.0])

# Reaction 3879
reaction('SC2H4COC2H3 + H2O2 => C2H5COC2H3 + HO2', [1.896000e+13, -0.53, 6065.0])

# Reaction 3880
reaction('C2H5COC2H3 + CH3O2 => C2H5CO + CH2CHO + CH3O', [3.970000e+11, 0.0, 17050.0])

# Reaction 3881
reaction('C2H5CO + CH2CHO + CH3O => C2H5COC2H3 + CH3O2', [1.000000e+00, 0.0, 0.0])

# Reaction 3882
reaction('C2H5COC2H3 + CH3O2 => PC2H4COC2H3 + CH3O2H', [3.010000e+12, 0.0, 19380.0])

# Reaction 3883
reaction('PC2H4COC2H3 + CH3O2H => C2H5COC2H3 + CH3O2', [1.250000e+12, 1.06, 5822.0])

# Reaction 3884
reaction('C2H5COC2H3 + CH3O2 => SC2H4COC2H3 + CH3O2H', [2.000000e+12, 0.0, 15250.0])

# Reaction 3885
reaction('SC2H4COC2H3 + CH3O2H => C2H5COC2H3 + CH3O2', [7.793000e+14, -0.71, 11900.0])

# Reaction 3886
reaction('PC2H4COC2H3 => C2H3CO + C2H4', [5.263000e+14, 0.38, 21460.0])

# Reaction 3887
reaction('C2H3CO + C2H4 => PC2H4COC2H3', [8.000000e+10, 0.0, 11300.0])

# Reaction 3888
reaction('SC2H4COC2H3 => CH3CHCO + C2H3', [1.637000e+16, -0.74, 54590.0])

# Reaction 3889
reaction('CH3CHCO + C2H3 => SC2H4COC2H3', [8.000000e+10, 0.0, 12600.0])

# Reaction 3890
reaction('NC7H16 => H + C7H15-1', [2.680000e+88, -21.17, 142800.0])

# Reaction 3891
reaction('H + C7H15-1 => NC7H16', [7.264000e+84, -20.81, 41560.0])

# Reaction 3892
reaction('NC7H16 => H + C7H15-2', [1.300000e+88, -21.01, 139500.0])

# Reaction 3893
reaction('H + C7H15-2 => NC7H16', [2.263000e+83, -20.31, 40830.0])

# Reaction 3894
reaction('NC7H16 => H + C7H15-3', [1.300000e+88, -21.01, 139500.0])

# Reaction 3895
reaction('H + C7H15-3 => NC7H16', [2.263000e+83, -20.31, 40830.0])

# Reaction 3896
reaction('NC7H16 => H + C7H15-4', [6.500000e+87, -21.01, 139500.0])

# Reaction 3897
reaction('H + C7H15-4 => NC7H16', [2.255000e+83, -20.31, 40830.0])

# Reaction 3898
falloff_reaction('NC7H16 (+ M) <=> C6H13-1 + CH3 (+ M)',
                kf=[4.325000e+24, -2.12, 89900.0],
                kf0=[4.963000e+42, -7.78, 42800.0],
                falloff=Troe(A=0.892, T3=10000000000.0, T1=2.228, T2=1798000000.0))

# Reaction 3899
falloff_reaction('NC7H16 (+ M) <=> C5H11-1 + C2H5 (+ M)',
                kf=[6.818000e+26, -2.7, 88910.0],
                kf0=[3.753000e+48, -9.46, 41310.0],
                falloff=Troe(A=0.09, T3=3.6556, T1=10000000000.0, T2=9330000000.0))

# Reaction 3900
falloff_reaction('NC7H16 (+ M) <=> PC4H9 + NC3H7 (+ M)',
                kf=[1.362000e+26, -2.53, 88760.0],
                kf0=[6.509000e+48, -9.57, 41290.0],
                falloff=Troe(A=0.911, T3=10000000000.0, T1=22.382, T2=5000000000.0))

# Reaction 3901
reaction('NC7H16 + H => C7H15-1 + H2', [1.880000e+05, 2.75, 6280.0])

# Reaction 3902
reaction('C7H15-1 + H2 => NC7H16 + H', [2.037000e+01, 3.38, 8618.0])

# Reaction 3903
reaction('NC7H16 + H => C7H15-2 + H2', [2.600000e+06, 2.4, 4471.0])

# Reaction 3904
reaction('C7H15-2 + H2 => NC7H16 + H', [1.807000e+01, 3.38, 9318.0])

# Reaction 3905
reaction('NC7H16 + H => C7H15-3 + H2', [2.600000e+06, 2.4, 4471.0])

# Reaction 3906
reaction('C7H15-3 + H2 => NC7H16 + H', [1.807000e+01, 3.38, 9318.0])

# Reaction 3907
reaction('NC7H16 + H => C7H15-4 + H2', [1.300000e+06, 2.4, 4471.0])

# Reaction 3908
reaction('C7H15-4 + H2 => NC7H16 + H', [1.800000e+01, 3.38, 9318.0])

# Reaction 3909
reaction('NC7H16 + O => C7H15-1 + OH', [1.930000e+05, 2.68, 3716.0])

# Reaction 3910
reaction('C7H15-1 + OH => NC7H16 + O', [1.085000e+01, 3.29, 4642.0])

# Reaction 3911
reaction('NC7H16 + O => C7H15-2 + OH', [9.540000e+04, 2.71, 2106.0])

# Reaction 3912
reaction('C7H15-2 + OH => NC7H16 + O', [3.481000e-01, 3.67, 5541.0])

# Reaction 3913
reaction('NC7H16 + O => C7H15-3 + OH', [9.540000e+04, 2.71, 2106.0])

# Reaction 3914
reaction('C7H15-3 + OH => NC7H16 + O', [3.481000e-01, 3.67, 5541.0])

# Reaction 3915
reaction('NC7H16 + O => C7H15-4 + OH', [4.770000e+04, 2.71, 2106.0])

# Reaction 3916
reaction('C7H15-4 + OH => NC7H16 + O', [3.468000e-01, 3.67, 5541.0])

# Reaction 3917
reaction('NC7H16 + OH => C7H15-1 + H2O', [2.570000e+07, 1.8, 954.0])

# Reaction 3918
reaction('C7H15-1 + H2O => NC7H16 + OH', [2.952000e+04, 2.33, 18180.0])

# Reaction 3919
reaction('NC7H16 + OH => C7H15-2 + H2O', [4.900000e+06, 2.0, -596.0])

# Reaction 3920
reaction('C7H15-2 + H2O => NC7H16 + OH', [3.624000e+02, 2.87, 19140.0])

# Reaction 3921
reaction('NC7H16 + OH => C7H15-3 + H2O', [4.900000e+06, 2.0, -596.0])

# Reaction 3922
reaction('C7H15-3 + H2O => NC7H16 + OH', [3.624000e+02, 2.87, 19140.0])

# Reaction 3923
reaction('NC7H16 + OH => C7H15-4 + H2O', [2.450000e+06, 2.0, -596.0])

# Reaction 3924
reaction('C7H15-4 + H2O => NC7H16 + OH', [3.610000e+02, 2.87, 19140.0])

# Reaction 3925
reaction('NC7H16 + HO2 => C7H15-1 + H2O2', [4.080000e+01, 3.59, 17160.0])

# Reaction 3926
reaction('C7H15-1 + H2O2 => NC7H16 + HO2', [2.557000e+00, 3.53, 3498.0])

# Reaction 3927
reaction('NC7H16 + HO2 => C7H15-2 + H2O2', [1.264000e+02, 3.37, 13720.0])

# Reaction 3928
reaction('C7H15-2 + H2O2 => NC7H16 + HO2', [4.982000e-01, 3.66, 2562.0])

# Reaction 3929
reaction('NC7H16 + HO2 => C7H15-3 + H2O2', [1.264000e+02, 3.37, 13720.0])

# Reaction 3930
reaction('C7H15-3 + H2O2 => NC7H16 + HO2', [4.982000e-01, 3.66, 2562.0])

# Reaction 3931
reaction('NC7H16 + HO2 => C7H15-4 + H2O2', [6.320000e+01, 3.37, 13720.0])

# Reaction 3932
reaction('C7H15-4 + H2O2 => NC7H16 + HO2', [4.952000e-01, 3.66, 2562.0])

# Reaction 3933
reaction('NC7H16 + CH3 => C7H15-1 + CH4', [9.040000e-01, 3.65, 7154.0])

# Reaction 3934
reaction('C7H15-1 + CH4 => NC7H16 + CH3', [8.935000e-02, 3.84, 11030.0])

# Reaction 3935
reaction('NC7H16 + CH3 => C7H15-2 + CH4', [5.410000e+04, 2.26, 7287.0])

# Reaction 3936
reaction('C7H15-2 + CH4 => NC7H16 + CH3', [3.432000e+02, 2.79, 13670.0])

# Reaction 3937
reaction('NC7H16 + CH3 => C7H15-3 + CH4', [5.410000e+04, 2.26, 7287.0])

# Reaction 3938
reaction('C7H15-3 + CH4 => NC7H16 + CH3', [3.432000e+02, 2.79, 13670.0])

# Reaction 3939
reaction('NC7H16 + CH3 => C7H15-4 + CH4', [2.705000e+04, 2.26, 7287.0])

# Reaction 3940
reaction('C7H15-4 + CH4 => NC7H16 + CH3', [3.419000e+02, 2.79, 13670.0])

# Reaction 3941
reaction('NC7H16 + O2 => C7H15-1 + HO2', [4.200000e+13, 0.0, 52800.0])

# Reaction 3942
reaction('C7H15-1 + HO2 => NC7H16 + O2', [2.386000e+10, 0.28, 453.0])

# Reaction 3943
reaction('NC7H16 + O2 => C7H15-2 + HO2', [2.800000e+13, 0.0, 50150.0])

# Reaction 3944
reaction('C7H15-2 + HO2 => NC7H16 + O2', [1.000000e+09, 0.63, 309.0])

# Reaction 3945
reaction('NC7H16 + O2 => C7H15-3 + HO2', [2.800000e+13, 0.0, 50150.0])

# Reaction 3946
reaction('C7H15-3 + HO2 => NC7H16 + O2', [1.000000e+09, 0.63, 309.0])

# Reaction 3947
reaction('NC7H16 + O2 => C7H15-4 + HO2', [1.400000e+13, 0.0, 50150.0])

# Reaction 3948
reaction('C7H15-4 + HO2 => NC7H16 + O2', [9.941000e+08, 0.63, 309.0])

# Reaction 3949
reaction('NC7H16 + C2H5 => C7H15-1 + C2H6', [1.000000e+11, 0.0, 13400.0])

# Reaction 3950
reaction('C7H15-1 + C2H6 => NC7H16 + C2H5', [3.200000e+11, 0.0, 12300.0])

# Reaction 3951
reaction('NC7H16 + C2H5 => C7H15-2 + C2H6', [1.000000e+11, 0.0, 10400.0])

# Reaction 3952
reaction('C7H15-2 + C2H6 => NC7H16 + C2H5', [1.000000e+11, 0.0, 12900.0])

# Reaction 3953
reaction('NC7H16 + C2H5 => C7H15-3 + C2H6', [1.000000e+11, 0.0, 10400.0])

# Reaction 3954
reaction('C7H15-3 + C2H6 => NC7H16 + C2H5', [1.000000e+11, 0.0, 12900.0])

# Reaction 3955
reaction('NC7H16 + C2H5 => C7H15-4 + C2H6', [5.000000e+10, 0.0, 10400.0])

# Reaction 3956
reaction('C7H15-4 + C2H6 => NC7H16 + C2H5', [1.000000e+11, 0.0, 12900.0])

# Reaction 3957
reaction('NC7H16 + CH3O => C7H15-1 + CH3OH', [3.160000e+11, 0.0, 7000.0])

# Reaction 3958
reaction('C7H15-1 + CH3OH => NC7H16 + CH3O', [1.200000e+10, 0.0, 9200.0])

# Reaction 3959
reaction('NC7H16 + CH3O => C7H15-2 + CH3OH', [2.190000e+11, 0.0, 5000.0])

# Reaction 3960
reaction('C7H15-2 + CH3OH => NC7H16 + CH3O', [8.900000e+09, 0.0, 7200.0])

# Reaction 3961
reaction('NC7H16 + CH3O => C7H15-3 + CH3OH', [2.190000e+11, 0.0, 5000.0])

# Reaction 3962
reaction('C7H15-3 + CH3OH => NC7H16 + CH3O', [8.900000e+09, 0.0, 7200.0])

# Reaction 3963
reaction('NC7H16 + CH3O => C7H15-4 + CH3OH', [1.095000e+11, 0.0, 5000.0])

# Reaction 3964
reaction('C7H15-4 + CH3OH => NC7H16 + CH3O', [8.900000e+09, 0.0, 7200.0])

# Reaction 3965
reaction('NC7H16 + C2H3 => C7H15-1 + C2H4', [1.000000e+12, 0.0, 18000.0])

# Reaction 3966
reaction('C7H15-1 + C2H4 => NC7H16 + C2H3', [2.570000e+12, 0.0, 25400.0])

# Reaction 3967
reaction('NC7H16 + C2H3 => C7H15-2 + C2H4', [8.000000e+11, 0.0, 16800.0])

# Reaction 3968
reaction('C7H15-2 + C2H4 => NC7H16 + C2H3', [2.000000e+12, 0.0, 24200.0])

# Reaction 3969
reaction('NC7H16 + C2H3 => C7H15-3 + C2H4', [8.000000e+11, 0.0, 16800.0])

# Reaction 3970
reaction('C7H15-3 + C2H4 => NC7H16 + C2H3', [2.000000e+12, 0.0, 24200.0])

# Reaction 3971
reaction('NC7H16 + C2H3 => C7H15-4 + C2H4', [4.000000e+11, 0.0, 16800.0])

# Reaction 3972
reaction('C7H15-4 + C2H4 => NC7H16 + C2H3', [2.000000e+12, 0.0, 24200.0])

# Reaction 3973
reaction('NC7H16 + CH3O2 => C7H15-1 + CH3O2H', [1.386000e+00, 3.97, 18280.0])

# Reaction 3974
reaction('C7H15-1 + CH3O2H => NC7H16 + CH3O2', [1.690000e+00, 3.46, 3024.0])

# Reaction 3975
reaction('NC7H16 + CH3O2 => C7H15-2 + CH3O2H', [2.037000e+01, 3.58, 14810.0])

# Reaction 3976
reaction('C7H15-2 + CH3O2H => NC7H16 + CH3O2', [1.562000e+00, 3.42, 2054.0])

# Reaction 3977
reaction('NC7H16 + CH3O2 => C7H15-3 + CH3O2H', [2.037000e+01, 3.58, 14810.0])

# Reaction 3978
reaction('C7H15-3 + CH3O2H => NC7H16 + CH3O2', [1.562000e+00, 3.42, 2054.0])

# Reaction 3979
reaction('NC7H16 + CH3O2 => C7H15-4 + CH3O2H', [1.019000e+01, 3.58, 14810.0])

# Reaction 3980
reaction('C7H15-4 + CH3O2H => NC7H16 + CH3O2', [1.553000e+00, 3.42, 2054.0])

# Reaction 3981
reaction('NC7H16 + O2CHO => C7H15-1 + HO2CHO', [1.680000e+13, 0.0, 20440.0])

# Reaction 3982
reaction('C7H15-1 + HO2CHO => NC7H16 + O2CHO', [1.170000e+03, 2.25, 3057.0])

# Reaction 3983
reaction('NC7H16 + O2CHO => C7H15-2 + HO2CHO', [1.120000e+13, 0.0, 17690.0])

# Reaction 3984
reaction('C7H15-2 + HO2CHO => NC7H16 + O2CHO', [2.482000e+01, 2.64, 2823.0])

# Reaction 3985
reaction('NC7H16 + O2CHO => C7H15-3 + HO2CHO', [1.120000e+13, 0.0, 17690.0])

# Reaction 3986
reaction('C7H15-3 + HO2CHO => NC7H16 + O2CHO', [2.482000e+01, 2.64, 2823.0])

# Reaction 3987
reaction('NC7H16 + O2CHO => C7H15-4 + HO2CHO', [5.600000e+12, 0.0, 17690.0])

# Reaction 3988
reaction('C7H15-4 + HO2CHO => NC7H16 + O2CHO', [2.472000e+01, 2.64, 2823.0])

# Reaction 3989
reaction('NC7H16 + C7H15O2-1 => C7H15-1 + C7H15O2H-1', [1.210000e+13, 0.0, 20430.0])

# Reaction 3990
reaction('C7H15-1 + C7H15O2H-1 => NC7H16 + C7H15O2-1', [1.440000e+10, 0.0, 15000.0])

# Reaction 3991
reaction('NC7H16 + C7H15O2-2 => C7H15-1 + C7H15O2H-2', [1.210000e+13, 0.0, 20430.0])

# Reaction 3992
reaction('C7H15-1 + C7H15O2H-2 => NC7H16 + C7H15O2-2', [1.440000e+10, 0.0, 15000.0])

# Reaction 3993
reaction('NC7H16 + C7H15O2-3 => C7H15-1 + C7H15O2H-3', [1.210000e+13, 0.0, 20430.0])

# Reaction 3994
reaction('C7H15-1 + C7H15O2H-3 => NC7H16 + C7H15O2-3', [1.440000e+10, 0.0, 15000.0])

# Reaction 3995
reaction('NC7H16 + C7H15O2-4 => C7H15-1 + C7H15O2H-4', [1.210000e+13, 0.0, 20430.0])

# Reaction 3996
reaction('C7H15-1 + C7H15O2H-4 => NC7H16 + C7H15O2-4', [1.440000e+10, 0.0, 15000.0])

# Reaction 3997
reaction('NC7H16 + C7H15O2-1 => C7H15-2 + C7H15O2H-1', [8.064000e+12, 0.0, 17700.0])

# Reaction 3998
reaction('C7H15-2 + C7H15O2H-1 => NC7H16 + C7H15O2-1', [1.440000e+10, 0.0, 15000.0])

# Reaction 3999
reaction('NC7H16 + C7H15O2-2 => C7H15-2 + C7H15O2H-2', [8.064000e+12, 0.0, 17700.0])

# Reaction 4000
reaction('C7H15-2 + C7H15O2H-2 => NC7H16 + C7H15O2-2', [1.440000e+10, 0.0, 15000.0])

# Reaction 4001
reaction('NC7H16 + C7H15O2-3 => C7H15-2 + C7H15O2H-3', [8.064000e+12, 0.0, 17700.0])

# Reaction 4002
reaction('C7H15-2 + C7H15O2H-3 => NC7H16 + C7H15O2-3', [1.440000e+10, 0.0, 15000.0])

# Reaction 4003
reaction('NC7H16 + C7H15O2-4 => C7H15-2 + C7H15O2H-4', [8.064000e+12, 0.0, 17700.0])

# Reaction 4004
reaction('C7H15-2 + C7H15O2H-4 => NC7H16 + C7H15O2-4', [1.440000e+10, 0.0, 15000.0])

# Reaction 4005
reaction('NC7H16 + C7H15O2-1 => C7H15-3 + C7H15O2H-1', [8.064000e+12, 0.0, 17700.0])

# Reaction 4006
reaction('C7H15-3 + C7H15O2H-1 => NC7H16 + C7H15O2-1', [1.440000e+10, 0.0, 15000.0])

# Reaction 4007
reaction('NC7H16 + C7H15O2-2 => C7H15-3 + C7H15O2H-2', [8.064000e+12, 0.0, 17700.0])

# Reaction 4008
reaction('C7H15-3 + C7H15O2H-2 => NC7H16 + C7H15O2-2', [1.440000e+10, 0.0, 15000.0])

# Reaction 4009
reaction('NC7H16 + C7H15O2-3 => C7H15-3 + C7H15O2H-3', [8.064000e+12, 0.0, 17700.0])

# Reaction 4010
reaction('C7H15-3 + C7H15O2H-3 => NC7H16 + C7H15O2-3', [1.440000e+10, 0.0, 15000.0])

# Reaction 4011
reaction('NC7H16 + C7H15O2-4 => C7H15-3 + C7H15O2H-4', [8.064000e+12, 0.0, 17700.0])

# Reaction 4012
reaction('C7H15-3 + C7H15O2H-4 => NC7H16 + C7H15O2-4', [1.440000e+10, 0.0, 15000.0])

# Reaction 4013
reaction('NC7H16 + C7H15O2-1 => C7H15-4 + C7H15O2H-1', [4.032000e+12, 0.0, 17700.0])

# Reaction 4014
reaction('C7H15-4 + C7H15O2H-1 => NC7H16 + C7H15O2-1', [1.440000e+10, 0.0, 15000.0])

# Reaction 4015
reaction('NC7H16 + C7H15O2-2 => C7H15-4 + C7H15O2H-2', [4.032000e+12, 0.0, 17700.0])

# Reaction 4016
reaction('C7H15-4 + C7H15O2H-2 => NC7H16 + C7H15O2-2', [1.440000e+10, 0.0, 15000.0])

# Reaction 4017
reaction('NC7H16 + C7H15O2-3 => C7H15-4 + C7H15O2H-3', [4.032000e+12, 0.0, 17700.0])

# Reaction 4018
reaction('C7H15-4 + C7H15O2H-3 => NC7H16 + C7H15O2-3', [1.440000e+10, 0.0, 15000.0])

# Reaction 4019
reaction('NC7H16 + C7H15O2-4 => C7H15-4 + C7H15O2H-4', [4.032000e+12, 0.0, 17700.0])

# Reaction 4020
reaction('C7H15-4 + C7H15O2H-4 => NC7H16 + C7H15O2-4', [1.440000e+10, 0.0, 15000.0])

# Reaction 4021
reaction('NC7H16 + C7H15-1 => C7H15-2 + NC7H16', [1.000000e+11, 0.0, 10400.0])

# Reaction 4022
reaction('C7H15-2 + NC7H16 => NC7H16 + C7H15-1', [1.500000e+11, 0.0, 12300.0])

# Reaction 4023
reaction('NC7H16 + C7H15-1 => C7H15-3 + NC7H16', [1.000000e+11, 0.0, 10400.0])

# Reaction 4024
reaction('C7H15-3 + NC7H16 => NC7H16 + C7H15-1', [1.500000e+11, 0.0, 12300.0])

# Reaction 4025
reaction('NC7H16 + C7H15-1 => C7H15-4 + NC7H16', [5.000000e+10, 0.0, 10400.0])

# Reaction 4026
reaction('C7H15-4 + NC7H16 => NC7H16 + C7H15-1', [1.500000e+11, 0.0, 12300.0])

# Reaction 4027
reaction('NC7H16 + C7H15-2 => C7H15-3 + NC7H16', [1.000000e+11, 0.0, 10400.0])

# Reaction 4028
reaction('C7H15-3 + NC7H16 => NC7H16 + C7H15-2', [1.000000e+11, 0.0, 10400.0])

# Reaction 4029
reaction('NC7H16 + C7H15-2 => C7H15-4 + NC7H16', [5.000000e+10, 0.0, 10400.0])

# Reaction 4030
reaction('C7H15-4 + NC7H16 => NC7H16 + C7H15-2', [1.000000e+11, 0.0, 10400.0])

# Reaction 4031
reaction('NC7H16 + C7H15-3 => C7H15-4 + NC7H16', [5.000000e+10, 0.0, 10400.0])

# Reaction 4032
reaction('C7H15-4 + NC7H16 => NC7H16 + C7H15-3', [1.000000e+11, 0.0, 10400.0])

# Reaction 4033
reaction('C7H15-1 => C5H11-1 + C2H4', [1.227000e+19, -1.91, 31400.0])

# Reaction 4034
reaction('C5H11-1 + C2H4 => C7H15-1', [1.000000e+11, 0.0, 8200.0])

# Reaction 4035
reaction('C7H15-1 => C7H14-1 + H', [9.648000e+13, -0.26, 36010.0])

# Reaction 4036
reaction('C7H14-1 + H => C7H15-1', [1.000000e+13, 0.0, 2900.0])

# Reaction 4037
reaction('C7H15-2 => PC4H9 + C3H6', [9.764000e+18, -1.79, 31360.0])

# Reaction 4038
reaction('PC4H9 + C3H6 => C7H15-2', [1.000000e+11, 0.0, 8200.0])

# Reaction 4039
reaction('C7H15-2 => C7H14-1 + H', [6.067000e+12, 0.09, 36810.0])

# Reaction 4040
reaction('C7H14-1 + H => C7H15-2', [1.000000e+13, 0.0, 1200.0])

# Reaction 4041
reaction('C7H15-2 => C7H14-2 + H', [1.229000e+13, -0.08, 35640.0])

# Reaction 4042
reaction('C7H14-2 + H => C7H15-2', [1.000000e+13, 0.0, 2900.0])

# Reaction 4043
reaction('C7H15-3 => C4H8-1 + NC3H7', [6.165000e+18, -1.71, 30960.0])

# Reaction 4044
reaction('C4H8-1 + NC3H7 => C7H15-3', [1.000000e+11, 0.0, 7700.0])

# Reaction 4045
reaction('C7H15-3 => C6H12-1 + CH3', [2.698000e+17, -1.35, 31480.0])

# Reaction 4046
reaction('C6H12-1 + CH3 => C7H15-3', [1.750000e+11, 0.0, 7200.0])

# Reaction 4047
reaction('C7H15-3 => C7H14-2 + H', [1.229000e+13, -0.08, 35640.0])

# Reaction 4048
reaction('C7H14-2 + H => C7H15-3', [1.000000e+13, 0.0, 2900.0])

# Reaction 4049
reaction('C7H15-3 => C7H14-3 + H', [9.152000e+12, -0.02, 35730.0])

# Reaction 4050
reaction('C7H14-3 + H => C7H15-3', [1.000000e+13, 0.0, 2900.0])

# Reaction 4051
reaction('C7H15-4 => C2H5 + C5H10-1', [1.143000e+18, -1.34, 31430.0])

# Reaction 4052
reaction('C2H5 + C5H10-1 => C7H15-4', [1.000000e+11, 0.0, 8200.0])

# Reaction 4053
reaction('C7H15-4 => C7H14-3 + H', [1.819000e+13, -0.02, 35730.0])

# Reaction 4054
reaction('C7H14-3 + H => C7H15-4', [1.000000e+13, 0.0, 2900.0])

# Reaction 4055
reaction('C7H15-1 + O2 => C7H14-1 + HO2', [3.000000e-09, 0.0, 3000.0])

# Reaction 4056
reaction('C7H14-1 + HO2 => C7H15-1 + O2', [6.517000e-10, 0.18, 18760.0])

# Reaction 4057
reaction('C7H15-2 + O2 => C7H14-1 + HO2', [4.500000e-09, 0.0, 5020.0])

# Reaction 4058
reaction('C7H14-1 + HO2 => C7H15-2 + O2', [1.532000e-08, -0.16, 18270.0])

# Reaction 4059
reaction('C7H15-2 + O2 => C7H14-2 + HO2', [3.000000e-09, 0.0, 3000.0])

# Reaction 4060
reaction('C7H14-2 + HO2 => C7H15-2 + O2', [5.040000e-09, 0.01, 19120.0])

# Reaction 4061
reaction('C7H15-3 + O2 => C7H14-2 + HO2', [3.000000e-09, 0.0, 3000.0])

# Reaction 4062
reaction('C7H14-2 + HO2 => C7H15-3 + O2', [5.040000e-09, 0.01, 19120.0])

# Reaction 4063
reaction('C7H15-3 + O2 => C7H14-3 + HO2', [3.000000e-09, 0.0, 3000.0])

# Reaction 4064
reaction('C7H14-3 + HO2 => C7H15-3 + O2', [6.769000e-09, -0.05, 19030.0])

# Reaction 4065
reaction('C7H15-4 + O2 => C7H14-3 + HO2', [6.000000e-09, 0.0, 3000.0])

# Reaction 4066
reaction('C7H14-3 + HO2 => C7H15-4 + O2', [6.795000e-09, -0.05, 19030.0])

# Reaction 4067
reaction('C7H15-1 => C7H15-3', [1.386000e+09, 0.98, 33760.0])

# Reaction 4068
reaction('C7H15-3 => C7H15-1', [8.716000e+07, 1.33, 36270.0])

# Reaction 4069
reaction('C7H15-1 => C7H15-4', [2.541000e+09, 0.35, 19760.0])

# Reaction 4070
reaction('C7H15-4 => C7H15-1', [3.176000e+08, 0.7, 22270.0])

# Reaction 4071
reaction('C7H15-2 => C7H15-3', [9.587000e+08, 1.39, 39700.0])

# Reaction 4072
reaction('C7H15-3 => C7H15-2', [9.587000e+08, 1.39, 39700.0])

# Reaction 4073
reaction('C7H15-1 => C7H15-2', [5.478000e+08, 1.62, 38760.0])

# Reaction 4074
reaction('C7H15-2 => C7H15-1', [3.445000e+07, 1.97, 41270.0])

# Reaction 4075
reaction('C7H14-1 + H => C7H131-3 + H2', [3.376000e+05, 2.36, 207.0])

# Reaction 4076
reaction('C7H131-3 + H2 => C7H14-1 + H', [4.393000e+06, 2.1, 20620.0])

# Reaction 4077
reaction('C7H14-1 + H => C7H131-4 + H2', [1.300000e+06, 2.4, 4471.0])

# Reaction 4078
reaction('C7H131-4 + H2 => C7H14-1 + H', [3.834000e+03, 2.74, 11540.0])

# Reaction 4079
reaction('C7H14-1 + H => C7H131-5 + H2', [1.300000e+06, 2.4, 4471.0])

# Reaction 4080
reaction('C7H131-5 + H2 => C7H14-1 + H', [3.834000e+03, 2.74, 11540.0])

# Reaction 4081
reaction('C7H14-1 + H => C7H131-6 + H2', [1.300000e+06, 2.4, 4471.0])

# Reaction 4082
reaction('C7H131-6 + H2 => C7H14-1 + H', [3.834000e+03, 2.74, 11540.0])

# Reaction 4083
reaction('C7H14-1 + H => C7H131-7 + H2', [6.650000e+05, 2.54, 6756.0])

# Reaction 4084
reaction('C7H131-7 + H2 => C7H14-1 + H', [3.056000e+04, 2.54, 11320.0])

# Reaction 4085
reaction('C7H14-1 + OH => C7H131-3 + H2O', [2.764000e+04, 2.64, -1919.0])

# Reaction 4086
reaction('C7H131-3 + H2O => C7H14-1 + OH', [1.557000e+06, 2.38, 33650.0])

# Reaction 4087
reaction('C7H14-1 + OH => C7H131-4 + H2O', [4.670000e+07, 1.61, -35.0])

# Reaction 4088
reaction('C7H131-4 + H2O => C7H14-1 + OH', [5.963000e+05, 1.95, 22190.0])

# Reaction 4089
reaction('C7H14-1 + OH => C7H131-5 + H2O', [4.670000e+07, 1.61, -35.0])

# Reaction 4090
reaction('C7H131-5 + H2O => C7H14-1 + OH', [5.963000e+05, 1.95, 22190.0])

# Reaction 4091
reaction('C7H14-1 + OH => C7H131-6 + H2O', [4.670000e+07, 1.61, -35.0])

# Reaction 4092
reaction('C7H131-6 + H2O => C7H14-1 + OH', [5.963000e+05, 1.95, 22190.0])

# Reaction 4093
reaction('C7H14-1 + OH => C7H131-7 + H2O', [5.270000e+09, 0.97, 1586.0])

# Reaction 4094
reaction('C7H131-7 + H2O => C7H14-1 + OH', [1.049000e+09, 0.97, 21300.0])

# Reaction 4095
reaction('C7H14-1 + CH3 => C7H131-3 + CH4', [3.690000e+00, 3.31, 4002.0])

# Reaction 4096
reaction('C7H131-3 + CH4 => C7H14-1 + CH3', [1.254000e+03, 3.05, 24890.0])

# Reaction 4097
reaction('C7H14-1 + CH3 => C7H131-4 + CH4', [1.510000e+00, 3.46, 5481.0])

# Reaction 4098
reaction('C7H131-4 + CH4 => C7H14-1 + CH3', [1.163000e-01, 3.8, 13030.0])

# Reaction 4099
reaction('C7H14-1 + CH3 => C7H131-5 + CH4', [1.510000e+00, 3.46, 5481.0])

# Reaction 4100
reaction('C7H131-5 + CH4 => C7H14-1 + CH3', [1.163000e-01, 3.8, 13030.0])

# Reaction 4101
reaction('C7H14-1 + CH3 => C7H131-6 + CH4', [1.510000e+00, 3.46, 5481.0])

# Reaction 4102
reaction('C7H131-6 + CH4 => C7H14-1 + CH3', [1.163000e-01, 3.8, 13030.0])

# Reaction 4103
reaction('C7H14-1 + CH3 => C7H131-7 + CH4', [4.521000e-01, 3.65, 7154.0])

# Reaction 4104
reaction('C7H131-7 + CH4 => C7H14-1 + CH3', [5.427000e-01, 3.65, 12190.0])

# Reaction 4105
reaction('C7H14-1 + HO2 => C7H131-3 + H2O2', [4.820000e+03, 2.55, 10530.0])

# Reaction 4106
reaction('C7H131-3 + H2O2 => C7H14-1 + HO2', [1.612000e+06, 1.96, 14630.0])

# Reaction 4107
reaction('C7H14-1 + HO2 => C7H131-4 + H2O2', [9.640000e+03, 2.6, 13910.0])

# Reaction 4108
reaction('C7H131-4 + H2O2 => C7H14-1 + HO2', [7.307000e+02, 2.62, 4672.0])

# Reaction 4109
reaction('C7H14-1 + HO2 => C7H131-5 + H2O2', [9.640000e+03, 2.6, 13910.0])

# Reaction 4110
reaction('C7H131-5 + H2O2 => C7H14-1 + HO2', [7.307000e+02, 2.62, 4672.0])

# Reaction 4111
reaction('C7H14-1 + HO2 => C7H131-6 + H2O2', [9.640000e+03, 2.6, 13910.0])

# Reaction 4112
reaction('C7H131-6 + H2O2 => C7H14-1 + HO2', [7.307000e+02, 2.62, 4672.0])

# Reaction 4113
reaction('C7H14-1 + HO2 => C7H131-7 + H2O2', [2.380000e+04, 2.55, 16490.0])

# Reaction 4114
reaction('C7H131-7 + H2O2 => C7H14-1 + HO2', [2.811000e+04, 2.22, 4746.0])

# Reaction 4115
reaction('C7H14-1 + CH3O2 => C7H131-3 + CH3O2H', [4.820000e+03, 2.55, 10530.0])

# Reaction 4116
reaction('C7H131-3 + CH3O2H => C7H14-1 + CH3O2', [3.357000e+06, 1.79, 11610.0])

# Reaction 4117
reaction('C7H14-1 + CH3O2 => C7H131-4 + CH3O2H', [9.640000e+03, 2.6, 13910.0])

# Reaction 4118
reaction('C7H131-4 + CH3O2H => C7H14-1 + CH3O2', [1.521000e+03, 2.44, 1652.0])

# Reaction 4119
reaction('C7H14-1 + CH3O2 => C7H131-5 + CH3O2H', [9.640000e+03, 2.6, 13910.0])

# Reaction 4120
reaction('C7H131-5 + CH3O2H => C7H14-1 + CH3O2', [1.521000e+03, 2.44, 1652.0])

# Reaction 4121
reaction('C7H14-1 + CH3O2 => C7H131-6 + CH3O2H', [9.640000e+03, 2.6, 13910.0])

# Reaction 4122
reaction('C7H131-6 + CH3O2H => C7H14-1 + CH3O2', [1.521000e+03, 2.44, 1652.0])

# Reaction 4123
reaction('C7H14-1 + CH3O2 => C7H131-7 + CH3O2H', [2.380000e+04, 2.55, 16490.0])

# Reaction 4124
reaction('C7H131-7 + CH3O2H => C7H14-1 + CH3O2', [5.853000e+04, 2.04, 1726.0])

# Reaction 4125
reaction('C7H14-1 + CH3O => C7H131-3 + CH3OH', [4.000000e+01, 2.9, 8609.0])

# Reaction 4126
reaction('C7H131-3 + CH3OH => C7H14-1 + CH3O', [2.510000e+02, 2.66, 27290.0])

# Reaction 4127
reaction('C7H14-1 + CH3O => C7H131-4 + CH3OH', [1.450000e+11, 0.0, 4571.0])

# Reaction 4128
reaction('C7H131-4 + CH3OH => C7H14-1 + CH3O', [2.062000e+08, 0.37, 9911.0])

# Reaction 4129
reaction('C7H14-1 + CH3O => C7H131-5 + CH3OH', [1.450000e+11, 0.0, 4571.0])

# Reaction 4130
reaction('C7H131-5 + CH3OH => C7H14-1 + CH3O', [2.062000e+08, 0.37, 9911.0])

# Reaction 4131
reaction('C7H14-1 + CH3O => C7H131-6 + CH3OH', [1.450000e+11, 0.0, 4571.0])

# Reaction 4132
reaction('C7H131-6 + CH3OH => C7H14-1 + CH3O', [2.062000e+08, 0.37, 9911.0])

# Reaction 4133
reaction('C7H14-1 + CH3O => C7H131-7 + CH3OH', [2.170000e+11, 0.0, 6458.0])

# Reaction 4134
reaction('C7H131-7 + CH3OH => C7H14-1 + CH3O', [4.808000e+09, 0.02, 9288.0])

# Reaction 4135
reaction('C7H14-2 + H => C7H131-3 + H2', [1.730000e+05, 2.5, 2492.0])

# Reaction 4136
reaction('C7H131-3 + H2 => C7H14-2 + H', [4.561000e+06, 2.06, 20030.0])

# Reaction 4137
reaction('C7H14-2 + H => C7H132-4 + H2', [3.376000e+05, 2.36, 207.0])

# Reaction 4138
reaction('C7H132-4 + H2 => C7H14-2 + H', [4.400000e+06, 2.1, 20590.0])

# Reaction 4139
reaction('C7H14-2 + H => C7H132-5 + H2', [1.300000e+06, 2.4, 4471.0])

# Reaction 4140
reaction('C7H132-5 + H2 => C7H14-2 + H', [3.843000e+03, 2.74, 11520.0])

# Reaction 4141
reaction('C7H14-2 + H => C7H132-6 + H2', [1.300000e+06, 2.4, 4471.0])

# Reaction 4142
reaction('C7H132-6 + H2 => C7H14-2 + H', [3.843000e+03, 2.74, 11520.0])

# Reaction 4143
reaction('C7H14-2 + H => C7H132-7 + H2', [6.651000e+05, 2.54, 6756.0])

# Reaction 4144
reaction('C7H132-7 + H2 => C7H14-2 + H', [3.149000e+04, 2.53, 11300.0])

# Reaction 4145
reaction('C7H14-2 + OH => C7H131-3 + H2O', [3.120000e+06, 2.0, -298.0])

# Reaction 4146
reaction('C7H131-3 + H2O => C7H14-2 + OH', [3.561000e+08, 1.56, 32400.0])

# Reaction 4147
reaction('C7H14-2 + OH => C7H132-4 + H2O', [2.764000e+04, 2.64, -1919.0])

# Reaction 4148
reaction('C7H132-4 + H2O => C7H14-2 + OH', [1.560000e+06, 2.38, 33620.0])

# Reaction 4149
reaction('C7H14-2 + OH => C7H132-5 + H2O', [4.670000e+05, 1.61, -35.0])

# Reaction 4150
reaction('C7H132-5 + H2O => C7H14-2 + OH', [5.978000e+03, 1.95, 22170.0])

# Reaction 4151
reaction('C7H14-2 + OH => C7H132-6 + H2O', [4.670000e+05, 1.61, -35.0])

# Reaction 4152
reaction('C7H132-6 + H2O => C7H14-2 + OH', [5.978000e+03, 1.95, 22170.0])

# Reaction 4153
reaction('C7H14-2 + OH => C7H132-7 + H2O', [5.270000e+09, 0.97, 1586.0])

# Reaction 4154
reaction('C7H132-7 + H2O => C7H14-2 + OH', [1.080000e+09, 0.96, 21280.0])

# Reaction 4155
reaction('C7H14-2 + CH3 => C7H131-3 + CH4', [2.210000e+00, 3.5, 5675.0])

# Reaction 4156
reaction('C7H131-3 + CH4 => C7H14-2 + CH3', [1.522000e+03, 3.06, 23700.0])

# Reaction 4157
reaction('C7H14-2 + CH3 => C7H132-4 + CH4', [3.690000e+00, 3.31, 4002.0])

# Reaction 4158
reaction('C7H132-4 + CH4 => C7H14-2 + CH3', [1.256000e+03, 3.05, 24860.0])

# Reaction 4159
reaction('C7H14-2 + CH3 => C7H132-5 + CH4', [1.510000e+00, 3.46, 5481.0])

# Reaction 4160
reaction('C7H132-5 + CH4 => C7H14-2 + CH3', [1.166000e-01, 3.8, 13010.0])

# Reaction 4161
reaction('C7H14-2 + CH3 => C7H132-6 + CH4', [1.510000e+00, 3.46, 5481.0])

# Reaction 4162
reaction('C7H132-6 + CH4 => C7H14-2 + CH3', [1.166000e-01, 3.8, 13010.0])

# Reaction 4163
reaction('C7H14-2 + CH3 => C7H132-7 + CH4', [4.521000e-01, 3.65, 7154.0])

# Reaction 4164
reaction('C7H132-7 + CH4 => C7H14-2 + CH3', [5.591000e-01, 3.64, 12170.0])

# Reaction 4165
reaction('C7H14-2 + HO2 => C7H131-3 + H2O2', [9.639000e+03, 2.6, 13910.0])

# Reaction 4166
reaction('C7H131-3 + H2O2 => C7H14-2 + HO2', [6.531000e+06, 1.84, 15140.0])

# Reaction 4167
reaction('C7H14-2 + HO2 => C7H132-4 + H2O2', [4.820000e+03, 2.55, 10530.0])

# Reaction 4168
reaction('C7H132-4 + H2O2 => C7H14-2 + HO2', [1.615000e+06, 1.96, 14600.0])

# Reaction 4169
reaction('C7H14-2 + HO2 => C7H132-5 + H2O2', [9.640000e+03, 2.6, 13910.0])

# Reaction 4170
reaction('C7H132-5 + H2O2 => C7H14-2 + HO2', [7.325000e+02, 2.62, 4652.0])

# Reaction 4171
reaction('C7H14-2 + HO2 => C7H132-6 + H2O2', [9.640000e+03, 2.6, 13910.0])

# Reaction 4172
reaction('C7H132-6 + H2O2 => C7H14-2 + HO2', [7.325000e+02, 2.62, 4652.0])

# Reaction 4173
reaction('C7H14-2 + HO2 => C7H132-7 + H2O2', [2.380000e+04, 2.55, 16490.0])

# Reaction 4174
reaction('C7H132-7 + H2O2 => C7H14-2 + HO2', [2.896000e+04, 2.21, 4726.0])

# Reaction 4175
reaction('C7H14-2 + CH3O2 => C7H131-3 + CH3O2H', [9.639000e+03, 2.6, 13910.0])

# Reaction 4176
reaction('C7H131-3 + CH3O2H => C7H14-2 + CH3O2', [1.360000e+07, 1.66, 12120.0])

# Reaction 4177
reaction('C7H14-2 + CH3O2 => C7H132-4 + CH3O2H', [4.820000e+03, 2.55, 10530.0])

# Reaction 4178
reaction('C7H132-4 + CH3O2H => C7H14-2 + CH3O2', [3.362000e+06, 1.79, 11580.0])

# Reaction 4179
reaction('C7H14-2 + CH3O2 => C7H132-5 + CH3O2H', [9.640000e+03, 2.6, 13910.0])

# Reaction 4180
reaction('C7H132-5 + CH3O2H => C7H14-2 + CH3O2', [1.525000e+03, 2.44, 1632.0])

# Reaction 4181
reaction('C7H14-2 + CH3O2 => C7H132-6 + CH3O2H', [9.640000e+03, 2.6, 13910.0])

# Reaction 4182
reaction('C7H132-6 + CH3O2H => C7H14-2 + CH3O2', [1.525000e+03, 2.44, 1632.0])

# Reaction 4183
reaction('C7H14-2 + CH3O2 => C7H132-7 + CH3O2H', [2.380000e+04, 2.55, 16490.0])

# Reaction 4184
reaction('C7H132-7 + CH3O2H => C7H14-2 + CH3O2', [6.030000e+04, 2.04, 1706.0])

# Reaction 4185
reaction('C7H14-2 + CH3O => C7H131-3 + CH3OH', [9.000000e+01, 2.95, 11990.0])

# Reaction 4186
reaction('C7H131-3 + CH3OH => C7H14-2 + CH3O', [1.144000e+03, 2.54, 27800.0])

# Reaction 4187
reaction('C7H14-2 + CH3O => C7H132-4 + CH3OH', [4.000000e+01, 2.9, 8609.0])

# Reaction 4188
reaction('C7H132-4 + CH3OH => C7H14-2 + CH3O', [2.513000e+02, 2.66, 27260.0])

# Reaction 4189
reaction('C7H14-2 + CH3O => C7H132-5 + CH3OH', [1.450000e+11, 0.0, 4571.0])

# Reaction 4190
reaction('C7H132-5 + CH3OH => C7H14-2 + CH3O', [2.067000e+08, 0.37, 9891.0])

# Reaction 4191
reaction('C7H14-2 + CH3O => C7H132-6 + CH3OH', [1.450000e+11, 0.0, 4571.0])

# Reaction 4192
reaction('C7H132-6 + CH3OH => C7H14-2 + CH3O', [2.067000e+08, 0.37, 9891.0])

# Reaction 4193
reaction('C7H14-2 + CH3O => C7H132-7 + CH3OH', [2.170000e+11, 0.0, 6458.0])

# Reaction 4194
reaction('C7H132-7 + CH3OH => C7H14-2 + CH3O', [4.953000e+09, 0.02, 9268.0])

# Reaction 4195
reaction('C7H14-3 + H => C7H133-1 + H2', [6.651000e+05, 2.54, 6756.0])

# Reaction 4196
reaction('C7H133-1 + H2 => C7H14-3 + H', [3.097000e+04, 2.53, 11260.0])

# Reaction 4197
reaction('C7H14-3 + H => C7H132-4 + H2', [3.376000e+05, 2.36, 207.0])

# Reaction 4198
reaction('C7H132-4 + H2 => C7H14-3 + H', [3.276000e+06, 2.16, 20680.0])

# Reaction 4199
reaction('C7H14-3 + H => C7H133-5 + H2', [3.376000e+05, 2.36, 207.0])

# Reaction 4200
reaction('C7H133-5 + H2 => C7H14-3 + H', [4.380000e+06, 2.1, 20560.0])

# Reaction 4201
reaction('C7H14-3 + H => C7H133-6 + H2', [1.300000e+06, 2.4, 4471.0])

# Reaction 4202
reaction('C7H133-6 + H2 => C7H14-3 + H', [3.916000e+03, 2.74, 11490.0])

# Reaction 4203
reaction('C7H14-3 + H => C7H133-7 + H2', [6.651000e+05, 2.54, 6756.0])

# Reaction 4204
reaction('C7H133-7 + H2 => C7H14-3 + H', [3.097000e+04, 2.53, 11260.0])

# Reaction 4205
reaction('C7H14-3 + OH => C7H133-1 + H2O', [1.054000e+10, 0.97, 1586.0])

# Reaction 4206
reaction('C7H133-1 + H2O => C7H14-3 + OH', [2.125000e+09, 0.96, 21240.0])

# Reaction 4207
reaction('C7H14-3 + OH => C7H132-4 + H2O', [2.764000e+04, 2.64, -1919.0])

# Reaction 4208
reaction('C7H132-4 + H2O => C7H14-3 + OH', [1.161000e+06, 2.44, 33710.0])

# Reaction 4209
reaction('C7H14-3 + OH => C7H133-5 + H2O', [2.764000e+04, 2.64, -1919.0])

# Reaction 4210
reaction('C7H133-5 + H2O => C7H14-3 + OH', [1.552000e+06, 2.38, 33590.0])

# Reaction 4211
reaction('C7H14-3 + OH => C7H133-6 + H2O', [4.670000e+07, 1.61, -35.0])

# Reaction 4212
reaction('C7H133-6 + H2O => C7H14-3 + OH', [6.091000e+05, 1.95, 22140.0])

# Reaction 4213
reaction('C7H14-3 + OH => C7H133-7 + H2O', [1.054000e+10, 0.97, 1586.0])

# Reaction 4214
reaction('C7H133-7 + H2O => C7H14-3 + OH', [2.125000e+09, 0.96, 21240.0])

# Reaction 4215
reaction('C7H14-3 + CH3 => C7H133-1 + CH4', [9.042000e-01, 3.65, 7154.0])

# Reaction 4216
reaction('C7H133-1 + CH4 => C7H14-3 + CH3', [1.100000e+00, 3.64, 12130.0])

# Reaction 4217
reaction('C7H14-3 + CH3 => C7H132-4 + CH4', [3.690000e+00, 3.31, 4002.0])

# Reaction 4218
reaction('C7H132-4 + CH4 => C7H14-3 + CH3', [9.355000e+02, 3.11, 24950.0])

# Reaction 4219
reaction('C7H14-3 + CH3 => C7H133-5 + CH4', [3.690000e+00, 3.31, 4002.0])

# Reaction 4220
reaction('C7H133-5 + CH4 => C7H14-3 + CH3', [1.251000e+03, 3.05, 24830.0])

# Reaction 4221
reaction('C7H14-3 + CH3 => C7H133-6 + CH4', [1.510000e+00, 3.46, 5481.0])

# Reaction 4222
reaction('C7H133-6 + CH4 => C7H14-3 + CH3', [1.188000e-01, 3.8, 12980.0])

# Reaction 4223
reaction('C7H14-3 + CH3 => C7H133-7 + CH4', [9.042000e-01, 3.65, 7154.0])

# Reaction 4224
reaction('C7H133-7 + CH4 => C7H14-3 + CH3', [1.100000e+00, 3.64, 12130.0])

# Reaction 4225
reaction('C7H14-3 + HO2 => C7H133-1 + H2O2', [4.760000e+04, 2.55, 16490.0])

# Reaction 4226
reaction('C7H133-1 + H2O2 => C7H14-3 + HO2', [5.697000e+04, 2.22, 4686.0])

# Reaction 4227
reaction('C7H14-3 + HO2 => C7H132-4 + H2O2', [4.820000e+03, 2.55, 10530.0])

# Reaction 4228
reaction('C7H132-4 + H2O2 => C7H14-3 + HO2', [1.202000e+06, 2.03, 14690.0])

# Reaction 4229
reaction('C7H14-3 + HO2 => C7H133-5 + H2O2', [4.820000e+03, 2.55, 10530.0])

# Reaction 4230
reaction('C7H133-5 + H2O2 => C7H14-3 + HO2', [1.607000e+06, 1.96, 14570.0])

# Reaction 4231
reaction('C7H14-3 + HO2 => C7H133-6 + H2O2', [9.640000e+03, 2.6, 13910.0])

# Reaction 4232
reaction('C7H133-6 + H2O2 => C7H14-3 + HO2', [7.464000e+02, 2.61, 4622.0])

# Reaction 4233
reaction('C7H14-3 + HO2 => C7H133-7 + H2O2', [4.760000e+04, 2.55, 16490.0])

# Reaction 4234
reaction('C7H133-7 + H2O2 => C7H14-3 + HO2', [5.697000e+04, 2.22, 4686.0])

# Reaction 4235
reaction('C7H14-3 + CH3O2 => C7H133-1 + CH3O2H', [4.760000e+04, 2.55, 16490.0])

# Reaction 4236
reaction('C7H133-1 + CH3O2H => C7H14-3 + CH3O2', [1.186000e+05, 2.04, 1666.0])

# Reaction 4237
reaction('C7H14-3 + CH3O2 => C7H132-4 + CH3O2H', [4.820000e+03, 2.55, 10530.0])

# Reaction 4238
reaction('C7H132-4 + CH3O2H => C7H14-3 + CH3O2', [2.503000e+06, 1.85, 11670.0])

# Reaction 4239
reaction('C7H14-3 + CH3O2 => C7H133-5 + CH3O2H', [4.820000e+03, 2.55, 10530.0])

# Reaction 4240
reaction('C7H133-5 + CH3O2H => C7H14-3 + CH3O2', [3.346000e+06, 1.79, 11550.0])

# Reaction 4241
reaction('C7H14-3 + CH3O2 => C7H133-6 + CH3O2H', [9.640000e+03, 2.6, 13910.0])

# Reaction 4242
reaction('C7H133-6 + CH3O2H => C7H14-3 + CH3O2', [1.554000e+03, 2.44, 1602.0])

# Reaction 4243
reaction('C7H14-3 + CH3O2 => C7H133-7 + CH3O2H', [4.760000e+04, 2.55, 16490.0])

# Reaction 4244
reaction('C7H133-7 + CH3O2H => C7H14-3 + CH3O2', [1.186000e+05, 2.04, 1666.0])

# Reaction 4245
reaction('C7H14-3 + CH3O => C7H133-1 + CH3OH', [4.340000e+11, 0.0, 6458.0])

# Reaction 4246
reaction('C7H133-1 + CH3OH => C7H14-3 + CH3O', [9.743000e+09, 0.02, 9228.0])

# Reaction 4247
reaction('C7H14-3 + CH3O => C7H132-4 + CH3OH', [4.000000e+01, 2.9, 8609.0])

# Reaction 4248
reaction('C7H132-4 + CH3OH => C7H14-3 + CH3O', [1.871000e+02, 2.73, 27350.0])

# Reaction 4249
reaction('C7H14-3 + CH3O => C7H133-5 + CH3OH', [4.000000e+01, 2.9, 8609.0])

# Reaction 4250
reaction('C7H133-5 + CH3OH => C7H14-3 + CH3O', [2.502000e+02, 2.66, 27230.0])

# Reaction 4251
reaction('C7H14-3 + CH3O => C7H133-6 + CH3OH', [1.450000e+11, 0.0, 4571.0])

# Reaction 4252
reaction('C7H133-6 + CH3OH => C7H14-3 + CH3O', [2.106000e+08, 0.37, 9861.0])

# Reaction 4253
reaction('C7H14-3 + CH3O => C7H133-7 + CH3OH', [4.340000e+11, 0.0, 6458.0])

# Reaction 4254
reaction('C7H133-7 + CH3OH => C7H14-3 + CH3O', [9.743000e+09, 0.02, 9228.0])

# Reaction 4255
reaction('C7H131-3 + HO2 => C7H13O1-3 + OH', [9.640000e+12, 0.0, 0.0])

# Reaction 4256
reaction('C7H13O1-3 + OH => C7H131-3 + HO2', [4.435000e+15, -1.02, 15400.0])

# Reaction 4257
reaction('C7H131-3 + CH3O2 => C7H13O1-3 + CH3O', [9.640000e+12, 0.0, 0.0])

# Reaction 4258
reaction('C7H13O1-3 + CH3O => C7H131-3 + CH3O2', [4.332000e+17, -1.6, 20160.0])

# Reaction 4259
reaction('C7H131-3 + C2H5O2 => C7H13O1-3 + C2H5O', [9.640000e+12, 0.0, 0.0])

# Reaction 4260
reaction('C7H13O1-3 + C2H5O => C7H131-3 + C2H5O2', [2.836000e+14, -0.68, 18000.0])

# Reaction 4261
reaction('C7H131-3 => C4H6 + NC3H7', [1.105000e+19, -1.53, 40700.0])

# Reaction 4262
reaction('C4H6 + NC3H7 => C7H131-3', [8.500000e+10, 0.0, 8300.0])

# Reaction 4263
reaction('C7H131-4 => C5H10-1 + C2H3', [2.492000e+12, 0.03, 37300.0])

# Reaction 4264
reaction('C5H10-1 + C2H3 => C7H131-4', [8.500000e+10, 0.0, 8300.0])

# Reaction 4265
reaction('C7H131-4 => C5H81-3 + C2H5', [2.289000e+14, -0.68, 22050.0])

# Reaction 4266
reaction('C5H81-3 + C2H5 => C7H131-4', [8.500000e+10, 0.0, 8300.0])

# Reaction 4267
reaction('C7H131-5 => C4H8-1 + C3H5-A', [6.177000e+15, -1.18, 17980.0])

# Reaction 4268
reaction('C4H8-1 + C3H5-A => C7H131-5', [5.000000e+10, 0.0, 8300.0])

# Reaction 4269
reaction('C7H131-6 => C3H6 + C4H71-4', [1.238000e+17, -1.17, 30740.0])

# Reaction 4270
reaction('C3H6 + C4H71-4 => C7H131-6', [1.000000e+11, 0.0, 8300.0])

# Reaction 4271
reaction('C7H131-7 => C2H4 + C5H91-5', [9.416000e+17, -1.56, 31180.0])

# Reaction 4272
reaction('C2H4 + C5H91-5 => C7H131-7', [1.000000e+11, 0.0, 8300.0])

# Reaction 4273
reaction('C7H132-4 + HO2 => C7H13O2-4 + OH', [9.640000e+12, 0.0, 0.0])

# Reaction 4274
reaction('C7H13O2-4 + OH => C7H132-4 + HO2', [4.273000e+15, -1.02, 15430.0])

# Reaction 4275
reaction('C7H132-4 + CH3O2 => C7H13O2-4 + CH3O', [9.640000e+12, 0.0, 0.0])

# Reaction 4276
reaction('C7H13O2-4 + CH3O => C7H132-4 + CH3O2', [4.173000e+17, -1.59, 20190.0])

# Reaction 4277
reaction('C7H132-4 + C2H5O2 => C7H13O2-4 + C2H5O', [9.640000e+12, 0.0, 0.0])

# Reaction 4278
reaction('C7H13O2-4 + C2H5O => C7H132-4 + C2H5O2', [2.732000e+14, -0.67, 18030.0])

# Reaction 4279
reaction('C7H132-5 => C4H8-1 + C3H5-S', [9.419000e+17, -1.49, 44260.0])

# Reaction 4280
reaction('C4H8-1 + C3H5-S => C7H132-5', [1.000000e+11, 0.0, 8300.0])

# Reaction 4281
reaction('C7H132-6 => C3H6 + C4H71-3', [2.189000e+14, -0.74, 17740.0])

# Reaction 4282
reaction('C3H6 + C4H71-3 => C7H132-6', [1.000000e+11, 0.0, 8300.0])

# Reaction 4283
reaction('C7H132-7 => C2H4 + C5H92-5', [9.786000e+17, -1.57, 31160.0])

# Reaction 4284
reaction('C2H4 + C5H92-5 => C7H132-7', [1.000000e+11, 0.0, 8300.0])

# Reaction 4285
reaction('C7H133-1 => C2H4 + C5H91-1', [1.007000e+17, -1.32, 43990.0])

# Reaction 4286
reaction('C2H4 + C5H91-1 => C7H133-1', [1.000000e+11, 0.0, 8300.0])

# Reaction 4287
reaction('C7H133-5 + HO2 => C7H13O3-5 + OH', [9.640000e+12, 0.0, 0.0])

# Reaction 4288
reaction('C7H13O3-5 + OH => C7H133-5 + HO2', [1.182000e+15, -0.84, 15270.0])

# Reaction 4289
reaction('C7H133-5 + CH3O2 => C7H13O3-5 + CH3O', [9.640000e+12, 0.0, 0.0])

# Reaction 4290
reaction('C7H13O3-5 + CH3O => C7H133-5 + CH3O2', [1.155000e+17, -1.42, 20030.0])

# Reaction 4291
reaction('C7H133-5 + C2H5O2 => C7H13O3-5 + C2H5O', [9.640000e+12, 0.0, 0.0])

# Reaction 4292
reaction('C7H13O3-5 + C2H5O => C7H133-5 + C2H5O2', [7.559000e+13, -0.5, 17870.0])

# Reaction 4293
reaction('C7H133-6 => C3H6 + C4H71-1', [1.354000e+16, -0.92, 43560.0])

# Reaction 4294
reaction('C3H6 + C4H71-1 => C7H133-6', [1.000000e+11, 0.0, 8300.0])

# Reaction 4295
reaction('C7H133-7 => C2H4 + C5H91-3', [2.274000e+15, -1.2, 18050.0])

# Reaction 4296
reaction('C2H4 + C5H91-3 => C7H133-7', [1.000000e+11, 0.0, 8300.0])

# Reaction 4297
reaction('C7H13O1-3 => C2H3CHO + PC4H9', [9.625000e+19, -1.96, 10850.0])

# Reaction 4298
reaction('C2H3CHO + PC4H9 => C7H13O1-3', [1.000000e+11, 0.0, 9600.0])

# Reaction 4299
reaction('C7H13O1-3 => NC4H9CHO + C2H3', [1.026000e+18, -1.51, 23300.0])

# Reaction 4300
reaction('NC4H9CHO + C2H3 => C7H13O1-3', [1.000000e+11, 0.0, 9600.0])

# Reaction 4301
reaction('C7H13O2-4 => SC3H5CHO + NC3H7', [6.319000e+19, -1.93, 11130.0])

# Reaction 4302
reaction('SC3H5CHO + NC3H7 => C7H13O2-4', [1.000000e+11, 0.0, 9600.0])

# Reaction 4303
reaction('C7H13O2-4 => NC3H7CHO + C3H5-S', [1.012000e+22, -2.46, 29190.0])

# Reaction 4304
reaction('NC3H7CHO + C3H5-S => C7H13O2-4', [1.000000e+11, 0.0, 9600.0])

# Reaction 4305
reaction('C7H13O3-5 => C4H7CHO1-1 + C2H5', [6.033000e+18, -1.62, 10450.0])

# Reaction 4306
reaction('C4H7CHO1-1 + C2H5 => C7H13O3-5', [1.000000e+11, 0.0, 9600.0])

# Reaction 4307
reaction('C7H13O3-5 => C2H5CHO + C4H71-1', [6.425000e+21, -2.43, 30090.0])

# Reaction 4308
reaction('C2H5CHO + C4H71-1 => C7H13O3-5', [1.000000e+11, 0.0, 9600.0])

# Reaction 4309
reaction('C7H14-1 + OH => CH2O + C6H13-1', [1.000000e+11, 0.0, -4000.0])

# Reaction 4310
reaction('CH2O + C6H13-1 => C7H14-1 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 4311
reaction('C7H14-1 + OH => CH3CHO + C5H11-1', [1.000000e+11, 0.0, -4000.0])

# Reaction 4312
reaction('CH3CHO + C5H11-1 => C7H14-1 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 4313
reaction('C7H14-2 + OH => CH3CHO + C5H11-1', [1.000000e+11, 0.0, -4000.0])

# Reaction 4314
reaction('CH3CHO + C5H11-1 => C7H14-2 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 4315
reaction('C7H14-2 + OH => C2H5CHO + PC4H9', [1.000000e+11, 0.0, -4000.0])

# Reaction 4316
reaction('C2H5CHO + PC4H9 => C7H14-2 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 4317
reaction('C7H14-3 + OH => C2H5CHO + PC4H9', [1.000000e+11, 0.0, -4000.0])

# Reaction 4318
reaction('C2H5CHO + PC4H9 => C7H14-3 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 4319
reaction('C7H14-1 + O => CH2CHO + C5H11-1', [1.000000e+11, 0.0, -1050.0])

# Reaction 4320
reaction('CH2CHO + C5H11-1 => C7H14-1 + O', [0.000000e+00, 0.0, 0.0])

# Reaction 4321
reaction('C7H14-2 + O => CH3CHO + C5H10-1', [1.000000e+11, 0.0, -1050.0])

# Reaction 4322
reaction('CH3CHO + C5H10-1 => C7H14-2 + O', [0.000000e+00, 0.0, 0.0])

# Reaction 4323
reaction('C7H14-3 + O => CH3CHO + C5H10-1', [1.000000e+11, 0.0, -1050.0])

# Reaction 4324
reaction('CH3CHO + C5H10-1 => C7H14-3 + O', [0.000000e+00, 0.0, 0.0])

# Reaction 4325
reaction('C7H14-1 => PC4H9 + C3H5-A', [3.170000e+21, -1.62, 75330.0])

# Reaction 4326
reaction('PC4H9 + C3H5-A => C7H14-1', [1.000000e+13, 0.0, 0.0])

# Reaction 4327
reaction('C7H14-2 => C4H71-3 + NC3H7', [3.735000e+21, -1.74, 75710.0])

# Reaction 4328
reaction('C4H71-3 + NC3H7 => C7H14-2', [1.000000e+13, 0.0, 0.0])

# Reaction 4329
reaction('C7H14-3 => C5H91-3 + C2H5', [5.947000e+21, -1.85, 75790.0])

# Reaction 4330
reaction('C5H91-3 + C2H5 => C7H14-3', [1.000000e+13, 0.0, 0.0])

# Reaction 4331
reaction('C7H15O2-1 => C7H15-1 + O2', [2.657000e+20, -1.67, 35400.0])

# Reaction 4332
reaction('C7H15-1 + O2 => C7H15O2-1', [4.520000e+12, 0.0, 0.0])

# Reaction 4333
reaction('C7H15O2-2 => C7H15-2 + O2', [1.357000e+23, -2.36, 37670.0])

# Reaction 4334
reaction('C7H15-2 + O2 => C7H15O2-2', [7.540000e+12, 0.0, 0.0])

# Reaction 4335
reaction('C7H15O2-3 => C7H15-3 + O2', [1.357000e+23, -2.36, 37670.0])

# Reaction 4336
reaction('C7H15-3 + O2 => C7H15O2-3', [7.540000e+12, 0.0, 0.0])

# Reaction 4337
reaction('C7H15O2-4 => C7H15-4 + O2', [1.357000e+23, -2.36, 37670.0])

# Reaction 4338
reaction('C7H15-4 + O2 => C7H15O2-4', [7.540000e+12, 0.0, 0.0])

# Reaction 4339
reaction('C7H15-1 + C7H15O2-1 => C7H15O-1 + C7H15O-1', [7.000000e+12, 0.0, -1000.0])

# Reaction 4340
reaction('C7H15O-1 + C7H15O-1 => C7H15-1 + C7H15O2-1', [1.116000e+14, -0.23, 30250.0])

# Reaction 4341
reaction('C7H15-1 + C7H15O2-2 => C7H15O-1 + C7H15O-2', [7.000000e+12, 0.0, -1000.0])

# Reaction 4342
reaction('C7H15O-1 + C7H15O-2 => C7H15-1 + C7H15O2-2', [6.348000e+13, -0.16, 29880.0])

# Reaction 4343
reaction('C7H15-1 + C7H15O2-3 => C7H15O-1 + C7H15O-3', [7.000000e+12, 0.0, -1000.0])

# Reaction 4344
reaction('C7H15O-1 + C7H15O-3 => C7H15-1 + C7H15O2-3', [6.348000e+13, -0.16, 29880.0])

# Reaction 4345
reaction('C7H15-1 + C7H15O2-4 => C7H15O-1 + C7H15O-4', [7.000000e+12, 0.0, -1000.0])

# Reaction 4346
reaction('C7H15O-1 + C7H15O-4 => C7H15-1 + C7H15O2-4', [6.348000e+13, -0.16, 29880.0])

# Reaction 4347
reaction('C7H15-2 + C7H15O2-1 => C7H15O-2 + C7H15O-1', [7.000000e+12, 0.0, -1000.0])

# Reaction 4348
reaction('C7H15O-2 + C7H15O-1 => C7H15-2 + C7H15O2-1', [9.689000e+15, -0.8, 32160.0])

# Reaction 4349
reaction('C7H15-2 + C7H15O2-2 => C7H15O-2 + C7H15O-2', [7.000000e+12, 0.0, -1000.0])

# Reaction 4350
reaction('C7H15O-2 + C7H15O-2 => C7H15-2 + C7H15O2-2', [5.510000e+15, -0.73, 31790.0])

# Reaction 4351
reaction('C7H15-2 + C7H15O2-3 => C7H15O-2 + C7H15O-3', [7.000000e+12, 0.0, -1000.0])

# Reaction 4352
reaction('C7H15O-2 + C7H15O-3 => C7H15-2 + C7H15O2-3', [5.510000e+15, -0.73, 31790.0])

# Reaction 4353
reaction('C7H15-2 + C7H15O2-4 => C7H15O-2 + C7H15O-4', [7.000000e+12, 0.0, -1000.0])

# Reaction 4354
reaction('C7H15O-2 + C7H15O-4 => C7H15-2 + C7H15O2-4', [5.510000e+15, -0.73, 31790.0])

# Reaction 4355
reaction('C7H15-3 + C7H15O2-1 => C7H15O-3 + C7H15O-1', [7.000000e+12, 0.0, -1000.0])

# Reaction 4356
reaction('C7H15O-3 + C7H15O-1 => C7H15-3 + C7H15O2-1', [9.689000e+15, -0.8, 32160.0])

# Reaction 4357
reaction('C7H15-3 + C7H15O2-2 => C7H15O-3 + C7H15O-2', [7.000000e+12, 0.0, -1000.0])

# Reaction 4358
reaction('C7H15O-3 + C7H15O-2 => C7H15-3 + C7H15O2-2', [5.510000e+15, -0.73, 31790.0])

# Reaction 4359
reaction('C7H15-3 + C7H15O2-3 => C7H15O-3 + C7H15O-3', [7.000000e+12, 0.0, -1000.0])

# Reaction 4360
reaction('C7H15O-3 + C7H15O-3 => C7H15-3 + C7H15O2-3', [5.510000e+15, -0.73, 31790.0])

# Reaction 4361
reaction('C7H15-3 + C7H15O2-4 => C7H15O-3 + C7H15O-4', [7.000000e+12, 0.0, -1000.0])

# Reaction 4362
reaction('C7H15O-3 + C7H15O-4 => C7H15-3 + C7H15O2-4', [5.510000e+15, -0.73, 31790.0])

# Reaction 4363
reaction('C7H15-4 + C7H15O2-1 => C7H15O-4 + C7H15O-1', [7.000000e+12, 0.0, -1000.0])

# Reaction 4364
reaction('C7H15O-4 + C7H15O-1 => C7H15-4 + C7H15O2-1', [9.689000e+15, -0.8, 32160.0])

# Reaction 4365
reaction('C7H15-4 + C7H15O2-2 => C7H15O-4 + C7H15O-2', [7.000000e+12, 0.0, -1000.0])

# Reaction 4366
reaction('C7H15O-4 + C7H15O-2 => C7H15-4 + C7H15O2-2', [5.510000e+15, -0.73, 31790.0])

# Reaction 4367
reaction('C7H15-4 + C7H15O2-3 => C7H15O-4 + C7H15O-3', [7.000000e+12, 0.0, -1000.0])

# Reaction 4368
reaction('C7H15O-4 + C7H15O-3 => C7H15-4 + C7H15O2-3', [5.510000e+15, -0.73, 31790.0])

# Reaction 4369
reaction('C7H15-4 + C7H15O2-4 => C7H15O-4 + C7H15O-4', [7.000000e+12, 0.0, -1000.0])

# Reaction 4370
reaction('C7H15O-4 + C7H15O-4 => C7H15-4 + C7H15O2-4', [5.511000e+15, -0.73, 31790.0])

# Reaction 4371
reaction('C7H15-1 + HO2 => C7H15O-1 + OH', [7.000000e+12, 0.0, -1000.0])

# Reaction 4372
reaction('C7H15O-1 + OH => C7H15-1 + HO2', [3.549000e+15, -0.57, 26160.0])

# Reaction 4373
reaction('C7H15-2 + HO2 => C7H15O-2 + OH', [7.000000e+12, 0.0, -1000.0])

# Reaction 4374
reaction('C7H15O-2 + OH => C7H15-2 + HO2', [3.081000e+17, -1.15, 28070.0])

# Reaction 4375
reaction('C7H15-3 + HO2 => C7H15O-3 + OH', [7.000000e+12, 0.0, -1000.0])

# Reaction 4376
reaction('C7H15O-3 + OH => C7H15-3 + HO2', [3.081000e+17, -1.15, 28070.0])

# Reaction 4377
reaction('C7H15-4 + HO2 => C7H15O-4 + OH', [7.000000e+12, 0.0, -1000.0])

# Reaction 4378
reaction('C7H15O-4 + OH => C7H15-4 + HO2', [3.081000e+17, -1.15, 28070.0])

# Reaction 4379
reaction('C7H15-1 + CH3O2 => C7H15O-1 + CH3O', [7.000000e+12, 0.0, -1000.0])

# Reaction 4380
reaction('C7H15O-1 + CH3O => C7H15-1 + CH3O2', [2.220000e+14, -0.28, 30400.0])

# Reaction 4381
reaction('C7H15-2 + CH3O2 => C7H15O-2 + CH3O', [7.000000e+12, 0.0, -1000.0])

# Reaction 4382
reaction('C7H15O-2 + CH3O => C7H15-2 + CH3O2', [1.927000e+16, -0.85, 32310.0])

# Reaction 4383
reaction('C7H15-3 + CH3O2 => C7H15O-3 + CH3O', [7.000000e+12, 0.0, -1000.0])

# Reaction 4384
reaction('C7H15O-3 + CH3O => C7H15-3 + CH3O2', [1.927000e+16, -0.85, 32310.0])

# Reaction 4385
reaction('C7H15-4 + CH3O2 => C7H15O-4 + CH3O', [7.000000e+12, 0.0, -1000.0])

# Reaction 4386
reaction('C7H15O-4 + CH3O => C7H15-4 + CH3O2', [1.927000e+16, -0.85, 32310.0])

# Reaction 4387
reaction('C7H15O2-1 <=> C7H14-1 + HO2', [1.004400e+39, -8.11, 41490.0])

# Reaction 4388
reaction('C7H15O2-2 <=> C7H14-1 + HO2', [1.007500e+43, -9.41, 42490.0])

# Reaction 4389
reaction('C7H15O2-2 <=> C7H14-2 + HO2', [1.004400e+39, -8.11, 41490.0])

# Reaction 4390
reaction('C7H15O2-3 <=> C7H14-2 + HO2', [1.004400e+39, -8.11, 41490.0])

# Reaction 4391
reaction('C7H15O2-3 <=> C7H14-3 + HO2', [1.004400e+39, -8.11, 41490.0])

# Reaction 4392
reaction('C7H15O2-4 <=> C7H14-3 + HO2', [2.009000e+39, -8.11, 41490.0])

# Reaction 4393
reaction('C7H15O2-1 <=> C7H14OOH1-2', [2.000000e+11, 0.0, 26450.0])

# Reaction 4394
reaction('C7H15O2-1 <=> C7H14OOH1-3', [2.500000e+10, 0.0, 20450.0])

# Reaction 4395
reaction('C7H15O2-1 <=> C7H14OOH1-4', [3.125000e+09, 0.0, 18650.0])

# Reaction 4396
reaction('C7H15O2-1 <=> C7H14OOH1-5', [3.906000e+08, 0.0, 21650.0])

# Reaction 4397
reaction('C7H15O2-2 <=> C7H14OOH2-1', [3.000000e+11, 0.0, 29000.0])

# Reaction 4398
reaction('C7H15O2-2 <=> C7H14OOH2-3', [2.000000e+11, 0.0, 26450.0])

# Reaction 4399
reaction('C7H15O2-2 <=> C7H14OOH2-4', [2.500000e+10, 0.0, 20450.0])

# Reaction 4400
reaction('C7H15O2-2 <=> C7H14OOH2-5', [3.125000e+09, 0.0, 18650.0])

# Reaction 4401
reaction('C7H15O2-2 <=> C7H14OOH2-6', [3.906000e+08, 0.0, 21650.0])

# Reaction 4402
reaction('C7H15O2-3 <=> C7H14OOH3-1', [3.750000e+10, 0.0, 24000.0])

# Reaction 4403
reaction('C7H15O2-3 <=> C7H14OOH3-2', [2.000000e+11, 0.0, 26450.0])

# Reaction 4404
reaction('C7H15O2-3 <=> C7H14OOH3-4', [2.000000e+11, 0.0, 26450.0])

# Reaction 4405
reaction('C7H15O2-3 <=> C7H14OOH3-5', [2.500000e+10, 0.0, 20450.0])

# Reaction 4406
reaction('C7H15O2-3 <=> C7H14OOH3-6', [3.125000e+09, 0.0, 18650.0])

# Reaction 4407
reaction('C7H15O2-3 <=> C7H14OOH3-7', [5.860000e+08, 0.0, 25150.0])

# Reaction 4408
reaction('C7H15O2-4 <=> C7H14OOH4-1', [9.376000e+09, 0.0, 21950.0])

# Reaction 4409
reaction('C7H15O2-4 <=> C7H14OOH4-2', [5.000000e+10, 0.0, 20450.0])

# Reaction 4410
reaction('C7H15O2-4 <=> C7H14OOH4-3', [4.000000e+11, 0.0, 26450.0])

# Reaction 4411
reaction('C7H15O2-1 + HO2 => C7H15O2H-1 + O2', [1.750000e+10, 0.0, -3275.0])

# Reaction 4412
reaction('C7H15O2H-1 + O2 => C7H15O2-1 + HO2', [3.782000e+13, -0.79, 34010.0])

# Reaction 4413
reaction('C7H15O2-2 + HO2 => C7H15O2H-2 + O2', [1.750000e+10, 0.0, -3275.0])

# Reaction 4414
reaction('C7H15O2H-2 + O2 => C7H15O2-2 + HO2', [4.496000e+13, -0.82, 34040.0])

# Reaction 4415
reaction('C7H15O2-3 + HO2 => C7H15O2H-3 + O2', [1.750000e+10, 0.0, -3275.0])

# Reaction 4416
reaction('C7H15O2H-3 + O2 => C7H15O2-3 + HO2', [4.496000e+13, -0.82, 34040.0])

# Reaction 4417
reaction('C7H15O2-4 + HO2 => C7H15O2H-4 + O2', [1.750000e+10, 0.0, -3275.0])

# Reaction 4418
reaction('C7H15O2H-4 + O2 => C7H15O2-4 + HO2', [4.497000e+13, -0.82, 34040.0])

# Reaction 4419
reaction('H2O2 + C7H15O2-1 => HO2 + C7H15O2H-1', [2.400000e+12, 0.0, 10000.0])

# Reaction 4420
reaction('HO2 + C7H15O2H-1 => H2O2 + C7H15O2-1', [2.400000e+12, 0.0, 10000.0])

# Reaction 4421
reaction('H2O2 + C7H15O2-2 => HO2 + C7H15O2H-2', [2.400000e+12, 0.0, 10000.0])

# Reaction 4422
reaction('HO2 + C7H15O2H-2 => H2O2 + C7H15O2-2', [2.400000e+12, 0.0, 10000.0])

# Reaction 4423
reaction('H2O2 + C7H15O2-3 => HO2 + C7H15O2H-3', [2.400000e+12, 0.0, 10000.0])

# Reaction 4424
reaction('HO2 + C7H15O2H-3 => H2O2 + C7H15O2-3', [2.400000e+12, 0.0, 10000.0])

# Reaction 4425
reaction('H2O2 + C7H15O2-4 => HO2 + C7H15O2H-4', [2.400000e+12, 0.0, 10000.0])

# Reaction 4426
reaction('HO2 + C7H15O2H-4 => H2O2 + C7H15O2-4', [2.400000e+12, 0.0, 10000.0])

# Reaction 4427
reaction('C7H15O2-1 + CH3O2 => C7H15O-1 + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 4428
reaction('C7H15O-1 + CH3O + O2 => C7H15O2-1 + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 4429
reaction('C7H15O2-2 + CH3O2 => C7H15O-2 + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 4430
reaction('C7H15O-2 + CH3O + O2 => C7H15O2-2 + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 4431
reaction('C7H15O2-3 + CH3O2 => C7H15O-3 + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 4432
reaction('C7H15O-3 + CH3O + O2 => C7H15O2-3 + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 4433
reaction('C7H15O2-4 + CH3O2 => C7H15O-4 + CH3O + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 4434
reaction('C7H15O-4 + CH3O + O2 => C7H15O2-4 + CH3O2', [0.000000e+00, 0.0, 0.0])

# Reaction 4435
reaction('C7H15O2-1 + C7H15O2-1 => O2 + C7H15O-1 + C7H15O-1', [1.400000e+16, -1.61, 1860.0])

# Reaction 4436
reaction('O2 + C7H15O-1 + C7H15O-1 => C7H15O2-1 + C7H15O2-1', [0.000000e+00, 0.0, 0.0])

# Reaction 4437
reaction('C7H15O2-1 + C7H15O2-2 => C7H15O-1 + C7H15O-2 + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 4438
reaction('C7H15O-1 + C7H15O-2 + O2 => C7H15O2-1 + C7H15O2-2', [0.000000e+00, 0.0, 0.0])

# Reaction 4439
reaction('C7H15O2-1 + C7H15O2-3 => C7H15O-1 + C7H15O-3 + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 4440
reaction('C7H15O-1 + C7H15O-3 + O2 => C7H15O2-1 + C7H15O2-3', [0.000000e+00, 0.0, 0.0])

# Reaction 4441
reaction('C7H15O2-1 + C7H15O2-4 => C7H15O-1 + C7H15O-4 + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 4442
reaction('C7H15O-1 + C7H15O-4 + O2 => C7H15O2-1 + C7H15O2-4', [0.000000e+00, 0.0, 0.0])

# Reaction 4443
reaction('C7H15O2-2 + C7H15O2-2 => O2 + C7H15O-2 + C7H15O-2', [1.400000e+16, -1.61, 1860.0])

# Reaction 4444
reaction('O2 + C7H15O-2 + C7H15O-2 => C7H15O2-2 + C7H15O2-2', [0.000000e+00, 0.0, 0.0])

# Reaction 4445
reaction('C7H15O2-2 + C7H15O2-3 => C7H15O-2 + C7H15O-3 + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 4446
reaction('C7H15O-2 + C7H15O-3 + O2 => C7H15O2-2 + C7H15O2-3', [0.000000e+00, 0.0, 0.0])

# Reaction 4447
reaction('C7H15O2-2 + C7H15O2-4 => C7H15O-2 + C7H15O-4 + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 4448
reaction('C7H15O-2 + C7H15O-4 + O2 => C7H15O2-2 + C7H15O2-4', [0.000000e+00, 0.0, 0.0])

# Reaction 4449
reaction('C7H15O2-3 + C7H15O2-3 => O2 + C7H15O-3 + C7H15O-3', [1.400000e+16, -1.61, 1860.0])

# Reaction 4450
reaction('O2 + C7H15O-3 + C7H15O-3 => C7H15O2-3 + C7H15O2-3', [0.000000e+00, 0.0, 0.0])

# Reaction 4451
reaction('C7H15O2-3 + C7H15O2-4 => C7H15O-3 + C7H15O-4 + O2', [1.400000e+16, -1.61, 1860.0])

# Reaction 4452
reaction('C7H15O-3 + C7H15O-4 + O2 => C7H15O2-3 + C7H15O2-4', [0.000000e+00, 0.0, 0.0])

# Reaction 4453
reaction('C7H15O2-4 + C7H15O2-4 => O2 + C7H15O-4 + C7H15O-4', [1.400000e+16, -1.61, 1860.0])

# Reaction 4454
reaction('O2 + C7H15O-4 + C7H15O-4 => C7H15O2-4 + C7H15O2-4', [0.000000e+00, 0.0, 0.0])

# Reaction 4455
reaction('C7H15O2H-1 => C7H15O-1 + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 4456
reaction('C7H15O-1 + OH => C7H15O2H-1', [1.989000e+07, 1.93, -6522.0])

# Reaction 4457
reaction('C7H15O2H-2 => C7H15O-2 + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 4458
reaction('C7H15O-2 + OH => C7H15O2H-2', [9.517000e+06, 2.03, -6922.0])

# Reaction 4459
reaction('C7H15O2H-3 => C7H15O-3 + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 4460
reaction('C7H15O-3 + OH => C7H15O2H-3', [9.517000e+06, 2.03, -6922.0])

# Reaction 4461
reaction('C7H15O2H-4 => C7H15O-4 + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 4462
reaction('C7H15O-4 + OH => C7H15O2H-4', [9.516000e+06, 2.03, -6922.0])

# Reaction 4463
reaction('C7H15O-1 => CH2O + C6H13-1', [6.023000e+20, -2.18, 24830.0])

# Reaction 4464
reaction('CH2O + C6H13-1 => C7H15O-1', [1.000000e+11, 0.0, 11900.0])

# Reaction 4465
reaction('C7H15O-2 => CH3CHO + C5H11-1', [1.475000e+22, -2.55, 20370.0])

# Reaction 4466
reaction('CH3CHO + C5H11-1 => C7H15O-2', [1.000000e+11, 0.0, 12900.0])

# Reaction 4467
reaction('C7H15O-3 => C2H5CHO + PC4H9', [6.573000e+17, -1.16, 19370.0])

# Reaction 4468
reaction('C2H5CHO + PC4H9 => C7H15O-3', [1.000000e+11, 0.0, 12900.0])

# Reaction 4469
reaction('C7H15O-4 => NC3H7CHO + NC3H7', [4.769000e+22, -2.58, 21820.0])

# Reaction 4470
reaction('NC3H7CHO + NC3H7 => C7H15O-4', [1.000000e+11, 0.0, 12900.0])

# Reaction 4471
reaction('C7H14-1 + HO2 <=> C7H14OOH1-2', [2.500000e+03, 2.5, 11200.0])

# Reaction 4472
reaction('C7H14-1 + HO2 <=> C7H14OOH2-1', [2.700000e+03, 2.5, 10500.0])

# Reaction 4473
reaction('C7H14-2 + HO2 <=> C7H14OOH2-3', [2.700000e+03, 2.5, 10500.0])

# Reaction 4474
reaction('C7H14-2 + HO2 <=> C7H14OOH3-2', [2.700000e+03, 2.5, 10500.0])

# Reaction 4475
reaction('C7H14-3 + HO2 <=> C7H14OOH3-4', [2.700000e+03, 2.5, 10500.0])

# Reaction 4476
reaction('C7H14-3 + HO2 <=> C7H14OOH4-3', [2.700000e+03, 2.5, 10500.0])

# Reaction 4477
reaction('C7H14OOH1-2 => C7H14O1-2 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 4478
reaction('C7H14O1-2 + OH => C7H14OOH1-2', [0.000000e+00, 0.0, 0.0])

# Reaction 4479
reaction('C7H14OOH1-3 => C7H14O1-3 + OH', [7.500000e+10, 0.0, 15250.0])

# Reaction 4480
reaction('C7H14O1-3 + OH => C7H14OOH1-3', [0.000000e+00, 0.0, 0.0])

# Reaction 4481
reaction('C7H14OOH1-4 => C7H14O1-4 + OH', [9.375000e+09, 0.0, 7000.0])

# Reaction 4482
reaction('C7H14O1-4 + OH => C7H14OOH1-4', [0.000000e+00, 0.0, 0.0])

# Reaction 4483
reaction('C7H14OOH1-5 => C7H14O1-5 + OH', [1.172000e+09, 0.0, 1800.0])

# Reaction 4484
reaction('C7H14O1-5 + OH => C7H14OOH1-5', [0.000000e+00, 0.0, 0.0])

# Reaction 4485
reaction('C7H14OOH2-1 => C7H14O1-2 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 4486
reaction('C7H14O1-2 + OH => C7H14OOH2-1', [0.000000e+00, 0.0, 0.0])

# Reaction 4487
reaction('C7H14OOH2-3 => C7H14O2-3 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 4488
reaction('C7H14O2-3 + OH => C7H14OOH2-3', [0.000000e+00, 0.0, 0.0])

# Reaction 4489
reaction('C7H14OOH2-4 => C7H14O2-4 + OH', [7.500000e+10, 0.0, 15250.0])

# Reaction 4490
reaction('C7H14O2-4 + OH => C7H14OOH2-4', [0.000000e+00, 0.0, 0.0])

# Reaction 4491
reaction('C7H14OOH2-5 => C7H14O2-5 + OH', [9.375000e+09, 0.0, 7000.0])

# Reaction 4492
reaction('C7H14O2-5 + OH => C7H14OOH2-5', [0.000000e+00, 0.0, 0.0])

# Reaction 4493
reaction('C7H14OOH2-6 => C7H14O2-6 + OH', [1.172000e+09, 0.0, 1800.0])

# Reaction 4494
reaction('C7H14O2-6 + OH => C7H14OOH2-6', [0.000000e+00, 0.0, 0.0])

# Reaction 4495
reaction('C7H14OOH3-1 => C7H14O1-3 + OH', [7.500000e+10, 0.0, 15250.0])

# Reaction 4496
reaction('C7H14O1-3 + OH => C7H14OOH3-1', [0.000000e+00, 0.0, 0.0])

# Reaction 4497
reaction('C7H14OOH3-2 => C7H14O2-3 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 4498
reaction('C7H14O2-3 + OH => C7H14OOH3-2', [0.000000e+00, 0.0, 0.0])

# Reaction 4499
reaction('C7H14OOH3-4 => C7H14O3-4 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 4500
reaction('C7H14O3-4 + OH => C7H14OOH3-4', [0.000000e+00, 0.0, 0.0])

# Reaction 4501
reaction('C7H14OOH3-5 => C7H14O3-5 + OH', [7.500000e+10, 0.0, 15250.0])

# Reaction 4502
reaction('C7H14O3-5 + OH => C7H14OOH3-5', [0.000000e+00, 0.0, 0.0])

# Reaction 4503
reaction('C7H14OOH3-6 => C7H14O2-5 + OH', [9.375000e+09, 0.0, 7000.0])

# Reaction 4504
reaction('C7H14O2-5 + OH => C7H14OOH3-6', [0.000000e+00, 0.0, 0.0])

# Reaction 4505
reaction('C7H14OOH3-7 => C7H14O1-5 + OH', [1.172000e+09, 0.0, 1800.0])

# Reaction 4506
reaction('C7H14O1-5 + OH => C7H14OOH3-7', [0.000000e+00, 0.0, 0.0])

# Reaction 4507
reaction('C7H14OOH4-1 => C7H14O1-4 + OH', [9.375000e+09, 0.0, 7000.0])

# Reaction 4508
reaction('C7H14O1-4 + OH => C7H14OOH4-1', [0.000000e+00, 0.0, 0.0])

# Reaction 4509
reaction('C7H14OOH4-2 => C7H14O2-4 + OH', [7.500000e+10, 0.0, 15250.0])

# Reaction 4510
reaction('C7H14O2-4 + OH => C7H14OOH4-2', [0.000000e+00, 0.0, 0.0])

# Reaction 4511
reaction('C7H14OOH4-3 => C7H14O3-4 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 4512
reaction('C7H14O3-4 + OH => C7H14OOH4-3', [0.000000e+00, 0.0, 0.0])

# Reaction 4513
reaction('C7H14OOH1-3 => OH + CH2O + C6H12-1', [2.147000e+09, 1.23, 30370.0])

# Reaction 4514
reaction('OH + CH2O + C6H12-1 => C7H14OOH1-3', [0.000000e+00, 0.0, 0.0])

# Reaction 4515
reaction('C7H14OOH2-4 => OH + CH3CHO + C5H10-1', [1.548000e+12, 0.59, 30090.0])

# Reaction 4516
reaction('OH + CH3CHO + C5H10-1 => C7H14OOH2-4', [0.000000e+00, 0.0, 0.0])

# Reaction 4517
reaction('C7H14OOH3-1 => OH + NC4H9CHO + C2H4', [8.182000e+13, -0.13, 31330.0])

# Reaction 4518
reaction('OH + NC4H9CHO + C2H4 => C7H14OOH3-1', [0.000000e+00, 0.0, 0.0])

# Reaction 4519
reaction('C7H14OOH3-5 => OH + C2H5CHO + C4H8-1', [2.661000e+13, 0.13, 30430.0])

# Reaction 4520
reaction('OH + C2H5CHO + C4H8-1 => C7H14OOH3-5', [0.000000e+00, 0.0, 0.0])

# Reaction 4521
reaction('C7H14OOH4-2 => OH + NC3H7CHO + C3H6', [6.186000e+13, 0.09, 30840.0])

# Reaction 4522
reaction('OH + NC3H7CHO + C3H6 => C7H14OOH4-2', [0.000000e+00, 0.0, 0.0])

# Reaction 4523
reaction('C7H14OOH1-3 => C4H7OOH1-4 + NC3H7', [1.609000e+12, 0.54, 27740.0])

# Reaction 4524
reaction('C4H7OOH1-4 + NC3H7 => C7H14OOH1-3', [2.200000e+03, 2.48, 6130.0])

# Reaction 4525
reaction('C7H14OOH1-4 => C5H10-1 + C2H4 + HO2', [1.446000e+11, 0.69, 30820.0])

# Reaction 4526
reaction('C5H10-1 + C2H4 + HO2 => C7H14OOH1-4', [0.000000e+00, 0.0, 0.0])

# Reaction 4527
reaction('C7H14OOH1-4 => C5H9OOH1-5 + C2H5', [1.721000e+12, 0.51, 27900.0])

# Reaction 4528
reaction('C5H9OOH1-5 + C2H5 => C7H14OOH1-4', [2.200000e+03, 2.48, 6130.0])

# Reaction 4529
reaction('C7H14OOH2-4 => C5H9OOH1-4 + C2H5', [4.080000e+12, 0.32, 29230.0])

# Reaction 4530
reaction('C5H9OOH1-4 + C2H5 => C7H14OOH2-4', [2.200000e+03, 2.48, 6130.0])

# Reaction 4531
reaction('C7H14OOH2-5 => C4H8-1 + C3H6OOH2-1', [6.856000e+13, 0.03, 31380.0])

# Reaction 4532
reaction('C4H8-1 + C3H6OOH2-1 => C7H14OOH2-5', [2.200000e+03, 2.48, 6130.0])

# Reaction 4533
reaction('C7H14OOH2-5 => C6H11OOH1-5 + CH3', [7.417000e+10, 0.75, 30260.0])

# Reaction 4534
reaction('C6H11OOH1-5 + CH3 => C7H14OOH2-5', [2.200000e+03, 2.48, 6130.0])

# Reaction 4535
reaction('C7H14OOH3-5 => C6H11OOH1-4 + CH3', [7.417000e+10, 0.75, 30260.0])

# Reaction 4536
reaction('C6H11OOH1-4 + CH3 => C7H14OOH3-5', [2.200000e+03, 2.48, 6130.0])

# Reaction 4537
reaction('C7H14OOH3-6 => C4H8OOH2-1 + C3H6', [4.578000e+11, 0.6, 29170.0])

# Reaction 4538
reaction('C4H8OOH2-1 + C3H6 => C7H14OOH3-6', [2.200000e+03, 2.48, 6130.0])

# Reaction 4539
reaction('C7H14OOH4-1 => C5H10OOH2-1 + C2H4', [1.643000e+12, 0.44, 29320.0])

# Reaction 4540
reaction('C5H10OOH2-1 + C2H4 => C7H14OOH4-1', [2.200000e+03, 2.48, 6130.0])

# Reaction 4541
reaction('C7H14OOH1-2O2 => C7H14OOH1-2 + O2', [1.367000e+23, -2.37, 37640.0])

# Reaction 4542
reaction('C7H14OOH1-2 + O2 => C7H14OOH1-2O2', [7.540000e+12, 0.0, 0.0])

# Reaction 4543
reaction('C7H14OOH1-3O2 => C7H14OOH1-3 + O2', [1.367000e+23, -2.37, 37640.0])

# Reaction 4544
reaction('C7H14OOH1-3 + O2 => C7H14OOH1-3O2', [7.540000e+12, 0.0, 0.0])

# Reaction 4545
reaction('C7H14OOH1-4O2 => C7H14OOH1-4 + O2', [1.367000e+23, -2.37, 37640.0])

# Reaction 4546
reaction('C7H14OOH1-4 + O2 => C7H14OOH1-4O2', [7.540000e+12, 0.0, 0.0])

# Reaction 4547
reaction('C7H14OOH1-5O2 => C7H14OOH1-5 + O2', [1.367000e+23, -2.37, 37640.0])

# Reaction 4548
reaction('C7H14OOH1-5 + O2 => C7H14OOH1-5O2', [7.540000e+12, 0.0, 0.0])

# Reaction 4549
reaction('C7H14OOH2-1O2 => C7H14OOH2-1 + O2', [3.323000e+20, -1.65, 35280.0])

# Reaction 4550
reaction('C7H14OOH2-1 + O2 => C7H14OOH2-1O2', [4.520000e+12, 0.0, 0.0])

# Reaction 4551
reaction('C7H14OOH2-3O2 => C7H14OOH2-3 + O2', [1.389000e+23, -2.38, 37600.0])

# Reaction 4552
reaction('C7H14OOH2-3 + O2 => C7H14OOH2-3O2', [7.540000e+12, 0.0, 0.0])

# Reaction 4553
reaction('C7H14OOH2-4O2 => C7H14OOH2-4 + O2', [1.389000e+23, -2.38, 37600.0])

# Reaction 4554
reaction('C7H14OOH2-4 + O2 => C7H14OOH2-4O2', [7.540000e+12, 0.0, 0.0])

# Reaction 4555
reaction('C7H14OOH2-5O2 => C7H14OOH2-5 + O2', [1.389000e+23, -2.38, 37600.0])

# Reaction 4556
reaction('C7H14OOH2-5 + O2 => C7H14OOH2-5O2', [7.540000e+12, 0.0, 0.0])

# Reaction 4557
reaction('C7H14OOH2-6O2 => C7H14OOH2-6 + O2', [1.389000e+23, -2.38, 37600.0])

# Reaction 4558
reaction('C7H14OOH2-6 + O2 => C7H14OOH2-6O2', [7.540000e+12, 0.0, 0.0])

# Reaction 4559
reaction('C7H14OOH3-1O2 => C7H14OOH3-1 + O2', [3.323000e+20, -1.65, 35280.0])

# Reaction 4560
reaction('C7H14OOH3-1 + O2 => C7H14OOH3-1O2', [4.520000e+12, 0.0, 0.0])

# Reaction 4561
reaction('C7H14OOH3-2O2 => C7H14OOH3-2 + O2', [1.389000e+23, -2.38, 37600.0])

# Reaction 4562
reaction('C7H14OOH3-2 + O2 => C7H14OOH3-2O2', [7.540000e+12, 0.0, 0.0])

# Reaction 4563
reaction('C7H14OOH3-4O2 => C7H14OOH3-4 + O2', [1.389000e+23, -2.38, 37600.0])

# Reaction 4564
reaction('C7H14OOH3-4 + O2 => C7H14OOH3-4O2', [7.540000e+12, 0.0, 0.0])

# Reaction 4565
reaction('C7H14OOH3-5O2 => C7H14OOH3-5 + O2', [1.389000e+23, -2.38, 37600.0])

# Reaction 4566
reaction('C7H14OOH3-5 + O2 => C7H14OOH3-5O2', [7.540000e+12, 0.0, 0.0])

# Reaction 4567
reaction('C7H14OOH3-6O2 => C7H14OOH3-6 + O2', [1.389000e+23, -2.38, 37600.0])

# Reaction 4568
reaction('C7H14OOH3-6 + O2 => C7H14OOH3-6O2', [7.540000e+12, 0.0, 0.0])

# Reaction 4569
reaction('C7H14OOH3-7O2 => C7H14OOH3-7 + O2', [3.323000e+20, -1.65, 35280.0])

# Reaction 4570
reaction('C7H14OOH3-7 + O2 => C7H14OOH3-7O2', [4.520000e+12, 0.0, 0.0])

# Reaction 4571
reaction('C7H14OOH4-1O2 => C7H14OOH4-1 + O2', [1.668000e+20, -1.65, 35280.0])

# Reaction 4572
reaction('C7H14OOH4-1 + O2 => C7H14OOH4-1O2', [4.520000e+12, 0.0, 0.0])

# Reaction 4573
reaction('C7H14OOH4-2O2 => C7H14OOH4-2 + O2', [6.969000e+22, -2.38, 37600.0])

# Reaction 4574
reaction('C7H14OOH4-2 + O2 => C7H14OOH4-2O2', [7.540000e+12, 0.0, 0.0])

# Reaction 4575
reaction('C7H14OOH4-3O2 => C7H14OOH4-3 + O2', [6.969000e+22, -2.38, 37600.0])

# Reaction 4576
reaction('C7H14OOH4-3 + O2 => C7H14OOH4-3O2', [7.540000e+12, 0.0, 0.0])

# Reaction 4577
reaction('C7H14OOH1-2O2 <=> NC7KET12 + OH', [2.000000e+11, 0.0, 26000.0])

# Reaction 4578
reaction('C7H14OOH1-3O2 <=> NC7KET13 + OH', [2.500000e+10, 0.0, 21000.0])

# Reaction 4579
reaction('C7H14OOH1-4O2 <=> NC7KET14 + OH', [3.125000e+09, 0.0, 18950.0])

# Reaction 4580
reaction('C7H14OOH1-5O2 <=> NC7KET15 + OH', [3.906000e+08, 0.0, 22150.0])

# Reaction 4581
reaction('C7H14OOH2-1O2 <=> NC7KET21 + OH', [1.000000e+11, 0.0, 23450.0])

# Reaction 4582
reaction('C7H14OOH2-3O2 <=> NC7KET23 + OH', [1.000000e+11, 0.0, 23450.0])

# Reaction 4583
reaction('C7H14OOH2-4O2 <=> NC7KET24 + OH', [1.250000e+10, 0.0, 17450.0])

# Reaction 4584
reaction('C7H14OOH2-5O2 <=> NC7KET25 + OH', [1.563000e+09, 0.0, 15650.0])

# Reaction 4585
reaction('C7H14OOH2-6O2 <=> NC7KET26 + OH', [1.953000e+08, 0.0, 18650.0])

# Reaction 4586
reaction('C7H14OOH3-1O2 <=> NC7KET31 + OH', [1.250000e+10, 0.0, 17450.0])

# Reaction 4587
reaction('C7H14OOH3-2O2 <=> NC7KET32 + OH', [1.000000e+11, 0.0, 23450.0])

# Reaction 4588
reaction('C7H14OOH3-4O2 <=> NC7KET34 + OH', [1.000000e+11, 0.0, 23450.0])

# Reaction 4589
reaction('C7H14OOH3-5O2 <=> NC7KET35 + OH', [1.250000e+10, 0.0, 17450.0])

# Reaction 4590
reaction('C7H14OOH3-6O2 <=> NC7KET36 + OH', [1.563000e+09, 0.0, 15650.0])

# Reaction 4591
reaction('C7H14OOH3-7O2 <=> NC7KET37 + OH', [1.953000e+08, 0.0, 18650.0])

# Reaction 4592
reaction('C7H14OOH4-1O2 <=> NC7KET41 + OH', [1.563000e+09, 0.0, 15650.0])

# Reaction 4593
reaction('C7H14OOH4-2O2 <=> NC7KET42 + OH', [1.250000e+10, 0.0, 17450.0])

# Reaction 4594
reaction('C7H14OOH4-3O2 <=> NC7KET43 + OH', [1.000000e+11, 0.0, 23450.0])

# Reaction 4595
reaction('NC7KET12 => NC5H11CHO + HCO + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 4596
reaction('NC5H11CHO + HCO + OH => NC7KET12', [0.000000e+00, 0.0, 0.0])

# Reaction 4597
reaction('NC7KET13 => NC4H9CHO + CH2CHO + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 4598
reaction('NC4H9CHO + CH2CHO + OH => NC7KET13', [0.000000e+00, 0.0, 0.0])

# Reaction 4599
reaction('NC7KET14 => NC3H7CHO + CH2CH2CHO + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 4600
reaction('NC3H7CHO + CH2CH2CHO + OH => NC7KET14', [0.000000e+00, 0.0, 0.0])

# Reaction 4601
reaction('NC7KET15 => C2H5CHO + C3H6CHO-1 + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 4602
reaction('C2H5CHO + C3H6CHO-1 + OH => NC7KET15', [0.000000e+00, 0.0, 0.0])

# Reaction 4603
reaction('NC7KET21 => CH2O + NC5H11CO + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 4604
reaction('CH2O + NC5H11CO + OH => NC7KET21', [0.000000e+00, 0.0, 0.0])

# Reaction 4605
reaction('NC7KET23 => NC4H9CHO + CH3CO + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 4606
reaction('NC4H9CHO + CH3CO + OH => NC7KET23', [0.000000e+00, 0.0, 0.0])

# Reaction 4607
reaction('NC7KET24 => NC3H7CHO + CH3COCH2 + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 4608
reaction('NC3H7CHO + CH3COCH2 + OH => NC7KET24', [0.000000e+00, 0.0, 0.0])

# Reaction 4609
reaction('NC7KET25 => C2H5CHO + CH2CH2COCH3 + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 4610
reaction('C2H5CHO + CH2CH2COCH3 + OH => NC7KET25', [0.000000e+00, 0.0, 0.0])

# Reaction 4611
reaction('NC7KET26 => CH3CHO + C3H6COCH3-1 + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 4612
reaction('CH3CHO + C3H6COCH3-1 + OH => NC7KET26', [0.000000e+00, 0.0, 0.0])

# Reaction 4613
reaction('NC7KET31 => CH2O + NC4H9COCH2 + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 4614
reaction('CH2O + NC4H9COCH2 + OH => NC7KET31', [0.000000e+00, 0.0, 0.0])

# Reaction 4615
reaction('NC7KET32 => CH3CHO + NC4H9CO + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 4616
reaction('CH3CHO + NC4H9CO + OH => NC7KET32', [0.000000e+00, 0.0, 0.0])

# Reaction 4617
reaction('NC7KET34 => NC3H7CHO + C2H5CO + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 4618
reaction('NC3H7CHO + C2H5CO + OH => NC7KET34', [0.000000e+00, 0.0, 0.0])

# Reaction 4619
reaction('NC7KET35 => C2H5CHO + C2H5COCH2 + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 4620
reaction('C2H5CHO + C2H5COCH2 + OH => NC7KET35', [0.000000e+00, 0.0, 0.0])

# Reaction 4621
reaction('NC7KET36 => CH3CHO + C2H5COC2H4P + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 4622
reaction('CH3CHO + C2H5COC2H4P + OH => NC7KET36', [0.000000e+00, 0.0, 0.0])

# Reaction 4623
reaction('NC7KET37 => CH2O + C3H6COC2H5-1 + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 4624
reaction('CH2O + C3H6COC2H5-1 + OH => NC7KET37', [0.000000e+00, 0.0, 0.0])

# Reaction 4625
reaction('NC7KET41 => CH2O + NC3H7COC2H4P + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 4626
reaction('CH2O + NC3H7COC2H4P + OH => NC7KET41', [0.000000e+00, 0.0, 0.0])

# Reaction 4627
reaction('NC7KET42 => CH3CHO + NC3H7COCH2 + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 4628
reaction('CH3CHO + NC3H7COCH2 + OH => NC7KET42', [0.000000e+00, 0.0, 0.0])

# Reaction 4629
reaction('NC7KET43 => C2H5CHO + NC3H7CO + OH', [1.000000e+16, 0.0, 39000.0])

# Reaction 4630
reaction('C2H5CHO + NC3H7CO + OH => NC7KET43', [0.000000e+00, 0.0, 0.0])

# Reaction 4631
reaction('C7H14O1-2 + OH => PC4H9 + C2H3CHO + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 4632
reaction('PC4H9 + C2H3CHO + H2O => C7H14O1-2 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 4633
reaction('C7H14O1-3 + OH => C6H12-1 + HCO + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 4634
reaction('C6H12-1 + HCO + H2O => C7H14O1-3 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 4635
reaction('C7H14O1-4 + OH => C5H10-1 + CH2CHO + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 4636
reaction('C5H10-1 + CH2CHO + H2O => C7H14O1-4 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 4637
reaction('C7H14O1-5 + OH => C4H8-1 + CH2CH2CHO + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 4638
reaction('C4H8-1 + CH2CH2CHO + H2O => C7H14O1-5 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 4639
reaction('C7H14O2-3 + OH => C2H3COCH3 + NC3H7 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 4640
reaction('C2H3COCH3 + NC3H7 + H2O => C7H14O2-3 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 4641
reaction('C7H14O2-4 + OH => CH3CO + C5H10-1 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 4642
reaction('CH3CO + C5H10-1 + H2O => C7H14O2-4 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 4643
reaction('C7H14O2-5 + OH => CH3COCH2 + C4H8-1 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 4644
reaction('CH3COCH2 + C4H8-1 + H2O => C7H14O2-5 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 4645
reaction('C7H14O2-6 + OH => CH2CH2COCH3 + C3H6 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 4646
reaction('CH2CH2COCH3 + C3H6 + H2O => C7H14O2-6 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 4647
reaction('C7H14O3-4 + OH => C2H5COC2H3 + C2H5 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 4648
reaction('C2H5COC2H3 + C2H5 + H2O => C7H14O3-4 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 4649
reaction('C7H14O3-5 + OH => C2H5CO + C4H8-1 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 4650
reaction('C2H5CO + C4H8-1 + H2O => C7H14O3-5 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 4651
reaction('C7H14O1-2 + OH => CH2CO + C5H11-1 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 4652
reaction('CH2CO + C5H11-1 + H2O => C7H14O1-2 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 4653
reaction('C7H14O1-3 + OH => C2H4 + NC4H9CO + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 4654
reaction('C2H4 + NC4H9CO + H2O => C7H14O1-3 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 4655
reaction('C7H14O1-4 + OH => C2H4 + NC3H7COCH2 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 4656
reaction('C2H4 + NC3H7COCH2 + H2O => C7H14O1-4 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 4657
reaction('C7H14O1-5 + OH => C2H4 + C2H5COC2H4P + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 4658
reaction('C2H4 + C2H5COC2H4P + H2O => C7H14O1-5 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 4659
reaction('C7H14O2-3 + OH => CH3CHCO + PC4H9 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 4660
reaction('CH3CHCO + PC4H9 + H2O => C7H14O2-3 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 4661
reaction('C7H14O2-4 + OH => C3H6 + NC3H7CO + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 4662
reaction('C3H6 + NC3H7CO + H2O => C7H14O2-4 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 4663
reaction('C7H14O2-5 + OH => C3H6 + C2H5COCH2 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 4664
reaction('C3H6 + C2H5COCH2 + H2O => C7H14O2-5 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 4665
reaction('C7H14O2-6 + OH => CH3CHO + C5H91-4 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 4666
reaction('CH3CHO + C5H91-4 + H2O => C7H14O2-6 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 4667
reaction('C7H14O3-4 + OH => C2H5CHCO + NC3H7 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 4668
reaction('C2H5CHCO + NC3H7 + H2O => C7H14O3-4 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 4669
reaction('C7H14O3-5 + OH => C2H5CHO + C4H71-2 + H2O', [2.500000e+12, 0.0, 0.0])

# Reaction 4670
reaction('C2H5CHO + C4H71-2 + H2O => C7H14O3-5 + OH', [0.000000e+00, 0.0, 0.0])

# Reaction 4671
reaction('C7H14O1-2 + HO2 => PC4H9 + C2H3CHO + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 4672
reaction('PC4H9 + C2H3CHO + H2O2 => C7H14O1-2 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 4673
reaction('C7H14O1-3 + HO2 => C6H12-1 + HCO + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 4674
reaction('C6H12-1 + HCO + H2O2 => C7H14O1-3 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 4675
reaction('C7H14O1-4 + HO2 => C5H10-1 + CH2CHO + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 4676
reaction('C5H10-1 + CH2CHO + H2O2 => C7H14O1-4 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 4677
reaction('C7H14O1-5 + HO2 => C4H8-1 + CH2CH2CHO + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 4678
reaction('C4H8-1 + CH2CH2CHO + H2O2 => C7H14O1-5 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 4679
reaction('C7H14O2-3 + HO2 => C2H3COCH3 + NC3H7 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 4680
reaction('C2H3COCH3 + NC3H7 + H2O2 => C7H14O2-3 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 4681
reaction('C7H14O2-4 + HO2 => CH3CO + C5H10-1 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 4682
reaction('CH3CO + C5H10-1 + H2O2 => C7H14O2-4 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 4683
reaction('C7H14O2-5 + HO2 => CH3COCH2 + C4H8-1 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 4684
reaction('CH3COCH2 + C4H8-1 + H2O2 => C7H14O2-5 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 4685
reaction('C7H14O2-6 + HO2 => CH2CH2COCH3 + C3H6 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 4686
reaction('CH2CH2COCH3 + C3H6 + H2O2 => C7H14O2-6 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 4687
reaction('C7H14O3-4 + HO2 => C2H5COC2H3 + C2H5 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 4688
reaction('C2H5COC2H3 + C2H5 + H2O2 => C7H14O3-4 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 4689
reaction('C7H14O3-5 + HO2 => C2H5CO + C4H8-1 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 4690
reaction('C2H5CO + C4H8-1 + H2O2 => C7H14O3-5 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 4691
reaction('C7H14O1-2 + HO2 => CH2CO + C5H11-1 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 4692
reaction('CH2CO + C5H11-1 + H2O2 => C7H14O1-2 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 4693
reaction('C7H14O1-3 + HO2 => C2H4 + NC4H9CO + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 4694
reaction('C2H4 + NC4H9CO + H2O2 => C7H14O1-3 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 4695
reaction('C7H14O1-4 + HO2 => C2H4 + NC3H7COCH2 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 4696
reaction('C2H4 + NC3H7COCH2 + H2O2 => C7H14O1-4 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 4697
reaction('C7H14O1-5 + HO2 => C2H4 + C2H5COC2H4P + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 4698
reaction('C2H4 + C2H5COC2H4P + H2O2 => C7H14O1-5 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 4699
reaction('C7H14O2-3 + HO2 => CH3CHCO + PC4H9 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 4700
reaction('CH3CHCO + PC4H9 + H2O2 => C7H14O2-3 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 4701
reaction('C7H14O2-4 + HO2 => C3H6 + NC3H7CO + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 4702
reaction('C3H6 + NC3H7CO + H2O2 => C7H14O2-4 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 4703
reaction('C7H14O2-5 + HO2 => C3H6 + C2H5COCH2 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 4704
reaction('C3H6 + C2H5COCH2 + H2O2 => C7H14O2-5 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 4705
reaction('C7H14O2-6 + HO2 => CH3CHO + C5H91-4 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 4706
reaction('CH3CHO + C5H91-4 + H2O2 => C7H14O2-6 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 4707
reaction('C7H14O3-4 + HO2 => C2H5CHCO + NC3H7 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 4708
reaction('C2H5CHCO + NC3H7 + H2O2 => C7H14O3-4 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 4709
reaction('C7H14O3-5 + HO2 => C2H5CHO + C4H71-2 + H2O2', [5.000000e+12, 0.0, 17700.0])

# Reaction 4710
reaction('C2H5CHO + C4H71-2 + H2O2 => C7H14O3-5 + HO2', [0.000000e+00, 0.0, 0.0])

# Reaction 4711
reaction('C7H14OH-1 => C7H14-1 + OH', [5.430000e+14, -0.53, 27830.0])

# Reaction 4712
reaction('C7H14-1 + OH => C7H14OH-1', [1.000000e+12, 0.0, -1042.0])

# Reaction 4713
reaction('O2C7H14OH-1 => C7H14OH-1 + O2', [2.983000e+21, -1.98, 37820.0])

# Reaction 4714
reaction('C7H14OH-1 + O2 => O2C7H14OH-1', [2.000000e+12, 0.0, 0.0])

# Reaction 4715
reaction('O2C7H14OH-1 => NC5H11CHO + CH2O + OH', [2.500000e+10, 0.0, 18860.0])

# Reaction 4716
reaction('NC5H11CHO + CH2O + OH => O2C7H14OH-1', [0.000000e+00, 0.0, 0.0])

# Reaction 4717
reaction('C7H14OH-2 => C7H14-2 + OH', [1.044000e+16, -1.0, 29400.0])

# Reaction 4718
reaction('C7H14-2 + OH => C7H14OH-2', [1.000000e+12, 0.0, -1042.0])

# Reaction 4719
reaction('O2C7H14OH-2 => C7H14OH-2 + O2', [4.775000e+21, -2.06, 37860.0])

# Reaction 4720
reaction('C7H14OH-2 + O2 => O2C7H14OH-2', [2.000000e+12, 0.0, 0.0])

# Reaction 4721
reaction('O2C7H14OH-1 => NC4H9CHO + CH3CHO + OH', [2.500000e+10, 0.0, 18860.0])

# Reaction 4722
reaction('NC4H9CHO + CH3CHO + OH => O2C7H14OH-1', [0.000000e+00, 0.0, 0.0])

# Reaction 4723
reaction('C7H14OH-3 => C7H14-3 + OH', [7.773000e+15, -0.93, 29490.0])

# Reaction 4724
reaction('C7H14-3 + OH => C7H14OH-3', [1.000000e+12, 0.0, -1042.0])

# Reaction 4725
reaction('O2C7H14OH-3 => C7H14OH-3 + O2', [4.775000e+21, -2.06, 37860.0])

# Reaction 4726
reaction('C7H14OH-3 + O2 => O2C7H14OH-3', [2.000000e+12, 0.0, 0.0])

# Reaction 4727
reaction('O2C7H14OH-3 => NC3H7CHO + C2H5CHO + OH', [2.500000e+10, 0.0, 18860.0])

# Reaction 4728
reaction('NC3H7CHO + C2H5CHO + OH => O2C7H14OH-3', [0.000000e+00, 0.0, 0.0])

# Reaction 4729
reaction('NC5H11CHO + O2 => NC5H11CO + HO2', [2.000000e+13, 0.5, 42200.0])

# Reaction 4730
reaction('NC5H11CO + HO2 => NC5H11CHO + O2', [1.000000e+07, 0.0, 40000.0])

# Reaction 4731
reaction('NC5H11CHO + OH => NC5H11CO + H2O', [2.690000e+10, 0.76, -340.0])

# Reaction 4732
reaction('NC5H11CO + H2O => NC5H11CHO + OH', [1.740000e+10, 0.76, 31200.0])

# Reaction 4733
reaction('NC5H11CHO + H => NC5H11CO + H2', [4.000000e+13, 0.0, 4200.0])

# Reaction 4734
reaction('NC5H11CO + H2 => NC5H11CHO + H', [1.800000e+13, 0.0, 24000.0])

# Reaction 4735
reaction('NC5H11CHO + O => NC5H11CO + OH', [5.000000e+12, 0.0, 1790.0])

# Reaction 4736
reaction('NC5H11CO + OH => NC5H11CHO + O', [1.000000e+12, 0.0, 19000.0])

# Reaction 4737
reaction('NC5H11CHO + HO2 => NC5H11CO + H2O2', [2.800000e+12, 0.0, 13600.0])

# Reaction 4738
reaction('NC5H11CO + H2O2 => NC5H11CHO + HO2', [1.000000e+12, 0.0, 10000.0])

# Reaction 4739
reaction('NC5H11CHO + CH3 => NC5H11CO + CH4', [1.700000e+12, 0.0, 8440.0])

# Reaction 4740
reaction('NC5H11CO + CH4 => NC5H11CHO + CH3', [1.500000e+13, 0.0, 28000.0])

# Reaction 4741
reaction('NC5H11CHO + CH3O => NC5H11CO + CH3OH', [1.150000e+11, 0.0, 1280.0])

# Reaction 4742
reaction('NC5H11CO + CH3OH => NC5H11CHO + CH3O', [3.000000e+11, 0.0, 18000.0])

# Reaction 4743
reaction('NC5H11CHO + CH3O2 => NC5H11CO + CH3O2H', [1.000000e+12, 0.0, 9500.0])

# Reaction 4744
reaction('NC5H11CO + CH3O2H => NC5H11CHO + CH3O2', [2.500000e+10, 0.0, 10000.0])

# Reaction 4745
reaction('NC5H11CHO + OH => C5H10CHO-1 + H2O', [5.270000e+09, 0.97, 1586.0])

# Reaction 4746
reaction('C5H10CHO-1 + H2O => NC5H11CHO + OH', [1.071000e+09, 0.96, 21010.0])

# Reaction 4747
reaction('NC5H11CHO + OH => C5H10CHO-2 + H2O', [4.670000e+07, 1.61, -35.0])

# Reaction 4748
reaction('C5H10CHO-2 + H2O => NC5H11CHO + OH', [1.752000e+06, 1.85, 23880.0])

# Reaction 4749
reaction('NC5H11CHO + OH => C5H10CHO-3 + H2O', [4.670000e+07, 1.61, -35.0])

# Reaction 4750
reaction('C5H10CHO-3 + H2O => NC5H11CHO + OH', [1.752000e+06, 1.85, 23880.0])

# Reaction 4751
reaction('NC5H11CHO + OH => C5H10CHO-4 + H2O', [4.670000e+07, 1.61, -35.0])

# Reaction 4752
reaction('C5H10CHO-4 + H2O => NC5H11CHO + OH', [1.752000e+06, 1.85, 23880.0])

# Reaction 4753
reaction('NC5H11CHO + OH => C5H10CHO-5 + H2O', [4.670000e+07, 1.61, -35.0])

# Reaction 4754
reaction('C5H10CHO-5 + H2O => NC5H11CHO + OH', [3.317000e+09, 1.25, 30840.0])

# Reaction 4755
reaction('NC5H11CO => C5H11-1 + CO', [1.000000e+11, 0.0, 9600.0])

# Reaction 4756
reaction('C5H11-1 + CO => NC5H11CO', [1.000000e+11, 0.0, 0.0])

# Reaction 4757
reaction('NC5H11CHO + HO2 => C5H10CHO-1 + H2O2', [2.760000e+04, 2.55, 16480.0])

# Reaction 4758
reaction('C5H10CHO-1 + H2O2 => NC5H11CHO + HO2', [3.330000e+04, 2.22, 4442.0])

# Reaction 4759
reaction('NC5H11CHO + HO2 => C5H10CHO-2 + H2O2', [1.475000e+04, 2.6, 13910.0])

# Reaction 4760
reaction('C5H10CHO-2 + H2O2 => NC5H11CHO + HO2', [3.285000e+03, 2.51, 6362.0])

# Reaction 4761
reaction('NC5H11CHO + HO2 => C5H10CHO-3 + H2O2', [1.475000e+04, 2.6, 13910.0])

# Reaction 4762
reaction('C5H10CHO-3 + H2O2 => NC5H11CHO + HO2', [3.285000e+03, 2.51, 6362.0])

# Reaction 4763
reaction('NC5H11CHO + HO2 => C5H10CHO-4 + H2O2', [1.475000e+04, 2.6, 13910.0])

# Reaction 4764
reaction('C5H10CHO-4 + H2O2 => NC5H11CHO + HO2', [3.285000e+03, 2.51, 6362.0])

# Reaction 4765
reaction('NC5H11CHO + HO2 => C5H10CHO-5 + H2O2', [2.950000e+04, 2.6, 13910.0])

# Reaction 4766
reaction('C5H10CHO-5 + H2O2 => NC5H11CHO + HO2', [1.244000e+07, 1.91, 13320.0])

# Reaction 4767
reaction('NC5H11CHO + CH3O2 => C5H10CHO-1 + CH3O2H', [6.030000e+12, 0.0, 19380.0])

# Reaction 4768
reaction('C5H10CHO-1 + CH3O2H => NC5H11CHO + CH3O2', [2.990000e+13, -0.51, 6617.0])

# Reaction 4769
reaction('NC5H11CHO + CH3O2 => C5H10CHO-2 + CH3O2H', [1.990000e+12, 0.0, 17050.0])

# Reaction 4770
reaction('C5H10CHO-2 + CH3O2H => NC5H11CHO + CH3O2', [1.821000e+12, -0.26, 8784.0])

# Reaction 4771
reaction('NC5H11CHO + CH3O2 => C5H10CHO-3 + CH3O2H', [1.990000e+12, 0.0, 17050.0])

# Reaction 4772
reaction('C5H10CHO-3 + CH3O2H => NC5H11CHO + CH3O2', [1.821000e+12, -0.26, 8784.0])

# Reaction 4773
reaction('NC5H11CHO + CH3O2 => C5H10CHO-4 + CH3O2H', [1.990000e+12, 0.0, 17050.0])

# Reaction 4774
reaction('C5H10CHO-4 + CH3O2H => NC5H11CHO + CH3O2', [1.821000e+12, -0.26, 8784.0])

# Reaction 4775
reaction('NC5H11CHO + CH3O2 => C5H10CHO-5 + CH3O2H', [3.980000e+12, 0.0, 17050.0])

# Reaction 4776
reaction('C5H10CHO-5 + CH3O2H => NC5H11CHO + CH3O2', [6.897000e+15, -0.86, 15740.0])

# Reaction 4777
reaction('C5H10CHO-1 => C2H4 + C3H6CHO-1', [2.680000e+18, -1.58, 30410.0])

# Reaction 4778
reaction('C2H4 + C3H6CHO-1 => C5H10CHO-1', [2.500000e+11, 0.0, 7800.0])

# Reaction 4779
reaction('C5H10CHO-2 => C3H6 + CH2CH2CHO', [9.380000e+17, -1.31, 31970.0])

# Reaction 4780
reaction('C3H6 + CH2CH2CHO => C5H10CHO-2', [1.000000e+11, 0.0, 7800.0])

# Reaction 4781
reaction('C5H10CHO-3 => C4H8-1 + CH2CHO', [6.270000e+16, -1.43, 25990.0])

# Reaction 4782
reaction('C4H8-1 + CH2CHO => C5H10CHO-3', [1.000000e+11, 0.0, 7800.0])

# Reaction 4783
reaction('C5H10CHO-3 => C4H7CHO1-4 + CH3', [2.373000e+14, -0.56, 31320.0])

# Reaction 4784
reaction('C4H7CHO1-4 + CH3 => C5H10CHO-3', [1.000000e+11, 0.0, 7800.0])

# Reaction 4785
reaction('C5H10CHO-4 => AC3H5CHO + C2H5', [7.188000e+17, -1.37, 33230.0])

# Reaction 4786
reaction('AC3H5CHO + C2H5 => C5H10CHO-4', [1.000000e+11, 0.0, 7800.0])

# Reaction 4787
reaction('C5H10CHO-4 => C5H10-1 + HCO', [1.057000e+14, -0.41, 26330.0])

# Reaction 4788
reaction('C5H10-1 + HCO => C5H10CHO-4', [1.000000e+11, 0.0, 7800.0])

# Reaction 4789
reaction('C5H10CHO-5 => C2H3CHO + NC3H7', [1.564000e+19, -1.53, 33310.0])

# Reaction 4790
reaction('C2H3CHO + NC3H7 => C5H10CHO-5', [1.000000e+11, 0.0, 7800.0])

# Reaction 4791
reaction('C4H7CHO1-4 + OH => C4H7CO1-4 + H2O', [3.370000e+12, 0.0, -616.0])

# Reaction 4792
reaction('C4H7CO1-4 + H2O => C4H7CHO1-4 + OH', [2.398000e+12, -0.02, 30940.0])

# Reaction 4793
reaction('C4H7CHO1-4 + OH => C4H6CHO1-43 + H2O', [2.080000e+06, 2.0, -298.0])

# Reaction 4794
reaction('C4H6CHO1-43 + H2O => C4H7CHO1-4 + OH', [1.187000e+08, 1.74, 34980.0])

# Reaction 4795
reaction('C4H7CHO1-4 + OH => C4H6CHO1-44 + H2O', [4.670000e+07, 1.61, -35.0])

# Reaction 4796
reaction('C4H6CHO1-44 + H2O => C4H7CHO1-4 + OH', [1.502000e+09, 1.31, 28900.0])

# Reaction 4797
reaction('C4H7CHO1-4 + OH => CH3CHO + CH2CH2CHO', [1.000000e+11, 0.0, 0.0])

# Reaction 4798
reaction('CH3CHO + CH2CH2CHO => C4H7CHO1-4 + OH', [5.344000e+06, 0.97, 19260.0])

# Reaction 4799
reaction('C4H7CHO1-4 + HO2 => C4H7CO1-4 + H2O2', [1.000000e+12, 0.0, 11920.0])

# Reaction 4800
reaction('C4H7CO1-4 + H2O2 => C4H7CHO1-4 + HO2', [4.225000e+12, -0.35, 12010.0])

# Reaction 4801
reaction('C4H7CHO1-4 + HO2 => C4H6CHO1-43 + H2O2', [9.640000e+03, 2.6, 13910.0])

# Reaction 4802
reaction('C4H6CHO1-43 + H2O2 => C4H7CHO1-4 + HO2', [3.265000e+06, 2.01, 17730.0])

# Reaction 4803
reaction('C4H7CHO1-4 + HO2 => C4H6CHO1-44 + H2O2', [1.475000e+04, 2.6, 13910.0])

# Reaction 4804
reaction('C4H6CHO1-44 + H2O2 => C4H7CHO1-4 + HO2', [2.815000e+06, 1.97, 11380.0])

# Reaction 4805
reaction('C4H7CHO1-4 + CH3O2 => C4H7CO1-4 + CH3O2H', [2.800000e+12, 0.0, 13600.0])

# Reaction 4806
reaction('C4H7CO1-4 + CH3O2H => C4H7CHO1-4 + CH3O2', [4.861000e+13, -0.52, 12970.0])

# Reaction 4807
reaction('C4H7CHO1-4 + CH3O2 => C4H6CHO1-43 + CH3O2H', [1.990000e+12, 0.0, 17050.0])

# Reaction 4808
reaction('C4H6CHO1-43 + CH3O2H => C4H7CHO1-4 + CH3O2', [2.770000e+15, -0.77, 20150.0])

# Reaction 4809
reaction('C4H7CHO1-4 + CH3O2 => C4H6CHO1-44 + CH3O2H', [1.990000e+12, 0.0, 17050.0])

# Reaction 4810
reaction('C4H6CHO1-44 + CH3O2H => C4H7CHO1-4 + CH3O2', [1.561000e+15, -0.8, 13800.0])

# Reaction 4811
reaction('C4H7CO1-4 => C4H71-4 + CO', [7.372000e+18, -1.76, 15230.0])

# Reaction 4812
reaction('C4H71-4 + CO => C4H7CO1-4', [1.500000e+11, 0.0, 4810.0])

# Reaction 4813
reaction('C4H6CHO1-43 => C4H6 + HCO', [4.834000e+15, -0.79, 33540.0])

# Reaction 4814
reaction('C4H6 + HCO => C4H6CHO1-43', [1.000000e+11, 0.0, 7800.0])

# Reaction 4815
reaction('C4H6CHO1-44 => C2H3CHO + C2H3', [2.118000e+14, -0.39, 37160.0])

# Reaction 4816
reaction('C2H3CHO + C2H3 => C4H6CHO1-44', [1.000000e+11, 0.0, 7800.0])

# Reaction 4817
reaction('NC4H9COCH3 + OH => C4H8COCH3-1 + H2O', [2.065000e+07, 1.73, 753.0])

# Reaction 4818
reaction('C4H8COCH3-1 + H2O => NC4H9COCH3 + OH', [4.278000e+06, 1.72, 20190.0])

# Reaction 4819
reaction('NC4H9COCH3 + OH => C4H8COCH3-2 + H2O', [3.615000e+07, 1.64, -247.0])

# Reaction 4820
reaction('C4H8COCH3-2 + H2O => NC4H9COCH3 + OH', [3.070000e+05, 2.04, 21630.0])

# Reaction 4821
reaction('NC4H9COCH3 + OH => C4H8COCH3-3 + H2O', [3.615000e+07, 1.64, -247.0])

# Reaction 4822
reaction('C4H8COCH3-3 + H2O => NC4H9COCH3 + OH', [3.070000e+05, 2.04, 21630.0])

# Reaction 4823
reaction('NC4H9COCH3 + OH => C4H8COCH3-4 + H2O', [8.450000e+11, 0.0, -228.0])

# Reaction 4824
reaction('C4H8COCH3-4 + H2O => NC4H9COCH3 + OH', [1.633000e+13, -0.23, 28640.0])

# Reaction 4825
reaction('NC4H9COCH3 + OH => NC4H9COCH2 + H2O', [5.100000e+11, 0.0, 1192.0])

# Reaction 4826
reaction('NC4H9COCH2 + H2O => NC4H9COCH3 + OH', [6.786000e+13, -0.7, 27680.0])

# Reaction 4827
reaction('NC4H9COCH3 + HO2 => C4H8COCH3-1 + H2O2', [2.380000e+04, 2.55, 16490.0])

# Reaction 4828
reaction('C4H8COCH3-1 + H2O2 => NC4H9COCH3 + HO2', [2.927000e+04, 2.21, 4462.0])

# Reaction 4829
reaction('NC4H9COCH3 + HO2 => C4H8COCH3-2 + H2O2', [5.600000e+12, 0.0, 17700.0])

# Reaction 4830
reaction('C4H8COCH3-2 + H2O2 => NC4H9COCH3 + HO2', [2.823000e+11, 0.07, 8112.0])

# Reaction 4831
reaction('NC4H9COCH3 + HO2 => C4H8COCH3-3 + H2O2', [5.600000e+12, 0.0, 17700.0])

# Reaction 4832
reaction('C4H8COCH3-3 + H2O2 => NC4H9COCH3 + HO2', [2.823000e+11, 0.07, 8112.0])

# Reaction 4833
reaction('NC4H9COCH3 + HO2 => C4H8COCH3-4 + H2O2', [2.000000e+11, 0.0, 8698.0])

# Reaction 4834
reaction('C4H8COCH3-4 + H2O2 => NC4H9COCH3 + HO2', [2.294000e+13, -0.56, 6100.0])

# Reaction 4835
reaction('NC4H9COCH3 + HO2 => NC4H9COCH2 + H2O2', [2.380000e+04, 2.55, 14690.0])

# Reaction 4836
reaction('NC4H9COCH2 + H2O2 => NC4H9COCH3 + HO2', [1.880000e+07, 1.52, 9712.0])

# Reaction 4837
reaction('NC4H9COCH3 + CH3O2 => C4H8COCH3-1 + CH3O2H', [3.010000e+12, 0.0, 19380.0])

# Reaction 4838
reaction('C4H8COCH3-1 + CH3O2H => NC4H9COCH3 + CH3O2', [1.521000e+13, -0.51, 6627.0])

# Reaction 4839
reaction('NC4H9COCH3 + CH3O2 => C4H8COCH3-2 + CH3O2H', [1.990000e+12, 0.0, 17050.0])

# Reaction 4840
reaction('C4H8COCH3-2 + CH3O2H => NC4H9COCH3 + CH3O2', [4.123000e+11, -0.11, 6744.0])

# Reaction 4841
reaction('NC4H9COCH3 + CH3O2 => C4H8COCH3-3 + CH3O2H', [1.990000e+12, 0.0, 17050.0])

# Reaction 4842
reaction('C4H8COCH3-3 + CH3O2H => NC4H9COCH3 + CH3O2', [4.123000e+11, -0.11, 6744.0])

# Reaction 4843
reaction('NC4H9COCH3 + CH3O2 => C4H8COCH3-4 + CH3O2H', [2.000000e+12, 0.0, 15250.0])

# Reaction 4844
reaction('C4H8COCH3-4 + CH3O2H => NC4H9COCH3 + CH3O2', [9.429000e+14, -0.73, 11930.0])

# Reaction 4845
reaction('NC4H9COCH3 + CH3O2 => NC4H9COCH2 + CH3O2H', [3.010000e+12, 0.0, 17580.0])

# Reaction 4846
reaction('NC4H9COCH2 + CH3O2H => NC4H9COCH3 + CH3O2', [9.770000e+15, -1.21, 11880.0])

# Reaction 4847
reaction('C4H8COCH3-1 => CH2CH2COCH3 + C2H4', [1.133000e+18, -1.59, 30910.0])

# Reaction 4848
reaction('CH2CH2COCH3 + C2H4 => C4H8COCH3-1', [1.000000e+11, 0.0, 8300.0])

# Reaction 4849
reaction('C4H8COCH3-2 => C3H6 + CH3COCH2', [3.175000e+15, -0.79, 26220.0])

# Reaction 4850
reaction('C3H6 + CH3COCH2 => C4H8COCH3-2', [1.000000e+11, 0.0, 8300.0])

# Reaction 4851
reaction('C4H8COCH3-3 => C4H8-1 + CH3CO', [1.714000e+18, -1.61, 28250.0])

# Reaction 4852
reaction('C4H8-1 + CH3CO => C4H8COCH3-3', [1.000000e+11, 0.0, 9800.0])

# Reaction 4853
reaction('C4H8COCH3-4 => C2H3COCH3 + C2H5', [4.691000e+19, -1.61, 33750.0])

# Reaction 4854
reaction('C2H3COCH3 + C2H5 => C4H8COCH3-4', [1.000000e+11, 0.0, 8300.0])

# Reaction 4855
reaction('NC4H9COCH2 => PC4H9 + CH2CO', [1.554000e+18, -1.41, 43140.0])

# Reaction 4856
reaction('PC4H9 + CH2CO => NC4H9COCH2', [1.000000e+11, 0.0, 11600.0])

# Reaction 4857
reaction('C4H7OOH1-4 => C4H7O1-4 + OH', [2.021000e+20, -1.53, 47040.0])

# Reaction 4858
reaction('C4H7O1-4 + OH => C4H7OOH1-4', [2.000000e+13, 0.0, 0.0])

# Reaction 4859
reaction('C5H9OOH1-4 => C5H9O1-4 + OH', [1.178000e+20, -1.38, 46050.0])

# Reaction 4860
reaction('C5H9O1-4 + OH => C5H9OOH1-4', [2.000000e+13, 0.0, 0.0])

# Reaction 4861
reaction('C5H9OOH1-5 => C5H9O1-5 + OH', [1.594000e+20, -1.5, 46990.0])

# Reaction 4862
reaction('C5H9O1-5 + OH => C5H9OOH1-5', [2.000000e+13, 0.0, 0.0])

# Reaction 4863
reaction('C6H11OOH1-4 => C6H11O1-4 + OH', [1.234000e+20, -1.39, 46050.0])

# Reaction 4864
reaction('C6H11O1-4 + OH => C6H11OOH1-4', [2.000000e+13, 0.0, 0.0])

# Reaction 4865
reaction('C6H11OOH1-5 => C6H11O1-5 + OH', [1.234000e+20, -1.39, 46050.0])

# Reaction 4866
reaction('C6H11O1-5 + OH => C6H11OOH1-5', [2.000000e+13, 0.0, 0.0])

# Reaction 4867
reaction('C4H7O1-4 => CH2O + C3H5-A', [2.412000e+16, -1.14, 7550.0])

# Reaction 4868
reaction('CH2O + C3H5-A => C4H7O1-4', [1.000000e+11, 0.0, 11900.0])

# Reaction 4869
reaction('C5H9O1-4 => CH3CHO + C3H5-A', [7.715000e+20, -2.43, 5890.0])

# Reaction 4870
reaction('CH3CHO + C3H5-A => C5H9O1-4', [5.000000e+10, 0.0, 9600.0])

# Reaction 4871
reaction('C5H9O1-4 => AC3H5CHO + CH3', [1.112000e+17, -1.21, 17960.0])

# Reaction 4872
reaction('AC3H5CHO + CH3 => C5H9O1-4', [1.000000e+11, 0.0, 9600.0])

# Reaction 4873
reaction('C5H9O1-5 => CH2O + C4H71-4', [4.310000e+17, -1.33, 17940.0])

# Reaction 4874
reaction('CH2O + C4H71-4 => C5H9O1-5', [1.000000e+11, 0.0, 9600.0])

# Reaction 4875
reaction('C6H11O1-4 => AC3H5CHO + C2H5', [7.268000e+20, -2.1, 18870.0])

# Reaction 4876
reaction('AC3H5CHO + C2H5 => C6H11O1-4', [1.000000e+11, 0.0, 9600.0])

# Reaction 4877
reaction('C6H11O1-4 => C2H5CHO + C3H5-A', [1.211000e+21, -2.46, 5641.0])

# Reaction 4878
reaction('C2H5CHO + C3H5-A => C6H11O1-4', [5.000000e+10, 0.0, 9600.0])

# Reaction 4879
reaction('C6H11O1-5 => C4H7CHO1-4 + CH3', [2.399000e+17, -1.3, 16960.0])

# Reaction 4880
reaction('C4H7CHO1-4 + CH3 => C6H11O1-5', [1.000000e+11, 0.0, 9600.0])

# Reaction 4881
reaction('C6H11O1-5 => CH3CHO + C4H71-4', [1.999000e+22, -2.58, 18530.0])

# Reaction 4882
reaction('CH3CHO + C4H71-4 => C6H11O1-5', [1.000000e+11, 0.0, 9600.0])

# Reaction 4883
reaction('C5H91-1 => C2H2 + NC3H7', [2.755000e+15, -0.67, 30800.0])

# Reaction 4884
reaction('C2H2 + NC3H7 => C5H91-1', [2.000000e+11, 0.0, 7800.0])

# Reaction 4885
reaction('C4H7CHO1-1 + OH => C4H7CO1-1 + H2O', [3.370000e+12, 0.0, -616.0])

# Reaction 4886
reaction('C4H7CO1-1 + H2O => C4H7CHO1-1 + OH', [5.203000e+12, 0.03, 36800.0])

# Reaction 4887
reaction('C4H7CHO1-1 + OH => C4H6CHO1-14 + H2O', [5.270000e+09, 0.97, 1586.0])

# Reaction 4888
reaction('C4H6CHO1-14 + H2O => C4H7CHO1-1 + OH', [1.067000e+09, 0.96, 21020.0])

# Reaction 4889
reaction('C4H7CHO1-1 + OH => C4H6CHO1-13 + H2O', [3.120000e+06, 2.0, -298.0])

# Reaction 4890
reaction('C4H6CHO1-13 + H2O => C4H7CHO1-1 + OH', [1.251000e+08, 1.79, 34930.0])

# Reaction 4891
reaction('C4H7CHO1-1 + OH => C2H5CHO + CH2CHO', [1.000000e+11, 0.0, 0.0])

# Reaction 4892
reaction('C2H5CHO + CH2CHO => C4H7CHO1-1 + OH', [1.076000e+07, 1.1, 17020.0])

# Reaction 4893
reaction('C4H7CHO1-1 + HO2 => C4H7CO1-1 + H2O2', [1.000000e+12, 0.0, 11920.0])

# Reaction 4894
reaction('C4H7CO1-1 + H2O2 => C4H7CHO1-1 + HO2', [9.165000e+12, -0.3, 17870.0])

# Reaction 4895
reaction('C4H7CHO1-1 + HO2 => C4H6CHO1-14 + H2O2', [2.760000e+04, 2.55, 16480.0])

# Reaction 4896
reaction('C4H6CHO1-14 + H2O2 => C4H7CHO1-1 + HO2', [3.317000e+04, 2.22, 4452.0])

# Reaction 4897
reaction('C4H7CHO1-1 + HO2 => C4H6CHO1-13 + H2O2', [9.640000e+03, 2.6, 13910.0])

# Reaction 4898
reaction('C4H6CHO1-13 + H2O2 => C4H7CHO1-1 + HO2', [2.295000e+06, 2.06, 17680.0])

# Reaction 4899
reaction('C4H7CHO1-1 + CH3O2 => C4H7CO1-1 + CH3O2H', [2.800000e+12, 0.0, 13600.0])

# Reaction 4900
reaction('C4H7CO1-1 + CH3O2H => C4H7CHO1-1 + CH3O2', [1.055000e+14, -0.47, 18830.0])

# Reaction 4901
reaction('C4H7CHO1-1 + CH3O2 => C4H6CHO1-14 + CH3O2H', [6.030000e+12, 0.0, 19380.0])

# Reaction 4902
reaction('C4H6CHO1-14 + CH3O2H => C4H7CHO1-1 + CH3O2', [2.978000e+13, -0.51, 6627.0])

# Reaction 4903
reaction('C4H7CHO1-1 + CH3O2 => C4H6CHO1-13 + CH3O2H', [1.990000e+12, 0.0, 17050.0])

# Reaction 4904
reaction('C4H6CHO1-13 + CH3O2H => C4H7CHO1-1 + CH3O2', [1.947000e+15, -0.72, 20100.0])

# Reaction 4905
reaction('C4H7CO1-1 => C4H71-1 + CO', [8.510000e+20, -2.12, 40320.0])

# Reaction 4906
reaction('C4H71-1 + CO => C4H7CO1-1', [1.500000e+11, 0.0, 4810.0])

# Reaction 4907
reaction('C4H6CHO1-14 => C2H4 + CHCHCHO', [9.050000e+16, -1.33, 46870.0])

# Reaction 4908
reaction('C2H4 + CHCHCHO => C4H6CHO1-14', [8.000000e+10, 0.0, 11400.0])

# Reaction 4909
reaction('C4H6CHO1-13 => C4H6 + HCO', [8.954000e+17, -1.28, 46230.0])

# Reaction 4910
reaction('C4H6 + HCO => C4H6CHO1-13', [8.000000e+10, 0.0, 11400.0])

# Reaction 4911
reaction('NC3H7COC2H5 + OH => C3H6COC2H5-1 + H2O', [2.065000e+07, 1.73, 753.0])

# Reaction 4912
reaction('C3H6COC2H5-1 + H2O => NC3H7COC2H5 + OH', [5.173000e+06, 1.7, 20220.0])

# Reaction 4913
reaction('NC3H7COC2H5 + OH => C3H6COC2H5-2 + H2O', [3.615000e+07, 1.64, -247.0])

# Reaction 4914
reaction('C3H6COC2H5-2 + H2O => NC3H7COC2H5 + OH', [7.233000e+05, 1.92, 21760.0])

# Reaction 4915
reaction('NC3H7COC2H5 + OH => C3H6COC2H5-3 + H2O', [8.450000e+11, 0.0, -228.0])

# Reaction 4916
reaction('C3H6COC2H5-3 + H2O => NC3H7COC2H5 + OH', [3.313000e+13, -0.32, 28740.0])

# Reaction 4917
reaction('NC3H7COC2H5 + OH => NC3H7COC2H4P + H2O', [5.100000e+11, 0.0, 1192.0])

# Reaction 4918
reaction('NC3H7COC2H4P + H2O => NC3H7COC2H5 + OH', [1.278000e+11, -0.03, 20660.0])

# Reaction 4919
reaction('NC3H7COC2H5 + OH => NC3H7COC2H4S + H2O', [8.450000e+11, 0.0, -228.0])

# Reaction 4920
reaction('NC3H7COC2H4S + H2O => NC3H7COC2H5 + OH', [3.313000e+13, -0.32, 28740.0])

# Reaction 4921
reaction('NC3H7COC2H5 + HO2 => C3H6COC2H5-1 + H2O2', [2.380000e+04, 2.55, 16490.0])

# Reaction 4922
reaction('C3H6COC2H5-1 + H2O2 => NC3H7COC2H5 + HO2', [3.539000e+04, 2.19, 4492.0])

# Reaction 4923
reaction('NC3H7COC2H5 + HO2 => C3H6COC2H5-2 + H2O2', [5.600000e+12, 0.0, 17700.0])

# Reaction 4924
reaction('C3H6COC2H5-2 + H2O2 => NC3H7COC2H5 + HO2', [6.652000e+11, -0.05, 8242.0])

# Reaction 4925
reaction('NC3H7COC2H5 + HO2 => C3H6COC2H5-3 + H2O2', [2.000000e+11, 0.0, 8698.0])

# Reaction 4926
reaction('C3H6COC2H5-3 + H2O2 => NC3H7COC2H5 + HO2', [4.654000e+13, -0.65, 6200.0])

# Reaction 4927
reaction('NC3H7COC2H5 + HO2 => NC3H7COC2H4P + H2O2', [2.380000e+04, 2.55, 14690.0])

# Reaction 4928
reaction('NC3H7COC2H4P + H2O2 => NC3H7COC2H5 + HO2', [3.539000e+04, 2.19, 2692.0])

# Reaction 4929
reaction('NC3H7COC2H5 + HO2 => NC3H7COC2H4S + H2O2', [2.000000e+11, 0.0, 8698.0])

# Reaction 4930
reaction('NC3H7COC2H4S + H2O2 => NC3H7COC2H5 + HO2', [4.654000e+13, -0.65, 6200.0])

# Reaction 4931
reaction('NC3H7COC2H5 + CH3O2 => C3H6COC2H5-1 + CH3O2H', [3.010000e+12, 0.0, 19380.0])

# Reaction 4932
reaction('C3H6COC2H5-1 + CH3O2H => NC3H7COC2H5 + CH3O2', [1.839000e+13, -0.54, 6657.0])

# Reaction 4933
reaction('NC3H7COC2H5 + CH3O2 => C3H6COC2H5-2 + CH3O2H', [1.990000e+12, 0.0, 17050.0])

# Reaction 4934
reaction('C3H6COC2H5-2 + CH3O2H => NC3H7COC2H5 + CH3O2', [9.713000e+11, -0.22, 6874.0])

# Reaction 4935
reaction('NC3H7COC2H5 + CH3O2 => C3H6COC2H5-3 + CH3O2H', [2.000000e+12, 0.0, 15250.0])

# Reaction 4936
reaction('C3H6COC2H5-3 + CH3O2H => NC3H7COC2H5 + CH3O2', [1.913000e+15, -0.83, 12030.0])

# Reaction 4937
reaction('NC3H7COC2H5 + CH3O2 => NC3H7COC2H4P + CH3O2H', [3.010000e+12, 0.0, 17580.0])

# Reaction 4938
reaction('NC3H7COC2H4P + CH3O2H => NC3H7COC2H5 + CH3O2', [1.839000e+13, -0.54, 4862.0])

# Reaction 4939
reaction('NC3H7COC2H5 + CH3O2 => NC3H7COC2H4S + CH3O2H', [2.000000e+12, 0.0, 15250.0])

# Reaction 4940
reaction('NC3H7COC2H4S + CH3O2H => NC3H7COC2H5 + CH3O2', [1.913000e+15, -0.83, 12030.0])

# Reaction 4941
reaction('C3H6COC2H5-1 => C2H4 + C2H5COCH2', [2.218000e+15, -0.84, 23590.0])

# Reaction 4942
reaction('C2H4 + C2H5COCH2 => C3H6COC2H5-1', [2.500000e+11, 0.0, 7800.0])

# Reaction 4943
reaction('C3H6COC2H5-2 => C3H6 + C2H5CO', [4.051000e+16, -1.11, 26150.0])

# Reaction 4944
reaction('C3H6 + C2H5CO => C3H6COC2H5-2', [1.000000e+11, 0.0, 8300.0])

# Reaction 4945
reaction('C3H6COC2H5-3 => C2H5COC2H3 + CH3', [2.915000e+15, -0.68, 32300.0])

# Reaction 4946
reaction('C2H5COC2H3 + CH3 => C3H6COC2H5-3', [1.000000e+11, 0.0, 7800.0])

# Reaction 4947
reaction('NC3H7COC2H4P => NC3H7CO + C2H4', [5.398000e+17, -1.45, 26040.0])

# Reaction 4948
reaction('NC3H7CO + C2H4 => NC3H7COC2H4P', [2.500000e+11, 0.0, 7800.0])

# Reaction 4949
reaction('NC3H7COC2H4S => CH3CHCO + NC3H7', [1.973000e+19, -1.49, 42860.0])

# Reaction 4950
reaction('CH3CHCO + NC3H7 => NC3H7COC2H4S', [1.000000e+11, 0.0, 10600.0])

# Reaction 4951
reaction('CHCHCHO + OH => CH2CHO + HCO', [1.000000e+12, 0.0, 0.0])

# Reaction 4952
reaction('CH2CHO + HCO => CHCHCHO + OH', [2.719000e+11, 0.29, 36200.0])

# Reaction 4953
reaction('O2 + C6H12-1 => CH2O + NC4H9CHO', [1.000000e+14, 0.0, 37000.0])

# Reaction 4954
reaction('C6H12-1 <=> C3H6 + C3H6', [4.000000e+12, 0.0, 58000.0])

# Reaction 4955
reaction('C6H111-3 + H <=> C6H12-1', [1.000000e+14, 0.0, 0.0])

# Reaction 4956
reaction('C5H91-5 + CH3 <=> C6H12-1', [1.000000e+13, 0.0, 0.0])

# Reaction 4957
reaction('PC4H9 + C2H3 <=> C6H12-1', [1.000000e+13, 0.0, 0.0])

# Reaction 4958
reaction('C4H71-4 + C2H5 <=> C6H12-1', [8.000000e+12, 0.0, 0.0])

# Reaction 4959
reaction('C6H111-3 + H <=> C6H12-2', [1.000000e+14, 0.0, 0.0])

# Reaction 4960
reaction('C6H112-4 + H <=> C6H12-2', [1.000000e+14, 0.0, 0.0])

# Reaction 4961
reaction('C5H92-5 + CH3 <=> C6H12-2', [1.000000e+13, 0.0, 0.0])

# Reaction 4962
reaction('NC3H7 + C3H5-T <=> C6H12-2', [1.000000e+13, 0.0, 0.0])

# Reaction 4963
reaction('C6H112-4 + H <=> C6H12-3', [1.000000e+14, 0.0, 0.0])

# Reaction 4964
reaction('C4H71-1 + C2H5 <=> C6H12-3', [1.000000e+13, 0.0, 0.0])

# Reaction 4965
reaction('C6H12-1 + O2 <=> C6H111-3 + HO2', [2.200000e+12, 0.0, 37220.0])

# Reaction 4966
reaction('C6H12-1 + O2 <=> C6H111-4 + HO2', [2.000000e+13, 0.0, 49640.0])

# Reaction 4967
reaction('C6H12-1 + O2 <=> C6H111-5 + HO2', [2.000000e+13, 0.0, 49640.0])

# Reaction 4968
reaction('C6H12-1 + O2 <=> C6H111-6 + HO2', [3.000000e+13, 0.0, 52290.0])

# Reaction 4969
reaction('C6H12-1 + O <=> C6H111-3 + OH', [6.600000e+05, 2.43, 1210.0])

# Reaction 4970
reaction('C6H12-1 + O <=> C6H111-4 + OH', [5.510000e+05, 2.45, 2830.0])

# Reaction 4971
reaction('C6H12-1 + O <=> C6H111-5 + OH', [5.510000e+05, 2.45, 2830.0])

# Reaction 4972
reaction('C6H12-1 + O <=> C6H111-6 + OH', [9.800000e+05, 2.43, 4750.0])

# Reaction 4973
reaction('C6H12-2 + O2 <=> C6H111-3 + HO2', [5.500000e+12, 0.0, 39900.0])

# Reaction 4974
reaction('C6H12-2 + O2 <=> C6H112-4 + HO2', [2.200000e+12, 0.0, 37220.0])

# Reaction 4975
reaction('C6H12-2 + O2 <=> C6H112-5 + HO2', [2.000000e+13, 0.0, 49640.0])

# Reaction 4976
reaction('C6H12-2 + O2 <=> C6H112-6 + HO2', [3.000000e+13, 0.0, 52290.0])

# Reaction 4977
reaction('C6H12-2 + O <=> C6H111-3 + OH', [4.410000e+05, 2.42, 3150.0])

# Reaction 4978
reaction('C6H12-2 + O <=> C6H112-4 + OH', [6.600000e+05, 2.43, 1210.0])

# Reaction 4979
reaction('C6H12-2 + O <=> C6H112-5 + OH', [5.510000e+05, 2.45, 2830.0])

# Reaction 4980
reaction('C6H12-2 + O <=> C6H112-6 + OH', [9.800000e+05, 2.43, 4750.0])

# Reaction 4981
reaction('C6H12-3 + O2 <=> C6H112-4 + HO2', [4.400000e+12, 0.0, 37220.0])

# Reaction 4982
reaction('C6H12-3 + O2 <=> C6H113-1 + HO2', [6.000000e+13, 0.0, 52290.0])

# Reaction 4983
reaction('C6H12-3 + O <=> C6H112-4 + OH', [1.320000e+06, 2.43, 1210.0])

# Reaction 4984
reaction('C6H12-3 + O <=> C6H113-1 + OH', [1.960000e+06, 2.43, 4750.0])

# Reaction 4985
reaction('C2H5 + C4H6 <=> C6H111-3', [8.500000e+10, 0.0, 8300.0])

# Reaction 4986
reaction('CH3 + C5H81-4 <=> C6H111-4', [1.000000e+11, 0.0, 7800.0])

# Reaction 4987
reaction('C3H5-A + C3H6 <=> C6H111-5', [4.000000e+11, 0.0, 16900.0])

# Reaction 4988
reaction('C4H71-4 + C2H4 <=> C6H111-6', [1.000000e+11, 0.0, 8200.0])

# Reaction 4989
reaction('CH3 + C5H81-3 <=> C6H112-4', [6.000000e+10, 0.0, 7500.0])

# Reaction 4990
reaction('C3H5-T + C3H6 <=> C6H112-5', [6.300000e+11, 0.0, 3100.0])

# Reaction 4991
reaction('C4H71-3 + C2H4 <=> C6H112-6', [4.000000e+11, 0.0, 13050.0])

# Reaction 4992
reaction('C4H71-1 + C2H4 <=> C6H113-1', [6.300000e+11, 0.0, 3100.0])

# Reaction 4993
reaction('C6H111-6 <=> C6H111-4', [3.800000e+10, 0.67, 36000.0])

# Reaction 4994
reaction('C6H111-6 <=> C6H111-5', [3.560000e+10, 0.88, 37300.0])

# Reaction 4995
reaction('C6H112-6 <=> C6H112-4', [3.800000e+10, 0.67, 28400.0])

# Reaction 4996
reaction('C6H112-6 <=> C6H112-5', [3.560000e+10, 0.88, 37300.0])

# Reaction 4997
reaction('C6H113-1 <=> C6H112-4', [3.560000e+10, 0.88, 29600.0])

# Reaction 4998
reaction('C6H111-3 + O2 <=> C6H112O2-1', [2.000000e+12, 0.0, 0.0])

# Reaction 4999
reaction('C6H111-3 + O2 <=> C6H111O2-3', [2.000000e+12, 0.0, 0.0])

# Reaction 5000
reaction('C6H111-4 + O2 <=> C6H111O2-4', [7.540000e+12, 0.0, 0.0])

# Reaction 5001
reaction('C6H111-5 + O2 <=> C6H111O2-5', [7.540000e+12, 0.0, 0.0])

# Reaction 5002
reaction('C6H111-6 + O2 <=> C6H111O2-6', [4.520000e+12, 0.0, 0.0])

# Reaction 5003
reaction('C6H112-4 + O2 <=> C6H113O2-2', [2.000000e+12, 0.0, 0.0])

# Reaction 5004
reaction('C6H112-4 + O2 <=> C6H112O2-4', [2.000000e+12, 0.0, 0.0])

# Reaction 5005
reaction('C6H112-5 + O2 <=> C6H112O2-5', [7.540000e+12, 0.0, 0.0])

# Reaction 5006
reaction('C6H112-6 + O2 <=> C6H112O2-6', [4.520000e+12, 0.0, 0.0])

# Reaction 5007
reaction('C6H113-1 + O2 <=> C6H113O2-1', [4.520000e+12, 0.0, 0.0])

# Reaction 5008
reaction('C6H111O2-3 => C6H101-3 + HO2', [1.004000e+39, -8.11, 40500.0])

# Reaction 5009
reaction('C6H111O2-4 => C6H101-3 + HO2', [2.040000e+38, -8.11, 38500.0])

# Reaction 5010
reaction('C6H111O2-5 => C6H101-4 + HO2', [1.004000e+39, -8.11, 40500.0])

# Reaction 5011
reaction('C6H111O2-6 => C6H101-5 + HO2', [1.004000e+39, -8.11, 40500.0])

# Reaction 5012
reaction('C6H111O2-5 => C6H101-5 + HO2', [1.008000e+43, -9.41, 41500.0])

# Reaction 5013
reaction('C6H112O2-4 => C6H101-3 + HO2', [5.040000e+38, -8.11, 40500.0])

# Reaction 5014
reaction('C6H112O2-5 => C6H102-4 + HO2', [1.040000e+38, -8.11, 37500.0])

# Reaction 5015
reaction('C6H112O2-6 => C6H101-4 + HO2', [5.040000e+38, -8.11, 40500.0])

# Reaction 5016
reaction('C6H113O2-2 => C6H101-3 + HO2', [5.080000e+42, -9.41, 41500.0])

# Reaction 5017
reaction('C6H113O2-1 => C6H101-3 + HO2', [1.040000e+38, -8.11, 37500.0])

# Reaction 5018
reaction('C6H111O2-3 <=> C6H101OOH3-4', [2.000000e+11, 0.0, 26450.0])

# Reaction 5019
reaction('C6H111O2-3 <=> C6H101OOH3-5', [2.500000e+10, 0.0, 20450.0])

# Reaction 5020
reaction('C6H111O2-3 <=> C6H101OOH3-6', [9.376000e+09, 0.0, 21950.0])

# Reaction 5021
reaction('C6H111O2-4 <=> C6H101OOH4-3', [2.000000e+11, 0.0, 24450.0])

# Reaction 5022
reaction('C6H111O2-4 <=> C6H101OOH4-5', [2.000000e+11, 0.0, 26450.0])

# Reaction 5023
reaction('C6H111O2-4 <=> C6H101OOH4-6', [3.750000e+10, 0.0, 24000.0])

# Reaction 5024
reaction('C6H111O2-5 <=> C6H101OOH5-3', [2.500000e+10, 0.0, 18450.0])

# Reaction 5025
reaction('C6H111O2-5 <=> C6H101OOH5-4', [2.000000e+11, 0.0, 26450.0])

# Reaction 5026
reaction('C6H111O2-5 <=> C6H101OOH5-6', [3.000000e+11, 0.0, 29000.0])

# Reaction 5027
reaction('C6H111O2-6 <=> C6H101OOH6-3', [3.125000e+09, 0.0, 16650.0])

# Reaction 5028
reaction('C6H111O2-6 <=> C6H101OOH6-4', [2.500000e+10, 0.0, 20450.0])

# Reaction 5029
reaction('C6H111O2-6 <=> C6H101OOH6-5', [2.000000e+11, 0.0, 26450.0])

# Reaction 5030
reaction('C6H112O2-4 <=> C6H102OOH4-5', [2.000000e+11, 0.0, 26450.0])

# Reaction 5031
reaction('C6H112O2-4 <=> C6H102OOH4-6', [3.750000e+10, 0.0, 24000.0])

# Reaction 5032
reaction('C6H112O2-5 <=> C6H102OOH5-4', [2.000000e+11, 0.0, 24450.0])

# Reaction 5033
reaction('C6H112O2-5 <=> C6H102OOH5-6', [3.000000e+11, 0.0, 29000.0])

# Reaction 5034
reaction('C6H112O2-6 <=> C6H102OOH6-4', [2.500000e+10, 0.0, 18450.0])

# Reaction 5035
reaction('C6H112O2-6 <=> C6H102OOH6-5', [2.000000e+11, 0.0, 26450.0])

# Reaction 5036
reaction('C6H113O2-1 <=> C6H103OOH1-2', [2.000000e+11, 0.0, 24450.0])

# Reaction 5037
reaction('C6H113O2-2 <=> C6H103OOH2-1', [3.000000e+11, 0.0, 29000.0])

# Reaction 5038
reaction('C6H101OOH3-4 => ETES1 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 5039
reaction('C6H101OOH3-5 => MVOX + OH', [7.500000e+10, 0.0, 15250.0])

# Reaction 5040
reaction('C6H101OOH3-4 => VTHF + OH', [9.375000e+09, 0.0, 7000.0])

# Reaction 5041
reaction('C6H101OOH4-3 => EDHF + OH', [9.375000e+09, 0.0, 7000.0])

# Reaction 5042
reaction('C6H101OOH4-3 => ETES1 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 5043
reaction('C6H101OOH4-5 => ETES1 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 5044
reaction('C6H101OOH4-6 => ETES1 + OH', [7.500000e+10, 0.0, 15250.0])

# Reaction 5045
reaction('C6H101OOH5-3 => MVOX + OH', [7.500000e+10, 0.0, 15250.0])

# Reaction 5046
reaction('C6H101OOH5-4 => ETES1 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 5047
reaction('C6H101OOH5-6 => ETES1 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 5048
reaction('C6H101OOH6-3 => VTHF + OH', [9.375000e+09, 0.0, 7000.0])

# Reaction 5049
reaction('C6H101OOH6-4 => ETES1 + OH', [7.500000e+10, 0.0, 15250.0])

# Reaction 5050
reaction('C6H101OOH6-5 => ETES1 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 5051
reaction('C6H102OOH4-5 => ETES1 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 5052
reaction('C6H102OOH4-6 => ETES1 + OH', [7.500000e+10, 0.0, 15250.0])

# Reaction 5053
reaction('C6H102OOH5-4 => ETES1 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 5054
reaction('C6H102OOH5-6 => ETES1 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 5055
reaction('C6H102OOH6-4 => ETES1 + OH', [7.500000e+10, 0.0, 15250.0])

# Reaction 5056
reaction('C6H102OOH6-5 => ETES1 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 5057
reaction('C6H103OOH2-1 => ETES1 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 5058
reaction('C6H103OOH1-2 => ETES1 + OH', [6.000000e+11, 0.0, 22000.0])

# Reaction 5059
reaction('HO2 + C6H101-3 <=> C6H101OOH3-4', [4.850000e+11, 0.0, 10000.0])

# Reaction 5060
reaction('HO2 + C6H101-3 <=> C6H101OOH4-3', [4.850000e+11, 0.0, 9000.0])

# Reaction 5061
reaction('HO2 + C6H101-4 <=> C6H101OOH4-5', [3.400000e+11, 0.0, 11800.0])

# Reaction 5062
reaction('HO2 + C6H101-4 <=> C6H101OOH5-4', [3.400000e+11, 0.0, 11800.0])

# Reaction 5063
reaction('HO2 + C6H101-5 <=> C6H101OOH5-6', [1.000000e+12, 0.0, 13700.0])

# Reaction 5064
reaction('HO2 + C6H101-5 <=> C6H101OOH6-5', [1.000000e+12, 0.0, 13700.0])

# Reaction 5065
reaction('HO2 + C6H102-4 <=> C6H102OOH4-5', [3.800000e+11, 0.0, 8800.0])

# Reaction 5066
reaction('HO2 + C6H102-4 <=> C6H102OOH5-4', [3.800000e+11, 0.0, 7800.0])

# Reaction 5067
reaction('HO2 + C6H101-4 <=> C6H102OOH5-6', [1.000000e+12, 0.0, 13700.0])

# Reaction 5068
reaction('HO2 + C6H101-4 <=> C6H102OOH6-5', [1.000000e+12, 0.0, 13700.0])

# Reaction 5069
reaction('HO2 + C6H101-3 <=> C6H103OOH2-1', [4.850000e+11, 0.0, 10000.0])

# Reaction 5070
reaction('HO2 + C6H101-3 <=> C6H103OOH1-2', [4.850000e+11, 0.0, 9000.0])

# Reaction 5071
reaction('C6H12-1 + HO2 <=> C6H12OOH1-2', [1.000000e+12, 0.0, 13700.0])

# Reaction 5072
reaction('C6H12-1 + HO2 <=> C6H12OOH2-1', [1.000000e+12, 0.0, 13700.0])

# Reaction 5073
reaction('C6H12-2 + HO2 <=> C6H12OOH2-3', [6.800000e+11, 0.0, 11800.0])

# Reaction 5074
reaction('C6H12-2 + HO2 <=> C6H12OOH3-2', [6.800000e+11, 0.0, 11800.0])

# Reaction 5075
reaction('C6H12-3 + HO2 <=> C6H12OOH3-4', [6.800000e+11, 0.0, 11800.0])

# Reaction 5076
reaction('C6H101OOH3-4 => CH3 + AC3H5OOH + C2H2', [3.000000e+13, 0.0, 33500.0])

# Reaction 5077
reaction('C6H101OOH3-5 => OH + C3H6 + C2H3CHO', [3.000000e+13, 0.0, 23000.0])

# Reaction 5078
reaction('C6H101OOH3-6 => HO2 + C4H6 + C2H4', [3.000000e+13, 0.0, 30000.0])

# Reaction 5079
reaction('C6H101OOH4-3 => C2H5 + CH3O2H + C3H2', [1.000000e+13, 0.0, 35000.0])

# Reaction 5080
reaction('C6H101OOH4-5 => C3H5-A + AC3H5OOH', [3.000000e+13, 0.0, 23000.0])

# Reaction 5081
reaction('C6H101OOH4-6 => OH + C2H3COCH3 + C2H4', [3.000000e+13, 0.0, 30000.0])

# Reaction 5082
reaction('C6H101OOH5-3 => OH + C4H6 + CH3CHO', [1.000000e+13, 0.0, 35000.0])

# Reaction 5083
reaction('C6H101OOH5-4 => CH3 + CH3O2H + 2 C2H2', [3.000000e+13, 0.0, 33500.0])

# Reaction 5084
reaction('C6H101OOH5-6 => C4H71-4 + C2H3OOH', [3.000000e+13, 0.0, 30000.0])

# Reaction 5085
reaction('C6H101OOH6-3 => HO2 + C4H6 + C2H4', [1.000000e+13, 0.0, 35000.0])

# Reaction 5086
reaction('C6H101OOH6-4 => OH + C5H81-4 + CH2O', [3.000000e+13, 0.0, 29500.0])

# Reaction 5087
reaction('C6H101OOH6-5 => C3H5-A + AC3H5OOH', [3.000000e+13, 0.0, 23000.0])

# Reaction 5088
reaction('C6H102OOH4-6 => OH + C2H3COCH3 + C2H4', [3.000000e+13, 0.0, 22000.0])

# Reaction 5089
reaction('C6H102OOH5-6 => C4H71-3 + CH2CHO + OH', [3.000000e+13, 0.0, 22000.0])

# Reaction 5090
reaction('C6H102OOH6-4 => OH + C5H81-3 + CH2O', [1.000000e+13, 0.0, 35000.0])

# Reaction 5091
reaction('C6H102OOH6-5 => C3H5-T + AC3H5OOH', [3.000000e+13, 0.0, 39000.0])

# Reaction 5092
reaction('C6H103OOH2-1 => C4H71-4 + CH2CHO + OH', [3.000000e+13, 0.0, 39000.0])

# Reaction 5093
reaction('C6H101OOH3-4 + O2 <=> C6H101OOH3-4O2', [7.540000e+12, 0.0, 0.0])

# Reaction 5094
reaction('C6H101OOH3-5 + O2 <=> C6H101OOH3-5O2', [7.540000e+12, 0.0, 0.0])

# Reaction 5095
reaction('C6H101OOH3-6 + O2 <=> C6H101OOH3-6O2', [4.520000e+12, 0.0, 0.0])

# Reaction 5096
reaction('C6H101OOH4-3 + O2 <=> C6H101OOH4-3O2', [2.000000e+12, 0.0, 0.0])

# Reaction 5097
reaction('C6H101OOH4-5 + O2 <=> C6H101OOH4-5O2', [7.540000e+12, 0.0, 0.0])

# Reaction 5098
reaction('C6H101OOH4-6 + O2 <=> C6H101OOH4-6O2', [4.520000e+12, 0.0, 0.0])

# Reaction 5099
reaction('C6H101OOH5-3 + O2 <=> C6H101OOH5-3O2', [2.000000e+12, 0.0, 0.0])

# Reaction 5100
reaction('C6H101OOH5-4 + O2 <=> C6H101OOH5-4O2', [7.540000e+12, 0.0, 0.0])

# Reaction 5101
reaction('C6H101OOH5-6 + O2 <=> C6H101OOH5-6O2', [4.520000e+12, 0.0, 0.0])

# Reaction 5102
reaction('C6H101OOH6-3 + O2 <=> C6H101OOH6-3O2', [2.000000e+12, 0.0, 0.0])

# Reaction 5103
reaction('C6H101OOH6-4 + O2 <=> C6H101OOH6-4O2', [7.540000e+12, 0.0, 0.0])

# Reaction 5104
reaction('C6H101OOH6-5 + O2 <=> C6H101OOH6-5O2', [7.540000e+12, 0.0, 0.0])

# Reaction 5105
reaction('C6H102OOH4-5 + O2 <=> C6H102OOH4-5O2', [7.540000e+12, 0.0, 0.0])

# Reaction 5106
reaction('C6H102OOH4-6 + O2 <=> C6H102OOH4-6O2', [4.520000e+12, 0.0, 0.0])

# Reaction 5107
reaction('C6H102OOH5-4 + O2 <=> C6H102OOH5-4O2', [1.000000e+12, 0.0, 0.0])

# Reaction 5108
reaction('C6H102OOH5-6 + O2 <=> C6H102OOH5-6O2', [4.520000e+12, 0.0, 0.0])

# Reaction 5109
reaction('C6H102OOH6-4 + O2 <=> C6H102OOH6-4O2', [1.000000e+12, 0.0, 0.0])

# Reaction 5110
reaction('C6H102OOH6-5 + O2 <=> C6H102OOH6-5O2', [7.540000e+12, 0.0, 0.0])

# Reaction 5111
reaction('C6H103OOH2-1 + O2 <=> C6H103OOH2-1O2', [4.520000e+12, 0.0, 0.0])

# Reaction 5112
reaction('C6H103OOH1-2 + O2 <=> C6H103OOH1-2O2', [1.000000e+12, 0.0, 0.0])

# Reaction 5113
reaction('C6H101OOH3-4O2 => NC6D1KET34 + OH', [1.000000e+11, 0.0, 21450.0])

# Reaction 5114
reaction('C6H101OOH3-5O2 => NC6D1KET35 + OH', [1.250000e+10, 0.0, 15450.0])

# Reaction 5115
reaction('C6H101OOH3-6O2 => NC6D1KET36 + OH', [1.563000e+09, 0.0, 13650.0])

# Reaction 5116
reaction('C6H101OOH4-3O2 => NC6D1KET43 + OH', [1.000000e+11, 0.0, 23450.0])

# Reaction 5117
reaction('C6H101OOH4-5O2 => NC6D1KET45 + OH', [1.000000e+11, 0.0, 23450.0])

# Reaction 5118
reaction('C6H101OOH4-6O2 => NC6D1KET46 + OH', [1.250000e+10, 0.0, 17450.0])

# Reaction 5119
reaction('C6H101OOH5-3O2 => NC6D1KET53 + OH', [1.250000e+10, 0.0, 17450.0])

# Reaction 5120
reaction('C6H101OOH5-4O2 => NC6D1KET54 + OH', [1.000000e+11, 0.0, 23450.0])

# Reaction 5121
reaction('C6H101OOH5-6O2 => NC6D1KET56 + OH', [1.000000e+11, 0.0, 23450.0])

# Reaction 5122
reaction('C6H101OOH6-3O2 => NC6D1KET63 + OH', [3.120000e+09, 0.0, 22150.0])

# Reaction 5123
reaction('C6H101OOH6-4O2 => NC6D1KET64 + OH', [2.500000e+10, 0.0, 21000.0])

# Reaction 5124
reaction('C6H101OOH6-5O2 => NC6D1KET65 + OH', [2.000000e+11, 0.0, 26150.0])

# Reaction 5125
reaction('C6H102OOH4-5O2 => NC6D2KET45 + OH', [1.000000e+11, 0.0, 21450.0])

# Reaction 5126
reaction('C6H102OOH4-6O2 => NC6D2KET46 + OH', [1.250000e+10, 0.0, 15450.0])

# Reaction 5127
reaction('C6H102OOH5-4O2 => NC6D2KET54 + OH', [1.000000e+11, 0.0, 23450.0])

# Reaction 5128
reaction('C6H102OOH5-6O2 => NC6D2KET56 + OH', [1.000000e+11, 0.0, 23450.0])

# Reaction 5129
reaction('C6H102OOH6-4O2 => NC6D2KET64 + OH', [2.500000e+10, 0.0, 21000.0])

# Reaction 5130
reaction('C6H102OOH6-5O2 => NC6D2KET65 + OH', [2.000000e+11, 0.0, 26000.0])

# Reaction 5131
reaction('C6H103OOH2-1O2 => NC6D3KET21 + OH', [1.000000e+11, 0.0, 21450.0])

# Reaction 5132
reaction('C6H103OOH1-2O2 => NC6D3KET12 + OH', [2.000000e+11, 0.0, 26000.0])

# Reaction 5133
reaction('NC6D1KET34 => OH + C2H3 + CO + C2H5CHO', [1.000000e+16, 0.0, 39000.0])

# Reaction 5134
reaction('NC6D1KET35 => OH + CH2CHO + C2H3COCH3', [5.000000e+15, 0.0, 39000.0])

# Reaction 5135
reaction('NC6D1KET35 => OH + CH3 + CO + C2H3COCH3', [5.000000e+15, 0.0, 39000.0])

# Reaction 5136
reaction('NC6D1KET36 => OH + C2H3 + CO + CH2O + C2H4', [1.000000e+16, 0.0, 39000.0])

# Reaction 5137
reaction('NC6D1KET43 => OH + C3H5O + C2H3CHO', [1.000000e+16, 0.0, 39000.0])

# Reaction 5138
reaction('NC6D1KET45 => OH + C2H3 + CH3CHO + CH2CO', [1.000000e+16, 0.0, 39000.0])

# Reaction 5139
reaction('NC6D1KET46 => OH + C3H5-A + CO + CH3CHO', [1.000000e+16, 0.0, 39000.0])

# Reaction 5140
reaction('NC6D1KET53 => OH + C2H5CO + C2H3CHO', [1.000000e+16, 0.0, 39000.0])

# Reaction 5141
reaction('NC6D1KET54 => OH + C3H5-A + CO + CH3CHO', [1.000000e+16, 0.0, 39000.0])

# Reaction 5142
reaction('NC6D1KET56 => OH + C3H5-A + CH2CO + CH2O', [1.000000e+16, 0.0, 39000.0])

# Reaction 5143
reaction('NC6D1KET63 => OH + C2H3CHO + HCO + C2H4', [1.000000e+16, 0.0, 39000.0])

# Reaction 5144
reaction('NC6D1KET64 => OH + C3H5-A + CH3CHO + CO', [5.000000e+15, 0.0, 39000.0])

# Reaction 5145
reaction('NC6D1KET64 => OH + CH2CHO + AC3H5CHO', [5.000000e+15, 0.0, 39000.0])

# Reaction 5146
reaction('NC6D1KET65 => OH + HCO + CH2CO + C3H6', [1.000000e+16, 0.0, 39000.0])

# Reaction 5147
reaction('NC6D2KET45 => OH + C3H5-T + CO + CH3CHO', [1.000000e+16, 0.0, 39000.0])

# Reaction 5148
reaction('NC6D2KET46 => OH + C3H5-T + CH2CO + CH2O', [1.000000e+16, 0.0, 39000.0])

# Reaction 5149
reaction('NC6D2KET54 => OH + CH2CHO + C2H3COCH3', [1.000000e+16, 0.0, 39000.0])

# Reaction 5150
reaction('NC6D2KET56 => OH + C3H5-T + CH2O + CH2CO', [1.000000e+16, 0.0, 39000.0])

# Reaction 5151
reaction('NC6D2KET64 => OH + CH2CHO + C2H3COCH3', [1.000000e+16, 0.0, 39000.0])

# Reaction 5152
reaction('NC6D2KET65 => OH + C4H71-3 + CO + CH2O', [1.000000e+16, 0.0, 39000.0])

# Reaction 5153
reaction('NC6D3KET21 => OH + C4H71-4 + CO + CH2O', [1.000000e+16, 0.0, 39000.0])

# Reaction 5154
reaction('NC6D3KET12 => OH + HCO + C4H7CHO1-4', [1.000000e+16, 0.0, 39000.0])

# Reaction 5155
reaction('C6H12OH-1 => CH3CHO + PC4H9', [1.500000e+13, 0.0, 30000.0])

# Reaction 5156
reaction('C6H12OH-1 => C2H5CHO + NC3H7', [1.500000e+13, 0.0, 30000.0])

# Reaction 5157
reaction('C6H12OH-2 => NC3H7CHO + C2H5', [1.500000e+13, 0.0, 30000.0])

# Reaction 5158
reaction('C6H12OH-2 => NC4H9CHO + CH3', [1.500000e+13, 0.0, 30000.0])

# Reaction 5159
reaction('C6H12OH-3 => NC3H7CHO + C2H5', [3.000000e+13, 0.0, 30000.0])

# Reaction 5160
reaction('C6H12-3 + O <=> NC3H7 + C2H5CO', [1.000000e+11, 0.0, -1050.0])

# Reaction 5161
reaction('C6H111-4 + HO2 <=> C6H111O2H-4', [9.000000e+12, 0.0, 0.0])

# Reaction 5162
reaction('C6H111-5 + HO2 <=> C6H111O2H-5', [9.000000e+12, 0.0, 0.0])

# Reaction 5163
reaction('C6H111-6 + HO2 <=> C6H111O2H-6', [9.000000e+12, 0.0, 0.0])

# Reaction 5164
reaction('C6H112-5 + HO2 <=> C6H112O2H-5', [9.000000e+12, 0.0, 0.0])

# Reaction 5165
reaction('C6H112-6 + HO2 <=> C6H112O2H-6', [9.000000e+12, 0.0, 0.0])

# Reaction 5166
reaction('C6H112-4 + HO2 <=> C6H113O2H-2', [4.000000e+12, 0.0, 0.0])

# Reaction 5167
reaction('C6H112-4 + HO2 <=> C6H112O2H-4', [4.000000e+12, 0.0, 0.0])

# Reaction 5168
reaction('C6H113-1 + HO2 <=> C6H113O2H-1', [9.000000e+12, 0.0, 0.0])

# Reaction 5169
reaction('C6H112O2-1 => C2H3COC3H7 + OH', [4.000000e+12, 0.0, 38000.0])

# Reaction 5170
reaction('C6H111O2-3 => C2H3COC3H7 + OH', [4.000000e+12, 0.0, 38000.0])

# Reaction 5171
reaction('C6H101OOH3-6 => C2H3COC3H7 + OH', [1.000000e+11, 0.0, 16700.0])

# Reaction 5172
reaction('C6H111O2H-4 => OH + C2H5CHO + C3H5-A', [1.000000e+16, 0.0, 39000.0])

# Reaction 5173
reaction('C6H111O2H-5 => OH + CH3CHO + C4H71-4', [1.000000e+16, 0.0, 39000.0])

# Reaction 5174
reaction('C6H111O2H-6 => OH + CH2O + C5H91-5', [1.000000e+16, 0.0, 39000.0])

# Reaction 5175
reaction('C6H112O2H-4 => OH + C2H3COCH3 + C2H5', [1.000000e+16, 0.0, 39000.0])

# Reaction 5176
reaction('C6H112O2H-5 => OH + CH3CHO + C4H71-3', [1.000000e+16, 0.0, 39000.0])

# Reaction 5177
reaction('C6H112O2H-6 => OH + CH2O + C5H91-5', [1.000000e+16, 0.0, 39000.0])

# Reaction 5178
reaction('C6H113O2H-2 => OH + CH3CHO + C4H71-4', [1.000000e+16, 0.0, 39000.0])

# Reaction 5179
reaction('C6H113O2H-1 => OH + CH2O + C5H91-3', [1.000000e+16, 0.0, 39000.0])

# Reaction 5180
reaction('O2 + ETES1 => HO2 + C2H3COCH3 + C2H3', [2.045000e+07, 2.0, 40722.49])

# Reaction 5181
reaction('H + ETES1 => H2 + C2H3COCH3 + C2H3', [2.574000e+07, 2.0, 3950.57])

# Reaction 5182
reaction('OH + ETES1 => H2O + C2H3COCH3 + C2H3', [4.793000e+06, 2.0, -2259.83])

# Reaction 5183
reaction('O + ETES1 => OH + C2H3COCH3 + C2H3', [1.624000e+07, 2.0, 2579.54])

# Reaction 5184
reaction('HO2 + ETES1 => H2O2 + C2H3COCH3 + C2H3', [6.465000e+05, 2.0, 11887.73])

# Reaction 5185
reaction('HCO + ETES1 => CH2O + C2H3COCH3 + C2H3', [1.516000e+06, 2.0, 12360.44])

# Reaction 5186
reaction('CH3 + ETES1 => CH4 + C2H3COCH3 + C2H3', [4.684000e+05, 2.0, 4871.29])

# Reaction 5187
reaction('C2H5 + ETES1 => C2H6 + C2H3COCH3 + C2H3', [2.758000e+05, 2.0, 7658.07])

# Reaction 5188
reaction('CH3O + ETES1 => CH3OH + C2H3COCH3 + C2H3', [5.136000e+05, 2.0, 1583.56])

# Reaction 5189
reaction('CH3O2 + ETES1 => CH3O2H + C2H3COCH3 + C2H3', [9.133000e+05, 2.0, 12360.44])

# Reaction 5190
reaction('O2 + MVOX => HO2 + C2H3COCH3 + C2H3', [2.045000e+07, 2.0, 40722.49])

# Reaction 5191
reaction('H + MVOX => H2 + C2H3COCH3 + C2H3', [2.574000e+07, 2.0, 3950.57])

# Reaction 5192
reaction('OH + MVOX => H2O + C2H3COCH3 + C2H3', [4.793000e+06, 2.0, -2259.83])

# Reaction 5193
reaction('O + MVOX => OH + C2H3COCH3 + C2H3', [1.624000e+07, 2.0, 2579.54])

# Reaction 5194
reaction('HO2 + MVOX => H2O2 + C2H3COCH3 + C2H3', [6.465000e+05, 2.0, 11887.73])

# Reaction 5195
reaction('HCO + MVOX => CH2O + C2H3COCH3 + C2H3', [1.516000e+06, 2.0, 12360.44])

# Reaction 5196
reaction('CH3 + MVOX => CH4 + C2H3COCH3 + C2H3', [4.684000e+05, 2.0, 4871.29])

# Reaction 5197
reaction('C2H5 + MVOX => C2H6 + C2H3COCH3 + C2H3', [2.758000e+05, 2.0, 7658.07])

# Reaction 5198
reaction('CH3O2 + MVOX => CH3O2H + C2H3COCH3 + C2H3', [9.133000e+05, 2.0, 12360.44])

# Reaction 5199
reaction('C2H3 + MVOX => C2H4 + C2H3COCH3 + C2H3', [8.139000e+05, 2.0, 4871.29])

# Reaction 5200
reaction('CH3O + MVOX => CH3OH + C2H3COCH3 + C2H3', [5.136000e+05, 2.0, 1583.56])

# Reaction 5201
reaction('O2 + VTHF => HO2 + C2H3COCH3 + C2H3', [2.045000e+07, 2.0, 40722.49])

# Reaction 5202
reaction('H + VTHF => H2 + C2H3COCH3 + C2H3', [2.574000e+07, 2.0, 3950.57])

# Reaction 5203
reaction('OH + VTHF => H2O + C2H3COCH3 + C2H3', [4.793000e+06, 2.0, -2259.83])

# Reaction 5204
reaction('O + VTHF => OH + C2H3COCH3 + C2H3', [1.624000e+07, 2.0, 2579.54])

# Reaction 5205
reaction('HO2 + VTHF => H2O2 + C2H3COCH3 + C2H3', [6.465000e+05, 2.0, 11887.73])

# Reaction 5206
reaction('HCO + VTHF => CH2O + C2H3COCH3 + C2H3', [1.516000e+06, 2.0, 12360.44])

# Reaction 5207
reaction('CH3 + VTHF => CH4 + C2H3COCH3 + C2H3', [4.684000e+05, 2.0, 4871.29])

# Reaction 5208
reaction('C2H5 + VTHF => C2H6 + C2H3COCH3 + C2H3', [2.758000e+05, 2.0, 7658.07])

# Reaction 5209
reaction('CH3O + VTHF => CH3OH + C2H3COCH3 + C2H3', [5.136000e+05, 2.0, 1583.56])

# Reaction 5210
reaction('CH3O2 + VTHF => CH3O2H + C2H3COCH3 + C2H3', [9.133000e+05, 2.0, 12360.44])

# Reaction 5211
reaction('C2H3 + VTHF => C2H4 + C2H3COCH3 + C2H3', [8.139000e+05, 2.0, 4871.29])

# Reaction 5212
reaction('O2 + EDHF => HO2 + C2H3COCH3 + C2H3', [2.045000e+07, 2.0, 40722.49])

# Reaction 5213
reaction('H + EDHF => H2 + C2H3COCH3 + C2H3', [2.574000e+07, 2.0, 3950.57])

# Reaction 5214
reaction('OH + EDHF => H2O + C2H3COCH3 + C2H3', [4.793000e+06, 2.0, -2259.83])

# Reaction 5215
reaction('O + EDHF => OH + C2H3COCH3 + C2H3', [1.624000e+07, 2.0, 2579.54])

# Reaction 5216
reaction('HO2 + EDHF => H2O2 + C2H3COCH3 + C2H3', [6.465000e+05, 2.0, 11887.73])

# Reaction 5217
reaction('HCO + EDHF => CH2O + C2H3COCH3 + C2H3', [1.516000e+06, 2.0, 12360.44])

# Reaction 5218
reaction('CH3 + EDHF => CH4 + C2H3COCH3 + C2H3', [4.684000e+05, 2.0, 4871.29])

# Reaction 5219
reaction('C2H5 + EDHF => C2H6 + C2H3COCH3 + C2H3', [2.758000e+05, 2.0, 7658.07])

# Reaction 5220
reaction('C2H3 + EDHF => C2H4 + C2H3COCH3 + C2H3', [8.139000e+05, 2.0, 4871.29])

# Reaction 5221
reaction('CH3O2 + EDHF => CH3O2H + C2H3COCH3 + C2H3', [9.133000e+05, 2.0, 12360.44])

# Reaction 5222
reaction('CH3O + EDHF => CH3OH + C2H3COCH3 + C2H3', [5.136000e+05, 2.0, 1583.56])

# Reaction 5223
reaction('O2 + C5H81-4 => HO2 + C3H5-A + C2H2', [5.111000e+06, 2.0, 40722.49])

# Reaction 5224
reaction('H + C5H81-4 => H2 + C3H5-A + C2H2', [6.435000e+06, 2.0, 3950.57])

# Reaction 5225
reaction('OH + C5H81-4 => H2O + C3H5-A + C2H2', [1.198000e+06, 2.0, -2259.83])

# Reaction 5226
reaction('O + C5H81-4 => OH + C3H5-A + C2H2', [4.060000e+06, 2.0, 2579.54])

# Reaction 5227
reaction('HO2 + C5H81-4 => H2O2 + C3H5-A + C2H2', [1.616000e+05, 2.0, 11887.73])

# Reaction 5228
reaction('HCO + C5H81-4 => CH2O + C3H5-A + C2H2', [3.789000e+05, 2.0, 12360.44])

# Reaction 5229
reaction('CH3 + C5H81-4 => CH4 + C3H5-A + C2H2', [1.171000e+05, 2.0, 4871.29])

# Reaction 5230
reaction('C2H5 + C5H81-4 => C2H6 + C3H5-A + C2H2', [6.895000e+04, 2.0, 7658.07])

# Reaction 5231
reaction('CH3O + C5H81-4 => CH3OH + C3H5-A + C2H2', [1.284000e+05, 2.0, 1583.56])

# Reaction 5232
reaction('CH3O2 + C5H81-4 => CH3O2H + C3H5-A + C2H2', [2.283000e+05, 2.0, 12360.44])

# Reaction 5233
reaction('C2H3 + C5H81-4 => C2H4 + C3H5-A + C2H2', [2.035000e+05, 2.0, 4871.29])

# Reaction 5234
reaction('C6H101-3 + H <=> C6H111-3', [2.500000e+11, 0.5, 2620.0])

# Reaction 5235
reaction('C6H101-3 + H <=> C6H111-4', [2.500000e+11, 0.5, 2620.0])

# Reaction 5236
reaction('C6H101-3 + H <=> C6H113-1', [4.250000e+11, 0.5, 1230.0])

# Reaction 5237
reaction('C6H101-3 + H <=> C6H112-4', [2.500000e+11, 0.5, 2620.0])

# Reaction 5238
reaction('C6H101-4 + H <=> C6H111-4', [2.500000e+11, 0.5, 2620.0])

# Reaction 5239
reaction('C6H101-4 + H <=> C6H111-5', [2.500000e+11, 0.5, 2620.0])

# Reaction 5240
reaction('C6H101-4 + H <=> C6H112-5', [4.250000e+11, 0.5, 1230.0])

# Reaction 5241
reaction('C6H101-4 + H <=> C6H112-6', [2.500000e+11, 0.5, 2620.0])

# Reaction 5242
reaction('C6H101-5 + H <=> C6H111-5', [4.250000e+11, 0.5, 1230.0])

# Reaction 5243
reaction('C6H101-5 + H <=> C6H111-6', [2.500000e+11, 0.5, 2620.0])

# Reaction 5244
reaction('C6H101-3 + O => C5H91-3 + HCO', [2.000000e+11, 0.0, -1050.0])

# Reaction 5245
reaction('C6H101-4 + O => C5H92-5 + HCO', [2.000000e+11, 0.0, -1050.0])

# Reaction 5246
reaction('C6H101-5 + O => C5H91-5 + HCO', [2.000000e+11, 0.0, -1050.0])

# Reaction 5247
reaction('C6H101-3 + OH => C5H91-3 + CH2O', [1.000000e+12, 0.0, 0.0])

# Reaction 5248
reaction('C6H101-4 + OH => C5H92-5 + CH2O', [1.000000e+12, 0.0, 0.0])

# Reaction 5249
reaction('C6H101-5 + OH => C5H91-5 + CH2O', [1.000000e+12, 0.0, 0.0])

# Reaction 5250
reaction('C6H101-3 + OH => C2H3 + C4H6 + H2O', [2.764000e+04, 2.64, -1919.0])

# Reaction 5251
reaction('C6H101-4 + OH => C2H3 + C4H6 + H2O', [2.764000e+04, 2.64, -1919.0])

# Reaction 5252
reaction('C6H101-5 + OH => C2H3 + C4H6 + H2O', [2.764000e+04, 2.64, -1919.0])

# Reaction 5253
reaction('C6H101-3 + HO2 => C2H3 + C4H6 + H2O2', [4.820000e+03, 2.55, 10530.0])

# Reaction 5254
reaction('C6H101-4 + HO2 => C2H3 + C4H6 + H2O2', [4.820000e+03, 2.55, 10530.0])

# Reaction 5255
reaction('C6H101-5 + HO2 => C2H3 + C4H6 + H2O2', [4.820000e+03, 2.55, 10530.0])

# Reaction 5256
reaction('C6H101-3 + H => C2H3 + C4H6 + H2', [3.376000e+05, 2.36, 207.0])

# Reaction 5257
reaction('C6H101-4 + H => C2H3 + C4H6 + H2', [3.376000e+05, 2.36, 207.0])

# Reaction 5258
reaction('C6H101-5 + H => C2H3 + C4H6 + H2', [3.376000e+05, 2.36, 207.0])

    """)




